
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION2D_MATRICES_CONST_HPP_
#define CCOMPUTATION2D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (4)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (15)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (15)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS (10)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (27)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*4+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_2D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[15],
		T io_hv[15]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,7.838367178,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,7.668115805,10.73312629,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,4.381780461,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,2.449489743,9.142857143,14.84614978,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,-2.993325910,.6110100926,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.},
			{4.732863826,3.346640106,-5.074680381,1.521277658,0.,0.,0.,0.,0.,4.364357806,17.00840129,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,10.95445115,-3.833259389,.5532833353,0.,0.,0.,0.,17.74823935,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,7.838367178,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,7.668115805,10.73312629,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,4.381780461,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,2.449489743,9.142857143,14.84614978,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,-2.993325910,.6110100926,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.},
			{4.732863826,3.346640106,-5.074680381,1.521277658,0.,0.,0.,0.,0.,4.364357806,17.00840129,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,10.95445115,-3.833259389,.5532833353,0.,0.,0.,0.,17.74823935,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,-2.828427124,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,8.854377448,-3.098386677,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,3.919183589,4.200000000,0.,0.,-5.091168823,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,3.834057902,5.366563145,4.647580015,0.,6.640783086,-6.196773354,13.41640786,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,-.9128709293,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,2.529822128,-1.807392228,0.,0.,2.190890230,6.260990336,0.,0.,8.082903771,0.,0.,0.,0.,0.},
			{2.,3.535533905,1.649572198,-2.571428571,0.,1.224744872,4.571428571,7.423074892,0.,-7.377111136,11.49977817,0.,0.,0.,0.},
			{2.645751311,-1.496662955,.3055050464,0.,0.,5.184592560,-1.058300524,0.,0.,6.831300514,0.,0.,0.,0.,0.},
			{2.366431913,1.673320053,-2.537340190,.7606388294,0.,2.898275350,5.747048932,-2.195775165,0.,2.182178903,8.504200642,0.,10.06230590,0.,0.},
			{2.683281572,-1.897366596,.6639400021,-.9583148473e-1,0.,5.477225576,-1.916629694,.2766416676,0.,7.423074892,-1.071428571,0.,8.874119673,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,-2.828427124,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,8.854377448,-3.098386677,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,3.919183589,4.200000000,0.,0.,-5.091168823,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,3.834057902,5.366563145,4.647580015,0.,6.640783086,-6.196773354,13.41640786,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,-.9128709293,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,2.529822128,-1.807392228,0.,0.,2.190890230,6.260990336,0.,0.,8.082903771,0.,0.,0.,0.,0.},
			{2.,3.535533905,1.649572198,-2.571428571,0.,1.224744872,4.571428571,7.423074892,0.,-7.377111136,11.49977817,0.,0.,0.,0.},
			{2.645751311,-1.496662955,.3055050464,0.,0.,5.184592560,-1.058300524,0.,0.,6.831300514,0.,0.,0.,0.,0.},
			{2.366431913,1.673320053,-2.537340190,.7606388294,0.,2.898275350,5.747048932,-2.195775165,0.,2.182178903,8.504200642,0.,10.06230590,0.,0.},
			{2.683281572,-1.897366596,.6639400021,-.9583148473e-1,0.,5.477225576,-1.916629694,.2766416676,0.,7.423074892,-1.071428571,0.,8.874119673,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,-1.92171958555151,2.19799643560933,-2.29634628619806,2.23836106649518,3.41890639763464,-3.91413586924262,4.09536544265748,-4.00097073404889,5.33523806427656,-5.59797201662203,5.49216459339238,7.19422103343416,-7.09892453011936,9.0013491098729},
			{1.414213562,-1.59519010006696,1.2388862761662,-.514957807193037,-.353775594121333,3.23038451124486,-2.62173967966463,1.30644820527577,.307822029727901,4.7630793830775,-2.90243304119459,.518849452524734,6.06854814809742,-2.66498561200528,7.17423567654378},
			{1.414213562,-1.03822870489707,-.0622566343920988,1.10967175893133,-1.41978880536247,2.90882269973526,-.707254507004454,-1.57948219926848,2.54719860772099,3.86201434387853,.544351948717181,-3.18953813043078,4.43071582310306,2.19278566166382,4.71658607960972},
			{1.414213562,-.300186182387741,-1.13611369814234,1.22655366483195,.100858250439743,2.48271365058504,1.26648273457434,-2.5074905585056,.461151554581851,2.81340752869054,3.19379348794833,-2.33714259619681,2.75487590413567,4.77320941254595,2.50302559510191},
			{1.414213562,.553376983055106,-1.45368745167752,-.194424990185747,1.49978166741586,1.9899087271236,2.74863027720701,-.854309601049289,-2.52312426315122,1.80736427869175,4.12997714955087,1.65834138994758,1.4184747657974,4.48820932686102,1.03297875354963},
			{1.414213562,1.44662301694489,-.724355119014703,-1.6918152058931,.160475142667942,1.4741928888764,3.38025060487184,2.16048862424162,-1.70770592621981,.991945941760336,3.46032388717692,4.72311662432345,.576746649127288,2.68886700736674,.31115444533373},
			{1.414213562,2.30018618238774,.987081447023724,-1.35561938023839,-2.20870753342451,.981387965414959,3.10522366683209,4.42237267757001,2.83495589139377,.43960319187862,2.03901059601618,4.55407145300732,.170154848292169,1.03685515398483,.0611113213546603},
			{1.414213562,3.03822870489707,3.26615690329693,1.66498726415444,-.923099517802526,.555278916264744,2.17523616991868,4.50024407433993,6.26847807110811,.140734880491405,.792696592497175,2.35258131414952,.0308216365902829,.225963663364185,.00626330408480523},
			{1.414213562,3.59519010006696,5.47681396366563,6.3309248170395,5.76936797267538,.233717104755142,1.04841335581819,2.66785366107377,5.04596922575242,.0249321870529789,.15913874220381,.559597243722218,.00229823078236389,.0189957623925217,.00019657159112013},
			{1.414213562,3.92171958555151,6.96914454009643,10.2353232484353,13.4470270492568,.0451952183653594,.217799593005347,.614368644454251,1.33333501160483,.000932318622845177,.00636097475878139,.0244039729141499,1.66187980931977e-005,.000146461329046052,2.74870867417466e-007}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,-1.92171958555151,2.19799643560933,-2.29634628619806,2.23836106649518,3.41890639763464,-3.91413586924262,4.09536544265748,-4.00097073404889,5.33523806427656,-5.59797201662203,5.49216459339238,7.19422103343416,-7.09892453011936,9.0013491098729},
			{1.414213562,-1.59519010006696,1.2388862761662,-.514957807193037,-.353775594121333,3.23038451124486,-2.62173967966463,1.30644820527577,.307822029727901,4.7630793830775,-2.90243304119459,.518849452524734,6.06854814809742,-2.66498561200528,7.17423567654378},
			{1.414213562,-1.03822870489707,-.0622566343920988,1.10967175893133,-1.41978880536247,2.90882269973526,-.707254507004454,-1.57948219926848,2.54719860772099,3.86201434387853,.544351948717181,-3.18953813043078,4.43071582310306,2.19278566166382,4.71658607960972},
			{1.414213562,-.300186182387741,-1.13611369814234,1.22655366483195,.100858250439743,2.48271365058504,1.26648273457434,-2.5074905585056,.461151554581851,2.81340752869054,3.19379348794833,-2.33714259619681,2.75487590413567,4.77320941254595,2.50302559510191},
			{1.414213562,.553376983055106,-1.45368745167752,-.194424990185747,1.49978166741586,1.9899087271236,2.74863027720701,-.854309601049289,-2.52312426315122,1.80736427869175,4.12997714955087,1.65834138994758,1.4184747657974,4.48820932686102,1.03297875354963},
			{1.414213562,1.44662301694489,-.724355119014703,-1.6918152058931,.160475142667942,1.4741928888764,3.38025060487184,2.16048862424162,-1.70770592621981,.991945941760336,3.46032388717692,4.72311662432345,.576746649127288,2.68886700736674,.31115444533373},
			{1.414213562,2.30018618238774,.987081447023724,-1.35561938023839,-2.20870753342451,.981387965414959,3.10522366683209,4.42237267757001,2.83495589139377,.43960319187862,2.03901059601618,4.55407145300732,.170154848292169,1.03685515398483,.0611113213546603},
			{1.414213562,3.03822870489707,3.26615690329693,1.66498726415444,-.923099517802526,.555278916264744,2.17523616991868,4.50024407433993,6.26847807110811,.140734880491405,.792696592497175,2.35258131414952,.0308216365902829,.225963663364185,.00626330408480523},
			{1.414213562,3.59519010006696,5.47681396366563,6.3309248170395,5.76936797267538,.233717104755142,1.04841335581819,2.66785366107377,5.04596922575242,.0249321870529789,.15913874220381,.559597243722218,.00229823078236389,.0189957623925217,.00019657159112013},
			{1.414213562,3.92171958555151,6.96914454009643,10.2353232484353,13.4470270492568,.0451952183653594,.217799593005347,.614368644454251,1.33333501160483,.000932318622845177,.00636097475878139,.0244039729141499,1.66187980931977e-005,.000146461329046052,2.74870867417466e-007}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.37371117926928,4.13193546224797,-4.77114810448709,5.33430574565371,5.05405999869821,-5.83592580124217,6.52476340023727,-6.34960212623238,7.09907099144837,7.15379270875817},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-2.99666740648972,3.67015303548283,-4.23792768749944,4.73814719602452,3.40958356039142,-3.93704797266052,4.40175344784644,-2.40062592164153,2.68398137439781,.178027784911327},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-2.35354378347051,2.88249067692314,-3.32841353768903,3.72127946338712,1.05379652310981,-1.21681941252376,1.36044546108577,1.7591790753139,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-1.50132568517008,1.83874093225775,-2.12319514516607,2.37380433681192,-1.19541898829253,1.38035094930968,-1.54327927740643,3.34189684440344,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-.515715838247201,.631620327664824,-.729332332662821,.81541833693141,-2.55652048864375,2.95201558448519,-3.30045375807884,1.60938069663531,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,.515715838247201,-.631620327664824,.729332332662821,-.81541833693141,-2.55652048864375,2.95201558448519,-3.30045375807885,-1.60938069663533,1.79934231949429,2.78946777547417},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,1.50132568517008,-1.83874093225775,2.12319514516607,-2.37380433681192,-1.19541898829252,1.38035094930968,-1.54327927740643,-3.34189684440345,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,2.35354378347051,-2.88249067692314,3.32841353768903,-3.72127946338712,1.0537965231098,-1.21681941252376,1.36044546108577,-1.75917907531388,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,2.99666740648972,-3.67015303548283,4.23792768749944,-4.73814719602452,3.40958356039142,-3.93704797266052,4.40175344784644,2.40062592164155,-2.68398137439783,.178027784911158},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,3.37371117926928,-4.13193546224797,4.77114810448709,-5.33430574565371,5.05405999869821,-5.83592580124217,6.52476340023727,6.34960212623239,-7.09907099144838,7.15379270875832}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.37371117926928,4.13193546224797,-4.77114810448709,5.33430574565371,5.05405999869821,-5.83592580124217,6.52476340023727,-6.34960212623238,7.09907099144837,7.15379270875817},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-2.99666740648972,3.67015303548283,-4.23792768749944,4.73814719602452,3.40958356039142,-3.93704797266052,4.40175344784644,-2.40062592164153,2.68398137439781,.178027784911327},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-2.35354378347051,2.88249067692314,-3.32841353768903,3.72127946338712,1.05379652310981,-1.21681941252376,1.36044546108577,1.7591790753139,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-1.50132568517008,1.83874093225775,-2.12319514516607,2.37380433681192,-1.19541898829253,1.38035094930968,-1.54327927740643,3.34189684440344,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-.515715838247201,.631620327664824,-.729332332662821,.81541833693141,-2.55652048864375,2.95201558448519,-3.30045375807884,1.60938069663531,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,.515715838247201,-.631620327664824,.729332332662821,-.81541833693141,-2.55652048864375,2.95201558448519,-3.30045375807885,-1.60938069663533,1.79934231949429,2.78946777547417},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,1.50132568517008,-1.83874093225775,2.12319514516607,-2.37380433681192,-1.19541898829252,1.38035094930968,-1.54327927740643,-3.34189684440345,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,2.35354378347051,-2.88249067692314,3.32841353768903,-3.72127946338712,1.0537965231098,-1.21681941252376,1.36044546108577,-1.75917907531388,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,2.99666740648972,-3.67015303548283,4.23792768749944,-4.73814719602452,3.40958356039142,-3.93704797266052,4.40175344784644,2.40062592164155,-2.68398137439783,.178027784911158},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,3.37371117926928,-4.13193546224797,4.77114810448709,-5.33430574565371,5.05405999869821,-5.83592580124217,6.52476340023727,6.34960212623239,-7.09907099144838,7.15379270875832}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,3.92171958555151,6.96914454009643,10.2353232484353,13.4470270492568,-.0451952183653594,-.217799593005347,-.614368644454251,-1.33333501160483,.000932318622845132,.00636097475878108,.0244039729141487,-1.66187980893617e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,3.59519010006696,5.47681396366563,6.3309248170395,5.76936797267538,-.233717104755142,-1.04841335581819,-2.66785366107377,-5.04596922575242,.0249321870529788,.159138742203809,.559597243722216,-.00229823078236368,-.0189957623925199,.00019657159111697},
			{1.414213562,3.03822870489707,3.26615690329693,1.66498726415444,-.923099517802526,-.555278916264744,-2.17523616991868,-4.50024407433993,-6.26847807110811,.140734880491405,.792696592497175,2.35258131414951,-.0308216365902829,-.225963663364185,.00626330408481366},
			{1.414213562,2.30018618238774,.987081447023724,-1.35561938023839,-2.20870753342451,-.981387965414959,-3.10522366683209,-4.42237267757001,-2.83495589139377,.439603191878621,2.03901059601619,4.55407145300733,-.170154848292169,-1.03685515398483,.0611113213546687},
			{1.414213562,1.44662301694489,-.724355119014703,-1.6918152058931,.160475142667942,-1.4741928888764,-3.38025060487184,-2.16048862424162,1.70770592621981,.991945941760336,3.46032388717692,4.72311662432345,-.576746649127296,-2.68886700736678,.311154445333742},
			{1.414213562,.553376983055106,-1.45368745167752,-.194424990185747,1.49978166741586,-1.9899087271236,-2.74863027720701,.854309601049289,2.52312426315122,1.80736427869175,4.12997714955087,1.65834138994758,-1.41847476579742,-4.48820932686106,1.03297875354965},
			{1.414213562,-.300186182387741,-1.13611369814234,1.22655366483195,.100858250439743,-2.48271365058504,-1.26648273457434,2.5074905585056,-.461151554581851,2.81340752869054,3.19379348794833,-2.33714259619681,-2.75487590413568,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.03822870489707,-.0622566343920988,1.10967175893133,-1.41978880536247,-2.90882269973526,.707254507004454,1.57948219926848,-2.54719860772099,3.86201434387853,.544351948717181,-3.18953813043078,-4.43071582310304,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.59519010006696,1.2388862761662,-.514957807193037,-.353775594121333,-3.23038451124486,2.62173967966463,-1.30644820527577,-.307822029727901,4.7630793830775,-2.90243304119458,.518849452524734,-6.0685481480974,2.66498561200528,7.17423567654393},
			{1.414213562,-1.92171958555151,2.19799643560933,-2.29634628619806,2.23836106649518,-3.41890639763464,3.91413586924262,-4.09536544265748,4.00097073404889,5.33523806427656,-5.59797201662203,5.49216459339238,-7.19422103343416,7.09892453011937,9.00134910987265}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,3.92171958555151,6.96914454009643,10.2353232484353,13.4470270492568,-.0451952183653594,-.217799593005347,-.614368644454251,-1.33333501160483,.000932318622845132,.00636097475878108,.0244039729141487,-1.66187980893617e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,3.59519010006696,5.47681396366563,6.3309248170395,5.76936797267538,-.233717104755142,-1.04841335581819,-2.66785366107377,-5.04596922575242,.0249321870529788,.159138742203809,.559597243722216,-.00229823078236368,-.0189957623925199,.00019657159111697},
			{1.414213562,3.03822870489707,3.26615690329693,1.66498726415444,-.923099517802526,-.555278916264744,-2.17523616991868,-4.50024407433993,-6.26847807110811,.140734880491405,.792696592497175,2.35258131414951,-.0308216365902829,-.225963663364185,.00626330408481366},
			{1.414213562,2.30018618238774,.987081447023724,-1.35561938023839,-2.20870753342451,-.981387965414959,-3.10522366683209,-4.42237267757001,-2.83495589139377,.439603191878621,2.03901059601619,4.55407145300733,-.170154848292169,-1.03685515398483,.0611113213546687},
			{1.414213562,1.44662301694489,-.724355119014703,-1.6918152058931,.160475142667942,-1.4741928888764,-3.38025060487184,-2.16048862424162,1.70770592621981,.991945941760336,3.46032388717692,4.72311662432345,-.576746649127296,-2.68886700736678,.311154445333742},
			{1.414213562,.553376983055106,-1.45368745167752,-.194424990185747,1.49978166741586,-1.9899087271236,-2.74863027720701,.854309601049289,2.52312426315122,1.80736427869175,4.12997714955087,1.65834138994758,-1.41847476579742,-4.48820932686106,1.03297875354965},
			{1.414213562,-.300186182387741,-1.13611369814234,1.22655366483195,.100858250439743,-2.48271365058504,-1.26648273457434,2.5074905585056,-.461151554581851,2.81340752869054,3.19379348794833,-2.33714259619681,-2.75487590413568,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.03822870489707,-.0622566343920988,1.10967175893133,-1.41978880536247,-2.90882269973526,.707254507004454,1.57948219926848,-2.54719860772099,3.86201434387853,.544351948717181,-3.18953813043078,-4.43071582310304,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.59519010006696,1.2388862761662,-.514957807193037,-.353775594121333,-3.23038451124486,2.62173967966463,-1.30644820527577,-.307822029727901,4.7630793830775,-2.90243304119458,.518849452524734,-6.0685481480974,2.66498561200528,7.17423567654393},
			{1.414213562,-1.92171958555151,2.19799643560933,-2.29634628619806,2.23836106649518,-3.41890639763464,3.91413586924262,-4.09536544265748,4.00097073404889,5.33523806427656,-5.59797201662203,5.49216459339238,-7.19422103343416,7.09892453011937,9.00134910987265}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0666713464862914,-.149451343590541,-.219086364157122,-.269266719261963,-.295524224714757,-.295524224714757,-.269266719261963,-.219086364157122,-.149451343590541,-.0666713464862914},
			{.0905970892324072,.168576592685474,.16083974734181,.0571555461131064,-.115637629421958,-.302296737069186,-.437956192288621,-.470674654782083,-.379932709650339,-.184884611630122},
			{-.103621819076306,-.130923096345775,.00964463927720944,.216316415190729,.303772972096487,.151366448890014,-.187940626427248,-.505984710909207,-.578779066349661,-.328551685958625},
			{.108258401003525,.0544197412932449,-.171907523449491,-.233536214191699,.0406284425731006,.353533858334046,.258110368117603,-.257935587575359,-.669039843415995,-.482531635271947},
			{-.105524759595668,.0373863179252117,.219950066547109,-.0192034434697929,-.313405150523241,-.0335340385442513,.420538627970391,.143004226828253,-.609695606023063,-.633943432044736},
			{-.161180106788003,-.341380763335387,-.450627406012926,-.472709481254668,-.415825621665808,-.308057935581016,-.186856585708771,-.0860223958211081,-.0246987698672132,-.00213067258216707},
			{.184527086816818,.277060359339501,.109566067415032,-.241138721975255,-.574373527114506,-.706361447936911,-.591235589736538,-.336982048861548,-.110794287988378,-.0102678920072701},
			{-.193070930476864,-.138062909860115,.244689360540118,.477427012543675,.178522670975115,-.451471222303063,-.842021186737924,-.697166353210365,-.281933595255738,-.0289636486661423},
			{.188620808842354,-.0325300344585138,-.394605522412419,-.0878034049814253,.527250170228969,.356854497200184,-.539776517640132,-.971096661439192,-.533248230919088,-.0628584273827533},
			{-.251523189241307,-.503352981601333,-.598293428384306,-.535673702440208,-.37767982241688,-.207284149381534,-.0837005897640986,-.0218022893353362,-.00263478512147966,-4.39530065166604e-005},
			{.263909455949514,.306723488668707,-.08432961782286,-.608099311980938,-.863029691810222,-.72309413582073,-.388228276260842,-.122802537682291,-.016817473305435,-.000299880275019948},
			{-.258921296192325,-.0548310028164179,.494115125970334,.444992705652827,-.346538929299739,-.986976377645973,-.867096676474837,-.364455907049916,-.0591371502798876,-.0011504950719751},
			{-.339162638412015,-.641312386098098,-.686395213472645,-.52452926221411,-.296414676313606,-.120521122714452,-.0323975380897773,-.00477480946005403,-.000242872567024778,-7.83472648717735e-007},
			{.334670002816657,.281630505360336,-.339700771258984,-.908820614314273,-.937888460850653,-.56188496465138,-.197417556313917,-.0350057153616537,-.0020074352890353,-6.90473792141432e-006},
			{-.424357452903393,-.758159297392549,-.730681505704979,-.47657688200433,-.215858660481788,-.0650210679957279,-.0116356153302471,-.000970295120692958,-2.0773304102378e-005,-1.29584465340518e-008}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0666713464862914,-.149451343590541,-.219086364157122,-.269266719261963,-.295524224714757,-.295524224714757,-.269266719261963,-.219086364157122,-.149451343590541,-.0666713464862914},
			{.0905970892324072,.168576592685474,.16083974734181,.0571555461131064,-.115637629421958,-.302296737069186,-.437956192288621,-.470674654782083,-.379932709650339,-.184884611630122},
			{-.103621819076306,-.130923096345775,.00964463927720944,.216316415190729,.303772972096487,.151366448890014,-.187940626427248,-.505984710909207,-.578779066349661,-.328551685958625},
			{.108258401003525,.0544197412932449,-.171907523449491,-.233536214191699,.0406284425731006,.353533858334046,.258110368117603,-.257935587575359,-.669039843415995,-.482531635271947},
			{-.105524759595668,.0373863179252117,.219950066547109,-.0192034434697929,-.313405150523241,-.0335340385442513,.420538627970391,.143004226828253,-.609695606023063,-.633943432044736},
			{-.161180106788003,-.341380763335387,-.450627406012926,-.472709481254668,-.415825621665808,-.308057935581016,-.186856585708771,-.0860223958211081,-.0246987698672132,-.00213067258216707},
			{.184527086816818,.277060359339501,.109566067415032,-.241138721975255,-.574373527114506,-.706361447936911,-.591235589736538,-.336982048861548,-.110794287988378,-.0102678920072701},
			{-.193070930476864,-.138062909860115,.244689360540118,.477427012543675,.178522670975115,-.451471222303063,-.842021186737924,-.697166353210365,-.281933595255738,-.0289636486661423},
			{.188620808842354,-.0325300344585138,-.394605522412419,-.0878034049814253,.527250170228969,.356854497200184,-.539776517640132,-.971096661439192,-.533248230919088,-.0628584273827533},
			{-.251523189241307,-.503352981601333,-.598293428384306,-.535673702440208,-.37767982241688,-.207284149381534,-.0837005897640986,-.0218022893353362,-.00263478512147966,-4.39530065166604e-005},
			{.263909455949514,.306723488668707,-.08432961782286,-.608099311980938,-.863029691810222,-.72309413582073,-.388228276260842,-.122802537682291,-.016817473305435,-.000299880275019948},
			{-.258921296192325,-.0548310028164179,.494115125970334,.444992705652827,-.346538929299739,-.986976377645973,-.867096676474837,-.364455907049916,-.0591371502798876,-.0011504950719751},
			{-.339162638412015,-.641312386098098,-.686395213472645,-.52452926221411,-.296414676313606,-.120521122714452,-.0323975380897773,-.00477480946005403,-.000242872567024778,-7.83472648717735e-007},
			{.334670002816657,.281630505360336,-.339700771258984,-.908820614314273,-.937888460850653,-.56188496465138,-.197417556313917,-.0350057153616537,-.0020074352890353,-6.90473792141432e-006},
			{-.424357452903393,-.758159297392549,-.730681505704979,-.47657688200433,-.215858660481788,-.0650210679957279,-.0116356153302471,-.000970295120692958,-2.0773304102378e-005,-1.29584465340518e-008}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{.0666713464862914,.149451343590541,.219086364157122,.269266719261963,.295524224714757,.295524224714757,.269266719261963,.219086364157122,.149451343590541,.0666713464862914},
			{-.0816553896850849,-.1830397666013,-.268324900917017,-.329783033481719,-.361941778623412,-.361941778623412,-.329783033481719,-.268324900917017,-.1830397666013,-.0816553896850849},
			{.0942875223977143,.211356116964865,.309834907440273,.380800646175515,.417934366491145,.417934366491145,.380800646175515,.309834907440273,.211356116964865,.0942875223977143},
			{-.105416654777859,-.236303322546677,-.346405957492347,-.425748065451799,-.467264826902148,-.467264826902148,-.425748065451799,-.346405957492347,-.236303322546677,-.105416654777859},
			{.112464933488868,.223927985096936,.257814675202576,.202128520894733,.0762032616355625,-.0762032616355625,-.202128520894733,-.257814675202576,-.223927985096936,-.112464933488868},
			{-.137740850431264,-.274254651167906,-.315757201061946,-.247555869200864,-.093329553823614,.093329553823614,.247555869200864,.315757201061946,.274254651167906,.137740850431264},
			{.159049434205836,.316681993468174,.364605010191818,.285852895545897,.107767686084793,-.107767686084793,-.285852895545897,-.364605010191818,-.316681993468174,-.159049434205836},
			{-.177822673316147,-.35406123228781,-.407640793823026,-.319593252971582,-.120487935919926,.120487935919926,.319593252971582,.407640793823026,.35406123228781,.177822673316147},
			{-.168480492667857,-.25478342209236,-.115436224404772,.160943274560492,.377756867686918,.377756867686918,.160943274560491,-.115436224404772,-.25478342209236,-.168480492667857},
			{.194544515581452,.294198554647266,.133294270462818,-.185841285775377,-.436196058475433,-.436196058475433,-.185841285775376,.133294270462818,.294198554647266,.194544515581452},
			{-.217507380699146,-.328923983467474,-.149027524851671,.207776873966102,.487682019031578,.487682019031578,.207776873966101,-.14902752485167,-.328923983467474,-.217507380699146},
			{.211668261704066,.179388384723805,-.192706073755905,-.44993079970221,-.237805491322023,.237805491322025,.449930799702211,.192706073755903,-.179388384723806,-.211668261704066},
			{-.236652310900817,-.20056231128787,.215451940275855,.503037926601612,.265874621982498,-.265874621982501,-.503037926601614,-.215451940275852,.200562311287872,.236652310900817},
			{-.23847649618836,-.0133032458257229,.440420079694959,.223537309141119,-.412177650856899,-.412177650856901,.223537309141112,.440420079694946,-.0133032458257102,-.238476496188365}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{.0666713464862914,.149451343590541,.219086364157122,.269266719261963,.295524224714757,.295524224714757,.269266719261963,.219086364157122,.149451343590541,.0666713464862914},
			{-.0816553896850849,-.1830397666013,-.268324900917017,-.329783033481719,-.361941778623412,-.361941778623412,-.329783033481719,-.268324900917017,-.1830397666013,-.0816553896850849},
			{.0942875223977143,.211356116964865,.309834907440273,.380800646175515,.417934366491145,.417934366491145,.380800646175515,.309834907440273,.211356116964865,.0942875223977143},
			{-.105416654777859,-.236303322546677,-.346405957492347,-.425748065451799,-.467264826902148,-.467264826902148,-.425748065451799,-.346405957492347,-.236303322546677,-.105416654777859},
			{.112464933488868,.223927985096936,.257814675202576,.202128520894733,.0762032616355625,-.0762032616355625,-.202128520894733,-.257814675202576,-.223927985096936,-.112464933488868},
			{-.137740850431264,-.274254651167906,-.315757201061946,-.247555869200864,-.093329553823614,.093329553823614,.247555869200864,.315757201061946,.274254651167906,.137740850431264},
			{.159049434205836,.316681993468174,.364605010191818,.285852895545897,.107767686084793,-.107767686084793,-.285852895545897,-.364605010191818,-.316681993468174,-.159049434205836},
			{-.177822673316147,-.35406123228781,-.407640793823026,-.319593252971582,-.120487935919926,.120487935919926,.319593252971582,.407640793823026,.35406123228781,.177822673316147},
			{-.168480492667857,-.25478342209236,-.115436224404772,.160943274560492,.377756867686918,.377756867686918,.160943274560491,-.115436224404772,-.25478342209236,-.168480492667857},
			{.194544515581452,.294198554647266,.133294270462818,-.185841285775377,-.436196058475433,-.436196058475433,-.185841285775376,.133294270462818,.294198554647266,.194544515581452},
			{-.217507380699146,-.328923983467474,-.149027524851671,.207776873966102,.487682019031578,.487682019031578,.207776873966101,-.14902752485167,-.328923983467474,-.217507380699146},
			{.211668261704066,.179388384723805,-.192706073755905,-.44993079970221,-.237805491322023,.237805491322025,.449930799702211,.192706073755903,-.179388384723806,-.211668261704066},
			{-.236652310900817,-.20056231128787,.215451940275855,.503037926601612,.265874621982498,-.265874621982501,-.503037926601614,-.215451940275852,.200562311287872,.236652310900817},
			{-.23847649618836,-.0133032458257229,.440420079694959,.223537309141119,-.412177650856899,-.412177650856901,.223537309141112,.440420079694946,-.0133032458257102,-.238476496188365}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{-.13073316265519,-.26865299545921,-.332817240216851,-.309681793511623,-.213756072768581,-.0817681519461757,.0404150742496602,.113730876059729,.119201651868669,.0640618161688986},
			{-.232321125172907,-.409258602732634,-.357785220355006,-.132894091442215,.107032242482493,.214799928567286,.152958804103683,.0068197898368121,-.0925766092644603,-.0732716909670717},
			{-.341201391347813,-.473082610038628,-.182388003035756,.182511591542363,.249986188541097,.0287286472449085,-.165135040664019,-.121556975536052,.0384805680887192,.0765502494498098},
			{-.448265699805767,-.431119897592287,.101119258555274,.297365715667201,-.0237121460614668,-.221610907252239,-.0135788851032055,.155528183618924,.0264361189354877,-.0746172731128625},
			{.00150661303158024,.0174646676628736,.060827019428775,.132127558885229,.217829855282603,.294033116918165,.334256079779963,.318641694631351,.241392652759809,.113971546520449},
			{.00726049606491637,.0783433923326487,.238282291825274,.41806669467124,.499472969673169,.406143415849555,.170510825470376,-.077474909236672,-.195911258835258,-.130480354366348},
			{.0204803923823613,.19935715707521,.492971056033385,.59539889112151,.319238362842028,-.126234591258232,-.337591878134575,-.173021506144313,.0976252198049579,.1365217642077},
			{.0444476202705045,.377063440252612,.686669034698865,.381679636063987,-.252334234943452,-.372822170863378,.0620863830924046,.279028240875838,.0230022079648012,-.133375053045642},
			{-3.10794689696649e-005,-.00186307442685919,-.0154165466384745,-.0591852546271217,-.146572027698836,-.267059963618762,-.378778507598704,-.423057340461501,-.355924306714669,-.177853752785117},
			{-.000212047376066624,-.0118917494198341,-.0868345071649731,-.274518846864829,-.511304767009956,-.610254147605341,-.429991147250034,-.0596300446331458,.216886258844056,.186612165970385},
			{-.000813522866900639,-.0418162799719245,-.257709243250501,-.613129939717922,-.697897689320284,-.245040026788325,.314657359662651,.349392156168293,-.0387713739005192,-.183085004282902},
			{5.53998822800774e-007,.000171736839152697,.00337630014896861,.0229085188830747,.0852213031700891,.209596827719864,.370898198340224,.485354710148538,.453476337188524,.239824201609527},
			{4.88238700670906e-006,.00141947110604439,.0247527787190417,.139595292831676,.397312668856578,.663187290839078,.642633219433286,.240204718994897,-.199142840181825,-.236647428513811},
			{-9.16300542796916e-009,-1.46889442020809e-005,-.000686102259776139,-.00822762250546759,-.0459768381119022,-.152635122644788,-.336990745110908,-.516669847707885,-.536099580547344,-.300066032674195}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{-.13073316265519,-.26865299545921,-.332817240216851,-.309681793511623,-.213756072768581,-.0817681519461757,.0404150742496602,.113730876059729,.119201651868669,.0640618161688986},
			{-.232321125172907,-.409258602732634,-.357785220355006,-.132894091442215,.107032242482493,.214799928567286,.152958804103683,.0068197898368121,-.0925766092644603,-.0732716909670717},
			{-.341201391347813,-.473082610038628,-.182388003035756,.182511591542363,.249986188541097,.0287286472449085,-.165135040664019,-.121556975536052,.0384805680887192,.0765502494498098},
			{-.448265699805767,-.431119897592287,.101119258555274,.297365715667201,-.0237121460614668,-.221610907252239,-.0135788851032055,.155528183618924,.0264361189354877,-.0746172731128625},
			{.00150661303158024,.0174646676628736,.060827019428775,.132127558885229,.217829855282603,.294033116918165,.334256079779963,.318641694631351,.241392652759809,.113971546520449},
			{.00726049606491637,.0783433923326487,.238282291825274,.41806669467124,.499472969673169,.406143415849555,.170510825470376,-.077474909236672,-.195911258835258,-.130480354366348},
			{.0204803923823613,.19935715707521,.492971056033385,.59539889112151,.319238362842028,-.126234591258232,-.337591878134575,-.173021506144313,.0976252198049579,.1365217642077},
			{.0444476202705045,.377063440252612,.686669034698865,.381679636063987,-.252334234943452,-.372822170863378,.0620863830924046,.279028240875838,.0230022079648012,-.133375053045642},
			{-3.10794689696649e-005,-.00186307442685919,-.0154165466384745,-.0591852546271217,-.146572027698836,-.267059963618762,-.378778507598704,-.423057340461501,-.355924306714669,-.177853752785117},
			{-.000212047376066624,-.0118917494198341,-.0868345071649731,-.274518846864829,-.511304767009956,-.610254147605341,-.429991147250034,-.0596300446331458,.216886258844056,.186612165970385},
			{-.000813522866900639,-.0418162799719245,-.257709243250501,-.613129939717922,-.697897689320284,-.245040026788325,.314657359662651,.349392156168293,-.0387713739005192,-.183085004282902},
			{5.53998822800774e-007,.000171736839152697,.00337630014896861,.0229085188830747,.0852213031700891,.209596827719864,.370898198340224,.485354710148538,.453476337188524,.239824201609527},
			{4.88238700670906e-006,.00141947110604439,.0247527787190417,.139595292831676,.397312668856578,.663187290839078,.642633219433286,.240204718994897,-.199142840181825,-.236647428513811},
			{-9.16300542796916e-009,-1.46889442020809e-005,-.000686102259776139,-.00822762250546759,-.0459768381119022,-.152635122644788,-.336990745110908,-.516669847707885,-.536099580547344,-.300066032674195}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,33.98823325,-.4082482906,1.750000000,-7.678758582,34.53410150,.3227486122,-2.459674775,14.75000000,-.2645751311,2.958039892,.2236067977},
			{0.,-.7500000000,2.857738034,-12.37436867,55.33985905,-.7216878367,2.828427124,-13.26806944,63.17066828,.9128709293,-5.850213671,33.43604921,-.9354143468,8.844691708,.9035079028},
			{0.,0.,1.166666667,-7.216878367,40.02082790,0.,1.587713241,-9.500000000,52.88300766,.9316949905,-6.067673910,36.47616523,-1.636634177,13.68699853,2.074812507},
			{0.,0.,0.,-1.875000000,15.87608264,0.,0.,-2.901185103,23.88339730,0.,-2.627379873,21.21428571,-1.218935425,11.40958244,2.747169230},
			{0.,0.,0.,0.,3.100000000,0.,0.,0.,5.080682370,0.,0.,5.494338457,0.,4.138710979,1.621428571},
			{0.,.4330127020,-.7071067810,2.245365598,-7.850689990,-.2500000000,0.,.4949747467,-3.162277660,.3162277660,-1.004158022,3.091022771,-.3240370350,2.070196679,.3129843186},
			{0.,0.,-.8660254040,3.500000000,-14.98165545,0.,-.2500000000,1.558845727,-8.714212528,.5809475019,-1.565247584,4.964285714,-1.020504077,5.704791219,1.293725044},
			{0.,0.,0.,1.659882024,-10.71525393,0.,0.,1.275000000,-8.720665110,0.,-.5486726408,1.773290113,-.9983468483,5.709015429,2.250018896},
			{0.,0.,0.,0.,-3.117691454,0.,0.,0.,-3.300000000,0.,0.,-.4979550017,0.,1.996693696,1.554721797},
			{0.,0.,.3726779962,-.6454972245,2.020725943,0.,-.3227486122,.2236067977,-.2500000000,.8333333333e-1,.2886751347,-1.152673615,-.1463850110,.1636634177,.1855768722},
			{0.,0.,0.,-.8385254914,3.500000000,0.,0.,.9682458365e-1,.2886751347,0.,.6250000000,-2.395787117,-.2746751327,-.1889822365,.6190476190},
			{0.,0.,0.,0.,1.788854382,0.,0.,0.,.6454972245,0.,0.,-1.357142857,0.,-.6338656908,.6016087651},
			{0.,0.,0.,.3307189139,-.5916079782,0.,0.,-.3436931772,.3415650256,0.,.1479019946,.1889822365,-.2500000000e-1,-.2236067977,.5634361697e-1},
			{0.,0.,0.,0.,-.7937253933,0.,0.,0.,.3055050464,0.,0.,.5070925528,0.,-.4500000000,.1070899340},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,-.3464101616,0.,0.,.1916629694,0.,-.5669467095e-1,.7142857143e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,33.98823325,-.4082482906,1.750000000,-7.678758582,34.53410150,.3227486122,-2.459674775,14.75000000,-.2645751311,2.958039892,.2236067977},
			{0.,-.7500000000,2.857738034,-12.37436867,55.33985905,-.7216878367,2.828427124,-13.26806944,63.17066828,.9128709293,-5.850213671,33.43604921,-.9354143468,8.844691708,.9035079028},
			{0.,0.,1.166666667,-7.216878367,40.02082790,0.,1.587713241,-9.500000000,52.88300766,.9316949905,-6.067673910,36.47616523,-1.636634177,13.68699853,2.074812507},
			{0.,0.,0.,-1.875000000,15.87608264,0.,0.,-2.901185103,23.88339730,0.,-2.627379873,21.21428571,-1.218935425,11.40958244,2.747169230},
			{0.,0.,0.,0.,3.100000000,0.,0.,0.,5.080682370,0.,0.,5.494338457,0.,4.138710979,1.621428571},
			{0.,.4330127020,-.7071067810,2.245365598,-7.850689990,-.2500000000,0.,.4949747467,-3.162277660,.3162277660,-1.004158022,3.091022771,-.3240370350,2.070196679,.3129843186},
			{0.,0.,-.8660254040,3.500000000,-14.98165545,0.,-.2500000000,1.558845727,-8.714212528,.5809475019,-1.565247584,4.964285714,-1.020504077,5.704791219,1.293725044},
			{0.,0.,0.,1.659882024,-10.71525393,0.,0.,1.275000000,-8.720665110,0.,-.5486726408,1.773290113,-.9983468483,5.709015429,2.250018896},
			{0.,0.,0.,0.,-3.117691454,0.,0.,0.,-3.300000000,0.,0.,-.4979550017,0.,1.996693696,1.554721797},
			{0.,0.,.3726779962,-.6454972245,2.020725943,0.,-.3227486122,.2236067977,-.2500000000,.8333333333e-1,.2886751347,-1.152673615,-.1463850110,.1636634177,.1855768722},
			{0.,0.,0.,-.8385254914,3.500000000,0.,0.,.9682458365e-1,.2886751347,0.,.6250000000,-2.395787117,-.2746751327,-.1889822365,.6190476190},
			{0.,0.,0.,0.,1.788854382,0.,0.,0.,.6454972245,0.,0.,-1.357142857,0.,-.6338656908,.6016087651},
			{0.,0.,0.,.3307189139,-.5916079782,0.,0.,-.3436931772,.3415650256,0.,.1479019946,.1889822365,-.2500000000e-1,-.2236067977,.5634361697e-1},
			{0.,0.,0.,0.,-.7937253933,0.,0.,0.,.3055050464,0.,0.,.5070925528,0.,-.4500000000,.1070899340},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,-.3464101616,0.,0.,.1916629694,0.,-.5669467095e-1,.7142857143e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.5000000000,0.,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,-.6454972245e-1,2.904737510,1.341640786,.7500000000,-11.64130577,-6.507687761,50.31152948},
			{0.,-.2500000000,0.,0.,0.,.7216878367,.3535533905,.1632993162,.9128709293e-1,-2.464751509,-1.581138830,-.9596449171,10.85080642,7.410417377,-50.14468862},
			{0.,0.,.1666666667,0.,0.,0.,-.4330127020,-.2000000000,-.1118033988,1.155301788,1.290994449,.9691236663,-6.197388082,-6.026182953,33.05867928},
			{0.,0.,0.,-.1250000000,0.,0.,0.,.3031088914,.1290994449,0.,-.7267220925,-.7857142857,1.974864371,3.507390157,-14.50782199},
			{0.,0.,0.,0.,.1000000000,0.,0.,0.,-.2309401077,0.,0.,.5111012519,0.,-1.266180985,3.502380952},
			{0.,-.4330127020,0.,0.,0.,-.7500000000,.6123724358,.2828427124,.1581138830,3.320391543,-1.278019301,-1.195584279,-15.76980237,1.794170455,75.89869728},
			{0.,0.,.2886751347,0.,0.,0.,-.2500000000,-.3464101616,-.1936491673,-1.484643616,.4472135954,1.107142857,8.919961562,-.8451542550e-1,-49.83237207},
			{0.,0.,0.,-.2165063510,0.,0.,0.,.3250000000,.2236067977,0.,.3227486122e-1,-.7010681843,-2.776822654,-.7563225568,21.71637090},
			{0.,0.,0.,0.,.1732050808,0.,0.,0.,-.3000000000,0.,0.,.2766416676,0.,.5237229368,-5.171408842},
			{0.,0.,.3726779962,0.,0.,0.,.9682458365,-.4472135954,-.2500000000,1.083333333,-4.041451885,-.4610694458e-1,-8.002380602,16.80277755,48.86857636},
			{0.,0.,0.,-.2795084971,0.,0.,0.,-.9682458365e-1,.2886751347,0.,1.375000000,.7985957060,2.387560769,-7.275816105,-21.01190476},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,-.1290994449,0.,0.,-.6428571429,0.,1.774823935,4.876757779},
			{0.,0.,0.,-.3307189139,0.,0.,0.,-1.031079532,.3415650256,0.,-1.922725929,3.968626966,-1.575000000,13.75181806,16.26921940},
			{0.,0.,0.,0.,.2645751311,0.,0.,0.,.3055050464,0.,0.,-.8451542550,0.,-3.450000000,-3.647357166},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,1.039230485,0.,0.,2.491618602,0.,3.571764270,2.292857143}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.5000000000,0.,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,-.6454972245e-1,2.904737510,1.341640786,.7500000000,-11.64130577,-6.507687761,50.31152948},
			{0.,-.2500000000,0.,0.,0.,.7216878367,.3535533905,.1632993162,.9128709293e-1,-2.464751509,-1.581138830,-.9596449171,10.85080642,7.410417377,-50.14468862},
			{0.,0.,.1666666667,0.,0.,0.,-.4330127020,-.2000000000,-.1118033988,1.155301788,1.290994449,.9691236663,-6.197388082,-6.026182953,33.05867928},
			{0.,0.,0.,-.1250000000,0.,0.,0.,.3031088914,.1290994449,0.,-.7267220925,-.7857142857,1.974864371,3.507390157,-14.50782199},
			{0.,0.,0.,0.,.1000000000,0.,0.,0.,-.2309401077,0.,0.,.5111012519,0.,-1.266180985,3.502380952},
			{0.,-.4330127020,0.,0.,0.,-.7500000000,.6123724358,.2828427124,.1581138830,3.320391543,-1.278019301,-1.195584279,-15.76980237,1.794170455,75.89869728},
			{0.,0.,.2886751347,0.,0.,0.,-.2500000000,-.3464101616,-.1936491673,-1.484643616,.4472135954,1.107142857,8.919961562,-.8451542550e-1,-49.83237207},
			{0.,0.,0.,-.2165063510,0.,0.,0.,.3250000000,.2236067977,0.,.3227486122e-1,-.7010681843,-2.776822654,-.7563225568,21.71637090},
			{0.,0.,0.,0.,.1732050808,0.,0.,0.,-.3000000000,0.,0.,.2766416676,0.,.5237229368,-5.171408842},
			{0.,0.,.3726779962,0.,0.,0.,.9682458365,-.4472135954,-.2500000000,1.083333333,-4.041451885,-.4610694458e-1,-8.002380602,16.80277755,48.86857636},
			{0.,0.,0.,-.2795084971,0.,0.,0.,-.9682458365e-1,.2886751347,0.,1.375000000,.7985957060,2.387560769,-7.275816105,-21.01190476},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,-.1290994449,0.,0.,-.6428571429,0.,1.774823935,4.876757779},
			{0.,0.,0.,-.3307189139,0.,0.,0.,-1.031079532,.3415650256,0.,-1.922725929,3.968626966,-1.575000000,13.75181806,16.26921940},
			{0.,0.,0.,0.,.2645751311,0.,0.,0.,.3055050464,0.,0.,-.8451542550,0.,-3.450000000,-3.647357166},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,1.039230485,0.,0.,2.491618602,0.,3.571764270,2.292857143}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,-.5590169942e-1,.4082482906,.1250000000,-.2453738645,.3227486122e-1,-.1613743061,.5590169942e-1,.1250000000,.6614378278e-1,-.7395099729e-1,-.2795084971e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,0.,.7216878367,-.7071067810,.3163924252,.9128709293e-1,.3651483717,.3952847075e-1,-.2525381361,-.1403121520,.8964214566e-1,.5646924394e-1},
			{0.,0.,.8333333333e-1,-.2165063510,.2420614591,0.,-.2165063510,.4250000000,-.3913118960,.5776508940,-.6132223630,.3608439183,.3218713882,-.3049687729e-2,-.1210307296},
			{0.,0.,0.,-.3125000000e-1,.1118033988,0.,0.,.7577722285e-1,-.2259240285,0.,-.1816805231,.3750000000,.4937160930,-.5387858374,.2888254471},
			{0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,-.2886751347e-1,0.,0.,.6388765648e-1,0.,-.1582726231,.4377976190},
			{0.,-.4330127020,.7071067810,-.4592793268,0.,-.7500000000,0.,.1237436867,.1581138830,-.3162277660,-.1141088661,.2916059217e-1,.1890216038,-.1725163898e-1,-.9780759959e-1},
			{0.,0.,.1443375673,-.3750000000,.4192627457,0.,-.1250000000,.3897114318,-.4357106264,-.7423218081,.2795084971,.8928571429e-2,-.3496171374,-.8451542550e-1,.1796840339},
			{0.,0.,0.,-.5412658775e-1,.1936491673,0.,0.,.8125000000e-1,-.2795084971,0.,.8068715303e-2,.1340277411,-.6942056633,.3964594048,-.3527181261},
			{0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,-.3750000000e-1,0.,0.,.3458020845e-1,0.,.6546536711e-1,-.6464261051},
			{0.,0.,.1863389981,-.4841229182,.5412658775,0.,.4841229182,-.3913118960,.6250000000e-1,.5416666667,.3608439183,-.1959545145,.1341862601,.9547032698e-1,-.1546473936},
			{0.,0.,0.,-.6987712428e-1,.2500000000,0.,0.,-.2420614591e-1,-.7216878367e-1,0.,.3437500000,-.5190872089,.5968901920,.1653594569,.1994047619},
			{0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,-.1613743061e-1,0.,0.,-.8035714286e-1,0.,.2218529919,.6095947223},
			{0.,0.,0.,-.8267972847e-1,.2958039891,0.,0.,-.2577698829,.4269562821,0.,-.4806814823,.4724555913e-1,-.3937500000,-.4192627457,-.1056442818e-1},
			{0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,.3818813080e-1,0.,0.,-.1056442819,0.,-.4312500000,-.4559196456},
			{0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,.1299038106,0.,0.,.3114523254,0.,.4464705337,.2866071429}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,-.5590169942e-1,.4082482906,.1250000000,-.2453738645,.3227486122e-1,-.1613743061,.5590169942e-1,.1250000000,.6614378278e-1,-.7395099729e-1,-.2795084971e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,0.,.7216878367,-.7071067810,.3163924252,.9128709293e-1,.3651483717,.3952847075e-1,-.2525381361,-.1403121520,.8964214566e-1,.5646924394e-1},
			{0.,0.,.8333333333e-1,-.2165063510,.2420614591,0.,-.2165063510,.4250000000,-.3913118960,.5776508940,-.6132223630,.3608439183,.3218713882,-.3049687729e-2,-.1210307296},
			{0.,0.,0.,-.3125000000e-1,.1118033988,0.,0.,.7577722285e-1,-.2259240285,0.,-.1816805231,.3750000000,.4937160930,-.5387858374,.2888254471},
			{0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,-.2886751347e-1,0.,0.,.6388765648e-1,0.,-.1582726231,.4377976190},
			{0.,-.4330127020,.7071067810,-.4592793268,0.,-.7500000000,0.,.1237436867,.1581138830,-.3162277660,-.1141088661,.2916059217e-1,.1890216038,-.1725163898e-1,-.9780759959e-1},
			{0.,0.,.1443375673,-.3750000000,.4192627457,0.,-.1250000000,.3897114318,-.4357106264,-.7423218081,.2795084971,.8928571429e-2,-.3496171374,-.8451542550e-1,.1796840339},
			{0.,0.,0.,-.5412658775e-1,.1936491673,0.,0.,.8125000000e-1,-.2795084971,0.,.8068715303e-2,.1340277411,-.6942056633,.3964594048,-.3527181261},
			{0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,-.3750000000e-1,0.,0.,.3458020845e-1,0.,.6546536711e-1,-.6464261051},
			{0.,0.,.1863389981,-.4841229182,.5412658775,0.,.4841229182,-.3913118960,.6250000000e-1,.5416666667,.3608439183,-.1959545145,.1341862601,.9547032698e-1,-.1546473936},
			{0.,0.,0.,-.6987712428e-1,.2500000000,0.,0.,-.2420614591e-1,-.7216878367e-1,0.,.3437500000,-.5190872089,.5968901920,.1653594569,.1994047619},
			{0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,-.1613743061e-1,0.,0.,-.8035714286e-1,0.,.2218529919,.6095947223},
			{0.,0.,0.,-.8267972847e-1,.2958039891,0.,0.,-.2577698829,.4269562821,0.,-.4806814823,.4724555913e-1,-.3937500000,-.4192627457,-.1056442818e-1},
			{0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,.3818813080e-1,0.,0.,-.1056442819,0.,-.4312500000,-.4559196456},
			{0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,.1299038106,0.,0.,.3114523254,0.,.4464705337,.2866071429}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,.5590169942e-1,-.4082482906,-.1250000000,.2453738645,-.3227486122e-1,.1613743061,-.5590169942e-1,-.1250000000,-.6614378278e-1,.7395099729e-1,.2795084971e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,0.,-.7216878367,-.3535533905,.5103103630e-1,-.9128709293e-1,.4564354646,.1185854122,-.1515228816,-.2338535867,.5976143047e-1,.1129384878},
			{0.,0.,.5833333333,.2165063510,-.2420614591,0.,.7938566203,.3750000000,-.1677050983,.4658474951,.9682458365e-1,.3092947872e-1,-.4091585443,-.1189378215,.2593515633},
			{0.,0.,0.,-.4687500000,-.1118033988,0.,0.,-.7252962758,-.2581988897,0.,-.6568449682,-.2857142857,-.3047338563,.6338656908e-1,.3433961537},
			{0.,0.,0.,0.,.3875000000,0.,0.,0.,.6350852963,0.,0.,.6867923071,0.,.5173388725,.2026785714},
			{0.,.4330127020,.7071067810,.2551551816,-.1825741858,-.2500000000,-.6123724358,-.4065863991,.1581138830,.1581138830,.2966830521,.2916059217e-1,-.8100925875e-1,-.1035098339,.3912303983e-1},
			{0.,0.,-.4330127020,-.6250000000,-.5590169942e-1,0.,-.1250000000,-.4330127020e-1,.4841229182e-1,.2904737510,.6149186937,.2767857143,-.2551260193,-.4278593414,.1617156305},
			{0.,0.,0.,.4149705060,.5163977793,0.,0.,.3187500000,.3354101966,0.,-.1371681602,-.3917733971,-.2495867120,-.4025587803,.2812523620},
			{0.,0.,0.,0.,-.3897114318,0.,0.,0.,-.4125000000,0.,0.,-.6224437520e-1,0.,.2495867120,.1943402247},
			{0.,0.,.1863389981,.4841229182,.4330127020,0.,-.1613743061,-.5031152948,-.5625000000,.4166666667e-1,.2165063510,.3803822929,-.3659625275e-1,-.1432054905,.2319710903e-1},
			{0.,0.,0.,-.2096313728,-.5000000000,0.,0.,.2420614591e-1,.1443375673,0.,.1562500000,.3992978531,-.6866878320e-1,-.3307189139,.7738095238e-1},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,.8068715303e-1,0.,0.,-.1696428571,0.,-.7923321136e-1,.7520109565e-1},
			{0.,0.,0.,.8267972847e-1,.2958039891,0.,0.,-.8592329430e-1,-.3415650256,0.,.3697549864e-1,.1889822365,-.6250000000e-2,-.5590169942e-1,.7042952119e-2},
			{0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,.3818813080e-1,0.,0.,.6338656908e-1,0.,-.5625000000e-1,.1338624175e-1},
			{0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,-.4330127020e-1,0.,0.,.2395787117e-1,0.,-.7086833870e-2,.8928571429e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,.5590169942e-1,-.4082482906,-.1250000000,.2453738645,-.3227486122e-1,.1613743061,-.5590169942e-1,-.1250000000,-.6614378278e-1,.7395099729e-1,.2795084971e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,0.,-.7216878367,-.3535533905,.5103103630e-1,-.9128709293e-1,.4564354646,.1185854122,-.1515228816,-.2338535867,.5976143047e-1,.1129384878},
			{0.,0.,.5833333333,.2165063510,-.2420614591,0.,.7938566203,.3750000000,-.1677050983,.4658474951,.9682458365e-1,.3092947872e-1,-.4091585443,-.1189378215,.2593515633},
			{0.,0.,0.,-.4687500000,-.1118033988,0.,0.,-.7252962758,-.2581988897,0.,-.6568449682,-.2857142857,-.3047338563,.6338656908e-1,.3433961537},
			{0.,0.,0.,0.,.3875000000,0.,0.,0.,.6350852963,0.,0.,.6867923071,0.,.5173388725,.2026785714},
			{0.,.4330127020,.7071067810,.2551551816,-.1825741858,-.2500000000,-.6123724358,-.4065863991,.1581138830,.1581138830,.2966830521,.2916059217e-1,-.8100925875e-1,-.1035098339,.3912303983e-1},
			{0.,0.,-.4330127020,-.6250000000,-.5590169942e-1,0.,-.1250000000,-.4330127020e-1,.4841229182e-1,.2904737510,.6149186937,.2767857143,-.2551260193,-.4278593414,.1617156305},
			{0.,0.,0.,.4149705060,.5163977793,0.,0.,.3187500000,.3354101966,0.,-.1371681602,-.3917733971,-.2495867120,-.4025587803,.2812523620},
			{0.,0.,0.,0.,-.3897114318,0.,0.,0.,-.4125000000,0.,0.,-.6224437520e-1,0.,.2495867120,.1943402247},
			{0.,0.,.1863389981,.4841229182,.4330127020,0.,-.1613743061,-.5031152948,-.5625000000,.4166666667e-1,.2165063510,.3803822929,-.3659625275e-1,-.1432054905,.2319710903e-1},
			{0.,0.,0.,-.2096313728,-.5000000000,0.,0.,.2420614591e-1,.1443375673,0.,.1562500000,.3992978531,-.6866878320e-1,-.3307189139,.7738095238e-1},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,.8068715303e-1,0.,0.,-.1696428571,0.,-.7923321136e-1,.7520109565e-1},
			{0.,0.,0.,.8267972847e-1,.2958039891,0.,0.,-.8592329430e-1,-.3415650256,0.,.3697549864e-1,.1889822365,-.6250000000e-2,-.5590169942e-1,.7042952119e-2},
			{0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,.3818813080e-1,0.,0.,.6338656908e-1,0.,-.5625000000e-1,.1338624175e-1},
			{0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,-.4330127020e-1,0.,0.,.2395787117e-1,0.,-.7086833870e-2,.8928571429e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.1178511302,.2357022604e-1,.7856742012e-2,.7071067810e-1,.1178511302e-1,.4714045206e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.1000000000,.1190476190e-1,.7619047619e-1},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,.5248906592e-1,0.,.5248906592e-1},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,0.,0.,.3367175148e-2,.3086577219e-2,0.,-.5050762721e-2,-.7482611439e-3,.1010152544e-1,-.3928371006e-2,.1795826746e-1},
			{0.,0.,0.,0.,.5019488348e-3,0.,0.,0.,-.6274360436e-3,0.,0.,.8365813917e-3,0.,-.1254872087e-2,.2509744174e-2},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,0.,.9622504490e-2,.5498573994e-2,0.,.4123930495e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,0.,.1346870059e-1,.5892556509e-2,0.,.7576144082e-2,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,0.,.4860098696e-2,0.,0.,.4860098696e-2,0.},
			{0.,0.,0.,0.,-.8694008848e-3,0.,0.,0.,.9780759957e-3,0.,0.,-.1086751106e-2,0.,.1086751106e-2,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,0.,0.,.4347004424e-2,.4890379978e-2,0.,0.,.1086751106e-2,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,0.,0.,.3764616260e-2,.3137180218e-2,0.,0.,.1673162783e-2,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,0.,0.,0.,-.9820927513e-3,0.,0.,.6547285010e-3,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,0.,0.,0.,.7423923386e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,0.,0.,0.,.6640158940e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.1178511302,.2357022604e-1,.7856742012e-2,.7071067810e-1,.1178511302e-1,.4714045206e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.1000000000,.1190476190e-1,.7619047619e-1},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,.5248906592e-1,0.,.5248906592e-1},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,0.,0.,.3367175148e-2,.3086577219e-2,0.,-.5050762721e-2,-.7482611439e-3,.1010152544e-1,-.3928371006e-2,.1795826746e-1},
			{0.,0.,0.,0.,.5019488348e-3,0.,0.,0.,-.6274360436e-3,0.,0.,.8365813917e-3,0.,-.1254872087e-2,.2509744174e-2},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,0.,.9622504490e-2,.5498573994e-2,0.,.4123930495e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,0.,.1346870059e-1,.5892556509e-2,0.,.7576144082e-2,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,0.,.4860098696e-2,0.,0.,.4860098696e-2,0.},
			{0.,0.,0.,0.,-.8694008848e-3,0.,0.,0.,.9780759957e-3,0.,0.,-.1086751106e-2,0.,.1086751106e-2,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,0.,0.,.4347004424e-2,.4890379978e-2,0.,0.,.1086751106e-2,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,0.,0.,.3764616260e-2,.3137180218e-2,0.,0.,.1673162783e-2,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,0.,0.,0.,-.9820927513e-3,0.,0.,.6547285010e-3,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,0.,0.,0.,.7423923386e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,0.,0.,0.,.6640158940e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,4.242640686,-4.898979486,5.477225575,5.477225575,-6.324555320,7.071067810,-7.483314774,8.366600265,9.486832980},
			{0.,6.,-19.59591794,42.42640686,-75.89466384,3.464101616,-25.45584412,63.68673332,-120.4989626,-10.95445115,56.92099788,-127.2792206,22.44994432,-100.3992032,-37.94733192},
			{0.,0.,24.49489743,-127.2792206,398.4469852,0.,21.21320343,-161.6663230,575.1086854,5.477225575,-94.86832980,487.9036789,-22.44994432,250.9980080,56.92099788},
			{0.,0.,0.,98.99494934,-708.3501958,0.,0.,102.8785692,-920.1738966,0.,44.27188724,-622.2539673,7.483314774,-234.2648074,-37.94733192},
			{0.,0.,0.,0.,398.4469852,0.,0.,0.,460.0869483,0.,0.,254.5584412,0.,75.29940238,9.486832980},
			{0.,0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-10.95445115,-32.86335345,37.94733192,-42.42640686,89.79977729,-100.3992032,-189.7366596},
			{0.,0.,0.,0.,0.,0.,42.42640686,-117.5755077,230.0434742,32.86335345,-303.5786554,721.2489166,-179.5995546,1104.391235,569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,205.7571384,-920.1738966,0.,265.6313234,-2206.173157,89.79977729,-1907.584860,-569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,1527.350647,0.,903.5928286,189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,42.42640686,-224.4994432,250.9980080,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-678.8225098,224.4994432,-2509.980080,-1707.629936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,0.,2258.982072,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-167.3320053,-1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,4.242640686,-4.898979486,5.477225575,5.477225575,-6.324555320,7.071067810,-7.483314774,8.366600265,9.486832980},
			{0.,6.,-19.59591794,42.42640686,-75.89466384,3.464101616,-25.45584412,63.68673332,-120.4989626,-10.95445115,56.92099788,-127.2792206,22.44994432,-100.3992032,-37.94733192},
			{0.,0.,24.49489743,-127.2792206,398.4469852,0.,21.21320343,-161.6663230,575.1086854,5.477225575,-94.86832980,487.9036789,-22.44994432,250.9980080,56.92099788},
			{0.,0.,0.,98.99494934,-708.3501958,0.,0.,102.8785692,-920.1738966,0.,44.27188724,-622.2539673,7.483314774,-234.2648074,-37.94733192},
			{0.,0.,0.,0.,398.4469852,0.,0.,0.,460.0869483,0.,0.,254.5584412,0.,75.29940238,9.486832980},
			{0.,0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-10.95445115,-32.86335345,37.94733192,-42.42640686,89.79977729,-100.3992032,-189.7366596},
			{0.,0.,0.,0.,0.,0.,42.42640686,-117.5755077,230.0434742,32.86335345,-303.5786554,721.2489166,-179.5995546,1104.391235,569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,205.7571384,-920.1738966,0.,265.6313234,-2206.173157,89.79977729,-1907.584860,-569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,1527.350647,0.,903.5928286,189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,42.42640686,-224.4994432,250.9980080,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-678.8225098,224.4994432,-2509.980080,-1707.629936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,0.,2258.982072,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-167.3320053,-1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[15],
		T o_output[1]
	) {
		static const T matrix[1][15] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[15],
		T o_output[1]
	) {
		static const T matrix[1][15] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 7875;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[15][15] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,7.838367178,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,7.668115805,10.73312629,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,4.381780461,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,2.449489743,9.142857143,14.84614978,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,-2.993325910,.6110100926,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.},
			{4.732863826,3.346640106,-5.074680381,1.521277658,0.,0.,0.,0.,0.,4.364357806,17.00840129,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,10.95445115,-3.833259389,.5532833353,0.,0.,0.,0.,17.74823935,0.,0.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,-2.828427124,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,8.854377448,-3.098386677,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,3.919183589,4.200000000,0.,0.,-5.091168823,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,3.834057902,5.366563145,4.647580015,0.,6.640783086,-6.196773354,13.41640786,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,-.9128709293,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,2.529822128,-1.807392228,0.,0.,2.190890230,6.260990336,0.,0.,8.082903771,0.,0.,0.,0.,0.},
			{2.,3.535533905,1.649572198,-2.571428571,0.,1.224744872,4.571428571,7.423074892,0.,-7.377111136,11.49977817,0.,0.,0.,0.},
			{2.645751311,-1.496662955,.3055050464,0.,0.,5.184592560,-1.058300524,0.,0.,6.831300514,0.,0.,0.,0.,0.},
			{2.366431913,1.673320053,-2.537340190,.7606388294,0.,2.898275350,5.747048932,-2.195775165,0.,2.182178903,8.504200642,0.,10.06230590,0.,0.},
			{2.683281572,-1.897366596,.6639400021,-.9583148473e-1,0.,5.477225576,-1.916629694,.2766416676,0.,7.423074892,-1.071428571,0.,8.874119673,0.,0.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[10][15] = 
		{
			{1.414213562,-1.92171958555151,2.19799643560933,-2.29634628619806,2.23836106649518,3.41890639763464,-3.91413586924262,4.09536544265748,-4.00097073404889,5.33523806427656,-5.59797201662203,5.49216459339238,7.19422103343416,-7.09892453011936,9.0013491098729},
			{1.414213562,-1.59519010006696,1.2388862761662,-.514957807193037,-.353775594121333,3.23038451124486,-2.62173967966463,1.30644820527577,.307822029727901,4.7630793830775,-2.90243304119459,.518849452524734,6.06854814809742,-2.66498561200528,7.17423567654378},
			{1.414213562,-1.03822870489707,-.0622566343920988,1.10967175893133,-1.41978880536247,2.90882269973526,-.707254507004454,-1.57948219926848,2.54719860772099,3.86201434387853,.544351948717181,-3.18953813043078,4.43071582310306,2.19278566166382,4.71658607960972},
			{1.414213562,-.300186182387741,-1.13611369814234,1.22655366483195,.100858250439743,2.48271365058504,1.26648273457434,-2.5074905585056,.461151554581851,2.81340752869054,3.19379348794833,-2.33714259619681,2.75487590413567,4.77320941254595,2.50302559510191},
			{1.414213562,.553376983055106,-1.45368745167752,-.194424990185747,1.49978166741586,1.9899087271236,2.74863027720701,-.854309601049289,-2.52312426315122,1.80736427869175,4.12997714955087,1.65834138994758,1.4184747657974,4.48820932686102,1.03297875354963},
			{1.414213562,1.44662301694489,-.724355119014703,-1.6918152058931,.160475142667942,1.4741928888764,3.38025060487184,2.16048862424162,-1.70770592621981,.991945941760336,3.46032388717692,4.72311662432345,.576746649127288,2.68886700736674,.31115444533373},
			{1.414213562,2.30018618238774,.987081447023724,-1.35561938023839,-2.20870753342451,.981387965414959,3.10522366683209,4.42237267757001,2.83495589139377,.43960319187862,2.03901059601618,4.55407145300732,.170154848292169,1.03685515398483,.0611113213546603},
			{1.414213562,3.03822870489707,3.26615690329693,1.66498726415444,-.923099517802526,.555278916264744,2.17523616991868,4.50024407433993,6.26847807110811,.140734880491405,.792696592497175,2.35258131414952,.0308216365902829,.225963663364185,.00626330408480523},
			{1.414213562,3.59519010006696,5.47681396366563,6.3309248170395,5.76936797267538,.233717104755142,1.04841335581819,2.66785366107377,5.04596922575242,.0249321870529789,.15913874220381,.559597243722218,.00229823078236389,.0189957623925217,.00019657159112013},
			{1.414213562,3.92171958555151,6.96914454009643,10.2353232484353,13.4470270492568,.0451952183653594,.217799593005347,.614368644454251,1.33333501160483,.000932318622845177,.00636097475878139,.0244039729141499,1.66187980931977e-005,.000146461329046052,2.74870867417466e-007}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[10][15] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.37371117926928,4.13193546224797,-4.77114810448709,5.33430574565371,5.05405999869821,-5.83592580124217,6.52476340023727,-6.34960212623238,7.09907099144837,7.15379270875817},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-2.99666740648972,3.67015303548283,-4.23792768749944,4.73814719602452,3.40958356039142,-3.93704797266052,4.40175344784644,-2.40062592164153,2.68398137439781,.178027784911327},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-2.35354378347051,2.88249067692314,-3.32841353768903,3.72127946338712,1.05379652310981,-1.21681941252376,1.36044546108577,1.7591790753139,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-1.50132568517008,1.83874093225775,-2.12319514516607,2.37380433681192,-1.19541898829253,1.38035094930968,-1.54327927740643,3.34189684440344,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-.515715838247201,.631620327664824,-.729332332662821,.81541833693141,-2.55652048864375,2.95201558448519,-3.30045375807884,1.60938069663531,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,.515715838247201,-.631620327664824,.729332332662821,-.81541833693141,-2.55652048864375,2.95201558448519,-3.30045375807885,-1.60938069663533,1.79934231949429,2.78946777547417},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,1.50132568517008,-1.83874093225775,2.12319514516607,-2.37380433681192,-1.19541898829252,1.38035094930968,-1.54327927740643,-3.34189684440345,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,2.35354378347051,-2.88249067692314,3.32841353768903,-3.72127946338712,1.0537965231098,-1.21681941252376,1.36044546108577,-1.75917907531388,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,2.99666740648972,-3.67015303548283,4.23792768749944,-4.73814719602452,3.40958356039142,-3.93704797266052,4.40175344784644,2.40062592164155,-2.68398137439783,.178027784911158},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,3.37371117926928,-4.13193546224797,4.77114810448709,-5.33430574565371,5.05405999869821,-5.83592580124217,6.52476340023727,6.34960212623239,-7.09907099144838,7.15379270875832}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[10][15] = 
		{
			{1.414213562,3.92171958555151,6.96914454009643,10.2353232484353,13.4470270492568,-.0451952183653594,-.217799593005347,-.614368644454251,-1.33333501160483,.000932318622845132,.00636097475878108,.0244039729141487,-1.66187980893617e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,3.59519010006696,5.47681396366563,6.3309248170395,5.76936797267538,-.233717104755142,-1.04841335581819,-2.66785366107377,-5.04596922575242,.0249321870529788,.159138742203809,.559597243722216,-.00229823078236368,-.0189957623925199,.00019657159111697},
			{1.414213562,3.03822870489707,3.26615690329693,1.66498726415444,-.923099517802526,-.555278916264744,-2.17523616991868,-4.50024407433993,-6.26847807110811,.140734880491405,.792696592497175,2.35258131414951,-.0308216365902829,-.225963663364185,.00626330408481366},
			{1.414213562,2.30018618238774,.987081447023724,-1.35561938023839,-2.20870753342451,-.981387965414959,-3.10522366683209,-4.42237267757001,-2.83495589139377,.439603191878621,2.03901059601619,4.55407145300733,-.170154848292169,-1.03685515398483,.0611113213546687},
			{1.414213562,1.44662301694489,-.724355119014703,-1.6918152058931,.160475142667942,-1.4741928888764,-3.38025060487184,-2.16048862424162,1.70770592621981,.991945941760336,3.46032388717692,4.72311662432345,-.576746649127296,-2.68886700736678,.311154445333742},
			{1.414213562,.553376983055106,-1.45368745167752,-.194424990185747,1.49978166741586,-1.9899087271236,-2.74863027720701,.854309601049289,2.52312426315122,1.80736427869175,4.12997714955087,1.65834138994758,-1.41847476579742,-4.48820932686106,1.03297875354965},
			{1.414213562,-.300186182387741,-1.13611369814234,1.22655366483195,.100858250439743,-2.48271365058504,-1.26648273457434,2.5074905585056,-.461151554581851,2.81340752869054,3.19379348794833,-2.33714259619681,-2.75487590413568,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.03822870489707,-.0622566343920988,1.10967175893133,-1.41978880536247,-2.90882269973526,.707254507004454,1.57948219926848,-2.54719860772099,3.86201434387853,.544351948717181,-3.18953813043078,-4.43071582310304,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.59519010006696,1.2388862761662,-.514957807193037,-.353775594121333,-3.23038451124486,2.62173967966463,-1.30644820527577,-.307822029727901,4.7630793830775,-2.90243304119458,.518849452524734,-6.0685481480974,2.66498561200528,7.17423567654393},
			{1.414213562,-1.92171958555151,2.19799643560933,-2.29634628619806,2.23836106649518,-3.41890639763464,3.91413586924262,-4.09536544265748,4.00097073404889,5.33523806427656,-5.59797201662203,5.49216459339238,-7.19422103343416,7.09892453011937,9.00134910987265}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[15][10] = 
		{
			{-.0666713464862914,-.149451343590541,-.219086364157122,-.269266719261963,-.295524224714757,-.295524224714757,-.269266719261963,-.219086364157122,-.149451343590541,-.0666713464862914},
			{.0905970892324072,.168576592685474,.16083974734181,.0571555461131064,-.115637629421958,-.302296737069186,-.437956192288621,-.470674654782083,-.379932709650339,-.184884611630122},
			{-.103621819076306,-.130923096345775,.00964463927720944,.216316415190729,.303772972096487,.151366448890014,-.187940626427248,-.505984710909207,-.578779066349661,-.328551685958625},
			{.108258401003525,.0544197412932449,-.171907523449491,-.233536214191699,.0406284425731006,.353533858334046,.258110368117603,-.257935587575359,-.669039843415995,-.482531635271947},
			{-.105524759595668,.0373863179252117,.219950066547109,-.0192034434697929,-.313405150523241,-.0335340385442513,.420538627970391,.143004226828253,-.609695606023063,-.633943432044736},
			{-.161180106788003,-.341380763335387,-.450627406012926,-.472709481254668,-.415825621665808,-.308057935581016,-.186856585708771,-.0860223958211081,-.0246987698672132,-.00213067258216707},
			{.184527086816818,.277060359339501,.109566067415032,-.241138721975255,-.574373527114506,-.706361447936911,-.591235589736538,-.336982048861548,-.110794287988378,-.0102678920072701},
			{-.193070930476864,-.138062909860115,.244689360540118,.477427012543675,.178522670975115,-.451471222303063,-.842021186737924,-.697166353210365,-.281933595255738,-.0289636486661423},
			{.188620808842354,-.0325300344585138,-.394605522412419,-.0878034049814253,.527250170228969,.356854497200184,-.539776517640132,-.971096661439192,-.533248230919088,-.0628584273827533},
			{-.251523189241307,-.503352981601333,-.598293428384306,-.535673702440208,-.37767982241688,-.207284149381534,-.0837005897640986,-.0218022893353362,-.00263478512147966,-4.39530065166604e-005},
			{.263909455949514,.306723488668707,-.08432961782286,-.608099311980938,-.863029691810222,-.72309413582073,-.388228276260842,-.122802537682291,-.016817473305435,-.000299880275019948},
			{-.258921296192325,-.0548310028164179,.494115125970334,.444992705652827,-.346538929299739,-.986976377645973,-.867096676474837,-.364455907049916,-.0591371502798876,-.0011504950719751},
			{-.339162638412015,-.641312386098098,-.686395213472645,-.52452926221411,-.296414676313606,-.120521122714452,-.0323975380897773,-.00477480946005403,-.000242872567024778,-7.83472648717735e-007},
			{.334670002816657,.281630505360336,-.339700771258984,-.908820614314273,-.937888460850653,-.56188496465138,-.197417556313917,-.0350057153616537,-.0020074352890353,-6.90473792141432e-006},
			{-.424357452903393,-.758159297392549,-.730681505704979,-.47657688200433,-.215858660481788,-.0650210679957279,-.0116356153302471,-.000970295120692958,-2.0773304102378e-005,-1.29584465340518e-008}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{.0666713464862914,.149451343590541,.219086364157122,.269266719261963,.295524224714757,.295524224714757,.269266719261963,.219086364157122,.149451343590541,.0666713464862914},
			{-.0816553896850849,-.1830397666013,-.268324900917017,-.329783033481719,-.361941778623412,-.361941778623412,-.329783033481719,-.268324900917017,-.1830397666013,-.0816553896850849},
			{.0942875223977143,.211356116964865,.309834907440273,.380800646175515,.417934366491145,.417934366491145,.380800646175515,.309834907440273,.211356116964865,.0942875223977143},
			{-.105416654777859,-.236303322546677,-.346405957492347,-.425748065451799,-.467264826902148,-.467264826902148,-.425748065451799,-.346405957492347,-.236303322546677,-.105416654777859},
			{.112464933488868,.223927985096936,.257814675202576,.202128520894733,.0762032616355625,-.0762032616355625,-.202128520894733,-.257814675202576,-.223927985096936,-.112464933488868},
			{-.137740850431264,-.274254651167906,-.315757201061946,-.247555869200864,-.093329553823614,.093329553823614,.247555869200864,.315757201061946,.274254651167906,.137740850431264},
			{.159049434205836,.316681993468174,.364605010191818,.285852895545897,.107767686084793,-.107767686084793,-.285852895545897,-.364605010191818,-.316681993468174,-.159049434205836},
			{-.177822673316147,-.35406123228781,-.407640793823026,-.319593252971582,-.120487935919926,.120487935919926,.319593252971582,.407640793823026,.35406123228781,.177822673316147},
			{-.168480492667857,-.25478342209236,-.115436224404772,.160943274560492,.377756867686918,.377756867686918,.160943274560491,-.115436224404772,-.25478342209236,-.168480492667857},
			{.194544515581452,.294198554647266,.133294270462818,-.185841285775377,-.436196058475433,-.436196058475433,-.185841285775376,.133294270462818,.294198554647266,.194544515581452},
			{-.217507380699146,-.328923983467474,-.149027524851671,.207776873966102,.487682019031578,.487682019031578,.207776873966101,-.14902752485167,-.328923983467474,-.217507380699146},
			{.211668261704066,.179388384723805,-.192706073755905,-.44993079970221,-.237805491322023,.237805491322025,.449930799702211,.192706073755903,-.179388384723806,-.211668261704066},
			{-.236652310900817,-.20056231128787,.215451940275855,.503037926601612,.265874621982498,-.265874621982501,-.503037926601614,-.215451940275852,.200562311287872,.236652310900817},
			{-.23847649618836,-.0133032458257229,.440420079694959,.223537309141119,-.412177650856899,-.412177650856901,.223537309141112,.440420079694946,-.0133032458257102,-.238476496188365}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{-.13073316265519,-.26865299545921,-.332817240216851,-.309681793511623,-.213756072768581,-.0817681519461757,.0404150742496602,.113730876059729,.119201651868669,.0640618161688986},
			{-.232321125172907,-.409258602732634,-.357785220355006,-.132894091442215,.107032242482493,.214799928567286,.152958804103683,.0068197898368121,-.0925766092644603,-.0732716909670717},
			{-.341201391347813,-.473082610038628,-.182388003035756,.182511591542363,.249986188541097,.0287286472449085,-.165135040664019,-.121556975536052,.0384805680887192,.0765502494498098},
			{-.448265699805767,-.431119897592287,.101119258555274,.297365715667201,-.0237121460614668,-.221610907252239,-.0135788851032055,.155528183618924,.0264361189354877,-.0746172731128625},
			{.00150661303158024,.0174646676628736,.060827019428775,.132127558885229,.217829855282603,.294033116918165,.334256079779963,.318641694631351,.241392652759809,.113971546520449},
			{.00726049606491637,.0783433923326487,.238282291825274,.41806669467124,.499472969673169,.406143415849555,.170510825470376,-.077474909236672,-.195911258835258,-.130480354366348},
			{.0204803923823613,.19935715707521,.492971056033385,.59539889112151,.319238362842028,-.126234591258232,-.337591878134575,-.173021506144313,.0976252198049579,.1365217642077},
			{.0444476202705045,.377063440252612,.686669034698865,.381679636063987,-.252334234943452,-.372822170863378,.0620863830924046,.279028240875838,.0230022079648012,-.133375053045642},
			{-3.10794689696649e-005,-.00186307442685919,-.0154165466384745,-.0591852546271217,-.146572027698836,-.267059963618762,-.378778507598704,-.423057340461501,-.355924306714669,-.177853752785117},
			{-.000212047376066624,-.0118917494198341,-.0868345071649731,-.274518846864829,-.511304767009956,-.610254147605341,-.429991147250034,-.0596300446331458,.216886258844056,.186612165970385},
			{-.000813522866900639,-.0418162799719245,-.257709243250501,-.613129939717922,-.697897689320284,-.245040026788325,.314657359662651,.349392156168293,-.0387713739005192,-.183085004282902},
			{5.53998822800774e-007,.000171736839152697,.00337630014896861,.0229085188830747,.0852213031700891,.209596827719864,.370898198340224,.485354710148538,.453476337188524,.239824201609527},
			{4.88238700670906e-006,.00141947110604439,.0247527787190417,.139595292831676,.397312668856578,.663187290839078,.642633219433286,.240204718994897,-.199142840181825,-.236647428513811},
			{-9.16300542796916e-009,-1.46889442020809e-005,-.000686102259776139,-.00822762250546759,-.0459768381119022,-.152635122644788,-.336990745110908,-.516669847707885,-.536099580547344,-.300066032674195}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[15][15] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,33.98823325,-.4082482906,1.750000000,-7.678758582,34.53410150,.3227486122,-2.459674775,14.75000000,-.2645751311,2.958039892,.2236067977},
			{0.,-.7500000000,2.857738034,-12.37436867,55.33985905,-.7216878367,2.828427124,-13.26806944,63.17066828,.9128709293,-5.850213671,33.43604921,-.9354143468,8.844691708,.9035079028},
			{0.,0.,1.166666667,-7.216878367,40.02082790,0.,1.587713241,-9.500000000,52.88300766,.9316949905,-6.067673910,36.47616523,-1.636634177,13.68699853,2.074812507},
			{0.,0.,0.,-1.875000000,15.87608264,0.,0.,-2.901185103,23.88339730,0.,-2.627379873,21.21428571,-1.218935425,11.40958244,2.747169230},
			{0.,0.,0.,0.,3.100000000,0.,0.,0.,5.080682370,0.,0.,5.494338457,0.,4.138710979,1.621428571},
			{0.,.4330127020,-.7071067810,2.245365598,-7.850689990,-.2500000000,0.,.4949747467,-3.162277660,.3162277660,-1.004158022,3.091022771,-.3240370350,2.070196679,.3129843186},
			{0.,0.,-.8660254040,3.500000000,-14.98165545,0.,-.2500000000,1.558845727,-8.714212528,.5809475019,-1.565247584,4.964285714,-1.020504077,5.704791219,1.293725044},
			{0.,0.,0.,1.659882024,-10.71525393,0.,0.,1.275000000,-8.720665110,0.,-.5486726408,1.773290113,-.9983468483,5.709015429,2.250018896},
			{0.,0.,0.,0.,-3.117691454,0.,0.,0.,-3.300000000,0.,0.,-.4979550017,0.,1.996693696,1.554721797},
			{0.,0.,.3726779962,-.6454972245,2.020725943,0.,-.3227486122,.2236067977,-.2500000000,.8333333333e-1,.2886751347,-1.152673615,-.1463850110,.1636634177,.1855768722},
			{0.,0.,0.,-.8385254914,3.500000000,0.,0.,.9682458365e-1,.2886751347,0.,.6250000000,-2.395787117,-.2746751327,-.1889822365,.6190476190},
			{0.,0.,0.,0.,1.788854382,0.,0.,0.,.6454972245,0.,0.,-1.357142857,0.,-.6338656908,.6016087651},
			{0.,0.,0.,.3307189139,-.5916079782,0.,0.,-.3436931772,.3415650256,0.,.1479019946,.1889822365,-.2500000000e-1,-.2236067977,.5634361697e-1},
			{0.,0.,0.,0.,-.7937253933,0.,0.,0.,.3055050464,0.,0.,.5070925528,0.,-.4500000000,.1070899340},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,-.3464101616,0.,0.,.1916629694,0.,-.5669467095e-1,.7142857143e-2}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[15][15] = 
		{
			{.5000000000,0.,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,-.6454972245e-1,2.904737510,1.341640786,.7500000000,-11.64130577,-6.507687761,50.31152948},
			{0.,-.2500000000,0.,0.,0.,.7216878367,.3535533905,.1632993162,.9128709293e-1,-2.464751509,-1.581138830,-.9596449171,10.85080642,7.410417377,-50.14468862},
			{0.,0.,.1666666667,0.,0.,0.,-.4330127020,-.2000000000,-.1118033988,1.155301788,1.290994449,.9691236663,-6.197388082,-6.026182953,33.05867928},
			{0.,0.,0.,-.1250000000,0.,0.,0.,.3031088914,.1290994449,0.,-.7267220925,-.7857142857,1.974864371,3.507390157,-14.50782199},
			{0.,0.,0.,0.,.1000000000,0.,0.,0.,-.2309401077,0.,0.,.5111012519,0.,-1.266180985,3.502380952},
			{0.,-.4330127020,0.,0.,0.,-.7500000000,.6123724358,.2828427124,.1581138830,3.320391543,-1.278019301,-1.195584279,-15.76980237,1.794170455,75.89869728},
			{0.,0.,.2886751347,0.,0.,0.,-.2500000000,-.3464101616,-.1936491673,-1.484643616,.4472135954,1.107142857,8.919961562,-.8451542550e-1,-49.83237207},
			{0.,0.,0.,-.2165063510,0.,0.,0.,.3250000000,.2236067977,0.,.3227486122e-1,-.7010681843,-2.776822654,-.7563225568,21.71637090},
			{0.,0.,0.,0.,.1732050808,0.,0.,0.,-.3000000000,0.,0.,.2766416676,0.,.5237229368,-5.171408842},
			{0.,0.,.3726779962,0.,0.,0.,.9682458365,-.4472135954,-.2500000000,1.083333333,-4.041451885,-.4610694458e-1,-8.002380602,16.80277755,48.86857636},
			{0.,0.,0.,-.2795084971,0.,0.,0.,-.9682458365e-1,.2886751347,0.,1.375000000,.7985957060,2.387560769,-7.275816105,-21.01190476},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,-.1290994449,0.,0.,-.6428571429,0.,1.774823935,4.876757779},
			{0.,0.,0.,-.3307189139,0.,0.,0.,-1.031079532,.3415650256,0.,-1.922725929,3.968626966,-1.575000000,13.75181806,16.26921940},
			{0.,0.,0.,0.,.2645751311,0.,0.,0.,.3055050464,0.,0.,-.8451542550,0.,-3.450000000,-3.647357166},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,1.039230485,0.,0.,2.491618602,0.,3.571764270,2.292857143}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[15][15] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,-.5590169942e-1,.4082482906,.1250000000,-.2453738645,.3227486122e-1,-.1613743061,.5590169942e-1,.1250000000,.6614378278e-1,-.7395099729e-1,-.2795084971e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,0.,.7216878367,-.7071067810,.3163924252,.9128709293e-1,.3651483717,.3952847075e-1,-.2525381361,-.1403121520,.8964214566e-1,.5646924394e-1},
			{0.,0.,.8333333333e-1,-.2165063510,.2420614591,0.,-.2165063510,.4250000000,-.3913118960,.5776508940,-.6132223630,.3608439183,.3218713882,-.3049687729e-2,-.1210307296},
			{0.,0.,0.,-.3125000000e-1,.1118033988,0.,0.,.7577722285e-1,-.2259240285,0.,-.1816805231,.3750000000,.4937160930,-.5387858374,.2888254471},
			{0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,-.2886751347e-1,0.,0.,.6388765648e-1,0.,-.1582726231,.4377976190},
			{0.,-.4330127020,.7071067810,-.4592793268,0.,-.7500000000,0.,.1237436867,.1581138830,-.3162277660,-.1141088661,.2916059217e-1,.1890216038,-.1725163898e-1,-.9780759959e-1},
			{0.,0.,.1443375673,-.3750000000,.4192627457,0.,-.1250000000,.3897114318,-.4357106264,-.7423218081,.2795084971,.8928571429e-2,-.3496171374,-.8451542550e-1,.1796840339},
			{0.,0.,0.,-.5412658775e-1,.1936491673,0.,0.,.8125000000e-1,-.2795084971,0.,.8068715303e-2,.1340277411,-.6942056633,.3964594048,-.3527181261},
			{0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,-.3750000000e-1,0.,0.,.3458020845e-1,0.,.6546536711e-1,-.6464261051},
			{0.,0.,.1863389981,-.4841229182,.5412658775,0.,.4841229182,-.3913118960,.6250000000e-1,.5416666667,.3608439183,-.1959545145,.1341862601,.9547032698e-1,-.1546473936},
			{0.,0.,0.,-.6987712428e-1,.2500000000,0.,0.,-.2420614591e-1,-.7216878367e-1,0.,.3437500000,-.5190872089,.5968901920,.1653594569,.1994047619},
			{0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,-.1613743061e-1,0.,0.,-.8035714286e-1,0.,.2218529919,.6095947223},
			{0.,0.,0.,-.8267972847e-1,.2958039891,0.,0.,-.2577698829,.4269562821,0.,-.4806814823,.4724555913e-1,-.3937500000,-.4192627457,-.1056442818e-1},
			{0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,.3818813080e-1,0.,0.,-.1056442819,0.,-.4312500000,-.4559196456},
			{0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,.1299038106,0.,0.,.3114523254,0.,.4464705337,.2866071429}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[15][15] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,.5590169942e-1,-.4082482906,-.1250000000,.2453738645,-.3227486122e-1,.1613743061,-.5590169942e-1,-.1250000000,-.6614378278e-1,.7395099729e-1,.2795084971e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,0.,-.7216878367,-.3535533905,.5103103630e-1,-.9128709293e-1,.4564354646,.1185854122,-.1515228816,-.2338535867,.5976143047e-1,.1129384878},
			{0.,0.,.5833333333,.2165063510,-.2420614591,0.,.7938566203,.3750000000,-.1677050983,.4658474951,.9682458365e-1,.3092947872e-1,-.4091585443,-.1189378215,.2593515633},
			{0.,0.,0.,-.4687500000,-.1118033988,0.,0.,-.7252962758,-.2581988897,0.,-.6568449682,-.2857142857,-.3047338563,.6338656908e-1,.3433961537},
			{0.,0.,0.,0.,.3875000000,0.,0.,0.,.6350852963,0.,0.,.6867923071,0.,.5173388725,.2026785714},
			{0.,.4330127020,.7071067810,.2551551816,-.1825741858,-.2500000000,-.6123724358,-.4065863991,.1581138830,.1581138830,.2966830521,.2916059217e-1,-.8100925875e-1,-.1035098339,.3912303983e-1},
			{0.,0.,-.4330127020,-.6250000000,-.5590169942e-1,0.,-.1250000000,-.4330127020e-1,.4841229182e-1,.2904737510,.6149186937,.2767857143,-.2551260193,-.4278593414,.1617156305},
			{0.,0.,0.,.4149705060,.5163977793,0.,0.,.3187500000,.3354101966,0.,-.1371681602,-.3917733971,-.2495867120,-.4025587803,.2812523620},
			{0.,0.,0.,0.,-.3897114318,0.,0.,0.,-.4125000000,0.,0.,-.6224437520e-1,0.,.2495867120,.1943402247},
			{0.,0.,.1863389981,.4841229182,.4330127020,0.,-.1613743061,-.5031152948,-.5625000000,.4166666667e-1,.2165063510,.3803822929,-.3659625275e-1,-.1432054905,.2319710903e-1},
			{0.,0.,0.,-.2096313728,-.5000000000,0.,0.,.2420614591e-1,.1443375673,0.,.1562500000,.3992978531,-.6866878320e-1,-.3307189139,.7738095238e-1},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,.8068715303e-1,0.,0.,-.1696428571,0.,-.7923321136e-1,.7520109565e-1},
			{0.,0.,0.,.8267972847e-1,.2958039891,0.,0.,-.8592329430e-1,-.3415650256,0.,.3697549864e-1,.1889822365,-.6250000000e-2,-.5590169942e-1,.7042952119e-2},
			{0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,.3818813080e-1,0.,0.,.6338656908e-1,0.,-.5625000000e-1,.1338624175e-1},
			{0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,-.4330127020e-1,0.,0.,.2395787117e-1,0.,-.7086833870e-2,.8928571429e-3}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[15][15] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.1178511302,.2357022604e-1,.7856742012e-2,.7071067810e-1,.1178511302e-1,.4714045206e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.1000000000,.1190476190e-1,.7619047619e-1},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,.5248906592e-1,0.,.5248906592e-1},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,0.,0.,.3367175148e-2,.3086577219e-2,0.,-.5050762721e-2,-.7482611439e-3,.1010152544e-1,-.3928371006e-2,.1795826746e-1},
			{0.,0.,0.,0.,.5019488348e-3,0.,0.,0.,-.6274360436e-3,0.,0.,.8365813917e-3,0.,-.1254872087e-2,.2509744174e-2},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,0.,.9622504490e-2,.5498573994e-2,0.,.4123930495e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,0.,.1346870059e-1,.5892556509e-2,0.,.7576144082e-2,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,0.,.4860098696e-2,0.,0.,.4860098696e-2,0.},
			{0.,0.,0.,0.,-.8694008848e-3,0.,0.,0.,.9780759957e-3,0.,0.,-.1086751106e-2,0.,.1086751106e-2,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,0.,0.,.4347004424e-2,.4890379978e-2,0.,0.,.1086751106e-2,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,0.,0.,.3764616260e-2,.3137180218e-2,0.,0.,.1673162783e-2,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,0.,0.,0.,-.9820927513e-3,0.,0.,.6547285010e-3,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,0.,0.,0.,.7423923386e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,0.,0.,0.,.6640158940e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[15][15] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,4.242640686,-4.898979486,5.477225575,5.477225575,-6.324555320,7.071067810,-7.483314774,8.366600265,9.486832980},
			{0.,6.,-19.59591794,42.42640686,-75.89466384,3.464101616,-25.45584412,63.68673332,-120.4989626,-10.95445115,56.92099788,-127.2792206,22.44994432,-100.3992032,-37.94733192},
			{0.,0.,24.49489743,-127.2792206,398.4469852,0.,21.21320343,-161.6663230,575.1086854,5.477225575,-94.86832980,487.9036789,-22.44994432,250.9980080,56.92099788},
			{0.,0.,0.,98.99494934,-708.3501958,0.,0.,102.8785692,-920.1738966,0.,44.27188724,-622.2539673,7.483314774,-234.2648074,-37.94733192},
			{0.,0.,0.,0.,398.4469852,0.,0.,0.,460.0869483,0.,0.,254.5584412,0.,75.29940238,9.486832980},
			{0.,0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-10.95445115,-32.86335345,37.94733192,-42.42640686,89.79977729,-100.3992032,-189.7366596},
			{0.,0.,0.,0.,0.,0.,42.42640686,-117.5755077,230.0434742,32.86335345,-303.5786554,721.2489166,-179.5995546,1104.391235,569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,205.7571384,-920.1738966,0.,265.6313234,-2206.173157,89.79977729,-1907.584860,-569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,1527.350647,0.,903.5928286,189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,42.42640686,-224.4994432,250.9980080,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-678.8225098,224.4994432,-2509.980080,-1707.629936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,0.,2258.982072,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-167.3320053,-1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][15] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
