
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION3D_MATRICES_CONST_HPP_
#define CCOMPUTATION3D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (2)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (10)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (10)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS (9)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (6)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*2+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_3D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[10],
		T io_hv[10]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,3.535533905,0.,0.,6.123724358,0.,0.,0.},
			{2.581988897,0.,0.,-.9128709293,0.,0.,4.743416490,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,3.535533905,0.,0.,6.123724358,0.,0.,0.},
			{2.581988897,0.,0.,-.9128709293,0.,0.,4.743416490,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_dofs_to_nodes
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_dofs_to_nodes(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,-1.732050808,0.,-3.181980514,1.369306394}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_dofs_to_nodes
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_dofs_to_nodes(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,-1.732050808,0.,-3.181980514,1.369306394}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[10],
		T o_output[9]
	) {
		static const T matrix[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,1.732050808,0.,3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,1.732050808,3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,4.,0.,7.348469229,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[10],
		T o_output[9]
	) {
		static const T matrix[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,1.732050808,0.,3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,1.732050808,3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,4.,0.,7.348469229,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[10],
		T o_output[9]
	) {
		static const T matrix[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[10],
		T o_output[9]
	) {
		static const T matrix[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[10],
		T o_output[9]
	) {
		static const T matrix[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,4.,0.,7.348469229,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,-1.732050808,0.,-3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,-1.732050808,-3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[10],
		T o_output[9]
	) {
		static const T matrix[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,4.,0.,7.348469229,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,-1.732050808,0.,-3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,-1.732050808,-3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[9],
		T o_output[10]
	) {
		static const T matrix[10][9] = 
		{
			{-.5555555556e-1,-.2222222222,-.5555555556e-1,-.2222222222,-.8888888889,-.2222222222,-.5555555556e-1,-.2222222222,-.5555555556e-1},
			{.9622504490e-1,0.,-.9622504490e-1,.3849001795,0.,-.3849001795,.9622504490e-1,0.,-.9622504490e-1},
			{-.4969039948e-1,.9938079897e-1,-.4969039948e-1,-.1987615980,.3975231960,-.1987615980,-.4969039948e-1,.9938079897e-1,-.4969039948e-1},
			{.7856742012e-1,.3142696804,.7856742012e-1,-.1571348402,-.6285393608,-.1571348402,-.1571348402,-.6285393608,-.1571348402},
			{-.1360827635,0.,.1360827635,.2721655270,0.,-.2721655270,.2721655270,0.,-.2721655270},
			{0.,0.,0.,0.,0.,0.,-.1924500898,-.7698003590,-.1924500898},
			{-.1360827635,-.5443310539,-.1360827635,-.2721655270,-1.088662108,-.2721655270,0.,0.,0.},
			{.2357022604,0.,-.2357022604,.4714045206,0.,-.4714045206,0.,0.,0.},
			{.8333333333e-1,.3333333333,.8333333333e-1,-.3333333333,-1.333333333,-.3333333333,-.8333333333e-1,-.3333333333,-.8333333333e-1},
			{-.1936491673,-.7745966692,-.1936491673,-.2581988897,-1.032795559,-.2581988897,.2151657415e-1,.8606629657e-1,.2151657415e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[9],
		T o_output[10]
	) {
		static const T matrix[10][9] = 
		{
			{-.5555555556e-1,-.2222222222,-.5555555556e-1,-.2222222222,-.8888888889,-.2222222222,-.5555555556e-1,-.2222222222,-.5555555556e-1},
			{.9622504490e-1,0.,-.9622504490e-1,.3849001795,0.,-.3849001795,.9622504490e-1,0.,-.9622504490e-1},
			{-.4969039948e-1,.9938079897e-1,-.4969039948e-1,-.1987615980,.3975231960,-.1987615980,-.4969039948e-1,.9938079897e-1,-.4969039948e-1},
			{.7856742012e-1,.3142696804,.7856742012e-1,-.1571348402,-.6285393608,-.1571348402,-.1571348402,-.6285393608,-.1571348402},
			{-.1360827635,0.,.1360827635,.2721655270,0.,-.2721655270,.2721655270,0.,-.2721655270},
			{0.,0.,0.,0.,0.,0.,-.1924500898,-.7698003590,-.1924500898},
			{-.1360827635,-.5443310539,-.1360827635,-.2721655270,-1.088662108,-.2721655270,0.,0.,0.},
			{.2357022604,0.,-.2357022604,.4714045206,0.,-.4714045206,0.,0.,0.},
			{.8333333333e-1,.3333333333,.8333333333e-1,-.3333333333,-1.333333333,-.3333333333,-.8333333333e-1,-.3333333333,-.8333333333e-1},
			{-.1936491673,-.7745966692,-.1936491673,-.2581988897,-1.032795559,-.2581988897,.2151657415e-1,.8606629657e-1,.2151657415e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[9],
		T o_output[10]
	) {
		static const T matrix[10][9] = 
		{
			{-.3928371006e-1,-.1571348402,-.3928371006e-1,-.1571348402,-.6285393608,-.1571348402,-.3928371006e-1,-.1571348402,-.3928371006e-1},
			{.6804138176e-1,0.,-.6804138176e-1,.2721655270,0.,-.2721655270,.6804138176e-1,0.,-.6804138176e-1},
			{-.3513641844e-1,.7027283688e-1,-.3513641844e-1,-.1405456738,.2810913476,-.1405456738,-.3513641844e-1,.7027283688e-1,-.3513641844e-1},
			{.5555555556e-1,.2222222222,.5555555556e-1,.2222222222,.8888888889,.2222222222,.5555555556e-1,.2222222222,.5555555556e-1},
			{-.9622504490e-1,0.,.9622504490e-1,-.3849001795,0.,.3849001795,-.9622504490e-1,0.,.9622504490e-1},
			{-.6804138176e-1,-.2721655270,-.6804138176e-1,-.2721655270,-1.088662108,-.2721655270,-.6804138176e-1,-.2721655270,-.6804138176e-1},
			{.9622504490e-1,.3849001795,.9622504490e-1,0.,0.,0.,-.9622504490e-1,-.3849001795,-.9622504490e-1},
			{-.1666666667,0.,.1666666667,0.,0.,0.,.1666666667,0.,-.1666666667},
			{-.1178511302,-.4714045206,-.1178511302,0.,0.,0.,.1178511302,.4714045206,.1178511302},
			{-.6085806194e-1,-.2434322478,-.6085806194e-1,.1217161239,.4868644956,.1217161239,-.6085806194e-1,-.2434322478,-.6085806194e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[9],
		T o_output[10]
	) {
		static const T matrix[10][9] = 
		{
			{-.3928371006e-1,-.1571348402,-.3928371006e-1,-.1571348402,-.6285393608,-.1571348402,-.3928371006e-1,-.1571348402,-.3928371006e-1},
			{.6804138176e-1,0.,-.6804138176e-1,.2721655270,0.,-.2721655270,.6804138176e-1,0.,-.6804138176e-1},
			{-.3513641844e-1,.7027283688e-1,-.3513641844e-1,-.1405456738,.2810913476,-.1405456738,-.3513641844e-1,.7027283688e-1,-.3513641844e-1},
			{.5555555556e-1,.2222222222,.5555555556e-1,.2222222222,.8888888889,.2222222222,.5555555556e-1,.2222222222,.5555555556e-1},
			{-.9622504490e-1,0.,.9622504490e-1,-.3849001795,0.,.3849001795,-.9622504490e-1,0.,.9622504490e-1},
			{-.6804138176e-1,-.2721655270,-.6804138176e-1,-.2721655270,-1.088662108,-.2721655270,-.6804138176e-1,-.2721655270,-.6804138176e-1},
			{.9622504490e-1,.3849001795,.9622504490e-1,0.,0.,0.,-.9622504490e-1,-.3849001795,-.9622504490e-1},
			{-.1666666667,0.,.1666666667,0.,0.,0.,.1666666667,0.,-.1666666667},
			{-.1178511302,-.4714045206,-.1178511302,0.,0.,0.,.1178511302,.4714045206,.1178511302},
			{-.6085806194e-1,-.2434322478,-.6085806194e-1,.1217161239,.4868644956,.1217161239,-.6085806194e-1,-.2434322478,-.6085806194e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[9],
		T o_output[10]
	) {
		static const T matrix[10][9] = 
		{
			{-.3928371006e-1,-.1571348402,-.3928371006e-1,-.1571348402,-.6285393608,-.1571348402,-.3928371006e-1,-.1571348402,-.3928371006e-1},
			{.6804138176e-1,0.,-.6804138176e-1,.2721655270,0.,-.2721655270,.6804138176e-1,0.,-.6804138176e-1},
			{-.3513641844e-1,.7027283688e-1,-.3513641844e-1,-.1405456738,.2810913476,-.1405456738,-.3513641844e-1,.7027283688e-1,-.3513641844e-1},
			{-.1111111111,-.4444444444,-.1111111111,-.1111111111,-.4444444444,-.1111111111,.5555555556e-1,.2222222222,.5555555556e-1},
			{.1924500898,0.,-.1924500898,.1924500898,0.,-.1924500898,-.9622504490e-1,0.,.9622504490e-1},
			{-.1360827635,-.5443310539,-.1360827635,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1924500898,.7698003590,.1924500898,.9622504490e-1,.3849001795,.9622504490e-1},
			{0.,0.,0.,-.3333333333,0.,.3333333333,-.1666666667,0.,.1666666667},
			{.5892556509e-1,.2357022604,.5892556509e-1,.2357022604,.9428090414,.2357022604,-.5892556509e-1,-.2357022604,-.5892556509e-1},
			{.1521451549e-1,.6085806194e-1,.1521451549e-1,-.1825741858,-.7302967432,-.1825741858,-.1369306394,-.5477225575,-.1369306394}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[9],
		T o_output[10]
	) {
		static const T matrix[10][9] = 
		{
			{-.3928371006e-1,-.1571348402,-.3928371006e-1,-.1571348402,-.6285393608,-.1571348402,-.3928371006e-1,-.1571348402,-.3928371006e-1},
			{.6804138176e-1,0.,-.6804138176e-1,.2721655270,0.,-.2721655270,.6804138176e-1,0.,-.6804138176e-1},
			{-.3513641844e-1,.7027283688e-1,-.3513641844e-1,-.1405456738,.2810913476,-.1405456738,-.3513641844e-1,.7027283688e-1,-.3513641844e-1},
			{-.1111111111,-.4444444444,-.1111111111,-.1111111111,-.4444444444,-.1111111111,.5555555556e-1,.2222222222,.5555555556e-1},
			{.1924500898,0.,-.1924500898,.1924500898,0.,-.1924500898,-.9622504490e-1,0.,.9622504490e-1},
			{-.1360827635,-.5443310539,-.1360827635,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1924500898,.7698003590,.1924500898,.9622504490e-1,.3849001795,.9622504490e-1},
			{0.,0.,0.,-.3333333333,0.,.3333333333,-.1666666667,0.,.1666666667},
			{.5892556509e-1,.2357022604,.5892556509e-1,.2357022604,.9428090414,.2357022604,-.5892556509e-1,-.2357022604,-.5892556509e-1},
			{.1521451549e-1,.6085806194e-1,.1521451549e-1,-.1825741858,-.7302967432,-.1825741858,-.1369306394,-.5477225575,-.1369306394}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.5000000000,0.,0.,-.7071067810,0.,2.309401077,-.4082482906,0.,1.750000000,.3227486122},
			{0.,.5000000000,0.,0.,-.7071067810,0.,0.,-.4082482906,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.7500000000,0.,2.857738034,-.7216878367,0.,2.828427124,.9128709293},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,1.166666667,0.,0.,1.587713241,.9316949905},
			{0.,0.,0.,.4330127020,0.,-.7071067810,-.2500000000,0.,0.,.3162277660},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,-.8660254040,0.,0.,-.2500000000,.5809475019},
			{0.,0.,0.,0.,0.,.3726779962,0.,0.,-.3227486122,.8333333333e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.5000000000,0.,0.,-.7071067810,0.,2.309401077,-.4082482906,0.,1.750000000,.3227486122},
			{0.,.5000000000,0.,0.,-.7071067810,0.,0.,-.4082482906,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.7500000000,0.,2.857738034,-.7216878367,0.,2.828427124,.9128709293},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,1.166666667,0.,0.,1.587713241,.9316949905},
			{0.,0.,0.,.4330127020,0.,-.7071067810,-.2500000000,0.,0.,.3162277660},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,-.8660254040,0.,0.,-.2500000000,.5809475019},
			{0.,0.,0.,0.,0.,.3726779962,0.,0.,-.3227486122,.8333333333e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.8164965809,0.,-.2500000000,2.904737510},
			{0.,.5000000000,0.,0.,0.,0.,0.,-.8164965809,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,.3535533905,-2.464751509},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.1666666667,0.,0.,-.4330127020,1.155301788},
			{0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,.6123724358,3.320391543},
			{0.,0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,0.},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,-.2500000000,-1.484643616},
			{0.,0.,0.,0.,0.,.3726779962,0.,0.,.9682458365,1.083333333}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.8164965809,0.,-.2500000000,2.904737510},
			{0.,.5000000000,0.,0.,0.,0.,0.,-.8164965809,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,.3535533905,-2.464751509},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.1666666667,0.,0.,-.4330127020,1.155301788},
			{0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,.6123724358,3.320391543},
			{0.,0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,0.},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,-.2500000000,-1.484643616},
			{0.,0.,0.,0.,0.,.3726779962,0.,0.,.9682458365,1.083333333}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,0.,0.,-.7071067810,0.,.1443375673,.4082482906,0.,.1250000000,-.1613743061},
			{0.,1.,0.,0.,-.7071067810,0.,0.,.4082482906,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.2500000000,0.,.4082482906,.7216878367,0.,-.7071067810,.3651483717},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,.5776508940},
			{0.,0.,0.,-.4330127020,0.,.7071067810,-.7500000000,0.,0.,-.3162277660},
			{0.,0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,0.},
			{0.,0.,0.,0.,0.,.1443375673,0.,0.,-.1250000000,-.7423218081},
			{0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,.5416666667}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,0.,0.,-.7071067810,0.,.1443375673,.4082482906,0.,.1250000000,-.1613743061},
			{0.,1.,0.,0.,-.7071067810,0.,0.,.4082482906,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.2500000000,0.,.4082482906,.7216878367,0.,-.7071067810,.3651483717},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,.5776508940},
			{0.,0.,0.,-.4330127020,0.,.7071067810,-.7500000000,0.,0.,-.3162277660},
			{0.,0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,0.},
			{0.,0.,0.,0.,0.,.1443375673,0.,0.,-.1250000000,-.7423218081},
			{0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,.5416666667}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,0.,0.,.7071067810,0.,-.1443375673,-.4082482906,0.,-.1250000000,.1613743061},
			{0.,1.,0.,0.,.7071067810,0.,0.,-.4082482906,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.7500000000,0.,-.4082482906,-.7216878367,0.,-.3535533905,.4564354646},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.5833333333,0.,0.,.7938566203,.4658474951},
			{0.,0.,0.,.4330127020,0.,.7071067810,-.2500000000,0.,-.6123724358,.1581138830},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.1250000000,.2904737510},
			{0.,0.,0.,0.,0.,.1863389981,0.,0.,-.1613743061,.4166666667e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,0.,0.,.7071067810,0.,-.1443375673,-.4082482906,0.,-.1250000000,.1613743061},
			{0.,1.,0.,0.,.7071067810,0.,0.,-.4082482906,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.7500000000,0.,-.4082482906,-.7216878367,0.,-.3535533905,.4564354646},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.5833333333,0.,0.,.7938566203,.4658474951},
			{0.,0.,0.,.4330127020,0.,.7071067810,-.2500000000,0.,-.6123724358,.1581138830},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.1250000000,.2904737510},
			{0.,0.,0.,0.,0.,.1863389981,0.,0.,-.1613743061,.4166666667e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.7071067810,.2357022604,.1178511302,.2357022604,.5892556509e-1,.1178511302,.3535533905,.1178511302,.1178511302,.2357022604},
			{0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.6804138176e-1,.2041241452},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,.1666666667,.1666666667e-1,.1333333333,0.,-.4166666667e-1,.8333333333e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,.4811252245e-1,0.},
			{0.,0.,.1360827635e-1,0.,-.2041241452e-1,.4082482906e-1,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,0.,.2886751347e-1,0.,0.,.7216878367e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,-.2357022604e-1,0.,.2357022604e-1,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.7071067810,.2357022604,.1178511302,.2357022604,.5892556509e-1,.1178511302,.3535533905,.1178511302,.1178511302,.2357022604},
			{0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.6804138176e-1,.2041241452},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,.1666666667,.1666666667e-1,.1333333333,0.,-.4166666667e-1,.8333333333e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,.4811252245e-1,0.},
			{0.,0.,.1360827635e-1,0.,-.2041241452e-1,.4082482906e-1,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,0.,.2886751347e-1,0.,0.,.7216878367e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,-.2357022604e-1,0.,.2357022604e-1,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{0.,0.,0.,0.,0.,0.,6.928203232,-12.,-8.485281372,-32.86335345},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345},
			{0.,0.,0.,6.,-10.39230485,-19.59591794,3.464101616,-6.,-25.45584412,-10.95445115},
			{0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,32.86335345},
			{0.,0.,0.,0.,0.,24.49489743,0.,0.,21.21320343,5.477225575},
			{0.,4.898979486,-18.97366596,0.,-6.928203232,0.,0.,-12.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,24.,0.,0.},
			{0.,0.,0.,0.,20.78460970,0.,0.,12.,0.,0.},
			{0.,0.,18.97366596,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{0.,0.,0.,0.,0.,0.,6.928203232,-12.,-8.485281372,-32.86335345},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345},
			{0.,0.,0.,6.,-10.39230485,-19.59591794,3.464101616,-6.,-25.45584412,-10.95445115},
			{0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,32.86335345},
			{0.,0.,0.,0.,0.,24.49489743,0.,0.,21.21320343,5.477225575},
			{0.,4.898979486,-18.97366596,0.,-6.928203232,0.,0.,-12.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,24.,0.,0.},
			{0.,0.,0.,0.,20.78460970,0.,0.,12.,0.,0.},
			{0.,0.,18.97366596,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[10],
		T o_output[1]
	) {
		static const T matrix[1][10] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[10],
		T o_output[1]
	) {
		static const T matrix[1][10] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_w(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_w(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_bottom_face(
		const T i_input[10],
		T o_output[6]
	) {
		static const T matrix[6][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 6; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_bottom_face(
		const T i_input[10],
		T o_output[6]
	) {
		static const T matrix[6][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 6; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_top_face(
		const T i_input[10],
		T o_output[6]
	) {
		static const T matrix[6][10] = 
		{
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,-1.732050808,-3.,-3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,1.732050808,3.,3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 6; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_top_face(
		const T i_input[10],
		T o_output[6]
	) {
		static const T matrix[6][10] = 
		{
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,-1.732050808,-3.,-3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,1.732050808,3.,3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 6; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_bottom_fluxes_to_dofs(
		const T i_input[6],
		T o_output[10]
	) {
		static const T matrix[10][6] = 
		{
			{0.,-.2357022604,0.,-.2357022604,-.2357022604,0.},
			{0.,.4082482906,0.,.4082482906,.4082482906,0.},
			{0.,-.5270462768,0.,-.5270462768,-.5270462768,0.},
			{.5000000000e-1,.1333333333,.5000000000e-1,-.6666666667e-1,-.6666666667e-1,-.1000000000},
			{-.8660254040e-1,-.2309401077,-.8660254040e-1,.1154700539,.1154700539,.1732050808},
			{-.2721655270e-1,-.2721655270e-1,-.2721655270e-1,.8164965809e-1,.8164965809e-1,-.8164965809e-1},
			{.8660254040e-1,0.,-.8660254040e-1,.1154700539,-.1154700539,0.},
			{-.1500000000,0.,.1500000000,-.2000000000,.2000000000,0.},
			{-.4714045206e-1,0.,.4714045206e-1,.9428090414e-1,-.9428090414e-1,0.},
			{-.6085806194e-1,.1217161239,-.6085806194e-1,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 6; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_bottom_fluxes_to_dofs(
		const T i_input[6],
		T o_output[10]
	) {
		static const T matrix[10][6] = 
		{
			{0.,-.2357022604,0.,-.2357022604,-.2357022604,0.},
			{0.,.4082482906,0.,.4082482906,.4082482906,0.},
			{0.,-.5270462768,0.,-.5270462768,-.5270462768,0.},
			{.5000000000e-1,.1333333333,.5000000000e-1,-.6666666667e-1,-.6666666667e-1,-.1000000000},
			{-.8660254040e-1,-.2309401077,-.8660254040e-1,.1154700539,.1154700539,.1732050808},
			{-.2721655270e-1,-.2721655270e-1,-.2721655270e-1,.8164965809e-1,.8164965809e-1,-.8164965809e-1},
			{.8660254040e-1,0.,-.8660254040e-1,.1154700539,-.1154700539,0.},
			{-.1500000000,0.,.1500000000,-.2000000000,.2000000000,0.},
			{-.4714045206e-1,0.,.4714045206e-1,.9428090414e-1,-.9428090414e-1,0.},
			{-.6085806194e-1,.1217161239,-.6085806194e-1,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 6; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_top_fluxes_to_dofs(
		const T i_input[6],
		T o_output[10]
	) {
		static const T matrix[10][6] = 
		{
			{0.,-.2357022604,0.,-.2357022604,-.2357022604,0.},
			{0.,-.4082482906,0.,-.4082482906,-.4082482906,0.},
			{0.,-.5270462768,0.,-.5270462768,-.5270462768,0.},
			{.5000000000e-1,.1333333333,.5000000000e-1,-.6666666667e-1,-.6666666667e-1,-.1000000000},
			{.8660254040e-1,.2309401077,.8660254040e-1,-.1154700539,-.1154700539,-.1732050808},
			{-.2721655270e-1,-.2721655270e-1,-.2721655270e-1,.8164965809e-1,.8164965809e-1,-.8164965809e-1},
			{.8660254040e-1,0.,-.8660254040e-1,.1154700539,-.1154700539,0.},
			{.1500000000,0.,-.1500000000,.2000000000,-.2000000000,0.},
			{-.4714045206e-1,0.,.4714045206e-1,.9428090414e-1,-.9428090414e-1,0.},
			{-.6085806194e-1,.1217161239,-.6085806194e-1,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 6; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_top_fluxes_to_dofs(
		const T i_input[6],
		T o_output[10]
	) {
		static const T matrix[10][6] = 
		{
			{0.,-.2357022604,0.,-.2357022604,-.2357022604,0.},
			{0.,-.4082482906,0.,-.4082482906,-.4082482906,0.},
			{0.,-.5270462768,0.,-.5270462768,-.5270462768,0.},
			{.5000000000e-1,.1333333333,.5000000000e-1,-.6666666667e-1,-.6666666667e-1,-.1000000000},
			{.8660254040e-1,.2309401077,.8660254040e-1,-.1154700539,-.1154700539,-.1732050808},
			{-.2721655270e-1,-.2721655270e-1,-.2721655270e-1,.8164965809e-1,.8164965809e-1,-.8164965809e-1},
			{.8660254040e-1,0.,-.8660254040e-1,.1154700539,-.1154700539,0.},
			{.1500000000,0.,-.1500000000,.2000000000,-.2000000000,0.},
			{-.4714045206e-1,0.,.4714045206e-1,.9428090414e-1,-.9428090414e-1,0.},
			{-.6085806194e-1,.1217161239,-.6085806194e-1,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 6; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 6400;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[10][10] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,3.535533905,0.,0.,6.123724358,0.,0.,0.},
			{2.581988897,0.,0.,-.9128709293,0.,0.,4.743416490,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_dofs_to_nodes:" << std::endl;
		static const T timestep_stiffness_dofs_to_nodes[10][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,-1.732050808,0.,-3.181980514,1.369306394}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_dofs_to_nodes[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_dofs_to_nodes[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,1.732050808,0.,3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,1.732050808,3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,4.,0.,7.348469229,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.}
		};
		for (int j = 0; j < 9; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,3.464101616,0.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575}
		};
		for (int j = 0; j < 9; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[9][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,0.,-1.581138830,4.,0.,7.348469229,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,0.,-1.581138830,1.,0.,-1.224744872,-1.732050808,0.,-3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,-1.732050808,-3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,0.,-1.581138830,-2.,0.,2.449489743,-3.464101616,0.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575}
		};
		for (int j = 0; j < 9; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[10][9] = 
		{
			{-.5555555556e-1,-.2222222222,-.5555555556e-1,-.2222222222,-.8888888889,-.2222222222,-.5555555556e-1,-.2222222222,-.5555555556e-1},
			{.9622504490e-1,0.,-.9622504490e-1,.3849001795,0.,-.3849001795,.9622504490e-1,0.,-.9622504490e-1},
			{-.4969039948e-1,.9938079897e-1,-.4969039948e-1,-.1987615980,.3975231960,-.1987615980,-.4969039948e-1,.9938079897e-1,-.4969039948e-1},
			{.7856742012e-1,.3142696804,.7856742012e-1,-.1571348402,-.6285393608,-.1571348402,-.1571348402,-.6285393608,-.1571348402},
			{-.1360827635,0.,.1360827635,.2721655270,0.,-.2721655270,.2721655270,0.,-.2721655270},
			{0.,0.,0.,0.,0.,0.,-.1924500898,-.7698003590,-.1924500898},
			{-.1360827635,-.5443310539,-.1360827635,-.2721655270,-1.088662108,-.2721655270,0.,0.,0.},
			{.2357022604,0.,-.2357022604,.4714045206,0.,-.4714045206,0.,0.,0.},
			{.8333333333e-1,.3333333333,.8333333333e-1,-.3333333333,-1.333333333,-.3333333333,-.8333333333e-1,-.3333333333,-.8333333333e-1},
			{-.1936491673,-.7745966692,-.1936491673,-.2581988897,-1.032795559,-.2581988897,.2151657415e-1,.8606629657e-1,.2151657415e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 9; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[10][9] = 
		{
			{-.3928371006e-1,-.1571348402,-.3928371006e-1,-.1571348402,-.6285393608,-.1571348402,-.3928371006e-1,-.1571348402,-.3928371006e-1},
			{.6804138176e-1,0.,-.6804138176e-1,.2721655270,0.,-.2721655270,.6804138176e-1,0.,-.6804138176e-1},
			{-.3513641844e-1,.7027283688e-1,-.3513641844e-1,-.1405456738,.2810913476,-.1405456738,-.3513641844e-1,.7027283688e-1,-.3513641844e-1},
			{.5555555556e-1,.2222222222,.5555555556e-1,.2222222222,.8888888889,.2222222222,.5555555556e-1,.2222222222,.5555555556e-1},
			{-.9622504490e-1,0.,.9622504490e-1,-.3849001795,0.,.3849001795,-.9622504490e-1,0.,.9622504490e-1},
			{-.6804138176e-1,-.2721655270,-.6804138176e-1,-.2721655270,-1.088662108,-.2721655270,-.6804138176e-1,-.2721655270,-.6804138176e-1},
			{.9622504490e-1,.3849001795,.9622504490e-1,0.,0.,0.,-.9622504490e-1,-.3849001795,-.9622504490e-1},
			{-.1666666667,0.,.1666666667,0.,0.,0.,.1666666667,0.,-.1666666667},
			{-.1178511302,-.4714045206,-.1178511302,0.,0.,0.,.1178511302,.4714045206,.1178511302},
			{-.6085806194e-1,-.2434322478,-.6085806194e-1,.1217161239,.4868644956,.1217161239,-.6085806194e-1,-.2434322478,-.6085806194e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 9; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[10][9] = 
		{
			{-.3928371006e-1,-.1571348402,-.3928371006e-1,-.1571348402,-.6285393608,-.1571348402,-.3928371006e-1,-.1571348402,-.3928371006e-1},
			{.6804138176e-1,0.,-.6804138176e-1,.2721655270,0.,-.2721655270,.6804138176e-1,0.,-.6804138176e-1},
			{-.3513641844e-1,.7027283688e-1,-.3513641844e-1,-.1405456738,.2810913476,-.1405456738,-.3513641844e-1,.7027283688e-1,-.3513641844e-1},
			{-.1111111111,-.4444444444,-.1111111111,-.1111111111,-.4444444444,-.1111111111,.5555555556e-1,.2222222222,.5555555556e-1},
			{.1924500898,0.,-.1924500898,.1924500898,0.,-.1924500898,-.9622504490e-1,0.,.9622504490e-1},
			{-.1360827635,-.5443310539,-.1360827635,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1924500898,.7698003590,.1924500898,.9622504490e-1,.3849001795,.9622504490e-1},
			{0.,0.,0.,-.3333333333,0.,.3333333333,-.1666666667,0.,.1666666667},
			{.5892556509e-1,.2357022604,.5892556509e-1,.2357022604,.9428090414,.2357022604,-.5892556509e-1,-.2357022604,-.5892556509e-1},
			{.1521451549e-1,.6085806194e-1,.1521451549e-1,-.1825741858,-.7302967432,-.1825741858,-.1369306394,-.5477225575,-.1369306394}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 9; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[10][10] = 
		{
			{.5000000000,0.,0.,-.7071067810,0.,2.309401077,-.4082482906,0.,1.750000000,.3227486122},
			{0.,.5000000000,0.,0.,-.7071067810,0.,0.,-.4082482906,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.7500000000,0.,2.857738034,-.7216878367,0.,2.828427124,.9128709293},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,1.166666667,0.,0.,1.587713241,.9316949905},
			{0.,0.,0.,.4330127020,0.,-.7071067810,-.2500000000,0.,0.,.3162277660},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,-.8660254040,0.,0.,-.2500000000,.5809475019},
			{0.,0.,0.,0.,0.,.3726779962,0.,0.,-.3227486122,.8333333333e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[10][10] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.8164965809,0.,-.2500000000,2.904737510},
			{0.,.5000000000,0.,0.,0.,0.,0.,-.8164965809,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,.3535533905,-2.464751509},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.1666666667,0.,0.,-.4330127020,1.155301788},
			{0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,.6123724358,3.320391543},
			{0.,0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,0.},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,-.2500000000,-1.484643616},
			{0.,0.,0.,0.,0.,.3726779962,0.,0.,.9682458365,1.083333333}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[10][10] = 
		{
			{1.,0.,0.,-.7071067810,0.,.1443375673,.4082482906,0.,.1250000000,-.1613743061},
			{0.,1.,0.,0.,-.7071067810,0.,0.,.4082482906,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.2500000000,0.,.4082482906,.7216878367,0.,-.7071067810,.3651483717},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,.5776508940},
			{0.,0.,0.,-.4330127020,0.,.7071067810,-.7500000000,0.,0.,-.3162277660},
			{0.,0.,0.,0.,-.4330127020,0.,0.,-.7500000000,0.,0.},
			{0.,0.,0.,0.,0.,.1443375673,0.,0.,-.1250000000,-.7423218081},
			{0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,.5416666667}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[10][10] = 
		{
			{1.,0.,0.,.7071067810,0.,-.1443375673,-.4082482906,0.,-.1250000000,.1613743061},
			{0.,1.,0.,0.,.7071067810,0.,0.,-.4082482906,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.7500000000,0.,-.4082482906,-.7216878367,0.,-.3535533905,.4564354646},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.7216878367,0.,0.},
			{0.,0.,0.,0.,0.,.5833333333,0.,0.,.7938566203,.4658474951},
			{0.,0.,0.,.4330127020,0.,.7071067810,-.2500000000,0.,-.6123724358,.1581138830},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.1250000000,.2904737510},
			{0.,0.,0.,0.,0.,.1863389981,0.,0.,-.1613743061,.4166666667e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[10][10] = 
		{
			{.7071067810,.2357022604,.1178511302,.2357022604,.5892556509e-1,.1178511302,.3535533905,.1178511302,.1178511302,.2357022604},
			{0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.6804138176e-1,.2041241452},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,.1666666667,.1666666667e-1,.1333333333,0.,-.4166666667e-1,.8333333333e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,.4811252245e-1,0.},
			{0.,0.,.1360827635e-1,0.,-.2041241452e-1,.4082482906e-1,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,0.,.2886751347e-1,0.,0.,.7216878367e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,-.2357022604e-1,0.,.2357022604e-1,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[10][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{0.,0.,0.,0.,0.,0.,6.928203232,-12.,-8.485281372,-32.86335345},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345},
			{0.,0.,0.,6.,-10.39230485,-19.59591794,3.464101616,-6.,-25.45584412,-10.95445115},
			{0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,32.86335345},
			{0.,0.,0.,0.,0.,24.49489743,0.,0.,21.21320343,5.477225575},
			{0.,4.898979486,-18.97366596,0.,-6.928203232,0.,0.,-12.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,24.,0.,0.},
			{0.,0.,0.,0.,20.78460970,0.,0.,12.,0.,0.},
			{0.,0.,18.97366596,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][10] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_w:" << std::endl;
		static const T timestep_stiffness_w[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_w[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_w[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_bottom_face:" << std::endl;
		static const T timestep_dofs_to_bottom_face[6][10] = 
		{
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,-3.464101616,6.,4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,-2.449489743,3.162277660,-2.,3.464101616,2.449489743,3.464101616,-6.,-4.242640686,5.477225575},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,-1.732050808,3.,-3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,1.,-1.732050808,-1.224744872,1.732050808,-3.,3.181980514,1.369306394},
			{1.414213562,-2.449489743,3.162277660,4.,-6.928203232,7.348469229,0.,0.,0.,0.}
		};
		for (int j = 0; j < 6; j++)
		{
			std::cout << timestep_dofs_to_bottom_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_bottom_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_top_face:" << std::endl;
		static const T timestep_dofs_to_top_face[6][10] = 
		{
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,-3.464101616,-6.,4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,0.,0.,0.,-2.738612788},
			{1.414213562,2.449489743,3.162277660,-2.,-3.464101616,2.449489743,3.464101616,6.,-4.242640686,5.477225575},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,-1.732050808,-3.,-3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,1.,1.732050808,-1.224744872,1.732050808,3.,3.181980514,1.369306394},
			{1.414213562,2.449489743,3.162277660,4.,6.928203232,7.348469229,0.,0.,0.,0.}
		};
		for (int j = 0; j < 6; j++)
		{
			std::cout << timestep_dofs_to_top_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_top_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_bottom_fluxes_to_dofs:" << std::endl;
		static const T timestep_bottom_fluxes_to_dofs[10][6] = 
		{
			{0.,-.2357022604,0.,-.2357022604,-.2357022604,0.},
			{0.,.4082482906,0.,.4082482906,.4082482906,0.},
			{0.,-.5270462768,0.,-.5270462768,-.5270462768,0.},
			{.5000000000e-1,.1333333333,.5000000000e-1,-.6666666667e-1,-.6666666667e-1,-.1000000000},
			{-.8660254040e-1,-.2309401077,-.8660254040e-1,.1154700539,.1154700539,.1732050808},
			{-.2721655270e-1,-.2721655270e-1,-.2721655270e-1,.8164965809e-1,.8164965809e-1,-.8164965809e-1},
			{.8660254040e-1,0.,-.8660254040e-1,.1154700539,-.1154700539,0.},
			{-.1500000000,0.,.1500000000,-.2000000000,.2000000000,0.},
			{-.4714045206e-1,0.,.4714045206e-1,.9428090414e-1,-.9428090414e-1,0.},
			{-.6085806194e-1,.1217161239,-.6085806194e-1,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_bottom_fluxes_to_dofs[j][0];
			for (int i = 1; i < 6; i++)
			{
				std::cout << ", " << timestep_bottom_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_top_fluxes_to_dofs:" << std::endl;
		static const T timestep_top_fluxes_to_dofs[10][6] = 
		{
			{0.,-.2357022604,0.,-.2357022604,-.2357022604,0.},
			{0.,-.4082482906,0.,-.4082482906,-.4082482906,0.},
			{0.,-.5270462768,0.,-.5270462768,-.5270462768,0.},
			{.5000000000e-1,.1333333333,.5000000000e-1,-.6666666667e-1,-.6666666667e-1,-.1000000000},
			{.8660254040e-1,.2309401077,.8660254040e-1,-.1154700539,-.1154700539,-.1732050808},
			{-.2721655270e-1,-.2721655270e-1,-.2721655270e-1,.8164965809e-1,.8164965809e-1,-.8164965809e-1},
			{.8660254040e-1,0.,-.8660254040e-1,.1154700539,-.1154700539,0.},
			{.1500000000,0.,-.1500000000,.2000000000,-.2000000000,0.},
			{-.4714045206e-1,0.,.4714045206e-1,.9428090414e-1,-.9428090414e-1,0.},
			{-.6085806194e-1,.1217161239,-.6085806194e-1,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_top_fluxes_to_dofs[j][0];
			for (int i = 1; i < 6; i++)
			{
				std::cout << ", " << timestep_top_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
