
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION3D_MATRICES_CONST_HPP_
#define CCOMPUTATION3D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (5)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (56)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (56)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS (144)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (52)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*5+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_3D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[56],
		T io_hv[56]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1959.99950,-5383.2139,9015.826058,-8519.319340,4200.00000,-835.7894471,-6584.2284,10097.4849,-9661.82590,4925.362810,-1028.78570,13230.11101,-11457.6123,5819.69066,-1229.634099,-14816.2076,6782.212097,-1419.859143,8640.1667,-1626.65301,-2047.25769,-6584.2284,10097.4849,-9661.82590,4925.362810,-1028.78570,10802.34025,-9355.10139,4751.757564,-1003.99205,-11476.58483,5253.47885,-1099.81816,6531.35214,-1229.63410,-1525.93584,13230.11101,-11457.6123,5819.69066,-1229.634099,-11476.58483,5253.47885,-1099.81816,6196.18489,-1166.53333,-1412.74207,-14816.2076,6782.212097,-1419.859143,6531.35214,-1229.63410,-1412.74207,8640.1667,-1626.65301,-1525.93584,-2047.25769},
			{-5383.2139,33096.0000,-83609.96448,103016.3016,-61834.21385,14476.29787,10097.4848,-40465.5705,63751.51762,-44899.88862,11879.39392,-11457.61230,30581.07650,-27719.99999,8519.154887,6782.21205,-12064.62597,4918.536370,-1626.653003,1878.297097,0.,10097.4848,-40465.5705,63751.51762,-44899.88862,11879.39392,-9355.10127,24969.34437,-22633.28523,6955.860840,5253.47884,-9345.21910,3809.881889,-1229.634091,1419.85915,0.,-11457.61230,30581.07650,-27719.99999,8519.154887,5253.47884,-9345.21910,3809.881889,-1166.53333,1346.99667,0.,6782.21205,-12064.62597,4918.536370,-1229.634091,1419.85915,0.,-1626.653003,1878.297097,0.,0.},
			{9015.826058,-83609.96448,277200.,-416728.6598,293014.3477,-78493.04477,-9661.825918,63751.51762,-141972.4255,128683.4255,-41407.82533,5819.690715,-27720.,39039.67213,-16497.27251,-1419.859148,4918.536370,-3809.881888,0.,0.,0.,-9661.825918,63751.51762,-141972.4255,128683.4255,-41407.82533,4751.757568,-22633.28523,31875.75880,-13469.96659,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,5819.690715,-27720.,39039.67213,-16497.27251,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,-1419.859148,4918.536370,-3809.881888,0.,0.,0.,0.,0.,0.,0.},
			{-8519.319340,103016.3016,-416728.6598,734000.,-588944.2418,176903.2820,4925.362927,-44899.88862,128683.4255,-143050.2010,54438.22186,-1229.634092,8519.154886,-16497.27251,9295.160030,0.,0.,0.,0.,0.,0.,4925.362927,-44899.88862,128683.4255,-143050.2010,54438.22186,-1003.992031,6955.860840,-13469.96659,7589.466382,0.,0.,0.,0.,0.,0.,-1229.634092,8519.154886,-16497.27251,9295.160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4200.00000,-61834.21385,293014.3477,-588944.2418,529200.,-175515.7839,-1028.785692,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1028.785692,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-835.7894471,14476.29787,-78493.04477,176903.2820,-175515.7839,63504.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-6584.2284,10097.4848,-9661.825918,4925.362927,-1028.785692,0.,39536.0001,-37827.9896,22915.22463,-7196.44356,840.,-106783.7919,62111.73800,-21382.90905,3011.976097,144131.6731,-46695.88421,6955.860840,-95851.44763,13281.56616,25073.68342,19768.0002,-18913.9949,11457.61232,-3598.221779,420.,-58125.73395,33809.34781,-11639.38143,1639.512123,83732.93531,-27127.85727,4040.989977,-57965.50700,8031.93625,15574.01680,-35594.59728,20703.91267,-7127.636351,1003.992032,55821.95698,-18085.23818,2693.993319,-41243.18141,5714.82284,11534.99025,36032.9184,-11673.97107,1738.965210,-28982.75339,4015.96813,8651.24267,-19170.28946,2656.313229,6229.60671,4178.947267},
			{10097.4849,-40465.5705,63751.51762,-44899.88862,11879.39392,0.,-37827.9896,112000.0001,-123625.6285,58656.96888,-9699.484525,62111.73799,-135471.9749,92590.71229,-20867.58252,-46695.88421,74658.13284,-24095.80875,13281.56616,-15336.23160,0.,-18913.9949,56000.0000,-61812.81421,29328.48444,-4849.742262,33809.34782,-73741.60290,50399.99999,-11358.87319,-27127.85727,43372.45579,-13998.39991,8031.93625,-9274.48112,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,-18085.23818,28914.97050,-9332.266609,5714.82284,-6598.90901,0.,-11673.97107,18664.53323,-6023.952191,4015.96813,-4637.24056,0.,2656.313229,-3067.24632,0.,0.},
			{-9661.82590,63751.51762,-141972.4255,128683.4255,-41407.82533,0.,22915.22463,-123625.6285,216720.,-147665.3514,33809.34781,-21382.90906,92590.71229,-114752.7317,40409.89977,6955.860840,-24095.80875,18664.53321,0.,0.,0.,11457.61232,-61812.81421,108360.,-73832.67568,16904.67391,-11639.38143,50400.,-62463.47540,21996.36334,4040.989977,-13998.39991,10843.11394,0.,0.,0.,-7127.636352,30863.57076,-38250.91056,13469.96659,2693.993318,-9332.266605,7228.742626,0.,0.,0.,1738.965210,-6023.952189,4666.133303,0.,0.,0.,0.,0.,0.,0.},
			{4925.362810,-44899.88862,128683.4255,-143050.2010,54438.22186,0.,-7196.44356,58656.96888,-147665.3514,140800.,-44448.62202,3011.976094,-20867.58252,40409.89977,-22768.39914,0.,0.,0.,0.,0.,0.,-3598.221779,29328.48444,-73832.67568,70400.,-22224.31101,1639.512123,-11358.87318,21996.36334,-12393.54671,0.,0.,0.,0.,0.,0.,1003.992031,-6955.860840,13469.96659,-7589.466382,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1028.78570,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,840.,-9699.484525,33809.34781,-44448.62202,19600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,420.,-4849.742262,16904.67391,-22224.31101,9800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{13230.11101,-11457.61230,5819.690715,-1229.634092,0.,0.,-106783.7919,62111.73799,-21382.90906,3011.976094,0.,347760.0000,-130070.0875,27047.47825,-1904.940944,-535300.7081,116428.4501,-10998.18167,392829.1232,-37800.,-111005.9285,-35594.59728,20703.91267,-7127.636352,1003.992031,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-248785.2872,54111.05382,-5111.492932,197967.2700,-19049.40944,-59099.23855,57960.00000,-21678.34791,4507.913042,-317.4901573,-124392.6436,27055.52691,-2555.746466,112684.9455,-10843.11394,-36476.84197,-53530.07079,11642.84501,-1099.818167,59390.18104,-5714.822832,-21886.10519,26188.60820,-2520.,-11819.84771,-5285.996592},
			{-11457.6123,30581.07650,-27720.,8519.154886,0.,0.,62111.73800,-135471.9749,92590.71229,-20867.58252,0.,-130070.0875,230832.,-105407.1147,13197.81800,116428.4501,-167230.2366,38098.81888,-37800.,43647.68036,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,54111.05382,-77721.59051,17706.73093,-19049.40944,21996.36334,0.,-21678.34791,38472.,-17567.85246,2199.636334,27055.52691,-38860.79526,8853.365465,-10843.11394,12520.54951,0.,11642.84501,-16723.02366,3809.881888,-5714.822832,6598.909002,0.,-2520.,2909.845357,0.,0.},
			{5819.69066,-27719.99999,39039.67213,-16497.27251,0.,0.,-21382.90905,92590.71229,-114752.7317,40409.89977,0.,27047.47825,-105407.1147,108864.,-25557.46466,-10998.18167,38098.81888,-29511.21822,0.,0.,0.,-7127.636351,30863.57076,-38250.91056,13469.96659,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-5111.492932,17706.73093,-13715.57480,0.,0.,0.,4507.913042,-17567.85246,18144.,-4259.577443,-2555.746466,8853.365465,-6857.787398,0.,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,0.,0.,0.,0.},
			{-1229.634099,8519.154887,-16497.27251,9295.160030,0.,0.,3011.976097,-20867.58252,40409.89977,-22768.39914,0.,-1904.940944,13197.81800,-25557.46466,14400.,0.,0.,0.,0.,0.,0.,1003.992032,-6955.860840,13469.96659,-7589.466382,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,-317.4901573,2199.636334,-4259.577443,2400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-14816.2076,6782.21205,-1419.859148,0.,0.,0.,144131.6731,-46695.88421,6955.860840,0.,0.,-535300.7081,116428.4501,-10998.18167,0.,909200.,-119165.0956,5366.563145,-720578.5174,43037.19323,216661.3874,36032.9184,-11673.97107,1738.965210,0.,0.,-174828.4790,38025.37259,-3591.991091,0.,352131.6458,-46152.44305,2078.460970,-311260.6624,18590.32006,100931.1647,-53530.07079,11642.84501,-1099.818167,0.,140852.6583,-18460.97722,831.3843878,-147643.8959,8818.163075,53396.62909,45460.,-5958.254780,268.3281572,-62252.13248,3718.064012,26698.31455,-20587.95764,1229.634092,11534.99025,3868.953347},
			{6782.212097,-12064.62597,4918.536370,0.,0.,0.,-46695.88421,74658.13284,-24095.80875,0.,0.,116428.4501,-167230.2366,38098.81888,0.,-119165.0956,153600.,-18590.32006,43037.19323,-49695.07017,0.,-11673.97107,18664.53323,-6023.952189,0.,0.,38025.37259,-54617.16651,12443.02214,0.,-46152.44305,59489.02419,-7200.,18590.32006,-21466.25258,0.,11642.84501,-16723.02366,3809.881888,0.,-18460.97722,23795.60968,-2880.,8818.163075,-10182.33765,0.,-5958.254780,7680.,-929.5160030,3718.064012,-4293.250516,0.,1229.634092,-1419.859148,0.,0.},
			{-1419.859143,4918.536370,-3809.881888,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,5366.563145,-18590.32006,14400.,0.,0.,0.,1738.965210,-6023.952191,4666.133303,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,268.3281572,-929.5160030,720.,0.,0.,0.,0.,0.,0.,0.},
			{8640.1667,-1626.653003,0.,0.,0.,0.,-95851.44763,13281.56616,0.,0.,0.,392829.1232,-37800.,0.,0.,-720578.5174,43037.19323,0.,607600.,-16974.09792,-192267.9080,-19170.28946,2656.313229,0.,0.,0.,106914.5452,-10287.85692,0.,0.,-239210.4513,14287.05708,0.,229651.2138,-6415.605974,-79615.57635,26188.60820,-2520.,0.,0.,-79736.81711,4762.352360,0.,93371.25892,-2608.447815,-36854.85042,-20587.95764,1229.634092,0.,32807.31626,-916.5151392,-15794.93589,8680.,-242.4871131,-5686.826882,-1525.935778},
			{-1626.65301,1878.297097,0.,0.,0.,0.,13281.56616,-15336.23160,0.,0.,0.,-37800.,43647.68036,0.,0.,43037.19323,-49695.07017,0.,-16974.09792,19600.,0.,2656.313229,-3067.24632,0.,0.,0.,-10287.85692,11879.39392,0.,0.,14287.05708,-16497.27251,0.,-6415.605974,7408.103671,0.,-2520.,2909.845357,0.,0.,4762.352360,-5499.090835,0.,-2608.447815,3011.976094,0.,1229.634092,-1419.859148,0.,-916.5151392,1058.300524,0.,-242.4871131,280.,0.,0.},
			{-2047.25769,0.,0.,0.,0.,0.,25073.68342,0.,0.,0.,0.,-111005.9285,0.,0.,0.,216661.3874,0.,0.,-192267.9080,0.,63504.,4178.947267,0.,0.,0.,0.,-25895.98888,0.,0.,0.,62934.44585,0.,0.,-64595.94539,0.,23666.54347,-5285.996592,0.,0.,0.,17981.27024,0.,0.,-22980.41043,0.,9738.205176,3868.953347,0.,0.,-6920.994149,0.,3651.826941,-1525.935778,0.,1126.978260,252.},
			{-6584.2284,10097.4848,-9661.825918,4925.362927,-1028.785692,0.,19768.0002,-18913.9949,11457.61232,-3598.221779,420.,-35594.59728,20703.91267,-7127.636351,1003.992032,36032.9184,-11673.97107,1738.965210,-19170.28946,2656.313232,4178.947267,39536.0001,-37827.9896,22915.22463,-7196.44356,840.,-58125.73395,33809.34781,-11639.38143,1639.512123,55821.95698,-18085.23818,2693.993319,-28982.75339,4015.968128,6229.60671,-106783.7919,62111.73800,-21382.90905,3011.976097,83732.93531,-27127.85727,4040.989977,-41243.18141,5714.82284,8651.24267,144131.6731,-46695.88421,6955.860840,-57965.50700,8031.936250,11534.99025,-95851.44763,13281.56616,15574.01680,25073.68342},
			{10097.4849,-40465.5705,63751.51762,-44899.88862,11879.39392,0.,-18913.9949,56000.0000,-61812.81421,29328.48444,-4849.742262,20703.91267,-45157.32496,30863.57076,-6955.860840,-11673.97107,18664.53323,-6023.952191,2656.313229,-3067.24632,0.,-37827.9896,112000.0001,-123625.6285,58656.96888,-9699.484525,33809.34782,-73741.60290,50399.99999,-11358.87319,-18085.23818,28914.97050,-9332.266609,4015.96813,-4637.24056,0.,62111.73799,-135471.9749,92590.71229,-20867.58252,-27127.85727,43372.45579,-13998.39991,5714.82284,-6598.90901,0.,-46695.88421,74658.13284,-24095.80875,8031.93625,-9274.48112,0.,13281.56616,-15336.23160,0.,0.},
			{-9661.82590,63751.51762,-141972.4255,128683.4255,-41407.82533,0.,11457.61232,-61812.81421,108360.,-73832.67568,16904.67391,-7127.636352,30863.57076,-38250.91056,13469.96659,1738.965210,-6023.952189,4666.133303,0.,0.,0.,22915.22463,-123625.6285,216720.,-147665.3514,33809.34781,-11639.38143,50400.,-62463.47540,21996.36334,2693.993318,-9332.266605,7228.742626,0.,0.,0.,-21382.90906,92590.71229,-114752.7317,40409.89977,4040.989977,-13998.39991,10843.11394,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,0.,0.,0.,0.,0.},
			{4925.362810,-44899.88862,128683.4255,-143050.2010,54438.22186,0.,-3598.221779,29328.48444,-73832.67568,70400.,-22224.31101,1003.992031,-6955.860840,13469.96659,-7589.466382,0.,0.,0.,0.,0.,0.,-7196.44356,58656.96888,-147665.3514,140800.,-44448.62202,1639.512123,-11358.87318,21996.36334,-12393.54671,0.,0.,0.,0.,0.,0.,3011.976094,-20867.58252,40409.89977,-22768.39914,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1028.78570,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,420.,-4849.742262,16904.67391,-22224.31101,9800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,840.,-9699.484525,33809.34781,-44448.62202,19600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{10802.34025,-9355.10127,4751.757568,-1003.992031,0.,0.,-58125.73395,33809.34782,-11639.38143,1639.512123,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-174828.4790,38025.37259,-3591.991091,106914.5452,-10287.85692,-25895.98888,-58125.73395,33809.34782,-11639.38143,1639.512123,0.,154560.0000,-57808.92777,12021.10144,-846.6404195,-203132.3364,44181.49041,-4173.516504,129311.6793,-12443.02214,-32169.55082,141972.4255,-53100.89083,11042.08676,-777.6888838,-203132.3364,44181.49041,-4173.516504,138010.3092,-13280.04820,-35739.86008,-174828.4790,38025.37259,-3591.991091,129311.6793,-12443.02214,-35739.86008,106914.5452,-10287.85692,-32169.55082,-25895.98888},
			{-9355.10139,24969.34437,-22633.28523,6955.860840,0.,0.,33809.34781,-73741.60290,50400.,-11358.87318,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,38025.37259,-54617.16651,12443.02214,-10287.85692,11879.39392,0.,33809.34781,-73741.60290,50400.,-11358.87318,0.,-57808.92777,102592.,-46847.60655,5865.696891,44181.49041,-63459.41291,14457.48525,-12443.02214,14367.96436,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,44181.49041,-63459.41291,14457.48525,-13280.04820,15334.47879,0.,38025.37259,-54617.16651,12443.02214,-12443.02214,14367.96436,0.,-10287.85692,11879.39392,0.,0.},
			{4751.757564,-22633.28523,31875.75880,-13469.96659,0.,0.,-11639.38143,50399.99999,-62463.47540,21996.36334,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-3591.991091,12443.02214,-9638.323502,0.,0.,0.,-11639.38143,50399.99999,-62463.47540,21996.36334,0.,12021.10144,-46847.60655,48384.,-11358.87318,-4173.516504,14457.48525,-11198.71993,0.,0.,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-4173.516504,14457.48525,-11198.71993,0.,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,0.,0.,0.,0.,0.,0.},
			{-1003.99205,6955.860840,-13469.96659,7589.466382,0.,0.,1639.512123,-11358.87319,21996.36334,-12393.54671,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,1639.512123,-11358.87319,21996.36334,-12393.54671,0.,-846.6404195,5865.696891,-11358.87318,6400.,0.,0.,0.,0.,0.,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-11476.58483,5253.47884,-1099.818167,0.,0.,0.,83732.93531,-27127.85727,4040.989977,0.,0.,-248785.2872,54111.05382,-5111.492932,0.,352131.6458,-46152.44305,2078.460970,-239210.4513,14287.05708,62934.44585,55821.95698,-18085.23818,2693.993318,0.,0.,-203132.3364,44181.49041,-4173.516504,0.,327312.,-42899.43441,1931.962732,-241101.4725,14400.,67012.23766,-124392.6436,27055.52691,-2555.746466,0.,245484.,-32174.57581,1448.972049,-205856.0459,12294.93554,62041.27659,140852.6583,-18460.97722,831.3843878,-144660.8835,8640.,49633.02127,-79736.81711,4762.352360,33506.11883,17981.27024},
			{5253.47885,-9345.21910,3809.881888,0.,0.,0.,-27127.85727,43372.45579,-13998.39991,0.,0.,54111.05382,-77721.59051,17706.73093,0.,-46152.44305,59489.02419,-7200.,14287.05708,-16497.27251,0.,-18085.23818,28914.97050,-9332.266605,0.,0.,44181.49041,-63459.41291,14457.48525,0.,-42899.43441,55296.,-6692.515222,14400.,-16627.68776,0.,27055.52691,-38860.79526,8853.365465,0.,-32174.57581,41472.,-5019.386416,12294.93554,-14196.96869,0.,-18460.97722,23795.60968,-2880.,8640.,-9976.612654,0.,4762.352360,-5499.090835,0.,0.},
			{-1099.81816,3809.881889,-2951.121822,0.,0.,0.,4040.989977,-13998.39991,10843.11394,0.,0.,-5111.492932,17706.73093,-13715.57480,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,2693.993319,-9332.266609,7228.742626,0.,0.,-4173.516504,14457.48525,-11198.71993,0.,1931.962732,-6692.515222,5184.,0.,0.,0.,-2555.746466,8853.365465,-6857.787398,0.,1448.972049,-5019.386416,3888.,0.,0.,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,0.,0.,0.,0.},
			{6531.35214,-1229.634091,0.,0.,0.,0.,-57965.50700,8031.93625,0.,0.,0.,197967.2700,-19049.40944,0.,0.,-311260.6624,18590.32006,0.,229651.2138,-6415.605974,-64595.94539,-28982.75339,4015.96813,0.,0.,0.,129311.6793,-12443.02214,0.,0.,-241101.4725,14400.,0.,198400.,-5542.562586,-60183.71873,59390.18104,-5714.822832,0.,0.,-144660.8835,8640.,0.,141164.0747,-3943.602413,-47759.39698,-62252.13248,3718.064012,0.,79360.,-2217.025034,-31839.59798,32807.31626,-916.5151392,-17195.34821,-6920.994149},
			{-1229.63410,1419.85915,0.,0.,0.,0.,8031.93625,-9274.48112,0.,0.,0.,-19049.40944,21996.36334,0.,0.,18590.32006,-21466.25258,0.,-6415.605974,7408.103671,0.,4015.96813,-4637.24056,0.,0.,0.,-12443.02214,14367.96436,0.,0.,14400.,-16627.68776,0.,-5542.562586,6400.,0.,-5714.822832,6598.909002,0.,0.,8640.,-9976.612654,0.,-3943.602413,4553.679829,0.,3718.064012,-4293.250516,0.,-2217.025034,2560.,0.,-916.5151392,1058.300524,0.,0.},
			{-1525.93584,0.,0.,0.,0.,0.,15574.01680,0.,0.,0.,0.,-59099.23855,0.,0.,0.,100931.1647,0.,0.,-79615.57635,0.,23666.54347,6229.60671,0.,0.,0.,0.,-32169.55082,0.,0.,0.,67012.23766,0.,0.,-60183.71873,0.,19600.,-11819.84771,0.,0.,0.,33506.11883,0.,0.,-36704.11422,0.,13609.55547,11534.99025,0.,0.,-17195.34821,0.,7776.888838,-5686.826882,0.,3500.,1126.978260},
			{13230.11101,-11457.61230,5819.690715,-1229.634092,0.,0.,-35594.59728,20703.91267,-7127.636352,1003.992031,0.,57960.00000,-21678.34791,4507.913042,-317.4901573,-53530.07079,11642.84501,-1099.818167,26188.60820,-2520.,-5285.996592,-106783.7919,62111.73799,-21382.90906,3011.976094,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-124392.6436,27055.52691,-2555.746466,59390.18104,-5714.822832,-11819.84771,347760.0000,-130070.0875,27047.47825,-1904.940944,-248785.2872,54111.05382,-5111.492932,112684.9455,-10843.11394,-21886.10519,-535300.7081,116428.4501,-10998.18167,197967.2700,-19049.40944,-36476.84197,392829.1232,-37800.,-59099.23855,-111005.9285},
			{-11457.6123,30581.07650,-27720.,8519.154886,0.,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,0.,-21678.34791,38472.,-17567.85246,2199.636334,11642.84501,-16723.02366,3809.881888,-2520.,2909.845357,0.,62111.73800,-135471.9749,92590.71229,-20867.58252,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,27055.52691,-38860.79526,8853.365465,-5714.822832,6598.909002,0.,-130070.0875,230832.,-105407.1147,13197.81800,54111.05382,-77721.59051,17706.73093,-10843.11394,12520.54951,0.,116428.4501,-167230.2366,38098.81888,-19049.40944,21996.36334,0.,-37800.,43647.68036,0.,0.},
			{5819.69066,-27719.99999,39039.67213,-16497.27251,0.,0.,-7127.636351,30863.57076,-38250.91056,13469.96659,0.,4507.913042,-17567.85246,18144.,-4259.577443,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,-21382.90905,92590.71229,-114752.7317,40409.89977,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-2555.746466,8853.365465,-6857.787398,0.,0.,0.,27047.47825,-105407.1147,108864.,-25557.46466,-5111.492932,17706.73093,-13715.57480,0.,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,0.,0.,0.,0.,0.,0.},
			{-1229.634099,8519.154887,-16497.27251,9295.160030,0.,0.,1003.992032,-6955.860840,13469.96659,-7589.466382,0.,-317.4901573,2199.636334,-4259.577443,2400.,0.,0.,0.,0.,0.,0.,3011.976097,-20867.58252,40409.89977,-22768.39914,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,-1904.940944,13197.81800,-25557.46466,14400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-11476.58483,5253.47884,-1099.818167,0.,0.,0.,55821.95698,-18085.23818,2693.993318,0.,0.,-124392.6436,27055.52691,-2555.746466,0.,140852.6583,-18460.97722,831.3843878,-79736.81711,4762.352360,17981.27024,83732.93531,-27127.85727,4040.989977,0.,0.,-203132.3364,44181.49041,-4173.516504,0.,245484.,-32174.57581,1448.972049,-144660.8835,8640.,33506.11883,-248785.2872,54111.05382,-5111.492932,0.,327312.,-42899.43441,1931.962732,-205856.0459,12294.93554,49633.02127,352131.6458,-46152.44305,2078.460970,-241101.4725,14400.,62041.27659,-239210.4513,14287.05708,67012.23766,62934.44585},
			{5253.47885,-9345.21910,3809.881888,0.,0.,0.,-18085.23818,28914.97050,-9332.266605,0.,0.,27055.52691,-38860.79526,8853.365465,0.,-18460.97722,23795.60968,-2880.,4762.352360,-5499.090835,0.,-27127.85727,43372.45579,-13998.39991,0.,0.,44181.49041,-63459.41291,14457.48525,0.,-32174.57581,41472.,-5019.386416,8640.,-9976.612654,0.,54111.05382,-77721.59051,17706.73093,0.,-42899.43441,55296.,-6692.515222,12294.93554,-14196.96869,0.,-46152.44305,59489.02419,-7200.,14400.,-16627.68776,0.,14287.05708,-16497.27251,0.,0.},
			{-1099.81816,3809.881889,-2951.121822,0.,0.,0.,2693.993319,-9332.266609,7228.742626,0.,0.,-2555.746466,8853.365465,-6857.787398,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,4040.989977,-13998.39991,10843.11394,0.,0.,-4173.516504,14457.48525,-11198.71993,0.,1448.972049,-5019.386416,3888.,0.,0.,0.,-5111.492932,17706.73093,-13715.57480,0.,1931.962732,-6692.515222,5184.,0.,0.,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,0.,0.,0.,0.},
			{6196.18489,-1166.53333,0.,0.,0.,0.,-41243.18141,5714.82284,0.,0.,0.,112684.9455,-10843.11394,0.,0.,-147643.8959,8818.163075,0.,93371.25892,-2608.447815,-22980.41043,-41243.18141,5714.82284,0.,0.,0.,138010.3092,-13280.04820,0.,0.,-205856.0459,12294.93554,0.,141164.0747,-3943.602413,-36704.11422,112684.9455,-10843.11394,0.,0.,-205856.0459,12294.93554,0.,160704.,-4489.475694,-45308.54223,-147643.8959,8818.163075,0.,141164.0747,-3943.602413,-45308.54223,93371.25892,-2608.447815,-36704.11422,-22980.41043},
			{-1166.53333,1346.99667,0.,0.,0.,0.,5714.82284,-6598.90901,0.,0.,0.,-10843.11394,12520.54951,0.,0.,8818.163075,-10182.33765,0.,-2608.447815,3011.976094,0.,5714.82284,-6598.90901,0.,0.,0.,-13280.04820,15334.47879,0.,0.,12294.93554,-14196.96869,0.,-3943.602413,4553.679829,0.,-10843.11394,12520.54951,0.,0.,12294.93554,-14196.96869,0.,-4489.475694,5184.,0.,8818.163075,-10182.33765,0.,-3943.602413,4553.679829,0.,-2608.447815,3011.976094,0.,0.},
			{-1412.74207,0.,0.,0.,0.,0.,11534.99025,0.,0.,0.,0.,-36476.84197,0.,0.,0.,53396.62909,0.,0.,-36854.85042,0.,9738.205176,8651.24267,0.,0.,0.,0.,-35739.86008,0.,0.,0.,62041.27659,0.,0.,-47759.39698,0.,13609.55547,-21886.10519,0.,0.,0.,49633.02127,0.,0.,-45308.54223,0.,14400.,26698.31455,0.,0.,-31839.59798,0.,12000.,-15794.93589,0.,7776.888838,3651.826941},
			{-14816.2076,6782.21205,-1419.859148,0.,0.,0.,36032.9184,-11673.97107,1738.965210,0.,0.,-53530.07079,11642.84501,-1099.818167,0.,45460.,-5958.254780,268.3281572,-20587.95764,1229.634092,3868.953347,144131.6731,-46695.88421,6955.860840,0.,0.,-174828.4790,38025.37259,-3591.991091,0.,140852.6583,-18460.97722,831.3843878,-62252.13248,3718.064012,11534.99025,-535300.7081,116428.4501,-10998.18167,0.,352131.6458,-46152.44305,2078.460970,-147643.8959,8818.163075,26698.31455,909200.,-119165.0956,5366.563145,-311260.6624,18590.32006,53396.62909,-720578.5174,43037.19323,100931.1647,216661.3874},
			{6782.212097,-12064.62597,4918.536370,0.,0.,0.,-11673.97107,18664.53323,-6023.952189,0.,0.,11642.84501,-16723.02366,3809.881888,0.,-5958.254780,7680.,-929.5160030,1229.634092,-1419.859148,0.,-46695.88421,74658.13284,-24095.80875,0.,0.,38025.37259,-54617.16651,12443.02214,0.,-18460.97722,23795.60968,-2880.,3718.064012,-4293.250516,0.,116428.4501,-167230.2366,38098.81888,0.,-46152.44305,59489.02419,-7200.,8818.163075,-10182.33765,0.,-119165.0956,153600.,-18590.32006,18590.32006,-21466.25258,0.,43037.19323,-49695.07017,0.,0.},
			{-1419.859143,4918.536370,-3809.881888,0.,0.,0.,1738.965210,-6023.952191,4666.133303,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,268.3281572,-929.5160030,720.,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,5366.563145,-18590.32006,14400.,0.,0.,0.,0.,0.,0.,0.},
			{6531.35214,-1229.634091,0.,0.,0.,0.,-28982.75339,4015.96813,0.,0.,0.,59390.18104,-5714.822832,0.,0.,-62252.13248,3718.064012,0.,32807.31626,-916.5151392,-6920.994149,-57965.50700,8031.93625,0.,0.,0.,129311.6793,-12443.02214,0.,0.,-144660.8835,8640.,0.,79360.,-2217.025034,-17195.34821,197967.2700,-19049.40944,0.,0.,-241101.4725,14400.,0.,141164.0747,-3943.602413,-31839.59798,-311260.6624,18590.32006,0.,198400.,-5542.562586,-47759.39698,229651.2138,-6415.605974,-60183.71873,-64595.94539},
			{-1229.63410,1419.85915,0.,0.,0.,0.,4015.96813,-4637.24056,0.,0.,0.,-5714.822832,6598.909002,0.,0.,3718.064012,-4293.250516,0.,-916.5151392,1058.300524,0.,8031.93625,-9274.48112,0.,0.,0.,-12443.02214,14367.96436,0.,0.,8640.,-9976.612654,0.,-2217.025034,2560.,0.,-19049.40944,21996.36334,0.,0.,14400.,-16627.68776,0.,-3943.602413,4553.679829,0.,18590.32006,-21466.25258,0.,-5542.562586,6400.,0.,-6415.605974,7408.103671,0.,0.},
			{-1412.74207,0.,0.,0.,0.,0.,8651.24267,0.,0.,0.,0.,-21886.10519,0.,0.,0.,26698.31455,0.,0.,-15794.93589,0.,3651.826941,11534.99025,0.,0.,0.,0.,-35739.86008,0.,0.,0.,49633.02127,0.,0.,-31839.59798,0.,7776.888838,-36476.84197,0.,0.,0.,62041.27659,0.,0.,-45308.54223,0.,12000.,53396.62909,0.,0.,-47759.39698,0.,14400.,-36854.85042,0.,13609.55547,9738.205176},
			{8640.1667,-1626.653004,0.,0.,0.,0.,-19170.28946,2656.313232,0.,0.,0.,26188.60820,-2520.,0.,0.,-20587.95764,1229.634092,0.,8680.,-242.4871131,-1525.935778,-95851.44763,13281.56616,0.,0.,0.,106914.5452,-10287.85692,0.,0.,-79736.81711,4762.352360,0.,32807.31626,-916.5151392,-5686.826882,392829.1232,-37800.,0.,0.,-239210.4513,14287.05708,0.,93371.25892,-2608.447815,-15794.93589,-720578.5174,43037.19323,0.,229651.2138,-6415.605974,-36854.85042,607600.,-16974.09792,-79615.57635,-192267.9080},
			{-1626.65301,1878.297097,0.,0.,0.,0.,2656.313229,-3067.24632,0.,0.,0.,-2520.,2909.845357,0.,0.,1229.634092,-1419.859148,0.,-242.4871131,280.,0.,13281.56616,-15336.23160,0.,0.,0.,-10287.85692,11879.39392,0.,0.,4762.352360,-5499.090835,0.,-916.5151392,1058.300524,0.,-37800.,43647.68036,0.,0.,14287.05708,-16497.27251,0.,-2608.447815,3011.976094,0.,43037.19323,-49695.07017,0.,-6415.605974,7408.103671,0.,-16974.09792,19600.,0.,0.},
			{-1525.93584,0.,0.,0.,0.,0.,6229.60671,0.,0.,0.,0.,-11819.84771,0.,0.,0.,11534.99025,0.,0.,-5686.826882,0.,1126.978260,15574.01680,0.,0.,0.,0.,-32169.55082,0.,0.,0.,33506.11883,0.,0.,-17195.34821,0.,3500.,-59099.23855,0.,0.,0.,67012.23766,0.,0.,-36704.11422,0.,7776.888838,100931.1647,0.,0.,-60183.71873,0.,13609.55547,-79615.57635,0.,19600.,23666.54347},
			{-2047.25769,0.,0.,0.,0.,0.,4178.947267,0.,0.,0.,0.,-5285.996592,0.,0.,0.,3868.953347,0.,0.,-1525.935778,0.,252.,25073.68342,0.,0.,0.,0.,-25895.98888,0.,0.,0.,17981.27024,0.,0.,-6920.994149,0.,1126.978260,-111005.9285,0.,0.,0.,62934.44585,0.,0.,-22980.41043,0.,3651.826941,216661.3874,0.,0.,-64595.94539,0.,9738.205176,-192267.9080,0.,23666.54347,63504.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1959.99950,-5383.2139,9015.826058,-8519.319340,4200.00000,-835.7894471,-6584.2284,10097.4849,-9661.82590,4925.362810,-1028.78570,13230.11101,-11457.6123,5819.69066,-1229.634099,-14816.2076,6782.212097,-1419.859143,8640.1667,-1626.65301,-2047.25769,-6584.2284,10097.4849,-9661.82590,4925.362810,-1028.78570,10802.34025,-9355.10139,4751.757564,-1003.99205,-11476.58483,5253.47885,-1099.81816,6531.35214,-1229.63410,-1525.93584,13230.11101,-11457.6123,5819.69066,-1229.634099,-11476.58483,5253.47885,-1099.81816,6196.18489,-1166.53333,-1412.74207,-14816.2076,6782.212097,-1419.859143,6531.35214,-1229.63410,-1412.74207,8640.1667,-1626.65301,-1525.93584,-2047.25769},
			{-5383.2139,33096.0000,-83609.96448,103016.3016,-61834.21385,14476.29787,10097.4848,-40465.5705,63751.51762,-44899.88862,11879.39392,-11457.61230,30581.07650,-27719.99999,8519.154887,6782.21205,-12064.62597,4918.536370,-1626.653003,1878.297097,0.,10097.4848,-40465.5705,63751.51762,-44899.88862,11879.39392,-9355.10127,24969.34437,-22633.28523,6955.860840,5253.47884,-9345.21910,3809.881889,-1229.634091,1419.85915,0.,-11457.61230,30581.07650,-27719.99999,8519.154887,5253.47884,-9345.21910,3809.881889,-1166.53333,1346.99667,0.,6782.21205,-12064.62597,4918.536370,-1229.634091,1419.85915,0.,-1626.653003,1878.297097,0.,0.},
			{9015.826058,-83609.96448,277200.,-416728.6598,293014.3477,-78493.04477,-9661.825918,63751.51762,-141972.4255,128683.4255,-41407.82533,5819.690715,-27720.,39039.67213,-16497.27251,-1419.859148,4918.536370,-3809.881888,0.,0.,0.,-9661.825918,63751.51762,-141972.4255,128683.4255,-41407.82533,4751.757568,-22633.28523,31875.75880,-13469.96659,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,5819.690715,-27720.,39039.67213,-16497.27251,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,-1419.859148,4918.536370,-3809.881888,0.,0.,0.,0.,0.,0.,0.},
			{-8519.319340,103016.3016,-416728.6598,734000.,-588944.2418,176903.2820,4925.362927,-44899.88862,128683.4255,-143050.2010,54438.22186,-1229.634092,8519.154886,-16497.27251,9295.160030,0.,0.,0.,0.,0.,0.,4925.362927,-44899.88862,128683.4255,-143050.2010,54438.22186,-1003.992031,6955.860840,-13469.96659,7589.466382,0.,0.,0.,0.,0.,0.,-1229.634092,8519.154886,-16497.27251,9295.160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4200.00000,-61834.21385,293014.3477,-588944.2418,529200.,-175515.7839,-1028.785692,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1028.785692,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-835.7894471,14476.29787,-78493.04477,176903.2820,-175515.7839,63504.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-6584.2284,10097.4848,-9661.825918,4925.362927,-1028.785692,0.,39536.0001,-37827.9896,22915.22463,-7196.44356,840.,-106783.7919,62111.73800,-21382.90905,3011.976097,144131.6731,-46695.88421,6955.860840,-95851.44763,13281.56616,25073.68342,19768.0002,-18913.9949,11457.61232,-3598.221779,420.,-58125.73395,33809.34781,-11639.38143,1639.512123,83732.93531,-27127.85727,4040.989977,-57965.50700,8031.93625,15574.01680,-35594.59728,20703.91267,-7127.636351,1003.992032,55821.95698,-18085.23818,2693.993319,-41243.18141,5714.82284,11534.99025,36032.9184,-11673.97107,1738.965210,-28982.75339,4015.96813,8651.24267,-19170.28946,2656.313229,6229.60671,4178.947267},
			{10097.4849,-40465.5705,63751.51762,-44899.88862,11879.39392,0.,-37827.9896,112000.0001,-123625.6285,58656.96888,-9699.484525,62111.73799,-135471.9749,92590.71229,-20867.58252,-46695.88421,74658.13284,-24095.80875,13281.56616,-15336.23160,0.,-18913.9949,56000.0000,-61812.81421,29328.48444,-4849.742262,33809.34782,-73741.60290,50399.99999,-11358.87319,-27127.85727,43372.45579,-13998.39991,8031.93625,-9274.48112,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,-18085.23818,28914.97050,-9332.266609,5714.82284,-6598.90901,0.,-11673.97107,18664.53323,-6023.952191,4015.96813,-4637.24056,0.,2656.313229,-3067.24632,0.,0.},
			{-9661.82590,63751.51762,-141972.4255,128683.4255,-41407.82533,0.,22915.22463,-123625.6285,216720.,-147665.3514,33809.34781,-21382.90906,92590.71229,-114752.7317,40409.89977,6955.860840,-24095.80875,18664.53321,0.,0.,0.,11457.61232,-61812.81421,108360.,-73832.67568,16904.67391,-11639.38143,50400.,-62463.47540,21996.36334,4040.989977,-13998.39991,10843.11394,0.,0.,0.,-7127.636352,30863.57076,-38250.91056,13469.96659,2693.993318,-9332.266605,7228.742626,0.,0.,0.,1738.965210,-6023.952189,4666.133303,0.,0.,0.,0.,0.,0.,0.},
			{4925.362810,-44899.88862,128683.4255,-143050.2010,54438.22186,0.,-7196.44356,58656.96888,-147665.3514,140800.,-44448.62202,3011.976094,-20867.58252,40409.89977,-22768.39914,0.,0.,0.,0.,0.,0.,-3598.221779,29328.48444,-73832.67568,70400.,-22224.31101,1639.512123,-11358.87318,21996.36334,-12393.54671,0.,0.,0.,0.,0.,0.,1003.992031,-6955.860840,13469.96659,-7589.466382,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1028.78570,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,840.,-9699.484525,33809.34781,-44448.62202,19600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,420.,-4849.742262,16904.67391,-22224.31101,9800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{13230.11101,-11457.61230,5819.690715,-1229.634092,0.,0.,-106783.7919,62111.73799,-21382.90906,3011.976094,0.,347760.0000,-130070.0875,27047.47825,-1904.940944,-535300.7081,116428.4501,-10998.18167,392829.1232,-37800.,-111005.9285,-35594.59728,20703.91267,-7127.636352,1003.992031,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-248785.2872,54111.05382,-5111.492932,197967.2700,-19049.40944,-59099.23855,57960.00000,-21678.34791,4507.913042,-317.4901573,-124392.6436,27055.52691,-2555.746466,112684.9455,-10843.11394,-36476.84197,-53530.07079,11642.84501,-1099.818167,59390.18104,-5714.822832,-21886.10519,26188.60820,-2520.,-11819.84771,-5285.996592},
			{-11457.6123,30581.07650,-27720.,8519.154886,0.,0.,62111.73800,-135471.9749,92590.71229,-20867.58252,0.,-130070.0875,230832.,-105407.1147,13197.81800,116428.4501,-167230.2366,38098.81888,-37800.,43647.68036,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,54111.05382,-77721.59051,17706.73093,-19049.40944,21996.36334,0.,-21678.34791,38472.,-17567.85246,2199.636334,27055.52691,-38860.79526,8853.365465,-10843.11394,12520.54951,0.,11642.84501,-16723.02366,3809.881888,-5714.822832,6598.909002,0.,-2520.,2909.845357,0.,0.},
			{5819.69066,-27719.99999,39039.67213,-16497.27251,0.,0.,-21382.90905,92590.71229,-114752.7317,40409.89977,0.,27047.47825,-105407.1147,108864.,-25557.46466,-10998.18167,38098.81888,-29511.21822,0.,0.,0.,-7127.636351,30863.57076,-38250.91056,13469.96659,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-5111.492932,17706.73093,-13715.57480,0.,0.,0.,4507.913042,-17567.85246,18144.,-4259.577443,-2555.746466,8853.365465,-6857.787398,0.,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,0.,0.,0.,0.},
			{-1229.634099,8519.154887,-16497.27251,9295.160030,0.,0.,3011.976097,-20867.58252,40409.89977,-22768.39914,0.,-1904.940944,13197.81800,-25557.46466,14400.,0.,0.,0.,0.,0.,0.,1003.992032,-6955.860840,13469.96659,-7589.466382,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,-317.4901573,2199.636334,-4259.577443,2400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-14816.2076,6782.21205,-1419.859148,0.,0.,0.,144131.6731,-46695.88421,6955.860840,0.,0.,-535300.7081,116428.4501,-10998.18167,0.,909200.,-119165.0956,5366.563145,-720578.5174,43037.19323,216661.3874,36032.9184,-11673.97107,1738.965210,0.,0.,-174828.4790,38025.37259,-3591.991091,0.,352131.6458,-46152.44305,2078.460970,-311260.6624,18590.32006,100931.1647,-53530.07079,11642.84501,-1099.818167,0.,140852.6583,-18460.97722,831.3843878,-147643.8959,8818.163075,53396.62909,45460.,-5958.254780,268.3281572,-62252.13248,3718.064012,26698.31455,-20587.95764,1229.634092,11534.99025,3868.953347},
			{6782.212097,-12064.62597,4918.536370,0.,0.,0.,-46695.88421,74658.13284,-24095.80875,0.,0.,116428.4501,-167230.2366,38098.81888,0.,-119165.0956,153600.,-18590.32006,43037.19323,-49695.07017,0.,-11673.97107,18664.53323,-6023.952189,0.,0.,38025.37259,-54617.16651,12443.02214,0.,-46152.44305,59489.02419,-7200.,18590.32006,-21466.25258,0.,11642.84501,-16723.02366,3809.881888,0.,-18460.97722,23795.60968,-2880.,8818.163075,-10182.33765,0.,-5958.254780,7680.,-929.5160030,3718.064012,-4293.250516,0.,1229.634092,-1419.859148,0.,0.},
			{-1419.859143,4918.536370,-3809.881888,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,5366.563145,-18590.32006,14400.,0.,0.,0.,1738.965210,-6023.952191,4666.133303,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,268.3281572,-929.5160030,720.,0.,0.,0.,0.,0.,0.,0.},
			{8640.1667,-1626.653003,0.,0.,0.,0.,-95851.44763,13281.56616,0.,0.,0.,392829.1232,-37800.,0.,0.,-720578.5174,43037.19323,0.,607600.,-16974.09792,-192267.9080,-19170.28946,2656.313229,0.,0.,0.,106914.5452,-10287.85692,0.,0.,-239210.4513,14287.05708,0.,229651.2138,-6415.605974,-79615.57635,26188.60820,-2520.,0.,0.,-79736.81711,4762.352360,0.,93371.25892,-2608.447815,-36854.85042,-20587.95764,1229.634092,0.,32807.31626,-916.5151392,-15794.93589,8680.,-242.4871131,-5686.826882,-1525.935778},
			{-1626.65301,1878.297097,0.,0.,0.,0.,13281.56616,-15336.23160,0.,0.,0.,-37800.,43647.68036,0.,0.,43037.19323,-49695.07017,0.,-16974.09792,19600.,0.,2656.313229,-3067.24632,0.,0.,0.,-10287.85692,11879.39392,0.,0.,14287.05708,-16497.27251,0.,-6415.605974,7408.103671,0.,-2520.,2909.845357,0.,0.,4762.352360,-5499.090835,0.,-2608.447815,3011.976094,0.,1229.634092,-1419.859148,0.,-916.5151392,1058.300524,0.,-242.4871131,280.,0.,0.},
			{-2047.25769,0.,0.,0.,0.,0.,25073.68342,0.,0.,0.,0.,-111005.9285,0.,0.,0.,216661.3874,0.,0.,-192267.9080,0.,63504.,4178.947267,0.,0.,0.,0.,-25895.98888,0.,0.,0.,62934.44585,0.,0.,-64595.94539,0.,23666.54347,-5285.996592,0.,0.,0.,17981.27024,0.,0.,-22980.41043,0.,9738.205176,3868.953347,0.,0.,-6920.994149,0.,3651.826941,-1525.935778,0.,1126.978260,252.},
			{-6584.2284,10097.4848,-9661.825918,4925.362927,-1028.785692,0.,19768.0002,-18913.9949,11457.61232,-3598.221779,420.,-35594.59728,20703.91267,-7127.636351,1003.992032,36032.9184,-11673.97107,1738.965210,-19170.28946,2656.313232,4178.947267,39536.0001,-37827.9896,22915.22463,-7196.44356,840.,-58125.73395,33809.34781,-11639.38143,1639.512123,55821.95698,-18085.23818,2693.993319,-28982.75339,4015.968128,6229.60671,-106783.7919,62111.73800,-21382.90905,3011.976097,83732.93531,-27127.85727,4040.989977,-41243.18141,5714.82284,8651.24267,144131.6731,-46695.88421,6955.860840,-57965.50700,8031.936250,11534.99025,-95851.44763,13281.56616,15574.01680,25073.68342},
			{10097.4849,-40465.5705,63751.51762,-44899.88862,11879.39392,0.,-18913.9949,56000.0000,-61812.81421,29328.48444,-4849.742262,20703.91267,-45157.32496,30863.57076,-6955.860840,-11673.97107,18664.53323,-6023.952191,2656.313229,-3067.24632,0.,-37827.9896,112000.0001,-123625.6285,58656.96888,-9699.484525,33809.34782,-73741.60290,50399.99999,-11358.87319,-18085.23818,28914.97050,-9332.266609,4015.96813,-4637.24056,0.,62111.73799,-135471.9749,92590.71229,-20867.58252,-27127.85727,43372.45579,-13998.39991,5714.82284,-6598.90901,0.,-46695.88421,74658.13284,-24095.80875,8031.93625,-9274.48112,0.,13281.56616,-15336.23160,0.,0.},
			{-9661.82590,63751.51762,-141972.4255,128683.4255,-41407.82533,0.,11457.61232,-61812.81421,108360.,-73832.67568,16904.67391,-7127.636352,30863.57076,-38250.91056,13469.96659,1738.965210,-6023.952189,4666.133303,0.,0.,0.,22915.22463,-123625.6285,216720.,-147665.3514,33809.34781,-11639.38143,50400.,-62463.47540,21996.36334,2693.993318,-9332.266605,7228.742626,0.,0.,0.,-21382.90906,92590.71229,-114752.7317,40409.89977,4040.989977,-13998.39991,10843.11394,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,0.,0.,0.,0.,0.},
			{4925.362810,-44899.88862,128683.4255,-143050.2010,54438.22186,0.,-3598.221779,29328.48444,-73832.67568,70400.,-22224.31101,1003.992031,-6955.860840,13469.96659,-7589.466382,0.,0.,0.,0.,0.,0.,-7196.44356,58656.96888,-147665.3514,140800.,-44448.62202,1639.512123,-11358.87318,21996.36334,-12393.54671,0.,0.,0.,0.,0.,0.,3011.976094,-20867.58252,40409.89977,-22768.39914,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1028.78570,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,420.,-4849.742262,16904.67391,-22224.31101,9800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,840.,-9699.484525,33809.34781,-44448.62202,19600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{10802.34025,-9355.10127,4751.757568,-1003.992031,0.,0.,-58125.73395,33809.34782,-11639.38143,1639.512123,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-174828.4790,38025.37259,-3591.991091,106914.5452,-10287.85692,-25895.98888,-58125.73395,33809.34782,-11639.38143,1639.512123,0.,154560.0000,-57808.92777,12021.10144,-846.6404195,-203132.3364,44181.49041,-4173.516504,129311.6793,-12443.02214,-32169.55082,141972.4255,-53100.89083,11042.08676,-777.6888838,-203132.3364,44181.49041,-4173.516504,138010.3092,-13280.04820,-35739.86008,-174828.4790,38025.37259,-3591.991091,129311.6793,-12443.02214,-35739.86008,106914.5452,-10287.85692,-32169.55082,-25895.98888},
			{-9355.10139,24969.34437,-22633.28523,6955.860840,0.,0.,33809.34781,-73741.60290,50400.,-11358.87318,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,38025.37259,-54617.16651,12443.02214,-10287.85692,11879.39392,0.,33809.34781,-73741.60290,50400.,-11358.87318,0.,-57808.92777,102592.,-46847.60655,5865.696891,44181.49041,-63459.41291,14457.48525,-12443.02214,14367.96436,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,44181.49041,-63459.41291,14457.48525,-13280.04820,15334.47879,0.,38025.37259,-54617.16651,12443.02214,-12443.02214,14367.96436,0.,-10287.85692,11879.39392,0.,0.},
			{4751.757564,-22633.28523,31875.75880,-13469.96659,0.,0.,-11639.38143,50399.99999,-62463.47540,21996.36334,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-3591.991091,12443.02214,-9638.323502,0.,0.,0.,-11639.38143,50399.99999,-62463.47540,21996.36334,0.,12021.10144,-46847.60655,48384.,-11358.87318,-4173.516504,14457.48525,-11198.71993,0.,0.,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-4173.516504,14457.48525,-11198.71993,0.,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,0.,0.,0.,0.,0.,0.},
			{-1003.99205,6955.860840,-13469.96659,7589.466382,0.,0.,1639.512123,-11358.87319,21996.36334,-12393.54671,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,1639.512123,-11358.87319,21996.36334,-12393.54671,0.,-846.6404195,5865.696891,-11358.87318,6400.,0.,0.,0.,0.,0.,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-11476.58483,5253.47884,-1099.818167,0.,0.,0.,83732.93531,-27127.85727,4040.989977,0.,0.,-248785.2872,54111.05382,-5111.492932,0.,352131.6458,-46152.44305,2078.460970,-239210.4513,14287.05708,62934.44585,55821.95698,-18085.23818,2693.993318,0.,0.,-203132.3364,44181.49041,-4173.516504,0.,327312.,-42899.43441,1931.962732,-241101.4725,14400.,67012.23766,-124392.6436,27055.52691,-2555.746466,0.,245484.,-32174.57581,1448.972049,-205856.0459,12294.93554,62041.27659,140852.6583,-18460.97722,831.3843878,-144660.8835,8640.,49633.02127,-79736.81711,4762.352360,33506.11883,17981.27024},
			{5253.47885,-9345.21910,3809.881888,0.,0.,0.,-27127.85727,43372.45579,-13998.39991,0.,0.,54111.05382,-77721.59051,17706.73093,0.,-46152.44305,59489.02419,-7200.,14287.05708,-16497.27251,0.,-18085.23818,28914.97050,-9332.266605,0.,0.,44181.49041,-63459.41291,14457.48525,0.,-42899.43441,55296.,-6692.515222,14400.,-16627.68776,0.,27055.52691,-38860.79526,8853.365465,0.,-32174.57581,41472.,-5019.386416,12294.93554,-14196.96869,0.,-18460.97722,23795.60968,-2880.,8640.,-9976.612654,0.,4762.352360,-5499.090835,0.,0.},
			{-1099.81816,3809.881889,-2951.121822,0.,0.,0.,4040.989977,-13998.39991,10843.11394,0.,0.,-5111.492932,17706.73093,-13715.57480,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,2693.993319,-9332.266609,7228.742626,0.,0.,-4173.516504,14457.48525,-11198.71993,0.,1931.962732,-6692.515222,5184.,0.,0.,0.,-2555.746466,8853.365465,-6857.787398,0.,1448.972049,-5019.386416,3888.,0.,0.,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,0.,0.,0.,0.},
			{6531.35214,-1229.634091,0.,0.,0.,0.,-57965.50700,8031.93625,0.,0.,0.,197967.2700,-19049.40944,0.,0.,-311260.6624,18590.32006,0.,229651.2138,-6415.605974,-64595.94539,-28982.75339,4015.96813,0.,0.,0.,129311.6793,-12443.02214,0.,0.,-241101.4725,14400.,0.,198400.,-5542.562586,-60183.71873,59390.18104,-5714.822832,0.,0.,-144660.8835,8640.,0.,141164.0747,-3943.602413,-47759.39698,-62252.13248,3718.064012,0.,79360.,-2217.025034,-31839.59798,32807.31626,-916.5151392,-17195.34821,-6920.994149},
			{-1229.63410,1419.85915,0.,0.,0.,0.,8031.93625,-9274.48112,0.,0.,0.,-19049.40944,21996.36334,0.,0.,18590.32006,-21466.25258,0.,-6415.605974,7408.103671,0.,4015.96813,-4637.24056,0.,0.,0.,-12443.02214,14367.96436,0.,0.,14400.,-16627.68776,0.,-5542.562586,6400.,0.,-5714.822832,6598.909002,0.,0.,8640.,-9976.612654,0.,-3943.602413,4553.679829,0.,3718.064012,-4293.250516,0.,-2217.025034,2560.,0.,-916.5151392,1058.300524,0.,0.},
			{-1525.93584,0.,0.,0.,0.,0.,15574.01680,0.,0.,0.,0.,-59099.23855,0.,0.,0.,100931.1647,0.,0.,-79615.57635,0.,23666.54347,6229.60671,0.,0.,0.,0.,-32169.55082,0.,0.,0.,67012.23766,0.,0.,-60183.71873,0.,19600.,-11819.84771,0.,0.,0.,33506.11883,0.,0.,-36704.11422,0.,13609.55547,11534.99025,0.,0.,-17195.34821,0.,7776.888838,-5686.826882,0.,3500.,1126.978260},
			{13230.11101,-11457.61230,5819.690715,-1229.634092,0.,0.,-35594.59728,20703.91267,-7127.636352,1003.992031,0.,57960.00000,-21678.34791,4507.913042,-317.4901573,-53530.07079,11642.84501,-1099.818167,26188.60820,-2520.,-5285.996592,-106783.7919,62111.73799,-21382.90906,3011.976094,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-124392.6436,27055.52691,-2555.746466,59390.18104,-5714.822832,-11819.84771,347760.0000,-130070.0875,27047.47825,-1904.940944,-248785.2872,54111.05382,-5111.492932,112684.9455,-10843.11394,-21886.10519,-535300.7081,116428.4501,-10998.18167,197967.2700,-19049.40944,-36476.84197,392829.1232,-37800.,-59099.23855,-111005.9285},
			{-11457.6123,30581.07650,-27720.,8519.154886,0.,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,0.,-21678.34791,38472.,-17567.85246,2199.636334,11642.84501,-16723.02366,3809.881888,-2520.,2909.845357,0.,62111.73800,-135471.9749,92590.71229,-20867.58252,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,27055.52691,-38860.79526,8853.365465,-5714.822832,6598.909002,0.,-130070.0875,230832.,-105407.1147,13197.81800,54111.05382,-77721.59051,17706.73093,-10843.11394,12520.54951,0.,116428.4501,-167230.2366,38098.81888,-19049.40944,21996.36334,0.,-37800.,43647.68036,0.,0.},
			{5819.69066,-27719.99999,39039.67213,-16497.27251,0.,0.,-7127.636351,30863.57076,-38250.91056,13469.96659,0.,4507.913042,-17567.85246,18144.,-4259.577443,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,-21382.90905,92590.71229,-114752.7317,40409.89977,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-2555.746466,8853.365465,-6857.787398,0.,0.,0.,27047.47825,-105407.1147,108864.,-25557.46466,-5111.492932,17706.73093,-13715.57480,0.,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,0.,0.,0.,0.,0.,0.},
			{-1229.634099,8519.154887,-16497.27251,9295.160030,0.,0.,1003.992032,-6955.860840,13469.96659,-7589.466382,0.,-317.4901573,2199.636334,-4259.577443,2400.,0.,0.,0.,0.,0.,0.,3011.976097,-20867.58252,40409.89977,-22768.39914,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,-1904.940944,13197.81800,-25557.46466,14400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-11476.58483,5253.47884,-1099.818167,0.,0.,0.,55821.95698,-18085.23818,2693.993318,0.,0.,-124392.6436,27055.52691,-2555.746466,0.,140852.6583,-18460.97722,831.3843878,-79736.81711,4762.352360,17981.27024,83732.93531,-27127.85727,4040.989977,0.,0.,-203132.3364,44181.49041,-4173.516504,0.,245484.,-32174.57581,1448.972049,-144660.8835,8640.,33506.11883,-248785.2872,54111.05382,-5111.492932,0.,327312.,-42899.43441,1931.962732,-205856.0459,12294.93554,49633.02127,352131.6458,-46152.44305,2078.460970,-241101.4725,14400.,62041.27659,-239210.4513,14287.05708,67012.23766,62934.44585},
			{5253.47885,-9345.21910,3809.881888,0.,0.,0.,-18085.23818,28914.97050,-9332.266605,0.,0.,27055.52691,-38860.79526,8853.365465,0.,-18460.97722,23795.60968,-2880.,4762.352360,-5499.090835,0.,-27127.85727,43372.45579,-13998.39991,0.,0.,44181.49041,-63459.41291,14457.48525,0.,-32174.57581,41472.,-5019.386416,8640.,-9976.612654,0.,54111.05382,-77721.59051,17706.73093,0.,-42899.43441,55296.,-6692.515222,12294.93554,-14196.96869,0.,-46152.44305,59489.02419,-7200.,14400.,-16627.68776,0.,14287.05708,-16497.27251,0.,0.},
			{-1099.81816,3809.881889,-2951.121822,0.,0.,0.,2693.993319,-9332.266609,7228.742626,0.,0.,-2555.746466,8853.365465,-6857.787398,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,4040.989977,-13998.39991,10843.11394,0.,0.,-4173.516504,14457.48525,-11198.71993,0.,1448.972049,-5019.386416,3888.,0.,0.,0.,-5111.492932,17706.73093,-13715.57480,0.,1931.962732,-6692.515222,5184.,0.,0.,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,0.,0.,0.,0.},
			{6196.18489,-1166.53333,0.,0.,0.,0.,-41243.18141,5714.82284,0.,0.,0.,112684.9455,-10843.11394,0.,0.,-147643.8959,8818.163075,0.,93371.25892,-2608.447815,-22980.41043,-41243.18141,5714.82284,0.,0.,0.,138010.3092,-13280.04820,0.,0.,-205856.0459,12294.93554,0.,141164.0747,-3943.602413,-36704.11422,112684.9455,-10843.11394,0.,0.,-205856.0459,12294.93554,0.,160704.,-4489.475694,-45308.54223,-147643.8959,8818.163075,0.,141164.0747,-3943.602413,-45308.54223,93371.25892,-2608.447815,-36704.11422,-22980.41043},
			{-1166.53333,1346.99667,0.,0.,0.,0.,5714.82284,-6598.90901,0.,0.,0.,-10843.11394,12520.54951,0.,0.,8818.163075,-10182.33765,0.,-2608.447815,3011.976094,0.,5714.82284,-6598.90901,0.,0.,0.,-13280.04820,15334.47879,0.,0.,12294.93554,-14196.96869,0.,-3943.602413,4553.679829,0.,-10843.11394,12520.54951,0.,0.,12294.93554,-14196.96869,0.,-4489.475694,5184.,0.,8818.163075,-10182.33765,0.,-3943.602413,4553.679829,0.,-2608.447815,3011.976094,0.,0.},
			{-1412.74207,0.,0.,0.,0.,0.,11534.99025,0.,0.,0.,0.,-36476.84197,0.,0.,0.,53396.62909,0.,0.,-36854.85042,0.,9738.205176,8651.24267,0.,0.,0.,0.,-35739.86008,0.,0.,0.,62041.27659,0.,0.,-47759.39698,0.,13609.55547,-21886.10519,0.,0.,0.,49633.02127,0.,0.,-45308.54223,0.,14400.,26698.31455,0.,0.,-31839.59798,0.,12000.,-15794.93589,0.,7776.888838,3651.826941},
			{-14816.2076,6782.21205,-1419.859148,0.,0.,0.,36032.9184,-11673.97107,1738.965210,0.,0.,-53530.07079,11642.84501,-1099.818167,0.,45460.,-5958.254780,268.3281572,-20587.95764,1229.634092,3868.953347,144131.6731,-46695.88421,6955.860840,0.,0.,-174828.4790,38025.37259,-3591.991091,0.,140852.6583,-18460.97722,831.3843878,-62252.13248,3718.064012,11534.99025,-535300.7081,116428.4501,-10998.18167,0.,352131.6458,-46152.44305,2078.460970,-147643.8959,8818.163075,26698.31455,909200.,-119165.0956,5366.563145,-311260.6624,18590.32006,53396.62909,-720578.5174,43037.19323,100931.1647,216661.3874},
			{6782.212097,-12064.62597,4918.536370,0.,0.,0.,-11673.97107,18664.53323,-6023.952189,0.,0.,11642.84501,-16723.02366,3809.881888,0.,-5958.254780,7680.,-929.5160030,1229.634092,-1419.859148,0.,-46695.88421,74658.13284,-24095.80875,0.,0.,38025.37259,-54617.16651,12443.02214,0.,-18460.97722,23795.60968,-2880.,3718.064012,-4293.250516,0.,116428.4501,-167230.2366,38098.81888,0.,-46152.44305,59489.02419,-7200.,8818.163075,-10182.33765,0.,-119165.0956,153600.,-18590.32006,18590.32006,-21466.25258,0.,43037.19323,-49695.07017,0.,0.},
			{-1419.859143,4918.536370,-3809.881888,0.,0.,0.,1738.965210,-6023.952191,4666.133303,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,268.3281572,-929.5160030,720.,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,5366.563145,-18590.32006,14400.,0.,0.,0.,0.,0.,0.,0.},
			{6531.35214,-1229.634091,0.,0.,0.,0.,-28982.75339,4015.96813,0.,0.,0.,59390.18104,-5714.822832,0.,0.,-62252.13248,3718.064012,0.,32807.31626,-916.5151392,-6920.994149,-57965.50700,8031.93625,0.,0.,0.,129311.6793,-12443.02214,0.,0.,-144660.8835,8640.,0.,79360.,-2217.025034,-17195.34821,197967.2700,-19049.40944,0.,0.,-241101.4725,14400.,0.,141164.0747,-3943.602413,-31839.59798,-311260.6624,18590.32006,0.,198400.,-5542.562586,-47759.39698,229651.2138,-6415.605974,-60183.71873,-64595.94539},
			{-1229.63410,1419.85915,0.,0.,0.,0.,4015.96813,-4637.24056,0.,0.,0.,-5714.822832,6598.909002,0.,0.,3718.064012,-4293.250516,0.,-916.5151392,1058.300524,0.,8031.93625,-9274.48112,0.,0.,0.,-12443.02214,14367.96436,0.,0.,8640.,-9976.612654,0.,-2217.025034,2560.,0.,-19049.40944,21996.36334,0.,0.,14400.,-16627.68776,0.,-3943.602413,4553.679829,0.,18590.32006,-21466.25258,0.,-5542.562586,6400.,0.,-6415.605974,7408.103671,0.,0.},
			{-1412.74207,0.,0.,0.,0.,0.,8651.24267,0.,0.,0.,0.,-21886.10519,0.,0.,0.,26698.31455,0.,0.,-15794.93589,0.,3651.826941,11534.99025,0.,0.,0.,0.,-35739.86008,0.,0.,0.,49633.02127,0.,0.,-31839.59798,0.,7776.888838,-36476.84197,0.,0.,0.,62041.27659,0.,0.,-45308.54223,0.,12000.,53396.62909,0.,0.,-47759.39698,0.,14400.,-36854.85042,0.,13609.55547,9738.205176},
			{8640.1667,-1626.653004,0.,0.,0.,0.,-19170.28946,2656.313232,0.,0.,0.,26188.60820,-2520.,0.,0.,-20587.95764,1229.634092,0.,8680.,-242.4871131,-1525.935778,-95851.44763,13281.56616,0.,0.,0.,106914.5452,-10287.85692,0.,0.,-79736.81711,4762.352360,0.,32807.31626,-916.5151392,-5686.826882,392829.1232,-37800.,0.,0.,-239210.4513,14287.05708,0.,93371.25892,-2608.447815,-15794.93589,-720578.5174,43037.19323,0.,229651.2138,-6415.605974,-36854.85042,607600.,-16974.09792,-79615.57635,-192267.9080},
			{-1626.65301,1878.297097,0.,0.,0.,0.,2656.313229,-3067.24632,0.,0.,0.,-2520.,2909.845357,0.,0.,1229.634092,-1419.859148,0.,-242.4871131,280.,0.,13281.56616,-15336.23160,0.,0.,0.,-10287.85692,11879.39392,0.,0.,4762.352360,-5499.090835,0.,-916.5151392,1058.300524,0.,-37800.,43647.68036,0.,0.,14287.05708,-16497.27251,0.,-2608.447815,3011.976094,0.,43037.19323,-49695.07017,0.,-6415.605974,7408.103671,0.,-16974.09792,19600.,0.,0.},
			{-1525.93584,0.,0.,0.,0.,0.,6229.60671,0.,0.,0.,0.,-11819.84771,0.,0.,0.,11534.99025,0.,0.,-5686.826882,0.,1126.978260,15574.01680,0.,0.,0.,0.,-32169.55082,0.,0.,0.,33506.11883,0.,0.,-17195.34821,0.,3500.,-59099.23855,0.,0.,0.,67012.23766,0.,0.,-36704.11422,0.,7776.888838,100931.1647,0.,0.,-60183.71873,0.,13609.55547,-79615.57635,0.,19600.,23666.54347},
			{-2047.25769,0.,0.,0.,0.,0.,4178.947267,0.,0.,0.,0.,-5285.996592,0.,0.,0.,3868.953347,0.,0.,-1525.935778,0.,252.,25073.68342,0.,0.,0.,0.,-25895.98888,0.,0.,0.,17981.27024,0.,0.,-6920.994149,0.,1126.978260,-111005.9285,0.,0.,0.,62934.44585,0.,0.,-22980.41043,0.,3651.826941,216661.3874,0.,0.,-64595.94539,0.,9738.205176,-192267.9080,0.,23666.54347,63504.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,1.354006401,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.581138830,1.369306394,1.178511302,1.045825033,1.296148140,1.122497216,.9660917834,1.095445115,.9486832980,.9476070828,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.936491673,1.677050983,1.443375673,1.280868846,1.673320053,1.449137675,1.247219129,1.449137675,1.254990040,1.271348312,1.581138830,1.369306394,1.178511302,1.045825033,1.673320053,1.449137675,1.247219129,1.527525232,1.322875656,1.373213125,1.296148140,1.122497216,.9660917834,1.449137675,1.254990040,1.373213125,1.095445115,.9486832980,1.271348312,.9476070828},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,2.010178183,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.122497216,1.296148140,1.254990040,.9486832980,1.095445115,.8206518070,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.677050983,1.936491673,1.875000000,1.774823935,1.449137675,1.673320053,1.620185174,1.254990040,1.449137675,1.101019935,1.369306394,1.581138830,1.530931089,1.449137675,1.449137675,1.673320053,1.620185174,1.322875656,1.527525232,1.189237451,1.122497216,1.296148140,1.254990040,1.254990040,1.449137675,1.189237451,.9486832980,1.095445115,1.101019935,.8206518070},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,2.270737765,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.178511302,1.530931089,1.581138830,1.559023911,.9660917834,1.254990040,1.296148140,.8164965809,1.060660172,.7063046176,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.443375673,1.875000000,1.936491673,1.909406540,1.247219129,1.620185175,1.673320053,1.080123450,1.403121520,.9476070828,1.178511302,1.530931089,1.581138830,1.559023911,1.247219129,1.620185175,1.673320053,1.138550085,1.479019946,1.023532632,.9660917834,1.254990040,1.296148140,1.080123450,1.403121520,1.023532632,.8164965809,1.060660172,.9476070828,.7063046176},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,2.388242806,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.045825033,1.449137675,1.559023911,1.581138830,.8573214100,1.187939392,1.278019301,.7245688375,1.003992032,.6267831704,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.280868845,1.774823935,1.909406540,1.936491673,1.106797181,1.533623161,1.649915823,.9585144758,1.328156617,.8409178659,1.045825033,1.449137675,1.559023911,1.581138830,1.106797181,1.533623161,1.649915823,1.010362971,1.400000000,.9082951065,.8573214100,1.187939392,1.278019301,.9585144758,1.328156617,.9082951065,.7245688375,1.003992032,.8409178659,.6267831704},
			{1.469693846,2.121320343,2.347382390,2.430277762,2.449489743,2.437211522,1.200000000,1.732050808,1.916629694,1.984313483,2.,.9486832980,1.369306394,1.515228816,1.568737550,.7776888840,1.122497216,1.242118007,.6572670691,.9486832980,.5685642496,1.200000000,1.732050808,1.916629694,1.984313483,2.,1.161895004,1.677050983,1.855768722,1.921303269,1.003992032,1.449137675,1.603567452,.8694826050,1.254990040,.7628089876,.9486832980,1.369306394,1.515228816,1.568737550,1.003992032,1.449137675,1.603567452,.9165151392,1.322875656,.8239278748,.7776888840,1.122497216,1.242118007,.8694826050,1.254990040,.8239278748,.6572670691,.9486832980,.7628089876,.5685642496},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.748014747,3.872983346,3.354101966,2.886751347,2.561737690,2.323790008,3.674234614,3.181980514,2.738612787,2.430277762,3.346640106,2.898275348,2.494438258,3.030457633,2.624453295,2.752549839,1.936491673,1.677050983,1.443375673,1.280868845,1.161895004,3.,2.598076212,2.236067977,1.984313483,3.240370349,2.806243040,2.415229457,3.207134903,2.777460299,3.077444275,1.224744872,1.060660172,.9128709292,.8100925872,2.160246899,1.870828694,1.610152971,2.535462764,2.195775163,2.659215782,.8366600264,.7245688370,.6236095646,1.603567452,1.388730150,1.994411836,.6060915266,.5248906592,1.230977710,.4587583064},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,2.595128875,3.354101966,3.872983346,3.750000000,3.549647869,3.354101966,3.181980514,3.674234614,3.557562366,3.367491648,2.898275350,3.346640106,3.240370350,2.624453295,3.030457633,2.383778084,1.677050983,1.936491673,1.875000000,1.774823935,1.677050983,2.598076212,3.,2.904737510,2.749545418,2.806243040,3.240370350,3.137475099,2.777460300,3.207134903,2.665144921,1.060660172,1.224744872,1.185854122,1.122497216,1.870828694,2.160246900,2.091650066,2.195775165,2.535462764,2.302948421,.7245688375,.8366600264,.8100925875,1.388730150,1.603567452,1.727211316,.5248906592,.6060915266,1.066057968,.3972963474},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.931509850,2.886751347,3.750000000,3.872983346,3.818813079,3.711537446,2.738612788,3.557562368,3.674234614,3.622844188,2.494438258,3.240370349,3.346640106,2.258769757,2.934227987,2.051629517,1.443375673,1.875000000,1.936491673,1.909406540,1.855768722,2.236067977,2.904737510,3.,2.958039892,2.415229459,3.137475099,3.240370349,2.390457218,3.105295018,2.293791533,.9128709293,1.185854122,1.224744872,1.207614729,1.610152972,2.091650066,2.160246899,1.889822365,2.454951265,1.982062418,.6236095646,.8100925872,.8366600265,1.195228610,1.552647508,1.486546813,.4517539516,.5868455971,.9175166129,.3419382528},
			{2.091650066,2.898275350,3.118047822,3.162277659,3.137475099,3.083208205,2.561737692,3.549647870,3.818813080,3.872983346,3.842606539,2.430277762,3.367491647,3.622844188,3.674234614,2.213594361,3.067246321,3.299831644,2.004459314,2.777460300,1.820640586,1.280868846,1.774823935,1.909406540,1.936491673,1.921303269,1.984313483,2.749545418,2.958039892,3.,2.143303525,2.969848480,3.195048251,2.121320343,2.939387692,2.035538055,.8100925874,1.122497216,1.207614729,1.224744872,1.428869017,1.979898987,2.130032168,1.677050983,2.323790008,1.758905910,.5533985903,.7668115804,.8249579111,1.060660172,1.469693846,1.319179432,.4008918626,.5554920599,.8142152223,.3034400977},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,5.019960159,4.347413025,3.741657387,3.320391543,3.011976095,5.291502622,4.582575695,3.944053189,3.500000000,5.163977793,4.472135954,3.849001795,4.909902528,4.252100321,4.624812309,1.673320053,1.449137675,1.247219129,1.106797181,1.003992032,3.240370349,2.806243040,2.415229459,2.143303525,4.,3.464101616,2.981423969,4.330127020,3.750000000,4.432026303,.8819171036,.7637626160,.6573421980,.5833333333,2.,1.732050808,1.490711985,2.738612788,2.371708245,3.191423693,.5163977793,.4472135954,.3849001795,1.299038106,1.125000000,1.914854215,.3273268354,.2834733546,.8864052607,.2202291575},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,4.347413025,5.019960159,4.860555525,4.600869484,4.347413025,4.582575696,5.291502622,5.123475385,4.849742262,4.472135954,5.163977793,5.,4.252100321,4.909902529,4.005204947,1.449137675,1.673320053,1.620185175,1.533623161,1.449137675,2.806243040,3.240370350,3.137475099,2.969848480,3.464101616,4.,3.872983346,3.750000000,4.330127020,3.838247368,.7637626162,.8819171036,.8539125641,.8082903771,1.732050808,2.,1.936491673,2.371708245,2.738612788,2.763853992,.4472135954,.5163977793,.5000000000,1.125000000,1.299038106,1.658312395,.2834733546,.3273268354,.7676494735,.1907240451},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,3.741657387,4.860555524,5.019960158,4.949747467,4.810702356,3.944053188,5.123475380,5.291502622,5.217491946,3.849001795,5.,5.163977793,3.659625271,4.753992682,3.447131570,1.247219129,1.620185174,1.673320053,1.649915823,1.603567452,2.415229457,3.137475099,3.240370349,3.195048251,2.981423969,3.872983346,4.,3.227486122,4.192627457,3.303437362,.6573421979,.8539125633,.8819171036,.8695819911,1.490711985,1.936491673,2.,2.041241452,2.651650429,2.378746774,.3849001795,.5000000000,.5163977793,.9682458365,1.257788237,1.427248064,.2439750181,.3169328455,.6606874726,.1641491224},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,5.796550700,5.019960159,4.320493800,3.834057903,3.477930420,6.546536711,5.669467095,4.879500367,4.330127020,6.708203931,5.809475019,5.,6.614378278,5.728219620,6.408327913,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,3.207134903,2.777460300,2.390457218,2.121320343,4.330127020,3.750000000,3.227486122,5.,4.330127020,5.373546315,.6546536711,.5669467095,.4879500367,.4330127020,1.732050808,1.500000000,1.290994449,2.635231383,2.282177324,3.316624790,.3354101966,.2904737510,.2500000000,1.,.8660254040,1.658312395,.1889822365,.1636634177,.6141195788,.1144344270},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,5.019960159,5.796550700,5.612486080,5.312626469,5.019960159,5.669467095,6.546536709,6.338656908,6.,5.809475019,6.708203931,6.495190530,5.728219619,6.614378278,5.549774770,1.254990040,1.449137675,1.403121520,1.328156617,1.254990040,2.777460299,3.207134903,3.105295018,2.939387692,3.750000000,4.330127020,4.192627457,4.330127020,5.,4.653627619,.5669467095,.6546536709,.6338656908,.6000000000,1.500000000,1.732050808,1.677050983,2.282177323,2.635231383,2.872281324,.2904737510,.3354101966,.3247595265,.8660254040,1.,1.436140662,.1636634177,.1889822365,.5318431565,.9910312088e-1},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,6.356741560,5.505099678,4.738035416,4.204589330,3.814044936,7.538168182,6.528245145,5.618618828,4.986029592,8.010409897,6.937218462,5.970607017,8.124038403,7.035623637,8.049844717,1.271348312,1.101019935,.9476070828,.8409178659,.7628089876,3.077444275,2.665144921,2.293791533,2.035538055,4.432026303,3.838247368,3.303437362,5.373546315,4.653627619,6.,.5025445453,.4352163430,.3745745885,.3324019728,1.477342101,1.279415789,1.101145788,2.427520781,2.102294664,3.240370350,.2288688541,.1982062418,.1705887719,.7676494735,.6648039455,1.388730150,.1160576915,.1005089091,.4285714286,.6388765648e-1},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.427248064,1.581138830,1.369306394,1.178511302,1.045825033,.9486832980,1.,.8660254040,.7453559923,.6614378278,.6831300514,.5916079782,.5091750773,.4948716594,.4285714286,.3745745885,3.162277660,2.738612788,2.357022604,2.091650066,1.897366596,2.449489743,2.121320343,1.825741858,1.620185175,1.763834207,1.527525232,1.314684396,1.309307342,1.133893419,1.005089091,3.,2.598076212,2.236067977,1.984313483,2.645751311,2.291287848,1.972026594,2.070196679,1.792842914,1.628430445,2.732520206,2.366431913,2.036700309,2.618614683,2.267786838,2.171240595,2.474358298,2.142857143,2.512722727,2.247447532},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,1.369306394,1.581138830,1.530931089,1.449137675,1.369306394,.8660254040,1.,.9682458365,.9165151390,.5916079782,.6831300507,.6614378278,.4285714286,.4948716594,.3243911095,2.738612788,3.162277660,3.061862179,2.898275350,2.738612788,2.121320343,2.449489743,2.371708245,2.244994432,1.527525232,1.763834207,1.707825127,1.133893419,1.309307341,.8704326857,2.598076212,3.,2.904737510,2.749545417,2.291287848,2.645751311,2.561737690,1.792842914,2.070196679,1.410262133,2.366431913,2.732520203,2.645751311,2.267786838,2.618614683,1.880349512,2.142857143,2.474358298,2.176081715,1.946346656},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.393567770,1.178511302,1.530931089,1.581138830,1.559023911,1.515228816,.7453559923,.9682458365,1.,.9860132974,.5091750773,.6614378278,.6831300514,.3688555568,.4791574237,.2791914141,2.357022604,3.061862179,3.162277659,3.118047822,3.030457633,1.825741858,2.371708244,2.449489743,2.415229459,1.314684396,1.707825129,1.763834207,.9759000734,1.267731382,.7491491772,2.236067977,2.904737510,3.,2.958039892,1.972026594,2.561737692,2.645751311,1.543033500,2.004459314,1.213760391,2.036700309,2.645751311,2.732520206,1.951800147,2.535462764,1.618347188,1.844277784,2.395787117,1.872872943,1.675148485},
			{1.707825129,2.366431913,2.545875386,2.581988897,2.561737692,2.517428959,1.045825033,1.449137675,1.559023911,1.581138830,1.568737550,.6614378278,.9165151390,.9860132974,1.,.4518480572,.6260990336,.6735753142,.3273268354,.4535573675,.2477578022,2.091650066,2.898275350,3.118047822,3.162277660,3.137475099,1.620185174,2.244994432,2.415229459,2.449489743,1.166666667,1.616580754,1.739163982,.8660254040,1.200000000,.6648039455,1.984313483,2.749545417,2.958039892,3.,1.750000000,2.424871131,2.608745974,1.369306394,1.897366596,1.077105496,1.807392228,2.504396134,2.694301258,1.732050808,2.400000000,1.436140662,1.636634177,2.267786838,1.662009863,1.486546813},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,2.645751311,2.291287848,1.972026594,1.750000000,2.065591118,1.788854382,1.539600718,1.636634177,1.417366774,1.321374945,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,4.320493799,3.741657387,3.220305944,2.857738034,4.,3.464101616,2.981423969,3.464101616,3.,2.954684202,2.645751311,2.291287848,1.972026594,1.750000000,4.,3.464101616,2.981423969,4.107919181,3.557562368,3.829708432,2.065591118,1.788854382,1.539600718,3.464101616,3.,3.829708432,1.636634177,1.417366774,2.954684202,1.321374945},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,2.291287848,2.645751311,2.561737692,2.424871131,1.788854382,2.065591118,2.,1.417366774,1.636634177,1.144344270,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,3.741657387,4.320493800,4.183300132,3.959797974,3.464101616,4.,3.872983346,3.,3.464101616,2.558831578,2.291287848,2.645751311,2.561737692,2.424871131,3.464101616,4.,3.872983346,3.557562368,4.107919182,3.316624790,1.788854382,2.065591118,2.,3.,3.464101616,3.316624790,1.417366774,1.636634177,2.558831578,1.144344270},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,1.972026594,2.561737690,2.645751311,2.608745974,1.539600718,2.,2.065591118,1.219875090,1.584664228,.9848947341,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,3.220305942,4.183300132,4.320493799,4.260064335,2.981423969,3.872983346,4.,2.581988897,3.354101966,2.202291575,1.972026594,2.561737690,2.645751311,2.608745974,2.981423969,3.872983346,4.,3.061862179,3.977475643,2.854496129,1.539600718,2.,2.065591118,2.581988897,3.354101966,2.854496129,1.219875090,1.584664228,2.202291575,.9848947341},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,2.068278942,4.582575696,3.968626966,3.415650256,3.031088914,2.749545418,4.140393357,3.585685828,3.086067000,2.738612788,3.535533905,3.061862179,2.635231383,2.988071523,2.587745849,2.533114026,3.055050464,2.645751311,2.277100171,2.020725943,1.833030278,5.070925527,4.391550330,3.779644731,3.354101966,5.477225575,4.743416490,4.082482906,5.270462768,4.564354646,4.855041564,2.070196679,1.792842914,1.543033500,1.369306394,4.107919181,3.557562368,3.061862179,5.,4.330127020,5.244044240,1.414213562,1.224744872,1.054092553,3.162277660,2.738612788,4.195235392,.9960238407,.8625819493,2.427520781,.7237468643},
			{3.240370348,3.741657386,3.622844188,3.429285640,3.240370348,3.070597895,3.968626966,4.582575695,4.437059836,4.200000000,3.968626966,3.585685828,4.140393357,4.008918626,3.794733192,3.061862179,3.535533905,3.423265985,2.587745849,2.988071523,2.193741096,2.645751311,3.055050463,2.958039891,2.800000000,2.645751311,4.391550326,5.070925527,4.909902528,4.647580015,4.743416490,5.477225576,5.303300858,4.564354646,5.270462768,4.204589330,1.792842914,2.070196679,2.004459314,1.897366596,3.557562368,4.107919182,3.977475643,4.330127020,5.,4.541475532,1.224744872,1.414213562,1.369306394,2.738612788,3.162277660,3.633180426,.8625819493,.9960238407,2.102294664,.6267831704},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,5.492852500,4.756949802,4.094130528,3.633180426,3.295711500,5.428101485,4.700873779,4.045867967,3.590351654,4.944132325,4.281744193,3.685138655,4.387482194,3.799671040,3.864367134,2.746426250,2.378474901,2.047065264,1.816590213,1.647855750,5.318431565,4.605896844,3.964124837,3.517811820,6.382847385,5.527707984,4.757493549,6.633249580,5.744562647,6.480740700,1.628430445,1.410262133,1.213760391,1.077105496,3.829708432,3.316624790,2.854496129,5.244044240,4.541475532,6.,.9888264647,.8563488387,.7370277310,2.653299832,2.297825059,4.,.6267831705,.5428101485,1.851640200,.4140393357},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.462494065,1.296148140,1.122497216,.9660917834,.8573214100,.7776888840,.6831300514,.5916079782,.5091750773,.4518480572,.4000000000,.3464101616,.2981423969,.2535462764,.2195775165,.1705887719,3.888444420,3.367491648,2.898275350,2.571964230,2.333066652,2.509980079,2.173706512,1.870828694,1.660195771,1.549193338,1.341640786,1.154700539,1.006230590,.8714212528,.6866065626,4.098780308,3.549647869,3.055050464,2.711088342,3.098386677,2.683281572,2.309401077,2.121320343,1.837117307,1.483239697,4.,3.464101616,2.981423969,3.354101966,2.904737510,2.472066162,3.803194146,3.293662747,3.433032812,3.582364211},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,2.171240594,1.122497216,1.296148140,1.254990040,1.187939392,1.122497216,.5916079782,.6831300507,.6614378278,.6260990336,.3464101616,.4000000000,.3872983346,.2195775163,.2535462764,.1477342101,3.367491648,3.888444419,3.764970119,3.563818176,3.367491648,2.173706511,2.509980079,2.430277762,2.300434741,1.341640786,1.549193338,1.500000000,.8714212528,1.006230590,.5946187254,3.549647869,4.098780304,3.968626966,3.756594201,2.683281572,3.098386677,3.,1.837117307,2.121320343,1.284523258,3.464101616,4.,3.872983346,2.904737510,3.354101966,2.140872097,3.293662744,3.803194146,2.973093626,3.102418412},
			{1.972026594,2.561737692,2.645751311,2.608745974,2.535462764,2.452677109,.9660917834,1.254990040,1.296148140,1.278019301,1.242118007,.5091750773,.6614378278,.6831300514,.6735753142,.2981423969,.3872983346,.4000000000,.1889822365,.2454951266,.1271493634,2.898275350,3.764970119,3.888444420,3.834057903,3.726354021,1.870828694,2.430277762,2.509980080,2.474873734,1.154700539,1.500000000,1.549193338,.7500000000,.9742785795,.5117663157,3.055050464,3.968626966,4.098780308,4.041451885,2.309401077,3.,3.098386677,1.581138830,2.053959591,1.105541597,2.981423969,3.872983346,4.,2.500000000,3.247595265,1.842569328,2.834733546,3.682426899,2.558831578,2.670136631},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,2.898275350,2.509980080,2.160246900,1.917028952,1.738965210,1.963961013,1.700840129,1.463850110,1.299038106,1.341640786,1.161895004,1.,.9449111824,.8183170887,.6866065625,4.347413025,3.764970119,3.240370350,2.875543427,2.608447815,4.810702356,4.166190450,3.585685828,3.181980514,3.897114318,3.375000000,2.904737510,3.,2.598076212,2.302948420,3.927922025,3.401680258,2.927700220,2.598076212,5.196152424,4.500000000,3.872983346,4.743416490,4.107919182,3.979949748,3.354101966,2.904737510,2.500000000,5.,4.330127020,4.974937185,2.834733546,2.454951266,4.605896842,2.403122968},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,2.509980080,2.898275350,2.806243040,2.656313234,2.509980080,1.700840129,1.963961012,1.901597073,1.800000000,1.161895004,1.341640786,1.299038106,.8183170885,.9449111824,.5946187254,3.764970119,4.347413025,4.209364560,3.984469852,3.764970119,4.166190449,4.810702356,4.657942526,4.409081537,3.375000000,3.897114318,3.773364711,2.598076212,3.,1.994411836,3.401680258,3.927922024,3.803194146,3.600000000,4.500000000,5.196152424,5.031152948,4.107919181,4.743416490,3.446737588,2.904737510,3.354101966,3.247595265,4.330127020,5.,4.308421985,2.454951265,2.834733546,3.988823673,2.081165539},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,3.256860892,2.820524267,2.427520781,2.154210993,2.472066162,2.140872097,1.842569328,1.880349512,1.628430445,1.449137675,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,5.983235509,5.181633950,4.459640440,3.957538298,5.744562647,4.974937185,4.281744193,4.974937185,4.308421985,4.166190450,3.256860892,2.820524267,2.427520781,2.154210993,5.744562647,4.974937185,4.281744193,6.292853088,5.449770639,6.,2.472066162,2.140872097,1.842569328,4.974937185,4.308421985,6.,1.880349512,1.628430445,4.166190450,1.449137675},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.483239697,1.095445115,.9486832980,.8164965809,.7245688375,.6572670691,.4948716594,.4285714286,.3688555568,.3273268354,.2535462764,.2195775163,.1889822365,.1428571429,.1237179149,.8650429586e-1,4.381780461,3.794733192,3.265986323,2.898275350,2.629068276,2.424366106,2.099562637,1.807015806,1.603567451,1.309307341,1.133893419,.9759000724,.7559289460,.6546536709,.4642307659,4.948716594,4.285714286,3.688555568,3.273268354,3.273268354,2.834733546,2.439750181,1.992047682,1.725163898,1.253566341,5.070925527,4.391550326,3.779644731,3.779644731,3.273268354,2.507132682,5.,4.330127020,4.062019202,4.844240569},
			{2.323790008,2.683281572,2.598076212,2.459268185,2.323790008,2.202039870,.9486832980,1.095445115,1.060660172,1.003992032,.9486832980,.4285714286,.4948716594,.4791574237,.4535573675,.2195775165,.2535462764,.2454951266,.1237179149,.1428571429,.7491491772e-1,3.794733192,4.381780460,4.242640686,4.015968127,3.794733192,2.099562637,2.424366106,2.347382389,2.221968240,1.133893419,1.309307342,1.267731382,.6546536711,.7559289460,.4020356364,4.285714286,4.948716594,4.791574237,4.535573675,2.834733546,3.273268354,3.169328455,1.725163898,1.992047682,1.085620297,4.391550330,5.070925527,4.909902529,3.273268354,3.779644731,2.171240595,4.330127020,5.,3.517811818,4.195235392},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,2.542696624,2.202039870,1.895214166,1.681835732,1.525617974,1.507633636,1.305649029,1.123723765,.9972059183,.9154754167,.7928249671,.6823550877,.5802884574,.5025445453,.3833259389,5.085393248,4.404079742,3.790428332,3.363671463,3.051235949,4.923910840,4.264231874,3.670066451,3.256860890,3.545621043,3.070597895,2.642749891,2.456478316,2.127372625,1.714285714,5.025445453,4.352163430,3.745745885,3.324019728,5.909368405,5.117663157,4.404583150,4.855041564,4.204589330,3.703280400,4.577377083,3.964124837,3.411775437,6.141195788,5.318431565,5.554920600,4.062019202,3.517811818,6.,3.577708763},
			{2.708012801,2.345207879,2.018433570,1.791182105,1.624807681,1.496910398,.9476070828,.8206518070,.7063046176,.6267831704,.5685642496,.3745745885,.3243911095,.2791914141,.2477578022,.1705887719,.1477342101,.1271493634,.8650429586e-1,.7491491772e-1,.4761904762e-1,4.738035416,4.103259034,3.531523088,3.133915852,2.842821248,2.293791533,1.986481737,1.709691264,1.517200488,1.101145788,.9536202253,.8207456118,.5721721350,.4955156045,.3194382825,5.618618828,4.865866643,4.187871213,3.716367033,3.303437362,2.860860676,2.462236835,1.809367160,1.566957926,1.035098339,5.970607017,5.170697355,4.450227718,4.005204947,3.468609231,2.415229459,6.055300709,5.244044240,4.472135954,6.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,1.354006401,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.581138830,1.369306394,1.178511302,1.045825033,1.296148140,1.122497216,.9660917834,1.095445115,.9486832980,.9476070828,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.936491673,1.677050983,1.443375673,1.280868846,1.673320053,1.449137675,1.247219129,1.449137675,1.254990040,1.271348312,1.581138830,1.369306394,1.178511302,1.045825033,1.673320053,1.449137675,1.247219129,1.527525232,1.322875656,1.373213125,1.296148140,1.122497216,.9660917834,1.449137675,1.254990040,1.373213125,1.095445115,.9486832980,1.271348312,.9476070828},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,2.010178183,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.122497216,1.296148140,1.254990040,.9486832980,1.095445115,.8206518070,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.677050983,1.936491673,1.875000000,1.774823935,1.449137675,1.673320053,1.620185174,1.254990040,1.449137675,1.101019935,1.369306394,1.581138830,1.530931089,1.449137675,1.449137675,1.673320053,1.620185174,1.322875656,1.527525232,1.189237451,1.122497216,1.296148140,1.254990040,1.254990040,1.449137675,1.189237451,.9486832980,1.095445115,1.101019935,.8206518070},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,2.270737765,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.178511302,1.530931089,1.581138830,1.559023911,.9660917834,1.254990040,1.296148140,.8164965809,1.060660172,.7063046176,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.443375673,1.875000000,1.936491673,1.909406540,1.247219129,1.620185175,1.673320053,1.080123450,1.403121520,.9476070828,1.178511302,1.530931089,1.581138830,1.559023911,1.247219129,1.620185175,1.673320053,1.138550085,1.479019946,1.023532632,.9660917834,1.254990040,1.296148140,1.080123450,1.403121520,1.023532632,.8164965809,1.060660172,.9476070828,.7063046176},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,2.388242806,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.045825033,1.449137675,1.559023911,1.581138830,.8573214100,1.187939392,1.278019301,.7245688375,1.003992032,.6267831704,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.280868845,1.774823935,1.909406540,1.936491673,1.106797181,1.533623161,1.649915823,.9585144758,1.328156617,.8409178659,1.045825033,1.449137675,1.559023911,1.581138830,1.106797181,1.533623161,1.649915823,1.010362971,1.400000000,.9082951065,.8573214100,1.187939392,1.278019301,.9585144758,1.328156617,.9082951065,.7245688375,1.003992032,.8409178659,.6267831704},
			{1.469693846,2.121320343,2.347382390,2.430277762,2.449489743,2.437211522,1.200000000,1.732050808,1.916629694,1.984313483,2.,.9486832980,1.369306394,1.515228816,1.568737550,.7776888840,1.122497216,1.242118007,.6572670691,.9486832980,.5685642496,1.200000000,1.732050808,1.916629694,1.984313483,2.,1.161895004,1.677050983,1.855768722,1.921303269,1.003992032,1.449137675,1.603567452,.8694826050,1.254990040,.7628089876,.9486832980,1.369306394,1.515228816,1.568737550,1.003992032,1.449137675,1.603567452,.9165151392,1.322875656,.8239278748,.7776888840,1.122497216,1.242118007,.8694826050,1.254990040,.8239278748,.6572670691,.9486832980,.7628089876,.5685642496},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.748014747,3.872983346,3.354101966,2.886751347,2.561737690,2.323790008,3.674234614,3.181980514,2.738612787,2.430277762,3.346640106,2.898275348,2.494438258,3.030457633,2.624453295,2.752549839,1.936491673,1.677050983,1.443375673,1.280868845,1.161895004,3.,2.598076212,2.236067977,1.984313483,3.240370349,2.806243040,2.415229457,3.207134903,2.777460299,3.077444275,1.224744872,1.060660172,.9128709292,.8100925872,2.160246899,1.870828694,1.610152971,2.535462764,2.195775163,2.659215782,.8366600264,.7245688370,.6236095646,1.603567452,1.388730150,1.994411836,.6060915266,.5248906592,1.230977710,.4587583064},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,2.595128875,3.354101966,3.872983346,3.750000000,3.549647869,3.354101966,3.181980514,3.674234614,3.557562366,3.367491648,2.898275350,3.346640106,3.240370350,2.624453295,3.030457633,2.383778084,1.677050983,1.936491673,1.875000000,1.774823935,1.677050983,2.598076212,3.,2.904737510,2.749545418,2.806243040,3.240370350,3.137475099,2.777460300,3.207134903,2.665144921,1.060660172,1.224744872,1.185854122,1.122497216,1.870828694,2.160246900,2.091650066,2.195775165,2.535462764,2.302948421,.7245688375,.8366600264,.8100925875,1.388730150,1.603567452,1.727211316,.5248906592,.6060915266,1.066057968,.3972963474},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.931509850,2.886751347,3.750000000,3.872983346,3.818813079,3.711537446,2.738612788,3.557562368,3.674234614,3.622844188,2.494438258,3.240370349,3.346640106,2.258769757,2.934227987,2.051629517,1.443375673,1.875000000,1.936491673,1.909406540,1.855768722,2.236067977,2.904737510,3.,2.958039892,2.415229459,3.137475099,3.240370349,2.390457218,3.105295018,2.293791533,.9128709293,1.185854122,1.224744872,1.207614729,1.610152972,2.091650066,2.160246899,1.889822365,2.454951265,1.982062418,.6236095646,.8100925872,.8366600265,1.195228610,1.552647508,1.486546813,.4517539516,.5868455971,.9175166129,.3419382528},
			{2.091650066,2.898275350,3.118047822,3.162277659,3.137475099,3.083208205,2.561737692,3.549647870,3.818813080,3.872983346,3.842606539,2.430277762,3.367491647,3.622844188,3.674234614,2.213594361,3.067246321,3.299831644,2.004459314,2.777460300,1.820640586,1.280868846,1.774823935,1.909406540,1.936491673,1.921303269,1.984313483,2.749545418,2.958039892,3.,2.143303525,2.969848480,3.195048251,2.121320343,2.939387692,2.035538055,.8100925874,1.122497216,1.207614729,1.224744872,1.428869017,1.979898987,2.130032168,1.677050983,2.323790008,1.758905910,.5533985903,.7668115804,.8249579111,1.060660172,1.469693846,1.319179432,.4008918626,.5554920599,.8142152223,.3034400977},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,5.019960159,4.347413025,3.741657387,3.320391543,3.011976095,5.291502622,4.582575695,3.944053189,3.500000000,5.163977793,4.472135954,3.849001795,4.909902528,4.252100321,4.624812309,1.673320053,1.449137675,1.247219129,1.106797181,1.003992032,3.240370349,2.806243040,2.415229459,2.143303525,4.,3.464101616,2.981423969,4.330127020,3.750000000,4.432026303,.8819171036,.7637626160,.6573421980,.5833333333,2.,1.732050808,1.490711985,2.738612788,2.371708245,3.191423693,.5163977793,.4472135954,.3849001795,1.299038106,1.125000000,1.914854215,.3273268354,.2834733546,.8864052607,.2202291575},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,4.347413025,5.019960159,4.860555525,4.600869484,4.347413025,4.582575696,5.291502622,5.123475385,4.849742262,4.472135954,5.163977793,5.,4.252100321,4.909902529,4.005204947,1.449137675,1.673320053,1.620185175,1.533623161,1.449137675,2.806243040,3.240370350,3.137475099,2.969848480,3.464101616,4.,3.872983346,3.750000000,4.330127020,3.838247368,.7637626162,.8819171036,.8539125641,.8082903771,1.732050808,2.,1.936491673,2.371708245,2.738612788,2.763853992,.4472135954,.5163977793,.5000000000,1.125000000,1.299038106,1.658312395,.2834733546,.3273268354,.7676494735,.1907240451},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,3.741657387,4.860555524,5.019960158,4.949747467,4.810702356,3.944053188,5.123475380,5.291502622,5.217491946,3.849001795,5.,5.163977793,3.659625271,4.753992682,3.447131570,1.247219129,1.620185174,1.673320053,1.649915823,1.603567452,2.415229457,3.137475099,3.240370349,3.195048251,2.981423969,3.872983346,4.,3.227486122,4.192627457,3.303437362,.6573421979,.8539125633,.8819171036,.8695819911,1.490711985,1.936491673,2.,2.041241452,2.651650429,2.378746774,.3849001795,.5000000000,.5163977793,.9682458365,1.257788237,1.427248064,.2439750181,.3169328455,.6606874726,.1641491224},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,5.796550700,5.019960159,4.320493800,3.834057903,3.477930420,6.546536711,5.669467095,4.879500367,4.330127020,6.708203931,5.809475019,5.,6.614378278,5.728219620,6.408327913,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,3.207134903,2.777460300,2.390457218,2.121320343,4.330127020,3.750000000,3.227486122,5.,4.330127020,5.373546315,.6546536711,.5669467095,.4879500367,.4330127020,1.732050808,1.500000000,1.290994449,2.635231383,2.282177324,3.316624790,.3354101966,.2904737510,.2500000000,1.,.8660254040,1.658312395,.1889822365,.1636634177,.6141195788,.1144344270},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,5.019960159,5.796550700,5.612486080,5.312626469,5.019960159,5.669467095,6.546536709,6.338656908,6.,5.809475019,6.708203931,6.495190530,5.728219619,6.614378278,5.549774770,1.254990040,1.449137675,1.403121520,1.328156617,1.254990040,2.777460299,3.207134903,3.105295018,2.939387692,3.750000000,4.330127020,4.192627457,4.330127020,5.,4.653627619,.5669467095,.6546536709,.6338656908,.6000000000,1.500000000,1.732050808,1.677050983,2.282177323,2.635231383,2.872281324,.2904737510,.3354101966,.3247595265,.8660254040,1.,1.436140662,.1636634177,.1889822365,.5318431565,.9910312088e-1},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,6.356741560,5.505099678,4.738035416,4.204589330,3.814044936,7.538168182,6.528245145,5.618618828,4.986029592,8.010409897,6.937218462,5.970607017,8.124038403,7.035623637,8.049844717,1.271348312,1.101019935,.9476070828,.8409178659,.7628089876,3.077444275,2.665144921,2.293791533,2.035538055,4.432026303,3.838247368,3.303437362,5.373546315,4.653627619,6.,.5025445453,.4352163430,.3745745885,.3324019728,1.477342101,1.279415789,1.101145788,2.427520781,2.102294664,3.240370350,.2288688541,.1982062418,.1705887719,.7676494735,.6648039455,1.388730150,.1160576915,.1005089091,.4285714286,.6388765648e-1},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.427248064,1.581138830,1.369306394,1.178511302,1.045825033,.9486832980,1.,.8660254040,.7453559923,.6614378278,.6831300514,.5916079782,.5091750773,.4948716594,.4285714286,.3745745885,3.162277660,2.738612788,2.357022604,2.091650066,1.897366596,2.449489743,2.121320343,1.825741858,1.620185175,1.763834207,1.527525232,1.314684396,1.309307342,1.133893419,1.005089091,3.,2.598076212,2.236067977,1.984313483,2.645751311,2.291287848,1.972026594,2.070196679,1.792842914,1.628430445,2.732520206,2.366431913,2.036700309,2.618614683,2.267786838,2.171240595,2.474358298,2.142857143,2.512722727,2.247447532},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,1.369306394,1.581138830,1.530931089,1.449137675,1.369306394,.8660254040,1.,.9682458365,.9165151390,.5916079782,.6831300507,.6614378278,.4285714286,.4948716594,.3243911095,2.738612788,3.162277660,3.061862179,2.898275350,2.738612788,2.121320343,2.449489743,2.371708245,2.244994432,1.527525232,1.763834207,1.707825127,1.133893419,1.309307341,.8704326857,2.598076212,3.,2.904737510,2.749545417,2.291287848,2.645751311,2.561737690,1.792842914,2.070196679,1.410262133,2.366431913,2.732520203,2.645751311,2.267786838,2.618614683,1.880349512,2.142857143,2.474358298,2.176081715,1.946346656},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.393567770,1.178511302,1.530931089,1.581138830,1.559023911,1.515228816,.7453559923,.9682458365,1.,.9860132974,.5091750773,.6614378278,.6831300514,.3688555568,.4791574237,.2791914141,2.357022604,3.061862179,3.162277659,3.118047822,3.030457633,1.825741858,2.371708244,2.449489743,2.415229459,1.314684396,1.707825129,1.763834207,.9759000734,1.267731382,.7491491772,2.236067977,2.904737510,3.,2.958039892,1.972026594,2.561737692,2.645751311,1.543033500,2.004459314,1.213760391,2.036700309,2.645751311,2.732520206,1.951800147,2.535462764,1.618347188,1.844277784,2.395787117,1.872872943,1.675148485},
			{1.707825129,2.366431913,2.545875386,2.581988897,2.561737692,2.517428959,1.045825033,1.449137675,1.559023911,1.581138830,1.568737550,.6614378278,.9165151390,.9860132974,1.,.4518480572,.6260990336,.6735753142,.3273268354,.4535573675,.2477578022,2.091650066,2.898275350,3.118047822,3.162277660,3.137475099,1.620185174,2.244994432,2.415229459,2.449489743,1.166666667,1.616580754,1.739163982,.8660254040,1.200000000,.6648039455,1.984313483,2.749545417,2.958039892,3.,1.750000000,2.424871131,2.608745974,1.369306394,1.897366596,1.077105496,1.807392228,2.504396134,2.694301258,1.732050808,2.400000000,1.436140662,1.636634177,2.267786838,1.662009863,1.486546813},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,2.645751311,2.291287848,1.972026594,1.750000000,2.065591118,1.788854382,1.539600718,1.636634177,1.417366774,1.321374945,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,4.320493799,3.741657387,3.220305944,2.857738034,4.,3.464101616,2.981423969,3.464101616,3.,2.954684202,2.645751311,2.291287848,1.972026594,1.750000000,4.,3.464101616,2.981423969,4.107919181,3.557562368,3.829708432,2.065591118,1.788854382,1.539600718,3.464101616,3.,3.829708432,1.636634177,1.417366774,2.954684202,1.321374945},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,2.291287848,2.645751311,2.561737692,2.424871131,1.788854382,2.065591118,2.,1.417366774,1.636634177,1.144344270,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,3.741657387,4.320493800,4.183300132,3.959797974,3.464101616,4.,3.872983346,3.,3.464101616,2.558831578,2.291287848,2.645751311,2.561737692,2.424871131,3.464101616,4.,3.872983346,3.557562368,4.107919182,3.316624790,1.788854382,2.065591118,2.,3.,3.464101616,3.316624790,1.417366774,1.636634177,2.558831578,1.144344270},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,1.972026594,2.561737690,2.645751311,2.608745974,1.539600718,2.,2.065591118,1.219875090,1.584664228,.9848947341,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,3.220305942,4.183300132,4.320493799,4.260064335,2.981423969,3.872983346,4.,2.581988897,3.354101966,2.202291575,1.972026594,2.561737690,2.645751311,2.608745974,2.981423969,3.872983346,4.,3.061862179,3.977475643,2.854496129,1.539600718,2.,2.065591118,2.581988897,3.354101966,2.854496129,1.219875090,1.584664228,2.202291575,.9848947341},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,2.068278942,4.582575696,3.968626966,3.415650256,3.031088914,2.749545418,4.140393357,3.585685828,3.086067000,2.738612788,3.535533905,3.061862179,2.635231383,2.988071523,2.587745849,2.533114026,3.055050464,2.645751311,2.277100171,2.020725943,1.833030278,5.070925527,4.391550330,3.779644731,3.354101966,5.477225575,4.743416490,4.082482906,5.270462768,4.564354646,4.855041564,2.070196679,1.792842914,1.543033500,1.369306394,4.107919181,3.557562368,3.061862179,5.,4.330127020,5.244044240,1.414213562,1.224744872,1.054092553,3.162277660,2.738612788,4.195235392,.9960238407,.8625819493,2.427520781,.7237468643},
			{3.240370348,3.741657386,3.622844188,3.429285640,3.240370348,3.070597895,3.968626966,4.582575695,4.437059836,4.200000000,3.968626966,3.585685828,4.140393357,4.008918626,3.794733192,3.061862179,3.535533905,3.423265985,2.587745849,2.988071523,2.193741096,2.645751311,3.055050463,2.958039891,2.800000000,2.645751311,4.391550326,5.070925527,4.909902528,4.647580015,4.743416490,5.477225576,5.303300858,4.564354646,5.270462768,4.204589330,1.792842914,2.070196679,2.004459314,1.897366596,3.557562368,4.107919182,3.977475643,4.330127020,5.,4.541475532,1.224744872,1.414213562,1.369306394,2.738612788,3.162277660,3.633180426,.8625819493,.9960238407,2.102294664,.6267831704},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,5.492852500,4.756949802,4.094130528,3.633180426,3.295711500,5.428101485,4.700873779,4.045867967,3.590351654,4.944132325,4.281744193,3.685138655,4.387482194,3.799671040,3.864367134,2.746426250,2.378474901,2.047065264,1.816590213,1.647855750,5.318431565,4.605896844,3.964124837,3.517811820,6.382847385,5.527707984,4.757493549,6.633249580,5.744562647,6.480740700,1.628430445,1.410262133,1.213760391,1.077105496,3.829708432,3.316624790,2.854496129,5.244044240,4.541475532,6.,.9888264647,.8563488387,.7370277310,2.653299832,2.297825059,4.,.6267831705,.5428101485,1.851640200,.4140393357},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.462494065,1.296148140,1.122497216,.9660917834,.8573214100,.7776888840,.6831300514,.5916079782,.5091750773,.4518480572,.4000000000,.3464101616,.2981423969,.2535462764,.2195775165,.1705887719,3.888444420,3.367491648,2.898275350,2.571964230,2.333066652,2.509980079,2.173706512,1.870828694,1.660195771,1.549193338,1.341640786,1.154700539,1.006230590,.8714212528,.6866065626,4.098780308,3.549647869,3.055050464,2.711088342,3.098386677,2.683281572,2.309401077,2.121320343,1.837117307,1.483239697,4.,3.464101616,2.981423969,3.354101966,2.904737510,2.472066162,3.803194146,3.293662747,3.433032812,3.582364211},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,2.171240594,1.122497216,1.296148140,1.254990040,1.187939392,1.122497216,.5916079782,.6831300507,.6614378278,.6260990336,.3464101616,.4000000000,.3872983346,.2195775163,.2535462764,.1477342101,3.367491648,3.888444419,3.764970119,3.563818176,3.367491648,2.173706511,2.509980079,2.430277762,2.300434741,1.341640786,1.549193338,1.500000000,.8714212528,1.006230590,.5946187254,3.549647869,4.098780304,3.968626966,3.756594201,2.683281572,3.098386677,3.,1.837117307,2.121320343,1.284523258,3.464101616,4.,3.872983346,2.904737510,3.354101966,2.140872097,3.293662744,3.803194146,2.973093626,3.102418412},
			{1.972026594,2.561737692,2.645751311,2.608745974,2.535462764,2.452677109,.9660917834,1.254990040,1.296148140,1.278019301,1.242118007,.5091750773,.6614378278,.6831300514,.6735753142,.2981423969,.3872983346,.4000000000,.1889822365,.2454951266,.1271493634,2.898275350,3.764970119,3.888444420,3.834057903,3.726354021,1.870828694,2.430277762,2.509980080,2.474873734,1.154700539,1.500000000,1.549193338,.7500000000,.9742785795,.5117663157,3.055050464,3.968626966,4.098780308,4.041451885,2.309401077,3.,3.098386677,1.581138830,2.053959591,1.105541597,2.981423969,3.872983346,4.,2.500000000,3.247595265,1.842569328,2.834733546,3.682426899,2.558831578,2.670136631},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,2.898275350,2.509980080,2.160246900,1.917028952,1.738965210,1.963961013,1.700840129,1.463850110,1.299038106,1.341640786,1.161895004,1.,.9449111824,.8183170887,.6866065625,4.347413025,3.764970119,3.240370350,2.875543427,2.608447815,4.810702356,4.166190450,3.585685828,3.181980514,3.897114318,3.375000000,2.904737510,3.,2.598076212,2.302948420,3.927922025,3.401680258,2.927700220,2.598076212,5.196152424,4.500000000,3.872983346,4.743416490,4.107919182,3.979949748,3.354101966,2.904737510,2.500000000,5.,4.330127020,4.974937185,2.834733546,2.454951266,4.605896842,2.403122968},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,2.509980080,2.898275350,2.806243040,2.656313234,2.509980080,1.700840129,1.963961012,1.901597073,1.800000000,1.161895004,1.341640786,1.299038106,.8183170885,.9449111824,.5946187254,3.764970119,4.347413025,4.209364560,3.984469852,3.764970119,4.166190449,4.810702356,4.657942526,4.409081537,3.375000000,3.897114318,3.773364711,2.598076212,3.,1.994411836,3.401680258,3.927922024,3.803194146,3.600000000,4.500000000,5.196152424,5.031152948,4.107919181,4.743416490,3.446737588,2.904737510,3.354101966,3.247595265,4.330127020,5.,4.308421985,2.454951265,2.834733546,3.988823673,2.081165539},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,3.256860892,2.820524267,2.427520781,2.154210993,2.472066162,2.140872097,1.842569328,1.880349512,1.628430445,1.449137675,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,5.983235509,5.181633950,4.459640440,3.957538298,5.744562647,4.974937185,4.281744193,4.974937185,4.308421985,4.166190450,3.256860892,2.820524267,2.427520781,2.154210993,5.744562647,4.974937185,4.281744193,6.292853088,5.449770639,6.,2.472066162,2.140872097,1.842569328,4.974937185,4.308421985,6.,1.880349512,1.628430445,4.166190450,1.449137675},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.483239697,1.095445115,.9486832980,.8164965809,.7245688375,.6572670691,.4948716594,.4285714286,.3688555568,.3273268354,.2535462764,.2195775163,.1889822365,.1428571429,.1237179149,.8650429586e-1,4.381780461,3.794733192,3.265986323,2.898275350,2.629068276,2.424366106,2.099562637,1.807015806,1.603567451,1.309307341,1.133893419,.9759000724,.7559289460,.6546536709,.4642307659,4.948716594,4.285714286,3.688555568,3.273268354,3.273268354,2.834733546,2.439750181,1.992047682,1.725163898,1.253566341,5.070925527,4.391550326,3.779644731,3.779644731,3.273268354,2.507132682,5.,4.330127020,4.062019202,4.844240569},
			{2.323790008,2.683281572,2.598076212,2.459268185,2.323790008,2.202039870,.9486832980,1.095445115,1.060660172,1.003992032,.9486832980,.4285714286,.4948716594,.4791574237,.4535573675,.2195775165,.2535462764,.2454951266,.1237179149,.1428571429,.7491491772e-1,3.794733192,4.381780460,4.242640686,4.015968127,3.794733192,2.099562637,2.424366106,2.347382389,2.221968240,1.133893419,1.309307342,1.267731382,.6546536711,.7559289460,.4020356364,4.285714286,4.948716594,4.791574237,4.535573675,2.834733546,3.273268354,3.169328455,1.725163898,1.992047682,1.085620297,4.391550330,5.070925527,4.909902529,3.273268354,3.779644731,2.171240595,4.330127020,5.,3.517811818,4.195235392},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,2.542696624,2.202039870,1.895214166,1.681835732,1.525617974,1.507633636,1.305649029,1.123723765,.9972059183,.9154754167,.7928249671,.6823550877,.5802884574,.5025445453,.3833259389,5.085393248,4.404079742,3.790428332,3.363671463,3.051235949,4.923910840,4.264231874,3.670066451,3.256860890,3.545621043,3.070597895,2.642749891,2.456478316,2.127372625,1.714285714,5.025445453,4.352163430,3.745745885,3.324019728,5.909368405,5.117663157,4.404583150,4.855041564,4.204589330,3.703280400,4.577377083,3.964124837,3.411775437,6.141195788,5.318431565,5.554920600,4.062019202,3.517811818,6.,3.577708763},
			{2.708012801,2.345207879,2.018433570,1.791182105,1.624807681,1.496910398,.9476070828,.8206518070,.7063046176,.6267831704,.5685642496,.3745745885,.3243911095,.2791914141,.2477578022,.1705887719,.1477342101,.1271493634,.8650429586e-1,.7491491772e-1,.4761904762e-1,4.738035416,4.103259034,3.531523088,3.133915852,2.842821248,2.293791533,1.986481737,1.709691264,1.517200488,1.101145788,.9536202253,.8207456118,.5721721350,.4955156045,.3194382825,5.618618828,4.865866643,4.187871213,3.716367033,3.303437362,2.860860676,2.462236835,1.809367160,1.566957926,1.035098339,5.970607017,5.170697355,4.450227718,4.005204947,3.468609231,2.415229459,6.055300709,5.244044240,4.472135954,6.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,1.354006401,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.581138830,1.369306394,1.178511302,1.045825033,1.296148140,1.122497216,.9660917834,1.095445115,.9486832980,.9476070828,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.936491673,1.677050983,1.443375673,1.280868846,1.673320053,1.449137675,1.247219129,1.449137675,1.254990040,1.271348312,1.581138830,1.369306394,1.178511302,1.045825033,1.673320053,1.449137675,1.247219129,1.527525232,1.322875656,1.373213125,1.296148140,1.122497216,.9660917834,1.449137675,1.254990040,1.373213125,1.095445115,.9486832980,1.271348312,.9476070828},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,2.010178183,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.122497216,1.296148140,1.254990040,.9486832980,1.095445115,.8206518070,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.677050983,1.936491673,1.875000000,1.774823935,1.449137675,1.673320053,1.620185174,1.254990040,1.449137675,1.101019935,1.369306394,1.581138830,1.530931089,1.449137675,1.449137675,1.673320053,1.620185174,1.322875656,1.527525232,1.189237451,1.122497216,1.296148140,1.254990040,1.254990040,1.449137675,1.189237451,.9486832980,1.095445115,1.101019935,.8206518070},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,2.270737765,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.178511302,1.530931089,1.581138830,1.559023911,.9660917834,1.254990040,1.296148140,.8164965809,1.060660172,.7063046176,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.443375673,1.875000000,1.936491673,1.909406540,1.247219129,1.620185175,1.673320053,1.080123450,1.403121520,.9476070828,1.178511302,1.530931089,1.581138830,1.559023911,1.247219129,1.620185175,1.673320053,1.138550085,1.479019946,1.023532632,.9660917834,1.254990040,1.296148140,1.080123450,1.403121520,1.023532632,.8164965809,1.060660172,.9476070828,.7063046176},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,2.388242806,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.045825033,1.449137675,1.559023911,1.581138830,.8573214100,1.187939392,1.278019301,.7245688375,1.003992032,.6267831704,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.280868845,1.774823935,1.909406540,1.936491673,1.106797181,1.533623161,1.649915823,.9585144758,1.328156617,.8409178659,1.045825033,1.449137675,1.559023911,1.581138830,1.106797181,1.533623161,1.649915823,1.010362971,1.400000000,.9082951065,.8573214100,1.187939392,1.278019301,.9585144758,1.328156617,.9082951065,.7245688375,1.003992032,.8409178659,.6267831704},
			{1.469693846,2.121320343,2.347382390,2.430277762,2.449489743,2.437211522,1.200000000,1.732050808,1.916629694,1.984313483,2.,.9486832980,1.369306394,1.515228816,1.568737550,.7776888840,1.122497216,1.242118007,.6572670691,.9486832980,.5685642496,1.200000000,1.732050808,1.916629694,1.984313483,2.,1.161895004,1.677050983,1.855768722,1.921303269,1.003992032,1.449137675,1.603567452,.8694826050,1.254990040,.7628089876,.9486832980,1.369306394,1.515228816,1.568737550,1.003992032,1.449137675,1.603567452,.9165151392,1.322875656,.8239278748,.7776888840,1.122497216,1.242118007,.8694826050,1.254990040,.8239278748,.6572670691,.9486832980,.7628089876,.5685642496},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.427248064,3.162277660,2.738612788,2.357022604,2.091650066,1.897366596,3.,2.598076212,2.236067977,1.984313483,2.732520206,2.366431913,2.036700309,2.474358298,2.142857143,2.247447532,1.581138830,1.369306394,1.178511302,1.045825033,.9486832980,2.449489743,2.121320343,1.825741858,1.620185175,2.645751311,2.291287848,1.972026594,2.618614683,2.267786838,2.512722727,1.,.8660254040,.7453559923,.6614378278,1.763834207,1.527525232,1.314684396,2.070196679,1.792842914,2.171240595,.6831300514,.5916079782,.5091750773,1.309307342,1.133893419,1.628430445,.4948716594,.4285714286,1.005089091,.3745745885},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,2.738612788,3.162277660,3.061862179,2.898275350,2.738612788,2.598076212,3.,2.904737510,2.749545417,2.366431913,2.732520203,2.645751311,2.142857143,2.474358298,1.946346656,1.369306394,1.581138830,1.530931089,1.449137675,1.369306394,2.121320343,2.449489743,2.371708245,2.244994432,2.291287848,2.645751311,2.561737690,2.267786838,2.618614683,2.176081715,.8660254040,1.,.9682458365,.9165151390,1.527525232,1.763834207,1.707825127,1.792842914,2.070196679,1.880349512,.5916079782,.6831300507,.6614378278,1.133893419,1.309307341,1.410262133,.4285714286,.4948716594,.8704326857,.3243911095},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.393567770,2.357022604,3.061862179,3.162277659,3.118047822,3.030457633,2.236067977,2.904737510,3.,2.958039892,2.036700309,2.645751311,2.732520206,1.844277784,2.395787117,1.675148485,1.178511302,1.530931089,1.581138830,1.559023911,1.515228816,1.825741858,2.371708244,2.449489743,2.415229459,1.972026594,2.561737692,2.645751311,1.951800147,2.535462764,1.872872943,.7453559923,.9682458365,1.,.9860132974,1.314684396,1.707825129,1.763834207,1.543033500,2.004459314,1.618347188,.5091750773,.6614378278,.6831300514,.9759000734,1.267731382,1.213760391,.3688555568,.4791574237,.7491491772,.2791914141},
			{1.707825129,2.366431913,2.545875386,2.581988897,2.561737692,2.517428959,2.091650066,2.898275350,3.118047822,3.162277660,3.137475099,1.984313483,2.749545417,2.958039892,3.,1.807392228,2.504396134,2.694301258,1.636634177,2.267786838,1.486546813,1.045825033,1.449137675,1.559023911,1.581138830,1.568737550,1.620185174,2.244994432,2.415229459,2.449489743,1.750000000,2.424871131,2.608745974,1.732050808,2.400000000,1.662009863,.6614378278,.9165151390,.9860132974,1.,1.166666667,1.616580754,1.739163982,1.369306394,1.897366596,1.436140662,.4518480572,.6260990336,.6735753142,.8660254040,1.200000000,1.077105496,.3273268354,.4535573675,.6648039455,.2477578022},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.462494065,3.888444420,3.367491648,2.898275350,2.571964230,2.333066652,4.098780308,3.549647869,3.055050464,2.711088342,4.,3.464101616,2.981423969,3.803194146,3.293662747,3.582364211,1.296148140,1.122497216,.9660917834,.8573214100,.7776888840,2.509980079,2.173706512,1.870828694,1.660195771,3.098386677,2.683281572,2.309401077,3.354101966,2.904737510,3.433032812,.6831300514,.5916079782,.5091750773,.4518480572,1.549193338,1.341640786,1.154700539,2.121320343,1.837117307,2.472066162,.4000000000,.3464101616,.2981423969,1.006230590,.8714212528,1.483239697,.2535462764,.2195775165,.6866065626,.1705887719},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,2.171240594,3.367491648,3.888444419,3.764970119,3.563818176,3.367491648,3.549647869,4.098780304,3.968626966,3.756594201,3.464101616,4.,3.872983346,3.293662744,3.803194146,3.102418412,1.122497216,1.296148140,1.254990040,1.187939392,1.122497216,2.173706511,2.509980079,2.430277762,2.300434741,2.683281572,3.098386677,3.,2.904737510,3.354101966,2.973093626,.5916079782,.6831300507,.6614378278,.6260990336,1.341640786,1.549193338,1.500000000,1.837117307,2.121320343,2.140872097,.3464101616,.4000000000,.3872983346,.8714212528,1.006230590,1.284523258,.2195775163,.2535462764,.5946187254,.1477342101},
			{1.972026594,2.561737692,2.645751311,2.608745974,2.535462764,2.452677109,2.898275350,3.764970119,3.888444420,3.834057903,3.726354021,3.055050464,3.968626966,4.098780308,4.041451885,2.981423969,3.872983346,4.,2.834733546,3.682426899,2.670136631,.9660917834,1.254990040,1.296148140,1.278019301,1.242118007,1.870828694,2.430277762,2.509980080,2.474873734,2.309401077,3.,3.098386677,2.500000000,3.247595265,2.558831578,.5091750773,.6614378278,.6831300514,.6735753142,1.154700539,1.500000000,1.549193338,1.581138830,2.053959591,1.842569328,.2981423969,.3872983346,.4000000000,.7500000000,.9742785795,1.105541597,.1889822365,.2454951266,.5117663157,.1271493634},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.483239697,4.381780461,3.794733192,3.265986323,2.898275350,2.629068276,4.948716594,4.285714286,3.688555568,3.273268354,5.070925527,4.391550326,3.779644731,5.,4.330127020,4.844240569,1.095445115,.9486832980,.8164965809,.7245688375,.6572670691,2.424366106,2.099562637,1.807015806,1.603567451,3.273268354,2.834733546,2.439750181,3.779644731,3.273268354,4.062019202,.4948716594,.4285714286,.3688555568,.3273268354,1.309307341,1.133893419,.9759000724,1.992047682,1.725163898,2.507132682,.2535462764,.2195775163,.1889822365,.7559289460,.6546536709,1.253566341,.1428571429,.1237179149,.4642307659,.8650429586e-1},
			{2.323790008,2.683281572,2.598076212,2.459268185,2.323790008,2.202039870,3.794733192,4.381780460,4.242640686,4.015968127,3.794733192,4.285714286,4.948716594,4.791574237,4.535573675,4.391550330,5.070925527,4.909902529,4.330127020,5.,4.195235392,.9486832980,1.095445115,1.060660172,1.003992032,.9486832980,2.099562637,2.424366106,2.347382389,2.221968240,2.834733546,3.273268354,3.169328455,3.273268354,3.779644731,3.517811818,.4285714286,.4948716594,.4791574237,.4535573675,1.133893419,1.309307342,1.267731382,1.725163898,1.992047682,2.171240595,.2195775165,.2535462764,.2454951266,.6546536711,.7559289460,1.085620297,.1237179149,.1428571429,.4020356364,.7491491772e-1},
			{2.708012801,2.345207879,2.018433570,1.791182105,1.624807681,1.496910398,4.738035416,4.103259034,3.531523088,3.133915852,2.842821248,5.618618828,4.865866643,4.187871213,3.716367033,5.970607017,5.170697355,4.450227718,6.055300709,5.244044240,6.,.9476070828,.8206518070,.7063046176,.6267831704,.5685642496,2.293791533,1.986481737,1.709691264,1.517200488,3.303437362,2.860860676,2.462236835,4.005204947,3.468609231,4.472135954,.3745745885,.3243911095,.2791914141,.2477578022,1.101145788,.9536202253,.8207456118,1.809367160,1.566957926,2.415229459,.1705887719,.1477342101,.1271493634,.5721721350,.4955156045,1.035098339,.8650429586e-1,.7491491772e-1,.3194382825,.4761904762e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.748014747,1.936491673,1.677050983,1.443375673,1.280868845,1.161895004,1.224744872,1.060660172,.9128709292,.8100925872,.8366600264,.7245688370,.6236095646,.6060915266,.5248906592,.4587583064,3.872983346,3.354101966,2.886751347,2.561737690,2.323790008,3.,2.598076212,2.236067977,1.984313483,2.160246899,1.870828694,1.610152971,1.603567452,1.388730150,1.230977710,3.674234614,3.181980514,2.738612787,2.430277762,3.240370349,2.806243040,2.415229457,2.535462764,2.195775163,1.994411836,3.346640106,2.898275348,2.494438258,3.207134903,2.777460299,2.659215782,3.030457633,2.624453295,3.077444275,2.752549839},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,2.595128875,1.677050983,1.936491673,1.875000000,1.774823935,1.677050983,1.060660172,1.224744872,1.185854122,1.122497216,.7245688375,.8366600264,.8100925875,.5248906592,.6060915266,.3972963474,3.354101966,3.872983346,3.750000000,3.549647869,3.354101966,2.598076212,3.,2.904737510,2.749545418,1.870828694,2.160246900,2.091650066,1.388730150,1.603567452,1.066057968,3.181980514,3.674234614,3.557562366,3.367491648,2.806243040,3.240370350,3.137475099,2.195775165,2.535462764,1.727211316,2.898275350,3.346640106,3.240370350,2.777460300,3.207134903,2.302948421,2.624453295,3.030457633,2.665144921,2.383778084},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.931509850,1.443375673,1.875000000,1.936491673,1.909406540,1.855768722,.9128709293,1.185854122,1.224744872,1.207614729,.6236095646,.8100925872,.8366600265,.4517539516,.5868455971,.3419382528,2.886751347,3.750000000,3.872983346,3.818813079,3.711537446,2.236067977,2.904737510,3.,2.958039892,1.610152972,2.091650066,2.160246899,1.195228610,1.552647508,.9175166129,2.738612788,3.557562368,3.674234614,3.622844188,2.415229459,3.137475099,3.240370349,1.889822365,2.454951265,1.486546813,2.494438258,3.240370349,3.346640106,2.390457218,3.105295018,1.982062418,2.258769757,2.934227987,2.293791533,2.051629517},
			{2.091650066,2.898275350,3.118047822,3.162277659,3.137475099,3.083208205,1.280868846,1.774823935,1.909406540,1.936491673,1.921303269,.8100925874,1.122497216,1.207614729,1.224744872,.5533985903,.7668115804,.8249579111,.4008918626,.5554920599,.3034400977,2.561737692,3.549647870,3.818813080,3.872983346,3.842606539,1.984313483,2.749545418,2.958039892,3.,1.428869017,1.979898987,2.130032168,1.060660172,1.469693846,.8142152223,2.430277762,3.367491647,3.622844188,3.674234614,2.143303525,2.969848480,3.195048251,1.677050983,2.323790008,1.319179432,2.213594361,3.067246321,3.299831644,2.121320343,2.939387692,1.758905910,2.004459314,2.777460300,2.035538055,1.820640586},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,2.645751311,2.291287848,1.972026594,1.750000000,2.065591118,1.788854382,1.539600718,1.636634177,1.417366774,1.321374945,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,4.320493799,3.741657387,3.220305944,2.857738034,4.,3.464101616,2.981423969,3.464101616,3.,2.954684202,2.645751311,2.291287848,1.972026594,1.750000000,4.,3.464101616,2.981423969,4.107919181,3.557562368,3.829708432,2.065591118,1.788854382,1.539600718,3.464101616,3.,3.829708432,1.636634177,1.417366774,2.954684202,1.321374945},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,2.291287848,2.645751311,2.561737692,2.424871131,1.788854382,2.065591118,2.,1.417366774,1.636634177,1.144344270,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,3.741657387,4.320493800,4.183300132,3.959797974,3.464101616,4.,3.872983346,3.,3.464101616,2.558831578,2.291287848,2.645751311,2.561737692,2.424871131,3.464101616,4.,3.872983346,3.557562368,4.107919182,3.316624790,1.788854382,2.065591118,2.,3.,3.464101616,3.316624790,1.417366774,1.636634177,2.558831578,1.144344270},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,1.972026594,2.561737690,2.645751311,2.608745974,1.539600718,2.,2.065591118,1.219875090,1.584664228,.9848947341,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,3.220305942,4.183300132,4.320493799,4.260064335,2.981423969,3.872983346,4.,2.581988897,3.354101966,2.202291575,1.972026594,2.561737690,2.645751311,2.608745974,2.981423969,3.872983346,4.,3.061862179,3.977475643,2.854496129,1.539600718,2.,2.065591118,2.581988897,3.354101966,2.854496129,1.219875090,1.584664228,2.202291575,.9848947341},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,4.347413025,3.764970119,3.240370350,2.875543427,2.608447815,3.927922025,3.401680258,2.927700220,2.598076212,3.354101966,2.904737510,2.500000000,2.834733546,2.454951266,2.403122968,2.898275350,2.509980080,2.160246900,1.917028952,1.738965210,4.810702356,4.166190450,3.585685828,3.181980514,5.196152424,4.500000000,3.872983346,5.,4.330127020,4.605896842,1.963961013,1.700840129,1.463850110,1.299038106,3.897114318,3.375000000,2.904737510,4.743416490,4.107919182,4.974937185,1.341640786,1.161895004,1.,3.,2.598076212,3.979949748,.9449111824,.8183170887,2.302948420,.6866065625},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,3.764970119,4.347413025,4.209364560,3.984469852,3.764970119,3.401680258,3.927922024,3.803194146,3.600000000,2.904737510,3.354101966,3.247595265,2.454951265,2.834733546,2.081165539,2.509980080,2.898275350,2.806243040,2.656313234,2.509980080,4.166190449,4.810702356,4.657942526,4.409081537,4.500000000,5.196152424,5.031152948,4.330127020,5.,3.988823673,1.700840129,1.963961012,1.901597073,1.800000000,3.375000000,3.897114318,3.773364711,4.107919181,4.743416490,4.308421985,1.161895004,1.341640786,1.299038106,2.598076212,3.,3.446737588,.8183170885,.9449111824,1.994411836,.5946187254},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,5.085393248,4.404079742,3.790428332,3.363671463,3.051235949,5.025445453,4.352163430,3.745745885,3.324019728,4.577377083,3.964124837,3.411775437,4.062019202,3.517811818,3.577708763,2.542696624,2.202039870,1.895214166,1.681835732,1.525617974,4.923910840,4.264231874,3.670066451,3.256860890,5.909368405,5.117663157,4.404583150,6.141195788,5.318431565,6.,1.507633636,1.305649029,1.123723765,.9972059183,3.545621043,3.070597895,2.642749891,4.855041564,4.204589330,5.554920600,.9154754167,.7928249671,.6823550877,2.456478316,2.127372625,3.703280400,.5802884574,.5025445453,1.714285714,.3833259389},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,1.673320053,1.449137675,1.247219129,1.106797181,1.003992032,.8819171036,.7637626160,.6573421980,.5833333333,.5163977793,.4472135954,.3849001795,.3273268354,.2834733546,.2202291575,5.019960159,4.347413025,3.741657387,3.320391543,3.011976095,3.240370349,2.806243040,2.415229459,2.143303525,2.,1.732050808,1.490711985,1.299038106,1.125000000,.8864052607,5.291502622,4.582575695,3.944053189,3.500000000,4.,3.464101616,2.981423969,2.738612788,2.371708245,1.914854215,5.163977793,4.472135954,3.849001795,4.330127020,3.750000000,3.191423693,4.909902528,4.252100321,4.432026303,4.624812309},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,1.449137675,1.673320053,1.620185175,1.533623161,1.449137675,.7637626162,.8819171036,.8539125641,.8082903771,.4472135954,.5163977793,.5000000000,.2834733546,.3273268354,.1907240451,4.347413025,5.019960159,4.860555525,4.600869484,4.347413025,2.806243040,3.240370350,3.137475099,2.969848480,1.732050808,2.,1.936491673,1.125000000,1.299038106,.7676494735,4.582575696,5.291502622,5.123475385,4.849742262,3.464101616,4.,3.872983346,2.371708245,2.738612788,1.658312395,4.472135954,5.163977793,5.,3.750000000,4.330127020,2.763853992,4.252100321,4.909902529,3.838247368,4.005204947},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,1.247219129,1.620185174,1.673320053,1.649915823,1.603567452,.6573421979,.8539125633,.8819171036,.8695819911,.3849001795,.5000000000,.5163977793,.2439750181,.3169328455,.1641491224,3.741657387,4.860555524,5.019960158,4.949747467,4.810702356,2.415229457,3.137475099,3.240370349,3.195048251,1.490711985,1.936491673,2.,.9682458365,1.257788237,.6606874726,3.944053188,5.123475380,5.291502622,5.217491946,2.981423969,3.872983346,4.,2.041241452,2.651650429,1.427248064,3.849001795,5.,5.163977793,3.227486122,4.192627457,2.378746774,3.659625271,4.753992682,3.303437362,3.447131570},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,2.068278942,3.055050464,2.645751311,2.277100171,2.020725943,1.833030278,2.070196679,1.792842914,1.543033500,1.369306394,1.414213562,1.224744872,1.054092553,.9960238407,.8625819493,.7237468643,4.582575696,3.968626966,3.415650256,3.031088914,2.749545418,5.070925527,4.391550330,3.779644731,3.354101966,4.107919181,3.557562368,3.061862179,3.162277660,2.738612788,2.427520781,4.140393357,3.585685828,3.086067000,2.738612788,5.477225575,4.743416490,4.082482906,5.,4.330127020,4.195235392,3.535533905,3.061862179,2.635231383,5.270462768,4.564354646,5.244044240,2.988071523,2.587745849,4.855041564,2.533114026},
			{3.240370348,3.741657386,3.622844188,3.429285640,3.240370348,3.070597895,2.645751311,3.055050463,2.958039891,2.800000000,2.645751311,1.792842914,2.070196679,2.004459314,1.897366596,1.224744872,1.414213562,1.369306394,.8625819493,.9960238407,.6267831704,3.968626966,4.582575695,4.437059836,4.200000000,3.968626966,4.391550326,5.070925527,4.909902528,4.647580015,3.557562368,4.107919182,3.977475643,2.738612788,3.162277660,2.102294664,3.585685828,4.140393357,4.008918626,3.794733192,4.743416490,5.477225576,5.303300858,4.330127020,5.,3.633180426,3.061862179,3.535533905,3.423265985,4.564354646,5.270462768,4.541475532,2.587745849,2.988071523,4.204589330,2.193741096},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,3.256860892,2.820524267,2.427520781,2.154210993,2.472066162,2.140872097,1.842569328,1.880349512,1.628430445,1.449137675,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,5.983235509,5.181633950,4.459640440,3.957538298,5.744562647,4.974937185,4.281744193,4.974937185,4.308421985,4.166190450,3.256860892,2.820524267,2.427520781,2.154210993,5.744562647,4.974937185,4.281744193,6.292853088,5.449770639,6.,2.472066162,2.140872097,1.842569328,4.974937185,4.308421985,6.,1.880349512,1.628430445,4.166190450,1.449137675},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,.6546536711,.5669467095,.4879500367,.4330127020,.3354101966,.2904737510,.2500000000,.1889822365,.1636634177,.1144344270,5.796550700,5.019960159,4.320493800,3.834057903,3.477930420,3.207134903,2.777460300,2.390457218,2.121320343,1.732050808,1.500000000,1.290994449,1.,.8660254040,.6141195788,6.546536711,5.669467095,4.879500367,4.330127020,4.330127020,3.750000000,3.227486122,2.635231383,2.282177324,1.658312395,6.708203931,5.809475019,5.,5.,4.330127020,3.316624790,6.614378278,5.728219620,5.373546315,6.408327913},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,1.254990040,1.449137675,1.403121520,1.328156617,1.254990040,.5669467095,.6546536709,.6338656908,.6000000000,.2904737510,.3354101966,.3247595265,.1636634177,.1889822365,.9910312088e-1,5.019960159,5.796550700,5.612486080,5.312626469,5.019960159,2.777460299,3.207134903,3.105295018,2.939387692,1.500000000,1.732050808,1.677050983,.8660254040,1.,.5318431565,5.669467095,6.546536709,6.338656908,6.,3.750000000,4.330127020,4.192627457,2.282177323,2.635231383,1.436140662,5.809475019,6.708203931,6.495190530,4.330127020,5.,2.872281324,5.728219619,6.614378278,4.653627619,5.549774770},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,2.746426250,2.378474901,2.047065264,1.816590213,1.647855750,1.628430445,1.410262133,1.213760391,1.077105496,.9888264647,.8563488387,.7370277310,.6267831705,.5428101485,.4140393357,5.492852500,4.756949802,4.094130528,3.633180426,3.295711500,5.318431565,4.605896844,3.964124837,3.517811820,3.829708432,3.316624790,2.854496129,2.653299832,2.297825059,1.851640200,5.428101485,4.700873779,4.045867967,3.590351654,6.382847385,5.527707984,4.757493549,5.244044240,4.541475532,4.,4.944132325,4.281744193,3.685138655,6.633249580,5.744562647,6.,4.387482194,3.799671040,6.480740700,3.864367134},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,1.271348312,1.101019935,.9476070828,.8409178659,.7628089876,.5025445453,.4352163430,.3745745885,.3324019728,.2288688541,.1982062418,.1705887719,.1160576915,.1005089091,.6388765648e-1,6.356741560,5.505099678,4.738035416,4.204589330,3.814044936,3.077444275,2.665144921,2.293791533,2.035538055,1.477342101,1.279415789,1.101145788,.7676494735,.6648039455,.4285714286,7.538168182,6.528245145,5.618618828,4.986029592,4.432026303,3.838247368,3.303437362,2.427520781,2.102294664,1.388730150,8.010409897,6.937218462,5.970607017,5.373546315,4.653627619,3.240370350,8.124038403,7.035623637,6.,8.049844717},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,1.354006401,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.581138830,1.369306394,1.178511302,1.045825033,1.296148140,1.122497216,.9660917834,1.095445115,.9486832980,.9476070828,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.936491673,1.677050983,1.443375673,1.280868846,1.673320053,1.449137675,1.247219129,1.449137675,1.254990040,1.271348312,1.581138830,1.369306394,1.178511302,1.045825033,1.673320053,1.449137675,1.247219129,1.527525232,1.322875656,1.373213125,1.296148140,1.122497216,.9660917834,1.449137675,1.254990040,1.373213125,1.095445115,.9486832980,1.271348312,.9476070828},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,2.010178183,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.122497216,1.296148140,1.254990040,.9486832980,1.095445115,.8206518070,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.677050983,1.936491673,1.875000000,1.774823935,1.449137675,1.673320053,1.620185174,1.254990040,1.449137675,1.101019935,1.369306394,1.581138830,1.530931089,1.449137675,1.449137675,1.673320053,1.620185174,1.322875656,1.527525232,1.189237451,1.122497216,1.296148140,1.254990040,1.254990040,1.449137675,1.189237451,.9486832980,1.095445115,1.101019935,.8206518070},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,2.270737765,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.178511302,1.530931089,1.581138830,1.559023911,.9660917834,1.254990040,1.296148140,.8164965809,1.060660172,.7063046176,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.443375673,1.875000000,1.936491673,1.909406540,1.247219129,1.620185175,1.673320053,1.080123450,1.403121520,.9476070828,1.178511302,1.530931089,1.581138830,1.559023911,1.247219129,1.620185175,1.673320053,1.138550085,1.479019946,1.023532632,.9660917834,1.254990040,1.296148140,1.080123450,1.403121520,1.023532632,.8164965809,1.060660172,.9476070828,.7063046176},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,2.388242806,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.045825033,1.449137675,1.559023911,1.581138830,.8573214100,1.187939392,1.278019301,.7245688375,1.003992032,.6267831704,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.280868845,1.774823935,1.909406540,1.936491673,1.106797181,1.533623161,1.649915823,.9585144758,1.328156617,.8409178659,1.045825033,1.449137675,1.559023911,1.581138830,1.106797181,1.533623161,1.649915823,1.010362971,1.400000000,.9082951065,.8573214100,1.187939392,1.278019301,.9585144758,1.328156617,.9082951065,.7245688375,1.003992032,.8409178659,.6267831704},
			{1.469693846,2.121320343,2.347382390,2.430277762,2.449489743,2.437211522,1.200000000,1.732050808,1.916629694,1.984313483,2.,.9486832980,1.369306394,1.515228816,1.568737550,.7776888840,1.122497216,1.242118007,.6572670691,.9486832980,.5685642496,1.200000000,1.732050808,1.916629694,1.984313483,2.,1.161895004,1.677050983,1.855768722,1.921303269,1.003992032,1.449137675,1.603567452,.8694826050,1.254990040,.7628089876,.9486832980,1.369306394,1.515228816,1.568737550,1.003992032,1.449137675,1.603567452,.9165151392,1.322875656,.8239278748,.7776888840,1.122497216,1.242118007,.8694826050,1.254990040,.8239278748,.6572670691,.9486832980,.7628089876,.5685642496},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.427248064,3.162277660,2.738612788,2.357022604,2.091650066,1.897366596,3.,2.598076212,2.236067977,1.984313483,2.732520206,2.366431913,2.036700309,2.474358298,2.142857143,2.247447532,1.581138830,1.369306394,1.178511302,1.045825033,.9486832980,2.449489743,2.121320343,1.825741858,1.620185175,2.645751311,2.291287848,1.972026594,2.618614683,2.267786838,2.512722727,1.,.8660254040,.7453559923,.6614378278,1.763834207,1.527525232,1.314684396,2.070196679,1.792842914,2.171240595,.6831300514,.5916079782,.5091750773,1.309307342,1.133893419,1.628430445,.4948716594,.4285714286,1.005089091,.3745745885},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,2.738612788,3.162277660,3.061862179,2.898275350,2.738612788,2.598076212,3.,2.904737510,2.749545417,2.366431913,2.732520203,2.645751311,2.142857143,2.474358298,1.946346656,1.369306394,1.581138830,1.530931089,1.449137675,1.369306394,2.121320343,2.449489743,2.371708245,2.244994432,2.291287848,2.645751311,2.561737690,2.267786838,2.618614683,2.176081715,.8660254040,1.,.9682458365,.9165151390,1.527525232,1.763834207,1.707825127,1.792842914,2.070196679,1.880349512,.5916079782,.6831300507,.6614378278,1.133893419,1.309307341,1.410262133,.4285714286,.4948716594,.8704326857,.3243911095},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.393567770,2.357022604,3.061862179,3.162277659,3.118047822,3.030457633,2.236067977,2.904737510,3.,2.958039892,2.036700309,2.645751311,2.732520206,1.844277784,2.395787117,1.675148485,1.178511302,1.530931089,1.581138830,1.559023911,1.515228816,1.825741858,2.371708244,2.449489743,2.415229459,1.972026594,2.561737692,2.645751311,1.951800147,2.535462764,1.872872943,.7453559923,.9682458365,1.,.9860132974,1.314684396,1.707825129,1.763834207,1.543033500,2.004459314,1.618347188,.5091750773,.6614378278,.6831300514,.9759000734,1.267731382,1.213760391,.3688555568,.4791574237,.7491491772,.2791914141},
			{1.707825129,2.366431913,2.545875386,2.581988897,2.561737692,2.517428959,2.091650066,2.898275350,3.118047822,3.162277660,3.137475099,1.984313483,2.749545417,2.958039892,3.,1.807392228,2.504396134,2.694301258,1.636634177,2.267786838,1.486546813,1.045825033,1.449137675,1.559023911,1.581138830,1.568737550,1.620185174,2.244994432,2.415229459,2.449489743,1.750000000,2.424871131,2.608745974,1.732050808,2.400000000,1.662009863,.6614378278,.9165151390,.9860132974,1.,1.166666667,1.616580754,1.739163982,1.369306394,1.897366596,1.436140662,.4518480572,.6260990336,.6735753142,.8660254040,1.200000000,1.077105496,.3273268354,.4535573675,.6648039455,.2477578022},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.462494065,3.888444420,3.367491648,2.898275350,2.571964230,2.333066652,4.098780308,3.549647869,3.055050464,2.711088342,4.,3.464101616,2.981423969,3.803194146,3.293662747,3.582364211,1.296148140,1.122497216,.9660917834,.8573214100,.7776888840,2.509980079,2.173706512,1.870828694,1.660195771,3.098386677,2.683281572,2.309401077,3.354101966,2.904737510,3.433032812,.6831300514,.5916079782,.5091750773,.4518480572,1.549193338,1.341640786,1.154700539,2.121320343,1.837117307,2.472066162,.4000000000,.3464101616,.2981423969,1.006230590,.8714212528,1.483239697,.2535462764,.2195775165,.6866065626,.1705887719},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,2.171240594,3.367491648,3.888444419,3.764970119,3.563818176,3.367491648,3.549647869,4.098780304,3.968626966,3.756594201,3.464101616,4.,3.872983346,3.293662744,3.803194146,3.102418412,1.122497216,1.296148140,1.254990040,1.187939392,1.122497216,2.173706511,2.509980079,2.430277762,2.300434741,2.683281572,3.098386677,3.,2.904737510,3.354101966,2.973093626,.5916079782,.6831300507,.6614378278,.6260990336,1.341640786,1.549193338,1.500000000,1.837117307,2.121320343,2.140872097,.3464101616,.4000000000,.3872983346,.8714212528,1.006230590,1.284523258,.2195775163,.2535462764,.5946187254,.1477342101},
			{1.972026594,2.561737692,2.645751311,2.608745974,2.535462764,2.452677109,2.898275350,3.764970119,3.888444420,3.834057903,3.726354021,3.055050464,3.968626966,4.098780308,4.041451885,2.981423969,3.872983346,4.,2.834733546,3.682426899,2.670136631,.9660917834,1.254990040,1.296148140,1.278019301,1.242118007,1.870828694,2.430277762,2.509980080,2.474873734,2.309401077,3.,3.098386677,2.500000000,3.247595265,2.558831578,.5091750773,.6614378278,.6831300514,.6735753142,1.154700539,1.500000000,1.549193338,1.581138830,2.053959591,1.842569328,.2981423969,.3872983346,.4000000000,.7500000000,.9742785795,1.105541597,.1889822365,.2454951266,.5117663157,.1271493634},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.483239697,4.381780461,3.794733192,3.265986323,2.898275350,2.629068276,4.948716594,4.285714286,3.688555568,3.273268354,5.070925527,4.391550326,3.779644731,5.,4.330127020,4.844240569,1.095445115,.9486832980,.8164965809,.7245688375,.6572670691,2.424366106,2.099562637,1.807015806,1.603567451,3.273268354,2.834733546,2.439750181,3.779644731,3.273268354,4.062019202,.4948716594,.4285714286,.3688555568,.3273268354,1.309307341,1.133893419,.9759000724,1.992047682,1.725163898,2.507132682,.2535462764,.2195775163,.1889822365,.7559289460,.6546536709,1.253566341,.1428571429,.1237179149,.4642307659,.8650429586e-1},
			{2.323790008,2.683281572,2.598076212,2.459268185,2.323790008,2.202039870,3.794733192,4.381780460,4.242640686,4.015968127,3.794733192,4.285714286,4.948716594,4.791574237,4.535573675,4.391550330,5.070925527,4.909902529,4.330127020,5.,4.195235392,.9486832980,1.095445115,1.060660172,1.003992032,.9486832980,2.099562637,2.424366106,2.347382389,2.221968240,2.834733546,3.273268354,3.169328455,3.273268354,3.779644731,3.517811818,.4285714286,.4948716594,.4791574237,.4535573675,1.133893419,1.309307342,1.267731382,1.725163898,1.992047682,2.171240595,.2195775165,.2535462764,.2454951266,.6546536711,.7559289460,1.085620297,.1237179149,.1428571429,.4020356364,.7491491772e-1},
			{2.708012801,2.345207879,2.018433570,1.791182105,1.624807681,1.496910398,4.738035416,4.103259034,3.531523088,3.133915852,2.842821248,5.618618828,4.865866643,4.187871213,3.716367033,5.970607017,5.170697355,4.450227718,6.055300709,5.244044240,6.,.9476070828,.8206518070,.7063046176,.6267831704,.5685642496,2.293791533,1.986481737,1.709691264,1.517200488,3.303437362,2.860860676,2.462236835,4.005204947,3.468609231,4.472135954,.3745745885,.3243911095,.2791914141,.2477578022,1.101145788,.9536202253,.8207456118,1.809367160,1.566957926,2.415229459,.1705887719,.1477342101,.1271493634,.5721721350,.4955156045,1.035098339,.8650429586e-1,.7491491772e-1,.3194382825,.4761904762e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.748014747,1.936491673,1.677050983,1.443375673,1.280868845,1.161895004,1.224744872,1.060660172,.9128709292,.8100925872,.8366600264,.7245688370,.6236095646,.6060915266,.5248906592,.4587583064,3.872983346,3.354101966,2.886751347,2.561737690,2.323790008,3.,2.598076212,2.236067977,1.984313483,2.160246899,1.870828694,1.610152971,1.603567452,1.388730150,1.230977710,3.674234614,3.181980514,2.738612787,2.430277762,3.240370349,2.806243040,2.415229457,2.535462764,2.195775163,1.994411836,3.346640106,2.898275348,2.494438258,3.207134903,2.777460299,2.659215782,3.030457633,2.624453295,3.077444275,2.752549839},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,2.595128875,1.677050983,1.936491673,1.875000000,1.774823935,1.677050983,1.060660172,1.224744872,1.185854122,1.122497216,.7245688375,.8366600264,.8100925875,.5248906592,.6060915266,.3972963474,3.354101966,3.872983346,3.750000000,3.549647869,3.354101966,2.598076212,3.,2.904737510,2.749545418,1.870828694,2.160246900,2.091650066,1.388730150,1.603567452,1.066057968,3.181980514,3.674234614,3.557562366,3.367491648,2.806243040,3.240370350,3.137475099,2.195775165,2.535462764,1.727211316,2.898275350,3.346640106,3.240370350,2.777460300,3.207134903,2.302948421,2.624453295,3.030457633,2.665144921,2.383778084},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.931509850,1.443375673,1.875000000,1.936491673,1.909406540,1.855768722,.9128709293,1.185854122,1.224744872,1.207614729,.6236095646,.8100925872,.8366600265,.4517539516,.5868455971,.3419382528,2.886751347,3.750000000,3.872983346,3.818813079,3.711537446,2.236067977,2.904737510,3.,2.958039892,1.610152972,2.091650066,2.160246899,1.195228610,1.552647508,.9175166129,2.738612788,3.557562368,3.674234614,3.622844188,2.415229459,3.137475099,3.240370349,1.889822365,2.454951265,1.486546813,2.494438258,3.240370349,3.346640106,2.390457218,3.105295018,1.982062418,2.258769757,2.934227987,2.293791533,2.051629517},
			{2.091650066,2.898275350,3.118047822,3.162277659,3.137475099,3.083208205,1.280868846,1.774823935,1.909406540,1.936491673,1.921303269,.8100925874,1.122497216,1.207614729,1.224744872,.5533985903,.7668115804,.8249579111,.4008918626,.5554920599,.3034400977,2.561737692,3.549647870,3.818813080,3.872983346,3.842606539,1.984313483,2.749545418,2.958039892,3.,1.428869017,1.979898987,2.130032168,1.060660172,1.469693846,.8142152223,2.430277762,3.367491647,3.622844188,3.674234614,2.143303525,2.969848480,3.195048251,1.677050983,2.323790008,1.319179432,2.213594361,3.067246321,3.299831644,2.121320343,2.939387692,1.758905910,2.004459314,2.777460300,2.035538055,1.820640586},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,2.645751311,2.291287848,1.972026594,1.750000000,2.065591118,1.788854382,1.539600718,1.636634177,1.417366774,1.321374945,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,4.320493799,3.741657387,3.220305944,2.857738034,4.,3.464101616,2.981423969,3.464101616,3.,2.954684202,2.645751311,2.291287848,1.972026594,1.750000000,4.,3.464101616,2.981423969,4.107919181,3.557562368,3.829708432,2.065591118,1.788854382,1.539600718,3.464101616,3.,3.829708432,1.636634177,1.417366774,2.954684202,1.321374945},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,2.291287848,2.645751311,2.561737692,2.424871131,1.788854382,2.065591118,2.,1.417366774,1.636634177,1.144344270,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,3.741657387,4.320493800,4.183300132,3.959797974,3.464101616,4.,3.872983346,3.,3.464101616,2.558831578,2.291287848,2.645751311,2.561737692,2.424871131,3.464101616,4.,3.872983346,3.557562368,4.107919182,3.316624790,1.788854382,2.065591118,2.,3.,3.464101616,3.316624790,1.417366774,1.636634177,2.558831578,1.144344270},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,1.972026594,2.561737690,2.645751311,2.608745974,1.539600718,2.,2.065591118,1.219875090,1.584664228,.9848947341,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,3.220305942,4.183300132,4.320493799,4.260064335,2.981423969,3.872983346,4.,2.581988897,3.354101966,2.202291575,1.972026594,2.561737690,2.645751311,2.608745974,2.981423969,3.872983346,4.,3.061862179,3.977475643,2.854496129,1.539600718,2.,2.065591118,2.581988897,3.354101966,2.854496129,1.219875090,1.584664228,2.202291575,.9848947341},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,4.347413025,3.764970119,3.240370350,2.875543427,2.608447815,3.927922025,3.401680258,2.927700220,2.598076212,3.354101966,2.904737510,2.500000000,2.834733546,2.454951266,2.403122968,2.898275350,2.509980080,2.160246900,1.917028952,1.738965210,4.810702356,4.166190450,3.585685828,3.181980514,5.196152424,4.500000000,3.872983346,5.,4.330127020,4.605896842,1.963961013,1.700840129,1.463850110,1.299038106,3.897114318,3.375000000,2.904737510,4.743416490,4.107919182,4.974937185,1.341640786,1.161895004,1.,3.,2.598076212,3.979949748,.9449111824,.8183170887,2.302948420,.6866065625},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,3.764970119,4.347413025,4.209364560,3.984469852,3.764970119,3.401680258,3.927922024,3.803194146,3.600000000,2.904737510,3.354101966,3.247595265,2.454951265,2.834733546,2.081165539,2.509980080,2.898275350,2.806243040,2.656313234,2.509980080,4.166190449,4.810702356,4.657942526,4.409081537,4.500000000,5.196152424,5.031152948,4.330127020,5.,3.988823673,1.700840129,1.963961012,1.901597073,1.800000000,3.375000000,3.897114318,3.773364711,4.107919181,4.743416490,4.308421985,1.161895004,1.341640786,1.299038106,2.598076212,3.,3.446737588,.8183170885,.9449111824,1.994411836,.5946187254},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,5.085393248,4.404079742,3.790428332,3.363671463,3.051235949,5.025445453,4.352163430,3.745745885,3.324019728,4.577377083,3.964124837,3.411775437,4.062019202,3.517811818,3.577708763,2.542696624,2.202039870,1.895214166,1.681835732,1.525617974,4.923910840,4.264231874,3.670066451,3.256860890,5.909368405,5.117663157,4.404583150,6.141195788,5.318431565,6.,1.507633636,1.305649029,1.123723765,.9972059183,3.545621043,3.070597895,2.642749891,4.855041564,4.204589330,5.554920600,.9154754167,.7928249671,.6823550877,2.456478316,2.127372625,3.703280400,.5802884574,.5025445453,1.714285714,.3833259389},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,1.673320053,1.449137675,1.247219129,1.106797181,1.003992032,.8819171036,.7637626160,.6573421980,.5833333333,.5163977793,.4472135954,.3849001795,.3273268354,.2834733546,.2202291575,5.019960159,4.347413025,3.741657387,3.320391543,3.011976095,3.240370349,2.806243040,2.415229459,2.143303525,2.,1.732050808,1.490711985,1.299038106,1.125000000,.8864052607,5.291502622,4.582575695,3.944053189,3.500000000,4.,3.464101616,2.981423969,2.738612788,2.371708245,1.914854215,5.163977793,4.472135954,3.849001795,4.330127020,3.750000000,3.191423693,4.909902528,4.252100321,4.432026303,4.624812309},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,1.449137675,1.673320053,1.620185175,1.533623161,1.449137675,.7637626162,.8819171036,.8539125641,.8082903771,.4472135954,.5163977793,.5000000000,.2834733546,.3273268354,.1907240451,4.347413025,5.019960159,4.860555525,4.600869484,4.347413025,2.806243040,3.240370350,3.137475099,2.969848480,1.732050808,2.,1.936491673,1.125000000,1.299038106,.7676494735,4.582575696,5.291502622,5.123475385,4.849742262,3.464101616,4.,3.872983346,2.371708245,2.738612788,1.658312395,4.472135954,5.163977793,5.,3.750000000,4.330127020,2.763853992,4.252100321,4.909902529,3.838247368,4.005204947},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,1.247219129,1.620185174,1.673320053,1.649915823,1.603567452,.6573421979,.8539125633,.8819171036,.8695819911,.3849001795,.5000000000,.5163977793,.2439750181,.3169328455,.1641491224,3.741657387,4.860555524,5.019960158,4.949747467,4.810702356,2.415229457,3.137475099,3.240370349,3.195048251,1.490711985,1.936491673,2.,.9682458365,1.257788237,.6606874726,3.944053188,5.123475380,5.291502622,5.217491946,2.981423969,3.872983346,4.,2.041241452,2.651650429,1.427248064,3.849001795,5.,5.163977793,3.227486122,4.192627457,2.378746774,3.659625271,4.753992682,3.303437362,3.447131570},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,2.068278942,3.055050464,2.645751311,2.277100171,2.020725943,1.833030278,2.070196679,1.792842914,1.543033500,1.369306394,1.414213562,1.224744872,1.054092553,.9960238407,.8625819493,.7237468643,4.582575696,3.968626966,3.415650256,3.031088914,2.749545418,5.070925527,4.391550330,3.779644731,3.354101966,4.107919181,3.557562368,3.061862179,3.162277660,2.738612788,2.427520781,4.140393357,3.585685828,3.086067000,2.738612788,5.477225575,4.743416490,4.082482906,5.,4.330127020,4.195235392,3.535533905,3.061862179,2.635231383,5.270462768,4.564354646,5.244044240,2.988071523,2.587745849,4.855041564,2.533114026},
			{3.240370348,3.741657386,3.622844188,3.429285640,3.240370348,3.070597895,2.645751311,3.055050463,2.958039891,2.800000000,2.645751311,1.792842914,2.070196679,2.004459314,1.897366596,1.224744872,1.414213562,1.369306394,.8625819493,.9960238407,.6267831704,3.968626966,4.582575695,4.437059836,4.200000000,3.968626966,4.391550326,5.070925527,4.909902528,4.647580015,3.557562368,4.107919182,3.977475643,2.738612788,3.162277660,2.102294664,3.585685828,4.140393357,4.008918626,3.794733192,4.743416490,5.477225576,5.303300858,4.330127020,5.,3.633180426,3.061862179,3.535533905,3.423265985,4.564354646,5.270462768,4.541475532,2.587745849,2.988071523,4.204589330,2.193741096},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,3.256860892,2.820524267,2.427520781,2.154210993,2.472066162,2.140872097,1.842569328,1.880349512,1.628430445,1.449137675,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,5.983235509,5.181633950,4.459640440,3.957538298,5.744562647,4.974937185,4.281744193,4.974937185,4.308421985,4.166190450,3.256860892,2.820524267,2.427520781,2.154210993,5.744562647,4.974937185,4.281744193,6.292853088,5.449770639,6.,2.472066162,2.140872097,1.842569328,4.974937185,4.308421985,6.,1.880349512,1.628430445,4.166190450,1.449137675},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,.6546536711,.5669467095,.4879500367,.4330127020,.3354101966,.2904737510,.2500000000,.1889822365,.1636634177,.1144344270,5.796550700,5.019960159,4.320493800,3.834057903,3.477930420,3.207134903,2.777460300,2.390457218,2.121320343,1.732050808,1.500000000,1.290994449,1.,.8660254040,.6141195788,6.546536711,5.669467095,4.879500367,4.330127020,4.330127020,3.750000000,3.227486122,2.635231383,2.282177324,1.658312395,6.708203931,5.809475019,5.,5.,4.330127020,3.316624790,6.614378278,5.728219620,5.373546315,6.408327913},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,1.254990040,1.449137675,1.403121520,1.328156617,1.254990040,.5669467095,.6546536709,.6338656908,.6000000000,.2904737510,.3354101966,.3247595265,.1636634177,.1889822365,.9910312088e-1,5.019960159,5.796550700,5.612486080,5.312626469,5.019960159,2.777460299,3.207134903,3.105295018,2.939387692,1.500000000,1.732050808,1.677050983,.8660254040,1.,.5318431565,5.669467095,6.546536709,6.338656908,6.,3.750000000,4.330127020,4.192627457,2.282177323,2.635231383,1.436140662,5.809475019,6.708203931,6.495190530,4.330127020,5.,2.872281324,5.728219619,6.614378278,4.653627619,5.549774770},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,2.746426250,2.378474901,2.047065264,1.816590213,1.647855750,1.628430445,1.410262133,1.213760391,1.077105496,.9888264647,.8563488387,.7370277310,.6267831705,.5428101485,.4140393357,5.492852500,4.756949802,4.094130528,3.633180426,3.295711500,5.318431565,4.605896844,3.964124837,3.517811820,3.829708432,3.316624790,2.854496129,2.653299832,2.297825059,1.851640200,5.428101485,4.700873779,4.045867967,3.590351654,6.382847385,5.527707984,4.757493549,5.244044240,4.541475532,4.,4.944132325,4.281744193,3.685138655,6.633249580,5.744562647,6.,4.387482194,3.799671040,6.480740700,3.864367134},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,1.271348312,1.101019935,.9476070828,.8409178659,.7628089876,.5025445453,.4352163430,.3745745885,.3324019728,.2288688541,.1982062418,.1705887719,.1160576915,.1005089091,.6388765648e-1,6.356741560,5.505099678,4.738035416,4.204589330,3.814044936,3.077444275,2.665144921,2.293791533,2.035538055,1.477342101,1.279415789,1.101145788,.7676494735,.6648039455,.4285714286,7.538168182,6.528245145,5.618618828,4.986029592,4.432026303,3.838247368,3.303437362,2.427520781,2.102294664,1.388730150,8.010409897,6.937218462,5.970607017,5.373546315,4.653627619,3.240370350,8.124038403,7.035623637,6.,8.049844717},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,6.8546484667622e-008,1.09461601509022e-009,7.65361974719019e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,.122554596919714,.00195706935488603,2.32941342728397e-005,2.54112819850642e-007,.00244089433681864,3.89785419960172e-005,4.63944411923424e-007,3.89785419960172e-005,6.22446745908012e-007,5.51740069821357e-007,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,.262307293803493,.00418877446399711,4.98571369236328e-005,.00662303395559898,.000105762958835416,.000128085303901421,7.27823584467216,.116225851033514,.00138338509726576,.450140785461579,.0071882798268767,.0137645079246526,9.14177937313726,.145984701575574,.68472894504081,10.9691712594614},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,5.01144631725624e-005,8.00275816365055e-007,2.90963994598978e-006,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,.612365059546825,.00977883263593246,.000116393136443702,1.26971828042803e-006,.0634197097850823,.00101274675595133,1.20542784326106e-005,.00526617557576491,8.40953427875793e-005,.000387613365377294,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,1.2594400120878,.0201119461265218,.000239383633673578,.165355473459674,.00264055480463631,.0166285878943821,6.45780747384476,.103124463877826,.00122744505825102,2.07683410541886,.0331648790465492,.330223979417038,7.79428043908953,.124466546112493,3.03569721712443,8.98680735137037},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.0113956405123916,.000181976518058871,2.16598631798674e-006,.00166206328995312,2.65414208197256e-005,.000231576690958128,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,1.36595759702526,.0218129211013044,.000259629589389913,2.83226696917689e-006,.339486246269225,.00542124200448682,6.45266550483906e-005,.067649468233801,.00108029159656201,.0119492023316399,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,2.61131941894657,.0417000531735786,.000496337360406356,.822758173894117,.0131385916878024,.198554676115027,5.18623489409549,.0828187732700635,.000985755369367884,4.0025799484364,.0639170357987087,1.5272774292324,5.81833621896758,.0929127735338078,5.43816471058122,6.23568135536041},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.0657431866024243,.00104985026256264,1.24959051250218e-005,.0171974308138302,.000274625070541602,.00429748986298374,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,2.19713005267825,.0350858801127846,.000417611772616793,4.55567499953034e-006,.979365349166059,.0156394452705675,.000186149426512757,.350018452610409,.00558943038088415,.110884150015163,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.76697803717186,.0601547184591987,.000715995110407768,2.12867470143779,.0339927193989142,.921342735777177,3.7410698100085,.0597409910483066,.000711070695350591,5.17830525859385,.0826921452798168,3.543802987655,3.76406389114815,.0601081826012263,6.30978459469567,3.6179005979755},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,.236321146004057,.00377380273155657,4.4917911225712e-005,.0946962283839188,.00151220020487266,.0362494356624709,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.90029193047496,.0463146434325445,.000551262795170185,6.01365741772922e-006,1.98037740107218,.0316245658532948,.000376413273965959,1.08420509999131,.0173136269811071,.526147189418459,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,4.28496925234002,.0684264990231684,.00081445047001826,3.70920808974538,.0592321925476606,2.4592936780291,2.39387697960251,.0382277237453785,.000455007753107355,5.07587527588423,.0810564450674401,5.32120685033247,2.07554270972356,.0331442567927967,5.32974583920556,1.71909544841957},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,.626153181688297,.00999901459257298,.000119013865260662,.347192252628355,.00554429890730429,.183906859961414,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,3.30149822057238,.0527214902997539,.000627520670660238,6.84554511742325e-006,3.11944010216456,.0498142116158778,.000592916613348927,2.36319452013471,.0377377568190159,1.58691847335237,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,4.0126111958998,.0640772243410266,.000762682969712426,4.80639993236566,.076753204286926,4.40969381865368,1.33269841309056,.0212818065446243,.000253307966815722,3.91021223386622,.0624420195354773,5.67229573503973,.95054294779362,.0151791815252261,3.37758202463881,.647664476709677},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,1.33269841309056,.0212818065446243,.000253307966815722,.95054294779362,.0151791815252261,.647664476709677,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,3.30149822057238,.0527214902997539,.000627520670660238,6.84554511742325e-006,4.0126111958998,.0640772243410266,.000762682969712426,3.91021223386622,.0624420195354773,3.37758202463881,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,3.11944010216456,.0498142116158778,.000592916613348927,4.80639993236566,.076753204286926,5.67229573503973,.626153181688297,.00999901459257298,.000119013865260662,2.36319452013471,.0377377568190159,4.40969381865368,.347192252628355,.00554429890730429,1.58691847335237,.183906859961414},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,2.39387697960251,.0382277237453785,.000455007753107355,2.07554270972356,.0331442567927967,1.71909544841957,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,2.90029193047496,.0463146434325445,.000551262795170185,6.01365741772922e-006,4.28496925234002,.0684264990231684,.00081445047001826,5.07587527588423,.0810564450674401,5.32974583920556,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,1.98037740107218,.0316245658532948,.000376413273965959,3.70920808974538,.0592321925476606,5.32120685033247,.236321146004057,.00377380273155657,4.4917911225712e-005,1.08420509999131,.0173136269811071,2.4592936780291,.0946962283839188,.00151220020487266,.526147189418459,.0362494356624709},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.7410698100085,.0597409910483066,.000711070695350591,3.76406389114815,.0601081826012263,3.6179005979755,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,2.19713005267825,.0350858801127846,.000417611772616793,4.55567499953034e-006,3.76697803717186,.0601547184591987,.000715995110407768,5.17830525859385,.0826921452798168,6.30978459469567,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.979365349166059,.0156394452705675,.000186149426512757,2.12867470143779,.0339927193989142,3.543802987655,.0657431866024243,.00104985026256264,1.24959051250218e-005,.350018452610409,.00558943038088415,.921342735777177,.0171974308138302,.000274625070541602,.110884150015163,.00429748986298374},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,5.18623489409549,.0828187732700635,.000985755369367884,5.81833621896758,.0929127735338078,6.23568135536041,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,1.36595759702526,.0218129211013044,.000259629589389913,2.83226696917689e-006,2.61131941894657,.0417000531735786,.000496337360406356,4.0025799484364,.0639170357987087,5.43816471058122,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.339486246269225,.00542124200448682,6.45266550483906e-005,.822758173894117,.0131385916878024,1.5272774292324,.0113956405123916,.000181976518058871,2.16598631798674e-006,.067649468233801,.00108029159656201,.198554676115027,.00166206328995312,2.65414208197256e-005,.0119492023316399,.000231576690958128},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,6.45780747384476,.103124463877826,.00122744505825102,7.79428043908953,.124466546112493,8.98680735137037,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.612365059546825,.00977883263593246,.000116393136443702,1.26971828042803e-006,1.2594400120878,.0201119461265218,.000239383633673578,2.07683410541886,.0331648790465492,3.03569721712443,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.0634197097850823,.00101274675595133,1.20542784326106e-005,.165355473459674,.00264055480463631,.330223979417038,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,.00526617557576491,8.40953427875793e-005,.0166285878943821,5.01144631725624e-005,8.00275816365055e-007,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,7.27823584467216,.116225851033514,.00138338509726576,9.14177937313726,.145984701575574,10.9691712594614,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.122554596919714,.00195706935488603,2.32941342728397e-005,2.54112819850642e-007,.262307293803493,.00418877446399711,4.98571369236328e-005,.450140785461579,.0071882798268767,.68472894504081,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,.00244089433681864,3.89785419960172e-005,4.63944411923424e-007,.00662303395559898,.000105762958835416,.0137645079246526,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,3.89785419960172e-005,6.22446745908012e-007,.000128085303901421,6.8546484667622e-008,1.09461601509022e-009,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,6.8546484667622e-008,5.69188702865245e-009,7.65361974719019e-010,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,.122554596919714,.0101765528017885,.000629847662741689,3.57281062205031e-005,.00244089433681864,.0002026842788013,1.25445444792834e-005,3.89785419960172e-005,3.23665697396975e-006,5.51740069821357e-007,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,.262307293803493,.0217811823588299,.0013480819160108,.00662303395559898,.000549956153763738,.000128085303901421,7.27823584467216,.60436208201944,.037405205103521,.450140785461579,.037378291693961,.0137645079246526,9.14177937313726,.759104944404841,.68472894504081,10.9691712594614},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,5.01144631725624e-005,4.16134925463982e-006,2.90963994598978e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,.612365059546825,.0508488911805658,.0031471418550943,.000178521609495778,.0634197097850823,.00526617557576491,.000325934375061529,.00526617557576491,.000437286851447401,.000387613365377294,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,1.2594400120878,.104579983939912,.00647266905916808,.165355473459674,.0137306045533387,.0166285878943821,6.45780747384476,.53623626019628,.0331887586818224,2.07683410541886,.172453848883548,.330223979417038,7.79428043908953,.647212821256298,3.03569721712443,8.98680735137037},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.0113956405123916,.000946258567734596,5.85658777413804e-005,.00166206328995312,.000138012569524986,.000231576690958128,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,1.36595759702526,.11342487316276,.00702009734040598,.000398214994344052,.339486246269225,.0281898826795267,.0017447294838579,.067649468233801,.00561740157357991,.0119492023316399,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,2.61131941894657,.216835847896165,.0134204140287723,.822758173894117,.0683192814389796,.198554676115027,5.18623489409549,.430648825529863,.0266537364367886,4.0025799484364,.332361798070895,1.5272774292324,5.81833621896758,.483136554903789,5.43816471058122,6.23568135536041},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.0657431866024243,.0054591098696972,.000337875473054763,.0171974308138302,.00142802120123345,.00429748986298374,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,2.19713005267825,.182442850414858,.0112917610860855,.000640525103711722,.979365349166059,.0813234544750617,.00503327489386337,.350018452610409,.029064444374942,.110884150015163,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.76697803717186,.31279814746903,.0193597169803656,2.12867470143779,.176758530797994,.921342735777177,3.7410698100085,.310646808870845,.0192265662361566,5.17830525859385,.429990373428657,3.543802987655,3.76406389114815,.31255616594975,6.30978459469567,3.6179005979755},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,.236321146004057,.0196233734207421,.0012145307084342,.0946962283839188,.00786328046747136,.0362494356624709,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.90029193047496,.240831227166574,.0149055371205287,.000845516535217038,1.98037740107218,.164444383862132,.0101777991856447,1.08420509999131,.0900290215693292,.526147189418459,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,4.28496925234002,.355810527926517,.0220218411618753,3.70920808974538,.308001110702629,2.4592936780291,2.39387697960251,.198780103635972,.0123029070952925,5.07587527588423,.421484906038046,5.32120685033247,2.07554270972356,.172346615354148,5.32974583920556,1.71909544841957},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,.626153181688297,.0519938139714517,.00321800346775215,.347192252628355,.028829765505355,.183906859961414,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,3.30149822057238,.274146150459585,.0169674658481853,.000962479434277651,3.11944010216456,.259028610060664,.016031810358111,2.36319452013471,.196232327654089,1.58691847335237,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,4.0126111958998,.333194761478692,.0206221051299754,4.80639993236566,.399108510989818,4.40969381865368,1.33269841309056,.110663133873164,.00684916764813009,3.91021223386622,.32469187015097,5.67229573503973,.95054294779362,.0789301318799674,3.37758202463881,.647664476709677},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,1.33269841309056,.110663133873164,.00684916764813009,.95054294779362,.0789301318799674,.647664476709677,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,3.30149822057238,.274146150459585,.0169674658481853,.000962479434277651,4.0126111958998,.333194761478692,.0206221051299754,3.91021223386622,.32469187015097,3.37758202463881,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,3.11944010216456,.259028610060664,.016031810358111,4.80639993236566,.399108510989818,5.67229573503973,.626153181688297,.0519938139714517,.00321800346775215,2.36319452013471,.196232327654089,4.40969381865368,.347192252628355,.028829765505355,1.58691847335237,.183906859961414},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,2.39387697960251,.198780103635972,.0123029070952925,2.07554270972356,.172346615354148,1.71909544841957,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,2.90029193047496,.240831227166574,.0149055371205287,.000845516535217038,4.28496925234002,.355810527926517,.0220218411618753,5.07587527588423,.421484906038046,5.32974583920556,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,1.98037740107218,.164444383862132,.0101777991856447,3.70920808974538,.308001110702629,5.32120685033247,.236321146004057,.0196233734207421,.0012145307084342,1.08420509999131,.0900290215693292,2.4592936780291,.0946962283839188,.00786328046747136,.526147189418459,.0362494356624709},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.7410698100085,.310646808870845,.0192265662361566,3.76406389114815,.31255616594975,3.6179005979755,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,2.19713005267825,.182442850414858,.0112917610860855,.000640525103711722,3.76697803717186,.31279814746903,.0193597169803656,5.17830525859385,.429990373428657,6.30978459469567,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.979365349166059,.0813234544750617,.00503327489386337,2.12867470143779,.176758530797994,3.543802987655,.0657431866024243,.0054591098696972,.000337875473054763,.350018452610409,.029064444374942,.921342735777177,.0171974308138302,.00142802120123345,.110884150015163,.00429748986298374},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,5.18623489409549,.430648825529863,.0266537364367886,5.81833621896758,.483136554903789,6.23568135536041,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,1.36595759702526,.11342487316276,.00702009734040598,.000398214994344052,2.61131941894657,.216835847896165,.0134204140287723,4.0025799484364,.332361798070895,5.43816471058122,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.339486246269225,.0281898826795267,.0017447294838579,.822758173894117,.0683192814389796,1.5272774292324,.0113956405123916,.000946258567734596,5.85658777413804e-005,.067649468233801,.00561740157357991,.198554676115027,.00166206328995312,.000138012569524986,.0119492023316399,.000231576690958128},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,6.45780747384476,.53623626019628,.0331887586818224,7.79428043908953,.647212821256298,8.98680735137037,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.612365059546825,.0508488911805658,.0031471418550943,.000178521609495778,1.2594400120878,.104579983939912,.00647266905916808,2.07683410541886,.172453848883548,3.03569721712443,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.0634197097850823,.00526617557576491,.000325934375061529,.165355473459674,.0137306045533387,.330223979417038,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,.00526617557576491,.000437286851447401,.0166285878943821,5.01144631725624e-005,4.16134925463982e-006,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,7.27823584467216,.60436208201944,.037405205103521,9.14177937313726,.759104944404841,10.9691712594614,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.122554596919714,.0101765528017885,.000629847662741689,3.57281062205031e-005,.262307293803493,.0217811823588299,.0013480819160108,.450140785461579,.037378291693961,.68472894504081,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,.00244089433681864,.0002026842788013,1.25445444792834e-005,.00662303395559898,.000549956153763738,.0137645079246526,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,3.89785419960172e-005,3.23665697396975e-006,.000128085303901421,6.8546484667622e-008,5.69188702865245e-009,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,6.8546484667622e-008,1.36592668764407e-008,7.65361974719019e-010,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,.122554596919714,.0244214704020099,.00362725225603475,.000493768468568094,.00244089433681864,.000486397330423004,7.22432263467335e-005,3.89785419960172e-005,7.76725911325977e-006,5.51740069821357e-007,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,.262307293803493,.0522700083866766,.00776351721285218,.00662303395559898,.00131977283399776,.000128085303901421,7.27823584467216,1.45033499905202,.215414174926983,.450140785461579,.0896996126351677,.0137645079246526,9.14177937313726,1.82168355953934,.68472894504081,10.9691712594614},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.000824565926330002,.000164311358893045,2.44047036237916e-005,5.01144631725624e-005,9.98631556618579e-006,2.90963994598978e-006,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,.612365059546825,.122026064732152,.0181241879095991,.00246719882612911,.0634197097850823,.0126376537772916,.00187703514229927,.00526617557576491,.00104939148922157,.000387613365377294,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,1.2594400120878,.250968774218159,.0372756855923469,.165355473459674,.0329504066094101,.0166285878943821,6.45780747384476,1.28684813137965,.191131930663375,2.07683410541886,.413851001242982,.330223979417038,7.79428043908953,1.55316727245144,3.03569721712443,8.98680735137037},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.0113956405123916,.00227081076025246,.000337277130218073,.00166206328995312,.000331199567025036,.000231576690958128,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,1.36595759702526,.272194547284149,.0404282899212952,.0055033985485992,.339486246269225,.067649468233801,.0100477850929001,.067649468233801,.0134805182990133,.0119492023316399,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,2.61131941894657,.520357958596769,.0772873028549269,.822758173894117,.16395112785691,.198554676115027,5.18623489409549,1.03346169878758,.153497157076452,4.0025799484364,.797594624863601,1.5272774292324,5.81833621896758,1.1594206118243,5.43816471058122,6.23568135536041},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.0657431866024243,.0131006533057736,.00194580315907169,.0171974308138302,.00342693426532766,.00429748986298374,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,2.19713005267825,.437822390179288,.0650285272089779,.00885216522776848,.979365349166059,.195158259887658,.0289863069976914,.350018452610409,.0697482223969557,.110884150015163,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.76697803717186,.750646201027513,.111491367273716,2.12867470143779,.424181283413309,.921342735777177,3.7410698100085,.745483465381799,.110724560685561,5.17830525859385,1.0318815597352,3.543802987655,3.76406389114815,.750065499032834,6.30978459469567,3.6179005979755},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,.236321146004057,.0470917453597858,.00699440438187064,.0946962283839188,.0188701296931614,.0362494356624709,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.90029193047496,.577941548644503,.0858400314014041,.0116851814693591,1.98037740107218,.394630061032062,.0586132921362272,1.08420509999131,.216049690735243,.526147189418459,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,4.28496925234002,.853866377517717,.126822369537332,3.70920808974538,.739134376376297,2.4592936780291,2.39387697960251,.477028175651108,.0708516521644813,5.07587527588423,1.01147032381073,5.32120685033247,2.07554270972356,.413593664547998,5.32974583920556,1.71909544841957},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,.626153181688297,.124773625580585,.0185322753878643,.347192252628355,.0691850451424154,.183906859961414,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,3.30149822057238,.657889978038246,.0977145465764208,.0133016285094573,3.11944010216456,.621611232907719,.0923261666744144,2.36319452013471,.470914078181719,1.58691847335237,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,4.0126111958998,.799593552359549,.118761379587061,4.80639993236566,.957771936938711,4.40969381865368,1.33269841309056,.26556698531571,.0394439167964793,3.91021223386622,.77918849841486,5.67229573503973,.95054294779362,.189414816301507,3.37758202463881,.647664476709677},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,1.33269841309056,.26556698531571,.0394439167964793,.95054294779362,.189414816301507,.647664476709677,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,3.30149822057238,.657889978038246,.0977145465764208,.0133016285094573,4.0126111958998,.799593552359549,.118761379587061,3.91021223386622,.77918849841486,3.37758202463881,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,3.11944010216456,.621611232907719,.0923261666744144,4.80639993236566,.957771936938711,5.67229573503973,.626153181688297,.124773625580585,.0185322753878643,2.36319452013471,.470914078181719,4.40969381865368,.347192252628355,.0691850451424154,1.58691847335237,.183906859961414},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,2.39387697960251,.477028175651108,.0708516521644813,2.07554270972356,.413593664547998,1.71909544841957,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,2.90029193047496,.577941548644503,.0858400314014041,.0116851814693591,4.28496925234002,.853866377517717,.126822369537332,5.07587527588423,1.01147032381073,5.32974583920556,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,1.98037740107218,.394630061032062,.0586132921362272,3.70920808974538,.739134376376297,5.32120685033247,.236321146004057,.0470917453597858,.00699440438187064,1.08420509999131,.216049690735243,2.4592936780291,.0946962283839188,.0188701296931614,.526147189418459,.0362494356624709},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.7410698100085,.745483465381799,.110724560685561,3.76406389114815,.750065499032834,3.6179005979755,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,2.19713005267825,.437822390179288,.0650285272089779,.00885216522776848,3.76697803717186,.750646201027513,.111491367273716,5.17830525859385,1.0318815597352,6.30978459469567,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.979365349166059,.195158259887658,.0289863069976914,2.12867470143779,.424181283413309,3.543802987655,.0657431866024243,.0131006533057736,.00194580315907169,.350018452610409,.0697482223969557,.921342735777177,.0171974308138302,.00342693426532766,.110884150015163,.00429748986298374},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,5.18623489409549,1.03346169878758,.153497157076452,5.81833621896758,1.1594206118243,6.23568135536041,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,1.36595759702526,.272194547284149,.0404282899212952,.0055033985485992,2.61131941894657,.520357958596769,.0772873028549269,4.0025799484364,.797594624863601,5.43816471058122,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.339486246269225,.067649468233801,.0100477850929001,.822758173894117,.16395112785691,1.5272774292324,.0113956405123916,.00227081076025246,.000337277130218073,.067649468233801,.0134805182990133,.198554676115027,.00166206328995312,.000331199567025036,.0119492023316399,.000231576690958128},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,6.45780747384476,1.28684813137965,.191131930663375,7.79428043908953,1.55316727245144,8.98680735137037,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.612365059546825,.122026064732152,.0181241879095991,.00246719882612911,1.2594400120878,.250968774218159,.0372756855923469,2.07683410541886,.413851001242982,3.03569721712443,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.0634197097850823,.0126376537772916,.00187703514229927,.165355473459674,.0329504066094101,.330223979417038,.000824565926330002,.000164311358893045,2.44047036237916e-005,.00526617557576491,.00104939148922157,.0166285878943821,5.01144631725624e-005,9.98631556618579e-006,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,7.27823584467216,1.45033499905202,.215414174926983,9.14177937313726,1.82168355953934,10.9691712594614,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.122554596919714,.0244214704020099,.00362725225603475,.000493768468568094,.262307293803493,.0522700083866766,.00776351721285218,.450140785461579,.0896996126351677,.68472894504081,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,.00244089433681864,.000486397330423004,7.22432263467335e-005,.00662303395559898,.00131977283399776,.0137645079246526,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,3.89785419960172e-005,7.76725911325977e-006,.000128085303901421,6.8546484667622e-008,1.36592668764407e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,6.8546484667622e-008,2.44980430684131e-008,7.65361974719019e-010,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,.122554596919714,.0438001716427634,.0116677137630873,.00284862553642039,.00244089433681864,.000872358879642074,.00023238342051707,3.89785419960172e-005,1.39306633365776e-005,5.51740069821357e-007,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,.262307293803493,.0937468261090737,.0249727590585012,.00662303395559898,.0023670268704951,.000128085303901421,7.27823584467216,2.60119152746664,.692918704338266,.450140785461579,.160877226679454,.0137645079246526,9.14177937313726,3.26720919228747,.68472894504081,10.9691712594614},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.000824565926330002,.000294694201614443,7.85021488046878e-005,5.01144631725624e-005,1.79105651165762e-005,2.90963994598978e-006,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,.612365059546825,.218855068600592,.0582997326325419,.0142336459837523,.0634197097850823,.0226657689113441,.00603782345911206,.00526617557576491,.00188209500049501,.000387613365377294,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,1.2594400120878,.450115214472287,.119903993192299,.165355473459674,.0590969110964026,.0166285878943821,6.45780747384476,2.30797605978546,.614810468242533,2.07683410541886,.74224625260403,.330223979417038,7.79428043908953,2.78562232344933,3.03569721712443,8.98680735137037},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.0113956405123916,.00407272369067121,.00108491296894855,.00166206328995312,.000594010010245393,.000231576690958128,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,1.36595759702526,.488183868334511,.130044915940983,.0317499448437878,.339486246269225,.121330053899551,.0323205203668776,.067649468233801,.0241774555590068,.0119492023316399,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,2.61131941894657,.933267634054527,.248608606068695,.822758173894117,.294048123349947,.198554676115027,5.18623489409549,1.85352474883709,.493751403486214,4.0025799484364,1.43049459738903,1.5272774292324,5.81833621896758,2.07943342346341,5.43816471058122,6.23568135536041},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.0657431866024243,.0234961635798143,.00625902823868614,.0171974308138302,.0061462437174735,.00429748986298374,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,2.19713005267825,.785239198263813,.209176034185977,.0510694901064878,.979365349166059,.350018452610409,.0932397057677362,.350018452610409,.125094192220009,.110884150015163,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.76697803717186,1.3462920908025,.358632173497365,2.12867470143779,.760773725550181,.921342735777177,3.7410698100085,1.33703266851113,.356165601240813,5.17830525859385,1.85069075233423,3.543802987655,3.76406389114815,1.34525059548623,6.30978459469567,3.6179005979755},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,.236321146004057,.0844595552914031,.0224987683542483,.0946962283839188,.0338437819621889,.0362494356624709,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.90029193047496,1.03654442641711,.276120006304964,.0674135925038969,1.98037740107218,.707773288179075,.188540268799867,1.08420509999131,.387487460083106,.526147189418459,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,4.28496925234002,1.53141859517937,.407947118664331,3.70920808974538,1.32564550862114,2.4592936780291,2.39387697960251,.855555199093713,.227907170151078,5.07587527588423,1.81408297965651,5.32120685033247,2.07554270972356,.741784716454702,5.32974583920556,1.71909544841957},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,.626153181688297,.223782848737467,.0596124199094679,.347192252628355,.12408412771496,.183906859961414,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,3.30149822057238,1.17993280035085,.314316534794822,.0767391217950806,3.11944010216456,1.11486647807198,.296983835024965,2.36319452013471,.844589499068632,1.58691847335237,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,4.0126111958998,1.43407969550076,.382017484674746,4.80639993236566,1.71777434126509,4.40969381865368,1.33269841309056,.476297264169728,.126878501518793,3.91021223386622,1.39748301027071,5.67229573503973,.95054294779362,.339717524338981,3.37758202463881,.647664476709677},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,1.33269841309056,.476297264169728,.126878501518793,.95054294779362,.339717524338981,.647664476709677,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,3.30149822057238,1.17993280035085,.314316534794822,.0767391217950806,4.0126111958998,1.43407969550076,.382017484674746,3.91021223386622,1.39748301027071,3.37758202463881,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,3.11944010216456,1.11486647807198,.296983835024965,4.80639993236566,1.71777434126509,5.67229573503973,.626153181688297,.223782848737467,.0596124199094679,2.36319452013471,.844589499068632,4.40969381865368,.347192252628355,.12408412771496,1.58691847335237,.183906859961414},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,2.39387697960251,.855555199093713,.227907170151078,2.07554270972356,.741784716454702,1.71909544841957,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,2.90029193047496,1.03654442641711,.276120006304964,.0674135925038969,4.28496925234002,1.53141859517937,.407947118664331,5.07587527588423,1.81408297965651,5.32974583920556,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,1.98037740107218,.707773288179075,.188540268799867,3.70920808974538,1.32564550862114,5.32120685033247,.236321146004057,.0844595552914031,.0224987683542483,1.08420509999131,.387487460083106,2.4592936780291,.0946962283839188,.0338437819621889,.526147189418459,.0362494356624709},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.7410698100085,1.33703266851113,.356165601240813,3.76406389114815,1.34525059548623,3.6179005979755,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,2.19713005267825,.785239198263813,.209176034185977,.0510694901064878,3.76697803717186,1.3462920908025,.358632173497365,5.17830525859385,1.85069075233423,6.30978459469567,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.979365349166059,.350018452610409,.0932397057677362,2.12867470143779,.760773725550181,3.543802987655,.0657431866024243,.0234961635798143,.00625902823868614,.350018452610409,.125094192220009,.921342735777177,.0171974308138302,.0061462437174735,.110884150015163,.00429748986298374},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,5.18623489409549,1.85352474883709,.493751403486214,5.81833621896758,2.07943342346341,6.23568135536041,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,1.36595759702526,.488183868334511,.130044915940983,.0317499448437878,2.61131941894657,.933267634054527,.248608606068695,4.0025799484364,1.43049459738903,5.43816471058122,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.339486246269225,.121330053899551,.0323205203668776,.822758173894117,.294048123349947,1.5272774292324,.0113956405123916,.00407272369067121,.00108491296894855,.067649468233801,.0241774555590068,.198554676115027,.00166206328995312,.000594010010245393,.0119492023316399,.000231576690958128},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,6.45780747384476,2.30797605978546,.614810468242533,7.79428043908953,2.78562232344933,8.98680735137037,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.612365059546825,.218855068600592,.0582997326325419,.0142336459837523,1.2594400120878,.450115214472287,.119903993192299,2.07683410541886,.74224625260403,3.03569721712443,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.0634197097850823,.0226657689113441,.00603782345911206,.165355473459674,.0590969110964026,.330223979417038,.000824565926330002,.000294694201614443,7.85021488046878e-005,.00526617557576491,.00188209500049501,.0166285878943821,5.01144631725624e-005,1.79105651165762e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,7.27823584467216,2.60119152746664,.692918704338266,9.14177937313726,3.26720919228747,10.9691712594614,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.122554596919714,.0438001716427634,.0116677137630873,.00284862553642039,.262307293803493,.0937468261090737,.0249727590585012,.450140785461579,.160877226679454,.68472894504081,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,.00244089433681864,.000872358879642074,.00023238342051707,.00662303395559898,.0023670268704951,.0137645079246526,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,3.89785419960172e-005,1.39306633365776e-005,.000128085303901421,6.8546484667622e-008,2.44980430684131e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,6.8546484667622e-008,3.7527416869482e-008,7.65361974719019e-010,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,.122554596919714,.0670954531185465,.0273791727414708,.0102396991398416,.00244089433681864,.00133632614020231,.000545305270748019,3.89785419960172e-005,2.13397375798647e-005,5.51740069821357e-007,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,.262307293803493,.143606418429484,.058600467750326,.00662303395559898,.00362593876834461,.000128085303901421,7.27823584467216,3.98464475452196,1.62598614322311,.450140785461579,.246440368054962,.0137645079246526,9.14177937313726,5.00488635016176,.68472894504081,10.9691712594614},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.000824565926330002,.000451428390509428,.000184211229066994,5.01144631725624e-005,2.74363646769969e-005,2.90963994598978e-006,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,.612365059546825,.335253937240527,.13680473166712,.051164412687175,.0634197097850823,.0347206410009163,.0141682093704184,.00526617557576491,.00288309410977452,.000387613365377294,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,1.2594400120878,.689510637467093,.281363788027607,.165355473459674,.0905278193943952,.0166285878943821,6.45780747384476,3.53548156799627,1.44269926011819,2.07683410541886,1.13701263662145,.330223979417038,7.79428043908953,4.26716574385505,3.03569721712443,8.98680735137037},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.0113956405123916,.00623881668046779,.00254583033055523,.00166206328995312,.000909936406629402,.000231576690958128,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,1.36595759702526,.747826244112006,.305160230186869,.114128683728465,.339486246269225,.18585988679871,.0758425453509597,.067649468233801,.0370363237207337,.0119492023316399,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,2.61131941894657,1.42962943840673,.58337889571007,.822758173894117,.450438693157099,.198554676115027,5.18623489409549,2.83932866528416,1.15862500931079,4.0025799484364,2.19130838066721,1.5272774292324,5.81833621896758,3.18538769471191,5.43816471058122,6.23568135536041},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.0657431866024243,.0359926841107618,.0146872831147845,.0171974308138302,.00941514591687767,.00429748986298374,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,2.19713005267825,1.20287153766575,.490847383612732,.183574923144401,.979365349166059,.536177046731461,.218794021040783,.350018452610409,.191625995851247,.110884150015163,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.76697803717186,2.06232246300417,.841557517454501,2.12867470143779,1.16539401381336,.921342735777177,3.7410698100085,2.04813838310324,.835769519114577,5.17830525859385,2.83498739737667,3.543802987655,3.76406389114815,2.06072704459494,6.30978459469567,3.6179005979755},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,.236321146004057,.129379678661691,.0527950614618356,.0946962283839188,.0518437211734884,.0362494356624709,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.90029193047496,1.58783437049499,.647936476974322,.242325604524073,1.98037740107218,1.08420509999131,.442423999509308,1.08420509999131,.593574082861527,.526147189418459,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,4.28496925234002,2.34590917578526,.957278765839435,3.70920808974538,2.03069492058023,2.4592936780291,2.39387697960251,1.31058536069927,.534801410737989,5.07587527588423,2.77890964691703,5.32120685033247,2.07554270972356,1.13630563078316,5.32974583920556,1.71909544841957},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,.626153181688297,.342802575265255,.139885051637275,.347192252628355,.190078724845143,.183906859961414,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,3.30149822057238,1.80748437551057,.737567519771828,.275847249626467,3.11944010216456,1.70781229176476,.69689502893851,2.36319452013471,1.29378751301169,1.58691847335237,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,4.0126111958998,2.19680022632121,.896432918697558,4.80639993236566,2.63137890697367,4.40969381865368,1.33269841309056,.729617706050062,.297729999278181,3.91021223386622,2.14073946020884,5.67229573503973,.95054294779362,.520397532085943,3.37758202463881,.647664476709677},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,1.33269841309056,.729617706050062,.297729999278181,.95054294779362,.520397532085943,.647664476709677,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,3.30149822057238,1.80748437551057,.737567519771828,.275847249626467,4.0126111958998,2.19680022632121,.896432918697558,3.91021223386622,2.14073946020884,3.37758202463881,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,3.11944010216456,1.70781229176476,.69689502893851,4.80639993236566,2.63137890697367,5.67229573503973,.626153181688297,.342802575265255,.139885051637275,2.36319452013471,1.29378751301169,4.40969381865368,.347192252628355,.190078724845143,1.58691847335237,.183906859961414},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,2.39387697960251,1.31058536069927,.534801410737989,2.07554270972356,1.13630563078316,1.71909544841957,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,2.90029193047496,1.58783437049499,.647936476974322,.242325604524073,4.28496925234002,2.34590917578526,.957278765839435,5.07587527588423,2.77890964691703,5.32974583920556,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,1.98037740107218,1.08420509999131,.442423999509308,3.70920808974538,2.03069492058023,5.32120685033247,.236321146004057,.129379678661691,.0527950614618356,1.08420509999131,.593574082861527,2.4592936780291,.0946962283839188,.0518437211734884,.526147189418459,.0362494356624709},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.7410698100085,2.04813838310324,.835769519114577,3.76406389114815,2.06072704459494,3.6179005979755,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,2.19713005267825,1.20287153766575,.490847383612732,.183574923144401,3.76697803717186,2.06232246300417,.841557517454501,5.17830525859385,2.83498739737667,6.30978459469567,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.979365349166059,.536177046731461,.218794021040783,2.12867470143779,1.16539401381336,3.543802987655,.0657431866024243,.0359926841107618,.0146872831147845,.350018452610409,.191625995851247,.921342735777177,.0171974308138302,.00941514591687767,.110884150015163,.00429748986298374},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,5.18623489409549,2.83932866528416,1.15862500931079,5.81833621896758,3.18538769471191,6.23568135536041,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,1.36595759702526,.747826244112006,.305160230186869,.114128683728465,2.61131941894657,1.42962943840673,.58337889571007,4.0025799484364,2.19130838066721,5.43816471058122,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.339486246269225,.18585988679871,.0758425453509597,.822758173894117,.450438693157099,1.5272774292324,.0113956405123916,.00623881668046779,.00254583033055523,.067649468233801,.0370363237207337,.198554676115027,.00166206328995312,.000909936406629402,.0119492023316399,.000231576690958128},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,6.45780747384476,3.53548156799627,1.44269926011819,7.79428043908953,4.26716574385505,8.98680735137037,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.612365059546825,.335253937240527,.13680473166712,.051164412687175,1.2594400120878,.689510637467093,.281363788027607,2.07683410541886,1.13701263662145,3.03569721712443,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.0634197097850823,.0347206410009163,.0141682093704184,.165355473459674,.0905278193943952,.330223979417038,.000824565926330002,.000451428390509428,.000184211229066994,.00526617557576491,.00288309410977452,.0166285878943821,5.01144631725624e-005,2.74363646769969e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,7.27823584467216,3.98464475452196,1.62598614322311,9.14177937313726,5.00488635016176,10.9691712594614,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.122554596919714,.0670954531185465,.0273791727414708,.0102396991398416,.262307293803493,.143606418429484,.058600467750326,.450140785461579,.246440368054962,.68472894504081,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,.00244089433681864,.00133632614020231,.000545305270748019,.00662303395559898,.00362593876834461,.0137645079246526,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,3.89785419960172e-005,2.13397375798647e-005,.000128085303901421,6.8546484667622e-008,3.7527416869482e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,6.8546484667622e-008,5.19287666029906e-008,7.65361974719019e-010,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,.122554596919714,.0928436971090411,.0524250669352539,.0271309626935909,.00244089433681864,.001849148543354,.00104413911950708,3.89785419960172e-005,2.95289776007745e-005,5.51740069821357e-007,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,.262307293803493,.19871616027429,.112206949174718,.00662303395559898,.00501741243403364,.000128085303901421,7.27823584467216,5.51377378765633,3.11340423610211,.450140785461579,.341013195733391,.0137645079246526,9.14177937313726,6.92553862835673,.68472894504081,10.9691712594614},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.000824565926330002,.000624666483447503,.000352723805983911,5.01144631725624e-005,3.79652184082228e-005,2.90963994598978e-006,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,.612365059546825,.463909453726739,.26195083695297,.135564507599069,.0634197097850823,.0480448752738173,.0271289907701198,.00526617557576491,.00398949710931715,.000387613365377294,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,1.2594400120878,.954114083786716,.538749492535305,.165355473459674,.125268361016134,.0166285878943821,6.45780747384476,4.89224179250545,2.76245034842036,2.07683410541886,1.57334740190396,.330223979417038,7.79428043908953,5.90471373874725,3.03569721712443,8.98680735137037},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.0113956405123916,.00863299641447793,.00487470263420339,.00166206328995312,.00125912943722616,.000231576690958128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,1.36595759702526,1.03480862072507,.584314421936207,.302393753783036,.339486246269225,.25718462626926,.145221718497785,.067649468233801,.0512492137952295,.0119492023316399,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,2.61131941894657,1.97825748822484,1.11704169972566,.822758173894117,.623296984369123,.198554676115027,5.18623489409549,3.92893643816517,2.21851091848272,4.0025799484364,3.03223485513554,1.5272774292324,5.81833621896758,4.40779749680161,5.43816471058122,6.23568135536041},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.0657431866024243,.0498050718253105,.0281229023119214,.0171974308138302,.0130282592204804,.00429748986298374,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,2.19713005267825,1.66448001337449,.939864296992246,.48639753212384,.979365349166059,.741937896083684,.41894221225766,.350018452610409,.265163510940112,.110884150015163,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.76697803717186,2.85374989208327,1.61139672111394,2.12867470143779,1.61262028650446,.921342735777177,3.7410698100085,2.8341225946006,1.60031398303209,5.17830525859385,3.92292918515564,3.543802987655,3.76406389114815,2.85154222320761,6.30978459469567,3.6179005979755},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,.547225517982271,.414561685224313,.234086155283846,.121144008364367,.236321146004057,.179029831969503,.101090878717283,.0946962283839188,.0717390303204678,.0362494356624709,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.90029193047496,2.19717442094167,1.24065520517793,.642062510456358,1.98037740107218,1.50027468671861,.847144214584193,1.08420509999131,.821361356090746,.526147189418459,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,4.28496925234002,3.24616454377475,1.83297734554319,3.70920808974538,2.80998510873911,2.4592936780291,2.39387697960251,1.81352960012534,1.02402654819954,5.07587527588423,3.84533129384947,5.32120685033247,2.07554270972356,1.57236907874115,5.32974583920556,1.71909544841957},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,.626153181688297,.474354922529452,.267848969162523,.347192252628355,.263022466294621,.183906859961414,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,3.30149822057238,2.50111630653613,1.41227885000115,.730880989425352,3.11944010216456,2.36319452013471,1.33440001580502,2.36319452013471,1.79028548729353,1.58691847335237,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,4.0126111958998,3.03983422634135,1.71647099090401,4.80639993236566,3.6411848323989,4.40969381865368,1.33269841309056,1.0096124574376,.570087171302651,3.91021223386622,2.96225983721782,5.67229573503973,.95054294779362,.720102907121204,3.37758202463881,.647664476709677},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,1.33269841309056,1.0096124574376,.570087171302651,.95054294779362,.720102907121204,.647664476709677,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,3.30149822057238,2.50111630653613,1.41227885000115,.730880989425352,4.0126111958998,3.03983422634135,1.71647099090401,3.91021223386622,2.96225983721782,3.37758202463881,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,3.11944010216456,2.36319452013471,1.33440001580502,4.80639993236566,3.6411848323989,5.67229573503973,.626153181688297,.474354922529452,.267848969162523,2.36319452013471,1.79028548729353,4.40969381865368,.347192252628355,.263022466294621,1.58691847335237,.183906859961414},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,2.39387697960251,1.81352960012534,1.02402654819954,2.07554270972356,1.57236907874115,1.71909544841957,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,2.90029193047496,2.19717442094167,1.24065520517793,.642062510456358,4.28496925234002,3.24616454377475,1.83297734554319,5.07587527588423,3.84533129384947,5.32974583920556,.547225517982271,.414561685224313,.234086155283846,.121144008364367,1.98037740107218,1.50027468671861,.847144214584193,3.70920808974538,2.80998510873911,5.32120685033247,.236321146004057,.179029831969503,.101090878717283,1.08420509999131,.821361356090746,2.4592936780291,.0946962283839188,.0717390303204678,.526147189418459,.0362494356624709},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.7410698100085,2.8341225946006,1.60031398303209,3.76406389114815,2.85154222320761,3.6179005979755,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,2.19713005267825,1.66448001337449,.939864296992246,.48639753212384,3.76697803717186,2.85374989208327,1.61139672111394,5.17830525859385,3.92292918515564,6.30978459469567,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.979365349166059,.741937896083684,.41894221225766,2.12867470143779,1.61262028650446,3.543802987655,.0657431866024243,.0498050718253105,.0281229023119214,.350018452610409,.265163510940112,.921342735777177,.0171974308138302,.0130282592204804,.110884150015163,.00429748986298374},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,5.18623489409549,3.92893643816517,2.21851091848272,5.81833621896758,4.40779749680161,6.23568135536041,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,1.36595759702526,1.03480862072507,.584314421936207,.302393753783036,2.61131941894657,1.97825748822484,1.11704169972566,4.0025799484364,3.03223485513554,5.43816471058122,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.339486246269225,.25718462626926,.145221718497785,.822758173894117,.623296984369123,1.5272774292324,.0113956405123916,.00863299641447793,.00487470263420339,.067649468233801,.0512492137952295,.198554676115027,.00166206328995312,.00125912943722616,.0119492023316399,.000231576690958128},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,6.45780747384476,4.89224179250545,2.76245034842036,7.79428043908953,5.90471373874725,8.98680735137037,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.612365059546825,.463909453726739,.26195083695297,.135564507599069,1.2594400120878,.954114083786716,.538749492535305,2.07683410541886,1.57334740190396,3.03569721712443,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.0634197097850823,.0480448752738173,.0271289907701198,.165355473459674,.125268361016134,.330223979417038,.000824565926330002,.000624666483447503,.000352723805983911,.00526617557576491,.00398949710931715,.0166285878943821,5.01144631725624e-005,3.79652184082228e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,7.27823584467216,5.51377378765633,3.11340423610211,9.14177937313726,6.92553862835673,10.9691712594614,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.122554596919714,.0928436971090411,.0524250669352539,.0271309626935909,.262307293803493,.19871616027429,.112206949174718,.450140785461579,.341013195733391,.68472894504081,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,.00244089433681864,.001849148543354,.00104413911950708,.00662303395559898,.00501741243403364,.0137645079246526,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,3.89785419960172e-005,2.95289776007745e-005,.000128085303901421,6.8546484667622e-008,5.19287666029906e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,6.8546484667622e-008,6.67972275267736e-008,7.65361974719019e-010,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,.122554596919714,.119427091493105,.0867440815154734,.057745280204241,.00244089433681864,.00237860446263913,.00172766377213932,3.89785419960172e-005,3.79838375617868e-005,5.51740069821357e-007,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,.262307293803493,.255613399651286,.18566096935718,.00662303395559898,.00645401887770827,.000128085303901421,7.27823584467216,7.09250048322984,5.15153163620379,.450140785461579,.438653515481796,.0137645079246526,9.14177937313726,8.90848771819592,.68472894504081,10.9691712594614},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.000824565926330002,.000803523594969987,.000583627344081487,5.01144631725624e-005,4.88355780044979e-005,2.90963994598978e-006,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,.612365059546825,.596737942368568,.433431678432723,.288534194877923,.0634197097850823,.0618012842418601,.0448884383820082,.00526617557576491,.00513178655225592,.000387613365377294,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,1.2594400120878,1.22730000557206,.891431001025123,.165355473459674,.161135720335381,.0166285878943821,6.45780747384476,6.29300885631198,4.57083285180669,2.07683410541886,2.02383478866576,.330223979417038,7.79428043908953,7.59537599078757,3.03569721712443,8.98680735137037},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.0113956405123916,.0111048319353399,.00806582856989507,.00166206328995312,.00161964862649386,.000231576690958128,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,1.36595759702526,1.33109933870947,.966824094086502,.643611958831769,.339486246269225,.33082280056131,.24028819272175,.067649468233801,.0659231023263137,.0119492023316399,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,2.61131941894657,2.54468041880833,1.84829055872937,.822758173894117,.801761975206599,.198554676115027,5.18623489409549,5.05388589728677,3.6708144253493,4.0025799484364,3.90043697901805,1.5272774292324,5.81833621896758,5.66985645040998,5.43816471058122,6.23568135536041},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.0657431866024243,.0640654676075239,.0465329941039433,.0171974308138302,.0167585647100291,.00429748986298374,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,2.19713005267825,2.14106086934751,1.55512753646012,1.0352438319403,.979365349166059,.954372647329225,.693194296828135,.350018452610409,.341086232751863,.110884150015163,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.76697803717186,3.67084745731307,2.66626514187773,2.12867470143779,2.07435244929784,.921342735777177,3.7410698100085,3.6456003901969,2.64792731244254,5.17830525859385,5.04615862255375,3.543802987655,3.76406389114815,3.66800767948205,6.30978459469567,3.6179005979755},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,.547225517982271,.533260715077649,.387325943916553,.257841742921929,.236321146004057,.230290399761936,.167267987177829,.0946962283839188,.092279648532812,.0362494356624709,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.90029193047496,2.82627856027675,2.05282515677973,1.36655967551418,1.98037740107218,1.92983958905795,1.40171011886371,1.08420509999131,1.05653696348977,.526147189418459,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,4.28496925234002,4.17561990789474,3.03289906094317,3.70920808974538,3.61455175876414,2.4592936780291,2.39387697960251,2.33278695510428,1.69438491095747,5.07587527588423,4.94634257953458,5.32120685033247,2.07554270972356,2.02257634793674,5.32974583920556,1.71909544841957},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,.626153181688297,.610174201341807,.443190904144467,.347192252628355,.338332155278326,.183906859961414,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,3.30149822057238,3.21724635356535,2.3367987653383,1.55560007239588,3.11944010216456,3.03983422634135,2.20793812029263,2.36319452013471,2.30288749099116,1.58691847335237,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,4.0126111958998,3.91021223386622,2.84012416689539,4.80639993236566,4.6837440522375,4.40969381865368,1.33269841309056,1.29868890491693,.943283260267015,3.91021223386622,3.8104264222726,5.67229573503973,.95054294779362,.926285773305773,3.37758202463881,.647664476709677},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,1.33269841309056,1.29868890491693,.943283260267015,.95054294779362,.926285773305773,.647664476709677,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,3.30149822057238,3.21724635356535,2.3367987653383,1.55560007239588,4.0126111958998,3.91021223386622,2.84012416689539,3.91021223386622,3.8104264222726,3.37758202463881,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,3.11944010216456,3.03983422634135,2.20793812029263,4.80639993236566,4.6837440522375,5.67229573503973,.626153181688297,.610174201341807,.443190904144467,2.36319452013471,2.30288749099116,4.40969381865368,.347192252628355,.338332155278326,1.58691847335237,.183906859961414},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,2.39387697960251,2.33278695510428,1.69438491095747,2.07554270972356,2.02257634793674,1.71909544841957,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,2.90029193047496,2.82627856027675,2.05282515677973,1.36655967551418,4.28496925234002,4.17561990789474,3.03289906094317,5.07587527588423,4.94634257953458,5.32974583920556,.547225517982271,.533260715077649,.387325943916553,.257841742921929,1.98037740107218,1.92983958905795,1.40171011886371,3.70920808974538,3.61455175876414,5.32120685033247,.236321146004057,.230290399761936,.167267987177829,1.08420509999131,1.05653696348977,2.4592936780291,.0946962283839188,.092279648532812,.526147189418459,.0362494356624709},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.7410698100085,3.6456003901969,2.64792731244254,3.76406389114815,3.66800767948205,3.6179005979755,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,2.19713005267825,2.14106086934751,1.55512753646012,1.0352438319403,3.76697803717186,3.67084745731307,2.66626514187773,5.17830525859385,5.04615862255375,6.30978459469567,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.979365349166059,.954372647329225,.693194296828135,2.12867470143779,2.07435244929784,3.543802987655,.0657431866024243,.0640654676075239,.0465329941039433,.350018452610409,.341086232751863,.921342735777177,.0171974308138302,.0167585647100291,.110884150015163,.00429748986298374},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,5.18623489409549,5.05388589728677,3.6708144253493,5.81833621896758,5.66985645040998,6.23568135536041,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,1.36595759702526,1.33109933870947,.966824094086502,.643611958831769,2.61131941894657,2.54468041880833,1.84829055872937,4.0025799484364,3.90043697901805,5.43816471058122,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.339486246269225,.33082280056131,.24028819272175,.822758173894117,.801761975206599,1.5272774292324,.0113956405123916,.0111048319353399,.00806582856989507,.067649468233801,.0659231023263137,.198554676115027,.00166206328995312,.00161964862649386,.0119492023316399,.000231576690958128},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,6.45780747384476,6.29300885631198,4.57083285180669,7.79428043908953,7.59537599078757,8.98680735137037,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.612365059546825,.596737942368568,.433431678432723,.288534194877923,1.2594400120878,1.22730000557206,.891431001025123,2.07683410541886,2.02383478866576,3.03569721712443,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.0634197097850823,.0618012842418601,.0448884383820082,.165355473459674,.161135720335381,.330223979417038,.000824565926330002,.000803523594969987,.000583627344081487,.00526617557576491,.00513178655225592,.0166285878943821,5.01144631725624e-005,4.88355780044979e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,7.27823584467216,7.09250048322984,5.15153163620379,9.14177937313726,8.90848771819592,10.9691712594614,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.122554596919714,.119427091493105,.0867440815154734,.057745280204241,.262307293803493,.255613399651286,.18566096935718,.450140785461579,.438653515481796,.68472894504081,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,.00244089433681864,.00237860446263913,.00172766377213932,.00662303395559898,.00645401887770827,.0137645079246526,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,3.89785419960172e-005,3.79838375617868e-005,.000128085303901421,6.8546484667622e-008,6.67972275267736e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,6.8546484667622e-008,8.11985772602823e-008,7.65361974719019e-010,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,.122554596919714,.1451753354836,.128179867438098,.103725715888757,.00244089433681864,.00289142686579082,.00255293167466323,3.89785419960172e-005,4.61730775826967e-005,5.51740069821357e-007,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,.262307293803493,.310723141496092,.274347229515445,.00662303395559898,.00784549254339729,.000128085303901421,7.27823584467216,8.62162951636421,7.61230772976691,.450140785461579,.533226343160225,.0137645079246526,9.14177937313726,10.8291399963909,.68472894504081,10.9691712594614},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.000824565926330002,.000976761687908062,.000862413599759767,5.01144631725624e-005,5.93644317357238e-005,2.90963994598978e-006,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,.612365059546825,.72539345885478,.640472688330538,.518283326641486,.0634197097850823,.0751255185147611,.066330681941484,.00526617557576491,.00623818955179855,.000387613365377294,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,1.2594400120878,1.49190345189168,1.31724845713225,.165355473459674,.19587626195712,.0166285878943821,6.45780747384476,7.64976908082116,6.7542215997961,2.07683410541886,2.46016955394827,.330223979417038,7.79428043908953,9.23292398567978,3.03569721712443,8.98680735137037},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.0113956405123916,.0134990116693501,.0119187017581499,.00166206328995312,.00196884165709061,.000231576690958128,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,1.36595759702526,1.61808171532254,1.4286552125618,1.15609641079355,.339486246269225,.402147540031861,.355068704998853,.067649468233801,.0801359924008096,.0119492023316399,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,2.61131941894657,3.09330846862644,2.73117928815418,.822758173894117,.974620266418623,.198554676115027,5.18623489409549,6.14349367016778,5.42428368841739,4.0025799484364,4.74136345348638,1.5272774292324,5.81833621896758,6.89226625249968,5.43816471058122,6.23568135536041},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,.23320173943763,.276245376362,.243905726875982,.197373399115395,.0657431866024243,.0778778553220726,.0687608066341364,.0171974308138302,.0203716780136318,.00429748986298374,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,2.19713005267825,2.60266934505625,2.29797858240319,1.85957029228411,.979365349166059,1.16013349668145,1.02431833416123,.350018452610409,.414623747840729,.110884150015163,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.76697803717186,4.46227488639218,3.93988277320967,2.12867470143779,2.52157872198894,.921342735777177,3.7410698100085,4.43158460169426,3.91278535624427,5.17830525859385,6.13410041033272,3.543802987655,3.76406389114815,4.45882285809472,6.30978459469567,3.6179005979755},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,.547225517982271,.648230667294543,.572343234018839,.463151607819505,.236321146004057,.279940553069749,.247168314523763,.0946962283839188,.112174957679791,.0362494356624709,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.90029193047496,3.43561861072344,3.03341567369851,2.45470071573063,1.98037740107218,2.34590917578526,2.07127695727047,1.08420509999131,1.28432423671899,.526147189418459,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,4.28496925234002,5.07587527588423,4.48164984622588,3.70920808974538,4.39384194692302,2.4592936780291,2.39387697960251,2.83573119453035,2.50375621577018,5.07587527588423,6.01276422646702,5.32120685033247,2.07554270972356,2.45863979589474,5.32974583920556,1.71909544841957},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,.626153181688297,.741726548606004,.654893686699248,.347192252628355,.411275896727804,.183906859961414,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,3.30149822057238,3.91087828459092,3.45303738004469,2.79426700459607,3.11944010216456,3.6952164547113,3.26262277063996,2.36319452013471,2.79938546527299,1.58691847335237,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,4.0126111958998,4.75324623388636,4.19679052288369,4.80639993236566,5.69354997766273,4.40969381865368,1.33269841309056,1.57868365630447,1.39386942769156,3.91021223386622,4.63194679928158,5.67229573503973,.95054294779362,1.12599114834103,3.37758202463881,.647664476709677},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,1.33269841309056,1.57868365630447,1.39386942769156,.95054294779362,1.12599114834103,.647664476709677,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,3.30149822057238,3.91087828459092,3.45303738004469,2.79426700459607,4.0126111958998,4.75324623388636,4.19679052288369,3.91021223386622,4.63194679928158,3.37758202463881,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,3.11944010216456,3.6952164547113,3.26262277063996,4.80639993236566,5.69354997766273,5.67229573503973,.626153181688297,.741726548606004,.654893686699248,2.36319452013471,2.79938546527299,4.40969381865368,.347192252628355,.411275896727804,1.58691847335237,.183906859961414},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,2.39387697960251,2.83573119453035,2.50375621577018,2.07554270972356,2.45863979589474,1.71909544841957,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,2.90029193047496,3.43561861072344,3.03341567369851,2.45470071573063,4.28496925234002,5.07587527588423,4.48164984622588,5.07587527588423,6.01276422646702,5.32974583920556,.547225517982271,.648230667294543,.572343234018839,.463151607819505,1.98037740107218,2.34590917578526,2.07127695727047,3.70920808974538,4.39384194692302,5.32120685033247,.236321146004057,.279940553069749,.247168314523763,1.08420509999131,1.28432423671899,2.4592936780291,.0946962283839188,.112174957679791,.526147189418459,.0362494356624709},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.7410698100085,4.43158460169426,3.91278535624427,3.76406389114815,4.45882285809472,3.6179005979755,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,2.19713005267825,2.60266934505625,2.29797858240319,1.85957029228411,3.76697803717186,4.46227488639218,3.93988277320967,5.17830525859385,6.13410041033272,6.30978459469567,.23320173943763,.276245376362,.243905726875982,.197373399115395,.979365349166059,1.16013349668145,1.02431833416123,2.12867470143779,2.52157872198894,3.543802987655,.0657431866024243,.0778778553220726,.0687608066341364,.350018452610409,.414623747840729,.921342735777177,.0171974308138302,.0203716780136318,.110884150015163,.00429748986298374},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,5.18623489409549,6.14349367016778,5.42428368841739,5.81833621896758,6.89226625249968,6.23568135536041,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,1.36595759702526,1.61808171532254,1.4286552125618,1.15609641079355,2.61131941894657,3.09330846862644,2.73117928815418,4.0025799484364,4.74136345348638,5.43816471058122,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.339486246269225,.402147540031861,.355068704998853,.822758173894117,.974620266418623,1.5272774292324,.0113956405123916,.0134990116693501,.0119187017581499,.067649468233801,.0801359924008096,.198554676115027,.00166206328995312,.00196884165709061,.0119492023316399,.000231576690958128},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,6.45780747384476,7.64976908082116,6.7542215997961,7.79428043908953,9.23292398567978,8.98680735137037,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.612365059546825,.72539345885478,.640472688330538,.518283326641486,1.2594400120878,1.49190345189168,1.31724845713225,2.07683410541886,2.46016955394827,3.03569721712443,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.0634197097850823,.0751255185147611,.066330681941484,.165355473459674,.19587626195712,.330223979417038,.000824565926330002,.000976761687908062,.000862413599759767,.00526617557576491,.00623818955179855,.0166285878943821,5.01144631725624e-005,5.93644317357238e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,7.27823584467216,8.62162951636421,7.61230772976691,9.14177937313726,10.8291399963909,10.9691712594614,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.122554596919714,.1451753354836,.128179867438098,.103725715888757,.262307293803493,.310723141496092,.274347229515445,.450140785461579,.533226343160225,.68472894504081,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,.00244089433681864,.00289142686579082,.00255293167466323,.00662303395559898,.00784549254339729,.0137645079246526,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,3.89785419960172e-005,4.61730775826967e-005,.000128085303901421,6.8546484667622e-008,8.11985772602823e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,6.8546484667622e-008,9.42279510613512e-008,7.65361974719019e-010,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,.122554596919714,.168470616959383,.172616566596884,.162099033300108,.00244089433681864,.00335539412635106,.00343796813996252,3.89785419960172e-005,5.35821518259837e-005,5.51740069821357e-007,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,.262307293803493,.360582733816502,.369456434624521,.00662303395559898,.0091044044412468,.000128085303901421,7.27823584467216,10.0050827434195,10.2513011633895,.450140785461579,.618789484535732,.0137645079246526,9.14177937313726,12.5668171542652,.68472894504081,10.9691712594614},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.000824565926330002,.00113349587680305,.00116139045508752,5.01144631725624e-005,6.88902312961445e-005,2.90963994598978e-006,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,.612365059546825,.841792327494715,.862508275818624,.809955617122475,.0634197097850823,.0871803906043333,.0893258419252021,.00526617557576491,.00723918866107805,.000387613365377294,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,1.2594400120878,1.73129887488649,1.77390498656132,.165355473459674,.227307170255112,.0166285878943821,6.45780747384476,8.87727458903197,9.0957384017764,2.07683410541886,2.85493593796569,.330223979417038,7.79428043908953,10.7144674060855,3.03569721712443,8.98680735137037},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.0113956405123916,.0156651046591466,.0160506124472133,.00166206328995312,.00228476805347462,.000231576690958128,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,1.36595759702526,1.87772409110003,1.92393362992246,1.80670828815818,.339486246269225,.46667737293102,.478161992112396,.067649468233801,.0929948605625365,.0119492023316399,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,2.61131941894657,3.58967027297864,3.67800966645072,.822758173894117,1.13101083622578,.198554676115027,5.18623489409549,7.12929758661485,7.30474484999971,4.0025799484364,5.50217723676456,1.5272774292324,5.81833621896758,7.99822052374818,5.43816471058122,6.23568135536041},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.0657431866024243,.0903743758530202,.0925984290266871,.0171974308138302,.023640580213036,.00429748986298374,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,2.19713005267825,3.02030168445819,3.09462929659509,2.90607342788681,.979365349166059,1.3462920908025,1.37942344206699,.350018452610409,.481155551471966,.110884150015163,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.76697803717186,5.17830525859385,5.30573990048865,2.12867470143779,2.92619901025212,.921342735777177,3.7410698100085,5.14269031628637,5.26924849841671,5.17830525859385,7.11839705537515,3.543802987655,3.76406389114815,5.17429930720343,6.30978459469567,3.6179005979755},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,.547225517982271,.752247756737171,.770760072903028,.723797635481732,.236321146004057,.324860676440036,.332855281233889,.0946962283839188,.130174896891091,.0362494356624709,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.90029193047496,3.98690855480132,4.08502371800223,3.83612308338039,1.98037740107218,2.72234098759749,2.78933598529375,1.08420509999131,1.49041085949741,.526147189418459,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,4.28496925234002,5.89036585649012,6.03532383522368,3.70920808974538,5.09889135888211,2.4592936780291,2.39387697960251,3.29076135613591,3.37174480048962,5.07587527588423,6.97759089372753,5.32120685033247,2.07554270972356,2.8531607102232,5.32974583920556,1.71909544841957},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,.626153181688297,.860746275133792,.881928667453122,.347192252628355,.477270493857987,.183906859961414,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,3.30149822057238,4.53842985975064,4.65011759480767,4.36678577097684,3.11944010216456,4.28816226840408,4.39369108444944,2.36319452013471,3.24858347921606,1.58691847335237,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,4.0126111958998,5.51596676470681,5.65171103126921,4.80639993236566,6.60715454337131,4.40969381865368,1.33269841309056,1.83200409818481,1.87708849838432,3.91021223386622,5.37520324921971,5.67229573503973,.95054294779362,1.306671156088,3.37758202463881,.647664476709677},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,1.33269841309056,1.83200409818481,1.87708849838432,.95054294779362,1.306671156088,.647664476709677,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,3.30149822057238,4.53842985975064,4.65011759480767,4.36678577097684,4.0126111958998,5.51596676470681,5.65171103126921,3.91021223386622,5.37520324921971,3.37758202463881,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,3.11944010216456,4.28816226840408,4.39369108444944,4.80639993236566,6.60715454337131,5.67229573503973,.626153181688297,.860746275133792,.881928667453122,2.36319452013471,3.24858347921606,4.40969381865368,.347192252628355,.477270493857987,1.58691847335237,.183906859961414},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,2.39387697960251,3.29076135613591,3.37174480048962,2.07554270972356,2.8531607102232,1.71909544841957,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,2.90029193047496,3.98690855480132,4.08502371800223,3.83612308338039,4.28496925234002,5.89036585649012,6.03532383522368,5.07587527588423,6.97759089372753,5.32974583920556,.547225517982271,.752247756737171,.770760072903028,.723797635481732,1.98037740107218,2.72234098759749,2.78933598529375,3.70920808974538,5.09889135888211,5.32120685033247,.236321146004057,.324860676440036,.332855281233889,1.08420509999131,1.49041085949741,2.4592936780291,.0946962283839188,.130174896891091,.526147189418459,.0362494356624709},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.7410698100085,5.14269031628637,5.26924849841671,3.76406389114815,5.17429930720343,3.6179005979755,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,2.19713005267825,3.02030168445819,3.09462929659509,2.90607342788681,3.76697803717186,5.17830525859385,5.30573990048865,5.17830525859385,7.11839705537515,6.30978459469567,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.979365349166059,1.3462920908025,1.37942344206699,2.12867470143779,2.92619901025212,3.543802987655,.0657431866024243,.0903743758530202,.0925984290266871,.350018452610409,.481155551471966,.921342735777177,.0171974308138302,.023640580213036,.110884150015163,.00429748986298374},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,5.18623489409549,7.12929758661485,7.30474484999971,5.81833621896758,7.99822052374818,6.23568135536041,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,1.36595759702526,1.87772409110003,1.92393362992246,1.80670828815818,2.61131941894657,3.58967027297864,3.67800966645072,4.0025799484364,5.50217723676456,5.43816471058122,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.339486246269225,.46667737293102,.478161992112396,.822758173894117,1.13101083622578,1.5272774292324,.0113956405123916,.0156651046591466,.0160506124472133,.067649468233801,.0929948605625365,.198554676115027,.00166206328995312,.00228476805347462,.0119492023316399,.000231576690958128},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,6.45780747384476,8.87727458903197,9.0957384017764,7.79428043908953,10.7144674060855,8.98680735137037,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.612365059546825,.841792327494715,.862508275818624,.809955617122475,1.2594400120878,1.73129887488649,1.77390498656132,2.07683410541886,2.85493593796569,3.03569721712443,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.0634197097850823,.0871803906043333,.0893258419252021,.165355473459674,.227307170255112,.330223979417038,.000824565926330002,.00113349587680305,.00116139045508752,.00526617557576491,.00723918866107805,.0166285878943821,5.01144631725624e-005,6.88902312961445e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,7.27823584467216,10.0050827434195,10.2513011633895,9.14177937313726,12.5668171542652,10.9691712594614,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.122554596919714,.168470616959383,.172616566596884,.162099033300108,.262307293803493,.360582733816502,.369456434624521,.450140785461579,.618789484535732,.68472894504081,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,.00244089433681864,.00335539412635106,.00343796813996252,.00662303395559898,.0091044044412468,.0137645079246526,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,3.89785419960172e-005,5.35821518259837e-005,.000128085303901421,6.8546484667622e-008,9.42279510613512e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,6.8546484667622e-008,1.05066727253324e-007,7.65361974719019e-010,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,.122554596919714,.187849318200136,.214611696543555,.22471744861619,.00244089433681864,.00374135567557013,.00427437638070464,3.89785419960172e-005,5.97455560493016e-005,5.51740069821357e-007,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,.262307293803493,.402059551538899,.459339875638177,.00662303395559898,.0101516584777441,.000128085303901421,7.27823584467216,11.1559392718342,12.7452954129935,.450140785461579,.689967098580019,.0137645079246526,9.14177937313726,14.0123427870133,.68472894504081,10.9691712594614},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.000824565926330002,.00126387871952445,.00144394006224153,5.01144631725624e-005,7.68144808465349e-005,2.90963994598978e-006,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,.612365059546825,.938621331363155,1.07234414405062,1.12283926724684,.0634197097850823,.0972085057383859,.111057535546505,.00526617557576491,.0080718921723515,.000387613365377294,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,1.2594400120878,1.93044531514061,2.2054705766571,.165355473459674,.253453674742105,.0166285878943821,6.45780747384476,9.89840251743778,11.3086008382978,2.07683410541886,3.18333118932674,.330223979417038,7.79428043908953,11.9469224570834,3.03569721712443,8.98680735137037},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.0113956405123916,.0174670175895654,.0199554957891379,.00166206328995312,.00254757849669498,.000231576690958128,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,1.36595759702526,2.0937134121504,2.39199903285711,2.50463478185567,.339486246269225,.520357958596769,.594491933034081,.067649468233801,.10369179782253,.0119492023316399,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,2.61131941894657,4.0025799484364,4.57281656090381,.822758173894117,1.26110783171881,.198554676115027,5.18623489409549,7.94936063666436,9.0818842940915,4.0025799484364,6.13507720928999,1.5272774292324,5.81833621896758,8.91823333538729,5.43816471058122,6.23568135536041},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.0657431866024243,.100769886127061,.115126296058794,.0171974308138302,.0263598896651818,.00429748986298374,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,2.19713005267825,3.36771849254271,3.8475081309354,4.02868168249338,.979365349166059,1.50115228352525,1.71501734155852,.350018452610409,.53650152129502,.110884150015163,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.76697803717186,5.77395114836884,6.59655016845447,2.12867470143779,3.262791452389,.921342735777177,3.7410698100085,5.7342395194157,6.55118093268348,5.17830525859385,7.93720624797419,3.543802987655,3.76406389114815,5.76948440365682,6.30978459469567,3.6179005979755},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,.236321146004057,.362228486371654,.41383418762992,.0946962283839188,.145148549160118,.0362494356624709,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.90029193047496,4.44551143257392,5.07885128191947,5.3180069882276,1.98037740107218,3.0354842147445,3.46794134442417,1.08420509999131,1.66184862884528,.526147189418459,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,4.28496925234002,6.56791807415177,7.50363138951749,3.70920808974538,5.68540249112695,2.4592936780291,2.39387697960251,3.66928837957851,4.19204185444651,5.07587527588423,7.78020354957332,5.32120685033247,2.07554270972356,3.1813517621299,5.32974583920556,1.71909544841957},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,.626153181688297,.959755498290674,1.09648923787555,.347192252628355,.532169576430532,.183906859961414,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,3.30149822057238,5.06047268206324,5.78142437787735,6.0536632275323,3.11944010216456,4.78141751356834,5.46261298270436,2.36319452013471,3.62225890010297,1.58691847335237,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,4.0126111958998,6.15045290784802,7.02669110333533,4.80639993236566,7.36715694769769,4.40969381865368,1.33269841309056,2.04273437703882,2.33375715403626,3.91021223386622,5.99349776107556,5.67229573503973,.95054294779362,1.45697386412547,3.37758202463881,.647664476709677},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,1.33269841309056,2.04273437703882,2.33375715403626,.95054294779362,1.45697386412547,.647664476709677,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,3.30149822057238,5.06047268206324,5.78142437787735,6.0536632275323,4.0126111958998,6.15045290784802,7.02669110333533,3.91021223386622,5.99349776107556,3.37758202463881,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,3.11944010216456,4.78141751356834,5.46261298270436,4.80639993236566,7.36715694769769,5.67229573503973,.626153181688297,.959755498290674,1.09648923787555,2.36319452013471,3.62225890010297,4.40969381865368,.347192252628355,.532169576430532,1.58691847335237,.183906859961414},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,2.39387697960251,3.66928837957851,4.19204185444651,2.07554270972356,3.1813517621299,1.71909544841957,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,2.90029193047496,4.44551143257392,5.07885128191947,5.3180069882276,4.28496925234002,6.56791807415177,7.50363138951749,5.07587527588423,7.78020354957332,5.32974583920556,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,1.98037740107218,3.0354842147445,3.46794134442417,3.70920808974538,5.68540249112695,5.32120685033247,.236321146004057,.362228486371654,.41383418762992,1.08420509999131,1.66184862884528,2.4592936780291,.0946962283839188,.145148549160118,.526147189418459,.0362494356624709},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.7410698100085,5.7342395194157,6.55118093268348,3.76406389114815,5.76948440365682,3.6179005979755,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,2.19713005267825,3.36771849254271,3.8475081309354,4.02868168249338,3.76697803717186,5.77395114836884,6.59655016845447,5.17830525859385,7.93720624797419,6.30978459469567,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.979365349166059,1.50115228352525,1.71501734155852,2.12867470143779,3.262791452389,3.543802987655,.0657431866024243,.100769886127061,.115126296058794,.350018452610409,.53650152129502,.921342735777177,.0171974308138302,.0263598896651818,.110884150015163,.00429748986298374},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,5.18623489409549,7.94936063666436,9.0818842940915,5.81833621896758,8.91823333538729,6.23568135536041,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,1.36595759702526,2.0937134121504,2.39199903285711,2.50463478185567,2.61131941894657,4.0025799484364,4.57281656090381,4.0025799484364,6.13507720928999,5.43816471058122,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.339486246269225,.520357958596769,.594491933034081,.822758173894117,1.26110783171881,1.5272774292324,.0113956405123916,.0174670175895654,.0199554957891379,.067649468233801,.10369179782253,.198554676115027,.00166206328995312,.00254757849669498,.0119492023316399,.000231576690958128},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,6.45780747384476,9.89840251743778,11.3086008382978,7.79428043908953,11.9469224570834,8.98680735137037,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.612365059546825,.938621331363155,1.07234414405062,1.12283926724684,1.2594400120878,1.93044531514061,2.2054705766571,2.07683410541886,3.18333118932674,3.03569721712443,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.0634197097850823,.0972085057383859,.111057535546505,.165355473459674,.253453674742105,.330223979417038,.000824565926330002,.00126387871952445,.00144394006224153,.00526617557576491,.0080718921723515,.0166285878943821,5.01144631725624e-005,7.68144808465349e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,7.27823584467216,11.1559392718342,12.7452954129935,9.14177937313726,14.0123427870133,10.9691712594614,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.122554596919714,.187849318200136,.214611696543555,.22471744861619,.262307293803493,.402059551538899,.459339875638177,.450140785461579,.689967098580019,.68472894504081,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,.00244089433681864,.00374135567557013,.00427437638070464,.00662303395559898,.0101516584777441,.0137645079246526,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,3.89785419960172e-005,5.97455560493016e-005,.000128085303901421,6.8546484667622e-008,1.05066727253324e-007,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,6.8546484667622e-008,1.13034107101112e-007,7.65361974719019e-010,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,.122554596919714,.202094235800358,.248394511041402,.279814171322846,.00244089433681864,.00402506872719183,.00494722164817595,3.89785419960172e-005,6.42761581885916e-005,5.51740069821357e-007,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,.262307293803493,.432548377566746,.531646250640432,.00662303395559898,.0109214751579782,.000128085303901421,7.27823584467216,12.0019121888667,14.7515791225583,.450140785461579,.742288419521226,.0137645079246526,9.14177937313726,15.0749214021478,.68472894504081,10.9691712594614},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.000824565926330002,.00135972068684592,.00167123596481511,5.01144631725624e-005,8.26394471580808e-005,2.90963994598978e-006,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,.612365059546825,1.00979850491474,1.24114576986956,1.39813949040533,.0634197097850823,.104579983939912,.128539509653141,.00526617557576491,.00868399681012567,.000387613365377294,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,1.2594400120878,2.07683410541886,2.55264179132826,.165355473459674,.272673476798176,.0166285878943821,6.45780747384476,10.6490143886212,13.0887291840651,2.07683410541886,3.42472834168617,.330223979417038,7.79428043908953,12.8528769082785,3.03569721712443,8.98680735137037},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.0113956405123916,.0187915697820833,.0230967635919399,.00166206328995312,.00274076549419503,.000231576690958128,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,1.36595759702526,2.25248308627179,2.76853237613478,3.11872669553275,.339486246269225,.559817544151044,.688073089222744,.067649468233801,.111554914547963,.0119492023316399,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,2.61131941894657,4.306102059137,5.29264039202637,.822758173894117,1.35673967813674,.198554676115027,5.18623489409549,8.55217350992208,10.5114970194908,4.0025799484364,6.60031003608269,1.5272774292324,5.81833621896758,9.5945173923078,5.43816471058122,6.23568135536041},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.0657431866024243,.108411429563137,.133248748684711,.0171974308138302,.0283588027292761,.00429748986298374,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,2.19713005267825,3.62309803230714,4.45315850116099,5.01644279717582,.979365349166059,1.61498708893785,1.98498451316933,.350018452610409,.577185299317033,.110884150015163,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.76697803717186,6.21179920192732,7.63493733120357,2.12867470143779,3.51021420500431,.921342735777177,3.7410698100085,6.16907617592665,7.58242635758397,5.17830525859385,8.53909743428073,3.543802987655,3.76406389114815,6.20699373673991,6.30978459469567,3.6179005979755},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,.236321146004057,.389696858310697,.47897734533629,.0946962283839188,.156155398385808,.0362494356624709,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.90029193047496,4.78262175405185,5.87833189496431,6.62188774242252,1.98037740107218,3.26566989191443,4.01384271427013,1.08420509999131,1.78786929801119,.526147189418459,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,4.28496925234002,7.06597392374297,8.68480553507875,3.70920808974538,6.11653575680062,2.4592936780291,2.39387697960251,3.94753645159365,4.85192654207924,5.07587527588423,8.370188967346,5.32120685033247,2.07554270972356,3.42259881132375,5.32974583920556,1.71909544841957},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,.626153181688297,1.03253530989981,1.26909163149447,.347192252628355,.572524856067592,.183906859961414,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,3.30149822057238,5.4442165096419,6.69149959948354,7.53791381092379,3.11944010216456,5.1440001364154,6.3225029329745,2.36319452013471,3.8969406506306,1.58691847335237,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,4.0126111958998,6.61685169872888,8.13278832870006,4.80639993236566,7.92582037364658,4.40969381865368,1.33269841309056,2.19763822848137,2.70112242949704,3.91021223386622,6.44799438933945,5.67229573503973,.95054294779362,1.56745854854701,3.37758202463881,.647664476709677},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,1.33269841309056,2.19763822848137,2.70112242949704,.95054294779362,1.56745854854701,.647664476709677,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,3.30149822057238,5.4442165096419,6.69149959948354,7.53791381092379,4.0126111958998,6.61685169872888,8.13278832870006,3.91021223386622,6.44799438933945,3.37758202463881,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,3.11944010216456,5.1440001364154,6.3225029329745,4.80639993236566,7.92582037364658,5.67229573503973,.626153181688297,1.03253530989981,1.26909163149447,2.36319452013471,3.8969406506306,4.40969381865368,.347192252628355,.572524856067592,1.58691847335237,.183906859961414},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,2.39387697960251,3.94753645159365,4.85192654207924,2.07554270972356,3.42259881132375,1.71909544841957,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,2.90029193047496,4.78262175405185,5.87833189496431,6.62188774242252,4.28496925234002,7.06597392374297,8.68480553507875,5.07587527588423,8.370188967346,5.32974583920556,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,1.98037740107218,3.26566989191443,4.01384271427013,3.70920808974538,6.11653575680062,5.32120685033247,.236321146004057,.389696858310697,.47897734533629,1.08420509999131,1.78786929801119,2.4592936780291,.0946962283839188,.156155398385808,.526147189418459,.0362494356624709},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.7410698100085,6.16907617592665,7.58242635758397,3.76406389114815,6.20699373673991,3.6179005979755,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,2.19713005267825,3.62309803230714,4.45315850116099,5.01644279717582,3.76697803717186,6.21179920192732,7.63493733120357,5.17830525859385,8.53909743428073,6.30978459469567,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.979365349166059,1.61498708893785,1.98498451316933,2.12867470143779,3.51021420500431,3.543802987655,.0657431866024243,.108411429563137,.133248748684711,.350018452610409,.577185299317033,.921342735777177,.0171974308138302,.0283588027292761,.110884150015163,.00429748986298374},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,5.18623489409549,8.55217350992208,10.5114970194908,5.81833621896758,9.5945173923078,6.23568135536041,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,1.36595759702526,2.25248308627179,2.76853237613478,3.11872669553275,2.61131941894657,4.306102059137,5.29264039202637,4.0025799484364,6.60031003608269,5.43816471058122,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.339486246269225,.559817544151044,.688073089222744,.822758173894117,1.35673967813674,1.5272774292324,.0113956405123916,.0187915697820833,.0230967635919399,.067649468233801,.111554914547963,.198554676115027,.00166206328995312,.00274076549419503,.0119492023316399,.000231576690958128},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,6.45780747384476,10.6490143886212,13.0887291840651,7.79428043908953,12.8528769082785,8.98680735137037,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.612365059546825,1.00979850491474,1.24114576986956,1.39813949040533,1.2594400120878,2.07683410541886,2.55264179132826,2.07683410541886,3.42472834168617,3.03569721712443,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.0634197097850823,.104579983939912,.128539509653141,.165355473459674,.272673476798176,.330223979417038,.000824565926330002,.00135972068684592,.00167123596481511,.00526617557576491,.00868399681012567,.0166285878943821,5.01144631725624e-005,8.26394471580808e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,7.27823584467216,12.0019121888667,14.7515791225583,9.14177937313726,15.0749214021478,10.9691712594614,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.122554596919714,.202094235800358,.248394511041402,.279814171322846,.262307293803493,.432548377566746,.531646250640432,.450140785461579,.742288419521226,.68472894504081,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,.00244089433681864,.00402506872719183,.00494722164817595,.00662303395559898,.0109214751579782,.0137645079246526,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,3.89785419960172e-005,6.42761581885916e-005,.000128085303901421,6.8546484667622e-008,1.13034107101112e-007,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,6.8546484667622e-008,1.17631378114674e-007,7.65361974719019e-010,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,.122554596919714,.21031371924726,.269010572516923,.315362998946248,.00244089433681864,.00418877446399711,.00535782744298286,3.89785419960172e-005,6.68903684166533e-005,5.51740069821357e-007,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,.262307293803493,.450140785461579,.575771427724586,.00662303395559898,.0113656683529065,.000128085303901421,7.27823584467216,12.4900484198527,15.9759196314393,.450140785461579,.77247843138831,.0137645079246526,9.14177937313726,15.6880416449771,.68472894504081,10.9691712594614},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.000824565926330002,.00141502261880692,.00180994395496465,5.01144631725624e-005,8.60005205963556e-005,2.90963994598978e-006,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,.612365059546825,1.05086856345938,1.34415745633734,1.57576530364745,.0634197097850823,.108833412759726,.139207935545212,.00526617557576491,.00903718831878549,.000387613365377294,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,1.2594400120878,2.16130214323225,2.76450404172331,.165355473459674,.283763526546879,.0166285878943821,6.45780747384476,11.0821261849396,14.1750577199246,2.07683410541886,3.56401731152317,.330223979417038,7.79428043908953,13.3756231834223,3.03569721712443,8.98680735137037},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.0113956405123916,.019555851831759,.0250137314673599,.00166206328995312,.00285223664290029,.000231576690958128,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,1.36595759702526,2.34409503833324,2.99831295149481,3.51494350320858,.339486246269225,.582586184826084,.745181263826102,.067649468233801,.116092024524981,.0119492023316399,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,2.61131941894657,4.48123785385959,5.73191499287154,.822758173894117,1.41192036788792,.198554676115027,5.18623489409549,8.90000356218188,11.3839223715851,4.0025799484364,6.86875479835488,1.5272774292324,5.81833621896758,9.98474117367778,5.43816471058122,6.23568135536041},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.0657431866024243,.112820689170272,.144308028468726,.0171974308138302,.0295121988599679,.00429748986298374,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,2.19713005267825,3.77045500260921,4.82275841315291,5.6537538362715,.979365349166059,1.68067109814234,2.14973277020564,.350018452610409,.600660313311091,.110884150015163,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.76697803717186,6.46444263093715,8.26861613804166,2.12867470143779,3.65298001640339,.921342735777177,3.7410698100085,6.41998199374919,8.21174689798638,5.17830525859385,8.88639566242957,3.543802987655,3.76406389114815,6.45944172008843,6.30978459469567,3.6179005979755},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,.236321146004057,.405546428999883,.518731148089159,.0946962283839188,.162506478648407,.0362494356624709,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.90029193047496,4.97713833778588,6.36621728024122,7.46316159493692,1.98037740107218,3.39848970992327,4.34698062381386,1.08420509999131,1.86058469259941,.526147189418459,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,4.28496925234002,7.35335795264632,9.40562051630952,3.70920808974538,6.36530467495559,2.4592936780291,2.39387697960251,4.10808883148424,5.25462310508641,5.07587527588423,8.71061742831661,5.32120685033247,2.07554270972356,3.5618011698851,5.32974583920556,1.71909544841957},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,.626153181688297,1.07453010927869,1.3744227476422,.347192252628355,.595810322665643,.183906859961414,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,3.30149822057238,5.66564116980173,7.24687566849574,8.49556365917046,3.11944010216456,5.35321453486018,6.84725329319339,2.36319452013471,4.05543522146567,1.58691847335237,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,4.0126111958998,6.88596923586654,8.80778739952872,4.80639993236566,8.24817568034947,4.40969381865368,1.33269841309056,2.28701955580991,2.92530816462441,3.91021223386622,6.71024423995495,5.67229573503973,.95054294779362,1.63120949890175,3.37758202463881,.647664476709677},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,1.33269841309056,2.28701955580991,2.92530816462441,.95054294779362,1.63120949890175,.647664476709677,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,3.30149822057238,5.66564116980173,7.24687566849574,8.49556365917046,4.0126111958998,6.88596923586654,8.80778739952872,3.91021223386622,6.71024423995495,3.37758202463881,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,3.11944010216456,5.35321453486018,6.84725329319339,4.80639993236566,8.24817568034947,5.67229573503973,.626153181688297,1.07453010927869,1.3744227476422,2.36319452013471,4.05543522146567,4.40969381865368,.347192252628355,.595810322665643,1.58691847335237,.183906859961414},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,2.39387697960251,4.10808883148424,5.25462310508641,2.07554270972356,3.5618011698851,1.71909544841957,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,2.90029193047496,4.97713833778588,6.36621728024122,7.46316159493692,4.28496925234002,7.35335795264632,9.40562051630952,5.07587527588423,8.71061742831661,5.32974583920556,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,1.98037740107218,3.39848970992327,4.34698062381386,3.70920808974538,6.36530467495559,5.32120685033247,.236321146004057,.405546428999883,.518731148089159,1.08420509999131,1.86058469259941,2.4592936780291,.0946962283839188,.162506478648407,.526147189418459,.0362494356624709},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.7410698100085,6.41998199374919,8.21174689798638,3.76406389114815,6.45944172008843,3.6179005979755,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,2.19713005267825,3.77045500260921,4.82275841315291,5.6537538362715,3.76697803717186,6.46444263093715,8.26861613804166,5.17830525859385,8.88639566242957,6.30978459469567,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.979365349166059,1.68067109814234,2.14973277020564,2.12867470143779,3.65298001640339,3.543802987655,.0657431866024243,.112820689170272,.144308028468726,.350018452610409,.600660313311091,.921342735777177,.0171974308138302,.0295121988599679,.110884150015163,.00429748986298374},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,5.18623489409549,8.90000356218188,11.3839223715851,5.81833621896758,9.98474117367778,6.23568135536041,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,1.36595759702526,2.34409503833324,2.99831295149481,3.51494350320858,2.61131941894657,4.48123785385959,5.73191499287154,4.0025799484364,6.86875479835488,5.43816471058122,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.339486246269225,.582586184826084,.745181263826102,.822758173894117,1.41192036788792,1.5272774292324,.0113956405123916,.019555851831759,.0250137314673599,.067649468233801,.116092024524981,.198554676115027,.00166206328995312,.00285223664290029,.0119492023316399,.000231576690958128},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,6.45780747384476,11.0821261849396,14.1750577199246,7.79428043908953,13.3756231834223,8.98680735137037,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.612365059546825,1.05086856345938,1.34415745633734,1.57576530364745,1.2594400120878,2.16130214323225,2.76450404172331,2.07683410541886,3.56401731152317,3.03569721712443,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.0634197097850823,.108833412759726,.139207935545212,.165355473459674,.283763526546879,.330223979417038,.000824565926330002,.00141502261880692,.00180994395496465,.00526617557576491,.00903718831878549,.0166285878943821,5.01144631725624e-005,8.60005205963556e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,7.27823584467216,12.4900484198527,15.9759196314393,9.14177937313726,15.6880416449771,10.9691712594614,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.122554596919714,.21031371924726,.269010572516923,.315362998946248,.262307293803493,.450140785461579,.575771427724586,.450140785461579,.77247843138831,.68472894504081,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,.00244089433681864,.00418877446399711,.00535782744298286,.00662303395559898,.0113656683529065,.0137645079246526,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,3.89785419960172e-005,6.68903684166533e-005,.000128085303901421,6.8546484667622e-008,1.17631378114674e-007,5.51740069821357e-007,7.65361974719019e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,6.8546484667622e-008,1.09461601509022e-009,7.65361974719019e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,.122554596919714,.00195706935488603,2.32941342728397e-005,2.54112819850642e-007,.00244089433681864,3.89785419960172e-005,4.63944411923424e-007,3.89785419960172e-005,6.22446745908012e-007,5.51740069821357e-007,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,.262307293803493,.00418877446399711,4.98571369236328e-005,.00662303395559898,.000105762958835416,.000128085303901421,7.27823584467216,.116225851033514,.00138338509726576,.450140785461579,.0071882798268767,.0137645079246526,9.14177937313726,.145984701575574,.68472894504081,10.9691712594614},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,5.01144631725624e-005,8.00275816365055e-007,2.90963994598978e-006,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,.612365059546825,.00977883263593246,.000116393136443702,1.26971828042803e-006,.0634197097850823,.00101274675595133,1.20542784326106e-005,.00526617557576491,8.40953427875793e-005,.000387613365377294,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,1.2594400120878,.0201119461265218,.000239383633673578,.165355473459674,.00264055480463631,.0166285878943821,6.45780747384476,.103124463877826,.00122744505825102,2.07683410541886,.0331648790465492,.330223979417038,7.79428043908953,.124466546112493,3.03569721712443,8.98680735137037},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.0113956405123916,.000181976518058871,2.16598631798674e-006,.00166206328995312,2.65414208197256e-005,.000231576690958128,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,1.36595759702526,.0218129211013044,.000259629589389913,2.83226696917689e-006,.339486246269225,.00542124200448682,6.45266550483906e-005,.067649468233801,.00108029159656201,.0119492023316399,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,2.61131941894657,.0417000531735786,.000496337360406356,.822758173894117,.0131385916878024,.198554676115027,5.18623489409549,.0828187732700635,.000985755369367884,4.0025799484364,.0639170357987087,1.5272774292324,5.81833621896758,.0929127735338078,5.43816471058122,6.23568135536041},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.0657431866024243,.00104985026256264,1.24959051250218e-005,.0171974308138302,.000274625070541602,.00429748986298374,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,2.19713005267825,.0350858801127846,.000417611772616793,4.55567499953034e-006,.979365349166059,.0156394452705675,.000186149426512757,.350018452610409,.00558943038088415,.110884150015163,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.76697803717186,.0601547184591987,.000715995110407768,2.12867470143779,.0339927193989142,.921342735777177,3.7410698100085,.0597409910483066,.000711070695350591,5.17830525859385,.0826921452798168,3.543802987655,3.76406389114815,.0601081826012263,6.30978459469567,3.6179005979755},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,.236321146004057,.00377380273155657,4.4917911225712e-005,.0946962283839188,.00151220020487266,.0362494356624709,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.90029193047496,.0463146434325445,.000551262795170185,6.01365741772922e-006,1.98037740107218,.0316245658532948,.000376413273965959,1.08420509999131,.0173136269811071,.526147189418459,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,4.28496925234002,.0684264990231684,.00081445047001826,3.70920808974538,.0592321925476606,2.4592936780291,2.39387697960251,.0382277237453785,.000455007753107355,5.07587527588423,.0810564450674401,5.32120685033247,2.07554270972356,.0331442567927967,5.32974583920556,1.71909544841957},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,.626153181688297,.00999901459257298,.000119013865260662,.347192252628355,.00554429890730429,.183906859961414,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,3.30149822057238,.0527214902997539,.000627520670660238,6.84554511742325e-006,3.11944010216456,.0498142116158778,.000592916613348927,2.36319452013471,.0377377568190159,1.58691847335237,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,4.0126111958998,.0640772243410266,.000762682969712426,4.80639993236566,.076753204286926,4.40969381865368,1.33269841309056,.0212818065446243,.000253307966815722,3.91021223386622,.0624420195354773,5.67229573503973,.95054294779362,.0151791815252261,3.37758202463881,.647664476709677},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,1.33269841309056,.0212818065446243,.000253307966815722,.95054294779362,.0151791815252261,.647664476709677,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,3.30149822057238,.0527214902997539,.000627520670660238,6.84554511742325e-006,4.0126111958998,.0640772243410266,.000762682969712426,3.91021223386622,.0624420195354773,3.37758202463881,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,3.11944010216456,.0498142116158778,.000592916613348927,4.80639993236566,.076753204286926,5.67229573503973,.626153181688297,.00999901459257298,.000119013865260662,2.36319452013471,.0377377568190159,4.40969381865368,.347192252628355,.00554429890730429,1.58691847335237,.183906859961414},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,2.39387697960251,.0382277237453785,.000455007753107355,2.07554270972356,.0331442567927967,1.71909544841957,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,2.90029193047496,.0463146434325445,.000551262795170185,6.01365741772922e-006,4.28496925234002,.0684264990231684,.00081445047001826,5.07587527588423,.0810564450674401,5.32974583920556,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,1.98037740107218,.0316245658532948,.000376413273965959,3.70920808974538,.0592321925476606,5.32120685033247,.236321146004057,.00377380273155657,4.4917911225712e-005,1.08420509999131,.0173136269811071,2.4592936780291,.0946962283839188,.00151220020487266,.526147189418459,.0362494356624709},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.7410698100085,.0597409910483066,.000711070695350591,3.76406389114815,.0601081826012263,3.6179005979755,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,2.19713005267825,.0350858801127846,.000417611772616793,4.55567499953034e-006,3.76697803717186,.0601547184591987,.000715995110407768,5.17830525859385,.0826921452798168,6.30978459469567,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.979365349166059,.0156394452705675,.000186149426512757,2.12867470143779,.0339927193989142,3.543802987655,.0657431866024243,.00104985026256264,1.24959051250218e-005,.350018452610409,.00558943038088415,.921342735777177,.0171974308138302,.000274625070541602,.110884150015163,.00429748986298374},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,5.18623489409549,.0828187732700635,.000985755369367884,5.81833621896758,.0929127735338078,6.23568135536041,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,1.36595759702526,.0218129211013044,.000259629589389913,2.83226696917689e-006,2.61131941894657,.0417000531735786,.000496337360406356,4.0025799484364,.0639170357987087,5.43816471058122,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.339486246269225,.00542124200448682,6.45266550483906e-005,.822758173894117,.0131385916878024,1.5272774292324,.0113956405123916,.000181976518058871,2.16598631798674e-006,.067649468233801,.00108029159656201,.198554676115027,.00166206328995312,2.65414208197256e-005,.0119492023316399,.000231576690958128},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,6.45780747384476,.103124463877826,.00122744505825102,7.79428043908953,.124466546112493,8.98680735137037,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.612365059546825,.00977883263593246,.000116393136443702,1.26971828042803e-006,1.2594400120878,.0201119461265218,.000239383633673578,2.07683410541886,.0331648790465492,3.03569721712443,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.0634197097850823,.00101274675595133,1.20542784326106e-005,.165355473459674,.00264055480463631,.330223979417038,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,.00526617557576491,8.40953427875793e-005,.0166285878943821,5.01144631725624e-005,8.00275816365055e-007,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,7.27823584467216,.116225851033514,.00138338509726576,9.14177937313726,.145984701575574,10.9691712594614,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.122554596919714,.00195706935488603,2.32941342728397e-005,2.54112819850642e-007,.262307293803493,.00418877446399711,4.98571369236328e-005,.450140785461579,.0071882798268767,.68472894504081,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,.00244089433681864,3.89785419960172e-005,4.63944411923424e-007,.00662303395559898,.000105762958835416,.0137645079246526,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,3.89785419960172e-005,6.22446745908012e-007,.000128085303901421,6.8546484667622e-008,1.09461601509022e-009,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,6.8546484667622e-008,5.69188702865245e-009,7.65361974719019e-010,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,.122554596919714,.0101765528017885,.000629847662741689,3.57281062205031e-005,.00244089433681864,.0002026842788013,1.25445444792834e-005,3.89785419960172e-005,3.23665697396975e-006,5.51740069821357e-007,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,.262307293803493,.0217811823588299,.0013480819160108,.00662303395559898,.000549956153763738,.000128085303901421,7.27823584467216,.60436208201944,.037405205103521,.450140785461579,.037378291693961,.0137645079246526,9.14177937313726,.759104944404841,.68472894504081,10.9691712594614},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,5.01144631725624e-005,4.16134925463982e-006,2.90963994598978e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,.612365059546825,.0508488911805658,.0031471418550943,.000178521609495778,.0634197097850823,.00526617557576491,.000325934375061529,.00526617557576491,.000437286851447401,.000387613365377294,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,1.2594400120878,.104579983939912,.00647266905916808,.165355473459674,.0137306045533387,.0166285878943821,6.45780747384476,.53623626019628,.0331887586818224,2.07683410541886,.172453848883548,.330223979417038,7.79428043908953,.647212821256298,3.03569721712443,8.98680735137037},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.0113956405123916,.000946258567734596,5.85658777413804e-005,.00166206328995312,.000138012569524986,.000231576690958128,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,1.36595759702526,.11342487316276,.00702009734040598,.000398214994344052,.339486246269225,.0281898826795267,.0017447294838579,.067649468233801,.00561740157357991,.0119492023316399,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,2.61131941894657,.216835847896165,.0134204140287723,.822758173894117,.0683192814389796,.198554676115027,5.18623489409549,.430648825529863,.0266537364367886,4.0025799484364,.332361798070895,1.5272774292324,5.81833621896758,.483136554903789,5.43816471058122,6.23568135536041},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.0657431866024243,.0054591098696972,.000337875473054763,.0171974308138302,.00142802120123345,.00429748986298374,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,2.19713005267825,.182442850414858,.0112917610860855,.000640525103711722,.979365349166059,.0813234544750617,.00503327489386337,.350018452610409,.029064444374942,.110884150015163,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.76697803717186,.31279814746903,.0193597169803656,2.12867470143779,.176758530797994,.921342735777177,3.7410698100085,.310646808870845,.0192265662361566,5.17830525859385,.429990373428657,3.543802987655,3.76406389114815,.31255616594975,6.30978459469567,3.6179005979755},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,.236321146004057,.0196233734207421,.0012145307084342,.0946962283839188,.00786328046747136,.0362494356624709,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.90029193047496,.240831227166574,.0149055371205287,.000845516535217038,1.98037740107218,.164444383862132,.0101777991856447,1.08420509999131,.0900290215693292,.526147189418459,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,4.28496925234002,.355810527926517,.0220218411618753,3.70920808974538,.308001110702629,2.4592936780291,2.39387697960251,.198780103635972,.0123029070952925,5.07587527588423,.421484906038046,5.32120685033247,2.07554270972356,.172346615354148,5.32974583920556,1.71909544841957},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,.626153181688297,.0519938139714517,.00321800346775215,.347192252628355,.028829765505355,.183906859961414,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,3.30149822057238,.274146150459585,.0169674658481853,.000962479434277651,3.11944010216456,.259028610060664,.016031810358111,2.36319452013471,.196232327654089,1.58691847335237,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,4.0126111958998,.333194761478692,.0206221051299754,4.80639993236566,.399108510989818,4.40969381865368,1.33269841309056,.110663133873164,.00684916764813009,3.91021223386622,.32469187015097,5.67229573503973,.95054294779362,.0789301318799674,3.37758202463881,.647664476709677},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,1.33269841309056,.110663133873164,.00684916764813009,.95054294779362,.0789301318799674,.647664476709677,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,3.30149822057238,.274146150459585,.0169674658481853,.000962479434277651,4.0126111958998,.333194761478692,.0206221051299754,3.91021223386622,.32469187015097,3.37758202463881,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,3.11944010216456,.259028610060664,.016031810358111,4.80639993236566,.399108510989818,5.67229573503973,.626153181688297,.0519938139714517,.00321800346775215,2.36319452013471,.196232327654089,4.40969381865368,.347192252628355,.028829765505355,1.58691847335237,.183906859961414},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,2.39387697960251,.198780103635972,.0123029070952925,2.07554270972356,.172346615354148,1.71909544841957,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,2.90029193047496,.240831227166574,.0149055371205287,.000845516535217038,4.28496925234002,.355810527926517,.0220218411618753,5.07587527588423,.421484906038046,5.32974583920556,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,1.98037740107218,.164444383862132,.0101777991856447,3.70920808974538,.308001110702629,5.32120685033247,.236321146004057,.0196233734207421,.0012145307084342,1.08420509999131,.0900290215693292,2.4592936780291,.0946962283839188,.00786328046747136,.526147189418459,.0362494356624709},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.7410698100085,.310646808870845,.0192265662361566,3.76406389114815,.31255616594975,3.6179005979755,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,2.19713005267825,.182442850414858,.0112917610860855,.000640525103711722,3.76697803717186,.31279814746903,.0193597169803656,5.17830525859385,.429990373428657,6.30978459469567,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.979365349166059,.0813234544750617,.00503327489386337,2.12867470143779,.176758530797994,3.543802987655,.0657431866024243,.0054591098696972,.000337875473054763,.350018452610409,.029064444374942,.921342735777177,.0171974308138302,.00142802120123345,.110884150015163,.00429748986298374},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,5.18623489409549,.430648825529863,.0266537364367886,5.81833621896758,.483136554903789,6.23568135536041,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,1.36595759702526,.11342487316276,.00702009734040598,.000398214994344052,2.61131941894657,.216835847896165,.0134204140287723,4.0025799484364,.332361798070895,5.43816471058122,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.339486246269225,.0281898826795267,.0017447294838579,.822758173894117,.0683192814389796,1.5272774292324,.0113956405123916,.000946258567734596,5.85658777413804e-005,.067649468233801,.00561740157357991,.198554676115027,.00166206328995312,.000138012569524986,.0119492023316399,.000231576690958128},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,6.45780747384476,.53623626019628,.0331887586818224,7.79428043908953,.647212821256298,8.98680735137037,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.612365059546825,.0508488911805658,.0031471418550943,.000178521609495778,1.2594400120878,.104579983939912,.00647266905916808,2.07683410541886,.172453848883548,3.03569721712443,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.0634197097850823,.00526617557576491,.000325934375061529,.165355473459674,.0137306045533387,.330223979417038,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,.00526617557576491,.000437286851447401,.0166285878943821,5.01144631725624e-005,4.16134925463982e-006,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,7.27823584467216,.60436208201944,.037405205103521,9.14177937313726,.759104944404841,10.9691712594614,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.122554596919714,.0101765528017885,.000629847662741689,3.57281062205031e-005,.262307293803493,.0217811823588299,.0013480819160108,.450140785461579,.037378291693961,.68472894504081,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,.00244089433681864,.0002026842788013,1.25445444792834e-005,.00662303395559898,.000549956153763738,.0137645079246526,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,3.89785419960172e-005,3.23665697396975e-006,.000128085303901421,6.8546484667622e-008,5.69188702865245e-009,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,6.8546484667622e-008,1.36592668764407e-008,7.65361974719019e-010,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,.122554596919714,.0244214704020099,.00362725225603475,.000493768468568094,.00244089433681864,.000486397330423004,7.22432263467335e-005,3.89785419960172e-005,7.76725911325977e-006,5.51740069821357e-007,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,.262307293803493,.0522700083866766,.00776351721285218,.00662303395559898,.00131977283399776,.000128085303901421,7.27823584467216,1.45033499905202,.215414174926983,.450140785461579,.0896996126351677,.0137645079246526,9.14177937313726,1.82168355953934,.68472894504081,10.9691712594614},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.000824565926330002,.000164311358893045,2.44047036237916e-005,5.01144631725624e-005,9.98631556618579e-006,2.90963994598978e-006,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,.612365059546825,.122026064732152,.0181241879095991,.00246719882612911,.0634197097850823,.0126376537772916,.00187703514229927,.00526617557576491,.00104939148922157,.000387613365377294,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,1.2594400120878,.250968774218159,.0372756855923469,.165355473459674,.0329504066094101,.0166285878943821,6.45780747384476,1.28684813137965,.191131930663375,2.07683410541886,.413851001242982,.330223979417038,7.79428043908953,1.55316727245144,3.03569721712443,8.98680735137037},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.0113956405123916,.00227081076025246,.000337277130218073,.00166206328995312,.000331199567025036,.000231576690958128,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,1.36595759702526,.272194547284149,.0404282899212952,.0055033985485992,.339486246269225,.067649468233801,.0100477850929001,.067649468233801,.0134805182990133,.0119492023316399,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,2.61131941894657,.520357958596769,.0772873028549269,.822758173894117,.16395112785691,.198554676115027,5.18623489409549,1.03346169878758,.153497157076452,4.0025799484364,.797594624863601,1.5272774292324,5.81833621896758,1.1594206118243,5.43816471058122,6.23568135536041},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.0657431866024243,.0131006533057736,.00194580315907169,.0171974308138302,.00342693426532766,.00429748986298374,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,2.19713005267825,.437822390179288,.0650285272089779,.00885216522776848,.979365349166059,.195158259887658,.0289863069976914,.350018452610409,.0697482223969557,.110884150015163,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.76697803717186,.750646201027513,.111491367273716,2.12867470143779,.424181283413309,.921342735777177,3.7410698100085,.745483465381799,.110724560685561,5.17830525859385,1.0318815597352,3.543802987655,3.76406389114815,.750065499032834,6.30978459469567,3.6179005979755},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,.236321146004057,.0470917453597858,.00699440438187064,.0946962283839188,.0188701296931614,.0362494356624709,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.90029193047496,.577941548644503,.0858400314014041,.0116851814693591,1.98037740107218,.394630061032062,.0586132921362272,1.08420509999131,.216049690735243,.526147189418459,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,4.28496925234002,.853866377517717,.126822369537332,3.70920808974538,.739134376376297,2.4592936780291,2.39387697960251,.477028175651108,.0708516521644813,5.07587527588423,1.01147032381073,5.32120685033247,2.07554270972356,.413593664547998,5.32974583920556,1.71909544841957},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,.626153181688297,.124773625580585,.0185322753878643,.347192252628355,.0691850451424154,.183906859961414,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,3.30149822057238,.657889978038246,.0977145465764208,.0133016285094573,3.11944010216456,.621611232907719,.0923261666744144,2.36319452013471,.470914078181719,1.58691847335237,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,4.0126111958998,.799593552359549,.118761379587061,4.80639993236566,.957771936938711,4.40969381865368,1.33269841309056,.26556698531571,.0394439167964793,3.91021223386622,.77918849841486,5.67229573503973,.95054294779362,.189414816301507,3.37758202463881,.647664476709677},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,1.33269841309056,.26556698531571,.0394439167964793,.95054294779362,.189414816301507,.647664476709677,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,3.30149822057238,.657889978038246,.0977145465764208,.0133016285094573,4.0126111958998,.799593552359549,.118761379587061,3.91021223386622,.77918849841486,3.37758202463881,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,3.11944010216456,.621611232907719,.0923261666744144,4.80639993236566,.957771936938711,5.67229573503973,.626153181688297,.124773625580585,.0185322753878643,2.36319452013471,.470914078181719,4.40969381865368,.347192252628355,.0691850451424154,1.58691847335237,.183906859961414},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,2.39387697960251,.477028175651108,.0708516521644813,2.07554270972356,.413593664547998,1.71909544841957,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,2.90029193047496,.577941548644503,.0858400314014041,.0116851814693591,4.28496925234002,.853866377517717,.126822369537332,5.07587527588423,1.01147032381073,5.32974583920556,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,1.98037740107218,.394630061032062,.0586132921362272,3.70920808974538,.739134376376297,5.32120685033247,.236321146004057,.0470917453597858,.00699440438187064,1.08420509999131,.216049690735243,2.4592936780291,.0946962283839188,.0188701296931614,.526147189418459,.0362494356624709},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.7410698100085,.745483465381799,.110724560685561,3.76406389114815,.750065499032834,3.6179005979755,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,2.19713005267825,.437822390179288,.0650285272089779,.00885216522776848,3.76697803717186,.750646201027513,.111491367273716,5.17830525859385,1.0318815597352,6.30978459469567,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.979365349166059,.195158259887658,.0289863069976914,2.12867470143779,.424181283413309,3.543802987655,.0657431866024243,.0131006533057736,.00194580315907169,.350018452610409,.0697482223969557,.921342735777177,.0171974308138302,.00342693426532766,.110884150015163,.00429748986298374},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,5.18623489409549,1.03346169878758,.153497157076452,5.81833621896758,1.1594206118243,6.23568135536041,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,1.36595759702526,.272194547284149,.0404282899212952,.0055033985485992,2.61131941894657,.520357958596769,.0772873028549269,4.0025799484364,.797594624863601,5.43816471058122,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.339486246269225,.067649468233801,.0100477850929001,.822758173894117,.16395112785691,1.5272774292324,.0113956405123916,.00227081076025246,.000337277130218073,.067649468233801,.0134805182990133,.198554676115027,.00166206328995312,.000331199567025036,.0119492023316399,.000231576690958128},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,6.45780747384476,1.28684813137965,.191131930663375,7.79428043908953,1.55316727245144,8.98680735137037,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.612365059546825,.122026064732152,.0181241879095991,.00246719882612911,1.2594400120878,.250968774218159,.0372756855923469,2.07683410541886,.413851001242982,3.03569721712443,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.0634197097850823,.0126376537772916,.00187703514229927,.165355473459674,.0329504066094101,.330223979417038,.000824565926330002,.000164311358893045,2.44047036237916e-005,.00526617557576491,.00104939148922157,.0166285878943821,5.01144631725624e-005,9.98631556618579e-006,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,7.27823584467216,1.45033499905202,.215414174926983,9.14177937313726,1.82168355953934,10.9691712594614,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.122554596919714,.0244214704020099,.00362725225603475,.000493768468568094,.262307293803493,.0522700083866766,.00776351721285218,.450140785461579,.0896996126351677,.68472894504081,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,.00244089433681864,.000486397330423004,7.22432263467335e-005,.00662303395559898,.00131977283399776,.0137645079246526,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,3.89785419960172e-005,7.76725911325977e-006,.000128085303901421,6.8546484667622e-008,1.36592668764407e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,6.8546484667622e-008,2.44980430684131e-008,7.65361974719019e-010,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,.122554596919714,.0438001716427634,.0116677137630873,.00284862553642039,.00244089433681864,.000872358879642074,.00023238342051707,3.89785419960172e-005,1.39306633365776e-005,5.51740069821357e-007,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,.262307293803493,.0937468261090737,.0249727590585012,.00662303395559898,.0023670268704951,.000128085303901421,7.27823584467216,2.60119152746664,.692918704338266,.450140785461579,.160877226679454,.0137645079246526,9.14177937313726,3.26720919228747,.68472894504081,10.9691712594614},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.000824565926330002,.000294694201614443,7.85021488046878e-005,5.01144631725624e-005,1.79105651165762e-005,2.90963994598978e-006,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,.612365059546825,.218855068600592,.0582997326325419,.0142336459837523,.0634197097850823,.0226657689113441,.00603782345911206,.00526617557576491,.00188209500049501,.000387613365377294,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,1.2594400120878,.450115214472287,.119903993192299,.165355473459674,.0590969110964026,.0166285878943821,6.45780747384476,2.30797605978546,.614810468242533,2.07683410541886,.74224625260403,.330223979417038,7.79428043908953,2.78562232344933,3.03569721712443,8.98680735137037},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.0113956405123916,.00407272369067121,.00108491296894855,.00166206328995312,.000594010010245393,.000231576690958128,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,1.36595759702526,.488183868334511,.130044915940983,.0317499448437878,.339486246269225,.121330053899551,.0323205203668776,.067649468233801,.0241774555590068,.0119492023316399,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,2.61131941894657,.933267634054527,.248608606068695,.822758173894117,.294048123349947,.198554676115027,5.18623489409549,1.85352474883709,.493751403486214,4.0025799484364,1.43049459738903,1.5272774292324,5.81833621896758,2.07943342346341,5.43816471058122,6.23568135536041},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.0657431866024243,.0234961635798143,.00625902823868614,.0171974308138302,.0061462437174735,.00429748986298374,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,2.19713005267825,.785239198263813,.209176034185977,.0510694901064878,.979365349166059,.350018452610409,.0932397057677362,.350018452610409,.125094192220009,.110884150015163,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.76697803717186,1.3462920908025,.358632173497365,2.12867470143779,.760773725550181,.921342735777177,3.7410698100085,1.33703266851113,.356165601240813,5.17830525859385,1.85069075233423,3.543802987655,3.76406389114815,1.34525059548623,6.30978459469567,3.6179005979755},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,.236321146004057,.0844595552914031,.0224987683542483,.0946962283839188,.0338437819621889,.0362494356624709,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.90029193047496,1.03654442641711,.276120006304964,.0674135925038969,1.98037740107218,.707773288179075,.188540268799867,1.08420509999131,.387487460083106,.526147189418459,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,4.28496925234002,1.53141859517937,.407947118664331,3.70920808974538,1.32564550862114,2.4592936780291,2.39387697960251,.855555199093713,.227907170151078,5.07587527588423,1.81408297965651,5.32120685033247,2.07554270972356,.741784716454702,5.32974583920556,1.71909544841957},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,.626153181688297,.223782848737467,.0596124199094679,.347192252628355,.12408412771496,.183906859961414,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,3.30149822057238,1.17993280035085,.314316534794822,.0767391217950806,3.11944010216456,1.11486647807198,.296983835024965,2.36319452013471,.844589499068632,1.58691847335237,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,4.0126111958998,1.43407969550076,.382017484674746,4.80639993236566,1.71777434126509,4.40969381865368,1.33269841309056,.476297264169728,.126878501518793,3.91021223386622,1.39748301027071,5.67229573503973,.95054294779362,.339717524338981,3.37758202463881,.647664476709677},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,1.33269841309056,.476297264169728,.126878501518793,.95054294779362,.339717524338981,.647664476709677,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,3.30149822057238,1.17993280035085,.314316534794822,.0767391217950806,4.0126111958998,1.43407969550076,.382017484674746,3.91021223386622,1.39748301027071,3.37758202463881,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,3.11944010216456,1.11486647807198,.296983835024965,4.80639993236566,1.71777434126509,5.67229573503973,.626153181688297,.223782848737467,.0596124199094679,2.36319452013471,.844589499068632,4.40969381865368,.347192252628355,.12408412771496,1.58691847335237,.183906859961414},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,2.39387697960251,.855555199093713,.227907170151078,2.07554270972356,.741784716454702,1.71909544841957,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,2.90029193047496,1.03654442641711,.276120006304964,.0674135925038969,4.28496925234002,1.53141859517937,.407947118664331,5.07587527588423,1.81408297965651,5.32974583920556,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,1.98037740107218,.707773288179075,.188540268799867,3.70920808974538,1.32564550862114,5.32120685033247,.236321146004057,.0844595552914031,.0224987683542483,1.08420509999131,.387487460083106,2.4592936780291,.0946962283839188,.0338437819621889,.526147189418459,.0362494356624709},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.7410698100085,1.33703266851113,.356165601240813,3.76406389114815,1.34525059548623,3.6179005979755,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,2.19713005267825,.785239198263813,.209176034185977,.0510694901064878,3.76697803717186,1.3462920908025,.358632173497365,5.17830525859385,1.85069075233423,6.30978459469567,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.979365349166059,.350018452610409,.0932397057677362,2.12867470143779,.760773725550181,3.543802987655,.0657431866024243,.0234961635798143,.00625902823868614,.350018452610409,.125094192220009,.921342735777177,.0171974308138302,.0061462437174735,.110884150015163,.00429748986298374},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,5.18623489409549,1.85352474883709,.493751403486214,5.81833621896758,2.07943342346341,6.23568135536041,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,1.36595759702526,.488183868334511,.130044915940983,.0317499448437878,2.61131941894657,.933267634054527,.248608606068695,4.0025799484364,1.43049459738903,5.43816471058122,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.339486246269225,.121330053899551,.0323205203668776,.822758173894117,.294048123349947,1.5272774292324,.0113956405123916,.00407272369067121,.00108491296894855,.067649468233801,.0241774555590068,.198554676115027,.00166206328995312,.000594010010245393,.0119492023316399,.000231576690958128},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,6.45780747384476,2.30797605978546,.614810468242533,7.79428043908953,2.78562232344933,8.98680735137037,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.612365059546825,.218855068600592,.0582997326325419,.0142336459837523,1.2594400120878,.450115214472287,.119903993192299,2.07683410541886,.74224625260403,3.03569721712443,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.0634197097850823,.0226657689113441,.00603782345911206,.165355473459674,.0590969110964026,.330223979417038,.000824565926330002,.000294694201614443,7.85021488046878e-005,.00526617557576491,.00188209500049501,.0166285878943821,5.01144631725624e-005,1.79105651165762e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,7.27823584467216,2.60119152746664,.692918704338266,9.14177937313726,3.26720919228747,10.9691712594614,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.122554596919714,.0438001716427634,.0116677137630873,.00284862553642039,.262307293803493,.0937468261090737,.0249727590585012,.450140785461579,.160877226679454,.68472894504081,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,.00244089433681864,.000872358879642074,.00023238342051707,.00662303395559898,.0023670268704951,.0137645079246526,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,3.89785419960172e-005,1.39306633365776e-005,.000128085303901421,6.8546484667622e-008,2.44980430684131e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,6.8546484667622e-008,3.7527416869482e-008,7.65361974719019e-010,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,.122554596919714,.0670954531185465,.0273791727414708,.0102396991398416,.00244089433681864,.00133632614020231,.000545305270748019,3.89785419960172e-005,2.13397375798647e-005,5.51740069821357e-007,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,.262307293803493,.143606418429484,.058600467750326,.00662303395559898,.00362593876834461,.000128085303901421,7.27823584467216,3.98464475452196,1.62598614322311,.450140785461579,.246440368054962,.0137645079246526,9.14177937313726,5.00488635016176,.68472894504081,10.9691712594614},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.000824565926330002,.000451428390509428,.000184211229066994,5.01144631725624e-005,2.74363646769969e-005,2.90963994598978e-006,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,.612365059546825,.335253937240527,.13680473166712,.051164412687175,.0634197097850823,.0347206410009163,.0141682093704184,.00526617557576491,.00288309410977452,.000387613365377294,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,1.2594400120878,.689510637467093,.281363788027607,.165355473459674,.0905278193943952,.0166285878943821,6.45780747384476,3.53548156799627,1.44269926011819,2.07683410541886,1.13701263662145,.330223979417038,7.79428043908953,4.26716574385505,3.03569721712443,8.98680735137037},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.0113956405123916,.00623881668046779,.00254583033055523,.00166206328995312,.000909936406629402,.000231576690958128,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,1.36595759702526,.747826244112006,.305160230186869,.114128683728465,.339486246269225,.18585988679871,.0758425453509597,.067649468233801,.0370363237207337,.0119492023316399,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,2.61131941894657,1.42962943840673,.58337889571007,.822758173894117,.450438693157099,.198554676115027,5.18623489409549,2.83932866528416,1.15862500931079,4.0025799484364,2.19130838066721,1.5272774292324,5.81833621896758,3.18538769471191,5.43816471058122,6.23568135536041},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.0657431866024243,.0359926841107618,.0146872831147845,.0171974308138302,.00941514591687767,.00429748986298374,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,2.19713005267825,1.20287153766575,.490847383612732,.183574923144401,.979365349166059,.536177046731461,.218794021040783,.350018452610409,.191625995851247,.110884150015163,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.76697803717186,2.06232246300417,.841557517454501,2.12867470143779,1.16539401381336,.921342735777177,3.7410698100085,2.04813838310324,.835769519114577,5.17830525859385,2.83498739737667,3.543802987655,3.76406389114815,2.06072704459494,6.30978459469567,3.6179005979755},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,.236321146004057,.129379678661691,.0527950614618356,.0946962283839188,.0518437211734884,.0362494356624709,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.90029193047496,1.58783437049499,.647936476974322,.242325604524073,1.98037740107218,1.08420509999131,.442423999509308,1.08420509999131,.593574082861527,.526147189418459,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,4.28496925234002,2.34590917578526,.957278765839435,3.70920808974538,2.03069492058023,2.4592936780291,2.39387697960251,1.31058536069927,.534801410737989,5.07587527588423,2.77890964691703,5.32120685033247,2.07554270972356,1.13630563078316,5.32974583920556,1.71909544841957},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,.626153181688297,.342802575265255,.139885051637275,.347192252628355,.190078724845143,.183906859961414,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,3.30149822057238,1.80748437551057,.737567519771828,.275847249626467,3.11944010216456,1.70781229176476,.69689502893851,2.36319452013471,1.29378751301169,1.58691847335237,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,4.0126111958998,2.19680022632121,.896432918697558,4.80639993236566,2.63137890697367,4.40969381865368,1.33269841309056,.729617706050062,.297729999278181,3.91021223386622,2.14073946020884,5.67229573503973,.95054294779362,.520397532085943,3.37758202463881,.647664476709677},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,1.33269841309056,.729617706050062,.297729999278181,.95054294779362,.520397532085943,.647664476709677,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,3.30149822057238,1.80748437551057,.737567519771828,.275847249626467,4.0126111958998,2.19680022632121,.896432918697558,3.91021223386622,2.14073946020884,3.37758202463881,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,3.11944010216456,1.70781229176476,.69689502893851,4.80639993236566,2.63137890697367,5.67229573503973,.626153181688297,.342802575265255,.139885051637275,2.36319452013471,1.29378751301169,4.40969381865368,.347192252628355,.190078724845143,1.58691847335237,.183906859961414},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,2.39387697960251,1.31058536069927,.534801410737989,2.07554270972356,1.13630563078316,1.71909544841957,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,2.90029193047496,1.58783437049499,.647936476974322,.242325604524073,4.28496925234002,2.34590917578526,.957278765839435,5.07587527588423,2.77890964691703,5.32974583920556,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,1.98037740107218,1.08420509999131,.442423999509308,3.70920808974538,2.03069492058023,5.32120685033247,.236321146004057,.129379678661691,.0527950614618356,1.08420509999131,.593574082861527,2.4592936780291,.0946962283839188,.0518437211734884,.526147189418459,.0362494356624709},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.7410698100085,2.04813838310324,.835769519114577,3.76406389114815,2.06072704459494,3.6179005979755,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,2.19713005267825,1.20287153766575,.490847383612732,.183574923144401,3.76697803717186,2.06232246300417,.841557517454501,5.17830525859385,2.83498739737667,6.30978459469567,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.979365349166059,.536177046731461,.218794021040783,2.12867470143779,1.16539401381336,3.543802987655,.0657431866024243,.0359926841107618,.0146872831147845,.350018452610409,.191625995851247,.921342735777177,.0171974308138302,.00941514591687767,.110884150015163,.00429748986298374},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,5.18623489409549,2.83932866528416,1.15862500931079,5.81833621896758,3.18538769471191,6.23568135536041,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,1.36595759702526,.747826244112006,.305160230186869,.114128683728465,2.61131941894657,1.42962943840673,.58337889571007,4.0025799484364,2.19130838066721,5.43816471058122,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.339486246269225,.18585988679871,.0758425453509597,.822758173894117,.450438693157099,1.5272774292324,.0113956405123916,.00623881668046779,.00254583033055523,.067649468233801,.0370363237207337,.198554676115027,.00166206328995312,.000909936406629402,.0119492023316399,.000231576690958128},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,6.45780747384476,3.53548156799627,1.44269926011819,7.79428043908953,4.26716574385505,8.98680735137037,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.612365059546825,.335253937240527,.13680473166712,.051164412687175,1.2594400120878,.689510637467093,.281363788027607,2.07683410541886,1.13701263662145,3.03569721712443,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.0634197097850823,.0347206410009163,.0141682093704184,.165355473459674,.0905278193943952,.330223979417038,.000824565926330002,.000451428390509428,.000184211229066994,.00526617557576491,.00288309410977452,.0166285878943821,5.01144631725624e-005,2.74363646769969e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,7.27823584467216,3.98464475452196,1.62598614322311,9.14177937313726,5.00488635016176,10.9691712594614,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.122554596919714,.0670954531185465,.0273791727414708,.0102396991398416,.262307293803493,.143606418429484,.058600467750326,.450140785461579,.246440368054962,.68472894504081,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,.00244089433681864,.00133632614020231,.000545305270748019,.00662303395559898,.00362593876834461,.0137645079246526,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,3.89785419960172e-005,2.13397375798647e-005,.000128085303901421,6.8546484667622e-008,3.7527416869482e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,6.8546484667622e-008,5.19287666029906e-008,7.65361974719019e-010,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,.122554596919714,.0928436971090411,.0524250669352539,.0271309626935909,.00244089433681864,.001849148543354,.00104413911950708,3.89785419960172e-005,2.95289776007745e-005,5.51740069821357e-007,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,.262307293803493,.19871616027429,.112206949174718,.00662303395559898,.00501741243403364,.000128085303901421,7.27823584467216,5.51377378765633,3.11340423610211,.450140785461579,.341013195733391,.0137645079246526,9.14177937313726,6.92553862835673,.68472894504081,10.9691712594614},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.000824565926330002,.000624666483447503,.000352723805983911,5.01144631725624e-005,3.79652184082228e-005,2.90963994598978e-006,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,.612365059546825,.463909453726739,.26195083695297,.135564507599069,.0634197097850823,.0480448752738173,.0271289907701198,.00526617557576491,.00398949710931715,.000387613365377294,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,1.2594400120878,.954114083786716,.538749492535305,.165355473459674,.125268361016134,.0166285878943821,6.45780747384476,4.89224179250545,2.76245034842036,2.07683410541886,1.57334740190396,.330223979417038,7.79428043908953,5.90471373874725,3.03569721712443,8.98680735137037},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.0113956405123916,.00863299641447793,.00487470263420339,.00166206328995312,.00125912943722616,.000231576690958128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,1.36595759702526,1.03480862072507,.584314421936207,.302393753783036,.339486246269225,.25718462626926,.145221718497785,.067649468233801,.0512492137952295,.0119492023316399,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,2.61131941894657,1.97825748822484,1.11704169972566,.822758173894117,.623296984369123,.198554676115027,5.18623489409549,3.92893643816517,2.21851091848272,4.0025799484364,3.03223485513554,1.5272774292324,5.81833621896758,4.40779749680161,5.43816471058122,6.23568135536041},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.0657431866024243,.0498050718253105,.0281229023119214,.0171974308138302,.0130282592204804,.00429748986298374,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,2.19713005267825,1.66448001337449,.939864296992246,.48639753212384,.979365349166059,.741937896083684,.41894221225766,.350018452610409,.265163510940112,.110884150015163,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.76697803717186,2.85374989208327,1.61139672111394,2.12867470143779,1.61262028650446,.921342735777177,3.7410698100085,2.8341225946006,1.60031398303209,5.17830525859385,3.92292918515564,3.543802987655,3.76406389114815,2.85154222320761,6.30978459469567,3.6179005979755},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,.547225517982271,.414561685224313,.234086155283846,.121144008364367,.236321146004057,.179029831969503,.101090878717283,.0946962283839188,.0717390303204678,.0362494356624709,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.90029193047496,2.19717442094167,1.24065520517793,.642062510456358,1.98037740107218,1.50027468671861,.847144214584193,1.08420509999131,.821361356090746,.526147189418459,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,4.28496925234002,3.24616454377475,1.83297734554319,3.70920808974538,2.80998510873911,2.4592936780291,2.39387697960251,1.81352960012534,1.02402654819954,5.07587527588423,3.84533129384947,5.32120685033247,2.07554270972356,1.57236907874115,5.32974583920556,1.71909544841957},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,.626153181688297,.474354922529452,.267848969162523,.347192252628355,.263022466294621,.183906859961414,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,3.30149822057238,2.50111630653613,1.41227885000115,.730880989425352,3.11944010216456,2.36319452013471,1.33440001580502,2.36319452013471,1.79028548729353,1.58691847335237,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,4.0126111958998,3.03983422634135,1.71647099090401,4.80639993236566,3.6411848323989,4.40969381865368,1.33269841309056,1.0096124574376,.570087171302651,3.91021223386622,2.96225983721782,5.67229573503973,.95054294779362,.720102907121204,3.37758202463881,.647664476709677},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,1.33269841309056,1.0096124574376,.570087171302651,.95054294779362,.720102907121204,.647664476709677,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,3.30149822057238,2.50111630653613,1.41227885000115,.730880989425352,4.0126111958998,3.03983422634135,1.71647099090401,3.91021223386622,2.96225983721782,3.37758202463881,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,3.11944010216456,2.36319452013471,1.33440001580502,4.80639993236566,3.6411848323989,5.67229573503973,.626153181688297,.474354922529452,.267848969162523,2.36319452013471,1.79028548729353,4.40969381865368,.347192252628355,.263022466294621,1.58691847335237,.183906859961414},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,2.39387697960251,1.81352960012534,1.02402654819954,2.07554270972356,1.57236907874115,1.71909544841957,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,2.90029193047496,2.19717442094167,1.24065520517793,.642062510456358,4.28496925234002,3.24616454377475,1.83297734554319,5.07587527588423,3.84533129384947,5.32974583920556,.547225517982271,.414561685224313,.234086155283846,.121144008364367,1.98037740107218,1.50027468671861,.847144214584193,3.70920808974538,2.80998510873911,5.32120685033247,.236321146004057,.179029831969503,.101090878717283,1.08420509999131,.821361356090746,2.4592936780291,.0946962283839188,.0717390303204678,.526147189418459,.0362494356624709},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.7410698100085,2.8341225946006,1.60031398303209,3.76406389114815,2.85154222320761,3.6179005979755,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,2.19713005267825,1.66448001337449,.939864296992246,.48639753212384,3.76697803717186,2.85374989208327,1.61139672111394,5.17830525859385,3.92292918515564,6.30978459469567,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.979365349166059,.741937896083684,.41894221225766,2.12867470143779,1.61262028650446,3.543802987655,.0657431866024243,.0498050718253105,.0281229023119214,.350018452610409,.265163510940112,.921342735777177,.0171974308138302,.0130282592204804,.110884150015163,.00429748986298374},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,5.18623489409549,3.92893643816517,2.21851091848272,5.81833621896758,4.40779749680161,6.23568135536041,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,1.36595759702526,1.03480862072507,.584314421936207,.302393753783036,2.61131941894657,1.97825748822484,1.11704169972566,4.0025799484364,3.03223485513554,5.43816471058122,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.339486246269225,.25718462626926,.145221718497785,.822758173894117,.623296984369123,1.5272774292324,.0113956405123916,.00863299641447793,.00487470263420339,.067649468233801,.0512492137952295,.198554676115027,.00166206328995312,.00125912943722616,.0119492023316399,.000231576690958128},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,6.45780747384476,4.89224179250545,2.76245034842036,7.79428043908953,5.90471373874725,8.98680735137037,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.612365059546825,.463909453726739,.26195083695297,.135564507599069,1.2594400120878,.954114083786716,.538749492535305,2.07683410541886,1.57334740190396,3.03569721712443,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.0634197097850823,.0480448752738173,.0271289907701198,.165355473459674,.125268361016134,.330223979417038,.000824565926330002,.000624666483447503,.000352723805983911,.00526617557576491,.00398949710931715,.0166285878943821,5.01144631725624e-005,3.79652184082228e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,7.27823584467216,5.51377378765633,3.11340423610211,9.14177937313726,6.92553862835673,10.9691712594614,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.122554596919714,.0928436971090411,.0524250669352539,.0271309626935909,.262307293803493,.19871616027429,.112206949174718,.450140785461579,.341013195733391,.68472894504081,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,.00244089433681864,.001849148543354,.00104413911950708,.00662303395559898,.00501741243403364,.0137645079246526,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,3.89785419960172e-005,2.95289776007745e-005,.000128085303901421,6.8546484667622e-008,5.19287666029906e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,6.8546484667622e-008,6.67972275267736e-008,7.65361974719019e-010,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,.122554596919714,.119427091493105,.0867440815154734,.057745280204241,.00244089433681864,.00237860446263913,.00172766377213932,3.89785419960172e-005,3.79838375617868e-005,5.51740069821357e-007,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,.262307293803493,.255613399651286,.18566096935718,.00662303395559898,.00645401887770827,.000128085303901421,7.27823584467216,7.09250048322984,5.15153163620379,.450140785461579,.438653515481796,.0137645079246526,9.14177937313726,8.90848771819592,.68472894504081,10.9691712594614},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.000824565926330002,.000803523594969987,.000583627344081487,5.01144631725624e-005,4.88355780044979e-005,2.90963994598978e-006,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,.612365059546825,.596737942368568,.433431678432723,.288534194877923,.0634197097850823,.0618012842418601,.0448884383820082,.00526617557576491,.00513178655225592,.000387613365377294,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,1.2594400120878,1.22730000557206,.891431001025123,.165355473459674,.161135720335381,.0166285878943821,6.45780747384476,6.29300885631198,4.57083285180669,2.07683410541886,2.02383478866576,.330223979417038,7.79428043908953,7.59537599078757,3.03569721712443,8.98680735137037},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.0113956405123916,.0111048319353399,.00806582856989507,.00166206328995312,.00161964862649386,.000231576690958128,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,1.36595759702526,1.33109933870947,.966824094086502,.643611958831769,.339486246269225,.33082280056131,.24028819272175,.067649468233801,.0659231023263137,.0119492023316399,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,2.61131941894657,2.54468041880833,1.84829055872937,.822758173894117,.801761975206599,.198554676115027,5.18623489409549,5.05388589728677,3.6708144253493,4.0025799484364,3.90043697901805,1.5272774292324,5.81833621896758,5.66985645040998,5.43816471058122,6.23568135536041},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.0657431866024243,.0640654676075239,.0465329941039433,.0171974308138302,.0167585647100291,.00429748986298374,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,2.19713005267825,2.14106086934751,1.55512753646012,1.0352438319403,.979365349166059,.954372647329225,.693194296828135,.350018452610409,.341086232751863,.110884150015163,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.76697803717186,3.67084745731307,2.66626514187773,2.12867470143779,2.07435244929784,.921342735777177,3.7410698100085,3.6456003901969,2.64792731244254,5.17830525859385,5.04615862255375,3.543802987655,3.76406389114815,3.66800767948205,6.30978459469567,3.6179005979755},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,.547225517982271,.533260715077649,.387325943916553,.257841742921929,.236321146004057,.230290399761936,.167267987177829,.0946962283839188,.092279648532812,.0362494356624709,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.90029193047496,2.82627856027675,2.05282515677973,1.36655967551418,1.98037740107218,1.92983958905795,1.40171011886371,1.08420509999131,1.05653696348977,.526147189418459,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,4.28496925234002,4.17561990789474,3.03289906094317,3.70920808974538,3.61455175876414,2.4592936780291,2.39387697960251,2.33278695510428,1.69438491095747,5.07587527588423,4.94634257953458,5.32120685033247,2.07554270972356,2.02257634793674,5.32974583920556,1.71909544841957},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,.626153181688297,.610174201341807,.443190904144467,.347192252628355,.338332155278326,.183906859961414,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,3.30149822057238,3.21724635356535,2.3367987653383,1.55560007239588,3.11944010216456,3.03983422634135,2.20793812029263,2.36319452013471,2.30288749099116,1.58691847335237,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,4.0126111958998,3.91021223386622,2.84012416689539,4.80639993236566,4.6837440522375,4.40969381865368,1.33269841309056,1.29868890491693,.943283260267015,3.91021223386622,3.8104264222726,5.67229573503973,.95054294779362,.926285773305773,3.37758202463881,.647664476709677},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,1.33269841309056,1.29868890491693,.943283260267015,.95054294779362,.926285773305773,.647664476709677,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,3.30149822057238,3.21724635356535,2.3367987653383,1.55560007239588,4.0126111958998,3.91021223386622,2.84012416689539,3.91021223386622,3.8104264222726,3.37758202463881,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,3.11944010216456,3.03983422634135,2.20793812029263,4.80639993236566,4.6837440522375,5.67229573503973,.626153181688297,.610174201341807,.443190904144467,2.36319452013471,2.30288749099116,4.40969381865368,.347192252628355,.338332155278326,1.58691847335237,.183906859961414},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,2.39387697960251,2.33278695510428,1.69438491095747,2.07554270972356,2.02257634793674,1.71909544841957,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,2.90029193047496,2.82627856027675,2.05282515677973,1.36655967551418,4.28496925234002,4.17561990789474,3.03289906094317,5.07587527588423,4.94634257953458,5.32974583920556,.547225517982271,.533260715077649,.387325943916553,.257841742921929,1.98037740107218,1.92983958905795,1.40171011886371,3.70920808974538,3.61455175876414,5.32120685033247,.236321146004057,.230290399761936,.167267987177829,1.08420509999131,1.05653696348977,2.4592936780291,.0946962283839188,.092279648532812,.526147189418459,.0362494356624709},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.7410698100085,3.6456003901969,2.64792731244254,3.76406389114815,3.66800767948205,3.6179005979755,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,2.19713005267825,2.14106086934751,1.55512753646012,1.0352438319403,3.76697803717186,3.67084745731307,2.66626514187773,5.17830525859385,5.04615862255375,6.30978459469567,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.979365349166059,.954372647329225,.693194296828135,2.12867470143779,2.07435244929784,3.543802987655,.0657431866024243,.0640654676075239,.0465329941039433,.350018452610409,.341086232751863,.921342735777177,.0171974308138302,.0167585647100291,.110884150015163,.00429748986298374},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,5.18623489409549,5.05388589728677,3.6708144253493,5.81833621896758,5.66985645040998,6.23568135536041,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,1.36595759702526,1.33109933870947,.966824094086502,.643611958831769,2.61131941894657,2.54468041880833,1.84829055872937,4.0025799484364,3.90043697901805,5.43816471058122,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.339486246269225,.33082280056131,.24028819272175,.822758173894117,.801761975206599,1.5272774292324,.0113956405123916,.0111048319353399,.00806582856989507,.067649468233801,.0659231023263137,.198554676115027,.00166206328995312,.00161964862649386,.0119492023316399,.000231576690958128},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,6.45780747384476,6.29300885631198,4.57083285180669,7.79428043908953,7.59537599078757,8.98680735137037,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.612365059546825,.596737942368568,.433431678432723,.288534194877923,1.2594400120878,1.22730000557206,.891431001025123,2.07683410541886,2.02383478866576,3.03569721712443,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.0634197097850823,.0618012842418601,.0448884383820082,.165355473459674,.161135720335381,.330223979417038,.000824565926330002,.000803523594969987,.000583627344081487,.00526617557576491,.00513178655225592,.0166285878943821,5.01144631725624e-005,4.88355780044979e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,7.27823584467216,7.09250048322984,5.15153163620379,9.14177937313726,8.90848771819592,10.9691712594614,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.122554596919714,.119427091493105,.0867440815154734,.057745280204241,.262307293803493,.255613399651286,.18566096935718,.450140785461579,.438653515481796,.68472894504081,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,.00244089433681864,.00237860446263913,.00172766377213932,.00662303395559898,.00645401887770827,.0137645079246526,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,3.89785419960172e-005,3.79838375617868e-005,.000128085303901421,6.8546484667622e-008,6.67972275267736e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,6.8546484667622e-008,8.11985772602823e-008,7.65361974719019e-010,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,.122554596919714,.1451753354836,.128179867438098,.103725715888757,.00244089433681864,.00289142686579082,.00255293167466323,3.89785419960172e-005,4.61730775826967e-005,5.51740069821357e-007,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,.262307293803493,.310723141496092,.274347229515445,.00662303395559898,.00784549254339729,.000128085303901421,7.27823584467216,8.62162951636421,7.61230772976691,.450140785461579,.533226343160225,.0137645079246526,9.14177937313726,10.8291399963909,.68472894504081,10.9691712594614},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.000824565926330002,.000976761687908062,.000862413599759767,5.01144631725624e-005,5.93644317357238e-005,2.90963994598978e-006,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,.612365059546825,.72539345885478,.640472688330538,.518283326641486,.0634197097850823,.0751255185147611,.066330681941484,.00526617557576491,.00623818955179855,.000387613365377294,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,1.2594400120878,1.49190345189168,1.31724845713225,.165355473459674,.19587626195712,.0166285878943821,6.45780747384476,7.64976908082116,6.7542215997961,2.07683410541886,2.46016955394827,.330223979417038,7.79428043908953,9.23292398567978,3.03569721712443,8.98680735137037},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.0113956405123916,.0134990116693501,.0119187017581499,.00166206328995312,.00196884165709061,.000231576690958128,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,1.36595759702526,1.61808171532254,1.4286552125618,1.15609641079355,.339486246269225,.402147540031861,.355068704998853,.067649468233801,.0801359924008096,.0119492023316399,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,2.61131941894657,3.09330846862644,2.73117928815418,.822758173894117,.974620266418623,.198554676115027,5.18623489409549,6.14349367016778,5.42428368841739,4.0025799484364,4.74136345348638,1.5272774292324,5.81833621896758,6.89226625249968,5.43816471058122,6.23568135536041},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,.23320173943763,.276245376362,.243905726875982,.197373399115395,.0657431866024243,.0778778553220726,.0687608066341364,.0171974308138302,.0203716780136318,.00429748986298374,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,2.19713005267825,2.60266934505625,2.29797858240319,1.85957029228411,.979365349166059,1.16013349668145,1.02431833416123,.350018452610409,.414623747840729,.110884150015163,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.76697803717186,4.46227488639218,3.93988277320967,2.12867470143779,2.52157872198894,.921342735777177,3.7410698100085,4.43158460169426,3.91278535624427,5.17830525859385,6.13410041033272,3.543802987655,3.76406389114815,4.45882285809472,6.30978459469567,3.6179005979755},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,.547225517982271,.648230667294543,.572343234018839,.463151607819505,.236321146004057,.279940553069749,.247168314523763,.0946962283839188,.112174957679791,.0362494356624709,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.90029193047496,3.43561861072344,3.03341567369851,2.45470071573063,1.98037740107218,2.34590917578526,2.07127695727047,1.08420509999131,1.28432423671899,.526147189418459,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,4.28496925234002,5.07587527588423,4.48164984622588,3.70920808974538,4.39384194692302,2.4592936780291,2.39387697960251,2.83573119453035,2.50375621577018,5.07587527588423,6.01276422646702,5.32120685033247,2.07554270972356,2.45863979589474,5.32974583920556,1.71909544841957},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,.626153181688297,.741726548606004,.654893686699248,.347192252628355,.411275896727804,.183906859961414,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,3.30149822057238,3.91087828459092,3.45303738004469,2.79426700459607,3.11944010216456,3.6952164547113,3.26262277063996,2.36319452013471,2.79938546527299,1.58691847335237,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,4.0126111958998,4.75324623388636,4.19679052288369,4.80639993236566,5.69354997766273,4.40969381865368,1.33269841309056,1.57868365630447,1.39386942769156,3.91021223386622,4.63194679928158,5.67229573503973,.95054294779362,1.12599114834103,3.37758202463881,.647664476709677},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,1.33269841309056,1.57868365630447,1.39386942769156,.95054294779362,1.12599114834103,.647664476709677,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,3.30149822057238,3.91087828459092,3.45303738004469,2.79426700459607,4.0126111958998,4.75324623388636,4.19679052288369,3.91021223386622,4.63194679928158,3.37758202463881,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,3.11944010216456,3.6952164547113,3.26262277063996,4.80639993236566,5.69354997766273,5.67229573503973,.626153181688297,.741726548606004,.654893686699248,2.36319452013471,2.79938546527299,4.40969381865368,.347192252628355,.411275896727804,1.58691847335237,.183906859961414},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,2.39387697960251,2.83573119453035,2.50375621577018,2.07554270972356,2.45863979589474,1.71909544841957,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,2.90029193047496,3.43561861072344,3.03341567369851,2.45470071573063,4.28496925234002,5.07587527588423,4.48164984622588,5.07587527588423,6.01276422646702,5.32974583920556,.547225517982271,.648230667294543,.572343234018839,.463151607819505,1.98037740107218,2.34590917578526,2.07127695727047,3.70920808974538,4.39384194692302,5.32120685033247,.236321146004057,.279940553069749,.247168314523763,1.08420509999131,1.28432423671899,2.4592936780291,.0946962283839188,.112174957679791,.526147189418459,.0362494356624709},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.7410698100085,4.43158460169426,3.91278535624427,3.76406389114815,4.45882285809472,3.6179005979755,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,2.19713005267825,2.60266934505625,2.29797858240319,1.85957029228411,3.76697803717186,4.46227488639218,3.93988277320967,5.17830525859385,6.13410041033272,6.30978459469567,.23320173943763,.276245376362,.243905726875982,.197373399115395,.979365349166059,1.16013349668145,1.02431833416123,2.12867470143779,2.52157872198894,3.543802987655,.0657431866024243,.0778778553220726,.0687608066341364,.350018452610409,.414623747840729,.921342735777177,.0171974308138302,.0203716780136318,.110884150015163,.00429748986298374},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,5.18623489409549,6.14349367016778,5.42428368841739,5.81833621896758,6.89226625249968,6.23568135536041,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,1.36595759702526,1.61808171532254,1.4286552125618,1.15609641079355,2.61131941894657,3.09330846862644,2.73117928815418,4.0025799484364,4.74136345348638,5.43816471058122,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.339486246269225,.402147540031861,.355068704998853,.822758173894117,.974620266418623,1.5272774292324,.0113956405123916,.0134990116693501,.0119187017581499,.067649468233801,.0801359924008096,.198554676115027,.00166206328995312,.00196884165709061,.0119492023316399,.000231576690958128},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,6.45780747384476,7.64976908082116,6.7542215997961,7.79428043908953,9.23292398567978,8.98680735137037,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.612365059546825,.72539345885478,.640472688330538,.518283326641486,1.2594400120878,1.49190345189168,1.31724845713225,2.07683410541886,2.46016955394827,3.03569721712443,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.0634197097850823,.0751255185147611,.066330681941484,.165355473459674,.19587626195712,.330223979417038,.000824565926330002,.000976761687908062,.000862413599759767,.00526617557576491,.00623818955179855,.0166285878943821,5.01144631725624e-005,5.93644317357238e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,7.27823584467216,8.62162951636421,7.61230772976691,9.14177937313726,10.8291399963909,10.9691712594614,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.122554596919714,.1451753354836,.128179867438098,.103725715888757,.262307293803493,.310723141496092,.274347229515445,.450140785461579,.533226343160225,.68472894504081,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,.00244089433681864,.00289142686579082,.00255293167466323,.00662303395559898,.00784549254339729,.0137645079246526,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,3.89785419960172e-005,4.61730775826967e-005,.000128085303901421,6.8546484667622e-008,8.11985772602823e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,6.8546484667622e-008,9.42279510613512e-008,7.65361974719019e-010,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,.122554596919714,.168470616959383,.172616566596884,.162099033300108,.00244089433681864,.00335539412635106,.00343796813996252,3.89785419960172e-005,5.35821518259837e-005,5.51740069821357e-007,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,.262307293803493,.360582733816502,.369456434624521,.00662303395559898,.0091044044412468,.000128085303901421,7.27823584467216,10.0050827434195,10.2513011633895,.450140785461579,.618789484535732,.0137645079246526,9.14177937313726,12.5668171542652,.68472894504081,10.9691712594614},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.000824565926330002,.00113349587680305,.00116139045508752,5.01144631725624e-005,6.88902312961445e-005,2.90963994598978e-006,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,.612365059546825,.841792327494715,.862508275818624,.809955617122475,.0634197097850823,.0871803906043333,.0893258419252021,.00526617557576491,.00723918866107805,.000387613365377294,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,1.2594400120878,1.73129887488649,1.77390498656132,.165355473459674,.227307170255112,.0166285878943821,6.45780747384476,8.87727458903197,9.0957384017764,2.07683410541886,2.85493593796569,.330223979417038,7.79428043908953,10.7144674060855,3.03569721712443,8.98680735137037},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.0113956405123916,.0156651046591466,.0160506124472133,.00166206328995312,.00228476805347462,.000231576690958128,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,1.36595759702526,1.87772409110003,1.92393362992246,1.80670828815818,.339486246269225,.46667737293102,.478161992112396,.067649468233801,.0929948605625365,.0119492023316399,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,2.61131941894657,3.58967027297864,3.67800966645072,.822758173894117,1.13101083622578,.198554676115027,5.18623489409549,7.12929758661485,7.30474484999971,4.0025799484364,5.50217723676456,1.5272774292324,5.81833621896758,7.99822052374818,5.43816471058122,6.23568135536041},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.0657431866024243,.0903743758530202,.0925984290266871,.0171974308138302,.023640580213036,.00429748986298374,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,2.19713005267825,3.02030168445819,3.09462929659509,2.90607342788681,.979365349166059,1.3462920908025,1.37942344206699,.350018452610409,.481155551471966,.110884150015163,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.76697803717186,5.17830525859385,5.30573990048865,2.12867470143779,2.92619901025212,.921342735777177,3.7410698100085,5.14269031628637,5.26924849841671,5.17830525859385,7.11839705537515,3.543802987655,3.76406389114815,5.17429930720343,6.30978459469567,3.6179005979755},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,.547225517982271,.752247756737171,.770760072903028,.723797635481732,.236321146004057,.324860676440036,.332855281233889,.0946962283839188,.130174896891091,.0362494356624709,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.90029193047496,3.98690855480132,4.08502371800223,3.83612308338039,1.98037740107218,2.72234098759749,2.78933598529375,1.08420509999131,1.49041085949741,.526147189418459,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,4.28496925234002,5.89036585649012,6.03532383522368,3.70920808974538,5.09889135888211,2.4592936780291,2.39387697960251,3.29076135613591,3.37174480048962,5.07587527588423,6.97759089372753,5.32120685033247,2.07554270972356,2.8531607102232,5.32974583920556,1.71909544841957},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,.626153181688297,.860746275133792,.881928667453122,.347192252628355,.477270493857987,.183906859961414,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,3.30149822057238,4.53842985975064,4.65011759480767,4.36678577097684,3.11944010216456,4.28816226840408,4.39369108444944,2.36319452013471,3.24858347921606,1.58691847335237,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,4.0126111958998,5.51596676470681,5.65171103126921,4.80639993236566,6.60715454337131,4.40969381865368,1.33269841309056,1.83200409818481,1.87708849838432,3.91021223386622,5.37520324921971,5.67229573503973,.95054294779362,1.306671156088,3.37758202463881,.647664476709677},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,1.33269841309056,1.83200409818481,1.87708849838432,.95054294779362,1.306671156088,.647664476709677,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,3.30149822057238,4.53842985975064,4.65011759480767,4.36678577097684,4.0126111958998,5.51596676470681,5.65171103126921,3.91021223386622,5.37520324921971,3.37758202463881,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,3.11944010216456,4.28816226840408,4.39369108444944,4.80639993236566,6.60715454337131,5.67229573503973,.626153181688297,.860746275133792,.881928667453122,2.36319452013471,3.24858347921606,4.40969381865368,.347192252628355,.477270493857987,1.58691847335237,.183906859961414},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,2.39387697960251,3.29076135613591,3.37174480048962,2.07554270972356,2.8531607102232,1.71909544841957,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,2.90029193047496,3.98690855480132,4.08502371800223,3.83612308338039,4.28496925234002,5.89036585649012,6.03532383522368,5.07587527588423,6.97759089372753,5.32974583920556,.547225517982271,.752247756737171,.770760072903028,.723797635481732,1.98037740107218,2.72234098759749,2.78933598529375,3.70920808974538,5.09889135888211,5.32120685033247,.236321146004057,.324860676440036,.332855281233889,1.08420509999131,1.49041085949741,2.4592936780291,.0946962283839188,.130174896891091,.526147189418459,.0362494356624709},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.7410698100085,5.14269031628637,5.26924849841671,3.76406389114815,5.17429930720343,3.6179005979755,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,2.19713005267825,3.02030168445819,3.09462929659509,2.90607342788681,3.76697803717186,5.17830525859385,5.30573990048865,5.17830525859385,7.11839705537515,6.30978459469567,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.979365349166059,1.3462920908025,1.37942344206699,2.12867470143779,2.92619901025212,3.543802987655,.0657431866024243,.0903743758530202,.0925984290266871,.350018452610409,.481155551471966,.921342735777177,.0171974308138302,.023640580213036,.110884150015163,.00429748986298374},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,5.18623489409549,7.12929758661485,7.30474484999971,5.81833621896758,7.99822052374818,6.23568135536041,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,1.36595759702526,1.87772409110003,1.92393362992246,1.80670828815818,2.61131941894657,3.58967027297864,3.67800966645072,4.0025799484364,5.50217723676456,5.43816471058122,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.339486246269225,.46667737293102,.478161992112396,.822758173894117,1.13101083622578,1.5272774292324,.0113956405123916,.0156651046591466,.0160506124472133,.067649468233801,.0929948605625365,.198554676115027,.00166206328995312,.00228476805347462,.0119492023316399,.000231576690958128},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,6.45780747384476,8.87727458903197,9.0957384017764,7.79428043908953,10.7144674060855,8.98680735137037,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.612365059546825,.841792327494715,.862508275818624,.809955617122475,1.2594400120878,1.73129887488649,1.77390498656132,2.07683410541886,2.85493593796569,3.03569721712443,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.0634197097850823,.0871803906043333,.0893258419252021,.165355473459674,.227307170255112,.330223979417038,.000824565926330002,.00113349587680305,.00116139045508752,.00526617557576491,.00723918866107805,.0166285878943821,5.01144631725624e-005,6.88902312961445e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,7.27823584467216,10.0050827434195,10.2513011633895,9.14177937313726,12.5668171542652,10.9691712594614,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.122554596919714,.168470616959383,.172616566596884,.162099033300108,.262307293803493,.360582733816502,.369456434624521,.450140785461579,.618789484535732,.68472894504081,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,.00244089433681864,.00335539412635106,.00343796813996252,.00662303395559898,.0091044044412468,.0137645079246526,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,3.89785419960172e-005,5.35821518259837e-005,.000128085303901421,6.8546484667622e-008,9.42279510613512e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,6.8546484667622e-008,1.05066727253324e-007,7.65361974719019e-010,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,.122554596919714,.187849318200136,.214611696543555,.22471744861619,.00244089433681864,.00374135567557013,.00427437638070464,3.89785419960172e-005,5.97455560493016e-005,5.51740069821357e-007,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,.262307293803493,.402059551538899,.459339875638177,.00662303395559898,.0101516584777441,.000128085303901421,7.27823584467216,11.1559392718342,12.7452954129935,.450140785461579,.689967098580019,.0137645079246526,9.14177937313726,14.0123427870133,.68472894504081,10.9691712594614},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.000824565926330002,.00126387871952445,.00144394006224153,5.01144631725624e-005,7.68144808465349e-005,2.90963994598978e-006,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,.612365059546825,.938621331363155,1.07234414405062,1.12283926724684,.0634197097850823,.0972085057383859,.111057535546505,.00526617557576491,.0080718921723515,.000387613365377294,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,1.2594400120878,1.93044531514061,2.2054705766571,.165355473459674,.253453674742105,.0166285878943821,6.45780747384476,9.89840251743778,11.3086008382978,2.07683410541886,3.18333118932674,.330223979417038,7.79428043908953,11.9469224570834,3.03569721712443,8.98680735137037},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.0113956405123916,.0174670175895654,.0199554957891379,.00166206328995312,.00254757849669498,.000231576690958128,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,1.36595759702526,2.0937134121504,2.39199903285711,2.50463478185567,.339486246269225,.520357958596769,.594491933034081,.067649468233801,.10369179782253,.0119492023316399,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,2.61131941894657,4.0025799484364,4.57281656090381,.822758173894117,1.26110783171881,.198554676115027,5.18623489409549,7.94936063666436,9.0818842940915,4.0025799484364,6.13507720928999,1.5272774292324,5.81833621896758,8.91823333538729,5.43816471058122,6.23568135536041},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.0657431866024243,.100769886127061,.115126296058794,.0171974308138302,.0263598896651818,.00429748986298374,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,2.19713005267825,3.36771849254271,3.8475081309354,4.02868168249338,.979365349166059,1.50115228352525,1.71501734155852,.350018452610409,.53650152129502,.110884150015163,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.76697803717186,5.77395114836884,6.59655016845447,2.12867470143779,3.262791452389,.921342735777177,3.7410698100085,5.7342395194157,6.55118093268348,5.17830525859385,7.93720624797419,3.543802987655,3.76406389114815,5.76948440365682,6.30978459469567,3.6179005979755},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,.236321146004057,.362228486371654,.41383418762992,.0946962283839188,.145148549160118,.0362494356624709,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.90029193047496,4.44551143257392,5.07885128191947,5.3180069882276,1.98037740107218,3.0354842147445,3.46794134442417,1.08420509999131,1.66184862884528,.526147189418459,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,4.28496925234002,6.56791807415177,7.50363138951749,3.70920808974538,5.68540249112695,2.4592936780291,2.39387697960251,3.66928837957851,4.19204185444651,5.07587527588423,7.78020354957332,5.32120685033247,2.07554270972356,3.1813517621299,5.32974583920556,1.71909544841957},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,.626153181688297,.959755498290674,1.09648923787555,.347192252628355,.532169576430532,.183906859961414,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,3.30149822057238,5.06047268206324,5.78142437787735,6.0536632275323,3.11944010216456,4.78141751356834,5.46261298270436,2.36319452013471,3.62225890010297,1.58691847335237,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,4.0126111958998,6.15045290784802,7.02669110333533,4.80639993236566,7.36715694769769,4.40969381865368,1.33269841309056,2.04273437703882,2.33375715403626,3.91021223386622,5.99349776107556,5.67229573503973,.95054294779362,1.45697386412547,3.37758202463881,.647664476709677},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,1.33269841309056,2.04273437703882,2.33375715403626,.95054294779362,1.45697386412547,.647664476709677,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,3.30149822057238,5.06047268206324,5.78142437787735,6.0536632275323,4.0126111958998,6.15045290784802,7.02669110333533,3.91021223386622,5.99349776107556,3.37758202463881,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,3.11944010216456,4.78141751356834,5.46261298270436,4.80639993236566,7.36715694769769,5.67229573503973,.626153181688297,.959755498290674,1.09648923787555,2.36319452013471,3.62225890010297,4.40969381865368,.347192252628355,.532169576430532,1.58691847335237,.183906859961414},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,2.39387697960251,3.66928837957851,4.19204185444651,2.07554270972356,3.1813517621299,1.71909544841957,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,2.90029193047496,4.44551143257392,5.07885128191947,5.3180069882276,4.28496925234002,6.56791807415177,7.50363138951749,5.07587527588423,7.78020354957332,5.32974583920556,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,1.98037740107218,3.0354842147445,3.46794134442417,3.70920808974538,5.68540249112695,5.32120685033247,.236321146004057,.362228486371654,.41383418762992,1.08420509999131,1.66184862884528,2.4592936780291,.0946962283839188,.145148549160118,.526147189418459,.0362494356624709},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.7410698100085,5.7342395194157,6.55118093268348,3.76406389114815,5.76948440365682,3.6179005979755,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,2.19713005267825,3.36771849254271,3.8475081309354,4.02868168249338,3.76697803717186,5.77395114836884,6.59655016845447,5.17830525859385,7.93720624797419,6.30978459469567,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.979365349166059,1.50115228352525,1.71501734155852,2.12867470143779,3.262791452389,3.543802987655,.0657431866024243,.100769886127061,.115126296058794,.350018452610409,.53650152129502,.921342735777177,.0171974308138302,.0263598896651818,.110884150015163,.00429748986298374},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,5.18623489409549,7.94936063666436,9.0818842940915,5.81833621896758,8.91823333538729,6.23568135536041,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,1.36595759702526,2.0937134121504,2.39199903285711,2.50463478185567,2.61131941894657,4.0025799484364,4.57281656090381,4.0025799484364,6.13507720928999,5.43816471058122,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.339486246269225,.520357958596769,.594491933034081,.822758173894117,1.26110783171881,1.5272774292324,.0113956405123916,.0174670175895654,.0199554957891379,.067649468233801,.10369179782253,.198554676115027,.00166206328995312,.00254757849669498,.0119492023316399,.000231576690958128},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,6.45780747384476,9.89840251743778,11.3086008382978,7.79428043908953,11.9469224570834,8.98680735137037,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.612365059546825,.938621331363155,1.07234414405062,1.12283926724684,1.2594400120878,1.93044531514061,2.2054705766571,2.07683410541886,3.18333118932674,3.03569721712443,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.0634197097850823,.0972085057383859,.111057535546505,.165355473459674,.253453674742105,.330223979417038,.000824565926330002,.00126387871952445,.00144394006224153,.00526617557576491,.0080718921723515,.0166285878943821,5.01144631725624e-005,7.68144808465349e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,7.27823584467216,11.1559392718342,12.7452954129935,9.14177937313726,14.0123427870133,10.9691712594614,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.122554596919714,.187849318200136,.214611696543555,.22471744861619,.262307293803493,.402059551538899,.459339875638177,.450140785461579,.689967098580019,.68472894504081,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,.00244089433681864,.00374135567557013,.00427437638070464,.00662303395559898,.0101516584777441,.0137645079246526,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,3.89785419960172e-005,5.97455560493016e-005,.000128085303901421,6.8546484667622e-008,1.05066727253324e-007,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,6.8546484667622e-008,1.13034107101112e-007,7.65361974719019e-010,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,.122554596919714,.202094235800358,.248394511041402,.279814171322846,.00244089433681864,.00402506872719183,.00494722164817595,3.89785419960172e-005,6.42761581885916e-005,5.51740069821357e-007,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,.262307293803493,.432548377566746,.531646250640432,.00662303395559898,.0109214751579782,.000128085303901421,7.27823584467216,12.0019121888667,14.7515791225583,.450140785461579,.742288419521226,.0137645079246526,9.14177937313726,15.0749214021478,.68472894504081,10.9691712594614},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.000824565926330002,.00135972068684592,.00167123596481511,5.01144631725624e-005,8.26394471580808e-005,2.90963994598978e-006,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,.612365059546825,1.00979850491474,1.24114576986956,1.39813949040533,.0634197097850823,.104579983939912,.128539509653141,.00526617557576491,.00868399681012567,.000387613365377294,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,1.2594400120878,2.07683410541886,2.55264179132826,.165355473459674,.272673476798176,.0166285878943821,6.45780747384476,10.6490143886212,13.0887291840651,2.07683410541886,3.42472834168617,.330223979417038,7.79428043908953,12.8528769082785,3.03569721712443,8.98680735137037},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.0113956405123916,.0187915697820833,.0230967635919399,.00166206328995312,.00274076549419503,.000231576690958128,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,1.36595759702526,2.25248308627179,2.76853237613478,3.11872669553275,.339486246269225,.559817544151044,.688073089222744,.067649468233801,.111554914547963,.0119492023316399,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,2.61131941894657,4.306102059137,5.29264039202637,.822758173894117,1.35673967813674,.198554676115027,5.18623489409549,8.55217350992208,10.5114970194908,4.0025799484364,6.60031003608269,1.5272774292324,5.81833621896758,9.5945173923078,5.43816471058122,6.23568135536041},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.0657431866024243,.108411429563137,.133248748684711,.0171974308138302,.0283588027292761,.00429748986298374,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,2.19713005267825,3.62309803230714,4.45315850116099,5.01644279717582,.979365349166059,1.61498708893785,1.98498451316933,.350018452610409,.577185299317033,.110884150015163,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.76697803717186,6.21179920192732,7.63493733120357,2.12867470143779,3.51021420500431,.921342735777177,3.7410698100085,6.16907617592665,7.58242635758397,5.17830525859385,8.53909743428073,3.543802987655,3.76406389114815,6.20699373673991,6.30978459469567,3.6179005979755},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,.236321146004057,.389696858310697,.47897734533629,.0946962283839188,.156155398385808,.0362494356624709,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.90029193047496,4.78262175405185,5.87833189496431,6.62188774242252,1.98037740107218,3.26566989191443,4.01384271427013,1.08420509999131,1.78786929801119,.526147189418459,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,4.28496925234002,7.06597392374297,8.68480553507875,3.70920808974538,6.11653575680062,2.4592936780291,2.39387697960251,3.94753645159365,4.85192654207924,5.07587527588423,8.370188967346,5.32120685033247,2.07554270972356,3.42259881132375,5.32974583920556,1.71909544841957},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,.626153181688297,1.03253530989981,1.26909163149447,.347192252628355,.572524856067592,.183906859961414,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,3.30149822057238,5.4442165096419,6.69149959948354,7.53791381092379,3.11944010216456,5.1440001364154,6.3225029329745,2.36319452013471,3.8969406506306,1.58691847335237,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,4.0126111958998,6.61685169872888,8.13278832870006,4.80639993236566,7.92582037364658,4.40969381865368,1.33269841309056,2.19763822848137,2.70112242949704,3.91021223386622,6.44799438933945,5.67229573503973,.95054294779362,1.56745854854701,3.37758202463881,.647664476709677},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,1.33269841309056,2.19763822848137,2.70112242949704,.95054294779362,1.56745854854701,.647664476709677,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,3.30149822057238,5.4442165096419,6.69149959948354,7.53791381092379,4.0126111958998,6.61685169872888,8.13278832870006,3.91021223386622,6.44799438933945,3.37758202463881,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,3.11944010216456,5.1440001364154,6.3225029329745,4.80639993236566,7.92582037364658,5.67229573503973,.626153181688297,1.03253530989981,1.26909163149447,2.36319452013471,3.8969406506306,4.40969381865368,.347192252628355,.572524856067592,1.58691847335237,.183906859961414},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,2.39387697960251,3.94753645159365,4.85192654207924,2.07554270972356,3.42259881132375,1.71909544841957,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,2.90029193047496,4.78262175405185,5.87833189496431,6.62188774242252,4.28496925234002,7.06597392374297,8.68480553507875,5.07587527588423,8.370188967346,5.32974583920556,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,1.98037740107218,3.26566989191443,4.01384271427013,3.70920808974538,6.11653575680062,5.32120685033247,.236321146004057,.389696858310697,.47897734533629,1.08420509999131,1.78786929801119,2.4592936780291,.0946962283839188,.156155398385808,.526147189418459,.0362494356624709},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.7410698100085,6.16907617592665,7.58242635758397,3.76406389114815,6.20699373673991,3.6179005979755,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,2.19713005267825,3.62309803230714,4.45315850116099,5.01644279717582,3.76697803717186,6.21179920192732,7.63493733120357,5.17830525859385,8.53909743428073,6.30978459469567,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.979365349166059,1.61498708893785,1.98498451316933,2.12867470143779,3.51021420500431,3.543802987655,.0657431866024243,.108411429563137,.133248748684711,.350018452610409,.577185299317033,.921342735777177,.0171974308138302,.0283588027292761,.110884150015163,.00429748986298374},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,5.18623489409549,8.55217350992208,10.5114970194908,5.81833621896758,9.5945173923078,6.23568135536041,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,1.36595759702526,2.25248308627179,2.76853237613478,3.11872669553275,2.61131941894657,4.306102059137,5.29264039202637,4.0025799484364,6.60031003608269,5.43816471058122,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.339486246269225,.559817544151044,.688073089222744,.822758173894117,1.35673967813674,1.5272774292324,.0113956405123916,.0187915697820833,.0230967635919399,.067649468233801,.111554914547963,.198554676115027,.00166206328995312,.00274076549419503,.0119492023316399,.000231576690958128},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,6.45780747384476,10.6490143886212,13.0887291840651,7.79428043908953,12.8528769082785,8.98680735137037,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.612365059546825,1.00979850491474,1.24114576986956,1.39813949040533,1.2594400120878,2.07683410541886,2.55264179132826,2.07683410541886,3.42472834168617,3.03569721712443,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.0634197097850823,.104579983939912,.128539509653141,.165355473459674,.272673476798176,.330223979417038,.000824565926330002,.00135972068684592,.00167123596481511,.00526617557576491,.00868399681012567,.0166285878943821,5.01144631725624e-005,8.26394471580808e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,7.27823584467216,12.0019121888667,14.7515791225583,9.14177937313726,15.0749214021478,10.9691712594614,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.122554596919714,.202094235800358,.248394511041402,.279814171322846,.262307293803493,.432548377566746,.531646250640432,.450140785461579,.742288419521226,.68472894504081,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,.00244089433681864,.00402506872719183,.00494722164817595,.00662303395559898,.0109214751579782,.0137645079246526,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,3.89785419960172e-005,6.42761581885916e-005,.000128085303901421,6.8546484667622e-008,1.13034107101112e-007,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,6.8546484667622e-008,1.17631378114674e-007,7.65361974719019e-010,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,.122554596919714,.21031371924726,.269010572516923,.315362998946248,.00244089433681864,.00418877446399711,.00535782744298286,3.89785419960172e-005,6.68903684166533e-005,5.51740069821357e-007,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,.262307293803493,.450140785461579,.575771427724586,.00662303395559898,.0113656683529065,.000128085303901421,7.27823584467216,12.4900484198527,15.9759196314393,.450140785461579,.77247843138831,.0137645079246526,9.14177937313726,15.6880416449771,.68472894504081,10.9691712594614},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.000824565926330002,.00141502261880692,.00180994395496465,5.01144631725624e-005,8.60005205963556e-005,2.90963994598978e-006,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,.612365059546825,1.05086856345938,1.34415745633734,1.57576530364745,.0634197097850823,.108833412759726,.139207935545212,.00526617557576491,.00903718831878549,.000387613365377294,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,1.2594400120878,2.16130214323225,2.76450404172331,.165355473459674,.283763526546879,.0166285878943821,6.45780747384476,11.0821261849396,14.1750577199246,2.07683410541886,3.56401731152317,.330223979417038,7.79428043908953,13.3756231834223,3.03569721712443,8.98680735137037},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.0113956405123916,.019555851831759,.0250137314673599,.00166206328995312,.00285223664290029,.000231576690958128,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,1.36595759702526,2.34409503833324,2.99831295149481,3.51494350320858,.339486246269225,.582586184826084,.745181263826102,.067649468233801,.116092024524981,.0119492023316399,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,2.61131941894657,4.48123785385959,5.73191499287154,.822758173894117,1.41192036788792,.198554676115027,5.18623489409549,8.90000356218188,11.3839223715851,4.0025799484364,6.86875479835488,1.5272774292324,5.81833621896758,9.98474117367778,5.43816471058122,6.23568135536041},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.0657431866024243,.112820689170272,.144308028468726,.0171974308138302,.0295121988599679,.00429748986298374,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,2.19713005267825,3.77045500260921,4.82275841315291,5.6537538362715,.979365349166059,1.68067109814234,2.14973277020564,.350018452610409,.600660313311091,.110884150015163,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.76697803717186,6.46444263093715,8.26861613804166,2.12867470143779,3.65298001640339,.921342735777177,3.7410698100085,6.41998199374919,8.21174689798638,5.17830525859385,8.88639566242957,3.543802987655,3.76406389114815,6.45944172008843,6.30978459469567,3.6179005979755},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,.236321146004057,.405546428999883,.518731148089159,.0946962283839188,.162506478648407,.0362494356624709,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.90029193047496,4.97713833778588,6.36621728024122,7.46316159493692,1.98037740107218,3.39848970992327,4.34698062381386,1.08420509999131,1.86058469259941,.526147189418459,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,4.28496925234002,7.35335795264632,9.40562051630952,3.70920808974538,6.36530467495559,2.4592936780291,2.39387697960251,4.10808883148424,5.25462310508641,5.07587527588423,8.71061742831661,5.32120685033247,2.07554270972356,3.5618011698851,5.32974583920556,1.71909544841957},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,.626153181688297,1.07453010927869,1.3744227476422,.347192252628355,.595810322665643,.183906859961414,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,3.30149822057238,5.66564116980173,7.24687566849574,8.49556365917046,3.11944010216456,5.35321453486018,6.84725329319339,2.36319452013471,4.05543522146567,1.58691847335237,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,4.0126111958998,6.88596923586654,8.80778739952872,4.80639993236566,8.24817568034947,4.40969381865368,1.33269841309056,2.28701955580991,2.92530816462441,3.91021223386622,6.71024423995495,5.67229573503973,.95054294779362,1.63120949890175,3.37758202463881,.647664476709677},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,1.33269841309056,2.28701955580991,2.92530816462441,.95054294779362,1.63120949890175,.647664476709677,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,3.30149822057238,5.66564116980173,7.24687566849574,8.49556365917046,4.0126111958998,6.88596923586654,8.80778739952872,3.91021223386622,6.71024423995495,3.37758202463881,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,3.11944010216456,5.35321453486018,6.84725329319339,4.80639993236566,8.24817568034947,5.67229573503973,.626153181688297,1.07453010927869,1.3744227476422,2.36319452013471,4.05543522146567,4.40969381865368,.347192252628355,.595810322665643,1.58691847335237,.183906859961414},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,2.39387697960251,4.10808883148424,5.25462310508641,2.07554270972356,3.5618011698851,1.71909544841957,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,2.90029193047496,4.97713833778588,6.36621728024122,7.46316159493692,4.28496925234002,7.35335795264632,9.40562051630952,5.07587527588423,8.71061742831661,5.32974583920556,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,1.98037740107218,3.39848970992327,4.34698062381386,3.70920808974538,6.36530467495559,5.32120685033247,.236321146004057,.405546428999883,.518731148089159,1.08420509999131,1.86058469259941,2.4592936780291,.0946962283839188,.162506478648407,.526147189418459,.0362494356624709},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.7410698100085,6.41998199374919,8.21174689798638,3.76406389114815,6.45944172008843,3.6179005979755,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,2.19713005267825,3.77045500260921,4.82275841315291,5.6537538362715,3.76697803717186,6.46444263093715,8.26861613804166,5.17830525859385,8.88639566242957,6.30978459469567,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.979365349166059,1.68067109814234,2.14973277020564,2.12867470143779,3.65298001640339,3.543802987655,.0657431866024243,.112820689170272,.144308028468726,.350018452610409,.600660313311091,.921342735777177,.0171974308138302,.0295121988599679,.110884150015163,.00429748986298374},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,5.18623489409549,8.90000356218188,11.3839223715851,5.81833621896758,9.98474117367778,6.23568135536041,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,1.36595759702526,2.34409503833324,2.99831295149481,3.51494350320858,2.61131941894657,4.48123785385959,5.73191499287154,4.0025799484364,6.86875479835488,5.43816471058122,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.339486246269225,.582586184826084,.745181263826102,.822758173894117,1.41192036788792,1.5272774292324,.0113956405123916,.019555851831759,.0250137314673599,.067649468233801,.116092024524981,.198554676115027,.00166206328995312,.00285223664290029,.0119492023316399,.000231576690958128},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,6.45780747384476,11.0821261849396,14.1750577199246,7.79428043908953,13.3756231834223,8.98680735137037,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.612365059546825,1.05086856345938,1.34415745633734,1.57576530364745,1.2594400120878,2.16130214323225,2.76450404172331,2.07683410541886,3.56401731152317,3.03569721712443,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.0634197097850823,.108833412759726,.139207935545212,.165355473459674,.283763526546879,.330223979417038,.000824565926330002,.00141502261880692,.00180994395496465,.00526617557576491,.00903718831878549,.0166285878943821,5.01144631725624e-005,8.60005205963556e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,7.27823584467216,12.4900484198527,15.9759196314393,9.14177937313726,15.6880416449771,10.9691712594614,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.122554596919714,.21031371924726,.269010572516923,.315362998946248,.262307293803493,.450140785461579,.575771427724586,.450140785461579,.77247843138831,.68472894504081,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,.00244089433681864,.00418877446399711,.00535782744298286,.00662303395559898,.0113656683529065,.0137645079246526,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,3.89785419960172e-005,6.68903684166533e-005,.000128085303901421,6.8546484667622e-008,1.17631378114674e-007,5.51740069821357e-007,7.65361974719019e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,0.,0.,0.,6.8546484667622e-008,1.09461601509022e-009,0.,7.65361974719019e-010},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,0.,0.,0.,0.,0.,0.,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,0.,0.,0.,5.01144631725624e-005,8.00275816365055e-007,0.,2.90963994598978e-006},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,0.,0.,0.,0.,0.,0.,.0113956405123916,.000181976518058871,2.16598631798674e-006,0.,0.,0.,.00166206328995312,2.65414208197256e-005,0.,.000231576690958128},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,0.,0.,0.,0.,0.,0.,.0657431866024243,.00104985026256264,1.24959051250218e-005,0.,0.,0.,.0171974308138302,.000274625070541602,0.,.00429748986298374},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,0.,0.,0.,0.,0.,0.,.236321146004057,.00377380273155657,4.4917911225712e-005,0.,0.,0.,.0946962283839188,.00151220020487266,0.,.0362494356624709},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,0.,0.,0.,0.,0.,0.,.626153181688297,.00999901459257298,.000119013865260662,0.,0.,0.,.347192252628355,.00554429890730429,0.,.183906859961414},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,0.,0.,0.,0.,0.,0.,1.33269841309056,.0212818065446243,.000253307966815722,0.,0.,0.,.95054294779362,.0151791815252261,0.,.647664476709677},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,0.,0.,0.,0.,0.,0.,2.39387697960251,.0382277237453785,.000455007753107355,0.,0.,0.,2.07554270972356,.0331442567927967,0.,1.71909544841957},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,0.,0.,0.,0.,0.,0.,3.7410698100085,.0597409910483066,.000711070695350591,0.,0.,0.,3.76406389114815,.0601081826012263,0.,3.6179005979755},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,0.,0.,0.,0.,0.,0.,5.18623489409549,.0828187732700635,.000985755369367884,0.,0.,0.,5.81833621896758,.0929127735338078,0.,6.23568135536041},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,0.,0.,0.,0.,0.,0.,6.45780747384476,.103124463877826,.00122744505825102,0.,0.,0.,7.79428043908953,.124466546112493,0.,8.98680735137037},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,0.,0.,0.,0.,0.,0.,7.27823584467216,.116225851033514,.00138338509726576,0.,0.,0.,9.14177937313726,.145984701575574,0.,10.9691712594614},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,0.,0.,0.,6.8546484667622e-008,5.69188702865245e-009,0.,7.65361974719019e-010},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,0.,0.,0.,0.,0.,0.,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,0.,0.,0.,5.01144631725624e-005,4.16134925463982e-006,0.,2.90963994598978e-006},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,0.,0.,0.,0.,0.,0.,.0113956405123916,.000946258567734596,5.85658777413804e-005,0.,0.,0.,.00166206328995312,.000138012569524986,0.,.000231576690958128},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,0.,0.,0.,0.,0.,0.,.0657431866024243,.0054591098696972,.000337875473054763,0.,0.,0.,.0171974308138302,.00142802120123345,0.,.00429748986298374},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,0.,0.,0.,0.,0.,0.,.236321146004057,.0196233734207421,.0012145307084342,0.,0.,0.,.0946962283839188,.00786328046747136,0.,.0362494356624709},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,0.,0.,0.,0.,0.,0.,.626153181688297,.0519938139714517,.00321800346775215,0.,0.,0.,.347192252628355,.028829765505355,0.,.183906859961414},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,0.,0.,0.,0.,0.,0.,1.33269841309056,.110663133873164,.00684916764813009,0.,0.,0.,.95054294779362,.0789301318799674,0.,.647664476709677},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,0.,0.,0.,0.,0.,0.,2.39387697960251,.198780103635972,.0123029070952925,0.,0.,0.,2.07554270972356,.172346615354148,0.,1.71909544841957},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,0.,0.,0.,0.,0.,0.,3.7410698100085,.310646808870845,.0192265662361566,0.,0.,0.,3.76406389114815,.31255616594975,0.,3.6179005979755},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,0.,0.,0.,0.,0.,0.,5.18623489409549,.430648825529863,.0266537364367886,0.,0.,0.,5.81833621896758,.483136554903789,0.,6.23568135536041},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,0.,0.,0.,0.,0.,0.,6.45780747384476,.53623626019628,.0331887586818224,0.,0.,0.,7.79428043908953,.647212821256298,0.,8.98680735137037},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,0.,0.,0.,0.,0.,0.,7.27823584467216,.60436208201944,.037405205103521,0.,0.,0.,9.14177937313726,.759104944404841,0.,10.9691712594614},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,0.,0.,0.,6.8546484667622e-008,1.36592668764407e-008,0.,7.65361974719019e-010},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,0.,0.,0.,0.,0.,0.,.000824565926330002,.000164311358893045,2.44047036237916e-005,0.,0.,0.,5.01144631725624e-005,9.98631556618579e-006,0.,2.90963994598978e-006},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,0.,0.,0.,0.,0.,0.,.0113956405123916,.00227081076025246,.000337277130218073,0.,0.,0.,.00166206328995312,.000331199567025036,0.,.000231576690958128},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,0.,0.,0.,0.,0.,0.,.0657431866024243,.0131006533057736,.00194580315907169,0.,0.,0.,.0171974308138302,.00342693426532766,0.,.00429748986298374},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,0.,0.,0.,0.,0.,0.,.236321146004057,.0470917453597858,.00699440438187064,0.,0.,0.,.0946962283839188,.0188701296931614,0.,.0362494356624709},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,0.,0.,0.,0.,0.,0.,.626153181688297,.124773625580585,.0185322753878643,0.,0.,0.,.347192252628355,.0691850451424154,0.,.183906859961414},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,0.,0.,0.,0.,0.,0.,1.33269841309056,.26556698531571,.0394439167964793,0.,0.,0.,.95054294779362,.189414816301507,0.,.647664476709677},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,0.,0.,0.,0.,0.,0.,2.39387697960251,.477028175651108,.0708516521644813,0.,0.,0.,2.07554270972356,.413593664547998,0.,1.71909544841957},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,0.,0.,0.,0.,0.,0.,3.7410698100085,.745483465381799,.110724560685561,0.,0.,0.,3.76406389114815,.750065499032834,0.,3.6179005979755},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,0.,0.,0.,0.,0.,0.,5.18623489409549,1.03346169878758,.153497157076452,0.,0.,0.,5.81833621896758,1.1594206118243,0.,6.23568135536041},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,0.,0.,0.,0.,0.,0.,6.45780747384476,1.28684813137965,.191131930663375,0.,0.,0.,7.79428043908953,1.55316727245144,0.,8.98680735137037},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,0.,0.,0.,0.,0.,0.,7.27823584467216,1.45033499905202,.215414174926983,0.,0.,0.,9.14177937313726,1.82168355953934,0.,10.9691712594614},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,0.,0.,0.,6.8546484667622e-008,2.44980430684131e-008,0.,7.65361974719019e-010},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,0.,0.,0.,0.,0.,0.,.000824565926330002,.000294694201614443,7.85021488046878e-005,0.,0.,0.,5.01144631725624e-005,1.79105651165762e-005,0.,2.90963994598978e-006},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,0.,0.,0.,0.,0.,0.,.0113956405123916,.00407272369067121,.00108491296894855,0.,0.,0.,.00166206328995312,.000594010010245393,0.,.000231576690958128},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,0.,0.,0.,0.,0.,0.,.0657431866024243,.0234961635798143,.00625902823868614,0.,0.,0.,.0171974308138302,.0061462437174735,0.,.00429748986298374},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,0.,0.,0.,0.,0.,0.,.236321146004057,.0844595552914031,.0224987683542483,0.,0.,0.,.0946962283839188,.0338437819621889,0.,.0362494356624709},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,0.,0.,0.,0.,0.,0.,.626153181688297,.223782848737467,.0596124199094679,0.,0.,0.,.347192252628355,.12408412771496,0.,.183906859961414},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,0.,0.,0.,0.,0.,0.,1.33269841309056,.476297264169728,.126878501518793,0.,0.,0.,.95054294779362,.339717524338981,0.,.647664476709677},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,0.,0.,0.,0.,0.,0.,2.39387697960251,.855555199093713,.227907170151078,0.,0.,0.,2.07554270972356,.741784716454702,0.,1.71909544841957},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,0.,0.,0.,0.,0.,0.,3.7410698100085,1.33703266851113,.356165601240813,0.,0.,0.,3.76406389114815,1.34525059548623,0.,3.6179005979755},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,0.,0.,0.,0.,0.,0.,5.18623489409549,1.85352474883709,.493751403486214,0.,0.,0.,5.81833621896758,2.07943342346341,0.,6.23568135536041},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,0.,0.,0.,0.,0.,0.,6.45780747384476,2.30797605978546,.614810468242533,0.,0.,0.,7.79428043908953,2.78562232344933,0.,8.98680735137037},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,0.,0.,0.,0.,0.,0.,7.27823584467216,2.60119152746664,.692918704338266,0.,0.,0.,9.14177937313726,3.26720919228747,0.,10.9691712594614},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,0.,0.,0.,6.8546484667622e-008,3.7527416869482e-008,0.,7.65361974719019e-010},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,0.,0.,0.,0.,0.,0.,.000824565926330002,.000451428390509428,.000184211229066994,0.,0.,0.,5.01144631725624e-005,2.74363646769969e-005,0.,2.90963994598978e-006},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,0.,0.,0.,0.,0.,0.,.0113956405123916,.00623881668046779,.00254583033055523,0.,0.,0.,.00166206328995312,.000909936406629402,0.,.000231576690958128},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,0.,0.,0.,0.,0.,0.,.0657431866024243,.0359926841107618,.0146872831147845,0.,0.,0.,.0171974308138302,.00941514591687767,0.,.00429748986298374},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,0.,0.,0.,0.,0.,0.,.236321146004057,.129379678661691,.0527950614618356,0.,0.,0.,.0946962283839188,.0518437211734884,0.,.0362494356624709},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,0.,0.,0.,0.,0.,0.,.626153181688297,.342802575265255,.139885051637275,0.,0.,0.,.347192252628355,.190078724845143,0.,.183906859961414},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,0.,0.,0.,0.,0.,0.,1.33269841309056,.729617706050062,.297729999278181,0.,0.,0.,.95054294779362,.520397532085943,0.,.647664476709677},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,0.,0.,0.,0.,0.,0.,2.39387697960251,1.31058536069927,.534801410737989,0.,0.,0.,2.07554270972356,1.13630563078316,0.,1.71909544841957},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,0.,0.,0.,0.,0.,0.,3.7410698100085,2.04813838310324,.835769519114577,0.,0.,0.,3.76406389114815,2.06072704459494,0.,3.6179005979755},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,0.,0.,0.,0.,0.,0.,5.18623489409549,2.83932866528416,1.15862500931079,0.,0.,0.,5.81833621896758,3.18538769471191,0.,6.23568135536041},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,0.,0.,0.,0.,0.,0.,6.45780747384476,3.53548156799627,1.44269926011819,0.,0.,0.,7.79428043908953,4.26716574385505,0.,8.98680735137037},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,0.,0.,0.,0.,0.,0.,7.27823584467216,3.98464475452196,1.62598614322311,0.,0.,0.,9.14177937313726,5.00488635016176,0.,10.9691712594614},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,0.,0.,0.,6.8546484667622e-008,5.19287666029906e-008,0.,7.65361974719019e-010},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,0.,0.,0.,0.,0.,0.,.000824565926330002,.000624666483447503,.000352723805983911,0.,0.,0.,5.01144631725624e-005,3.79652184082228e-005,0.,2.90963994598978e-006},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,0.,0.,0.,0.,0.,0.,.0113956405123916,.00863299641447793,.00487470263420339,0.,0.,0.,.00166206328995312,.00125912943722616,0.,.000231576690958128},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,0.,0.,0.,0.,0.,0.,.0657431866024243,.0498050718253105,.0281229023119214,0.,0.,0.,.0171974308138302,.0130282592204804,0.,.00429748986298374},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.414561685224313,.234086155283846,.121144008364367,0.,0.,0.,0.,0.,0.,.236321146004057,.179029831969503,.101090878717283,0.,0.,0.,.0946962283839188,.0717390303204678,0.,.0362494356624709},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,0.,0.,0.,0.,0.,0.,.626153181688297,.474354922529452,.267848969162523,0.,0.,0.,.347192252628355,.263022466294621,0.,.183906859961414},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,0.,0.,0.,0.,0.,0.,1.33269841309056,1.0096124574376,.570087171302651,0.,0.,0.,.95054294779362,.720102907121204,0.,.647664476709677},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,0.,0.,0.,0.,0.,0.,2.39387697960251,1.81352960012534,1.02402654819954,0.,0.,0.,2.07554270972356,1.57236907874115,0.,1.71909544841957},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,0.,0.,0.,0.,0.,0.,3.7410698100085,2.8341225946006,1.60031398303209,0.,0.,0.,3.76406389114815,2.85154222320761,0.,3.6179005979755},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,0.,0.,0.,0.,0.,0.,5.18623489409549,3.92893643816517,2.21851091848272,0.,0.,0.,5.81833621896758,4.40779749680161,0.,6.23568135536041},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,0.,0.,0.,0.,0.,0.,6.45780747384476,4.89224179250545,2.76245034842036,0.,0.,0.,7.79428043908953,5.90471373874725,0.,8.98680735137037},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,0.,0.,0.,0.,0.,0.,7.27823584467216,5.51377378765633,3.11340423610211,0.,0.,0.,9.14177937313726,6.92553862835673,0.,10.9691712594614},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,0.,0.,0.,6.8546484667622e-008,6.67972275267736e-008,0.,7.65361974719019e-010},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,0.,0.,0.,0.,0.,0.,.000824565926330002,.000803523594969987,.000583627344081487,0.,0.,0.,5.01144631725624e-005,4.88355780044979e-005,0.,2.90963994598978e-006},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,0.,0.,0.,0.,0.,0.,.0113956405123916,.0111048319353399,.00806582856989507,0.,0.,0.,.00166206328995312,.00161964862649386,0.,.000231576690958128},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.227250598233051,.16506007282648,.109880005542746,0.,0.,0.,0.,0.,0.,.0657431866024243,.0640654676075239,.0465329941039433,0.,0.,0.,.0171974308138302,.0167585647100291,0.,.00429748986298374},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.533260715077649,.387325943916553,.257841742921929,0.,0.,0.,0.,0.,0.,.236321146004057,.230290399761936,.167267987177829,0.,0.,0.,.0946962283839188,.092279648532812,0.,.0362494356624709},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,0.,0.,0.,0.,0.,0.,.626153181688297,.610174201341807,.443190904144467,0.,0.,0.,.347192252628355,.338332155278326,0.,.183906859961414},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,0.,0.,0.,0.,0.,0.,1.33269841309056,1.29868890491693,.943283260267015,0.,0.,0.,.95054294779362,.926285773305773,0.,.647664476709677},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,0.,0.,0.,0.,0.,0.,2.39387697960251,2.33278695510428,1.69438491095747,0.,0.,0.,2.07554270972356,2.02257634793674,0.,1.71909544841957},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,0.,0.,0.,0.,0.,0.,3.7410698100085,3.6456003901969,2.64792731244254,0.,0.,0.,3.76406389114815,3.66800767948205,0.,3.6179005979755},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,0.,0.,0.,0.,0.,0.,5.18623489409549,5.05388589728677,3.6708144253493,0.,0.,0.,5.81833621896758,5.66985645040998,0.,6.23568135536041},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,0.,0.,0.,0.,0.,0.,6.45780747384476,6.29300885631198,4.57083285180669,0.,0.,0.,7.79428043908953,7.59537599078757,0.,8.98680735137037},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,0.,0.,0.,0.,0.,0.,7.27823584467216,7.09250048322984,5.15153163620379,0.,0.,0.,9.14177937313726,8.90848771819592,0.,10.9691712594614},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,0.,0.,0.,6.8546484667622e-008,8.11985772602823e-008,0.,7.65361974719019e-010},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,0.,0.,0.,0.,0.,0.,.000824565926330002,.000976761687908062,.000862413599759767,0.,0.,0.,5.01144631725624e-005,5.93644317357238e-005,0.,2.90963994598978e-006},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,0.,0.,0.,0.,0.,0.,.0113956405123916,.0134990116693501,.0119187017581499,0.,0.,0.,.00166206328995312,.00196884165709061,0.,.000231576690958128},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.276245376362,.243905726875982,.197373399115395,0.,0.,0.,0.,0.,0.,.0657431866024243,.0778778553220726,.0687608066341364,0.,0.,0.,.0171974308138302,.0203716780136318,0.,.00429748986298374},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.648230667294543,.572343234018839,.463151607819505,0.,0.,0.,0.,0.,0.,.236321146004057,.279940553069749,.247168314523763,0.,0.,0.,.0946962283839188,.112174957679791,0.,.0362494356624709},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,0.,0.,0.,0.,0.,0.,.626153181688297,.741726548606004,.654893686699248,0.,0.,0.,.347192252628355,.411275896727804,0.,.183906859961414},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,0.,0.,0.,0.,0.,0.,1.33269841309056,1.57868365630447,1.39386942769156,0.,0.,0.,.95054294779362,1.12599114834103,0.,.647664476709677},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,0.,0.,0.,0.,0.,0.,2.39387697960251,2.83573119453035,2.50375621577018,0.,0.,0.,2.07554270972356,2.45863979589474,0.,1.71909544841957},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,0.,0.,0.,0.,0.,0.,3.7410698100085,4.43158460169426,3.91278535624427,0.,0.,0.,3.76406389114815,4.45882285809472,0.,3.6179005979755},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,0.,0.,0.,0.,0.,0.,5.18623489409549,6.14349367016778,5.42428368841739,0.,0.,0.,5.81833621896758,6.89226625249968,0.,6.23568135536041},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,0.,0.,0.,0.,0.,0.,6.45780747384476,7.64976908082116,6.7542215997961,0.,0.,0.,7.79428043908953,9.23292398567978,0.,8.98680735137037},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,0.,0.,0.,0.,0.,0.,7.27823584467216,8.62162951636421,7.61230772976691,0.,0.,0.,9.14177937313726,10.8291399963909,0.,10.9691712594614},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,0.,0.,0.,6.8546484667622e-008,9.42279510613512e-008,0.,7.65361974719019e-010},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,0.,0.,0.,0.,0.,0.,.000824565926330002,.00113349587680305,.00116139045508752,0.,0.,0.,5.01144631725624e-005,6.88902312961445e-005,0.,2.90963994598978e-006},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,0.,0.,0.,0.,0.,0.,.0113956405123916,.0156651046591466,.0160506124472133,0.,0.,0.,.00166206328995312,.00228476805347462,0.,.000231576690958128},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.320572560296515,.3284616374485,.308448458722372,0.,0.,0.,0.,0.,0.,.0657431866024243,.0903743758530202,.0925984290266871,0.,0.,0.,.0171974308138302,.023640580213036,0.,.00429748986298374},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.752247756737171,.770760072903028,.723797635481732,0.,0.,0.,0.,0.,0.,.236321146004057,.324860676440036,.332855281233889,0.,0.,0.,.0946962283839188,.130174896891091,0.,.0362494356624709},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,0.,0.,0.,0.,0.,0.,.626153181688297,.860746275133792,.881928667453122,0.,0.,0.,.347192252628355,.477270493857987,0.,.183906859961414},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,0.,0.,0.,0.,0.,0.,1.33269841309056,1.83200409818481,1.87708849838432,0.,0.,0.,.95054294779362,1.306671156088,0.,.647664476709677},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,0.,0.,0.,0.,0.,0.,2.39387697960251,3.29076135613591,3.37174480048962,0.,0.,0.,2.07554270972356,2.8531607102232,0.,1.71909544841957},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,0.,0.,0.,0.,0.,0.,3.7410698100085,5.14269031628637,5.26924849841671,0.,0.,0.,3.76406389114815,5.17429930720343,0.,3.6179005979755},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,0.,0.,0.,0.,0.,0.,5.18623489409549,7.12929758661485,7.30474484999971,0.,0.,0.,5.81833621896758,7.99822052374818,0.,6.23568135536041},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,0.,0.,0.,0.,0.,0.,6.45780747384476,8.87727458903197,9.0957384017764,0.,0.,0.,7.79428043908953,10.7144674060855,0.,8.98680735137037},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,0.,0.,0.,0.,0.,0.,7.27823584467216,10.0050827434195,10.2513011633895,0.,0.,0.,9.14177937313726,12.5668171542652,0.,10.9691712594614},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,0.,0.,0.,6.8546484667622e-008,1.05066727253324e-007,0.,7.65361974719019e-010},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,0.,0.,0.,0.,0.,0.,.000824565926330002,.00126387871952445,.00144394006224153,0.,0.,0.,5.01144631725624e-005,7.68144808465349e-005,0.,2.90963994598978e-006},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,0.,0.,0.,0.,0.,0.,.0113956405123916,.0174670175895654,.0199554957891379,0.,0.,0.,.00166206328995312,.00254757849669498,0.,.000231576690958128},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.357447120288584,.408371633453457,.427601258703136,0.,0.,0.,0.,0.,0.,.0657431866024243,.100769886127061,.115126296058794,0.,0.,0.,.0171974308138302,.0263598896651818,0.,.00429748986298374},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,0.,0.,0.,0.,0.,0.,.236321146004057,.362228486371654,.41383418762992,0.,0.,0.,.0946962283839188,.145148549160118,0.,.0362494356624709},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,0.,0.,0.,0.,0.,0.,.626153181688297,.959755498290674,1.09648923787555,0.,0.,0.,.347192252628355,.532169576430532,0.,.183906859961414},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,0.,0.,0.,0.,0.,0.,1.33269841309056,2.04273437703882,2.33375715403626,0.,0.,0.,.95054294779362,1.45697386412547,0.,.647664476709677},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,0.,0.,0.,0.,0.,0.,2.39387697960251,3.66928837957851,4.19204185444651,0.,0.,0.,2.07554270972356,3.1813517621299,0.,1.71909544841957},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,0.,0.,0.,0.,0.,0.,3.7410698100085,5.7342395194157,6.55118093268348,0.,0.,0.,3.76406389114815,5.76948440365682,0.,3.6179005979755},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,0.,0.,0.,0.,0.,0.,5.18623489409549,7.94936063666436,9.0818842940915,0.,0.,0.,5.81833621896758,8.91823333538729,0.,6.23568135536041},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,0.,0.,0.,0.,0.,0.,6.45780747384476,9.89840251743778,11.3086008382978,0.,0.,0.,7.79428043908953,11.9469224570834,0.,8.98680735137037},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,0.,0.,0.,0.,0.,0.,7.27823584467216,11.1559392718342,12.7452954129935,0.,0.,0.,9.14177937313726,14.0123427870133,0.,10.9691712594614},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,0.,0.,0.,6.8546484667622e-008,1.13034107101112e-007,0.,7.65361974719019e-010},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,0.,0.,0.,0.,0.,0.,.000824565926330002,.00135972068684592,.00167123596481511,0.,0.,0.,5.01144631725624e-005,8.26394471580808e-005,0.,2.90963994598978e-006},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,0.,0.,0.,0.,0.,0.,.0113956405123916,.0187915697820833,.0230967635919399,0.,0.,0.,.00166206328995312,.00274076549419503,0.,.000231576690958128},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.384552913504838,.472654910466463,.532441484172332,0.,0.,0.,0.,0.,0.,.0657431866024243,.108411429563137,.133248748684711,0.,0.,0.,.0171974308138302,.0283588027292761,0.,.00429748986298374},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,0.,0.,0.,0.,0.,0.,.236321146004057,.389696858310697,.47897734533629,0.,0.,0.,.0946962283839188,.156155398385808,0.,.0362494356624709},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,0.,0.,0.,0.,0.,0.,.626153181688297,1.03253530989981,1.26909163149447,0.,0.,0.,.347192252628355,.572524856067592,0.,.183906859961414},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,0.,0.,0.,0.,0.,0.,1.33269841309056,2.19763822848137,2.70112242949704,0.,0.,0.,.95054294779362,1.56745854854701,0.,.647664476709677},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,0.,0.,0.,0.,0.,0.,2.39387697960251,3.94753645159365,4.85192654207924,0.,0.,0.,2.07554270972356,3.42259881132375,0.,1.71909544841957},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,0.,0.,0.,0.,0.,0.,3.7410698100085,6.16907617592665,7.58242635758397,0.,0.,0.,3.76406389114815,6.20699373673991,0.,3.6179005979755},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,0.,0.,0.,0.,0.,0.,5.18623489409549,8.55217350992208,10.5114970194908,0.,0.,0.,5.81833621896758,9.5945173923078,0.,6.23568135536041},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,0.,0.,0.,0.,0.,0.,6.45780747384476,10.6490143886212,13.0887291840651,0.,0.,0.,7.79428043908953,12.8528769082785,0.,8.98680735137037},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,0.,0.,0.,0.,0.,0.,7.27823584467216,12.0019121888667,14.7515791225583,0.,0.,0.,9.14177937313726,15.0749214021478,0.,10.9691712594614},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,0.,0.,0.,6.8546484667622e-008,1.17631378114674e-007,0.,7.65361974719019e-010},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,0.,0.,0.,0.,0.,0.,.000824565926330002,.00141502261880692,.00180994395496465,0.,0.,0.,5.01144631725624e-005,8.60005205963556e-005,0.,2.90963994598978e-006},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,0.,0.,0.,0.,0.,0.,.0113956405123916,.019555851831759,.0250137314673599,0.,0.,0.,.00166206328995312,.00285223664290029,0.,.000231576690958128},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.400193272045958,.511883968508167,.600085200896572,0.,0.,0.,0.,0.,0.,.0657431866024243,.112820689170272,.144308028468726,0.,0.,0.,.0171974308138302,.0295121988599679,0.,.00429748986298374},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,0.,0.,0.,0.,0.,0.,.236321146004057,.405546428999883,.518731148089159,0.,0.,0.,.0946962283839188,.162506478648407,0.,.0362494356624709},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,0.,0.,0.,0.,0.,0.,.626153181688297,1.07453010927869,1.3744227476422,0.,0.,0.,.347192252628355,.595810322665643,0.,.183906859961414},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,0.,0.,0.,0.,0.,0.,1.33269841309056,2.28701955580991,2.92530816462441,0.,0.,0.,.95054294779362,1.63120949890175,0.,.647664476709677},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,0.,0.,0.,0.,0.,0.,2.39387697960251,4.10808883148424,5.25462310508641,0.,0.,0.,2.07554270972356,3.5618011698851,0.,1.71909544841957},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,0.,0.,0.,0.,0.,0.,3.7410698100085,6.41998199374919,8.21174689798638,0.,0.,0.,3.76406389114815,6.45944172008843,0.,3.6179005979755},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,0.,0.,0.,0.,0.,0.,5.18623489409549,8.90000356218188,11.3839223715851,0.,0.,0.,5.81833621896758,9.98474117367778,0.,6.23568135536041},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,0.,0.,0.,0.,0.,0.,6.45780747384476,11.0821261849396,14.1750577199246,0.,0.,0.,7.79428043908953,13.3756231834223,0.,8.98680735137037},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,0.,0.,0.,0.,0.,0.,7.27823584467216,12.4900484198527,15.9759196314393,0.,0.,0.,9.14177937313726,15.6880416449771,0.,10.9691712594614}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,0.,0.,0.,6.8546484667622e-008,1.09461601509022e-009,0.,7.65361974719019e-010},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,0.,0.,0.,0.,0.,0.,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,0.,0.,0.,5.01144631725624e-005,8.00275816365055e-007,0.,2.90963994598978e-006},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,0.,0.,0.,0.,0.,0.,.0113956405123916,.000181976518058871,2.16598631798674e-006,0.,0.,0.,.00166206328995312,2.65414208197256e-005,0.,.000231576690958128},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,0.,0.,0.,0.,0.,0.,.0657431866024243,.00104985026256264,1.24959051250218e-005,0.,0.,0.,.0171974308138302,.000274625070541602,0.,.00429748986298374},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,0.,0.,0.,0.,0.,0.,.236321146004057,.00377380273155657,4.4917911225712e-005,0.,0.,0.,.0946962283839188,.00151220020487266,0.,.0362494356624709},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,0.,0.,0.,0.,0.,0.,.626153181688297,.00999901459257298,.000119013865260662,0.,0.,0.,.347192252628355,.00554429890730429,0.,.183906859961414},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,0.,0.,0.,0.,0.,0.,1.33269841309056,.0212818065446243,.000253307966815722,0.,0.,0.,.95054294779362,.0151791815252261,0.,.647664476709677},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,0.,0.,0.,0.,0.,0.,2.39387697960251,.0382277237453785,.000455007753107355,0.,0.,0.,2.07554270972356,.0331442567927967,0.,1.71909544841957},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,0.,0.,0.,0.,0.,0.,3.7410698100085,.0597409910483066,.000711070695350591,0.,0.,0.,3.76406389114815,.0601081826012263,0.,3.6179005979755},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,0.,0.,0.,0.,0.,0.,5.18623489409549,.0828187732700635,.000985755369367884,0.,0.,0.,5.81833621896758,.0929127735338078,0.,6.23568135536041},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,0.,0.,0.,0.,0.,0.,6.45780747384476,.103124463877826,.00122744505825102,0.,0.,0.,7.79428043908953,.124466546112493,0.,8.98680735137037},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,0.,0.,0.,0.,0.,0.,7.27823584467216,.116225851033514,.00138338509726576,0.,0.,0.,9.14177937313726,.145984701575574,0.,10.9691712594614},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,0.,0.,0.,6.8546484667622e-008,5.69188702865245e-009,0.,7.65361974719019e-010},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,0.,0.,0.,0.,0.,0.,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,0.,0.,0.,5.01144631725624e-005,4.16134925463982e-006,0.,2.90963994598978e-006},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,0.,0.,0.,0.,0.,0.,.0113956405123916,.000946258567734596,5.85658777413804e-005,0.,0.,0.,.00166206328995312,.000138012569524986,0.,.000231576690958128},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,0.,0.,0.,0.,0.,0.,.0657431866024243,.0054591098696972,.000337875473054763,0.,0.,0.,.0171974308138302,.00142802120123345,0.,.00429748986298374},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,0.,0.,0.,0.,0.,0.,.236321146004057,.0196233734207421,.0012145307084342,0.,0.,0.,.0946962283839188,.00786328046747136,0.,.0362494356624709},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,0.,0.,0.,0.,0.,0.,.626153181688297,.0519938139714517,.00321800346775215,0.,0.,0.,.347192252628355,.028829765505355,0.,.183906859961414},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,0.,0.,0.,0.,0.,0.,1.33269841309056,.110663133873164,.00684916764813009,0.,0.,0.,.95054294779362,.0789301318799674,0.,.647664476709677},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,0.,0.,0.,0.,0.,0.,2.39387697960251,.198780103635972,.0123029070952925,0.,0.,0.,2.07554270972356,.172346615354148,0.,1.71909544841957},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,0.,0.,0.,0.,0.,0.,3.7410698100085,.310646808870845,.0192265662361566,0.,0.,0.,3.76406389114815,.31255616594975,0.,3.6179005979755},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,0.,0.,0.,0.,0.,0.,5.18623489409549,.430648825529863,.0266537364367886,0.,0.,0.,5.81833621896758,.483136554903789,0.,6.23568135536041},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,0.,0.,0.,0.,0.,0.,6.45780747384476,.53623626019628,.0331887586818224,0.,0.,0.,7.79428043908953,.647212821256298,0.,8.98680735137037},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,0.,0.,0.,0.,0.,0.,7.27823584467216,.60436208201944,.037405205103521,0.,0.,0.,9.14177937313726,.759104944404841,0.,10.9691712594614},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,0.,0.,0.,6.8546484667622e-008,1.36592668764407e-008,0.,7.65361974719019e-010},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,0.,0.,0.,0.,0.,0.,.000824565926330002,.000164311358893045,2.44047036237916e-005,0.,0.,0.,5.01144631725624e-005,9.98631556618579e-006,0.,2.90963994598978e-006},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,0.,0.,0.,0.,0.,0.,.0113956405123916,.00227081076025246,.000337277130218073,0.,0.,0.,.00166206328995312,.000331199567025036,0.,.000231576690958128},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,0.,0.,0.,0.,0.,0.,.0657431866024243,.0131006533057736,.00194580315907169,0.,0.,0.,.0171974308138302,.00342693426532766,0.,.00429748986298374},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,0.,0.,0.,0.,0.,0.,.236321146004057,.0470917453597858,.00699440438187064,0.,0.,0.,.0946962283839188,.0188701296931614,0.,.0362494356624709},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,0.,0.,0.,0.,0.,0.,.626153181688297,.124773625580585,.0185322753878643,0.,0.,0.,.347192252628355,.0691850451424154,0.,.183906859961414},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,0.,0.,0.,0.,0.,0.,1.33269841309056,.26556698531571,.0394439167964793,0.,0.,0.,.95054294779362,.189414816301507,0.,.647664476709677},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,0.,0.,0.,0.,0.,0.,2.39387697960251,.477028175651108,.0708516521644813,0.,0.,0.,2.07554270972356,.413593664547998,0.,1.71909544841957},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,0.,0.,0.,0.,0.,0.,3.7410698100085,.745483465381799,.110724560685561,0.,0.,0.,3.76406389114815,.750065499032834,0.,3.6179005979755},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,0.,0.,0.,0.,0.,0.,5.18623489409549,1.03346169878758,.153497157076452,0.,0.,0.,5.81833621896758,1.1594206118243,0.,6.23568135536041},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,0.,0.,0.,0.,0.,0.,6.45780747384476,1.28684813137965,.191131930663375,0.,0.,0.,7.79428043908953,1.55316727245144,0.,8.98680735137037},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,0.,0.,0.,0.,0.,0.,7.27823584467216,1.45033499905202,.215414174926983,0.,0.,0.,9.14177937313726,1.82168355953934,0.,10.9691712594614},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,0.,0.,0.,6.8546484667622e-008,2.44980430684131e-008,0.,7.65361974719019e-010},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,0.,0.,0.,0.,0.,0.,.000824565926330002,.000294694201614443,7.85021488046878e-005,0.,0.,0.,5.01144631725624e-005,1.79105651165762e-005,0.,2.90963994598978e-006},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,0.,0.,0.,0.,0.,0.,.0113956405123916,.00407272369067121,.00108491296894855,0.,0.,0.,.00166206328995312,.000594010010245393,0.,.000231576690958128},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,0.,0.,0.,0.,0.,0.,.0657431866024243,.0234961635798143,.00625902823868614,0.,0.,0.,.0171974308138302,.0061462437174735,0.,.00429748986298374},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,0.,0.,0.,0.,0.,0.,.236321146004057,.0844595552914031,.0224987683542483,0.,0.,0.,.0946962283839188,.0338437819621889,0.,.0362494356624709},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,0.,0.,0.,0.,0.,0.,.626153181688297,.223782848737467,.0596124199094679,0.,0.,0.,.347192252628355,.12408412771496,0.,.183906859961414},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,0.,0.,0.,0.,0.,0.,1.33269841309056,.476297264169728,.126878501518793,0.,0.,0.,.95054294779362,.339717524338981,0.,.647664476709677},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,0.,0.,0.,0.,0.,0.,2.39387697960251,.855555199093713,.227907170151078,0.,0.,0.,2.07554270972356,.741784716454702,0.,1.71909544841957},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,0.,0.,0.,0.,0.,0.,3.7410698100085,1.33703266851113,.356165601240813,0.,0.,0.,3.76406389114815,1.34525059548623,0.,3.6179005979755},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,0.,0.,0.,0.,0.,0.,5.18623489409549,1.85352474883709,.493751403486214,0.,0.,0.,5.81833621896758,2.07943342346341,0.,6.23568135536041},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,0.,0.,0.,0.,0.,0.,6.45780747384476,2.30797605978546,.614810468242533,0.,0.,0.,7.79428043908953,2.78562232344933,0.,8.98680735137037},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,0.,0.,0.,0.,0.,0.,7.27823584467216,2.60119152746664,.692918704338266,0.,0.,0.,9.14177937313726,3.26720919228747,0.,10.9691712594614},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,0.,0.,0.,6.8546484667622e-008,3.7527416869482e-008,0.,7.65361974719019e-010},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,0.,0.,0.,0.,0.,0.,.000824565926330002,.000451428390509428,.000184211229066994,0.,0.,0.,5.01144631725624e-005,2.74363646769969e-005,0.,2.90963994598978e-006},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,0.,0.,0.,0.,0.,0.,.0113956405123916,.00623881668046779,.00254583033055523,0.,0.,0.,.00166206328995312,.000909936406629402,0.,.000231576690958128},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,0.,0.,0.,0.,0.,0.,.0657431866024243,.0359926841107618,.0146872831147845,0.,0.,0.,.0171974308138302,.00941514591687767,0.,.00429748986298374},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,0.,0.,0.,0.,0.,0.,.236321146004057,.129379678661691,.0527950614618356,0.,0.,0.,.0946962283839188,.0518437211734884,0.,.0362494356624709},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,0.,0.,0.,0.,0.,0.,.626153181688297,.342802575265255,.139885051637275,0.,0.,0.,.347192252628355,.190078724845143,0.,.183906859961414},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,0.,0.,0.,0.,0.,0.,1.33269841309056,.729617706050062,.297729999278181,0.,0.,0.,.95054294779362,.520397532085943,0.,.647664476709677},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,0.,0.,0.,0.,0.,0.,2.39387697960251,1.31058536069927,.534801410737989,0.,0.,0.,2.07554270972356,1.13630563078316,0.,1.71909544841957},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,0.,0.,0.,0.,0.,0.,3.7410698100085,2.04813838310324,.835769519114577,0.,0.,0.,3.76406389114815,2.06072704459494,0.,3.6179005979755},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,0.,0.,0.,0.,0.,0.,5.18623489409549,2.83932866528416,1.15862500931079,0.,0.,0.,5.81833621896758,3.18538769471191,0.,6.23568135536041},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,0.,0.,0.,0.,0.,0.,6.45780747384476,3.53548156799627,1.44269926011819,0.,0.,0.,7.79428043908953,4.26716574385505,0.,8.98680735137037},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,0.,0.,0.,0.,0.,0.,7.27823584467216,3.98464475452196,1.62598614322311,0.,0.,0.,9.14177937313726,5.00488635016176,0.,10.9691712594614},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,0.,0.,0.,6.8546484667622e-008,5.19287666029906e-008,0.,7.65361974719019e-010},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,0.,0.,0.,0.,0.,0.,.000824565926330002,.000624666483447503,.000352723805983911,0.,0.,0.,5.01144631725624e-005,3.79652184082228e-005,0.,2.90963994598978e-006},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,0.,0.,0.,0.,0.,0.,.0113956405123916,.00863299641447793,.00487470263420339,0.,0.,0.,.00166206328995312,.00125912943722616,0.,.000231576690958128},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,0.,0.,0.,0.,0.,0.,.0657431866024243,.0498050718253105,.0281229023119214,0.,0.,0.,.0171974308138302,.0130282592204804,0.,.00429748986298374},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.414561685224313,.234086155283846,.121144008364367,0.,0.,0.,0.,0.,0.,.236321146004057,.179029831969503,.101090878717283,0.,0.,0.,.0946962283839188,.0717390303204678,0.,.0362494356624709},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,0.,0.,0.,0.,0.,0.,.626153181688297,.474354922529452,.267848969162523,0.,0.,0.,.347192252628355,.263022466294621,0.,.183906859961414},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,0.,0.,0.,0.,0.,0.,1.33269841309056,1.0096124574376,.570087171302651,0.,0.,0.,.95054294779362,.720102907121204,0.,.647664476709677},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,0.,0.,0.,0.,0.,0.,2.39387697960251,1.81352960012534,1.02402654819954,0.,0.,0.,2.07554270972356,1.57236907874115,0.,1.71909544841957},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,0.,0.,0.,0.,0.,0.,3.7410698100085,2.8341225946006,1.60031398303209,0.,0.,0.,3.76406389114815,2.85154222320761,0.,3.6179005979755},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,0.,0.,0.,0.,0.,0.,5.18623489409549,3.92893643816517,2.21851091848272,0.,0.,0.,5.81833621896758,4.40779749680161,0.,6.23568135536041},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,0.,0.,0.,0.,0.,0.,6.45780747384476,4.89224179250545,2.76245034842036,0.,0.,0.,7.79428043908953,5.90471373874725,0.,8.98680735137037},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,0.,0.,0.,0.,0.,0.,7.27823584467216,5.51377378765633,3.11340423610211,0.,0.,0.,9.14177937313726,6.92553862835673,0.,10.9691712594614},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,0.,0.,0.,6.8546484667622e-008,6.67972275267736e-008,0.,7.65361974719019e-010},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,0.,0.,0.,0.,0.,0.,.000824565926330002,.000803523594969987,.000583627344081487,0.,0.,0.,5.01144631725624e-005,4.88355780044979e-005,0.,2.90963994598978e-006},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,0.,0.,0.,0.,0.,0.,.0113956405123916,.0111048319353399,.00806582856989507,0.,0.,0.,.00166206328995312,.00161964862649386,0.,.000231576690958128},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.227250598233051,.16506007282648,.109880005542746,0.,0.,0.,0.,0.,0.,.0657431866024243,.0640654676075239,.0465329941039433,0.,0.,0.,.0171974308138302,.0167585647100291,0.,.00429748986298374},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.533260715077649,.387325943916553,.257841742921929,0.,0.,0.,0.,0.,0.,.236321146004057,.230290399761936,.167267987177829,0.,0.,0.,.0946962283839188,.092279648532812,0.,.0362494356624709},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,0.,0.,0.,0.,0.,0.,.626153181688297,.610174201341807,.443190904144467,0.,0.,0.,.347192252628355,.338332155278326,0.,.183906859961414},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,0.,0.,0.,0.,0.,0.,1.33269841309056,1.29868890491693,.943283260267015,0.,0.,0.,.95054294779362,.926285773305773,0.,.647664476709677},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,0.,0.,0.,0.,0.,0.,2.39387697960251,2.33278695510428,1.69438491095747,0.,0.,0.,2.07554270972356,2.02257634793674,0.,1.71909544841957},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,0.,0.,0.,0.,0.,0.,3.7410698100085,3.6456003901969,2.64792731244254,0.,0.,0.,3.76406389114815,3.66800767948205,0.,3.6179005979755},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,0.,0.,0.,0.,0.,0.,5.18623489409549,5.05388589728677,3.6708144253493,0.,0.,0.,5.81833621896758,5.66985645040998,0.,6.23568135536041},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,0.,0.,0.,0.,0.,0.,6.45780747384476,6.29300885631198,4.57083285180669,0.,0.,0.,7.79428043908953,7.59537599078757,0.,8.98680735137037},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,0.,0.,0.,0.,0.,0.,7.27823584467216,7.09250048322984,5.15153163620379,0.,0.,0.,9.14177937313726,8.90848771819592,0.,10.9691712594614},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,0.,0.,0.,6.8546484667622e-008,8.11985772602823e-008,0.,7.65361974719019e-010},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,0.,0.,0.,0.,0.,0.,.000824565926330002,.000976761687908062,.000862413599759767,0.,0.,0.,5.01144631725624e-005,5.93644317357238e-005,0.,2.90963994598978e-006},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,0.,0.,0.,0.,0.,0.,.0113956405123916,.0134990116693501,.0119187017581499,0.,0.,0.,.00166206328995312,.00196884165709061,0.,.000231576690958128},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.276245376362,.243905726875982,.197373399115395,0.,0.,0.,0.,0.,0.,.0657431866024243,.0778778553220726,.0687608066341364,0.,0.,0.,.0171974308138302,.0203716780136318,0.,.00429748986298374},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.648230667294543,.572343234018839,.463151607819505,0.,0.,0.,0.,0.,0.,.236321146004057,.279940553069749,.247168314523763,0.,0.,0.,.0946962283839188,.112174957679791,0.,.0362494356624709},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,0.,0.,0.,0.,0.,0.,.626153181688297,.741726548606004,.654893686699248,0.,0.,0.,.347192252628355,.411275896727804,0.,.183906859961414},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,0.,0.,0.,0.,0.,0.,1.33269841309056,1.57868365630447,1.39386942769156,0.,0.,0.,.95054294779362,1.12599114834103,0.,.647664476709677},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,0.,0.,0.,0.,0.,0.,2.39387697960251,2.83573119453035,2.50375621577018,0.,0.,0.,2.07554270972356,2.45863979589474,0.,1.71909544841957},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,0.,0.,0.,0.,0.,0.,3.7410698100085,4.43158460169426,3.91278535624427,0.,0.,0.,3.76406389114815,4.45882285809472,0.,3.6179005979755},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,0.,0.,0.,0.,0.,0.,5.18623489409549,6.14349367016778,5.42428368841739,0.,0.,0.,5.81833621896758,6.89226625249968,0.,6.23568135536041},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,0.,0.,0.,0.,0.,0.,6.45780747384476,7.64976908082116,6.7542215997961,0.,0.,0.,7.79428043908953,9.23292398567978,0.,8.98680735137037},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,0.,0.,0.,0.,0.,0.,7.27823584467216,8.62162951636421,7.61230772976691,0.,0.,0.,9.14177937313726,10.8291399963909,0.,10.9691712594614},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,0.,0.,0.,6.8546484667622e-008,9.42279510613512e-008,0.,7.65361974719019e-010},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,0.,0.,0.,0.,0.,0.,.000824565926330002,.00113349587680305,.00116139045508752,0.,0.,0.,5.01144631725624e-005,6.88902312961445e-005,0.,2.90963994598978e-006},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,0.,0.,0.,0.,0.,0.,.0113956405123916,.0156651046591466,.0160506124472133,0.,0.,0.,.00166206328995312,.00228476805347462,0.,.000231576690958128},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.320572560296515,.3284616374485,.308448458722372,0.,0.,0.,0.,0.,0.,.0657431866024243,.0903743758530202,.0925984290266871,0.,0.,0.,.0171974308138302,.023640580213036,0.,.00429748986298374},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.752247756737171,.770760072903028,.723797635481732,0.,0.,0.,0.,0.,0.,.236321146004057,.324860676440036,.332855281233889,0.,0.,0.,.0946962283839188,.130174896891091,0.,.0362494356624709},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,0.,0.,0.,0.,0.,0.,.626153181688297,.860746275133792,.881928667453122,0.,0.,0.,.347192252628355,.477270493857987,0.,.183906859961414},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,0.,0.,0.,0.,0.,0.,1.33269841309056,1.83200409818481,1.87708849838432,0.,0.,0.,.95054294779362,1.306671156088,0.,.647664476709677},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,0.,0.,0.,0.,0.,0.,2.39387697960251,3.29076135613591,3.37174480048962,0.,0.,0.,2.07554270972356,2.8531607102232,0.,1.71909544841957},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,0.,0.,0.,0.,0.,0.,3.7410698100085,5.14269031628637,5.26924849841671,0.,0.,0.,3.76406389114815,5.17429930720343,0.,3.6179005979755},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,0.,0.,0.,0.,0.,0.,5.18623489409549,7.12929758661485,7.30474484999971,0.,0.,0.,5.81833621896758,7.99822052374818,0.,6.23568135536041},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,0.,0.,0.,0.,0.,0.,6.45780747384476,8.87727458903197,9.0957384017764,0.,0.,0.,7.79428043908953,10.7144674060855,0.,8.98680735137037},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,0.,0.,0.,0.,0.,0.,7.27823584467216,10.0050827434195,10.2513011633895,0.,0.,0.,9.14177937313726,12.5668171542652,0.,10.9691712594614},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,0.,0.,0.,6.8546484667622e-008,1.05066727253324e-007,0.,7.65361974719019e-010},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,0.,0.,0.,0.,0.,0.,.000824565926330002,.00126387871952445,.00144394006224153,0.,0.,0.,5.01144631725624e-005,7.68144808465349e-005,0.,2.90963994598978e-006},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,0.,0.,0.,0.,0.,0.,.0113956405123916,.0174670175895654,.0199554957891379,0.,0.,0.,.00166206328995312,.00254757849669498,0.,.000231576690958128},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.357447120288584,.408371633453457,.427601258703136,0.,0.,0.,0.,0.,0.,.0657431866024243,.100769886127061,.115126296058794,0.,0.,0.,.0171974308138302,.0263598896651818,0.,.00429748986298374},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,0.,0.,0.,0.,0.,0.,.236321146004057,.362228486371654,.41383418762992,0.,0.,0.,.0946962283839188,.145148549160118,0.,.0362494356624709},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,0.,0.,0.,0.,0.,0.,.626153181688297,.959755498290674,1.09648923787555,0.,0.,0.,.347192252628355,.532169576430532,0.,.183906859961414},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,0.,0.,0.,0.,0.,0.,1.33269841309056,2.04273437703882,2.33375715403626,0.,0.,0.,.95054294779362,1.45697386412547,0.,.647664476709677},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,0.,0.,0.,0.,0.,0.,2.39387697960251,3.66928837957851,4.19204185444651,0.,0.,0.,2.07554270972356,3.1813517621299,0.,1.71909544841957},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,0.,0.,0.,0.,0.,0.,3.7410698100085,5.7342395194157,6.55118093268348,0.,0.,0.,3.76406389114815,5.76948440365682,0.,3.6179005979755},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,0.,0.,0.,0.,0.,0.,5.18623489409549,7.94936063666436,9.0818842940915,0.,0.,0.,5.81833621896758,8.91823333538729,0.,6.23568135536041},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,0.,0.,0.,0.,0.,0.,6.45780747384476,9.89840251743778,11.3086008382978,0.,0.,0.,7.79428043908953,11.9469224570834,0.,8.98680735137037},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,0.,0.,0.,0.,0.,0.,7.27823584467216,11.1559392718342,12.7452954129935,0.,0.,0.,9.14177937313726,14.0123427870133,0.,10.9691712594614},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,0.,0.,0.,6.8546484667622e-008,1.13034107101112e-007,0.,7.65361974719019e-010},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,0.,0.,0.,0.,0.,0.,.000824565926330002,.00135972068684592,.00167123596481511,0.,0.,0.,5.01144631725624e-005,8.26394471580808e-005,0.,2.90963994598978e-006},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,0.,0.,0.,0.,0.,0.,.0113956405123916,.0187915697820833,.0230967635919399,0.,0.,0.,.00166206328995312,.00274076549419503,0.,.000231576690958128},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.384552913504838,.472654910466463,.532441484172332,0.,0.,0.,0.,0.,0.,.0657431866024243,.108411429563137,.133248748684711,0.,0.,0.,.0171974308138302,.0283588027292761,0.,.00429748986298374},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,0.,0.,0.,0.,0.,0.,.236321146004057,.389696858310697,.47897734533629,0.,0.,0.,.0946962283839188,.156155398385808,0.,.0362494356624709},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,0.,0.,0.,0.,0.,0.,.626153181688297,1.03253530989981,1.26909163149447,0.,0.,0.,.347192252628355,.572524856067592,0.,.183906859961414},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,0.,0.,0.,0.,0.,0.,1.33269841309056,2.19763822848137,2.70112242949704,0.,0.,0.,.95054294779362,1.56745854854701,0.,.647664476709677},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,0.,0.,0.,0.,0.,0.,2.39387697960251,3.94753645159365,4.85192654207924,0.,0.,0.,2.07554270972356,3.42259881132375,0.,1.71909544841957},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,0.,0.,0.,0.,0.,0.,3.7410698100085,6.16907617592665,7.58242635758397,0.,0.,0.,3.76406389114815,6.20699373673991,0.,3.6179005979755},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,0.,0.,0.,0.,0.,0.,5.18623489409549,8.55217350992208,10.5114970194908,0.,0.,0.,5.81833621896758,9.5945173923078,0.,6.23568135536041},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,0.,0.,0.,0.,0.,0.,6.45780747384476,10.6490143886212,13.0887291840651,0.,0.,0.,7.79428043908953,12.8528769082785,0.,8.98680735137037},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,0.,0.,0.,0.,0.,0.,7.27823584467216,12.0019121888667,14.7515791225583,0.,0.,0.,9.14177937313726,15.0749214021478,0.,10.9691712594614},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,0.,0.,0.,6.8546484667622e-008,1.17631378114674e-007,0.,7.65361974719019e-010},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,0.,0.,0.,0.,0.,0.,.000824565926330002,.00141502261880692,.00180994395496465,0.,0.,0.,5.01144631725624e-005,8.60005205963556e-005,0.,2.90963994598978e-006},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,0.,0.,0.,0.,0.,0.,.0113956405123916,.019555851831759,.0250137314673599,0.,0.,0.,.00166206328995312,.00285223664290029,0.,.000231576690958128},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.400193272045958,.511883968508167,.600085200896572,0.,0.,0.,0.,0.,0.,.0657431866024243,.112820689170272,.144308028468726,0.,0.,0.,.0171974308138302,.0295121988599679,0.,.00429748986298374},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,0.,0.,0.,0.,0.,0.,.236321146004057,.405546428999883,.518731148089159,0.,0.,0.,.0946962283839188,.162506478648407,0.,.0362494356624709},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,0.,0.,0.,0.,0.,0.,.626153181688297,1.07453010927869,1.3744227476422,0.,0.,0.,.347192252628355,.595810322665643,0.,.183906859961414},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,0.,0.,0.,0.,0.,0.,1.33269841309056,2.28701955580991,2.92530816462441,0.,0.,0.,.95054294779362,1.63120949890175,0.,.647664476709677},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,0.,0.,0.,0.,0.,0.,2.39387697960251,4.10808883148424,5.25462310508641,0.,0.,0.,2.07554270972356,3.5618011698851,0.,1.71909544841957},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,0.,0.,0.,0.,0.,0.,3.7410698100085,6.41998199374919,8.21174689798638,0.,0.,0.,3.76406389114815,6.45944172008843,0.,3.6179005979755},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,0.,0.,0.,0.,0.,0.,5.18623489409549,8.90000356218188,11.3839223715851,0.,0.,0.,5.81833621896758,9.98474117367778,0.,6.23568135536041},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,0.,0.,0.,0.,0.,0.,6.45780747384476,11.0821261849396,14.1750577199246,0.,0.,0.,7.79428043908953,13.3756231834223,0.,8.98680735137037},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,0.,0.,0.,0.,0.,0.,7.27823584467216,12.4900484198527,15.9759196314393,0.,0.,0.,9.14177937313726,15.6880416449771,0.,10.9691712594614}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,7.27823584467216,.116225851033514,.00138338509726576,9.14177937313726,.145984701575574,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,6.45780747384476,.103124463877826,.00122744505825102,7.79428043908953,.124466546112493,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,5.18623489409549,.0828187732700635,.000985755369367884,5.81833621896758,.0929127735338078,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.7410698100085,.0597409910483066,.000711070695350591,3.76406389114815,.0601081826012263,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,2.39387697960251,.0382277237453785,.000455007753107355,2.07554270972356,.0331442567927967,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,1.33269841309056,.0212818065446243,.000253307966815722,.95054294779362,.0151791815252261,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,.626153181688297,.00999901459257298,.000119013865260662,.347192252628355,.00554429890730429,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,.236321146004057,.00377380273155657,4.4917911225712e-005,.0946962283839188,.00151220020487266,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.0657431866024243,.00104985026256264,1.24959051250218e-005,.0171974308138302,.000274625070541602,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.0113956405123916,.000181976518058871,2.16598631798674e-006,.00166206328995312,2.65414208197256e-005,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,5.01144631725624e-005,8.00275816365055e-007,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,6.8546484667622e-008,1.09461601509022e-009,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,7.27823584467216,.60436208201944,.037405205103521,9.14177937313726,.759104944404841,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,6.45780747384476,.53623626019628,.0331887586818224,7.79428043908953,.647212821256298,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,5.18623489409549,.430648825529863,.0266537364367886,5.81833621896758,.483136554903789,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.7410698100085,.310646808870845,.0192265662361566,3.76406389114815,.31255616594975,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,2.39387697960251,.198780103635972,.0123029070952925,2.07554270972356,.172346615354148,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,1.33269841309056,.110663133873164,.00684916764813009,.95054294779362,.0789301318799674,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,.626153181688297,.0519938139714517,.00321800346775215,.347192252628355,.028829765505355,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,.236321146004057,.0196233734207421,.0012145307084342,.0946962283839188,.00786328046747136,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.0657431866024243,.0054591098696972,.000337875473054763,.0171974308138302,.00142802120123345,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.0113956405123916,.000946258567734596,5.85658777413804e-005,.00166206328995312,.000138012569524986,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,5.01144631725624e-005,4.16134925463982e-006,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,6.8546484667622e-008,5.69188702865245e-009,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,7.27823584467216,1.45033499905202,.215414174926983,9.14177937313726,1.82168355953934,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,6.45780747384476,1.28684813137965,.191131930663375,7.79428043908953,1.55316727245144,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,5.18623489409549,1.03346169878758,.153497157076452,5.81833621896758,1.1594206118243,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.7410698100085,.745483465381799,.110724560685561,3.76406389114815,.750065499032834,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,2.39387697960251,.477028175651108,.0708516521644813,2.07554270972356,.413593664547998,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,1.33269841309056,.26556698531571,.0394439167964793,.95054294779362,.189414816301507,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,.626153181688297,.124773625580585,.0185322753878643,.347192252628355,.0691850451424154,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,.236321146004057,.0470917453597858,.00699440438187064,.0946962283839188,.0188701296931614,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.0657431866024243,.0131006533057736,.00194580315907169,.0171974308138302,.00342693426532766,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.0113956405123916,.00227081076025246,.000337277130218073,.00166206328995312,.000331199567025036,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.000824565926330002,.000164311358893045,2.44047036237916e-005,5.01144631725624e-005,9.98631556618579e-006,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,6.8546484667622e-008,1.36592668764407e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,7.27823584467216,2.60119152746664,.692918704338266,9.14177937313726,3.26720919228747,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,6.45780747384476,2.30797605978546,.614810468242533,7.79428043908953,2.78562232344933,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,5.18623489409549,1.85352474883709,.493751403486214,5.81833621896758,2.07943342346341,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.7410698100085,1.33703266851113,.356165601240813,3.76406389114815,1.34525059548623,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,2.39387697960251,.855555199093713,.227907170151078,2.07554270972356,.741784716454702,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,1.33269841309056,.476297264169728,.126878501518793,.95054294779362,.339717524338981,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,.626153181688297,.223782848737467,.0596124199094679,.347192252628355,.12408412771496,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,.236321146004057,.0844595552914031,.0224987683542483,.0946962283839188,.0338437819621889,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.0657431866024243,.0234961635798143,.00625902823868614,.0171974308138302,.0061462437174735,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.0113956405123916,.00407272369067121,.00108491296894855,.00166206328995312,.000594010010245393,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.000824565926330002,.000294694201614443,7.85021488046878e-005,5.01144631725624e-005,1.79105651165762e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,6.8546484667622e-008,2.44980430684131e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,7.27823584467216,3.98464475452196,1.62598614322311,9.14177937313726,5.00488635016176,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,6.45780747384476,3.53548156799627,1.44269926011819,7.79428043908953,4.26716574385505,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,5.18623489409549,2.83932866528416,1.15862500931079,5.81833621896758,3.18538769471191,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.7410698100085,2.04813838310324,.835769519114577,3.76406389114815,2.06072704459494,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,2.39387697960251,1.31058536069927,.534801410737989,2.07554270972356,1.13630563078316,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,1.33269841309056,.729617706050062,.297729999278181,.95054294779362,.520397532085943,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,.626153181688297,.342802575265255,.139885051637275,.347192252628355,.190078724845143,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,.236321146004057,.129379678661691,.0527950614618356,.0946962283839188,.0518437211734884,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.0657431866024243,.0359926841107618,.0146872831147845,.0171974308138302,.00941514591687767,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.0113956405123916,.00623881668046779,.00254583033055523,.00166206328995312,.000909936406629402,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.000824565926330002,.000451428390509428,.000184211229066994,5.01144631725624e-005,2.74363646769969e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,6.8546484667622e-008,3.7527416869482e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,7.27823584467216,5.51377378765633,3.11340423610211,9.14177937313726,6.92553862835673,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,6.45780747384476,4.89224179250545,2.76245034842036,7.79428043908953,5.90471373874725,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,5.18623489409549,3.92893643816517,2.21851091848272,5.81833621896758,4.40779749680161,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.7410698100085,2.8341225946006,1.60031398303209,3.76406389114815,2.85154222320761,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,2.39387697960251,1.81352960012534,1.02402654819954,2.07554270972356,1.57236907874115,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,1.33269841309056,1.0096124574376,.570087171302651,.95054294779362,.720102907121204,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,.626153181688297,.474354922529452,.267848969162523,.347192252628355,.263022466294621,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,.547225517982271,.414561685224313,.234086155283846,.121144008364367,.236321146004057,.179029831969503,.101090878717283,.0946962283839188,.0717390303204678,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.0657431866024243,.0498050718253105,.0281229023119214,.0171974308138302,.0130282592204804,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.0113956405123916,.00863299641447793,.00487470263420339,.00166206328995312,.00125912943722616,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.000824565926330002,.000624666483447503,.000352723805983911,5.01144631725624e-005,3.79652184082228e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,6.8546484667622e-008,5.19287666029906e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,7.27823584467216,7.09250048322984,5.15153163620379,9.14177937313726,8.90848771819592,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,6.45780747384476,6.29300885631198,4.57083285180669,7.79428043908953,7.59537599078757,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,5.18623489409549,5.05388589728677,3.6708144253493,5.81833621896758,5.66985645040998,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.7410698100085,3.6456003901969,2.64792731244254,3.76406389114815,3.66800767948205,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,2.39387697960251,2.33278695510428,1.69438491095747,2.07554270972356,2.02257634793674,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,1.33269841309056,1.29868890491693,.943283260267015,.95054294779362,.926285773305773,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,.626153181688297,.610174201341807,.443190904144467,.347192252628355,.338332155278326,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,.547225517982271,.533260715077649,.387325943916553,.257841742921929,.236321146004057,.230290399761936,.167267987177829,.0946962283839188,.092279648532812,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.0657431866024243,.0640654676075239,.0465329941039433,.0171974308138302,.0167585647100291,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.0113956405123916,.0111048319353399,.00806582856989507,.00166206328995312,.00161964862649386,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.000824565926330002,.000803523594969987,.000583627344081487,5.01144631725624e-005,4.88355780044979e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,6.8546484667622e-008,6.67972275267736e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,7.27823584467216,8.62162951636421,7.61230772976691,9.14177937313726,10.8291399963909,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,6.45780747384476,7.64976908082116,6.7542215997961,7.79428043908953,9.23292398567978,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,5.18623489409549,6.14349367016778,5.42428368841739,5.81833621896758,6.89226625249968,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.7410698100085,4.43158460169426,3.91278535624427,3.76406389114815,4.45882285809472,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,2.39387697960251,2.83573119453035,2.50375621577018,2.07554270972356,2.45863979589474,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,1.33269841309056,1.57868365630447,1.39386942769156,.95054294779362,1.12599114834103,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,.626153181688297,.741726548606004,.654893686699248,.347192252628355,.411275896727804,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,.547225517982271,.648230667294543,.572343234018839,.463151607819505,.236321146004057,.279940553069749,.247168314523763,.0946962283839188,.112174957679791,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,.23320173943763,.276245376362,.243905726875982,.197373399115395,.0657431866024243,.0778778553220726,.0687608066341364,.0171974308138302,.0203716780136318,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.0113956405123916,.0134990116693501,.0119187017581499,.00166206328995312,.00196884165709061,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.000824565926330002,.000976761687908062,.000862413599759767,5.01144631725624e-005,5.93644317357238e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,6.8546484667622e-008,8.11985772602823e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,7.27823584467216,10.0050827434195,10.2513011633895,9.14177937313726,12.5668171542652,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,6.45780747384476,8.87727458903197,9.0957384017764,7.79428043908953,10.7144674060855,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,5.18623489409549,7.12929758661485,7.30474484999971,5.81833621896758,7.99822052374818,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.7410698100085,5.14269031628637,5.26924849841671,3.76406389114815,5.17429930720343,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,2.39387697960251,3.29076135613591,3.37174480048962,2.07554270972356,2.8531607102232,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,1.33269841309056,1.83200409818481,1.87708849838432,.95054294779362,1.306671156088,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,.626153181688297,.860746275133792,.881928667453122,.347192252628355,.477270493857987,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,.547225517982271,.752247756737171,.770760072903028,.723797635481732,.236321146004057,.324860676440036,.332855281233889,.0946962283839188,.130174896891091,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.0657431866024243,.0903743758530202,.0925984290266871,.0171974308138302,.023640580213036,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.0113956405123916,.0156651046591466,.0160506124472133,.00166206328995312,.00228476805347462,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.000824565926330002,.00113349587680305,.00116139045508752,5.01144631725624e-005,6.88902312961445e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,6.8546484667622e-008,9.42279510613512e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,7.27823584467216,11.1559392718342,12.7452954129935,9.14177937313726,14.0123427870133,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,6.45780747384476,9.89840251743778,11.3086008382978,7.79428043908953,11.9469224570834,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,5.18623489409549,7.94936063666436,9.0818842940915,5.81833621896758,8.91823333538729,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.7410698100085,5.7342395194157,6.55118093268348,3.76406389114815,5.76948440365682,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,2.39387697960251,3.66928837957851,4.19204185444651,2.07554270972356,3.1813517621299,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,1.33269841309056,2.04273437703882,2.33375715403626,.95054294779362,1.45697386412547,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,.626153181688297,.959755498290674,1.09648923787555,.347192252628355,.532169576430532,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,.236321146004057,.362228486371654,.41383418762992,.0946962283839188,.145148549160118,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.0657431866024243,.100769886127061,.115126296058794,.0171974308138302,.0263598896651818,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.0113956405123916,.0174670175895654,.0199554957891379,.00166206328995312,.00254757849669498,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.000824565926330002,.00126387871952445,.00144394006224153,5.01144631725624e-005,7.68144808465349e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,6.8546484667622e-008,1.05066727253324e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,7.27823584467216,12.0019121888667,14.7515791225583,9.14177937313726,15.0749214021478,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,6.45780747384476,10.6490143886212,13.0887291840651,7.79428043908953,12.8528769082785,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,5.18623489409549,8.55217350992208,10.5114970194908,5.81833621896758,9.5945173923078,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.7410698100085,6.16907617592665,7.58242635758397,3.76406389114815,6.20699373673991,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,2.39387697960251,3.94753645159365,4.85192654207924,2.07554270972356,3.42259881132375,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,1.33269841309056,2.19763822848137,2.70112242949704,.95054294779362,1.56745854854701,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,.626153181688297,1.03253530989981,1.26909163149447,.347192252628355,.572524856067592,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,.236321146004057,.389696858310697,.47897734533629,.0946962283839188,.156155398385808,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.0657431866024243,.108411429563137,.133248748684711,.0171974308138302,.0283588027292761,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.0113956405123916,.0187915697820833,.0230967635919399,.00166206328995312,.00274076549419503,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.000824565926330002,.00135972068684592,.00167123596481511,5.01144631725624e-005,8.26394471580808e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,6.8546484667622e-008,1.13034107101112e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,7.27823584467216,12.4900484198527,15.9759196314393,9.14177937313726,15.6880416449771,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,6.45780747384476,11.0821261849396,14.1750577199246,7.79428043908953,13.3756231834223,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,5.18623489409549,8.90000356218188,11.3839223715851,5.81833621896758,9.98474117367778,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.7410698100085,6.41998199374919,8.21174689798638,3.76406389114815,6.45944172008843,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,2.39387697960251,4.10808883148424,5.25462310508641,2.07554270972356,3.5618011698851,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,1.33269841309056,2.28701955580991,2.92530816462441,.95054294779362,1.63120949890175,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,.626153181688297,1.07453010927869,1.3744227476422,.347192252628355,.595810322665643,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,.236321146004057,.405546428999883,.518731148089159,.0946962283839188,.162506478648407,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.0657431866024243,.112820689170272,.144308028468726,.0171974308138302,.0295121988599679,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.0113956405123916,.019555851831759,.0250137314673599,.00166206328995312,.00285223664290029,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.000824565926330002,.00141502261880692,.00180994395496465,5.01144631725624e-005,8.60005205963556e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,6.8546484667622e-008,1.17631378114674e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,7.27823584467216,.116225851033514,.00138338509726576,9.14177937313726,.145984701575574,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,6.45780747384476,.103124463877826,.00122744505825102,7.79428043908953,.124466546112493,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,5.18623489409549,.0828187732700635,.000985755369367884,5.81833621896758,.0929127735338078,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.7410698100085,.0597409910483066,.000711070695350591,3.76406389114815,.0601081826012263,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,2.39387697960251,.0382277237453785,.000455007753107355,2.07554270972356,.0331442567927967,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,1.33269841309056,.0212818065446243,.000253307966815722,.95054294779362,.0151791815252261,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,.626153181688297,.00999901459257298,.000119013865260662,.347192252628355,.00554429890730429,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,.236321146004057,.00377380273155657,4.4917911225712e-005,.0946962283839188,.00151220020487266,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.0657431866024243,.00104985026256264,1.24959051250218e-005,.0171974308138302,.000274625070541602,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.0113956405123916,.000181976518058871,2.16598631798674e-006,.00166206328995312,2.65414208197256e-005,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,5.01144631725624e-005,8.00275816365055e-007,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,6.8546484667622e-008,1.09461601509022e-009,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,7.27823584467216,.60436208201944,.037405205103521,9.14177937313726,.759104944404841,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,6.45780747384476,.53623626019628,.0331887586818224,7.79428043908953,.647212821256298,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,5.18623489409549,.430648825529863,.0266537364367886,5.81833621896758,.483136554903789,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.7410698100085,.310646808870845,.0192265662361566,3.76406389114815,.31255616594975,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,2.39387697960251,.198780103635972,.0123029070952925,2.07554270972356,.172346615354148,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,1.33269841309056,.110663133873164,.00684916764813009,.95054294779362,.0789301318799674,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,.626153181688297,.0519938139714517,.00321800346775215,.347192252628355,.028829765505355,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,.236321146004057,.0196233734207421,.0012145307084342,.0946962283839188,.00786328046747136,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.0657431866024243,.0054591098696972,.000337875473054763,.0171974308138302,.00142802120123345,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.0113956405123916,.000946258567734596,5.85658777413804e-005,.00166206328995312,.000138012569524986,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,5.01144631725624e-005,4.16134925463982e-006,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,6.8546484667622e-008,5.69188702865245e-009,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,7.27823584467216,1.45033499905202,.215414174926983,9.14177937313726,1.82168355953934,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,6.45780747384476,1.28684813137965,.191131930663375,7.79428043908953,1.55316727245144,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,5.18623489409549,1.03346169878758,.153497157076452,5.81833621896758,1.1594206118243,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.7410698100085,.745483465381799,.110724560685561,3.76406389114815,.750065499032834,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,2.39387697960251,.477028175651108,.0708516521644813,2.07554270972356,.413593664547998,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,1.33269841309056,.26556698531571,.0394439167964793,.95054294779362,.189414816301507,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,.626153181688297,.124773625580585,.0185322753878643,.347192252628355,.0691850451424154,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,.236321146004057,.0470917453597858,.00699440438187064,.0946962283839188,.0188701296931614,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.0657431866024243,.0131006533057736,.00194580315907169,.0171974308138302,.00342693426532766,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.0113956405123916,.00227081076025246,.000337277130218073,.00166206328995312,.000331199567025036,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.000824565926330002,.000164311358893045,2.44047036237916e-005,5.01144631725624e-005,9.98631556618579e-006,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,6.8546484667622e-008,1.36592668764407e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,7.27823584467216,2.60119152746664,.692918704338266,9.14177937313726,3.26720919228747,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,6.45780747384476,2.30797605978546,.614810468242533,7.79428043908953,2.78562232344933,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,5.18623489409549,1.85352474883709,.493751403486214,5.81833621896758,2.07943342346341,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.7410698100085,1.33703266851113,.356165601240813,3.76406389114815,1.34525059548623,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,2.39387697960251,.855555199093713,.227907170151078,2.07554270972356,.741784716454702,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,1.33269841309056,.476297264169728,.126878501518793,.95054294779362,.339717524338981,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,.626153181688297,.223782848737467,.0596124199094679,.347192252628355,.12408412771496,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,.236321146004057,.0844595552914031,.0224987683542483,.0946962283839188,.0338437819621889,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.0657431866024243,.0234961635798143,.00625902823868614,.0171974308138302,.0061462437174735,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.0113956405123916,.00407272369067121,.00108491296894855,.00166206328995312,.000594010010245393,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.000824565926330002,.000294694201614443,7.85021488046878e-005,5.01144631725624e-005,1.79105651165762e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,6.8546484667622e-008,2.44980430684131e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,7.27823584467216,3.98464475452196,1.62598614322311,9.14177937313726,5.00488635016176,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,6.45780747384476,3.53548156799627,1.44269926011819,7.79428043908953,4.26716574385505,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,5.18623489409549,2.83932866528416,1.15862500931079,5.81833621896758,3.18538769471191,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.7410698100085,2.04813838310324,.835769519114577,3.76406389114815,2.06072704459494,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,2.39387697960251,1.31058536069927,.534801410737989,2.07554270972356,1.13630563078316,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,1.33269841309056,.729617706050062,.297729999278181,.95054294779362,.520397532085943,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,.626153181688297,.342802575265255,.139885051637275,.347192252628355,.190078724845143,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,.236321146004057,.129379678661691,.0527950614618356,.0946962283839188,.0518437211734884,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.0657431866024243,.0359926841107618,.0146872831147845,.0171974308138302,.00941514591687767,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.0113956405123916,.00623881668046779,.00254583033055523,.00166206328995312,.000909936406629402,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.000824565926330002,.000451428390509428,.000184211229066994,5.01144631725624e-005,2.74363646769969e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,6.8546484667622e-008,3.7527416869482e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,7.27823584467216,5.51377378765633,3.11340423610211,9.14177937313726,6.92553862835673,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,6.45780747384476,4.89224179250545,2.76245034842036,7.79428043908953,5.90471373874725,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,5.18623489409549,3.92893643816517,2.21851091848272,5.81833621896758,4.40779749680161,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.7410698100085,2.8341225946006,1.60031398303209,3.76406389114815,2.85154222320761,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,2.39387697960251,1.81352960012534,1.02402654819954,2.07554270972356,1.57236907874115,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,1.33269841309056,1.0096124574376,.570087171302651,.95054294779362,.720102907121204,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,.626153181688297,.474354922529452,.267848969162523,.347192252628355,.263022466294621,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,.547225517982271,.414561685224313,.234086155283846,.121144008364367,.236321146004057,.179029831969503,.101090878717283,.0946962283839188,.0717390303204678,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.0657431866024243,.0498050718253105,.0281229023119214,.0171974308138302,.0130282592204804,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.0113956405123916,.00863299641447793,.00487470263420339,.00166206328995312,.00125912943722616,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.000824565926330002,.000624666483447503,.000352723805983911,5.01144631725624e-005,3.79652184082228e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,6.8546484667622e-008,5.19287666029906e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,7.27823584467216,7.09250048322984,5.15153163620379,9.14177937313726,8.90848771819592,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,6.45780747384476,6.29300885631198,4.57083285180669,7.79428043908953,7.59537599078757,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,5.18623489409549,5.05388589728677,3.6708144253493,5.81833621896758,5.66985645040998,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.7410698100085,3.6456003901969,2.64792731244254,3.76406389114815,3.66800767948205,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,2.39387697960251,2.33278695510428,1.69438491095747,2.07554270972356,2.02257634793674,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,1.33269841309056,1.29868890491693,.943283260267015,.95054294779362,.926285773305773,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,.626153181688297,.610174201341807,.443190904144467,.347192252628355,.338332155278326,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,.547225517982271,.533260715077649,.387325943916553,.257841742921929,.236321146004057,.230290399761936,.167267987177829,.0946962283839188,.092279648532812,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.0657431866024243,.0640654676075239,.0465329941039433,.0171974308138302,.0167585647100291,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.0113956405123916,.0111048319353399,.00806582856989507,.00166206328995312,.00161964862649386,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.000824565926330002,.000803523594969987,.000583627344081487,5.01144631725624e-005,4.88355780044979e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,6.8546484667622e-008,6.67972275267736e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,7.27823584467216,8.62162951636421,7.61230772976691,9.14177937313726,10.8291399963909,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,6.45780747384476,7.64976908082116,6.7542215997961,7.79428043908953,9.23292398567978,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,5.18623489409549,6.14349367016778,5.42428368841739,5.81833621896758,6.89226625249968,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.7410698100085,4.43158460169426,3.91278535624427,3.76406389114815,4.45882285809472,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,2.39387697960251,2.83573119453035,2.50375621577018,2.07554270972356,2.45863979589474,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,1.33269841309056,1.57868365630447,1.39386942769156,.95054294779362,1.12599114834103,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,.626153181688297,.741726548606004,.654893686699248,.347192252628355,.411275896727804,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,.547225517982271,.648230667294543,.572343234018839,.463151607819505,.236321146004057,.279940553069749,.247168314523763,.0946962283839188,.112174957679791,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,.23320173943763,.276245376362,.243905726875982,.197373399115395,.0657431866024243,.0778778553220726,.0687608066341364,.0171974308138302,.0203716780136318,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.0113956405123916,.0134990116693501,.0119187017581499,.00166206328995312,.00196884165709061,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.000824565926330002,.000976761687908062,.000862413599759767,5.01144631725624e-005,5.93644317357238e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,6.8546484667622e-008,8.11985772602823e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,7.27823584467216,10.0050827434195,10.2513011633895,9.14177937313726,12.5668171542652,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,6.45780747384476,8.87727458903197,9.0957384017764,7.79428043908953,10.7144674060855,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,5.18623489409549,7.12929758661485,7.30474484999971,5.81833621896758,7.99822052374818,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.7410698100085,5.14269031628637,5.26924849841671,3.76406389114815,5.17429930720343,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,2.39387697960251,3.29076135613591,3.37174480048962,2.07554270972356,2.8531607102232,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,1.33269841309056,1.83200409818481,1.87708849838432,.95054294779362,1.306671156088,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,.626153181688297,.860746275133792,.881928667453122,.347192252628355,.477270493857987,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,.547225517982271,.752247756737171,.770760072903028,.723797635481732,.236321146004057,.324860676440036,.332855281233889,.0946962283839188,.130174896891091,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.0657431866024243,.0903743758530202,.0925984290266871,.0171974308138302,.023640580213036,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.0113956405123916,.0156651046591466,.0160506124472133,.00166206328995312,.00228476805347462,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.000824565926330002,.00113349587680305,.00116139045508752,5.01144631725624e-005,6.88902312961445e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,6.8546484667622e-008,9.42279510613512e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,7.27823584467216,11.1559392718342,12.7452954129935,9.14177937313726,14.0123427870133,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,6.45780747384476,9.89840251743778,11.3086008382978,7.79428043908953,11.9469224570834,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,5.18623489409549,7.94936063666436,9.0818842940915,5.81833621896758,8.91823333538729,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.7410698100085,5.7342395194157,6.55118093268348,3.76406389114815,5.76948440365682,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,2.39387697960251,3.66928837957851,4.19204185444651,2.07554270972356,3.1813517621299,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,1.33269841309056,2.04273437703882,2.33375715403626,.95054294779362,1.45697386412547,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,.626153181688297,.959755498290674,1.09648923787555,.347192252628355,.532169576430532,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,.236321146004057,.362228486371654,.41383418762992,.0946962283839188,.145148549160118,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.0657431866024243,.100769886127061,.115126296058794,.0171974308138302,.0263598896651818,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.0113956405123916,.0174670175895654,.0199554957891379,.00166206328995312,.00254757849669498,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.000824565926330002,.00126387871952445,.00144394006224153,5.01144631725624e-005,7.68144808465349e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,6.8546484667622e-008,1.05066727253324e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,7.27823584467216,12.0019121888667,14.7515791225583,9.14177937313726,15.0749214021478,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,6.45780747384476,10.6490143886212,13.0887291840651,7.79428043908953,12.8528769082785,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,5.18623489409549,8.55217350992208,10.5114970194908,5.81833621896758,9.5945173923078,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.7410698100085,6.16907617592665,7.58242635758397,3.76406389114815,6.20699373673991,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,2.39387697960251,3.94753645159365,4.85192654207924,2.07554270972356,3.42259881132375,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,1.33269841309056,2.19763822848137,2.70112242949704,.95054294779362,1.56745854854701,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,.626153181688297,1.03253530989981,1.26909163149447,.347192252628355,.572524856067592,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,.236321146004057,.389696858310697,.47897734533629,.0946962283839188,.156155398385808,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.0657431866024243,.108411429563137,.133248748684711,.0171974308138302,.0283588027292761,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.0113956405123916,.0187915697820833,.0230967635919399,.00166206328995312,.00274076549419503,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.000824565926330002,.00135972068684592,.00167123596481511,5.01144631725624e-005,8.26394471580808e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,6.8546484667622e-008,1.13034107101112e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,7.27823584467216,12.4900484198527,15.9759196314393,9.14177937313726,15.6880416449771,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,6.45780747384476,11.0821261849396,14.1750577199246,7.79428043908953,13.3756231834223,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,5.18623489409549,8.90000356218188,11.3839223715851,5.81833621896758,9.98474117367778,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.7410698100085,6.41998199374919,8.21174689798638,3.76406389114815,6.45944172008843,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,2.39387697960251,4.10808883148424,5.25462310508641,2.07554270972356,3.5618011698851,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,1.33269841309056,2.28701955580991,2.92530816462441,.95054294779362,1.63120949890175,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,.626153181688297,1.07453010927869,1.3744227476422,.347192252628355,.595810322665643,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,.236321146004057,.405546428999883,.518731148089159,.0946962283839188,.162506478648407,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.0657431866024243,.112820689170272,.144308028468726,.0171974308138302,.0295121988599679,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.0113956405123916,.019555851831759,.0250137314673599,.00166206328995312,.00285223664290029,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.000824565926330002,.00141502261880692,.00180994395496465,5.01144631725624e-005,8.60005205963556e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,6.8546484667622e-008,1.17631378114674e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505},
			{-1.77695599756377e-005,-4.02808883404862e-005,-6.02967929985932e-005,-7.65271814689107e-005,-8.7949757770287e-005,-9.38463500146473e-005,-9.38463500146473e-005,-8.7949757770287e-005,-7.65271814689107e-005,-6.02967929985932e-005,-4.02808883404862e-005,-1.77695599756377e-005,-.000209456341481459,-.000474805651630584,-.000710740484393843,-.000902054045026543,-.00103669615466006,-.00110620145700885,-.00110620145700885,-.00103669615466006,-.000902054045026543,-.000710740484393843,-.000474805651630584,-.000209456341481459,-.000752419091267727,-.00170561957877169,-.00255315597328018,-.00324040169914516,-.00372406952729629,-.00397374979986132,-.00397374979986132,-.00372406952729629,-.00324040169914516,-.00255315597328018,-.00170561957877169,-.000752419091267727,-.00171271583978708,-.00388246617226489,-.00581169553993682,-.00737605861122963,-.00847702157201008,-.00904536355411474,-.00904536355411474,-.00847702157201008,-.00737605861122963,-.00581169553993682,-.00388246617226489,-.00171271583978708,-.00301523746950065,-.00683508449255492,-.0102314941838372,-.012985556497555,-.0149238025829595,-.0159243690517937,-.0159243690517937,-.0149238025829595,-.012985556497555,-.0102314941838372,-.00683508449255492,-.00301523746950065,-.00445208652099934,-.0100922026364427,-.0151071011176736,-.0191735548643253,-.0220354320327057,-.0235127977574022,-.0235127977574022,-.0220354320327057,-.0191735548643253,-.0151071011176736,-.0100922026364427,-.00445208652099934,-.00572682649263902,-.012981844165618,-.0194326292850773,-.0246634070198705,-.0283447087891094,-.0302450800266409,-.0302450800266409,-.0283447087891094,-.0246634070198705,-.0194326292850773,-.012981844165618,-.00572682649263902,-.00652410991880577,-.0147891643642708,-.022138021752572,-.0280970234695782,-.0322908326268278,-.0344557717700214,-.0344557717700214,-.0322908326268278,-.0280970234695782,-.022138021752572,-.0147891643642708,-.00652410991880577,-.0065876977962188,-.0149333084057324,-.0223537921535752,-.028370873865462,-.0326055583951231,-.034791598330081,-.034791598330081,-.0326055583951231,-.028370873865462,-.0223537921535752,-.0149333084057324,-.0065876977962188,-.00578758817420658,-.0131195816511808,-.0196388096598727,-.0249250859943669,-.0286454464091129,-.0305659806028896,-.0305659806028896,-.0286454464091129,-.0249250859943669,-.0196388096598727,-.0131195816511809,-.00578758817420658,-.00415955383809286,-.00942907556113789,-.0141144607455587,-.0179137205329708,-.0205875527021657,-.0219678453450519,-.0219678453450519,-.0205875527021657,-.0179137205329708,-.0141144607455587,-.0094290755611379,-.00415955383809286,-.00190958089376513,-.00432872929123383,-.00647971047247572,-.00822388645453697,-.00945139763053613,-.0100850666636237,-.0100850666636237,-.00945139763053613,-.00822388645453697,-.00647971047247572,-.00432872929123383,-.00190958089376513},
			{-2.11503243320183e-007,-4.79445666606905e-007,-7.17686160968121e-007,-9.10869324002267e-007,-1.0468272170579e-006,-1.11701175656978e-006,-1.11701175656978e-006,-1.0468272170579e-006,-9.10869324002267e-007,-7.17686160968121e-007,-4.79445666606905e-007,-2.11503243320183e-007,-1.29636812850387e-005,-2.93866926947107e-005,-4.39891819502224e-005,-5.58299694289237e-005,-6.41632449195998e-005,-6.84650702111951e-005,-6.84650702111951e-005,-6.41632449195998e-005,-5.58299694289237e-005,-4.39891819502224e-005,-2.93866926947107e-005,-1.29636812850387e-005,-.000111754689657294,-.00025333087492222,-.000379213069886165,-.000481287743035328,-.000553125564084201,-.000590209872148528,-.000590209872148528,-.000553125564084201,-.000481287743035328,-.000379213069886165,-.00025333087492222,-.000111754689657294,-.000456242005972593,-.0010342311978472,-.00154814918484807,-.00196487222152227,-.00225815236645889,-.00240955021073066,-.00240955021073066,-.00225815236645889,-.00196487222152227,-.00154814918484807,-.0010342311978472,-.000456242005972593,-.00123040688572486,-.0027891451698972,-.00417509434078899,-.00529892530561013,-.00608985184251931,-.00649814601017004,-.00649814601017004,-.00608985184251931,-.00529892530561013,-.00417509434078899,-.0027891451698972,-.00123040688572486,-.00251391253350483,-.00569865715294523,-.00853036675399556,-.0108265281139349,-.0124425139778682,-.0132767224314477,-.0132767224314477,-.0124425139778682,-.0108265281139349,-.00853036675399556,-.00569865715294523,-.00251391253350483,-.00415959476024,-.00942916832541804,-.0141145996051725,-.0179138967701193,-.0205877552447685,-.0219680614671252,-.0219680614671252,-.0205877552447685,-.0179138967701193,-.0141145996051725,-.00942916832541803,-.00415959476024,-.00576034173845761,-.0130578181276279,-.0195463553335665,-.0248077452759809,-.0285105912170028,-.030422084043298,-.030422084043298,-.0285105912170028,-.0248077452759809,-.0195463553335665,-.0130578181276279,-.00576034173845761,-.00674981664852096,-.0153008071730865,-.0229039040804791,-.0290690621631277,-.0334079594567994,-.0356477963776399,-.0356477963776399,-.0334079594567994,-.0290690621631277,-.0229039040804791,-.0153008071730865,-.00674981664852096,-.006612130023256,-.0149886925463945,-.0224366971291726,-.0284760947867864,-.0327264847688013,-.0349206322134925,-.0349206322134925,-.0327264847688013,-.0284760947867864,-.0224366971291726,-.0149886925463945,-.006612130023256,-.00511251762065263,-.0115892994367119,-.0173481176290686,-.0220177969659965,-.0253042105122013,-.02700073151736,-.02700073151736,-.0253042105122013,-.0220177969659965,-.0173481176290686,-.0115892994367119,-.00511251762065263,-.00244252943287481,-.005536842526707,-.00828814510932977,-.01051910646122,-.0120892060502712,-.0128997269708937,-.0128997269708937,-.0120892060502712,-.01051910646122,-.00828814510932977,-.005536842526707,-.00244252943287481},
			{-2.30726263295083e-009,-5.23021327582048e-009,-7.82914926217426e-009,-9.9365604128826e-009,-1.14197081952885e-008,-1.21853426266313e-008,-1.21853426266313e-008,-1.14197081952885e-008,-9.9365604128826e-009,-7.82914926217426e-009,-5.23021327582048e-009,-2.30726263295083e-009,-7.35364770593756e-007,-1.66696002908487e-006,-2.49528617544501e-006,-3.16695479923119e-006,-3.63965981910605e-006,-3.88368084208039e-006,-3.88368084208039e-006,-3.63965981910605e-006,-3.16695479923119e-006,-2.49528617544501e-006,-1.66696002908487e-006,-7.35364770593756e-007,-1.52128768774711e-005,-3.44852768261633e-005,-5.16212944637669e-005,-6.55164557289296e-005,-7.52955525177414e-005,-8.03437434652432e-005,-8.03437434652432e-005,-7.52955525177414e-005,-6.55164557289296e-005,-5.16212944637669e-005,-3.44852768261633e-005,-1.52128768774711e-005,-.000111389656582294,-.000252503400490955,-.000377974416605436,-.000479715675273885,-.000551318846834677,-.000588282023525919,-.000588282023525919,-.000551318846834677,-.000479715675273885,-.000377974416605436,-.000252503400490955,-.000111389656582294,-.000460167166116478,-.00104312893857298,-.00156146828610925,-.00198177649169201,-.00227757979653243,-.00243028015301575,-.00243028015301575,-.00227757979653243,-.00198177649169201,-.00156146828610925,-.00104312893857298,-.000460167166116478,-.00130099723581861,-.00294916275130836,-.00441462597427551,-.00560293286342162,-.0064392360816438,-.00687095471852507,-.00687095471852507,-.0064392360816438,-.00560293286342162,-.00441462597427551,-.00294916275130836,-.00130099723581861,-.00276903001105212,-.00627696965144704,-.00939604748863851,-.011925228449054,-.0137052081798638,-.0146240739767634,-.0146240739767634,-.0137052081798638,-.011925228449054,-.00939604748863851,-.00627696965144704,-.00276903001105212,-.00466138390221013,-.0105666479493299,-.0158173022080392,-.0200749243239431,-.0230713414195831,-.0246181596977759,-.0246181596977759,-.0230713414195831,-.0200749243239431,-.0158173022080392,-.0105666479493299,-.00466138390221013,-.00633855009047991,-.0143685284714563,-.0215083684256039,-.0272978832165217,-.0313724113504692,-.0334757749315177,-.0334757749315177,-.0313724113504692,-.0272978832165217,-.0215083684256039,-.0143685284714563,-.00633855009047991,-.00692348559220745,-.0156944882399467,-.0234932085068186,-.0298169926007543,-.0342675272542232,-.036564993905222,-.036564993905222,-.0342675272542232,-.0298169926007543,-.0234932085068186,-.0156944882399467,-.00692348559220746,-.00575920488574972,-.0130552410555424,-.0195424976931696,-.0248028492552804,-.0285049644079869,-.0304160799848247,-.0304160799848247,-.0285049644079869,-.0248028492552804,-.0195424976931696,-.0130552410555424,-.00575920488574972,-.00286339454936457,-.00649087969965739,-.0097162511988505,-.0123316229887931,-.014172261854691,-.0151224412691208,-.0151224412691208,-.014172261854691,-.0123316229887931,-.0097162511988505,-.00649087969965739,-.00286339454936457},
			{-2.41204413706157e-011,-5.46773700026945e-011,-8.18470047853013e-011,-1.03878171232717e-010,-1.19383202440944e-010,-1.27387250245815e-010,-1.27387250245815e-010,-1.19383202440944e-010,-1.03878171232717e-010,-8.18470047853013e-011,-5.46773700026945e-011,-2.41204413706157e-011,-3.99747274956451e-008,-9.06166239851226e-008,-1.35644769610767e-007,-1.72156948704618e-007,-1.97853385508442e-007,-2.11118467392541e-007,-2.11118467392541e-007,-1.97853385508442e-007,-1.72156948704618e-007,-1.35644769610767e-007,-9.06166239851226e-008,-3.99747274956451e-008,-1.98456423704136e-006,-4.49870511967568e-006,-6.73414868783207e-006,-8.54681307319903e-006,-9.82252482147343e-006,-1.04810760801775e-005,-1.04810760801775e-005,-9.82252482147343e-006,-8.54681307319903e-006,-6.73414868783207e-006,-4.49870511967567e-006,-1.98456423704136e-006,-2.60616962562586e-005,-5.90778994134517e-005,-8.84341934471251e-005,-.00011223846631682,-.000128991369283385,-.000137639596714465,-.000137639596714465,-.000128991369283385,-.00011223846631682,-8.84341934471251e-005,-5.90778994134517e-005,-2.60616962562586e-005,-.000164926605038453,-.000373863515530918,-.00055963937078906,-.000710280291134566,-.000816297926504351,-.000871026627805438,-.000871026627805438,-.000816297926504351,-.000710280291134566,-.00055963937078906,-.000373863515530918,-.000164926605038453,-.00064522442811524,-.00146262559000318,-.0021894162732808,-.00277875237013443,-.00319351364006811,-.00340762278874215,-.00340762278874215,-.00319351364006811,-.00277875237013443,-.0021894162732808,-.00146262559000318,-.00064522442811524,-.00176649543659496,-.00400437633419262,-.00599418386382994,-.00760766822733064,-.00874320163042015,-.00932938965676434,-.00932938965676434,-.00874320163042015,-.00760766822733064,-.00599418386382994,-.00400437633419262,-.00176649543659496,-.00361484526436693,-.00819430411680041,-.0122661212166373,-.0155678542354177,-.0178915384406828,-.0190910767848879,-.0190910767848879,-.0178915384406828,-.0155678542354177,-.0122661212166373,-.00819430411680041,-.00361484526436693,-.00570421773357935,-.0129305935493704,-.0193559118161673,-.0245660391272654,-.028232807600475,-.0301256764218797,-.0301256764218797,-.028232807600475,-.0245660391272654,-.0193559118161673,-.0129305935493704,-.00570421773357935,-.00694730582706703,-.0157484850874735,-.0235740368319136,-.0299195778892043,-.0343854245960146,-.0366907956752786,-.0366907956752786,-.0343854245960146,-.0299195778892043,-.0235740368319136,-.0157484850874735,-.00694730582706703,-.00621725180422221,-.01409356371536,-.0210967714210865,-.0267754945936724,-.0307720501198624,-.03283516247719,-.03283516247719,-.0307720501198624,-.0267754945936724,-.0210967714210865,-.01409356371536,-.00621725180422221,-.00321684991523739,-.00729210922619538,-.0109156182658738,-.0138538296704664,-.0159216756755027,-.0169891445541618,-.0169891445541618,-.0159216756755027,-.0138538296704664,-.0109156182658738,-.00729210922619538,-.00321684991523739},
			{-2.45853458176161e-013,-5.57312376361246e-013,-8.34245479120496e-013,-1.05880349510068e-012,-1.21684229228106e-012,-1.29842549393142e-012,-1.29842549393142e-012,-1.21684229228106e-012,-1.05880349510068e-012,-8.34245479120496e-013,-5.57312376361246e-013,-2.45853458176161e-013,-2.11870775687607e-009,-4.80278806553758e-009,-7.18933295005722e-009,-9.12452155328201e-009,-1.04864630445972e-008,-1.11895280470165e-008,-1.11895280470165e-008,-1.04864630445972e-008,-9.12452155328201e-009,-7.18933295005722e-009,-4.80278806553757e-009,-2.11870775687607e-009,-2.52418873530302e-007,-5.72195173861672e-007,-8.56523661084552e-007,-1.08707840640484e-006,-1.24933756458098e-006,-1.33309941203413e-006,-1.33309941203413e-006,-1.24933756458098e-006,-1.08707840640484e-006,-8.56523661084552e-007,-5.72195173861672e-007,-2.52418873530302e-007,-5.94515724243734e-006,-1.34767667504191e-005,-2.01734829721616e-005,-2.56036799885296e-005,-2.94253284884657e-005,-3.1398149882767e-005,-3.1398149882767e-005,-2.94253284884657e-005,-2.56036799885296e-005,-2.01734829721616e-005,-1.34767667504191e-005,-5.94515724243734e-006,-5.76326598948289e-005,-.000130644469596975,-.000195562780868023,-.000248203389861898,-.000285250646856585,-.000304375278857807,-.000304375278857807,-.000285250646856585,-.000248203389861898,-.000195562780868023,-.000130644469596975,-5.7632659894829e-005,-.000311995321652138,-.000707245915568759,-.00105868222690829,-.00134365300155203,-.00154420891695623,-.00164774041669247,-.00164774041669247,-.00154420891695623,-.00134365300155203,-.00105868222690829,-.000707245915568759,-.000311995321652138,-.00109875339493098,-.00249070674094487,-.00372835940234233,-.00473194049592369,-.00543823792293946,-.00580284462990898,-.00580284462990898,-.00543823792293946,-.00473194049592369,-.00372835940234233,-.00249070674094487,-.00109875339493098,-.00273317477899627,-.00619569129673009,-.00927438125108664,-.0117708127035932,-.0135277440795465,-.0144347117943581,-.0144347117943581,-.0135277440795465,-.0117708127035932,-.00927438125108664,-.00619569129673009,-.00273317477899627,-.00500501190886783,-.0113456003480285,-.0169833224592084,-.0215548080610388,-.0247721151016589,-.0264329617655731,-.0264329617655731,-.0247721151016589,-.0215548080610388,-.0169833224592084,-.0113456003480285,-.00500501190886783,-.00679690022731426,-.0154075384811573,-.0230636710532136,-.0292718344086679,-.0336409978300346,-.0358964588105553,-.0358964588105553,-.0336409978300346,-.0292718344086679,-.0230636710532136,-.0154075384811573,-.00679690022731426,-.00654390879914259,-.0148340454130535,-.0222052045635651,-.0281822902128456,-.0323888264281449,-.0345603355665664,-.0345603355665664,-.0323888264281449,-.0281822902128456,-.0222052045635651,-.0148340454130535,-.00654390879914259,-.00352357286688459,-.00798740348129818,-.0119564099539485,-.0151747764476273,-.0174397891986819,-.0186090400111835,-.0186090400111835,-.0174397891986819,-.0151747764476273,-.0119564099539485,-.00798740348129818,-.00352357286688459},
			{-2.51299527189796e-005,-.00029621599889564,-.00106408128362993,-.00242214596950661,-.00426418972382729,-.00629620113986558,-.00809895569654666,-.00922648473106781,-.00931641156972225,-.00818488569070588,-.00588249745219549,-.00270055519884434,-5.69657786046887e-005,-.000671476592135066,-.00241211034093489,-.00549063631715242,-.00966626919088759,-.014272529844964,-.0183591000865756,-.0209150368234713,-.0211188872818742,-.0185538903067371,-.0133347265413407,-.00612174767248644,-8.52725424398906e-005,-.00100513883251861,-.00361070780484583,-.00821897865428013,-.0144695178404447,-.0213646672921825,-.0274818878919328,-.0313078906116195,-.0316130360391258,-.0277734709743103,-.0199608618151533,-.00916369443189591,-.000108225777940879,-.00127569706467466,-.00458262003120292,-.0104313221265329,-.0183643501167498,-.0271155013323826,-.034879324707423,-.0397351916593658,-.0401224746033481,-.035249394662203,-.025333826534352,-.011630331761288,-.000124379740266118,-.00146610976221544,-.00526662963356749,-.0119883188775888,-.0211054440183851,-.0311628068384012,-.0400854715975038,-.045666133448504,-.0461112228985269,-.0405107788184967,-.0291151962521434,-.013366294714629,-.000132718780991214,-.00156440510346971,-.00561973086134261,-.0127920758168761,-.0225204586888932,-.0332521174831873,-.0427730023755956,-.0487278197470118,-.0492027502229294,-.0432268243227726,-.0310672248280699,-.014262438055421,-.000132718780991214,-.00156440510346971,-.00561973086134261,-.0127920758168761,-.0225204586888932,-.0332521174831873,-.0427730023755956,-.0487278197470118,-.0492027502229294,-.0432268243227726,-.0310672248280699,-.014262438055421,-.000124379740266118,-.00146610976221544,-.00526662963356749,-.0119883188775888,-.0211054440183851,-.0311628068384012,-.0400854715975038,-.045666133448504,-.0461112228985269,-.0405107788184967,-.0291151962521434,-.013366294714629,-.000108225777940879,-.00127569706467466,-.00458262003120292,-.0104313221265329,-.0183643501167498,-.0271155013323826,-.034879324707423,-.0397351916593658,-.0401224746033481,-.035249394662203,-.025333826534352,-.011630331761288,-8.52725424398906e-005,-.00100513883251861,-.00361070780484583,-.00821897865428013,-.0144695178404447,-.0213646672921825,-.0274818878919328,-.0313078906116195,-.0316130360391258,-.0277734709743103,-.0199608618151533,-.00916369443189591,-5.69657786046887e-005,-.000671476592135066,-.00241211034093489,-.00549063631715242,-.00966626919088758,-.014272529844964,-.0183591000865756,-.0209150368234713,-.0211188872818742,-.0185538903067371,-.0133347265413407,-.00612174767248644,-2.51299527189796e-005,-.00029621599889564,-.00106408128362993,-.00242214596950661,-.00426418972382729,-.00629620113986558,-.00809895569654666,-.00922648473106781,-.00931641156972225,-.00818488569070588,-.00588249745219549,-.00270055519884434},
			{-4.01299188864111e-007,-4.73026119128408e-006,-1.69922705697598e-005,-3.86791500860789e-005,-6.80946715845654e-005,-.000100543778916268,-.000129331892821903,-.000147337357934948,-.000148773395949532,-.000130704105389941,-9.39373616201976e-005,-4.31250557013781e-005,-4.73026119128408e-006,-5.5757329091799e-005,-.000200294145262156,-.000455925373489819,-.000802656948401598,-.00118514651569285,-.00152448260646228,-.00173671964859619,-.00175364675704319,-.00154065738088721,-.00110727424428985,-.000508330898783026,-1.69922705697598e-005,-.000200294145262156,-.000719506211645245,-.00163779694031119,-.00288334269301423,-.00425734001677702,-.00547631935750839,-.00623872741470196,-.00629953378353948,-.00553442315607657,-.00397760351765944,-.00182605057560164,-3.86791500860789e-005,-.000455925373489819,-.00163779694031119,-.00372808291892168,-.0065632926638522,-.00969089403328137,-.0124656312103062,-.0142010847243002,-.0143394969898535,-.0125978916716827,-.0090541356913339,-.0041566007314033,-6.80946715845654e-005,-.000802656948401598,-.00288334269301423,-.0065632926638522,-.0115546814618157,-.0170608259252993,-.0219457527239106,-.0250010198852289,-.0252446947784246,-.0221785973614222,-.015939812405686,-.00731769858135901,-.000100543778916268,-.00118514651569285,-.00425734001677702,-.00969089403328137,-.0170608259252993,-.0251908096484753,-.0324035472773186,-.0369147388118314,-.0372745319354167,-.0327473492108772,-.0235356003221748,-.0108047964872896,-.000129331892821903,-.00152448260646228,-.00547631935750839,-.0124656312103062,-.0219457527239106,-.0324035472773186,-.0416814660110368,-.0474843207110434,-.0479471313016055,-.0421237067473831,-.0302744115168138,-.0138984708583546,-.000147337357934948,-.00173671964859619,-.00623872741470196,-.0142010847243002,-.0250010198852289,-.0369147388118314,-.0474843207110434,-.0540950434131128,-.0546222860610777,-.0479881297889341,-.0344891867628131,-.0158334029675552,-.000148773395949532,-.00175364675704319,-.00629953378353948,-.0143394969898535,-.0252446947784246,-.0372745319354167,-.0479471313016055,-.0546222860610777,-.0551546675312396,-.0484558508041033,-.0348253390053783,-.0159877247830154,-.000130704105389941,-.00154065738088721,-.00553442315607657,-.0125978916716827,-.0221785973614222,-.0327473492108772,-.0421237067473831,-.0479881297889341,-.0484558508041033,-.0425706396619946,-.0305956233004421,-.0140459337615274,-9.39373616201976e-005,-.00110727424428985,-.00397760351765944,-.0090541356913339,-.015939812405686,-.0235356003221748,-.0302744115168138,-.0344891867628131,-.0348253390053783,-.0305956233004421,-.0219891496246005,-.0100948470984408,-4.31250557013781e-005,-.000508330898783026,-.00182605057560164,-.0041566007314033,-.00731769858135901,-.0108047964872896,-.0138984708583546,-.0158334029675552,-.0159877247830154,-.0140459337615274,-.0100948470984408,-.00463437375617701},
			{-4.77648743752639e-009,-5.63022148645256e-008,-2.02251510004898e-007,-4.60380881913306e-007,-8.10500874190154e-007,-1.19672830207889e-006,-1.53938053820615e-006,-1.75369165645863e-006,-1.77078418424541e-006,-1.55571337982341e-006,-1.11809502771057e-006,-5.13298537640478e-007,-2.92765537918228e-007,-3.45093511422209e-006,-1.23966142266297e-005,-2.82181536753749e-005,-4.96780798691598e-005,-7.33511413319109e-005,-9.43533458892899e-005,-.000107489130425036,-.00010853678378063,-9.53544357538297e-005,-6.85314672157709e-005,-3.14616387984846e-005,-2.52381411676342e-006,-2.97491255946283e-005,-.000106866232302167,-.00024325736936566,-.000428254773970375,-.000632330728850647,-.000813382299066098,-.000926620621724616,-.000935652020526277,-.000822012292713028,-.000590782254057447,-.000271218151906612,-1.03035498453369e-005,-.000121451733066827,-.000436284726357267,-.000993105797236321,-.0017483634713027,-.00258150992189795,-.00332065860400601,-.0037829575879782,-.0038198285552606,-.00335589082222505,-.00241188697772547,-.00110725656401101,-2.7786916836135e-005,-.000327534612535476,-.00117658550597994,-.00267823697769347,-.00471503812818561,-.00696189202635823,-.00895524997265581,-.0102019914951301,-.0103014262061688,-.00905026525693164,-.00650444787225885,-.00298608212919577,-5.67729905547983e-005,-.000669204128457476,-.00240394708818667,-.00547205447566498,-.00963355584556567,-.0142242276315438,-.0182969677820565,-.0208442545176493,-.021047415089364,-.0184910987778976,-.0132895981153155,-.00610102997451746,-9.39382857945563e-005,-.00110728513787305,-.00397764265011023,-.00905422476771312,-.0159399692247158,-.0235358318700697,-.0302747093624672,-.0344895260742596,-.0348256816239572,-.0305959243062406,-.0219893659582073,-.0100949464135448,-.000130088785011913,-.0015334043731933,-.00550836855481976,-.0125385841278022,-.0220741863877053,-.0325931833471828,-.0419253994709405,-.0477622143590056,-.0482277334672722,-.0423702283435318,-.0304515871935496,-.0139798092182615,-.000152434610085198,-.00179680283515025,-.00645456111210986,-.0146923824553127,-.0258659498945255,-.0381918333275861,-.0491270782522392,-.0559665041221956,-.0565119871455732,-.0496483170026962,-.0356823673916168,-.0163811720361289,-.000149325161617318,-.00176015062198346,-.00632289727835408,-.0143926788244271,-.0253383214364947,-.0374127744408963,-.0481249559775311,-.0548248673220092,-.0553592232703062,-.04863556220147,-.0349544980281293,-.016047019508292,-.000115458636973295,-.001360953435314,-.0048888820448881,-.0111284599424805,-.0195916617438277,-.0289276629306809,-.0372103519687145,-.0423907423550166,-.0428039078843937,-.0376051541440983,-.027026916660946,-.0124076008346097,-5.51609089712794e-005,-.000650201929690796,-.00233568648062049,-.00531667428240711,-.00936000890342516,-.0138203275519371,-.0177774213479609,-.0202523773150973,-.0204497690975875,-.0179660399513751,-.0129122370468795,-.00592778988329973},
			{-5.21061086714588e-011,-6.14193874587905e-010,-2.20633662228147e-009,-5.02223790536332e-009,-8.84165344957659e-009,-1.30549605277777e-008,-1.67929112469387e-008,-1.91308046389391e-008,-1.9317264903302e-008,-1.6971084189159e-008,-1.21971598964522e-008,-5.59951004436151e-009,-1.66071240016209e-008,-1.95754280954478e-007,-7.03197893870279e-007,-1.60067465766628e-006,-2.81798888768115e-006,-4.16084320723466e-006,-5.35219317920171e-006,-6.09732050598647e-006,-6.15674863851456e-006,-5.40897999788056e-006,-3.88744721171111e-006,-1.78466133868042e-006,-3.43560288483173e-007,-4.04967152831355e-006,-1.45474238197567e-005,-3.31139965656822e-005,-5.82972148036964e-005,-8.60775467486887e-005,-.000110723628756234,-.000126138468756338,-.00012736788971769,-.000111898407472009,-8.04216603301225e-005,-3.69202255792134e-005,-2.51557038538368e-006,-2.96519536996954e-005,-.000106517166772019,-.00024246279880047,-.000426855931918081,-.000630265297550785,-.000810725484866511,-.000923593928337934,-.000932595827175548,-.000819327289690436,-.000588852535789363,-.00027033225085123,-1.03921937720791e-005,-.000122496611646568,-.000440038189183426,-.00100164972615974,-.00176340506432601,-.00260371927496907,-.00334922703163026,-.00381550328536708,-.00385269146248218,-.00338476236112819,-.0024326370236585,-.0011167825595386,-2.93810953216598e-005,-.000346325780898917,-.0012440880400349,-.00283189158407368,-.00498554717338517,-.00736130656207001,-.00946902654323017,-.0107872955591597,-.0108924349936834,-.00956949300162804,-.0068776181278623,-.00315739828904347,-6.253444086072e-005,-.000737116463054399,-.00264790502577821,-.00602737082635835,-.0106111906809188,-.0156677341271221,-.0201537850748433,-.0229595761766608,-.0231833539384622,-.0203676169191767,-.0146382563131457,-.00672017616832581,-.00010527045022929,-.00124086152956873,-.00445748215529598,-.010146473397001,-.0178628737232513,-.026375056735671,-.0339268729271671,-.0386501404333965,-.039026847819149,-.0342868373598303,-.0246420342366014,-.0113127416048292,-.000143146764099252,-.00168732357718739,-.00606128448364179,-.0137971750917403,-.0242899366861979,-.0358648036217386,-.0461337637005683,-.0525564631192487,-.0530687098434273,-.0466232433562013,-.0335082395307456,-.0153830666658602,-.000156356665903898,-.00184303355014811,-.00662063329842476,-.0150704090994405,-.0265314660737799,-.039174487477097,-.0503910900341255,-.0574064904417204,-.0579660084330403,-.0509257399611718,-.0366004544098804,-.0168026502756542,-.000130063110871032,-.0015331017426483,-.00550728143089673,-.0125361095303546,-.0220698298571165,-.0325867508020486,-.0419171251326689,-.0477527880751015,-.0482182153091671,-.0423618662144552,-.0304455773108406,-.0139770501826433,-6.46655241703613e-005,-.000762236326125946,-.0027381417997588,-.00623277490756985,-.0109728047138306,-.0162016678481028,-.0208405969245687,-.0237420053295095,-.0239734091138143,-.0210617158489217,-.015137106919017,-.00694919005368325},
			{-5.44724437242996e-013,-6.4208673651397e-012,-2.30653469542085e-011,-5.25031668349854e-011,-9.24318630275365e-011,-1.36478355571743e-010,-1.75555407262082e-010,-1.99996066807805e-010,-2.01945348094836e-010,-1.77418051741924e-010,-1.27510789616824e-010,-5.85380492906022e-011,-9.02770003435125e-010,-1.06412821914539e-008,-3.82261230182231e-008,-8.70133242853409e-008,-1.53187020074262e-007,-2.26185126101403e-007,-2.9094739428102e-007,-3.31452818296366e-007,-3.34683355709183e-007,-2.94034348800585e-007,-2.11323208782435e-007,-9.70149149662404e-008,-4.48184434349486e-008,-5.28291483055767e-007,-1.89775394143646e-006,-4.31981760329904e-006,-7.60504200188573e-006,-1.12290674717028e-005,-1.44442208796392e-005,-1.64551317961879e-005,-1.6615513352675e-005,-1.45974741955391e-005,-1.04912405632317e-005,-4.81635129901971e-006,-5.88564803133452e-007,-6.93762988830925e-006,-2.49216860143408e-005,-5.67287126102093e-005,-9.98709394081982e-005,-.000147462369938119,-.00018968441040975,-.0002160920965543,-.000218198259374459,-.000191696963742466,-.000137773078747959,-6.3249292855151e-005,-3.7246230587395e-006,-4.39035024137147e-005,-.000157712261925103,-.000358997123096203,-.000632014693772654,-.000933188224038868,-.00120038256642976,-.00136749870422512,-.00138082716451303,-.00121311863645989,-.000871871344066475,-.000400261404288981,-1.45714378978438e-005,-.000171758899848453,-.000617000537806862,-.00140446541897081,-.00247256238164128,-.00365081084424778,-.00469612622392589,-.00534991652302709,-.00540206000930644,-.00474595216619707,-.00341092748034312,-.00156589918054001,-3.98936826157651e-005,-.000470241515286636,-.00168922406982674,-.00384514541818346,-.00676938127811489,-.00999518991411039,-.0128570543562206,-.0146470014343651,-.0147897598708826,-.0129934678207622,-.00933843724141566,-.00428711876993237,-8.16359254002543e-005,-.000962272689432589,-.00345672199472992,-.00786846397549439,-.01385243649599,-.0204535285962103,-.0263098681663614,-.0299727033964202,-.0302648352907365,-.0265890161087102,-.0191095911936167,-.00877289047128288,-.000128821307499823,-.00151846414940855,-.0054546995681281,-.0124164185359414,-.0218591137752465,-.0322756223296346,-.0415169130590646,-.0472968588510568,-.0477578423262084,-.0419574078870825,-.0301548921174003,-.0138436015212964,-.000156894610627413,-.00184937450253303,-.00664341157096352,-.015122258804788,-.0266227475173882,-.0393092672046696,-.0505644604551389,-.0576039973305382,-.0581654403421276,-.0511009498439408,-.036726378183032,-.0168604597524974,-.000140407421995964,-.00165503394391524,-.00594529211811034,-.0135331440961878,-.0238251099283638,-.0351784732865376,-.0452509204027591,-.0515507112035252,-.0520531552679814,-.045731033082912,-.0328670058156056,-.0150886871005195,-7.26477900154669e-005,-.000856326230599691,-.00307613605632261,-.00700214416426768,-.0123272798443727,-.0182015900873001,-.0234131452361595,-.0266727014101101,-.0269326695112991,-.0236615589216733,-.0170056205219542,-.00780699308131397},
			{-3.66334363412211e-007,-2.24537546398006e-005,-.000193564800479007,-.000790234334926298,-.00213112724015142,-.00435422423400534,-.00720462946394863,-.00997720456040535,-.0116910253775248,-.0114525451470328,-.00885514027394949,-.00423058507690692,-8.30424254068273e-007,-5.08992448158822e-005,-.000438781946510387,-.00179134098152266,-.00483094114415894,-.00987036372424941,-.0163317986134588,-.0226168044340364,-.0265017754217613,-.0259611770305218,-.0200732554493939,-.00959009257018474,-1.2430688947403e-006,-7.61914981245153e-005,-.000656816303965783,-.00268147304597068,-.00723147552495667,-.0147750286257641,-.0244472036457215,-.0338552805420154,-.0396707255608121,-.0388614993834644,-.0300478211505447,-.0143554884304917,-1.57767194829697e-006,-9.670034364015e-005,-.000833614824033866,-.00340325851820642,-.00917800785523132,-.0187520967637297,-.03102777936875,-.0429682752411084,-.0503490925971213,-.0493220429740223,-.0381359430175126,-.0182196268418573,-1.81315792676965e-006,-.000111134000184102,-.000958041580001006,-.00391123463011006,-.0105479328022725,-.0215510663844978,-.035659038099294,-.0493817925438395,-.057864283158913,-.0566839343751764,-.0438281782544715,-.0209391191031562,-1.93472111511538e-006,-.000118584980154756,-.00102227348573477,-.00417346338855665,-.0112551190451086,-.0229959578102643,-.038049798606524,-.0526925952374313,-.0617437945066473,-.0604843092288456,-.0467666388316428,-.0223429825183335,-1.93472111511538e-006,-.000118584980154756,-.00102227348573477,-.00417346338855665,-.0112551190451086,-.0229959578102643,-.038049798606524,-.0526925952374313,-.0617437945066473,-.0604843092288456,-.0467666388316428,-.0223429825183335,-1.81315792676965e-006,-.000111134000184102,-.000958041580001006,-.00391123463011006,-.0105479328022725,-.0215510663844978,-.035659038099294,-.0493817925438395,-.057864283158913,-.0566839343751764,-.0438281782544715,-.0209391191031562,-1.57767194829697e-006,-9.670034364015e-005,-.000833614824033866,-.00340325851820642,-.00917800785523132,-.0187520967637297,-.03102777936875,-.0429682752411084,-.0503490925971213,-.0493220429740223,-.0381359430175126,-.0182196268418573,-1.2430688947403e-006,-7.61914981245153e-005,-.000656816303965783,-.00268147304597068,-.00723147552495667,-.0147750286257641,-.0244472036457215,-.0338552805420154,-.0396707255608121,-.0388614993834644,-.0300478211505447,-.0143554884304917,-8.30424254068273e-007,-5.08992448158822e-005,-.000438781946510386,-.00179134098152266,-.00483094114415894,-.00987036372424941,-.0163317986134588,-.0226168044340364,-.0265017754217613,-.0259611770305218,-.020073255449394,-.00959009257018474,-3.66334363412211e-007,-2.24537546398006e-005,-.000193564800479007,-.000790234334926298,-.00213112724015142,-.00435422423400534,-.00720462946394863,-.00997720456040535,-.0116910253775248,-.0114525451470328,-.00885514027394949,-.00423058507690693},
			{-5.84997849237861e-009,-3.58563091085661e-007,-3.09102839585264e-006,-1.26192198302079e-005,-3.40318838868514e-005,-6.95324290155738e-005,-.000115050433754236,-.000159325572268953,-.00018669350692448,-.000182885225860587,-.00014140737349095,-6.75580403638936e-005,-6.89558489033362e-008,-4.22651508264995e-006,-3.64350890004165e-005,-.000148747387195263,-.000401146336906179,-.000819604324242625,-.00135614179377857,-.0018780291418385,-.00220062505725282,-.00215573544715162,-.00166682074010411,-.000796331478757963,-2.4770649960978e-007,-1.51826896966895e-005,-.000130883869937062,-.000534337481072711,-.00144101706420306,-.00294422186735428,-.00487159743595245,-.00674634613725228,-.00790519061914258,-.00774393601400238,-.00598763321131698,-.00286062003831949,-5.63849124010232e-007,-3.45600389940658e-005,-.000297928215558894,-.00121630123191481,-.0032801570028024,-.00670187065504814,-.0110891153486868,-.0153565666050425,-.0179944200606735,-.0176273595758088,-.013629524241044,-.00651156956023967,-9.92656788923096e-007,-6.0842973540354e-005,-.000524502838069343,-.00214129919480704,-.0057747187659129,-.0117986481151247,-.0195223955581165,-.0270352465684925,-.0316791894770017,-.031032979229028,-.0239947872427992,-.0114636229006531,-1.46568685042074e-006,-8.98364341570052e-005,-.000774443817183375,-.00316169103729199,-.00852654155446687,-.0174210498412452,-.0288253893768114,-.0399183341468002,-.0467752519970179,-.0458211036210322,-.0354290068157069,-.0169263752901921,-1.88534841928859e-006,-.00011555877647658,-.00099618579926221,-.00406695959489997,-.0109678964760433,-.0224091174530978,-.0370787950245641,-.0513479179830939,-.0601681371359018,-.0589407930194454,-.045573187736373,-.0217728056224891,-2.1478248622506e-006,-.000131646761218445,-.00113487385418326,-.00463315790456963,-.012494836761645,-.0255288938184798,-.042240870176189,-.0584965270825159,-.0685446888933965,-.0671464750773758,-.0519178442938048,-.0248039952501605,-2.16875884799657e-006,-.000132929869292705,-.00114593501354709,-.004678315432652,-.012616618914028,-.0257777136867565,-.042652574961671,-.0590666692229003,-.0692127664287096,-.0678009247845504,-.0524238666569454,-.0250457498234202,-1.90535198329458e-006,-.000116784856153978,-.00100675533972143,-.0041101100734715,-.0110842659581639,-.0226468783946324,-.0374722013795649,-.0518927198634471,-.0608065216222948,-.0595661553735154,-.0460567196758888,-.0220038152896698,-1.36938115091738e-006,-8.39335630014821e-005,-.000723557535766244,-.00295394620634634,-.00796628917278541,-.0162763671345931,-.0269313631824755,-.0372954252410452,-.0437017964620066,-.042810342192032,-.0331010775700317,-.015814196101363,-6.28659751468171e-007,-3.85324807640305e-005,-.00033217304057603,-.00135610679808757,-.00365718877328773,-.00747220517149892,-.0123637338469658,-.0171216996431039,-.0200627564386384,-.0196535048439074,-.0151961454884669,-.00726003025826054},
			{-6.96297165792815e-011,-4.26781849550599e-009,-3.67911491336473e-008,-1.50201013794113e-007,-4.05066519951766e-007,-8.27613867594724e-007,-1.36939462342768e-006,-1.89638209018582e-006,-2.22213055847631e-006,-2.17680226684608e-006,-1.68310966462922e-006,-8.04113589360719e-007,-4.26781849550599e-009,-2.61587661208472e-007,-2.25504216385422e-006,-9.20627996502583e-006,-2.48277670898173e-005,-5.07269876251226e-005,-8.39343887728836e-005,-.00011623506394466,-.000136201184821656,-.000133422875058339,-.000103162944064127,-4.92865836277514e-005,-3.67911491336473e-008,-2.25504216385422e-006,-1.94398127850062e-005,-7.93636419908757e-005,-.000214030208318087,-.000437296986453802,-.000723564654409836,-.00100201580190209,-.00117413571083906,-.001150185018243,-.000889326306666056,-.000424879842112779,-1.50201013794113e-007,-9.20627996502583e-006,-7.93636419908757e-005,-.000324004543650437,-.000873785000712069,-.00178527858577817,-.00295397526843083,-.00409076076250558,-.00479344565888028,-.00469566620937377,-.0036307023836027,-.00173458520673544,-4.05066519951766e-007,-2.48277670898173e-005,-.000214030208318087,-.000873785000712069,-.00235644913761801,-.00481459189667545,-.00796636754827121,-.0110320841661877,-.0129271055006463,-.0126634109999641,-.0097913851734879,-.00467788049829846,-8.27613867594724e-007,-5.07269876251226e-005,-.000437296986453802,-.00178527858577817,-.00481459189667545,-.0098369596701605,-.016276527267894,-.0225402628819985,-.0264120860481604,-.025873317192124,-.0200053219740424,-.00955763703157627,-1.36939462342768e-006,-8.39343887728836e-005,-.000723564654409836,-.00295397526843083,-.00796636754827121,-.016276527267894,-.0269316281438179,-.037295792168106,-.0437022264174669,-.0428107633770146,-.0331014032314506,-.0158143516876436,-1.89638209018582e-006,-.00011623506394466,-.00100201580190209,-.00409076076250558,-.0110320841661877,-.0225402628819985,-.037295792168106,-.0516484226656698,-.0605202606038321,-.0592857336712341,-.0458398968231796,-.0219002271480226,-2.22213055847631e-006,-.000136201184821656,-.00117413571083906,-.00479344565888028,-.0129271055006463,-.0264120860481604,-.0437022264174669,-.0605202606038321,-.0709160465028163,-.0694694603763257,-.0537139830919874,-.0256621090417623,-2.17680226684608e-006,-.000133422875058339,-.001150185018243,-.00469566620937377,-.0126634109999641,-.025873317192124,-.0428107633770146,-.0592857336712341,-.0694694603763257,-.0680523825420278,-.0526182945056766,-.0251386386461752,-1.68310966462922e-006,-.000103162944064127,-.000889326306666056,-.0036307023836027,-.0097913851734879,-.0200053219740424,-.0331014032314506,-.0458398968231796,-.0537139830919874,-.0526182945056766,-.0406846140173892,-.0194372664460252,-8.04113589360719e-007,-4.92865836277514e-005,-.000424879842112779,-.00173458520673544,-.00467788049829846,-.00955763703157627,-.0158143516876436,-.0219002271480226,-.0256621090417623,-.0251386386461752,-.0194372664460252,-.00928624582089677},
			{-7.59581936873143e-013,-4.65571023163414e-011,-4.01350079989265e-010,-1.63852421901125e-009,-4.41882039598848e-009,-9.02833697182859e-009,-1.49385560003323e-008,-2.06873968742175e-008,-2.42409464882796e-008,-2.37464657802842e-008,-1.83608344516711e-008,-8.77197535304017e-009,-2.42091987612831e-010,-1.48385590679724e-008,-1.2791725800267e-007,-5.22226195326764e-007,-1.40835499192186e-006,-2.8774881763854e-006,-4.76117787775939e-006,-6.59343355166069e-006,-7.72601168100707e-006,-7.56841207045341e-006,-5.85191760211322e-006,-2.79578126521843e-006,-5.00828398075496e-009,-3.06973057681128e-007,-2.64629143836613e-006,-1.08035673306481e-005,-2.91353787244659e-005,-5.95281078102025e-005,-9.84969850916432e-005,-.000136401819657762,-.000159832057717645,-.000156571711876826,-.000121061690113565,-5.7837794064797e-005,-3.66709753307118e-008,-2.24767634356886e-006,-1.93763150067161e-005,-7.91044103306138e-005,-.00021333110473795,-.000435868609164376,-.000721201218685817,-.000998742839454559,-.00117030053969446,-.00114642807911561,-.0008864214307152,-.000423492023911399,-1.51493229393287e-007,-9.28548381512419e-006,-8.00464265713343e-005,-.000326792032995098,-.000881302384115301,-.00180063776864345,-.00297938903112802,-.00412595456537195,-.00483468482962708,-.00473606415990016,-.00366193819311915,-.00174950826221703,-4.28305814057175e-007,-2.62521745709628e-005,-.000226309453117529,-.000923915400575431,-.00249164227715486,-.00509081249643689,-.00842341040243774,-.011665011934605,-.0136687535803171,-.0133899305174026,-.0103531321176025,-.00494625775323145,-9.11601977606186e-007,-5.58748760112742e-005,-.000481674864645602,-.00196645265756991,-.00530318747211388,-.0108352363826531,-.0179283057316278,-.0248276992732258,-.0290924437312574,-.0284989994042999,-.0220355068809891,-.010527567456728,-1.53459036799021e-006,-9.40597416919936e-005,-.000810851255203673,-.00331032553849853,-.00892738345711226,-.0182400321589689,-.0301805019800266,-.0417949380321651,-.0489742069762943,-.0479752030574144,-.0370945625876605,-.0177221024244434,-2.08673606807239e-006,-.00012790244200432,-.00110259558209743,-.00450138085207253,-.0121394565234153,-.0248027967481441,-.0410394482774791,-.0568327590696347,-.0665951293806949,-.0652366837960255,-.0504411882777659,-.0240985158662193,-2.27930471413928e-006,-.000139705563856799,-.00120434555501086,-.00491677828990747,-.0132597125742255,-.027091654003041,-.044826659852073,-.0620774124945382,-.0727406760531766,-.0712568700882474,-.0550960133326791,-.02632238051474,-1.89600782307658e-006,-.000116212123967743,-.00100181804557465,-.00408995341613109,-.011029906890695,-.0225358143697108,-.0372884315268124,-.0516382294108676,-.0605083164164748,-.0592740331282503,-.0458308499302774,-.0218959049522221,-9.42668054679449e-007,-5.77790109816517e-005,-.000498089647453146,-.00203346652033204,-.00548391243189871,-.0112044855690737,-.0185392764637387,-.025673791359598,-.0300838721412506,-.0294702040898467,-.0227864450885946,-.0108863317310921},
			{-4.61452526590166e-009,-1.47072954118751e-006,-3.04257537549422e-005,-.000222779313164588,-.000920334332232957,-.00260199447163722,-.00553806002210424,-.00932276780442027,-.0126771001809598,-.0138469711844149,-.0115184097714994,-.00572678909872914,-1.0460426551641e-008,-3.33392005816974e-006,-6.89705536523267e-005,-.000505006800981909,-.00208625787714596,-.00589832550261673,-.0125539393028941,-.0211332958986598,-.0287370569429126,-.0313889764798934,-.0261104821110847,-.0129817593993148,-1.56582985243485e-008,-4.99057235089002e-006,-.000103242588927534,-.000755948833210871,-.0031229365722185,-.00882925194855102,-.018792094977277,-.0316346044160784,-.0430167368512079,-.0469864170136373,-.0390849953863391,-.019432502397701,-1.98731208257652e-008,-6.33390959846238e-006,-.000131032911457859,-.000959431350547771,-.00396355298338403,-.0112058657268432,-.023850456898108,-.0401498486478861,-.0545957664330435,-.0596339852015086,-.0496056985105609,-.0246632459775863,-2.2839416390577e-008,-7.2793196382121e-006,-.000150591105035483,-.00110263769366935,-.00455515959306486,-.0128784721632876,-.0274104163597275,-.0461426828391663,-.0627448227009384,-.0685350545084463,-.0570099288159739,-.028344523709382,-2.43706852532627e-008,-7.76736168416077e-006,-.000160687486930486,-.00117656404705184,-.00486056030603151,-.0137419094370501,-.0292481479535269,-.0492363193955517,-.0669515498630354,-.073129987810444,-.0608321599696494,-.0302448825382416,-2.43706852532627e-008,-7.76736168416077e-006,-.000160687486930486,-.00117656404705184,-.00486056030603151,-.0137419094370501,-.0292481479535269,-.0492363193955517,-.0669515498630354,-.073129987810444,-.0608321599696494,-.0302448825382416,-2.2839416390577e-008,-7.2793196382121e-006,-.000150591105035483,-.00110263769366935,-.00455515959306486,-.0128784721632876,-.0274104163597275,-.0461426828391663,-.0627448227009384,-.0685350545084463,-.0570099288159739,-.028344523709382,-1.98731208257652e-008,-6.33390959846238e-006,-.000131032911457859,-.000959431350547771,-.00396355298338403,-.0112058657268432,-.023850456898108,-.0401498486478861,-.0545957664330435,-.0596339852015086,-.0496056985105609,-.0246632459775863,-1.56582985243485e-008,-4.99057235089002e-006,-.000103242588927534,-.000755948833210871,-.0031229365722185,-.00882925194855102,-.018792094977277,-.0316346044160784,-.0430167368512079,-.0469864170136373,-.0390849953863391,-.019432502397701,-1.0460426551641e-008,-3.33392005816974e-006,-6.89705536523267e-005,-.000505006800981909,-.00208625787714596,-.00589832550261673,-.0125539393028941,-.0211332958986598,-.0287370569429126,-.0313889764798934,-.0261104821110847,-.0129817593993148,-4.61452526590166e-009,-1.47072954118751e-006,-3.04257537549422e-005,-.000222779313164588,-.000920334332232957,-.00260199447163722,-.00553806002210424,-.00932276780442027,-.0126771001809598,-.0138469711844149,-.0115184097714994,-.00572678909872914},
			{-7.36891655603157e-011,-2.34860199933996e-008,-4.85867619430457e-007,-3.55755375585554e-006,-1.46967813742169e-005,-4.15511434782526e-005,-8.84370543742305e-005,-.000148874898420579,-.000202440095184293,-.000221121717472685,-.0001839370153449,-9.14508612930534e-005,-8.68601307305661e-010,-2.76838358998326e-007,-5.72710039808159e-006,-4.19341950699251e-005,-.000173236369522755,-.000489778616311158,-.00104244009902437,-.00175484320401577,-.00238623588679712,-.00260644304233894,-.00216813327679371,-.00107796495006253,-3.12023117416446e-009,-9.94471998471375e-007,-2.05731640619964e-005,-.000150638021863614,-.000622307975060138,-.00175940617887584,-.00374470319905967,-.00630383171759815,-.00857195072154575,-.00936298940145173,-.00778847209081444,-.00387231726863855,-7.10251695871362e-009,-2.26369580965429e-006,-4.68302630441653e-005,-.000342894178409679,-.0014165466273796,-.00400489948507659,-.00852398956746932,-.0143492802872553,-.0195121521357594,-.0213127769373929,-.0177287361159355,-.00881447480486433,-1.25039862212867e-008,-3.98523810328614e-006,-8.24447118179819e-005,-.000603665448054148,-.00249383135774626,-.00705062842794049,-.0150064897727934,-.0252619182805468,-.0343511578883119,-.0375211591484582,-.0312114527009358,-.0155178892424445,-1.84625022332888e-008,-5.88432089415219e-006,-.000121732034018966,-.00089132973162475,-.00368221511100628,-.01041045957611,-.0221575220926239,-.0372999629412389,-.0507204916900297,-.055401091485105,-.0460846249345851,-.0229126184021882,-2.37487628354249e-008,-7.56914418191853e-006,-.000156586857450866,-.00114653897597652,-.00473652229125437,-.0133912257589351,-.0285017561839274,-.0479798438190263,-.0652429943039245,-.0712637629438524,-.0592797668540085,-.0294730548165372,-2.70550433775344e-008,-8.62291335307587e-006,-.000178386733239242,-.00130615905948616,-.00539593649305364,-.0152555396799836,-.0324697440130155,-.0546595527845674,-.074326062928209,-.0811850373450308,-.0675326405320358,-.03357626593249,-2.7318738012221e-008,-8.7069574240775e-006,-.000180125397036526,-.0013188896669085,-.00544852851748593,-.0154042292942165,-.0327862135586896,-.055192297478506,-.0750504891927709,-.0819763152766897,-.0681908540387014,-.0339035203026146,-2.40007374267426e-008,-7.64945287105235e-006,-.000158248245443382,-.00115870376501852,-.00478677683617009,-.0135333067869547,-.0288041600782291,-.0484889104016337,-.0659352230713762,-.0720198721217481,-.0599087257235696,-.0297857641982319,-1.72493889467441e-008,-5.49768056941236e-006,-.000113733402739168,-.000832763242290672,-.0034402682709391,-.00972642083254007,-.0207016206060418,-.0348490990193348,-.0473878067921877,-.0517608590117147,-.0430565483441448,-.0214071018984119,-7.91890304680574e-009,-2.52389226922161e-006,-5.22130837362085e-005,-.000382307535472911,-.00157936904180661,-.00446523548185483,-.00950376427809313,-.0159986326040109,-.0217549415081703,-.0237625359019177,-.0197665339288291,-.00982763882071387},
			{-8.77089671184217e-013,-2.79543748348947e-010,-5.78306820717772e-009,-4.23439949444061e-008,-1.74929313488118e-007,-4.94564953932593e-007,-1.05262729400979e-006,-1.77199232357178e-006,-2.40955526052686e-006,-2.63191437974773e-006,-2.18932125341295e-006,-1.08849930992072e-006,-5.37595110937315e-011,-1.71340921393568e-008,-3.5446195486468e-007,-2.5953930832331e-006,-1.07219531571797e-005,-3.03134001015112e-005,-6.45187493924991e-005,-.000108610834338575,-.000147689018597559,-.000161318089751033,-.000134190202070998,-6.67174550672731e-005,-4.63439153301422e-010,-1.47706126638825e-007,-3.05567414766182e-006,-2.23738413632665e-005,-9.24296518291858e-005,-.000261319647275767,-.000556190225361124,-.000936290380644839,-.00127316771187585,-.00139065846034338,-.0011567998362318,-.000575144383900619,-1.89200479726517e-009,-6.03014868716438e-007,-1.24748850093275e-005,-9.13419051691941e-005,-.000377346936322897,-.00106684561013735,-.00227066394170372,-.00382243468035404,-.00519774689175974,-.00567740653672345,-.00472267141011844,-.00234804488509053,-5.1024142886996e-009,-1.62622826690742e-006,-3.36426374888076e-005,-.000246333541419145,-.00101764033710382,-.00287710067800549,-.00612359342731638,-.0103084544811151,-.0140174369789623,-.0153109972435626,-.0127362394210907,-.00633227663561512,-1.04250255588739e-008,-3.32263714544313e-006,-6.87371381162648e-005,-.000503297717511878,-.00207919740024719,-.00587836392863835,-.0125114532807165,-.0210617749476202,-.0286398027496958,-.0312827474533958,-.0260221169776175,-.012937825514943,-1.7249558650954e-008,-5.49773465709859e-006,-.000113734521679514,-.000832771435243388,-.00344030211724,-.00972651652373048,-.0207018242742517,-.0348494418743096,-.0473882730064016,-.0517613682492133,-.0430569719462525,-.0214073125073398,-2.388774830107e-008,-7.61344126955836e-006,-.000157503254546897,-.00115324889407038,-.00476424195651651,-.0134695955569983,-.0286685577088058,-.0482606374326126,-.0656248174753565,-.071680821606866,-.0596266912835726,-.0296455406961615,-2.79910339523387e-008,-8.92122984487403e-006,-.000184558163040781,-.00135134666284035,-.00558261317396905,-.0157833171133724,-.0335930604290494,-.0565505431450506,-.076897431725972,-.0839936977751664,-.0698689855213075,-.0347378633474854,-2.74200568127716e-008,-8.73924948977462e-006,-.000180793439944228,-.00132378111976588,-.00546873583357464,-.0154613599726102,-.0329078099454216,-.0553969927824272,-.0753288338784716,-.0822803462291326,-.068443757944315,-.0340292605183584,-2.12012654199985e-008,-6.75721240365912e-006,-.000139789998708784,-.00102355130296691,-.00422844200182213,-.0119547672264766,-.0254444116512108,-.0428331113779068,-.0582444672431377,-.0636193962384749,-.0529209070727989,-.0263115204035591,-1.0129004659729e-008,-3.22829013115797e-006,-6.67853319250079e-005,-.000489006467861309,-.00202015813166747,-.00571144648888444,-.0121561878064161,-.0204637212045026,-.0278265692364139,-.0303944669425635,-.0252832132289538,-.0125704531070497},
			{-5.3934946534894e-011,-8.93862080339512e-008,-4.43762053833248e-006,-5.82757244194688e-005,-.000368787100047311,-.00144276568155165,-.00395000387691709,-.00808303973670482,-.0127550186066991,-.0155346480848768,-.0139021976630662,-.00719309508160456,-1.22262316118168e-010,-2.02624931058027e-007,-1.00594104551317e-005,-.000132102199014488,-.000835984234769121,-.00327053024384089,-.00895405768790711,-.0183230210276625,-.028913686156645,-.0352146831870672,-.0315141665027775,-.0163056519239563,-1.83015466396657e-010,-3.03310925545804e-007,-1.50580142318092e-005,-.00019774486802044,-.00125139167557278,-.00489568361654787,-.0134034025849063,-.0274278808519569,-.0432811345737185,-.0527131488435292,-.0471738149883672,-.0244080644511933,-2.32278652181071e-010,-3.84954639980414e-007,-1.91112550165974e-005,-.000250972840295157,-.00158823501355166,-.00621347919028916,-.017011263301183,-.0348107803231648,-.0549313334090684,-.0669022099971482,-.0598718260236465,-.0309781048820443,-2.6694895594493e-010,-4.42413619435074e-007,-2.19638332045296e-005,-.000288433470136975,-.00182529765297712,-.00714091358400096,-.0195503931804077,-.0400066961627327,-.0631304769703183,-.0768881468075034,-.0688083958532261,-.0356019491148408,-2.84846550926105e-010,-4.72075244245615e-007,-2.3436398585193e-005,-.000307771494551617,-.00194767474936783,-.00761967619488893,-.0208611494544942,-.0426889454411423,-.0673630603261273,-.0820431132524654,-.0734216553279677,-.0379888820906312,-2.84846550926105e-010,-4.72075244245615e-007,-2.3436398585193e-005,-.000307771494551617,-.00194767474936783,-.00761967619488893,-.0208611494544942,-.0426889454411423,-.0673630603261273,-.0820431132524654,-.0734216553279677,-.0379888820906312,-2.6694895594493e-010,-4.42413619435074e-007,-2.19638332045296e-005,-.000288433470136975,-.00182529765297712,-.00714091358400096,-.0195503931804077,-.0400066961627327,-.0631304769703183,-.0768881468075034,-.0688083958532261,-.0356019491148408,-2.32278652181071e-010,-3.84954639980414e-007,-1.91112550165974e-005,-.000250972840295156,-.00158823501355166,-.00621347919028916,-.017011263301183,-.0348107803231648,-.0549313334090684,-.0669022099971482,-.0598718260236465,-.0309781048820443,-1.83015466396657e-010,-3.03310925545804e-007,-1.50580142318092e-005,-.00019774486802044,-.00125139167557278,-.00489568361654787,-.0134034025849063,-.0274278808519569,-.0432811345737185,-.0527131488435292,-.0471738149883672,-.0244080644511933,-1.22262316118168e-010,-2.02624931058027e-007,-1.00594104551317e-005,-.000132102199014488,-.00083598423476912,-.00327053024384089,-.00895405768790711,-.0183230210276625,-.028913686156645,-.0352146831870672,-.0315141665027775,-.0163056519239563,-5.3934946534894e-011,-8.93862080339512e-008,-4.43762053833248e-006,-5.82757244194688e-005,-.000368787100047311,-.00144276568155165,-.00395000387691709,-.00808303973670482,-.0127550186066991,-.0155346480848768,-.0139021976630662,-.00719309508160456},
			{-8.61284959198155e-013,-1.42740470669776e-009,-7.08641812006219e-008,-9.30602664014744e-007,-5.88914614407857e-006,-2.30394662644891e-005,-6.30774506425453e-005,-.000129077731546852,-.000203684371377566,-.000248072161029854,-.000222003626892483,-.000114866241683583,-1.015228271122e-011,-1.68253444704445e-008,-8.35302177246445e-007,-1.09693560023246e-005,-6.94175324250859e-005,-.000271574665892772,-.000743517119145248,-.00152148671399835,-.00240090262810037,-.00292411783656719,-.00261683843315572,-.00135397065407101,-3.64695156892415e-011,-6.04408074120732e-008,-3.00061244597533e-006,-3.94046454582605e-005,-.000249364981245768,-.00097556350825723,-.00267089776882591,-.00546555736926158,-.0086246372913827,-.0105041559963674,-.00940033222170924,-.00486379816401721,-8.30147957637377e-011,-1.37580145726718e-007,-6.83023134969281e-006,-8.96959702655164e-005,-.000567624291069279,-.00222065480887151,-.00607970872643953,-.0124411339215587,-.0196320814726822,-.0239103905886559,-.0213977796180793,-.0110713620291109,-1.46147607732105e-010,-2.42209945641729e-007,-1.20246272104963e-005,-.000157909820254493,-.00099930297324955,-.0039094643904085,-.0107033315917073,-.0219026252294268,-.0345622935723416,-.0420942598523953,-.0376708064290284,-.0194911408262166,-2.15791227404772e-010,-3.57630085574026e-007,-1.77547146005579e-005,-.000233158479025165,-.00147550013642398,-.00577243878564134,-.015803782883183,-.0323398682673508,-.0510322397173339,-.0621534087434013,-.0556220500820546,-.0287792408488707,-2.77577471181613e-010,-4.60028222490692e-007,-2.28383184972103e-005,-.000299917386683,-.00189797148624789,-.00742522752170763,-.0203287878778718,-.041599554161483,-.0656440033259579,-.0799494318272223,-.0715479873273676,-.0370194330577289,-3.16221547012239e-010,-5.24072921213834e-007,-2.60178478303842e-005,-.000341671604648038,-.00216220530078157,-.00845896075008184,-.0231589427061906,-.0473910051704331,-.0747829000510975,-.0910799169238503,-.0815088311812927,-.0421732438918713,-3.19303631345124e-010,-5.29180849357959e-007,-2.62714333369098e-005,-.00034500174046461,-.00218327944688257,-.00854140684095277,-.0233846636134322,-.0478529062519234,-.0755117789234986,-.0919676362701244,-.0823032650013445,-.0425842895511383,-2.80522570694571e-010,-4.64909126147018e-007,-2.30806332657561e-005,-.000303099512904151,-.00191810898111,-.00750400925375776,-.0205444765035491,-.0420409258123548,-.0663404868028083,-.0807976960315661,-.072307112128577,-.0374122095817176,-2.01612260665772e-010,-3.34131331017576e-007,-1.65881078259945e-005,-.000217838364492369,-.00137854963658506,-.00539314988437847,-.0147653657308921,-.0302149166554566,-.0476790708315039,-.058069502618547,-.0519672991102317,-.026888246929992,-9.25567827468422e-011,-1.53394049110811e-007,-7.61532005623914e-006,-.00010000591288281,-.000632868848341838,-.0024759039977092,-.00677852995457166,-.0138711577726375,-.0218886559078795,-.0266587276008582,-.0238573090634585,-.0123439399038752},
			{-6.02215524056038e-013,-5.18975291864507e-009,-6.18297441623807e-007,-1.45626016852063e-005,-.000141170609267733,-.00076422934021594,-.00269138517084675,-.00669488358667141,-.0122597253338038,-.0166489373902391,-.016029237481894,-.00863095559575211,-1.36513094948138e-012,-1.1764380103799e-008,-1.40158620930416e-006,-3.30112019224449e-005,-.000320012288242827,-.00173239161588508,-.00610096061448634,-.0151762822814405,-.0277909316794019,-.0377406074727463,-.0363358420848087,-.0195650628997474,-2.0434757441563e-012,-1.76101973193715e-008,-2.09804592236745e-006,-4.94147396186346e-005,-.000479029025826867,-.00259323125578963,-.0091325781138374,-.0227175017461691,-.0416004741639895,-.0564942256781886,-.0543914208171814,-.0292871035439603,-2.59352830098303e-012,-2.23504219535243e-008,-2.66278740620364e-006,-6.27159515120887e-005,-.00060797165761674,-.0032912642453023,-.0115908397087212,-.0288324849829067,-.0527982812554331,-.0717010581395466,-.0690322308074567,-.0371704592590807,-2.98064271365472e-012,-2.56864836669144e-008,-3.06023954984572e-006,-7.20770403136054e-005,-.00069871853362736,-.00378252390296041,-.0133209080116245,-.0331360703672624,-.0606790418511532,-.0824032791251857,-.0793360981199193,-.0427185847602993,-3.18047992928925e-012,-2.7408634178924e-008,-3.26541333602756e-006,-7.69094460834964e-005,-.000745564123550859,-.00403612324963012,-.0142140084005345,-.0353576784818238,-.0647472687189207,-.087928007662455,-.0846551874810701,-.0455826526323854,-3.18047992928925e-012,-2.7408634178924e-008,-3.26541333602756e-006,-7.69094460834964e-005,-.000745564123550859,-.00403612324963012,-.0142140084005345,-.0353576784818238,-.0647472687189207,-.087928007662455,-.0846551874810701,-.0455826526323854,-2.98064271365472e-012,-2.56864836669144e-008,-3.06023954984572e-006,-7.20770403136054e-005,-.00069871853362736,-.00378252390296041,-.0133209080116245,-.0331360703672624,-.0606790418511532,-.0824032791251857,-.0793360981199193,-.0427185847602993,-2.59352830098303e-012,-2.23504219535243e-008,-2.66278740620364e-006,-6.27159515120887e-005,-.00060797165761674,-.0032912642453023,-.0115908397087212,-.0288324849829067,-.0527982812554331,-.0717010581395466,-.0690322308074567,-.0371704592590807,-2.0434757441563e-012,-1.76101973193715e-008,-2.09804592236745e-006,-4.94147396186346e-005,-.000479029025826867,-.00259323125578962,-.0091325781138374,-.0227175017461691,-.0416004741639895,-.0564942256781886,-.0543914208171814,-.0292871035439603,-1.36513094948138e-012,-1.1764380103799e-008,-1.40158620930416e-006,-3.30112019224449e-005,-.000320012288242827,-.00173239161588507,-.00610096061448634,-.0151762822814405,-.0277909316794019,-.0377406074727463,-.0363358420848087,-.0195650628997474,-6.02215524056038e-013,-5.18975291864507e-009,-6.18297441623807e-007,-1.45626016852063e-005,-.000141170609267733,-.00076422934021594,-.00269138517084675,-.00669488358667141,-.0122597253338038,-.0166489373902391,-.016029237481894,-.00863095559575211},
			{-.00270055519884434,-.00588249745219549,-.00818488569070588,-.00931641156972225,-.00922648473106781,-.00809895569654666,-.00629620113986558,-.00426418972382729,-.00242214596950661,-.00106408128362993,-.00029621599889564,-2.51299527189796e-005,-.00612174767248644,-.0133347265413407,-.0185538903067371,-.0211188872818742,-.0209150368234713,-.0183591000865756,-.014272529844964,-.00966626919088759,-.00549063631715242,-.00241211034093489,-.000671476592135066,-5.69657786046887e-005,-.00916369443189591,-.0199608618151533,-.0277734709743103,-.0316130360391258,-.0313078906116195,-.0274818878919328,-.0213646672921825,-.0144695178404447,-.00821897865428013,-.00361070780484583,-.00100513883251861,-8.52725424398906e-005,-.011630331761288,-.025333826534352,-.035249394662203,-.0401224746033481,-.0397351916593658,-.034879324707423,-.0271155013323826,-.0183643501167498,-.0104313221265329,-.00458262003120292,-.00127569706467466,-.000108225777940879,-.013366294714629,-.0291151962521434,-.0405107788184967,-.0461112228985269,-.045666133448504,-.0400854715975038,-.0311628068384012,-.0211054440183851,-.0119883188775888,-.00526662963356749,-.00146610976221544,-.000124379740266118,-.014262438055421,-.03106722482807,-.0432268243227726,-.0492027502229294,-.0487278197470118,-.0427730023755956,-.0332521174831873,-.0225204586888932,-.0127920758168761,-.00561973086134261,-.00156440510346971,-.000132718780991214,-.014262438055421,-.03106722482807,-.0432268243227726,-.0492027502229294,-.0487278197470118,-.0427730023755956,-.0332521174831873,-.0225204586888932,-.0127920758168761,-.00561973086134261,-.00156440510346971,-.000132718780991214,-.013366294714629,-.0291151962521434,-.0405107788184967,-.0461112228985269,-.045666133448504,-.0400854715975038,-.0311628068384012,-.0211054440183851,-.0119883188775888,-.00526662963356749,-.00146610976221544,-.000124379740266118,-.011630331761288,-.025333826534352,-.035249394662203,-.0401224746033481,-.0397351916593658,-.034879324707423,-.0271155013323826,-.0183643501167498,-.0104313221265329,-.00458262003120292,-.00127569706467466,-.000108225777940879,-.00916369443189591,-.0199608618151533,-.0277734709743103,-.0316130360391258,-.0313078906116195,-.0274818878919328,-.0213646672921825,-.0144695178404447,-.00821897865428013,-.00361070780484583,-.00100513883251861,-8.52725424398906e-005,-.00612174767248644,-.0133347265413407,-.0185538903067371,-.0211188872818742,-.0209150368234713,-.0183591000865756,-.014272529844964,-.00966626919088758,-.00549063631715242,-.00241211034093489,-.000671476592135066,-5.69657786046887e-005,-.00270055519884434,-.00588249745219549,-.00818488569070588,-.00931641156972225,-.00922648473106781,-.00809895569654666,-.00629620113986558,-.00426418972382729,-.00242214596950661,-.00106408128362993,-.00029621599889564,-2.51299527189796e-005},
			{-4.31250557013781e-005,-9.39373616201976e-005,-.000130704105389941,-.000148773395949532,-.000147337357934948,-.000129331892821903,-.000100543778916268,-6.80946715845654e-005,-3.86791500860789e-005,-1.69922705697598e-005,-4.73026119128408e-006,-4.01299188864111e-007,-.000508330898783026,-.00110727424428985,-.00154065738088721,-.00175364675704319,-.00173671964859619,-.00152448260646228,-.00118514651569285,-.000802656948401598,-.000455925373489819,-.000200294145262156,-5.5757329091799e-005,-4.73026119128408e-006,-.00182605057560164,-.00397760351765945,-.00553442315607657,-.00629953378353948,-.00623872741470196,-.00547631935750839,-.00425734001677702,-.00288334269301423,-.00163779694031119,-.000719506211645245,-.000200294145262156,-1.69922705697598e-005,-.0041566007314033,-.0090541356913339,-.0125978916716827,-.0143394969898535,-.0142010847243002,-.0124656312103062,-.00969089403328137,-.0065632926638522,-.00372808291892168,-.00163779694031119,-.000455925373489819,-3.86791500860789e-005,-.00731769858135901,-.015939812405686,-.0221785973614222,-.0252446947784247,-.0250010198852289,-.0219457527239106,-.0170608259252993,-.0115546814618157,-.0065632926638522,-.00288334269301423,-.000802656948401598,-6.80946715845654e-005,-.0108047964872896,-.0235356003221748,-.0327473492108772,-.0372745319354167,-.0369147388118314,-.0324035472773186,-.0251908096484753,-.0170608259252993,-.00969089403328137,-.00425734001677702,-.00118514651569285,-.000100543778916268,-.0138984708583546,-.0302744115168138,-.0421237067473831,-.0479471313016055,-.0474843207110434,-.0416814660110368,-.0324035472773186,-.0219457527239106,-.0124656312103062,-.00547631935750839,-.00152448260646228,-.000129331892821903,-.0158334029675552,-.0344891867628131,-.0479881297889341,-.0546222860610778,-.0540950434131128,-.0474843207110434,-.0369147388118314,-.0250010198852289,-.0142010847243002,-.00623872741470196,-.00173671964859619,-.000147337357934948,-.0159877247830154,-.0348253390053783,-.0484558508041033,-.0551546675312396,-.0546222860610777,-.0479471313016055,-.0372745319354167,-.0252446947784246,-.0143394969898535,-.00629953378353948,-.00175364675704319,-.000148773395949532,-.0140459337615274,-.0305956233004421,-.0425706396619946,-.0484558508041033,-.0479881297889341,-.0421237067473831,-.0327473492108772,-.0221785973614222,-.0125978916716827,-.00553442315607657,-.00154065738088721,-.000130704105389941,-.0100948470984408,-.0219891496246005,-.0305956233004421,-.0348253390053783,-.0344891867628131,-.0302744115168138,-.0235356003221748,-.015939812405686,-.0090541356913339,-.00397760351765944,-.00110727424428985,-9.39373616201976e-005,-.00463437375617701,-.0100948470984408,-.0140459337615274,-.0159877247830154,-.0158334029675552,-.0138984708583546,-.0108047964872896,-.00731769858135901,-.0041566007314033,-.00182605057560164,-.000508330898783026,-4.31250557013781e-005},
			{-5.13298537640478e-007,-1.11809502771057e-006,-1.55571337982341e-006,-1.77078418424541e-006,-1.75369165645863e-006,-1.53938053820615e-006,-1.19672830207889e-006,-8.10500874190154e-007,-4.60380881913306e-007,-2.02251510004898e-007,-5.63022148645256e-008,-4.77648743752639e-009,-3.14616387984846e-005,-6.85314672157709e-005,-9.53544357538297e-005,-.00010853678378063,-.000107489130425036,-9.43533458892899e-005,-7.33511413319109e-005,-4.96780798691598e-005,-2.82181536753749e-005,-1.23966142266297e-005,-3.45093511422209e-006,-2.92765537918228e-007,-.000271218151906612,-.000590782254057447,-.000822012292713028,-.000935652020526277,-.000926620621724616,-.000813382299066098,-.000632330728850647,-.000428254773970375,-.00024325736936566,-.000106866232302167,-2.97491255946283e-005,-2.52381411676342e-006,-.00110725656401101,-.00241188697772547,-.00335589082222505,-.0038198285552606,-.0037829575879782,-.00332065860400601,-.00258150992189795,-.0017483634713027,-.000993105797236321,-.000436284726357267,-.000121451733066827,-1.03035498453369e-005,-.00298608212919577,-.00650444787225884,-.00905026525693164,-.0103014262061688,-.0102019914951301,-.00895524997265581,-.00696189202635823,-.00471503812818561,-.00267823697769347,-.00117658550597994,-.000327534612535476,-2.7786916836135e-005,-.00610102997451746,-.0132895981153155,-.0184910987778976,-.021047415089364,-.0208442545176493,-.0182969677820565,-.0142242276315438,-.00963355584556567,-.00547205447566498,-.00240394708818667,-.000669204128457476,-5.67729905547983e-005,-.0100949464135448,-.0219893659582073,-.0305959243062406,-.0348256816239572,-.0344895260742596,-.0302747093624672,-.0235358318700697,-.0159399692247158,-.00905422476771311,-.00397764265011023,-.00110728513787305,-9.39382857945563e-005,-.0139798092182615,-.0304515871935496,-.0423702283435318,-.0482277334672722,-.0477622143590056,-.0419253994709405,-.0325931833471828,-.0220741863877053,-.0125385841278022,-.00550836855481976,-.0015334043731933,-.000130088785011913,-.0163811720361289,-.0356823673916168,-.0496483170026962,-.0565119871455732,-.0559665041221956,-.0491270782522392,-.0381918333275861,-.0258659498945255,-.0146923824553127,-.00645456111210986,-.00179680283515025,-.000152434610085198,-.0160470195082919,-.0349544980281293,-.04863556220147,-.0553592232703062,-.0548248673220092,-.0481249559775311,-.0374127744408963,-.0253383214364947,-.0143926788244271,-.00632289727835408,-.00176015062198346,-.000149325161617318,-.0124076008346097,-.027026916660946,-.0376051541440983,-.0428039078843937,-.0423907423550166,-.0372103519687145,-.0289276629306809,-.0195916617438277,-.0111284599424805,-.0048888820448881,-.00136095343531401,-.000115458636973295,-.00592778988329973,-.0129122370468795,-.0179660399513751,-.0204497690975875,-.0202523773150973,-.0177774213479609,-.0138203275519371,-.00936000890342516,-.00531667428240711,-.00233568648062049,-.000650201929690796,-5.51609089712794e-005},
			{-5.59951004436151e-009,-1.21971598964522e-008,-1.6971084189159e-008,-1.9317264903302e-008,-1.91308046389391e-008,-1.67929112469387e-008,-1.30549605277777e-008,-8.84165344957659e-009,-5.02223790536332e-009,-2.20633662228147e-009,-6.14193874587905e-010,-5.21061086714588e-011,-1.78466133868042e-006,-3.88744721171111e-006,-5.40897999788056e-006,-6.15674863851456e-006,-6.09732050598647e-006,-5.35219317920171e-006,-4.16084320723466e-006,-2.81798888768115e-006,-1.60067465766628e-006,-7.03197893870279e-007,-1.95754280954478e-007,-1.66071240016209e-008,-3.69202255792134e-005,-8.04216603301225e-005,-.000111898407472009,-.00012736788971769,-.000126138468756338,-.000110723628756234,-8.60775467486887e-005,-5.82972148036964e-005,-3.31139965656822e-005,-1.45474238197567e-005,-4.04967152831355e-006,-3.43560288483173e-007,-.00027033225085123,-.000588852535789363,-.000819327289690437,-.000932595827175549,-.000923593928337934,-.000810725484866511,-.000630265297550785,-.000426855931918081,-.00024246279880047,-.000106517166772019,-2.96519536996954e-005,-2.51557038538368e-006,-.0011167825595386,-.0024326370236585,-.00338476236112819,-.00385269146248218,-.00381550328536708,-.00334922703163026,-.00260371927496907,-.00176340506432601,-.00100164972615974,-.000440038189183426,-.000122496611646568,-1.03921937720791e-005,-.00315739828904347,-.0068776181278623,-.00956949300162804,-.0108924349936834,-.0107872955591597,-.00946902654323017,-.00736130656207001,-.00498554717338517,-.00283189158407368,-.0012440880400349,-.000346325780898917,-2.93810953216598e-005,-.00672017616832581,-.0146382563131457,-.0203676169191767,-.0231833539384622,-.0229595761766608,-.0201537850748433,-.0156677341271221,-.0106111906809188,-.00602737082635835,-.00264790502577821,-.000737116463054399,-6.253444086072e-005,-.0113127416048293,-.0246420342366014,-.0342868373598303,-.039026847819149,-.0386501404333965,-.0339268729271671,-.026375056735671,-.0178628737232513,-.010146473397001,-.00445748215529598,-.00124086152956873,-.00010527045022929,-.0153830666658602,-.0335082395307456,-.0466232433562013,-.0530687098434273,-.0525564631192487,-.0461337637005683,-.0358648036217386,-.0242899366861979,-.0137971750917403,-.00606128448364179,-.00168732357718739,-.000143146764099252,-.0168026502756542,-.0366004544098804,-.0509257399611718,-.0579660084330403,-.0574064904417204,-.0503910900341255,-.039174487477097,-.0265314660737799,-.0150704090994405,-.00662063329842476,-.00184303355014811,-.000156356665903898,-.0139770501826433,-.0304455773108406,-.0423618662144552,-.0482182153091671,-.0477527880751015,-.0419171251326689,-.0325867508020486,-.0220698298571165,-.0125361095303546,-.00550728143089673,-.0015331017426483,-.000130063110871032,-.00694919005368325,-.015137106919017,-.0210617158489217,-.0239734091138143,-.0237420053295095,-.0208405969245687,-.0162016678481028,-.0109728047138306,-.00623277490756985,-.0027381417997588,-.000762236326125946,-6.46655241703613e-005},
			{-5.85380492906022e-011,-1.27510789616824e-010,-1.77418051741924e-010,-2.01945348094836e-010,-1.99996066807805e-010,-1.75555407262082e-010,-1.36478355571743e-010,-9.24318630275365e-011,-5.25031668349854e-011,-2.30653469542085e-011,-6.4208673651397e-012,-5.44724437242996e-013,-9.70149149662404e-008,-2.11323208782435e-007,-2.94034348800585e-007,-3.34683355709183e-007,-3.31452818296366e-007,-2.9094739428102e-007,-2.26185126101403e-007,-1.53187020074262e-007,-8.70133242853409e-008,-3.82261230182231e-008,-1.06412821914539e-008,-9.02770003435125e-010,-4.81635129901971e-006,-1.04912405632317e-005,-1.45974741955391e-005,-1.6615513352675e-005,-1.64551317961879e-005,-1.44442208796392e-005,-1.12290674717028e-005,-7.60504200188573e-006,-4.31981760329904e-006,-1.89775394143646e-006,-5.28291483055767e-007,-4.48184434349486e-008,-6.3249292855151e-005,-.000137773078747959,-.000191696963742466,-.000218198259374459,-.0002160920965543,-.00018968441040975,-.000147462369938119,-9.98709394081982e-005,-5.67287126102093e-005,-2.49216860143408e-005,-6.93762988830925e-006,-5.88564803133452e-007,-.000400261404288981,-.000871871344066475,-.00121311863645989,-.00138082716451303,-.00136749870422512,-.00120038256642976,-.000933188224038868,-.000632014693772654,-.000358997123096203,-.000157712261925103,-4.39035024137147e-005,-3.7246230587395e-006,-.00156589918054001,-.00341092748034313,-.00474595216619707,-.00540206000930644,-.0053499165230271,-.00469612622392589,-.00365081084424778,-.00247256238164128,-.00140446541897081,-.000617000537806862,-.000171758899848453,-1.45714378978438e-005,-.00428711876993237,-.00933843724141566,-.0129934678207622,-.0147897598708826,-.0146470014343651,-.0128570543562206,-.00999518991411039,-.00676938127811489,-.00384514541818346,-.00168922406982674,-.000470241515286636,-3.98936826157651e-005,-.00877289047128288,-.0191095911936167,-.0265890161087102,-.0302648352907365,-.0299727033964202,-.0263098681663614,-.0204535285962103,-.01385243649599,-.00786846397549439,-.00345672199472992,-.000962272689432589,-8.16359254002543e-005,-.0138436015212964,-.0301548921174003,-.0419574078870825,-.0477578423262084,-.0472968588510568,-.0415169130590646,-.0322756223296346,-.0218591137752465,-.0124164185359414,-.0054546995681281,-.00151846414940855,-.000128821307499823,-.0168604597524974,-.036726378183032,-.0511009498439408,-.0581654403421276,-.0576039973305382,-.0505644604551389,-.0393092672046696,-.0266227475173882,-.015122258804788,-.00664341157096352,-.00184937450253303,-.000156894610627413,-.0150886871005195,-.0328670058156056,-.045731033082912,-.0520531552679814,-.0515507112035252,-.0452509204027591,-.0351784732865376,-.0238251099283638,-.0135331440961878,-.00594529211811034,-.00165503394391524,-.000140407421995964,-.00780699308131397,-.0170056205219542,-.0236615589216733,-.0269326695112991,-.0266727014101101,-.0234131452361595,-.0182015900873001,-.0123272798443727,-.00700214416426768,-.00307613605632261,-.000856326230599691,-7.26477900154669e-005},
			{-9.64305560601928e-005,-.00109223938839564,-.0036470340953271,-.00744526010016809,-.0112949614634506,-.0137194545508608,-.0137194545508608,-.0112949614634506,-.00744526010016809,-.0036470340953271,-.00109223938839564,-9.64305560601928e-005,-.000218593396043406,-.00247594047941341,-.00826727129817659,-.0168772716472393,-.025603958800506,-.0310999156767677,-.0310999156767677,-.025603958800506,-.0168772716472393,-.00826727129817659,-.00247594047941341,-.000218593396043406,-.000327214252095852,-.00370625566402853,-.0123753463904701,-.0252637267155745,-.0383267764774421,-.0465537195204054,-.0465537195204054,-.0383267764774421,-.0252637267155745,-.0123753463904701,-.00370625566402854,-.000327214252095852,-.000415292144143345,-.00470388698413707,-.0157064800939949,-.0320640900252994,-.0486433860367337,-.0590848163661698,-.0590848163661698,-.0486433860367337,-.0320640900252994,-.0157064800939949,-.00470388698413707,-.000415292144143345,-.000477279350685986,-.00540599709662285,-.0180508558289433,-.0368500302339686,-.0559039799577005,-.0679039158056228,-.0679039158056228,-.0559039799577005,-.0368500302339686,-.0180508558289433,-.00540599709662285,-.000477279350685986,-.000509278548739481,-.00576844221712266,-.019261075608775,-.0393206408188163,-.059652062760935,-.0724565344080177,-.0724565344080177,-.059652062760935,-.0393206408188163,-.019261075608775,-.00576844221712266,-.000509278548739481,-.000509278548739481,-.00576844221712266,-.019261075608775,-.0393206408188163,-.059652062760935,-.0724565344080177,-.0724565344080177,-.059652062760935,-.0393206408188163,-.019261075608775,-.00576844221712266,-.000509278548739481,-.000477279350685986,-.00540599709662285,-.0180508558289433,-.0368500302339686,-.0559039799577005,-.0679039158056228,-.0679039158056228,-.0559039799577005,-.0368500302339686,-.0180508558289433,-.00540599709662285,-.000477279350685986,-.000415292144143345,-.00470388698413707,-.0157064800939949,-.0320640900252994,-.0486433860367337,-.0590848163661699,-.0590848163661699,-.0486433860367337,-.0320640900252994,-.0157064800939949,-.00470388698413707,-.000415292144143345,-.000327214252095852,-.00370625566402853,-.0123753463904701,-.0252637267155745,-.0383267764774421,-.0465537195204054,-.0465537195204054,-.0383267764774421,-.0252637267155745,-.0123753463904701,-.00370625566402854,-.000327214252095852,-.000218593396043406,-.00247594047941341,-.0082672712981766,-.0168772716472393,-.025603958800506,-.0310999156767677,-.0310999156767677,-.025603958800506,-.0168772716472393,-.0082672712981766,-.00247594047941341,-.000218593396043406,-9.64305560601928e-005,-.00109223938839564,-.0036470340953271,-.00744526010016809,-.0112949614634506,-.0137194545508608,-.0137194545508608,-.0112949614634506,-.00744526010016809,-.0036470340953271,-.00109223938839564,-9.64305560601928e-005},
			{-1.53989561292143e-006,-1.74419262023187e-005,-5.82393385771156e-005,-.000118893054584807,-.000180368778489956,-.000219085409622217,-.000219085409622217,-.000180368778489956,-.000118893054584807,-5.82393385771156e-005,-1.74419262023187e-005,-1.53989561292143e-006,-1.81513161714798e-005,-.000205594401647313,-.000686488512117749,-.00140143617934442,-.00212607315610414,-.00258244033247591,-.00258244033247591,-.00212607315610414,-.00140143617934442,-.000686488512117749,-.000205594401647313,-1.81513161714798e-005,-6.52040263973911e-005,-.000738546046221744,-.00246603687813908,-.00503430609488318,-.00763738171291507,-.00927676571867675,-.00927676571867675,-.00763738171291507,-.00503430609488318,-.00246603687813908,-.000738546046221744,-6.52040263973911e-005,-.000148422561475078,-.00168113691751877,-.00561338816585801,-.0114594856657818,-.0173848122489431,-.0211165077717546,-.0211165077717546,-.0173848122489431,-.0114594856657818,-.005613388165858,-.00168113691751877,-.000148422561475078,-.000261298026375791,-.00295964275410309,-.00988237390894377,-.0201744327681154,-.0306059744853902,-.0371756271409923,-.0371756271409923,-.0306059744853902,-.0201744327681154,-.00988237390894377,-.00295964275410309,-.000261298026375791,-.000385814196380375,-.0043699992938526,-.0145916147966848,-.0297881414330556,-.0451906185986146,-.0548909033461696,-.0548909033461696,-.0451906185986146,-.0297881414330556,-.0145916147966848,-.0043699992938526,-.000385814196380375,-.000496282125391267,-.0056212357084272,-.0187695467718091,-.0383172062628759,-.0581297848971723,-.0706074956102091,-.070607495610209,-.0581297848971723,-.0383172062628759,-.0187695467718091,-.0056212357084272,-.000496282125391267,-.00056537405855632,-.00640381888440892,-.0213826255122005,-.0436516918684151,-.066222559163116,-.080437405087215,-.080437405087215,-.066222559163116,-.0436516918684151,-.0213826255122005,-.00640381888440892,-.00056537405855632,-.000570884532287664,-.00646623433345328,-.0215910333696339,-.0440771473659381,-.0668680038331856,-.0812213961477219,-.0812213961477219,-.0668680038331856,-.0440771473659381,-.0215910333696339,-.00646623433345328,-.000570884532287664,-.0005015476832896,-.00568087707080664,-.0189686918350622,-.038723752172464,-.0587465424476329,-.0713566417856807,-.0713566417856807,-.0587465424476329,-.038723752172464,-.0189686918350622,-.00568087707080664,-.0005015476832896,-.000360463552039073,-.00408286030594201,-.0136328454187137,-.027830855811784,-.0422212843688097,-.0512841937399486,-.0512841937399486,-.0422212843688097,-.027830855811784,-.0136328454187137,-.00408286030594202,-.000360463552039073,-.000165482726913838,-.00187437218884842,-.00625860901253902,-.0127766757166634,-.0193830783490639,-.0235437069286394,-.0235437069286394,-.0193830783490639,-.0127766757166634,-.00625860901253902,-.00187437218884842,-.000165482726913838},
			{-1.83286990233606e-008,-2.076035629087e-007,-6.93197187616125e-007,-1.41513164604626e-006,-2.14685010231423e-006,-2.6076771046559e-006,-2.6076771046559e-006,-2.14685010231423e-006,-1.41513164604626e-006,-6.93197187616125e-007,-2.076035629087e-007,-1.83286990233606e-008,-1.1234220751336e-006,-1.27246579340282e-005,-4.24881778022507e-005,-8.67377509097766e-005,-.000131587015197788,-.000159832512959983,-.000159832512959983,-.000131587015197788,-8.67377509097766e-005,-4.24881778022507e-005,-1.27246579340282e-005,-1.1234220751336e-006,-9.68457050125128e-006,-.000109694165348983,-.00036627351598611,-.000747731313456525,-.00113435880773542,-.00137785189949114,-.00137785189949114,-.00113435880773542,-.000747731313456525,-.00036627351598611,-.000109694165348983,-9.68457050125128e-006,-3.95375611173333e-005,-.000447829851219516,-.00149532305248749,-.00305263640770766,-.00463105521138285,-.00562512335262842,-.00562512335262842,-.00463105521138285,-.00305263640770766,-.00149532305248749,-.000447829851219516,-3.95375611173333e-005,-.000106626059868886,-.001207720738907,-.00403263127041889,-.00823243981590666,-.0124891661566992,-.015169998412074,-.015169998412074,-.0124891661566992,-.00823243981590666,-.00403263127041889,-.001207720738907,-.000106626059868886,-.000217853615265421,-.00246756120900883,-.00823929255543563,-.0168201542714382,-.0255173078910835,-.0309946649224858,-.0309946649224858,-.0255173078910835,-.0168201542714382,-.00823929255543563,-.00246756120900883,-.000217853615265422,-.000360467098389462,-.00408290047436723,-.0136329795428058,-.0278311296202336,-.0422216997546855,-.0512846982894915,-.0512846982894915,-.0422216997546855,-.0278311296202336,-.0136329795428058,-.00408290047436723,-.000360467098389462,-.000499186529428583,-.00565413300383832,-.0188793922500835,-.038541450987549,-.0584699792610344,-.0710207135860798,-.0710207135860798,-.0584699792610344,-.038541450987549,-.0188793922500835,-.00565413300383832,-.000499186529428583,-.000584933620267579,-.00662536405218185,-.0221223743155372,-.0451618565956071,-.0685135808557914,-.0832202005920431,-.0832202005920431,-.0685135808557914,-.0451618565956071,-.0221223743155372,-.00662536405218185,-.000584933620267579,-.000573001809320343,-.00649021608224453,-.0216711094559173,-.0442406191829269,-.0671160015993357,-.0815226272844922,-.0815226272844922,-.0671160015993357,-.0442406191829269,-.0216711094559173,-.00649021608224453,-.000573001809320343,-.000443046618338206,-.00501825341691917,-.0167561630764652,-.0342069717815257,-.0518942821144625,-.0630335258090745,-.0630335258090745,-.0518942821144625,-.0342069717815257,-.0167561630764652,-.00501825341691917,-.000443046618338206,-.000211667613829872,-.00239749426445693,-.00800533602681095,-.016342542282569,-.0247927383077269,-.0301145645786321,-.0301145645786321,-.0247927383077269,-.016342542282569,-.00800533602681095,-.00239749426445693,-.000211667613829872},
			{-1.99945502952237e-010,-2.26472150301291e-009,-7.56200209007358e-009,-1.54374955010019e-008,-2.34197213300945e-008,-2.84468259074429e-008,-2.84468259074429e-008,-2.34197213300945e-008,-1.54374955010019e-008,-7.56200209007358e-009,-2.26472150301291e-009,-1.99945502952237e-010,-6.37261128446745e-008,-7.21806171840919e-007,-2.4101417206608e-006,-4.92019858316594e-006,-7.46427292555284e-006,-9.0664986763096e-006,-9.0664986763096e-006,-7.46427292555284e-006,-4.92019858316594e-006,-2.4101417206608e-006,-7.21806171840918e-007,-6.37261128446745e-008,-1.31833553544195e-006,-1.49323830304632e-005,-4.98598664497681e-005,-.000101786729867376,-.0001544173307413,-.000187563415583651,-.000187563415583651,-.0001544173307413,-.000101786729867376,-4.98598664497681e-005,-1.49323830304632e-005,-1.31833553544195e-006,-9.65293703064032e-006,-.000109335862711265,-.000365077128133699,-.000745288940144662,-.0011306535627788,-.00137335131399144,-.00137335131399144,-.0011306535627788,-.000745288940144662,-.000365077128133699,-.000109335862711265,-9.65293703064033e-006,-3.98777122973611e-005,-.000451682639505874,-.00150818767757949,-.0030788989754264,-.00467089729699484,-.00567351764636142,-.00567351764636142,-.00467089729699484,-.0030788989754264,-.00150818767757949,-.000451682639505874,-3.98777122973611e-005,-.000112743362173094,-.00127700954942894,-.00426398957619228,-.0087047476468162,-.0132056889748729,-.0160403247315041,-.0160403247315041,-.0132056889748729,-.0087047476468162,-.00426398957619228,-.00127700954942894,-.000112743362173094,-.000239961888318527,-.00271797484991882,-.00907543442693066,-.0185271012182458,-.0281068614761773,-.0341400729730291,-.0341400729730291,-.0281068614761773,-.0185271012182458,-.00907543442693066,-.00271797484991882,-.000239961888318527,-.000403951737210287,-.004575437666423,-.0152775823210325,-.0311885140387246,-.0473150782417566,-.0574713838208962,-.0574713838208962,-.0473150782417566,-.0311885140387246,-.0152775823210325,-.004575437666423,-.000403951737210287,-.000549293594811998,-.00622168039425801,-.0207744572931185,-.042410143131176,-.0643390460348379,-.0781495909283993,-.0781495909283993,-.0643390460348379,-.042410143131176,-.0207744572931185,-.00622168039425801,-.000549293594811998,-.000599983629581883,-.00679583090045496,-.0226915704225263,-.0463238454758227,-.0702763963177652,-.0853614090140783,-.0853614090140783,-.0702763963177652,-.0463238454758227,-.0226915704225263,-.00679583090045496,-.000599983629581883,-.000499088010632534,-.00565301711159485,-.0188756662380921,-.0385338444975389,-.058458439702929,-.0710066970315766,-.0710066970315766,-.058458439702929,-.0385338444975389,-.0188756662380921,-.00565301711159485,-.000499088010632534,-.000248139442448811,-.00281059950217384,-.00938471210765786,-.0191584780345857,-.0290647026682284,-.0353035172878956,-.0353035172878956,-.0290647026682284,-.0191584780345857,-.00938471210765786,-.00281059950217384,-.000248139442448811},
			{-1.92058726559064e-006,-.000113117990564565,-.000906410211953001,-.00331870648655728,-.00771242584002074,-.0129629077002378,-.0166745014700866,-.0166874796528932,-.0127648935683974,-.00697208388856876,-.00224638876284317,-.000206393222578878,-4.35367906124191e-006,-.000256421270615505,-.00205469401540682,-.00752300257311598,-.0174828957229149,-.0293849390956891,-.0377985573514871,-.0378279769169244,-.0289360711920697,-.0158046531822386,-.00509222147590975,-.000467862027215704,-6.51705798837019e-006,-.000383839108612685,-.00307568836802217,-.0112612444155839,-.0261702903746801,-.0439865569734266,-.0565809713282148,-.0566250097176437,-.0433146429173246,-.0236581153148544,-.0076225882021403,-.000700346515907722,-8.27128698753606e-006,-.000487158995675827,-.00390358674443833,-.0142924897345245,-.0332146779455411,-.0558266378740331,-.071811153548789,-.0718670459711192,-.0549738613606283,-.0300262881966959,-.00967439828220347,-.000888862280208502,-9.50587324711357e-006,-.000559873169805834,-.00448624269205744,-.0164258108812803,-.0381723568496402,-.0641594160912017,-.0825298075610779,-.0825940425807241,-.063179352715725,-.0345080626641537,-.0111184153023947,-.00102153536475457,-1.01431945983696e-005,-.000597409871152132,-.00478702287082038,-.0175270795090219,-.0407316228334812,-.0684609846789651,-.0880630191983885,-.0881315608554635,-.067415212946293,-.0368216560137206,-.0118638495492168,-.00109002421181751,-1.01431945983696e-005,-.000597409871152132,-.00478702287082038,-.0175270795090219,-.0407316228334812,-.0684609846789651,-.0880630191983885,-.0881315608554635,-.067415212946293,-.0368216560137206,-.0118638495492168,-.00109002421181751,-9.50587324711357e-006,-.000559873169805834,-.00448624269205744,-.0164258108812803,-.0381723568496403,-.0641594160912017,-.0825298075610779,-.0825940425807241,-.063179352715725,-.0345080626641537,-.0111184153023947,-.00102153536475457,-8.27128698753606e-006,-.000487158995675827,-.00390358674443833,-.0142924897345245,-.0332146779455411,-.0558266378740331,-.071811153548789,-.0718670459711192,-.0549738613606283,-.0300262881966959,-.00967439828220347,-.000888862280208502,-6.51705798837019e-006,-.000383839108612685,-.00307568836802217,-.0112612444155839,-.0261702903746801,-.0439865569734266,-.0565809713282148,-.0566250097176437,-.0433146429173246,-.0236581153148544,-.0076225882021403,-.000700346515907722,-4.35367906124191e-006,-.000256421270615505,-.00205469401540682,-.00752300257311598,-.0174828957229149,-.0293849390956891,-.0377985573514871,-.0378279769169244,-.0289360711920697,-.0158046531822386,-.00509222147590975,-.000467862027215704,-1.92058726559064e-006,-.000113117990564565,-.000906410211953001,-.00331870648655728,-.00771242584002074,-.0129629077002378,-.0166745014700866,-.0166874796528932,-.0127648935683974,-.00697208388856876,-.00224638876284317,-.000206393222578878},
			{-3.06697796211909e-008,-1.80637657239712e-006,-1.44744276633774e-005,-5.29962883716334e-005,-.000123159413319067,-.000207004143752805,-.000266274433108618,-.000266481681179921,-.000203842064166924,-.000111336922926592,-3.58724904274172e-005,-3.29588494374136e-006,-3.61515976889952e-007,-2.1292425288575e-005,-.000170615404521307,-.000624686750242411,-.00145172531948917,-.00244003400654881,-.00313867471481826,-.0031411176240886,-.00240276141102928,-.00131236927532137,-.000422842243423972,-3.88498085043389e-005,-1.29865498884588e-006,-7.6487668853577e-005,-.000612892819182374,-.00224402963223718,-.00521495715018533,-.00876520690127838,-.0112748974798531,-.0112836730154145,-.00863131450035304,-.00471435570086283,-.00151895413763271,-.000139558140870839,-2.95610118832487e-006,-.00017410728078841,-.0013951151049833,-.00510803771553511,-.0118706978844523,-.0199520571355211,-.025664813306616,-.0256847888746645,-.0196472806638003,-.0107311892759846,-.00345756353291146,-.000317673277053615,-5.20421827112933e-006,-.000306515993225909,-.00245610114713186,-.00899270407717994,-.0208983721752537,-.0351256109572267,-.0451829223109582,-.0452180893128947,-.034589051759252,-.0188922664493339,-.00608704308994404,-.000559264033051857,-7.68418084863826e-006,-.000452579850083181,-.00362650534890298,-.0132779912077635,-.0308570591913224,-.0518639943508112,-.0667139094133409,-.0667658345226861,-.051071748964773,-.0278949852742899,-.00898769757526597,-.000825769740663992,-9.88434754146089e-006,-.000582164399381254,-.0046648614778342,-.0170798010009427,-.0396921810095751,-.0667139094133409,-.0858157140598609,-.0858825065829686,-.0656948250258218,-.0358819937409303,-.0115610926110904,-.0010622075751108,-1.12604371581811e-005,-.000663212782377944,-.00531429912823647,-.0194576349160754,-.0452180893128947,-.0760017574629395,-.0977628974803007,-.0978389887959829,-.0748407971004666,-.0408774513376,-.0131706171076067,-.00121008711989414,-1.13701881137203e-005,-.000669676850830108,-.00536609546608301,-.0196472806638003,-.0456588118567917,-.0767425160486888,-.0987157531522446,-.0987925860994006,-.0755702402721397,-.0412758674276048,-.0132989858194334,-.0012218813527315,-9.98922055945599e-006,-.000588341168993646,-.00471435570086283,-.017261017846082,-.040113315413826,-.067421744603583,-.0867262195724516,-.0867937207648416,-.0663918477213734,-.0362627020233443,-.01168375590964,-.0010734776072168,-7.17927735473982e-006,-.000422842243423972,-.00338821902309026,-.0124055359279088,-.0288295383268322,-.0484561734690455,-.0623303470508818,-.06237886032367,-.0477159840498455,-.0260620930240078,-.00839714407356662,-.000771510993319342,-3.29588494374136e-006,-.000194119730276028,-.00155547411146173,-.00569517196000836,-.0132351539871473,-.0222454106000665,-.0286148092951926,-.0286370809191255,-.0219056021427923,-.0119646666030396,-.00385498697918716,-.000354187662792186},
			{-3.65048874112027e-010,-2.15005044744533e-008,-1.72282735226472e-007,-6.30791470989705e-007,-1.46591223424896e-006,-2.46387911966847e-006,-3.16934726012697e-006,-3.1718140426088e-006,-2.4262422795302e-006,-1.32519433994785e-006,-4.26974448589619e-007,-3.92294663599167e-008,-2.23749630651595e-008,-1.31783174148499e-006,-1.05597362732147e-005,-3.86631403796111e-005,-8.98502486218331e-005,-.00015101869423293,-.000194258996301528,-.000194410192951716,-.000148711817078381,-8.12252180824206e-005,-2.61705711055768e-005,-2.40449409138482e-006,-1.92885569959679e-007,-1.1360498152647e-005,-9.10312452249138e-005,-.000333299404644144,-.000774562906136249,-.00130187150820567,-.00167462878541062,-.00167593219100564,-.00128198484678476,-.000700209981991998,-.000225605982417451,-2.07281778267435e-005,-7.87461355147049e-007,-4.63795880236092e-005,-.000371637897746933,-.00136070521452511,-.00316217722165629,-.00531493103550084,-.00683672424538222,-.00684204543937499,-.00523374312001741,-.00285862908989082,-.000921043459502259,-8.46234324557743e-005,-2.12364898657175e-006,-.00012507783964276,-.00100224403364473,-.00366958991823735,-.00852785271078683,-.0143334626308517,-.0184374794017228,-.0184518297544813,-.0141145126680233,-.00770923516443232,-.00248389714184165,-.000228214966233319,-4.33894499945229e-006,-.000255553469660892,-.00204774035893469,-.00749754263838111,-.0174237287374253,-.0292854922824927,-.0376706365121829,-.0376999565136218,-.0288381434728016,-.0157511658366454,-.00507498798101565,-.000466278651886231,-7.17934798611665e-006,-.000422846403442921,-.00338825235716593,-.0124056579764159,-.0288298219584409,-.0484566501919911,-.0623309602711284,-.0623794740212014,-.0477164534906379,-.0260623494288551,-.00839722668659549,-.000771518583621402,-9.94219394991092e-006,-.000585571414309697,-.00469216176055673,-.0171797575373556,-.0399244725296093,-.0671043408542665,-.0863179354585031,-.0863851188731245,-.0660792924543047,-.0360919868089839,-.0116287518756503,-.00106842395843715,-1.16500008667718e-005,-.000686157151895624,-.00549815149985173,-.0201307871491398,-.0467824447922306,-.0786310982319312,-.101145082058928,-.101223805813716,-.0774299735296571,-.0422916390211788,-.0136262649987847,-.00125195103863209,-1.14123574777427e-005,-.000672160525384795,-.00538599705705276,-.0197201478250184,-.0458281496934892,-.0770271360622618,-.0990818667545695,-.0991589846571268,-.0758505125894862,-.0414289499502059,-.0133483086422873,-.00122641302441892,-8.82406705447926e-006,-.000519716417833643,-.00416446814598573,-.015247674029803,-.0354344548588437,-.0595575992912496,-.0766103793918513,-.0766700071720531,-.0586478306968206,-.0320329811846719,-.0103209499661349,-.000948266016470748,-4.21573969958762e-006,-.000248296972548114,-.00198959431997835,-.00728464826218446,-.0169289781185245,-.0284539242725722,-.0366009704831913,-.0366294579367253,-.028019278030959,-.0153039080100614,-.00493088257841441,-.00045303856676568},
			{-3.06697796211909e-008,-9.39296633663941e-006,-.000180620480252622,-.00118608291592308,-.00422235247914266,-.00982031115806883,-.0162489801425508,-.0197676016789817,-.0175473588745023,-.0106866754670937,-.00370432632910279,-.000354187662792186,-6.95237231559063e-008,-2.1292425288575e-005,-.000409439142389331,-.00268866947546018,-.00957143052386794,-.0222611509665594,-.0368339652566651,-.0448101448253052,-.0397771922581837,-.024225067014107,-.00839714407356662,-.000802889532215207,-1.04070632906319e-007,-3.18728065084108e-005,-.000612892819182374,-.00402469144754472,-.0143275530598965,-.0333228999419048,-.0551370654893481,-.0670766742761412,-.0595428061909519,-.0362627020233444,-.0125697540159339,-.00120185194317225,-1.32083844163853e-007,-4.04521687853206e-005,-.000777868235805777,-.00510803771553511,-.0181841719682459,-.0422925910998948,-.0699785843745455,-.085132037201067,-.0755702402721397,-.0460237143683289,-.0159532173895116,-.00152536042432723,-1.51798901731387e-007,-4.64901277906996e-005,-.000893974161901044,-.00587047204848792,-.0208983721752537,-.0486052546469995,-.0804237060180922,-.0978389887959829,-.0868499819149511,-.0528932916734667,-.0183344215497376,-.00175303829641842,-1.61976260365957e-007,-4.96070587967628e-005,-.000953910667053138,-.00626405789601532,-.0222995037123093,-.0518639943508112,-.0858157140598609,-.104398604617067,-.0926728396779039,-.0564395228555362,-.0195636529956887,-.00187057076364506,-1.61976260365957e-007,-4.96070587967628e-005,-.000953910667053138,-.00626405789601532,-.0222995037123093,-.0518639943508112,-.0858157140598609,-.104398604617067,-.0926728396779039,-.0564395228555362,-.0195636529956887,-.00187057076364506,-1.51798901731387e-007,-4.64901277906996e-005,-.000893974161901044,-.00587047204848792,-.0208983721752537,-.0486052546469995,-.0804237060180922,-.0978389887959829,-.0868499819149511,-.0528932916734667,-.0183344215497376,-.00175303829641842,-1.32083844163853e-007,-4.04521687853206e-005,-.000777868235805777,-.00510803771553511,-.0181841719682459,-.0422925910998948,-.0699785843745455,-.085132037201067,-.0755702402721397,-.0460237143683289,-.0159532173895116,-.00152536042432723,-1.04070632906319e-007,-3.18728065084108e-005,-.000612892819182374,-.00402469144754472,-.0143275530598965,-.0333228999419048,-.0551370654893481,-.0670766742761412,-.0595428061909519,-.0362627020233443,-.0125697540159339,-.00120185194317225,-6.95237231559063e-008,-2.1292425288575e-005,-.000409439142389331,-.00268866947546018,-.00957143052386794,-.0222611509665594,-.0368339652566651,-.0448101448253052,-.0397771922581837,-.024225067014107,-.00839714407356662,-.000802889532215207,-3.06697796211909e-008,-9.39296633663941e-006,-.000180620480252622,-.00118608291592308,-.00422235247914266,-.00982031115806883,-.0162489801425508,-.0197676016789817,-.0175473588745023,-.0106866754670937,-.00370432632910279,-.000354187662792186},
			{-4.89764458734161e-010,-1.49995895979437e-007,-2.8843210756594e-006,-1.89405096647495e-005,-6.74265743697526e-005,-.000156820147987287,-.00025947930056204,-.000315668024236135,-.00028021305752868,-.000170655084270396,-5.91542359272188e-005,-5.6560083280767e-006,-5.77303387673914e-009,-1.76805681469661e-006,-3.3998553762376e-005,-.000223258756300761,-.000794781840714042,-.00184849678399644,-.00305857798732252,-.00372089514708473,-.00330297440936967,-.00201157426832706,-.000697272743806459,-6.66694512082365e-005,-2.07381685017084e-008,-6.35129828208852e-006,-.000122131231479417,-.000802000786155484,-.00285505335439599,-.00664025858844208,-.0109871701831305,-.0133663775728643,-.0118651027035077,-.00722607333006517,-.00250477651118074,-.000239493192417188,-4.72058591990341e-008,-1.44573274351898e-005,-.000278004767708733,-.00182557761480488,-.00649889823405786,-.0151150817366226,-.0250098656792164,-.030425605696751,-.0270082851124352,-.0164485595800252,-.00570157038226907,-.00054515334463801,-8.31059491191454e-008,-2.54521353623046e-005,-.000489427593779599,-.00321393070570632,-.0114413150217837,-.0266101122837137,-.044029886541206,-.0535643007426736,-.0475481053927164,-.0289577009874395,-.0100376187644472,-.000959742855872972,-1.22708370279532e-007,-3.75808240390211e-005,-.000722654190693695,-.00474546290931794,-.0168934370530446,-.0392906109117239,-.0650114182959186,-.0790892603834303,-.0702061715732879,-.0427568944566942,-.0148208383771267,-.00141708846333921,-1.57842742379509e-007,-4.83411222371217e-005,-.000929567551026009,-.00610420363142845,-.0217304363719754,-.0505404624146264,-.0836257585887017,-.101734427107537,-.0903078952792718,-.0549992267137261,-.0190644026033435,-.00182283513943086,-1.79817461292618e-007,-5.50711280460104e-005,-.00105898107575704,-.00695402514977553,-.0247557273922317,-.0575766583052787,-.0952680584574148,-.115897798864594,-.102880475966099,-.062656167598386,-.0217185308967278,-.0020766085420598,-1.81570069821245e-007,-5.56078841986148e-005,-.00106930253870999,-.00702180323816678,-.0249970115180785,-.0581378348543761,-.0961965979360501,-.11702740813212,-.103883210618848,-.0632668520831664,-.0219302126889477,-.00209684841095391,-1.59517455322627e-007,-4.88540219870558e-005,-.000939430271260956,-.00616896928789685,-.0219609965010198,-.0510766971839986,-.0845130286536747,-.102813830311442,-.0912660628757668,-.0555827690131719,-.0192666760896712,-.00184217543696469,-1.14645586997212e-007,-3.51115056126793e-005,-.000675170843678735,-.00443365338137005,-.0157834221327753,-.0367089477367546,-.0607397213008226,-.0738925524084332,-.0655931435851624,-.0399474726296093,-.0138470074345037,-.00132397601187602,-5.2631852118241e-008,-1.61190990378048e-005,-.000309959527704379,-.00203541536332734,-.00724590244919067,-.0168524577290889,-.0278845798860605,-.0339228224379417,-.0301127040609837,-.0183392097943569,-.00635692717584237,-.000607815019228362},
			{-4.34129792523551e-010,-6.91363066088356e-007,-3.19037343548278e-005,-.000375745321421736,-.00204903932812416,-.00659447754217011,-.0140356226122099,-.020756280852719,-.0213815233314377,-.0145196103881163,-.00541459382780606,-.000538770430325433,-9.84106174935942e-010,-1.56721486103326e-006,-7.23209107571181e-005,-.000851757463741354,-.00464486033951284,-.0149486770580804,-.0318166205582331,-.0470513300576056,-.0484686596091265,-.032913747194243,-.0122740602291765,-.0012213105768994,-1.47311662585924e-009,-2.34597681315439e-006,-.000108257766028679,-.00127500275172926,-.00695292963828415,-.0223767976156558,-.0476265609511484,-.0704315228803041,-.0725531351410981,-.04926885883459,-.0183731416903004,-.00182818984576065,-1.86964277444457e-009,-2.97745509129004e-006,-.000137398048789906,-.00161820160082515,-.00882448438314045,-.0284000718224987,-.0604464398750644,-.0893899271345745,-.0920826243480438,-.0625308032698983,-.0233187318655643,-.00232029282368979,-2.14870881134135e-009,-3.42187506483889e-006,-.000157906313511514,-.00185973710365748,-.0101416418199092,-.0326391144885199,-.0694687775381858,-.102732418569255,-.105827032314693,-.0718642565322132,-.0267993251511533,-.00266662364773693,-2.29276901154607e-009,-3.65129470714422e-006,-.000168493143619897,-.00198442319330673,-.0108215882804389,-.0348274041920398,-.0741263123085118,-.109620114430362,-.112922206580213,-.0766823962117034,-.0285960861297693,-.00284540745247398,-2.29276901154607e-009,-3.65129470714422e-006,-.000168493143619897,-.00198442319330673,-.0108215882804389,-.0348274041920399,-.0741263123085118,-.109620114430362,-.112922206580213,-.0766823962117034,-.0285960861297693,-.00284540745247398,-2.14870881134135e-009,-3.42187506483889e-006,-.000157906313511514,-.00185973710365748,-.0101416418199092,-.0326391144885199,-.0694687775381858,-.102732418569255,-.105827032314693,-.0718642565322132,-.0267993251511533,-.00266662364773693,-1.86964277444457e-009,-2.97745509129004e-006,-.000137398048789906,-.00161820160082515,-.00882448438314046,-.0284000718224987,-.0604464398750644,-.0893899271345745,-.0920826243480438,-.0625308032698983,-.0233187318655643,-.00232029282368979,-1.47311662585924e-009,-2.34597681315439e-006,-.000108257766028679,-.00127500275172926,-.00695292963828415,-.0223767976156558,-.0476265609511484,-.0704315228803041,-.0725531351410981,-.04926885883459,-.0183731416903004,-.00182818984576065,-9.84106174935942e-010,-1.56721486103326e-006,-7.23209107571182e-005,-.000851757463741354,-.00464486033951284,-.0149486770580804,-.0318166205582331,-.0470513300576056,-.0484686596091265,-.032913747194243,-.0122740602291765,-.0012213105768994,-4.34129792523551e-010,-6.91363066088356e-007,-3.19037343548278e-005,-.000375745321421736,-.00204903932812416,-.00659447754217011,-.0140356226122099,-.020756280852719,-.0213815233314377,-.0145196103881163,-.00541459382780606,-.000538770430325433},
			{-.00423058507690693,-.00885514027394949,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394863,-.00435422423400534,-.00213112724015142,-.000790234334926298,-.000193564800479007,-2.24537546398006e-005,-3.66334363412211e-007,-.00959009257018474,-.020073255449394,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134588,-.00987036372424941,-.00483094114415894,-.00179134098152266,-.000438781946510387,-5.08992448158822e-005,-8.30424254068273e-007,-.0143554884304917,-.0300478211505447,-.0388614993834644,-.0396707255608121,-.0338552805420154,-.0244472036457215,-.0147750286257641,-.00723147552495667,-.00268147304597068,-.000656816303965783,-7.61914981245153e-005,-1.2430688947403e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971213,-.0429682752411084,-.03102777936875,-.0187520967637297,-.00917800785523132,-.00340325851820642,-.000833614824033866,-9.670034364015e-005,-1.57767194829697e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.057864283158913,-.0493817925438395,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011006,-.000958041580001006,-.000111134000184102,-1.81315792676965e-006,-.0223429825183335,-.0467666388316429,-.0604843092288455,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102643,-.0112551190451086,-.00417346338855665,-.00102227348573477,-.000118584980154756,-1.93472111511538e-006,-.0223429825183335,-.0467666388316429,-.0604843092288455,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102643,-.0112551190451086,-.00417346338855665,-.00102227348573477,-.000118584980154756,-1.93472111511538e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.057864283158913,-.0493817925438395,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011006,-.000958041580001006,-.000111134000184102,-1.81315792676965e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971213,-.0429682752411084,-.03102777936875,-.0187520967637297,-.00917800785523132,-.00340325851820642,-.000833614824033866,-9.670034364015e-005,-1.57767194829697e-006,-.0143554884304917,-.0300478211505447,-.0388614993834644,-.0396707255608121,-.0338552805420154,-.0244472036457215,-.0147750286257641,-.00723147552495667,-.00268147304597068,-.000656816303965783,-7.61914981245153e-005,-1.2430688947403e-006,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134588,-.00987036372424941,-.00483094114415894,-.00179134098152266,-.000438781946510386,-5.08992448158822e-005,-8.30424254068273e-007,-.00423058507690692,-.00885514027394949,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394863,-.00435422423400534,-.00213112724015142,-.000790234334926298,-.000193564800479007,-2.24537546398006e-005,-3.66334363412211e-007},
			{-6.75580403638936e-005,-.00014140737349095,-.000182885225860587,-.00018669350692448,-.000159325572268953,-.000115050433754236,-6.95324290155738e-005,-3.40318838868514e-005,-1.26192198302079e-005,-3.09102839585264e-006,-3.58563091085661e-007,-5.84997849237861e-009,-.000796331478757963,-.00166682074010411,-.00215573544715162,-.00220062505725282,-.0018780291418385,-.00135614179377857,-.000819604324242625,-.000401146336906179,-.000148747387195263,-3.64350890004165e-005,-4.22651508264995e-006,-6.89558489033362e-008,-.00286062003831949,-.00598763321131698,-.00774393601400238,-.00790519061914258,-.00674634613725228,-.00487159743595245,-.00294422186735428,-.00144101706420306,-.000534337481072711,-.000130883869937062,-1.51826896966895e-005,-2.4770649960978e-007,-.00651156956023967,-.013629524241044,-.0176273595758088,-.0179944200606735,-.0153565666050425,-.0110891153486868,-.00670187065504814,-.0032801570028024,-.00121630123191481,-.000297928215558894,-3.45600389940658e-005,-5.63849124010232e-007,-.0114636229006531,-.0239947872427992,-.031032979229028,-.0316791894770017,-.0270352465684925,-.0195223955581165,-.0117986481151247,-.0057747187659129,-.00214129919480704,-.000524502838069343,-6.0842973540354e-005,-9.92656788923096e-007,-.0169263752901921,-.0354290068157069,-.0458211036210322,-.0467752519970179,-.0399183341468002,-.0288253893768114,-.0174210498412452,-.00852654155446687,-.00316169103729199,-.000774443817183375,-8.98364341570052e-005,-1.46568685042074e-006,-.0217728056224891,-.045573187736373,-.0589407930194454,-.0601681371359018,-.0513479179830939,-.0370787950245641,-.0224091174530978,-.0109678964760433,-.00406695959489998,-.000996185799262211,-.00011555877647658,-1.88534841928859e-006,-.0248039952501605,-.0519178442938048,-.0671464750773758,-.0685446888933965,-.0584965270825159,-.042240870176189,-.0255288938184798,-.012494836761645,-.00463315790456963,-.00113487385418326,-.000131646761218445,-2.1478248622506e-006,-.0250457498234202,-.0524238666569454,-.0678009247845504,-.0692127664287096,-.0590666692229003,-.042652574961671,-.0257777136867565,-.012616618914028,-.004678315432652,-.00114593501354709,-.000132929869292705,-2.16875884799657e-006,-.0220038152896698,-.0460567196758888,-.0595661553735154,-.0608065216222948,-.0518927198634471,-.0374722013795649,-.0226468783946324,-.0110842659581639,-.0041101100734715,-.00100675533972143,-.000116784856153978,-1.90535198329458e-006,-.015814196101363,-.0331010775700317,-.042810342192032,-.0437017964620066,-.0372954252410452,-.0269313631824755,-.0162763671345931,-.00796628917278541,-.00295394620634634,-.000723557535766244,-8.39335630014821e-005,-1.36938115091738e-006,-.00726003025826053,-.0151961454884669,-.0196535048439074,-.0200627564386384,-.0171216996431039,-.0123637338469658,-.00747220517149892,-.00365718877328773,-.00135610679808757,-.00033217304057603,-3.85324807640305e-005,-6.28659751468172e-007},
			{-8.04113589360719e-007,-1.68310966462922e-006,-2.17680226684608e-006,-2.22213055847631e-006,-1.89638209018582e-006,-1.36939462342768e-006,-8.27613867594724e-007,-4.05066519951766e-007,-1.50201013794113e-007,-3.67911491336473e-008,-4.26781849550599e-009,-6.96297165792815e-011,-4.92865836277514e-005,-.000103162944064127,-.000133422875058339,-.000136201184821656,-.00011623506394466,-8.39343887728836e-005,-5.07269876251226e-005,-2.48277670898173e-005,-9.20627996502583e-006,-2.25504216385422e-006,-2.61587661208472e-007,-4.26781849550599e-009,-.000424879842112779,-.000889326306666056,-.001150185018243,-.00117413571083906,-.00100201580190209,-.000723564654409836,-.000437296986453802,-.000214030208318087,-7.93636419908757e-005,-1.94398127850062e-005,-2.25504216385422e-006,-3.67911491336473e-008,-.00173458520673544,-.0036307023836027,-.00469566620937377,-.00479344565888028,-.00409076076250558,-.00295397526843083,-.00178527858577817,-.000873785000712069,-.000324004543650437,-7.93636419908757e-005,-9.20627996502583e-006,-1.50201013794113e-007,-.00467788049829846,-.0097913851734879,-.0126634109999641,-.0129271055006463,-.0110320841661877,-.00796636754827121,-.00481459189667545,-.00235644913761801,-.000873785000712069,-.000214030208318087,-2.48277670898173e-005,-4.05066519951766e-007,-.00955763703157626,-.0200053219740424,-.025873317192124,-.0264120860481604,-.0225402628819985,-.016276527267894,-.0098369596701605,-.00481459189667545,-.00178527858577817,-.000437296986453802,-5.07269876251226e-005,-8.27613867594725e-007,-.0158143516876436,-.0331014032314506,-.0428107633770146,-.0437022264174669,-.037295792168106,-.0269316281438179,-.016276527267894,-.00796636754827121,-.00295397526843083,-.000723564654409835,-8.39343887728836e-005,-1.36939462342768e-006,-.0219002271480226,-.0458398968231796,-.0592857336712341,-.0605202606038321,-.0516484226656698,-.037295792168106,-.0225402628819985,-.0110320841661877,-.00409076076250558,-.00100201580190209,-.00011623506394466,-1.89638209018582e-006,-.0256621090417623,-.0537139830919874,-.0694694603763257,-.0709160465028163,-.0605202606038321,-.0437022264174669,-.0264120860481604,-.0129271055006463,-.00479344565888028,-.00117413571083906,-.000136201184821656,-2.22213055847631e-006,-.0251386386461752,-.0526182945056766,-.0680523825420277,-.0694694603763257,-.0592857336712341,-.0428107633770146,-.025873317192124,-.0126634109999641,-.00469566620937377,-.001150185018243,-.000133422875058339,-2.17680226684608e-006,-.0194372664460252,-.0406846140173892,-.0526182945056766,-.0537139830919874,-.0458398968231796,-.0331014032314506,-.0200053219740424,-.0097913851734879,-.0036307023836027,-.000889326306666056,-.000103162944064127,-1.68310966462922e-006,-.00928624582089677,-.0194372664460252,-.0251386386461752,-.0256621090417623,-.0219002271480226,-.0158143516876436,-.00955763703157627,-.00467788049829846,-.00173458520673544,-.000424879842112779,-4.92865836277514e-005,-8.04113589360719e-007},
			{-8.77197535304017e-009,-1.83608344516711e-008,-2.37464657802842e-008,-2.42409464882796e-008,-2.06873968742175e-008,-1.49385560003323e-008,-9.02833697182859e-009,-4.41882039598848e-009,-1.63852421901125e-009,-4.01350079989265e-010,-4.65571023163414e-011,-7.59581936873143e-013,-2.79578126521843e-006,-5.85191760211322e-006,-7.56841207045341e-006,-7.72601168100707e-006,-6.59343355166069e-006,-4.76117787775939e-006,-2.8774881763854e-006,-1.40835499192186e-006,-5.22226195326764e-007,-1.2791725800267e-007,-1.48385590679724e-008,-2.42091987612831e-010,-5.7837794064797e-005,-.000121061690113565,-.000156571711876826,-.000159832057717645,-.000136401819657762,-9.84969850916432e-005,-5.95281078102025e-005,-2.91353787244659e-005,-1.08035673306481e-005,-2.64629143836613e-006,-3.06973057681128e-007,-5.00828398075496e-009,-.000423492023911399,-.000886421430715201,-.00114642807911561,-.00117030053969446,-.000998742839454559,-.000721201218685817,-.000435868609164376,-.00021333110473795,-7.91044103306138e-005,-1.93763150067161e-005,-2.24767634356886e-006,-3.66709753307118e-008,-.00174950826221703,-.00366193819311915,-.00473606415990016,-.00483468482962708,-.00412595456537195,-.00297938903112802,-.00180063776864345,-.000881302384115301,-.000326792032995098,-8.00464265713343e-005,-9.28548381512419e-006,-1.51493229393287e-007,-.00494625775323145,-.0103531321176025,-.0133899305174026,-.0136687535803171,-.011665011934605,-.00842341040243774,-.00509081249643689,-.00249164227715485,-.000923915400575431,-.000226309453117529,-2.62521745709628e-005,-4.28305814057175e-007,-.010527567456728,-.0220355068809891,-.0284989994042999,-.0290924437312573,-.0248276992732258,-.0179283057316278,-.0108352363826531,-.00530318747211388,-.00196645265756991,-.000481674864645602,-5.58748760112742e-005,-9.11601977606186e-007,-.0177221024244434,-.0370945625876605,-.0479752030574144,-.0489742069762943,-.0417949380321651,-.0301805019800266,-.0182400321589689,-.00892738345711226,-.00331032553849853,-.000810851255203673,-9.40597416919936e-005,-1.53459036799021e-006,-.0240985158662193,-.0504411882777659,-.0652366837960255,-.0665951293806949,-.0568327590696347,-.0410394482774791,-.0248027967481441,-.0121394565234153,-.00450138085207253,-.00110259558209743,-.00012790244200432,-2.08673606807239e-006,-.02632238051474,-.0550960133326791,-.0712568700882474,-.0727406760531766,-.0620774124945382,-.044826659852073,-.027091654003041,-.0132597125742255,-.00491677828990747,-.00120434555501086,-.000139705563856799,-2.27930471413928e-006,-.0218959049522221,-.0458308499302774,-.0592740331282503,-.0605083164164748,-.0516382294108676,-.0372884315268124,-.0225358143697108,-.011029906890695,-.00408995341613109,-.00100181804557465,-.000116212123967743,-1.89600782307658e-006,-.0108863317310921,-.0227864450885946,-.0294702040898467,-.0300838721412506,-.025673791359598,-.0185392764637387,-.0112044855690737,-.00548391243189871,-.00203346652033204,-.000498089647453146,-5.77790109816517e-005,-9.42668054679449e-007},
			{-.000206393222578878,-.00224638876284317,-.00697208388856876,-.0127648935683974,-.0166874796528932,-.0166745014700866,-.0129629077002378,-.00771242584002074,-.00331870648655728,-.000906410211953001,-.000113117990564565,-1.92058726559064e-006,-.000467862027215704,-.00509222147590975,-.0158046531822386,-.0289360711920697,-.0378279769169244,-.0377985573514871,-.0293849390956891,-.0174828957229149,-.00752300257311598,-.00205469401540682,-.000256421270615505,-4.35367906124192e-006,-.000700346515907722,-.0076225882021403,-.0236581153148544,-.0433146429173246,-.0566250097176437,-.0565809713282148,-.0439865569734266,-.0261702903746801,-.0112612444155839,-.00307568836802217,-.000383839108612685,-6.51705798837019e-006,-.000888862280208502,-.00967439828220347,-.0300262881966959,-.0549738613606283,-.0718670459711192,-.071811153548789,-.0558266378740331,-.0332146779455411,-.0142924897345245,-.00390358674443833,-.000487158995675827,-8.27128698753606e-006,-.00102153536475457,-.0111184153023947,-.0345080626641537,-.063179352715725,-.0825940425807241,-.0825298075610779,-.0641594160912017,-.0381723568496403,-.0164258108812803,-.00448624269205744,-.000559873169805834,-9.50587324711357e-006,-.00109002421181751,-.0118638495492168,-.0368216560137206,-.067415212946293,-.0881315608554635,-.0880630191983885,-.0684609846789651,-.0407316228334812,-.0175270795090219,-.00478702287082038,-.000597409871152132,-1.01431945983696e-005,-.00109002421181751,-.0118638495492168,-.0368216560137206,-.067415212946293,-.0881315608554635,-.0880630191983885,-.0684609846789651,-.0407316228334812,-.0175270795090219,-.00478702287082038,-.000597409871152132,-1.01431945983696e-005,-.00102153536475457,-.0111184153023947,-.0345080626641537,-.063179352715725,-.0825940425807241,-.0825298075610779,-.0641594160912017,-.0381723568496402,-.0164258108812803,-.00448624269205744,-.000559873169805834,-9.50587324711357e-006,-.000888862280208502,-.00967439828220347,-.0300262881966959,-.0549738613606283,-.0718670459711192,-.071811153548789,-.0558266378740331,-.0332146779455411,-.0142924897345245,-.00390358674443833,-.000487158995675827,-8.27128698753606e-006,-.000700346515907722,-.0076225882021403,-.0236581153148544,-.0433146429173246,-.0566250097176437,-.0565809713282148,-.0439865569734266,-.0261702903746801,-.0112612444155839,-.00307568836802217,-.000383839108612685,-6.51705798837019e-006,-.000467862027215704,-.00509222147590975,-.0158046531822386,-.0289360711920697,-.0378279769169244,-.0377985573514871,-.0293849390956891,-.0174828957229149,-.00752300257311598,-.00205469401540682,-.000256421270615505,-4.35367906124192e-006,-.000206393222578878,-.00224638876284317,-.00697208388856876,-.0127648935683974,-.0166874796528932,-.0166745014700866,-.0129629077002378,-.00771242584002074,-.00331870648655728,-.000906410211953001,-.000113117990564565,-1.92058726559064e-006},
			{-3.29588494374136e-006,-3.58724904274172e-005,-.000111336922926592,-.000203842064166924,-.000266481681179921,-.000266274433108618,-.000207004143752805,-.000123159413319067,-5.29962883716334e-005,-1.44744276633774e-005,-1.80637657239712e-006,-3.06697796211909e-008,-3.88498085043389e-005,-.000422842243423972,-.00131236927532137,-.00240276141102928,-.0031411176240886,-.00313867471481826,-.00244003400654881,-.00145172531948917,-.000624686750242411,-.000170615404521307,-2.1292425288575e-005,-3.61515976889952e-007,-.000139558140870839,-.00151895413763271,-.00471435570086283,-.00863131450035304,-.0112836730154145,-.0112748974798531,-.00876520690127838,-.00521495715018533,-.00224402963223718,-.000612892819182374,-7.6487668853577e-005,-1.29865498884588e-006,-.000317673277053615,-.00345756353291146,-.0107311892759846,-.0196472806638003,-.0256847888746645,-.025664813306616,-.0199520571355211,-.0118706978844523,-.00510803771553511,-.0013951151049833,-.00017410728078841,-2.95610118832487e-006,-.000559264033051857,-.00608704308994404,-.0188922664493339,-.034589051759252,-.0452180893128947,-.0451829223109582,-.0351256109572267,-.0208983721752537,-.00899270407717994,-.00245610114713186,-.000306515993225909,-5.20421827112933e-006,-.000825769740663992,-.00898769757526597,-.0278949852742899,-.051071748964773,-.0667658345226861,-.0667139094133409,-.0518639943508112,-.0308570591913224,-.0132779912077635,-.00362650534890298,-.000452579850083181,-7.68418084863826e-006,-.0010622075751108,-.0115610926110904,-.0358819937409303,-.0656948250258218,-.0858825065829686,-.0858157140598609,-.0667139094133409,-.0396921810095751,-.0170798010009427,-.0046648614778342,-.000582164399381254,-9.88434754146089e-006,-.00121008711989414,-.0131706171076067,-.0408774513376,-.0748407971004666,-.0978389887959829,-.0977628974803007,-.0760017574629395,-.0452180893128947,-.0194576349160754,-.00531429912823647,-.000663212782377944,-1.12604371581811e-005,-.0012218813527315,-.0132989858194334,-.0412758674276049,-.0755702402721397,-.0987925860994006,-.0987157531522446,-.0767425160486888,-.0456588118567917,-.0196472806638003,-.00536609546608301,-.000669676850830108,-1.13701881137203e-005,-.0010734776072168,-.01168375590964,-.0362627020233444,-.0663918477213734,-.0867937207648416,-.0867262195724516,-.067421744603583,-.040113315413826,-.017261017846082,-.00471435570086283,-.000588341168993646,-9.98922055945599e-006,-.000771510993319342,-.00839714407356662,-.0260620930240078,-.0477159840498455,-.06237886032367,-.0623303470508818,-.0484561734690455,-.0288295383268322,-.0124055359279088,-.00338821902309026,-.000422842243423972,-7.17927735473982e-006,-.000354187662792186,-.00385498697918716,-.0119646666030396,-.0219056021427923,-.0286370809191255,-.0286148092951926,-.0222454106000665,-.0132351539871473,-.00569517196000836,-.00155547411146173,-.000194119730276028,-3.29588494374136e-006},
			{-3.92294663599167e-008,-4.26974448589619e-007,-1.32519433994785e-006,-2.4262422795302e-006,-3.1718140426088e-006,-3.16934726012697e-006,-2.46387911966847e-006,-1.46591223424896e-006,-6.30791470989705e-007,-1.72282735226472e-007,-2.15005044744533e-008,-3.65048874112027e-010,-2.40449409138482e-006,-2.61705711055768e-005,-8.12252180824206e-005,-.000148711817078381,-.000194410192951716,-.000194258996301528,-.00015101869423293,-8.98502486218331e-005,-3.86631403796111e-005,-1.05597362732147e-005,-1.31783174148499e-006,-2.23749630651595e-008,-2.07281778267435e-005,-.000225605982417451,-.000700209981991998,-.00128198484678476,-.00167593219100564,-.00167462878541062,-.00130187150820567,-.000774562906136249,-.000333299404644144,-9.10312452249138e-005,-1.1360498152647e-005,-1.92885569959679e-007,-8.46234324557744e-005,-.000921043459502259,-.00285862908989082,-.00523374312001741,-.00684204543937499,-.00683672424538222,-.00531493103550084,-.00316217722165629,-.00136070521452511,-.000371637897746933,-4.63795880236092e-005,-7.87461355147049e-007,-.000228214966233319,-.00248389714184165,-.00770923516443232,-.0141145126680233,-.0184518297544813,-.0184374794017228,-.0143334626308517,-.00852785271078683,-.00366958991823735,-.00100224403364473,-.00012507783964276,-2.12364898657175e-006,-.000466278651886231,-.00507498798101565,-.0157511658366454,-.0288381434728016,-.0376999565136218,-.0376706365121829,-.0292854922824927,-.0174237287374253,-.00749754263838111,-.00204774035893469,-.000255553469660892,-4.3389449994523e-006,-.000771518583621402,-.00839722668659549,-.0260623494288551,-.0477164534906379,-.0623794740212014,-.0623309602711284,-.0484566501919911,-.0288298219584409,-.0124056579764159,-.00338825235716593,-.000422846403442921,-7.17934798611665e-006,-.00106842395843715,-.0116287518756503,-.0360919868089839,-.0660792924543047,-.0863851188731245,-.0863179354585031,-.0671043408542665,-.0399244725296093,-.0171797575373556,-.00469216176055673,-.000585571414309697,-9.94219394991092e-006,-.00125195103863209,-.0136262649987847,-.0422916390211788,-.077429973529657,-.101223805813716,-.101145082058928,-.0786310982319312,-.0467824447922306,-.0201307871491398,-.00549815149985173,-.000686157151895624,-1.16500008667718e-005,-.00122641302441892,-.0133483086422873,-.0414289499502059,-.0758505125894862,-.0991589846571268,-.0990818667545695,-.0770271360622618,-.0458281496934892,-.0197201478250184,-.00538599705705277,-.000672160525384795,-1.14123574777427e-005,-.000948266016470748,-.0103209499661349,-.0320329811846719,-.0586478306968206,-.0766700071720531,-.0766103793918513,-.0595575992912496,-.0354344548588437,-.015247674029803,-.00416446814598573,-.000519716417833643,-8.82406705447926e-006,-.00045303856676568,-.00493088257841441,-.0153039080100614,-.028019278030959,-.0366294579367253,-.0366009704831913,-.0284539242725722,-.0169289781185245,-.00728464826218446,-.00198959431997835,-.000248296972548114,-4.21573969958762e-006},
			{-5.21125166412433e-006,-.000294934791565546,-.00219672054164462,-.00721329026542284,-.0144452225630722,-.0199731094854017,-.0199731094854017,-.0144452225630722,-.00721329026542284,-.00219672054164462,-.000294934791565546,-5.21125166412433e-006,-1.18131144881787e-005,-.000668572290088461,-.00497963117683019,-.0163514313324231,-.032745121315009,-.0452759997488288,-.0452759997488288,-.032745121315009,-.0163514313324231,-.00497963117683019,-.000668572290088461,-1.18131144881787e-005,-1.76831482201068e-005,-.00100079135890211,-.0074540508575846,-.0244765920241849,-.049016441369304,-.0677740163420273,-.0677740163420273,-.049016441369304,-.0244765920241849,-.0074540508575846,-.00100079135890211,-1.76831482201068e-005,-2.24430094120152e-005,-.00127017935990404,-.00946049490012139,-.0310650783636037,-.0622104413366804,-.0860170863084584,-.0860170863084584,-.0622104413366804,-.0310650783636037,-.00946049490012139,-.00127017935990404,-2.24430094120152e-005,-2.57928908857684e-005,-.00145976847551561,-.0108725843403854,-.0357019044050888,-.0714960768359992,-.0988561418273082,-.0988561418273082,-.0714960768359992,-.0357019044050888,-.0108725843403854,-.00145976847551561,-2.57928908857684e-005,-2.75221754706548e-005,-.00155763866515012,-.0116015368483061,-.0380955388841488,-.0762895319046813,-.105483952681873,-.105483952681873,-.0762895319046813,-.0380955388841488,-.0116015368483061,-.00155763866515012,-2.75221754706548e-005,-2.75221754706548e-005,-.00155763866515012,-.0116015368483061,-.0380955388841488,-.0762895319046813,-.105483952681873,-.105483952681873,-.0762895319046813,-.0380955388841488,-.0116015368483061,-.00155763866515012,-2.75221754706548e-005,-2.57928908857684e-005,-.00145976847551561,-.0108725843403854,-.0357019044050888,-.0714960768359992,-.0988561418273082,-.0988561418273082,-.0714960768359992,-.0357019044050888,-.0108725843403854,-.00145976847551561,-2.57928908857684e-005,-2.24430094120152e-005,-.00127017935990404,-.00946049490012139,-.0310650783636037,-.0622104413366804,-.0860170863084584,-.0860170863084584,-.0622104413366804,-.0310650783636037,-.00946049490012139,-.00127017935990404,-2.24430094120152e-005,-1.76831482201068e-005,-.00100079135890211,-.0074540508575846,-.0244765920241849,-.049016441369304,-.0677740163420273,-.0677740163420273,-.049016441369304,-.0244765920241849,-.0074540508575846,-.00100079135890211,-1.76831482201068e-005,-1.18131144881787e-005,-.000668572290088461,-.00497963117683019,-.0163514313324231,-.032745121315009,-.0452759997488288,-.0452759997488288,-.032745121315009,-.0163514313324231,-.00497963117683019,-.000668572290088461,-1.18131144881787e-005,-5.21125166412433e-006,-.000294934791565546,-.00219672054164462,-.00721329026542284,-.0144452225630722,-.0199731094854017,-.0199731094854017,-.0144452225630722,-.00721329026542284,-.00219672054164462,-.000294934791565546,-5.21125166412433e-006},
			{-8.32182650623191e-008,-4.7098016450767e-006,-3.50793406430393e-005,-.000115188737748337,-.000230675169348194,-.000318949769921382,-.000318949769921382,-.000230675169348194,-.000115188737748337,-3.50793406430393e-005,-4.7098016450767e-006,-8.32182650623191e-008,-9.80924309227986e-007,-5.55161648928631e-005,-.000413493095087696,-.00135777203384397,-.00271905309464788,-.00375957829094687,-.00375957829094687,-.00271905309464788,-.00135777203384397,-.000413493095087696,-5.55161648928631e-005,-9.80924309227986e-007,-3.52372323574208e-006,-.000199427823688284,-.00148536967967095,-.00487745366231389,-.00976751261916201,-.0135053370130334,-.0135053370130334,-.00976751261916201,-.00487745366231389,-.00148536967967095,-.000199427823688284,-3.52372323574208e-006,-8.02097749900636e-006,-.000453953383811275,-.00338111593370866,-.011102445754264,-.0222335846768227,-.0307419161639197,-.0307419161639197,-.0222335846768227,-.011102445754264,-.00338111593370866,-.000453953383811275,-8.02097749900635e-006,-1.41209366639782e-005,-.000799185259152228,-.00595245703648738,-.0195458637467361,-.039142241836921,-.0541211655479415,-.0541211655479415,-.039142241836921,-.0195458637467361,-.00595245703648738,-.000799185259152228,-1.41209366639782e-005,-2.08499769658255e-005,-.00118002046473714,-.00878897732169731,-.0288600408453219,-.0577946675996675,-.0799114875939334,-.0799114875939334,-.0577946675996675,-.0288600408453219,-.00878897732169731,-.00118002046473714,-2.08499769658255e-005,-2.68198292857976e-005,-.0015178888432285,-.0113054739461356,-.0371233680459864,-.0743426777493866,-.102792077840484,-.102792077840484,-.0743426777493866,-.0371233680459864,-.0113054739461356,-.0015178888432285,-2.68198292857976e-005,-3.05536608257741e-005,-.00172920790781427,-.0128794114512791,-.0422916486119502,-.0846925957890945,-.117102694743021,-.117102694743021,-.0846925957890945,-.0422916486119502,-.0128794114512791,-.00172920790781427,-3.0553660825774e-005,-3.08514550786741e-005,-.00174606180234275,-.0130049418986058,-.0427038483144384,-.0855180604832598,-.11824404764644,-.11824404764644,-.0855180604832598,-.0427038483144384,-.0130049418986058,-.00174606180234275,-3.08514550786741e-005,-2.7104387920297e-005,-.00153399365776509,-.0114254251283075,-.037517247321153,-.0751314542415842,-.103882702728358,-.103882702728358,-.0751314542415842,-.037517247321153,-.0114254251283075,-.00153399365776509,-2.7104387920297e-005,-1.94799901806221e-005,-.00110248501011246,-.00821148110642986,-.0269637378113508,-.0539971607248856,-.0746607536401705,-.0746607536401705,-.0539971607248856,-.0269637378113508,-.00821148110642986,-.00110248501011246,-1.94799901806221e-005,-8.94293438853622e-006,-.00050613224228358,-.00376975224764522,-.0123785964921314,-.0247891842373852,-.0342754906451081,-.0342754906451081,-.0247891842373852,-.0123785964921314,-.00376975224764522,-.00050613224228358,-8.94293438853622e-006},
			{-1.00782323868636e-007,-2.96594300874774e-005,-.000530130419242249,-.00312208933690556,-.00957752805114946,-.0183245877738932,-.0235713601331056,-.0207230264243475,-.0120086359724961,-.00407774946276834,-.000589001007959142,-1.08304313867452e-005,-2.28458191424719e-007,-6.72334145152776e-005,-.00120172498658768,-.00707730141552942,-.0217107986092594,-.0415390518964802,-.0534326863949788,-.0469759473290577,-.0272217503074272,-.00924363749104905,-.00133517565243957,-2.45509398075641e-005,-3.41980945423285e-007,-.000100642251068274,-.00179887201456491,-.0105940706876594,-.0324990729725514,-.0621801483718135,-.0799838276575082,-.0703186818540462,-.0407484618877329,-.0138368769735303,-.00199863541368309,-3.67505036876191e-005,-4.34033661954923e-007,-.00012773262768919,-.00228308336566367,-.0134457295270728,-.0412470105167922,-.0789174890001959,-.101513473425809,-.0892467121265429,-.0517169285858085,-.0175614181527209,-.00253661807513852,-4.66428200392335e-005,-4.98818259086074e-007,-.00014679821533987,-.00262386024318778,-.0154526618157401,-.0474036089408846,-.0906968466390758,-.116665545847229,-.102567827047345,-.0594362846564907,-.020182665073873,-.00291523797142323,-5.3604806102002e-005,-5.32261533433169e-007,-.000156640302913549,-.00279977697510962,-.0164886856562537,-.0505817842982924,-.0967776174795907,-.124487388343073,-.109444487868491,-.0634211908577612,-.0215358120223382,-.00311069012556852,-5.7198740774876e-005,-5.32261533433169e-007,-.000156640302913549,-.00279977697510962,-.0164886856562537,-.0505817842982924,-.0967776174795907,-.124487388343073,-.109444487868491,-.0634211908577612,-.0215358120223382,-.00311069012556852,-5.7198740774876e-005,-4.98818259086074e-007,-.00014679821533987,-.00262386024318778,-.0154526618157401,-.0474036089408846,-.0906968466390758,-.116665545847229,-.102567827047345,-.0594362846564907,-.020182665073873,-.00291523797142323,-5.3604806102002e-005,-4.34033661954923e-007,-.00012773262768919,-.00228308336566367,-.0134457295270728,-.0412470105167922,-.0789174890001959,-.101513473425809,-.0892467121265429,-.0517169285858086,-.0175614181527209,-.00253661807513852,-4.66428200392335e-005,-3.41980945423285e-007,-.000100642251068274,-.00179887201456491,-.0105940706876594,-.0324990729725514,-.0621801483718135,-.0799838276575082,-.0703186818540462,-.0407484618877329,-.0138368769735303,-.00199863541368309,-3.67505036876191e-005,-2.28458191424719e-007,-6.72334145152776e-005,-.00120172498658768,-.00707730141552942,-.0217107986092594,-.0415390518964802,-.0534326863949788,-.0469759473290577,-.0272217503074272,-.00924363749104905,-.00133517565243957,-2.45509398075641e-005,-1.00782323868636e-007,-2.96594300874774e-005,-.000530130419242249,-.00312208933690556,-.00957752805114946,-.0183245877738932,-.0235713601331056,-.0207230264243475,-.0120086359724961,-.00407774946276834,-.000589001007959142,-1.08304313867452e-005},
			{-.00572678909872914,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442027,-.00553806002210424,-.00260199447163722,-.000920334332232957,-.000222779313164588,-3.04257537549422e-005,-1.47072954118751e-006,-4.61452526590166e-009,-.0129817593993148,-.0261104821110847,-.0313889764798934,-.0287370569429126,-.0211332958986598,-.0125539393028941,-.00589832550261673,-.00208625787714596,-.000505006800981909,-6.89705536523267e-005,-3.33392005816974e-006,-1.0460426551641e-008,-.019432502397701,-.0390849953863391,-.0469864170136373,-.0430167368512079,-.0316346044160784,-.018792094977277,-.00882925194855102,-.0031229365722185,-.000755948833210871,-.000103242588927534,-4.99057235089002e-006,-1.56582985243485e-008,-.0246632459775863,-.0496056985105609,-.0596339852015086,-.0545957664330435,-.0401498486478861,-.023850456898108,-.0112058657268432,-.00396355298338403,-.000959431350547771,-.000131032911457859,-6.33390959846238e-006,-1.98731208257652e-008,-.028344523709382,-.0570099288159739,-.0685350545084464,-.0627448227009384,-.0461426828391663,-.0274104163597275,-.0128784721632876,-.00455515959306486,-.00110263769366935,-.000150591105035483,-7.2793196382121e-006,-2.2839416390577e-008,-.0302448825382416,-.0608321599696494,-.073129987810444,-.0669515498630354,-.0492363193955517,-.0292481479535269,-.0137419094370501,-.00486056030603151,-.00117656404705184,-.000160687486930486,-7.76736168416077e-006,-2.43706852532627e-008,-.0302448825382416,-.0608321599696494,-.073129987810444,-.0669515498630354,-.0492363193955517,-.0292481479535269,-.0137419094370501,-.00486056030603151,-.00117656404705184,-.000160687486930486,-7.76736168416077e-006,-2.43706852532627e-008,-.028344523709382,-.0570099288159739,-.0685350545084464,-.0627448227009384,-.0461426828391663,-.0274104163597275,-.0128784721632876,-.00455515959306486,-.00110263769366935,-.000150591105035483,-7.2793196382121e-006,-2.2839416390577e-008,-.0246632459775863,-.0496056985105609,-.0596339852015086,-.0545957664330435,-.0401498486478861,-.023850456898108,-.0112058657268432,-.00396355298338403,-.000959431350547771,-.000131032911457859,-6.33390959846238e-006,-1.98731208257652e-008,-.019432502397701,-.0390849953863391,-.0469864170136373,-.0430167368512079,-.0316346044160784,-.018792094977277,-.00882925194855102,-.0031229365722185,-.000755948833210871,-.000103242588927534,-4.99057235089002e-006,-1.56582985243485e-008,-.0129817593993148,-.0261104821110847,-.0313889764798934,-.0287370569429126,-.0211332958986598,-.0125539393028941,-.00589832550261673,-.00208625787714596,-.000505006800981909,-6.89705536523267e-005,-3.33392005816974e-006,-1.0460426551641e-008,-.00572678909872914,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442027,-.00553806002210424,-.00260199447163722,-.000920334332232957,-.000222779313164588,-3.04257537549422e-005,-1.47072954118751e-006,-4.61452526590166e-009},
			{-9.14508612930534e-005,-.0001839370153449,-.000221121717472685,-.000202440095184293,-.000148874898420579,-8.84370543742305e-005,-4.15511434782526e-005,-1.46967813742169e-005,-3.55755375585554e-006,-4.85867619430457e-007,-2.34860199933996e-008,-7.36891655603157e-011,-.00107796495006253,-.00216813327679371,-.00260644304233894,-.00238623588679712,-.00175484320401577,-.00104244009902437,-.000489778616311158,-.000173236369522755,-4.19341950699251e-005,-5.72710039808159e-006,-2.76838358998326e-007,-8.68601307305661e-010,-.00387231726863855,-.00778847209081444,-.00936298940145173,-.00857195072154575,-.00630383171759815,-.00374470319905967,-.00175940617887584,-.000622307975060138,-.000150638021863614,-2.05731640619964e-005,-9.94471998471375e-007,-3.12023117416446e-009,-.00881447480486433,-.0177287361159355,-.0213127769373929,-.0195121521357594,-.0143492802872553,-.00852398956746932,-.00400489948507659,-.0014165466273796,-.000342894178409679,-4.68302630441653e-005,-2.26369580965429e-006,-7.10251695871362e-009,-.0155178892424445,-.0312114527009358,-.0375211591484582,-.0343511578883119,-.0252619182805468,-.0150064897727934,-.00705062842794049,-.00249383135774626,-.000603665448054148,-8.24447118179819e-005,-3.98523810328615e-006,-1.25039862212867e-008,-.0229126184021882,-.0460846249345851,-.055401091485105,-.0507204916900297,-.0372999629412389,-.0221575220926239,-.01041045957611,-.00368221511100628,-.00089132973162475,-.000121732034018965,-5.88432089415219e-006,-1.84625022332888e-008,-.0294730548165372,-.0592797668540085,-.0712637629438524,-.0652429943039245,-.0479798438190263,-.0285017561839274,-.0133912257589351,-.00473652229125437,-.00114653897597652,-.000156586857450866,-7.56914418191853e-006,-2.37487628354249e-008,-.03357626593249,-.0675326405320358,-.0811850373450308,-.074326062928209,-.0546595527845674,-.0324697440130155,-.0152555396799836,-.00539593649305364,-.00130615905948616,-.000178386733239242,-8.62291335307587e-006,-2.70550433775344e-008,-.0339035203026146,-.0681908540387014,-.0819763152766897,-.0750504891927709,-.055192297478506,-.0327862135586896,-.0154042292942165,-.00544852851748593,-.0013188896669085,-.000180125397036526,-8.7069574240775e-006,-2.7318738012221e-008,-.0297857641982319,-.0599087257235696,-.0720198721217481,-.0659352230713762,-.0484889104016337,-.0288041600782291,-.0135333067869547,-.00478677683617009,-.00115870376501852,-.000158248245443382,-7.64945287105236e-006,-2.40007374267426e-008,-.0214071018984119,-.0430565483441448,-.0517608590117147,-.0473878067921877,-.0348490990193348,-.0207016206060418,-.00972642083254007,-.0034402682709391,-.000832763242290672,-.000113733402739168,-5.49768056941236e-006,-1.72493889467441e-008,-.00982763882071387,-.0197665339288291,-.0237625359019177,-.0217549415081703,-.0159986326040109,-.00950376427809313,-.00446523548185483,-.00157936904180661,-.000382307535472911,-5.22130837362085e-005,-2.52389226922161e-006,-7.91890304680574e-009},
			{-1.08849930992072e-006,-2.18932125341295e-006,-2.63191437974773e-006,-2.40955526052686e-006,-1.77199232357178e-006,-1.05262729400979e-006,-4.94564953932593e-007,-1.74929313488118e-007,-4.23439949444061e-008,-5.78306820717772e-009,-2.79543748348947e-010,-8.77089671184216e-013,-6.67174550672731e-005,-.000134190202070998,-.000161318089751033,-.000147689018597559,-.000108610834338575,-6.45187493924991e-005,-3.03134001015112e-005,-1.07219531571797e-005,-2.5953930832331e-006,-3.5446195486468e-007,-1.71340921393568e-008,-5.37595110937315e-011,-.000575144383900619,-.0011567998362318,-.00139065846034338,-.00127316771187585,-.000936290380644839,-.000556190225361124,-.000261319647275767,-9.24296518291858e-005,-2.23738413632665e-005,-3.05567414766182e-006,-1.47706126638825e-007,-4.63439153301422e-010,-.00234804488509053,-.00472267141011844,-.00567740653672345,-.00519774689175974,-.00382243468035404,-.00227066394170372,-.00106684561013735,-.000377346936322897,-9.13419051691941e-005,-1.24748850093275e-005,-6.03014868716438e-007,-1.89200479726517e-009,-.00633227663561512,-.0127362394210907,-.0153109972435626,-.0140174369789623,-.0103084544811151,-.00612359342731638,-.00287710067800549,-.00101764033710382,-.000246333541419145,-3.36426374888076e-005,-1.62622826690742e-006,-5.1024142886996e-009,-.012937825514943,-.0260221169776175,-.0312827474533958,-.0286398027496958,-.0210617749476202,-.0125114532807165,-.00587836392863835,-.00207919740024719,-.000503297717511878,-6.87371381162648e-005,-3.32263714544313e-006,-1.04250255588739e-008,-.0214073125073398,-.0430569719462525,-.0517613682492133,-.0473882730064016,-.0348494418743096,-.0207018242742517,-.00972651652373047,-.00344030211724,-.000832771435243388,-.000113734521679514,-5.49773465709859e-006,-1.7249558650954e-008,-.0296455406961615,-.0596266912835726,-.071680821606866,-.0656248174753565,-.0482606374326126,-.0286685577088058,-.0134695955569983,-.00476424195651651,-.00115324889407038,-.000157503254546897,-7.61344126955836e-006,-2.388774830107e-008,-.0347378633474854,-.0698689855213075,-.0839936977751664,-.076897431725972,-.0565505431450506,-.0335930604290494,-.0157833171133724,-.00558261317396905,-.00135134666284035,-.000184558163040781,-8.92122984487403e-006,-2.79910339523387e-008,-.0340292605183584,-.068443757944315,-.0822803462291326,-.0753288338784716,-.0553969927824272,-.0329078099454216,-.0154613599726102,-.00546873583357464,-.00132378111976588,-.000180793439944228,-8.73924948977462e-006,-2.74200568127716e-008,-.0263115204035591,-.0529209070727989,-.0636193962384749,-.0582444672431377,-.0428331113779068,-.0254444116512108,-.0119547672264766,-.00422844200182213,-.00102355130296691,-.000139789998708784,-6.75721240365912e-006,-2.12012654199985e-008,-.0125704531070497,-.0252832132289538,-.0303944669425635,-.0278265692364139,-.0204637212045026,-.0121561878064161,-.00571144648888444,-.00202015813166747,-.000489006467861309,-6.67853319250079e-005,-3.22829013115797e-006,-1.0129004659729e-008},
			{-.000354187662792186,-.00370432632910279,-.0106866754670937,-.0175473588745023,-.0197676016789817,-.0162489801425508,-.00982031115806883,-.00422235247914266,-.00118608291592308,-.000180620480252622,-9.39296633663941e-006,-3.06697796211909e-008,-.000802889532215207,-.00839714407356662,-.024225067014107,-.0397771922581837,-.0448101448253052,-.0368339652566651,-.0222611509665594,-.00957143052386794,-.00268866947546018,-.000409439142389331,-2.1292425288575e-005,-6.95237231559063e-008,-.00120185194317225,-.0125697540159339,-.0362627020233443,-.0595428061909519,-.0670766742761412,-.0551370654893481,-.0333228999419048,-.0143275530598965,-.00402469144754472,-.000612892819182374,-3.18728065084108e-005,-1.04070632906319e-007,-.00152536042432723,-.0159532173895116,-.0460237143683289,-.0755702402721397,-.085132037201067,-.0699785843745455,-.0422925910998948,-.0181841719682459,-.00510803771553511,-.000777868235805777,-4.04521687853206e-005,-1.32083844163853e-007,-.00175303829641842,-.0183344215497376,-.0528932916734667,-.0868499819149511,-.0978389887959829,-.0804237060180922,-.0486052546469995,-.0208983721752537,-.00587047204848792,-.000893974161901044,-4.64901277906996e-005,-1.51798901731387e-007,-.00187057076364506,-.0195636529956887,-.0564395228555362,-.0926728396779039,-.104398604617067,-.0858157140598609,-.0518639943508112,-.0222995037123093,-.00626405789601532,-.000953910667053138,-4.96070587967628e-005,-1.61976260365957e-007,-.00187057076364506,-.0195636529956887,-.0564395228555362,-.0926728396779039,-.104398604617067,-.0858157140598609,-.0518639943508112,-.0222995037123093,-.00626405789601532,-.000953910667053138,-4.96070587967628e-005,-1.61976260365957e-007,-.00175303829641842,-.0183344215497376,-.0528932916734667,-.0868499819149511,-.0978389887959829,-.0804237060180922,-.0486052546469995,-.0208983721752537,-.00587047204848792,-.000893974161901044,-4.64901277906996e-005,-1.51798901731387e-007,-.00152536042432723,-.0159532173895116,-.0460237143683289,-.0755702402721397,-.085132037201067,-.0699785843745455,-.0422925910998948,-.0181841719682459,-.00510803771553511,-.000777868235805777,-4.04521687853206e-005,-1.32083844163853e-007,-.00120185194317225,-.0125697540159339,-.0362627020233444,-.0595428061909519,-.0670766742761412,-.0551370654893481,-.0333228999419048,-.0143275530598965,-.00402469144754472,-.000612892819182374,-3.18728065084108e-005,-1.04070632906319e-007,-.000802889532215207,-.00839714407356662,-.024225067014107,-.0397771922581837,-.0448101448253052,-.0368339652566651,-.0222611509665594,-.00957143052386794,-.00268866947546018,-.000409439142389331,-2.1292425288575e-005,-6.95237231559063e-008,-.000354187662792186,-.00370432632910279,-.0106866754670937,-.0175473588745023,-.0197676016789817,-.0162489801425508,-.00982031115806883,-.00422235247914266,-.00118608291592308,-.000180620480252622,-9.39296633663941e-006,-3.06697796211909e-008},
			{-5.6560083280767e-006,-5.91542359272188e-005,-.000170655084270396,-.00028021305752868,-.000315668024236135,-.00025947930056204,-.000156820147987287,-6.74265743697526e-005,-1.89405096647495e-005,-2.8843210756594e-006,-1.49995895979437e-007,-4.89764458734161e-010,-6.66694512082365e-005,-.000697272743806459,-.00201157426832706,-.00330297440936967,-.00372089514708473,-.00305857798732252,-.00184849678399644,-.000794781840714042,-.000223258756300761,-3.3998553762376e-005,-1.76805681469661e-006,-5.77303387673915e-009,-.000239493192417188,-.00250477651118074,-.00722607333006517,-.0118651027035077,-.0133663775728643,-.0109871701831305,-.00664025858844208,-.00285505335439599,-.000802000786155484,-.000122131231479417,-6.35129828208852e-006,-2.07381685017084e-008,-.00054515334463801,-.00570157038226907,-.0164485595800252,-.0270082851124352,-.030425605696751,-.0250098656792164,-.0151150817366226,-.00649889823405786,-.00182557761480488,-.000278004767708733,-1.44573274351898e-005,-4.72058591990341e-008,-.000959742855872972,-.0100376187644472,-.0289577009874395,-.0475481053927164,-.0535643007426736,-.044029886541206,-.0266101122837137,-.0114413150217837,-.00321393070570632,-.000489427593779599,-2.54521353623046e-005,-8.31059491191454e-008,-.00141708846333921,-.0148208383771267,-.0427568944566942,-.0702061715732879,-.0790892603834303,-.0650114182959186,-.0392906109117239,-.0168934370530446,-.00474546290931794,-.000722654190693695,-3.75808240390211e-005,-1.22708370279532e-007,-.00182283513943086,-.0190644026033435,-.0549992267137261,-.0903078952792718,-.101734427107537,-.0836257585887017,-.0505404624146264,-.0217304363719754,-.00610420363142845,-.000929567551026009,-4.83411222371217e-005,-1.57842742379509e-007,-.0020766085420598,-.0217185308967278,-.062656167598386,-.102880475966099,-.115897798864594,-.0952680584574148,-.0575766583052787,-.0247557273922317,-.00695402514977553,-.00105898107575704,-5.50711280460104e-005,-1.79817461292618e-007,-.00209684841095391,-.0219302126889477,-.0632668520831664,-.103883210618848,-.11702740813212,-.0961965979360501,-.0581378348543761,-.0249970115180785,-.00702180323816678,-.00106930253870999,-5.56078841986148e-005,-1.81570069821245e-007,-.00184217543696469,-.0192666760896712,-.0555827690131719,-.0912660628757668,-.102813830311442,-.0845130286536747,-.0510766971839986,-.0219609965010198,-.00616896928789685,-.000939430271260956,-4.88540219870558e-005,-1.59517455322627e-007,-.00132397601187602,-.0138470074345037,-.0399474726296093,-.0655931435851624,-.0738925524084332,-.0607397213008226,-.0367089477367546,-.0157834221327753,-.00443365338137005,-.000675170843678735,-3.51115056126793e-005,-1.14645586997212e-007,-.000607815019228362,-.00635692717584237,-.0183392097943569,-.0301127040609837,-.0339228224379417,-.0278845798860605,-.0168524577290889,-.00724590244919067,-.00203541536332734,-.000309959527704379,-1.61190990378048e-005,-5.2631852118241e-008},
			{-1.08304313867452e-005,-.000589001007959142,-.00407774946276834,-.0120086359724961,-.0207230264243475,-.0235713601331056,-.0183245877738932,-.00957752805114946,-.00312208933690556,-.000530130419242249,-2.96594300874774e-005,-1.00782323868636e-007,-2.45509398075641e-005,-.00133517565243957,-.00924363749104905,-.0272217503074272,-.0469759473290577,-.0534326863949788,-.0415390518964802,-.0217107986092594,-.00707730141552942,-.00120172498658768,-6.72334145152776e-005,-2.28458191424719e-007,-3.67505036876191e-005,-.00199863541368309,-.0138368769735303,-.0407484618877329,-.0703186818540462,-.0799838276575082,-.0621801483718135,-.0324990729725514,-.0105940706876594,-.00179887201456491,-.000100642251068273,-3.41980945423285e-007,-4.66428200392335e-005,-.00253661807513852,-.0175614181527209,-.0517169285858086,-.0892467121265429,-.101513473425809,-.0789174890001959,-.0412470105167922,-.0134457295270728,-.00228308336566367,-.00012773262768919,-4.34033661954923e-007,-5.3604806102002e-005,-.00291523797142323,-.020182665073873,-.0594362846564907,-.102567827047345,-.116665545847229,-.0906968466390758,-.0474036089408846,-.0154526618157401,-.00262386024318778,-.00014679821533987,-4.98818259086074e-007,-5.7198740774876e-005,-.00311069012556852,-.0215358120223382,-.0634211908577612,-.109444487868491,-.124487388343073,-.0967776174795907,-.0505817842982924,-.0164886856562537,-.00279977697510962,-.000156640302913549,-5.32261533433169e-007,-5.7198740774876e-005,-.00311069012556852,-.0215358120223382,-.0634211908577612,-.109444487868491,-.124487388343073,-.0967776174795907,-.0505817842982924,-.0164886856562537,-.00279977697510962,-.000156640302913549,-5.32261533433169e-007,-5.3604806102002e-005,-.00291523797142323,-.020182665073873,-.0594362846564907,-.102567827047345,-.116665545847229,-.0906968466390758,-.0474036089408846,-.0154526618157401,-.00262386024318778,-.00014679821533987,-4.98818259086074e-007,-4.66428200392335e-005,-.00253661807513852,-.0175614181527209,-.0517169285858085,-.0892467121265429,-.101513473425809,-.0789174890001959,-.0412470105167922,-.0134457295270728,-.00228308336566367,-.00012773262768919,-4.34033661954923e-007,-3.67505036876191e-005,-.00199863541368309,-.0138368769735303,-.0407484618877329,-.0703186818540462,-.0799838276575082,-.0621801483718135,-.0324990729725514,-.0105940706876594,-.00179887201456491,-.000100642251068273,-3.41980945423285e-007,-2.45509398075641e-005,-.00133517565243957,-.00924363749104905,-.0272217503074272,-.0469759473290577,-.0534326863949788,-.0415390518964802,-.0217107986092594,-.00707730141552942,-.00120172498658768,-6.72334145152776e-005,-2.28458191424719e-007,-1.08304313867452e-005,-.000589001007959142,-.00407774946276834,-.0120086359724961,-.0207230264243475,-.0235713601331056,-.0183245877738932,-.00957752805114946,-.00312208933690556,-.000530130419242249,-2.96594300874774e-005,-1.00782323868636e-007},
			{-.00719309508160456,-.0139021976630662,-.0155346480848768,-.0127550186066991,-.00808303973670482,-.00395000387691709,-.00144276568155165,-.000368787100047311,-5.82757244194688e-005,-4.43762053833248e-006,-8.93862080339512e-008,-5.3934946534894e-011,-.0163056519239563,-.0315141665027775,-.0352146831870672,-.028913686156645,-.0183230210276625,-.00895405768790711,-.00327053024384089,-.000835984234769121,-.000132102199014488,-1.00594104551317e-005,-2.02624931058027e-007,-1.22262316118168e-010,-.0244080644511933,-.0471738149883672,-.0527131488435292,-.0432811345737185,-.0274278808519569,-.0134034025849063,-.00489568361654787,-.00125139167557278,-.00019774486802044,-1.50580142318092e-005,-3.03310925545804e-007,-1.83015466396657e-010,-.0309781048820443,-.0598718260236465,-.0669022099971482,-.0549313334090684,-.0348107803231648,-.017011263301183,-.00621347919028916,-.00158823501355166,-.000250972840295156,-1.91112550165974e-005,-3.84954639980414e-007,-2.32278652181071e-010,-.0356019491148408,-.0688083958532261,-.0768881468075034,-.0631304769703183,-.0400066961627327,-.0195503931804077,-.00714091358400096,-.00182529765297712,-.000288433470136975,-2.19638332045296e-005,-4.42413619435074e-007,-2.6694895594493e-010,-.0379888820906312,-.0734216553279677,-.0820431132524654,-.0673630603261273,-.0426889454411423,-.0208611494544942,-.00761967619488893,-.00194767474936783,-.000307771494551617,-2.3436398585193e-005,-4.72075244245615e-007,-2.84846550926105e-010,-.0379888820906312,-.0734216553279677,-.0820431132524654,-.0673630603261273,-.0426889454411423,-.0208611494544942,-.00761967619488893,-.00194767474936783,-.000307771494551617,-2.3436398585193e-005,-4.72075244245615e-007,-2.84846550926105e-010,-.0356019491148408,-.0688083958532261,-.0768881468075034,-.0631304769703183,-.0400066961627327,-.0195503931804077,-.00714091358400096,-.00182529765297712,-.000288433470136975,-2.19638332045296e-005,-4.42413619435074e-007,-2.6694895594493e-010,-.0309781048820443,-.0598718260236465,-.0669022099971482,-.0549313334090684,-.0348107803231648,-.017011263301183,-.00621347919028916,-.00158823501355166,-.000250972840295157,-1.91112550165974e-005,-3.84954639980414e-007,-2.32278652181071e-010,-.0244080644511933,-.0471738149883672,-.0527131488435292,-.0432811345737185,-.0274278808519569,-.0134034025849063,-.00489568361654787,-.00125139167557278,-.00019774486802044,-1.50580142318092e-005,-3.03310925545804e-007,-1.83015466396657e-010,-.0163056519239563,-.0315141665027775,-.0352146831870672,-.028913686156645,-.0183230210276625,-.00895405768790711,-.00327053024384089,-.00083598423476912,-.000132102199014488,-1.00594104551317e-005,-2.02624931058027e-007,-1.22262316118168e-010,-.00719309508160456,-.0139021976630662,-.0155346480848768,-.0127550186066991,-.00808303973670482,-.00395000387691709,-.00144276568155165,-.000368787100047311,-5.82757244194688e-005,-4.43762053833248e-006,-8.93862080339512e-008,-5.3934946534894e-011},
			{-.000114866241683583,-.000222003626892483,-.000248072161029854,-.000203684371377566,-.000129077731546852,-6.30774506425453e-005,-2.30394662644891e-005,-5.88914614407857e-006,-9.30602664014744e-007,-7.08641812006219e-008,-1.42740470669776e-009,-8.61284959198155e-013,-.00135397065407101,-.00261683843315572,-.00292411783656719,-.00240090262810037,-.00152148671399835,-.000743517119145248,-.000271574665892772,-6.94175324250859e-005,-1.09693560023246e-005,-8.35302177246445e-007,-1.68253444704445e-008,-1.015228271122e-011,-.00486379816401721,-.00940033222170924,-.0105041559963674,-.0086246372913827,-.00546555736926158,-.00267089776882591,-.00097556350825723,-.000249364981245768,-3.94046454582605e-005,-3.00061244597533e-006,-6.04408074120732e-008,-3.64695156892415e-011,-.0110713620291109,-.0213977796180793,-.0239103905886559,-.0196320814726822,-.0124411339215587,-.00607970872643953,-.00222065480887151,-.000567624291069279,-8.96959702655164e-005,-6.8302313496928e-006,-1.37580145726718e-007,-8.30147957637377e-011,-.0194911408262166,-.0376708064290284,-.0420942598523953,-.0345622935723416,-.0219026252294268,-.0107033315917073,-.0039094643904085,-.00099930297324955,-.000157909820254493,-1.20246272104963e-005,-2.42209945641729e-007,-1.46147607732105e-010,-.0287792408488707,-.0556220500820546,-.0621534087434013,-.0510322397173339,-.0323398682673508,-.015803782883183,-.00577243878564134,-.00147550013642398,-.000233158479025165,-1.77547146005579e-005,-3.57630085574026e-007,-2.15791227404772e-010,-.0370194330577289,-.0715479873273676,-.0799494318272223,-.0656440033259579,-.041599554161483,-.0203287878778718,-.00742522752170763,-.00189797148624789,-.000299917386683,-2.28383184972103e-005,-4.60028222490692e-007,-2.77577471181613e-010,-.0421732438918713,-.0815088311812927,-.0910799169238503,-.0747829000510975,-.0473910051704331,-.0231589427061906,-.00845896075008184,-.00216220530078157,-.000341671604648038,-2.60178478303842e-005,-5.24072921213834e-007,-3.16221547012239e-010,-.0425842895511383,-.0823032650013445,-.0919676362701244,-.0755117789234986,-.0478529062519234,-.0233846636134322,-.00854140684095277,-.00218327944688257,-.00034500174046461,-2.62714333369098e-005,-5.29180849357959e-007,-3.19303631345124e-010,-.0374122095817176,-.072307112128577,-.0807976960315661,-.0663404868028083,-.0420409258123548,-.0205444765035491,-.00750400925375776,-.00191810898111,-.000303099512904151,-2.30806332657561e-005,-4.64909126147018e-007,-2.80522570694571e-010,-.026888246929992,-.0519672991102317,-.058069502618547,-.0476790708315039,-.0302149166554566,-.0147653657308921,-.00539314988437847,-.00137854963658506,-.000217838364492369,-1.65881078259945e-005,-3.34131331017576e-007,-2.01612260665772e-010,-.0123439399038752,-.0238573090634585,-.0266587276008582,-.0218886559078795,-.0138711577726375,-.00677852995457166,-.0024759039977092,-.000632868848341838,-.00010000591288281,-7.61532005623914e-006,-1.53394049110811e-007,-9.25567827468422e-011},
			{-.000538770430325433,-.00541459382780606,-.0145196103881163,-.0213815233314377,-.020756280852719,-.0140356226122099,-.00659447754217011,-.00204903932812416,-.000375745321421736,-3.19037343548278e-005,-6.91363066088356e-007,-4.34129792523551e-010,-.0012213105768994,-.0122740602291765,-.032913747194243,-.0484686596091265,-.0470513300576056,-.0318166205582331,-.0149486770580804,-.00464486033951284,-.000851757463741354,-7.23209107571181e-005,-1.56721486103326e-006,-9.84106174935942e-010,-.00182818984576065,-.0183731416903004,-.04926885883459,-.0725531351410981,-.0704315228803041,-.0476265609511484,-.0223767976156558,-.00695292963828415,-.00127500275172926,-.000108257766028679,-2.34597681315439e-006,-1.47311662585924e-009,-.00232029282368979,-.0233187318655643,-.0625308032698983,-.0920826243480438,-.0893899271345745,-.0604464398750644,-.0284000718224987,-.00882448438314045,-.00161820160082515,-.000137398048789906,-2.97745509129004e-006,-1.86964277444457e-009,-.00266662364773693,-.0267993251511533,-.0718642565322132,-.105827032314693,-.102732418569255,-.0694687775381858,-.0326391144885199,-.0101416418199092,-.00185973710365748,-.000157906313511514,-3.4218750648389e-006,-2.14870881134135e-009,-.00284540745247398,-.0285960861297693,-.0766823962117034,-.112922206580213,-.109620114430362,-.0741263123085118,-.0348274041920399,-.0108215882804389,-.00198442319330673,-.000168493143619897,-3.65129470714422e-006,-2.29276901154607e-009,-.00284540745247398,-.0285960861297693,-.0766823962117034,-.112922206580213,-.109620114430362,-.0741263123085118,-.0348274041920398,-.0108215882804389,-.00198442319330673,-.000168493143619897,-3.65129470714422e-006,-2.29276901154607e-009,-.00266662364773693,-.0267993251511533,-.0718642565322132,-.105827032314693,-.102732418569255,-.0694687775381858,-.0326391144885199,-.0101416418199092,-.00185973710365748,-.000157906313511514,-3.4218750648389e-006,-2.14870881134135e-009,-.00232029282368979,-.0233187318655643,-.0625308032698983,-.0920826243480438,-.0893899271345745,-.0604464398750644,-.0284000718224987,-.00882448438314046,-.00161820160082515,-.000137398048789906,-2.97745509129004e-006,-1.86964277444457e-009,-.00182818984576065,-.0183731416903004,-.04926885883459,-.0725531351410981,-.0704315228803041,-.0476265609511484,-.0223767976156558,-.00695292963828415,-.00127500275172926,-.000108257766028679,-2.34597681315439e-006,-1.47311662585924e-009,-.0012213105768994,-.0122740602291765,-.032913747194243,-.0484686596091265,-.0470513300576056,-.0318166205582331,-.0149486770580804,-.00464486033951284,-.000851757463741354,-7.23209107571182e-005,-1.56721486103326e-006,-9.84106174935942e-010,-.000538770430325433,-.00541459382780606,-.0145196103881163,-.0213815233314377,-.020756280852719,-.0140356226122099,-.00659447754217011,-.00204903932812416,-.000375745321421736,-3.19037343548278e-005,-6.91363066088356e-007,-4.34129792523551e-010},
			{-.00863095559575211,-.016029237481894,-.0166489373902391,-.0122597253338038,-.00669488358667141,-.00269138517084675,-.00076422934021594,-.000141170609267733,-1.45626016852063e-005,-6.18297441623807e-007,-5.18975291864507e-009,-6.02215524056038e-013,-.0195650628997474,-.0363358420848087,-.0377406074727463,-.0277909316794019,-.0151762822814405,-.00610096061448634,-.00173239161588508,-.000320012288242827,-3.30112019224449e-005,-1.40158620930416e-006,-1.1764380103799e-008,-1.36513094948138e-012,-.0292871035439603,-.0543914208171814,-.0564942256781886,-.0416004741639895,-.0227175017461691,-.0091325781138374,-.00259323125578963,-.000479029025826867,-4.94147396186346e-005,-2.09804592236745e-006,-1.76101973193715e-008,-2.0434757441563e-012,-.0371704592590807,-.0690322308074568,-.0717010581395466,-.0527982812554331,-.0288324849829067,-.0115908397087212,-.0032912642453023,-.00060797165761674,-6.27159515120887e-005,-2.66278740620364e-006,-2.23504219535243e-008,-2.59352830098302e-012,-.0427185847602993,-.0793360981199193,-.0824032791251857,-.0606790418511532,-.0331360703672624,-.0133209080116245,-.00378252390296041,-.00069871853362736,-7.20770403136054e-005,-3.06023954984572e-006,-2.56864836669144e-008,-2.98064271365472e-012,-.0455826526323854,-.0846551874810701,-.087928007662455,-.0647472687189207,-.0353576784818238,-.0142140084005345,-.00403612324963012,-.000745564123550858,-7.69094460834964e-005,-3.26541333602756e-006,-2.7408634178924e-008,-3.18047992928925e-012,-.0455826526323854,-.0846551874810701,-.087928007662455,-.0647472687189207,-.0353576784818238,-.0142140084005345,-.00403612324963012,-.000745564123550858,-7.69094460834964e-005,-3.26541333602756e-006,-2.7408634178924e-008,-3.18047992928925e-012,-.0427185847602993,-.0793360981199193,-.0824032791251857,-.0606790418511532,-.0331360703672624,-.0133209080116245,-.00378252390296041,-.00069871853362736,-7.20770403136054e-005,-3.06023954984572e-006,-2.56864836669144e-008,-2.98064271365472e-012,-.0371704592590807,-.0690322308074568,-.0717010581395466,-.0527982812554331,-.0288324849829067,-.0115908397087212,-.0032912642453023,-.00060797165761674,-6.27159515120887e-005,-2.66278740620364e-006,-2.23504219535243e-008,-2.59352830098302e-012,-.0292871035439603,-.0543914208171814,-.0564942256781886,-.0416004741639895,-.0227175017461691,-.0091325781138374,-.00259323125578962,-.000479029025826867,-4.94147396186346e-005,-2.09804592236745e-006,-1.76101973193715e-008,-2.0434757441563e-012,-.0195650628997474,-.0363358420848087,-.0377406074727463,-.0277909316794019,-.0151762822814405,-.00610096061448634,-.00173239161588507,-.000320012288242827,-3.30112019224449e-005,-1.40158620930416e-006,-1.1764380103799e-008,-1.36513094948138e-012,-.00863095559575211,-.016029237481894,-.0166489373902391,-.0122597253338038,-.00669488358667141,-.00269138517084675,-.00076422934021594,-.000141170609267733,-1.45626016852063e-005,-6.18297441623807e-007,-5.18975291864507e-009,-6.02215524056038e-013}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505},
			{-1.77695599756377e-005,-4.02808883404862e-005,-6.02967929985932e-005,-7.65271814689107e-005,-8.7949757770287e-005,-9.38463500146473e-005,-9.38463500146473e-005,-8.7949757770287e-005,-7.65271814689107e-005,-6.02967929985932e-005,-4.02808883404862e-005,-1.77695599756377e-005,-.000209456341481459,-.000474805651630584,-.000710740484393843,-.000902054045026543,-.00103669615466006,-.00110620145700885,-.00110620145700885,-.00103669615466006,-.000902054045026543,-.000710740484393843,-.000474805651630584,-.000209456341481459,-.000752419091267727,-.00170561957877169,-.00255315597328018,-.00324040169914516,-.00372406952729629,-.00397374979986132,-.00397374979986132,-.00372406952729629,-.00324040169914516,-.00255315597328018,-.00170561957877169,-.000752419091267727,-.00171271583978708,-.00388246617226489,-.00581169553993682,-.00737605861122963,-.00847702157201008,-.00904536355411474,-.00904536355411474,-.00847702157201008,-.00737605861122963,-.00581169553993682,-.00388246617226489,-.00171271583978708,-.00301523746950065,-.00683508449255492,-.0102314941838372,-.012985556497555,-.0149238025829595,-.0159243690517937,-.0159243690517937,-.0149238025829595,-.012985556497555,-.0102314941838372,-.00683508449255492,-.00301523746950065,-.00445208652099934,-.0100922026364427,-.0151071011176736,-.0191735548643253,-.0220354320327057,-.0235127977574022,-.0235127977574022,-.0220354320327057,-.0191735548643253,-.0151071011176736,-.0100922026364427,-.00445208652099934,-.00572682649263902,-.012981844165618,-.0194326292850773,-.0246634070198705,-.0283447087891094,-.0302450800266409,-.0302450800266409,-.0283447087891094,-.0246634070198705,-.0194326292850773,-.012981844165618,-.00572682649263902,-.00652410991880577,-.0147891643642708,-.022138021752572,-.0280970234695782,-.0322908326268278,-.0344557717700214,-.0344557717700214,-.0322908326268278,-.0280970234695782,-.022138021752572,-.0147891643642708,-.00652410991880577,-.0065876977962188,-.0149333084057324,-.0223537921535752,-.028370873865462,-.0326055583951231,-.034791598330081,-.034791598330081,-.0326055583951231,-.028370873865462,-.0223537921535752,-.0149333084057324,-.0065876977962188,-.00578758817420658,-.0131195816511808,-.0196388096598727,-.0249250859943669,-.0286454464091129,-.0305659806028896,-.0305659806028896,-.0286454464091129,-.0249250859943669,-.0196388096598727,-.0131195816511809,-.00578758817420658,-.00415955383809286,-.00942907556113789,-.0141144607455587,-.0179137205329708,-.0205875527021657,-.0219678453450519,-.0219678453450519,-.0205875527021657,-.0179137205329708,-.0141144607455587,-.0094290755611379,-.00415955383809286,-.00190958089376513,-.00432872929123383,-.00647971047247572,-.00822388645453697,-.00945139763053613,-.0100850666636237,-.0100850666636237,-.00945139763053613,-.00822388645453697,-.00647971047247572,-.00432872929123383,-.00190958089376513},
			{-2.11503243320183e-007,-4.79445666606905e-007,-7.17686160968121e-007,-9.10869324002267e-007,-1.0468272170579e-006,-1.11701175656978e-006,-1.11701175656978e-006,-1.0468272170579e-006,-9.10869324002267e-007,-7.17686160968121e-007,-4.79445666606905e-007,-2.11503243320183e-007,-1.29636812850387e-005,-2.93866926947107e-005,-4.39891819502224e-005,-5.58299694289237e-005,-6.41632449195998e-005,-6.84650702111951e-005,-6.84650702111951e-005,-6.41632449195998e-005,-5.58299694289237e-005,-4.39891819502224e-005,-2.93866926947107e-005,-1.29636812850387e-005,-.000111754689657294,-.00025333087492222,-.000379213069886165,-.000481287743035328,-.000553125564084201,-.000590209872148528,-.000590209872148528,-.000553125564084201,-.000481287743035328,-.000379213069886165,-.00025333087492222,-.000111754689657294,-.000456242005972593,-.0010342311978472,-.00154814918484807,-.00196487222152227,-.00225815236645889,-.00240955021073066,-.00240955021073066,-.00225815236645889,-.00196487222152227,-.00154814918484807,-.0010342311978472,-.000456242005972593,-.00123040688572486,-.0027891451698972,-.00417509434078899,-.00529892530561013,-.00608985184251931,-.00649814601017004,-.00649814601017004,-.00608985184251931,-.00529892530561013,-.00417509434078899,-.0027891451698972,-.00123040688572486,-.00251391253350483,-.00569865715294523,-.00853036675399556,-.0108265281139349,-.0124425139778682,-.0132767224314477,-.0132767224314477,-.0124425139778682,-.0108265281139349,-.00853036675399556,-.00569865715294523,-.00251391253350483,-.00415959476024,-.00942916832541804,-.0141145996051725,-.0179138967701193,-.0205877552447685,-.0219680614671252,-.0219680614671252,-.0205877552447685,-.0179138967701193,-.0141145996051725,-.00942916832541803,-.00415959476024,-.00576034173845761,-.0130578181276279,-.0195463553335665,-.0248077452759809,-.0285105912170028,-.030422084043298,-.030422084043298,-.0285105912170028,-.0248077452759809,-.0195463553335665,-.0130578181276279,-.00576034173845761,-.00674981664852096,-.0153008071730865,-.0229039040804791,-.0290690621631277,-.0334079594567994,-.0356477963776399,-.0356477963776399,-.0334079594567994,-.0290690621631277,-.0229039040804791,-.0153008071730865,-.00674981664852096,-.006612130023256,-.0149886925463945,-.0224366971291726,-.0284760947867864,-.0327264847688013,-.0349206322134925,-.0349206322134925,-.0327264847688013,-.0284760947867864,-.0224366971291726,-.0149886925463945,-.006612130023256,-.00511251762065263,-.0115892994367119,-.0173481176290686,-.0220177969659965,-.0253042105122013,-.02700073151736,-.02700073151736,-.0253042105122013,-.0220177969659965,-.0173481176290686,-.0115892994367119,-.00511251762065263,-.00244252943287481,-.005536842526707,-.00828814510932977,-.01051910646122,-.0120892060502712,-.0128997269708937,-.0128997269708937,-.0120892060502712,-.01051910646122,-.00828814510932977,-.005536842526707,-.00244252943287481},
			{-2.30726263295083e-009,-5.23021327582048e-009,-7.82914926217426e-009,-9.9365604128826e-009,-1.14197081952885e-008,-1.21853426266313e-008,-1.21853426266313e-008,-1.14197081952885e-008,-9.9365604128826e-009,-7.82914926217426e-009,-5.23021327582048e-009,-2.30726263295083e-009,-7.35364770593756e-007,-1.66696002908487e-006,-2.49528617544501e-006,-3.16695479923119e-006,-3.63965981910605e-006,-3.88368084208039e-006,-3.88368084208039e-006,-3.63965981910605e-006,-3.16695479923119e-006,-2.49528617544501e-006,-1.66696002908487e-006,-7.35364770593756e-007,-1.52128768774711e-005,-3.44852768261633e-005,-5.16212944637669e-005,-6.55164557289296e-005,-7.52955525177414e-005,-8.03437434652432e-005,-8.03437434652432e-005,-7.52955525177414e-005,-6.55164557289296e-005,-5.16212944637669e-005,-3.44852768261633e-005,-1.52128768774711e-005,-.000111389656582294,-.000252503400490955,-.000377974416605436,-.000479715675273885,-.000551318846834677,-.000588282023525919,-.000588282023525919,-.000551318846834677,-.000479715675273885,-.000377974416605436,-.000252503400490955,-.000111389656582294,-.000460167166116478,-.00104312893857298,-.00156146828610925,-.00198177649169201,-.00227757979653243,-.00243028015301575,-.00243028015301575,-.00227757979653243,-.00198177649169201,-.00156146828610925,-.00104312893857298,-.000460167166116478,-.00130099723581861,-.00294916275130836,-.00441462597427551,-.00560293286342162,-.0064392360816438,-.00687095471852507,-.00687095471852507,-.0064392360816438,-.00560293286342162,-.00441462597427551,-.00294916275130836,-.00130099723581861,-.00276903001105212,-.00627696965144704,-.00939604748863851,-.011925228449054,-.0137052081798638,-.0146240739767634,-.0146240739767634,-.0137052081798638,-.011925228449054,-.00939604748863851,-.00627696965144704,-.00276903001105212,-.00466138390221013,-.0105666479493299,-.0158173022080392,-.0200749243239431,-.0230713414195831,-.0246181596977759,-.0246181596977759,-.0230713414195831,-.0200749243239431,-.0158173022080392,-.0105666479493299,-.00466138390221013,-.00633855009047991,-.0143685284714563,-.0215083684256039,-.0272978832165217,-.0313724113504692,-.0334757749315177,-.0334757749315177,-.0313724113504692,-.0272978832165217,-.0215083684256039,-.0143685284714563,-.00633855009047991,-.00692348559220745,-.0156944882399467,-.0234932085068186,-.0298169926007543,-.0342675272542232,-.036564993905222,-.036564993905222,-.0342675272542232,-.0298169926007543,-.0234932085068186,-.0156944882399467,-.00692348559220746,-.00575920488574972,-.0130552410555424,-.0195424976931696,-.0248028492552804,-.0285049644079869,-.0304160799848247,-.0304160799848247,-.0285049644079869,-.0248028492552804,-.0195424976931696,-.0130552410555424,-.00575920488574972,-.00286339454936457,-.00649087969965739,-.0097162511988505,-.0123316229887931,-.014172261854691,-.0151224412691208,-.0151224412691208,-.014172261854691,-.0123316229887931,-.0097162511988505,-.00649087969965739,-.00286339454936457},
			{-2.41204413706157e-011,-5.46773700026945e-011,-8.18470047853013e-011,-1.03878171232717e-010,-1.19383202440944e-010,-1.27387250245815e-010,-1.27387250245815e-010,-1.19383202440944e-010,-1.03878171232717e-010,-8.18470047853013e-011,-5.46773700026945e-011,-2.41204413706157e-011,-3.99747274956451e-008,-9.06166239851226e-008,-1.35644769610767e-007,-1.72156948704618e-007,-1.97853385508442e-007,-2.11118467392541e-007,-2.11118467392541e-007,-1.97853385508442e-007,-1.72156948704618e-007,-1.35644769610767e-007,-9.06166239851226e-008,-3.99747274956451e-008,-1.98456423704136e-006,-4.49870511967568e-006,-6.73414868783207e-006,-8.54681307319903e-006,-9.82252482147343e-006,-1.04810760801775e-005,-1.04810760801775e-005,-9.82252482147343e-006,-8.54681307319903e-006,-6.73414868783207e-006,-4.49870511967567e-006,-1.98456423704136e-006,-2.60616962562586e-005,-5.90778994134517e-005,-8.84341934471251e-005,-.00011223846631682,-.000128991369283385,-.000137639596714465,-.000137639596714465,-.000128991369283385,-.00011223846631682,-8.84341934471251e-005,-5.90778994134517e-005,-2.60616962562586e-005,-.000164926605038453,-.000373863515530918,-.00055963937078906,-.000710280291134566,-.000816297926504351,-.000871026627805438,-.000871026627805438,-.000816297926504351,-.000710280291134566,-.00055963937078906,-.000373863515530918,-.000164926605038453,-.00064522442811524,-.00146262559000318,-.0021894162732808,-.00277875237013443,-.00319351364006811,-.00340762278874215,-.00340762278874215,-.00319351364006811,-.00277875237013443,-.0021894162732808,-.00146262559000318,-.00064522442811524,-.00176649543659496,-.00400437633419262,-.00599418386382994,-.00760766822733064,-.00874320163042015,-.00932938965676434,-.00932938965676434,-.00874320163042015,-.00760766822733064,-.00599418386382994,-.00400437633419262,-.00176649543659496,-.00361484526436693,-.00819430411680041,-.0122661212166373,-.0155678542354177,-.0178915384406828,-.0190910767848879,-.0190910767848879,-.0178915384406828,-.0155678542354177,-.0122661212166373,-.00819430411680041,-.00361484526436693,-.00570421773357935,-.0129305935493704,-.0193559118161673,-.0245660391272654,-.028232807600475,-.0301256764218797,-.0301256764218797,-.028232807600475,-.0245660391272654,-.0193559118161673,-.0129305935493704,-.00570421773357935,-.00694730582706703,-.0157484850874735,-.0235740368319136,-.0299195778892043,-.0343854245960146,-.0366907956752786,-.0366907956752786,-.0343854245960146,-.0299195778892043,-.0235740368319136,-.0157484850874735,-.00694730582706703,-.00621725180422221,-.01409356371536,-.0210967714210865,-.0267754945936724,-.0307720501198624,-.03283516247719,-.03283516247719,-.0307720501198624,-.0267754945936724,-.0210967714210865,-.01409356371536,-.00621725180422221,-.00321684991523739,-.00729210922619538,-.0109156182658738,-.0138538296704664,-.0159216756755027,-.0169891445541618,-.0169891445541618,-.0159216756755027,-.0138538296704664,-.0109156182658738,-.00729210922619538,-.00321684991523739},
			{-2.45853458176161e-013,-5.57312376361246e-013,-8.34245479120496e-013,-1.05880349510068e-012,-1.21684229228106e-012,-1.29842549393142e-012,-1.29842549393142e-012,-1.21684229228106e-012,-1.05880349510068e-012,-8.34245479120496e-013,-5.57312376361246e-013,-2.45853458176161e-013,-2.11870775687607e-009,-4.80278806553758e-009,-7.18933295005722e-009,-9.12452155328201e-009,-1.04864630445972e-008,-1.11895280470165e-008,-1.11895280470165e-008,-1.04864630445972e-008,-9.12452155328201e-009,-7.18933295005722e-009,-4.80278806553757e-009,-2.11870775687607e-009,-2.52418873530302e-007,-5.72195173861672e-007,-8.56523661084552e-007,-1.08707840640484e-006,-1.24933756458098e-006,-1.33309941203413e-006,-1.33309941203413e-006,-1.24933756458098e-006,-1.08707840640484e-006,-8.56523661084552e-007,-5.72195173861672e-007,-2.52418873530302e-007,-5.94515724243734e-006,-1.34767667504191e-005,-2.01734829721616e-005,-2.56036799885296e-005,-2.94253284884657e-005,-3.1398149882767e-005,-3.1398149882767e-005,-2.94253284884657e-005,-2.56036799885296e-005,-2.01734829721616e-005,-1.34767667504191e-005,-5.94515724243734e-006,-5.76326598948289e-005,-.000130644469596975,-.000195562780868023,-.000248203389861898,-.000285250646856585,-.000304375278857807,-.000304375278857807,-.000285250646856585,-.000248203389861898,-.000195562780868023,-.000130644469596975,-5.7632659894829e-005,-.000311995321652138,-.000707245915568759,-.00105868222690829,-.00134365300155203,-.00154420891695623,-.00164774041669247,-.00164774041669247,-.00154420891695623,-.00134365300155203,-.00105868222690829,-.000707245915568759,-.000311995321652138,-.00109875339493098,-.00249070674094487,-.00372835940234233,-.00473194049592369,-.00543823792293946,-.00580284462990898,-.00580284462990898,-.00543823792293946,-.00473194049592369,-.00372835940234233,-.00249070674094487,-.00109875339493098,-.00273317477899627,-.00619569129673009,-.00927438125108664,-.0117708127035932,-.0135277440795465,-.0144347117943581,-.0144347117943581,-.0135277440795465,-.0117708127035932,-.00927438125108664,-.00619569129673009,-.00273317477899627,-.00500501190886783,-.0113456003480285,-.0169833224592084,-.0215548080610388,-.0247721151016589,-.0264329617655731,-.0264329617655731,-.0247721151016589,-.0215548080610388,-.0169833224592084,-.0113456003480285,-.00500501190886783,-.00679690022731426,-.0154075384811573,-.0230636710532136,-.0292718344086679,-.0336409978300346,-.0358964588105553,-.0358964588105553,-.0336409978300346,-.0292718344086679,-.0230636710532136,-.0154075384811573,-.00679690022731426,-.00654390879914259,-.0148340454130535,-.0222052045635651,-.0281822902128456,-.0323888264281449,-.0345603355665664,-.0345603355665664,-.0323888264281449,-.0281822902128456,-.0222052045635651,-.0148340454130535,-.00654390879914259,-.00352357286688459,-.00798740348129818,-.0119564099539485,-.0151747764476273,-.0174397891986819,-.0186090400111835,-.0186090400111835,-.0174397891986819,-.0151747764476273,-.0119564099539485,-.00798740348129818,-.00352357286688459},
			{-2.51299527189796e-005,-.00029621599889564,-.00106408128362993,-.00242214596950661,-.00426418972382729,-.00629620113986558,-.00809895569654666,-.00922648473106781,-.00931641156972225,-.00818488569070588,-.00588249745219549,-.00270055519884434,-5.69657786046887e-005,-.000671476592135066,-.00241211034093489,-.00549063631715242,-.00966626919088759,-.014272529844964,-.0183591000865756,-.0209150368234713,-.0211188872818742,-.0185538903067371,-.0133347265413407,-.00612174767248644,-8.52725424398906e-005,-.00100513883251861,-.00361070780484583,-.00821897865428013,-.0144695178404447,-.0213646672921825,-.0274818878919328,-.0313078906116195,-.0316130360391258,-.0277734709743103,-.0199608618151533,-.00916369443189591,-.000108225777940879,-.00127569706467466,-.00458262003120292,-.0104313221265329,-.0183643501167498,-.0271155013323826,-.034879324707423,-.0397351916593658,-.0401224746033481,-.035249394662203,-.025333826534352,-.011630331761288,-.000124379740266118,-.00146610976221544,-.00526662963356749,-.0119883188775888,-.0211054440183851,-.0311628068384012,-.0400854715975038,-.045666133448504,-.0461112228985269,-.0405107788184967,-.0291151962521434,-.013366294714629,-.000132718780991214,-.00156440510346971,-.00561973086134261,-.0127920758168761,-.0225204586888932,-.0332521174831873,-.0427730023755956,-.0487278197470118,-.0492027502229294,-.0432268243227726,-.0310672248280699,-.014262438055421,-.000132718780991214,-.00156440510346971,-.00561973086134261,-.0127920758168761,-.0225204586888932,-.0332521174831873,-.0427730023755956,-.0487278197470118,-.0492027502229294,-.0432268243227726,-.0310672248280699,-.014262438055421,-.000124379740266118,-.00146610976221544,-.00526662963356749,-.0119883188775888,-.0211054440183851,-.0311628068384012,-.0400854715975038,-.045666133448504,-.0461112228985269,-.0405107788184967,-.0291151962521434,-.013366294714629,-.000108225777940879,-.00127569706467466,-.00458262003120292,-.0104313221265329,-.0183643501167498,-.0271155013323826,-.034879324707423,-.0397351916593658,-.0401224746033481,-.035249394662203,-.025333826534352,-.011630331761288,-8.52725424398906e-005,-.00100513883251861,-.00361070780484583,-.00821897865428013,-.0144695178404447,-.0213646672921825,-.0274818878919328,-.0313078906116195,-.0316130360391258,-.0277734709743103,-.0199608618151533,-.00916369443189591,-5.69657786046887e-005,-.000671476592135066,-.00241211034093489,-.00549063631715242,-.00966626919088758,-.014272529844964,-.0183591000865756,-.0209150368234713,-.0211188872818742,-.0185538903067371,-.0133347265413407,-.00612174767248644,-2.51299527189796e-005,-.00029621599889564,-.00106408128362993,-.00242214596950661,-.00426418972382729,-.00629620113986558,-.00809895569654666,-.00922648473106781,-.00931641156972225,-.00818488569070588,-.00588249745219549,-.00270055519884434},
			{-4.01299188864111e-007,-4.73026119128408e-006,-1.69922705697598e-005,-3.86791500860789e-005,-6.80946715845654e-005,-.000100543778916268,-.000129331892821903,-.000147337357934948,-.000148773395949532,-.000130704105389941,-9.39373616201976e-005,-4.31250557013781e-005,-4.73026119128408e-006,-5.5757329091799e-005,-.000200294145262156,-.000455925373489819,-.000802656948401598,-.00118514651569285,-.00152448260646228,-.00173671964859619,-.00175364675704319,-.00154065738088721,-.00110727424428985,-.000508330898783026,-1.69922705697598e-005,-.000200294145262156,-.000719506211645245,-.00163779694031119,-.00288334269301423,-.00425734001677702,-.00547631935750839,-.00623872741470196,-.00629953378353948,-.00553442315607657,-.00397760351765944,-.00182605057560164,-3.86791500860789e-005,-.000455925373489819,-.00163779694031119,-.00372808291892168,-.0065632926638522,-.00969089403328137,-.0124656312103062,-.0142010847243002,-.0143394969898535,-.0125978916716827,-.0090541356913339,-.0041566007314033,-6.80946715845654e-005,-.000802656948401598,-.00288334269301423,-.0065632926638522,-.0115546814618157,-.0170608259252993,-.0219457527239106,-.0250010198852289,-.0252446947784246,-.0221785973614222,-.015939812405686,-.00731769858135901,-.000100543778916268,-.00118514651569285,-.00425734001677702,-.00969089403328137,-.0170608259252993,-.0251908096484753,-.0324035472773186,-.0369147388118314,-.0372745319354167,-.0327473492108772,-.0235356003221748,-.0108047964872896,-.000129331892821903,-.00152448260646228,-.00547631935750839,-.0124656312103062,-.0219457527239106,-.0324035472773186,-.0416814660110368,-.0474843207110434,-.0479471313016055,-.0421237067473831,-.0302744115168138,-.0138984708583546,-.000147337357934948,-.00173671964859619,-.00623872741470196,-.0142010847243002,-.0250010198852289,-.0369147388118314,-.0474843207110434,-.0540950434131128,-.0546222860610777,-.0479881297889341,-.0344891867628131,-.0158334029675552,-.000148773395949532,-.00175364675704319,-.00629953378353948,-.0143394969898535,-.0252446947784246,-.0372745319354167,-.0479471313016055,-.0546222860610777,-.0551546675312396,-.0484558508041033,-.0348253390053783,-.0159877247830154,-.000130704105389941,-.00154065738088721,-.00553442315607657,-.0125978916716827,-.0221785973614222,-.0327473492108772,-.0421237067473831,-.0479881297889341,-.0484558508041033,-.0425706396619946,-.0305956233004421,-.0140459337615274,-9.39373616201976e-005,-.00110727424428985,-.00397760351765944,-.0090541356913339,-.015939812405686,-.0235356003221748,-.0302744115168138,-.0344891867628131,-.0348253390053783,-.0305956233004421,-.0219891496246005,-.0100948470984408,-4.31250557013781e-005,-.000508330898783026,-.00182605057560164,-.0041566007314033,-.00731769858135901,-.0108047964872896,-.0138984708583546,-.0158334029675552,-.0159877247830154,-.0140459337615274,-.0100948470984408,-.00463437375617701},
			{-4.77648743752639e-009,-5.63022148645256e-008,-2.02251510004898e-007,-4.60380881913306e-007,-8.10500874190154e-007,-1.19672830207889e-006,-1.53938053820615e-006,-1.75369165645863e-006,-1.77078418424541e-006,-1.55571337982341e-006,-1.11809502771057e-006,-5.13298537640478e-007,-2.92765537918228e-007,-3.45093511422209e-006,-1.23966142266297e-005,-2.82181536753749e-005,-4.96780798691598e-005,-7.33511413319109e-005,-9.43533458892899e-005,-.000107489130425036,-.00010853678378063,-9.53544357538297e-005,-6.85314672157709e-005,-3.14616387984846e-005,-2.52381411676342e-006,-2.97491255946283e-005,-.000106866232302167,-.00024325736936566,-.000428254773970375,-.000632330728850647,-.000813382299066098,-.000926620621724616,-.000935652020526277,-.000822012292713028,-.000590782254057447,-.000271218151906612,-1.03035498453369e-005,-.000121451733066827,-.000436284726357267,-.000993105797236321,-.0017483634713027,-.00258150992189795,-.00332065860400601,-.0037829575879782,-.0038198285552606,-.00335589082222505,-.00241188697772547,-.00110725656401101,-2.7786916836135e-005,-.000327534612535476,-.00117658550597994,-.00267823697769347,-.00471503812818561,-.00696189202635823,-.00895524997265581,-.0102019914951301,-.0103014262061688,-.00905026525693164,-.00650444787225885,-.00298608212919577,-5.67729905547983e-005,-.000669204128457476,-.00240394708818667,-.00547205447566498,-.00963355584556567,-.0142242276315438,-.0182969677820565,-.0208442545176493,-.021047415089364,-.0184910987778976,-.0132895981153155,-.00610102997451746,-9.39382857945563e-005,-.00110728513787305,-.00397764265011023,-.00905422476771312,-.0159399692247158,-.0235358318700697,-.0302747093624672,-.0344895260742596,-.0348256816239572,-.0305959243062406,-.0219893659582073,-.0100949464135448,-.000130088785011913,-.0015334043731933,-.00550836855481976,-.0125385841278022,-.0220741863877053,-.0325931833471828,-.0419253994709405,-.0477622143590056,-.0482277334672722,-.0423702283435318,-.0304515871935496,-.0139798092182615,-.000152434610085198,-.00179680283515025,-.00645456111210986,-.0146923824553127,-.0258659498945255,-.0381918333275861,-.0491270782522392,-.0559665041221956,-.0565119871455732,-.0496483170026962,-.0356823673916168,-.0163811720361289,-.000149325161617318,-.00176015062198346,-.00632289727835408,-.0143926788244271,-.0253383214364947,-.0374127744408963,-.0481249559775311,-.0548248673220092,-.0553592232703062,-.04863556220147,-.0349544980281293,-.016047019508292,-.000115458636973295,-.001360953435314,-.0048888820448881,-.0111284599424805,-.0195916617438277,-.0289276629306809,-.0372103519687145,-.0423907423550166,-.0428039078843937,-.0376051541440983,-.027026916660946,-.0124076008346097,-5.51609089712794e-005,-.000650201929690796,-.00233568648062049,-.00531667428240711,-.00936000890342516,-.0138203275519371,-.0177774213479609,-.0202523773150973,-.0204497690975875,-.0179660399513751,-.0129122370468795,-.00592778988329973},
			{-5.21061086714588e-011,-6.14193874587905e-010,-2.20633662228147e-009,-5.02223790536332e-009,-8.84165344957659e-009,-1.30549605277777e-008,-1.67929112469387e-008,-1.91308046389391e-008,-1.9317264903302e-008,-1.6971084189159e-008,-1.21971598964522e-008,-5.59951004436151e-009,-1.66071240016209e-008,-1.95754280954478e-007,-7.03197893870279e-007,-1.60067465766628e-006,-2.81798888768115e-006,-4.16084320723466e-006,-5.35219317920171e-006,-6.09732050598647e-006,-6.15674863851456e-006,-5.40897999788056e-006,-3.88744721171111e-006,-1.78466133868042e-006,-3.43560288483173e-007,-4.04967152831355e-006,-1.45474238197567e-005,-3.31139965656822e-005,-5.82972148036964e-005,-8.60775467486887e-005,-.000110723628756234,-.000126138468756338,-.00012736788971769,-.000111898407472009,-8.04216603301225e-005,-3.69202255792134e-005,-2.51557038538368e-006,-2.96519536996954e-005,-.000106517166772019,-.00024246279880047,-.000426855931918081,-.000630265297550785,-.000810725484866511,-.000923593928337934,-.000932595827175548,-.000819327289690436,-.000588852535789363,-.00027033225085123,-1.03921937720791e-005,-.000122496611646568,-.000440038189183426,-.00100164972615974,-.00176340506432601,-.00260371927496907,-.00334922703163026,-.00381550328536708,-.00385269146248218,-.00338476236112819,-.0024326370236585,-.0011167825595386,-2.93810953216598e-005,-.000346325780898917,-.0012440880400349,-.00283189158407368,-.00498554717338517,-.00736130656207001,-.00946902654323017,-.0107872955591597,-.0108924349936834,-.00956949300162804,-.0068776181278623,-.00315739828904347,-6.253444086072e-005,-.000737116463054399,-.00264790502577821,-.00602737082635835,-.0106111906809188,-.0156677341271221,-.0201537850748433,-.0229595761766608,-.0231833539384622,-.0203676169191767,-.0146382563131457,-.00672017616832581,-.00010527045022929,-.00124086152956873,-.00445748215529598,-.010146473397001,-.0178628737232513,-.026375056735671,-.0339268729271671,-.0386501404333965,-.039026847819149,-.0342868373598303,-.0246420342366014,-.0113127416048292,-.000143146764099252,-.00168732357718739,-.00606128448364179,-.0137971750917403,-.0242899366861979,-.0358648036217386,-.0461337637005683,-.0525564631192487,-.0530687098434273,-.0466232433562013,-.0335082395307456,-.0153830666658602,-.000156356665903898,-.00184303355014811,-.00662063329842476,-.0150704090994405,-.0265314660737799,-.039174487477097,-.0503910900341255,-.0574064904417204,-.0579660084330403,-.0509257399611718,-.0366004544098804,-.0168026502756542,-.000130063110871032,-.0015331017426483,-.00550728143089673,-.0125361095303546,-.0220698298571165,-.0325867508020486,-.0419171251326689,-.0477527880751015,-.0482182153091671,-.0423618662144552,-.0304455773108406,-.0139770501826433,-6.46655241703613e-005,-.000762236326125946,-.0027381417997588,-.00623277490756985,-.0109728047138306,-.0162016678481028,-.0208405969245687,-.0237420053295095,-.0239734091138143,-.0210617158489217,-.015137106919017,-.00694919005368325},
			{-5.44724437242996e-013,-6.4208673651397e-012,-2.30653469542085e-011,-5.25031668349854e-011,-9.24318630275365e-011,-1.36478355571743e-010,-1.75555407262082e-010,-1.99996066807805e-010,-2.01945348094836e-010,-1.77418051741924e-010,-1.27510789616824e-010,-5.85380492906022e-011,-9.02770003435125e-010,-1.06412821914539e-008,-3.82261230182231e-008,-8.70133242853409e-008,-1.53187020074262e-007,-2.26185126101403e-007,-2.9094739428102e-007,-3.31452818296366e-007,-3.34683355709183e-007,-2.94034348800585e-007,-2.11323208782435e-007,-9.70149149662404e-008,-4.48184434349486e-008,-5.28291483055767e-007,-1.89775394143646e-006,-4.31981760329904e-006,-7.60504200188573e-006,-1.12290674717028e-005,-1.44442208796392e-005,-1.64551317961879e-005,-1.6615513352675e-005,-1.45974741955391e-005,-1.04912405632317e-005,-4.81635129901971e-006,-5.88564803133452e-007,-6.93762988830925e-006,-2.49216860143408e-005,-5.67287126102093e-005,-9.98709394081982e-005,-.000147462369938119,-.00018968441040975,-.0002160920965543,-.000218198259374459,-.000191696963742466,-.000137773078747959,-6.3249292855151e-005,-3.7246230587395e-006,-4.39035024137147e-005,-.000157712261925103,-.000358997123096203,-.000632014693772654,-.000933188224038868,-.00120038256642976,-.00136749870422512,-.00138082716451303,-.00121311863645989,-.000871871344066475,-.000400261404288981,-1.45714378978438e-005,-.000171758899848453,-.000617000537806862,-.00140446541897081,-.00247256238164128,-.00365081084424778,-.00469612622392589,-.00534991652302709,-.00540206000930644,-.00474595216619707,-.00341092748034312,-.00156589918054001,-3.98936826157651e-005,-.000470241515286636,-.00168922406982674,-.00384514541818346,-.00676938127811489,-.00999518991411039,-.0128570543562206,-.0146470014343651,-.0147897598708826,-.0129934678207622,-.00933843724141566,-.00428711876993237,-8.16359254002543e-005,-.000962272689432589,-.00345672199472992,-.00786846397549439,-.01385243649599,-.0204535285962103,-.0263098681663614,-.0299727033964202,-.0302648352907365,-.0265890161087102,-.0191095911936167,-.00877289047128288,-.000128821307499823,-.00151846414940855,-.0054546995681281,-.0124164185359414,-.0218591137752465,-.0322756223296346,-.0415169130590646,-.0472968588510568,-.0477578423262084,-.0419574078870825,-.0301548921174003,-.0138436015212964,-.000156894610627413,-.00184937450253303,-.00664341157096352,-.015122258804788,-.0266227475173882,-.0393092672046696,-.0505644604551389,-.0576039973305382,-.0581654403421276,-.0511009498439408,-.036726378183032,-.0168604597524974,-.000140407421995964,-.00165503394391524,-.00594529211811034,-.0135331440961878,-.0238251099283638,-.0351784732865376,-.0452509204027591,-.0515507112035252,-.0520531552679814,-.045731033082912,-.0328670058156056,-.0150886871005195,-7.26477900154669e-005,-.000856326230599691,-.00307613605632261,-.00700214416426768,-.0123272798443727,-.0182015900873001,-.0234131452361595,-.0266727014101101,-.0269326695112991,-.0236615589216733,-.0170056205219542,-.00780699308131397},
			{-3.66334363412211e-007,-2.24537546398006e-005,-.000193564800479007,-.000790234334926298,-.00213112724015142,-.00435422423400534,-.00720462946394863,-.00997720456040535,-.0116910253775248,-.0114525451470328,-.00885514027394949,-.00423058507690692,-8.30424254068273e-007,-5.08992448158822e-005,-.000438781946510387,-.00179134098152266,-.00483094114415894,-.00987036372424941,-.0163317986134588,-.0226168044340364,-.0265017754217613,-.0259611770305218,-.0200732554493939,-.00959009257018474,-1.2430688947403e-006,-7.61914981245153e-005,-.000656816303965783,-.00268147304597068,-.00723147552495667,-.0147750286257641,-.0244472036457215,-.0338552805420154,-.0396707255608121,-.0388614993834644,-.0300478211505447,-.0143554884304917,-1.57767194829697e-006,-9.670034364015e-005,-.000833614824033866,-.00340325851820642,-.00917800785523132,-.0187520967637297,-.03102777936875,-.0429682752411084,-.0503490925971213,-.0493220429740223,-.0381359430175126,-.0182196268418573,-1.81315792676965e-006,-.000111134000184102,-.000958041580001006,-.00391123463011006,-.0105479328022725,-.0215510663844978,-.035659038099294,-.0493817925438395,-.057864283158913,-.0566839343751764,-.0438281782544715,-.0209391191031562,-1.93472111511538e-006,-.000118584980154756,-.00102227348573477,-.00417346338855665,-.0112551190451086,-.0229959578102643,-.038049798606524,-.0526925952374313,-.0617437945066473,-.0604843092288456,-.0467666388316428,-.0223429825183335,-1.93472111511538e-006,-.000118584980154756,-.00102227348573477,-.00417346338855665,-.0112551190451086,-.0229959578102643,-.038049798606524,-.0526925952374313,-.0617437945066473,-.0604843092288456,-.0467666388316428,-.0223429825183335,-1.81315792676965e-006,-.000111134000184102,-.000958041580001006,-.00391123463011006,-.0105479328022725,-.0215510663844978,-.035659038099294,-.0493817925438395,-.057864283158913,-.0566839343751764,-.0438281782544715,-.0209391191031562,-1.57767194829697e-006,-9.670034364015e-005,-.000833614824033866,-.00340325851820642,-.00917800785523132,-.0187520967637297,-.03102777936875,-.0429682752411084,-.0503490925971213,-.0493220429740223,-.0381359430175126,-.0182196268418573,-1.2430688947403e-006,-7.61914981245153e-005,-.000656816303965783,-.00268147304597068,-.00723147552495667,-.0147750286257641,-.0244472036457215,-.0338552805420154,-.0396707255608121,-.0388614993834644,-.0300478211505447,-.0143554884304917,-8.30424254068273e-007,-5.08992448158822e-005,-.000438781946510386,-.00179134098152266,-.00483094114415894,-.00987036372424941,-.0163317986134588,-.0226168044340364,-.0265017754217613,-.0259611770305218,-.020073255449394,-.00959009257018474,-3.66334363412211e-007,-2.24537546398006e-005,-.000193564800479007,-.000790234334926298,-.00213112724015142,-.00435422423400534,-.00720462946394863,-.00997720456040535,-.0116910253775248,-.0114525451470328,-.00885514027394949,-.00423058507690693},
			{-5.84997849237861e-009,-3.58563091085661e-007,-3.09102839585264e-006,-1.26192198302079e-005,-3.40318838868514e-005,-6.95324290155738e-005,-.000115050433754236,-.000159325572268953,-.00018669350692448,-.000182885225860587,-.00014140737349095,-6.75580403638936e-005,-6.89558489033362e-008,-4.22651508264995e-006,-3.64350890004165e-005,-.000148747387195263,-.000401146336906179,-.000819604324242625,-.00135614179377857,-.0018780291418385,-.00220062505725282,-.00215573544715162,-.00166682074010411,-.000796331478757963,-2.4770649960978e-007,-1.51826896966895e-005,-.000130883869937062,-.000534337481072711,-.00144101706420306,-.00294422186735428,-.00487159743595245,-.00674634613725228,-.00790519061914258,-.00774393601400238,-.00598763321131698,-.00286062003831949,-5.63849124010232e-007,-3.45600389940658e-005,-.000297928215558894,-.00121630123191481,-.0032801570028024,-.00670187065504814,-.0110891153486868,-.0153565666050425,-.0179944200606735,-.0176273595758088,-.013629524241044,-.00651156956023967,-9.92656788923096e-007,-6.0842973540354e-005,-.000524502838069343,-.00214129919480704,-.0057747187659129,-.0117986481151247,-.0195223955581165,-.0270352465684925,-.0316791894770017,-.031032979229028,-.0239947872427992,-.0114636229006531,-1.46568685042074e-006,-8.98364341570052e-005,-.000774443817183375,-.00316169103729199,-.00852654155446687,-.0174210498412452,-.0288253893768114,-.0399183341468002,-.0467752519970179,-.0458211036210322,-.0354290068157069,-.0169263752901921,-1.88534841928859e-006,-.00011555877647658,-.00099618579926221,-.00406695959489997,-.0109678964760433,-.0224091174530978,-.0370787950245641,-.0513479179830939,-.0601681371359018,-.0589407930194454,-.045573187736373,-.0217728056224891,-2.1478248622506e-006,-.000131646761218445,-.00113487385418326,-.00463315790456963,-.012494836761645,-.0255288938184798,-.042240870176189,-.0584965270825159,-.0685446888933965,-.0671464750773758,-.0519178442938048,-.0248039952501605,-2.16875884799657e-006,-.000132929869292705,-.00114593501354709,-.004678315432652,-.012616618914028,-.0257777136867565,-.042652574961671,-.0590666692229003,-.0692127664287096,-.0678009247845504,-.0524238666569454,-.0250457498234202,-1.90535198329458e-006,-.000116784856153978,-.00100675533972143,-.0041101100734715,-.0110842659581639,-.0226468783946324,-.0374722013795649,-.0518927198634471,-.0608065216222948,-.0595661553735154,-.0460567196758888,-.0220038152896698,-1.36938115091738e-006,-8.39335630014821e-005,-.000723557535766244,-.00295394620634634,-.00796628917278541,-.0162763671345931,-.0269313631824755,-.0372954252410452,-.0437017964620066,-.042810342192032,-.0331010775700317,-.015814196101363,-6.28659751468171e-007,-3.85324807640305e-005,-.00033217304057603,-.00135610679808757,-.00365718877328773,-.00747220517149892,-.0123637338469658,-.0171216996431039,-.0200627564386384,-.0196535048439074,-.0151961454884669,-.00726003025826054},
			{-6.96297165792815e-011,-4.26781849550599e-009,-3.67911491336473e-008,-1.50201013794113e-007,-4.05066519951766e-007,-8.27613867594724e-007,-1.36939462342768e-006,-1.89638209018582e-006,-2.22213055847631e-006,-2.17680226684608e-006,-1.68310966462922e-006,-8.04113589360719e-007,-4.26781849550599e-009,-2.61587661208472e-007,-2.25504216385422e-006,-9.20627996502583e-006,-2.48277670898173e-005,-5.07269876251226e-005,-8.39343887728836e-005,-.00011623506394466,-.000136201184821656,-.000133422875058339,-.000103162944064127,-4.92865836277514e-005,-3.67911491336473e-008,-2.25504216385422e-006,-1.94398127850062e-005,-7.93636419908757e-005,-.000214030208318087,-.000437296986453802,-.000723564654409836,-.00100201580190209,-.00117413571083906,-.001150185018243,-.000889326306666056,-.000424879842112779,-1.50201013794113e-007,-9.20627996502583e-006,-7.93636419908757e-005,-.000324004543650437,-.000873785000712069,-.00178527858577817,-.00295397526843083,-.00409076076250558,-.00479344565888028,-.00469566620937377,-.0036307023836027,-.00173458520673544,-4.05066519951766e-007,-2.48277670898173e-005,-.000214030208318087,-.000873785000712069,-.00235644913761801,-.00481459189667545,-.00796636754827121,-.0110320841661877,-.0129271055006463,-.0126634109999641,-.0097913851734879,-.00467788049829846,-8.27613867594724e-007,-5.07269876251226e-005,-.000437296986453802,-.00178527858577817,-.00481459189667545,-.0098369596701605,-.016276527267894,-.0225402628819985,-.0264120860481604,-.025873317192124,-.0200053219740424,-.00955763703157627,-1.36939462342768e-006,-8.39343887728836e-005,-.000723564654409836,-.00295397526843083,-.00796636754827121,-.016276527267894,-.0269316281438179,-.037295792168106,-.0437022264174669,-.0428107633770146,-.0331014032314506,-.0158143516876436,-1.89638209018582e-006,-.00011623506394466,-.00100201580190209,-.00409076076250558,-.0110320841661877,-.0225402628819985,-.037295792168106,-.0516484226656698,-.0605202606038321,-.0592857336712341,-.0458398968231796,-.0219002271480226,-2.22213055847631e-006,-.000136201184821656,-.00117413571083906,-.00479344565888028,-.0129271055006463,-.0264120860481604,-.0437022264174669,-.0605202606038321,-.0709160465028163,-.0694694603763257,-.0537139830919874,-.0256621090417623,-2.17680226684608e-006,-.000133422875058339,-.001150185018243,-.00469566620937377,-.0126634109999641,-.025873317192124,-.0428107633770146,-.0592857336712341,-.0694694603763257,-.0680523825420278,-.0526182945056766,-.0251386386461752,-1.68310966462922e-006,-.000103162944064127,-.000889326306666056,-.0036307023836027,-.0097913851734879,-.0200053219740424,-.0331014032314506,-.0458398968231796,-.0537139830919874,-.0526182945056766,-.0406846140173892,-.0194372664460252,-8.04113589360719e-007,-4.92865836277514e-005,-.000424879842112779,-.00173458520673544,-.00467788049829846,-.00955763703157627,-.0158143516876436,-.0219002271480226,-.0256621090417623,-.0251386386461752,-.0194372664460252,-.00928624582089677},
			{-7.59581936873143e-013,-4.65571023163414e-011,-4.01350079989265e-010,-1.63852421901125e-009,-4.41882039598848e-009,-9.02833697182859e-009,-1.49385560003323e-008,-2.06873968742175e-008,-2.42409464882796e-008,-2.37464657802842e-008,-1.83608344516711e-008,-8.77197535304017e-009,-2.42091987612831e-010,-1.48385590679724e-008,-1.2791725800267e-007,-5.22226195326764e-007,-1.40835499192186e-006,-2.8774881763854e-006,-4.76117787775939e-006,-6.59343355166069e-006,-7.72601168100707e-006,-7.56841207045341e-006,-5.85191760211322e-006,-2.79578126521843e-006,-5.00828398075496e-009,-3.06973057681128e-007,-2.64629143836613e-006,-1.08035673306481e-005,-2.91353787244659e-005,-5.95281078102025e-005,-9.84969850916432e-005,-.000136401819657762,-.000159832057717645,-.000156571711876826,-.000121061690113565,-5.7837794064797e-005,-3.66709753307118e-008,-2.24767634356886e-006,-1.93763150067161e-005,-7.91044103306138e-005,-.00021333110473795,-.000435868609164376,-.000721201218685817,-.000998742839454559,-.00117030053969446,-.00114642807911561,-.0008864214307152,-.000423492023911399,-1.51493229393287e-007,-9.28548381512419e-006,-8.00464265713343e-005,-.000326792032995098,-.000881302384115301,-.00180063776864345,-.00297938903112802,-.00412595456537195,-.00483468482962708,-.00473606415990016,-.00366193819311915,-.00174950826221703,-4.28305814057175e-007,-2.62521745709628e-005,-.000226309453117529,-.000923915400575431,-.00249164227715486,-.00509081249643689,-.00842341040243774,-.011665011934605,-.0136687535803171,-.0133899305174026,-.0103531321176025,-.00494625775323145,-9.11601977606186e-007,-5.58748760112742e-005,-.000481674864645602,-.00196645265756991,-.00530318747211388,-.0108352363826531,-.0179283057316278,-.0248276992732258,-.0290924437312574,-.0284989994042999,-.0220355068809891,-.010527567456728,-1.53459036799021e-006,-9.40597416919936e-005,-.000810851255203673,-.00331032553849853,-.00892738345711226,-.0182400321589689,-.0301805019800266,-.0417949380321651,-.0489742069762943,-.0479752030574144,-.0370945625876605,-.0177221024244434,-2.08673606807239e-006,-.00012790244200432,-.00110259558209743,-.00450138085207253,-.0121394565234153,-.0248027967481441,-.0410394482774791,-.0568327590696347,-.0665951293806949,-.0652366837960255,-.0504411882777659,-.0240985158662193,-2.27930471413928e-006,-.000139705563856799,-.00120434555501086,-.00491677828990747,-.0132597125742255,-.027091654003041,-.044826659852073,-.0620774124945382,-.0727406760531766,-.0712568700882474,-.0550960133326791,-.02632238051474,-1.89600782307658e-006,-.000116212123967743,-.00100181804557465,-.00408995341613109,-.011029906890695,-.0225358143697108,-.0372884315268124,-.0516382294108676,-.0605083164164748,-.0592740331282503,-.0458308499302774,-.0218959049522221,-9.42668054679449e-007,-5.77790109816517e-005,-.000498089647453146,-.00203346652033204,-.00548391243189871,-.0112044855690737,-.0185392764637387,-.025673791359598,-.0300838721412506,-.0294702040898467,-.0227864450885946,-.0108863317310921},
			{-4.61452526590166e-009,-1.47072954118751e-006,-3.04257537549422e-005,-.000222779313164588,-.000920334332232957,-.00260199447163722,-.00553806002210424,-.00932276780442027,-.0126771001809598,-.0138469711844149,-.0115184097714994,-.00572678909872914,-1.0460426551641e-008,-3.33392005816974e-006,-6.89705536523267e-005,-.000505006800981909,-.00208625787714596,-.00589832550261673,-.0125539393028941,-.0211332958986598,-.0287370569429126,-.0313889764798934,-.0261104821110847,-.0129817593993148,-1.56582985243485e-008,-4.99057235089002e-006,-.000103242588927534,-.000755948833210871,-.0031229365722185,-.00882925194855102,-.018792094977277,-.0316346044160784,-.0430167368512079,-.0469864170136373,-.0390849953863391,-.019432502397701,-1.98731208257652e-008,-6.33390959846238e-006,-.000131032911457859,-.000959431350547771,-.00396355298338403,-.0112058657268432,-.023850456898108,-.0401498486478861,-.0545957664330435,-.0596339852015086,-.0496056985105609,-.0246632459775863,-2.2839416390577e-008,-7.2793196382121e-006,-.000150591105035483,-.00110263769366935,-.00455515959306486,-.0128784721632876,-.0274104163597275,-.0461426828391663,-.0627448227009384,-.0685350545084463,-.0570099288159739,-.028344523709382,-2.43706852532627e-008,-7.76736168416077e-006,-.000160687486930486,-.00117656404705184,-.00486056030603151,-.0137419094370501,-.0292481479535269,-.0492363193955517,-.0669515498630354,-.073129987810444,-.0608321599696494,-.0302448825382416,-2.43706852532627e-008,-7.76736168416077e-006,-.000160687486930486,-.00117656404705184,-.00486056030603151,-.0137419094370501,-.0292481479535269,-.0492363193955517,-.0669515498630354,-.073129987810444,-.0608321599696494,-.0302448825382416,-2.2839416390577e-008,-7.2793196382121e-006,-.000150591105035483,-.00110263769366935,-.00455515959306486,-.0128784721632876,-.0274104163597275,-.0461426828391663,-.0627448227009384,-.0685350545084463,-.0570099288159739,-.028344523709382,-1.98731208257652e-008,-6.33390959846238e-006,-.000131032911457859,-.000959431350547771,-.00396355298338403,-.0112058657268432,-.023850456898108,-.0401498486478861,-.0545957664330435,-.0596339852015086,-.0496056985105609,-.0246632459775863,-1.56582985243485e-008,-4.99057235089002e-006,-.000103242588927534,-.000755948833210871,-.0031229365722185,-.00882925194855102,-.018792094977277,-.0316346044160784,-.0430167368512079,-.0469864170136373,-.0390849953863391,-.019432502397701,-1.0460426551641e-008,-3.33392005816974e-006,-6.89705536523267e-005,-.000505006800981909,-.00208625787714596,-.00589832550261673,-.0125539393028941,-.0211332958986598,-.0287370569429126,-.0313889764798934,-.0261104821110847,-.0129817593993148,-4.61452526590166e-009,-1.47072954118751e-006,-3.04257537549422e-005,-.000222779313164588,-.000920334332232957,-.00260199447163722,-.00553806002210424,-.00932276780442027,-.0126771001809598,-.0138469711844149,-.0115184097714994,-.00572678909872914},
			{-7.36891655603157e-011,-2.34860199933996e-008,-4.85867619430457e-007,-3.55755375585554e-006,-1.46967813742169e-005,-4.15511434782526e-005,-8.84370543742305e-005,-.000148874898420579,-.000202440095184293,-.000221121717472685,-.0001839370153449,-9.14508612930534e-005,-8.68601307305661e-010,-2.76838358998326e-007,-5.72710039808159e-006,-4.19341950699251e-005,-.000173236369522755,-.000489778616311158,-.00104244009902437,-.00175484320401577,-.00238623588679712,-.00260644304233894,-.00216813327679371,-.00107796495006253,-3.12023117416446e-009,-9.94471998471375e-007,-2.05731640619964e-005,-.000150638021863614,-.000622307975060138,-.00175940617887584,-.00374470319905967,-.00630383171759815,-.00857195072154575,-.00936298940145173,-.00778847209081444,-.00387231726863855,-7.10251695871362e-009,-2.26369580965429e-006,-4.68302630441653e-005,-.000342894178409679,-.0014165466273796,-.00400489948507659,-.00852398956746932,-.0143492802872553,-.0195121521357594,-.0213127769373929,-.0177287361159355,-.00881447480486433,-1.25039862212867e-008,-3.98523810328614e-006,-8.24447118179819e-005,-.000603665448054148,-.00249383135774626,-.00705062842794049,-.0150064897727934,-.0252619182805468,-.0343511578883119,-.0375211591484582,-.0312114527009358,-.0155178892424445,-1.84625022332888e-008,-5.88432089415219e-006,-.000121732034018966,-.00089132973162475,-.00368221511100628,-.01041045957611,-.0221575220926239,-.0372999629412389,-.0507204916900297,-.055401091485105,-.0460846249345851,-.0229126184021882,-2.37487628354249e-008,-7.56914418191853e-006,-.000156586857450866,-.00114653897597652,-.00473652229125437,-.0133912257589351,-.0285017561839274,-.0479798438190263,-.0652429943039245,-.0712637629438524,-.0592797668540085,-.0294730548165372,-2.70550433775344e-008,-8.62291335307587e-006,-.000178386733239242,-.00130615905948616,-.00539593649305364,-.0152555396799836,-.0324697440130155,-.0546595527845674,-.074326062928209,-.0811850373450308,-.0675326405320358,-.03357626593249,-2.7318738012221e-008,-8.7069574240775e-006,-.000180125397036526,-.0013188896669085,-.00544852851748593,-.0154042292942165,-.0327862135586896,-.055192297478506,-.0750504891927709,-.0819763152766897,-.0681908540387014,-.0339035203026146,-2.40007374267426e-008,-7.64945287105235e-006,-.000158248245443382,-.00115870376501852,-.00478677683617009,-.0135333067869547,-.0288041600782291,-.0484889104016337,-.0659352230713762,-.0720198721217481,-.0599087257235696,-.0297857641982319,-1.72493889467441e-008,-5.49768056941236e-006,-.000113733402739168,-.000832763242290672,-.0034402682709391,-.00972642083254007,-.0207016206060418,-.0348490990193348,-.0473878067921877,-.0517608590117147,-.0430565483441448,-.0214071018984119,-7.91890304680574e-009,-2.52389226922161e-006,-5.22130837362085e-005,-.000382307535472911,-.00157936904180661,-.00446523548185483,-.00950376427809313,-.0159986326040109,-.0217549415081703,-.0237625359019177,-.0197665339288291,-.00982763882071387},
			{-8.77089671184217e-013,-2.79543748348947e-010,-5.78306820717772e-009,-4.23439949444061e-008,-1.74929313488118e-007,-4.94564953932593e-007,-1.05262729400979e-006,-1.77199232357178e-006,-2.40955526052686e-006,-2.63191437974773e-006,-2.18932125341295e-006,-1.08849930992072e-006,-5.37595110937315e-011,-1.71340921393568e-008,-3.5446195486468e-007,-2.5953930832331e-006,-1.07219531571797e-005,-3.03134001015112e-005,-6.45187493924991e-005,-.000108610834338575,-.000147689018597559,-.000161318089751033,-.000134190202070998,-6.67174550672731e-005,-4.63439153301422e-010,-1.47706126638825e-007,-3.05567414766182e-006,-2.23738413632665e-005,-9.24296518291858e-005,-.000261319647275767,-.000556190225361124,-.000936290380644839,-.00127316771187585,-.00139065846034338,-.0011567998362318,-.000575144383900619,-1.89200479726517e-009,-6.03014868716438e-007,-1.24748850093275e-005,-9.13419051691941e-005,-.000377346936322897,-.00106684561013735,-.00227066394170372,-.00382243468035404,-.00519774689175974,-.00567740653672345,-.00472267141011844,-.00234804488509053,-5.1024142886996e-009,-1.62622826690742e-006,-3.36426374888076e-005,-.000246333541419145,-.00101764033710382,-.00287710067800549,-.00612359342731638,-.0103084544811151,-.0140174369789623,-.0153109972435626,-.0127362394210907,-.00633227663561512,-1.04250255588739e-008,-3.32263714544313e-006,-6.87371381162648e-005,-.000503297717511878,-.00207919740024719,-.00587836392863835,-.0125114532807165,-.0210617749476202,-.0286398027496958,-.0312827474533958,-.0260221169776175,-.012937825514943,-1.7249558650954e-008,-5.49773465709859e-006,-.000113734521679514,-.000832771435243388,-.00344030211724,-.00972651652373048,-.0207018242742517,-.0348494418743096,-.0473882730064016,-.0517613682492133,-.0430569719462525,-.0214073125073398,-2.388774830107e-008,-7.61344126955836e-006,-.000157503254546897,-.00115324889407038,-.00476424195651651,-.0134695955569983,-.0286685577088058,-.0482606374326126,-.0656248174753565,-.071680821606866,-.0596266912835726,-.0296455406961615,-2.79910339523387e-008,-8.92122984487403e-006,-.000184558163040781,-.00135134666284035,-.00558261317396905,-.0157833171133724,-.0335930604290494,-.0565505431450506,-.076897431725972,-.0839936977751664,-.0698689855213075,-.0347378633474854,-2.74200568127716e-008,-8.73924948977462e-006,-.000180793439944228,-.00132378111976588,-.00546873583357464,-.0154613599726102,-.0329078099454216,-.0553969927824272,-.0753288338784716,-.0822803462291326,-.068443757944315,-.0340292605183584,-2.12012654199985e-008,-6.75721240365912e-006,-.000139789998708784,-.00102355130296691,-.00422844200182213,-.0119547672264766,-.0254444116512108,-.0428331113779068,-.0582444672431377,-.0636193962384749,-.0529209070727989,-.0263115204035591,-1.0129004659729e-008,-3.22829013115797e-006,-6.67853319250079e-005,-.000489006467861309,-.00202015813166747,-.00571144648888444,-.0121561878064161,-.0204637212045026,-.0278265692364139,-.0303944669425635,-.0252832132289538,-.0125704531070497},
			{-5.3934946534894e-011,-8.93862080339512e-008,-4.43762053833248e-006,-5.82757244194688e-005,-.000368787100047311,-.00144276568155165,-.00395000387691709,-.00808303973670482,-.0127550186066991,-.0155346480848768,-.0139021976630662,-.00719309508160456,-1.22262316118168e-010,-2.02624931058027e-007,-1.00594104551317e-005,-.000132102199014488,-.000835984234769121,-.00327053024384089,-.00895405768790711,-.0183230210276625,-.028913686156645,-.0352146831870672,-.0315141665027775,-.0163056519239563,-1.83015466396657e-010,-3.03310925545804e-007,-1.50580142318092e-005,-.00019774486802044,-.00125139167557278,-.00489568361654787,-.0134034025849063,-.0274278808519569,-.0432811345737185,-.0527131488435292,-.0471738149883672,-.0244080644511933,-2.32278652181071e-010,-3.84954639980414e-007,-1.91112550165974e-005,-.000250972840295157,-.00158823501355166,-.00621347919028916,-.017011263301183,-.0348107803231648,-.0549313334090684,-.0669022099971482,-.0598718260236465,-.0309781048820443,-2.6694895594493e-010,-4.42413619435074e-007,-2.19638332045296e-005,-.000288433470136975,-.00182529765297712,-.00714091358400096,-.0195503931804077,-.0400066961627327,-.0631304769703183,-.0768881468075034,-.0688083958532261,-.0356019491148408,-2.84846550926105e-010,-4.72075244245615e-007,-2.3436398585193e-005,-.000307771494551617,-.00194767474936783,-.00761967619488893,-.0208611494544942,-.0426889454411423,-.0673630603261273,-.0820431132524654,-.0734216553279677,-.0379888820906312,-2.84846550926105e-010,-4.72075244245615e-007,-2.3436398585193e-005,-.000307771494551617,-.00194767474936783,-.00761967619488893,-.0208611494544942,-.0426889454411423,-.0673630603261273,-.0820431132524654,-.0734216553279677,-.0379888820906312,-2.6694895594493e-010,-4.42413619435074e-007,-2.19638332045296e-005,-.000288433470136975,-.00182529765297712,-.00714091358400096,-.0195503931804077,-.0400066961627327,-.0631304769703183,-.0768881468075034,-.0688083958532261,-.0356019491148408,-2.32278652181071e-010,-3.84954639980414e-007,-1.91112550165974e-005,-.000250972840295156,-.00158823501355166,-.00621347919028916,-.017011263301183,-.0348107803231648,-.0549313334090684,-.0669022099971482,-.0598718260236465,-.0309781048820443,-1.83015466396657e-010,-3.03310925545804e-007,-1.50580142318092e-005,-.00019774486802044,-.00125139167557278,-.00489568361654787,-.0134034025849063,-.0274278808519569,-.0432811345737185,-.0527131488435292,-.0471738149883672,-.0244080644511933,-1.22262316118168e-010,-2.02624931058027e-007,-1.00594104551317e-005,-.000132102199014488,-.00083598423476912,-.00327053024384089,-.00895405768790711,-.0183230210276625,-.028913686156645,-.0352146831870672,-.0315141665027775,-.0163056519239563,-5.3934946534894e-011,-8.93862080339512e-008,-4.43762053833248e-006,-5.82757244194688e-005,-.000368787100047311,-.00144276568155165,-.00395000387691709,-.00808303973670482,-.0127550186066991,-.0155346480848768,-.0139021976630662,-.00719309508160456},
			{-8.61284959198155e-013,-1.42740470669776e-009,-7.08641812006219e-008,-9.30602664014744e-007,-5.88914614407857e-006,-2.30394662644891e-005,-6.30774506425453e-005,-.000129077731546852,-.000203684371377566,-.000248072161029854,-.000222003626892483,-.000114866241683583,-1.015228271122e-011,-1.68253444704445e-008,-8.35302177246445e-007,-1.09693560023246e-005,-6.94175324250859e-005,-.000271574665892772,-.000743517119145248,-.00152148671399835,-.00240090262810037,-.00292411783656719,-.00261683843315572,-.00135397065407101,-3.64695156892415e-011,-6.04408074120732e-008,-3.00061244597533e-006,-3.94046454582605e-005,-.000249364981245768,-.00097556350825723,-.00267089776882591,-.00546555736926158,-.0086246372913827,-.0105041559963674,-.00940033222170924,-.00486379816401721,-8.30147957637377e-011,-1.37580145726718e-007,-6.83023134969281e-006,-8.96959702655164e-005,-.000567624291069279,-.00222065480887151,-.00607970872643953,-.0124411339215587,-.0196320814726822,-.0239103905886559,-.0213977796180793,-.0110713620291109,-1.46147607732105e-010,-2.42209945641729e-007,-1.20246272104963e-005,-.000157909820254493,-.00099930297324955,-.0039094643904085,-.0107033315917073,-.0219026252294268,-.0345622935723416,-.0420942598523953,-.0376708064290284,-.0194911408262166,-2.15791227404772e-010,-3.57630085574026e-007,-1.77547146005579e-005,-.000233158479025165,-.00147550013642398,-.00577243878564134,-.015803782883183,-.0323398682673508,-.0510322397173339,-.0621534087434013,-.0556220500820546,-.0287792408488707,-2.77577471181613e-010,-4.60028222490692e-007,-2.28383184972103e-005,-.000299917386683,-.00189797148624789,-.00742522752170763,-.0203287878778718,-.041599554161483,-.0656440033259579,-.0799494318272223,-.0715479873273676,-.0370194330577289,-3.16221547012239e-010,-5.24072921213834e-007,-2.60178478303842e-005,-.000341671604648038,-.00216220530078157,-.00845896075008184,-.0231589427061906,-.0473910051704331,-.0747829000510975,-.0910799169238503,-.0815088311812927,-.0421732438918713,-3.19303631345124e-010,-5.29180849357959e-007,-2.62714333369098e-005,-.00034500174046461,-.00218327944688257,-.00854140684095277,-.0233846636134322,-.0478529062519234,-.0755117789234986,-.0919676362701244,-.0823032650013445,-.0425842895511383,-2.80522570694571e-010,-4.64909126147018e-007,-2.30806332657561e-005,-.000303099512904151,-.00191810898111,-.00750400925375776,-.0205444765035491,-.0420409258123548,-.0663404868028083,-.0807976960315661,-.072307112128577,-.0374122095817176,-2.01612260665772e-010,-3.34131331017576e-007,-1.65881078259945e-005,-.000217838364492369,-.00137854963658506,-.00539314988437847,-.0147653657308921,-.0302149166554566,-.0476790708315039,-.058069502618547,-.0519672991102317,-.026888246929992,-9.25567827468422e-011,-1.53394049110811e-007,-7.61532005623914e-006,-.00010000591288281,-.000632868848341838,-.0024759039977092,-.00677852995457166,-.0138711577726375,-.0218886559078795,-.0266587276008582,-.0238573090634585,-.0123439399038752},
			{-6.02215524056038e-013,-5.18975291864507e-009,-6.18297441623807e-007,-1.45626016852063e-005,-.000141170609267733,-.00076422934021594,-.00269138517084675,-.00669488358667141,-.0122597253338038,-.0166489373902391,-.016029237481894,-.00863095559575211,-1.36513094948138e-012,-1.1764380103799e-008,-1.40158620930416e-006,-3.30112019224449e-005,-.000320012288242827,-.00173239161588508,-.00610096061448634,-.0151762822814405,-.0277909316794019,-.0377406074727463,-.0363358420848087,-.0195650628997474,-2.0434757441563e-012,-1.76101973193715e-008,-2.09804592236745e-006,-4.94147396186346e-005,-.000479029025826867,-.00259323125578963,-.0091325781138374,-.0227175017461691,-.0416004741639895,-.0564942256781886,-.0543914208171814,-.0292871035439603,-2.59352830098303e-012,-2.23504219535243e-008,-2.66278740620364e-006,-6.27159515120887e-005,-.00060797165761674,-.0032912642453023,-.0115908397087212,-.0288324849829067,-.0527982812554331,-.0717010581395466,-.0690322308074567,-.0371704592590807,-2.98064271365472e-012,-2.56864836669144e-008,-3.06023954984572e-006,-7.20770403136054e-005,-.00069871853362736,-.00378252390296041,-.0133209080116245,-.0331360703672624,-.0606790418511532,-.0824032791251857,-.0793360981199193,-.0427185847602993,-3.18047992928925e-012,-2.7408634178924e-008,-3.26541333602756e-006,-7.69094460834964e-005,-.000745564123550859,-.00403612324963012,-.0142140084005345,-.0353576784818238,-.0647472687189207,-.087928007662455,-.0846551874810701,-.0455826526323854,-3.18047992928925e-012,-2.7408634178924e-008,-3.26541333602756e-006,-7.69094460834964e-005,-.000745564123550859,-.00403612324963012,-.0142140084005345,-.0353576784818238,-.0647472687189207,-.087928007662455,-.0846551874810701,-.0455826526323854,-2.98064271365472e-012,-2.56864836669144e-008,-3.06023954984572e-006,-7.20770403136054e-005,-.00069871853362736,-.00378252390296041,-.0133209080116245,-.0331360703672624,-.0606790418511532,-.0824032791251857,-.0793360981199193,-.0427185847602993,-2.59352830098303e-012,-2.23504219535243e-008,-2.66278740620364e-006,-6.27159515120887e-005,-.00060797165761674,-.0032912642453023,-.0115908397087212,-.0288324849829067,-.0527982812554331,-.0717010581395466,-.0690322308074567,-.0371704592590807,-2.0434757441563e-012,-1.76101973193715e-008,-2.09804592236745e-006,-4.94147396186346e-005,-.000479029025826867,-.00259323125578962,-.0091325781138374,-.0227175017461691,-.0416004741639895,-.0564942256781886,-.0543914208171814,-.0292871035439603,-1.36513094948138e-012,-1.1764380103799e-008,-1.40158620930416e-006,-3.30112019224449e-005,-.000320012288242827,-.00173239161588507,-.00610096061448634,-.0151762822814405,-.0277909316794019,-.0377406074727463,-.0363358420848087,-.0195650628997474,-6.02215524056038e-013,-5.18975291864507e-009,-6.18297441623807e-007,-1.45626016852063e-005,-.000141170609267733,-.00076422934021594,-.00269138517084675,-.00669488358667141,-.0122597253338038,-.0166489373902391,-.016029237481894,-.00863095559575211},
			{-.00270055519884434,-.00588249745219549,-.00818488569070588,-.00931641156972225,-.00922648473106781,-.00809895569654666,-.00629620113986558,-.00426418972382729,-.00242214596950661,-.00106408128362993,-.00029621599889564,-2.51299527189796e-005,-.00612174767248644,-.0133347265413407,-.0185538903067371,-.0211188872818742,-.0209150368234713,-.0183591000865756,-.014272529844964,-.00966626919088759,-.00549063631715242,-.00241211034093489,-.000671476592135066,-5.69657786046887e-005,-.00916369443189591,-.0199608618151533,-.0277734709743103,-.0316130360391258,-.0313078906116195,-.0274818878919328,-.0213646672921825,-.0144695178404447,-.00821897865428013,-.00361070780484583,-.00100513883251861,-8.52725424398906e-005,-.011630331761288,-.025333826534352,-.035249394662203,-.0401224746033481,-.0397351916593658,-.034879324707423,-.0271155013323826,-.0183643501167498,-.0104313221265329,-.00458262003120292,-.00127569706467466,-.000108225777940879,-.013366294714629,-.0291151962521434,-.0405107788184967,-.0461112228985269,-.045666133448504,-.0400854715975038,-.0311628068384012,-.0211054440183851,-.0119883188775888,-.00526662963356749,-.00146610976221544,-.000124379740266118,-.014262438055421,-.03106722482807,-.0432268243227726,-.0492027502229294,-.0487278197470118,-.0427730023755956,-.0332521174831873,-.0225204586888932,-.0127920758168761,-.00561973086134261,-.00156440510346971,-.000132718780991214,-.014262438055421,-.03106722482807,-.0432268243227726,-.0492027502229294,-.0487278197470118,-.0427730023755956,-.0332521174831873,-.0225204586888932,-.0127920758168761,-.00561973086134261,-.00156440510346971,-.000132718780991214,-.013366294714629,-.0291151962521434,-.0405107788184967,-.0461112228985269,-.045666133448504,-.0400854715975038,-.0311628068384012,-.0211054440183851,-.0119883188775888,-.00526662963356749,-.00146610976221544,-.000124379740266118,-.011630331761288,-.025333826534352,-.035249394662203,-.0401224746033481,-.0397351916593658,-.034879324707423,-.0271155013323826,-.0183643501167498,-.0104313221265329,-.00458262003120292,-.00127569706467466,-.000108225777940879,-.00916369443189591,-.0199608618151533,-.0277734709743103,-.0316130360391258,-.0313078906116195,-.0274818878919328,-.0213646672921825,-.0144695178404447,-.00821897865428013,-.00361070780484583,-.00100513883251861,-8.52725424398906e-005,-.00612174767248644,-.0133347265413407,-.0185538903067371,-.0211188872818742,-.0209150368234713,-.0183591000865756,-.014272529844964,-.00966626919088758,-.00549063631715242,-.00241211034093489,-.000671476592135066,-5.69657786046887e-005,-.00270055519884434,-.00588249745219549,-.00818488569070588,-.00931641156972225,-.00922648473106781,-.00809895569654666,-.00629620113986558,-.00426418972382729,-.00242214596950661,-.00106408128362993,-.00029621599889564,-2.51299527189796e-005},
			{-4.31250557013781e-005,-9.39373616201976e-005,-.000130704105389941,-.000148773395949532,-.000147337357934948,-.000129331892821903,-.000100543778916268,-6.80946715845654e-005,-3.86791500860789e-005,-1.69922705697598e-005,-4.73026119128408e-006,-4.01299188864111e-007,-.000508330898783026,-.00110727424428985,-.00154065738088721,-.00175364675704319,-.00173671964859619,-.00152448260646228,-.00118514651569285,-.000802656948401598,-.000455925373489819,-.000200294145262156,-5.5757329091799e-005,-4.73026119128408e-006,-.00182605057560164,-.00397760351765945,-.00553442315607657,-.00629953378353948,-.00623872741470196,-.00547631935750839,-.00425734001677702,-.00288334269301423,-.00163779694031119,-.000719506211645245,-.000200294145262156,-1.69922705697598e-005,-.0041566007314033,-.0090541356913339,-.0125978916716827,-.0143394969898535,-.0142010847243002,-.0124656312103062,-.00969089403328137,-.0065632926638522,-.00372808291892168,-.00163779694031119,-.000455925373489819,-3.86791500860789e-005,-.00731769858135901,-.015939812405686,-.0221785973614222,-.0252446947784247,-.0250010198852289,-.0219457527239106,-.0170608259252993,-.0115546814618157,-.0065632926638522,-.00288334269301423,-.000802656948401598,-6.80946715845654e-005,-.0108047964872896,-.0235356003221748,-.0327473492108772,-.0372745319354167,-.0369147388118314,-.0324035472773186,-.0251908096484753,-.0170608259252993,-.00969089403328137,-.00425734001677702,-.00118514651569285,-.000100543778916268,-.0138984708583546,-.0302744115168138,-.0421237067473831,-.0479471313016055,-.0474843207110434,-.0416814660110368,-.0324035472773186,-.0219457527239106,-.0124656312103062,-.00547631935750839,-.00152448260646228,-.000129331892821903,-.0158334029675552,-.0344891867628131,-.0479881297889341,-.0546222860610778,-.0540950434131128,-.0474843207110434,-.0369147388118314,-.0250010198852289,-.0142010847243002,-.00623872741470196,-.00173671964859619,-.000147337357934948,-.0159877247830154,-.0348253390053783,-.0484558508041033,-.0551546675312396,-.0546222860610777,-.0479471313016055,-.0372745319354167,-.0252446947784246,-.0143394969898535,-.00629953378353948,-.00175364675704319,-.000148773395949532,-.0140459337615274,-.0305956233004421,-.0425706396619946,-.0484558508041033,-.0479881297889341,-.0421237067473831,-.0327473492108772,-.0221785973614222,-.0125978916716827,-.00553442315607657,-.00154065738088721,-.000130704105389941,-.0100948470984408,-.0219891496246005,-.0305956233004421,-.0348253390053783,-.0344891867628131,-.0302744115168138,-.0235356003221748,-.015939812405686,-.0090541356913339,-.00397760351765944,-.00110727424428985,-9.39373616201976e-005,-.00463437375617701,-.0100948470984408,-.0140459337615274,-.0159877247830154,-.0158334029675552,-.0138984708583546,-.0108047964872896,-.00731769858135901,-.0041566007314033,-.00182605057560164,-.000508330898783026,-4.31250557013781e-005},
			{-5.13298537640478e-007,-1.11809502771057e-006,-1.55571337982341e-006,-1.77078418424541e-006,-1.75369165645863e-006,-1.53938053820615e-006,-1.19672830207889e-006,-8.10500874190154e-007,-4.60380881913306e-007,-2.02251510004898e-007,-5.63022148645256e-008,-4.77648743752639e-009,-3.14616387984846e-005,-6.85314672157709e-005,-9.53544357538297e-005,-.00010853678378063,-.000107489130425036,-9.43533458892899e-005,-7.33511413319109e-005,-4.96780798691598e-005,-2.82181536753749e-005,-1.23966142266297e-005,-3.45093511422209e-006,-2.92765537918228e-007,-.000271218151906612,-.000590782254057447,-.000822012292713028,-.000935652020526277,-.000926620621724616,-.000813382299066098,-.000632330728850647,-.000428254773970375,-.00024325736936566,-.000106866232302167,-2.97491255946283e-005,-2.52381411676342e-006,-.00110725656401101,-.00241188697772547,-.00335589082222505,-.0038198285552606,-.0037829575879782,-.00332065860400601,-.00258150992189795,-.0017483634713027,-.000993105797236321,-.000436284726357267,-.000121451733066827,-1.03035498453369e-005,-.00298608212919577,-.00650444787225884,-.00905026525693164,-.0103014262061688,-.0102019914951301,-.00895524997265581,-.00696189202635823,-.00471503812818561,-.00267823697769347,-.00117658550597994,-.000327534612535476,-2.7786916836135e-005,-.00610102997451746,-.0132895981153155,-.0184910987778976,-.021047415089364,-.0208442545176493,-.0182969677820565,-.0142242276315438,-.00963355584556567,-.00547205447566498,-.00240394708818667,-.000669204128457476,-5.67729905547983e-005,-.0100949464135448,-.0219893659582073,-.0305959243062406,-.0348256816239572,-.0344895260742596,-.0302747093624672,-.0235358318700697,-.0159399692247158,-.00905422476771311,-.00397764265011023,-.00110728513787305,-9.39382857945563e-005,-.0139798092182615,-.0304515871935496,-.0423702283435318,-.0482277334672722,-.0477622143590056,-.0419253994709405,-.0325931833471828,-.0220741863877053,-.0125385841278022,-.00550836855481976,-.0015334043731933,-.000130088785011913,-.0163811720361289,-.0356823673916168,-.0496483170026962,-.0565119871455732,-.0559665041221956,-.0491270782522392,-.0381918333275861,-.0258659498945255,-.0146923824553127,-.00645456111210986,-.00179680283515025,-.000152434610085198,-.0160470195082919,-.0349544980281293,-.04863556220147,-.0553592232703062,-.0548248673220092,-.0481249559775311,-.0374127744408963,-.0253383214364947,-.0143926788244271,-.00632289727835408,-.00176015062198346,-.000149325161617318,-.0124076008346097,-.027026916660946,-.0376051541440983,-.0428039078843937,-.0423907423550166,-.0372103519687145,-.0289276629306809,-.0195916617438277,-.0111284599424805,-.0048888820448881,-.00136095343531401,-.000115458636973295,-.00592778988329973,-.0129122370468795,-.0179660399513751,-.0204497690975875,-.0202523773150973,-.0177774213479609,-.0138203275519371,-.00936000890342516,-.00531667428240711,-.00233568648062049,-.000650201929690796,-5.51609089712794e-005},
			{-5.59951004436151e-009,-1.21971598964522e-008,-1.6971084189159e-008,-1.9317264903302e-008,-1.91308046389391e-008,-1.67929112469387e-008,-1.30549605277777e-008,-8.84165344957659e-009,-5.02223790536332e-009,-2.20633662228147e-009,-6.14193874587905e-010,-5.21061086714588e-011,-1.78466133868042e-006,-3.88744721171111e-006,-5.40897999788056e-006,-6.15674863851456e-006,-6.09732050598647e-006,-5.35219317920171e-006,-4.16084320723466e-006,-2.81798888768115e-006,-1.60067465766628e-006,-7.03197893870279e-007,-1.95754280954478e-007,-1.66071240016209e-008,-3.69202255792134e-005,-8.04216603301225e-005,-.000111898407472009,-.00012736788971769,-.000126138468756338,-.000110723628756234,-8.60775467486887e-005,-5.82972148036964e-005,-3.31139965656822e-005,-1.45474238197567e-005,-4.04967152831355e-006,-3.43560288483173e-007,-.00027033225085123,-.000588852535789363,-.000819327289690437,-.000932595827175549,-.000923593928337934,-.000810725484866511,-.000630265297550785,-.000426855931918081,-.00024246279880047,-.000106517166772019,-2.96519536996954e-005,-2.51557038538368e-006,-.0011167825595386,-.0024326370236585,-.00338476236112819,-.00385269146248218,-.00381550328536708,-.00334922703163026,-.00260371927496907,-.00176340506432601,-.00100164972615974,-.000440038189183426,-.000122496611646568,-1.03921937720791e-005,-.00315739828904347,-.0068776181278623,-.00956949300162804,-.0108924349936834,-.0107872955591597,-.00946902654323017,-.00736130656207001,-.00498554717338517,-.00283189158407368,-.0012440880400349,-.000346325780898917,-2.93810953216598e-005,-.00672017616832581,-.0146382563131457,-.0203676169191767,-.0231833539384622,-.0229595761766608,-.0201537850748433,-.0156677341271221,-.0106111906809188,-.00602737082635835,-.00264790502577821,-.000737116463054399,-6.253444086072e-005,-.0113127416048293,-.0246420342366014,-.0342868373598303,-.039026847819149,-.0386501404333965,-.0339268729271671,-.026375056735671,-.0178628737232513,-.010146473397001,-.00445748215529598,-.00124086152956873,-.00010527045022929,-.0153830666658602,-.0335082395307456,-.0466232433562013,-.0530687098434273,-.0525564631192487,-.0461337637005683,-.0358648036217386,-.0242899366861979,-.0137971750917403,-.00606128448364179,-.00168732357718739,-.000143146764099252,-.0168026502756542,-.0366004544098804,-.0509257399611718,-.0579660084330403,-.0574064904417204,-.0503910900341255,-.039174487477097,-.0265314660737799,-.0150704090994405,-.00662063329842476,-.00184303355014811,-.000156356665903898,-.0139770501826433,-.0304455773108406,-.0423618662144552,-.0482182153091671,-.0477527880751015,-.0419171251326689,-.0325867508020486,-.0220698298571165,-.0125361095303546,-.00550728143089673,-.0015331017426483,-.000130063110871032,-.00694919005368325,-.015137106919017,-.0210617158489217,-.0239734091138143,-.0237420053295095,-.0208405969245687,-.0162016678481028,-.0109728047138306,-.00623277490756985,-.0027381417997588,-.000762236326125946,-6.46655241703613e-005},
			{-5.85380492906022e-011,-1.27510789616824e-010,-1.77418051741924e-010,-2.01945348094836e-010,-1.99996066807805e-010,-1.75555407262082e-010,-1.36478355571743e-010,-9.24318630275365e-011,-5.25031668349854e-011,-2.30653469542085e-011,-6.4208673651397e-012,-5.44724437242996e-013,-9.70149149662404e-008,-2.11323208782435e-007,-2.94034348800585e-007,-3.34683355709183e-007,-3.31452818296366e-007,-2.9094739428102e-007,-2.26185126101403e-007,-1.53187020074262e-007,-8.70133242853409e-008,-3.82261230182231e-008,-1.06412821914539e-008,-9.02770003435125e-010,-4.81635129901971e-006,-1.04912405632317e-005,-1.45974741955391e-005,-1.6615513352675e-005,-1.64551317961879e-005,-1.44442208796392e-005,-1.12290674717028e-005,-7.60504200188573e-006,-4.31981760329904e-006,-1.89775394143646e-006,-5.28291483055767e-007,-4.48184434349486e-008,-6.3249292855151e-005,-.000137773078747959,-.000191696963742466,-.000218198259374459,-.0002160920965543,-.00018968441040975,-.000147462369938119,-9.98709394081982e-005,-5.67287126102093e-005,-2.49216860143408e-005,-6.93762988830925e-006,-5.88564803133452e-007,-.000400261404288981,-.000871871344066475,-.00121311863645989,-.00138082716451303,-.00136749870422512,-.00120038256642976,-.000933188224038868,-.000632014693772654,-.000358997123096203,-.000157712261925103,-4.39035024137147e-005,-3.7246230587395e-006,-.00156589918054001,-.00341092748034313,-.00474595216619707,-.00540206000930644,-.0053499165230271,-.00469612622392589,-.00365081084424778,-.00247256238164128,-.00140446541897081,-.000617000537806862,-.000171758899848453,-1.45714378978438e-005,-.00428711876993237,-.00933843724141566,-.0129934678207622,-.0147897598708826,-.0146470014343651,-.0128570543562206,-.00999518991411039,-.00676938127811489,-.00384514541818346,-.00168922406982674,-.000470241515286636,-3.98936826157651e-005,-.00877289047128288,-.0191095911936167,-.0265890161087102,-.0302648352907365,-.0299727033964202,-.0263098681663614,-.0204535285962103,-.01385243649599,-.00786846397549439,-.00345672199472992,-.000962272689432589,-8.16359254002543e-005,-.0138436015212964,-.0301548921174003,-.0419574078870825,-.0477578423262084,-.0472968588510568,-.0415169130590646,-.0322756223296346,-.0218591137752465,-.0124164185359414,-.0054546995681281,-.00151846414940855,-.000128821307499823,-.0168604597524974,-.036726378183032,-.0511009498439408,-.0581654403421276,-.0576039973305382,-.0505644604551389,-.0393092672046696,-.0266227475173882,-.015122258804788,-.00664341157096352,-.00184937450253303,-.000156894610627413,-.0150886871005195,-.0328670058156056,-.045731033082912,-.0520531552679814,-.0515507112035252,-.0452509204027591,-.0351784732865376,-.0238251099283638,-.0135331440961878,-.00594529211811034,-.00165503394391524,-.000140407421995964,-.00780699308131397,-.0170056205219542,-.0236615589216733,-.0269326695112991,-.0266727014101101,-.0234131452361595,-.0182015900873001,-.0123272798443727,-.00700214416426768,-.00307613605632261,-.000856326230599691,-7.26477900154669e-005},
			{-9.64305560601928e-005,-.00109223938839564,-.0036470340953271,-.00744526010016809,-.0112949614634506,-.0137194545508608,-.0137194545508608,-.0112949614634506,-.00744526010016809,-.0036470340953271,-.00109223938839564,-9.64305560601928e-005,-.000218593396043406,-.00247594047941341,-.00826727129817659,-.0168772716472393,-.025603958800506,-.0310999156767677,-.0310999156767677,-.025603958800506,-.0168772716472393,-.00826727129817659,-.00247594047941341,-.000218593396043406,-.000327214252095852,-.00370625566402853,-.0123753463904701,-.0252637267155745,-.0383267764774421,-.0465537195204054,-.0465537195204054,-.0383267764774421,-.0252637267155745,-.0123753463904701,-.00370625566402854,-.000327214252095852,-.000415292144143345,-.00470388698413707,-.0157064800939949,-.0320640900252994,-.0486433860367337,-.0590848163661698,-.0590848163661698,-.0486433860367337,-.0320640900252994,-.0157064800939949,-.00470388698413707,-.000415292144143345,-.000477279350685986,-.00540599709662285,-.0180508558289433,-.0368500302339686,-.0559039799577005,-.0679039158056228,-.0679039158056228,-.0559039799577005,-.0368500302339686,-.0180508558289433,-.00540599709662285,-.000477279350685986,-.000509278548739481,-.00576844221712266,-.019261075608775,-.0393206408188163,-.059652062760935,-.0724565344080177,-.0724565344080177,-.059652062760935,-.0393206408188163,-.019261075608775,-.00576844221712266,-.000509278548739481,-.000509278548739481,-.00576844221712266,-.019261075608775,-.0393206408188163,-.059652062760935,-.0724565344080177,-.0724565344080177,-.059652062760935,-.0393206408188163,-.019261075608775,-.00576844221712266,-.000509278548739481,-.000477279350685986,-.00540599709662285,-.0180508558289433,-.0368500302339686,-.0559039799577005,-.0679039158056228,-.0679039158056228,-.0559039799577005,-.0368500302339686,-.0180508558289433,-.00540599709662285,-.000477279350685986,-.000415292144143345,-.00470388698413707,-.0157064800939949,-.0320640900252994,-.0486433860367337,-.0590848163661699,-.0590848163661699,-.0486433860367337,-.0320640900252994,-.0157064800939949,-.00470388698413707,-.000415292144143345,-.000327214252095852,-.00370625566402853,-.0123753463904701,-.0252637267155745,-.0383267764774421,-.0465537195204054,-.0465537195204054,-.0383267764774421,-.0252637267155745,-.0123753463904701,-.00370625566402854,-.000327214252095852,-.000218593396043406,-.00247594047941341,-.0082672712981766,-.0168772716472393,-.025603958800506,-.0310999156767677,-.0310999156767677,-.025603958800506,-.0168772716472393,-.0082672712981766,-.00247594047941341,-.000218593396043406,-9.64305560601928e-005,-.00109223938839564,-.0036470340953271,-.00744526010016809,-.0112949614634506,-.0137194545508608,-.0137194545508608,-.0112949614634506,-.00744526010016809,-.0036470340953271,-.00109223938839564,-9.64305560601928e-005},
			{-1.53989561292143e-006,-1.74419262023187e-005,-5.82393385771156e-005,-.000118893054584807,-.000180368778489956,-.000219085409622217,-.000219085409622217,-.000180368778489956,-.000118893054584807,-5.82393385771156e-005,-1.74419262023187e-005,-1.53989561292143e-006,-1.81513161714798e-005,-.000205594401647313,-.000686488512117749,-.00140143617934442,-.00212607315610414,-.00258244033247591,-.00258244033247591,-.00212607315610414,-.00140143617934442,-.000686488512117749,-.000205594401647313,-1.81513161714798e-005,-6.52040263973911e-005,-.000738546046221744,-.00246603687813908,-.00503430609488318,-.00763738171291507,-.00927676571867675,-.00927676571867675,-.00763738171291507,-.00503430609488318,-.00246603687813908,-.000738546046221744,-6.52040263973911e-005,-.000148422561475078,-.00168113691751877,-.00561338816585801,-.0114594856657818,-.0173848122489431,-.0211165077717546,-.0211165077717546,-.0173848122489431,-.0114594856657818,-.005613388165858,-.00168113691751877,-.000148422561475078,-.000261298026375791,-.00295964275410309,-.00988237390894377,-.0201744327681154,-.0306059744853902,-.0371756271409923,-.0371756271409923,-.0306059744853902,-.0201744327681154,-.00988237390894377,-.00295964275410309,-.000261298026375791,-.000385814196380375,-.0043699992938526,-.0145916147966848,-.0297881414330556,-.0451906185986146,-.0548909033461696,-.0548909033461696,-.0451906185986146,-.0297881414330556,-.0145916147966848,-.0043699992938526,-.000385814196380375,-.000496282125391267,-.0056212357084272,-.0187695467718091,-.0383172062628759,-.0581297848971723,-.0706074956102091,-.070607495610209,-.0581297848971723,-.0383172062628759,-.0187695467718091,-.0056212357084272,-.000496282125391267,-.00056537405855632,-.00640381888440892,-.0213826255122005,-.0436516918684151,-.066222559163116,-.080437405087215,-.080437405087215,-.066222559163116,-.0436516918684151,-.0213826255122005,-.00640381888440892,-.00056537405855632,-.000570884532287664,-.00646623433345328,-.0215910333696339,-.0440771473659381,-.0668680038331856,-.0812213961477219,-.0812213961477219,-.0668680038331856,-.0440771473659381,-.0215910333696339,-.00646623433345328,-.000570884532287664,-.0005015476832896,-.00568087707080664,-.0189686918350622,-.038723752172464,-.0587465424476329,-.0713566417856807,-.0713566417856807,-.0587465424476329,-.038723752172464,-.0189686918350622,-.00568087707080664,-.0005015476832896,-.000360463552039073,-.00408286030594201,-.0136328454187137,-.027830855811784,-.0422212843688097,-.0512841937399486,-.0512841937399486,-.0422212843688097,-.027830855811784,-.0136328454187137,-.00408286030594202,-.000360463552039073,-.000165482726913838,-.00187437218884842,-.00625860901253902,-.0127766757166634,-.0193830783490639,-.0235437069286394,-.0235437069286394,-.0193830783490639,-.0127766757166634,-.00625860901253902,-.00187437218884842,-.000165482726913838},
			{-1.83286990233606e-008,-2.076035629087e-007,-6.93197187616125e-007,-1.41513164604626e-006,-2.14685010231423e-006,-2.6076771046559e-006,-2.6076771046559e-006,-2.14685010231423e-006,-1.41513164604626e-006,-6.93197187616125e-007,-2.076035629087e-007,-1.83286990233606e-008,-1.1234220751336e-006,-1.27246579340282e-005,-4.24881778022507e-005,-8.67377509097766e-005,-.000131587015197788,-.000159832512959983,-.000159832512959983,-.000131587015197788,-8.67377509097766e-005,-4.24881778022507e-005,-1.27246579340282e-005,-1.1234220751336e-006,-9.68457050125128e-006,-.000109694165348983,-.00036627351598611,-.000747731313456525,-.00113435880773542,-.00137785189949114,-.00137785189949114,-.00113435880773542,-.000747731313456525,-.00036627351598611,-.000109694165348983,-9.68457050125128e-006,-3.95375611173333e-005,-.000447829851219516,-.00149532305248749,-.00305263640770766,-.00463105521138285,-.00562512335262842,-.00562512335262842,-.00463105521138285,-.00305263640770766,-.00149532305248749,-.000447829851219516,-3.95375611173333e-005,-.000106626059868886,-.001207720738907,-.00403263127041889,-.00823243981590666,-.0124891661566992,-.015169998412074,-.015169998412074,-.0124891661566992,-.00823243981590666,-.00403263127041889,-.001207720738907,-.000106626059868886,-.000217853615265421,-.00246756120900883,-.00823929255543563,-.0168201542714382,-.0255173078910835,-.0309946649224858,-.0309946649224858,-.0255173078910835,-.0168201542714382,-.00823929255543563,-.00246756120900883,-.000217853615265422,-.000360467098389462,-.00408290047436723,-.0136329795428058,-.0278311296202336,-.0422216997546855,-.0512846982894915,-.0512846982894915,-.0422216997546855,-.0278311296202336,-.0136329795428058,-.00408290047436723,-.000360467098389462,-.000499186529428583,-.00565413300383832,-.0188793922500835,-.038541450987549,-.0584699792610344,-.0710207135860798,-.0710207135860798,-.0584699792610344,-.038541450987549,-.0188793922500835,-.00565413300383832,-.000499186529428583,-.000584933620267579,-.00662536405218185,-.0221223743155372,-.0451618565956071,-.0685135808557914,-.0832202005920431,-.0832202005920431,-.0685135808557914,-.0451618565956071,-.0221223743155372,-.00662536405218185,-.000584933620267579,-.000573001809320343,-.00649021608224453,-.0216711094559173,-.0442406191829269,-.0671160015993357,-.0815226272844922,-.0815226272844922,-.0671160015993357,-.0442406191829269,-.0216711094559173,-.00649021608224453,-.000573001809320343,-.000443046618338206,-.00501825341691917,-.0167561630764652,-.0342069717815257,-.0518942821144625,-.0630335258090745,-.0630335258090745,-.0518942821144625,-.0342069717815257,-.0167561630764652,-.00501825341691917,-.000443046618338206,-.000211667613829872,-.00239749426445693,-.00800533602681095,-.016342542282569,-.0247927383077269,-.0301145645786321,-.0301145645786321,-.0247927383077269,-.016342542282569,-.00800533602681095,-.00239749426445693,-.000211667613829872},
			{-1.99945502952237e-010,-2.26472150301291e-009,-7.56200209007358e-009,-1.54374955010019e-008,-2.34197213300945e-008,-2.84468259074429e-008,-2.84468259074429e-008,-2.34197213300945e-008,-1.54374955010019e-008,-7.56200209007358e-009,-2.26472150301291e-009,-1.99945502952237e-010,-6.37261128446745e-008,-7.21806171840919e-007,-2.4101417206608e-006,-4.92019858316594e-006,-7.46427292555284e-006,-9.0664986763096e-006,-9.0664986763096e-006,-7.46427292555284e-006,-4.92019858316594e-006,-2.4101417206608e-006,-7.21806171840918e-007,-6.37261128446745e-008,-1.31833553544195e-006,-1.49323830304632e-005,-4.98598664497681e-005,-.000101786729867376,-.0001544173307413,-.000187563415583651,-.000187563415583651,-.0001544173307413,-.000101786729867376,-4.98598664497681e-005,-1.49323830304632e-005,-1.31833553544195e-006,-9.65293703064032e-006,-.000109335862711265,-.000365077128133699,-.000745288940144662,-.0011306535627788,-.00137335131399144,-.00137335131399144,-.0011306535627788,-.000745288940144662,-.000365077128133699,-.000109335862711265,-9.65293703064033e-006,-3.98777122973611e-005,-.000451682639505874,-.00150818767757949,-.0030788989754264,-.00467089729699484,-.00567351764636142,-.00567351764636142,-.00467089729699484,-.0030788989754264,-.00150818767757949,-.000451682639505874,-3.98777122973611e-005,-.000112743362173094,-.00127700954942894,-.00426398957619228,-.0087047476468162,-.0132056889748729,-.0160403247315041,-.0160403247315041,-.0132056889748729,-.0087047476468162,-.00426398957619228,-.00127700954942894,-.000112743362173094,-.000239961888318527,-.00271797484991882,-.00907543442693066,-.0185271012182458,-.0281068614761773,-.0341400729730291,-.0341400729730291,-.0281068614761773,-.0185271012182458,-.00907543442693066,-.00271797484991882,-.000239961888318527,-.000403951737210287,-.004575437666423,-.0152775823210325,-.0311885140387246,-.0473150782417566,-.0574713838208962,-.0574713838208962,-.0473150782417566,-.0311885140387246,-.0152775823210325,-.004575437666423,-.000403951737210287,-.000549293594811998,-.00622168039425801,-.0207744572931185,-.042410143131176,-.0643390460348379,-.0781495909283993,-.0781495909283993,-.0643390460348379,-.042410143131176,-.0207744572931185,-.00622168039425801,-.000549293594811998,-.000599983629581883,-.00679583090045496,-.0226915704225263,-.0463238454758227,-.0702763963177652,-.0853614090140783,-.0853614090140783,-.0702763963177652,-.0463238454758227,-.0226915704225263,-.00679583090045496,-.000599983629581883,-.000499088010632534,-.00565301711159485,-.0188756662380921,-.0385338444975389,-.058458439702929,-.0710066970315766,-.0710066970315766,-.058458439702929,-.0385338444975389,-.0188756662380921,-.00565301711159485,-.000499088010632534,-.000248139442448811,-.00281059950217384,-.00938471210765786,-.0191584780345857,-.0290647026682284,-.0353035172878956,-.0353035172878956,-.0290647026682284,-.0191584780345857,-.00938471210765786,-.00281059950217384,-.000248139442448811},
			{-1.92058726559064e-006,-.000113117990564565,-.000906410211953001,-.00331870648655728,-.00771242584002074,-.0129629077002378,-.0166745014700866,-.0166874796528932,-.0127648935683974,-.00697208388856876,-.00224638876284317,-.000206393222578878,-4.35367906124191e-006,-.000256421270615505,-.00205469401540682,-.00752300257311598,-.0174828957229149,-.0293849390956891,-.0377985573514871,-.0378279769169244,-.0289360711920697,-.0158046531822386,-.00509222147590975,-.000467862027215704,-6.51705798837019e-006,-.000383839108612685,-.00307568836802217,-.0112612444155839,-.0261702903746801,-.0439865569734266,-.0565809713282148,-.0566250097176437,-.0433146429173246,-.0236581153148544,-.0076225882021403,-.000700346515907722,-8.27128698753606e-006,-.000487158995675827,-.00390358674443833,-.0142924897345245,-.0332146779455411,-.0558266378740331,-.071811153548789,-.0718670459711192,-.0549738613606283,-.0300262881966959,-.00967439828220347,-.000888862280208502,-9.50587324711357e-006,-.000559873169805834,-.00448624269205744,-.0164258108812803,-.0381723568496402,-.0641594160912017,-.0825298075610779,-.0825940425807241,-.063179352715725,-.0345080626641537,-.0111184153023947,-.00102153536475457,-1.01431945983696e-005,-.000597409871152132,-.00478702287082038,-.0175270795090219,-.0407316228334812,-.0684609846789651,-.0880630191983885,-.0881315608554635,-.067415212946293,-.0368216560137206,-.0118638495492168,-.00109002421181751,-1.01431945983696e-005,-.000597409871152132,-.00478702287082038,-.0175270795090219,-.0407316228334812,-.0684609846789651,-.0880630191983885,-.0881315608554635,-.067415212946293,-.0368216560137206,-.0118638495492168,-.00109002421181751,-9.50587324711357e-006,-.000559873169805834,-.00448624269205744,-.0164258108812803,-.0381723568496403,-.0641594160912017,-.0825298075610779,-.0825940425807241,-.063179352715725,-.0345080626641537,-.0111184153023947,-.00102153536475457,-8.27128698753606e-006,-.000487158995675827,-.00390358674443833,-.0142924897345245,-.0332146779455411,-.0558266378740331,-.071811153548789,-.0718670459711192,-.0549738613606283,-.0300262881966959,-.00967439828220347,-.000888862280208502,-6.51705798837019e-006,-.000383839108612685,-.00307568836802217,-.0112612444155839,-.0261702903746801,-.0439865569734266,-.0565809713282148,-.0566250097176437,-.0433146429173246,-.0236581153148544,-.0076225882021403,-.000700346515907722,-4.35367906124191e-006,-.000256421270615505,-.00205469401540682,-.00752300257311598,-.0174828957229149,-.0293849390956891,-.0377985573514871,-.0378279769169244,-.0289360711920697,-.0158046531822386,-.00509222147590975,-.000467862027215704,-1.92058726559064e-006,-.000113117990564565,-.000906410211953001,-.00331870648655728,-.00771242584002074,-.0129629077002378,-.0166745014700866,-.0166874796528932,-.0127648935683974,-.00697208388856876,-.00224638876284317,-.000206393222578878},
			{-3.06697796211909e-008,-1.80637657239712e-006,-1.44744276633774e-005,-5.29962883716334e-005,-.000123159413319067,-.000207004143752805,-.000266274433108618,-.000266481681179921,-.000203842064166924,-.000111336922926592,-3.58724904274172e-005,-3.29588494374136e-006,-3.61515976889952e-007,-2.1292425288575e-005,-.000170615404521307,-.000624686750242411,-.00145172531948917,-.00244003400654881,-.00313867471481826,-.0031411176240886,-.00240276141102928,-.00131236927532137,-.000422842243423972,-3.88498085043389e-005,-1.29865498884588e-006,-7.6487668853577e-005,-.000612892819182374,-.00224402963223718,-.00521495715018533,-.00876520690127838,-.0112748974798531,-.0112836730154145,-.00863131450035304,-.00471435570086283,-.00151895413763271,-.000139558140870839,-2.95610118832487e-006,-.00017410728078841,-.0013951151049833,-.00510803771553511,-.0118706978844523,-.0199520571355211,-.025664813306616,-.0256847888746645,-.0196472806638003,-.0107311892759846,-.00345756353291146,-.000317673277053615,-5.20421827112933e-006,-.000306515993225909,-.00245610114713186,-.00899270407717994,-.0208983721752537,-.0351256109572267,-.0451829223109582,-.0452180893128947,-.034589051759252,-.0188922664493339,-.00608704308994404,-.000559264033051857,-7.68418084863826e-006,-.000452579850083181,-.00362650534890298,-.0132779912077635,-.0308570591913224,-.0518639943508112,-.0667139094133409,-.0667658345226861,-.051071748964773,-.0278949852742899,-.00898769757526597,-.000825769740663992,-9.88434754146089e-006,-.000582164399381254,-.0046648614778342,-.0170798010009427,-.0396921810095751,-.0667139094133409,-.0858157140598609,-.0858825065829686,-.0656948250258218,-.0358819937409303,-.0115610926110904,-.0010622075751108,-1.12604371581811e-005,-.000663212782377944,-.00531429912823647,-.0194576349160754,-.0452180893128947,-.0760017574629395,-.0977628974803007,-.0978389887959829,-.0748407971004666,-.0408774513376,-.0131706171076067,-.00121008711989414,-1.13701881137203e-005,-.000669676850830108,-.00536609546608301,-.0196472806638003,-.0456588118567917,-.0767425160486888,-.0987157531522446,-.0987925860994006,-.0755702402721397,-.0412758674276048,-.0132989858194334,-.0012218813527315,-9.98922055945599e-006,-.000588341168993646,-.00471435570086283,-.017261017846082,-.040113315413826,-.067421744603583,-.0867262195724516,-.0867937207648416,-.0663918477213734,-.0362627020233443,-.01168375590964,-.0010734776072168,-7.17927735473982e-006,-.000422842243423972,-.00338821902309026,-.0124055359279088,-.0288295383268322,-.0484561734690455,-.0623303470508818,-.06237886032367,-.0477159840498455,-.0260620930240078,-.00839714407356662,-.000771510993319342,-3.29588494374136e-006,-.000194119730276028,-.00155547411146173,-.00569517196000836,-.0132351539871473,-.0222454106000665,-.0286148092951926,-.0286370809191255,-.0219056021427923,-.0119646666030396,-.00385498697918716,-.000354187662792186},
			{-3.65048874112027e-010,-2.15005044744533e-008,-1.72282735226472e-007,-6.30791470989705e-007,-1.46591223424896e-006,-2.46387911966847e-006,-3.16934726012697e-006,-3.1718140426088e-006,-2.4262422795302e-006,-1.32519433994785e-006,-4.26974448589619e-007,-3.92294663599167e-008,-2.23749630651595e-008,-1.31783174148499e-006,-1.05597362732147e-005,-3.86631403796111e-005,-8.98502486218331e-005,-.00015101869423293,-.000194258996301528,-.000194410192951716,-.000148711817078381,-8.12252180824206e-005,-2.61705711055768e-005,-2.40449409138482e-006,-1.92885569959679e-007,-1.1360498152647e-005,-9.10312452249138e-005,-.000333299404644144,-.000774562906136249,-.00130187150820567,-.00167462878541062,-.00167593219100564,-.00128198484678476,-.000700209981991998,-.000225605982417451,-2.07281778267435e-005,-7.87461355147049e-007,-4.63795880236092e-005,-.000371637897746933,-.00136070521452511,-.00316217722165629,-.00531493103550084,-.00683672424538222,-.00684204543937499,-.00523374312001741,-.00285862908989082,-.000921043459502259,-8.46234324557743e-005,-2.12364898657175e-006,-.00012507783964276,-.00100224403364473,-.00366958991823735,-.00852785271078683,-.0143334626308517,-.0184374794017228,-.0184518297544813,-.0141145126680233,-.00770923516443232,-.00248389714184165,-.000228214966233319,-4.33894499945229e-006,-.000255553469660892,-.00204774035893469,-.00749754263838111,-.0174237287374253,-.0292854922824927,-.0376706365121829,-.0376999565136218,-.0288381434728016,-.0157511658366454,-.00507498798101565,-.000466278651886231,-7.17934798611665e-006,-.000422846403442921,-.00338825235716593,-.0124056579764159,-.0288298219584409,-.0484566501919911,-.0623309602711284,-.0623794740212014,-.0477164534906379,-.0260623494288551,-.00839722668659549,-.000771518583621402,-9.94219394991092e-006,-.000585571414309697,-.00469216176055673,-.0171797575373556,-.0399244725296093,-.0671043408542665,-.0863179354585031,-.0863851188731245,-.0660792924543047,-.0360919868089839,-.0116287518756503,-.00106842395843715,-1.16500008667718e-005,-.000686157151895624,-.00549815149985173,-.0201307871491398,-.0467824447922306,-.0786310982319312,-.101145082058928,-.101223805813716,-.0774299735296571,-.0422916390211788,-.0136262649987847,-.00125195103863209,-1.14123574777427e-005,-.000672160525384795,-.00538599705705276,-.0197201478250184,-.0458281496934892,-.0770271360622618,-.0990818667545695,-.0991589846571268,-.0758505125894862,-.0414289499502059,-.0133483086422873,-.00122641302441892,-8.82406705447926e-006,-.000519716417833643,-.00416446814598573,-.015247674029803,-.0354344548588437,-.0595575992912496,-.0766103793918513,-.0766700071720531,-.0586478306968206,-.0320329811846719,-.0103209499661349,-.000948266016470748,-4.21573969958762e-006,-.000248296972548114,-.00198959431997835,-.00728464826218446,-.0169289781185245,-.0284539242725722,-.0366009704831913,-.0366294579367253,-.028019278030959,-.0153039080100614,-.00493088257841441,-.00045303856676568},
			{-3.06697796211909e-008,-9.39296633663941e-006,-.000180620480252622,-.00118608291592308,-.00422235247914266,-.00982031115806883,-.0162489801425508,-.0197676016789817,-.0175473588745023,-.0106866754670937,-.00370432632910279,-.000354187662792186,-6.95237231559063e-008,-2.1292425288575e-005,-.000409439142389331,-.00268866947546018,-.00957143052386794,-.0222611509665594,-.0368339652566651,-.0448101448253052,-.0397771922581837,-.024225067014107,-.00839714407356662,-.000802889532215207,-1.04070632906319e-007,-3.18728065084108e-005,-.000612892819182374,-.00402469144754472,-.0143275530598965,-.0333228999419048,-.0551370654893481,-.0670766742761412,-.0595428061909519,-.0362627020233444,-.0125697540159339,-.00120185194317225,-1.32083844163853e-007,-4.04521687853206e-005,-.000777868235805777,-.00510803771553511,-.0181841719682459,-.0422925910998948,-.0699785843745455,-.085132037201067,-.0755702402721397,-.0460237143683289,-.0159532173895116,-.00152536042432723,-1.51798901731387e-007,-4.64901277906996e-005,-.000893974161901044,-.00587047204848792,-.0208983721752537,-.0486052546469995,-.0804237060180922,-.0978389887959829,-.0868499819149511,-.0528932916734667,-.0183344215497376,-.00175303829641842,-1.61976260365957e-007,-4.96070587967628e-005,-.000953910667053138,-.00626405789601532,-.0222995037123093,-.0518639943508112,-.0858157140598609,-.104398604617067,-.0926728396779039,-.0564395228555362,-.0195636529956887,-.00187057076364506,-1.61976260365957e-007,-4.96070587967628e-005,-.000953910667053138,-.00626405789601532,-.0222995037123093,-.0518639943508112,-.0858157140598609,-.104398604617067,-.0926728396779039,-.0564395228555362,-.0195636529956887,-.00187057076364506,-1.51798901731387e-007,-4.64901277906996e-005,-.000893974161901044,-.00587047204848792,-.0208983721752537,-.0486052546469995,-.0804237060180922,-.0978389887959829,-.0868499819149511,-.0528932916734667,-.0183344215497376,-.00175303829641842,-1.32083844163853e-007,-4.04521687853206e-005,-.000777868235805777,-.00510803771553511,-.0181841719682459,-.0422925910998948,-.0699785843745455,-.085132037201067,-.0755702402721397,-.0460237143683289,-.0159532173895116,-.00152536042432723,-1.04070632906319e-007,-3.18728065084108e-005,-.000612892819182374,-.00402469144754472,-.0143275530598965,-.0333228999419048,-.0551370654893481,-.0670766742761412,-.0595428061909519,-.0362627020233443,-.0125697540159339,-.00120185194317225,-6.95237231559063e-008,-2.1292425288575e-005,-.000409439142389331,-.00268866947546018,-.00957143052386794,-.0222611509665594,-.0368339652566651,-.0448101448253052,-.0397771922581837,-.024225067014107,-.00839714407356662,-.000802889532215207,-3.06697796211909e-008,-9.39296633663941e-006,-.000180620480252622,-.00118608291592308,-.00422235247914266,-.00982031115806883,-.0162489801425508,-.0197676016789817,-.0175473588745023,-.0106866754670937,-.00370432632910279,-.000354187662792186},
			{-4.89764458734161e-010,-1.49995895979437e-007,-2.8843210756594e-006,-1.89405096647495e-005,-6.74265743697526e-005,-.000156820147987287,-.00025947930056204,-.000315668024236135,-.00028021305752868,-.000170655084270396,-5.91542359272188e-005,-5.6560083280767e-006,-5.77303387673914e-009,-1.76805681469661e-006,-3.3998553762376e-005,-.000223258756300761,-.000794781840714042,-.00184849678399644,-.00305857798732252,-.00372089514708473,-.00330297440936967,-.00201157426832706,-.000697272743806459,-6.66694512082365e-005,-2.07381685017084e-008,-6.35129828208852e-006,-.000122131231479417,-.000802000786155484,-.00285505335439599,-.00664025858844208,-.0109871701831305,-.0133663775728643,-.0118651027035077,-.00722607333006517,-.00250477651118074,-.000239493192417188,-4.72058591990341e-008,-1.44573274351898e-005,-.000278004767708733,-.00182557761480488,-.00649889823405786,-.0151150817366226,-.0250098656792164,-.030425605696751,-.0270082851124352,-.0164485595800252,-.00570157038226907,-.00054515334463801,-8.31059491191454e-008,-2.54521353623046e-005,-.000489427593779599,-.00321393070570632,-.0114413150217837,-.0266101122837137,-.044029886541206,-.0535643007426736,-.0475481053927164,-.0289577009874395,-.0100376187644472,-.000959742855872972,-1.22708370279532e-007,-3.75808240390211e-005,-.000722654190693695,-.00474546290931794,-.0168934370530446,-.0392906109117239,-.0650114182959186,-.0790892603834303,-.0702061715732879,-.0427568944566942,-.0148208383771267,-.00141708846333921,-1.57842742379509e-007,-4.83411222371217e-005,-.000929567551026009,-.00610420363142845,-.0217304363719754,-.0505404624146264,-.0836257585887017,-.101734427107537,-.0903078952792718,-.0549992267137261,-.0190644026033435,-.00182283513943086,-1.79817461292618e-007,-5.50711280460104e-005,-.00105898107575704,-.00695402514977553,-.0247557273922317,-.0575766583052787,-.0952680584574148,-.115897798864594,-.102880475966099,-.062656167598386,-.0217185308967278,-.0020766085420598,-1.81570069821245e-007,-5.56078841986148e-005,-.00106930253870999,-.00702180323816678,-.0249970115180785,-.0581378348543761,-.0961965979360501,-.11702740813212,-.103883210618848,-.0632668520831664,-.0219302126889477,-.00209684841095391,-1.59517455322627e-007,-4.88540219870558e-005,-.000939430271260956,-.00616896928789685,-.0219609965010198,-.0510766971839986,-.0845130286536747,-.102813830311442,-.0912660628757668,-.0555827690131719,-.0192666760896712,-.00184217543696469,-1.14645586997212e-007,-3.51115056126793e-005,-.000675170843678735,-.00443365338137005,-.0157834221327753,-.0367089477367546,-.0607397213008226,-.0738925524084332,-.0655931435851624,-.0399474726296093,-.0138470074345037,-.00132397601187602,-5.2631852118241e-008,-1.61190990378048e-005,-.000309959527704379,-.00203541536332734,-.00724590244919067,-.0168524577290889,-.0278845798860605,-.0339228224379417,-.0301127040609837,-.0183392097943569,-.00635692717584237,-.000607815019228362},
			{-4.34129792523551e-010,-6.91363066088356e-007,-3.19037343548278e-005,-.000375745321421736,-.00204903932812416,-.00659447754217011,-.0140356226122099,-.020756280852719,-.0213815233314377,-.0145196103881163,-.00541459382780606,-.000538770430325433,-9.84106174935942e-010,-1.56721486103326e-006,-7.23209107571181e-005,-.000851757463741354,-.00464486033951284,-.0149486770580804,-.0318166205582331,-.0470513300576056,-.0484686596091265,-.032913747194243,-.0122740602291765,-.0012213105768994,-1.47311662585924e-009,-2.34597681315439e-006,-.000108257766028679,-.00127500275172926,-.00695292963828415,-.0223767976156558,-.0476265609511484,-.0704315228803041,-.0725531351410981,-.04926885883459,-.0183731416903004,-.00182818984576065,-1.86964277444457e-009,-2.97745509129004e-006,-.000137398048789906,-.00161820160082515,-.00882448438314045,-.0284000718224987,-.0604464398750644,-.0893899271345745,-.0920826243480438,-.0625308032698983,-.0233187318655643,-.00232029282368979,-2.14870881134135e-009,-3.42187506483889e-006,-.000157906313511514,-.00185973710365748,-.0101416418199092,-.0326391144885199,-.0694687775381858,-.102732418569255,-.105827032314693,-.0718642565322132,-.0267993251511533,-.00266662364773693,-2.29276901154607e-009,-3.65129470714422e-006,-.000168493143619897,-.00198442319330673,-.0108215882804389,-.0348274041920398,-.0741263123085118,-.109620114430362,-.112922206580213,-.0766823962117034,-.0285960861297693,-.00284540745247398,-2.29276901154607e-009,-3.65129470714422e-006,-.000168493143619897,-.00198442319330673,-.0108215882804389,-.0348274041920399,-.0741263123085118,-.109620114430362,-.112922206580213,-.0766823962117034,-.0285960861297693,-.00284540745247398,-2.14870881134135e-009,-3.42187506483889e-006,-.000157906313511514,-.00185973710365748,-.0101416418199092,-.0326391144885199,-.0694687775381858,-.102732418569255,-.105827032314693,-.0718642565322132,-.0267993251511533,-.00266662364773693,-1.86964277444457e-009,-2.97745509129004e-006,-.000137398048789906,-.00161820160082515,-.00882448438314046,-.0284000718224987,-.0604464398750644,-.0893899271345745,-.0920826243480438,-.0625308032698983,-.0233187318655643,-.00232029282368979,-1.47311662585924e-009,-2.34597681315439e-006,-.000108257766028679,-.00127500275172926,-.00695292963828415,-.0223767976156558,-.0476265609511484,-.0704315228803041,-.0725531351410981,-.04926885883459,-.0183731416903004,-.00182818984576065,-9.84106174935942e-010,-1.56721486103326e-006,-7.23209107571182e-005,-.000851757463741354,-.00464486033951284,-.0149486770580804,-.0318166205582331,-.0470513300576056,-.0484686596091265,-.032913747194243,-.0122740602291765,-.0012213105768994,-4.34129792523551e-010,-6.91363066088356e-007,-3.19037343548278e-005,-.000375745321421736,-.00204903932812416,-.00659447754217011,-.0140356226122099,-.020756280852719,-.0213815233314377,-.0145196103881163,-.00541459382780606,-.000538770430325433},
			{-.00423058507690693,-.00885514027394949,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394863,-.00435422423400534,-.00213112724015142,-.000790234334926298,-.000193564800479007,-2.24537546398006e-005,-3.66334363412211e-007,-.00959009257018474,-.020073255449394,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134588,-.00987036372424941,-.00483094114415894,-.00179134098152266,-.000438781946510387,-5.08992448158822e-005,-8.30424254068273e-007,-.0143554884304917,-.0300478211505447,-.0388614993834644,-.0396707255608121,-.0338552805420154,-.0244472036457215,-.0147750286257641,-.00723147552495667,-.00268147304597068,-.000656816303965783,-7.61914981245153e-005,-1.2430688947403e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971213,-.0429682752411084,-.03102777936875,-.0187520967637297,-.00917800785523132,-.00340325851820642,-.000833614824033866,-9.670034364015e-005,-1.57767194829697e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.057864283158913,-.0493817925438395,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011006,-.000958041580001006,-.000111134000184102,-1.81315792676965e-006,-.0223429825183335,-.0467666388316429,-.0604843092288455,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102643,-.0112551190451086,-.00417346338855665,-.00102227348573477,-.000118584980154756,-1.93472111511538e-006,-.0223429825183335,-.0467666388316429,-.0604843092288455,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102643,-.0112551190451086,-.00417346338855665,-.00102227348573477,-.000118584980154756,-1.93472111511538e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.057864283158913,-.0493817925438395,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011006,-.000958041580001006,-.000111134000184102,-1.81315792676965e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971213,-.0429682752411084,-.03102777936875,-.0187520967637297,-.00917800785523132,-.00340325851820642,-.000833614824033866,-9.670034364015e-005,-1.57767194829697e-006,-.0143554884304917,-.0300478211505447,-.0388614993834644,-.0396707255608121,-.0338552805420154,-.0244472036457215,-.0147750286257641,-.00723147552495667,-.00268147304597068,-.000656816303965783,-7.61914981245153e-005,-1.2430688947403e-006,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134588,-.00987036372424941,-.00483094114415894,-.00179134098152266,-.000438781946510386,-5.08992448158822e-005,-8.30424254068273e-007,-.00423058507690692,-.00885514027394949,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394863,-.00435422423400534,-.00213112724015142,-.000790234334926298,-.000193564800479007,-2.24537546398006e-005,-3.66334363412211e-007},
			{-6.75580403638936e-005,-.00014140737349095,-.000182885225860587,-.00018669350692448,-.000159325572268953,-.000115050433754236,-6.95324290155738e-005,-3.40318838868514e-005,-1.26192198302079e-005,-3.09102839585264e-006,-3.58563091085661e-007,-5.84997849237861e-009,-.000796331478757963,-.00166682074010411,-.00215573544715162,-.00220062505725282,-.0018780291418385,-.00135614179377857,-.000819604324242625,-.000401146336906179,-.000148747387195263,-3.64350890004165e-005,-4.22651508264995e-006,-6.89558489033362e-008,-.00286062003831949,-.00598763321131698,-.00774393601400238,-.00790519061914258,-.00674634613725228,-.00487159743595245,-.00294422186735428,-.00144101706420306,-.000534337481072711,-.000130883869937062,-1.51826896966895e-005,-2.4770649960978e-007,-.00651156956023967,-.013629524241044,-.0176273595758088,-.0179944200606735,-.0153565666050425,-.0110891153486868,-.00670187065504814,-.0032801570028024,-.00121630123191481,-.000297928215558894,-3.45600389940658e-005,-5.63849124010232e-007,-.0114636229006531,-.0239947872427992,-.031032979229028,-.0316791894770017,-.0270352465684925,-.0195223955581165,-.0117986481151247,-.0057747187659129,-.00214129919480704,-.000524502838069343,-6.0842973540354e-005,-9.92656788923096e-007,-.0169263752901921,-.0354290068157069,-.0458211036210322,-.0467752519970179,-.0399183341468002,-.0288253893768114,-.0174210498412452,-.00852654155446687,-.00316169103729199,-.000774443817183375,-8.98364341570052e-005,-1.46568685042074e-006,-.0217728056224891,-.045573187736373,-.0589407930194454,-.0601681371359018,-.0513479179830939,-.0370787950245641,-.0224091174530978,-.0109678964760433,-.00406695959489998,-.000996185799262211,-.00011555877647658,-1.88534841928859e-006,-.0248039952501605,-.0519178442938048,-.0671464750773758,-.0685446888933965,-.0584965270825159,-.042240870176189,-.0255288938184798,-.012494836761645,-.00463315790456963,-.00113487385418326,-.000131646761218445,-2.1478248622506e-006,-.0250457498234202,-.0524238666569454,-.0678009247845504,-.0692127664287096,-.0590666692229003,-.042652574961671,-.0257777136867565,-.012616618914028,-.004678315432652,-.00114593501354709,-.000132929869292705,-2.16875884799657e-006,-.0220038152896698,-.0460567196758888,-.0595661553735154,-.0608065216222948,-.0518927198634471,-.0374722013795649,-.0226468783946324,-.0110842659581639,-.0041101100734715,-.00100675533972143,-.000116784856153978,-1.90535198329458e-006,-.015814196101363,-.0331010775700317,-.042810342192032,-.0437017964620066,-.0372954252410452,-.0269313631824755,-.0162763671345931,-.00796628917278541,-.00295394620634634,-.000723557535766244,-8.39335630014821e-005,-1.36938115091738e-006,-.00726003025826053,-.0151961454884669,-.0196535048439074,-.0200627564386384,-.0171216996431039,-.0123637338469658,-.00747220517149892,-.00365718877328773,-.00135610679808757,-.00033217304057603,-3.85324807640305e-005,-6.28659751468172e-007},
			{-8.04113589360719e-007,-1.68310966462922e-006,-2.17680226684608e-006,-2.22213055847631e-006,-1.89638209018582e-006,-1.36939462342768e-006,-8.27613867594724e-007,-4.05066519951766e-007,-1.50201013794113e-007,-3.67911491336473e-008,-4.26781849550599e-009,-6.96297165792815e-011,-4.92865836277514e-005,-.000103162944064127,-.000133422875058339,-.000136201184821656,-.00011623506394466,-8.39343887728836e-005,-5.07269876251226e-005,-2.48277670898173e-005,-9.20627996502583e-006,-2.25504216385422e-006,-2.61587661208472e-007,-4.26781849550599e-009,-.000424879842112779,-.000889326306666056,-.001150185018243,-.00117413571083906,-.00100201580190209,-.000723564654409836,-.000437296986453802,-.000214030208318087,-7.93636419908757e-005,-1.94398127850062e-005,-2.25504216385422e-006,-3.67911491336473e-008,-.00173458520673544,-.0036307023836027,-.00469566620937377,-.00479344565888028,-.00409076076250558,-.00295397526843083,-.00178527858577817,-.000873785000712069,-.000324004543650437,-7.93636419908757e-005,-9.20627996502583e-006,-1.50201013794113e-007,-.00467788049829846,-.0097913851734879,-.0126634109999641,-.0129271055006463,-.0110320841661877,-.00796636754827121,-.00481459189667545,-.00235644913761801,-.000873785000712069,-.000214030208318087,-2.48277670898173e-005,-4.05066519951766e-007,-.00955763703157626,-.0200053219740424,-.025873317192124,-.0264120860481604,-.0225402628819985,-.016276527267894,-.0098369596701605,-.00481459189667545,-.00178527858577817,-.000437296986453802,-5.07269876251226e-005,-8.27613867594725e-007,-.0158143516876436,-.0331014032314506,-.0428107633770146,-.0437022264174669,-.037295792168106,-.0269316281438179,-.016276527267894,-.00796636754827121,-.00295397526843083,-.000723564654409835,-8.39343887728836e-005,-1.36939462342768e-006,-.0219002271480226,-.0458398968231796,-.0592857336712341,-.0605202606038321,-.0516484226656698,-.037295792168106,-.0225402628819985,-.0110320841661877,-.00409076076250558,-.00100201580190209,-.00011623506394466,-1.89638209018582e-006,-.0256621090417623,-.0537139830919874,-.0694694603763257,-.0709160465028163,-.0605202606038321,-.0437022264174669,-.0264120860481604,-.0129271055006463,-.00479344565888028,-.00117413571083906,-.000136201184821656,-2.22213055847631e-006,-.0251386386461752,-.0526182945056766,-.0680523825420277,-.0694694603763257,-.0592857336712341,-.0428107633770146,-.025873317192124,-.0126634109999641,-.00469566620937377,-.001150185018243,-.000133422875058339,-2.17680226684608e-006,-.0194372664460252,-.0406846140173892,-.0526182945056766,-.0537139830919874,-.0458398968231796,-.0331014032314506,-.0200053219740424,-.0097913851734879,-.0036307023836027,-.000889326306666056,-.000103162944064127,-1.68310966462922e-006,-.00928624582089677,-.0194372664460252,-.0251386386461752,-.0256621090417623,-.0219002271480226,-.0158143516876436,-.00955763703157627,-.00467788049829846,-.00173458520673544,-.000424879842112779,-4.92865836277514e-005,-8.04113589360719e-007},
			{-8.77197535304017e-009,-1.83608344516711e-008,-2.37464657802842e-008,-2.42409464882796e-008,-2.06873968742175e-008,-1.49385560003323e-008,-9.02833697182859e-009,-4.41882039598848e-009,-1.63852421901125e-009,-4.01350079989265e-010,-4.65571023163414e-011,-7.59581936873143e-013,-2.79578126521843e-006,-5.85191760211322e-006,-7.56841207045341e-006,-7.72601168100707e-006,-6.59343355166069e-006,-4.76117787775939e-006,-2.8774881763854e-006,-1.40835499192186e-006,-5.22226195326764e-007,-1.2791725800267e-007,-1.48385590679724e-008,-2.42091987612831e-010,-5.7837794064797e-005,-.000121061690113565,-.000156571711876826,-.000159832057717645,-.000136401819657762,-9.84969850916432e-005,-5.95281078102025e-005,-2.91353787244659e-005,-1.08035673306481e-005,-2.64629143836613e-006,-3.06973057681128e-007,-5.00828398075496e-009,-.000423492023911399,-.000886421430715201,-.00114642807911561,-.00117030053969446,-.000998742839454559,-.000721201218685817,-.000435868609164376,-.00021333110473795,-7.91044103306138e-005,-1.93763150067161e-005,-2.24767634356886e-006,-3.66709753307118e-008,-.00174950826221703,-.00366193819311915,-.00473606415990016,-.00483468482962708,-.00412595456537195,-.00297938903112802,-.00180063776864345,-.000881302384115301,-.000326792032995098,-8.00464265713343e-005,-9.28548381512419e-006,-1.51493229393287e-007,-.00494625775323145,-.0103531321176025,-.0133899305174026,-.0136687535803171,-.011665011934605,-.00842341040243774,-.00509081249643689,-.00249164227715485,-.000923915400575431,-.000226309453117529,-2.62521745709628e-005,-4.28305814057175e-007,-.010527567456728,-.0220355068809891,-.0284989994042999,-.0290924437312573,-.0248276992732258,-.0179283057316278,-.0108352363826531,-.00530318747211388,-.00196645265756991,-.000481674864645602,-5.58748760112742e-005,-9.11601977606186e-007,-.0177221024244434,-.0370945625876605,-.0479752030574144,-.0489742069762943,-.0417949380321651,-.0301805019800266,-.0182400321589689,-.00892738345711226,-.00331032553849853,-.000810851255203673,-9.40597416919936e-005,-1.53459036799021e-006,-.0240985158662193,-.0504411882777659,-.0652366837960255,-.0665951293806949,-.0568327590696347,-.0410394482774791,-.0248027967481441,-.0121394565234153,-.00450138085207253,-.00110259558209743,-.00012790244200432,-2.08673606807239e-006,-.02632238051474,-.0550960133326791,-.0712568700882474,-.0727406760531766,-.0620774124945382,-.044826659852073,-.027091654003041,-.0132597125742255,-.00491677828990747,-.00120434555501086,-.000139705563856799,-2.27930471413928e-006,-.0218959049522221,-.0458308499302774,-.0592740331282503,-.0605083164164748,-.0516382294108676,-.0372884315268124,-.0225358143697108,-.011029906890695,-.00408995341613109,-.00100181804557465,-.000116212123967743,-1.89600782307658e-006,-.0108863317310921,-.0227864450885946,-.0294702040898467,-.0300838721412506,-.025673791359598,-.0185392764637387,-.0112044855690737,-.00548391243189871,-.00203346652033204,-.000498089647453146,-5.77790109816517e-005,-9.42668054679449e-007},
			{-.000206393222578878,-.00224638876284317,-.00697208388856876,-.0127648935683974,-.0166874796528932,-.0166745014700866,-.0129629077002378,-.00771242584002074,-.00331870648655728,-.000906410211953001,-.000113117990564565,-1.92058726559064e-006,-.000467862027215704,-.00509222147590975,-.0158046531822386,-.0289360711920697,-.0378279769169244,-.0377985573514871,-.0293849390956891,-.0174828957229149,-.00752300257311598,-.00205469401540682,-.000256421270615505,-4.35367906124192e-006,-.000700346515907722,-.0076225882021403,-.0236581153148544,-.0433146429173246,-.0566250097176437,-.0565809713282148,-.0439865569734266,-.0261702903746801,-.0112612444155839,-.00307568836802217,-.000383839108612685,-6.51705798837019e-006,-.000888862280208502,-.00967439828220347,-.0300262881966959,-.0549738613606283,-.0718670459711192,-.071811153548789,-.0558266378740331,-.0332146779455411,-.0142924897345245,-.00390358674443833,-.000487158995675827,-8.27128698753606e-006,-.00102153536475457,-.0111184153023947,-.0345080626641537,-.063179352715725,-.0825940425807241,-.0825298075610779,-.0641594160912017,-.0381723568496403,-.0164258108812803,-.00448624269205744,-.000559873169805834,-9.50587324711357e-006,-.00109002421181751,-.0118638495492168,-.0368216560137206,-.067415212946293,-.0881315608554635,-.0880630191983885,-.0684609846789651,-.0407316228334812,-.0175270795090219,-.00478702287082038,-.000597409871152132,-1.01431945983696e-005,-.00109002421181751,-.0118638495492168,-.0368216560137206,-.067415212946293,-.0881315608554635,-.0880630191983885,-.0684609846789651,-.0407316228334812,-.0175270795090219,-.00478702287082038,-.000597409871152132,-1.01431945983696e-005,-.00102153536475457,-.0111184153023947,-.0345080626641537,-.063179352715725,-.0825940425807241,-.0825298075610779,-.0641594160912017,-.0381723568496402,-.0164258108812803,-.00448624269205744,-.000559873169805834,-9.50587324711357e-006,-.000888862280208502,-.00967439828220347,-.0300262881966959,-.0549738613606283,-.0718670459711192,-.071811153548789,-.0558266378740331,-.0332146779455411,-.0142924897345245,-.00390358674443833,-.000487158995675827,-8.27128698753606e-006,-.000700346515907722,-.0076225882021403,-.0236581153148544,-.0433146429173246,-.0566250097176437,-.0565809713282148,-.0439865569734266,-.0261702903746801,-.0112612444155839,-.00307568836802217,-.000383839108612685,-6.51705798837019e-006,-.000467862027215704,-.00509222147590975,-.0158046531822386,-.0289360711920697,-.0378279769169244,-.0377985573514871,-.0293849390956891,-.0174828957229149,-.00752300257311598,-.00205469401540682,-.000256421270615505,-4.35367906124192e-006,-.000206393222578878,-.00224638876284317,-.00697208388856876,-.0127648935683974,-.0166874796528932,-.0166745014700866,-.0129629077002378,-.00771242584002074,-.00331870648655728,-.000906410211953001,-.000113117990564565,-1.92058726559064e-006},
			{-3.29588494374136e-006,-3.58724904274172e-005,-.000111336922926592,-.000203842064166924,-.000266481681179921,-.000266274433108618,-.000207004143752805,-.000123159413319067,-5.29962883716334e-005,-1.44744276633774e-005,-1.80637657239712e-006,-3.06697796211909e-008,-3.88498085043389e-005,-.000422842243423972,-.00131236927532137,-.00240276141102928,-.0031411176240886,-.00313867471481826,-.00244003400654881,-.00145172531948917,-.000624686750242411,-.000170615404521307,-2.1292425288575e-005,-3.61515976889952e-007,-.000139558140870839,-.00151895413763271,-.00471435570086283,-.00863131450035304,-.0112836730154145,-.0112748974798531,-.00876520690127838,-.00521495715018533,-.00224402963223718,-.000612892819182374,-7.6487668853577e-005,-1.29865498884588e-006,-.000317673277053615,-.00345756353291146,-.0107311892759846,-.0196472806638003,-.0256847888746645,-.025664813306616,-.0199520571355211,-.0118706978844523,-.00510803771553511,-.0013951151049833,-.00017410728078841,-2.95610118832487e-006,-.000559264033051857,-.00608704308994404,-.0188922664493339,-.034589051759252,-.0452180893128947,-.0451829223109582,-.0351256109572267,-.0208983721752537,-.00899270407717994,-.00245610114713186,-.000306515993225909,-5.20421827112933e-006,-.000825769740663992,-.00898769757526597,-.0278949852742899,-.051071748964773,-.0667658345226861,-.0667139094133409,-.0518639943508112,-.0308570591913224,-.0132779912077635,-.00362650534890298,-.000452579850083181,-7.68418084863826e-006,-.0010622075751108,-.0115610926110904,-.0358819937409303,-.0656948250258218,-.0858825065829686,-.0858157140598609,-.0667139094133409,-.0396921810095751,-.0170798010009427,-.0046648614778342,-.000582164399381254,-9.88434754146089e-006,-.00121008711989414,-.0131706171076067,-.0408774513376,-.0748407971004666,-.0978389887959829,-.0977628974803007,-.0760017574629395,-.0452180893128947,-.0194576349160754,-.00531429912823647,-.000663212782377944,-1.12604371581811e-005,-.0012218813527315,-.0132989858194334,-.0412758674276049,-.0755702402721397,-.0987925860994006,-.0987157531522446,-.0767425160486888,-.0456588118567917,-.0196472806638003,-.00536609546608301,-.000669676850830108,-1.13701881137203e-005,-.0010734776072168,-.01168375590964,-.0362627020233444,-.0663918477213734,-.0867937207648416,-.0867262195724516,-.067421744603583,-.040113315413826,-.017261017846082,-.00471435570086283,-.000588341168993646,-9.98922055945599e-006,-.000771510993319342,-.00839714407356662,-.0260620930240078,-.0477159840498455,-.06237886032367,-.0623303470508818,-.0484561734690455,-.0288295383268322,-.0124055359279088,-.00338821902309026,-.000422842243423972,-7.17927735473982e-006,-.000354187662792186,-.00385498697918716,-.0119646666030396,-.0219056021427923,-.0286370809191255,-.0286148092951926,-.0222454106000665,-.0132351539871473,-.00569517196000836,-.00155547411146173,-.000194119730276028,-3.29588494374136e-006},
			{-3.92294663599167e-008,-4.26974448589619e-007,-1.32519433994785e-006,-2.4262422795302e-006,-3.1718140426088e-006,-3.16934726012697e-006,-2.46387911966847e-006,-1.46591223424896e-006,-6.30791470989705e-007,-1.72282735226472e-007,-2.15005044744533e-008,-3.65048874112027e-010,-2.40449409138482e-006,-2.61705711055768e-005,-8.12252180824206e-005,-.000148711817078381,-.000194410192951716,-.000194258996301528,-.00015101869423293,-8.98502486218331e-005,-3.86631403796111e-005,-1.05597362732147e-005,-1.31783174148499e-006,-2.23749630651595e-008,-2.07281778267435e-005,-.000225605982417451,-.000700209981991998,-.00128198484678476,-.00167593219100564,-.00167462878541062,-.00130187150820567,-.000774562906136249,-.000333299404644144,-9.10312452249138e-005,-1.1360498152647e-005,-1.92885569959679e-007,-8.46234324557744e-005,-.000921043459502259,-.00285862908989082,-.00523374312001741,-.00684204543937499,-.00683672424538222,-.00531493103550084,-.00316217722165629,-.00136070521452511,-.000371637897746933,-4.63795880236092e-005,-7.87461355147049e-007,-.000228214966233319,-.00248389714184165,-.00770923516443232,-.0141145126680233,-.0184518297544813,-.0184374794017228,-.0143334626308517,-.00852785271078683,-.00366958991823735,-.00100224403364473,-.00012507783964276,-2.12364898657175e-006,-.000466278651886231,-.00507498798101565,-.0157511658366454,-.0288381434728016,-.0376999565136218,-.0376706365121829,-.0292854922824927,-.0174237287374253,-.00749754263838111,-.00204774035893469,-.000255553469660892,-4.3389449994523e-006,-.000771518583621402,-.00839722668659549,-.0260623494288551,-.0477164534906379,-.0623794740212014,-.0623309602711284,-.0484566501919911,-.0288298219584409,-.0124056579764159,-.00338825235716593,-.000422846403442921,-7.17934798611665e-006,-.00106842395843715,-.0116287518756503,-.0360919868089839,-.0660792924543047,-.0863851188731245,-.0863179354585031,-.0671043408542665,-.0399244725296093,-.0171797575373556,-.00469216176055673,-.000585571414309697,-9.94219394991092e-006,-.00125195103863209,-.0136262649987847,-.0422916390211788,-.077429973529657,-.101223805813716,-.101145082058928,-.0786310982319312,-.0467824447922306,-.0201307871491398,-.00549815149985173,-.000686157151895624,-1.16500008667718e-005,-.00122641302441892,-.0133483086422873,-.0414289499502059,-.0758505125894862,-.0991589846571268,-.0990818667545695,-.0770271360622618,-.0458281496934892,-.0197201478250184,-.00538599705705277,-.000672160525384795,-1.14123574777427e-005,-.000948266016470748,-.0103209499661349,-.0320329811846719,-.0586478306968206,-.0766700071720531,-.0766103793918513,-.0595575992912496,-.0354344548588437,-.015247674029803,-.00416446814598573,-.000519716417833643,-8.82406705447926e-006,-.00045303856676568,-.00493088257841441,-.0153039080100614,-.028019278030959,-.0366294579367253,-.0366009704831913,-.0284539242725722,-.0169289781185245,-.00728464826218446,-.00198959431997835,-.000248296972548114,-4.21573969958762e-006},
			{-5.21125166412433e-006,-.000294934791565546,-.00219672054164462,-.00721329026542284,-.0144452225630722,-.0199731094854017,-.0199731094854017,-.0144452225630722,-.00721329026542284,-.00219672054164462,-.000294934791565546,-5.21125166412433e-006,-1.18131144881787e-005,-.000668572290088461,-.00497963117683019,-.0163514313324231,-.032745121315009,-.0452759997488288,-.0452759997488288,-.032745121315009,-.0163514313324231,-.00497963117683019,-.000668572290088461,-1.18131144881787e-005,-1.76831482201068e-005,-.00100079135890211,-.0074540508575846,-.0244765920241849,-.049016441369304,-.0677740163420273,-.0677740163420273,-.049016441369304,-.0244765920241849,-.0074540508575846,-.00100079135890211,-1.76831482201068e-005,-2.24430094120152e-005,-.00127017935990404,-.00946049490012139,-.0310650783636037,-.0622104413366804,-.0860170863084584,-.0860170863084584,-.0622104413366804,-.0310650783636037,-.00946049490012139,-.00127017935990404,-2.24430094120152e-005,-2.57928908857684e-005,-.00145976847551561,-.0108725843403854,-.0357019044050888,-.0714960768359992,-.0988561418273082,-.0988561418273082,-.0714960768359992,-.0357019044050888,-.0108725843403854,-.00145976847551561,-2.57928908857684e-005,-2.75221754706548e-005,-.00155763866515012,-.0116015368483061,-.0380955388841488,-.0762895319046813,-.105483952681873,-.105483952681873,-.0762895319046813,-.0380955388841488,-.0116015368483061,-.00155763866515012,-2.75221754706548e-005,-2.75221754706548e-005,-.00155763866515012,-.0116015368483061,-.0380955388841488,-.0762895319046813,-.105483952681873,-.105483952681873,-.0762895319046813,-.0380955388841488,-.0116015368483061,-.00155763866515012,-2.75221754706548e-005,-2.57928908857684e-005,-.00145976847551561,-.0108725843403854,-.0357019044050888,-.0714960768359992,-.0988561418273082,-.0988561418273082,-.0714960768359992,-.0357019044050888,-.0108725843403854,-.00145976847551561,-2.57928908857684e-005,-2.24430094120152e-005,-.00127017935990404,-.00946049490012139,-.0310650783636037,-.0622104413366804,-.0860170863084584,-.0860170863084584,-.0622104413366804,-.0310650783636037,-.00946049490012139,-.00127017935990404,-2.24430094120152e-005,-1.76831482201068e-005,-.00100079135890211,-.0074540508575846,-.0244765920241849,-.049016441369304,-.0677740163420273,-.0677740163420273,-.049016441369304,-.0244765920241849,-.0074540508575846,-.00100079135890211,-1.76831482201068e-005,-1.18131144881787e-005,-.000668572290088461,-.00497963117683019,-.0163514313324231,-.032745121315009,-.0452759997488288,-.0452759997488288,-.032745121315009,-.0163514313324231,-.00497963117683019,-.000668572290088461,-1.18131144881787e-005,-5.21125166412433e-006,-.000294934791565546,-.00219672054164462,-.00721329026542284,-.0144452225630722,-.0199731094854017,-.0199731094854017,-.0144452225630722,-.00721329026542284,-.00219672054164462,-.000294934791565546,-5.21125166412433e-006},
			{-8.32182650623191e-008,-4.7098016450767e-006,-3.50793406430393e-005,-.000115188737748337,-.000230675169348194,-.000318949769921382,-.000318949769921382,-.000230675169348194,-.000115188737748337,-3.50793406430393e-005,-4.7098016450767e-006,-8.32182650623191e-008,-9.80924309227986e-007,-5.55161648928631e-005,-.000413493095087696,-.00135777203384397,-.00271905309464788,-.00375957829094687,-.00375957829094687,-.00271905309464788,-.00135777203384397,-.000413493095087696,-5.55161648928631e-005,-9.80924309227986e-007,-3.52372323574208e-006,-.000199427823688284,-.00148536967967095,-.00487745366231389,-.00976751261916201,-.0135053370130334,-.0135053370130334,-.00976751261916201,-.00487745366231389,-.00148536967967095,-.000199427823688284,-3.52372323574208e-006,-8.02097749900636e-006,-.000453953383811275,-.00338111593370866,-.011102445754264,-.0222335846768227,-.0307419161639197,-.0307419161639197,-.0222335846768227,-.011102445754264,-.00338111593370866,-.000453953383811275,-8.02097749900635e-006,-1.41209366639782e-005,-.000799185259152228,-.00595245703648738,-.0195458637467361,-.039142241836921,-.0541211655479415,-.0541211655479415,-.039142241836921,-.0195458637467361,-.00595245703648738,-.000799185259152228,-1.41209366639782e-005,-2.08499769658255e-005,-.00118002046473714,-.00878897732169731,-.0288600408453219,-.0577946675996675,-.0799114875939334,-.0799114875939334,-.0577946675996675,-.0288600408453219,-.00878897732169731,-.00118002046473714,-2.08499769658255e-005,-2.68198292857976e-005,-.0015178888432285,-.0113054739461356,-.0371233680459864,-.0743426777493866,-.102792077840484,-.102792077840484,-.0743426777493866,-.0371233680459864,-.0113054739461356,-.0015178888432285,-2.68198292857976e-005,-3.05536608257741e-005,-.00172920790781427,-.0128794114512791,-.0422916486119502,-.0846925957890945,-.117102694743021,-.117102694743021,-.0846925957890945,-.0422916486119502,-.0128794114512791,-.00172920790781427,-3.0553660825774e-005,-3.08514550786741e-005,-.00174606180234275,-.0130049418986058,-.0427038483144384,-.0855180604832598,-.11824404764644,-.11824404764644,-.0855180604832598,-.0427038483144384,-.0130049418986058,-.00174606180234275,-3.08514550786741e-005,-2.7104387920297e-005,-.00153399365776509,-.0114254251283075,-.037517247321153,-.0751314542415842,-.103882702728358,-.103882702728358,-.0751314542415842,-.037517247321153,-.0114254251283075,-.00153399365776509,-2.7104387920297e-005,-1.94799901806221e-005,-.00110248501011246,-.00821148110642986,-.0269637378113508,-.0539971607248856,-.0746607536401705,-.0746607536401705,-.0539971607248856,-.0269637378113508,-.00821148110642986,-.00110248501011246,-1.94799901806221e-005,-8.94293438853622e-006,-.00050613224228358,-.00376975224764522,-.0123785964921314,-.0247891842373852,-.0342754906451081,-.0342754906451081,-.0247891842373852,-.0123785964921314,-.00376975224764522,-.00050613224228358,-8.94293438853622e-006},
			{-1.00782323868636e-007,-2.96594300874774e-005,-.000530130419242249,-.00312208933690556,-.00957752805114946,-.0183245877738932,-.0235713601331056,-.0207230264243475,-.0120086359724961,-.00407774946276834,-.000589001007959142,-1.08304313867452e-005,-2.28458191424719e-007,-6.72334145152776e-005,-.00120172498658768,-.00707730141552942,-.0217107986092594,-.0415390518964802,-.0534326863949788,-.0469759473290577,-.0272217503074272,-.00924363749104905,-.00133517565243957,-2.45509398075641e-005,-3.41980945423285e-007,-.000100642251068274,-.00179887201456491,-.0105940706876594,-.0324990729725514,-.0621801483718135,-.0799838276575082,-.0703186818540462,-.0407484618877329,-.0138368769735303,-.00199863541368309,-3.67505036876191e-005,-4.34033661954923e-007,-.00012773262768919,-.00228308336566367,-.0134457295270728,-.0412470105167922,-.0789174890001959,-.101513473425809,-.0892467121265429,-.0517169285858085,-.0175614181527209,-.00253661807513852,-4.66428200392335e-005,-4.98818259086074e-007,-.00014679821533987,-.00262386024318778,-.0154526618157401,-.0474036089408846,-.0906968466390758,-.116665545847229,-.102567827047345,-.0594362846564907,-.020182665073873,-.00291523797142323,-5.3604806102002e-005,-5.32261533433169e-007,-.000156640302913549,-.00279977697510962,-.0164886856562537,-.0505817842982924,-.0967776174795907,-.124487388343073,-.109444487868491,-.0634211908577612,-.0215358120223382,-.00311069012556852,-5.7198740774876e-005,-5.32261533433169e-007,-.000156640302913549,-.00279977697510962,-.0164886856562537,-.0505817842982924,-.0967776174795907,-.124487388343073,-.109444487868491,-.0634211908577612,-.0215358120223382,-.00311069012556852,-5.7198740774876e-005,-4.98818259086074e-007,-.00014679821533987,-.00262386024318778,-.0154526618157401,-.0474036089408846,-.0906968466390758,-.116665545847229,-.102567827047345,-.0594362846564907,-.020182665073873,-.00291523797142323,-5.3604806102002e-005,-4.34033661954923e-007,-.00012773262768919,-.00228308336566367,-.0134457295270728,-.0412470105167922,-.0789174890001959,-.101513473425809,-.0892467121265429,-.0517169285858086,-.0175614181527209,-.00253661807513852,-4.66428200392335e-005,-3.41980945423285e-007,-.000100642251068274,-.00179887201456491,-.0105940706876594,-.0324990729725514,-.0621801483718135,-.0799838276575082,-.0703186818540462,-.0407484618877329,-.0138368769735303,-.00199863541368309,-3.67505036876191e-005,-2.28458191424719e-007,-6.72334145152776e-005,-.00120172498658768,-.00707730141552942,-.0217107986092594,-.0415390518964802,-.0534326863949788,-.0469759473290577,-.0272217503074272,-.00924363749104905,-.00133517565243957,-2.45509398075641e-005,-1.00782323868636e-007,-2.96594300874774e-005,-.000530130419242249,-.00312208933690556,-.00957752805114946,-.0183245877738932,-.0235713601331056,-.0207230264243475,-.0120086359724961,-.00407774946276834,-.000589001007959142,-1.08304313867452e-005},
			{-.00572678909872914,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442027,-.00553806002210424,-.00260199447163722,-.000920334332232957,-.000222779313164588,-3.04257537549422e-005,-1.47072954118751e-006,-4.61452526590166e-009,-.0129817593993148,-.0261104821110847,-.0313889764798934,-.0287370569429126,-.0211332958986598,-.0125539393028941,-.00589832550261673,-.00208625787714596,-.000505006800981909,-6.89705536523267e-005,-3.33392005816974e-006,-1.0460426551641e-008,-.019432502397701,-.0390849953863391,-.0469864170136373,-.0430167368512079,-.0316346044160784,-.018792094977277,-.00882925194855102,-.0031229365722185,-.000755948833210871,-.000103242588927534,-4.99057235089002e-006,-1.56582985243485e-008,-.0246632459775863,-.0496056985105609,-.0596339852015086,-.0545957664330435,-.0401498486478861,-.023850456898108,-.0112058657268432,-.00396355298338403,-.000959431350547771,-.000131032911457859,-6.33390959846238e-006,-1.98731208257652e-008,-.028344523709382,-.0570099288159739,-.0685350545084464,-.0627448227009384,-.0461426828391663,-.0274104163597275,-.0128784721632876,-.00455515959306486,-.00110263769366935,-.000150591105035483,-7.2793196382121e-006,-2.2839416390577e-008,-.0302448825382416,-.0608321599696494,-.073129987810444,-.0669515498630354,-.0492363193955517,-.0292481479535269,-.0137419094370501,-.00486056030603151,-.00117656404705184,-.000160687486930486,-7.76736168416077e-006,-2.43706852532627e-008,-.0302448825382416,-.0608321599696494,-.073129987810444,-.0669515498630354,-.0492363193955517,-.0292481479535269,-.0137419094370501,-.00486056030603151,-.00117656404705184,-.000160687486930486,-7.76736168416077e-006,-2.43706852532627e-008,-.028344523709382,-.0570099288159739,-.0685350545084464,-.0627448227009384,-.0461426828391663,-.0274104163597275,-.0128784721632876,-.00455515959306486,-.00110263769366935,-.000150591105035483,-7.2793196382121e-006,-2.2839416390577e-008,-.0246632459775863,-.0496056985105609,-.0596339852015086,-.0545957664330435,-.0401498486478861,-.023850456898108,-.0112058657268432,-.00396355298338403,-.000959431350547771,-.000131032911457859,-6.33390959846238e-006,-1.98731208257652e-008,-.019432502397701,-.0390849953863391,-.0469864170136373,-.0430167368512079,-.0316346044160784,-.018792094977277,-.00882925194855102,-.0031229365722185,-.000755948833210871,-.000103242588927534,-4.99057235089002e-006,-1.56582985243485e-008,-.0129817593993148,-.0261104821110847,-.0313889764798934,-.0287370569429126,-.0211332958986598,-.0125539393028941,-.00589832550261673,-.00208625787714596,-.000505006800981909,-6.89705536523267e-005,-3.33392005816974e-006,-1.0460426551641e-008,-.00572678909872914,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442027,-.00553806002210424,-.00260199447163722,-.000920334332232957,-.000222779313164588,-3.04257537549422e-005,-1.47072954118751e-006,-4.61452526590166e-009},
			{-9.14508612930534e-005,-.0001839370153449,-.000221121717472685,-.000202440095184293,-.000148874898420579,-8.84370543742305e-005,-4.15511434782526e-005,-1.46967813742169e-005,-3.55755375585554e-006,-4.85867619430457e-007,-2.34860199933996e-008,-7.36891655603157e-011,-.00107796495006253,-.00216813327679371,-.00260644304233894,-.00238623588679712,-.00175484320401577,-.00104244009902437,-.000489778616311158,-.000173236369522755,-4.19341950699251e-005,-5.72710039808159e-006,-2.76838358998326e-007,-8.68601307305661e-010,-.00387231726863855,-.00778847209081444,-.00936298940145173,-.00857195072154575,-.00630383171759815,-.00374470319905967,-.00175940617887584,-.000622307975060138,-.000150638021863614,-2.05731640619964e-005,-9.94471998471375e-007,-3.12023117416446e-009,-.00881447480486433,-.0177287361159355,-.0213127769373929,-.0195121521357594,-.0143492802872553,-.00852398956746932,-.00400489948507659,-.0014165466273796,-.000342894178409679,-4.68302630441653e-005,-2.26369580965429e-006,-7.10251695871362e-009,-.0155178892424445,-.0312114527009358,-.0375211591484582,-.0343511578883119,-.0252619182805468,-.0150064897727934,-.00705062842794049,-.00249383135774626,-.000603665448054148,-8.24447118179819e-005,-3.98523810328615e-006,-1.25039862212867e-008,-.0229126184021882,-.0460846249345851,-.055401091485105,-.0507204916900297,-.0372999629412389,-.0221575220926239,-.01041045957611,-.00368221511100628,-.00089132973162475,-.000121732034018965,-5.88432089415219e-006,-1.84625022332888e-008,-.0294730548165372,-.0592797668540085,-.0712637629438524,-.0652429943039245,-.0479798438190263,-.0285017561839274,-.0133912257589351,-.00473652229125437,-.00114653897597652,-.000156586857450866,-7.56914418191853e-006,-2.37487628354249e-008,-.03357626593249,-.0675326405320358,-.0811850373450308,-.074326062928209,-.0546595527845674,-.0324697440130155,-.0152555396799836,-.00539593649305364,-.00130615905948616,-.000178386733239242,-8.62291335307587e-006,-2.70550433775344e-008,-.0339035203026146,-.0681908540387014,-.0819763152766897,-.0750504891927709,-.055192297478506,-.0327862135586896,-.0154042292942165,-.00544852851748593,-.0013188896669085,-.000180125397036526,-8.7069574240775e-006,-2.7318738012221e-008,-.0297857641982319,-.0599087257235696,-.0720198721217481,-.0659352230713762,-.0484889104016337,-.0288041600782291,-.0135333067869547,-.00478677683617009,-.00115870376501852,-.000158248245443382,-7.64945287105236e-006,-2.40007374267426e-008,-.0214071018984119,-.0430565483441448,-.0517608590117147,-.0473878067921877,-.0348490990193348,-.0207016206060418,-.00972642083254007,-.0034402682709391,-.000832763242290672,-.000113733402739168,-5.49768056941236e-006,-1.72493889467441e-008,-.00982763882071387,-.0197665339288291,-.0237625359019177,-.0217549415081703,-.0159986326040109,-.00950376427809313,-.00446523548185483,-.00157936904180661,-.000382307535472911,-5.22130837362085e-005,-2.52389226922161e-006,-7.91890304680574e-009},
			{-1.08849930992072e-006,-2.18932125341295e-006,-2.63191437974773e-006,-2.40955526052686e-006,-1.77199232357178e-006,-1.05262729400979e-006,-4.94564953932593e-007,-1.74929313488118e-007,-4.23439949444061e-008,-5.78306820717772e-009,-2.79543748348947e-010,-8.77089671184216e-013,-6.67174550672731e-005,-.000134190202070998,-.000161318089751033,-.000147689018597559,-.000108610834338575,-6.45187493924991e-005,-3.03134001015112e-005,-1.07219531571797e-005,-2.5953930832331e-006,-3.5446195486468e-007,-1.71340921393568e-008,-5.37595110937315e-011,-.000575144383900619,-.0011567998362318,-.00139065846034338,-.00127316771187585,-.000936290380644839,-.000556190225361124,-.000261319647275767,-9.24296518291858e-005,-2.23738413632665e-005,-3.05567414766182e-006,-1.47706126638825e-007,-4.63439153301422e-010,-.00234804488509053,-.00472267141011844,-.00567740653672345,-.00519774689175974,-.00382243468035404,-.00227066394170372,-.00106684561013735,-.000377346936322897,-9.13419051691941e-005,-1.24748850093275e-005,-6.03014868716438e-007,-1.89200479726517e-009,-.00633227663561512,-.0127362394210907,-.0153109972435626,-.0140174369789623,-.0103084544811151,-.00612359342731638,-.00287710067800549,-.00101764033710382,-.000246333541419145,-3.36426374888076e-005,-1.62622826690742e-006,-5.1024142886996e-009,-.012937825514943,-.0260221169776175,-.0312827474533958,-.0286398027496958,-.0210617749476202,-.0125114532807165,-.00587836392863835,-.00207919740024719,-.000503297717511878,-6.87371381162648e-005,-3.32263714544313e-006,-1.04250255588739e-008,-.0214073125073398,-.0430569719462525,-.0517613682492133,-.0473882730064016,-.0348494418743096,-.0207018242742517,-.00972651652373047,-.00344030211724,-.000832771435243388,-.000113734521679514,-5.49773465709859e-006,-1.7249558650954e-008,-.0296455406961615,-.0596266912835726,-.071680821606866,-.0656248174753565,-.0482606374326126,-.0286685577088058,-.0134695955569983,-.00476424195651651,-.00115324889407038,-.000157503254546897,-7.61344126955836e-006,-2.388774830107e-008,-.0347378633474854,-.0698689855213075,-.0839936977751664,-.076897431725972,-.0565505431450506,-.0335930604290494,-.0157833171133724,-.00558261317396905,-.00135134666284035,-.000184558163040781,-8.92122984487403e-006,-2.79910339523387e-008,-.0340292605183584,-.068443757944315,-.0822803462291326,-.0753288338784716,-.0553969927824272,-.0329078099454216,-.0154613599726102,-.00546873583357464,-.00132378111976588,-.000180793439944228,-8.73924948977462e-006,-2.74200568127716e-008,-.0263115204035591,-.0529209070727989,-.0636193962384749,-.0582444672431377,-.0428331113779068,-.0254444116512108,-.0119547672264766,-.00422844200182213,-.00102355130296691,-.000139789998708784,-6.75721240365912e-006,-2.12012654199985e-008,-.0125704531070497,-.0252832132289538,-.0303944669425635,-.0278265692364139,-.0204637212045026,-.0121561878064161,-.00571144648888444,-.00202015813166747,-.000489006467861309,-6.67853319250079e-005,-3.22829013115797e-006,-1.0129004659729e-008},
			{-.000354187662792186,-.00370432632910279,-.0106866754670937,-.0175473588745023,-.0197676016789817,-.0162489801425508,-.00982031115806883,-.00422235247914266,-.00118608291592308,-.000180620480252622,-9.39296633663941e-006,-3.06697796211909e-008,-.000802889532215207,-.00839714407356662,-.024225067014107,-.0397771922581837,-.0448101448253052,-.0368339652566651,-.0222611509665594,-.00957143052386794,-.00268866947546018,-.000409439142389331,-2.1292425288575e-005,-6.95237231559063e-008,-.00120185194317225,-.0125697540159339,-.0362627020233443,-.0595428061909519,-.0670766742761412,-.0551370654893481,-.0333228999419048,-.0143275530598965,-.00402469144754472,-.000612892819182374,-3.18728065084108e-005,-1.04070632906319e-007,-.00152536042432723,-.0159532173895116,-.0460237143683289,-.0755702402721397,-.085132037201067,-.0699785843745455,-.0422925910998948,-.0181841719682459,-.00510803771553511,-.000777868235805777,-4.04521687853206e-005,-1.32083844163853e-007,-.00175303829641842,-.0183344215497376,-.0528932916734667,-.0868499819149511,-.0978389887959829,-.0804237060180922,-.0486052546469995,-.0208983721752537,-.00587047204848792,-.000893974161901044,-4.64901277906996e-005,-1.51798901731387e-007,-.00187057076364506,-.0195636529956887,-.0564395228555362,-.0926728396779039,-.104398604617067,-.0858157140598609,-.0518639943508112,-.0222995037123093,-.00626405789601532,-.000953910667053138,-4.96070587967628e-005,-1.61976260365957e-007,-.00187057076364506,-.0195636529956887,-.0564395228555362,-.0926728396779039,-.104398604617067,-.0858157140598609,-.0518639943508112,-.0222995037123093,-.00626405789601532,-.000953910667053138,-4.96070587967628e-005,-1.61976260365957e-007,-.00175303829641842,-.0183344215497376,-.0528932916734667,-.0868499819149511,-.0978389887959829,-.0804237060180922,-.0486052546469995,-.0208983721752537,-.00587047204848792,-.000893974161901044,-4.64901277906996e-005,-1.51798901731387e-007,-.00152536042432723,-.0159532173895116,-.0460237143683289,-.0755702402721397,-.085132037201067,-.0699785843745455,-.0422925910998948,-.0181841719682459,-.00510803771553511,-.000777868235805777,-4.04521687853206e-005,-1.32083844163853e-007,-.00120185194317225,-.0125697540159339,-.0362627020233444,-.0595428061909519,-.0670766742761412,-.0551370654893481,-.0333228999419048,-.0143275530598965,-.00402469144754472,-.000612892819182374,-3.18728065084108e-005,-1.04070632906319e-007,-.000802889532215207,-.00839714407356662,-.024225067014107,-.0397771922581837,-.0448101448253052,-.0368339652566651,-.0222611509665594,-.00957143052386794,-.00268866947546018,-.000409439142389331,-2.1292425288575e-005,-6.95237231559063e-008,-.000354187662792186,-.00370432632910279,-.0106866754670937,-.0175473588745023,-.0197676016789817,-.0162489801425508,-.00982031115806883,-.00422235247914266,-.00118608291592308,-.000180620480252622,-9.39296633663941e-006,-3.06697796211909e-008},
			{-5.6560083280767e-006,-5.91542359272188e-005,-.000170655084270396,-.00028021305752868,-.000315668024236135,-.00025947930056204,-.000156820147987287,-6.74265743697526e-005,-1.89405096647495e-005,-2.8843210756594e-006,-1.49995895979437e-007,-4.89764458734161e-010,-6.66694512082365e-005,-.000697272743806459,-.00201157426832706,-.00330297440936967,-.00372089514708473,-.00305857798732252,-.00184849678399644,-.000794781840714042,-.000223258756300761,-3.3998553762376e-005,-1.76805681469661e-006,-5.77303387673915e-009,-.000239493192417188,-.00250477651118074,-.00722607333006517,-.0118651027035077,-.0133663775728643,-.0109871701831305,-.00664025858844208,-.00285505335439599,-.000802000786155484,-.000122131231479417,-6.35129828208852e-006,-2.07381685017084e-008,-.00054515334463801,-.00570157038226907,-.0164485595800252,-.0270082851124352,-.030425605696751,-.0250098656792164,-.0151150817366226,-.00649889823405786,-.00182557761480488,-.000278004767708733,-1.44573274351898e-005,-4.72058591990341e-008,-.000959742855872972,-.0100376187644472,-.0289577009874395,-.0475481053927164,-.0535643007426736,-.044029886541206,-.0266101122837137,-.0114413150217837,-.00321393070570632,-.000489427593779599,-2.54521353623046e-005,-8.31059491191454e-008,-.00141708846333921,-.0148208383771267,-.0427568944566942,-.0702061715732879,-.0790892603834303,-.0650114182959186,-.0392906109117239,-.0168934370530446,-.00474546290931794,-.000722654190693695,-3.75808240390211e-005,-1.22708370279532e-007,-.00182283513943086,-.0190644026033435,-.0549992267137261,-.0903078952792718,-.101734427107537,-.0836257585887017,-.0505404624146264,-.0217304363719754,-.00610420363142845,-.000929567551026009,-4.83411222371217e-005,-1.57842742379509e-007,-.0020766085420598,-.0217185308967278,-.062656167598386,-.102880475966099,-.115897798864594,-.0952680584574148,-.0575766583052787,-.0247557273922317,-.00695402514977553,-.00105898107575704,-5.50711280460104e-005,-1.79817461292618e-007,-.00209684841095391,-.0219302126889477,-.0632668520831664,-.103883210618848,-.11702740813212,-.0961965979360501,-.0581378348543761,-.0249970115180785,-.00702180323816678,-.00106930253870999,-5.56078841986148e-005,-1.81570069821245e-007,-.00184217543696469,-.0192666760896712,-.0555827690131719,-.0912660628757668,-.102813830311442,-.0845130286536747,-.0510766971839986,-.0219609965010198,-.00616896928789685,-.000939430271260956,-4.88540219870558e-005,-1.59517455322627e-007,-.00132397601187602,-.0138470074345037,-.0399474726296093,-.0655931435851624,-.0738925524084332,-.0607397213008226,-.0367089477367546,-.0157834221327753,-.00443365338137005,-.000675170843678735,-3.51115056126793e-005,-1.14645586997212e-007,-.000607815019228362,-.00635692717584237,-.0183392097943569,-.0301127040609837,-.0339228224379417,-.0278845798860605,-.0168524577290889,-.00724590244919067,-.00203541536332734,-.000309959527704379,-1.61190990378048e-005,-5.2631852118241e-008},
			{-1.08304313867452e-005,-.000589001007959142,-.00407774946276834,-.0120086359724961,-.0207230264243475,-.0235713601331056,-.0183245877738932,-.00957752805114946,-.00312208933690556,-.000530130419242249,-2.96594300874774e-005,-1.00782323868636e-007,-2.45509398075641e-005,-.00133517565243957,-.00924363749104905,-.0272217503074272,-.0469759473290577,-.0534326863949788,-.0415390518964802,-.0217107986092594,-.00707730141552942,-.00120172498658768,-6.72334145152776e-005,-2.28458191424719e-007,-3.67505036876191e-005,-.00199863541368309,-.0138368769735303,-.0407484618877329,-.0703186818540462,-.0799838276575082,-.0621801483718135,-.0324990729725514,-.0105940706876594,-.00179887201456491,-.000100642251068273,-3.41980945423285e-007,-4.66428200392335e-005,-.00253661807513852,-.0175614181527209,-.0517169285858086,-.0892467121265429,-.101513473425809,-.0789174890001959,-.0412470105167922,-.0134457295270728,-.00228308336566367,-.00012773262768919,-4.34033661954923e-007,-5.3604806102002e-005,-.00291523797142323,-.020182665073873,-.0594362846564907,-.102567827047345,-.116665545847229,-.0906968466390758,-.0474036089408846,-.0154526618157401,-.00262386024318778,-.00014679821533987,-4.98818259086074e-007,-5.7198740774876e-005,-.00311069012556852,-.0215358120223382,-.0634211908577612,-.109444487868491,-.124487388343073,-.0967776174795907,-.0505817842982924,-.0164886856562537,-.00279977697510962,-.000156640302913549,-5.32261533433169e-007,-5.7198740774876e-005,-.00311069012556852,-.0215358120223382,-.0634211908577612,-.109444487868491,-.124487388343073,-.0967776174795907,-.0505817842982924,-.0164886856562537,-.00279977697510962,-.000156640302913549,-5.32261533433169e-007,-5.3604806102002e-005,-.00291523797142323,-.020182665073873,-.0594362846564907,-.102567827047345,-.116665545847229,-.0906968466390758,-.0474036089408846,-.0154526618157401,-.00262386024318778,-.00014679821533987,-4.98818259086074e-007,-4.66428200392335e-005,-.00253661807513852,-.0175614181527209,-.0517169285858085,-.0892467121265429,-.101513473425809,-.0789174890001959,-.0412470105167922,-.0134457295270728,-.00228308336566367,-.00012773262768919,-4.34033661954923e-007,-3.67505036876191e-005,-.00199863541368309,-.0138368769735303,-.0407484618877329,-.0703186818540462,-.0799838276575082,-.0621801483718135,-.0324990729725514,-.0105940706876594,-.00179887201456491,-.000100642251068273,-3.41980945423285e-007,-2.45509398075641e-005,-.00133517565243957,-.00924363749104905,-.0272217503074272,-.0469759473290577,-.0534326863949788,-.0415390518964802,-.0217107986092594,-.00707730141552942,-.00120172498658768,-6.72334145152776e-005,-2.28458191424719e-007,-1.08304313867452e-005,-.000589001007959142,-.00407774946276834,-.0120086359724961,-.0207230264243475,-.0235713601331056,-.0183245877738932,-.00957752805114946,-.00312208933690556,-.000530130419242249,-2.96594300874774e-005,-1.00782323868636e-007},
			{-.00719309508160456,-.0139021976630662,-.0155346480848768,-.0127550186066991,-.00808303973670482,-.00395000387691709,-.00144276568155165,-.000368787100047311,-5.82757244194688e-005,-4.43762053833248e-006,-8.93862080339512e-008,-5.3934946534894e-011,-.0163056519239563,-.0315141665027775,-.0352146831870672,-.028913686156645,-.0183230210276625,-.00895405768790711,-.00327053024384089,-.000835984234769121,-.000132102199014488,-1.00594104551317e-005,-2.02624931058027e-007,-1.22262316118168e-010,-.0244080644511933,-.0471738149883672,-.0527131488435292,-.0432811345737185,-.0274278808519569,-.0134034025849063,-.00489568361654787,-.00125139167557278,-.00019774486802044,-1.50580142318092e-005,-3.03310925545804e-007,-1.83015466396657e-010,-.0309781048820443,-.0598718260236465,-.0669022099971482,-.0549313334090684,-.0348107803231648,-.017011263301183,-.00621347919028916,-.00158823501355166,-.000250972840295156,-1.91112550165974e-005,-3.84954639980414e-007,-2.32278652181071e-010,-.0356019491148408,-.0688083958532261,-.0768881468075034,-.0631304769703183,-.0400066961627327,-.0195503931804077,-.00714091358400096,-.00182529765297712,-.000288433470136975,-2.19638332045296e-005,-4.42413619435074e-007,-2.6694895594493e-010,-.0379888820906312,-.0734216553279677,-.0820431132524654,-.0673630603261273,-.0426889454411423,-.0208611494544942,-.00761967619488893,-.00194767474936783,-.000307771494551617,-2.3436398585193e-005,-4.72075244245615e-007,-2.84846550926105e-010,-.0379888820906312,-.0734216553279677,-.0820431132524654,-.0673630603261273,-.0426889454411423,-.0208611494544942,-.00761967619488893,-.00194767474936783,-.000307771494551617,-2.3436398585193e-005,-4.72075244245615e-007,-2.84846550926105e-010,-.0356019491148408,-.0688083958532261,-.0768881468075034,-.0631304769703183,-.0400066961627327,-.0195503931804077,-.00714091358400096,-.00182529765297712,-.000288433470136975,-2.19638332045296e-005,-4.42413619435074e-007,-2.6694895594493e-010,-.0309781048820443,-.0598718260236465,-.0669022099971482,-.0549313334090684,-.0348107803231648,-.017011263301183,-.00621347919028916,-.00158823501355166,-.000250972840295157,-1.91112550165974e-005,-3.84954639980414e-007,-2.32278652181071e-010,-.0244080644511933,-.0471738149883672,-.0527131488435292,-.0432811345737185,-.0274278808519569,-.0134034025849063,-.00489568361654787,-.00125139167557278,-.00019774486802044,-1.50580142318092e-005,-3.03310925545804e-007,-1.83015466396657e-010,-.0163056519239563,-.0315141665027775,-.0352146831870672,-.028913686156645,-.0183230210276625,-.00895405768790711,-.00327053024384089,-.00083598423476912,-.000132102199014488,-1.00594104551317e-005,-2.02624931058027e-007,-1.22262316118168e-010,-.00719309508160456,-.0139021976630662,-.0155346480848768,-.0127550186066991,-.00808303973670482,-.00395000387691709,-.00144276568155165,-.000368787100047311,-5.82757244194688e-005,-4.43762053833248e-006,-8.93862080339512e-008,-5.3934946534894e-011},
			{-.000114866241683583,-.000222003626892483,-.000248072161029854,-.000203684371377566,-.000129077731546852,-6.30774506425453e-005,-2.30394662644891e-005,-5.88914614407857e-006,-9.30602664014744e-007,-7.08641812006219e-008,-1.42740470669776e-009,-8.61284959198155e-013,-.00135397065407101,-.00261683843315572,-.00292411783656719,-.00240090262810037,-.00152148671399835,-.000743517119145248,-.000271574665892772,-6.94175324250859e-005,-1.09693560023246e-005,-8.35302177246445e-007,-1.68253444704445e-008,-1.015228271122e-011,-.00486379816401721,-.00940033222170924,-.0105041559963674,-.0086246372913827,-.00546555736926158,-.00267089776882591,-.00097556350825723,-.000249364981245768,-3.94046454582605e-005,-3.00061244597533e-006,-6.04408074120732e-008,-3.64695156892415e-011,-.0110713620291109,-.0213977796180793,-.0239103905886559,-.0196320814726822,-.0124411339215587,-.00607970872643953,-.00222065480887151,-.000567624291069279,-8.96959702655164e-005,-6.8302313496928e-006,-1.37580145726718e-007,-8.30147957637377e-011,-.0194911408262166,-.0376708064290284,-.0420942598523953,-.0345622935723416,-.0219026252294268,-.0107033315917073,-.0039094643904085,-.00099930297324955,-.000157909820254493,-1.20246272104963e-005,-2.42209945641729e-007,-1.46147607732105e-010,-.0287792408488707,-.0556220500820546,-.0621534087434013,-.0510322397173339,-.0323398682673508,-.015803782883183,-.00577243878564134,-.00147550013642398,-.000233158479025165,-1.77547146005579e-005,-3.57630085574026e-007,-2.15791227404772e-010,-.0370194330577289,-.0715479873273676,-.0799494318272223,-.0656440033259579,-.041599554161483,-.0203287878778718,-.00742522752170763,-.00189797148624789,-.000299917386683,-2.28383184972103e-005,-4.60028222490692e-007,-2.77577471181613e-010,-.0421732438918713,-.0815088311812927,-.0910799169238503,-.0747829000510975,-.0473910051704331,-.0231589427061906,-.00845896075008184,-.00216220530078157,-.000341671604648038,-2.60178478303842e-005,-5.24072921213834e-007,-3.16221547012239e-010,-.0425842895511383,-.0823032650013445,-.0919676362701244,-.0755117789234986,-.0478529062519234,-.0233846636134322,-.00854140684095277,-.00218327944688257,-.00034500174046461,-2.62714333369098e-005,-5.29180849357959e-007,-3.19303631345124e-010,-.0374122095817176,-.072307112128577,-.0807976960315661,-.0663404868028083,-.0420409258123548,-.0205444765035491,-.00750400925375776,-.00191810898111,-.000303099512904151,-2.30806332657561e-005,-4.64909126147018e-007,-2.80522570694571e-010,-.026888246929992,-.0519672991102317,-.058069502618547,-.0476790708315039,-.0302149166554566,-.0147653657308921,-.00539314988437847,-.00137854963658506,-.000217838364492369,-1.65881078259945e-005,-3.34131331017576e-007,-2.01612260665772e-010,-.0123439399038752,-.0238573090634585,-.0266587276008582,-.0218886559078795,-.0138711577726375,-.00677852995457166,-.0024759039977092,-.000632868848341838,-.00010000591288281,-7.61532005623914e-006,-1.53394049110811e-007,-9.25567827468422e-011},
			{-.000538770430325433,-.00541459382780606,-.0145196103881163,-.0213815233314377,-.020756280852719,-.0140356226122099,-.00659447754217011,-.00204903932812416,-.000375745321421736,-3.19037343548278e-005,-6.91363066088356e-007,-4.34129792523551e-010,-.0012213105768994,-.0122740602291765,-.032913747194243,-.0484686596091265,-.0470513300576056,-.0318166205582331,-.0149486770580804,-.00464486033951284,-.000851757463741354,-7.23209107571181e-005,-1.56721486103326e-006,-9.84106174935942e-010,-.00182818984576065,-.0183731416903004,-.04926885883459,-.0725531351410981,-.0704315228803041,-.0476265609511484,-.0223767976156558,-.00695292963828415,-.00127500275172926,-.000108257766028679,-2.34597681315439e-006,-1.47311662585924e-009,-.00232029282368979,-.0233187318655643,-.0625308032698983,-.0920826243480438,-.0893899271345745,-.0604464398750644,-.0284000718224987,-.00882448438314045,-.00161820160082515,-.000137398048789906,-2.97745509129004e-006,-1.86964277444457e-009,-.00266662364773693,-.0267993251511533,-.0718642565322132,-.105827032314693,-.102732418569255,-.0694687775381858,-.0326391144885199,-.0101416418199092,-.00185973710365748,-.000157906313511514,-3.4218750648389e-006,-2.14870881134135e-009,-.00284540745247398,-.0285960861297693,-.0766823962117034,-.112922206580213,-.109620114430362,-.0741263123085118,-.0348274041920399,-.0108215882804389,-.00198442319330673,-.000168493143619897,-3.65129470714422e-006,-2.29276901154607e-009,-.00284540745247398,-.0285960861297693,-.0766823962117034,-.112922206580213,-.109620114430362,-.0741263123085118,-.0348274041920398,-.0108215882804389,-.00198442319330673,-.000168493143619897,-3.65129470714422e-006,-2.29276901154607e-009,-.00266662364773693,-.0267993251511533,-.0718642565322132,-.105827032314693,-.102732418569255,-.0694687775381858,-.0326391144885199,-.0101416418199092,-.00185973710365748,-.000157906313511514,-3.4218750648389e-006,-2.14870881134135e-009,-.00232029282368979,-.0233187318655643,-.0625308032698983,-.0920826243480438,-.0893899271345745,-.0604464398750644,-.0284000718224987,-.00882448438314046,-.00161820160082515,-.000137398048789906,-2.97745509129004e-006,-1.86964277444457e-009,-.00182818984576065,-.0183731416903004,-.04926885883459,-.0725531351410981,-.0704315228803041,-.0476265609511484,-.0223767976156558,-.00695292963828415,-.00127500275172926,-.000108257766028679,-2.34597681315439e-006,-1.47311662585924e-009,-.0012213105768994,-.0122740602291765,-.032913747194243,-.0484686596091265,-.0470513300576056,-.0318166205582331,-.0149486770580804,-.00464486033951284,-.000851757463741354,-7.23209107571182e-005,-1.56721486103326e-006,-9.84106174935942e-010,-.000538770430325433,-.00541459382780606,-.0145196103881163,-.0213815233314377,-.020756280852719,-.0140356226122099,-.00659447754217011,-.00204903932812416,-.000375745321421736,-3.19037343548278e-005,-6.91363066088356e-007,-4.34129792523551e-010},
			{-.00863095559575211,-.016029237481894,-.0166489373902391,-.0122597253338038,-.00669488358667141,-.00269138517084675,-.00076422934021594,-.000141170609267733,-1.45626016852063e-005,-6.18297441623807e-007,-5.18975291864507e-009,-6.02215524056038e-013,-.0195650628997474,-.0363358420848087,-.0377406074727463,-.0277909316794019,-.0151762822814405,-.00610096061448634,-.00173239161588508,-.000320012288242827,-3.30112019224449e-005,-1.40158620930416e-006,-1.1764380103799e-008,-1.36513094948138e-012,-.0292871035439603,-.0543914208171814,-.0564942256781886,-.0416004741639895,-.0227175017461691,-.0091325781138374,-.00259323125578963,-.000479029025826867,-4.94147396186346e-005,-2.09804592236745e-006,-1.76101973193715e-008,-2.0434757441563e-012,-.0371704592590807,-.0690322308074568,-.0717010581395466,-.0527982812554331,-.0288324849829067,-.0115908397087212,-.0032912642453023,-.00060797165761674,-6.27159515120887e-005,-2.66278740620364e-006,-2.23504219535243e-008,-2.59352830098302e-012,-.0427185847602993,-.0793360981199193,-.0824032791251857,-.0606790418511532,-.0331360703672624,-.0133209080116245,-.00378252390296041,-.00069871853362736,-7.20770403136054e-005,-3.06023954984572e-006,-2.56864836669144e-008,-2.98064271365472e-012,-.0455826526323854,-.0846551874810701,-.087928007662455,-.0647472687189207,-.0353576784818238,-.0142140084005345,-.00403612324963012,-.000745564123550858,-7.69094460834964e-005,-3.26541333602756e-006,-2.7408634178924e-008,-3.18047992928925e-012,-.0455826526323854,-.0846551874810701,-.087928007662455,-.0647472687189207,-.0353576784818238,-.0142140084005345,-.00403612324963012,-.000745564123550858,-7.69094460834964e-005,-3.26541333602756e-006,-2.7408634178924e-008,-3.18047992928925e-012,-.0427185847602993,-.0793360981199193,-.0824032791251857,-.0606790418511532,-.0331360703672624,-.0133209080116245,-.00378252390296041,-.00069871853362736,-7.20770403136054e-005,-3.06023954984572e-006,-2.56864836669144e-008,-2.98064271365472e-012,-.0371704592590807,-.0690322308074568,-.0717010581395466,-.0527982812554331,-.0288324849829067,-.0115908397087212,-.0032912642453023,-.00060797165761674,-6.27159515120887e-005,-2.66278740620364e-006,-2.23504219535243e-008,-2.59352830098302e-012,-.0292871035439603,-.0543914208171814,-.0564942256781886,-.0416004741639895,-.0227175017461691,-.0091325781138374,-.00259323125578962,-.000479029025826867,-4.94147396186346e-005,-2.09804592236745e-006,-1.76101973193715e-008,-2.0434757441563e-012,-.0195650628997474,-.0363358420848087,-.0377406074727463,-.0277909316794019,-.0151762822814405,-.00610096061448634,-.00173239161588507,-.000320012288242827,-3.30112019224449e-005,-1.40158620930416e-006,-1.1764380103799e-008,-1.36513094948138e-012,-.00863095559575211,-.016029237481894,-.0166489373902391,-.0122597253338038,-.00669488358667141,-.00269138517084675,-.00076422934021594,-.000141170609267733,-1.45626016852063e-005,-6.18297441623807e-007,-5.18975291864507e-009,-6.02215524056038e-013}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{-1.25649763607893e-005,-2.84828893052902e-005,-4.2636271224355e-005,-5.41128889760362e-005,-6.2189870139491e-005,-6.63593905024702e-005,-6.63593905024702e-005,-6.2189870139491e-005,-5.41128889760362e-005,-4.2636271224355e-005,-2.84828893052902e-005,-1.25649763607893e-005,-.000148107999463138,-.000335738296102257,-.000502569416311285,-.000637848532403299,-.000733054881183536,-.000782202551815753,-.000782202551815753,-.000733054881183536,-.000637848532403299,-.000502569416311285,-.000335738296102257,-.000148107999463138,-.000532040641869991,-.00120605517059218,-.00180535390260964,-.00229131001583844,-.0026333148170561,-.00280986543096192,-.00280986543096192,-.0026333148170561,-.00229131001583844,-.00180535390260964,-.00120605517059218,-.000532040641869991,-.00121107298487856,-.00274531815886015,-.00410948932756509,-.00521566106380589,-.00599415943941434,-.00639603790909957,-.00639603790909957,-.00599415943941434,-.00521566106380588,-.00410948932756509,-.00274531815886015,-.00121107298487856,-.00213209486213416,-.00483313459594366,-.00723475892097063,-.00918217505932459,-.010552722010284,-.0112602293456112,-.0112602293456112,-.010552722010284,-.00918217505932459,-.00723475892097063,-.00483313459594367,-.00213209486213416,-.00314810057025839,-.0071362649232201,-.0106823336471961,-.0135577506675935,-.0155814034208121,-.0166260587433132,-.0166260587433132,-.0155814034208121,-.0135577506675935,-.0106823336471961,-.0071362649232201,-.00314810057025839,-.00404947784869215,-.00917955004423722,-.0137409439473876,-.0174396623555152,-.0200427358008248,-.0213865011900097,-.0213865011900097,-.0200427358008248,-.0174396623555152,-.0137409439473876,-.00917955004423722,-.00404947784869215,-.00461324236601103,-.0104575184128172,-.0156539453074288,-.0198675958317378,-.0228330667266135,-.0243639098760258,-.0243639098760258,-.0228330667266135,-.0198675958317378,-.0156539453074288,-.0104575184128172,-.00461324236601103,-.00465820578534291,-.0105594436420292,-.0158065180211977,-.0200612373037489,-.023055611451648,-.0246013751140091,-.0246013751140091,-.023055611451648,-.0200612373037489,-.0158065180211977,-.0105594436420292,-.00465820578534291,-.00409244284577621,-.00927694515432801,-.0138867354885914,-.0176246973329244,-.0202553894113433,-.0216134121636217,-.0216134121636217,-.0202553894113433,-.0176246973329244,-.0138867354885914,-.00927694515432801,-.00409244284577621,-.00294124872640195,-.00666736327135993,-.0099804309086089,-.0126669132684861,-.0145575981275773,-.0155336124156416,-.0155336124156416,-.0145575981275773,-.0126669132684861,-.0099804309086089,-.00666736327135993,-.00294124872640195,-.00135027759956182,-.0030608738365598,-.00458184721642184,-.00581516588124543,-.0066831473580057,-.00713121902844807,-.00713121902844807,-.0066831473580057,-.00581516588124543,-.00458184721642184,-.0030608738365598,-.00135027759956182},
			{-1.49555377634105e-007,-3.39019282157687e-007,-5.07480751318181e-007,-6.4408187594673e-007,-7.40218624107564e-007,-7.89846587943962e-007,-7.89846587943962e-007,-7.40218624107564e-007,-6.4408187594673e-007,-5.07480751318181e-007,-3.39019282157687e-007,-1.49555377634105e-007,-9.16670694821036e-006,-2.07795296865571e-005,-3.11050488640572e-005,-3.94777499870446e-005,-4.53702655975518e-005,-4.84121154335211e-005,-4.84121154335211e-005,-4.53702655975518e-005,-3.94777499870446e-005,-3.11050488640572e-005,-2.07795296865571e-005,-9.16670694821036e-006,-7.90224989069185e-005,-.000179131979588681,-.000268144133301817,-.000340321826892032,-.000391118837314757,-.000417341403029572,-.000417341403029572,-.000391118837314757,-.000340321826892032,-.000268144133301817,-.000179131979588681,-7.90224989069185e-005,-.000322611816370485,-.000731311893505376,-.0010947067871833,-.00138937447237002,-.00159675485169678,-.00170380929406662,-.00170380929406662,-.00159675485169678,-.00138937447237002,-.0010947067871833,-.000731311893505376,-.000322611816370485,-.000870029052744199,-.00197222346386832,-.00295223752124433,-.00374690601758642,-.00430617553540284,-.00459488311014376,-.00459488311014376,-.00430617553540284,-.00374690601758643,-.00295223752124433,-.00197222346386832,-.000870029052744198,-.00177760460022009,-.00402955911756787,-.00603188017934985,-.00765551144808981,-.0087981860110801,-.00938806046568492,-.00938806046568492,-.0087981860110801,-.00765551144808981,-.00603188017934986,-.00402955911756787,-.00177760460022009,-.0029412776627297,-.00666742886561148,-.00998052909718346,-.0126670378869689,-.0145577413468253,-.0155337652370252,-.0155337652370252,-.0145577413468253,-.0126670378869689,-.00998052909718346,-.00666742886561148,-.0029412776627297,-.00407317670628986,-.00923327174798222,-.013821360407493,-.0175417249152225,-.0201600323904989,-.0215116619305183,-.0215116619305183,-.0201600323904989,-.0175417249152225,-.013821360407493,-.00923327174798222,-.00407317670628986,-.00477284112519419,-.0108193045125715,-.0161955058952257,-.0205549309837037,-.0236229946837403,-.0252067985596364,-.0252067985596364,-.0236229946837403,-.0205549309837037,-.0161955058952257,-.0108193045125715,-.00477284112519419,-.00467548197876496,-.0105986061434719,-.0158651406916522,-.0201356397307597,-.0231411193105227,-.0246926158479963,-.0246926158479963,-.0231411193105227,-.0201356397307597,-.0158651406916522,-.0105986061434719,-.00467548197876496,-.00361509587945292,-.0081948722230624,-.0122669716195725,-.015568933545552,-.0178927788504699,-.0190924003579595,-.0190924003579595,-.0178927788504699,-.015568933545552,-.0122669716195725,-.00819487222306241,-.00361509587945292,-.00172712912568916,-.00391513889802946,-.00586060361181134,-.00743813151271418,-.00854835957956341,-.00912148441898038,-.00912148441898038,-.00854835957956341,-.00743813151271418,-.00586060361181134,-.00391513889802946,-.00172712912568916},
			{-1.63148105416827e-009,-3.69831927536025e-009,-5.53604453566559e-009,-7.02620925147273e-009,-8.07495310619039e-009,-8.61633840464567e-009,-8.61633840464567e-009,-8.07495310619039e-009,-7.02620925147273e-009,-5.53604453566559e-009,-3.69831927536025e-009,-1.63148105416827e-009,-5.19981416069715e-007,-1.1787187408438e-006,-1.7644337761237e-006,-2.23937521483844e-006,-2.57362813998106e-006,-2.74617706012381e-006,-2.74617706012381e-006,-2.57362813998106e-006,-2.23937521483844e-006,-1.7644337761237e-006,-1.1787187408438e-006,-5.19981416069715e-007,-1.07571284042538e-005,-2.43847731013085e-005,-3.6501767378587e-005,-4.63271301374563e-005,-5.3241995792529e-005,-5.68116058451738e-005,-5.68116058451738e-005,-5.3241995792529e-005,-4.63271301374563e-005,-3.6501767378587e-005,-2.43847731013085e-005,-1.07571284042538e-005,-7.87643815441602e-005,-.000178546866806921,-.000267268273167243,-.000339210207117138,-.000389841295295595,-.000415978208195064,-.000415978208195064,-.000389841295295595,-.000339210207117138,-.000267268273167243,-.000178546866806921,-7.87643815441602e-005,-.000325387323726201,-.000737603546311472,-.00110412481400688,-.0014013275964412,-.00161049211924643,-.00171846757683388,-.00171846757683388,-.00161049211924644,-.0014013275964412,-.00110412481400688,-.000737603546311472,-.000325387323726201,-.000919943967994991,-.00208537298082308,-.00312161196363602,-.0039618718233036,-.00455322750019265,-.00485849867597654,-.00485849867597654,-.00455322750019265,-.0039618718233036,-.00312161196363602,-.00208537298082308,-.000919943967994991,-.00195799989864057,-.00443848780701132,-.00664400889731993,-.00843240990574942,-.00969104564411167,-.0103407818802712,-.0103407818802712,-.00969104564411167,-.00843240990574942,-.00664400889731993,-.00443848780701132,-.00195799989864057,-.00329609616783617,-.00747174842135327,-.0111845216543321,-.0141951151250118,-.0163139019731612,-.0174076676672231,-.0174076676672231,-.0163139019731612,-.0141951151250118,-.0111845216543321,-.00747174842135327,-.00329609616783617,-.00448203175305139,-.0101600839205191,-.0152087131700155,-.0193025183395333,-.022183644813943,-.0236709474657956,-.0236709474657956,-.022183644813943,-.0193025183395333,-.0152087131700155,-.0101600839205191,-.00448203175305139,-.00489564361298881,-.0110976790646466,-.0166122070513835,-.0210837976681447,-.0242308009023486,-.0258553551512484,-.0258553551512484,-.0242308009023486,-.0210837976681447,-.0166122070513835,-.0110976790646466,-.00489564361298881,-.00407237283003069,-.00923144948283444,-.0138186326438083,-.0175382629057834,-.0201560536356863,-.021507416420056,-.021507416420056,-.0201560536356863,-.0175382629057834,-.0138186326438083,-.00923144948283444,-.00407237283003069,-.00202472570360245,-.0045897450527047,-.00687042711223165,-.00871977424071199,-.0100213024648473,-.0106931807723117,-.0106931807723117,-.0100213024648473,-.00871977424071199,-.00687042711223165,-.0045897450527047,-.00202472570360245},
			{-1.70557276538753e-011,-3.86627390961513e-011,-5.7874572088226e-011,-7.34529592765333e-011,-8.44166719834875e-011,-9.00763884617597e-011,-9.00763884617597e-011,-8.44166719834875e-011,-7.34529592765333e-011,-5.7874572088226e-011,-3.86627390961513e-011,-1.70557276538753e-011,-2.82664008807978e-008,-6.40756292912074e-008,-9.59153363989562e-008,-1.21733345825304e-007,-1.39903470536826e-007,-1.49283299887593e-007,-1.49283299887593e-007,-1.39903470536826e-007,-1.21733345825304e-007,-9.59153363989562e-008,-6.40756292912074e-008,-2.82664008807978e-008,-1.40329882934204e-006,-3.18106489584209e-006,-4.76176220142831e-006,-6.04350947999848e-006,-6.94557390780468e-006,-7.41123996847038e-006,-7.41123996847038e-006,-6.94557390780468e-006,-6.04350947999848e-006,-4.76176220142831e-006,-3.18106489584208e-006,-1.40329882934204e-006,-1.84284021471628e-005,-4.17743832824876e-005,-6.2532417858728e-005,-7.93645806216638e-005,-9.12106719107567e-005,-9.73258921709034e-005,-9.73258921709034e-005,-9.12106719107567e-005,-7.93645806216638e-005,-6.2532417858728e-005,-4.17743832824876e-005,-1.84284021471628e-005,-.000116620720789999,-.000264361427000411,-.000395724793999518,-.000502244010271906,-.000577209799147466,-.000615908834952788,-.000615908834952788,-.000577209799147466,-.000502244010271906,-.000395724793999518,-.000264361427000411,-.000116620720789999,-.000456242568387133,-.00103423247275538,-.0015481510932686,-.00196487464364187,-.00225815515010815,-.00240955318100971,-.00240955318100971,-.00225815515010815,-.00196487464364187,-.0015481510932686,-.00103423247275538,-.000456242568387133,-.00124910090182185,-.00283152165958353,-.00423852805667493,-.00537943379114375,-.00618237716052034,-.00659687468888933,-.00659687468888933,-.00618237716052034,-.00537943379114375,-.00423852805667493,-.00283152165958353,-.00124910090182185,-.0025560815986996,-.00579424800656579,-.00867345748885218,-.0110081352954835,-.012651228153929,-.0134994298511859,-.0134994298511859,-.012651228153929,-.0110081352954835,-.00867345748885218,-.00579424800656579,-.0025560815986996,-.00403349103971441,-.00914331038111467,-.0136866964976499,-.0173708128492007,-.0199636097009642,-.021302070080123,-.021302070080123,-.0199636097009642,-.0173708128492007,-.0136866964976499,-.00914331038111467,-.00403349103971441,-.00491248705999991,-.0111358605958299,-.0166693612993899,-.021156336410114,-.0243141668994061,-.025944310422275,-.025944310422275,-.0243141668994061,-.021156336410114,-.0166693612993899,-.0111358605958299,-.00491248705999991,-.00439626090995001,-.00996565447158663,-.0149176701290573,-.0189331337918146,-.0217591253050265,-.0232179660428578,-.0232179660428578,-.0217591253050265,-.0189331337918146,-.0149176701290573,-.00996565447158663,-.00439626090995001,-.00227465638852363,-.00515629988163542,-.00771850769460683,-.00979613690280579,-.0112583248350307,-.012013139317637,-.012013139317637,-.0112583248350307,-.00979613690280579,-.00771850769460683,-.00515629988163542,-.00227465638852363},
			{-1.7384464750039e-013,-3.94079360668192e-013,-5.8990063561595e-013,-7.48687131527224e-013,-8.60437436733519e-013,-9.18125471866621e-013,-9.18125471866621e-013,-8.60437436733519e-013,-7.48687131527224e-013,-5.8990063561595e-013,-3.94079360668192e-013,-1.7384464750039e-013,-1.49815262263485e-009,-3.39608401063939e-009,-5.0836260825345e-009,-6.45201106711068e-009,-7.41504913145304e-009,-7.91219116240981e-009,-7.91219116240981e-009,-7.41504913145304e-009,-6.45201106711068e-009,-5.0836260825345e-009,-3.39608401063939e-009,-1.49815262263485e-009,-1.78487097219834e-007,-4.04603087706546e-007,-6.05653689159397e-007,-7.68680513053121e-007,-8.83415064139356e-007,-9.42643634493817e-007,-9.42643634493817e-007,-8.83415064139356e-007,-7.68680513053121e-007,-6.05653689159397e-007,-4.04603087706546e-007,-1.78487097219834e-007,-4.20386100245681e-006,-9.52951316020476e-006,-1.42648066135301e-005,-1.81045357479959e-005,-2.0806849318325e-005,-2.22018447046734e-005,-2.22018447046734e-005,-2.0806849318325e-005,-1.81045357479959e-005,-1.42648066135301e-005,-9.52951316020476e-006,-4.20386100245681e-006,-4.07524446402028e-005,-9.23795904009122e-005,-.00013828376853596,-.000175506300131138,-.000201702666783354,-.000215225823762682,-.000215225823762682,-.000201702666783354,-.000175506300131138,-.00013828376853596,-9.23795904009122e-005,-4.07524446402028e-005,-.000220614007696907,-.000500098382997093,-.000748601381966019,-.000950106149209752,-.00109192059703656,-.00116512842258577,-.00116512842258577,-.00109192059703656,-.000950106149209752,-.00074860138196602,-.000500098382997093,-.000220614007696907,-.000776935976612406,-.0017611956269338,-.0026363482167924,-.00334598721372161,-.00384541491403083,-.00410323078906309,-.00410323078906309,-.00384541491403083,-.00334598721372161,-.0026363482167924,-.0017611956269338,-.000776935976612406,-.00193264642090617,-.00438101533121211,-.00655797787568285,-.00832322148498332,-.00956555957532708,-.0102068825969568,-.0102068825969568,-.00956555957532708,-.00832322148498332,-.00655797787568285,-.00438101533121211,-.00193264642090617,-.00353907786161354,-.00802255094483993,-.0120090224811522,-.015241550951156,-.0175165305773379,-.0186909265162125,-.0186909265162125,-.0175165305773379,-.015241550951156,-.0120090224811522,-.00802255094483993,-.00353907786161354,-.00480613424305024,-.0108947749442933,-.0163084782050857,-.0206983126135993,-.0237877776977751,-.0253826294522236,-.0253826294522236,-.0237877776977751,-.0206983126135993,-.0163084782050857,-.0108947749442933,-.00480613424305024,-.00462724228856079,-.0104892541067666,-.0157014507286737,-.0199278885241277,-.0229023588080574,-.0244378476456489,-.0244378476456489,-.0229023588080574,-.0199278885241277,-.0157014507286737,-.0104892541067666,-.00462724228856079,-.00249154226883633,-.00564794716718901,-.00845445855931378,-.010730187331938,-.0123317932081052,-.0131585783867511,-.0131585783867511,-.0123317932081052,-.010730187331938,-.00845445855931378,-.00564794716718901,-.00249154226883633},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.77695599831757e-005,-.000209456341570312,-.000752419091586911,-.00171271584051363,-.00301523747077974,-.00445208652288796,-.0057268264950684,-.00652410992157336,-.00658769779901337,-.00578758817666173,-.00415955383985738,-.0019095808945752,-4.02808883575737e-005,-.000474805651832001,-.00170561957949523,-.00388246617391187,-.00683508449545443,-.010092202640724,-.012981844171125,-.0147891643705445,-.0149333084120673,-.0131195816567463,-.0094290755651378,-.00432872929307012,-6.02967930241717e-005,-.000710740484695346,-.00255315597436325,-.0058116955424022,-.0102314941881775,-.0151071011240822,-.0194326292933208,-.0221380217619632,-.0223537921630579,-.0196388096682036,-.0141144607515462,-.00647971047522447,-7.65271815013742e-005,-.000902054045409203,-.00324040170051977,-.00737605861435863,-.0129855565030636,-.0191735548724589,-.0246634070303329,-.0280970234814972,-.0283708738774972,-.0249250860049403,-.0179137205405699,-.00822388645802562,-8.79497578075961e-005,-.00103669615509983,-.00372406952887607,-.00847702157560611,-.0149238025892903,-.0220354320420534,-.0283447088011335,-.0322908326405259,-.0326055584089547,-.0286454464212646,-.0205875527108991,-.00945139763454551,-9.38463500544579e-005,-.00110620145747811,-.00397374980154702,-.00904536355795187,-.015924369058549,-.0235127977673766,-.0302450800394712,-.0344557717846379,-.0347915983448399,-.030565980615856,-.0219678453543708,-.0100850666679019,-9.38463500544579e-005,-.00110620145747811,-.00397374980154702,-.00904536355795187,-.015924369058549,-.0235127977673766,-.0302450800394711,-.0344557717846379,-.0347915983448399,-.030565980615856,-.0219678453543708,-.0100850666679019,-8.79497578075961e-005,-.00103669615509983,-.00372406952887607,-.00847702157560611,-.0149238025892903,-.0220354320420534,-.0283447088011335,-.0322908326405259,-.0326055584089547,-.0286454464212646,-.0205875527108991,-.00945139763454551,-7.65271815013742e-005,-.000902054045409203,-.00324040170051977,-.00737605861435863,-.0129855565030636,-.0191735548724589,-.0246634070303329,-.0280970234814972,-.0283708738774972,-.0249250860049403,-.0179137205405699,-.00822388645802562,-6.02967930241717e-005,-.000710740484695346,-.00255315597436325,-.0058116955424022,-.0102314941881775,-.0151071011240822,-.0194326292933208,-.0221380217619632,-.0223537921630579,-.0196388096682036,-.0141144607515462,-.00647971047522447,-4.02808883575737e-005,-.000474805651832001,-.00170561957949523,-.00388246617391187,-.00683508449545443,-.010092202640724,-.012981844171125,-.0147891643705445,-.0149333084120673,-.0131195816567463,-.0094290755651378,-.00432872929307012,-1.77695599831757e-005,-.000209456341570312,-.000752419091586911,-.00171271584051363,-.00301523747077974,-.00445208652288796,-.0057268264950684,-.00652410992157336,-.00658769779901337,-.00578758817666173,-.00415955383985738,-.0019095808945752},
			{-2.83761377805336e-007,-3.34479976602295e-006,-1.20153497508037e-005,-2.73502893236142e-005,-4.815020405282e-005,-7.10951878965703e-005,-9.14514584621857e-005,-.000104183244945397,-.000105198677163819,-9.2421759274531e-005,-6.64237454259384e-005,-3.04940193335369e-005,-3.34479976602295e-006,-3.94263855120624e-005,-.000141629348384198,-.000322387923394712,-.000567564171331005,-.00083802513816711,-.00107797198911481,-.00122804624086627,-.00124001551403817,-.00108940928179785,-.000782961126977123,-.000359444225710958,-1.20153497508037e-005,-.000141629348384198,-.000508767721494418,-.00115809732300615,-.00203883117125292,-.00301039399647408,-.00387234255465894,-.004411446462074,-.00445444305782967,-.00391342814465003,-.00281259042095047,-.00129121274513816,-2.73502893236142e-005,-.000322387923394712,-.00115809732300615,-.00263615271349073,-.00464094875074617,-.00685249688850132,-.00881453236290365,-.0100416833114065,-.0101395555630045,-.00890805463205047,-.00640224074681438,-.00293916056463564,-4.815020405282e-005,-.000567564171331005,-.00203883117125292,-.00464094875074617,-.00817039361825585,-.0120638257076051,-.0155179905734142,-.0176783907020889,-.017850694871518,-.0156826365956051,-.0112711494458756,-.00517439429092323,-7.10951878965703e-005,-.00083802513816711,-.00301039399647408,-.00685249688850132,-.0120638257076051,-.0178125923307157,-.0229127680203357,-.0261026621464626,-.0263570743040411,-.0231558726990041,-.0166421825914966,-.00764014486751864,-9.14514584621857e-005,-.00107797198911481,-.00387234255465894,-.00881453236290365,-.0155179905734142,-.0229127680203357,-.0294732472739763,-.0335764851836737,-.0339037416907514,-.0297859586976462,-.0214072416856188,-.00982770299465891,-.000104183244945397,-.00122804624086627,-.004411446462074,-.0100416833114065,-.0176783907020889,-.0261026621464626,-.0335764851836737,-.038250972036084,-.0386237888878892,-.0339327319991675,-.0243875378440283,-.0111959066105712,-.000105198677163819,-.00124001551403817,-.00445444305782967,-.0101395555630045,-.017850694871518,-.0263570743040411,-.0339037416907514,-.0386237888878892,-.039000239435718,-.0342634607007844,-.02462523337432,-.0113050286127969,-9.2421759274531e-005,-.00108940928179785,-.00391342814465003,-.00890805463205047,-.0156826365956051,-.0231558726990041,-.0297859586976462,-.0339327319991675,-.0342634607007844,-.0301019879923868,-.0216343727160793,-.00993197501349335,-6.64237454259384e-005,-.000782961126977123,-.00281259042095047,-.00640224074681438,-.0112711494458756,-.0166421825914966,-.0214072416856188,-.0243875378440283,-.02462523337432,-.0216343727160793,-.0155486768161827,-.00713813484023199,-3.04940193335369e-005,-.000359444225710958,-.00129121274513816,-.00293916056463564,-.00517439429092323,-.00764014486751864,-.00982770299465891,-.0111959066105712,-.0113050286127969,-.00993197501349335,-.00713813484023199,-.00327699711041027},
			{-3.37748665821831e-009,-3.98116779370311e-008,-1.43013414267412e-007,-3.25538443615425e-007,-5.73110664448679e-007,-8.46214697861095e-007,-1.08850641767933e-006,-1.24004726271932e-006,-1.25213350502815e-006,-1.10005548074599e-006,-7.90612576313682e-007,-3.62956876834475e-007,-2.07016497214321e-007,-2.44017962134497e-006,-8.76572998571608e-006,-1.99532478216857e-005,-3.51277071610772e-005,-5.18670894572505e-005,-6.67178907235581e-005,-7.60062930474401e-005,-7.67470958397088e-005,-6.74257681555381e-005,-4.84590652057195e-005,-2.22467381475193e-005,-1.78460607688856e-006,-2.10358084478815e-005,-7.5565837560655e-005,-.000172008935497438,-.000302821854829854,-.000447125346440885,-.000575148139518477,-.000655219725381629,-.000661605888719569,-.000581250466549428,-.000417746138158903,-.000191780194444643,-7.28570996785344e-006,-8.58793440610684e-005,-.000308499888616729,-.000702231843846736,-.00123627966686315,-.00182540317195597,-.00234806021751757,-.00267495496410619,-.00270102667510736,-.00237297315794306,-.00170546163785514,-.000782948625132055,-1.96483173282813e-005,-.000231601945658244,-.000831971590143711,-.00189379952905124,-.00333403543487275,-.00492280106302518,-.00633231798455615,-.00721389736971712,-.00728420832819645,-.00639950393640167,-.00459933919956203,-.00211147892329134,-4.01445666201286e-005,-.000473198777355153,-.0016998472881189,-.0038693268277857,-.0068119526671359,-.0100580478180592,-.0129379099972572,-.0147391137220966,-.0148827699400637,-.0130751813408912,-.00939716494906272,-.00431407966834182,-6.64243989158946e-005,-.000782968829903678,-.00281261809177179,-.00640230373332618,-.011271260333675,-.0166423463205833,-.0214074522943001,-.0243877777734532,-.0246254756422405,-.0216345855593206,-.0155488297871431,-.00713820506661556,-9.19866620625105e-005,-.00108428063087214,-.00389500475941537,-.00886611786558599,-.0156088068880401,-.0230468609713296,-.029645734277678,-.0337729856666483,-.0341021573849623,-.0299602757900378,-.021532523808133,-.00988521790053479,-.000107787546507207,-.00127053146952515,-.00456406393315995,-.0103890832686787,-.0182899885770756,-.0270057043390071,-.034738090181205,-.0395742945945498,-.039960009339504,-.0351066616363676,-.0252312439580584,-.0115832378335861,-.000105588834409239,-.00124461444104258,-.00447096354344966,-.0101771607988773,-.0179168989163566,-.0264548265171398,-.0340294827249939,-.0387670354712729,-.0391448821859808,-.0343905358485524,-.0247165625951827,-.0113469563151396,-8.16415851719115e-005,-.000962339403243543,-.00345696164727354,-.00786900949156683,-.0138533968774284,-.0204549466275595,-.0263116922143577,-.0299747813866719,-.0302669335343241,-.0265908595098725,-.0191109160505604,-.00877349869072301,-3.90046528002956e-005,-.000459762193746235,-.0016515797496082,-.003759456439442,-.00661852576932447,-.00977244733277217,-.01257053519047,-.014320593338432,-.0144601704064181,-.0127039086840373,-.0091303303785454,-.00419158042503607},
			{-3.68445827925519e-011,-4.34300653798924e-010,-1.56011558760704e-009,-3.55125848055141e-009,-6.25199311274643e-009,-9.2312511197497e-009,-1.18743814217069e-008,-1.35275216933177e-008,-1.36593690107054e-008,-1.20003687174081e-008,-8.62469447627331e-009,-3.95945152473478e-009,-1.174301000065e-008,-1.38419179545726e-007,-4.97235999402952e-007,-1.13184790520788e-006,-1.99261905231337e-006,-2.9421604480658e-006,-3.78457209223235e-006,-4.31145667798826e-006,-4.35347871350321e-006,-3.82472643681277e-006,-2.74884028563086e-006,-1.26194613503531e-006,-2.42933809796949e-007,-2.86355020000406e-006,-1.02865820344585e-005,-2.34151315299593e-005,-4.12223559228577e-005,-6.08660170299574e-005,-7.82934287517701e-005,-8.91933666496246e-005,-9.0062698548559e-005,-7.91241227483076e-005,-5.68667013887132e-005,-2.61065418768862e-005,-1.77877687852613e-006,-2.09670975420156e-005,-7.53190109571436e-005,-.000171447089262513,-.000301832724128607,-.000445664865962291,-.000573269488181101,-.000653079529962769,-.000659444833676081,-.000579351882704146,-.000416381621285402,-.000191153767800758,-7.34839068958039e-006,-8.66181847905149e-005,-.000311153987634738,-.000708273313928056,-.00124691567929258,-.00184110755612247,-.00236826114642384,-.00269796824743438,-.00272426425965937,-.00239338841889015,-.00172013413604819,-.000789684521168945,-2.07755717461149e-005,-.000244889308237957,-.000879703089733841,-.00200244974321189,-.00352531421515612,-.00520522978980597,-.00669561288171989,-.0076277698425577,-.00770211464969917,-.00676665339575356,-.00486321041790597,-.00223261774167844,-4.42185272019898e-005,-.000521220049687516,-.00187235160015967,-.00426199478516835,-.00750324488892071,-.0110787610500387,-.0142508780967575,-.016234872011969,-.0163931067848593,-.0144020800439593,-.0103508103065028,-.00475188214064504,-7.44374492353296e-005,-.000877421602303038,-.00315191585985935,-.00717464014604113,-.0126309591445223,-.0186499814768931,-.0239899219175831,-.0273297764014771,-.027596148748536,-.0242444552089724,-.0174245495155288,-.00799931630469631,-.000101220047626196,-.00119311794379991,-.00428597536221463,-.00975607607116152,-.0171755789499336,-.0253602458535457,-.0326214971629358,-.0371630314766057,-.0375252446090086,-.032967611546778,-.0236939034040642,-.0108774707577442,-.000110560858773534,-.00130322152160786,-.00468149470230067,-.0106563884722811,-.01876057958054,-.0277005457518707,-.0356318814839124,-.0405925186861703,-.0409881576521328,-.0360099360729874,-.0258804295145632,-.0118812679549556,-9.19685077033871e-005,-.00108406663876152,-.0038942360467172,-.00886436806094961,-.0156057263557175,-.0230423124750437,-.0296398834369748,-.0337663202773765,-.03409542703082,-.0299543628718611,-.0215282741793143,-.00988326696773915,-4.57254306619082e-005,-.00053898247521257,-.00193615863497058,-.00440723740391471,-.00775894462383229,-.0114563092049479,-.0147365274132258,-.0167881329718923,-.0169517601570097,-.0148928821041257,-.0107035509528065,-.00491381941200989},
			{-3.85178343554166e-013,-4.54023885618748e-012,-1.63096632460441e-011,-3.71253453125811e-011,-6.53591971617201e-011,-9.65047707354281e-011,-1.24136418981734e-010,-1.41418575087745e-010,-1.42796925104609e-010,-1.25453507524717e-010,-9.01637440362945e-011,-4.13926516217373e-011,-6.38354791449189e-010,-7.5245228000818e-009,-2.70299508117877e-008,-6.15277116719808e-008,-1.08319580712847e-007,-1.5993703651203e-007,-2.05730875518941e-007,-2.34372535522585e-007,-2.36656870434667e-007,-2.07913681993516e-007,-1.49428073991582e-007,-6.85999042669627e-008,-3.16914252834386e-008,-3.73558490210382e-007,-1.34191468136724e-006,-3.05457232158762e-006,-5.37757677216062e-006,-7.94014975773712e-006,-1.02136065356437e-005,-1.16355352814725e-005,-1.17489421676717e-005,-1.03219729945844e-005,-7.41842734727761e-006,-3.40567466501195e-006,-4.16178163573184e-007,-4.90564514068014e-006,-1.76222931839914e-005,-4.01132573852444e-005,-7.06194185176384e-005,-.000104271641780592,-.000134127132921491,-.000152800186874675,-.000154289468887485,-.00013555022303093,-9.74202782733315e-005,-4.47240038949301e-005,-2.6337062228933e-006,-3.10444642827677e-005,-.00011151940991293,-.000253849300234756,-.000446901875894084,-.000659863721515399,-.000848798652964526,-.000966967607276502,-.000976392251931343,-.000857804414450873,-.00061650613987428,-.000283027553294656,-1.03035625519223e-005,-.000121451882844024,-.000436285264394079,-.000993107021958261,-.00174836562742656,-.00258151310547804,-.00332066269912202,-.00378296225321243,-.00381983326596499,-.00335589496079028,-.00241188995212247,-.00110725792950642,-2.82090935115534e-005,-.00033251096434234,-.00119446179503315,-.00271892840056321,-.00478667540745511,-.00706766656937942,-.0090913103237661,-.010356994041021,-.0104579394995808,-.00918776920961406,-.00660327230083207,-.00303145075477106,-5.77253164541879e-005,-.00068042954422791,-.00244427156379506,-.00556384423606199,-.00979515178485467,-.0144628287733888,-.0186038861974662,-.0211939018276931,-.0214004702712195,-.0188012736005075,-.013512521522274,-.00620337034448753,-9.10904201184736e-005,-.00107371629731872,-.00385705505497627,-.00877973374713072,-.0154567275852829,-.0228223114223215,-.0293568907657418,-.0334439296312284,-.033769894172609,-.029668367645793,-.0213227287077879,-.00978890451433557,-.000110941243135535,-.0013077052520396,-.00469760137328079,-.0106930517505446,-.0188251253083289,-.0277958494112289,-.0357544728843004,-.0407321771466212,-.0411291773074706,-.0361338281692569,-.0259694710684948,-.011922145428059,-9.92830402484601e-005,-.00117028572514512,-.00420395637395983,-.00956937796371369,-.0168468967973055,-.024874937019263,-.0319972326801651,-.0364518574766186,-.0368071390818563,-.0323367236121244,-.0232404826956437,-.0106693129707941,-5.13697449717053e-005,-.000605514084724702,-.00217515666585201,-.00495126362270572,-.00871670317383983,-.0128704677825032,-.0165555937697616,-.0188604480446285,-.01904427325192,-.0167312487713743,-.0120247895925313,-.00552037774992995},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.5903751261877e-007,-1.58772021730906e-005,-.000136870983053836,-.000558780057100243,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567,-5.8719862146837e-007,-3.59912011760796e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-8.78982445184828e-007,-5.38755250068202e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-1.11558253342289e-006,-6.83774687490587e-005,-.000589454695127486,-.00240646717698951,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-1.2820962657192e-006,-7.85836051712971e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.3680544205638e-006,-8.38522436364219e-005,-.00072285651418097,-.00295108426386074,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.3680544205638e-006,-8.38522436364219e-005,-.00072285651418097,-.00295108426386074,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.2820962657192e-006,-7.85836051712971e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.11558253342289e-006,-6.83774687490587e-005,-.000589454695127486,-.00240646717698951,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-8.78982445184828e-007,-5.38755250068202e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-5.8719862146837e-007,-3.59912011760796e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-2.5903751261877e-007,-1.58772021730906e-005,-.000136870983053836,-.000558780057100243,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567},
			{-4.13655946284767e-009,-2.53542393256769e-007,-2.1856871401242e-006,-8.92313591757784e-006,-2.40641758792944e-005,-4.91668520822556e-005,-8.13529419075361e-005,-.000112660192597526,-.000132012244784625,-.000129319383418971,-9.99901127316091e-005,-4.77707484775864e-005,-4.87591483748876e-008,-2.98859747651745e-006,-2.57634985121267e-005,-.0001051802861973,-.000283653295149336,-.000579547775714673,-.0009589370588843,-.0013279671418103,-.00155607690124302,-.00152433515352727,-.00117862024866094,-.00056309138885062,-1.75154945664267e-007,-1.07357828440125e-005,-9.25488720048507e-005,-.000377833656408332,-.00101895293817233,-.00208187924827317,-.00344473958308176,-.00477038710314128,-.0055898138948427,-.00547578967002042,-.00423389604809699,-.00202276382802748,-3.98701539258914e-007,-2.44376379372225e-005,-.000210667061584079,-.000860054849279415,-.00231942126065002,-.00473893818806989,-.00784118866248488,-.0108587323850331,-.0127239764517783,-.0124644254937564,-.00963752901773119,-.00460437499342809,-7.01914347023563e-007,-4.30224791892881e-005,-.000370879513648267,-.00151412718159681,-.00408334279989948,-.00834290409323953,-.0138044182877922,-.0191168061846747,-.0224005697075911,-.0219436300590561,-.0169668767769879,-.00810600549215572,-1.03639711130188e-006,-6.35239518067959e-005,-.00054761447492283,-.0022356531730757,-.00602917535482301,-.0123185424813832,-.0203826283040634,-.0282265247763196,-.0330750978875268,-.0324004131004311,-.0250520909767,-.0119687547517608,-1.33314265253001e-006,-8.17123944937675e-005,-.000704409734165886,-.00287577470912415,-.00775547397560826,-.0158456389156716,-.0262186674070122,-.0363084610152352,-.0425452977913825,-.0416774344435578,-.0322251100971785,-.0153956985051803,-1.51874152529913e-006,-9.30883176033672e-005,-.000802476998296003,-.00327613737349354,-.00883518380630904,-.0180516539400008,-.0298688057526838,-.0413632909868219,-.0484684143436297,-.047479727872512,-.0367114597744218,-.0175390732465345,-1.53354408858128e-006,-9.39956120239112e-005,-.000810298419092016,-.00330806856783063,-.00892129679210925,-.0182275961561996,-.0301599249984219,-.0417664423606343,-.0489408164993326,-.0479424936985227,-.037069271618925,-.0177100195447144,-1.34728730829027e-006,-8.25793637481591e-005,-.000711883527900599,-.00290628670514157,-.00783775962556066,-.0160137612897764,-.0264968477084686,-.0366936941193377,-.0429967037908343,-.0421196324049362,-.0325670188106206,-.0155590470074065,-9.68298698098172e-007,-5.93499915831539e-005,-.000511632440253917,-.00208875539431883,-.00563301709645563,-.0115091295769889,-.0190433495379502,-.0263718481021364,-.0309018366364716,-.0302714832768886,-.0234059964205263,-.0111823253052378,-4.44529573439469e-007,-2.72465784513743e-005,-.000234881809580631,-.000958912313193877,-.0025860229823533,-.005283646948578,-.00874248004628156,-.012106869926272,-.0141865111307979,-.0138971265528759,-.0107452975256271,-.00513361662858989},
			{-4.92356447782966e-011,-3.0178033998418e-009,-2.60152710469109e-008,-1.06208155422931e-007,-2.86425283165094e-007,-5.85211378134644e-007,-9.68308224601569e-007,-1.34094463604488e-006,-1.57128358699498e-006,-1.53923164459519e-006,-1.19013825765391e-006,-5.68594172031225e-007,-3.0178033998418e-009,-1.84970409164038e-007,-1.59455560634358e-006,-6.50982299448902e-006,-1.75558824755616e-005,-3.58693969483533e-005,-5.93505754917118e-005,-8.21906019486044e-005,-9.63087814184433e-005,-9.4344219744047e-005,-7.29472173341576e-005,-3.48508775138952e-005,-2.60152710469109e-008,-1.59455560634358e-006,-1.37460234489013e-005,-5.61185694462147e-005,-.000151342211720415,-.000309215664595502,-.000511637473895075,-.00070853216856797,-.000830239323386619,-.000813303626233361,-.000628848662297057,-.000300435417626676,-1.06208155422931e-007,-6.50982299448902e-006,-5.61185694462147e-005,-.000229105810010919,-.000617859299465599,-.00126238259464391,-.00208877594431585,-.00289260467614267,-.00338947793153767,-.00332033741971269,-.00256729427659293,-.00122653696255208,-2.86425283165094e-007,-1.75558824755616e-005,-.000151342211720415,-.000617859299465599,-.00166626116517048,-.00340443057968316,-.00563307251629313,-.00780086152659004,-.00914084396303246,-.00895438379338928,-.00692355485520927,-.0033077610227998,-5.85211378134644e-007,-3.58693969483533e-005,-.000309215664595502,-.00126238259464391,-.00340443057968316,-.00695578089086413,-.0115092428083319,-.0159383727377934,-.0186761651548639,-.018295198043168,-.0141458988313976,-.0067582699589302,-9.68308224601569e-007,-5.93505754917118e-005,-.000511637473895075,-.00208877594431585,-.00563307251629313,-.0115092428083319,-.0190435368939121,-.0263721075587493,-.0309021406608933,-.030271781099646,-.023406226697924,-.0111824353213518,-1.34094463604488e-006,-8.21906019486044e-005,-.00070853216856797,-.00289260467614267,-.00780086152659004,-.0159383727377934,-.0263721075587493,-.036520949914119,-.0427942866834367,-.0419213443176089,-.0324137019011133,-.0154857991299779,-1.57128358699498e-006,-9.63087814184433e-005,-.000830239323386619,-.00338947793153767,-.00914084396303246,-.0186761651548639,-.0309021406608933,-.0427942866834367,-.0501452173903112,-.0491223265304294,-.037981521698904,-.0181458513277659,-1.53923164459519e-006,-9.4344219744047e-005,-.000813303626233361,-.00332033741971269,-.00895438379338928,-.018295198043168,-.030271781099646,-.0419213443176089,-.0491223265304294,-.0481203011840638,-.0372067528692506,-.0177757018611982,-1.19013825765391e-006,-7.29472173341576e-005,-.000628848662297057,-.00256729427659293,-.00692355485520927,-.0141458988313976,-.023406226697924,-.0324137019011133,-.037981521698904,-.0372067528692506,-.0287683664692428,-.0137442229153402,-5.68594172031225e-007,-3.48508775138952e-005,-.000300435417626676,-.00122653696255208,-.0033077610227998,-.0067582699589302,-.0111824353213518,-.0154857991299779,-.0181458513277659,-.0177757018611982,-.0137442229153402,-.00656636739345367},
			{-5.37105538571509e-013,-3.2920842768966e-011,-2.83797363265043e-010,-1.15861158670691e-009,-3.1245778676732e-009,-6.38399829730143e-009,-1.05631542517567e-008,-1.46281986187158e-008,-1.71409376487648e-008,-1.67912869868831e-008,-1.29830705524454e-008,-6.20272325817235e-009,-1.71184886157124e-010,-1.04924457427685e-008,-9.045116058834e-008,-3.69269684126226e-007,-9.95857365368599e-007,-2.03469140284302e-006,-3.36666116468723e-006,-4.66226157691217e-006,-5.46311525260784e-006,-5.35167549924359e-006,-4.13793062049084e-006,-1.9769158918718e-006,-3.54139156583407e-009,-2.1706273078516e-007,-1.87121062155825e-006,-7.6392757225221e-006,-2.06018238739433e-005,-4.20927287149029e-005,-6.96478861031055e-005,-9.64506516716335e-005,-.000113018331892963,-.000110712919239298,-8.56035420437903e-005,-4.08974964028785e-005,-2.59302953359118e-008,-1.58934718486943e-006,-1.37011237392702e-005,-5.59352649812969e-005,-.000150847870838019,-.000308205649327789,-.000509966272467282,-.000706217834626139,-.000827527447862548,-.00081064706909918,-.000626794604813159,-.000299454081965167,-1.07121889836103e-007,-6.56582857400443e-006,-5.6601371053274e-005,-.000231076862629534,-.000623174892248206,-.0012732431770043,-.00210674618825924,-.00291749045281179,-.00341863842883094,-.00334890308448347,-.00258938132932369,-.00123708915628192,-3.02857945621352e-007,-1.85630906649181e-005,-.000160024948988241,-.000653306845161927,-.0017618571509321,-.00359974803892942,-.0059562506178524,-.00824840904375729,-.00966526834956002,-.00946811067096992,-.00732076992880829,-.00349753239972921,-6.44599940278459e-007,-3.95095037359528e-005,-.000340595563207873,-.00139049200941683,-.00374991982442457,-.00766166912395449,-.0126772265613641,-.0175558345219898,-.0205714642490873,-.0201518357411283,-.0155814563465409,-.0074441143400151,-1.08511925583571e-006,-6.65102812046103e-005,-.000573358421239403,-.00234075363682485,-.00631261338244207,-.0128976504320703,-.0213408376153209,-.0295534841096122,-.0346299938653073,-.0339235914196489,-.0262298167578035,-.0125314188045125,-1.47554522466982e-006,-9.04406840954337e-005,-.00077965281321311,-.00318295692604349,-.0085838920298908,-.0175382257776313,-.0290192721808159,-.040186829342282,-.047089867591508,-.0461293015064619,-.0356673062917253,-.0170402239900308,-1.61171182018355e-006,-9.87867515986945e-005,-.000851600909064724,-.00347668727130158,-.00937603268029293,-.0191566922641638,-.0316972351677059,-.0438953593449828,-.0514354253188647,-.0503862160588214,-.0389587646541599,-.0186127337638557,-1.34067998923389e-006,-8.2174380935362e-005,-.000708392333727775,-.00289203379604635,-.00779932196032425,-.0159352271645878,-.0263669027993754,-.0365137421945241,-.0427858408675583,-.0419130707843193,-.0324073047817918,-.015482742876017,-6.66566974047622e-007,-4.08559304861565e-005,-.000352202567445854,-.00143787796622194,-.0038777116690518,-.00792276772768902,-.0131092481092602,-.0181541119739297,-.0212725100010394,-.0208385811603797,-.0161124498455309,-.0076977989913323},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.26296210833654e-009,-1.03996283213943e-006,-2.15142568085075e-005,-.00015752876308832,-.000650774647452403,-.00183988793598998,-.00391599979728114,-.00659219233567233,-.00896406350610278,-.00979128722597762,-.00814474566006138,-.00404945140720489,-7.3966385507205e-009,-2.35743748168761e-006,-4.87695462026171e-005,-.000357093733613841,-.00147520709262294,-.00417074596164616,-.00887697561402263,-.0149434968427065,-.0203201678410382,-.0221953581292932,-.0184628989656689,-.0091794901054094,-1.10720890713312e-008,-3.5288675522474e-006,-7.3003534757174e-005,-.000534536546334486,-.00220824962801375,-.00624322392727204,-.0132880177946399,-.0223690433086643,-.030417426340031,-.0332244141027671,-.0276372652876165,-.0137408542244633,-1.40524185029455e-008,-4.47875042967689e-006,-9.26542602749126e-005,-.000678420414234276,-.00280265519288241,-.00792374364660721,-.0168648198114988,-.0283902302500237,-.0386050366790666,-.0421675953362895,-.0350765258115668,-.017439548481424,-1.61499062123808e-008,-5.14725627996212e-006,-.000106483991585058,-.000779682590591189,-.00322098423849287,-.00910645500038529,-.0193820912882233,-.0326278039463224,-.0443672896278861,-.0484616018046971,-.0403121072713726,-.0200426049296945,-1.72326768092913e-008,-5.49235412024762e-006,-.000113623211690348,-.000831956416390127,-.00343693515366777,-.00971699735195308,-.0206815637605425,-.0348153353344463,-.0473418949315912,-.0517107103024967,-.043014832840112,-.0213863615446234,-1.72326768092913e-008,-5.49235412024762e-006,-.000113623211690348,-.000831956416390127,-.00343693515366777,-.00971699735195308,-.0206815637605425,-.0348153353344463,-.0473418949315912,-.0517107103024967,-.043014832840112,-.0213863615446234,-1.61499062123808e-008,-5.14725627996212e-006,-.000106483991585058,-.000779682590591189,-.00322098423849287,-.00910645500038529,-.0193820912882233,-.0326278039463224,-.0443672896278861,-.0484616018046971,-.0403121072713726,-.0200426049296945,-1.40524185029455e-008,-4.47875042967689e-006,-9.26542602749126e-005,-.000678420414234276,-.00280265519288241,-.00792374364660721,-.0168648198114988,-.0283902302500237,-.0386050366790666,-.0421675953362895,-.0350765258115668,-.017439548481424,-1.10720890713312e-008,-3.5288675522474e-006,-7.3003534757174e-005,-.000534536546334486,-.00220824962801375,-.00624322392727204,-.0132880177946399,-.0223690433086643,-.030417426340031,-.0332244141027671,-.0276372652876165,-.0137408542244633,-7.3966385507205e-009,-2.35743748168761e-006,-4.87695462026171e-005,-.000357093733613841,-.00147520709262294,-.00417074596164616,-.00887697561402263,-.0149434968427065,-.0203201678410382,-.0221953581292932,-.0184628989656689,-.0091794901054094,-3.26296210833654e-009,-1.03996283213943e-006,-2.15142568085075e-005,-.00015752876308832,-.000650774647452403,-.00183988793598998,-.00391599979728114,-.00659219233567233,-.00896406350610278,-.00979128722597762,-.00814474566006138,-.00404945140720489},
			{-5.21061086814239e-011,-1.66071240047969e-008,-3.43560288548878e-007,-2.51557038586478e-006,-1.03921937740666e-005,-2.93810953272788e-005,-6.25344408726795e-005,-.000105270450249422,-.000143146764126628,-.000156356665933801,-.000130063110895906,-6.46655241827284e-005,-6.14193874705368e-010,-1.95754280991916e-007,-4.04967152908804e-006,-2.96519537053663e-005,-.000122496611669995,-.000346325780965151,-.00073711646319537,-.00124086152980604,-.00168732357751008,-.00184303355050058,-.0015331017429415,-.000762236326271721,-2.20633662270343e-009,-7.03197894004764e-007,-1.45474238225389e-005,-.00010651716679239,-.000440038189267583,-.00124408804027283,-.00264790502628461,-.00445748215614846,-.00606128448480099,-.00662063329969093,-.00550728143194998,-.00273814180028246,-5.02223790632381e-009,-1.6006746579724e-006,-3.31139965720151e-005,-.00024246279884684,-.0010016497263513,-.00283189158461527,-.00602737082751107,-.0101464733989415,-.013797175094379,-.0150704091023226,-.0125361095327521,-.00623277490876185,-8.84165345126754e-009,-2.81798888822008e-006,-5.82972148148455e-005,-.000426855931999716,-.00176340506466325,-.00498554717433864,-.0106111906829482,-.0178628737266676,-.0242899366908432,-.0265314660788539,-.0220698298613373,-.0109728047159291,-1.30549605302744e-008,-4.16084320803041e-006,-8.60775467651508e-005,-.000630265297671321,-.00260371927546702,-.00736130656347784,-.0156677341301185,-.0263750567407152,-.0358648036285977,-.039174487484589,-.0325867508082808,-.0162016678512013,-1.67929112501503e-008,-5.3521931802253e-006,-.00011072362877741,-.00081072548502156,-.00334922703227079,-.0094690265450411,-.0201537850786976,-.0339268729336555,-.0461337637093912,-.0503910900437627,-.0419171251406854,-.0208405969285544,-1.91308046425978e-008,-6.09732050715256e-006,-.000126138468780461,-.000923593928514568,-.00381550328609678,-.0107872955612227,-.0229595761810517,-.0386501404407882,-.0525564631293,-.0574064904526992,-.0477527880842341,-.0237420053340501,-1.93172649069964e-008,-6.15674863969202e-006,-.000127367889742049,-.000932595827353905,-.00385269146321899,-.0108924349957666,-.0231833539428959,-.0390268478266128,-.0530687098535765,-.0579660084441261,-.0482182153183887,-.0239734091183992,-1.69710841924047e-008,-5.40897999891501e-006,-.00011189840749341,-.000819327289847131,-.00338476236177552,-.00956949300345818,-.020367616923072,-.0342868373663876,-.0466232433651179,-.0509257399709112,-.0423618662225568,-.0210617158529497,-1.21971598987849e-008,-3.88744721245458e-006,-8.04216603455029e-005,-.00058885253590198,-.00243263702412373,-.00687761812917763,-.0146382563159452,-.0246420342413141,-.0335082395371539,-.0366004544168802,-.0304455773166633,-.0151371069219119,-5.5995100454324e-009,-1.78466133902173e-006,-3.69202255862743e-005,-.000270332250902931,-.00111678255975218,-.00315739828964731,-.00672017616961103,-.0113127416069928,-.0153830666688022,-.0168026502788677,-.0139770501853164,-.00694919005501226},
			{-6.20196054366658e-013,-1.97667280147994e-010,-4.08924674643852e-009,-2.99417259756176e-008,-1.23693703828388e-007,-3.49710232755209e-007,-7.44319897852733e-007,-1.25298778853868e-006,-1.70381286481175e-006,-1.86104450591298e-006,-1.54808390489254e-006,-7.69685243564878e-007,-3.80137148576797e-011,-1.21156327444106e-008,-2.50642452023579e-007,-1.83522004948295e-006,-7.58156578700641e-006,-2.14348107782544e-005,-4.56216452211473e-005,-7.67994574913963e-005,-.000104431906584671,-.000114069115221109,-9.48868018782288e-005,-4.71763649140235e-005,-3.27700968053241e-010,-1.04444003796667e-007,-2.16068791147813e-006,-1.58206949533316e-005,-6.53576336083714e-005,-.00018478089469472,-.000393285880086281,-.000662057277488362,-.000900265522892682,-.000983344027882673,-.00081798100889079,-.000406688494124778,-1.33784942253663e-009,-4.263959029382e-007,-8.82107578694501e-006,-6.45884805686753e-005,-.000266824577604282,-.000754373765606237,-.00160560187139806,-.00270286948383404,-.00367536207502424,-.00401453266272909,-.00333943298029159,-.00166031846121592,-3.60795174491446e-009,-1.14991703559086e-006,-2.37889371116125e-005,-.000174184117617127,-.000719580383364912,-.00203441740011081,-.00433003443882712,-.00728917806907233,-.00991182474529423,-.0108265099805079,-.00900588126384461,-.00447759575057386,-7.37160626866756e-009,-2.34945925758498e-006,-4.86044964941899e-005,-.000355885229102249,-.0014702145815281,-.00415663099731916,-.00884693345962729,-.0148929238932162,-.0202513987414977,-.0221202428642781,-.0184004153805573,-.00914842415783806,-1.21972798977825e-008,-3.8874854582243e-006,-8.04224515558099e-005,-.000588858329194405,-.00243266095707262,-.00687768579306728,-.0146384003311175,-.0246422766763918,-.0335085692003862,-.0366008145021687,-.0304458768485863,-.0151372558449131,-1.68911888154202e-008,-5.3835159512905e-006,-.000111371619378443,-.000815470113608189,-.00336882779555505,-.00952444236070643,-.0202717315680825,-.0341254240019886,-.0464037534631961,-.05068599505261,-.0421624377574542,-.0209625628637279,-1.97926499253432e-008,-6.30826212149848e-006,-.000130502328643898,-.00095554638928031,-.00394750363309629,-.0111604905634277,-.023753880836457,-.0399872725481971,-.0543746954436094,-.0593925132894224,-.0494048334697752,-.0245633787434188,-1.93889081179464e-008,-6.179582578331e-006,-.000127840267412333,-.000936054606840123,-.00386698019345867,-.0109328324858832,-.0232693355725446,-.0391715892641307,-.0532655292683946,-.0581809907923459,-.0483970453850838,-.0240623208776429,-1.49915585521718e-008,-4.77807071380575e-006,-9.88464560551176e-005,-.000723760067411173,-.00298996001413125,-.00845329697557845,-.0179919160266198,-.030287583522627,-.0411850577650857,-.0449857065070876,-.0374207322675915,-.0186050545055932,-7.16228788345407e-009,-2.28274584398164e-006,-4.72243611004261e-005,-.000345779789560036,-.00142846751434807,-.00403860254373974,-.00859572283356178,-.0144700360358322,-.0196763558094162,-.021492133691307,-.0178779315290952,-.00888865263692734},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.81377664478188e-011,-6.32055938620332e-008,-3.13787157581542e-006,-4.12071599264361e-005,-.000260771859326372,-.00102018939735755,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404,-8.64525128335375e-011,-1.43277462826387e-007,-7.11307734943901e-006,-9.34103607574427e-005,-.000591130121526243,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-1.29411477385236e-010,-2.14473212317988e-007,-1.06476239773248e-005,-.000139826737158981,-.000884867539951351,-.00346177108471816,-.00947763686126095,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-1.64245810125438e-010,-2.72204036451189e-007,-1.35136980227866e-005,-.000177464597313172,-.00112305174849657,-.00439359327137407,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.88761417029128e-010,-3.12833670474357e-007,-1.55307754038705e-005,-.000203953262708829,-.00129068034844452,-.00504938842044633,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-2.01416927810585e-010,-3.33807606524435e-007,-1.65720363705528e-005,-.000217627310910781,-.00137721402318714,-.00538792470927311,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-2.01416927810585e-010,-3.33807606524435e-007,-1.65720363705528e-005,-.000217627310910781,-.00137721402318714,-.00538792470927311,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-1.88761417029128e-010,-3.12833670474357e-007,-1.55307754038705e-005,-.000203953262708829,-.00129068034844452,-.00504938842044633,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-1.64245810125438e-010,-2.72204036451189e-007,-1.35136980227866e-005,-.000177464597313172,-.00112305174849657,-.00439359327137407,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.29411477385236e-010,-2.14473212317988e-007,-1.06476239773248e-005,-.000139826737158981,-.000884867539951351,-.00346177108471816,-.00947763686126095,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-8.64525128335375e-011,-1.43277462826387e-007,-7.11307734943901e-006,-9.34103607574427e-005,-.000591130121526243,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-3.81377664478188e-011,-6.32055938620331e-008,-3.13787157581542e-006,-4.12071599264361e-005,-.000260771859326372,-.00102018939735755,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404},
			{-6.09020435343665e-013,-1.00932754786986e-009,-5.01085430834115e-008,-6.58035454488694e-007,-4.16425517497517e-006,-1.62913628348369e-005,-4.46024931010705e-005,-9.12717393010351e-005,-.000144026600260793,-.000175413507334088,-.000156980270065097,-8.12226984453021e-005,-7.17874795152048e-012,-1.18973151739896e-008,-5.90647834026672e-007,-7.75650601653937e-006,-4.90856079239649e-005,-.000192032287901912,-.000525745997014592,-.0010758535732373,-.00169769452974626,-.00206766355177068,-.00185038420184216,-.000957401831273779,-2.57878418572552e-011,-4.27381047927422e-008,-2.12175340882166e-006,-2.78632920211386e-005,-.000176327669275857,-.000689827572348814,-.00188860992469107,-.00386473267978856,-.00609853951561999,-.00742755993763223,-.00664703866113062,-.00343922466500658,-5.87003250388414e-011,-9.72838540256606e-008,-4.82970290571489e-006,-6.34246288365862e-005,-.000401370985487183,-.00157024007444182,-.00429900326923858,-.00879721016390503,-.0138819779418027,-.016907199330518,-.0151305150742709,-.00782863516982087,-1.03341964508827e-010,-1.71268295079276e-007,-8.50269544402536e-006,-.000111659104747358,-.000706613909031054,-.00276440878199448,-.0075683983517811,-.0154874948296011,-.0244392321648105,-.0297651365985099,-.0266372826857591,-.0137823178549157,-1.52587440258738e-010,-2.52882658732434e-007,-1.25544790953984e-005,-.00016486794165333,-.00104333615238232,-.00408173061038807,-.0111749620480468,-.0228677401605563,-.0360852427727842,-.0439490968079129,-.0393307288068947,-.0203499963670061,-1.96276912228913e-010,-3.25289075746179e-007,-1.6149129884536e-005,-.000212073617975246,-.00134206850877865,-.0052504287338376,-.0143746237655383,-.029415326849682,-.0464173199082629,-.0565327854119549,-.0505920670327779,-.0261766921577074,-2.23602400308645e-010,-3.7057551652431e-007,-1.83973966375979e-005,-.000241598308649254,-.0015289100309035,-.00598138850975171,-.0163758454369784,-.0335105011321007,-.0528794957568774,-.0644032269037527,-.0576354472700868,-.0298209867484436,-2.25781763041192e-010,-3.74187367153787e-007,-1.85767086689201e-005,-.000243953070268046,-.00154381170252316,-.00603968669970425,-.0165354542211866,-.033837114519146,-.0533948909503518,-.0650309392734592,-.0581971968115976,-.0301116399215653,-1.98359412066345e-010,-3.28740395820796e-007,-1.63204723006016e-005,-.000214323721005407,-.00135630786795552,-.00530613583081857,-.0145271386554198,-.0297274237371193,-.046909808097858,-.0571325987832424,-.0511288493276215,-.026454427101384,-1.42561396724728e-010,-2.36266530031746e-007,-1.17295635339088e-005,-.000154034984775778,-.000974781796488713,-.00381353285620554,-.0104406902377677,-.0213651724656962,-.0337141943145245,-.0410613390925373,-.0367464296104888,-.0190128617434316,-6.54475287423684e-011,-1.08465972348532e-007,-5.38484445409324e-006,-7.0714859176842e-005,-.000447505854382295,-.00175072850680893,-.00479314449861828,-.00980838972652806,-.0154776170276039,-.0188505670693449,-.0168696650240853,-.0087284836148921},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-4.25830680908177e-013,-3.66970948242474e-009,-4.37202313877829e-007,-1.02973144060446e-005,-9.98226951437858e-005,-.000540391748990977,-.00190309670559272,-.00473399758463878,-.00866893492130421,-.0117725765312941,-.011334382523687,-.00610300723148638,-9.65293351840578e-013,-8.31867295004696e-009,-9.91071113277978e-007,-2.33424447406373e-005,-.000226282859139225,-.00122498585958623,-.00431403062339345,-.0107312521172389,-.0196511562511815,-.026686639477119,-.0256933203450701,-.0138345886544025,-1.44495555626435e-012,-1.24522899458459e-008,-1.48354249933819e-006,-3.49414974841223e-005,-.000338724672636725,-.00183369140663751,-.00645770791571394,-.0160636995405713,-.0294159773896932,-.0399474500854692,-.0384605425083467,-.0207091095227104,-1.83390144930814e-012,-1.58041349298872e-008,-1.88287503228147e-006,-4.43468746144642e-005,-.000429900881983438,-.00232727526714406,-.00819596135984531,-.0203876456552513,-.0373340227205607,-.0507003044420999,-.0488131585372653,-.0262834838088483,-2.10763267567556e-012,-1.81630867905044e-008,-2.16391613832206e-006,-5.09661639870523e-005,-.000494068613398971,-.00267464830248913,-.00941930438906687,-.0234307400647473,-.0429065619801723,-.0582679174767988,-.0560990929882762,-.030206600974669,-2.24893892602145e-012,-1.93808310960915e-008,-2.30899591389123e-006,-5.43831908772887e-005,-.0005271934477113,-.00285397012027107,-.0100508217305121,-.0250016542281072,-.0457832327865349,-.0621744904907474,-.0598602571462754,-.0322318027893339,-2.24893892602145e-012,-1.93808310960915e-008,-2.30899591389123e-006,-5.43831908772887e-005,-.0005271934477113,-.00285397012027107,-.0100508217305121,-.0250016542281072,-.0457832327865349,-.0621744904907474,-.0598602571462754,-.0322318027893339,-2.10763267567556e-012,-1.81630867905044e-008,-2.16391613832206e-006,-5.09661639870523e-005,-.000494068613398971,-.00267464830248913,-.00941930438906688,-.0234307400647473,-.0429065619801723,-.0582679174767988,-.0560990929882762,-.030206600974669,-1.83390144930814e-012,-1.58041349298872e-008,-1.88287503228147e-006,-4.43468746144641e-005,-.000429900881983438,-.00232727526714406,-.00819596135984532,-.0203876456552513,-.0373340227205607,-.0507003044420999,-.0488131585372653,-.0262834838088483,-1.44495555626435e-012,-1.24522899458459e-008,-1.48354249933819e-006,-3.49414974841223e-005,-.000338724672636725,-.00183369140663751,-.00645770791571394,-.0160636995405713,-.0294159773896932,-.0399474500854692,-.0384605425083467,-.0207091095227104,-9.65293351840578e-013,-8.31867295004696e-009,-9.91071113277978e-007,-2.33424447406373e-005,-.000226282859139225,-.00122498585958623,-.00431403062339345,-.0107312521172389,-.0196511562511815,-.026686639477119,-.0256933203450701,-.0138345886544025,-4.25830680908177e-013,-3.66970948242474e-009,-4.37202313877829e-007,-1.02973144060446e-005,-9.98226951437858e-005,-.000540391748990977,-.00190309670559272,-.00473399758463878,-.0086689349213042,-.0117725765312941,-.011334382523687,-.00610300723148638}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{-1.25649763607893e-005,-2.84828893052902e-005,-4.2636271224355e-005,-5.41128889760362e-005,-6.2189870139491e-005,-6.63593905024702e-005,-6.63593905024702e-005,-6.2189870139491e-005,-5.41128889760362e-005,-4.2636271224355e-005,-2.84828893052902e-005,-1.25649763607893e-005,-.000148107999463138,-.000335738296102257,-.000502569416311285,-.000637848532403299,-.000733054881183536,-.000782202551815753,-.000782202551815753,-.000733054881183536,-.000637848532403299,-.000502569416311285,-.000335738296102257,-.000148107999463138,-.000532040641869991,-.00120605517059218,-.00180535390260964,-.00229131001583844,-.0026333148170561,-.00280986543096192,-.00280986543096192,-.0026333148170561,-.00229131001583844,-.00180535390260964,-.00120605517059218,-.000532040641869991,-.00121107298487856,-.00274531815886015,-.00410948932756509,-.00521566106380589,-.00599415943941434,-.00639603790909957,-.00639603790909957,-.00599415943941434,-.00521566106380588,-.00410948932756509,-.00274531815886015,-.00121107298487856,-.00213209486213416,-.00483313459594366,-.00723475892097063,-.00918217505932459,-.010552722010284,-.0112602293456112,-.0112602293456112,-.010552722010284,-.00918217505932459,-.00723475892097063,-.00483313459594367,-.00213209486213416,-.00314810057025839,-.0071362649232201,-.0106823336471961,-.0135577506675935,-.0155814034208121,-.0166260587433132,-.0166260587433132,-.0155814034208121,-.0135577506675935,-.0106823336471961,-.0071362649232201,-.00314810057025839,-.00404947784869215,-.00917955004423722,-.0137409439473876,-.0174396623555152,-.0200427358008248,-.0213865011900097,-.0213865011900097,-.0200427358008248,-.0174396623555152,-.0137409439473876,-.00917955004423722,-.00404947784869215,-.00461324236601103,-.0104575184128172,-.0156539453074288,-.0198675958317378,-.0228330667266135,-.0243639098760258,-.0243639098760258,-.0228330667266135,-.0198675958317378,-.0156539453074288,-.0104575184128172,-.00461324236601103,-.00465820578534291,-.0105594436420292,-.0158065180211977,-.0200612373037489,-.023055611451648,-.0246013751140091,-.0246013751140091,-.023055611451648,-.0200612373037489,-.0158065180211977,-.0105594436420292,-.00465820578534291,-.00409244284577621,-.00927694515432801,-.0138867354885914,-.0176246973329244,-.0202553894113433,-.0216134121636217,-.0216134121636217,-.0202553894113433,-.0176246973329244,-.0138867354885914,-.00927694515432801,-.00409244284577621,-.00294124872640195,-.00666736327135993,-.0099804309086089,-.0126669132684861,-.0145575981275773,-.0155336124156416,-.0155336124156416,-.0145575981275773,-.0126669132684861,-.0099804309086089,-.00666736327135993,-.00294124872640195,-.00135027759956182,-.0030608738365598,-.00458184721642184,-.00581516588124543,-.0066831473580057,-.00713121902844807,-.00713121902844807,-.0066831473580057,-.00581516588124543,-.00458184721642184,-.0030608738365598,-.00135027759956182},
			{-1.49555377634105e-007,-3.39019282157687e-007,-5.07480751318181e-007,-6.4408187594673e-007,-7.40218624107564e-007,-7.89846587943962e-007,-7.89846587943962e-007,-7.40218624107564e-007,-6.4408187594673e-007,-5.07480751318181e-007,-3.39019282157687e-007,-1.49555377634105e-007,-9.16670694821036e-006,-2.07795296865571e-005,-3.11050488640572e-005,-3.94777499870446e-005,-4.53702655975518e-005,-4.84121154335211e-005,-4.84121154335211e-005,-4.53702655975518e-005,-3.94777499870446e-005,-3.11050488640572e-005,-2.07795296865571e-005,-9.16670694821036e-006,-7.90224989069185e-005,-.000179131979588681,-.000268144133301817,-.000340321826892032,-.000391118837314757,-.000417341403029572,-.000417341403029572,-.000391118837314757,-.000340321826892032,-.000268144133301817,-.000179131979588681,-7.90224989069185e-005,-.000322611816370485,-.000731311893505376,-.0010947067871833,-.00138937447237002,-.00159675485169678,-.00170380929406662,-.00170380929406662,-.00159675485169678,-.00138937447237002,-.0010947067871833,-.000731311893505376,-.000322611816370485,-.000870029052744199,-.00197222346386832,-.00295223752124433,-.00374690601758642,-.00430617553540284,-.00459488311014376,-.00459488311014376,-.00430617553540284,-.00374690601758643,-.00295223752124433,-.00197222346386832,-.000870029052744198,-.00177760460022009,-.00402955911756787,-.00603188017934985,-.00765551144808981,-.0087981860110801,-.00938806046568492,-.00938806046568492,-.0087981860110801,-.00765551144808981,-.00603188017934986,-.00402955911756787,-.00177760460022009,-.0029412776627297,-.00666742886561148,-.00998052909718346,-.0126670378869689,-.0145577413468253,-.0155337652370252,-.0155337652370252,-.0145577413468253,-.0126670378869689,-.00998052909718346,-.00666742886561148,-.0029412776627297,-.00407317670628986,-.00923327174798222,-.013821360407493,-.0175417249152225,-.0201600323904989,-.0215116619305183,-.0215116619305183,-.0201600323904989,-.0175417249152225,-.013821360407493,-.00923327174798222,-.00407317670628986,-.00477284112519419,-.0108193045125715,-.0161955058952257,-.0205549309837037,-.0236229946837403,-.0252067985596364,-.0252067985596364,-.0236229946837403,-.0205549309837037,-.0161955058952257,-.0108193045125715,-.00477284112519419,-.00467548197876496,-.0105986061434719,-.0158651406916522,-.0201356397307597,-.0231411193105227,-.0246926158479963,-.0246926158479963,-.0231411193105227,-.0201356397307597,-.0158651406916522,-.0105986061434719,-.00467548197876496,-.00361509587945292,-.0081948722230624,-.0122669716195725,-.015568933545552,-.0178927788504699,-.0190924003579595,-.0190924003579595,-.0178927788504699,-.015568933545552,-.0122669716195725,-.00819487222306241,-.00361509587945292,-.00172712912568916,-.00391513889802946,-.00586060361181134,-.00743813151271418,-.00854835957956341,-.00912148441898038,-.00912148441898038,-.00854835957956341,-.00743813151271418,-.00586060361181134,-.00391513889802946,-.00172712912568916},
			{-1.63148105416827e-009,-3.69831927536025e-009,-5.53604453566559e-009,-7.02620925147273e-009,-8.07495310619039e-009,-8.61633840464567e-009,-8.61633840464567e-009,-8.07495310619039e-009,-7.02620925147273e-009,-5.53604453566559e-009,-3.69831927536025e-009,-1.63148105416827e-009,-5.19981416069715e-007,-1.1787187408438e-006,-1.7644337761237e-006,-2.23937521483844e-006,-2.57362813998106e-006,-2.74617706012381e-006,-2.74617706012381e-006,-2.57362813998106e-006,-2.23937521483844e-006,-1.7644337761237e-006,-1.1787187408438e-006,-5.19981416069715e-007,-1.07571284042538e-005,-2.43847731013085e-005,-3.6501767378587e-005,-4.63271301374563e-005,-5.3241995792529e-005,-5.68116058451738e-005,-5.68116058451738e-005,-5.3241995792529e-005,-4.63271301374563e-005,-3.6501767378587e-005,-2.43847731013085e-005,-1.07571284042538e-005,-7.87643815441602e-005,-.000178546866806921,-.000267268273167243,-.000339210207117138,-.000389841295295595,-.000415978208195064,-.000415978208195064,-.000389841295295595,-.000339210207117138,-.000267268273167243,-.000178546866806921,-7.87643815441602e-005,-.000325387323726201,-.000737603546311472,-.00110412481400688,-.0014013275964412,-.00161049211924643,-.00171846757683388,-.00171846757683388,-.00161049211924644,-.0014013275964412,-.00110412481400688,-.000737603546311472,-.000325387323726201,-.000919943967994991,-.00208537298082308,-.00312161196363602,-.0039618718233036,-.00455322750019265,-.00485849867597654,-.00485849867597654,-.00455322750019265,-.0039618718233036,-.00312161196363602,-.00208537298082308,-.000919943967994991,-.00195799989864057,-.00443848780701132,-.00664400889731993,-.00843240990574942,-.00969104564411167,-.0103407818802712,-.0103407818802712,-.00969104564411167,-.00843240990574942,-.00664400889731993,-.00443848780701132,-.00195799989864057,-.00329609616783617,-.00747174842135327,-.0111845216543321,-.0141951151250118,-.0163139019731612,-.0174076676672231,-.0174076676672231,-.0163139019731612,-.0141951151250118,-.0111845216543321,-.00747174842135327,-.00329609616783617,-.00448203175305139,-.0101600839205191,-.0152087131700155,-.0193025183395333,-.022183644813943,-.0236709474657956,-.0236709474657956,-.022183644813943,-.0193025183395333,-.0152087131700155,-.0101600839205191,-.00448203175305139,-.00489564361298881,-.0110976790646466,-.0166122070513835,-.0210837976681447,-.0242308009023486,-.0258553551512484,-.0258553551512484,-.0242308009023486,-.0210837976681447,-.0166122070513835,-.0110976790646466,-.00489564361298881,-.00407237283003069,-.00923144948283444,-.0138186326438083,-.0175382629057834,-.0201560536356863,-.021507416420056,-.021507416420056,-.0201560536356863,-.0175382629057834,-.0138186326438083,-.00923144948283444,-.00407237283003069,-.00202472570360245,-.0045897450527047,-.00687042711223165,-.00871977424071199,-.0100213024648473,-.0106931807723117,-.0106931807723117,-.0100213024648473,-.00871977424071199,-.00687042711223165,-.0045897450527047,-.00202472570360245},
			{-1.70557276538753e-011,-3.86627390961513e-011,-5.7874572088226e-011,-7.34529592765333e-011,-8.44166719834875e-011,-9.00763884617597e-011,-9.00763884617597e-011,-8.44166719834875e-011,-7.34529592765333e-011,-5.7874572088226e-011,-3.86627390961513e-011,-1.70557276538753e-011,-2.82664008807978e-008,-6.40756292912074e-008,-9.59153363989562e-008,-1.21733345825304e-007,-1.39903470536826e-007,-1.49283299887593e-007,-1.49283299887593e-007,-1.39903470536826e-007,-1.21733345825304e-007,-9.59153363989562e-008,-6.40756292912074e-008,-2.82664008807978e-008,-1.40329882934204e-006,-3.18106489584209e-006,-4.76176220142831e-006,-6.04350947999848e-006,-6.94557390780468e-006,-7.41123996847038e-006,-7.41123996847038e-006,-6.94557390780468e-006,-6.04350947999848e-006,-4.76176220142831e-006,-3.18106489584208e-006,-1.40329882934204e-006,-1.84284021471628e-005,-4.17743832824876e-005,-6.2532417858728e-005,-7.93645806216638e-005,-9.12106719107567e-005,-9.73258921709034e-005,-9.73258921709034e-005,-9.12106719107567e-005,-7.93645806216638e-005,-6.2532417858728e-005,-4.17743832824876e-005,-1.84284021471628e-005,-.000116620720789999,-.000264361427000411,-.000395724793999518,-.000502244010271906,-.000577209799147466,-.000615908834952788,-.000615908834952788,-.000577209799147466,-.000502244010271906,-.000395724793999518,-.000264361427000411,-.000116620720789999,-.000456242568387133,-.00103423247275538,-.0015481510932686,-.00196487464364187,-.00225815515010815,-.00240955318100971,-.00240955318100971,-.00225815515010815,-.00196487464364187,-.0015481510932686,-.00103423247275538,-.000456242568387133,-.00124910090182185,-.00283152165958353,-.00423852805667493,-.00537943379114375,-.00618237716052034,-.00659687468888933,-.00659687468888933,-.00618237716052034,-.00537943379114375,-.00423852805667493,-.00283152165958353,-.00124910090182185,-.0025560815986996,-.00579424800656579,-.00867345748885218,-.0110081352954835,-.012651228153929,-.0134994298511859,-.0134994298511859,-.012651228153929,-.0110081352954835,-.00867345748885218,-.00579424800656579,-.0025560815986996,-.00403349103971441,-.00914331038111467,-.0136866964976499,-.0173708128492007,-.0199636097009642,-.021302070080123,-.021302070080123,-.0199636097009642,-.0173708128492007,-.0136866964976499,-.00914331038111467,-.00403349103971441,-.00491248705999991,-.0111358605958299,-.0166693612993899,-.021156336410114,-.0243141668994061,-.025944310422275,-.025944310422275,-.0243141668994061,-.021156336410114,-.0166693612993899,-.0111358605958299,-.00491248705999991,-.00439626090995001,-.00996565447158663,-.0149176701290573,-.0189331337918146,-.0217591253050265,-.0232179660428578,-.0232179660428578,-.0217591253050265,-.0189331337918146,-.0149176701290573,-.00996565447158663,-.00439626090995001,-.00227465638852363,-.00515629988163542,-.00771850769460683,-.00979613690280579,-.0112583248350307,-.012013139317637,-.012013139317637,-.0112583248350307,-.00979613690280579,-.00771850769460683,-.00515629988163542,-.00227465638852363},
			{-1.7384464750039e-013,-3.94079360668192e-013,-5.8990063561595e-013,-7.48687131527224e-013,-8.60437436733519e-013,-9.18125471866621e-013,-9.18125471866621e-013,-8.60437436733519e-013,-7.48687131527224e-013,-5.8990063561595e-013,-3.94079360668192e-013,-1.7384464750039e-013,-1.49815262263485e-009,-3.39608401063939e-009,-5.0836260825345e-009,-6.45201106711068e-009,-7.41504913145304e-009,-7.91219116240981e-009,-7.91219116240981e-009,-7.41504913145304e-009,-6.45201106711068e-009,-5.0836260825345e-009,-3.39608401063939e-009,-1.49815262263485e-009,-1.78487097219834e-007,-4.04603087706546e-007,-6.05653689159397e-007,-7.68680513053121e-007,-8.83415064139356e-007,-9.42643634493817e-007,-9.42643634493817e-007,-8.83415064139356e-007,-7.68680513053121e-007,-6.05653689159397e-007,-4.04603087706546e-007,-1.78487097219834e-007,-4.20386100245681e-006,-9.52951316020476e-006,-1.42648066135301e-005,-1.81045357479959e-005,-2.0806849318325e-005,-2.22018447046734e-005,-2.22018447046734e-005,-2.0806849318325e-005,-1.81045357479959e-005,-1.42648066135301e-005,-9.52951316020476e-006,-4.20386100245681e-006,-4.07524446402028e-005,-9.23795904009122e-005,-.00013828376853596,-.000175506300131138,-.000201702666783354,-.000215225823762682,-.000215225823762682,-.000201702666783354,-.000175506300131138,-.00013828376853596,-9.23795904009122e-005,-4.07524446402028e-005,-.000220614007696907,-.000500098382997093,-.000748601381966019,-.000950106149209752,-.00109192059703656,-.00116512842258577,-.00116512842258577,-.00109192059703656,-.000950106149209752,-.00074860138196602,-.000500098382997093,-.000220614007696907,-.000776935976612406,-.0017611956269338,-.0026363482167924,-.00334598721372161,-.00384541491403083,-.00410323078906309,-.00410323078906309,-.00384541491403083,-.00334598721372161,-.0026363482167924,-.0017611956269338,-.000776935976612406,-.00193264642090617,-.00438101533121211,-.00655797787568285,-.00832322148498332,-.00956555957532708,-.0102068825969568,-.0102068825969568,-.00956555957532708,-.00832322148498332,-.00655797787568285,-.00438101533121211,-.00193264642090617,-.00353907786161354,-.00802255094483993,-.0120090224811522,-.015241550951156,-.0175165305773379,-.0186909265162125,-.0186909265162125,-.0175165305773379,-.015241550951156,-.0120090224811522,-.00802255094483993,-.00353907786161354,-.00480613424305024,-.0108947749442933,-.0163084782050857,-.0206983126135993,-.0237877776977751,-.0253826294522236,-.0253826294522236,-.0237877776977751,-.0206983126135993,-.0163084782050857,-.0108947749442933,-.00480613424305024,-.00462724228856079,-.0104892541067666,-.0157014507286737,-.0199278885241277,-.0229023588080574,-.0244378476456489,-.0244378476456489,-.0229023588080574,-.0199278885241277,-.0157014507286737,-.0104892541067666,-.00462724228856079,-.00249154226883633,-.00564794716718901,-.00845445855931378,-.010730187331938,-.0123317932081052,-.0131585783867511,-.0131585783867511,-.0123317932081052,-.010730187331938,-.00845445855931378,-.00564794716718901,-.00249154226883633},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.77695599831757e-005,-.000209456341570312,-.000752419091586911,-.00171271584051363,-.00301523747077974,-.00445208652288796,-.0057268264950684,-.00652410992157336,-.00658769779901337,-.00578758817666173,-.00415955383985738,-.0019095808945752,-4.02808883575737e-005,-.000474805651832001,-.00170561957949523,-.00388246617391187,-.00683508449545443,-.010092202640724,-.012981844171125,-.0147891643705445,-.0149333084120673,-.0131195816567463,-.0094290755651378,-.00432872929307012,-6.02967930241717e-005,-.000710740484695346,-.00255315597436325,-.0058116955424022,-.0102314941881775,-.0151071011240822,-.0194326292933208,-.0221380217619632,-.0223537921630579,-.0196388096682036,-.0141144607515462,-.00647971047522447,-7.65271815013742e-005,-.000902054045409203,-.00324040170051977,-.00737605861435863,-.0129855565030636,-.0191735548724589,-.0246634070303329,-.0280970234814972,-.0283708738774972,-.0249250860049403,-.0179137205405699,-.00822388645802562,-8.79497578075961e-005,-.00103669615509983,-.00372406952887607,-.00847702157560611,-.0149238025892903,-.0220354320420534,-.0283447088011335,-.0322908326405259,-.0326055584089547,-.0286454464212646,-.0205875527108991,-.00945139763454551,-9.38463500544579e-005,-.00110620145747811,-.00397374980154702,-.00904536355795187,-.015924369058549,-.0235127977673766,-.0302450800394712,-.0344557717846379,-.0347915983448399,-.030565980615856,-.0219678453543708,-.0100850666679019,-9.38463500544579e-005,-.00110620145747811,-.00397374980154702,-.00904536355795187,-.015924369058549,-.0235127977673766,-.0302450800394711,-.0344557717846379,-.0347915983448399,-.030565980615856,-.0219678453543708,-.0100850666679019,-8.79497578075961e-005,-.00103669615509983,-.00372406952887607,-.00847702157560611,-.0149238025892903,-.0220354320420534,-.0283447088011335,-.0322908326405259,-.0326055584089547,-.0286454464212646,-.0205875527108991,-.00945139763454551,-7.65271815013742e-005,-.000902054045409203,-.00324040170051977,-.00737605861435863,-.0129855565030636,-.0191735548724589,-.0246634070303329,-.0280970234814972,-.0283708738774972,-.0249250860049403,-.0179137205405699,-.00822388645802562,-6.02967930241717e-005,-.000710740484695346,-.00255315597436325,-.0058116955424022,-.0102314941881775,-.0151071011240822,-.0194326292933208,-.0221380217619632,-.0223537921630579,-.0196388096682036,-.0141144607515462,-.00647971047522447,-4.02808883575737e-005,-.000474805651832001,-.00170561957949523,-.00388246617391187,-.00683508449545443,-.010092202640724,-.012981844171125,-.0147891643705445,-.0149333084120673,-.0131195816567463,-.0094290755651378,-.00432872929307012,-1.77695599831757e-005,-.000209456341570312,-.000752419091586911,-.00171271584051363,-.00301523747077974,-.00445208652288796,-.0057268264950684,-.00652410992157336,-.00658769779901337,-.00578758817666173,-.00415955383985738,-.0019095808945752},
			{-2.83761377805336e-007,-3.34479976602295e-006,-1.20153497508037e-005,-2.73502893236142e-005,-4.815020405282e-005,-7.10951878965703e-005,-9.14514584621857e-005,-.000104183244945397,-.000105198677163819,-9.2421759274531e-005,-6.64237454259384e-005,-3.04940193335369e-005,-3.34479976602295e-006,-3.94263855120624e-005,-.000141629348384198,-.000322387923394712,-.000567564171331005,-.00083802513816711,-.00107797198911481,-.00122804624086627,-.00124001551403817,-.00108940928179785,-.000782961126977123,-.000359444225710958,-1.20153497508037e-005,-.000141629348384198,-.000508767721494418,-.00115809732300615,-.00203883117125292,-.00301039399647408,-.00387234255465894,-.004411446462074,-.00445444305782967,-.00391342814465003,-.00281259042095047,-.00129121274513816,-2.73502893236142e-005,-.000322387923394712,-.00115809732300615,-.00263615271349073,-.00464094875074617,-.00685249688850132,-.00881453236290365,-.0100416833114065,-.0101395555630045,-.00890805463205047,-.00640224074681438,-.00293916056463564,-4.815020405282e-005,-.000567564171331005,-.00203883117125292,-.00464094875074617,-.00817039361825585,-.0120638257076051,-.0155179905734142,-.0176783907020889,-.017850694871518,-.0156826365956051,-.0112711494458756,-.00517439429092323,-7.10951878965703e-005,-.00083802513816711,-.00301039399647408,-.00685249688850132,-.0120638257076051,-.0178125923307157,-.0229127680203357,-.0261026621464626,-.0263570743040411,-.0231558726990041,-.0166421825914966,-.00764014486751864,-9.14514584621857e-005,-.00107797198911481,-.00387234255465894,-.00881453236290365,-.0155179905734142,-.0229127680203357,-.0294732472739763,-.0335764851836737,-.0339037416907514,-.0297859586976462,-.0214072416856188,-.00982770299465891,-.000104183244945397,-.00122804624086627,-.004411446462074,-.0100416833114065,-.0176783907020889,-.0261026621464626,-.0335764851836737,-.038250972036084,-.0386237888878892,-.0339327319991675,-.0243875378440283,-.0111959066105712,-.000105198677163819,-.00124001551403817,-.00445444305782967,-.0101395555630045,-.017850694871518,-.0263570743040411,-.0339037416907514,-.0386237888878892,-.039000239435718,-.0342634607007844,-.02462523337432,-.0113050286127969,-9.2421759274531e-005,-.00108940928179785,-.00391342814465003,-.00890805463205047,-.0156826365956051,-.0231558726990041,-.0297859586976462,-.0339327319991675,-.0342634607007844,-.0301019879923868,-.0216343727160793,-.00993197501349335,-6.64237454259384e-005,-.000782961126977123,-.00281259042095047,-.00640224074681438,-.0112711494458756,-.0166421825914966,-.0214072416856188,-.0243875378440283,-.02462523337432,-.0216343727160793,-.0155486768161827,-.00713813484023199,-3.04940193335369e-005,-.000359444225710958,-.00129121274513816,-.00293916056463564,-.00517439429092323,-.00764014486751864,-.00982770299465891,-.0111959066105712,-.0113050286127969,-.00993197501349335,-.00713813484023199,-.00327699711041027},
			{-3.37748665821831e-009,-3.98116779370311e-008,-1.43013414267412e-007,-3.25538443615425e-007,-5.73110664448679e-007,-8.46214697861095e-007,-1.08850641767933e-006,-1.24004726271932e-006,-1.25213350502815e-006,-1.10005548074599e-006,-7.90612576313682e-007,-3.62956876834475e-007,-2.07016497214321e-007,-2.44017962134497e-006,-8.76572998571608e-006,-1.99532478216857e-005,-3.51277071610772e-005,-5.18670894572505e-005,-6.67178907235581e-005,-7.60062930474401e-005,-7.67470958397088e-005,-6.74257681555381e-005,-4.84590652057195e-005,-2.22467381475193e-005,-1.78460607688856e-006,-2.10358084478815e-005,-7.5565837560655e-005,-.000172008935497438,-.000302821854829854,-.000447125346440885,-.000575148139518477,-.000655219725381629,-.000661605888719569,-.000581250466549428,-.000417746138158903,-.000191780194444643,-7.28570996785344e-006,-8.58793440610684e-005,-.000308499888616729,-.000702231843846736,-.00123627966686315,-.00182540317195597,-.00234806021751757,-.00267495496410619,-.00270102667510736,-.00237297315794306,-.00170546163785514,-.000782948625132055,-1.96483173282813e-005,-.000231601945658244,-.000831971590143711,-.00189379952905124,-.00333403543487275,-.00492280106302518,-.00633231798455615,-.00721389736971712,-.00728420832819645,-.00639950393640167,-.00459933919956203,-.00211147892329134,-4.01445666201286e-005,-.000473198777355153,-.0016998472881189,-.0038693268277857,-.0068119526671359,-.0100580478180592,-.0129379099972572,-.0147391137220966,-.0148827699400637,-.0130751813408912,-.00939716494906272,-.00431407966834182,-6.64243989158946e-005,-.000782968829903678,-.00281261809177179,-.00640230373332618,-.011271260333675,-.0166423463205833,-.0214074522943001,-.0243877777734532,-.0246254756422405,-.0216345855593206,-.0155488297871431,-.00713820506661556,-9.19866620625105e-005,-.00108428063087214,-.00389500475941537,-.00886611786558599,-.0156088068880401,-.0230468609713296,-.029645734277678,-.0337729856666483,-.0341021573849623,-.0299602757900378,-.021532523808133,-.00988521790053479,-.000107787546507207,-.00127053146952515,-.00456406393315995,-.0103890832686787,-.0182899885770756,-.0270057043390071,-.034738090181205,-.0395742945945498,-.039960009339504,-.0351066616363676,-.0252312439580584,-.0115832378335861,-.000105588834409239,-.00124461444104258,-.00447096354344966,-.0101771607988773,-.0179168989163566,-.0264548265171398,-.0340294827249939,-.0387670354712729,-.0391448821859808,-.0343905358485524,-.0247165625951827,-.0113469563151396,-8.16415851719115e-005,-.000962339403243543,-.00345696164727354,-.00786900949156683,-.0138533968774284,-.0204549466275595,-.0263116922143577,-.0299747813866719,-.0302669335343241,-.0265908595098725,-.0191109160505604,-.00877349869072301,-3.90046528002956e-005,-.000459762193746235,-.0016515797496082,-.003759456439442,-.00661852576932447,-.00977244733277217,-.01257053519047,-.014320593338432,-.0144601704064181,-.0127039086840373,-.0091303303785454,-.00419158042503607},
			{-3.68445827925519e-011,-4.34300653798924e-010,-1.56011558760704e-009,-3.55125848055141e-009,-6.25199311274643e-009,-9.2312511197497e-009,-1.18743814217069e-008,-1.35275216933177e-008,-1.36593690107054e-008,-1.20003687174081e-008,-8.62469447627331e-009,-3.95945152473478e-009,-1.174301000065e-008,-1.38419179545726e-007,-4.97235999402952e-007,-1.13184790520788e-006,-1.99261905231337e-006,-2.9421604480658e-006,-3.78457209223235e-006,-4.31145667798826e-006,-4.35347871350321e-006,-3.82472643681277e-006,-2.74884028563086e-006,-1.26194613503531e-006,-2.42933809796949e-007,-2.86355020000406e-006,-1.02865820344585e-005,-2.34151315299593e-005,-4.12223559228577e-005,-6.08660170299574e-005,-7.82934287517701e-005,-8.91933666496246e-005,-9.0062698548559e-005,-7.91241227483076e-005,-5.68667013887132e-005,-2.61065418768862e-005,-1.77877687852613e-006,-2.09670975420156e-005,-7.53190109571436e-005,-.000171447089262513,-.000301832724128607,-.000445664865962291,-.000573269488181101,-.000653079529962769,-.000659444833676081,-.000579351882704146,-.000416381621285402,-.000191153767800758,-7.34839068958039e-006,-8.66181847905149e-005,-.000311153987634738,-.000708273313928056,-.00124691567929258,-.00184110755612247,-.00236826114642384,-.00269796824743438,-.00272426425965937,-.00239338841889015,-.00172013413604819,-.000789684521168945,-2.07755717461149e-005,-.000244889308237957,-.000879703089733841,-.00200244974321189,-.00352531421515612,-.00520522978980597,-.00669561288171989,-.0076277698425577,-.00770211464969917,-.00676665339575356,-.00486321041790597,-.00223261774167844,-4.42185272019898e-005,-.000521220049687516,-.00187235160015967,-.00426199478516835,-.00750324488892071,-.0110787610500387,-.0142508780967575,-.016234872011969,-.0163931067848593,-.0144020800439593,-.0103508103065028,-.00475188214064504,-7.44374492353296e-005,-.000877421602303038,-.00315191585985935,-.00717464014604113,-.0126309591445223,-.0186499814768931,-.0239899219175831,-.0273297764014771,-.027596148748536,-.0242444552089724,-.0174245495155288,-.00799931630469631,-.000101220047626196,-.00119311794379991,-.00428597536221463,-.00975607607116152,-.0171755789499336,-.0253602458535457,-.0326214971629358,-.0371630314766057,-.0375252446090086,-.032967611546778,-.0236939034040642,-.0108774707577442,-.000110560858773534,-.00130322152160786,-.00468149470230067,-.0106563884722811,-.01876057958054,-.0277005457518707,-.0356318814839124,-.0405925186861703,-.0409881576521328,-.0360099360729874,-.0258804295145632,-.0118812679549556,-9.19685077033871e-005,-.00108406663876152,-.0038942360467172,-.00886436806094961,-.0156057263557175,-.0230423124750437,-.0296398834369748,-.0337663202773765,-.03409542703082,-.0299543628718611,-.0215282741793143,-.00988326696773915,-4.57254306619082e-005,-.00053898247521257,-.00193615863497058,-.00440723740391471,-.00775894462383229,-.0114563092049479,-.0147365274132258,-.0167881329718923,-.0169517601570097,-.0148928821041257,-.0107035509528065,-.00491381941200989},
			{-3.85178343554166e-013,-4.54023885618748e-012,-1.63096632460441e-011,-3.71253453125811e-011,-6.53591971617201e-011,-9.65047707354281e-011,-1.24136418981734e-010,-1.41418575087745e-010,-1.42796925104609e-010,-1.25453507524717e-010,-9.01637440362945e-011,-4.13926516217373e-011,-6.38354791449189e-010,-7.5245228000818e-009,-2.70299508117877e-008,-6.15277116719808e-008,-1.08319580712847e-007,-1.5993703651203e-007,-2.05730875518941e-007,-2.34372535522585e-007,-2.36656870434667e-007,-2.07913681993516e-007,-1.49428073991582e-007,-6.85999042669627e-008,-3.16914252834386e-008,-3.73558490210382e-007,-1.34191468136724e-006,-3.05457232158762e-006,-5.37757677216062e-006,-7.94014975773712e-006,-1.02136065356437e-005,-1.16355352814725e-005,-1.17489421676717e-005,-1.03219729945844e-005,-7.41842734727761e-006,-3.40567466501195e-006,-4.16178163573184e-007,-4.90564514068014e-006,-1.76222931839914e-005,-4.01132573852444e-005,-7.06194185176384e-005,-.000104271641780592,-.000134127132921491,-.000152800186874675,-.000154289468887485,-.00013555022303093,-9.74202782733315e-005,-4.47240038949301e-005,-2.6337062228933e-006,-3.10444642827677e-005,-.00011151940991293,-.000253849300234756,-.000446901875894084,-.000659863721515399,-.000848798652964526,-.000966967607276502,-.000976392251931343,-.000857804414450873,-.00061650613987428,-.000283027553294656,-1.03035625519223e-005,-.000121451882844024,-.000436285264394079,-.000993107021958261,-.00174836562742656,-.00258151310547804,-.00332066269912202,-.00378296225321243,-.00381983326596499,-.00335589496079028,-.00241188995212247,-.00110725792950642,-2.82090935115534e-005,-.00033251096434234,-.00119446179503315,-.00271892840056321,-.00478667540745511,-.00706766656937942,-.0090913103237661,-.010356994041021,-.0104579394995808,-.00918776920961406,-.00660327230083207,-.00303145075477106,-5.77253164541879e-005,-.00068042954422791,-.00244427156379506,-.00556384423606199,-.00979515178485467,-.0144628287733888,-.0186038861974662,-.0211939018276931,-.0214004702712195,-.0188012736005075,-.013512521522274,-.00620337034448753,-9.10904201184736e-005,-.00107371629731872,-.00385705505497627,-.00877973374713072,-.0154567275852829,-.0228223114223215,-.0293568907657418,-.0334439296312284,-.033769894172609,-.029668367645793,-.0213227287077879,-.00978890451433557,-.000110941243135535,-.0013077052520396,-.00469760137328079,-.0106930517505446,-.0188251253083289,-.0277958494112289,-.0357544728843004,-.0407321771466212,-.0411291773074706,-.0361338281692569,-.0259694710684948,-.011922145428059,-9.92830402484601e-005,-.00117028572514512,-.00420395637395983,-.00956937796371369,-.0168468967973055,-.024874937019263,-.0319972326801651,-.0364518574766186,-.0368071390818563,-.0323367236121244,-.0232404826956437,-.0106693129707941,-5.13697449717053e-005,-.000605514084724702,-.00217515666585201,-.00495126362270572,-.00871670317383983,-.0128704677825032,-.0165555937697616,-.0188604480446285,-.01904427325192,-.0167312487713743,-.0120247895925313,-.00552037774992995},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.5903751261877e-007,-1.58772021730906e-005,-.000136870983053836,-.000558780057100243,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567,-5.8719862146837e-007,-3.59912011760796e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-8.78982445184828e-007,-5.38755250068202e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-1.11558253342289e-006,-6.83774687490587e-005,-.000589454695127486,-.00240646717698951,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-1.2820962657192e-006,-7.85836051712971e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.3680544205638e-006,-8.38522436364219e-005,-.00072285651418097,-.00295108426386074,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.3680544205638e-006,-8.38522436364219e-005,-.00072285651418097,-.00295108426386074,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.2820962657192e-006,-7.85836051712971e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.11558253342289e-006,-6.83774687490587e-005,-.000589454695127486,-.00240646717698951,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-8.78982445184828e-007,-5.38755250068202e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-5.8719862146837e-007,-3.59912011760796e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-2.5903751261877e-007,-1.58772021730906e-005,-.000136870983053836,-.000558780057100243,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567},
			{-4.13655946284767e-009,-2.53542393256769e-007,-2.1856871401242e-006,-8.92313591757784e-006,-2.40641758792944e-005,-4.91668520822556e-005,-8.13529419075361e-005,-.000112660192597526,-.000132012244784625,-.000129319383418971,-9.99901127316091e-005,-4.77707484775864e-005,-4.87591483748876e-008,-2.98859747651745e-006,-2.57634985121267e-005,-.0001051802861973,-.000283653295149336,-.000579547775714673,-.0009589370588843,-.0013279671418103,-.00155607690124302,-.00152433515352727,-.00117862024866094,-.00056309138885062,-1.75154945664267e-007,-1.07357828440125e-005,-9.25488720048507e-005,-.000377833656408332,-.00101895293817233,-.00208187924827317,-.00344473958308176,-.00477038710314128,-.0055898138948427,-.00547578967002042,-.00423389604809699,-.00202276382802748,-3.98701539258914e-007,-2.44376379372225e-005,-.000210667061584079,-.000860054849279415,-.00231942126065002,-.00473893818806989,-.00784118866248488,-.0108587323850331,-.0127239764517783,-.0124644254937564,-.00963752901773119,-.00460437499342809,-7.01914347023563e-007,-4.30224791892881e-005,-.000370879513648267,-.00151412718159681,-.00408334279989948,-.00834290409323953,-.0138044182877922,-.0191168061846747,-.0224005697075911,-.0219436300590561,-.0169668767769879,-.00810600549215572,-1.03639711130188e-006,-6.35239518067959e-005,-.00054761447492283,-.0022356531730757,-.00602917535482301,-.0123185424813832,-.0203826283040634,-.0282265247763196,-.0330750978875268,-.0324004131004311,-.0250520909767,-.0119687547517608,-1.33314265253001e-006,-8.17123944937675e-005,-.000704409734165886,-.00287577470912415,-.00775547397560826,-.0158456389156716,-.0262186674070122,-.0363084610152352,-.0425452977913825,-.0416774344435578,-.0322251100971785,-.0153956985051803,-1.51874152529913e-006,-9.30883176033672e-005,-.000802476998296003,-.00327613737349354,-.00883518380630904,-.0180516539400008,-.0298688057526838,-.0413632909868219,-.0484684143436297,-.047479727872512,-.0367114597744218,-.0175390732465345,-1.53354408858128e-006,-9.39956120239112e-005,-.000810298419092016,-.00330806856783063,-.00892129679210925,-.0182275961561996,-.0301599249984219,-.0417664423606343,-.0489408164993326,-.0479424936985227,-.037069271618925,-.0177100195447144,-1.34728730829027e-006,-8.25793637481591e-005,-.000711883527900599,-.00290628670514157,-.00783775962556066,-.0160137612897764,-.0264968477084686,-.0366936941193377,-.0429967037908343,-.0421196324049362,-.0325670188106206,-.0155590470074065,-9.68298698098172e-007,-5.93499915831539e-005,-.000511632440253917,-.00208875539431883,-.00563301709645563,-.0115091295769889,-.0190433495379502,-.0263718481021364,-.0309018366364716,-.0302714832768886,-.0234059964205263,-.0111823253052378,-4.44529573439469e-007,-2.72465784513743e-005,-.000234881809580631,-.000958912313193877,-.0025860229823533,-.005283646948578,-.00874248004628156,-.012106869926272,-.0141865111307979,-.0138971265528759,-.0107452975256271,-.00513361662858989},
			{-4.92356447782966e-011,-3.0178033998418e-009,-2.60152710469109e-008,-1.06208155422931e-007,-2.86425283165094e-007,-5.85211378134644e-007,-9.68308224601569e-007,-1.34094463604488e-006,-1.57128358699498e-006,-1.53923164459519e-006,-1.19013825765391e-006,-5.68594172031225e-007,-3.0178033998418e-009,-1.84970409164038e-007,-1.59455560634358e-006,-6.50982299448902e-006,-1.75558824755616e-005,-3.58693969483533e-005,-5.93505754917118e-005,-8.21906019486044e-005,-9.63087814184433e-005,-9.4344219744047e-005,-7.29472173341576e-005,-3.48508775138952e-005,-2.60152710469109e-008,-1.59455560634358e-006,-1.37460234489013e-005,-5.61185694462147e-005,-.000151342211720415,-.000309215664595502,-.000511637473895075,-.00070853216856797,-.000830239323386619,-.000813303626233361,-.000628848662297057,-.000300435417626676,-1.06208155422931e-007,-6.50982299448902e-006,-5.61185694462147e-005,-.000229105810010919,-.000617859299465599,-.00126238259464391,-.00208877594431585,-.00289260467614267,-.00338947793153767,-.00332033741971269,-.00256729427659293,-.00122653696255208,-2.86425283165094e-007,-1.75558824755616e-005,-.000151342211720415,-.000617859299465599,-.00166626116517048,-.00340443057968316,-.00563307251629313,-.00780086152659004,-.00914084396303246,-.00895438379338928,-.00692355485520927,-.0033077610227998,-5.85211378134644e-007,-3.58693969483533e-005,-.000309215664595502,-.00126238259464391,-.00340443057968316,-.00695578089086413,-.0115092428083319,-.0159383727377934,-.0186761651548639,-.018295198043168,-.0141458988313976,-.0067582699589302,-9.68308224601569e-007,-5.93505754917118e-005,-.000511637473895075,-.00208877594431585,-.00563307251629313,-.0115092428083319,-.0190435368939121,-.0263721075587493,-.0309021406608933,-.030271781099646,-.023406226697924,-.0111824353213518,-1.34094463604488e-006,-8.21906019486044e-005,-.00070853216856797,-.00289260467614267,-.00780086152659004,-.0159383727377934,-.0263721075587493,-.036520949914119,-.0427942866834367,-.0419213443176089,-.0324137019011133,-.0154857991299779,-1.57128358699498e-006,-9.63087814184433e-005,-.000830239323386619,-.00338947793153767,-.00914084396303246,-.0186761651548639,-.0309021406608933,-.0427942866834367,-.0501452173903112,-.0491223265304294,-.037981521698904,-.0181458513277659,-1.53923164459519e-006,-9.4344219744047e-005,-.000813303626233361,-.00332033741971269,-.00895438379338928,-.018295198043168,-.030271781099646,-.0419213443176089,-.0491223265304294,-.0481203011840638,-.0372067528692506,-.0177757018611982,-1.19013825765391e-006,-7.29472173341576e-005,-.000628848662297057,-.00256729427659293,-.00692355485520927,-.0141458988313976,-.023406226697924,-.0324137019011133,-.037981521698904,-.0372067528692506,-.0287683664692428,-.0137442229153402,-5.68594172031225e-007,-3.48508775138952e-005,-.000300435417626676,-.00122653696255208,-.0033077610227998,-.0067582699589302,-.0111824353213518,-.0154857991299779,-.0181458513277659,-.0177757018611982,-.0137442229153402,-.00656636739345367},
			{-5.37105538571509e-013,-3.2920842768966e-011,-2.83797363265043e-010,-1.15861158670691e-009,-3.1245778676732e-009,-6.38399829730143e-009,-1.05631542517567e-008,-1.46281986187158e-008,-1.71409376487648e-008,-1.67912869868831e-008,-1.29830705524454e-008,-6.20272325817235e-009,-1.71184886157124e-010,-1.04924457427685e-008,-9.045116058834e-008,-3.69269684126226e-007,-9.95857365368599e-007,-2.03469140284302e-006,-3.36666116468723e-006,-4.66226157691217e-006,-5.46311525260784e-006,-5.35167549924359e-006,-4.13793062049084e-006,-1.9769158918718e-006,-3.54139156583407e-009,-2.1706273078516e-007,-1.87121062155825e-006,-7.6392757225221e-006,-2.06018238739433e-005,-4.20927287149029e-005,-6.96478861031055e-005,-9.64506516716335e-005,-.000113018331892963,-.000110712919239298,-8.56035420437903e-005,-4.08974964028785e-005,-2.59302953359118e-008,-1.58934718486943e-006,-1.37011237392702e-005,-5.59352649812969e-005,-.000150847870838019,-.000308205649327789,-.000509966272467282,-.000706217834626139,-.000827527447862548,-.00081064706909918,-.000626794604813159,-.000299454081965167,-1.07121889836103e-007,-6.56582857400443e-006,-5.6601371053274e-005,-.000231076862629534,-.000623174892248206,-.0012732431770043,-.00210674618825924,-.00291749045281179,-.00341863842883094,-.00334890308448347,-.00258938132932369,-.00123708915628192,-3.02857945621352e-007,-1.85630906649181e-005,-.000160024948988241,-.000653306845161927,-.0017618571509321,-.00359974803892942,-.0059562506178524,-.00824840904375729,-.00966526834956002,-.00946811067096992,-.00732076992880829,-.00349753239972921,-6.44599940278459e-007,-3.95095037359528e-005,-.000340595563207873,-.00139049200941683,-.00374991982442457,-.00766166912395449,-.0126772265613641,-.0175558345219898,-.0205714642490873,-.0201518357411283,-.0155814563465409,-.0074441143400151,-1.08511925583571e-006,-6.65102812046103e-005,-.000573358421239403,-.00234075363682485,-.00631261338244207,-.0128976504320703,-.0213408376153209,-.0295534841096122,-.0346299938653073,-.0339235914196489,-.0262298167578035,-.0125314188045125,-1.47554522466982e-006,-9.04406840954337e-005,-.00077965281321311,-.00318295692604349,-.0085838920298908,-.0175382257776313,-.0290192721808159,-.040186829342282,-.047089867591508,-.0461293015064619,-.0356673062917253,-.0170402239900308,-1.61171182018355e-006,-9.87867515986945e-005,-.000851600909064724,-.00347668727130158,-.00937603268029293,-.0191566922641638,-.0316972351677059,-.0438953593449828,-.0514354253188647,-.0503862160588214,-.0389587646541599,-.0186127337638557,-1.34067998923389e-006,-8.2174380935362e-005,-.000708392333727775,-.00289203379604635,-.00779932196032425,-.0159352271645878,-.0263669027993754,-.0365137421945241,-.0427858408675583,-.0419130707843193,-.0324073047817918,-.015482742876017,-6.66566974047622e-007,-4.08559304861565e-005,-.000352202567445854,-.00143787796622194,-.0038777116690518,-.00792276772768902,-.0131092481092602,-.0181541119739297,-.0212725100010394,-.0208385811603797,-.0161124498455309,-.0076977989913323},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.26296210833654e-009,-1.03996283213943e-006,-2.15142568085075e-005,-.00015752876308832,-.000650774647452403,-.00183988793598998,-.00391599979728114,-.00659219233567233,-.00896406350610278,-.00979128722597762,-.00814474566006138,-.00404945140720489,-7.3966385507205e-009,-2.35743748168761e-006,-4.87695462026171e-005,-.000357093733613841,-.00147520709262294,-.00417074596164616,-.00887697561402263,-.0149434968427065,-.0203201678410382,-.0221953581292932,-.0184628989656689,-.0091794901054094,-1.10720890713312e-008,-3.5288675522474e-006,-7.3003534757174e-005,-.000534536546334486,-.00220824962801375,-.00624322392727204,-.0132880177946399,-.0223690433086643,-.030417426340031,-.0332244141027671,-.0276372652876165,-.0137408542244633,-1.40524185029455e-008,-4.47875042967689e-006,-9.26542602749126e-005,-.000678420414234276,-.00280265519288241,-.00792374364660721,-.0168648198114988,-.0283902302500237,-.0386050366790666,-.0421675953362895,-.0350765258115668,-.017439548481424,-1.61499062123808e-008,-5.14725627996212e-006,-.000106483991585058,-.000779682590591189,-.00322098423849287,-.00910645500038529,-.0193820912882233,-.0326278039463224,-.0443672896278861,-.0484616018046971,-.0403121072713726,-.0200426049296945,-1.72326768092913e-008,-5.49235412024762e-006,-.000113623211690348,-.000831956416390127,-.00343693515366777,-.00971699735195308,-.0206815637605425,-.0348153353344463,-.0473418949315912,-.0517107103024967,-.043014832840112,-.0213863615446234,-1.72326768092913e-008,-5.49235412024762e-006,-.000113623211690348,-.000831956416390127,-.00343693515366777,-.00971699735195308,-.0206815637605425,-.0348153353344463,-.0473418949315912,-.0517107103024967,-.043014832840112,-.0213863615446234,-1.61499062123808e-008,-5.14725627996212e-006,-.000106483991585058,-.000779682590591189,-.00322098423849287,-.00910645500038529,-.0193820912882233,-.0326278039463224,-.0443672896278861,-.0484616018046971,-.0403121072713726,-.0200426049296945,-1.40524185029455e-008,-4.47875042967689e-006,-9.26542602749126e-005,-.000678420414234276,-.00280265519288241,-.00792374364660721,-.0168648198114988,-.0283902302500237,-.0386050366790666,-.0421675953362895,-.0350765258115668,-.017439548481424,-1.10720890713312e-008,-3.5288675522474e-006,-7.3003534757174e-005,-.000534536546334486,-.00220824962801375,-.00624322392727204,-.0132880177946399,-.0223690433086643,-.030417426340031,-.0332244141027671,-.0276372652876165,-.0137408542244633,-7.3966385507205e-009,-2.35743748168761e-006,-4.87695462026171e-005,-.000357093733613841,-.00147520709262294,-.00417074596164616,-.00887697561402263,-.0149434968427065,-.0203201678410382,-.0221953581292932,-.0184628989656689,-.0091794901054094,-3.26296210833654e-009,-1.03996283213943e-006,-2.15142568085075e-005,-.00015752876308832,-.000650774647452403,-.00183988793598998,-.00391599979728114,-.00659219233567233,-.00896406350610278,-.00979128722597762,-.00814474566006138,-.00404945140720489},
			{-5.21061086814239e-011,-1.66071240047969e-008,-3.43560288548878e-007,-2.51557038586478e-006,-1.03921937740666e-005,-2.93810953272788e-005,-6.25344408726795e-005,-.000105270450249422,-.000143146764126628,-.000156356665933801,-.000130063110895906,-6.46655241827284e-005,-6.14193874705368e-010,-1.95754280991916e-007,-4.04967152908804e-006,-2.96519537053663e-005,-.000122496611669995,-.000346325780965151,-.00073711646319537,-.00124086152980604,-.00168732357751008,-.00184303355050058,-.0015331017429415,-.000762236326271721,-2.20633662270343e-009,-7.03197894004764e-007,-1.45474238225389e-005,-.00010651716679239,-.000440038189267583,-.00124408804027283,-.00264790502628461,-.00445748215614846,-.00606128448480099,-.00662063329969093,-.00550728143194998,-.00273814180028246,-5.02223790632381e-009,-1.6006746579724e-006,-3.31139965720151e-005,-.00024246279884684,-.0010016497263513,-.00283189158461527,-.00602737082751107,-.0101464733989415,-.013797175094379,-.0150704091023226,-.0125361095327521,-.00623277490876185,-8.84165345126754e-009,-2.81798888822008e-006,-5.82972148148455e-005,-.000426855931999716,-.00176340506466325,-.00498554717433864,-.0106111906829482,-.0178628737266676,-.0242899366908432,-.0265314660788539,-.0220698298613373,-.0109728047159291,-1.30549605302744e-008,-4.16084320803041e-006,-8.60775467651508e-005,-.000630265297671321,-.00260371927546702,-.00736130656347784,-.0156677341301185,-.0263750567407152,-.0358648036285977,-.039174487484589,-.0325867508082808,-.0162016678512013,-1.67929112501503e-008,-5.3521931802253e-006,-.00011072362877741,-.00081072548502156,-.00334922703227079,-.0094690265450411,-.0201537850786976,-.0339268729336555,-.0461337637093912,-.0503910900437627,-.0419171251406854,-.0208405969285544,-1.91308046425978e-008,-6.09732050715256e-006,-.000126138468780461,-.000923593928514568,-.00381550328609678,-.0107872955612227,-.0229595761810517,-.0386501404407882,-.0525564631293,-.0574064904526992,-.0477527880842341,-.0237420053340501,-1.93172649069964e-008,-6.15674863969202e-006,-.000127367889742049,-.000932595827353905,-.00385269146321899,-.0108924349957666,-.0231833539428959,-.0390268478266128,-.0530687098535765,-.0579660084441261,-.0482182153183887,-.0239734091183992,-1.69710841924047e-008,-5.40897999891501e-006,-.00011189840749341,-.000819327289847131,-.00338476236177552,-.00956949300345818,-.020367616923072,-.0342868373663876,-.0466232433651179,-.0509257399709112,-.0423618662225568,-.0210617158529497,-1.21971598987849e-008,-3.88744721245458e-006,-8.04216603455029e-005,-.00058885253590198,-.00243263702412373,-.00687761812917763,-.0146382563159452,-.0246420342413141,-.0335082395371539,-.0366004544168802,-.0304455773166633,-.0151371069219119,-5.5995100454324e-009,-1.78466133902173e-006,-3.69202255862743e-005,-.000270332250902931,-.00111678255975218,-.00315739828964731,-.00672017616961103,-.0113127416069928,-.0153830666688022,-.0168026502788677,-.0139770501853164,-.00694919005501226},
			{-6.20196054366658e-013,-1.97667280147994e-010,-4.08924674643852e-009,-2.99417259756176e-008,-1.23693703828388e-007,-3.49710232755209e-007,-7.44319897852733e-007,-1.25298778853868e-006,-1.70381286481175e-006,-1.86104450591298e-006,-1.54808390489254e-006,-7.69685243564878e-007,-3.80137148576797e-011,-1.21156327444106e-008,-2.50642452023579e-007,-1.83522004948295e-006,-7.58156578700641e-006,-2.14348107782544e-005,-4.56216452211473e-005,-7.67994574913963e-005,-.000104431906584671,-.000114069115221109,-9.48868018782288e-005,-4.71763649140235e-005,-3.27700968053241e-010,-1.04444003796667e-007,-2.16068791147813e-006,-1.58206949533316e-005,-6.53576336083714e-005,-.00018478089469472,-.000393285880086281,-.000662057277488362,-.000900265522892682,-.000983344027882673,-.00081798100889079,-.000406688494124778,-1.33784942253663e-009,-4.263959029382e-007,-8.82107578694501e-006,-6.45884805686753e-005,-.000266824577604282,-.000754373765606237,-.00160560187139806,-.00270286948383404,-.00367536207502424,-.00401453266272909,-.00333943298029159,-.00166031846121592,-3.60795174491446e-009,-1.14991703559086e-006,-2.37889371116125e-005,-.000174184117617127,-.000719580383364912,-.00203441740011081,-.00433003443882712,-.00728917806907233,-.00991182474529423,-.0108265099805079,-.00900588126384461,-.00447759575057386,-7.37160626866756e-009,-2.34945925758498e-006,-4.86044964941899e-005,-.000355885229102249,-.0014702145815281,-.00415663099731916,-.00884693345962729,-.0148929238932162,-.0202513987414977,-.0221202428642781,-.0184004153805573,-.00914842415783806,-1.21972798977825e-008,-3.8874854582243e-006,-8.04224515558099e-005,-.000588858329194405,-.00243266095707262,-.00687768579306728,-.0146384003311175,-.0246422766763918,-.0335085692003862,-.0366008145021687,-.0304458768485863,-.0151372558449131,-1.68911888154202e-008,-5.3835159512905e-006,-.000111371619378443,-.000815470113608189,-.00336882779555505,-.00952444236070643,-.0202717315680825,-.0341254240019886,-.0464037534631961,-.05068599505261,-.0421624377574542,-.0209625628637279,-1.97926499253432e-008,-6.30826212149848e-006,-.000130502328643898,-.00095554638928031,-.00394750363309629,-.0111604905634277,-.023753880836457,-.0399872725481971,-.0543746954436094,-.0593925132894224,-.0494048334697752,-.0245633787434188,-1.93889081179464e-008,-6.179582578331e-006,-.000127840267412333,-.000936054606840123,-.00386698019345867,-.0109328324858832,-.0232693355725446,-.0391715892641307,-.0532655292683946,-.0581809907923459,-.0483970453850838,-.0240623208776429,-1.49915585521718e-008,-4.77807071380575e-006,-9.88464560551176e-005,-.000723760067411173,-.00298996001413125,-.00845329697557845,-.0179919160266198,-.030287583522627,-.0411850577650857,-.0449857065070876,-.0374207322675915,-.0186050545055932,-7.16228788345407e-009,-2.28274584398164e-006,-4.72243611004261e-005,-.000345779789560036,-.00142846751434807,-.00403860254373974,-.00859572283356178,-.0144700360358322,-.0196763558094162,-.021492133691307,-.0178779315290952,-.00888865263692734},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.81377664478188e-011,-6.32055938620332e-008,-3.13787157581542e-006,-4.12071599264361e-005,-.000260771859326372,-.00102018939735755,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404,-8.64525128335375e-011,-1.43277462826387e-007,-7.11307734943901e-006,-9.34103607574427e-005,-.000591130121526243,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-1.29411477385236e-010,-2.14473212317988e-007,-1.06476239773248e-005,-.000139826737158981,-.000884867539951351,-.00346177108471816,-.00947763686126095,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-1.64245810125438e-010,-2.72204036451189e-007,-1.35136980227866e-005,-.000177464597313172,-.00112305174849657,-.00439359327137407,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.88761417029128e-010,-3.12833670474357e-007,-1.55307754038705e-005,-.000203953262708829,-.00129068034844452,-.00504938842044633,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-2.01416927810585e-010,-3.33807606524435e-007,-1.65720363705528e-005,-.000217627310910781,-.00137721402318714,-.00538792470927311,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-2.01416927810585e-010,-3.33807606524435e-007,-1.65720363705528e-005,-.000217627310910781,-.00137721402318714,-.00538792470927311,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-1.88761417029128e-010,-3.12833670474357e-007,-1.55307754038705e-005,-.000203953262708829,-.00129068034844452,-.00504938842044633,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-1.64245810125438e-010,-2.72204036451189e-007,-1.35136980227866e-005,-.000177464597313172,-.00112305174849657,-.00439359327137407,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.29411477385236e-010,-2.14473212317988e-007,-1.06476239773248e-005,-.000139826737158981,-.000884867539951351,-.00346177108471816,-.00947763686126095,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-8.64525128335375e-011,-1.43277462826387e-007,-7.11307734943901e-006,-9.34103607574427e-005,-.000591130121526243,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-3.81377664478188e-011,-6.32055938620331e-008,-3.13787157581542e-006,-4.12071599264361e-005,-.000260771859326372,-.00102018939735755,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404},
			{-6.09020435343665e-013,-1.00932754786986e-009,-5.01085430834115e-008,-6.58035454488694e-007,-4.16425517497517e-006,-1.62913628348369e-005,-4.46024931010705e-005,-9.12717393010351e-005,-.000144026600260793,-.000175413507334088,-.000156980270065097,-8.12226984453021e-005,-7.17874795152048e-012,-1.18973151739896e-008,-5.90647834026672e-007,-7.75650601653937e-006,-4.90856079239649e-005,-.000192032287901912,-.000525745997014592,-.0010758535732373,-.00169769452974626,-.00206766355177068,-.00185038420184216,-.000957401831273779,-2.57878418572552e-011,-4.27381047927422e-008,-2.12175340882166e-006,-2.78632920211386e-005,-.000176327669275857,-.000689827572348814,-.00188860992469107,-.00386473267978856,-.00609853951561999,-.00742755993763223,-.00664703866113062,-.00343922466500658,-5.87003250388414e-011,-9.72838540256606e-008,-4.82970290571489e-006,-6.34246288365862e-005,-.000401370985487183,-.00157024007444182,-.00429900326923858,-.00879721016390503,-.0138819779418027,-.016907199330518,-.0151305150742709,-.00782863516982087,-1.03341964508827e-010,-1.71268295079276e-007,-8.50269544402536e-006,-.000111659104747358,-.000706613909031054,-.00276440878199448,-.0075683983517811,-.0154874948296011,-.0244392321648105,-.0297651365985099,-.0266372826857591,-.0137823178549157,-1.52587440258738e-010,-2.52882658732434e-007,-1.25544790953984e-005,-.00016486794165333,-.00104333615238232,-.00408173061038807,-.0111749620480468,-.0228677401605563,-.0360852427727842,-.0439490968079129,-.0393307288068947,-.0203499963670061,-1.96276912228913e-010,-3.25289075746179e-007,-1.6149129884536e-005,-.000212073617975246,-.00134206850877865,-.0052504287338376,-.0143746237655383,-.029415326849682,-.0464173199082629,-.0565327854119549,-.0505920670327779,-.0261766921577074,-2.23602400308645e-010,-3.7057551652431e-007,-1.83973966375979e-005,-.000241598308649254,-.0015289100309035,-.00598138850975171,-.0163758454369784,-.0335105011321007,-.0528794957568774,-.0644032269037527,-.0576354472700868,-.0298209867484436,-2.25781763041192e-010,-3.74187367153787e-007,-1.85767086689201e-005,-.000243953070268046,-.00154381170252316,-.00603968669970425,-.0165354542211866,-.033837114519146,-.0533948909503518,-.0650309392734592,-.0581971968115976,-.0301116399215653,-1.98359412066345e-010,-3.28740395820796e-007,-1.63204723006016e-005,-.000214323721005407,-.00135630786795552,-.00530613583081857,-.0145271386554198,-.0297274237371193,-.046909808097858,-.0571325987832424,-.0511288493276215,-.026454427101384,-1.42561396724728e-010,-2.36266530031746e-007,-1.17295635339088e-005,-.000154034984775778,-.000974781796488713,-.00381353285620554,-.0104406902377677,-.0213651724656962,-.0337141943145245,-.0410613390925373,-.0367464296104888,-.0190128617434316,-6.54475287423684e-011,-1.08465972348532e-007,-5.38484445409324e-006,-7.0714859176842e-005,-.000447505854382295,-.00175072850680893,-.00479314449861828,-.00980838972652806,-.0154776170276039,-.0188505670693449,-.0168696650240853,-.0087284836148921},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-4.25830680908177e-013,-3.66970948242474e-009,-4.37202313877829e-007,-1.02973144060446e-005,-9.98226951437858e-005,-.000540391748990977,-.00190309670559272,-.00473399758463878,-.00866893492130421,-.0117725765312941,-.011334382523687,-.00610300723148638,-9.65293351840578e-013,-8.31867295004696e-009,-9.91071113277978e-007,-2.33424447406373e-005,-.000226282859139225,-.00122498585958623,-.00431403062339345,-.0107312521172389,-.0196511562511815,-.026686639477119,-.0256933203450701,-.0138345886544025,-1.44495555626435e-012,-1.24522899458459e-008,-1.48354249933819e-006,-3.49414974841223e-005,-.000338724672636725,-.00183369140663751,-.00645770791571394,-.0160636995405713,-.0294159773896932,-.0399474500854692,-.0384605425083467,-.0207091095227104,-1.83390144930814e-012,-1.58041349298872e-008,-1.88287503228147e-006,-4.43468746144642e-005,-.000429900881983438,-.00232727526714406,-.00819596135984531,-.0203876456552513,-.0373340227205607,-.0507003044420999,-.0488131585372653,-.0262834838088483,-2.10763267567556e-012,-1.81630867905044e-008,-2.16391613832206e-006,-5.09661639870523e-005,-.000494068613398971,-.00267464830248913,-.00941930438906687,-.0234307400647473,-.0429065619801723,-.0582679174767988,-.0560990929882762,-.030206600974669,-2.24893892602145e-012,-1.93808310960915e-008,-2.30899591389123e-006,-5.43831908772887e-005,-.0005271934477113,-.00285397012027107,-.0100508217305121,-.0250016542281072,-.0457832327865349,-.0621744904907474,-.0598602571462754,-.0322318027893339,-2.24893892602145e-012,-1.93808310960915e-008,-2.30899591389123e-006,-5.43831908772887e-005,-.0005271934477113,-.00285397012027107,-.0100508217305121,-.0250016542281072,-.0457832327865349,-.0621744904907474,-.0598602571462754,-.0322318027893339,-2.10763267567556e-012,-1.81630867905044e-008,-2.16391613832206e-006,-5.09661639870523e-005,-.000494068613398971,-.00267464830248913,-.00941930438906688,-.0234307400647473,-.0429065619801723,-.0582679174767988,-.0560990929882762,-.030206600974669,-1.83390144930814e-012,-1.58041349298872e-008,-1.88287503228147e-006,-4.43468746144641e-005,-.000429900881983438,-.00232727526714406,-.00819596135984532,-.0203876456552513,-.0373340227205607,-.0507003044420999,-.0488131585372653,-.0262834838088483,-1.44495555626435e-012,-1.24522899458459e-008,-1.48354249933819e-006,-3.49414974841223e-005,-.000338724672636725,-.00183369140663751,-.00645770791571394,-.0160636995405713,-.0294159773896932,-.0399474500854692,-.0384605425083467,-.0207091095227104,-9.65293351840578e-013,-8.31867295004696e-009,-9.91071113277978e-007,-2.33424447406373e-005,-.000226282859139225,-.00122498585958623,-.00431403062339345,-.0107312521172389,-.0196511562511815,-.026686639477119,-.0256933203450701,-.0138345886544025,-4.25830680908177e-013,-3.66970948242474e-009,-4.37202313877829e-007,-1.02973144060446e-005,-9.98226951437858e-005,-.000540391748990977,-.00190309670559272,-.00473399758463878,-.0086689349213042,-.0117725765312941,-.011334382523687,-.00610300723148638}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{-1.25649763607893e-005,-2.84828893052902e-005,-4.2636271224355e-005,-5.41128889760362e-005,-6.2189870139491e-005,-6.63593905024702e-005,-6.63593905024702e-005,-6.2189870139491e-005,-5.41128889760362e-005,-4.2636271224355e-005,-2.84828893052902e-005,-1.25649763607893e-005,-.000148107999463138,-.000335738296102257,-.000502569416311285,-.000637848532403299,-.000733054881183536,-.000782202551815753,-.000782202551815753,-.000733054881183536,-.000637848532403299,-.000502569416311285,-.000335738296102257,-.000148107999463138,-.000532040641869991,-.00120605517059218,-.00180535390260964,-.00229131001583844,-.0026333148170561,-.00280986543096192,-.00280986543096192,-.0026333148170561,-.00229131001583844,-.00180535390260964,-.00120605517059218,-.000532040641869991,-.00121107298487856,-.00274531815886015,-.00410948932756509,-.00521566106380589,-.00599415943941434,-.00639603790909957,-.00639603790909957,-.00599415943941434,-.00521566106380588,-.00410948932756509,-.00274531815886015,-.00121107298487856,-.00213209486213416,-.00483313459594366,-.00723475892097063,-.00918217505932459,-.010552722010284,-.0112602293456112,-.0112602293456112,-.010552722010284,-.00918217505932459,-.00723475892097063,-.00483313459594367,-.00213209486213416,-.00314810057025839,-.0071362649232201,-.0106823336471961,-.0135577506675935,-.0155814034208121,-.0166260587433132,-.0166260587433132,-.0155814034208121,-.0135577506675935,-.0106823336471961,-.0071362649232201,-.00314810057025839,-.00404947784869215,-.00917955004423722,-.0137409439473876,-.0174396623555152,-.0200427358008248,-.0213865011900097,-.0213865011900097,-.0200427358008248,-.0174396623555152,-.0137409439473876,-.00917955004423722,-.00404947784869215,-.00461324236601103,-.0104575184128172,-.0156539453074288,-.0198675958317378,-.0228330667266135,-.0243639098760258,-.0243639098760258,-.0228330667266135,-.0198675958317378,-.0156539453074288,-.0104575184128172,-.00461324236601103,-.00465820578534291,-.0105594436420292,-.0158065180211977,-.0200612373037489,-.023055611451648,-.0246013751140091,-.0246013751140091,-.023055611451648,-.0200612373037489,-.0158065180211977,-.0105594436420292,-.00465820578534291,-.00409244284577621,-.00927694515432801,-.0138867354885914,-.0176246973329244,-.0202553894113433,-.0216134121636217,-.0216134121636217,-.0202553894113433,-.0176246973329244,-.0138867354885914,-.00927694515432801,-.00409244284577621,-.00294124872640195,-.00666736327135993,-.0099804309086089,-.0126669132684861,-.0145575981275773,-.0155336124156416,-.0155336124156416,-.0145575981275773,-.0126669132684861,-.0099804309086089,-.00666736327135993,-.00294124872640195,-.00135027759956182,-.0030608738365598,-.00458184721642184,-.00581516588124543,-.0066831473580057,-.00713121902844807,-.00713121902844807,-.0066831473580057,-.00581516588124543,-.00458184721642184,-.0030608738365598,-.00135027759956182},
			{-1.49555377634105e-007,-3.39019282157687e-007,-5.07480751318181e-007,-6.4408187594673e-007,-7.40218624107564e-007,-7.89846587943962e-007,-7.89846587943962e-007,-7.40218624107564e-007,-6.4408187594673e-007,-5.07480751318181e-007,-3.39019282157687e-007,-1.49555377634105e-007,-9.16670694821036e-006,-2.07795296865571e-005,-3.11050488640572e-005,-3.94777499870446e-005,-4.53702655975518e-005,-4.84121154335211e-005,-4.84121154335211e-005,-4.53702655975518e-005,-3.94777499870446e-005,-3.11050488640572e-005,-2.07795296865571e-005,-9.16670694821036e-006,-7.90224989069185e-005,-.000179131979588681,-.000268144133301817,-.000340321826892032,-.000391118837314757,-.000417341403029572,-.000417341403029572,-.000391118837314757,-.000340321826892032,-.000268144133301817,-.000179131979588681,-7.90224989069185e-005,-.000322611816370485,-.000731311893505376,-.0010947067871833,-.00138937447237002,-.00159675485169678,-.00170380929406662,-.00170380929406662,-.00159675485169678,-.00138937447237002,-.0010947067871833,-.000731311893505376,-.000322611816370485,-.000870029052744199,-.00197222346386832,-.00295223752124433,-.00374690601758642,-.00430617553540284,-.00459488311014376,-.00459488311014376,-.00430617553540284,-.00374690601758643,-.00295223752124433,-.00197222346386832,-.000870029052744198,-.00177760460022009,-.00402955911756787,-.00603188017934985,-.00765551144808981,-.0087981860110801,-.00938806046568492,-.00938806046568492,-.0087981860110801,-.00765551144808981,-.00603188017934986,-.00402955911756787,-.00177760460022009,-.0029412776627297,-.00666742886561148,-.00998052909718346,-.0126670378869689,-.0145577413468253,-.0155337652370252,-.0155337652370252,-.0145577413468253,-.0126670378869689,-.00998052909718346,-.00666742886561148,-.0029412776627297,-.00407317670628986,-.00923327174798222,-.013821360407493,-.0175417249152225,-.0201600323904989,-.0215116619305183,-.0215116619305183,-.0201600323904989,-.0175417249152225,-.013821360407493,-.00923327174798222,-.00407317670628986,-.00477284112519419,-.0108193045125715,-.0161955058952257,-.0205549309837037,-.0236229946837403,-.0252067985596364,-.0252067985596364,-.0236229946837403,-.0205549309837037,-.0161955058952257,-.0108193045125715,-.00477284112519419,-.00467548197876496,-.0105986061434719,-.0158651406916522,-.0201356397307597,-.0231411193105227,-.0246926158479963,-.0246926158479963,-.0231411193105227,-.0201356397307597,-.0158651406916522,-.0105986061434719,-.00467548197876496,-.00361509587945292,-.0081948722230624,-.0122669716195725,-.015568933545552,-.0178927788504699,-.0190924003579595,-.0190924003579595,-.0178927788504699,-.015568933545552,-.0122669716195725,-.00819487222306241,-.00361509587945292,-.00172712912568916,-.00391513889802946,-.00586060361181134,-.00743813151271418,-.00854835957956341,-.00912148441898038,-.00912148441898038,-.00854835957956341,-.00743813151271418,-.00586060361181134,-.00391513889802946,-.00172712912568916},
			{-1.63148105416827e-009,-3.69831927536025e-009,-5.53604453566559e-009,-7.02620925147273e-009,-8.07495310619039e-009,-8.61633840464567e-009,-8.61633840464567e-009,-8.07495310619039e-009,-7.02620925147273e-009,-5.53604453566559e-009,-3.69831927536025e-009,-1.63148105416827e-009,-5.19981416069715e-007,-1.1787187408438e-006,-1.7644337761237e-006,-2.23937521483844e-006,-2.57362813998106e-006,-2.74617706012381e-006,-2.74617706012381e-006,-2.57362813998106e-006,-2.23937521483844e-006,-1.7644337761237e-006,-1.1787187408438e-006,-5.19981416069715e-007,-1.07571284042538e-005,-2.43847731013085e-005,-3.6501767378587e-005,-4.63271301374563e-005,-5.3241995792529e-005,-5.68116058451738e-005,-5.68116058451738e-005,-5.3241995792529e-005,-4.63271301374563e-005,-3.6501767378587e-005,-2.43847731013085e-005,-1.07571284042538e-005,-7.87643815441602e-005,-.000178546866806921,-.000267268273167243,-.000339210207117138,-.000389841295295595,-.000415978208195064,-.000415978208195064,-.000389841295295595,-.000339210207117138,-.000267268273167243,-.000178546866806921,-7.87643815441602e-005,-.000325387323726201,-.000737603546311472,-.00110412481400688,-.0014013275964412,-.00161049211924643,-.00171846757683388,-.00171846757683388,-.00161049211924644,-.0014013275964412,-.00110412481400688,-.000737603546311472,-.000325387323726201,-.000919943967994991,-.00208537298082308,-.00312161196363602,-.0039618718233036,-.00455322750019265,-.00485849867597654,-.00485849867597654,-.00455322750019265,-.0039618718233036,-.00312161196363602,-.00208537298082308,-.000919943967994991,-.00195799989864057,-.00443848780701132,-.00664400889731993,-.00843240990574942,-.00969104564411167,-.0103407818802712,-.0103407818802712,-.00969104564411167,-.00843240990574942,-.00664400889731993,-.00443848780701132,-.00195799989864057,-.00329609616783617,-.00747174842135327,-.0111845216543321,-.0141951151250118,-.0163139019731612,-.0174076676672231,-.0174076676672231,-.0163139019731612,-.0141951151250118,-.0111845216543321,-.00747174842135327,-.00329609616783617,-.00448203175305139,-.0101600839205191,-.0152087131700155,-.0193025183395333,-.022183644813943,-.0236709474657956,-.0236709474657956,-.022183644813943,-.0193025183395333,-.0152087131700155,-.0101600839205191,-.00448203175305139,-.00489564361298881,-.0110976790646466,-.0166122070513835,-.0210837976681447,-.0242308009023486,-.0258553551512484,-.0258553551512484,-.0242308009023486,-.0210837976681447,-.0166122070513835,-.0110976790646466,-.00489564361298881,-.00407237283003069,-.00923144948283444,-.0138186326438083,-.0175382629057834,-.0201560536356863,-.021507416420056,-.021507416420056,-.0201560536356863,-.0175382629057834,-.0138186326438083,-.00923144948283444,-.00407237283003069,-.00202472570360245,-.0045897450527047,-.00687042711223165,-.00871977424071199,-.0100213024648473,-.0106931807723117,-.0106931807723117,-.0100213024648473,-.00871977424071199,-.00687042711223165,-.0045897450527047,-.00202472570360245},
			{-1.70557276538753e-011,-3.86627390961513e-011,-5.7874572088226e-011,-7.34529592765333e-011,-8.44166719834875e-011,-9.00763884617597e-011,-9.00763884617597e-011,-8.44166719834875e-011,-7.34529592765333e-011,-5.7874572088226e-011,-3.86627390961513e-011,-1.70557276538753e-011,-2.82664008807978e-008,-6.40756292912074e-008,-9.59153363989562e-008,-1.21733345825304e-007,-1.39903470536826e-007,-1.49283299887593e-007,-1.49283299887593e-007,-1.39903470536826e-007,-1.21733345825304e-007,-9.59153363989562e-008,-6.40756292912074e-008,-2.82664008807978e-008,-1.40329882934204e-006,-3.18106489584209e-006,-4.76176220142831e-006,-6.04350947999848e-006,-6.94557390780468e-006,-7.41123996847038e-006,-7.41123996847038e-006,-6.94557390780468e-006,-6.04350947999848e-006,-4.76176220142831e-006,-3.18106489584208e-006,-1.40329882934204e-006,-1.84284021471628e-005,-4.17743832824876e-005,-6.2532417858728e-005,-7.93645806216638e-005,-9.12106719107567e-005,-9.73258921709034e-005,-9.73258921709034e-005,-9.12106719107567e-005,-7.93645806216638e-005,-6.2532417858728e-005,-4.17743832824876e-005,-1.84284021471628e-005,-.000116620720789999,-.000264361427000411,-.000395724793999518,-.000502244010271906,-.000577209799147466,-.000615908834952788,-.000615908834952788,-.000577209799147466,-.000502244010271906,-.000395724793999518,-.000264361427000411,-.000116620720789999,-.000456242568387133,-.00103423247275538,-.0015481510932686,-.00196487464364187,-.00225815515010815,-.00240955318100971,-.00240955318100971,-.00225815515010815,-.00196487464364187,-.0015481510932686,-.00103423247275538,-.000456242568387133,-.00124910090182185,-.00283152165958353,-.00423852805667493,-.00537943379114375,-.00618237716052034,-.00659687468888933,-.00659687468888933,-.00618237716052034,-.00537943379114375,-.00423852805667493,-.00283152165958353,-.00124910090182185,-.0025560815986996,-.00579424800656579,-.00867345748885218,-.0110081352954835,-.012651228153929,-.0134994298511859,-.0134994298511859,-.012651228153929,-.0110081352954835,-.00867345748885218,-.00579424800656579,-.0025560815986996,-.00403349103971441,-.00914331038111467,-.0136866964976499,-.0173708128492007,-.0199636097009642,-.021302070080123,-.021302070080123,-.0199636097009642,-.0173708128492007,-.0136866964976499,-.00914331038111467,-.00403349103971441,-.00491248705999991,-.0111358605958299,-.0166693612993899,-.021156336410114,-.0243141668994061,-.025944310422275,-.025944310422275,-.0243141668994061,-.021156336410114,-.0166693612993899,-.0111358605958299,-.00491248705999991,-.00439626090995001,-.00996565447158663,-.0149176701290573,-.0189331337918146,-.0217591253050265,-.0232179660428578,-.0232179660428578,-.0217591253050265,-.0189331337918146,-.0149176701290573,-.00996565447158663,-.00439626090995001,-.00227465638852363,-.00515629988163542,-.00771850769460683,-.00979613690280579,-.0112583248350307,-.012013139317637,-.012013139317637,-.0112583248350307,-.00979613690280579,-.00771850769460683,-.00515629988163542,-.00227465638852363},
			{-1.7384464750039e-013,-3.94079360668192e-013,-5.8990063561595e-013,-7.48687131527224e-013,-8.60437436733519e-013,-9.18125471866621e-013,-9.18125471866621e-013,-8.60437436733519e-013,-7.48687131527224e-013,-5.8990063561595e-013,-3.94079360668192e-013,-1.7384464750039e-013,-1.49815262263485e-009,-3.39608401063939e-009,-5.0836260825345e-009,-6.45201106711068e-009,-7.41504913145304e-009,-7.91219116240981e-009,-7.91219116240981e-009,-7.41504913145304e-009,-6.45201106711068e-009,-5.0836260825345e-009,-3.39608401063939e-009,-1.49815262263485e-009,-1.78487097219834e-007,-4.04603087706546e-007,-6.05653689159397e-007,-7.68680513053121e-007,-8.83415064139356e-007,-9.42643634493817e-007,-9.42643634493817e-007,-8.83415064139356e-007,-7.68680513053121e-007,-6.05653689159397e-007,-4.04603087706546e-007,-1.78487097219834e-007,-4.20386100245681e-006,-9.52951316020476e-006,-1.42648066135301e-005,-1.81045357479959e-005,-2.0806849318325e-005,-2.22018447046734e-005,-2.22018447046734e-005,-2.0806849318325e-005,-1.81045357479959e-005,-1.42648066135301e-005,-9.52951316020476e-006,-4.20386100245681e-006,-4.07524446402028e-005,-9.23795904009122e-005,-.00013828376853596,-.000175506300131138,-.000201702666783354,-.000215225823762682,-.000215225823762682,-.000201702666783354,-.000175506300131138,-.00013828376853596,-9.23795904009122e-005,-4.07524446402028e-005,-.000220614007696907,-.000500098382997093,-.000748601381966019,-.000950106149209752,-.00109192059703656,-.00116512842258577,-.00116512842258577,-.00109192059703656,-.000950106149209752,-.00074860138196602,-.000500098382997093,-.000220614007696907,-.000776935976612406,-.0017611956269338,-.0026363482167924,-.00334598721372161,-.00384541491403083,-.00410323078906309,-.00410323078906309,-.00384541491403083,-.00334598721372161,-.0026363482167924,-.0017611956269338,-.000776935976612406,-.00193264642090617,-.00438101533121211,-.00655797787568285,-.00832322148498332,-.00956555957532708,-.0102068825969568,-.0102068825969568,-.00956555957532708,-.00832322148498332,-.00655797787568285,-.00438101533121211,-.00193264642090617,-.00353907786161354,-.00802255094483993,-.0120090224811522,-.015241550951156,-.0175165305773379,-.0186909265162125,-.0186909265162125,-.0175165305773379,-.015241550951156,-.0120090224811522,-.00802255094483993,-.00353907786161354,-.00480613424305024,-.0108947749442933,-.0163084782050857,-.0206983126135993,-.0237877776977751,-.0253826294522236,-.0253826294522236,-.0237877776977751,-.0206983126135993,-.0163084782050857,-.0108947749442933,-.00480613424305024,-.00462724228856079,-.0104892541067666,-.0157014507286737,-.0199278885241277,-.0229023588080574,-.0244378476456489,-.0244378476456489,-.0229023588080574,-.0199278885241277,-.0157014507286737,-.0104892541067666,-.00462724228856079,-.00249154226883633,-.00564794716718901,-.00845445855931378,-.010730187331938,-.0123317932081052,-.0131585783867511,-.0131585783867511,-.0123317932081052,-.010730187331938,-.00845445855931378,-.00564794716718901,-.00249154226883633},
			{-.0019095808945752,-.00415955383985738,-.00578758817666173,-.00658769779901337,-.00652410992157336,-.0057268264950684,-.00445208652288796,-.00301523747077974,-.00171271584051363,-.000752419091586911,-.000209456341570312,-1.77695599831757e-005,-.00432872929307012,-.0094290755651378,-.0131195816567463,-.0149333084120673,-.0147891643705445,-.012981844171125,-.010092202640724,-.00683508449545443,-.00388246617391187,-.00170561957949523,-.000474805651832001,-4.02808883575737e-005,-.00647971047522447,-.0141144607515462,-.0196388096682036,-.0223537921630579,-.0221380217619632,-.0194326292933208,-.0151071011240822,-.0102314941881775,-.0058116955424022,-.00255315597436325,-.000710740484695346,-6.02967930241717e-005,-.00822388645802562,-.0179137205405699,-.0249250860049403,-.0283708738774972,-.0280970234814972,-.0246634070303329,-.0191735548724589,-.0129855565030636,-.00737605861435863,-.00324040170051977,-.000902054045409203,-7.65271815013742e-005,-.00945139763454551,-.0205875527108991,-.0286454464212646,-.0326055584089547,-.0322908326405259,-.0283447088011335,-.0220354320420534,-.0149238025892903,-.00847702157560611,-.00372406952887607,-.00103669615509983,-8.79497578075961e-005,-.0100850666679019,-.0219678453543708,-.030565980615856,-.0347915983448399,-.0344557717846379,-.0302450800394711,-.0235127977673766,-.015924369058549,-.00904536355795187,-.00397374980154702,-.00110620145747811,-9.38463500544579e-005,-.0100850666679019,-.0219678453543708,-.030565980615856,-.0347915983448399,-.0344557717846379,-.0302450800394712,-.0235127977673766,-.015924369058549,-.00904536355795187,-.00397374980154702,-.00110620145747811,-9.38463500544579e-005,-.00945139763454551,-.0205875527108991,-.0286454464212646,-.0326055584089547,-.0322908326405259,-.0283447088011335,-.0220354320420534,-.0149238025892903,-.00847702157560611,-.00372406952887607,-.00103669615509983,-8.79497578075961e-005,-.00822388645802562,-.0179137205405699,-.0249250860049403,-.0283708738774972,-.0280970234814972,-.0246634070303329,-.0191735548724589,-.0129855565030636,-.00737605861435863,-.00324040170051977,-.000902054045409203,-7.65271815013742e-005,-.00647971047522447,-.0141144607515462,-.0196388096682036,-.0223537921630579,-.0221380217619632,-.0194326292933208,-.0151071011240822,-.0102314941881775,-.0058116955424022,-.00255315597436325,-.000710740484695346,-6.02967930241717e-005,-.00432872929307012,-.0094290755651378,-.0131195816567463,-.0149333084120673,-.0147891643705445,-.012981844171125,-.010092202640724,-.00683508449545443,-.00388246617391187,-.00170561957949523,-.000474805651832001,-4.02808883575737e-005,-.0019095808945752,-.00415955383985738,-.00578758817666173,-.00658769779901337,-.00652410992157336,-.0057268264950684,-.00445208652288796,-.00301523747077974,-.00171271584051363,-.000752419091586911,-.000209456341570312,-1.77695599831757e-005},
			{-3.04940193335369e-005,-6.64237454259384e-005,-9.2421759274531e-005,-.000105198677163819,-.000104183244945397,-9.14514584621857e-005,-7.10951878965703e-005,-4.815020405282e-005,-2.73502893236142e-005,-1.20153497508037e-005,-3.34479976602295e-006,-2.83761377805336e-007,-.000359444225710958,-.000782961126977123,-.00108940928179785,-.00124001551403817,-.00122804624086627,-.00107797198911481,-.00083802513816711,-.000567564171331005,-.000322387923394712,-.000141629348384198,-3.94263855120624e-005,-3.34479976602295e-006,-.00129121274513816,-.00281259042095047,-.00391342814465003,-.00445444305782967,-.004411446462074,-.00387234255465894,-.00301039399647408,-.00203883117125292,-.00115809732300615,-.000508767721494418,-.000141629348384198,-1.20153497508037e-005,-.00293916056463564,-.00640224074681438,-.00890805463205047,-.0101395555630045,-.0100416833114065,-.00881453236290365,-.00685249688850132,-.00464094875074617,-.00263615271349073,-.00115809732300615,-.000322387923394712,-2.73502893236141e-005,-.00517439429092323,-.0112711494458756,-.0156826365956051,-.017850694871518,-.0176783907020889,-.0155179905734142,-.0120638257076051,-.00817039361825585,-.00464094875074617,-.00203883117125292,-.000567564171331005,-4.815020405282e-005,-.00764014486751864,-.0166421825914966,-.0231558726990041,-.0263570743040411,-.0261026621464626,-.0229127680203357,-.0178125923307157,-.0120638257076051,-.00685249688850132,-.00301039399647408,-.00083802513816711,-7.10951878965703e-005,-.00982770299465891,-.0214072416856188,-.0297859586976462,-.0339037416907514,-.0335764851836737,-.0294732472739763,-.0229127680203357,-.0155179905734142,-.00881453236290365,-.00387234255465894,-.00107797198911481,-9.14514584621857e-005,-.0111959066105712,-.0243875378440283,-.0339327319991675,-.0386237888878892,-.0382509720360841,-.0335764851836737,-.0261026621464626,-.0176783907020889,-.0100416833114065,-.004411446462074,-.00122804624086627,-.000104183244945397,-.0113050286127969,-.02462523337432,-.0342634607007844,-.039000239435718,-.0386237888878892,-.0339037416907514,-.0263570743040411,-.017850694871518,-.0101395555630045,-.00445444305782967,-.00124001551403817,-.000105198677163819,-.00993197501349335,-.0216343727160793,-.0301019879923868,-.0342634607007844,-.0339327319991675,-.0297859586976462,-.0231558726990041,-.0156826365956051,-.00890805463205047,-.00391342814465003,-.00108940928179785,-9.2421759274531e-005,-.00713813484023199,-.0155486768161827,-.0216343727160793,-.02462523337432,-.0243875378440283,-.0214072416856188,-.0166421825914966,-.0112711494458756,-.00640224074681438,-.00281259042095047,-.000782961126977123,-6.64237454259384e-005,-.00327699711041027,-.00713813484023199,-.00993197501349335,-.0113050286127969,-.0111959066105712,-.00982770299465891,-.00764014486751864,-.00517439429092323,-.00293916056463564,-.00129121274513816,-.000359444225710958,-3.04940193335369e-005},
			{-3.62956876834475e-007,-7.90612576313682e-007,-1.10005548074599e-006,-1.25213350502815e-006,-1.24004726271932e-006,-1.08850641767933e-006,-8.46214697861095e-007,-5.73110664448679e-007,-3.25538443615425e-007,-1.43013414267412e-007,-3.98116779370311e-008,-3.37748665821831e-009,-2.22467381475193e-005,-4.84590652057195e-005,-6.74257681555381e-005,-7.67470958397088e-005,-7.60062930474401e-005,-6.67178907235581e-005,-5.18670894572505e-005,-3.51277071610772e-005,-1.99532478216857e-005,-8.76572998571608e-006,-2.44017962134497e-006,-2.07016497214321e-007,-.000191780194444643,-.000417746138158903,-.000581250466549428,-.000661605888719569,-.000655219725381629,-.000575148139518477,-.000447125346440885,-.000302821854829854,-.000172008935497438,-7.5565837560655e-005,-2.10358084478815e-005,-1.78460607688856e-006,-.000782948625132055,-.00170546163785514,-.00237297315794306,-.00270102667510736,-.00267495496410619,-.00234806021751757,-.00182540317195597,-.00123627966686315,-.000702231843846736,-.000308499888616729,-8.58793440610684e-005,-7.28570996785344e-006,-.00211147892329134,-.00459933919956203,-.00639950393640168,-.00728420832819645,-.00721389736971712,-.00633231798455615,-.00492280106302518,-.00333403543487275,-.00189379952905124,-.000831971590143711,-.000231601945658244,-1.96483173282813e-005,-.00431407966834182,-.00939716494906272,-.0130751813408912,-.0148827699400637,-.0147391137220966,-.0129379099972572,-.0100580478180592,-.0068119526671359,-.0038693268277857,-.0016998472881189,-.000473198777355153,-4.01445666201286e-005,-.00713820506661555,-.0155488297871431,-.0216345855593206,-.0246254756422405,-.0243877777734532,-.0214074522943001,-.0166423463205833,-.011271260333675,-.00640230373332618,-.00281261809177179,-.000782968829903678,-6.64243989158946e-005,-.00988521790053478,-.021532523808133,-.0299602757900378,-.0341021573849623,-.0337729856666483,-.029645734277678,-.0230468609713296,-.0156088068880401,-.00886611786558599,-.00389500475941537,-.00108428063087214,-9.19866620625105e-005,-.0115832378335861,-.0252312439580584,-.0351066616363676,-.039960009339504,-.0395742945945498,-.034738090181205,-.0270057043390071,-.0182899885770756,-.0103890832686787,-.00456406393315995,-.00127053146952515,-.000107787546507207,-.0113469563151396,-.0247165625951827,-.0343905358485524,-.0391448821859808,-.0387670354712729,-.0340294827249939,-.0264548265171398,-.0179168989163566,-.0101771607988773,-.00447096354344966,-.00124461444104258,-.000105588834409239,-.00877349869072301,-.0191109160505604,-.0265908595098725,-.0302669335343241,-.0299747813866719,-.0263116922143577,-.0204549466275595,-.0138533968774284,-.00786900949156683,-.00345696164727354,-.000962339403243543,-8.16415851719116e-005,-.00419158042503607,-.0091303303785454,-.0127039086840373,-.0144601704064181,-.014320593338432,-.01257053519047,-.00977244733277217,-.00661852576932447,-.003759456439442,-.0016515797496082,-.000459762193746235,-3.90046528002957e-005},
			{-3.95945152473478e-009,-8.62469447627331e-009,-1.20003687174081e-008,-1.36593690107054e-008,-1.35275216933177e-008,-1.18743814217069e-008,-9.2312511197497e-009,-6.25199311274643e-009,-3.55125848055141e-009,-1.56011558760704e-009,-4.34300653798924e-010,-3.68445827925519e-011,-1.26194613503531e-006,-2.74884028563086e-006,-3.82472643681277e-006,-4.35347871350321e-006,-4.31145667798826e-006,-3.78457209223235e-006,-2.9421604480658e-006,-1.99261905231337e-006,-1.13184790520788e-006,-4.97235999402952e-007,-1.38419179545726e-007,-1.174301000065e-008,-2.61065418768862e-005,-5.68667013887132e-005,-7.91241227483076e-005,-9.0062698548559e-005,-8.91933666496246e-005,-7.82934287517701e-005,-6.08660170299574e-005,-4.12223559228577e-005,-2.34151315299593e-005,-1.02865820344585e-005,-2.86355020000406e-006,-2.42933809796949e-007,-.000191153767800758,-.000416381621285402,-.000579351882704146,-.000659444833676081,-.000653079529962769,-.000573269488181101,-.000445664865962291,-.000301832724128607,-.000171447089262513,-7.53190109571436e-005,-2.09670975420157e-005,-1.77877687852613e-006,-.000789684521168946,-.00172013413604819,-.00239338841889015,-.00272426425965937,-.00269796824743438,-.00236826114642384,-.00184110755612247,-.00124691567929258,-.000708273313928056,-.000311153987634738,-8.66181847905149e-005,-7.34839068958039e-006,-.00223261774167844,-.00486321041790597,-.00676665339575356,-.00770211464969917,-.0076277698425577,-.00669561288171989,-.00520522978980597,-.00352531421515612,-.00200244974321189,-.000879703089733842,-.000244889308237957,-2.0775571746115e-005,-.00475188214064504,-.0103508103065028,-.0144020800439593,-.0163931067848593,-.016234872011969,-.0142508780967576,-.0110787610500387,-.00750324488892071,-.00426199478516835,-.00187235160015967,-.000521220049687516,-4.42185272019898e-005,-.00799931630469631,-.0174245495155288,-.0242444552089724,-.027596148748536,-.0273297764014771,-.0239899219175831,-.0186499814768931,-.0126309591445223,-.00717464014604113,-.00315191585985935,-.000877421602303038,-7.44374492353296e-005,-.0108774707577442,-.0236939034040642,-.032967611546778,-.0375252446090086,-.0371630314766057,-.0326214971629358,-.0253602458535457,-.0171755789499336,-.00975607607116152,-.00428597536221463,-.00119311794379991,-.000101220047626196,-.0118812679549556,-.0258804295145632,-.0360099360729874,-.0409881576521328,-.0405925186861703,-.0356318814839124,-.0277005457518707,-.01876057958054,-.0106563884722811,-.00468149470230067,-.00130322152160786,-.000110560858773534,-.00988326696773915,-.0215282741793143,-.0299543628718611,-.03409542703082,-.0337663202773765,-.0296398834369748,-.0230423124750437,-.0156057263557175,-.00886436806094961,-.0038942360467172,-.00108406663876152,-9.19685077033871e-005,-.00491381941200989,-.0107035509528065,-.0148928821041257,-.0169517601570097,-.0167881329718923,-.0147365274132258,-.0114563092049479,-.00775894462383229,-.00440723740391471,-.00193615863497058,-.00053898247521257,-4.57254306619082e-005},
			{-4.13926516217373e-011,-9.01637440362945e-011,-1.25453507524717e-010,-1.42796925104609e-010,-1.41418575087745e-010,-1.24136418981734e-010,-9.65047707354281e-011,-6.53591971617201e-011,-3.71253453125811e-011,-1.63096632460441e-011,-4.54023885618748e-012,-3.85178343554166e-013,-6.85999042669627e-008,-1.49428073991582e-007,-2.07913681993516e-007,-2.36656870434667e-007,-2.34372535522585e-007,-2.05730875518941e-007,-1.5993703651203e-007,-1.08319580712847e-007,-6.15277116719808e-008,-2.70299508117877e-008,-7.5245228000818e-009,-6.38354791449189e-010,-3.40567466501195e-006,-7.41842734727761e-006,-1.03219729945844e-005,-1.17489421676717e-005,-1.16355352814725e-005,-1.02136065356437e-005,-7.94014975773712e-006,-5.37757677216062e-006,-3.05457232158762e-006,-1.34191468136724e-006,-3.73558490210382e-007,-3.16914252834386e-008,-4.47240038949301e-005,-9.74202782733315e-005,-.00013555022303093,-.000154289468887485,-.000152800186874675,-.000134127132921491,-.000104271641780592,-7.06194185176384e-005,-4.01132573852444e-005,-1.76222931839914e-005,-4.90564514068014e-006,-4.16178163573184e-007,-.000283027553294656,-.00061650613987428,-.000857804414450873,-.000976392251931343,-.000966967607276502,-.000848798652964526,-.000659863721515399,-.000446901875894084,-.000253849300234756,-.00011151940991293,-3.10444642827677e-005,-2.6337062228933e-006,-.00110725792950642,-.00241188995212247,-.00335589496079028,-.00381983326596499,-.00378296225321243,-.00332066269912202,-.00258151310547804,-.00174836562742656,-.000993107021958261,-.000436285264394079,-.000121451882844024,-1.03035625519223e-005,-.00303145075477106,-.00660327230083207,-.00918776920961406,-.0104579394995808,-.010356994041021,-.0090913103237661,-.00706766656937942,-.00478667540745511,-.00271892840056321,-.00119446179503315,-.00033251096434234,-2.82090935115534e-005,-.00620337034448753,-.013512521522274,-.0188012736005075,-.0214004702712195,-.0211939018276931,-.0186038861974662,-.0144628287733888,-.00979515178485467,-.00556384423606199,-.00244427156379506,-.00068042954422791,-5.77253164541879e-005,-.00978890451433557,-.0213227287077879,-.029668367645793,-.033769894172609,-.0334439296312284,-.0293568907657418,-.0228223114223215,-.0154567275852829,-.00877973374713072,-.00385705505497627,-.00107371629731872,-9.10904201184736e-005,-.011922145428059,-.0259694710684948,-.0361338281692569,-.0411291773074706,-.0407321771466212,-.0357544728843004,-.0277958494112289,-.0188251253083289,-.0106930517505446,-.00469760137328079,-.0013077052520396,-.000110941243135535,-.0106693129707941,-.0232404826956437,-.0323367236121244,-.0368071390818563,-.0364518574766186,-.0319972326801651,-.024874937019263,-.0168468967973055,-.00956937796371369,-.00420395637395983,-.00117028572514512,-9.92830402484601e-005,-.00552037774992994,-.0120247895925313,-.0167312487713743,-.01904427325192,-.0188604480446285,-.0165555937697616,-.0128704677825032,-.00871670317383983,-.00495126362270572,-.00217515666585201,-.000605514084724702,-5.13697449717053e-005},
			{-.0029914753970567,-.00626152973771969,-.00809817233744839,-.00826680332565275,-.00705494900380919,-.00509444235123848,-.00307890148348425,-.00150693452348,-.000558780057100243,-.000136870983053836,-1.58772021730906e-005,-2.5903751261877e-007,-.00678121949037336,-.0141939350525009,-.0183573243307094,-.0187395851191542,-.0159924957882962,-.0115483255515964,-.00697940112403575,-.00341599124344908,-.00126666935578621,-.000310265689921581,-3.59912011760796e-005,-5.8719862146837e-007,-.0101508632191237,-.0212470181010361,-.027479229748374,-.0280514390660412,-.0239392984565476,-.0172867834834987,-.0104475229362594,-.00511342538303042,-.00189608777487504,-.000464439262650617,-5.38755250068202e-005,-8.78982445184828e-007,-.0128832216939646,-.0269661839217411,-.0348759510581064,-.0356021848114063,-.0303831588068934,-.0219399532025914,-.0132597347865977,-.0064898315939296,-.00240646717698951,-.000589454695127486,-6.83774687490587e-005,-1.11558253342289e-006,-.0148061931138207,-.0309912020589656,-.0400815943915947,-.0409162270209603,-.0349182003841082,-.0252147476572524,-.0152389051863002,-.00745851481395462,-.00276566053049211,-.000677437698056106,-7.85836051712971e-005,-1.2820962657192e-006,-.0157988744548141,-.033069007459881,-.042768865222384,-.0436594558033572,-.0372592914205353,-.0269052706245537,-.0162605977118075,-.00795857100195779,-.00295108426386074,-.00072285651418097,-8.38522436364219e-005,-1.3680544205638e-006,-.0157988744548141,-.033069007459881,-.042768865222384,-.0436594558033572,-.0372592914205353,-.0269052706245537,-.0162605977118075,-.00795857100195779,-.00295108426386074,-.00072285651418097,-8.38522436364219e-005,-1.3680544205638e-006,-.0148061931138207,-.0309912020589656,-.0400815943915947,-.0409162270209603,-.0349182003841082,-.0252147476572524,-.0152389051863002,-.00745851481395462,-.00276566053049211,-.000677437698056106,-7.85836051712971e-005,-1.2820962657192e-006,-.0128832216939646,-.0269661839217411,-.0348759510581064,-.0356021848114063,-.0303831588068934,-.0219399532025914,-.0132597347865977,-.0064898315939296,-.00240646717698951,-.000589454695127486,-6.83774687490587e-005,-1.11558253342289e-006,-.0101508632191237,-.0212470181010361,-.027479229748374,-.0280514390660412,-.0239392984565476,-.0172867834834987,-.0104475229362594,-.00511342538303042,-.00189608777487504,-.000464439262650617,-5.38755250068202e-005,-8.78982445184828e-007,-.00678121949037336,-.0141939350525009,-.0183573243307094,-.0187395851191542,-.0159924957882962,-.0115483255515964,-.00697940112403575,-.00341599124344908,-.00126666935578621,-.000310265689921581,-3.59912011760796e-005,-5.8719862146837e-007,-.0029914753970567,-.00626152973771969,-.00809817233744839,-.00826680332565275,-.00705494900380919,-.00509444235123848,-.00307890148348425,-.00150693452348,-.000558780057100243,-.000136870983053836,-1.58772021730906e-005,-2.5903751261877e-007},
			{-4.77707484775864e-005,-9.99901127316091e-005,-.000129319383418971,-.000132012244784625,-.000112660192597526,-8.13529419075361e-005,-4.91668520822556e-005,-2.40641758792944e-005,-8.92313591757784e-006,-2.1856871401242e-006,-2.53542393256769e-007,-4.13655946284767e-009,-.000563091388850621,-.00117862024866094,-.00152433515352727,-.00155607690124302,-.0013279671418103,-.0009589370588843,-.000579547775714673,-.000283653295149336,-.0001051802861973,-2.57634985121267e-005,-2.98859747651745e-006,-4.87591483748876e-008,-.00202276382802748,-.004233896048097,-.00547578967002042,-.0055898138948427,-.00477038710314128,-.00344473958308176,-.00208187924827317,-.00101895293817233,-.000377833656408332,-9.25488720048507e-005,-1.07357828440126e-005,-1.75154945664267e-007,-.00460437499342809,-.00963752901773119,-.0124644254937564,-.0127239764517783,-.0108587323850331,-.00784118866248488,-.00473893818806989,-.00231942126065002,-.000860054849279415,-.000210667061584079,-2.44376379372225e-005,-3.98701539258914e-007,-.00810600549215572,-.0169668767769879,-.0219436300590561,-.0224005697075911,-.0191168061846747,-.0138044182877922,-.00834290409323953,-.00408334279989948,-.00151412718159681,-.000370879513648267,-4.30224791892881e-005,-7.01914347023563e-007,-.0119687547517608,-.0250520909767,-.0324004131004311,-.0330750978875267,-.0282265247763196,-.0203826283040634,-.0123185424813832,-.00602917535482301,-.0022356531730757,-.00054761447492283,-6.35239518067959e-005,-1.03639711130188e-006,-.0153956985051803,-.0322251100971785,-.0416774344435578,-.0425452977913825,-.0363084610152352,-.0262186674070122,-.0158456389156716,-.00775547397560826,-.00287577470912415,-.000704409734165886,-8.17123944937675e-005,-1.33314265253001e-006,-.0175390732465345,-.0367114597744218,-.047479727872512,-.0484684143436297,-.0413632909868219,-.0298688057526838,-.0180516539400008,-.00883518380630904,-.00327613737349354,-.000802476998296004,-9.30883176033672e-005,-1.51874152529913e-006,-.0177100195447144,-.037069271618925,-.0479424936985227,-.0489408164993326,-.0417664423606343,-.0301599249984219,-.0182275961561996,-.00892129679210925,-.00330806856783064,-.000810298419092016,-9.39956120239112e-005,-1.53354408858128e-006,-.0155590470074065,-.0325670188106206,-.0421196324049362,-.0429967037908343,-.0366936941193377,-.0264968477084686,-.0160137612897764,-.00783775962556066,-.00290628670514157,-.000711883527900599,-8.25793637481591e-005,-1.34728730829027e-006,-.0111823253052378,-.0234059964205263,-.0302714832768886,-.0309018366364716,-.0263718481021364,-.0190433495379502,-.0115091295769889,-.00563301709645563,-.00208875539431883,-.000511632440253917,-5.93499915831539e-005,-9.68298698098172e-007,-.00513361662858989,-.0107452975256271,-.0138971265528759,-.0141865111307979,-.012106869926272,-.00874248004628156,-.005283646948578,-.0025860229823533,-.000958912313193877,-.000234881809580631,-2.72465784513743e-005,-4.44529573439469e-007},
			{-5.68594172031225e-007,-1.19013825765391e-006,-1.53923164459519e-006,-1.57128358699498e-006,-1.34094463604488e-006,-9.68308224601569e-007,-5.85211378134644e-007,-2.86425283165094e-007,-1.06208155422931e-007,-2.60152710469109e-008,-3.0178033998418e-009,-4.92356447782966e-011,-3.48508775138951e-005,-7.29472173341577e-005,-9.4344219744047e-005,-9.63087814184433e-005,-8.21906019486044e-005,-5.93505754917118e-005,-3.58693969483533e-005,-1.75558824755616e-005,-6.50982299448902e-006,-1.59455560634358e-006,-1.84970409164038e-007,-3.0178033998418e-009,-.000300435417626676,-.000628848662297057,-.000813303626233361,-.000830239323386619,-.00070853216856797,-.000511637473895075,-.000309215664595502,-.000151342211720415,-5.61185694462147e-005,-1.37460234489013e-005,-1.59455560634358e-006,-2.60152710469109e-008,-.00122653696255208,-.00256729427659293,-.00332033741971269,-.00338947793153767,-.00289260467614267,-.00208877594431585,-.00126238259464391,-.000617859299465599,-.000229105810010919,-5.61185694462147e-005,-6.50982299448902e-006,-1.06208155422931e-007,-.0033077610227998,-.00692355485520927,-.00895438379338928,-.00914084396303245,-.00780086152659004,-.00563307251629313,-.00340443057968316,-.00166626116517048,-.000617859299465599,-.000151342211720415,-1.75558824755616e-005,-2.86425283165094e-007,-.0067582699589302,-.0141458988313976,-.018295198043168,-.0186761651548639,-.0159383727377934,-.0115092428083319,-.00695578089086413,-.00340443057968316,-.00126238259464391,-.000309215664595502,-3.58693969483533e-005,-5.85211378134644e-007,-.0111824353213518,-.023406226697924,-.030271781099646,-.0309021406608933,-.0263721075587493,-.0190435368939121,-.0115092428083319,-.00563307251629313,-.00208877594431585,-.000511637473895075,-5.93505754917118e-005,-9.68308224601569e-007,-.0154857991299779,-.0324137019011133,-.0419213443176089,-.0427942866834367,-.036520949914119,-.0263721075587493,-.0159383727377934,-.00780086152659004,-.00289260467614267,-.00070853216856797,-8.21906019486044e-005,-1.34094463604488e-006,-.0181458513277659,-.037981521698904,-.0491223265304294,-.0501452173903112,-.0427942866834367,-.0309021406608933,-.0186761651548639,-.00914084396303246,-.00338947793153767,-.000830239323386619,-9.63087814184433e-005,-1.57128358699498e-006,-.0177757018611982,-.0372067528692506,-.0481203011840638,-.0491223265304294,-.0419213443176089,-.030271781099646,-.018295198043168,-.00895438379338928,-.00332033741971269,-.000813303626233361,-9.4344219744047e-005,-1.53923164459519e-006,-.0137442229153402,-.0287683664692428,-.0372067528692506,-.037981521698904,-.0324137019011133,-.023406226697924,-.0141458988313976,-.00692355485520927,-.00256729427659293,-.000628848662297057,-7.29472173341577e-005,-1.19013825765392e-006,-.00656636739345367,-.0137442229153402,-.0177757018611982,-.0181458513277659,-.0154857991299779,-.0111824353213518,-.0067582699589302,-.0033077610227998,-.00122653696255208,-.000300435417626676,-3.48508775138952e-005,-5.68594172031225e-007},
			{-6.20272325817235e-009,-1.29830705524454e-008,-1.67912869868831e-008,-1.71409376487648e-008,-1.46281986187158e-008,-1.05631542517567e-008,-6.38399829730143e-009,-3.1245778676732e-009,-1.15861158670691e-009,-2.83797363265043e-010,-3.2920842768966e-011,-5.37105538571509e-013,-1.9769158918718e-006,-4.13793062049084e-006,-5.35167549924359e-006,-5.46311525260784e-006,-4.66226157691217e-006,-3.36666116468723e-006,-2.03469140284302e-006,-9.95857365368599e-007,-3.69269684126226e-007,-9.045116058834e-008,-1.04924457427685e-008,-1.71184886157124e-010,-4.08974964028785e-005,-8.56035420437903e-005,-.000110712919239298,-.000113018331892963,-9.64506516716335e-005,-6.96478861031055e-005,-4.20927287149029e-005,-2.06018238739433e-005,-7.6392757225221e-006,-1.87121062155825e-006,-2.1706273078516e-007,-3.54139156583407e-009,-.000299454081965167,-.000626794604813159,-.00081064706909918,-.000827527447862548,-.000706217834626139,-.000509966272467282,-.000308205649327789,-.000150847870838019,-5.59352649812969e-005,-1.37011237392702e-005,-1.58934718486943e-006,-2.59302953359118e-008,-.00123708915628192,-.00258938132932369,-.00334890308448347,-.00341863842883094,-.00291749045281179,-.00210674618825924,-.0012732431770043,-.000623174892248206,-.000231076862629534,-5.6601371053274e-005,-6.56582857400444e-006,-1.07121889836103e-007,-.00349753239972921,-.00732076992880828,-.00946811067096992,-.00966526834956002,-.00824840904375729,-.00595625061785239,-.00359974803892942,-.0017618571509321,-.000653306845161927,-.000160024948988241,-1.85630906649181e-005,-3.02857945621352e-007,-.0074441143400151,-.0155814563465409,-.0201518357411283,-.0205714642490873,-.0175558345219898,-.0126772265613641,-.00766166912395449,-.00374991982442457,-.00139049200941683,-.000340595563207872,-3.95095037359528e-005,-6.44599940278459e-007,-.0125314188045125,-.0262298167578035,-.033923591419649,-.0346299938653073,-.0295534841096122,-.0213408376153209,-.0128976504320703,-.00631261338244207,-.00234075363682485,-.000573358421239403,-6.65102812046103e-005,-1.08511925583571e-006,-.0170402239900308,-.0356673062917253,-.0461293015064619,-.047089867591508,-.040186829342282,-.0290192721808159,-.0175382257776313,-.0085838920298908,-.00318295692604349,-.00077965281321311,-9.04406840954337e-005,-1.47554522466982e-006,-.0186127337638557,-.0389587646541599,-.0503862160588214,-.0514354253188647,-.0438953593449828,-.0316972351677059,-.0191566922641638,-.00937603268029293,-.00347668727130158,-.000851600909064724,-9.87867515986945e-005,-1.61171182018355e-006,-.015482742876017,-.0324073047817918,-.0419130707843193,-.0427858408675583,-.0365137421945241,-.0263669027993754,-.0159352271645878,-.00779932196032425,-.00289203379604635,-.000708392333727775,-8.2174380935362e-005,-1.34067998923389e-006,-.00769779899133229,-.0161124498455309,-.0208385811603797,-.0212725100010394,-.0181541119739297,-.0131092481092602,-.00792276772768902,-.0038777116690518,-.00143787796622194,-.000352202567445854,-4.08559304861565e-005,-6.66566974047622e-007},
			{-.00404945140720489,-.00814474566006138,-.00979128722597762,-.00896406350610278,-.00659219233567233,-.00391599979728114,-.00183988793598998,-.000650774647452403,-.00015752876308832,-2.15142568085075e-005,-1.03996283213943e-006,-3.26296210833654e-009,-.00917949010540939,-.0184628989656689,-.0221953581292932,-.0203201678410382,-.0149434968427065,-.00887697561402263,-.00417074596164616,-.00147520709262294,-.000357093733613841,-4.87695462026171e-005,-2.35743748168761e-006,-7.3966385507205e-009,-.0137408542244633,-.0276372652876165,-.0332244141027671,-.030417426340031,-.0223690433086643,-.0132880177946399,-.00624322392727204,-.00220824962801375,-.000534536546334486,-7.3003534757174e-005,-3.5288675522474e-006,-1.10720890713312e-008,-.017439548481424,-.0350765258115668,-.0421675953362895,-.0386050366790666,-.0283902302500237,-.0168648198114988,-.00792374364660721,-.00280265519288241,-.000678420414234276,-9.26542602749126e-005,-4.47875042967689e-006,-1.40524185029455e-008,-.0200426049296945,-.0403121072713726,-.0484616018046971,-.044367289627886,-.0326278039463224,-.0193820912882233,-.00910645500038529,-.00322098423849287,-.000779682590591189,-.000106483991585058,-5.14725627996212e-006,-1.61499062123808e-008,-.0213863615446234,-.043014832840112,-.0517107103024967,-.0473418949315912,-.0348153353344463,-.0206815637605425,-.00971699735195308,-.00343693515366777,-.000831956416390127,-.000113623211690348,-5.49235412024762e-006,-1.72326768092913e-008,-.0213863615446234,-.043014832840112,-.0517107103024967,-.0473418949315912,-.0348153353344463,-.0206815637605425,-.00971699735195308,-.00343693515366777,-.000831956416390127,-.000113623211690348,-5.49235412024762e-006,-1.72326768092913e-008,-.0200426049296945,-.0403121072713726,-.0484616018046971,-.044367289627886,-.0326278039463224,-.0193820912882233,-.00910645500038529,-.00322098423849287,-.000779682590591189,-.000106483991585058,-5.14725627996212e-006,-1.61499062123808e-008,-.017439548481424,-.0350765258115668,-.0421675953362895,-.0386050366790666,-.0283902302500237,-.0168648198114988,-.00792374364660721,-.00280265519288241,-.000678420414234276,-9.26542602749126e-005,-4.47875042967689e-006,-1.40524185029455e-008,-.0137408542244633,-.0276372652876165,-.0332244141027671,-.030417426340031,-.0223690433086643,-.0132880177946399,-.00624322392727204,-.00220824962801375,-.000534536546334486,-7.3003534757174e-005,-3.5288675522474e-006,-1.10720890713312e-008,-.00917949010540939,-.0184628989656689,-.0221953581292932,-.0203201678410382,-.0149434968427065,-.00887697561402263,-.00417074596164616,-.00147520709262294,-.000357093733613841,-4.87695462026171e-005,-2.35743748168761e-006,-7.3966385507205e-009,-.00404945140720489,-.00814474566006138,-.00979128722597762,-.00896406350610278,-.00659219233567233,-.00391599979728114,-.00183988793598998,-.000650774647452403,-.00015752876308832,-2.15142568085075e-005,-1.03996283213943e-006,-3.26296210833654e-009},
			{-6.46655241827284e-005,-.000130063110895906,-.000156356665933801,-.000143146764126628,-.000105270450249422,-6.25344408726795e-005,-2.93810953272788e-005,-1.03921937740666e-005,-2.51557038586478e-006,-3.43560288548878e-007,-1.66071240047969e-008,-5.21061086814239e-011,-.000762236326271721,-.0015331017429415,-.00184303355050058,-.00168732357751008,-.00124086152980604,-.00073711646319537,-.000346325780965151,-.000122496611669995,-2.96519537053663e-005,-4.04967152908804e-006,-1.95754280991916e-007,-6.14193874705368e-010,-.00273814180028246,-.00550728143194998,-.00662063329969093,-.00606128448480099,-.00445748215614846,-.00264790502628461,-.00124408804027283,-.000440038189267583,-.00010651716679239,-1.45474238225389e-005,-7.03197894004764e-007,-2.20633662270343e-009,-.00623277490876185,-.0125361095327521,-.0150704091023226,-.013797175094379,-.0101464733989415,-.00602737082751107,-.00283189158461527,-.0010016497263513,-.00024246279884684,-3.31139965720151e-005,-1.6006746579724e-006,-5.02223790632381e-009,-.0109728047159291,-.0220698298613373,-.0265314660788539,-.0242899366908432,-.0178628737266676,-.0106111906829482,-.00498554717433864,-.00176340506466325,-.000426855931999716,-5.82972148148455e-005,-2.81798888822008e-006,-8.84165345126754e-009,-.0162016678512013,-.0325867508082808,-.039174487484589,-.0358648036285977,-.0263750567407152,-.0156677341301185,-.00736130656347784,-.00260371927546702,-.000630265297671321,-8.60775467651508e-005,-4.16084320803041e-006,-1.30549605302744e-008,-.0208405969285544,-.0419171251406854,-.0503910900437627,-.0461337637093912,-.0339268729336555,-.0201537850786976,-.00946902654504109,-.00334922703227079,-.00081072548502156,-.00011072362877741,-5.3521931802253e-006,-1.67929112501503e-008,-.0237420053340501,-.0477527880842341,-.0574064904526992,-.0525564631293,-.0386501404407882,-.0229595761810517,-.0107872955612227,-.00381550328609678,-.000923593928514568,-.000126138468780461,-6.09732050715256e-006,-1.91308046425978e-008,-.0239734091183992,-.0482182153183887,-.0579660084441261,-.0530687098535765,-.0390268478266128,-.0231833539428959,-.0108924349957666,-.00385269146321899,-.000932595827353905,-.000127367889742049,-6.15674863969202e-006,-1.93172649069964e-008,-.0210617158529497,-.0423618662225568,-.0509257399709112,-.0466232433651179,-.0342868373663876,-.020367616923072,-.00956949300345818,-.00338476236177552,-.000819327289847131,-.00011189840749341,-5.40897999891501e-006,-1.69710841924047e-008,-.0151371069219119,-.0304455773166633,-.0366004544168802,-.0335082395371539,-.0246420342413141,-.0146382563159452,-.00687761812917763,-.00243263702412373,-.00058885253590198,-8.04216603455029e-005,-3.88744721245458e-006,-1.21971598987849e-008,-.00694919005501226,-.0139770501853164,-.0168026502788677,-.0153830666688022,-.0113127416069928,-.00672017616961103,-.00315739828964731,-.00111678255975218,-.000270332250902931,-3.69202255862743e-005,-1.78466133902173e-006,-5.5995100454324e-009},
			{-7.69685243564879e-007,-1.54808390489254e-006,-1.86104450591298e-006,-1.70381286481175e-006,-1.25298778853868e-006,-7.44319897852733e-007,-3.49710232755209e-007,-1.23693703828388e-007,-2.99417259756176e-008,-4.08924674643852e-009,-1.97667280147994e-010,-6.20196054366658e-013,-4.71763649140235e-005,-9.48868018782288e-005,-.000114069115221109,-.000104431906584671,-7.67994574913963e-005,-4.56216452211473e-005,-2.14348107782544e-005,-7.58156578700641e-006,-1.83522004948295e-006,-2.50642452023579e-007,-1.21156327444106e-008,-3.80137148576797e-011,-.000406688494124778,-.00081798100889079,-.000983344027882673,-.000900265522892682,-.000662057277488362,-.000393285880086281,-.00018478089469472,-6.53576336083714e-005,-1.58206949533316e-005,-2.16068791147813e-006,-1.04444003796667e-007,-3.27700968053241e-010,-.00166031846121593,-.00333943298029159,-.00401453266272909,-.00367536207502424,-.00270286948383404,-.00160560187139806,-.000754373765606237,-.000266824577604282,-6.45884805686753e-005,-8.82107578694501e-006,-4.263959029382e-007,-1.33784942253663e-009,-.00447759575057386,-.00900588126384461,-.0108265099805079,-.00991182474529423,-.00728917806907233,-.00433003443882712,-.00203441740011081,-.000719580383364912,-.000174184117617127,-2.37889371116125e-005,-1.14991703559086e-006,-3.60795174491446e-009,-.00914842415783806,-.0184004153805573,-.0221202428642781,-.0202513987414977,-.0148929238932162,-.00884693345962729,-.00415663099731916,-.0014702145815281,-.000355885229102249,-4.86044964941899e-005,-2.34945925758498e-006,-7.37160626866756e-009,-.0151372558449131,-.0304458768485863,-.0366008145021687,-.0335085692003862,-.0246422766763918,-.0146384003311175,-.00687768579306728,-.00243266095707262,-.000588858329194405,-8.04224515558099e-005,-3.8874854582243e-006,-1.21972798977825e-008,-.0209625628637279,-.0421624377574542,-.05068599505261,-.0464037534631962,-.0341254240019886,-.0202717315680825,-.00952444236070643,-.00336882779555505,-.000815470113608189,-.000111371619378443,-5.3835159512905e-006,-1.68911888154202e-008,-.0245633787434188,-.0494048334697752,-.0593925132894224,-.0543746954436094,-.0399872725481971,-.023753880836457,-.0111604905634277,-.00394750363309629,-.00095554638928031,-.000130502328643898,-6.30826212149847e-006,-1.97926499253432e-008,-.0240623208776429,-.0483970453850838,-.0581809907923459,-.0532655292683946,-.0391715892641307,-.0232693355725446,-.0109328324858832,-.00386698019345867,-.000936054606840123,-.000127840267412333,-6.179582578331e-006,-1.93889081179464e-008,-.0186050545055932,-.0374207322675915,-.0449857065070876,-.0411850577650857,-.030287583522627,-.0179919160266198,-.00845329697557845,-.00298996001413125,-.000723760067411173,-9.88464560551176e-005,-4.77807071380575e-006,-1.49915585521718e-008,-.00888865263692734,-.0178779315290952,-.021492133691307,-.0196763558094162,-.0144700360358322,-.00859572283356178,-.00403860254373974,-.00142846751434807,-.000345779789560036,-4.72243611004261e-005,-2.28274584398164e-006,-7.16228788345407e-009},
			{-.00508628631126404,-.00983033824354329,-.0109846550070609,-.00901916015333691,-.00571557221193217,-.00279307452781809,-.00102018939735755,-.000260771859326372,-4.12071599264361e-005,-3.13787157581542e-006,-6.32055938620332e-008,-3.81377664478188e-011,-.0115298370501388,-.0222838808434348,-.0249005412854803,-.0204450635558571,-.012956332423902,-.00633147491192501,-.00231261411410569,-.000591130121526243,-9.34103607574427e-005,-7.11307734943901e-006,-1.43277462826387e-007,-8.64525128335375e-011,-.0172591078936303,-.0333569244815142,-.0372738250147888,-.0306043837625979,-.019394440549112,-.00947763686126095,-.00346177108471816,-.000884867539951351,-.000139826737158981,-1.06476239773248e-005,-2.14473212317988e-007,-1.29411477385236e-010,-.0219048280361805,-.0423357741945106,-.0473070063778303,-.0388423183634187,-.0246149388313989,-.0120287796399897,-.00439359327137407,-.00112305174849657,-.000177464597313172,-1.35136980227866e-005,-2.72204036451189e-007,-1.64245810125438e-010,-.0251743796492038,-.0486548833232205,-.0543681300147957,-.0446399883770301,-.0282890061570013,-.0138242155963766,-.00504938842044633,-.00129068034844452,-.000203953262708829,-1.55307754038705e-005,-3.12833670474356e-007,-1.88761417029128e-010,-.0268621961430683,-.051916950382044,-.0580132417457791,-.0476328767706495,-.0301856428110978,-.0147510602465105,-.00538792470927311,-.00137721402318714,-.000217627310910781,-1.65720363705528e-005,-3.33807606524435e-007,-2.01416927810585e-010,-.0268621961430683,-.051916950382044,-.0580132417457791,-.0476328767706495,-.0301856428110978,-.0147510602465105,-.00538792470927311,-.00137721402318714,-.000217627310910781,-1.65720363705528e-005,-3.33807606524435e-007,-2.01416927810585e-010,-.0251743796492038,-.0486548833232205,-.0543681300147957,-.0446399883770301,-.0282890061570013,-.0138242155963766,-.00504938842044633,-.00129068034844452,-.000203953262708829,-1.55307754038705e-005,-3.12833670474356e-007,-1.88761417029128e-010,-.0219048280361805,-.0423357741945106,-.0473070063778303,-.0388423183634187,-.0246149388313989,-.0120287796399897,-.00439359327137407,-.00112305174849657,-.000177464597313172,-1.35136980227866e-005,-2.72204036451189e-007,-1.64245810125438e-010,-.0172591078936303,-.0333569244815142,-.0372738250147888,-.0306043837625979,-.019394440549112,-.00947763686126095,-.00346177108471816,-.000884867539951351,-.000139826737158981,-1.06476239773248e-005,-2.14473212317988e-007,-1.29411477385236e-010,-.0115298370501388,-.0222838808434348,-.0249005412854803,-.0204450635558571,-.012956332423902,-.00633147491192501,-.00231261411410569,-.000591130121526243,-9.34103607574427e-005,-7.11307734943901e-006,-1.43277462826387e-007,-8.64525128335375e-011,-.00508628631126404,-.00983033824354329,-.0109846550070609,-.00901916015333691,-.00571557221193217,-.00279307452781809,-.00102018939735755,-.000260771859326372,-4.12071599264361e-005,-3.13787157581542e-006,-6.32055938620331e-008,-3.81377664478188e-011},
			{-8.12226984453021e-005,-.000156980270065097,-.000175413507334088,-.000144026600260793,-9.12717393010351e-005,-4.46024931010705e-005,-1.62913628348369e-005,-4.16425517497517e-006,-6.58035454488694e-007,-5.01085430834115e-008,-1.00932754786986e-009,-6.09020435343665e-013,-.000957401831273778,-.00185038420184216,-.00206766355177068,-.00169769452974626,-.0010758535732373,-.000525745997014592,-.000192032287901912,-4.90856079239649e-005,-7.75650601653937e-006,-5.90647834026672e-007,-1.18973151739896e-008,-7.17874795152048e-012,-.00343922466500658,-.00664703866113062,-.00742755993763223,-.00609853951561999,-.00386473267978856,-.00188860992469107,-.000689827572348814,-.000176327669275857,-2.78632920211386e-005,-2.12175340882166e-006,-4.27381047927422e-008,-2.57878418572552e-011,-.00782863516982087,-.0151305150742709,-.016907199330518,-.0138819779418027,-.00879721016390503,-.00429900326923858,-.00157024007444182,-.000401370985487183,-6.34246288365862e-005,-4.82970290571489e-006,-9.72838540256606e-008,-5.87003250388414e-011,-.0137823178549157,-.0266372826857591,-.0297651365985099,-.0244392321648105,-.0154874948296011,-.0075683983517811,-.00276440878199448,-.000706613909031054,-.000111659104747359,-8.50269544402536e-006,-1.71268295079276e-007,-1.03341964508827e-010,-.0203499963670061,-.0393307288068947,-.0439490968079129,-.0360852427727842,-.0228677401605563,-.0111749620480468,-.00408173061038807,-.00104333615238232,-.00016486794165333,-1.25544790953984e-005,-2.52882658732434e-007,-1.52587440258738e-010,-.0261766921577074,-.0505920670327779,-.0565327854119549,-.0464173199082629,-.029415326849682,-.0143746237655383,-.0052504287338376,-.00134206850877865,-.000212073617975246,-1.6149129884536e-005,-3.25289075746179e-007,-1.96276912228913e-010,-.0298209867484436,-.0576354472700868,-.0644032269037527,-.0528794957568774,-.0335105011321007,-.0163758454369784,-.00598138850975171,-.0015289100309035,-.000241598308649254,-1.83973966375979e-005,-3.7057551652431e-007,-2.23602400308645e-010,-.0301116399215653,-.0581971968115976,-.0650309392734592,-.0533948909503518,-.033837114519146,-.0165354542211866,-.00603968669970425,-.00154381170252316,-.000243953070268046,-1.85767086689201e-005,-3.74187367153787e-007,-2.25781763041192e-010,-.026454427101384,-.0511288493276215,-.0571325987832424,-.046909808097858,-.0297274237371193,-.0145271386554198,-.00530613583081857,-.00135630786795552,-.000214323721005407,-1.63204723006016e-005,-3.28740395820796e-007,-1.98359412066345e-010,-.0190128617434316,-.0367464296104888,-.0410613390925373,-.0337141943145245,-.0213651724656962,-.0104406902377677,-.00381353285620554,-.000974781796488713,-.000154034984775778,-1.17295635339088e-005,-2.36266530031746e-007,-1.42561396724728e-010,-.0087284836148921,-.0168696650240853,-.0188505670693449,-.0154776170276039,-.00980838972652806,-.00479314449861828,-.00175072850680893,-.000447505854382295,-7.0714859176842e-005,-5.38484445409324e-006,-1.08465972348532e-007,-6.54475287423684e-011},
			{-.00610300723148638,-.011334382523687,-.0117725765312941,-.00866893492130421,-.00473399758463878,-.00190309670559272,-.000540391748990977,-9.98226951437858e-005,-1.02973144060446e-005,-4.37202313877829e-007,-3.66970948242474e-009,-4.25830680908177e-013,-.0138345886544025,-.0256933203450701,-.026686639477119,-.0196511562511815,-.0107312521172389,-.00431403062339345,-.00122498585958623,-.000226282859139225,-2.33424447406373e-005,-9.91071113277978e-007,-8.31867295004696e-009,-9.65293351840578e-013,-.0207091095227104,-.0384605425083467,-.0399474500854692,-.0294159773896932,-.0160636995405713,-.00645770791571394,-.00183369140663751,-.000338724672636725,-3.49414974841223e-005,-1.48354249933819e-006,-1.24522899458459e-008,-1.44495555626435e-012,-.0262834838088483,-.0488131585372653,-.0507003044420999,-.0373340227205607,-.0203876456552513,-.00819596135984531,-.00232727526714406,-.000429900881983438,-4.43468746144641e-005,-1.88287503228147e-006,-1.58041349298872e-008,-1.83390144930814e-012,-.030206600974669,-.0560990929882762,-.0582679174767988,-.0429065619801723,-.0234307400647473,-.00941930438906687,-.00267464830248913,-.000494068613398971,-5.09661639870523e-005,-2.16391613832206e-006,-1.81630867905044e-008,-2.10763267567556e-012,-.0322318027893339,-.0598602571462754,-.0621744904907474,-.0457832327865349,-.0250016542281072,-.0100508217305121,-.00285397012027107,-.0005271934477113,-5.43831908772887e-005,-2.30899591389123e-006,-1.93808310960915e-008,-2.24893892602145e-012,-.0322318027893339,-.0598602571462754,-.0621744904907474,-.0457832327865349,-.0250016542281072,-.0100508217305121,-.00285397012027107,-.0005271934477113,-5.43831908772887e-005,-2.30899591389123e-006,-1.93808310960915e-008,-2.24893892602145e-012,-.030206600974669,-.0560990929882762,-.0582679174767988,-.0429065619801723,-.0234307400647473,-.00941930438906688,-.00267464830248913,-.000494068613398971,-5.09661639870523e-005,-2.16391613832206e-006,-1.81630867905044e-008,-2.10763267567556e-012,-.0262834838088483,-.0488131585372653,-.0507003044420999,-.0373340227205607,-.0203876456552513,-.00819596135984532,-.00232727526714406,-.000429900881983438,-4.43468746144642e-005,-1.88287503228147e-006,-1.58041349298872e-008,-1.83390144930814e-012,-.0207091095227104,-.0384605425083467,-.0399474500854692,-.0294159773896932,-.0160636995405713,-.00645770791571394,-.00183369140663751,-.000338724672636725,-3.49414974841223e-005,-1.48354249933819e-006,-1.24522899458459e-008,-1.44495555626435e-012,-.0138345886544025,-.0256933203450701,-.026686639477119,-.0196511562511815,-.0107312521172389,-.00431403062339345,-.00122498585958623,-.000226282859139225,-2.33424447406373e-005,-9.91071113277978e-007,-8.31867295004696e-009,-9.65293351840578e-013,-.00610300723148638,-.011334382523687,-.0117725765312941,-.0086689349213042,-.00473399758463878,-.00190309670559272,-.000540391748990977,-9.98226951437858e-005,-1.02973144060446e-005,-4.37202313877829e-007,-3.66970948242474e-009,-4.25830680908177e-013},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{-1.25649763607893e-005,-2.84828893052902e-005,-4.2636271224355e-005,-5.41128889760362e-005,-6.2189870139491e-005,-6.63593905024702e-005,-6.63593905024702e-005,-6.2189870139491e-005,-5.41128889760362e-005,-4.2636271224355e-005,-2.84828893052902e-005,-1.25649763607893e-005,-.000148107999463138,-.000335738296102257,-.000502569416311285,-.000637848532403299,-.000733054881183536,-.000782202551815753,-.000782202551815753,-.000733054881183536,-.000637848532403299,-.000502569416311285,-.000335738296102257,-.000148107999463138,-.000532040641869991,-.00120605517059218,-.00180535390260964,-.00229131001583844,-.0026333148170561,-.00280986543096192,-.00280986543096192,-.0026333148170561,-.00229131001583844,-.00180535390260964,-.00120605517059218,-.000532040641869991,-.00121107298487856,-.00274531815886015,-.00410948932756509,-.00521566106380589,-.00599415943941434,-.00639603790909957,-.00639603790909957,-.00599415943941434,-.00521566106380588,-.00410948932756509,-.00274531815886015,-.00121107298487856,-.00213209486213416,-.00483313459594366,-.00723475892097063,-.00918217505932459,-.010552722010284,-.0112602293456112,-.0112602293456112,-.010552722010284,-.00918217505932459,-.00723475892097063,-.00483313459594367,-.00213209486213416,-.00314810057025839,-.0071362649232201,-.0106823336471961,-.0135577506675935,-.0155814034208121,-.0166260587433132,-.0166260587433132,-.0155814034208121,-.0135577506675935,-.0106823336471961,-.0071362649232201,-.00314810057025839,-.00404947784869215,-.00917955004423722,-.0137409439473876,-.0174396623555152,-.0200427358008248,-.0213865011900097,-.0213865011900097,-.0200427358008248,-.0174396623555152,-.0137409439473876,-.00917955004423722,-.00404947784869215,-.00461324236601103,-.0104575184128172,-.0156539453074288,-.0198675958317378,-.0228330667266135,-.0243639098760258,-.0243639098760258,-.0228330667266135,-.0198675958317378,-.0156539453074288,-.0104575184128172,-.00461324236601103,-.00465820578534291,-.0105594436420292,-.0158065180211977,-.0200612373037489,-.023055611451648,-.0246013751140091,-.0246013751140091,-.023055611451648,-.0200612373037489,-.0158065180211977,-.0105594436420292,-.00465820578534291,-.00409244284577621,-.00927694515432801,-.0138867354885914,-.0176246973329244,-.0202553894113433,-.0216134121636217,-.0216134121636217,-.0202553894113433,-.0176246973329244,-.0138867354885914,-.00927694515432801,-.00409244284577621,-.00294124872640195,-.00666736327135993,-.0099804309086089,-.0126669132684861,-.0145575981275773,-.0155336124156416,-.0155336124156416,-.0145575981275773,-.0126669132684861,-.0099804309086089,-.00666736327135993,-.00294124872640195,-.00135027759956182,-.0030608738365598,-.00458184721642184,-.00581516588124543,-.0066831473580057,-.00713121902844807,-.00713121902844807,-.0066831473580057,-.00581516588124543,-.00458184721642184,-.0030608738365598,-.00135027759956182},
			{-1.49555377634105e-007,-3.39019282157687e-007,-5.07480751318181e-007,-6.4408187594673e-007,-7.40218624107564e-007,-7.89846587943962e-007,-7.89846587943962e-007,-7.40218624107564e-007,-6.4408187594673e-007,-5.07480751318181e-007,-3.39019282157687e-007,-1.49555377634105e-007,-9.16670694821036e-006,-2.07795296865571e-005,-3.11050488640572e-005,-3.94777499870446e-005,-4.53702655975518e-005,-4.84121154335211e-005,-4.84121154335211e-005,-4.53702655975518e-005,-3.94777499870446e-005,-3.11050488640572e-005,-2.07795296865571e-005,-9.16670694821036e-006,-7.90224989069185e-005,-.000179131979588681,-.000268144133301817,-.000340321826892032,-.000391118837314757,-.000417341403029572,-.000417341403029572,-.000391118837314757,-.000340321826892032,-.000268144133301817,-.000179131979588681,-7.90224989069185e-005,-.000322611816370485,-.000731311893505376,-.0010947067871833,-.00138937447237002,-.00159675485169678,-.00170380929406662,-.00170380929406662,-.00159675485169678,-.00138937447237002,-.0010947067871833,-.000731311893505376,-.000322611816370485,-.000870029052744199,-.00197222346386832,-.00295223752124433,-.00374690601758642,-.00430617553540284,-.00459488311014376,-.00459488311014376,-.00430617553540284,-.00374690601758643,-.00295223752124433,-.00197222346386832,-.000870029052744198,-.00177760460022009,-.00402955911756787,-.00603188017934985,-.00765551144808981,-.0087981860110801,-.00938806046568492,-.00938806046568492,-.0087981860110801,-.00765551144808981,-.00603188017934986,-.00402955911756787,-.00177760460022009,-.0029412776627297,-.00666742886561148,-.00998052909718346,-.0126670378869689,-.0145577413468253,-.0155337652370252,-.0155337652370252,-.0145577413468253,-.0126670378869689,-.00998052909718346,-.00666742886561148,-.0029412776627297,-.00407317670628986,-.00923327174798222,-.013821360407493,-.0175417249152225,-.0201600323904989,-.0215116619305183,-.0215116619305183,-.0201600323904989,-.0175417249152225,-.013821360407493,-.00923327174798222,-.00407317670628986,-.00477284112519419,-.0108193045125715,-.0161955058952257,-.0205549309837037,-.0236229946837403,-.0252067985596364,-.0252067985596364,-.0236229946837403,-.0205549309837037,-.0161955058952257,-.0108193045125715,-.00477284112519419,-.00467548197876496,-.0105986061434719,-.0158651406916522,-.0201356397307597,-.0231411193105227,-.0246926158479963,-.0246926158479963,-.0231411193105227,-.0201356397307597,-.0158651406916522,-.0105986061434719,-.00467548197876496,-.00361509587945292,-.0081948722230624,-.0122669716195725,-.015568933545552,-.0178927788504699,-.0190924003579595,-.0190924003579595,-.0178927788504699,-.015568933545552,-.0122669716195725,-.00819487222306241,-.00361509587945292,-.00172712912568916,-.00391513889802946,-.00586060361181134,-.00743813151271418,-.00854835957956341,-.00912148441898038,-.00912148441898038,-.00854835957956341,-.00743813151271418,-.00586060361181134,-.00391513889802946,-.00172712912568916},
			{-1.63148105416827e-009,-3.69831927536025e-009,-5.53604453566559e-009,-7.02620925147273e-009,-8.07495310619039e-009,-8.61633840464567e-009,-8.61633840464567e-009,-8.07495310619039e-009,-7.02620925147273e-009,-5.53604453566559e-009,-3.69831927536025e-009,-1.63148105416827e-009,-5.19981416069715e-007,-1.1787187408438e-006,-1.7644337761237e-006,-2.23937521483844e-006,-2.57362813998106e-006,-2.74617706012381e-006,-2.74617706012381e-006,-2.57362813998106e-006,-2.23937521483844e-006,-1.7644337761237e-006,-1.1787187408438e-006,-5.19981416069715e-007,-1.07571284042538e-005,-2.43847731013085e-005,-3.6501767378587e-005,-4.63271301374563e-005,-5.3241995792529e-005,-5.68116058451738e-005,-5.68116058451738e-005,-5.3241995792529e-005,-4.63271301374563e-005,-3.6501767378587e-005,-2.43847731013085e-005,-1.07571284042538e-005,-7.87643815441602e-005,-.000178546866806921,-.000267268273167243,-.000339210207117138,-.000389841295295595,-.000415978208195064,-.000415978208195064,-.000389841295295595,-.000339210207117138,-.000267268273167243,-.000178546866806921,-7.87643815441602e-005,-.000325387323726201,-.000737603546311472,-.00110412481400688,-.0014013275964412,-.00161049211924643,-.00171846757683388,-.00171846757683388,-.00161049211924644,-.0014013275964412,-.00110412481400688,-.000737603546311472,-.000325387323726201,-.000919943967994991,-.00208537298082308,-.00312161196363602,-.0039618718233036,-.00455322750019265,-.00485849867597654,-.00485849867597654,-.00455322750019265,-.0039618718233036,-.00312161196363602,-.00208537298082308,-.000919943967994991,-.00195799989864057,-.00443848780701132,-.00664400889731993,-.00843240990574942,-.00969104564411167,-.0103407818802712,-.0103407818802712,-.00969104564411167,-.00843240990574942,-.00664400889731993,-.00443848780701132,-.00195799989864057,-.00329609616783617,-.00747174842135327,-.0111845216543321,-.0141951151250118,-.0163139019731612,-.0174076676672231,-.0174076676672231,-.0163139019731612,-.0141951151250118,-.0111845216543321,-.00747174842135327,-.00329609616783617,-.00448203175305139,-.0101600839205191,-.0152087131700155,-.0193025183395333,-.022183644813943,-.0236709474657956,-.0236709474657956,-.022183644813943,-.0193025183395333,-.0152087131700155,-.0101600839205191,-.00448203175305139,-.00489564361298881,-.0110976790646466,-.0166122070513835,-.0210837976681447,-.0242308009023486,-.0258553551512484,-.0258553551512484,-.0242308009023486,-.0210837976681447,-.0166122070513835,-.0110976790646466,-.00489564361298881,-.00407237283003069,-.00923144948283444,-.0138186326438083,-.0175382629057834,-.0201560536356863,-.021507416420056,-.021507416420056,-.0201560536356863,-.0175382629057834,-.0138186326438083,-.00923144948283444,-.00407237283003069,-.00202472570360245,-.0045897450527047,-.00687042711223165,-.00871977424071199,-.0100213024648473,-.0106931807723117,-.0106931807723117,-.0100213024648473,-.00871977424071199,-.00687042711223165,-.0045897450527047,-.00202472570360245},
			{-1.70557276538753e-011,-3.86627390961513e-011,-5.7874572088226e-011,-7.34529592765333e-011,-8.44166719834875e-011,-9.00763884617597e-011,-9.00763884617597e-011,-8.44166719834875e-011,-7.34529592765333e-011,-5.7874572088226e-011,-3.86627390961513e-011,-1.70557276538753e-011,-2.82664008807978e-008,-6.40756292912074e-008,-9.59153363989562e-008,-1.21733345825304e-007,-1.39903470536826e-007,-1.49283299887593e-007,-1.49283299887593e-007,-1.39903470536826e-007,-1.21733345825304e-007,-9.59153363989562e-008,-6.40756292912074e-008,-2.82664008807978e-008,-1.40329882934204e-006,-3.18106489584209e-006,-4.76176220142831e-006,-6.04350947999848e-006,-6.94557390780468e-006,-7.41123996847038e-006,-7.41123996847038e-006,-6.94557390780468e-006,-6.04350947999848e-006,-4.76176220142831e-006,-3.18106489584208e-006,-1.40329882934204e-006,-1.84284021471628e-005,-4.17743832824876e-005,-6.2532417858728e-005,-7.93645806216638e-005,-9.12106719107567e-005,-9.73258921709034e-005,-9.73258921709034e-005,-9.12106719107567e-005,-7.93645806216638e-005,-6.2532417858728e-005,-4.17743832824876e-005,-1.84284021471628e-005,-.000116620720789999,-.000264361427000411,-.000395724793999518,-.000502244010271906,-.000577209799147466,-.000615908834952788,-.000615908834952788,-.000577209799147466,-.000502244010271906,-.000395724793999518,-.000264361427000411,-.000116620720789999,-.000456242568387133,-.00103423247275538,-.0015481510932686,-.00196487464364187,-.00225815515010815,-.00240955318100971,-.00240955318100971,-.00225815515010815,-.00196487464364187,-.0015481510932686,-.00103423247275538,-.000456242568387133,-.00124910090182185,-.00283152165958353,-.00423852805667493,-.00537943379114375,-.00618237716052034,-.00659687468888933,-.00659687468888933,-.00618237716052034,-.00537943379114375,-.00423852805667493,-.00283152165958353,-.00124910090182185,-.0025560815986996,-.00579424800656579,-.00867345748885218,-.0110081352954835,-.012651228153929,-.0134994298511859,-.0134994298511859,-.012651228153929,-.0110081352954835,-.00867345748885218,-.00579424800656579,-.0025560815986996,-.00403349103971441,-.00914331038111467,-.0136866964976499,-.0173708128492007,-.0199636097009642,-.021302070080123,-.021302070080123,-.0199636097009642,-.0173708128492007,-.0136866964976499,-.00914331038111467,-.00403349103971441,-.00491248705999991,-.0111358605958299,-.0166693612993899,-.021156336410114,-.0243141668994061,-.025944310422275,-.025944310422275,-.0243141668994061,-.021156336410114,-.0166693612993899,-.0111358605958299,-.00491248705999991,-.00439626090995001,-.00996565447158663,-.0149176701290573,-.0189331337918146,-.0217591253050265,-.0232179660428578,-.0232179660428578,-.0217591253050265,-.0189331337918146,-.0149176701290573,-.00996565447158663,-.00439626090995001,-.00227465638852363,-.00515629988163542,-.00771850769460683,-.00979613690280579,-.0112583248350307,-.012013139317637,-.012013139317637,-.0112583248350307,-.00979613690280579,-.00771850769460683,-.00515629988163542,-.00227465638852363},
			{-1.7384464750039e-013,-3.94079360668192e-013,-5.8990063561595e-013,-7.48687131527224e-013,-8.60437436733519e-013,-9.18125471866621e-013,-9.18125471866621e-013,-8.60437436733519e-013,-7.48687131527224e-013,-5.8990063561595e-013,-3.94079360668192e-013,-1.7384464750039e-013,-1.49815262263485e-009,-3.39608401063939e-009,-5.0836260825345e-009,-6.45201106711068e-009,-7.41504913145304e-009,-7.91219116240981e-009,-7.91219116240981e-009,-7.41504913145304e-009,-6.45201106711068e-009,-5.0836260825345e-009,-3.39608401063939e-009,-1.49815262263485e-009,-1.78487097219834e-007,-4.04603087706546e-007,-6.05653689159397e-007,-7.68680513053121e-007,-8.83415064139356e-007,-9.42643634493817e-007,-9.42643634493817e-007,-8.83415064139356e-007,-7.68680513053121e-007,-6.05653689159397e-007,-4.04603087706546e-007,-1.78487097219834e-007,-4.20386100245681e-006,-9.52951316020476e-006,-1.42648066135301e-005,-1.81045357479959e-005,-2.0806849318325e-005,-2.22018447046734e-005,-2.22018447046734e-005,-2.0806849318325e-005,-1.81045357479959e-005,-1.42648066135301e-005,-9.52951316020476e-006,-4.20386100245681e-006,-4.07524446402028e-005,-9.23795904009122e-005,-.00013828376853596,-.000175506300131138,-.000201702666783354,-.000215225823762682,-.000215225823762682,-.000201702666783354,-.000175506300131138,-.00013828376853596,-9.23795904009122e-005,-4.07524446402028e-005,-.000220614007696907,-.000500098382997093,-.000748601381966019,-.000950106149209752,-.00109192059703656,-.00116512842258577,-.00116512842258577,-.00109192059703656,-.000950106149209752,-.00074860138196602,-.000500098382997093,-.000220614007696907,-.000776935976612406,-.0017611956269338,-.0026363482167924,-.00334598721372161,-.00384541491403083,-.00410323078906309,-.00410323078906309,-.00384541491403083,-.00334598721372161,-.0026363482167924,-.0017611956269338,-.000776935976612406,-.00193264642090617,-.00438101533121211,-.00655797787568285,-.00832322148498332,-.00956555957532708,-.0102068825969568,-.0102068825969568,-.00956555957532708,-.00832322148498332,-.00655797787568285,-.00438101533121211,-.00193264642090617,-.00353907786161354,-.00802255094483993,-.0120090224811522,-.015241550951156,-.0175165305773379,-.0186909265162125,-.0186909265162125,-.0175165305773379,-.015241550951156,-.0120090224811522,-.00802255094483993,-.00353907786161354,-.00480613424305024,-.0108947749442933,-.0163084782050857,-.0206983126135993,-.0237877776977751,-.0253826294522236,-.0253826294522236,-.0237877776977751,-.0206983126135993,-.0163084782050857,-.0108947749442933,-.00480613424305024,-.00462724228856079,-.0104892541067666,-.0157014507286737,-.0199278885241277,-.0229023588080574,-.0244378476456489,-.0244378476456489,-.0229023588080574,-.0199278885241277,-.0157014507286737,-.0104892541067666,-.00462724228856079,-.00249154226883633,-.00564794716718901,-.00845445855931378,-.010730187331938,-.0123317932081052,-.0131585783867511,-.0131585783867511,-.0123317932081052,-.010730187331938,-.00845445855931378,-.00564794716718901,-.00249154226883633},
			{-.0019095808945752,-.00415955383985738,-.00578758817666173,-.00658769779901337,-.00652410992157336,-.0057268264950684,-.00445208652288796,-.00301523747077974,-.00171271584051363,-.000752419091586911,-.000209456341570312,-1.77695599831757e-005,-.00432872929307012,-.0094290755651378,-.0131195816567463,-.0149333084120673,-.0147891643705445,-.012981844171125,-.010092202640724,-.00683508449545443,-.00388246617391187,-.00170561957949523,-.000474805651832001,-4.02808883575737e-005,-.00647971047522447,-.0141144607515462,-.0196388096682036,-.0223537921630579,-.0221380217619632,-.0194326292933208,-.0151071011240822,-.0102314941881775,-.0058116955424022,-.00255315597436325,-.000710740484695346,-6.02967930241717e-005,-.00822388645802562,-.0179137205405699,-.0249250860049403,-.0283708738774972,-.0280970234814972,-.0246634070303329,-.0191735548724589,-.0129855565030636,-.00737605861435863,-.00324040170051977,-.000902054045409203,-7.65271815013742e-005,-.00945139763454551,-.0205875527108991,-.0286454464212646,-.0326055584089547,-.0322908326405259,-.0283447088011335,-.0220354320420534,-.0149238025892903,-.00847702157560611,-.00372406952887607,-.00103669615509983,-8.79497578075961e-005,-.0100850666679019,-.0219678453543708,-.030565980615856,-.0347915983448399,-.0344557717846379,-.0302450800394711,-.0235127977673766,-.015924369058549,-.00904536355795187,-.00397374980154702,-.00110620145747811,-9.38463500544579e-005,-.0100850666679019,-.0219678453543708,-.030565980615856,-.0347915983448399,-.0344557717846379,-.0302450800394712,-.0235127977673766,-.015924369058549,-.00904536355795187,-.00397374980154702,-.00110620145747811,-9.38463500544579e-005,-.00945139763454551,-.0205875527108991,-.0286454464212646,-.0326055584089547,-.0322908326405259,-.0283447088011335,-.0220354320420534,-.0149238025892903,-.00847702157560611,-.00372406952887607,-.00103669615509983,-8.79497578075961e-005,-.00822388645802562,-.0179137205405699,-.0249250860049403,-.0283708738774972,-.0280970234814972,-.0246634070303329,-.0191735548724589,-.0129855565030636,-.00737605861435863,-.00324040170051977,-.000902054045409203,-7.65271815013742e-005,-.00647971047522447,-.0141144607515462,-.0196388096682036,-.0223537921630579,-.0221380217619632,-.0194326292933208,-.0151071011240822,-.0102314941881775,-.0058116955424022,-.00255315597436325,-.000710740484695346,-6.02967930241717e-005,-.00432872929307012,-.0094290755651378,-.0131195816567463,-.0149333084120673,-.0147891643705445,-.012981844171125,-.010092202640724,-.00683508449545443,-.00388246617391187,-.00170561957949523,-.000474805651832001,-4.02808883575737e-005,-.0019095808945752,-.00415955383985738,-.00578758817666173,-.00658769779901337,-.00652410992157336,-.0057268264950684,-.00445208652288796,-.00301523747077974,-.00171271584051363,-.000752419091586911,-.000209456341570312,-1.77695599831757e-005},
			{-3.04940193335369e-005,-6.64237454259384e-005,-9.2421759274531e-005,-.000105198677163819,-.000104183244945397,-9.14514584621857e-005,-7.10951878965703e-005,-4.815020405282e-005,-2.73502893236142e-005,-1.20153497508037e-005,-3.34479976602295e-006,-2.83761377805336e-007,-.000359444225710958,-.000782961126977123,-.00108940928179785,-.00124001551403817,-.00122804624086627,-.00107797198911481,-.00083802513816711,-.000567564171331005,-.000322387923394712,-.000141629348384198,-3.94263855120624e-005,-3.34479976602295e-006,-.00129121274513816,-.00281259042095047,-.00391342814465003,-.00445444305782967,-.004411446462074,-.00387234255465894,-.00301039399647408,-.00203883117125292,-.00115809732300615,-.000508767721494418,-.000141629348384198,-1.20153497508037e-005,-.00293916056463564,-.00640224074681438,-.00890805463205047,-.0101395555630045,-.0100416833114065,-.00881453236290365,-.00685249688850132,-.00464094875074617,-.00263615271349073,-.00115809732300615,-.000322387923394712,-2.73502893236141e-005,-.00517439429092323,-.0112711494458756,-.0156826365956051,-.017850694871518,-.0176783907020889,-.0155179905734142,-.0120638257076051,-.00817039361825585,-.00464094875074617,-.00203883117125292,-.000567564171331005,-4.815020405282e-005,-.00764014486751864,-.0166421825914966,-.0231558726990041,-.0263570743040411,-.0261026621464626,-.0229127680203357,-.0178125923307157,-.0120638257076051,-.00685249688850132,-.00301039399647408,-.00083802513816711,-7.10951878965703e-005,-.00982770299465891,-.0214072416856188,-.0297859586976462,-.0339037416907514,-.0335764851836737,-.0294732472739763,-.0229127680203357,-.0155179905734142,-.00881453236290365,-.00387234255465894,-.00107797198911481,-9.14514584621857e-005,-.0111959066105712,-.0243875378440283,-.0339327319991675,-.0386237888878892,-.0382509720360841,-.0335764851836737,-.0261026621464626,-.0176783907020889,-.0100416833114065,-.004411446462074,-.00122804624086627,-.000104183244945397,-.0113050286127969,-.02462523337432,-.0342634607007844,-.039000239435718,-.0386237888878892,-.0339037416907514,-.0263570743040411,-.017850694871518,-.0101395555630045,-.00445444305782967,-.00124001551403817,-.000105198677163819,-.00993197501349335,-.0216343727160793,-.0301019879923868,-.0342634607007844,-.0339327319991675,-.0297859586976462,-.0231558726990041,-.0156826365956051,-.00890805463205047,-.00391342814465003,-.00108940928179785,-9.2421759274531e-005,-.00713813484023199,-.0155486768161827,-.0216343727160793,-.02462523337432,-.0243875378440283,-.0214072416856188,-.0166421825914966,-.0112711494458756,-.00640224074681438,-.00281259042095047,-.000782961126977123,-6.64237454259384e-005,-.00327699711041027,-.00713813484023199,-.00993197501349335,-.0113050286127969,-.0111959066105712,-.00982770299465891,-.00764014486751864,-.00517439429092323,-.00293916056463564,-.00129121274513816,-.000359444225710958,-3.04940193335369e-005},
			{-3.62956876834475e-007,-7.90612576313682e-007,-1.10005548074599e-006,-1.25213350502815e-006,-1.24004726271932e-006,-1.08850641767933e-006,-8.46214697861095e-007,-5.73110664448679e-007,-3.25538443615425e-007,-1.43013414267412e-007,-3.98116779370311e-008,-3.37748665821831e-009,-2.22467381475193e-005,-4.84590652057195e-005,-6.74257681555381e-005,-7.67470958397088e-005,-7.60062930474401e-005,-6.67178907235581e-005,-5.18670894572505e-005,-3.51277071610772e-005,-1.99532478216857e-005,-8.76572998571608e-006,-2.44017962134497e-006,-2.07016497214321e-007,-.000191780194444643,-.000417746138158903,-.000581250466549428,-.000661605888719569,-.000655219725381629,-.000575148139518477,-.000447125346440885,-.000302821854829854,-.000172008935497438,-7.5565837560655e-005,-2.10358084478815e-005,-1.78460607688856e-006,-.000782948625132055,-.00170546163785514,-.00237297315794306,-.00270102667510736,-.00267495496410619,-.00234806021751757,-.00182540317195597,-.00123627966686315,-.000702231843846736,-.000308499888616729,-8.58793440610684e-005,-7.28570996785344e-006,-.00211147892329134,-.00459933919956203,-.00639950393640168,-.00728420832819645,-.00721389736971712,-.00633231798455615,-.00492280106302518,-.00333403543487275,-.00189379952905124,-.000831971590143711,-.000231601945658244,-1.96483173282813e-005,-.00431407966834182,-.00939716494906272,-.0130751813408912,-.0148827699400637,-.0147391137220966,-.0129379099972572,-.0100580478180592,-.0068119526671359,-.0038693268277857,-.0016998472881189,-.000473198777355153,-4.01445666201286e-005,-.00713820506661555,-.0155488297871431,-.0216345855593206,-.0246254756422405,-.0243877777734532,-.0214074522943001,-.0166423463205833,-.011271260333675,-.00640230373332618,-.00281261809177179,-.000782968829903678,-6.64243989158946e-005,-.00988521790053478,-.021532523808133,-.0299602757900378,-.0341021573849623,-.0337729856666483,-.029645734277678,-.0230468609713296,-.0156088068880401,-.00886611786558599,-.00389500475941537,-.00108428063087214,-9.19866620625105e-005,-.0115832378335861,-.0252312439580584,-.0351066616363676,-.039960009339504,-.0395742945945498,-.034738090181205,-.0270057043390071,-.0182899885770756,-.0103890832686787,-.00456406393315995,-.00127053146952515,-.000107787546507207,-.0113469563151396,-.0247165625951827,-.0343905358485524,-.0391448821859808,-.0387670354712729,-.0340294827249939,-.0264548265171398,-.0179168989163566,-.0101771607988773,-.00447096354344966,-.00124461444104258,-.000105588834409239,-.00877349869072301,-.0191109160505604,-.0265908595098725,-.0302669335343241,-.0299747813866719,-.0263116922143577,-.0204549466275595,-.0138533968774284,-.00786900949156683,-.00345696164727354,-.000962339403243543,-8.16415851719116e-005,-.00419158042503607,-.0091303303785454,-.0127039086840373,-.0144601704064181,-.014320593338432,-.01257053519047,-.00977244733277217,-.00661852576932447,-.003759456439442,-.0016515797496082,-.000459762193746235,-3.90046528002957e-005},
			{-3.95945152473478e-009,-8.62469447627331e-009,-1.20003687174081e-008,-1.36593690107054e-008,-1.35275216933177e-008,-1.18743814217069e-008,-9.2312511197497e-009,-6.25199311274643e-009,-3.55125848055141e-009,-1.56011558760704e-009,-4.34300653798924e-010,-3.68445827925519e-011,-1.26194613503531e-006,-2.74884028563086e-006,-3.82472643681277e-006,-4.35347871350321e-006,-4.31145667798826e-006,-3.78457209223235e-006,-2.9421604480658e-006,-1.99261905231337e-006,-1.13184790520788e-006,-4.97235999402952e-007,-1.38419179545726e-007,-1.174301000065e-008,-2.61065418768862e-005,-5.68667013887132e-005,-7.91241227483076e-005,-9.0062698548559e-005,-8.91933666496246e-005,-7.82934287517701e-005,-6.08660170299574e-005,-4.12223559228577e-005,-2.34151315299593e-005,-1.02865820344585e-005,-2.86355020000406e-006,-2.42933809796949e-007,-.000191153767800758,-.000416381621285402,-.000579351882704146,-.000659444833676081,-.000653079529962769,-.000573269488181101,-.000445664865962291,-.000301832724128607,-.000171447089262513,-7.53190109571436e-005,-2.09670975420157e-005,-1.77877687852613e-006,-.000789684521168946,-.00172013413604819,-.00239338841889015,-.00272426425965937,-.00269796824743438,-.00236826114642384,-.00184110755612247,-.00124691567929258,-.000708273313928056,-.000311153987634738,-8.66181847905149e-005,-7.34839068958039e-006,-.00223261774167844,-.00486321041790597,-.00676665339575356,-.00770211464969917,-.0076277698425577,-.00669561288171989,-.00520522978980597,-.00352531421515612,-.00200244974321189,-.000879703089733842,-.000244889308237957,-2.0775571746115e-005,-.00475188214064504,-.0103508103065028,-.0144020800439593,-.0163931067848593,-.016234872011969,-.0142508780967576,-.0110787610500387,-.00750324488892071,-.00426199478516835,-.00187235160015967,-.000521220049687516,-4.42185272019898e-005,-.00799931630469631,-.0174245495155288,-.0242444552089724,-.027596148748536,-.0273297764014771,-.0239899219175831,-.0186499814768931,-.0126309591445223,-.00717464014604113,-.00315191585985935,-.000877421602303038,-7.44374492353296e-005,-.0108774707577442,-.0236939034040642,-.032967611546778,-.0375252446090086,-.0371630314766057,-.0326214971629358,-.0253602458535457,-.0171755789499336,-.00975607607116152,-.00428597536221463,-.00119311794379991,-.000101220047626196,-.0118812679549556,-.0258804295145632,-.0360099360729874,-.0409881576521328,-.0405925186861703,-.0356318814839124,-.0277005457518707,-.01876057958054,-.0106563884722811,-.00468149470230067,-.00130322152160786,-.000110560858773534,-.00988326696773915,-.0215282741793143,-.0299543628718611,-.03409542703082,-.0337663202773765,-.0296398834369748,-.0230423124750437,-.0156057263557175,-.00886436806094961,-.0038942360467172,-.00108406663876152,-9.19685077033871e-005,-.00491381941200989,-.0107035509528065,-.0148928821041257,-.0169517601570097,-.0167881329718923,-.0147365274132258,-.0114563092049479,-.00775894462383229,-.00440723740391471,-.00193615863497058,-.00053898247521257,-4.57254306619082e-005},
			{-4.13926516217373e-011,-9.01637440362945e-011,-1.25453507524717e-010,-1.42796925104609e-010,-1.41418575087745e-010,-1.24136418981734e-010,-9.65047707354281e-011,-6.53591971617201e-011,-3.71253453125811e-011,-1.63096632460441e-011,-4.54023885618748e-012,-3.85178343554166e-013,-6.85999042669627e-008,-1.49428073991582e-007,-2.07913681993516e-007,-2.36656870434667e-007,-2.34372535522585e-007,-2.05730875518941e-007,-1.5993703651203e-007,-1.08319580712847e-007,-6.15277116719808e-008,-2.70299508117877e-008,-7.5245228000818e-009,-6.38354791449189e-010,-3.40567466501195e-006,-7.41842734727761e-006,-1.03219729945844e-005,-1.17489421676717e-005,-1.16355352814725e-005,-1.02136065356437e-005,-7.94014975773712e-006,-5.37757677216062e-006,-3.05457232158762e-006,-1.34191468136724e-006,-3.73558490210382e-007,-3.16914252834386e-008,-4.47240038949301e-005,-9.74202782733315e-005,-.00013555022303093,-.000154289468887485,-.000152800186874675,-.000134127132921491,-.000104271641780592,-7.06194185176384e-005,-4.01132573852444e-005,-1.76222931839914e-005,-4.90564514068014e-006,-4.16178163573184e-007,-.000283027553294656,-.00061650613987428,-.000857804414450873,-.000976392251931343,-.000966967607276502,-.000848798652964526,-.000659863721515399,-.000446901875894084,-.000253849300234756,-.00011151940991293,-3.10444642827677e-005,-2.6337062228933e-006,-.00110725792950642,-.00241188995212247,-.00335589496079028,-.00381983326596499,-.00378296225321243,-.00332066269912202,-.00258151310547804,-.00174836562742656,-.000993107021958261,-.000436285264394079,-.000121451882844024,-1.03035625519223e-005,-.00303145075477106,-.00660327230083207,-.00918776920961406,-.0104579394995808,-.010356994041021,-.0090913103237661,-.00706766656937942,-.00478667540745511,-.00271892840056321,-.00119446179503315,-.00033251096434234,-2.82090935115534e-005,-.00620337034448753,-.013512521522274,-.0188012736005075,-.0214004702712195,-.0211939018276931,-.0186038861974662,-.0144628287733888,-.00979515178485467,-.00556384423606199,-.00244427156379506,-.00068042954422791,-5.77253164541879e-005,-.00978890451433557,-.0213227287077879,-.029668367645793,-.033769894172609,-.0334439296312284,-.0293568907657418,-.0228223114223215,-.0154567275852829,-.00877973374713072,-.00385705505497627,-.00107371629731872,-9.10904201184736e-005,-.011922145428059,-.0259694710684948,-.0361338281692569,-.0411291773074706,-.0407321771466212,-.0357544728843004,-.0277958494112289,-.0188251253083289,-.0106930517505446,-.00469760137328079,-.0013077052520396,-.000110941243135535,-.0106693129707941,-.0232404826956437,-.0323367236121244,-.0368071390818563,-.0364518574766186,-.0319972326801651,-.024874937019263,-.0168468967973055,-.00956937796371369,-.00420395637395983,-.00117028572514512,-9.92830402484601e-005,-.00552037774992994,-.0120247895925313,-.0167312487713743,-.01904427325192,-.0188604480446285,-.0165555937697616,-.0128704677825032,-.00871670317383983,-.00495126362270572,-.00217515666585201,-.000605514084724702,-5.13697449717053e-005},
			{-.0029914753970567,-.00626152973771969,-.00809817233744839,-.00826680332565275,-.00705494900380919,-.00509444235123848,-.00307890148348425,-.00150693452348,-.000558780057100243,-.000136870983053836,-1.58772021730906e-005,-2.5903751261877e-007,-.00678121949037336,-.0141939350525009,-.0183573243307094,-.0187395851191542,-.0159924957882962,-.0115483255515964,-.00697940112403575,-.00341599124344908,-.00126666935578621,-.000310265689921581,-3.59912011760796e-005,-5.8719862146837e-007,-.0101508632191237,-.0212470181010361,-.027479229748374,-.0280514390660412,-.0239392984565476,-.0172867834834987,-.0104475229362594,-.00511342538303042,-.00189608777487504,-.000464439262650617,-5.38755250068202e-005,-8.78982445184828e-007,-.0128832216939646,-.0269661839217411,-.0348759510581064,-.0356021848114063,-.0303831588068934,-.0219399532025914,-.0132597347865977,-.0064898315939296,-.00240646717698951,-.000589454695127486,-6.83774687490587e-005,-1.11558253342289e-006,-.0148061931138207,-.0309912020589656,-.0400815943915947,-.0409162270209603,-.0349182003841082,-.0252147476572524,-.0152389051863002,-.00745851481395462,-.00276566053049211,-.000677437698056106,-7.85836051712971e-005,-1.2820962657192e-006,-.0157988744548141,-.033069007459881,-.042768865222384,-.0436594558033572,-.0372592914205353,-.0269052706245537,-.0162605977118075,-.00795857100195779,-.00295108426386074,-.00072285651418097,-8.38522436364219e-005,-1.3680544205638e-006,-.0157988744548141,-.033069007459881,-.042768865222384,-.0436594558033572,-.0372592914205353,-.0269052706245537,-.0162605977118075,-.00795857100195779,-.00295108426386074,-.00072285651418097,-8.38522436364219e-005,-1.3680544205638e-006,-.0148061931138207,-.0309912020589656,-.0400815943915947,-.0409162270209603,-.0349182003841082,-.0252147476572524,-.0152389051863002,-.00745851481395462,-.00276566053049211,-.000677437698056106,-7.85836051712971e-005,-1.2820962657192e-006,-.0128832216939646,-.0269661839217411,-.0348759510581064,-.0356021848114063,-.0303831588068934,-.0219399532025914,-.0132597347865977,-.0064898315939296,-.00240646717698951,-.000589454695127486,-6.83774687490587e-005,-1.11558253342289e-006,-.0101508632191237,-.0212470181010361,-.027479229748374,-.0280514390660412,-.0239392984565476,-.0172867834834987,-.0104475229362594,-.00511342538303042,-.00189608777487504,-.000464439262650617,-5.38755250068202e-005,-8.78982445184828e-007,-.00678121949037336,-.0141939350525009,-.0183573243307094,-.0187395851191542,-.0159924957882962,-.0115483255515964,-.00697940112403575,-.00341599124344908,-.00126666935578621,-.000310265689921581,-3.59912011760796e-005,-5.8719862146837e-007,-.0029914753970567,-.00626152973771969,-.00809817233744839,-.00826680332565275,-.00705494900380919,-.00509444235123848,-.00307890148348425,-.00150693452348,-.000558780057100243,-.000136870983053836,-1.58772021730906e-005,-2.5903751261877e-007},
			{-4.77707484775864e-005,-9.99901127316091e-005,-.000129319383418971,-.000132012244784625,-.000112660192597526,-8.13529419075361e-005,-4.91668520822556e-005,-2.40641758792944e-005,-8.92313591757784e-006,-2.1856871401242e-006,-2.53542393256769e-007,-4.13655946284767e-009,-.000563091388850621,-.00117862024866094,-.00152433515352727,-.00155607690124302,-.0013279671418103,-.0009589370588843,-.000579547775714673,-.000283653295149336,-.0001051802861973,-2.57634985121267e-005,-2.98859747651745e-006,-4.87591483748876e-008,-.00202276382802748,-.004233896048097,-.00547578967002042,-.0055898138948427,-.00477038710314128,-.00344473958308176,-.00208187924827317,-.00101895293817233,-.000377833656408332,-9.25488720048507e-005,-1.07357828440126e-005,-1.75154945664267e-007,-.00460437499342809,-.00963752901773119,-.0124644254937564,-.0127239764517783,-.0108587323850331,-.00784118866248488,-.00473893818806989,-.00231942126065002,-.000860054849279415,-.000210667061584079,-2.44376379372225e-005,-3.98701539258914e-007,-.00810600549215572,-.0169668767769879,-.0219436300590561,-.0224005697075911,-.0191168061846747,-.0138044182877922,-.00834290409323953,-.00408334279989948,-.00151412718159681,-.000370879513648267,-4.30224791892881e-005,-7.01914347023563e-007,-.0119687547517608,-.0250520909767,-.0324004131004311,-.0330750978875267,-.0282265247763196,-.0203826283040634,-.0123185424813832,-.00602917535482301,-.0022356531730757,-.00054761447492283,-6.35239518067959e-005,-1.03639711130188e-006,-.0153956985051803,-.0322251100971785,-.0416774344435578,-.0425452977913825,-.0363084610152352,-.0262186674070122,-.0158456389156716,-.00775547397560826,-.00287577470912415,-.000704409734165886,-8.17123944937675e-005,-1.33314265253001e-006,-.0175390732465345,-.0367114597744218,-.047479727872512,-.0484684143436297,-.0413632909868219,-.0298688057526838,-.0180516539400008,-.00883518380630904,-.00327613737349354,-.000802476998296004,-9.30883176033672e-005,-1.51874152529913e-006,-.0177100195447144,-.037069271618925,-.0479424936985227,-.0489408164993326,-.0417664423606343,-.0301599249984219,-.0182275961561996,-.00892129679210925,-.00330806856783064,-.000810298419092016,-9.39956120239112e-005,-1.53354408858128e-006,-.0155590470074065,-.0325670188106206,-.0421196324049362,-.0429967037908343,-.0366936941193377,-.0264968477084686,-.0160137612897764,-.00783775962556066,-.00290628670514157,-.000711883527900599,-8.25793637481591e-005,-1.34728730829027e-006,-.0111823253052378,-.0234059964205263,-.0302714832768886,-.0309018366364716,-.0263718481021364,-.0190433495379502,-.0115091295769889,-.00563301709645563,-.00208875539431883,-.000511632440253917,-5.93499915831539e-005,-9.68298698098172e-007,-.00513361662858989,-.0107452975256271,-.0138971265528759,-.0141865111307979,-.012106869926272,-.00874248004628156,-.005283646948578,-.0025860229823533,-.000958912313193877,-.000234881809580631,-2.72465784513743e-005,-4.44529573439469e-007},
			{-5.68594172031225e-007,-1.19013825765391e-006,-1.53923164459519e-006,-1.57128358699498e-006,-1.34094463604488e-006,-9.68308224601569e-007,-5.85211378134644e-007,-2.86425283165094e-007,-1.06208155422931e-007,-2.60152710469109e-008,-3.0178033998418e-009,-4.92356447782966e-011,-3.48508775138951e-005,-7.29472173341577e-005,-9.4344219744047e-005,-9.63087814184433e-005,-8.21906019486044e-005,-5.93505754917118e-005,-3.58693969483533e-005,-1.75558824755616e-005,-6.50982299448902e-006,-1.59455560634358e-006,-1.84970409164038e-007,-3.0178033998418e-009,-.000300435417626676,-.000628848662297057,-.000813303626233361,-.000830239323386619,-.00070853216856797,-.000511637473895075,-.000309215664595502,-.000151342211720415,-5.61185694462147e-005,-1.37460234489013e-005,-1.59455560634358e-006,-2.60152710469109e-008,-.00122653696255208,-.00256729427659293,-.00332033741971269,-.00338947793153767,-.00289260467614267,-.00208877594431585,-.00126238259464391,-.000617859299465599,-.000229105810010919,-5.61185694462147e-005,-6.50982299448902e-006,-1.06208155422931e-007,-.0033077610227998,-.00692355485520927,-.00895438379338928,-.00914084396303245,-.00780086152659004,-.00563307251629313,-.00340443057968316,-.00166626116517048,-.000617859299465599,-.000151342211720415,-1.75558824755616e-005,-2.86425283165094e-007,-.0067582699589302,-.0141458988313976,-.018295198043168,-.0186761651548639,-.0159383727377934,-.0115092428083319,-.00695578089086413,-.00340443057968316,-.00126238259464391,-.000309215664595502,-3.58693969483533e-005,-5.85211378134644e-007,-.0111824353213518,-.023406226697924,-.030271781099646,-.0309021406608933,-.0263721075587493,-.0190435368939121,-.0115092428083319,-.00563307251629313,-.00208877594431585,-.000511637473895075,-5.93505754917118e-005,-9.68308224601569e-007,-.0154857991299779,-.0324137019011133,-.0419213443176089,-.0427942866834367,-.036520949914119,-.0263721075587493,-.0159383727377934,-.00780086152659004,-.00289260467614267,-.00070853216856797,-8.21906019486044e-005,-1.34094463604488e-006,-.0181458513277659,-.037981521698904,-.0491223265304294,-.0501452173903112,-.0427942866834367,-.0309021406608933,-.0186761651548639,-.00914084396303246,-.00338947793153767,-.000830239323386619,-9.63087814184433e-005,-1.57128358699498e-006,-.0177757018611982,-.0372067528692506,-.0481203011840638,-.0491223265304294,-.0419213443176089,-.030271781099646,-.018295198043168,-.00895438379338928,-.00332033741971269,-.000813303626233361,-9.4344219744047e-005,-1.53923164459519e-006,-.0137442229153402,-.0287683664692428,-.0372067528692506,-.037981521698904,-.0324137019011133,-.023406226697924,-.0141458988313976,-.00692355485520927,-.00256729427659293,-.000628848662297057,-7.29472173341577e-005,-1.19013825765392e-006,-.00656636739345367,-.0137442229153402,-.0177757018611982,-.0181458513277659,-.0154857991299779,-.0111824353213518,-.0067582699589302,-.0033077610227998,-.00122653696255208,-.000300435417626676,-3.48508775138952e-005,-5.68594172031225e-007},
			{-6.20272325817235e-009,-1.29830705524454e-008,-1.67912869868831e-008,-1.71409376487648e-008,-1.46281986187158e-008,-1.05631542517567e-008,-6.38399829730143e-009,-3.1245778676732e-009,-1.15861158670691e-009,-2.83797363265043e-010,-3.2920842768966e-011,-5.37105538571509e-013,-1.9769158918718e-006,-4.13793062049084e-006,-5.35167549924359e-006,-5.46311525260784e-006,-4.66226157691217e-006,-3.36666116468723e-006,-2.03469140284302e-006,-9.95857365368599e-007,-3.69269684126226e-007,-9.045116058834e-008,-1.04924457427685e-008,-1.71184886157124e-010,-4.08974964028785e-005,-8.56035420437903e-005,-.000110712919239298,-.000113018331892963,-9.64506516716335e-005,-6.96478861031055e-005,-4.20927287149029e-005,-2.06018238739433e-005,-7.6392757225221e-006,-1.87121062155825e-006,-2.1706273078516e-007,-3.54139156583407e-009,-.000299454081965167,-.000626794604813159,-.00081064706909918,-.000827527447862548,-.000706217834626139,-.000509966272467282,-.000308205649327789,-.000150847870838019,-5.59352649812969e-005,-1.37011237392702e-005,-1.58934718486943e-006,-2.59302953359118e-008,-.00123708915628192,-.00258938132932369,-.00334890308448347,-.00341863842883094,-.00291749045281179,-.00210674618825924,-.0012732431770043,-.000623174892248206,-.000231076862629534,-5.6601371053274e-005,-6.56582857400444e-006,-1.07121889836103e-007,-.00349753239972921,-.00732076992880828,-.00946811067096992,-.00966526834956002,-.00824840904375729,-.00595625061785239,-.00359974803892942,-.0017618571509321,-.000653306845161927,-.000160024948988241,-1.85630906649181e-005,-3.02857945621352e-007,-.0074441143400151,-.0155814563465409,-.0201518357411283,-.0205714642490873,-.0175558345219898,-.0126772265613641,-.00766166912395449,-.00374991982442457,-.00139049200941683,-.000340595563207872,-3.95095037359528e-005,-6.44599940278459e-007,-.0125314188045125,-.0262298167578035,-.033923591419649,-.0346299938653073,-.0295534841096122,-.0213408376153209,-.0128976504320703,-.00631261338244207,-.00234075363682485,-.000573358421239403,-6.65102812046103e-005,-1.08511925583571e-006,-.0170402239900308,-.0356673062917253,-.0461293015064619,-.047089867591508,-.040186829342282,-.0290192721808159,-.0175382257776313,-.0085838920298908,-.00318295692604349,-.00077965281321311,-9.04406840954337e-005,-1.47554522466982e-006,-.0186127337638557,-.0389587646541599,-.0503862160588214,-.0514354253188647,-.0438953593449828,-.0316972351677059,-.0191566922641638,-.00937603268029293,-.00347668727130158,-.000851600909064724,-9.87867515986945e-005,-1.61171182018355e-006,-.015482742876017,-.0324073047817918,-.0419130707843193,-.0427858408675583,-.0365137421945241,-.0263669027993754,-.0159352271645878,-.00779932196032425,-.00289203379604635,-.000708392333727775,-8.2174380935362e-005,-1.34067998923389e-006,-.00769779899133229,-.0161124498455309,-.0208385811603797,-.0212725100010394,-.0181541119739297,-.0131092481092602,-.00792276772768902,-.0038777116690518,-.00143787796622194,-.000352202567445854,-4.08559304861565e-005,-6.66566974047622e-007},
			{-.00404945140720489,-.00814474566006138,-.00979128722597762,-.00896406350610278,-.00659219233567233,-.00391599979728114,-.00183988793598998,-.000650774647452403,-.00015752876308832,-2.15142568085075e-005,-1.03996283213943e-006,-3.26296210833654e-009,-.00917949010540939,-.0184628989656689,-.0221953581292932,-.0203201678410382,-.0149434968427065,-.00887697561402263,-.00417074596164616,-.00147520709262294,-.000357093733613841,-4.87695462026171e-005,-2.35743748168761e-006,-7.3966385507205e-009,-.0137408542244633,-.0276372652876165,-.0332244141027671,-.030417426340031,-.0223690433086643,-.0132880177946399,-.00624322392727204,-.00220824962801375,-.000534536546334486,-7.3003534757174e-005,-3.5288675522474e-006,-1.10720890713312e-008,-.017439548481424,-.0350765258115668,-.0421675953362895,-.0386050366790666,-.0283902302500237,-.0168648198114988,-.00792374364660721,-.00280265519288241,-.000678420414234276,-9.26542602749126e-005,-4.47875042967689e-006,-1.40524185029455e-008,-.0200426049296945,-.0403121072713726,-.0484616018046971,-.044367289627886,-.0326278039463224,-.0193820912882233,-.00910645500038529,-.00322098423849287,-.000779682590591189,-.000106483991585058,-5.14725627996212e-006,-1.61499062123808e-008,-.0213863615446234,-.043014832840112,-.0517107103024967,-.0473418949315912,-.0348153353344463,-.0206815637605425,-.00971699735195308,-.00343693515366777,-.000831956416390127,-.000113623211690348,-5.49235412024762e-006,-1.72326768092913e-008,-.0213863615446234,-.043014832840112,-.0517107103024967,-.0473418949315912,-.0348153353344463,-.0206815637605425,-.00971699735195308,-.00343693515366777,-.000831956416390127,-.000113623211690348,-5.49235412024762e-006,-1.72326768092913e-008,-.0200426049296945,-.0403121072713726,-.0484616018046971,-.044367289627886,-.0326278039463224,-.0193820912882233,-.00910645500038529,-.00322098423849287,-.000779682590591189,-.000106483991585058,-5.14725627996212e-006,-1.61499062123808e-008,-.017439548481424,-.0350765258115668,-.0421675953362895,-.0386050366790666,-.0283902302500237,-.0168648198114988,-.00792374364660721,-.00280265519288241,-.000678420414234276,-9.26542602749126e-005,-4.47875042967689e-006,-1.40524185029455e-008,-.0137408542244633,-.0276372652876165,-.0332244141027671,-.030417426340031,-.0223690433086643,-.0132880177946399,-.00624322392727204,-.00220824962801375,-.000534536546334486,-7.3003534757174e-005,-3.5288675522474e-006,-1.10720890713312e-008,-.00917949010540939,-.0184628989656689,-.0221953581292932,-.0203201678410382,-.0149434968427065,-.00887697561402263,-.00417074596164616,-.00147520709262294,-.000357093733613841,-4.87695462026171e-005,-2.35743748168761e-006,-7.3966385507205e-009,-.00404945140720489,-.00814474566006138,-.00979128722597762,-.00896406350610278,-.00659219233567233,-.00391599979728114,-.00183988793598998,-.000650774647452403,-.00015752876308832,-2.15142568085075e-005,-1.03996283213943e-006,-3.26296210833654e-009},
			{-6.46655241827284e-005,-.000130063110895906,-.000156356665933801,-.000143146764126628,-.000105270450249422,-6.25344408726795e-005,-2.93810953272788e-005,-1.03921937740666e-005,-2.51557038586478e-006,-3.43560288548878e-007,-1.66071240047969e-008,-5.21061086814239e-011,-.000762236326271721,-.0015331017429415,-.00184303355050058,-.00168732357751008,-.00124086152980604,-.00073711646319537,-.000346325780965151,-.000122496611669995,-2.96519537053663e-005,-4.04967152908804e-006,-1.95754280991916e-007,-6.14193874705368e-010,-.00273814180028246,-.00550728143194998,-.00662063329969093,-.00606128448480099,-.00445748215614846,-.00264790502628461,-.00124408804027283,-.000440038189267583,-.00010651716679239,-1.45474238225389e-005,-7.03197894004764e-007,-2.20633662270343e-009,-.00623277490876185,-.0125361095327521,-.0150704091023226,-.013797175094379,-.0101464733989415,-.00602737082751107,-.00283189158461527,-.0010016497263513,-.00024246279884684,-3.31139965720151e-005,-1.6006746579724e-006,-5.02223790632381e-009,-.0109728047159291,-.0220698298613373,-.0265314660788539,-.0242899366908432,-.0178628737266676,-.0106111906829482,-.00498554717433864,-.00176340506466325,-.000426855931999716,-5.82972148148455e-005,-2.81798888822008e-006,-8.84165345126754e-009,-.0162016678512013,-.0325867508082808,-.039174487484589,-.0358648036285977,-.0263750567407152,-.0156677341301185,-.00736130656347784,-.00260371927546702,-.000630265297671321,-8.60775467651508e-005,-4.16084320803041e-006,-1.30549605302744e-008,-.0208405969285544,-.0419171251406854,-.0503910900437627,-.0461337637093912,-.0339268729336555,-.0201537850786976,-.00946902654504109,-.00334922703227079,-.00081072548502156,-.00011072362877741,-5.3521931802253e-006,-1.67929112501503e-008,-.0237420053340501,-.0477527880842341,-.0574064904526992,-.0525564631293,-.0386501404407882,-.0229595761810517,-.0107872955612227,-.00381550328609678,-.000923593928514568,-.000126138468780461,-6.09732050715256e-006,-1.91308046425978e-008,-.0239734091183992,-.0482182153183887,-.0579660084441261,-.0530687098535765,-.0390268478266128,-.0231833539428959,-.0108924349957666,-.00385269146321899,-.000932595827353905,-.000127367889742049,-6.15674863969202e-006,-1.93172649069964e-008,-.0210617158529497,-.0423618662225568,-.0509257399709112,-.0466232433651179,-.0342868373663876,-.020367616923072,-.00956949300345818,-.00338476236177552,-.000819327289847131,-.00011189840749341,-5.40897999891501e-006,-1.69710841924047e-008,-.0151371069219119,-.0304455773166633,-.0366004544168802,-.0335082395371539,-.0246420342413141,-.0146382563159452,-.00687761812917763,-.00243263702412373,-.00058885253590198,-8.04216603455029e-005,-3.88744721245458e-006,-1.21971598987849e-008,-.00694919005501226,-.0139770501853164,-.0168026502788677,-.0153830666688022,-.0113127416069928,-.00672017616961103,-.00315739828964731,-.00111678255975218,-.000270332250902931,-3.69202255862743e-005,-1.78466133902173e-006,-5.5995100454324e-009},
			{-7.69685243564879e-007,-1.54808390489254e-006,-1.86104450591298e-006,-1.70381286481175e-006,-1.25298778853868e-006,-7.44319897852733e-007,-3.49710232755209e-007,-1.23693703828388e-007,-2.99417259756176e-008,-4.08924674643852e-009,-1.97667280147994e-010,-6.20196054366658e-013,-4.71763649140235e-005,-9.48868018782288e-005,-.000114069115221109,-.000104431906584671,-7.67994574913963e-005,-4.56216452211473e-005,-2.14348107782544e-005,-7.58156578700641e-006,-1.83522004948295e-006,-2.50642452023579e-007,-1.21156327444106e-008,-3.80137148576797e-011,-.000406688494124778,-.00081798100889079,-.000983344027882673,-.000900265522892682,-.000662057277488362,-.000393285880086281,-.00018478089469472,-6.53576336083714e-005,-1.58206949533316e-005,-2.16068791147813e-006,-1.04444003796667e-007,-3.27700968053241e-010,-.00166031846121593,-.00333943298029159,-.00401453266272909,-.00367536207502424,-.00270286948383404,-.00160560187139806,-.000754373765606237,-.000266824577604282,-6.45884805686753e-005,-8.82107578694501e-006,-4.263959029382e-007,-1.33784942253663e-009,-.00447759575057386,-.00900588126384461,-.0108265099805079,-.00991182474529423,-.00728917806907233,-.00433003443882712,-.00203441740011081,-.000719580383364912,-.000174184117617127,-2.37889371116125e-005,-1.14991703559086e-006,-3.60795174491446e-009,-.00914842415783806,-.0184004153805573,-.0221202428642781,-.0202513987414977,-.0148929238932162,-.00884693345962729,-.00415663099731916,-.0014702145815281,-.000355885229102249,-4.86044964941899e-005,-2.34945925758498e-006,-7.37160626866756e-009,-.0151372558449131,-.0304458768485863,-.0366008145021687,-.0335085692003862,-.0246422766763918,-.0146384003311175,-.00687768579306728,-.00243266095707262,-.000588858329194405,-8.04224515558099e-005,-3.8874854582243e-006,-1.21972798977825e-008,-.0209625628637279,-.0421624377574542,-.05068599505261,-.0464037534631962,-.0341254240019886,-.0202717315680825,-.00952444236070643,-.00336882779555505,-.000815470113608189,-.000111371619378443,-5.3835159512905e-006,-1.68911888154202e-008,-.0245633787434188,-.0494048334697752,-.0593925132894224,-.0543746954436094,-.0399872725481971,-.023753880836457,-.0111604905634277,-.00394750363309629,-.00095554638928031,-.000130502328643898,-6.30826212149847e-006,-1.97926499253432e-008,-.0240623208776429,-.0483970453850838,-.0581809907923459,-.0532655292683946,-.0391715892641307,-.0232693355725446,-.0109328324858832,-.00386698019345867,-.000936054606840123,-.000127840267412333,-6.179582578331e-006,-1.93889081179464e-008,-.0186050545055932,-.0374207322675915,-.0449857065070876,-.0411850577650857,-.030287583522627,-.0179919160266198,-.00845329697557845,-.00298996001413125,-.000723760067411173,-9.88464560551176e-005,-4.77807071380575e-006,-1.49915585521718e-008,-.00888865263692734,-.0178779315290952,-.021492133691307,-.0196763558094162,-.0144700360358322,-.00859572283356178,-.00403860254373974,-.00142846751434807,-.000345779789560036,-4.72243611004261e-005,-2.28274584398164e-006,-7.16228788345407e-009},
			{-.00508628631126404,-.00983033824354329,-.0109846550070609,-.00901916015333691,-.00571557221193217,-.00279307452781809,-.00102018939735755,-.000260771859326372,-4.12071599264361e-005,-3.13787157581542e-006,-6.32055938620332e-008,-3.81377664478188e-011,-.0115298370501388,-.0222838808434348,-.0249005412854803,-.0204450635558571,-.012956332423902,-.00633147491192501,-.00231261411410569,-.000591130121526243,-9.34103607574427e-005,-7.11307734943901e-006,-1.43277462826387e-007,-8.64525128335375e-011,-.0172591078936303,-.0333569244815142,-.0372738250147888,-.0306043837625979,-.019394440549112,-.00947763686126095,-.00346177108471816,-.000884867539951351,-.000139826737158981,-1.06476239773248e-005,-2.14473212317988e-007,-1.29411477385236e-010,-.0219048280361805,-.0423357741945106,-.0473070063778303,-.0388423183634187,-.0246149388313989,-.0120287796399897,-.00439359327137407,-.00112305174849657,-.000177464597313172,-1.35136980227866e-005,-2.72204036451189e-007,-1.64245810125438e-010,-.0251743796492038,-.0486548833232205,-.0543681300147957,-.0446399883770301,-.0282890061570013,-.0138242155963766,-.00504938842044633,-.00129068034844452,-.000203953262708829,-1.55307754038705e-005,-3.12833670474356e-007,-1.88761417029128e-010,-.0268621961430683,-.051916950382044,-.0580132417457791,-.0476328767706495,-.0301856428110978,-.0147510602465105,-.00538792470927311,-.00137721402318714,-.000217627310910781,-1.65720363705528e-005,-3.33807606524435e-007,-2.01416927810585e-010,-.0268621961430683,-.051916950382044,-.0580132417457791,-.0476328767706495,-.0301856428110978,-.0147510602465105,-.00538792470927311,-.00137721402318714,-.000217627310910781,-1.65720363705528e-005,-3.33807606524435e-007,-2.01416927810585e-010,-.0251743796492038,-.0486548833232205,-.0543681300147957,-.0446399883770301,-.0282890061570013,-.0138242155963766,-.00504938842044633,-.00129068034844452,-.000203953262708829,-1.55307754038705e-005,-3.12833670474356e-007,-1.88761417029128e-010,-.0219048280361805,-.0423357741945106,-.0473070063778303,-.0388423183634187,-.0246149388313989,-.0120287796399897,-.00439359327137407,-.00112305174849657,-.000177464597313172,-1.35136980227866e-005,-2.72204036451189e-007,-1.64245810125438e-010,-.0172591078936303,-.0333569244815142,-.0372738250147888,-.0306043837625979,-.019394440549112,-.00947763686126095,-.00346177108471816,-.000884867539951351,-.000139826737158981,-1.06476239773248e-005,-2.14473212317988e-007,-1.29411477385236e-010,-.0115298370501388,-.0222838808434348,-.0249005412854803,-.0204450635558571,-.012956332423902,-.00633147491192501,-.00231261411410569,-.000591130121526243,-9.34103607574427e-005,-7.11307734943901e-006,-1.43277462826387e-007,-8.64525128335375e-011,-.00508628631126404,-.00983033824354329,-.0109846550070609,-.00901916015333691,-.00571557221193217,-.00279307452781809,-.00102018939735755,-.000260771859326372,-4.12071599264361e-005,-3.13787157581542e-006,-6.32055938620331e-008,-3.81377664478188e-011},
			{-8.12226984453021e-005,-.000156980270065097,-.000175413507334088,-.000144026600260793,-9.12717393010351e-005,-4.46024931010705e-005,-1.62913628348369e-005,-4.16425517497517e-006,-6.58035454488694e-007,-5.01085430834115e-008,-1.00932754786986e-009,-6.09020435343665e-013,-.000957401831273778,-.00185038420184216,-.00206766355177068,-.00169769452974626,-.0010758535732373,-.000525745997014592,-.000192032287901912,-4.90856079239649e-005,-7.75650601653937e-006,-5.90647834026672e-007,-1.18973151739896e-008,-7.17874795152048e-012,-.00343922466500658,-.00664703866113062,-.00742755993763223,-.00609853951561999,-.00386473267978856,-.00188860992469107,-.000689827572348814,-.000176327669275857,-2.78632920211386e-005,-2.12175340882166e-006,-4.27381047927422e-008,-2.57878418572552e-011,-.00782863516982087,-.0151305150742709,-.016907199330518,-.0138819779418027,-.00879721016390503,-.00429900326923858,-.00157024007444182,-.000401370985487183,-6.34246288365862e-005,-4.82970290571489e-006,-9.72838540256606e-008,-5.87003250388414e-011,-.0137823178549157,-.0266372826857591,-.0297651365985099,-.0244392321648105,-.0154874948296011,-.0075683983517811,-.00276440878199448,-.000706613909031054,-.000111659104747359,-8.50269544402536e-006,-1.71268295079276e-007,-1.03341964508827e-010,-.0203499963670061,-.0393307288068947,-.0439490968079129,-.0360852427727842,-.0228677401605563,-.0111749620480468,-.00408173061038807,-.00104333615238232,-.00016486794165333,-1.25544790953984e-005,-2.52882658732434e-007,-1.52587440258738e-010,-.0261766921577074,-.0505920670327779,-.0565327854119549,-.0464173199082629,-.029415326849682,-.0143746237655383,-.0052504287338376,-.00134206850877865,-.000212073617975246,-1.6149129884536e-005,-3.25289075746179e-007,-1.96276912228913e-010,-.0298209867484436,-.0576354472700868,-.0644032269037527,-.0528794957568774,-.0335105011321007,-.0163758454369784,-.00598138850975171,-.0015289100309035,-.000241598308649254,-1.83973966375979e-005,-3.7057551652431e-007,-2.23602400308645e-010,-.0301116399215653,-.0581971968115976,-.0650309392734592,-.0533948909503518,-.033837114519146,-.0165354542211866,-.00603968669970425,-.00154381170252316,-.000243953070268046,-1.85767086689201e-005,-3.74187367153787e-007,-2.25781763041192e-010,-.026454427101384,-.0511288493276215,-.0571325987832424,-.046909808097858,-.0297274237371193,-.0145271386554198,-.00530613583081857,-.00135630786795552,-.000214323721005407,-1.63204723006016e-005,-3.28740395820796e-007,-1.98359412066345e-010,-.0190128617434316,-.0367464296104888,-.0410613390925373,-.0337141943145245,-.0213651724656962,-.0104406902377677,-.00381353285620554,-.000974781796488713,-.000154034984775778,-1.17295635339088e-005,-2.36266530031746e-007,-1.42561396724728e-010,-.0087284836148921,-.0168696650240853,-.0188505670693449,-.0154776170276039,-.00980838972652806,-.00479314449861828,-.00175072850680893,-.000447505854382295,-7.0714859176842e-005,-5.38484445409324e-006,-1.08465972348532e-007,-6.54475287423684e-011},
			{-.00610300723148638,-.011334382523687,-.0117725765312941,-.00866893492130421,-.00473399758463878,-.00190309670559272,-.000540391748990977,-9.98226951437858e-005,-1.02973144060446e-005,-4.37202313877829e-007,-3.66970948242474e-009,-4.25830680908177e-013,-.0138345886544025,-.0256933203450701,-.026686639477119,-.0196511562511815,-.0107312521172389,-.00431403062339345,-.00122498585958623,-.000226282859139225,-2.33424447406373e-005,-9.91071113277978e-007,-8.31867295004696e-009,-9.65293351840578e-013,-.0207091095227104,-.0384605425083467,-.0399474500854692,-.0294159773896932,-.0160636995405713,-.00645770791571394,-.00183369140663751,-.000338724672636725,-3.49414974841223e-005,-1.48354249933819e-006,-1.24522899458459e-008,-1.44495555626435e-012,-.0262834838088483,-.0488131585372653,-.0507003044420999,-.0373340227205607,-.0203876456552513,-.00819596135984531,-.00232727526714406,-.000429900881983438,-4.43468746144641e-005,-1.88287503228147e-006,-1.58041349298872e-008,-1.83390144930814e-012,-.030206600974669,-.0560990929882762,-.0582679174767988,-.0429065619801723,-.0234307400647473,-.00941930438906687,-.00267464830248913,-.000494068613398971,-5.09661639870523e-005,-2.16391613832206e-006,-1.81630867905044e-008,-2.10763267567556e-012,-.0322318027893339,-.0598602571462754,-.0621744904907474,-.0457832327865349,-.0250016542281072,-.0100508217305121,-.00285397012027107,-.0005271934477113,-5.43831908772887e-005,-2.30899591389123e-006,-1.93808310960915e-008,-2.24893892602145e-012,-.0322318027893339,-.0598602571462754,-.0621744904907474,-.0457832327865349,-.0250016542281072,-.0100508217305121,-.00285397012027107,-.0005271934477113,-5.43831908772887e-005,-2.30899591389123e-006,-1.93808310960915e-008,-2.24893892602145e-012,-.030206600974669,-.0560990929882762,-.0582679174767988,-.0429065619801723,-.0234307400647473,-.00941930438906688,-.00267464830248913,-.000494068613398971,-5.09661639870523e-005,-2.16391613832206e-006,-1.81630867905044e-008,-2.10763267567556e-012,-.0262834838088483,-.0488131585372653,-.0507003044420999,-.0373340227205607,-.0203876456552513,-.00819596135984532,-.00232727526714406,-.000429900881983438,-4.43468746144642e-005,-1.88287503228147e-006,-1.58041349298872e-008,-1.83390144930814e-012,-.0207091095227104,-.0384605425083467,-.0399474500854692,-.0294159773896932,-.0160636995405713,-.00645770791571394,-.00183369140663751,-.000338724672636725,-3.49414974841223e-005,-1.48354249933819e-006,-1.24522899458459e-008,-1.44495555626435e-012,-.0138345886544025,-.0256933203450701,-.026686639477119,-.0196511562511815,-.0107312521172389,-.00431403062339345,-.00122498585958623,-.000226282859139225,-2.33424447406373e-005,-9.91071113277978e-007,-8.31867295004696e-009,-9.65293351840578e-013,-.00610300723148638,-.011334382523687,-.0117725765312941,-.0086689349213042,-.00473399758463878,-.00190309670559272,-.000540391748990977,-9.98226951437858e-005,-1.02973144060446e-005,-4.37202313877829e-007,-3.66970948242474e-009,-4.25830680908177e-013},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,3.354101966,0.,0.,4.062019202},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,3.354101966,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,-7.245688375,0.,0.,-5.477225576,0.,-11.12429773,-8.291561975},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370349,0.,0.,-7.245688375,0.,0.,-5.477225575,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,7.245688375,0.,0.,2.049390154,0.,0.,13.74772709,0.,15.19868416,5.196152424,0.,28.14249455,10.48808848},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,7.245688375,0.,0.,2.049390152,0.,0.,13.74772708,0.,0.,5.196152424,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-11.12429773,-.5000000000,0.,0.,-10.06230590,0.,-33.37289319,-2.535462764,0.,-30.89729531,-7.676494735},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,-10.06230590,0.,0.,-2.535462764,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,17.54992877,0.,0.,0.,2.645751311,0.,26.32489316,.5000000000,0.,16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,-3.354101966,-.5000000000},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,7.245688375,0.,0.,-5.477225576,0.,11.12429773,-8.291561975},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370349,0.,0.,7.245688375,0.,0.,-5.477225575,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,-5.916079782,0.,0.,1.673320053,0.,0.,0.,0.,-12.40967365,4.242640686,0.,0.,8.563488387},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246900,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,-5.916079782,0.,0.,1.673320053,0.,0.,0.,0.,0.,4.242640686,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,2.738612788,0.,8.616843970,-.3872983346,0.,0.,-2.598076212,0.,8.616843970,-1.963961012,0.,-7.977647347,-5.946187254},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,-2.598076212,0.,0.,-1.963961013,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-3.070597895,0.,0.,0.,0.,0.,0.,0.,0.,0.,-6.633249580,0.,0.,0.,1.,0.,0.,.3779644731,0.,6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,-1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,7.245688375,0.,0.,2.049390154,0.,0.,-13.74772709,0.,15.19868416,5.196152424,0.,-28.14249455,10.48808848},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,7.245688375,0.,0.,2.049390152,0.,0.,-13.74772708,0.,0.,5.196152424,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,-8.616843970,-.3872983346,0.,0.,2.598076212,0.,8.616843970,-1.963961012,0.,7.977647347,-5.946187254},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,2.598076212,0.,0.,-1.963961013,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.913024938,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,-6.292853088,.3585685828,0.,0.,2.171240594},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,-10.60660172,0.,11.12429773,-.5000000000,0.,0.,10.06230590,0.,-33.37289319,-2.535462764,0.,30.89729531,-7.676494735},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,10.06230590,0.,0.,-2.535462764,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,-3.070597895,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.633249580,0.,0.,0.,-1.,0.,0.,.3779644731,0.,-6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,1.,0.,0.,.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,-17.54992877,0.,0.,0.,-2.645751311,0.,26.32489316,.5000000000,0.,-16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,3.354101966,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,3.354101966,0.,0.,4.062019202},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,3.354101966,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,-7.245688375,0.,0.,-5.477225576,0.,-11.12429773,-8.291561975},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370349,0.,0.,-7.245688375,0.,0.,-5.477225575,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,7.245688375,0.,0.,2.049390154,0.,0.,13.74772709,0.,15.19868416,5.196152424,0.,28.14249455,10.48808848},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,7.245688375,0.,0.,2.049390152,0.,0.,13.74772708,0.,0.,5.196152424,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-11.12429773,-.5000000000,0.,0.,-10.06230590,0.,-33.37289319,-2.535462764,0.,-30.89729531,-7.676494735},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,-10.06230590,0.,0.,-2.535462764,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,17.54992877,0.,0.,0.,2.645751311,0.,26.32489316,.5000000000,0.,16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,-3.354101966,-.5000000000},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,7.245688375,0.,0.,-5.477225576,0.,11.12429773,-8.291561975},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370349,0.,0.,7.245688375,0.,0.,-5.477225575,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,-5.916079782,0.,0.,1.673320053,0.,0.,0.,0.,-12.40967365,4.242640686,0.,0.,8.563488387},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246900,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,-5.916079782,0.,0.,1.673320053,0.,0.,0.,0.,0.,4.242640686,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,2.738612788,0.,8.616843970,-.3872983346,0.,0.,-2.598076212,0.,8.616843970,-1.963961012,0.,-7.977647347,-5.946187254},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,-2.598076212,0.,0.,-1.963961013,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-3.070597895,0.,0.,0.,0.,0.,0.,0.,0.,0.,-6.633249580,0.,0.,0.,1.,0.,0.,.3779644731,0.,6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,-1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,7.245688375,0.,0.,2.049390154,0.,0.,-13.74772709,0.,15.19868416,5.196152424,0.,-28.14249455,10.48808848},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,7.245688375,0.,0.,2.049390152,0.,0.,-13.74772708,0.,0.,5.196152424,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,-8.616843970,-.3872983346,0.,0.,2.598076212,0.,8.616843970,-1.963961012,0.,7.977647347,-5.946187254},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,2.598076212,0.,0.,-1.963961013,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.913024938,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,-6.292853088,.3585685828,0.,0.,2.171240594},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,-10.60660172,0.,11.12429773,-.5000000000,0.,0.,10.06230590,0.,-33.37289319,-2.535462764,0.,30.89729531,-7.676494735},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,10.06230590,0.,0.,-2.535462764,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,-3.070597895,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.633249580,0.,0.,0.,-1.,0.,0.,.3779644731,0.,-6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,1.,0.,0.,.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,-17.54992877,0.,0.,0.,-2.645751311,0.,26.32489316,.5000000000,0.,-16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,3.354101966,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,3.354101966,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,-5.477225576,0.,-8.291561975,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,7.245688375,0.,11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370349,0.,0.,-5.477225575,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,5.196152424,0.,10.48808848,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,-13.74772709,0.,-28.14249455,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,15.19868416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390152,0.,0.,5.196152424,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,-13.74772708,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,-7.676494735,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,10.06230590,0.,30.89729531,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-33.37289319,.5000000000,0.,0.,3.354101966,0.,11.12429773,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,.5000000000,0.,0.,3.354101966,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,3.027650356,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,-16.24807681,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,26.32489316,0.,0.,0.,-2.645751311,0.,-17.54992877,.5000000000,0.,4.062019202,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,7.245688375,0.,0.,-3.354101966,.5000000000},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,-5.477225576,0.,-8.291561975,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,-7.245688375,0.,-11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370349,0.,0.,-5.477225575,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,-7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,4.242640686,0.,8.563488387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246899,0.,0.,-5.916079782,0.,-12.40967365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246900,0.,0.,-5.916079782,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961012,0.,-5.946187254,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,7.977647347,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,8.616843970,-.3872983346,0.,0.,-2.598076212,0.,-8.616843970,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961013,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,-2.598076212,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,2.288688541,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.141195788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,6.633249580,-.3779644731,0.,-3.070597895,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,-.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,5.196152424,0.,10.48808848,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,13.74772709,0.,28.14249455,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,15.19868416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390152,0.,0.,5.196152424,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,13.74772708,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961012,0.,-5.946187254,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.598076212,0.,-7.977647347,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,8.616843970,.3872983346,0.,0.,2.598076212,0.,8.616843970,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961013,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,.3872983346,0.,0.,2.598076212,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,2.171240594,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.292853088,0.,0.,0.,0.,0.,0.,.3585685828,0.,2.913024938,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,-1.,0.,0.,-.4629100500,.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,-7.676494735,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.06230590,0.,-30.89729531,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-33.37289319,-.5000000000,0.,0.,-3.354101966,0.,-11.12429773,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.06230590,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,-3.354101966,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,2.288688541,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,6.141195788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.633249580,-.3779644731,0.,-3.070597895,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,-.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,3.027650356,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,16.24807681,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,26.32489316,0.,0.,0.,2.645751311,0.,17.54992877,.5000000000,0.,4.062019202,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,1.080123450,0.,0.,1.500000000,.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,-3.354101966,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,3.354101966,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,-5.477225576,0.,-8.291561975,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,7.245688375,0.,11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370349,0.,0.,-5.477225575,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,5.196152424,0.,10.48808848,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,-13.74772709,0.,-28.14249455,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,15.19868416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390152,0.,0.,5.196152424,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,-13.74772708,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,-7.676494735,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,10.06230590,0.,30.89729531,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-33.37289319,.5000000000,0.,0.,3.354101966,0.,11.12429773,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,.5000000000,0.,0.,3.354101966,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,3.027650356,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,-16.24807681,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,26.32489316,0.,0.,0.,-2.645751311,0.,-17.54992877,.5000000000,0.,4.062019202,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,7.245688375,0.,0.,-3.354101966,.5000000000},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,-5.477225576,0.,-8.291561975,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,-7.245688375,0.,-11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370349,0.,0.,-5.477225575,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,-7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,4.242640686,0.,8.563488387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246899,0.,0.,-5.916079782,0.,-12.40967365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246900,0.,0.,-5.916079782,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961012,0.,-5.946187254,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,7.977647347,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,8.616843970,-.3872983346,0.,0.,-2.598076212,0.,-8.616843970,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961013,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,-2.598076212,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,2.288688541,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.141195788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,6.633249580,-.3779644731,0.,-3.070597895,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,-.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,5.196152424,0.,10.48808848,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,13.74772709,0.,28.14249455,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,15.19868416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390152,0.,0.,5.196152424,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,13.74772708,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961012,0.,-5.946187254,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.598076212,0.,-7.977647347,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,8.616843970,.3872983346,0.,0.,2.598076212,0.,8.616843970,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961013,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,.3872983346,0.,0.,2.598076212,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,2.171240594,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.292853088,0.,0.,0.,0.,0.,0.,.3585685828,0.,2.913024938,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,-1.,0.,0.,-.4629100500,.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,-7.676494735,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.06230590,0.,-30.89729531,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-33.37289319,-.5000000000,0.,0.,-3.354101966,0.,-11.12429773,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.06230590,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,-3.354101966,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,2.288688541,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,6.141195788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.633249580,-.3779644731,0.,-3.070597895,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,-.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,3.027650356,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,16.24807681,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,26.32489316,0.,0.,0.,2.645751311,0.,17.54992877,.5000000000,0.,4.062019202,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,1.080123450,0.,0.,1.500000000,.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,-3.354101966,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,.2538762001,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,1.703053324,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,3.679015663,.6614378278,0.,0.,2.218529919,0.,3.679015663,.4192627457,0.,1.703053324,.2538762001},
			{0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,0.,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,0.,.6614378278,0.,0.,2.218529919,0.,0.,.4192627457,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708245,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,-1.811422094,0.,-2.085805824,.7905694150,0.,0.,0.,1.045825033,0.,0.,0.,0.,-1.501951855,.8100925875,0.,0.,1.811422094,0.,1.501951855,.6846531970,0.,2.085805824,.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,-1.811422094,0.,0.,.7905694150,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,.8100925872,0.,0.,1.811422094,0.,0.,.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,.7905694148,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,.8794529546,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,-1.899835520,.5123475385,0.,0.,0.,0.,-1.899835520,.6495190530,0.,.8794529546,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,0.,.5123475380,0.,0.,0.,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,.8385254914,0.,.6436936521,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,1.390537216,.1250000000,0.,0.,-.8385254914,0.,-1.390537216,.3169328455,0.,-.6436936521,.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,.8385254914,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.8385254914,0.,0.,.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,-.7616286003,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,.5484352742,0.,0.,0.,-.3307189139,0.,.5484352742,.6250000000e-1,0.,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,-.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,.4528555234,0.,0.,-.2096313728,.3125000000e-1},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,-3.476343041,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219620,0.,-7.509759277,-.8100925875,0.,0.,-3.622844188,0.,-7.509759277,-.6846531970,0.,-3.476343041,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,0.,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219619,0.,0.,-.8100925872,0.,0.,-3.622844188,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,.5303300858,0.,.5352180242,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,0.,1.403121520,0.,2.154210993,-.2041241452,0.,0.,0.,-.5400617248,0.,0.,0.,0.,1.551209206,-.4183300132,0.,0.,-1.403121520,0.,-1.551209206,-.5303300858,0.,-2.154210993,-.5352180242},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,.5303300858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617250,0.,0.,1.403121520,0.,0.,-.2041241452,0.,0.,0.,-.5400617250,0.,0.,0.,0.,0.,-.4183300132,0.,0.,-1.403121520,0.,0.,-.5303300858,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.5400617248,0.,0.,0.,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.4986029592,0.,0.,0.,0.,.1250000000,0.,0.,.6846531969,0.,1.077105496,-.9682458365e-1,0.,0.,0.,0.,1.077105496,-.2454951265,0.,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6846531970,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.1919123684,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4145780988,0.,0.,0.,.1250000000,0.,.4145780988,-.4724555913e-1,0.,.1919123684,-.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931772,0.,4.397264773,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,9.499177600,.5123475385,0.,0.,3.436931772,0.,9.499177600,.6495190530,0.,4.397264773,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931771,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,0.,.5123475380,0.,0.,3.436931771,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.6495190530,0.,-1.495808878,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-1.077105496,.9682458365e-1,0.,0.,.6495190530,0.,1.077105496,.2454951265,0.,1.495808878,.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.6495190530,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.6495190530,0.,0.,.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,.1357025371,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1820640586,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3933033180,0.,0.,0.,0.,0.,-.3933033180,.4482107285e-1,0.,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,-.6250000000e-1,0.,0.,-.2893187812e-1,.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,-3.218468261,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,-6.952686082,-.1250000000,0.,0.,-1.677050983,0.,-6.952686082,-.3169328455,0.,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,0.,-.1250000000,0.,0.,-1.677050983,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.5757371053,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4145780988,0.,0.,0.,-.1250000000,0.,-.4145780988,-.4724555913e-1,0.,-.5757371053,-.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,1.269381000,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,2.742176371,0.,0.,0.,.3307189139,0.,2.742176371,.6250000000e-1,0.,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,.6750771565e-1,0.,0.,.9375000000e-1,.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,-.2096313728,-.3125000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,.2538762001,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,1.703053324,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,3.679015663,.6614378278,0.,0.,2.218529919,0.,3.679015663,.4192627457,0.,1.703053324,.2538762001},
			{0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,0.,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,0.,.6614378278,0.,0.,2.218529919,0.,0.,.4192627457,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708245,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,-1.811422094,0.,-2.085805824,.7905694150,0.,0.,0.,1.045825033,0.,0.,0.,0.,-1.501951855,.8100925875,0.,0.,1.811422094,0.,1.501951855,.6846531970,0.,2.085805824,.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,-1.811422094,0.,0.,.7905694150,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,.8100925872,0.,0.,1.811422094,0.,0.,.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,.7905694148,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,.8794529546,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,-1.899835520,.5123475385,0.,0.,0.,0.,-1.899835520,.6495190530,0.,.8794529546,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,0.,.5123475380,0.,0.,0.,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,.8385254914,0.,.6436936521,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,1.390537216,.1250000000,0.,0.,-.8385254914,0.,-1.390537216,.3169328455,0.,-.6436936521,.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,.8385254914,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.8385254914,0.,0.,.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,-.7616286003,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,.5484352742,0.,0.,0.,-.3307189139,0.,.5484352742,.6250000000e-1,0.,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,-.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,.4528555234,0.,0.,-.2096313728,.3125000000e-1},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,-3.476343041,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219620,0.,-7.509759277,-.8100925875,0.,0.,-3.622844188,0.,-7.509759277,-.6846531970,0.,-3.476343041,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,0.,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219619,0.,0.,-.8100925872,0.,0.,-3.622844188,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,.5303300858,0.,.5352180242,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,0.,1.403121520,0.,2.154210993,-.2041241452,0.,0.,0.,-.5400617248,0.,0.,0.,0.,1.551209206,-.4183300132,0.,0.,-1.403121520,0.,-1.551209206,-.5303300858,0.,-2.154210993,-.5352180242},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,.5303300858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617250,0.,0.,1.403121520,0.,0.,-.2041241452,0.,0.,0.,-.5400617250,0.,0.,0.,0.,0.,-.4183300132,0.,0.,-1.403121520,0.,0.,-.5303300858,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.5400617248,0.,0.,0.,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.4986029592,0.,0.,0.,0.,.1250000000,0.,0.,.6846531969,0.,1.077105496,-.9682458365e-1,0.,0.,0.,0.,1.077105496,-.2454951265,0.,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6846531970,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.1919123684,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4145780988,0.,0.,0.,.1250000000,0.,.4145780988,-.4724555913e-1,0.,.1919123684,-.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931772,0.,4.397264773,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,9.499177600,.5123475385,0.,0.,3.436931772,0.,9.499177600,.6495190530,0.,4.397264773,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931771,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,0.,.5123475380,0.,0.,3.436931771,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.6495190530,0.,-1.495808878,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-1.077105496,.9682458365e-1,0.,0.,.6495190530,0.,1.077105496,.2454951265,0.,1.495808878,.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.6495190530,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.6495190530,0.,0.,.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,.1357025371,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1820640586,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3933033180,0.,0.,0.,0.,0.,-.3933033180,.4482107285e-1,0.,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,-.6250000000e-1,0.,0.,-.2893187812e-1,.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,-3.218468261,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,-6.952686082,-.1250000000,0.,0.,-1.677050983,0.,-6.952686082,-.3169328455,0.,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,0.,-.1250000000,0.,0.,-1.677050983,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.5757371053,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4145780988,0.,0.,0.,-.1250000000,0.,-.4145780988,-.4724555913e-1,0.,-.5757371053,-.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,1.269381000,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,2.742176371,0.,0.,0.,.3307189139,0.,2.742176371,.6250000000e-1,0.,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,.6750771565e-1,0.,0.,.9375000000e-1,.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,-.2096313728,-.3125000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,.2538762001,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,1.703053324,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,3.679015663,.6614378278,0.,0.,2.218529919,0.,3.679015663,.4192627457,0.,1.703053324,.2538762001},
			{0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,0.,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,0.,.6614378278,0.,0.,2.218529919,0.,0.,.4192627457,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708245,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,-3.476343041,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219620,0.,-7.509759277,-.8100925875,0.,0.,-3.622844188,0.,-7.509759277,-.6846531970,0.,-3.476343041,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,0.,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219619,0.,0.,-.8100925872,0.,0.,-3.622844188,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931772,0.,4.397264773,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,9.499177600,.5123475385,0.,0.,3.436931772,0.,9.499177600,.6495190530,0.,4.397264773,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931771,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,0.,.5123475380,0.,0.,3.436931771,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,-3.218468261,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,-6.952686082,-.1250000000,0.,0.,-1.677050983,0.,-6.952686082,-.3169328455,0.,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,0.,-.1250000000,0.,0.,-1.677050983,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,1.269381000,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,2.742176371,0.,0.,0.,.3307189139,0.,2.742176371,.6250000000e-1,0.,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,-.2096313728,-.3125000000e-1},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,.8100925875,0.,0.,.6846531970,0.,.5182226234,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,1.811422094,0.,2.085805824,-.7905694150,0.,0.,0.,-1.045825033,0.,0.,0.,0.,1.501951855,-.8100925875,0.,0.,-1.811422094,0.,-1.501951855,-.6846531970,0.,-2.085805824,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,.8100925872,0.,0.,.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,1.811422094,0.,0.,-.7905694150,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,-.8100925872,0.,0.,-1.811422094,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694148,0.,0.,0.,.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,-.5303300858,0.,-.5352180242,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,0.,-1.403121520,0.,-2.154210993,.2041241452,0.,0.,0.,.5400617248,0.,0.,0.,0.,-1.551209206,.4183300132,0.,0.,1.403121520,0.,1.551209206,.5303300858,0.,2.154210993,.5352180242},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,-.5303300858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617250,0.,0.,-1.403121520,0.,0.,.2041241452,0.,0.,0.,.5400617250,0.,0.,0.,0.,0.,.4183300132,0.,0.,1.403121520,0.,0.,.5303300858,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.5400617248,0.,0.,0.,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2454951265,0.,.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6495190530,0.,1.495808878,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,1.077105496,-.9682458365e-1,0.,0.,-.6495190530,0.,-1.077105496,-.2454951265,0.,-1.495808878,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.6495190530,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,-.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.5757371053,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4145780988,0.,0.,0.,.1250000000,0.,.4145780988,.4724555913e-1,0.,.5757371053,.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,-.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,.8794529546,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,-1.899835520,.5123475385,0.,0.,0.,0.,-1.899835520,.6495190530,0.,.8794529546,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,0.,.5123475380,0.,0.,0.,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.4986029592,0.,0.,0.,0.,.1250000000,0.,0.,.6846531969,0.,1.077105496,-.9682458365e-1,0.,0.,0.,0.,1.077105496,-.2454951265,0.,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6846531970,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,.1357025371,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1820640586,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3933033180,0.,0.,0.,0.,0.,-.3933033180,.4482107285e-1,0.,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.3169328455,0.,.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.8385254914,0.,-.6436936521,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,-1.390537216,-.1250000000,0.,0.,.8385254914,0.,1.390537216,-.3169328455,0.,.6436936521,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.8385254914,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,-.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.1919123684,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4145780988,0.,0.,0.,-.1250000000,0.,-.4145780988,.4724555913e-1,0.,-.1919123684,.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,-.7616286003,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,.5484352742,0.,0.,0.,-.3307189139,0.,.5484352742,.6250000000e-1,0.,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,-.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,.2096313728,-.3125000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,.2538762001,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,1.703053324,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,3.679015663,.6614378278,0.,0.,2.218529919,0.,3.679015663,.4192627457,0.,1.703053324,.2538762001},
			{0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,0.,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,0.,.6614378278,0.,0.,2.218529919,0.,0.,.4192627457,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708245,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,-3.476343041,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219620,0.,-7.509759277,-.8100925875,0.,0.,-3.622844188,0.,-7.509759277,-.6846531970,0.,-3.476343041,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,0.,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219619,0.,0.,-.8100925872,0.,0.,-3.622844188,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931772,0.,4.397264773,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,9.499177600,.5123475385,0.,0.,3.436931772,0.,9.499177600,.6495190530,0.,4.397264773,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931771,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,0.,.5123475380,0.,0.,3.436931771,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,-3.218468261,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,-6.952686082,-.1250000000,0.,0.,-1.677050983,0.,-6.952686082,-.3169328455,0.,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,0.,-.1250000000,0.,0.,-1.677050983,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,1.269381000,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,2.742176371,0.,0.,0.,.3307189139,0.,2.742176371,.6250000000e-1,0.,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,-.2096313728,-.3125000000e-1},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,.8100925875,0.,0.,.6846531970,0.,.5182226234,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,1.811422094,0.,2.085805824,-.7905694150,0.,0.,0.,-1.045825033,0.,0.,0.,0.,1.501951855,-.8100925875,0.,0.,-1.811422094,0.,-1.501951855,-.6846531970,0.,-2.085805824,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,.8100925872,0.,0.,.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,1.811422094,0.,0.,-.7905694150,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,-.8100925872,0.,0.,-1.811422094,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694148,0.,0.,0.,.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,-.5303300858,0.,-.5352180242,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,0.,-1.403121520,0.,-2.154210993,.2041241452,0.,0.,0.,.5400617248,0.,0.,0.,0.,-1.551209206,.4183300132,0.,0.,1.403121520,0.,1.551209206,.5303300858,0.,2.154210993,.5352180242},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,-.5303300858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617250,0.,0.,-1.403121520,0.,0.,.2041241452,0.,0.,0.,.5400617250,0.,0.,0.,0.,0.,.4183300132,0.,0.,1.403121520,0.,0.,.5303300858,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.5400617248,0.,0.,0.,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2454951265,0.,.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6495190530,0.,1.495808878,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,1.077105496,-.9682458365e-1,0.,0.,-.6495190530,0.,-1.077105496,-.2454951265,0.,-1.495808878,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.6495190530,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,-.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.5757371053,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4145780988,0.,0.,0.,.1250000000,0.,.4145780988,.4724555913e-1,0.,.5757371053,.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,-.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,.8794529546,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,-1.899835520,.5123475385,0.,0.,0.,0.,-1.899835520,.6495190530,0.,.8794529546,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,0.,.5123475380,0.,0.,0.,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.4986029592,0.,0.,0.,0.,.1250000000,0.,0.,.6846531969,0.,1.077105496,-.9682458365e-1,0.,0.,0.,0.,1.077105496,-.2454951265,0.,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6846531970,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,.1357025371,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1820640586,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3933033180,0.,0.,0.,0.,0.,-.3933033180,.4482107285e-1,0.,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.3169328455,0.,.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.8385254914,0.,-.6436936521,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,-1.390537216,-.1250000000,0.,0.,.8385254914,0.,1.390537216,-.3169328455,0.,.6436936521,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.8385254914,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,-.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.1919123684,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4145780988,0.,0.,0.,-.1250000000,0.,-.4145780988,.4724555913e-1,0.,-.1919123684,.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,-.7616286003,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,.5484352742,0.,0.,0.,-.3307189139,0.,.5484352742,.6250000000e-1,0.,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,-.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,.2096313728,-.3125000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2357022604,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2132007163},
			{0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9622504490e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6900655595e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5976143047e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6085806194e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8703882798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9622504490e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2817180849e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1543033500e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1150109266e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6900655595e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1543033500e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7273929673e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5976143047e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1150109266e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6085806194e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.8703882798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2357022604,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2132007163},
			{0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9622504490e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6900655595e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5976143047e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6085806194e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8703882798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9622504490e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2817180849e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1543033500e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1150109266e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6900655595e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1543033500e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7273929673e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5976143047e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1150109266e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6085806194e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.8703882798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,4.690415760,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,14.49137675,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.73320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.43167672,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,11.48912529,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,30.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,35.49647870,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,64.80740698,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,86.94826050,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,14.49137675,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,64.80740698,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,137.4772708,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.73320053,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,86.94826050,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.43167672,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,11.48912529}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,4.690415760,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,14.49137675,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.73320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.43167672,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,11.48912529,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,30.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,35.49647870,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,64.80740698,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,86.94826050,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,14.49137675,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,64.80740698,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,137.4772708,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.73320053,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,86.94826050,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.43167672,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,11.48912529}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[56],
		T o_output[1]
	) {
		static const T matrix[1][56] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[56],
		T o_output[1]
	) {
		static const T matrix[1][56] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_w(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,.9574271080,1.414213562,1.224744872,1.054092553,.9354143465,.8485281372,1.118033988,.9682458365,.8333333333,.7395099729,.9165151392,.7937253933,.6831300514,.7745966692,.6708203931,.6700593942,1.414213562,1.224744872,1.054092553,.9354143465,.8485281372,1.369306394,1.185854122,1.020620726,.9057110469,1.183215957,1.024695077,.8819171036,1.024695077,.8874119674,.8989790125,1.118033988,.9682458365,.8333333333,.7395099729,1.183215957,1.024695077,.8819171036,1.080123450,.9354143465,.9710083125,.9165151392,.7937253933,.6831300514,1.024695077,.8874119674,.9710083125,.7745966692,.6708203931,.8989790125,.6700593942},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,1.825741858,2.108185107,2.041241452,1.932183566,1.825741858,1.443375673,1.666666667,1.613743061,1.527525232,1.183215956,1.366260101,1.322875656,1.,1.154700539,.8650429586,1.825741858,2.108185107,2.041241452,1.932183566,1.825741858,1.767766952,2.041241452,1.976423538,1.870828693,1.527525232,1.763834207,1.707825127,1.322875656,1.527525232,1.160576915,1.443375673,1.666666667,1.613743061,1.527525232,1.527525232,1.763834207,1.707825127,1.394433378,1.610152972,1.253566341,1.183215956,1.366260101,1.322875656,1.322875656,1.527525232,1.253566341,1.,1.154700539,1.160576915,.8650429586},
			{2.645751311,3.436931772,3.549647869,3.500000000,3.401680258,3.290611645,2.160246900,2.806243040,2.898275350,2.857738034,2.777460300,1.707825129,2.218529918,2.291287848,2.259240285,1.400000000,1.818653348,1.878297101,1.183215956,1.537042616,1.023532631,2.160246900,2.806243040,2.898275350,2.857738034,2.777460300,2.091650066,2.717133141,2.806243040,2.766992952,1.807392228,2.347871376,2.424871131,1.565247584,2.033316257,1.373213125,1.707825129,2.218529918,2.291287848,2.259240285,1.807392228,2.347871376,2.424871131,1.649915823,2.143303525,1.483239697,1.400000000,1.818653348,1.878297101,1.565247584,2.033316257,1.483239697,1.183215956,1.537042616,1.373213125,1.023532631},
			{3.,4.156921939,4.472135954,4.535573675,4.500000000,4.422166386,2.449489743,3.394112549,3.651483716,3.703280398,3.674234614,1.936491673,2.683281572,2.886751347,2.927700220,1.587450787,2.199636334,2.366431913,1.341640786,1.859032006,1.160576915,2.449489743,3.394112549,3.651483716,3.703280398,3.674234614,2.371708244,3.286335344,3.535533905,3.585685827,2.049390154,2.839718296,3.055050463,1.774823935,2.459268185,1.557077324,1.936491673,2.683281572,2.886751347,2.927700220,2.049390154,2.839718296,3.055050463,1.870828693,2.592296279,1.681835732,1.587450787,2.199636334,2.366431913,1.774823935,2.459268185,1.681835732,1.341640786,1.859032006,1.557077324,1.160576915},
			{3.316624790,4.787135539,5.297284632,5.484352741,5.527707984,5.500000000,2.708012802,3.908679800,4.325214791,4.477955262,4.513354671,2.140872097,3.090082702,3.419382528,3.540134475,1.754992877,2.533114026,2.803059554,1.483239697,2.140872097,1.283066056,2.708012802,3.908679800,4.325214791,4.477955262,4.513354671,2.622022120,3.784562941,4.187871214,4.335761539,2.265686063,3.270236146,3.618734322,1.962141687,2.832107579,1.721413752,2.140872097,3.090082702,3.419382528,3.540134475,2.265686063,3.270236146,3.618734322,2.068278942,2.985303509,1.859339360,1.754992877,2.533114026,2.803059554,1.962141687,2.832107579,1.859339360,1.483239697,2.140872097,1.721413752,1.283066056},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.7817359602,1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.643167672,1.423024947,1.224744872,1.086853256,1.496662955,1.296148140,1.115546702,1.355261854,1.173691195,1.230977710,.8660254040,.7500000000,.6454972245,.5728219619,.5196152424,1.341640786,1.161895004,1.,.8874119674,1.449137675,1.254990040,1.080123450,1.434274331,1.242118007,1.376274919,.5477225575,.4743416490,.4082482906,.3622844188,.9660917834,.8366600265,.7200822997,1.133893419,.9819805061,1.189237451,.3741657387,.3240370349,.2788866755,.7171371655,.6210590036,.8919280880,.2710523708,.2347382389,.5505099678,.2051629517},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.730085916,2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.121320343,2.449489743,2.371708244,2.244994432,1.932183566,2.231093404,2.160246900,1.749635531,2.020305089,1.589185390,1.118033988,1.290994449,1.250000000,1.183215956,1.118033988,1.732050808,2.,1.936491673,1.833030278,1.870828694,2.160246900,2.091650066,1.851640200,2.138089935,1.776763280,.7071067810,.8164965809,.7905694148,.7483314774,1.247219129,1.440164600,1.394433378,1.463850110,1.690308509,1.535298948,.4830458916,.5577733510,.5400617250,.9258201003,1.069044968,1.151474211,.3499271062,.4040610177,.7107053122,.2648642316},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.686773158,2.645751311,3.436931771,3.549647869,3.500000000,3.401680258,2.509980080,3.260559769,3.367491647,3.320391543,2.286190427,2.969848480,3.067246323,2.070196679,2.689264371,1.880349511,1.322875656,1.718465886,1.774823935,1.750000000,1.700840129,2.049390152,2.662235902,2.749545417,2.711088342,2.213594362,2.875543427,2.969848480,2.190890230,2.846049894,2.102294664,.8366600265,1.086853256,1.122497216,1.106797181,1.475729575,1.917028952,1.979898987,1.732050808,2.250000000,1.816590213,.5715476066,.7424621200,.7668115806,1.095445115,1.423024947,1.362442660,.4140393357,.5378528742,.8409178659,.3133915852},
			{2.449489743,3.394112549,3.651483718,3.703280400,3.674234614,3.610683736,3.,4.156921939,4.472135954,4.535573675,4.500000000,2.846049894,3.943602415,4.242640686,4.302822994,2.592296280,3.591991092,3.864367134,2.347382390,3.252628452,2.132115937,1.500000000,2.078460970,2.236067977,2.267786838,2.250000000,2.323790008,3.219937887,3.464101616,3.513240264,2.509980080,3.477930420,3.741657387,2.484236014,3.442258395,2.383778085,.9486832980,1.314534138,1.414213562,1.434274331,1.673320053,2.318620280,2.494438258,1.963961013,2.721344207,2.059819687,.6480740700,.8979977729,.9660917834,1.242118007,1.721129197,1.544864765,.4694764779,.6505256900,.9535112343,.3553526559},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.6180165404,1.643167673,1.423024947,1.224744872,1.086853256,.9859006037,1.732050808,1.500000000,1.290994449,1.145643924,1.690308509,1.463850109,1.259881577,1.607142857,1.391826542,1.513825177,.5477225576,.4743416490,.4082482906,.3622844188,.3286335346,1.060660172,.9185586536,.7905694150,.7015607599,1.309307341,1.133893419,.9759000724,1.417366774,1.227475633,1.450721144,.2886751347,.2500000000,.2151657415,.1909406540,.6546536709,.5669467095,.4879500362,.8964214566,.7763237544,1.044638617,.1690308509,.1463850109,.1259881577,.4252100321,.3682426898,.6267831705,.1071428571,.9278843614e-1,.2901442286,.7208691321e-1},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,1.367753011,2.121320343,2.449489743,2.371708244,2.244994432,2.121320343,2.236067977,2.581988897,2.500000000,2.366431913,2.182178903,2.519763153,2.439750183,2.074812507,2.395787117,1.954339899,.7071067810,.8164965809,.7905694148,.7483314772,.7071067810,1.369306394,1.581138830,1.530931089,1.449137675,1.690308509,1.951800147,1.889822365,1.829812638,2.112885637,1.872872943,.3726779962,.4303314828,.4166666667,.3944053189,.8451542550,.9759000734,.9449111824,1.157275125,1.336306209,1.348622656,.2182178903,.2519763153,.2439750183,.5489437912,.6338656908,.8091735938,.1383208338,.1597191412,.3745745885,.9306380472e-1},
			{1.707825129,2.218529919,2.291287848,2.259240285,2.195775165,2.124080684,2.509980080,3.260559769,3.367491648,3.320391543,3.227117245,2.645751311,3.436931771,3.549647870,3.500000000,2.581988897,3.354101966,3.464101616,2.454951265,3.189075241,2.312406154,.8366600265,1.086853256,1.122497216,1.106797181,1.075705749,1.620185174,2.104682280,2.173706512,2.143303525,2.,2.598076212,2.683281572,2.165063510,2.812500000,2.216013152,.4409585519,.5728219619,.5916079783,.5833333333,1.,1.299038106,1.341640786,1.369306394,1.778781184,1.595711847,.2581988897,.3354101966,.3464101616,.6495190530,.8437500000,.9574271080,.1636634177,.2126050161,.4432026303,.1101145788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.5066228049,1.496662955,1.296148140,1.115546702,.9899494934,.8979977729,1.690308509,1.463850109,1.259881577,1.118033988,1.732050808,1.500000000,1.290994449,1.707825127,1.479019946,1.654623153,.3741657387,.3240370349,.2788866754,.2474873734,.2244994432,.8280786708,.7171371654,.6172133998,.5477225574,1.118033988,.9682458365,.8333333333,1.290994449,1.118033988,1.387443692,.1690308509,.1463850109,.1259881577,.1118033988,.4472135954,.3872983346,.3333333333,.6804138176,.5892556509,.8563488387,.8660254040e-1,.7500000000e-1,.6454972245e-1,.2581988897,.2236067977,.4281744193,.4879500362e-1,.4225771273e-1,.1585649934,.2954684202e-1},
			{1.183215957,1.366260102,1.322875656,1.252198067,1.183215957,1.121223821,1.932183566,2.231093404,2.160246900,2.044830882,1.932183566,2.182178903,2.519763153,2.439750183,2.309401077,2.236067977,2.581988897,2.500000000,2.204792759,2.545875387,2.136109305,.4830458916,.5577733510,.5400617250,.5112077204,.4830458916,1.069044968,1.234426800,1.195228610,1.131370850,1.443375673,1.666666667,1.613743061,1.666666667,1.924500898,1.791182105,.2182178903,.2519763153,.2439750183,.2309401077,.5773502693,.6666666667,.6454972245,.8784104612,1.014301033,1.105541597,.1118033988,.1290994449,.1250000000,.3333333333,.3849001795,.5527707984,.6299407883e-1,.7273929675e-1,.2047065263,.3814480901e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.7745966692,.6708203931,.5773502693,.5123475385,.4647580015,.4281744193,1.355261854,1.173691195,1.010152544,.8964214566,.8131571124,1.607142857,1.391826542,1.197893559,1.063025080,1.707825129,1.479019946,1.272937693,1.732050808,1.500000000,1.716232660,.2710523708,.2347382389,.2020305089,.1792842914,.1626314225,.6561133241,.5682108062,.4890379978,.4339781718,.9449111824,.8183170887,.7042952119,1.145643924,.9921567416,1.279204298,.1071428571,.9278843614e-1,.7985957060e-1,.7086833870e-1,.3149703940,.2727723628,.2347650707,.5175491696,.4482107285,.6908492799,.4879500367e-1,.4225771273e-1,.3636964838e-1,.1636634177,.1417366774,.2960782628,.2474358298e-1,.2142857143e-1,.9137173554e-1,.1362089413e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.7817359602,.8660254040,.7500000000,.6454972245,.5728219619,.5196152424,.5477225575,.4743416490,.4082482906,.3622844188,.3741657387,.3240370349,.2788866755,.2710523708,.2347382389,.2051629517,1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.341640786,1.161895004,1.,.8874119674,.9660917834,.8366600265,.7200822997,.7171371655,.6210590036,.5505099678,1.643167672,1.423024947,1.224744872,1.086853256,1.449137675,1.254990040,1.080123450,1.133893419,.9819805061,.8919280880,1.496662955,1.296148140,1.115546702,1.434274331,1.242118007,1.189237451,1.355261854,1.173691195,1.376274919,1.230977710},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.730085916,1.118033988,1.290994449,1.250000000,1.183215956,1.118033988,.7071067810,.8164965809,.7905694148,.7483314774,.4830458916,.5577733510,.5400617250,.3499271062,.4040610177,.2648642316,2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.732050808,2.,1.936491673,1.833030278,1.247219129,1.440164600,1.394433378,.9258201003,1.069044968,.7107053122,2.121320343,2.449489743,2.371708244,2.244994432,1.870828694,2.160246900,2.091650066,1.463850110,1.690308509,1.151474211,1.932183566,2.231093404,2.160246900,1.851640200,2.138089935,1.535298948,1.749635531,2.020305089,1.776763280,1.589185390},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.686773158,1.322875656,1.718465886,1.774823935,1.750000000,1.700840129,.8366600265,1.086853256,1.122497216,1.106797181,.5715476066,.7424621200,.7668115806,.4140393357,.5378528742,.3133915852,2.645751311,3.436931771,3.549647869,3.500000000,3.401680258,2.049390152,2.662235902,2.749545417,2.711088342,1.475729575,1.917028952,1.979898987,1.095445115,1.423024947,.8409178659,2.509980080,3.260559769,3.367491647,3.320391543,2.213594362,2.875543427,2.969848480,1.732050808,2.250000000,1.362442660,2.286190427,2.969848480,3.067246323,2.190890230,2.846049894,1.816590213,2.070196679,2.689264371,2.102294664,1.880349511},
			{2.449489743,3.394112549,3.651483718,3.703280400,3.674234614,3.610683736,1.500000000,2.078460970,2.236067977,2.267786838,2.250000000,.9486832980,1.314534138,1.414213562,1.434274331,.6480740700,.8979977729,.9660917834,.4694764779,.6505256900,.3553526559,3.,4.156921939,4.472135954,4.535573675,4.500000000,2.323790008,3.219937887,3.464101616,3.513240264,1.673320053,2.318620280,2.494438258,1.242118007,1.721129197,.9535112343,2.846049894,3.943602415,4.242640686,4.302822994,2.509980080,3.477930420,3.741657387,1.963961013,2.721344207,1.544864765,2.592296280,3.591991092,3.864367134,2.484236014,3.442258395,2.059819687,2.347382390,3.252628452,2.383778085,2.132115937},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,.8215838360,.7569125884,1.341640786,1.161895004,1.,.8874119673,.8049844717,1.060660172,.9185586536,.7905694148,.7015607601,.8280786714,.7171371654,.6172134000,.6561133241,.5682108062,.5297284632,1.341640786,1.161895004,1.,.8874119673,.8049844717,1.732050808,1.500000000,1.290994449,1.145643924,1.603567452,1.388730150,1.195228610,1.388730150,1.202675589,1.184508853,1.060660172,.9185586536,.7905694148,.7015607601,1.603567452,1.388730150,1.195228610,1.646831374,1.426197805,1.535298948,.8280786714,.7171371654,.6172134000,1.388730150,1.202675589,1.535298948,.6561133241,.5682108062,1.184508853,.5297284632},
			{1.767766952,2.041241452,1.976423538,1.870828694,1.767766952,1.675148486,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.069044968,1.234426800,1.195228610,.8470386591,.9780759957,.6838765053,1.732050808,2.,1.936491673,1.833030278,1.732050808,2.236067977,2.581988897,2.500000000,2.366431913,2.070196679,2.390457218,2.314550249,1.792842914,2.070196679,1.529194354,1.369306394,1.581138830,1.530931089,1.449137675,2.070196679,2.390457218,2.314550249,2.126050161,2.454951265,1.982062418,1.069044968,1.234426800,1.195228610,1.792842914,2.070196679,1.982062418,.8470386591,.9780759957,1.529194354,.6838765053},
			{2.091650066,2.717133141,2.806243040,2.766992952,2.689264370,2.601456923,2.049390154,2.662235902,2.749545418,2.711088342,2.634930198,1.620185175,2.104682280,2.173706512,2.143303525,1.264911064,1.643167672,1.697056274,1.002229657,1.301934515,.8091735938,2.049390154,2.662235902,2.749545418,2.711088342,2.634930198,2.645751311,3.436931772,3.549647870,3.500000000,2.449489743,3.181980514,3.286335344,2.121320343,2.755675961,1.809367160,1.620185175,2.104682280,2.173706512,2.143303525,2.449489743,3.181980514,3.286335344,2.515576474,3.267829698,2.345207880,1.264911064,1.643167672,1.697056274,2.121320343,2.755675961,2.345207880,1.002229657,1.301934515,1.809367160,.8091735938},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.6540472289,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,1.309307342,1.133893419,.9759000734,.8660254040,1.118033988,.9682458365,.8333333333,.9449111824,.8183170887,.8010409895,.9660917834,.8366600265,.7200822999,.6390096507,.5796550700,1.603567452,1.388730150,1.195228610,1.060660172,1.732050808,1.500000000,1.290994449,1.666666667,1.443375673,1.535298947,.6546536711,.5669467095,.4879500367,.4330127020,1.299038106,1.125000000,.9682458365,1.581138830,1.369306394,1.658312395,.4472135954,.3872983346,.3333333333,1.,.8660254040,1.326649916,.3149703940,.2727723628,.7676494735,.2288688541},
			{1.527525232,1.763834207,1.707825127,1.616580754,1.527525232,1.447493729,1.870828694,2.160246899,2.091650066,1.979898987,1.870828694,1.690308509,1.951800145,1.889822365,1.788854382,1.443375673,1.666666667,1.613743061,1.219875090,1.408590424,1.034139471,1.247219129,1.440164599,1.394433378,1.319932658,1.247219129,2.070196678,2.390457218,2.314550249,2.190890230,2.236067977,2.581988897,2.500000000,2.151657415,2.484519974,1.982062418,.8451542548,.9759000724,.9449111824,.8944271908,1.677050983,1.936491673,1.875000000,2.041241452,2.357022604,2.140872097,.5773502693,.6666666667,.6454972245,1.290994449,1.490711985,1.712697677,.4066250301,.4695301414,.9910312088,.2954684202},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.024695077,.8874119674,.7637626160,.6777720856,.6148170462,.5664215159,1.434274331,1.242118007,1.069044968,.9486832980,.8605645990,1.417366774,1.227475633,1.056442819,.9375000000,1.290994449,1.118033988,.9622504490,1.145643924,.9921567416,1.009049958,.7171371655,.6210590036,.5345224840,.4743416490,.4302822994,1.388730150,1.202675589,1.035098339,.9185586536,1.666666667,1.443375673,1.242259987,1.732050808,1.500000000,1.692228225,.4252100321,.3682426898,.3169328455,.2812500000,1.,.8660254040,.7453559923,1.369306394,1.185854122,1.566698904,.2581988897,.2236067977,.1924500898,.6928203232,.6000000000,1.044465936,.1636634177,.1417366774,.4834937786,.1081124955},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.6180165404,.5477225576,.4743416490,.4082482906,.3622844188,.3286335346,.2886751347,.2500000000,.2151657415,.1909406540,.1690308509,.1463850109,.1259881577,.1071428571,.9278843614e-1,.7208691321e-1,1.643167673,1.423024947,1.224744872,1.086853256,.9859006037,1.060660172,.9185586536,.7905694150,.7015607599,.6546536709,.5669467095,.4879500362,.4252100321,.3682426898,.2901442286,1.732050808,1.500000000,1.290994449,1.145643924,1.309307341,1.133893419,.9759000724,.8964214566,.7763237544,.6267831705,1.690308509,1.463850109,1.259881577,1.417366774,1.227475633,1.044638617,1.607142857,1.391826542,1.450721144,1.513825177},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,1.367753011,.7071067810,.8164965809,.7905694148,.7483314772,.7071067810,.3726779962,.4303314828,.4166666667,.3944053189,.2182178903,.2519763153,.2439750183,.1383208338,.1597191412,.9306380472e-1,2.121320343,2.449489743,2.371708244,2.244994432,2.121320343,1.369306394,1.581138830,1.530931089,1.449137675,.8451542550,.9759000734,.9449111824,.5489437912,.6338656908,.3745745885,2.236067977,2.581988897,2.500000000,2.366431913,1.690308509,1.951800147,1.889822365,1.157275125,1.336306209,.8091735938,2.182178903,2.519763153,2.439750183,1.829812638,2.112885637,1.348622656,2.074812507,2.395787117,1.872872943,1.954339899},
			{1.707825129,2.218529919,2.291287848,2.259240285,2.195775165,2.124080684,.8366600265,1.086853256,1.122497216,1.106797181,1.075705749,.4409585519,.5728219619,.5916079783,.5833333333,.2581988897,.3354101966,.3464101616,.1636634177,.2126050161,.1101145788,2.509980080,3.260559769,3.367491648,3.320391543,3.227117245,1.620185174,2.104682280,2.173706512,2.143303525,1.,1.299038106,1.341640786,.6495190530,.8437500000,.4432026303,2.645751311,3.436931771,3.549647870,3.500000000,2.,2.598076212,2.683281572,1.369306394,1.778781184,.9574271080,2.581988897,3.354101966,3.464101616,2.165063510,2.812500000,1.595711847,2.454951265,3.189075241,2.216013152,2.312406154},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.6540472289,.9660917834,.8366600265,.7200822999,.6390096507,.5796550700,.6546536711,.5669467095,.4879500367,.4330127020,.4472135954,.3872983346,.3333333333,.3149703940,.2727723628,.2288688541,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,1.603567452,1.388730150,1.195228610,1.060660172,1.299038106,1.125000000,.9682458365,1.,.8660254040,.7676494735,1.309307342,1.133893419,.9759000734,.8660254040,1.732050808,1.500000000,1.290994449,1.581138830,1.369306394,1.326649916,1.118033988,.9682458365,.8333333333,1.666666667,1.443375673,1.658312395,.9449111824,.8183170887,1.535298947,.8010409895},
			{1.527525232,1.763834207,1.707825127,1.616580754,1.527525232,1.447493729,1.247219129,1.440164599,1.394433378,1.319932658,1.247219129,.8451542548,.9759000724,.9449111824,.8944271908,.5773502693,.6666666667,.6454972245,.4066250301,.4695301414,.2954684202,1.870828694,2.160246899,2.091650066,1.979898987,1.870828694,2.070196678,2.390457218,2.314550249,2.190890230,1.677050983,1.936491673,1.875000000,1.290994449,1.490711985,.9910312088,1.690308509,1.951800145,1.889822365,1.788854382,2.236067977,2.581988897,2.500000000,2.041241452,2.357022604,1.712697677,1.443375673,1.666666667,1.613743061,2.151657415,2.484519974,2.140872097,1.219875090,1.408590424,1.982062418,1.034139471},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.080123450,.9354143465,.8050764862,.7144345085,.6480740697,.5970607017,1.133893419,.9819805061,.8451542548,.7500000000,.6803360513,.8964214566,.7763237544,.6681531047,.5929270612,.6804138176,.5892556509,.5071505163,.5175491696,.4482107285,.3988620175,1.133893419,.9819805061,.8451542548,.7500000000,.6803360513,1.646831372,1.426197805,1.227475633,1.089276566,1.581138830,1.369306394,1.178511302,1.369306394,1.185854122,1.146706181,.8964214566,.7763237544,.6681531047,.5929270612,1.581138830,1.369306394,1.178511302,1.732050808,1.500000000,1.651445648,.6804138176,.5892556509,.5071505163,1.369306394,1.185854122,1.651445648,.5175491696,.4482107285,1.146706181,.3988620175},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.5066228049,.3741657387,.3240370349,.2788866754,.2474873734,.2244994432,.1690308509,.1463850109,.1259881577,.1118033988,.8660254040e-1,.7500000000e-1,.6454972245e-1,.4879500362e-1,.4225771273e-1,.2954684202e-1,1.496662955,1.296148140,1.115546702,.9899494934,.8979977729,.8280786708,.7171371654,.6172133998,.5477225574,.4472135954,.3872983346,.3333333333,.2581988897,.2236067977,.1585649934,1.690308509,1.463850109,1.259881577,1.118033988,1.118033988,.9682458365,.8333333333,.6804138176,.5892556509,.4281744193,1.732050808,1.500000000,1.290994449,1.290994449,1.118033988,.8563488387,1.707825127,1.479019946,1.387443692,1.654623153},
			{1.183215957,1.366260102,1.322875656,1.252198067,1.183215957,1.121223821,.4830458916,.5577733510,.5400617250,.5112077204,.4830458916,.2182178903,.2519763153,.2439750183,.2309401077,.1118033988,.1290994449,.1250000000,.6299407883e-1,.7273929675e-1,.3814480901e-1,1.932183566,2.231093404,2.160246900,2.044830882,1.932183566,1.069044968,1.234426800,1.195228610,1.131370850,.5773502693,.6666666667,.6454972245,.3333333333,.3849001795,.2047065263,2.182178903,2.519763153,2.439750183,2.309401077,1.443375673,1.666666667,1.613743061,.8784104612,1.014301033,.5527707984,2.236067977,2.581988897,2.500000000,1.666666667,1.924500898,1.105541597,2.204792759,2.545875387,1.791182105,2.136109305},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.024695077,.8874119674,.7637626160,.6777720856,.6148170462,.5664215159,.7171371655,.6210590036,.5345224840,.4743416490,.4302822994,.4252100321,.3682426898,.3169328455,.2812500000,.2581988897,.2236067977,.1924500898,.1636634177,.1417366774,.1081124955,1.434274331,1.242118007,1.069044968,.9486832980,.8605645990,1.388730150,1.202675589,1.035098339,.9185586536,1.,.8660254040,.7453559923,.6928203232,.6000000000,.4834937786,1.417366774,1.227475633,1.056442819,.9375000000,1.666666667,1.443375673,1.242259987,1.369306394,1.185854122,1.044465936,1.290994449,1.118033988,.9622504490,1.732050808,1.500000000,1.566698904,1.145643924,.9921567416,1.692228225,1.009049958},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.7745966692,.6708203931,.5773502693,.5123475385,.4647580015,.4281744193,.2710523708,.2347382389,.2020305089,.1792842914,.1626314225,.1071428571,.9278843614e-1,.7985957060e-1,.7086833870e-1,.4879500367e-1,.4225771273e-1,.3636964838e-1,.2474358298e-1,.2142857143e-1,.1362089413e-1,1.355261854,1.173691195,1.010152544,.8964214566,.8131571124,.6561133241,.5682108062,.4890379978,.4339781718,.3149703940,.2727723628,.2347650707,.1636634177,.1417366774,.9137173554e-1,1.607142857,1.391826542,1.197893559,1.063025080,.9449111824,.8183170887,.7042952119,.5175491696,.4482107285,.2960782628,1.707825129,1.479019946,1.272937693,1.145643924,.9921567416,.6908492799,1.732050808,1.500000000,1.279204298,1.716232660},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_w(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,.9574271080,1.414213562,1.224744872,1.054092553,.9354143465,.8485281372,1.118033988,.9682458365,.8333333333,.7395099729,.9165151392,.7937253933,.6831300514,.7745966692,.6708203931,.6700593942,1.414213562,1.224744872,1.054092553,.9354143465,.8485281372,1.369306394,1.185854122,1.020620726,.9057110469,1.183215957,1.024695077,.8819171036,1.024695077,.8874119674,.8989790125,1.118033988,.9682458365,.8333333333,.7395099729,1.183215957,1.024695077,.8819171036,1.080123450,.9354143465,.9710083125,.9165151392,.7937253933,.6831300514,1.024695077,.8874119674,.9710083125,.7745966692,.6708203931,.8989790125,.6700593942},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,1.825741858,2.108185107,2.041241452,1.932183566,1.825741858,1.443375673,1.666666667,1.613743061,1.527525232,1.183215956,1.366260101,1.322875656,1.,1.154700539,.8650429586,1.825741858,2.108185107,2.041241452,1.932183566,1.825741858,1.767766952,2.041241452,1.976423538,1.870828693,1.527525232,1.763834207,1.707825127,1.322875656,1.527525232,1.160576915,1.443375673,1.666666667,1.613743061,1.527525232,1.527525232,1.763834207,1.707825127,1.394433378,1.610152972,1.253566341,1.183215956,1.366260101,1.322875656,1.322875656,1.527525232,1.253566341,1.,1.154700539,1.160576915,.8650429586},
			{2.645751311,3.436931772,3.549647869,3.500000000,3.401680258,3.290611645,2.160246900,2.806243040,2.898275350,2.857738034,2.777460300,1.707825129,2.218529918,2.291287848,2.259240285,1.400000000,1.818653348,1.878297101,1.183215956,1.537042616,1.023532631,2.160246900,2.806243040,2.898275350,2.857738034,2.777460300,2.091650066,2.717133141,2.806243040,2.766992952,1.807392228,2.347871376,2.424871131,1.565247584,2.033316257,1.373213125,1.707825129,2.218529918,2.291287848,2.259240285,1.807392228,2.347871376,2.424871131,1.649915823,2.143303525,1.483239697,1.400000000,1.818653348,1.878297101,1.565247584,2.033316257,1.483239697,1.183215956,1.537042616,1.373213125,1.023532631},
			{3.,4.156921939,4.472135954,4.535573675,4.500000000,4.422166386,2.449489743,3.394112549,3.651483716,3.703280398,3.674234614,1.936491673,2.683281572,2.886751347,2.927700220,1.587450787,2.199636334,2.366431913,1.341640786,1.859032006,1.160576915,2.449489743,3.394112549,3.651483716,3.703280398,3.674234614,2.371708244,3.286335344,3.535533905,3.585685827,2.049390154,2.839718296,3.055050463,1.774823935,2.459268185,1.557077324,1.936491673,2.683281572,2.886751347,2.927700220,2.049390154,2.839718296,3.055050463,1.870828693,2.592296279,1.681835732,1.587450787,2.199636334,2.366431913,1.774823935,2.459268185,1.681835732,1.341640786,1.859032006,1.557077324,1.160576915},
			{3.316624790,4.787135539,5.297284632,5.484352741,5.527707984,5.500000000,2.708012802,3.908679800,4.325214791,4.477955262,4.513354671,2.140872097,3.090082702,3.419382528,3.540134475,1.754992877,2.533114026,2.803059554,1.483239697,2.140872097,1.283066056,2.708012802,3.908679800,4.325214791,4.477955262,4.513354671,2.622022120,3.784562941,4.187871214,4.335761539,2.265686063,3.270236146,3.618734322,1.962141687,2.832107579,1.721413752,2.140872097,3.090082702,3.419382528,3.540134475,2.265686063,3.270236146,3.618734322,2.068278942,2.985303509,1.859339360,1.754992877,2.533114026,2.803059554,1.962141687,2.832107579,1.859339360,1.483239697,2.140872097,1.721413752,1.283066056},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.7817359602,1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.643167672,1.423024947,1.224744872,1.086853256,1.496662955,1.296148140,1.115546702,1.355261854,1.173691195,1.230977710,.8660254040,.7500000000,.6454972245,.5728219619,.5196152424,1.341640786,1.161895004,1.,.8874119674,1.449137675,1.254990040,1.080123450,1.434274331,1.242118007,1.376274919,.5477225575,.4743416490,.4082482906,.3622844188,.9660917834,.8366600265,.7200822997,1.133893419,.9819805061,1.189237451,.3741657387,.3240370349,.2788866755,.7171371655,.6210590036,.8919280880,.2710523708,.2347382389,.5505099678,.2051629517},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.730085916,2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.121320343,2.449489743,2.371708244,2.244994432,1.932183566,2.231093404,2.160246900,1.749635531,2.020305089,1.589185390,1.118033988,1.290994449,1.250000000,1.183215956,1.118033988,1.732050808,2.,1.936491673,1.833030278,1.870828694,2.160246900,2.091650066,1.851640200,2.138089935,1.776763280,.7071067810,.8164965809,.7905694148,.7483314774,1.247219129,1.440164600,1.394433378,1.463850110,1.690308509,1.535298948,.4830458916,.5577733510,.5400617250,.9258201003,1.069044968,1.151474211,.3499271062,.4040610177,.7107053122,.2648642316},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.686773158,2.645751311,3.436931771,3.549647869,3.500000000,3.401680258,2.509980080,3.260559769,3.367491647,3.320391543,2.286190427,2.969848480,3.067246323,2.070196679,2.689264371,1.880349511,1.322875656,1.718465886,1.774823935,1.750000000,1.700840129,2.049390152,2.662235902,2.749545417,2.711088342,2.213594362,2.875543427,2.969848480,2.190890230,2.846049894,2.102294664,.8366600265,1.086853256,1.122497216,1.106797181,1.475729575,1.917028952,1.979898987,1.732050808,2.250000000,1.816590213,.5715476066,.7424621200,.7668115806,1.095445115,1.423024947,1.362442660,.4140393357,.5378528742,.8409178659,.3133915852},
			{2.449489743,3.394112549,3.651483718,3.703280400,3.674234614,3.610683736,3.,4.156921939,4.472135954,4.535573675,4.500000000,2.846049894,3.943602415,4.242640686,4.302822994,2.592296280,3.591991092,3.864367134,2.347382390,3.252628452,2.132115937,1.500000000,2.078460970,2.236067977,2.267786838,2.250000000,2.323790008,3.219937887,3.464101616,3.513240264,2.509980080,3.477930420,3.741657387,2.484236014,3.442258395,2.383778085,.9486832980,1.314534138,1.414213562,1.434274331,1.673320053,2.318620280,2.494438258,1.963961013,2.721344207,2.059819687,.6480740700,.8979977729,.9660917834,1.242118007,1.721129197,1.544864765,.4694764779,.6505256900,.9535112343,.3553526559},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.6180165404,1.643167673,1.423024947,1.224744872,1.086853256,.9859006037,1.732050808,1.500000000,1.290994449,1.145643924,1.690308509,1.463850109,1.259881577,1.607142857,1.391826542,1.513825177,.5477225576,.4743416490,.4082482906,.3622844188,.3286335346,1.060660172,.9185586536,.7905694150,.7015607599,1.309307341,1.133893419,.9759000724,1.417366774,1.227475633,1.450721144,.2886751347,.2500000000,.2151657415,.1909406540,.6546536709,.5669467095,.4879500362,.8964214566,.7763237544,1.044638617,.1690308509,.1463850109,.1259881577,.4252100321,.3682426898,.6267831705,.1071428571,.9278843614e-1,.2901442286,.7208691321e-1},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,1.367753011,2.121320343,2.449489743,2.371708244,2.244994432,2.121320343,2.236067977,2.581988897,2.500000000,2.366431913,2.182178903,2.519763153,2.439750183,2.074812507,2.395787117,1.954339899,.7071067810,.8164965809,.7905694148,.7483314772,.7071067810,1.369306394,1.581138830,1.530931089,1.449137675,1.690308509,1.951800147,1.889822365,1.829812638,2.112885637,1.872872943,.3726779962,.4303314828,.4166666667,.3944053189,.8451542550,.9759000734,.9449111824,1.157275125,1.336306209,1.348622656,.2182178903,.2519763153,.2439750183,.5489437912,.6338656908,.8091735938,.1383208338,.1597191412,.3745745885,.9306380472e-1},
			{1.707825129,2.218529919,2.291287848,2.259240285,2.195775165,2.124080684,2.509980080,3.260559769,3.367491648,3.320391543,3.227117245,2.645751311,3.436931771,3.549647870,3.500000000,2.581988897,3.354101966,3.464101616,2.454951265,3.189075241,2.312406154,.8366600265,1.086853256,1.122497216,1.106797181,1.075705749,1.620185174,2.104682280,2.173706512,2.143303525,2.,2.598076212,2.683281572,2.165063510,2.812500000,2.216013152,.4409585519,.5728219619,.5916079783,.5833333333,1.,1.299038106,1.341640786,1.369306394,1.778781184,1.595711847,.2581988897,.3354101966,.3464101616,.6495190530,.8437500000,.9574271080,.1636634177,.2126050161,.4432026303,.1101145788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.5066228049,1.496662955,1.296148140,1.115546702,.9899494934,.8979977729,1.690308509,1.463850109,1.259881577,1.118033988,1.732050808,1.500000000,1.290994449,1.707825127,1.479019946,1.654623153,.3741657387,.3240370349,.2788866754,.2474873734,.2244994432,.8280786708,.7171371654,.6172133998,.5477225574,1.118033988,.9682458365,.8333333333,1.290994449,1.118033988,1.387443692,.1690308509,.1463850109,.1259881577,.1118033988,.4472135954,.3872983346,.3333333333,.6804138176,.5892556509,.8563488387,.8660254040e-1,.7500000000e-1,.6454972245e-1,.2581988897,.2236067977,.4281744193,.4879500362e-1,.4225771273e-1,.1585649934,.2954684202e-1},
			{1.183215957,1.366260102,1.322875656,1.252198067,1.183215957,1.121223821,1.932183566,2.231093404,2.160246900,2.044830882,1.932183566,2.182178903,2.519763153,2.439750183,2.309401077,2.236067977,2.581988897,2.500000000,2.204792759,2.545875387,2.136109305,.4830458916,.5577733510,.5400617250,.5112077204,.4830458916,1.069044968,1.234426800,1.195228610,1.131370850,1.443375673,1.666666667,1.613743061,1.666666667,1.924500898,1.791182105,.2182178903,.2519763153,.2439750183,.2309401077,.5773502693,.6666666667,.6454972245,.8784104612,1.014301033,1.105541597,.1118033988,.1290994449,.1250000000,.3333333333,.3849001795,.5527707984,.6299407883e-1,.7273929675e-1,.2047065263,.3814480901e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.7745966692,.6708203931,.5773502693,.5123475385,.4647580015,.4281744193,1.355261854,1.173691195,1.010152544,.8964214566,.8131571124,1.607142857,1.391826542,1.197893559,1.063025080,1.707825129,1.479019946,1.272937693,1.732050808,1.500000000,1.716232660,.2710523708,.2347382389,.2020305089,.1792842914,.1626314225,.6561133241,.5682108062,.4890379978,.4339781718,.9449111824,.8183170887,.7042952119,1.145643924,.9921567416,1.279204298,.1071428571,.9278843614e-1,.7985957060e-1,.7086833870e-1,.3149703940,.2727723628,.2347650707,.5175491696,.4482107285,.6908492799,.4879500367e-1,.4225771273e-1,.3636964838e-1,.1636634177,.1417366774,.2960782628,.2474358298e-1,.2142857143e-1,.9137173554e-1,.1362089413e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.7817359602,.8660254040,.7500000000,.6454972245,.5728219619,.5196152424,.5477225575,.4743416490,.4082482906,.3622844188,.3741657387,.3240370349,.2788866755,.2710523708,.2347382389,.2051629517,1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.341640786,1.161895004,1.,.8874119674,.9660917834,.8366600265,.7200822997,.7171371655,.6210590036,.5505099678,1.643167672,1.423024947,1.224744872,1.086853256,1.449137675,1.254990040,1.080123450,1.133893419,.9819805061,.8919280880,1.496662955,1.296148140,1.115546702,1.434274331,1.242118007,1.189237451,1.355261854,1.173691195,1.376274919,1.230977710},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.730085916,1.118033988,1.290994449,1.250000000,1.183215956,1.118033988,.7071067810,.8164965809,.7905694148,.7483314774,.4830458916,.5577733510,.5400617250,.3499271062,.4040610177,.2648642316,2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.732050808,2.,1.936491673,1.833030278,1.247219129,1.440164600,1.394433378,.9258201003,1.069044968,.7107053122,2.121320343,2.449489743,2.371708244,2.244994432,1.870828694,2.160246900,2.091650066,1.463850110,1.690308509,1.151474211,1.932183566,2.231093404,2.160246900,1.851640200,2.138089935,1.535298948,1.749635531,2.020305089,1.776763280,1.589185390},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.686773158,1.322875656,1.718465886,1.774823935,1.750000000,1.700840129,.8366600265,1.086853256,1.122497216,1.106797181,.5715476066,.7424621200,.7668115806,.4140393357,.5378528742,.3133915852,2.645751311,3.436931771,3.549647869,3.500000000,3.401680258,2.049390152,2.662235902,2.749545417,2.711088342,1.475729575,1.917028952,1.979898987,1.095445115,1.423024947,.8409178659,2.509980080,3.260559769,3.367491647,3.320391543,2.213594362,2.875543427,2.969848480,1.732050808,2.250000000,1.362442660,2.286190427,2.969848480,3.067246323,2.190890230,2.846049894,1.816590213,2.070196679,2.689264371,2.102294664,1.880349511},
			{2.449489743,3.394112549,3.651483718,3.703280400,3.674234614,3.610683736,1.500000000,2.078460970,2.236067977,2.267786838,2.250000000,.9486832980,1.314534138,1.414213562,1.434274331,.6480740700,.8979977729,.9660917834,.4694764779,.6505256900,.3553526559,3.,4.156921939,4.472135954,4.535573675,4.500000000,2.323790008,3.219937887,3.464101616,3.513240264,1.673320053,2.318620280,2.494438258,1.242118007,1.721129197,.9535112343,2.846049894,3.943602415,4.242640686,4.302822994,2.509980080,3.477930420,3.741657387,1.963961013,2.721344207,1.544864765,2.592296280,3.591991092,3.864367134,2.484236014,3.442258395,2.059819687,2.347382390,3.252628452,2.383778085,2.132115937},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,.8215838360,.7569125884,1.341640786,1.161895004,1.,.8874119673,.8049844717,1.060660172,.9185586536,.7905694148,.7015607601,.8280786714,.7171371654,.6172134000,.6561133241,.5682108062,.5297284632,1.341640786,1.161895004,1.,.8874119673,.8049844717,1.732050808,1.500000000,1.290994449,1.145643924,1.603567452,1.388730150,1.195228610,1.388730150,1.202675589,1.184508853,1.060660172,.9185586536,.7905694148,.7015607601,1.603567452,1.388730150,1.195228610,1.646831374,1.426197805,1.535298948,.8280786714,.7171371654,.6172134000,1.388730150,1.202675589,1.535298948,.6561133241,.5682108062,1.184508853,.5297284632},
			{1.767766952,2.041241452,1.976423538,1.870828694,1.767766952,1.675148486,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.069044968,1.234426800,1.195228610,.8470386591,.9780759957,.6838765053,1.732050808,2.,1.936491673,1.833030278,1.732050808,2.236067977,2.581988897,2.500000000,2.366431913,2.070196679,2.390457218,2.314550249,1.792842914,2.070196679,1.529194354,1.369306394,1.581138830,1.530931089,1.449137675,2.070196679,2.390457218,2.314550249,2.126050161,2.454951265,1.982062418,1.069044968,1.234426800,1.195228610,1.792842914,2.070196679,1.982062418,.8470386591,.9780759957,1.529194354,.6838765053},
			{2.091650066,2.717133141,2.806243040,2.766992952,2.689264370,2.601456923,2.049390154,2.662235902,2.749545418,2.711088342,2.634930198,1.620185175,2.104682280,2.173706512,2.143303525,1.264911064,1.643167672,1.697056274,1.002229657,1.301934515,.8091735938,2.049390154,2.662235902,2.749545418,2.711088342,2.634930198,2.645751311,3.436931772,3.549647870,3.500000000,2.449489743,3.181980514,3.286335344,2.121320343,2.755675961,1.809367160,1.620185175,2.104682280,2.173706512,2.143303525,2.449489743,3.181980514,3.286335344,2.515576474,3.267829698,2.345207880,1.264911064,1.643167672,1.697056274,2.121320343,2.755675961,2.345207880,1.002229657,1.301934515,1.809367160,.8091735938},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.6540472289,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,1.309307342,1.133893419,.9759000734,.8660254040,1.118033988,.9682458365,.8333333333,.9449111824,.8183170887,.8010409895,.9660917834,.8366600265,.7200822999,.6390096507,.5796550700,1.603567452,1.388730150,1.195228610,1.060660172,1.732050808,1.500000000,1.290994449,1.666666667,1.443375673,1.535298947,.6546536711,.5669467095,.4879500367,.4330127020,1.299038106,1.125000000,.9682458365,1.581138830,1.369306394,1.658312395,.4472135954,.3872983346,.3333333333,1.,.8660254040,1.326649916,.3149703940,.2727723628,.7676494735,.2288688541},
			{1.527525232,1.763834207,1.707825127,1.616580754,1.527525232,1.447493729,1.870828694,2.160246899,2.091650066,1.979898987,1.870828694,1.690308509,1.951800145,1.889822365,1.788854382,1.443375673,1.666666667,1.613743061,1.219875090,1.408590424,1.034139471,1.247219129,1.440164599,1.394433378,1.319932658,1.247219129,2.070196678,2.390457218,2.314550249,2.190890230,2.236067977,2.581988897,2.500000000,2.151657415,2.484519974,1.982062418,.8451542548,.9759000724,.9449111824,.8944271908,1.677050983,1.936491673,1.875000000,2.041241452,2.357022604,2.140872097,.5773502693,.6666666667,.6454972245,1.290994449,1.490711985,1.712697677,.4066250301,.4695301414,.9910312088,.2954684202},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.024695077,.8874119674,.7637626160,.6777720856,.6148170462,.5664215159,1.434274331,1.242118007,1.069044968,.9486832980,.8605645990,1.417366774,1.227475633,1.056442819,.9375000000,1.290994449,1.118033988,.9622504490,1.145643924,.9921567416,1.009049958,.7171371655,.6210590036,.5345224840,.4743416490,.4302822994,1.388730150,1.202675589,1.035098339,.9185586536,1.666666667,1.443375673,1.242259987,1.732050808,1.500000000,1.692228225,.4252100321,.3682426898,.3169328455,.2812500000,1.,.8660254040,.7453559923,1.369306394,1.185854122,1.566698904,.2581988897,.2236067977,.1924500898,.6928203232,.6000000000,1.044465936,.1636634177,.1417366774,.4834937786,.1081124955},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.6180165404,.5477225576,.4743416490,.4082482906,.3622844188,.3286335346,.2886751347,.2500000000,.2151657415,.1909406540,.1690308509,.1463850109,.1259881577,.1071428571,.9278843614e-1,.7208691321e-1,1.643167673,1.423024947,1.224744872,1.086853256,.9859006037,1.060660172,.9185586536,.7905694150,.7015607599,.6546536709,.5669467095,.4879500362,.4252100321,.3682426898,.2901442286,1.732050808,1.500000000,1.290994449,1.145643924,1.309307341,1.133893419,.9759000724,.8964214566,.7763237544,.6267831705,1.690308509,1.463850109,1.259881577,1.417366774,1.227475633,1.044638617,1.607142857,1.391826542,1.450721144,1.513825177},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,1.367753011,.7071067810,.8164965809,.7905694148,.7483314772,.7071067810,.3726779962,.4303314828,.4166666667,.3944053189,.2182178903,.2519763153,.2439750183,.1383208338,.1597191412,.9306380472e-1,2.121320343,2.449489743,2.371708244,2.244994432,2.121320343,1.369306394,1.581138830,1.530931089,1.449137675,.8451542550,.9759000734,.9449111824,.5489437912,.6338656908,.3745745885,2.236067977,2.581988897,2.500000000,2.366431913,1.690308509,1.951800147,1.889822365,1.157275125,1.336306209,.8091735938,2.182178903,2.519763153,2.439750183,1.829812638,2.112885637,1.348622656,2.074812507,2.395787117,1.872872943,1.954339899},
			{1.707825129,2.218529919,2.291287848,2.259240285,2.195775165,2.124080684,.8366600265,1.086853256,1.122497216,1.106797181,1.075705749,.4409585519,.5728219619,.5916079783,.5833333333,.2581988897,.3354101966,.3464101616,.1636634177,.2126050161,.1101145788,2.509980080,3.260559769,3.367491648,3.320391543,3.227117245,1.620185174,2.104682280,2.173706512,2.143303525,1.,1.299038106,1.341640786,.6495190530,.8437500000,.4432026303,2.645751311,3.436931771,3.549647870,3.500000000,2.,2.598076212,2.683281572,1.369306394,1.778781184,.9574271080,2.581988897,3.354101966,3.464101616,2.165063510,2.812500000,1.595711847,2.454951265,3.189075241,2.216013152,2.312406154},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.6540472289,.9660917834,.8366600265,.7200822999,.6390096507,.5796550700,.6546536711,.5669467095,.4879500367,.4330127020,.4472135954,.3872983346,.3333333333,.3149703940,.2727723628,.2288688541,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,1.603567452,1.388730150,1.195228610,1.060660172,1.299038106,1.125000000,.9682458365,1.,.8660254040,.7676494735,1.309307342,1.133893419,.9759000734,.8660254040,1.732050808,1.500000000,1.290994449,1.581138830,1.369306394,1.326649916,1.118033988,.9682458365,.8333333333,1.666666667,1.443375673,1.658312395,.9449111824,.8183170887,1.535298947,.8010409895},
			{1.527525232,1.763834207,1.707825127,1.616580754,1.527525232,1.447493729,1.247219129,1.440164599,1.394433378,1.319932658,1.247219129,.8451542548,.9759000724,.9449111824,.8944271908,.5773502693,.6666666667,.6454972245,.4066250301,.4695301414,.2954684202,1.870828694,2.160246899,2.091650066,1.979898987,1.870828694,2.070196678,2.390457218,2.314550249,2.190890230,1.677050983,1.936491673,1.875000000,1.290994449,1.490711985,.9910312088,1.690308509,1.951800145,1.889822365,1.788854382,2.236067977,2.581988897,2.500000000,2.041241452,2.357022604,1.712697677,1.443375673,1.666666667,1.613743061,2.151657415,2.484519974,2.140872097,1.219875090,1.408590424,1.982062418,1.034139471},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.080123450,.9354143465,.8050764862,.7144345085,.6480740697,.5970607017,1.133893419,.9819805061,.8451542548,.7500000000,.6803360513,.8964214566,.7763237544,.6681531047,.5929270612,.6804138176,.5892556509,.5071505163,.5175491696,.4482107285,.3988620175,1.133893419,.9819805061,.8451542548,.7500000000,.6803360513,1.646831372,1.426197805,1.227475633,1.089276566,1.581138830,1.369306394,1.178511302,1.369306394,1.185854122,1.146706181,.8964214566,.7763237544,.6681531047,.5929270612,1.581138830,1.369306394,1.178511302,1.732050808,1.500000000,1.651445648,.6804138176,.5892556509,.5071505163,1.369306394,1.185854122,1.651445648,.5175491696,.4482107285,1.146706181,.3988620175},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.5066228049,.3741657387,.3240370349,.2788866754,.2474873734,.2244994432,.1690308509,.1463850109,.1259881577,.1118033988,.8660254040e-1,.7500000000e-1,.6454972245e-1,.4879500362e-1,.4225771273e-1,.2954684202e-1,1.496662955,1.296148140,1.115546702,.9899494934,.8979977729,.8280786708,.7171371654,.6172133998,.5477225574,.4472135954,.3872983346,.3333333333,.2581988897,.2236067977,.1585649934,1.690308509,1.463850109,1.259881577,1.118033988,1.118033988,.9682458365,.8333333333,.6804138176,.5892556509,.4281744193,1.732050808,1.500000000,1.290994449,1.290994449,1.118033988,.8563488387,1.707825127,1.479019946,1.387443692,1.654623153},
			{1.183215957,1.366260102,1.322875656,1.252198067,1.183215957,1.121223821,.4830458916,.5577733510,.5400617250,.5112077204,.4830458916,.2182178903,.2519763153,.2439750183,.2309401077,.1118033988,.1290994449,.1250000000,.6299407883e-1,.7273929675e-1,.3814480901e-1,1.932183566,2.231093404,2.160246900,2.044830882,1.932183566,1.069044968,1.234426800,1.195228610,1.131370850,.5773502693,.6666666667,.6454972245,.3333333333,.3849001795,.2047065263,2.182178903,2.519763153,2.439750183,2.309401077,1.443375673,1.666666667,1.613743061,.8784104612,1.014301033,.5527707984,2.236067977,2.581988897,2.500000000,1.666666667,1.924500898,1.105541597,2.204792759,2.545875387,1.791182105,2.136109305},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.024695077,.8874119674,.7637626160,.6777720856,.6148170462,.5664215159,.7171371655,.6210590036,.5345224840,.4743416490,.4302822994,.4252100321,.3682426898,.3169328455,.2812500000,.2581988897,.2236067977,.1924500898,.1636634177,.1417366774,.1081124955,1.434274331,1.242118007,1.069044968,.9486832980,.8605645990,1.388730150,1.202675589,1.035098339,.9185586536,1.,.8660254040,.7453559923,.6928203232,.6000000000,.4834937786,1.417366774,1.227475633,1.056442819,.9375000000,1.666666667,1.443375673,1.242259987,1.369306394,1.185854122,1.044465936,1.290994449,1.118033988,.9622504490,1.732050808,1.500000000,1.566698904,1.145643924,.9921567416,1.692228225,1.009049958},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.7745966692,.6708203931,.5773502693,.5123475385,.4647580015,.4281744193,.2710523708,.2347382389,.2020305089,.1792842914,.1626314225,.1071428571,.9278843614e-1,.7985957060e-1,.7086833870e-1,.4879500367e-1,.4225771273e-1,.3636964838e-1,.2474358298e-1,.2142857143e-1,.1362089413e-1,1.355261854,1.173691195,1.010152544,.8964214566,.8131571124,.6561133241,.5682108062,.4890379978,.4339781718,.3149703940,.2727723628,.2347650707,.1636634177,.1417366774,.9137173554e-1,1.607142857,1.391826542,1.197893559,1.063025080,.9449111824,.8183170887,.7042952119,.5175491696,.4482107285,.2960782628,1.707825129,1.479019946,1.272937693,1.145643924,.9921567416,.6908492799,1.732050808,1.500000000,1.279204298,1.716232660},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_bottom_face(
		const T i_input[56],
		T o_output[52]
	) {
		static const T matrix[52][56] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.154700539,2.,2.581988897,3.055050463,3.464101616,.6085806194,1.054092553,1.360827635,1.610152972,.2771598064,.4800548665,.6197481677,.1171213948,.2028602065,.4728035100e-1,1.154700539,2.,2.581988897,3.055050463,3.464101616,1.490711985,2.581988897,3.333333333,3.944053189,1.073435315,1.859244503,2.400274333,.6197481677,1.073435315,.3171662365,.6085806194,1.054092553,1.360827635,1.610152972,1.073435315,1.859244503,2.400274333,.9799078930,1.697250258,.6851573789,.2771598064,.4800548665,.6197481677,.6197481677,1.073435315,.6851573789,.1171213948,.2028602065,.3171662365,.4728035100e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,1.354996613,2.346922978,3.029864538,3.584984067,.9207895610,1.594854303,2.058948051,.5805991817,1.005627282,.3497284165,.1814813775e-1,.3143349662e-1,.4058046965e-1,.4801545922e-1,.5444441322e-1,.3495960260e-1,.6055180793e-1,.7817204790e-1,.9249441444e-1,.3756286152e-1,.6506078461e-1,.8399311177e-1,.3235996830e-1,.5604910924e-1,.2471097213e-1,.1503292582e-3,.2603779131e-3,.3361464403e-3,.3977338322e-3,.3956501378e-3,.6852861406e-3,.8847006032e-3,.5389279955e-3,.9334506698e-3,.5622712446e-3,.1076015046e-5,.1863712730e-5,.2406042788e-5,.3590156597e-5,.6218333635e-5,.5922410770e-5,.7146395199e-8,.1237791958e-7,.2887671511e-7,.4534132255e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,1.354996613,2.346922978,3.029864538,3.584984067,.9207895610,1.594854303,2.058948051,.5805991817,1.005627282,.3497284165,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,3.319050305,5.748763763,7.421622105,8.781381698,3.566202636,6.176844154,7.974271513,3.072242092,5.321279398,2.346049539,1.354996613,2.346922978,3.029864538,3.584984067,3.566202636,6.176844154,7.974271513,4.857641268,8.413681482,5.068046242,.9207895610,1.594854303,2.058948051,3.072242092,5.321279398,5.068046242,.5805991817,1.005627282,2.346049539,.3497284165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1814813775e-1,.3143349662e-1,.4058046965e-1,.4801545922e-1,.5444441322e-1,.1503292582e-3,.2603779131e-3,.3361464403e-3,.3977338322e-3,.1076015046e-5,.1863712730e-5,.2406042788e-5,.7146395199e-8,.1237791958e-7,.4534132255e-10,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,.3495960260e-1,.6055180793e-1,.7817204790e-1,.9249441444e-1,.3956501378e-3,.6852861406e-3,.8847006032e-3,.3590156597e-5,.6218333635e-5,.2887671511e-7,1.354996613,2.346922978,3.029864538,3.584984067,.3756286152e-1,.6506078461e-1,.8399311177e-1,.5389279955e-3,.9334506698e-3,.5922410770e-5,.9207895610,1.594854303,2.058948051,.3235996830e-1,.5604910924e-1,.5622712446e-3,.5805991817,1.005627282,.2471097213e-1,.3497284165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9363699245,1.621840284,2.093786803,2.477401956,.5289617350,.9161886001,1.182794397,.2772649114,.4802369137,.1388366216,.5995013155,1.038366737,1.340525694,1.586131391,1.798503946,.9600175758,1.662799218,2.146664559,2.539967760,.8574839288,1.485205731,1.917392354,.6140871797,1.063630196,.3898220464,.1640437399,.2841320922,.3668129538,.4340189401,.3589071350,.6216453928,.8025407512,.4064022821,.7039094010,.3524730288,.3878756937e-1,.6718204084e-1,.8673164178e-1,.1075826497,.1863386154,.1475305609,.8509799423e-2,.1473940496e-1,.2858477025e-1,.1783546975e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9363699245,1.621840284,2.093786803,2.477401956,.5289617350,.9161886001,1.182794397,.2772649114,.4802369137,.1388366216,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,2.293628526,3.972681141,5.128709298,6.068370681,2.048659991,3.548383191,4.580943002,1.467148005,2.541174888,.9313443707,.9363699245,1.621840284,2.093786803,2.477401956,2.048659991,3.548383191,4.580943002,2.319764682,4.017950291,2.011933789,.5289617350,.9161886001,1.182794397,1.467148005,2.541174888,2.011933789,.2772649114,.4802369137,.9313443707,.1388366216},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5995013155,1.038366737,1.340525694,1.586131391,1.798503946,.1640437399,.2841320922,.3668129538,.4340189401,.3878756937e-1,.6718204084e-1,.8673164178e-1,.8509799423e-2,.1473940496e-1,.1783546975e-2,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9600175758,1.662799218,2.146664559,2.539967760,.3589071350,.6216453928,.8025407512,.1075826497,.1863386154,.2858477025e-1,.9363699245,1.621840284,2.093786803,2.477401956,.8574839288,1.485205731,1.917392354,.4064022821,.7039094010,.1475305609,.5289617350,.9161886001,1.182794397,.6140871797,1.063630196,.3524730288,.2772649114,.4802369137,.3898220464,.1388366216},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,1.212281421,2.099733015,2.710743667,3.207395161,.7792166912,1.349642899,1.742381491,.4647366240,.8049474449,.2647853914,.2046688294,.3544968112,.4576534152,.5415028235,.6140064883,.3729226366,.6459209540,.8338803660,.9866605549,.3790037856,.6564538126,.8474782279,.3088343411,.5349167702,.2230693744,.1911977567e-1,.3311642289e-1,.4275311811e-1,.5058617157e-1,.4759739883e-1,.8244111306e-1,.1064310193,.6132467595e-1,.1062174545,.6051778323e-1,.1543395612e-2,.2673239616e-2,.3451137504e-2,.4870851947e-2,.8436563051e-2,.7600159080e-2,.1156020631e-3,.2002286468e-3,.4418338282e-3,.8271654356e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,1.212281421,2.099733015,2.710743667,3.207395161,.7792166912,1.349642899,1.742381491,.4647366240,.8049474449,.2647853914,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,2.969470906,5.143274482,6.639938805,7.856481545,3.017893269,5.227144472,6.748214496,2.459155065,4.259381518,1.776234404,1.212281421,2.099733015,2.710743667,3.207395161,3.017893269,5.227144472,6.748214496,3.888265562,6.734673506,3.837104864,.7792166912,1.349642899,1.742381491,2.459155065,4.259381518,3.837104864,.4647366240,.8049474449,1.776234404,.2647853914},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2046688294,.3544968112,.4576534152,.5415028235,.6140064883,.1911977567e-1,.3311642289e-1,.4275311811e-1,.5058617157e-1,.1543395612e-2,.2673239616e-2,.3451137504e-2,.1156020631e-3,.2002286468e-3,.8271654356e-5,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,.3729226366,.6459209540,.8338803660,.9866605549,.4759739883e-1,.8244111306e-1,.1064310193,.4870851947e-2,.8436563051e-2,.4418338282e-3,1.212281421,2.099733015,2.710743667,3.207395161,.3790037856,.6564538126,.8474782279,.6132467595e-1,.1062174545,.7600159080e-2,.7792166912,1.349642899,1.742381491,.3088343411,.5349167702,.6051778323e-1,.4647366240,.8049474449,.2230693744,.2647853914},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.3169457192,.5489660888,.7087121734,.8385595524,.1041670603,.1804226409,.2329246278,.3176653023e-1,.5502124434e-1,.9254386281e-2,1.797496348,3.113355001,4.019324022,4.755728318,5.392489045,1.674656646,2.900590397,3.744646101,4.430725019,.8702445254,1.507307732,1.945925915,.3625885274,.6280217519,.1339118908,1.474739846,2.554324341,3.297618546,3.901794882,1.877182782,3.251375951,4.197508304,1.236657185,2.141953076,.6240046052,1.045505193,1.810868114,2.337820683,1.687114596,2.922168200,1.346024785,.6877503643,1.191218574,1.344048463,.4321892234},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.3169457192,.5489660888,.7087121734,.8385595524,.1041670603,.1804226409,.2329246278,.3176653023e-1,.5502124434e-1,.9254386281e-2,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.7763552880,1.344686804,1.735983199,2.054043021,.4034372898,.6987738835,.9021132044,.1680926781,.2911450589,.6208031047e-1,.3169457192,.5489660888,.7087121734,.8385595524,.4034372898,.6987738835,.9021132044,.2657778603,.4603407575,.1341087982,.1041670603,.1804226409,.2329246278,.1680926781,.2911450589,.1341087982,.3176653023e-1,.5502124434e-1,.6208031047e-1,.9254386281e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.797496348,3.113355001,4.019324022,4.755728318,5.392489045,1.474739846,2.554324341,3.297618546,3.901794882,1.045505193,1.810868114,2.337820683,.6877503643,1.191218574,.4321892234,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,1.674656646,2.900590397,3.744646101,4.430725019,1.877182782,3.251375951,4.197508304,1.687114596,2.922168200,1.344048463,.3169457192,.5489660888,.7087121734,.8385595524,.8702445254,1.507307732,1.945925915,1.236657185,2.141953076,1.346024785,.1041670603,.1804226409,.2329246278,.3625885274,.6280217519,.6240046052,.3176653023e-1,.5502124434e-1,.1339118908,.9254386281e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.1199176989,.2077035473,.2681441264,.3172724093,.2424253337e-1,.4198929950e-1,.5420795256e-1,.4547435761e-2,.7876389783e-2,.8148799146e-3,2.438964522,4.224410470,5.453690464,6.452893580,7.316893568,1.397695098,2.420878923,3.125341251,3.697953638,.4467629626,.7738161499,.9989923539,.1144983212,.1983169097,.2601078148e-1,2.715128242,4.702740065,6.071211317,7.183554108,2.125840578,3.682063888,4.753524039,.8614352786,1.492049670,.2673685503,2.611792036,4.523756504,5.840144534,2.592424872,4.490211595,1.272224784,2.331206226,4.037767626,2.802295901,1.987747725},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.1199176989,.2077035473,.2681441264,.3172724093,.2424253337e-1,.4198929950e-1,.5420795256e-1,.4547435761e-2,.7876389783e-2,.8148799146e-3,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.2937371735,.5087677086,.6568162875,.7771555119,.9389092801e-1,.1626238576,.2099464974,.2406276826e-1,.4167793720e-1,.5466380647e-2,.1199176989,.2077035473,.2681441264,.3172724093,.9389092801e-1,.1626238576,.2099464974,.3804657725e-1,.6589860485e-1,.1180873184e-1,.2424253337e-1,.4198929950e-1,.5420795256e-1,.2406276826e-1,.4167793720e-1,.1180873184e-1,.4547435761e-2,.7876389783e-2,.5466380647e-2,.8148799146e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.438964522,4.224410470,5.453690464,6.452893580,7.316893568,2.715128242,4.702740065,6.071211317,7.183554108,2.611792036,4.523756504,5.840144534,2.331206226,4.037767626,1.987747725,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,1.397695098,2.420878923,3.125341251,3.697953638,2.125840578,3.682063888,4.753524039,2.592424872,4.490211595,2.802295901,.1199176989,.2077035473,.2681441264,.3172724093,.4467629626,.7738161499,.9989923539,.8614352786,1.492049670,1.272224784,.2424253337e-1,.4198929950e-1,.5420795256e-1,.1144983212,.1983169097,.2673685503,.4547435761e-2,.7876389783e-2,.2601078148e-1,.8148799146e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.3216132869e-2,.5570505532e-2,.7191491718e-2,.3076861022e-3,.5329279618e-3,.2812031928e-4,2.941263458,5.094417748,6.576865031,7.781851649,8.823790377,.8596589978,1.488973062,1.922255957,2.274443921,.1401445651,.2427375071,.3133727741,.1831823152e-1,.3172810771e-1,.2122380039e-2,3.948637231,6.839240305,8.829421268,10.44711213,1.576786555,2.731074426,3.525801923,.3258743636,.5644309546,.5158497099e-1,4.580616431,7.933860388,10.24256972,2.318872338,4.016404707,.5803898901,4.930539837,8.539945507,3.022825632,5.069947794},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.3216132869e-2,.5570505532e-2,.7191491718e-2,.3076861022e-3,.5329279618e-3,.2812031928e-4,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.7640636963e-1,.1323397142,.1708498365,.2021522527,.1245602905e-1,.2157447516e-1,.2785252767e-1,.1628121817e-2,.2819989708e-2,.1886368363e-3,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.1245602905e-1,.2157447516e-1,.2785252767e-1,.2574286624e-2,.4458795226e-2,.4075021408e-3,.3216132869e-2,.5570505532e-2,.7191491718e-2,.1628121817e-2,.2819989708e-2,.4075021408e-3,.3076861022e-3,.5329279618e-3,.1886368363e-3,.2812031928e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.941263458,5.094417748,6.576865031,7.781851649,8.823790377,3.948637231,6.839240305,8.829421268,10.44711213,4.580616431,7.933860388,10.24256972,4.930539837,8.539945507,5.069947794,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.8596589978,1.488973062,1.922255957,2.274443921,1.576786555,2.731074426,3.525801923,2.318872338,4.016404707,3.022825632,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.1401445651,.2427375071,.3133727741,.3258743636,.5644309546,.5803898901,.3216132869e-2,.5570505532e-2,.7191491718e-2,.1831823152e-1,.3172810771e-1,.5158497099e-1,.3076861022e-3,.5329279618e-3,.2122380039e-2,.2812031928e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.3420864716e-4,.5925111493e-4,.7649286046e-4,.7197424251e-6,.1246630449e-5,.1446627697e-7,3.349118367,5.800843169,7.488856330,8.860934307,10.04735510,.2152732603,.3728642244,.4813656438,.5695595107,.7718047390e-2,.1336805021e-1,.1725807861e-1,.2218615436e-3,.3842754658e-3,.5653129632e-5,5.119651213,8.867496018,11.44788813,13.54532391,.4496072984,.7787426841,1.005352482,.2043515547e-1,.3539472753e-1,.7114071839e-3,6.762598981,11.71316502,15.12163102,.7528940303,1.304050714,.4144232937e-1,8.288591754,14.35626204,1.117549271,9.704794288},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.3420864716e-4,.5925111493e-4,.7649286046e-4,.7197424251e-6,.1246630449e-5,.1446627697e-7,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.3695423161e-2,.6400660671e-2,.8263217397e-2,.9777170675e-2,.1324895208e-3,.2294785814e-3,.2962555747e-3,.3808518930e-5,.6596548290e-5,.9704273609e-7,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.1324895208e-3,.2294785814e-3,.2962555747e-3,.6021797165e-5,.1043005864e-4,.2096362696e-6,.3420864716e-4,.5925111493e-4,.7649286046e-4,.3808518930e-5,.6596548290e-5,.2096362696e-6,.7197424251e-6,.1246630449e-5,.9704273609e-7,.1446627697e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.349118367,5.800843169,7.488856330,8.860934307,10.04735510,5.119651213,8.867496018,11.44788813,13.54532391,6.762598981,11.71316502,15.12163102,8.288591754,14.35626204,9.704794288,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.2152732603,.3728642244,.4813656438,.5695595107,.4496072984,.7787426841,1.005352482,.7528940303,1.304050714,1.117549271,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.7718047390e-2,.1336805021e-1,.1725807861e-1,.2043515547e-1,.3539472753e-1,.4144232937e-1,.3420864716e-4,.5925111493e-4,.7649286046e-4,.2218615436e-3,.3842754658e-3,.7114071839e-3,.7197424251e-6,.1246630449e-5,.5653129632e-5,.1446627697e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4816958184,.8343216312,1.077104595,1.274447343,.1951696965,.3380438305,.4364127086,.7337460098e-1,.1270885369,.2635224308e-1,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.4120977289,.7137742043,.9214785357,1.090308107,.2640035315,.4572675299,.5903298427,.1356052452,.2348751745,.6174127088e-1,.5875925975e-1,.1017740233,.1313896991,.1554623886,.9220648278e-1,.1597063129,.2061799634,.7488558248e-1,.1297056336,.4658334547e-1,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1654169192e-1,.2865105086e-1,.1626980677e-1,.1091823989e-2,.1891094623e-2,.2630454079e-2,.1369544625e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,1.970938383,3.413765418,4.407152207,5.214612814,1.615335150,2.797842550,3.611999201,1.228417782,2.127682012,.8924173718,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,2.386704142,4.133892837,5.336832705,6.314625614,3.092845228,5.356965074,6.915812173,3.213475734,5.565903242,2.959539815,.4816958184,.8343216312,1.077104595,1.274447343,1.529002171,2.648309444,3.418952791,2.511857186,4.350664268,3.160657901,.1951696965,.3380438305,.4364127086,.7853709873,1.360302453,1.562526553,.7337460098e-1,.1270885369,.3575805848,.2635224308e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.5875925975e-1,.1017740233,.1313896991,.1554623886,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1091823989e-2,.1891094623e-2,.1369544625e-3,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,.8335857979,1.443812955,1.863954510,2.205460718,.1865140454,.3230518028,.4170580840,.3346031412e-1,.5795496410e-1,.5320847480e-2,1.970938383,3.413765418,4.407152207,5.214612814,1.080214264,1.870985987,2.415432523,.3064067886,.5307121257,.6657061456e-1,1.615335150,2.797842550,3.611999201,1.122345952,1.943960214,.3855501990,1.228417782,2.127682012,1.033655707,.8924173718},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.5875925975e-1,.1017740233,.1313896991,.1554623886,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1091823989e-2,.1891094623e-2,.1369544625e-3,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4120977289,.7137742043,.9214785357,1.090308107,.9220648278e-1,.1597063129,.2061799634,.1654169192e-1,.2865105086e-1,.2630454079e-2,.4816958184,.8343216312,1.077104595,1.274447343,.2640035315,.4572675299,.5903298427,.7488558248e-1,.1297056336,.1626980677e-1,.1951696965,.3380438305,.4364127086,.1356052452,.2348751745,.4658334547e-1,.7337460098e-1,.1270885369,.6174127088e-1,.2635224308e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4816958184,.8343216312,1.077104595,1.274447343,.1951696965,.3380438305,.4364127086,.7337460098e-1,.1270885369,.2635224308e-1,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,2.386704142,4.133892837,5.336832705,6.314625614,1.529002171,2.648309444,3.418952791,.7853709873,1.360302453,.3575805848,1.970938383,3.413765418,4.407152207,5.214612814,3.092845228,5.356965074,6.915812173,2.511857186,4.350664268,1.562526553,1.615335150,2.797842550,3.611999201,3.213475734,5.565903242,3.160657901,1.228417782,2.127682012,2.959539815,.8924173718},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,1.970938383,3.413765418,4.407152207,5.214612814,1.615335150,2.797842550,3.611999201,1.228417782,2.127682012,.8924173718,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.8335857979,1.443812955,1.863954510,2.205460718,1.080214264,1.870985987,2.415432523,1.122345952,1.943960214,1.033655707,.5875925975e-1,.1017740233,.1313896991,.1554623886,.1865140454,.3230518028,.4170580840,.3064067886,.5307121257,.3855501990,.8315101345e-2,.1440217800e-1,.1859313185e-1,.3346031412e-1,.5795496410e-1,.6657061456e-1,.1091823989e-2,.1891094623e-2,.5320847480e-2,.1369544625e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.6247151922,1.082038453,1.396905637,1.652841039,.2882545515,.4992715287,.6445567719,.1234139165,.2137591738,.5047665480e-1,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.9099854801e-1,.1576141086,.2034789393,.2407595278,.6638937401e-1,.1149897688,.1484511532,.3883469697e-1,.6726366826e-1,.2013601810e-1,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.3947983634e-2,.6838108240e-2,.8827959777e-2,.3651461244e-2,.6324516397e-2,.2586745805e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1373325428e-3,.2378669418e-3,.1538262752e-3,.1543374128e-5,.2673202406e-5,.4234517225e-5,.3753828413e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.258876421,3.912488728,5.051001230,5.976425253,1.981945831,3.432830877,4.431765606,1.613559350,2.794766775,1.254921551,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,2.909798360,5.039918600,6.506506935,7.698602827,4.036753290,6.991861793,9.026454761,4.490128369,7.777130471,4.427082670,.6247151922,1.082038453,1.396905637,1.652841039,2.122887626,3.676949226,4.746921039,3.733563649,6.466721933,5.029395863,.2882545515,.4992715287,.6445567719,1.241790556,2.150844335,2.644908288,.1234139165,.2137591738,.6438756851,.5047665480e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1543374128e-5,.2673202406e-5,.3753828413e-7,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,.1730372742,.2997093506,.3869231079,.4578135952,.7507244251e-2,.1300292847e-1,.1678670846e-1,.2611431652e-3,.4523132303e-3,.8052099006e-5,2.258876421,3.912488728,5.051001230,5.976425253,.2400540172,.4157857543,.5367771006,.1320313609e-1,.2286850254e-1,.5562127353e-3,1.981945831,3.432830877,4.431765606,.2670149191,.4624834065,.1778563440e-1,1.613559350,2.794766775,.2632657742,1.254921551},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1543374128e-5,.2673202406e-5,.3753828413e-7,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.9099854801e-1,.1576141086,.2034789393,.2407595278,.3947983634e-2,.6838108240e-2,.8827959777e-2,.1373325428e-3,.2378669418e-3,.4234517225e-5,.6247151922,1.082038453,1.396905637,1.652841039,.6638937401e-1,.1149897688,.1484511532,.3651461244e-2,.6324516397e-2,.1538262752e-3,.2882545515,.4992715287,.6445567719,.3883469697e-1,.6726366826e-1,.2586745805e-2,.1234139165,.2137591738,.2013601810e-1,.5047665480e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.6247151922,1.082038453,1.396905637,1.652841039,.2882545515,.4992715287,.6445567719,.1234139165,.2137591738,.5047665480e-1,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.909798360,5.039918600,6.506506935,7.698602827,2.122887626,3.676949226,4.746921039,1.241790556,2.150844335,.6438756851,2.258876421,3.912488728,5.051001230,5.976425253,4.036753290,6.991861793,9.026454761,3.733563649,6.466721933,2.644908288,1.981945831,3.432830877,4.431765606,4.490128369,7.777130471,5.029395863,1.613559350,2.794766775,4.427082670,1.254921551},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.258876421,3.912488728,5.051001230,5.976425253,1.981945831,3.432830877,4.431765606,1.613559350,2.794766775,1.254921551,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.1730372742,.2997093506,.3869231079,.4578135952,.2400540172,.4157857543,.5367771006,.2670149191,.4624834065,.2632657742,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.7507244251e-2,.1300292847e-1,.1678670846e-1,.1320313609e-1,.2286850254e-1,.1778563440e-1,.6061868131e-4,.1049946359e-3,.1355474921e-3,.2611431652e-3,.4523132303e-3,.5562127353e-3,.1543374128e-5,.2673202406e-5,.8052099006e-5,.3753828413e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,.2296154773,.3977056730,.5134358159,.6075054502,.6423249772e-1,.1112539495,.1436282312,.1667256107e-1,.2887772287e-1,.4134169639e-2,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,-.1192467538e-1,-.2065414363e-1,-.2666438477e-1,-.3154972553e-1,-.5274365494e-2,-.9135469009e-2,-.1179383978e-1,-.1870471581e-2,-.3239751814e-2,-.5879824074e-3,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,.1118252823e-3,.1936870705e-3,.2500489327e-3,.6270337257e-4,.1086054271e-3,.2693007136e-4,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,-.8407960796e-6,-.1456301529e-5,-.5709621067e-6,.3368848215e-8,.5835016271e-8,.5603684524e-8,-.1771076019e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,3.501856062,6.065392619,7.830388205,9.265040271,3.825608374,6.626148073,8.554320378,3.877899792,6.716719466,3.755183165,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,2.196470374,3.804398286,4.911457068,5.811314373,3.794002148,6.571404483,8.483646708,5.254447192,9.100969504,6.450432215,.2296154773,.3977056730,.5134358159,.6075054502,.9715138718,1.682711386,2.172371058,2.127396643,3.684759074,3.568154409,.6423249772e-1,.1112539495,.1436282312,.3445322647,.5967473875,.9136820088,.1667256107e-1,.2887772287e-1,.1083036558,.4134169639e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,.3368848215e-8,.5835016271e-8,-.1771076019e-10,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,-.4656880908e-1,-.8065954338e-1,-.1041310228,-.1232094877,.4367054075e-3,.7563959535e-3,.9765029769e-3,-.3283516813e-5,-.5687217949e-5,.2188377514e-7,3.501856062,6.065392619,7.830388205,9.265040271,-.8043912806e-1,-.1393246566,-.1798673583,.9562865187e-3,.1656336837e-2,-.8707719261e-5,3.825608374,6.626148073,8.554320378,-.1114029814,-.1929556239,.1603921848e-2,3.877899792,6.716719466,-.1367598443,3.755183165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,.3368848215e-8,.5835016271e-8,-.1771076019e-10,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,-.1192467538e-1,-.2065414363e-1,-.2666438477e-1,-.3154972553e-1,.1118252823e-3,.1936870705e-3,.2500489327e-3,-.8407960796e-6,-.1456301529e-5,.5603684524e-8,.2296154773,.3977056730,.5134358159,.6075054502,-.5274365494e-2,-.9135469009e-2,-.1179383978e-1,.6270337257e-4,.1086054271e-3,-.5709621067e-6,.6423249772e-1,.1112539495,.1436282312,-.1870471581e-2,-.3239751814e-2,.2693007136e-4,.1667256107e-1,.2887772287e-1,-.5879824074e-3,.4134169639e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,.2296154773,.3977056730,.5134358159,.6075054502,.6423249772e-1,.1112539495,.1436282312,.1667256107e-1,.2887772287e-1,.4134169639e-2,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,2.196470374,3.804398286,4.911457068,5.811314373,.9715138718,1.682711386,2.172371058,.3445322647,.5967473875,.1083036558,3.501856062,6.065392619,7.830388205,9.265040271,3.794002148,6.571404483,8.483646708,2.127396643,3.684759074,.9136820088,3.825608374,6.626148073,8.554320378,5.254447192,9.100969504,3.568154409,3.877899792,6.716719466,6.450432215,3.755183165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,3.501856062,6.065392619,7.830388205,9.265040271,3.825608374,6.626148073,8.554320378,3.877899792,6.716719466,3.755183165,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,-.4656880908e-1,-.8065954338e-1,-.1041310228,-.1232094877,-.8043912806e-1,-.1393246566,-.1798673583,-.1114029814,-.1929556239,-.1367598443,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,.4367054075e-3,.7563959535e-3,.9765029769e-3,.9562865187e-3,.1656336837e-2,.1603921848e-2,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,-.3283516813e-5,-.5687217949e-5,-.8707719261e-5,.3368848215e-8,.5835016271e-8,.2188377514e-7,-.1771076019e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1963948976,.3401659410,.4391523414,.5196120579,.5080990387e-1,.8800533503e-1,.1136143990,.1219720719e-1,.2112618256e-1,.2797118711e-2,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.1065535599,.1845561795,.2382610033,.2819142209,.4358691597e-1,.7549475298e-1,.9746330700e-1,.1429557258e-1,.2476065805e-1,.4156035932e-2,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.9654244188e-2,.1672164144e-1,.2158754627e-1,.5006493970e-2,.8671501924e-2,.1988588036e-2,.5521216640e-3,.9563027739e-3,.1234581572e-2,.7013355149e-3,.1214748745e-2,.4404604930e-3,.2935681623e-4,.5084749726e-4,.4516122448e-4,.1491150883e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,3.236488548,5.605762603,7.237008402,8.562943823,3.399101902,5.887417193,7.600622914,3.312440982,5.737316079,3.083689087,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,1.952889915,3.382504555,4.366794603,5.166861053,3.242932261,5.616923440,7.251416980,4.317727482,7.478523374,5.095712344,.1963948976,.3401659410,.4391523414,.5196120579,.7988512884,1.383651019,1.786285784,1.681718474,2.912821842,2.711667007,.5080990387e-1,.8800533503e-1,.1136143990,.2620060705,.4538078261,.6679814775,.1219720719e-1,.2112618256e-1,.7617090096e-1,.2797118711e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2935681623e-4,.5084749726e-4,.1491150883e-5,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,.4325535702,.7492047606,.9672191871,1.144429176,.3919134934e-1,.6788140826e-1,.8763452124e-1,.2847067532e-2,.4931265619e-2,.1833317338e-3,3.236488548,5.605762603,7.237008402,8.562943823,.7182903227,1.244115334,1.606145989,.8250448766e-1,.1429019645,.7258566080e-2,3.399101902,5.887417193,7.600622914,.9563511098,1.656448712,.1330333825,3.312440982,5.737316079,1.128670157,3.083689087},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2935681623e-4,.5084749726e-4,.1491150883e-5,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1065535599,.1845561795,.2382610033,.2819142209,.9654244188e-2,.1672164144e-1,.2158754627e-1,.7013355149e-3,.1214748745e-2,.4516122448e-4,.1963948976,.3401659410,.4391523414,.5196120579,.4358691597e-1,.7549475298e-1,.9746330700e-1,.5006493970e-2,.8671501924e-2,.4404604930e-3,.5080990387e-1,.8800533503e-1,.1136143990,.1429557258e-1,.2476065805e-1,.1988588036e-2,.1219720719e-1,.2112618256e-1,.4156035932e-2,.2797118711e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1963948976,.3401659410,.4391523414,.5196120579,.5080990387e-1,.8800533503e-1,.1136143990,.1219720719e-1,.2112618256e-1,.2797118711e-2,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,1.952889915,3.382504555,4.366794603,5.166861053,.7988512884,1.383651019,1.786285784,.2620060705,.4538078261,.7617090096e-1,3.236488548,5.605762603,7.237008402,8.562943823,3.242932261,5.616923440,7.251416980,1.681718474,2.912821842,.6679814775,3.399101902,5.887417193,7.600622914,4.317727482,7.478523374,2.711667007,3.312440982,5.737316079,5.095712344,3.083689087},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,3.236488548,5.605762603,7.237008402,8.562943823,3.399101902,5.887417193,7.600622914,3.312440982,5.737316079,3.083689087,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.4325535702,.7492047606,.9672191871,1.144429176,.7182903227,1.244115334,1.606145989,.9563511098,1.656448712,1.128670157,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.3919134934e-1,.6788140826e-1,.8763452124e-1,.8250448766e-1,.1429019645,.1330333825,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2847067532e-2,.4931265619e-2,.7258566080e-2,.2935681623e-4,.5084749726e-4,.1833317338e-3,.1491150883e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.4641847056e-2,.8039914942e-2,.1037948556e-1,.5018612597e-3,.8692492001e-3,.5183403142e-4,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1638187454e-1,.2837423903e-1,.3663098508e-1,.4334236605e-1,.3018092809e-2,.5227490086e-2,.6748660682e-2,.4458192555e-3,.7721816018e-3,.5837372111e-4,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.5066741239e-3,.8775853253e-3,.1132957783e-2,.1183382006e-3,.2049677758e-3,.2116980724e-4,.2196237131e-4,.3803994296e-4,.4910935520e-4,.1256467012e-4,.2176264704e-4,.3553964112e-5,.3986283756e-6,.6904445999e-6,.2761886212e-6,.6911871564e-8},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,4.440487963,7.691150763,9.929232940,11.74842685,5.462596453,9.461494597,12.21473700,6.235358013,10.79995688,6.799267692,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,1.030236404,1.784421795,2.303678632,2.725749316,2.003896641,3.470850795,4.480849108,3.125150808,5.412919982,4.320151396,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.1898042300,.3287505697,.4244151605,.4680276955,.8106477480,.8839606376,.4641847056e-2,.8039914942e-2,.1037948556e-1,.2803703724e-1,.4856157301e-1,.8372660778e-1,.5018612597e-3,.8692492001e-3,.3671053173e-2,.5183403142e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.2196237131e-4,.3803994296e-4,.4910935520e-4,.3986283756e-6,.6904445999e-6,.6911871564e-8,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,.1729549620,.2995667817,.3867390522,.4575958176,.5349314793e-2,.9265285003e-2,.1196143151e-1,.1326540524e-3,.2297635587e-3,.2915917367e-5,4.440487963,7.691150763,9.929232940,11.74842685,.3364119790,.5826826397,.7522400533,.1319057787e-1,.2284675106e-1,.3961429206e-3,5.462596453,9.461494597,12.21473700,.5246469034,.9087150930,.2491295222e-1,6.235358013,10.79995688,.7252622967,6.799267692},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.2196237131e-4,.3803994296e-4,.4910935520e-4,.3986283756e-6,.6904445999e-6,.6911871564e-8,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.1638187454e-1,.2837423903e-1,.3663098508e-1,.4334236605e-1,.5066741239e-3,.8775853253e-3,.1132957783e-2,.1256467012e-4,.2176264704e-4,.2761886212e-6,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.3018092809e-2,.5227490086e-2,.6748660682e-2,.1183382006e-3,.2049677758e-3,.3553964112e-5,.4641847056e-2,.8039914942e-2,.1037948556e-1,.4458192555e-3,.7721816018e-3,.2116980724e-4,.5018612597e-3,.8692492001e-3,.5837372111e-4,.5183403142e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.4641847056e-2,.8039914942e-2,.1037948556e-1,.5018612597e-3,.8692492001e-3,.5183403142e-4,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,1.030236404,1.784421795,2.303678632,2.725749316,.1898042300,.3287505697,.4244151605,.2803703724e-1,.4856157301e-1,.3671053173e-2,4.440487963,7.691150763,9.929232940,11.74842685,2.003896641,3.470850795,4.480849108,.4680276955,.8106477480,.8372660778e-1,5.462596453,9.461494597,12.21473700,3.125150808,5.412919982,.8839606376,6.235358013,10.79995688,4.320151396,6.799267692},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,4.440487963,7.691150763,9.929232940,11.74842685,5.462596453,9.461494597,12.21473700,6.235358013,10.79995688,6.799267692,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1729549620,.2995667817,.3867390522,.4575958176,.3364119790,.5826826397,.7522400533,.5246469034,.9087150930,.7252622967,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.5349314793e-2,.9265285003e-2,.1196143151e-1,.1319057787e-1,.2284675106e-1,.2491295222e-1,.2196237131e-4,.3803994296e-4,.4910935520e-4,.1326540524e-3,.2297635587e-3,.3961429206e-3,.3986283756e-6,.6904445999e-6,.2915917367e-5,.6911871564e-8}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 52; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_bottom_face(
		const T i_input[56],
		T o_output[52]
	) {
		static const T matrix[52][56] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.154700539,2.,2.581988897,3.055050463,3.464101616,.6085806194,1.054092553,1.360827635,1.610152972,.2771598064,.4800548665,.6197481677,.1171213948,.2028602065,.4728035100e-1,1.154700539,2.,2.581988897,3.055050463,3.464101616,1.490711985,2.581988897,3.333333333,3.944053189,1.073435315,1.859244503,2.400274333,.6197481677,1.073435315,.3171662365,.6085806194,1.054092553,1.360827635,1.610152972,1.073435315,1.859244503,2.400274333,.9799078930,1.697250258,.6851573789,.2771598064,.4800548665,.6197481677,.6197481677,1.073435315,.6851573789,.1171213948,.2028602065,.3171662365,.4728035100e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,1.354996613,2.346922978,3.029864538,3.584984067,.9207895610,1.594854303,2.058948051,.5805991817,1.005627282,.3497284165,.1814813775e-1,.3143349662e-1,.4058046965e-1,.4801545922e-1,.5444441322e-1,.3495960260e-1,.6055180793e-1,.7817204790e-1,.9249441444e-1,.3756286152e-1,.6506078461e-1,.8399311177e-1,.3235996830e-1,.5604910924e-1,.2471097213e-1,.1503292582e-3,.2603779131e-3,.3361464403e-3,.3977338322e-3,.3956501378e-3,.6852861406e-3,.8847006032e-3,.5389279955e-3,.9334506698e-3,.5622712446e-3,.1076015046e-5,.1863712730e-5,.2406042788e-5,.3590156597e-5,.6218333635e-5,.5922410770e-5,.7146395199e-8,.1237791958e-7,.2887671511e-7,.4534132255e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,1.354996613,2.346922978,3.029864538,3.584984067,.9207895610,1.594854303,2.058948051,.5805991817,1.005627282,.3497284165,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,3.319050305,5.748763763,7.421622105,8.781381698,3.566202636,6.176844154,7.974271513,3.072242092,5.321279398,2.346049539,1.354996613,2.346922978,3.029864538,3.584984067,3.566202636,6.176844154,7.974271513,4.857641268,8.413681482,5.068046242,.9207895610,1.594854303,2.058948051,3.072242092,5.321279398,5.068046242,.5805991817,1.005627282,2.346049539,.3497284165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1814813775e-1,.3143349662e-1,.4058046965e-1,.4801545922e-1,.5444441322e-1,.1503292582e-3,.2603779131e-3,.3361464403e-3,.3977338322e-3,.1076015046e-5,.1863712730e-5,.2406042788e-5,.7146395199e-8,.1237791958e-7,.4534132255e-10,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,.3495960260e-1,.6055180793e-1,.7817204790e-1,.9249441444e-1,.3956501378e-3,.6852861406e-3,.8847006032e-3,.3590156597e-5,.6218333635e-5,.2887671511e-7,1.354996613,2.346922978,3.029864538,3.584984067,.3756286152e-1,.6506078461e-1,.8399311177e-1,.5389279955e-3,.9334506698e-3,.5922410770e-5,.9207895610,1.594854303,2.058948051,.3235996830e-1,.5604910924e-1,.5622712446e-3,.5805991817,1.005627282,.2471097213e-1,.3497284165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9363699245,1.621840284,2.093786803,2.477401956,.5289617350,.9161886001,1.182794397,.2772649114,.4802369137,.1388366216,.5995013155,1.038366737,1.340525694,1.586131391,1.798503946,.9600175758,1.662799218,2.146664559,2.539967760,.8574839288,1.485205731,1.917392354,.6140871797,1.063630196,.3898220464,.1640437399,.2841320922,.3668129538,.4340189401,.3589071350,.6216453928,.8025407512,.4064022821,.7039094010,.3524730288,.3878756937e-1,.6718204084e-1,.8673164178e-1,.1075826497,.1863386154,.1475305609,.8509799423e-2,.1473940496e-1,.2858477025e-1,.1783546975e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9363699245,1.621840284,2.093786803,2.477401956,.5289617350,.9161886001,1.182794397,.2772649114,.4802369137,.1388366216,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,2.293628526,3.972681141,5.128709298,6.068370681,2.048659991,3.548383191,4.580943002,1.467148005,2.541174888,.9313443707,.9363699245,1.621840284,2.093786803,2.477401956,2.048659991,3.548383191,4.580943002,2.319764682,4.017950291,2.011933789,.5289617350,.9161886001,1.182794397,1.467148005,2.541174888,2.011933789,.2772649114,.4802369137,.9313443707,.1388366216},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5995013155,1.038366737,1.340525694,1.586131391,1.798503946,.1640437399,.2841320922,.3668129538,.4340189401,.3878756937e-1,.6718204084e-1,.8673164178e-1,.8509799423e-2,.1473940496e-1,.1783546975e-2,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9600175758,1.662799218,2.146664559,2.539967760,.3589071350,.6216453928,.8025407512,.1075826497,.1863386154,.2858477025e-1,.9363699245,1.621840284,2.093786803,2.477401956,.8574839288,1.485205731,1.917392354,.4064022821,.7039094010,.1475305609,.5289617350,.9161886001,1.182794397,.6140871797,1.063630196,.3524730288,.2772649114,.4802369137,.3898220464,.1388366216},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,1.212281421,2.099733015,2.710743667,3.207395161,.7792166912,1.349642899,1.742381491,.4647366240,.8049474449,.2647853914,.2046688294,.3544968112,.4576534152,.5415028235,.6140064883,.3729226366,.6459209540,.8338803660,.9866605549,.3790037856,.6564538126,.8474782279,.3088343411,.5349167702,.2230693744,.1911977567e-1,.3311642289e-1,.4275311811e-1,.5058617157e-1,.4759739883e-1,.8244111306e-1,.1064310193,.6132467595e-1,.1062174545,.6051778323e-1,.1543395612e-2,.2673239616e-2,.3451137504e-2,.4870851947e-2,.8436563051e-2,.7600159080e-2,.1156020631e-3,.2002286468e-3,.4418338282e-3,.8271654356e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,1.212281421,2.099733015,2.710743667,3.207395161,.7792166912,1.349642899,1.742381491,.4647366240,.8049474449,.2647853914,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,2.969470906,5.143274482,6.639938805,7.856481545,3.017893269,5.227144472,6.748214496,2.459155065,4.259381518,1.776234404,1.212281421,2.099733015,2.710743667,3.207395161,3.017893269,5.227144472,6.748214496,3.888265562,6.734673506,3.837104864,.7792166912,1.349642899,1.742381491,2.459155065,4.259381518,3.837104864,.4647366240,.8049474449,1.776234404,.2647853914},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2046688294,.3544968112,.4576534152,.5415028235,.6140064883,.1911977567e-1,.3311642289e-1,.4275311811e-1,.5058617157e-1,.1543395612e-2,.2673239616e-2,.3451137504e-2,.1156020631e-3,.2002286468e-3,.8271654356e-5,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,.3729226366,.6459209540,.8338803660,.9866605549,.4759739883e-1,.8244111306e-1,.1064310193,.4870851947e-2,.8436563051e-2,.4418338282e-3,1.212281421,2.099733015,2.710743667,3.207395161,.3790037856,.6564538126,.8474782279,.6132467595e-1,.1062174545,.7600159080e-2,.7792166912,1.349642899,1.742381491,.3088343411,.5349167702,.6051778323e-1,.4647366240,.8049474449,.2230693744,.2647853914},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.3169457192,.5489660888,.7087121734,.8385595524,.1041670603,.1804226409,.2329246278,.3176653023e-1,.5502124434e-1,.9254386281e-2,1.797496348,3.113355001,4.019324022,4.755728318,5.392489045,1.674656646,2.900590397,3.744646101,4.430725019,.8702445254,1.507307732,1.945925915,.3625885274,.6280217519,.1339118908,1.474739846,2.554324341,3.297618546,3.901794882,1.877182782,3.251375951,4.197508304,1.236657185,2.141953076,.6240046052,1.045505193,1.810868114,2.337820683,1.687114596,2.922168200,1.346024785,.6877503643,1.191218574,1.344048463,.4321892234},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.3169457192,.5489660888,.7087121734,.8385595524,.1041670603,.1804226409,.2329246278,.3176653023e-1,.5502124434e-1,.9254386281e-2,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.7763552880,1.344686804,1.735983199,2.054043021,.4034372898,.6987738835,.9021132044,.1680926781,.2911450589,.6208031047e-1,.3169457192,.5489660888,.7087121734,.8385595524,.4034372898,.6987738835,.9021132044,.2657778603,.4603407575,.1341087982,.1041670603,.1804226409,.2329246278,.1680926781,.2911450589,.1341087982,.3176653023e-1,.5502124434e-1,.6208031047e-1,.9254386281e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.797496348,3.113355001,4.019324022,4.755728318,5.392489045,1.474739846,2.554324341,3.297618546,3.901794882,1.045505193,1.810868114,2.337820683,.6877503643,1.191218574,.4321892234,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,1.674656646,2.900590397,3.744646101,4.430725019,1.877182782,3.251375951,4.197508304,1.687114596,2.922168200,1.344048463,.3169457192,.5489660888,.7087121734,.8385595524,.8702445254,1.507307732,1.945925915,1.236657185,2.141953076,1.346024785,.1041670603,.1804226409,.2329246278,.3625885274,.6280217519,.6240046052,.3176653023e-1,.5502124434e-1,.1339118908,.9254386281e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.1199176989,.2077035473,.2681441264,.3172724093,.2424253337e-1,.4198929950e-1,.5420795256e-1,.4547435761e-2,.7876389783e-2,.8148799146e-3,2.438964522,4.224410470,5.453690464,6.452893580,7.316893568,1.397695098,2.420878923,3.125341251,3.697953638,.4467629626,.7738161499,.9989923539,.1144983212,.1983169097,.2601078148e-1,2.715128242,4.702740065,6.071211317,7.183554108,2.125840578,3.682063888,4.753524039,.8614352786,1.492049670,.2673685503,2.611792036,4.523756504,5.840144534,2.592424872,4.490211595,1.272224784,2.331206226,4.037767626,2.802295901,1.987747725},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.1199176989,.2077035473,.2681441264,.3172724093,.2424253337e-1,.4198929950e-1,.5420795256e-1,.4547435761e-2,.7876389783e-2,.8148799146e-3,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.2937371735,.5087677086,.6568162875,.7771555119,.9389092801e-1,.1626238576,.2099464974,.2406276826e-1,.4167793720e-1,.5466380647e-2,.1199176989,.2077035473,.2681441264,.3172724093,.9389092801e-1,.1626238576,.2099464974,.3804657725e-1,.6589860485e-1,.1180873184e-1,.2424253337e-1,.4198929950e-1,.5420795256e-1,.2406276826e-1,.4167793720e-1,.1180873184e-1,.4547435761e-2,.7876389783e-2,.5466380647e-2,.8148799146e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.438964522,4.224410470,5.453690464,6.452893580,7.316893568,2.715128242,4.702740065,6.071211317,7.183554108,2.611792036,4.523756504,5.840144534,2.331206226,4.037767626,1.987747725,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,1.397695098,2.420878923,3.125341251,3.697953638,2.125840578,3.682063888,4.753524039,2.592424872,4.490211595,2.802295901,.1199176989,.2077035473,.2681441264,.3172724093,.4467629626,.7738161499,.9989923539,.8614352786,1.492049670,1.272224784,.2424253337e-1,.4198929950e-1,.5420795256e-1,.1144983212,.1983169097,.2673685503,.4547435761e-2,.7876389783e-2,.2601078148e-1,.8148799146e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.3216132869e-2,.5570505532e-2,.7191491718e-2,.3076861022e-3,.5329279618e-3,.2812031928e-4,2.941263458,5.094417748,6.576865031,7.781851649,8.823790377,.8596589978,1.488973062,1.922255957,2.274443921,.1401445651,.2427375071,.3133727741,.1831823152e-1,.3172810771e-1,.2122380039e-2,3.948637231,6.839240305,8.829421268,10.44711213,1.576786555,2.731074426,3.525801923,.3258743636,.5644309546,.5158497099e-1,4.580616431,7.933860388,10.24256972,2.318872338,4.016404707,.5803898901,4.930539837,8.539945507,3.022825632,5.069947794},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.3216132869e-2,.5570505532e-2,.7191491718e-2,.3076861022e-3,.5329279618e-3,.2812031928e-4,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.7640636963e-1,.1323397142,.1708498365,.2021522527,.1245602905e-1,.2157447516e-1,.2785252767e-1,.1628121817e-2,.2819989708e-2,.1886368363e-3,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.1245602905e-1,.2157447516e-1,.2785252767e-1,.2574286624e-2,.4458795226e-2,.4075021408e-3,.3216132869e-2,.5570505532e-2,.7191491718e-2,.1628121817e-2,.2819989708e-2,.4075021408e-3,.3076861022e-3,.5329279618e-3,.1886368363e-3,.2812031928e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.941263458,5.094417748,6.576865031,7.781851649,8.823790377,3.948637231,6.839240305,8.829421268,10.44711213,4.580616431,7.933860388,10.24256972,4.930539837,8.539945507,5.069947794,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.8596589978,1.488973062,1.922255957,2.274443921,1.576786555,2.731074426,3.525801923,2.318872338,4.016404707,3.022825632,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.1401445651,.2427375071,.3133727741,.3258743636,.5644309546,.5803898901,.3216132869e-2,.5570505532e-2,.7191491718e-2,.1831823152e-1,.3172810771e-1,.5158497099e-1,.3076861022e-3,.5329279618e-3,.2122380039e-2,.2812031928e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.3420864716e-4,.5925111493e-4,.7649286046e-4,.7197424251e-6,.1246630449e-5,.1446627697e-7,3.349118367,5.800843169,7.488856330,8.860934307,10.04735510,.2152732603,.3728642244,.4813656438,.5695595107,.7718047390e-2,.1336805021e-1,.1725807861e-1,.2218615436e-3,.3842754658e-3,.5653129632e-5,5.119651213,8.867496018,11.44788813,13.54532391,.4496072984,.7787426841,1.005352482,.2043515547e-1,.3539472753e-1,.7114071839e-3,6.762598981,11.71316502,15.12163102,.7528940303,1.304050714,.4144232937e-1,8.288591754,14.35626204,1.117549271,9.704794288},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.3420864716e-4,.5925111493e-4,.7649286046e-4,.7197424251e-6,.1246630449e-5,.1446627697e-7,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.3695423161e-2,.6400660671e-2,.8263217397e-2,.9777170675e-2,.1324895208e-3,.2294785814e-3,.2962555747e-3,.3808518930e-5,.6596548290e-5,.9704273609e-7,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.1324895208e-3,.2294785814e-3,.2962555747e-3,.6021797165e-5,.1043005864e-4,.2096362696e-6,.3420864716e-4,.5925111493e-4,.7649286046e-4,.3808518930e-5,.6596548290e-5,.2096362696e-6,.7197424251e-6,.1246630449e-5,.9704273609e-7,.1446627697e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.349118367,5.800843169,7.488856330,8.860934307,10.04735510,5.119651213,8.867496018,11.44788813,13.54532391,6.762598981,11.71316502,15.12163102,8.288591754,14.35626204,9.704794288,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.2152732603,.3728642244,.4813656438,.5695595107,.4496072984,.7787426841,1.005352482,.7528940303,1.304050714,1.117549271,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.7718047390e-2,.1336805021e-1,.1725807861e-1,.2043515547e-1,.3539472753e-1,.4144232937e-1,.3420864716e-4,.5925111493e-4,.7649286046e-4,.2218615436e-3,.3842754658e-3,.7114071839e-3,.7197424251e-6,.1246630449e-5,.5653129632e-5,.1446627697e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4816958184,.8343216312,1.077104595,1.274447343,.1951696965,.3380438305,.4364127086,.7337460098e-1,.1270885369,.2635224308e-1,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.4120977289,.7137742043,.9214785357,1.090308107,.2640035315,.4572675299,.5903298427,.1356052452,.2348751745,.6174127088e-1,.5875925975e-1,.1017740233,.1313896991,.1554623886,.9220648278e-1,.1597063129,.2061799634,.7488558248e-1,.1297056336,.4658334547e-1,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1654169192e-1,.2865105086e-1,.1626980677e-1,.1091823989e-2,.1891094623e-2,.2630454079e-2,.1369544625e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,1.970938383,3.413765418,4.407152207,5.214612814,1.615335150,2.797842550,3.611999201,1.228417782,2.127682012,.8924173718,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,2.386704142,4.133892837,5.336832705,6.314625614,3.092845228,5.356965074,6.915812173,3.213475734,5.565903242,2.959539815,.4816958184,.8343216312,1.077104595,1.274447343,1.529002171,2.648309444,3.418952791,2.511857186,4.350664268,3.160657901,.1951696965,.3380438305,.4364127086,.7853709873,1.360302453,1.562526553,.7337460098e-1,.1270885369,.3575805848,.2635224308e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.5875925975e-1,.1017740233,.1313896991,.1554623886,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1091823989e-2,.1891094623e-2,.1369544625e-3,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,.8335857979,1.443812955,1.863954510,2.205460718,.1865140454,.3230518028,.4170580840,.3346031412e-1,.5795496410e-1,.5320847480e-2,1.970938383,3.413765418,4.407152207,5.214612814,1.080214264,1.870985987,2.415432523,.3064067886,.5307121257,.6657061456e-1,1.615335150,2.797842550,3.611999201,1.122345952,1.943960214,.3855501990,1.228417782,2.127682012,1.033655707,.8924173718},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.5875925975e-1,.1017740233,.1313896991,.1554623886,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1091823989e-2,.1891094623e-2,.1369544625e-3,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4120977289,.7137742043,.9214785357,1.090308107,.9220648278e-1,.1597063129,.2061799634,.1654169192e-1,.2865105086e-1,.2630454079e-2,.4816958184,.8343216312,1.077104595,1.274447343,.2640035315,.4572675299,.5903298427,.7488558248e-1,.1297056336,.1626980677e-1,.1951696965,.3380438305,.4364127086,.1356052452,.2348751745,.4658334547e-1,.7337460098e-1,.1270885369,.6174127088e-1,.2635224308e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4816958184,.8343216312,1.077104595,1.274447343,.1951696965,.3380438305,.4364127086,.7337460098e-1,.1270885369,.2635224308e-1,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,2.386704142,4.133892837,5.336832705,6.314625614,1.529002171,2.648309444,3.418952791,.7853709873,1.360302453,.3575805848,1.970938383,3.413765418,4.407152207,5.214612814,3.092845228,5.356965074,6.915812173,2.511857186,4.350664268,1.562526553,1.615335150,2.797842550,3.611999201,3.213475734,5.565903242,3.160657901,1.228417782,2.127682012,2.959539815,.8924173718},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,1.970938383,3.413765418,4.407152207,5.214612814,1.615335150,2.797842550,3.611999201,1.228417782,2.127682012,.8924173718,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.8335857979,1.443812955,1.863954510,2.205460718,1.080214264,1.870985987,2.415432523,1.122345952,1.943960214,1.033655707,.5875925975e-1,.1017740233,.1313896991,.1554623886,.1865140454,.3230518028,.4170580840,.3064067886,.5307121257,.3855501990,.8315101345e-2,.1440217800e-1,.1859313185e-1,.3346031412e-1,.5795496410e-1,.6657061456e-1,.1091823989e-2,.1891094623e-2,.5320847480e-2,.1369544625e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.6247151922,1.082038453,1.396905637,1.652841039,.2882545515,.4992715287,.6445567719,.1234139165,.2137591738,.5047665480e-1,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.9099854801e-1,.1576141086,.2034789393,.2407595278,.6638937401e-1,.1149897688,.1484511532,.3883469697e-1,.6726366826e-1,.2013601810e-1,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.3947983634e-2,.6838108240e-2,.8827959777e-2,.3651461244e-2,.6324516397e-2,.2586745805e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1373325428e-3,.2378669418e-3,.1538262752e-3,.1543374128e-5,.2673202406e-5,.4234517225e-5,.3753828413e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.258876421,3.912488728,5.051001230,5.976425253,1.981945831,3.432830877,4.431765606,1.613559350,2.794766775,1.254921551,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,2.909798360,5.039918600,6.506506935,7.698602827,4.036753290,6.991861793,9.026454761,4.490128369,7.777130471,4.427082670,.6247151922,1.082038453,1.396905637,1.652841039,2.122887626,3.676949226,4.746921039,3.733563649,6.466721933,5.029395863,.2882545515,.4992715287,.6445567719,1.241790556,2.150844335,2.644908288,.1234139165,.2137591738,.6438756851,.5047665480e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1543374128e-5,.2673202406e-5,.3753828413e-7,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,.1730372742,.2997093506,.3869231079,.4578135952,.7507244251e-2,.1300292847e-1,.1678670846e-1,.2611431652e-3,.4523132303e-3,.8052099006e-5,2.258876421,3.912488728,5.051001230,5.976425253,.2400540172,.4157857543,.5367771006,.1320313609e-1,.2286850254e-1,.5562127353e-3,1.981945831,3.432830877,4.431765606,.2670149191,.4624834065,.1778563440e-1,1.613559350,2.794766775,.2632657742,1.254921551},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1543374128e-5,.2673202406e-5,.3753828413e-7,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.9099854801e-1,.1576141086,.2034789393,.2407595278,.3947983634e-2,.6838108240e-2,.8827959777e-2,.1373325428e-3,.2378669418e-3,.4234517225e-5,.6247151922,1.082038453,1.396905637,1.652841039,.6638937401e-1,.1149897688,.1484511532,.3651461244e-2,.6324516397e-2,.1538262752e-3,.2882545515,.4992715287,.6445567719,.3883469697e-1,.6726366826e-1,.2586745805e-2,.1234139165,.2137591738,.2013601810e-1,.5047665480e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.6247151922,1.082038453,1.396905637,1.652841039,.2882545515,.4992715287,.6445567719,.1234139165,.2137591738,.5047665480e-1,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.909798360,5.039918600,6.506506935,7.698602827,2.122887626,3.676949226,4.746921039,1.241790556,2.150844335,.6438756851,2.258876421,3.912488728,5.051001230,5.976425253,4.036753290,6.991861793,9.026454761,3.733563649,6.466721933,2.644908288,1.981945831,3.432830877,4.431765606,4.490128369,7.777130471,5.029395863,1.613559350,2.794766775,4.427082670,1.254921551},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.258876421,3.912488728,5.051001230,5.976425253,1.981945831,3.432830877,4.431765606,1.613559350,2.794766775,1.254921551,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.1730372742,.2997093506,.3869231079,.4578135952,.2400540172,.4157857543,.5367771006,.2670149191,.4624834065,.2632657742,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.7507244251e-2,.1300292847e-1,.1678670846e-1,.1320313609e-1,.2286850254e-1,.1778563440e-1,.6061868131e-4,.1049946359e-3,.1355474921e-3,.2611431652e-3,.4523132303e-3,.5562127353e-3,.1543374128e-5,.2673202406e-5,.8052099006e-5,.3753828413e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,.2296154773,.3977056730,.5134358159,.6075054502,.6423249772e-1,.1112539495,.1436282312,.1667256107e-1,.2887772287e-1,.4134169639e-2,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,-.1192467538e-1,-.2065414363e-1,-.2666438477e-1,-.3154972553e-1,-.5274365494e-2,-.9135469009e-2,-.1179383978e-1,-.1870471581e-2,-.3239751814e-2,-.5879824074e-3,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,.1118252823e-3,.1936870705e-3,.2500489327e-3,.6270337257e-4,.1086054271e-3,.2693007136e-4,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,-.8407960796e-6,-.1456301529e-5,-.5709621067e-6,.3368848215e-8,.5835016271e-8,.5603684524e-8,-.1771076019e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,3.501856062,6.065392619,7.830388205,9.265040271,3.825608374,6.626148073,8.554320378,3.877899792,6.716719466,3.755183165,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,2.196470374,3.804398286,4.911457068,5.811314373,3.794002148,6.571404483,8.483646708,5.254447192,9.100969504,6.450432215,.2296154773,.3977056730,.5134358159,.6075054502,.9715138718,1.682711386,2.172371058,2.127396643,3.684759074,3.568154409,.6423249772e-1,.1112539495,.1436282312,.3445322647,.5967473875,.9136820088,.1667256107e-1,.2887772287e-1,.1083036558,.4134169639e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,.3368848215e-8,.5835016271e-8,-.1771076019e-10,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,-.4656880908e-1,-.8065954338e-1,-.1041310228,-.1232094877,.4367054075e-3,.7563959535e-3,.9765029769e-3,-.3283516813e-5,-.5687217949e-5,.2188377514e-7,3.501856062,6.065392619,7.830388205,9.265040271,-.8043912806e-1,-.1393246566,-.1798673583,.9562865187e-3,.1656336837e-2,-.8707719261e-5,3.825608374,6.626148073,8.554320378,-.1114029814,-.1929556239,.1603921848e-2,3.877899792,6.716719466,-.1367598443,3.755183165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,.3368848215e-8,.5835016271e-8,-.1771076019e-10,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,-.1192467538e-1,-.2065414363e-1,-.2666438477e-1,-.3154972553e-1,.1118252823e-3,.1936870705e-3,.2500489327e-3,-.8407960796e-6,-.1456301529e-5,.5603684524e-8,.2296154773,.3977056730,.5134358159,.6075054502,-.5274365494e-2,-.9135469009e-2,-.1179383978e-1,.6270337257e-4,.1086054271e-3,-.5709621067e-6,.6423249772e-1,.1112539495,.1436282312,-.1870471581e-2,-.3239751814e-2,.2693007136e-4,.1667256107e-1,.2887772287e-1,-.5879824074e-3,.4134169639e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,.2296154773,.3977056730,.5134358159,.6075054502,.6423249772e-1,.1112539495,.1436282312,.1667256107e-1,.2887772287e-1,.4134169639e-2,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,2.196470374,3.804398286,4.911457068,5.811314373,.9715138718,1.682711386,2.172371058,.3445322647,.5967473875,.1083036558,3.501856062,6.065392619,7.830388205,9.265040271,3.794002148,6.571404483,8.483646708,2.127396643,3.684759074,.9136820088,3.825608374,6.626148073,8.554320378,5.254447192,9.100969504,3.568154409,3.877899792,6.716719466,6.450432215,3.755183165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,3.501856062,6.065392619,7.830388205,9.265040271,3.825608374,6.626148073,8.554320378,3.877899792,6.716719466,3.755183165,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,-.4656880908e-1,-.8065954338e-1,-.1041310228,-.1232094877,-.8043912806e-1,-.1393246566,-.1798673583,-.1114029814,-.1929556239,-.1367598443,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,.4367054075e-3,.7563959535e-3,.9765029769e-3,.9562865187e-3,.1656336837e-2,.1603921848e-2,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,-.3283516813e-5,-.5687217949e-5,-.8707719261e-5,.3368848215e-8,.5835016271e-8,.2188377514e-7,-.1771076019e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1963948976,.3401659410,.4391523414,.5196120579,.5080990387e-1,.8800533503e-1,.1136143990,.1219720719e-1,.2112618256e-1,.2797118711e-2,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.1065535599,.1845561795,.2382610033,.2819142209,.4358691597e-1,.7549475298e-1,.9746330700e-1,.1429557258e-1,.2476065805e-1,.4156035932e-2,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.9654244188e-2,.1672164144e-1,.2158754627e-1,.5006493970e-2,.8671501924e-2,.1988588036e-2,.5521216640e-3,.9563027739e-3,.1234581572e-2,.7013355149e-3,.1214748745e-2,.4404604930e-3,.2935681623e-4,.5084749726e-4,.4516122448e-4,.1491150883e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,3.236488548,5.605762603,7.237008402,8.562943823,3.399101902,5.887417193,7.600622914,3.312440982,5.737316079,3.083689087,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,1.952889915,3.382504555,4.366794603,5.166861053,3.242932261,5.616923440,7.251416980,4.317727482,7.478523374,5.095712344,.1963948976,.3401659410,.4391523414,.5196120579,.7988512884,1.383651019,1.786285784,1.681718474,2.912821842,2.711667007,.5080990387e-1,.8800533503e-1,.1136143990,.2620060705,.4538078261,.6679814775,.1219720719e-1,.2112618256e-1,.7617090096e-1,.2797118711e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2935681623e-4,.5084749726e-4,.1491150883e-5,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,.4325535702,.7492047606,.9672191871,1.144429176,.3919134934e-1,.6788140826e-1,.8763452124e-1,.2847067532e-2,.4931265619e-2,.1833317338e-3,3.236488548,5.605762603,7.237008402,8.562943823,.7182903227,1.244115334,1.606145989,.8250448766e-1,.1429019645,.7258566080e-2,3.399101902,5.887417193,7.600622914,.9563511098,1.656448712,.1330333825,3.312440982,5.737316079,1.128670157,3.083689087},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2935681623e-4,.5084749726e-4,.1491150883e-5,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1065535599,.1845561795,.2382610033,.2819142209,.9654244188e-2,.1672164144e-1,.2158754627e-1,.7013355149e-3,.1214748745e-2,.4516122448e-4,.1963948976,.3401659410,.4391523414,.5196120579,.4358691597e-1,.7549475298e-1,.9746330700e-1,.5006493970e-2,.8671501924e-2,.4404604930e-3,.5080990387e-1,.8800533503e-1,.1136143990,.1429557258e-1,.2476065805e-1,.1988588036e-2,.1219720719e-1,.2112618256e-1,.4156035932e-2,.2797118711e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1963948976,.3401659410,.4391523414,.5196120579,.5080990387e-1,.8800533503e-1,.1136143990,.1219720719e-1,.2112618256e-1,.2797118711e-2,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,1.952889915,3.382504555,4.366794603,5.166861053,.7988512884,1.383651019,1.786285784,.2620060705,.4538078261,.7617090096e-1,3.236488548,5.605762603,7.237008402,8.562943823,3.242932261,5.616923440,7.251416980,1.681718474,2.912821842,.6679814775,3.399101902,5.887417193,7.600622914,4.317727482,7.478523374,2.711667007,3.312440982,5.737316079,5.095712344,3.083689087},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,3.236488548,5.605762603,7.237008402,8.562943823,3.399101902,5.887417193,7.600622914,3.312440982,5.737316079,3.083689087,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.4325535702,.7492047606,.9672191871,1.144429176,.7182903227,1.244115334,1.606145989,.9563511098,1.656448712,1.128670157,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.3919134934e-1,.6788140826e-1,.8763452124e-1,.8250448766e-1,.1429019645,.1330333825,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2847067532e-2,.4931265619e-2,.7258566080e-2,.2935681623e-4,.5084749726e-4,.1833317338e-3,.1491150883e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.4641847056e-2,.8039914942e-2,.1037948556e-1,.5018612597e-3,.8692492001e-3,.5183403142e-4,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1638187454e-1,.2837423903e-1,.3663098508e-1,.4334236605e-1,.3018092809e-2,.5227490086e-2,.6748660682e-2,.4458192555e-3,.7721816018e-3,.5837372111e-4,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.5066741239e-3,.8775853253e-3,.1132957783e-2,.1183382006e-3,.2049677758e-3,.2116980724e-4,.2196237131e-4,.3803994296e-4,.4910935520e-4,.1256467012e-4,.2176264704e-4,.3553964112e-5,.3986283756e-6,.6904445999e-6,.2761886212e-6,.6911871564e-8},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,4.440487963,7.691150763,9.929232940,11.74842685,5.462596453,9.461494597,12.21473700,6.235358013,10.79995688,6.799267692,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,1.030236404,1.784421795,2.303678632,2.725749316,2.003896641,3.470850795,4.480849108,3.125150808,5.412919982,4.320151396,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.1898042300,.3287505697,.4244151605,.4680276955,.8106477480,.8839606376,.4641847056e-2,.8039914942e-2,.1037948556e-1,.2803703724e-1,.4856157301e-1,.8372660778e-1,.5018612597e-3,.8692492001e-3,.3671053173e-2,.5183403142e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.2196237131e-4,.3803994296e-4,.4910935520e-4,.3986283756e-6,.6904445999e-6,.6911871564e-8,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,.1729549620,.2995667817,.3867390522,.4575958176,.5349314793e-2,.9265285003e-2,.1196143151e-1,.1326540524e-3,.2297635587e-3,.2915917367e-5,4.440487963,7.691150763,9.929232940,11.74842685,.3364119790,.5826826397,.7522400533,.1319057787e-1,.2284675106e-1,.3961429206e-3,5.462596453,9.461494597,12.21473700,.5246469034,.9087150930,.2491295222e-1,6.235358013,10.79995688,.7252622967,6.799267692},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.2196237131e-4,.3803994296e-4,.4910935520e-4,.3986283756e-6,.6904445999e-6,.6911871564e-8,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.1638187454e-1,.2837423903e-1,.3663098508e-1,.4334236605e-1,.5066741239e-3,.8775853253e-3,.1132957783e-2,.1256467012e-4,.2176264704e-4,.2761886212e-6,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.3018092809e-2,.5227490086e-2,.6748660682e-2,.1183382006e-3,.2049677758e-3,.3553964112e-5,.4641847056e-2,.8039914942e-2,.1037948556e-1,.4458192555e-3,.7721816018e-3,.2116980724e-4,.5018612597e-3,.8692492001e-3,.5837372111e-4,.5183403142e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.4641847056e-2,.8039914942e-2,.1037948556e-1,.5018612597e-3,.8692492001e-3,.5183403142e-4,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,1.030236404,1.784421795,2.303678632,2.725749316,.1898042300,.3287505697,.4244151605,.2803703724e-1,.4856157301e-1,.3671053173e-2,4.440487963,7.691150763,9.929232940,11.74842685,2.003896641,3.470850795,4.480849108,.4680276955,.8106477480,.8372660778e-1,5.462596453,9.461494597,12.21473700,3.125150808,5.412919982,.8839606376,6.235358013,10.79995688,4.320151396,6.799267692},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,4.440487963,7.691150763,9.929232940,11.74842685,5.462596453,9.461494597,12.21473700,6.235358013,10.79995688,6.799267692,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1729549620,.2995667817,.3867390522,.4575958176,.3364119790,.5826826397,.7522400533,.5246469034,.9087150930,.7252622967,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.5349314793e-2,.9265285003e-2,.1196143151e-1,.1319057787e-1,.2284675106e-1,.2491295222e-1,.2196237131e-4,.3803994296e-4,.4910935520e-4,.1326540524e-3,.2297635587e-3,.3961429206e-3,.3986283756e-6,.6904445999e-6,.2915917367e-5,.6911871564e-8}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 52; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_top_face(
		const T i_input[56],
		T o_output[52]
	) {
		static const T matrix[52][56] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,1.154700539,0.,0.,0.,0.,.6085806194,0.,0.,0.,.2771598064,0.,0.,.1171213948,0.,.4728035100e-1,1.154700539,0.,0.,0.,0.,1.490711985,0.,0.,0.,1.073435315,0.,0.,.6197481677,0.,.3171662365,.6085806194,0.,0.,0.,1.073435315,0.,0.,.9799078930,0.,.6851573789,.2771598064,0.,0.,.6197481677,0.,.6851573789,.1171213948,0.,.3171662365,.4728035100e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.722976739,0.,0.,0.,0.,1.354996613,0.,0.,0.,.9207895610,0.,0.,.5805991817,0.,.3497284165,.1814813775e-1,0.,0.,0.,0.,.3495960260e-1,0.,0.,0.,.3756286152e-1,0.,0.,.3235996830e-1,0.,.2471097213e-1,.1503292582e-3,0.,0.,0.,.3956501378e-3,0.,0.,.5389279955e-3,0.,.5622712446e-3,.1076015046e-5,0.,0.,.3590156597e-5,0.,.5922410770e-5,.7146395199e-8,0.,.2887671511e-7,.4534132255e-10},
			{1.414213562,0.,0.,0.,0.,0.,1.722976739,0.,0.,0.,0.,1.354996613,0.,0.,0.,.9207895610,0.,0.,.5805991817,0.,.3497284165,1.722976739,0.,0.,0.,0.,3.319050305,0.,0.,0.,3.566202636,0.,0.,3.072242092,0.,2.346049539,1.354996613,0.,0.,0.,3.566202636,0.,0.,4.857641268,0.,5.068046242,.9207895610,0.,0.,3.072242092,0.,5.068046242,.5805991817,0.,2.346049539,.3497284165},
			{1.414213562,0.,0.,0.,0.,0.,.1814813775e-1,0.,0.,0.,0.,.1503292582e-3,0.,0.,0.,.1076015046e-5,0.,0.,.7146395199e-8,0.,.4534132255e-10,1.722976739,0.,0.,0.,0.,.3495960260e-1,0.,0.,0.,.3956501378e-3,0.,0.,.3590156597e-5,0.,.2887671511e-7,1.354996613,0.,0.,0.,.3756286152e-1,0.,0.,.5389279955e-3,0.,.5922410770e-5,.9207895610,0.,0.,.3235996830e-1,0.,.5622712446e-3,.5805991817,0.,.2471097213e-1,.3497284165},
			{1.414213562,0.,0.,0.,0.,0.,1.432300150,0.,0.,0.,0.,.9363699245,0.,0.,0.,.5289617350,0.,0.,.2772649114,0.,.1388366216,.5995013155,0.,0.,0.,0.,.9600175758,0.,0.,0.,.8574839288,0.,0.,.6140871797,0.,.3898220464,.1640437399,0.,0.,0.,.3589071350,0.,0.,.4064022821,0.,.3524730288,.3878756937e-1,0.,0.,.1075826497,0.,.1475305609,.8509799423e-2,0.,.2858477025e-1,.1783546975e-2},
			{1.414213562,0.,0.,0.,0.,0.,1.432300150,0.,0.,0.,0.,.9363699245,0.,0.,0.,.5289617350,0.,0.,.2772649114,0.,.1388366216,1.432300150,0.,0.,0.,0.,2.293628526,0.,0.,0.,2.048659991,0.,0.,1.467148005,0.,.9313443707,.9363699245,0.,0.,0.,2.048659991,0.,0.,2.319764682,0.,2.011933789,.5289617350,0.,0.,1.467148005,0.,2.011933789,.2772649114,0.,.9313443707,.1388366216},
			{1.414213562,0.,0.,0.,0.,0.,.5995013155,0.,0.,0.,0.,.1640437399,0.,0.,0.,.3878756937e-1,0.,0.,.8509799423e-2,0.,.1783546975e-2,1.432300150,0.,0.,0.,0.,.9600175758,0.,0.,0.,.3589071350,0.,0.,.1075826497,0.,.2858477025e-1,.9363699245,0.,0.,0.,.8574839288,0.,0.,.4064022821,0.,.1475305609,.5289617350,0.,0.,.6140871797,0.,.3524730288,.2772649114,0.,.3898220464,.1388366216},
			{1.414213562,0.,0.,0.,0.,0.,1.629716393,0.,0.,0.,0.,1.212281421,0.,0.,0.,.7792166912,0.,0.,.4647366240,0.,.2647853914,.2046688294,0.,0.,0.,0.,.3729226366,0.,0.,0.,.3790037856,0.,0.,.3088343411,0.,.2230693744,.1911977567e-1,0.,0.,0.,.4759739883e-1,0.,0.,.6132467595e-1,0.,.6051778323e-1,.1543395612e-2,0.,0.,.4870851947e-2,0.,.7600159080e-2,.1156020631e-3,0.,.4418338282e-3,.8271654356e-5},
			{1.414213562,0.,0.,0.,0.,0.,1.629716393,0.,0.,0.,0.,1.212281421,0.,0.,0.,.7792166912,0.,0.,.4647366240,0.,.2647853914,1.629716393,0.,0.,0.,0.,2.969470906,0.,0.,0.,3.017893269,0.,0.,2.459155065,0.,1.776234404,1.212281421,0.,0.,0.,3.017893269,0.,0.,3.888265562,0.,3.837104864,.7792166912,0.,0.,2.459155065,0.,3.837104864,.4647366240,0.,1.776234404,.2647853914},
			{1.414213562,0.,0.,0.,0.,0.,.2046688294,0.,0.,0.,0.,.1911977567e-1,0.,0.,0.,.1543395612e-2,0.,0.,.1156020631e-3,0.,.8271654356e-5,1.629716393,0.,0.,0.,0.,.3729226366,0.,0.,0.,.4759739883e-1,0.,0.,.4870851947e-2,0.,.4418338282e-3,1.212281421,0.,0.,0.,.3790037856,0.,0.,.6132467595e-1,0.,.7600159080e-2,.7792166912,0.,0.,.3088343411,0.,.6051778323e-1,.4647366240,0.,.2230693744,.2647853914},
			{1.414213562,0.,0.,0.,0.,0.,.8333026341,0.,0.,0.,0.,.3169457192,0.,0.,0.,.1041670603,0.,0.,.3176653023e-1,0.,.9254386281e-2,1.797496348,0.,0.,0.,0.,1.674656646,0.,0.,0.,.8702445254,0.,0.,.3625885274,0.,.1339118908,1.474739846,0.,0.,0.,1.877182782,0.,0.,1.236657185,0.,.6240046052,1.045505193,0.,0.,1.687114596,0.,1.346024785,.6877503643,0.,1.344048463,.4321892234},
			{1.414213562,0.,0.,0.,0.,0.,.8333026341,0.,0.,0.,0.,.3169457192,0.,0.,0.,.1041670603,0.,0.,.3176653023e-1,0.,.9254386281e-2,.8333026341,0.,0.,0.,0.,.7763552880,0.,0.,0.,.4034372898,0.,0.,.1680926781,0.,.6208031047e-1,.3169457192,0.,0.,0.,.4034372898,0.,0.,.2657778603,0.,.1341087982,.1041670603,0.,0.,.1680926781,0.,.1341087982,.3176653023e-1,0.,.6208031047e-1,.9254386281e-2},
			{1.414213562,0.,0.,0.,0.,0.,1.797496348,0.,0.,0.,0.,1.474739846,0.,0.,0.,1.045505193,0.,0.,.6877503643,0.,.4321892234,.8333026341,0.,0.,0.,0.,1.674656646,0.,0.,0.,1.877182782,0.,0.,1.687114596,0.,1.344048463,.3169457192,0.,0.,0.,.8702445254,0.,0.,1.236657185,0.,1.346024785,.1041670603,0.,0.,.3625885274,0.,.6240046052,.3176653023e-1,0.,.1339118908,.9254386281e-2},
			{1.414213562,0.,0.,0.,0.,0.,.5125685468,0.,0.,0.,0.,.1199176989,0.,0.,0.,.2424253337e-1,0.,0.,.4547435761e-2,0.,.8148799146e-3,2.438964522,0.,0.,0.,0.,1.397695098,0.,0.,0.,.4467629626,0.,0.,.1144983212,0.,.2601078148e-1,2.715128242,0.,0.,0.,2.125840578,0.,0.,.8614352786,0.,.2673685503,2.611792036,0.,0.,2.592424872,0.,1.272224784,2.331206226,0.,2.802295901,1.987747725},
			{1.414213562,0.,0.,0.,0.,0.,.5125685468,0.,0.,0.,0.,.1199176989,0.,0.,0.,.2424253337e-1,0.,0.,.4547435761e-2,0.,.8148799146e-3,.5125685468,0.,0.,0.,0.,.2937371735,0.,0.,0.,.9389092801e-1,0.,0.,.2406276826e-1,0.,.5466380647e-2,.1199176989,0.,0.,0.,.9389092801e-1,0.,0.,.3804657725e-1,0.,.1180873184e-1,.2424253337e-1,0.,0.,.2406276826e-1,0.,.1180873184e-1,.4547435761e-2,0.,.5466380647e-2,.8148799146e-3},
			{1.414213562,0.,0.,0.,0.,0.,2.438964522,0.,0.,0.,0.,2.715128242,0.,0.,0.,2.611792036,0.,0.,2.331206226,0.,1.987747725,.5125685468,0.,0.,0.,0.,1.397695098,0.,0.,0.,2.125840578,0.,0.,2.592424872,0.,2.802295901,.1199176989,0.,0.,0.,.4467629626,0.,0.,.8614352786,0.,1.272224784,.2424253337e-1,0.,0.,.1144983212,0.,.2673685503,.4547435761e-2,0.,.2601078148e-1,.8148799146e-3},
			{1.414213562,0.,0.,0.,0.,0.,.2614190785,0.,0.,0.,0.,.3119276979e-1,0.,0.,0.,.3216132869e-2,0.,0.,.3076861022e-3,0.,.2812031928e-4,2.941263458,0.,0.,0.,0.,.8596589978,0.,0.,0.,.1401445651,0.,0.,.1831823152e-1,0.,.2122380039e-2,3.948637231,0.,0.,0.,1.576786555,0.,0.,.3258743636,0.,.5158497099e-1,4.580616431,0.,0.,2.318872338,0.,.5803898901,4.930539837,0.,3.022825632,5.069947794},
			{1.414213562,0.,0.,0.,0.,0.,.2614190785,0.,0.,0.,0.,.3119276979e-1,0.,0.,0.,.3216132869e-2,0.,0.,.3076861022e-3,0.,.2812031928e-4,.2614190785,0.,0.,0.,0.,.7640636963e-1,0.,0.,0.,.1245602905e-1,0.,0.,.1628121817e-2,0.,.1886368363e-3,.3119276979e-1,0.,0.,0.,.1245602905e-1,0.,0.,.2574286624e-2,0.,.4075021408e-3,.3216132869e-2,0.,0.,.1628121817e-2,0.,.4075021408e-3,.3076861022e-3,0.,.1886368363e-3,.2812031928e-4},
			{1.414213562,0.,0.,0.,0.,0.,2.941263458,0.,0.,0.,0.,3.948637231,0.,0.,0.,4.580616431,0.,0.,4.930539837,0.,5.069947794,.2614190785,0.,0.,0.,0.,.8596589978,0.,0.,0.,1.576786555,0.,0.,2.318872338,0.,3.022825632,.3119276979e-1,0.,0.,0.,.1401445651,0.,0.,.3258743636,0.,.5803898901,.3216132869e-2,0.,0.,.1831823152e-1,0.,.5158497099e-1,.3076861022e-3,0.,.2122380039e-2,.2812031928e-4},
			{1.414213562,0.,0.,0.,0.,0.,.5749162514e-1,0.,0.,0.,0.,.1508650189e-2,0.,0.,0.,.3420864716e-4,0.,0.,.7197424251e-6,0.,.1446627697e-7,3.349118367,0.,0.,0.,0.,.2152732603,0.,0.,0.,.7718047390e-2,0.,0.,.2218615436e-3,0.,.5653129632e-5,5.119651213,0.,0.,0.,.4496072984,0.,0.,.2043515547e-1,0.,.7114071839e-3,6.762598981,0.,0.,.7528940303,0.,.4144232937e-1,8.288591754,0.,1.117549271,9.704794288},
			{1.414213562,0.,0.,0.,0.,0.,.5749162514e-1,0.,0.,0.,0.,.1508650189e-2,0.,0.,0.,.3420864716e-4,0.,0.,.7197424251e-6,0.,.1446627697e-7,.5749162514e-1,0.,0.,0.,0.,.3695423161e-2,0.,0.,0.,.1324895208e-3,0.,0.,.3808518930e-5,0.,.9704273609e-7,.1508650189e-2,0.,0.,0.,.1324895208e-3,0.,0.,.6021797165e-5,0.,.2096362696e-6,.3420864716e-4,0.,0.,.3808518930e-5,0.,.2096362696e-6,.7197424251e-6,0.,.9704273609e-7,.1446627697e-7},
			{1.414213562,0.,0.,0.,0.,0.,3.349118367,0.,0.,0.,0.,5.119651213,0.,0.,0.,6.762598981,0.,0.,8.288591754,0.,9.704794288,.5749162514e-1,0.,0.,0.,0.,.2152732603,0.,0.,0.,.4496072984,0.,0.,.7528940303,0.,1.117549271,.1508650189e-2,0.,0.,0.,.7718047390e-2,0.,0.,.2043515547e-1,0.,.4144232937e-1,.3420864716e-4,0.,0.,.2218615436e-3,0.,.7114071839e-3,.7197424251e-6,0.,.5653129632e-5,.1446627697e-7},
			{1.414213562,0.,0.,0.,0.,0.,1.027298721,0.,0.,0.,0.,.4816958184,0.,0.,0.,.1951696965,0.,0.,.7337460098e-1,0.,.2635224308e-1,.3587967227,0.,0.,0.,0.,.4120977289,0.,0.,0.,.2640035315,0.,0.,.1356052452,0.,.6174127088e-1,.5875925975e-1,0.,0.,0.,.9220648278e-1,0.,0.,.7488558248e-1,0.,.4658334547e-1,.8315101345e-2,0.,0.,.1654169192e-1,0.,.1626980677e-1,.1091823989e-2,0.,.2630454079e-2,.1369544625e-3},
			{1.414213562,0.,0.,0.,0.,0.,2.078006171,0.,0.,0.,0.,1.970938383,0.,0.,0.,1.615335150,0.,0.,1.228417782,0.,.8924173718,1.027298721,0.,0.,0.,0.,2.386704142,0.,0.,0.,3.092845228,0.,0.,3.213475734,0.,2.959539815,.4816958184,0.,0.,0.,1.529002171,0.,0.,2.511857186,0.,3.160657901,.1951696965,0.,0.,.7853709873,0.,1.562526553,.7337460098e-1,0.,.3575805848,.2635224308e-1},
			{1.414213562,0.,0.,0.,0.,0.,.3587967227,0.,0.,0.,0.,.5875925975e-1,0.,0.,0.,.8315101345e-2,0.,0.,.1091823989e-2,0.,.1369544625e-3,2.078006171,0.,0.,0.,0.,.8335857979,0.,0.,0.,.1865140454,0.,0.,.3346031412e-1,0.,.5320847480e-2,1.970938383,0.,0.,0.,1.080214264,0.,0.,.3064067886,0.,.6657061456e-1,1.615335150,0.,0.,1.122345952,0.,.3855501990,1.228417782,0.,1.033655707,.8924173718},
			{1.414213562,0.,0.,0.,0.,0.,.3587967227,0.,0.,0.,0.,.5875925975e-1,0.,0.,0.,.8315101345e-2,0.,0.,.1091823989e-2,0.,.1369544625e-3,1.027298721,0.,0.,0.,0.,.4120977289,0.,0.,0.,.9220648278e-1,0.,0.,.1654169192e-1,0.,.2630454079e-2,.4816958184,0.,0.,0.,.2640035315,0.,0.,.7488558248e-1,0.,.1626980677e-1,.1951696965,0.,0.,.1356052452,0.,.4658334547e-1,.7337460098e-1,0.,.6174127088e-1,.2635224308e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.027298721,0.,0.,0.,0.,.4816958184,0.,0.,0.,.1951696965,0.,0.,.7337460098e-1,0.,.2635224308e-1,2.078006171,0.,0.,0.,0.,2.386704142,0.,0.,0.,1.529002171,0.,0.,.7853709873,0.,.3575805848,1.970938383,0.,0.,0.,3.092845228,0.,0.,2.511857186,0.,1.562526553,1.615335150,0.,0.,3.213475734,0.,3.160657901,1.228417782,0.,2.959539815,.8924173718},
			{1.414213562,0.,0.,0.,0.,0.,2.078006171,0.,0.,0.,0.,1.970938383,0.,0.,0.,1.615335150,0.,0.,1.228417782,0.,.8924173718,.3587967227,0.,0.,0.,0.,.8335857979,0.,0.,0.,1.080214264,0.,0.,1.122345952,0.,1.033655707,.5875925975e-1,0.,0.,0.,.1865140454,0.,0.,.3064067886,0.,.3855501990,.8315101345e-2,0.,0.,.3346031412e-1,0.,.6657061456e-1,.1091823989e-2,0.,.5320847480e-2,.1369544625e-3},
			{1.414213562,0.,0.,0.,0.,0.,1.169907010,0.,0.,0.,0.,.6247151922,0.,0.,0.,.2882545515,0.,0.,.1234139165,0.,.5047665480e-1,.6957097879e-1,0.,0.,0.,0.,.9099854801e-1,0.,0.,0.,.6638937401e-1,0.,0.,.3883469697e-1,0.,.2013601810e-1,.2209202917e-2,0.,0.,0.,.3947983634e-2,0.,0.,.3651461244e-2,0.,.2586745805e-2,.6061868131e-4,0.,0.,.1373325428e-3,0.,.1538262752e-3,.1543374128e-5,0.,.4234517225e-5,.3753828413e-7},
			{1.414213562,0.,0.,0.,0.,0.,2.224623628,0.,0.,0.,0.,2.258876421,0.,0.,0.,1.981945831,0.,0.,1.613559350,0.,1.254921551,1.169907010,0.,0.,0.,0.,2.909798360,0.,0.,0.,4.036753290,0.,0.,4.490128369,0.,4.427082670,.6247151922,0.,0.,0.,2.122887626,0.,0.,3.733563649,0.,5.029395863,.2882545515,0.,0.,1.241790556,0.,2.644908288,.1234139165,0.,.6438756851,.5047665480e-1},
			{1.414213562,0.,0.,0.,0.,0.,.6957097879e-1,0.,0.,0.,0.,.2209202917e-2,0.,0.,0.,.6061868131e-4,0.,0.,.1543374128e-5,0.,.3753828413e-7,2.224623628,0.,0.,0.,0.,.1730372742,0.,0.,0.,.7507244251e-2,0.,0.,.2611431652e-3,0.,.8052099006e-5,2.258876421,0.,0.,0.,.2400540172,0.,0.,.1320313609e-1,0.,.5562127353e-3,1.981945831,0.,0.,.2670149191,0.,.1778563440e-1,1.613559350,0.,.2632657742,1.254921551},
			{1.414213562,0.,0.,0.,0.,0.,.6957097879e-1,0.,0.,0.,0.,.2209202917e-2,0.,0.,0.,.6061868131e-4,0.,0.,.1543374128e-5,0.,.3753828413e-7,1.169907010,0.,0.,0.,0.,.9099854801e-1,0.,0.,0.,.3947983634e-2,0.,0.,.1373325428e-3,0.,.4234517225e-5,.6247151922,0.,0.,0.,.6638937401e-1,0.,0.,.3651461244e-2,0.,.1538262752e-3,.2882545515,0.,0.,.3883469697e-1,0.,.2586745805e-2,.1234139165,0.,.2013601810e-1,.5047665480e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.169907010,0.,0.,0.,0.,.6247151922,0.,0.,0.,.2882545515,0.,0.,.1234139165,0.,.5047665480e-1,2.224623628,0.,0.,0.,0.,2.909798360,0.,0.,0.,2.122887626,0.,0.,1.241790556,0.,.6438756851,2.258876421,0.,0.,0.,4.036753290,0.,0.,3.733563649,0.,2.644908288,1.981945831,0.,0.,4.490128369,0.,5.029395863,1.613559350,0.,4.427082670,1.254921551},
			{1.414213562,0.,0.,0.,0.,0.,2.224623628,0.,0.,0.,0.,2.258876421,0.,0.,0.,1.981945831,0.,0.,1.613559350,0.,1.254921551,.6957097879e-1,0.,0.,0.,0.,.1730372742,0.,0.,0.,.2400540172,0.,0.,.2670149191,0.,.2632657742,.2209202917e-2,0.,0.,0.,.7507244251e-2,0.,0.,.1320313609e-1,0.,.1778563440e-1,.6061868131e-4,0.,0.,.2611431652e-3,0.,.5562127353e-3,.1543374128e-5,0.,.8052099006e-5,.3753828413e-7},
			{1.414213562,0.,0.,0.,0.,0.,.7092688532,0.,0.,0.,0.,.2296154773,0.,0.,0.,.6423249772e-1,0.,0.,.1667256107e-1,0.,.4134169639e-2,-.1503767417e-1,-0.,-0.,-0.,-0.,-.1192467538e-1,-0.,-0.,-0.,-.5274365494e-2,-0.,-0.,-.1870471581e-2,-0.,-.5879824074e-3,.1032145021e-3,0.,0.,0.,.1118252823e-3,0.,0.,.6270337257e-4,0.,.2693007136e-4,-.6121588826e-6,-0.,-0.,-.8407960796e-6,-0.,-.5709621067e-6,.3368848215e-8,0.,.5603684524e-8,-.1771076019e-10},
			{1.414213562,0.,0.,0.,0.,0.,2.769870437,0.,0.,0.,0.,3.501856062,0.,0.,0.,3.825608374,0.,0.,3.877899792,0.,3.755183165,.7092688532,0.,0.,0.,0.,2.196470374,0.,0.,0.,3.794002148,0.,0.,5.254447192,0.,6.450432215,.2296154773,0.,0.,0.,.9715138718,0.,0.,2.127396643,0.,3.568154409,.6423249772e-1,0.,0.,.3445322647,0.,.9136820088,.1667256107e-1,0.,.1083036558,.4134169639e-2},
			{1.414213562,0.,0.,0.,0.,0.,-.1503767417e-1,-0.,-0.,-0.,-0.,.1032145021e-3,0.,0.,0.,-.6121588826e-6,-0.,-0.,.3368848215e-8,0.,-.1771076019e-10,2.769870437,0.,0.,0.,0.,-.4656880908e-1,-0.,-0.,-0.,.4367054075e-3,0.,0.,-.3283516813e-5,-0.,.2188377514e-7,3.501856062,0.,0.,0.,-.8043912806e-1,-0.,-0.,.9562865187e-3,0.,-.8707719261e-5,3.825608374,0.,0.,-.1114029814,-0.,.1603921848e-2,3.877899792,0.,-.1367598443,3.755183165},
			{1.414213562,0.,0.,0.,0.,0.,-.1503767417e-1,-0.,-0.,-0.,-0.,.1032145021e-3,0.,0.,0.,-.6121588826e-6,-0.,-0.,.3368848215e-8,0.,-.1771076019e-10,.7092688532,0.,0.,0.,0.,-.1192467538e-1,-0.,-0.,-0.,.1118252823e-3,0.,0.,-.8407960796e-6,-0.,.5603684524e-8,.2296154773,0.,0.,0.,-.5274365494e-2,-0.,-0.,.6270337257e-4,0.,-.5709621067e-6,.6423249772e-1,0.,0.,-.1870471581e-2,-0.,.2693007136e-4,.1667256107e-1,0.,-.5879824074e-3,.4134169639e-2},
			{1.414213562,0.,0.,0.,0.,0.,.7092688532,0.,0.,0.,0.,.2296154773,0.,0.,0.,.6423249772e-1,0.,0.,.1667256107e-1,0.,.4134169639e-2,2.769870437,0.,0.,0.,0.,2.196470374,0.,0.,0.,.9715138718,0.,0.,.3445322647,0.,.1083036558,3.501856062,0.,0.,0.,3.794002148,0.,0.,2.127396643,0.,.9136820088,3.825608374,0.,0.,5.254447192,0.,3.568154409,3.877899792,0.,6.450432215,3.755183165},
			{1.414213562,0.,0.,0.,0.,0.,2.769870437,0.,0.,0.,0.,3.501856062,0.,0.,0.,3.825608374,0.,0.,3.877899792,0.,3.755183165,-.1503767417e-1,-0.,-0.,-0.,-0.,-.4656880908e-1,-0.,-0.,-0.,-.8043912806e-1,-0.,-0.,-.1114029814,-0.,-.1367598443,.1032145021e-3,0.,0.,0.,.4367054075e-3,0.,0.,.9562865187e-3,0.,.1603921848e-2,-.6121588826e-6,-0.,-0.,-.3283516813e-5,-0.,-.8707719261e-5,.3368848215e-8,0.,.2188377514e-7,-.1771076019e-10},
			{1.414213562,0.,0.,0.,0.,0.,.6559570585,0.,0.,0.,0.,.1963948976,0.,0.,0.,.5080990387e-1,0.,0.,.1219720719e-1,0.,.2797118711e-2,.1452906103,0.,0.,0.,0.,.1065535599,0.,0.,0.,.4358691597e-1,0.,0.,.1429557258e-1,0.,.4156035932e-2,.9635061189e-2,0.,0.,0.,.9654244188e-2,0.,0.,.5006493970e-2,0.,.1988588036e-2,.5521216640e-3,0.,0.,.7013355149e-3,0.,.4404604930e-3,.2935681623e-4,0.,.4516122448e-4,.1491150883e-5},
			{1.414213562,0.,0.,0.,0.,0.,2.662853947,0.,0.,0.,0.,3.236488548,0.,0.,0.,3.399101902,0.,0.,3.312440982,0.,3.083689087,.6559570585,0.,0.,0.,0.,1.952889915,0.,0.,0.,3.242932261,0.,0.,4.317727482,0.,5.095712344,.1963948976,0.,0.,0.,.7988512884,0.,0.,1.681718474,0.,2.711667007,.5080990387e-1,0.,0.,.2620060705,0.,.6679814775,.1219720719e-1,0.,.7617090096e-1,.2797118711e-2},
			{1.414213562,0.,0.,0.,0.,0.,.1452906103,0.,0.,0.,0.,.9635061189e-2,0.,0.,0.,.5521216640e-3,0.,0.,.2935681623e-4,0.,.1491150883e-5,2.662853947,0.,0.,0.,0.,.4325535702,0.,0.,0.,.3919134934e-1,0.,0.,.2847067532e-2,0.,.1833317338e-3,3.236488548,0.,0.,0.,.7182903227,0.,0.,.8250448766e-1,0.,.7258566080e-2,3.399101902,0.,0.,.9563511098,0.,.1330333825,3.312440982,0.,1.128670157,3.083689087},
			{1.414213562,0.,0.,0.,0.,0.,.1452906103,0.,0.,0.,0.,.9635061189e-2,0.,0.,0.,.5521216640e-3,0.,0.,.2935681623e-4,0.,.1491150883e-5,.6559570585,0.,0.,0.,0.,.1065535599,0.,0.,0.,.9654244188e-2,0.,0.,.7013355149e-3,0.,.4516122448e-4,.1963948976,0.,0.,0.,.4358691597e-1,0.,0.,.5006493970e-2,0.,.4404604930e-3,.5080990387e-1,0.,0.,.1429557258e-1,0.,.1988588036e-2,.1219720719e-1,0.,.4156035932e-2,.2797118711e-2},
			{1.414213562,0.,0.,0.,0.,0.,.6559570585,0.,0.,0.,0.,.1963948976,0.,0.,0.,.5080990387e-1,0.,0.,.1219720719e-1,0.,.2797118711e-2,2.662853947,0.,0.,0.,0.,1.952889915,0.,0.,0.,.7988512884,0.,0.,.2620060705,0.,.7617090096e-1,3.236488548,0.,0.,0.,3.242932261,0.,0.,1.681718474,0.,.6679814775,3.399101902,0.,0.,4.317727482,0.,2.711667007,3.312440982,0.,5.095712344,3.083689087},
			{1.414213562,0.,0.,0.,0.,0.,2.662853947,0.,0.,0.,0.,3.236488548,0.,0.,0.,3.399101902,0.,0.,3.312440982,0.,3.083689087,.1452906103,0.,0.,0.,0.,.4325535702,0.,0.,0.,.7182903227,0.,0.,.9563511098,0.,1.128670157,.9635061189e-2,0.,0.,0.,.3919134934e-1,0.,0.,.8250448766e-1,0.,.1330333825,.5521216640e-3,0.,0.,.2847067532e-2,0.,.7258566080e-2,.2935681623e-4,0.,.1833317338e-3,.1491150883e-5},
			{1.414213562,0.,0.,0.,0.,0.,.2954311110,0.,0.,0.,0.,.3983747795e-1,0.,0.,0.,.4641847056e-2,0.,0.,.5018612597e-3,0.,.5183403142e-4,.4959665215e-1,0.,0.,0.,0.,.1638187454e-1,0.,0.,0.,.3018092809e-2,0.,0.,.4458192555e-3,0.,.5837372111e-4,.1122752692e-2,0.,0.,0.,.5066741239e-3,0.,0.,.1183382006e-3,0.,.2116980724e-4,.2196237131e-4,0.,0.,.1256467012e-4,0.,.3553964112e-5,.3986283756e-6,0.,.2761886212e-6,.6911871564e-8},
			{1.414213562,0.,0.,0.,0.,0.,3.119073853,0.,0.,0.,0.,4.440487963,0.,0.,0.,5.462596453,0.,0.,6.235358013,0.,6.799267692,.2954311110,0.,0.,0.,0.,1.030236404,0.,0.,0.,2.003896641,0.,0.,3.125150808,0.,4.320151396,.3983747795e-1,0.,0.,0.,.1898042300,0.,0.,.4680276955,0.,.8839606376,.4641847056e-2,0.,0.,.2803703724e-1,0.,.8372660778e-1,.5018612597e-3,0.,.3671053173e-2,.5183403142e-4},
			{1.414213562,0.,0.,0.,0.,0.,.4959665215e-1,0.,0.,0.,0.,.1122752692e-2,0.,0.,0.,.2196237131e-4,0.,0.,.3986283756e-6,0.,.6911871564e-8,3.119073853,0.,0.,0.,0.,.1729549620,0.,0.,0.,.5349314793e-2,0.,0.,.1326540524e-3,0.,.2915917367e-5,4.440487963,0.,0.,0.,.3364119790,0.,0.,.1319057787e-1,0.,.3961429206e-3,5.462596453,0.,0.,.5246469034,0.,.2491295222e-1,6.235358013,0.,.7252622967,6.799267692},
			{1.414213562,0.,0.,0.,0.,0.,.4959665215e-1,0.,0.,0.,0.,.1122752692e-2,0.,0.,0.,.2196237131e-4,0.,0.,.3986283756e-6,0.,.6911871564e-8,.2954311110,0.,0.,0.,0.,.1638187454e-1,0.,0.,0.,.5066741239e-3,0.,0.,.1256467012e-4,0.,.2761886212e-6,.3983747795e-1,0.,0.,0.,.3018092809e-2,0.,0.,.1183382006e-3,0.,.3553964112e-5,.4641847056e-2,0.,0.,.4458192555e-3,0.,.2116980724e-4,.5018612597e-3,0.,.5837372111e-4,.5183403142e-4},
			{1.414213562,0.,0.,0.,0.,0.,.2954311110,0.,0.,0.,0.,.3983747795e-1,0.,0.,0.,.4641847056e-2,0.,0.,.5018612597e-3,0.,.5183403142e-4,3.119073853,0.,0.,0.,0.,1.030236404,0.,0.,0.,.1898042300,0.,0.,.2803703724e-1,0.,.3671053173e-2,4.440487963,0.,0.,0.,2.003896641,0.,0.,.4680276955,0.,.8372660778e-1,5.462596453,0.,0.,3.125150808,0.,.8839606376,6.235358013,0.,4.320151396,6.799267692},
			{1.414213562,0.,0.,0.,0.,0.,3.119073853,0.,0.,0.,0.,4.440487963,0.,0.,0.,5.462596453,0.,0.,6.235358013,0.,6.799267692,.4959665215e-1,0.,0.,0.,0.,.1729549620,0.,0.,0.,.3364119790,0.,0.,.5246469034,0.,.7252622967,.1122752692e-2,0.,0.,0.,.5349314793e-2,0.,0.,.1319057787e-1,0.,.2491295222e-1,.2196237131e-4,0.,0.,.1326540524e-3,0.,.3961429206e-3,.3986283756e-6,0.,.2915917367e-5,.6911871564e-8}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 52; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_top_face(
		const T i_input[56],
		T o_output[52]
	) {
		static const T matrix[52][56] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,1.154700539,0.,0.,0.,0.,.6085806194,0.,0.,0.,.2771598064,0.,0.,.1171213948,0.,.4728035100e-1,1.154700539,0.,0.,0.,0.,1.490711985,0.,0.,0.,1.073435315,0.,0.,.6197481677,0.,.3171662365,.6085806194,0.,0.,0.,1.073435315,0.,0.,.9799078930,0.,.6851573789,.2771598064,0.,0.,.6197481677,0.,.6851573789,.1171213948,0.,.3171662365,.4728035100e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.722976739,0.,0.,0.,0.,1.354996613,0.,0.,0.,.9207895610,0.,0.,.5805991817,0.,.3497284165,.1814813775e-1,0.,0.,0.,0.,.3495960260e-1,0.,0.,0.,.3756286152e-1,0.,0.,.3235996830e-1,0.,.2471097213e-1,.1503292582e-3,0.,0.,0.,.3956501378e-3,0.,0.,.5389279955e-3,0.,.5622712446e-3,.1076015046e-5,0.,0.,.3590156597e-5,0.,.5922410770e-5,.7146395199e-8,0.,.2887671511e-7,.4534132255e-10},
			{1.414213562,0.,0.,0.,0.,0.,1.722976739,0.,0.,0.,0.,1.354996613,0.,0.,0.,.9207895610,0.,0.,.5805991817,0.,.3497284165,1.722976739,0.,0.,0.,0.,3.319050305,0.,0.,0.,3.566202636,0.,0.,3.072242092,0.,2.346049539,1.354996613,0.,0.,0.,3.566202636,0.,0.,4.857641268,0.,5.068046242,.9207895610,0.,0.,3.072242092,0.,5.068046242,.5805991817,0.,2.346049539,.3497284165},
			{1.414213562,0.,0.,0.,0.,0.,.1814813775e-1,0.,0.,0.,0.,.1503292582e-3,0.,0.,0.,.1076015046e-5,0.,0.,.7146395199e-8,0.,.4534132255e-10,1.722976739,0.,0.,0.,0.,.3495960260e-1,0.,0.,0.,.3956501378e-3,0.,0.,.3590156597e-5,0.,.2887671511e-7,1.354996613,0.,0.,0.,.3756286152e-1,0.,0.,.5389279955e-3,0.,.5922410770e-5,.9207895610,0.,0.,.3235996830e-1,0.,.5622712446e-3,.5805991817,0.,.2471097213e-1,.3497284165},
			{1.414213562,0.,0.,0.,0.,0.,1.432300150,0.,0.,0.,0.,.9363699245,0.,0.,0.,.5289617350,0.,0.,.2772649114,0.,.1388366216,.5995013155,0.,0.,0.,0.,.9600175758,0.,0.,0.,.8574839288,0.,0.,.6140871797,0.,.3898220464,.1640437399,0.,0.,0.,.3589071350,0.,0.,.4064022821,0.,.3524730288,.3878756937e-1,0.,0.,.1075826497,0.,.1475305609,.8509799423e-2,0.,.2858477025e-1,.1783546975e-2},
			{1.414213562,0.,0.,0.,0.,0.,1.432300150,0.,0.,0.,0.,.9363699245,0.,0.,0.,.5289617350,0.,0.,.2772649114,0.,.1388366216,1.432300150,0.,0.,0.,0.,2.293628526,0.,0.,0.,2.048659991,0.,0.,1.467148005,0.,.9313443707,.9363699245,0.,0.,0.,2.048659991,0.,0.,2.319764682,0.,2.011933789,.5289617350,0.,0.,1.467148005,0.,2.011933789,.2772649114,0.,.9313443707,.1388366216},
			{1.414213562,0.,0.,0.,0.,0.,.5995013155,0.,0.,0.,0.,.1640437399,0.,0.,0.,.3878756937e-1,0.,0.,.8509799423e-2,0.,.1783546975e-2,1.432300150,0.,0.,0.,0.,.9600175758,0.,0.,0.,.3589071350,0.,0.,.1075826497,0.,.2858477025e-1,.9363699245,0.,0.,0.,.8574839288,0.,0.,.4064022821,0.,.1475305609,.5289617350,0.,0.,.6140871797,0.,.3524730288,.2772649114,0.,.3898220464,.1388366216},
			{1.414213562,0.,0.,0.,0.,0.,1.629716393,0.,0.,0.,0.,1.212281421,0.,0.,0.,.7792166912,0.,0.,.4647366240,0.,.2647853914,.2046688294,0.,0.,0.,0.,.3729226366,0.,0.,0.,.3790037856,0.,0.,.3088343411,0.,.2230693744,.1911977567e-1,0.,0.,0.,.4759739883e-1,0.,0.,.6132467595e-1,0.,.6051778323e-1,.1543395612e-2,0.,0.,.4870851947e-2,0.,.7600159080e-2,.1156020631e-3,0.,.4418338282e-3,.8271654356e-5},
			{1.414213562,0.,0.,0.,0.,0.,1.629716393,0.,0.,0.,0.,1.212281421,0.,0.,0.,.7792166912,0.,0.,.4647366240,0.,.2647853914,1.629716393,0.,0.,0.,0.,2.969470906,0.,0.,0.,3.017893269,0.,0.,2.459155065,0.,1.776234404,1.212281421,0.,0.,0.,3.017893269,0.,0.,3.888265562,0.,3.837104864,.7792166912,0.,0.,2.459155065,0.,3.837104864,.4647366240,0.,1.776234404,.2647853914},
			{1.414213562,0.,0.,0.,0.,0.,.2046688294,0.,0.,0.,0.,.1911977567e-1,0.,0.,0.,.1543395612e-2,0.,0.,.1156020631e-3,0.,.8271654356e-5,1.629716393,0.,0.,0.,0.,.3729226366,0.,0.,0.,.4759739883e-1,0.,0.,.4870851947e-2,0.,.4418338282e-3,1.212281421,0.,0.,0.,.3790037856,0.,0.,.6132467595e-1,0.,.7600159080e-2,.7792166912,0.,0.,.3088343411,0.,.6051778323e-1,.4647366240,0.,.2230693744,.2647853914},
			{1.414213562,0.,0.,0.,0.,0.,.8333026341,0.,0.,0.,0.,.3169457192,0.,0.,0.,.1041670603,0.,0.,.3176653023e-1,0.,.9254386281e-2,1.797496348,0.,0.,0.,0.,1.674656646,0.,0.,0.,.8702445254,0.,0.,.3625885274,0.,.1339118908,1.474739846,0.,0.,0.,1.877182782,0.,0.,1.236657185,0.,.6240046052,1.045505193,0.,0.,1.687114596,0.,1.346024785,.6877503643,0.,1.344048463,.4321892234},
			{1.414213562,0.,0.,0.,0.,0.,.8333026341,0.,0.,0.,0.,.3169457192,0.,0.,0.,.1041670603,0.,0.,.3176653023e-1,0.,.9254386281e-2,.8333026341,0.,0.,0.,0.,.7763552880,0.,0.,0.,.4034372898,0.,0.,.1680926781,0.,.6208031047e-1,.3169457192,0.,0.,0.,.4034372898,0.,0.,.2657778603,0.,.1341087982,.1041670603,0.,0.,.1680926781,0.,.1341087982,.3176653023e-1,0.,.6208031047e-1,.9254386281e-2},
			{1.414213562,0.,0.,0.,0.,0.,1.797496348,0.,0.,0.,0.,1.474739846,0.,0.,0.,1.045505193,0.,0.,.6877503643,0.,.4321892234,.8333026341,0.,0.,0.,0.,1.674656646,0.,0.,0.,1.877182782,0.,0.,1.687114596,0.,1.344048463,.3169457192,0.,0.,0.,.8702445254,0.,0.,1.236657185,0.,1.346024785,.1041670603,0.,0.,.3625885274,0.,.6240046052,.3176653023e-1,0.,.1339118908,.9254386281e-2},
			{1.414213562,0.,0.,0.,0.,0.,.5125685468,0.,0.,0.,0.,.1199176989,0.,0.,0.,.2424253337e-1,0.,0.,.4547435761e-2,0.,.8148799146e-3,2.438964522,0.,0.,0.,0.,1.397695098,0.,0.,0.,.4467629626,0.,0.,.1144983212,0.,.2601078148e-1,2.715128242,0.,0.,0.,2.125840578,0.,0.,.8614352786,0.,.2673685503,2.611792036,0.,0.,2.592424872,0.,1.272224784,2.331206226,0.,2.802295901,1.987747725},
			{1.414213562,0.,0.,0.,0.,0.,.5125685468,0.,0.,0.,0.,.1199176989,0.,0.,0.,.2424253337e-1,0.,0.,.4547435761e-2,0.,.8148799146e-3,.5125685468,0.,0.,0.,0.,.2937371735,0.,0.,0.,.9389092801e-1,0.,0.,.2406276826e-1,0.,.5466380647e-2,.1199176989,0.,0.,0.,.9389092801e-1,0.,0.,.3804657725e-1,0.,.1180873184e-1,.2424253337e-1,0.,0.,.2406276826e-1,0.,.1180873184e-1,.4547435761e-2,0.,.5466380647e-2,.8148799146e-3},
			{1.414213562,0.,0.,0.,0.,0.,2.438964522,0.,0.,0.,0.,2.715128242,0.,0.,0.,2.611792036,0.,0.,2.331206226,0.,1.987747725,.5125685468,0.,0.,0.,0.,1.397695098,0.,0.,0.,2.125840578,0.,0.,2.592424872,0.,2.802295901,.1199176989,0.,0.,0.,.4467629626,0.,0.,.8614352786,0.,1.272224784,.2424253337e-1,0.,0.,.1144983212,0.,.2673685503,.4547435761e-2,0.,.2601078148e-1,.8148799146e-3},
			{1.414213562,0.,0.,0.,0.,0.,.2614190785,0.,0.,0.,0.,.3119276979e-1,0.,0.,0.,.3216132869e-2,0.,0.,.3076861022e-3,0.,.2812031928e-4,2.941263458,0.,0.,0.,0.,.8596589978,0.,0.,0.,.1401445651,0.,0.,.1831823152e-1,0.,.2122380039e-2,3.948637231,0.,0.,0.,1.576786555,0.,0.,.3258743636,0.,.5158497099e-1,4.580616431,0.,0.,2.318872338,0.,.5803898901,4.930539837,0.,3.022825632,5.069947794},
			{1.414213562,0.,0.,0.,0.,0.,.2614190785,0.,0.,0.,0.,.3119276979e-1,0.,0.,0.,.3216132869e-2,0.,0.,.3076861022e-3,0.,.2812031928e-4,.2614190785,0.,0.,0.,0.,.7640636963e-1,0.,0.,0.,.1245602905e-1,0.,0.,.1628121817e-2,0.,.1886368363e-3,.3119276979e-1,0.,0.,0.,.1245602905e-1,0.,0.,.2574286624e-2,0.,.4075021408e-3,.3216132869e-2,0.,0.,.1628121817e-2,0.,.4075021408e-3,.3076861022e-3,0.,.1886368363e-3,.2812031928e-4},
			{1.414213562,0.,0.,0.,0.,0.,2.941263458,0.,0.,0.,0.,3.948637231,0.,0.,0.,4.580616431,0.,0.,4.930539837,0.,5.069947794,.2614190785,0.,0.,0.,0.,.8596589978,0.,0.,0.,1.576786555,0.,0.,2.318872338,0.,3.022825632,.3119276979e-1,0.,0.,0.,.1401445651,0.,0.,.3258743636,0.,.5803898901,.3216132869e-2,0.,0.,.1831823152e-1,0.,.5158497099e-1,.3076861022e-3,0.,.2122380039e-2,.2812031928e-4},
			{1.414213562,0.,0.,0.,0.,0.,.5749162514e-1,0.,0.,0.,0.,.1508650189e-2,0.,0.,0.,.3420864716e-4,0.,0.,.7197424251e-6,0.,.1446627697e-7,3.349118367,0.,0.,0.,0.,.2152732603,0.,0.,0.,.7718047390e-2,0.,0.,.2218615436e-3,0.,.5653129632e-5,5.119651213,0.,0.,0.,.4496072984,0.,0.,.2043515547e-1,0.,.7114071839e-3,6.762598981,0.,0.,.7528940303,0.,.4144232937e-1,8.288591754,0.,1.117549271,9.704794288},
			{1.414213562,0.,0.,0.,0.,0.,.5749162514e-1,0.,0.,0.,0.,.1508650189e-2,0.,0.,0.,.3420864716e-4,0.,0.,.7197424251e-6,0.,.1446627697e-7,.5749162514e-1,0.,0.,0.,0.,.3695423161e-2,0.,0.,0.,.1324895208e-3,0.,0.,.3808518930e-5,0.,.9704273609e-7,.1508650189e-2,0.,0.,0.,.1324895208e-3,0.,0.,.6021797165e-5,0.,.2096362696e-6,.3420864716e-4,0.,0.,.3808518930e-5,0.,.2096362696e-6,.7197424251e-6,0.,.9704273609e-7,.1446627697e-7},
			{1.414213562,0.,0.,0.,0.,0.,3.349118367,0.,0.,0.,0.,5.119651213,0.,0.,0.,6.762598981,0.,0.,8.288591754,0.,9.704794288,.5749162514e-1,0.,0.,0.,0.,.2152732603,0.,0.,0.,.4496072984,0.,0.,.7528940303,0.,1.117549271,.1508650189e-2,0.,0.,0.,.7718047390e-2,0.,0.,.2043515547e-1,0.,.4144232937e-1,.3420864716e-4,0.,0.,.2218615436e-3,0.,.7114071839e-3,.7197424251e-6,0.,.5653129632e-5,.1446627697e-7},
			{1.414213562,0.,0.,0.,0.,0.,1.027298721,0.,0.,0.,0.,.4816958184,0.,0.,0.,.1951696965,0.,0.,.7337460098e-1,0.,.2635224308e-1,.3587967227,0.,0.,0.,0.,.4120977289,0.,0.,0.,.2640035315,0.,0.,.1356052452,0.,.6174127088e-1,.5875925975e-1,0.,0.,0.,.9220648278e-1,0.,0.,.7488558248e-1,0.,.4658334547e-1,.8315101345e-2,0.,0.,.1654169192e-1,0.,.1626980677e-1,.1091823989e-2,0.,.2630454079e-2,.1369544625e-3},
			{1.414213562,0.,0.,0.,0.,0.,2.078006171,0.,0.,0.,0.,1.970938383,0.,0.,0.,1.615335150,0.,0.,1.228417782,0.,.8924173718,1.027298721,0.,0.,0.,0.,2.386704142,0.,0.,0.,3.092845228,0.,0.,3.213475734,0.,2.959539815,.4816958184,0.,0.,0.,1.529002171,0.,0.,2.511857186,0.,3.160657901,.1951696965,0.,0.,.7853709873,0.,1.562526553,.7337460098e-1,0.,.3575805848,.2635224308e-1},
			{1.414213562,0.,0.,0.,0.,0.,.3587967227,0.,0.,0.,0.,.5875925975e-1,0.,0.,0.,.8315101345e-2,0.,0.,.1091823989e-2,0.,.1369544625e-3,2.078006171,0.,0.,0.,0.,.8335857979,0.,0.,0.,.1865140454,0.,0.,.3346031412e-1,0.,.5320847480e-2,1.970938383,0.,0.,0.,1.080214264,0.,0.,.3064067886,0.,.6657061456e-1,1.615335150,0.,0.,1.122345952,0.,.3855501990,1.228417782,0.,1.033655707,.8924173718},
			{1.414213562,0.,0.,0.,0.,0.,.3587967227,0.,0.,0.,0.,.5875925975e-1,0.,0.,0.,.8315101345e-2,0.,0.,.1091823989e-2,0.,.1369544625e-3,1.027298721,0.,0.,0.,0.,.4120977289,0.,0.,0.,.9220648278e-1,0.,0.,.1654169192e-1,0.,.2630454079e-2,.4816958184,0.,0.,0.,.2640035315,0.,0.,.7488558248e-1,0.,.1626980677e-1,.1951696965,0.,0.,.1356052452,0.,.4658334547e-1,.7337460098e-1,0.,.6174127088e-1,.2635224308e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.027298721,0.,0.,0.,0.,.4816958184,0.,0.,0.,.1951696965,0.,0.,.7337460098e-1,0.,.2635224308e-1,2.078006171,0.,0.,0.,0.,2.386704142,0.,0.,0.,1.529002171,0.,0.,.7853709873,0.,.3575805848,1.970938383,0.,0.,0.,3.092845228,0.,0.,2.511857186,0.,1.562526553,1.615335150,0.,0.,3.213475734,0.,3.160657901,1.228417782,0.,2.959539815,.8924173718},
			{1.414213562,0.,0.,0.,0.,0.,2.078006171,0.,0.,0.,0.,1.970938383,0.,0.,0.,1.615335150,0.,0.,1.228417782,0.,.8924173718,.3587967227,0.,0.,0.,0.,.8335857979,0.,0.,0.,1.080214264,0.,0.,1.122345952,0.,1.033655707,.5875925975e-1,0.,0.,0.,.1865140454,0.,0.,.3064067886,0.,.3855501990,.8315101345e-2,0.,0.,.3346031412e-1,0.,.6657061456e-1,.1091823989e-2,0.,.5320847480e-2,.1369544625e-3},
			{1.414213562,0.,0.,0.,0.,0.,1.169907010,0.,0.,0.,0.,.6247151922,0.,0.,0.,.2882545515,0.,0.,.1234139165,0.,.5047665480e-1,.6957097879e-1,0.,0.,0.,0.,.9099854801e-1,0.,0.,0.,.6638937401e-1,0.,0.,.3883469697e-1,0.,.2013601810e-1,.2209202917e-2,0.,0.,0.,.3947983634e-2,0.,0.,.3651461244e-2,0.,.2586745805e-2,.6061868131e-4,0.,0.,.1373325428e-3,0.,.1538262752e-3,.1543374128e-5,0.,.4234517225e-5,.3753828413e-7},
			{1.414213562,0.,0.,0.,0.,0.,2.224623628,0.,0.,0.,0.,2.258876421,0.,0.,0.,1.981945831,0.,0.,1.613559350,0.,1.254921551,1.169907010,0.,0.,0.,0.,2.909798360,0.,0.,0.,4.036753290,0.,0.,4.490128369,0.,4.427082670,.6247151922,0.,0.,0.,2.122887626,0.,0.,3.733563649,0.,5.029395863,.2882545515,0.,0.,1.241790556,0.,2.644908288,.1234139165,0.,.6438756851,.5047665480e-1},
			{1.414213562,0.,0.,0.,0.,0.,.6957097879e-1,0.,0.,0.,0.,.2209202917e-2,0.,0.,0.,.6061868131e-4,0.,0.,.1543374128e-5,0.,.3753828413e-7,2.224623628,0.,0.,0.,0.,.1730372742,0.,0.,0.,.7507244251e-2,0.,0.,.2611431652e-3,0.,.8052099006e-5,2.258876421,0.,0.,0.,.2400540172,0.,0.,.1320313609e-1,0.,.5562127353e-3,1.981945831,0.,0.,.2670149191,0.,.1778563440e-1,1.613559350,0.,.2632657742,1.254921551},
			{1.414213562,0.,0.,0.,0.,0.,.6957097879e-1,0.,0.,0.,0.,.2209202917e-2,0.,0.,0.,.6061868131e-4,0.,0.,.1543374128e-5,0.,.3753828413e-7,1.169907010,0.,0.,0.,0.,.9099854801e-1,0.,0.,0.,.3947983634e-2,0.,0.,.1373325428e-3,0.,.4234517225e-5,.6247151922,0.,0.,0.,.6638937401e-1,0.,0.,.3651461244e-2,0.,.1538262752e-3,.2882545515,0.,0.,.3883469697e-1,0.,.2586745805e-2,.1234139165,0.,.2013601810e-1,.5047665480e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.169907010,0.,0.,0.,0.,.6247151922,0.,0.,0.,.2882545515,0.,0.,.1234139165,0.,.5047665480e-1,2.224623628,0.,0.,0.,0.,2.909798360,0.,0.,0.,2.122887626,0.,0.,1.241790556,0.,.6438756851,2.258876421,0.,0.,0.,4.036753290,0.,0.,3.733563649,0.,2.644908288,1.981945831,0.,0.,4.490128369,0.,5.029395863,1.613559350,0.,4.427082670,1.254921551},
			{1.414213562,0.,0.,0.,0.,0.,2.224623628,0.,0.,0.,0.,2.258876421,0.,0.,0.,1.981945831,0.,0.,1.613559350,0.,1.254921551,.6957097879e-1,0.,0.,0.,0.,.1730372742,0.,0.,0.,.2400540172,0.,0.,.2670149191,0.,.2632657742,.2209202917e-2,0.,0.,0.,.7507244251e-2,0.,0.,.1320313609e-1,0.,.1778563440e-1,.6061868131e-4,0.,0.,.2611431652e-3,0.,.5562127353e-3,.1543374128e-5,0.,.8052099006e-5,.3753828413e-7},
			{1.414213562,0.,0.,0.,0.,0.,.7092688532,0.,0.,0.,0.,.2296154773,0.,0.,0.,.6423249772e-1,0.,0.,.1667256107e-1,0.,.4134169639e-2,-.1503767417e-1,-0.,-0.,-0.,-0.,-.1192467538e-1,-0.,-0.,-0.,-.5274365494e-2,-0.,-0.,-.1870471581e-2,-0.,-.5879824074e-3,.1032145021e-3,0.,0.,0.,.1118252823e-3,0.,0.,.6270337257e-4,0.,.2693007136e-4,-.6121588826e-6,-0.,-0.,-.8407960796e-6,-0.,-.5709621067e-6,.3368848215e-8,0.,.5603684524e-8,-.1771076019e-10},
			{1.414213562,0.,0.,0.,0.,0.,2.769870437,0.,0.,0.,0.,3.501856062,0.,0.,0.,3.825608374,0.,0.,3.877899792,0.,3.755183165,.7092688532,0.,0.,0.,0.,2.196470374,0.,0.,0.,3.794002148,0.,0.,5.254447192,0.,6.450432215,.2296154773,0.,0.,0.,.9715138718,0.,0.,2.127396643,0.,3.568154409,.6423249772e-1,0.,0.,.3445322647,0.,.9136820088,.1667256107e-1,0.,.1083036558,.4134169639e-2},
			{1.414213562,0.,0.,0.,0.,0.,-.1503767417e-1,-0.,-0.,-0.,-0.,.1032145021e-3,0.,0.,0.,-.6121588826e-6,-0.,-0.,.3368848215e-8,0.,-.1771076019e-10,2.769870437,0.,0.,0.,0.,-.4656880908e-1,-0.,-0.,-0.,.4367054075e-3,0.,0.,-.3283516813e-5,-0.,.2188377514e-7,3.501856062,0.,0.,0.,-.8043912806e-1,-0.,-0.,.9562865187e-3,0.,-.8707719261e-5,3.825608374,0.,0.,-.1114029814,-0.,.1603921848e-2,3.877899792,0.,-.1367598443,3.755183165},
			{1.414213562,0.,0.,0.,0.,0.,-.1503767417e-1,-0.,-0.,-0.,-0.,.1032145021e-3,0.,0.,0.,-.6121588826e-6,-0.,-0.,.3368848215e-8,0.,-.1771076019e-10,.7092688532,0.,0.,0.,0.,-.1192467538e-1,-0.,-0.,-0.,.1118252823e-3,0.,0.,-.8407960796e-6,-0.,.5603684524e-8,.2296154773,0.,0.,0.,-.5274365494e-2,-0.,-0.,.6270337257e-4,0.,-.5709621067e-6,.6423249772e-1,0.,0.,-.1870471581e-2,-0.,.2693007136e-4,.1667256107e-1,0.,-.5879824074e-3,.4134169639e-2},
			{1.414213562,0.,0.,0.,0.,0.,.7092688532,0.,0.,0.,0.,.2296154773,0.,0.,0.,.6423249772e-1,0.,0.,.1667256107e-1,0.,.4134169639e-2,2.769870437,0.,0.,0.,0.,2.196470374,0.,0.,0.,.9715138718,0.,0.,.3445322647,0.,.1083036558,3.501856062,0.,0.,0.,3.794002148,0.,0.,2.127396643,0.,.9136820088,3.825608374,0.,0.,5.254447192,0.,3.568154409,3.877899792,0.,6.450432215,3.755183165},
			{1.414213562,0.,0.,0.,0.,0.,2.769870437,0.,0.,0.,0.,3.501856062,0.,0.,0.,3.825608374,0.,0.,3.877899792,0.,3.755183165,-.1503767417e-1,-0.,-0.,-0.,-0.,-.4656880908e-1,-0.,-0.,-0.,-.8043912806e-1,-0.,-0.,-.1114029814,-0.,-.1367598443,.1032145021e-3,0.,0.,0.,.4367054075e-3,0.,0.,.9562865187e-3,0.,.1603921848e-2,-.6121588826e-6,-0.,-0.,-.3283516813e-5,-0.,-.8707719261e-5,.3368848215e-8,0.,.2188377514e-7,-.1771076019e-10},
			{1.414213562,0.,0.,0.,0.,0.,.6559570585,0.,0.,0.,0.,.1963948976,0.,0.,0.,.5080990387e-1,0.,0.,.1219720719e-1,0.,.2797118711e-2,.1452906103,0.,0.,0.,0.,.1065535599,0.,0.,0.,.4358691597e-1,0.,0.,.1429557258e-1,0.,.4156035932e-2,.9635061189e-2,0.,0.,0.,.9654244188e-2,0.,0.,.5006493970e-2,0.,.1988588036e-2,.5521216640e-3,0.,0.,.7013355149e-3,0.,.4404604930e-3,.2935681623e-4,0.,.4516122448e-4,.1491150883e-5},
			{1.414213562,0.,0.,0.,0.,0.,2.662853947,0.,0.,0.,0.,3.236488548,0.,0.,0.,3.399101902,0.,0.,3.312440982,0.,3.083689087,.6559570585,0.,0.,0.,0.,1.952889915,0.,0.,0.,3.242932261,0.,0.,4.317727482,0.,5.095712344,.1963948976,0.,0.,0.,.7988512884,0.,0.,1.681718474,0.,2.711667007,.5080990387e-1,0.,0.,.2620060705,0.,.6679814775,.1219720719e-1,0.,.7617090096e-1,.2797118711e-2},
			{1.414213562,0.,0.,0.,0.,0.,.1452906103,0.,0.,0.,0.,.9635061189e-2,0.,0.,0.,.5521216640e-3,0.,0.,.2935681623e-4,0.,.1491150883e-5,2.662853947,0.,0.,0.,0.,.4325535702,0.,0.,0.,.3919134934e-1,0.,0.,.2847067532e-2,0.,.1833317338e-3,3.236488548,0.,0.,0.,.7182903227,0.,0.,.8250448766e-1,0.,.7258566080e-2,3.399101902,0.,0.,.9563511098,0.,.1330333825,3.312440982,0.,1.128670157,3.083689087},
			{1.414213562,0.,0.,0.,0.,0.,.1452906103,0.,0.,0.,0.,.9635061189e-2,0.,0.,0.,.5521216640e-3,0.,0.,.2935681623e-4,0.,.1491150883e-5,.6559570585,0.,0.,0.,0.,.1065535599,0.,0.,0.,.9654244188e-2,0.,0.,.7013355149e-3,0.,.4516122448e-4,.1963948976,0.,0.,0.,.4358691597e-1,0.,0.,.5006493970e-2,0.,.4404604930e-3,.5080990387e-1,0.,0.,.1429557258e-1,0.,.1988588036e-2,.1219720719e-1,0.,.4156035932e-2,.2797118711e-2},
			{1.414213562,0.,0.,0.,0.,0.,.6559570585,0.,0.,0.,0.,.1963948976,0.,0.,0.,.5080990387e-1,0.,0.,.1219720719e-1,0.,.2797118711e-2,2.662853947,0.,0.,0.,0.,1.952889915,0.,0.,0.,.7988512884,0.,0.,.2620060705,0.,.7617090096e-1,3.236488548,0.,0.,0.,3.242932261,0.,0.,1.681718474,0.,.6679814775,3.399101902,0.,0.,4.317727482,0.,2.711667007,3.312440982,0.,5.095712344,3.083689087},
			{1.414213562,0.,0.,0.,0.,0.,2.662853947,0.,0.,0.,0.,3.236488548,0.,0.,0.,3.399101902,0.,0.,3.312440982,0.,3.083689087,.1452906103,0.,0.,0.,0.,.4325535702,0.,0.,0.,.7182903227,0.,0.,.9563511098,0.,1.128670157,.9635061189e-2,0.,0.,0.,.3919134934e-1,0.,0.,.8250448766e-1,0.,.1330333825,.5521216640e-3,0.,0.,.2847067532e-2,0.,.7258566080e-2,.2935681623e-4,0.,.1833317338e-3,.1491150883e-5},
			{1.414213562,0.,0.,0.,0.,0.,.2954311110,0.,0.,0.,0.,.3983747795e-1,0.,0.,0.,.4641847056e-2,0.,0.,.5018612597e-3,0.,.5183403142e-4,.4959665215e-1,0.,0.,0.,0.,.1638187454e-1,0.,0.,0.,.3018092809e-2,0.,0.,.4458192555e-3,0.,.5837372111e-4,.1122752692e-2,0.,0.,0.,.5066741239e-3,0.,0.,.1183382006e-3,0.,.2116980724e-4,.2196237131e-4,0.,0.,.1256467012e-4,0.,.3553964112e-5,.3986283756e-6,0.,.2761886212e-6,.6911871564e-8},
			{1.414213562,0.,0.,0.,0.,0.,3.119073853,0.,0.,0.,0.,4.440487963,0.,0.,0.,5.462596453,0.,0.,6.235358013,0.,6.799267692,.2954311110,0.,0.,0.,0.,1.030236404,0.,0.,0.,2.003896641,0.,0.,3.125150808,0.,4.320151396,.3983747795e-1,0.,0.,0.,.1898042300,0.,0.,.4680276955,0.,.8839606376,.4641847056e-2,0.,0.,.2803703724e-1,0.,.8372660778e-1,.5018612597e-3,0.,.3671053173e-2,.5183403142e-4},
			{1.414213562,0.,0.,0.,0.,0.,.4959665215e-1,0.,0.,0.,0.,.1122752692e-2,0.,0.,0.,.2196237131e-4,0.,0.,.3986283756e-6,0.,.6911871564e-8,3.119073853,0.,0.,0.,0.,.1729549620,0.,0.,0.,.5349314793e-2,0.,0.,.1326540524e-3,0.,.2915917367e-5,4.440487963,0.,0.,0.,.3364119790,0.,0.,.1319057787e-1,0.,.3961429206e-3,5.462596453,0.,0.,.5246469034,0.,.2491295222e-1,6.235358013,0.,.7252622967,6.799267692},
			{1.414213562,0.,0.,0.,0.,0.,.4959665215e-1,0.,0.,0.,0.,.1122752692e-2,0.,0.,0.,.2196237131e-4,0.,0.,.3986283756e-6,0.,.6911871564e-8,.2954311110,0.,0.,0.,0.,.1638187454e-1,0.,0.,0.,.5066741239e-3,0.,0.,.1256467012e-4,0.,.2761886212e-6,.3983747795e-1,0.,0.,0.,.3018092809e-2,0.,0.,.1183382006e-3,0.,.3553964112e-5,.4641847056e-2,0.,0.,.4458192555e-3,0.,.2116980724e-4,.5018612597e-3,0.,.5837372111e-4,.5183403142e-4},
			{1.414213562,0.,0.,0.,0.,0.,.2954311110,0.,0.,0.,0.,.3983747795e-1,0.,0.,0.,.4641847056e-2,0.,0.,.5018612597e-3,0.,.5183403142e-4,3.119073853,0.,0.,0.,0.,1.030236404,0.,0.,0.,.1898042300,0.,0.,.2803703724e-1,0.,.3671053173e-2,4.440487963,0.,0.,0.,2.003896641,0.,0.,.4680276955,0.,.8372660778e-1,5.462596453,0.,0.,3.125150808,0.,.8839606376,6.235358013,0.,4.320151396,6.799267692},
			{1.414213562,0.,0.,0.,0.,0.,3.119073853,0.,0.,0.,0.,4.440487963,0.,0.,0.,5.462596453,0.,0.,6.235358013,0.,6.799267692,.4959665215e-1,0.,0.,0.,0.,.1729549620,0.,0.,0.,.3364119790,0.,0.,.5246469034,0.,.7252622967,.1122752692e-2,0.,0.,0.,.5349314793e-2,0.,0.,.1319057787e-1,0.,.2491295222e-1,.2196237131e-4,0.,0.,.1326540524e-3,0.,.3961429206e-3,.3986283756e-6,0.,.2915917367e-5,.6911871564e-8}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 52; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_bottom_fluxes_to_dofs(
		const T i_input[52],
		T o_output[56]
	) {
		static const T matrix[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-.1148215401,-.1569113387e-1,-.1569113387e-1,-.1569113387e-1,-.1021689440,-.1021689440,-.1021689440,-.6587041484e-1,-.6587041484e-1,-.6587041484e-1,-.1032031823,-.1032031823,-.1032031823,-.7348534591e-1,-.7348534591e-1,-.7348534591e-1,-.3478299668e-1,-.3478299668e-1,-.3478299668e-1,-.8775204783e-2,-.8775204783e-2,-.8775204783e-2,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{-.1753927996,-.2396860291e-1,-.2396860291e-1,-.2396860291e-1,-.1560656399,-.1560656399,-.1560656399,-.1006187207,-.1006187207,-.1006187207,-.1576454650,-.1576454650,-.1576454650,-.1122507200,-.1122507200,-.1122507200,-.5313190506e-1,-.5313190506e-1,-.5313190506e-1,-.1340434672e-1,-.1340434672e-1,-.1340434672e-1,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{-.2198665100,-.3004623384e-1,-.3004623384e-1,-.3004623384e-1,-.1956386332,-.1956386332,-.1956386332,-.1261322415,-.1261322415,-.1261322415,-.1976190488,-.1976190488,-.1976190488,-.1407137244,-.1407137244,-.1407137244,-.6660436781e-1,-.6660436781e-1,-.6660436781e-1,-.1680323787e-1,-.1680323787e-1,-.1680323787e-1,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1},
			{-.5412739307e-1,-.1103717978e-1,-.1103717978e-1,-.1162547668e-3,-.5974166429e-1,-.5974166429e-1,-.2500537776e-1,-.4382549273e-1,-.4382549273e-1,-.5503848602e-2,-.3510914219e-1,-.3510914219e-1,-.7573305578e-1,-.1537719318e-1,-.1537719318e-1,-.7316958648e-1,-.3712176777e-2,-.3712176777e-2,-.4176623208e-1,-.2059615826e-3,-.2059615826e-3,-.1199809046e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1},
			{-.9375139486e-1,-.1911695615e-1,-.1911695615e-1,-.2013591627e-3,-.1034755978,-.1034755978,-.4331058473e-1,-.7590798006e-1,-.7590798006e-1,-.9532945415e-2,-.6081081808e-1,-.6081081808e-1,-.1311735004,-.2663407986e-1,-.2663407986e-1,-.1267334413,-.6429678784e-2,-.6429678784e-2,-.7234123601e-1,-.3567359253e-3,-.3567359253e-3,-.2078130226e-1,-.5831436180e-1,-.1179575144,-.2036700860e-1,-.2036700860e-1,-.5831436180e-1,-.1179575144,-.3099954510e-1,-.5894683929e-1,-.1843453093e-2,-.1843453093e-2,-.3099954510e-1,-.5894683929e-1,-.2931476389e-2,-.1144814092e-1,.6215215367e-4,.6215215367e-4,-.2931476389e-2,-.1144814092e-1,-.2168320548e-1,-.8802284926e-1,-.4802701816e-2,-.4802701816e-2,-.2168320548e-1,-.8802284926e-1,-.3505184516e-2,-.3700669622e-1,-.5884465473e-3,-.5884465473e-3,-.3505184516e-2,-.3700669622e-1},
			{-.1210325303,-.2467988427e-1,-.2467988427e-1,-.2599535613e-3,-.1335864224,-.1335864224,-.5591372447e-1,-.9799678087e-1,-.9799678087e-1,-.1230697961e-1,-.7850642855e-1,-.7850642855e-1,-.1693442608,-.3438444924e-1,-.3438444924e-1,-.1636121692,-.8300679616e-2,-.8300679616e-2,-.9339213407e-1,-.4605440992e-3,-.4605440992e-3,-.2682854586e-1,-.7528351732e-1,-.1522824962,-.2629369504e-1,-.2629369504e-1,-.7528351732e-1,-.1522824962,-.4002024064e-1,-.7610004229e-1,-.2379887709e-2,-.2379887709e-2,-.4002024064e-1,-.7610004229e-1,-.3784519743e-2,-.1477948638e-1,.8023808537e-4,.8023808537e-4,-.3784519743e-2,-.1477948638e-1,-.2799289790e-1,-.1136370098,-.6200261386e-2,-.6200261386e-2,-.2799289790e-1,-.1136370098,-.4525173753e-2,-.4777543936e-1,-.7596812259e-3,-.7596812259e-3,-.4525173753e-2,-.4777543936e-1},
			{-.1432076211,-.2920163288e-1,-.2920163288e-1,-.3075812017e-3,-.1580615866,-.1580615866,-.6615801099e-1,-.1159513548,-.1159513548,-.1456181465e-1,-.9289005896e-1,-.9289005896e-1,-.2003708316,-.4068422900e-1,-.4068422900e-1,-.1935885293,-.9821496573e-2,-.9821496573e-2,-.1105030633,-.5449231269e-3,-.5449231269e-3,-.3174396355e-1,-.8907665897e-1,-.1801830794,-.3111111953e-1,-.3111111953e-1,-.8907665897e-1,-.1801830794,-.4735258731e-1,-.9004278434e-1,-.2815921112e-2,-.2815921112e-2,-.4735258731e-1,-.9004278434e-1,-.4477904149e-2,-.1748732412e-1,.9493898295e-4,.9493898295e-4,-.4477904149e-2,-.1748732412e-1,-.3312164346e-1,-.1344571232,-.7336248207e-2,-.7336248207e-2,-.3312164346e-1,-.1344571232,-.5354257791e-2,-.5652866219e-1,-.8988669485e-3,-.8988669485e-3,-.5354257791e-2,-.5652866219e-1},
			{-.1623821792,-.3311153935e-1,-.3311153935e-1,-.3487643005e-3,-.1792249928,-.1792249928,-.7501613328e-1,-.1314764782,-.1314764782,-.1651154581e-1,-.1053274266,-.1053274266,-.2271991674,-.4613157954e-1,-.4613157954e-1,-.2195087594,-.1113653033e-1,-.1113653033e-1,-.1252986963,-.6178847477e-3,-.6178847477e-3,-.3599427137e-1,-.1010034375,-.2043084081,-.3527669370e-1,-.3527669370e-1,-.1010034375,-.2043084081,-.5369278714e-1,-.1020989206,-.3192954419e-2,-.3192954419e-2,-.5369278714e-1,-.1020989206,-.5077466048e-2,-.1982876173e-1,.1076506880e-3,.1076506880e-3,-.5077466048e-2,-.1982876173e-1,-.3755641357e-1,-.1524600472,-.8318523561e-2,-.8318523561e-2,-.3755641357e-1,-.1524600472,-.6071157673e-2,-.6409747809e-1,-.1019219318e-2,-.1019219318e-2,-.6071157673e-2,-.6409747809e-1},
			{-.2852764098e-1,-.8679943781e-2,-.8679943781e-2,-.9629909754e-6,-.3905626741e-1,-.3905626741e-1,-.6842313070e-2,-.3260004675e-1,-.3260004675e-1,-.5141591464e-3,-.1335372273e-1,-.1335372273e-1,-.6213451013e-1,-.3597562967e-2,-.3597562967e-2,-.8145457175e-1,-.4429404170e-3,-.4429404170e-3,-.5607103930e-1,-.5404682501e-5,-.5404682501e-5,-.1834095772e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1},
			{-.4941132360e-1,-.1503410363e-1,-.1503410363e-1,-.1667949296e-5,-.6764743950e-1,-.6764743950e-1,-.1185123388e-1,-.5646493729e-1,-.5646493729e-1,-.8905497646e-3,-.2312932623e-1,-.2312932623e-1,-.1076201284,-.6231161842e-2,-.6231161842e-2,-.1410834568,-.7671953069e-3,-.7671953069e-3,-.9711788894e-1,-.9361184690e-5,-.9361184690e-5,-.3176747063e-1,-.2734334585e-1,-.1118798374,-.3335455075e-2,-.3335455075e-2,-.2734334585e-1,-.1118798374,-.1655335563e-1,-.5985445076e-1,-.5853822990e-4,-.5853822990e-4,-.1655335563e-1,-.5985445076e-1,-.9490228526e-3,-.1447350792e-1,-.4265954647e-6,-.4265954647e-6,-.9490228526e-3,-.1447350792e-1,-.6491996486e-2,-.1069848175,-.3184949515e-3,-.3184949515e-3,-.6491996486e-2,-.1069848175,-.4726574345e-3,-.5268480225e-1,-.1332105931e-4,-.1332105931e-4,-.4726574345e-3,-.5268480225e-1},
			{-.6378974448e-1,-.1940894434e-1,-.1940894434e-1,-.2153313281e-5,-.8733246888e-1,-.8733246888e-1,-.1529987715e-1,-.7289592064e-1,-.7289592064e-1,-.1149694802e-2,-.2985983179e-1,-.2985983179e-1,-.1389369884,-.8044395345e-2,-.8044395345e-2,-.1821379596,-.9904448821e-3,-.9904448821e-3,-.1253786555,-.1208523748e-4,-.1208523748e-4,-.4101162825e-1,-.3530010772e-1,-.1444362491,-.4306053985e-2,-.4306053985e-2,-.3530010772e-1,-.1444362491,-.2137029023e-1,-.7727176368e-1,-.7557252986e-4,-.7557252986e-4,-.2137029023e-1,-.7727176368e-1,-.1225183234e-2,-.1868521838e-1,-.5507323768e-6,-.5507323768e-6,-.1225183234e-2,-.1868521838e-1,-.8381131425e-2,-.1381168056,-.4111752145e-3,-.4111752145e-3,-.8381131425e-2,-.1381168056,-.6101981241e-3,-.6801578726e-1,-.1719741363e-4,-.1719741363e-4,-.6101981241e-3,-.6801578726e-1},
			{-.7547704355e-1,-.2296497265e-1,-.2296497265e-1,-.2547834636e-5,-.1033331707,-.1033331707,-.1810305878e-1,-.8625161644e-1,-.8625161644e-1,-.1360337236e-2,-.3533062943e-1,-.3533062943e-1,-.1643924617,-.9518256939e-2,-.9518256939e-2,-.2155085401,-.1171910189e-2,-.1171910189e-2,-.1483500258,-.1429944582e-4,-.1429944582e-4,-.4852561294e-1,-.4176765072e-1,-.1708992746,-.5094991785e-2,-.5094991785e-2,-.4176765072e-1,-.1708992746,-.2528566840e-1,-.9142918380e-1,-.8941862322e-4,-.8941862322e-4,-.2528566840e-1,-.9142918380e-1,-.1449656353e-2,-.2210864854e-1,-.6516353360e-6,-.6516353360e-6,-.1449656353e-2,-.2210864854e-1,-.9916688441e-2,-.1634220082,-.4865090747e-3,-.4865090747e-3,-.9916688441e-2,-.1634220082,-.7219961572e-3,-.8047736478e-1,-.2034825423e-4,-.2034825423e-4,-.7219961572e-3,-.8047736478e-1},
			{-.1299205923e-1,-.5898466126e-2,-.5898466126e-2,-.6892821735e-8,-.2206315093e-1,-.2206315093e-1,-.1617841028e-2,-.2095429338e-1,-.2095429338e-1,-.4150419878e-4,-.4388821038e-2,-.4388821038e-2,-.4404977136e-1,-.7272824702e-3,-.7272824702e-3,-.7835445799e-1,-.4566940490e-4,-.4566940490e-4,-.6504520647e-1,-.1225511905e-6,-.1225511905e-6,-.2422675624e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1},
			{-.2250290667e-1,-.1021644301e-1,-.1021644301e-1,-.1193871745e-7,-.3821449838e-1,-.3821449838e-1,-.2802182859e-2,-.3629390077e-1,-.3629390077e-1,-.7188738101e-4,-.7601661022e-2,-.7601661022e-2,-.7629644205e-1,-.1259690190e-2,-.1259690190e-2,-.1357139022,-.7910172963e-4,-.7910172963e-4,-.1126616024,-.2122648885e-6,-.2122648885e-6,-.4196197271e-1,-.1107876030e-1,-.9169410646e-1,-.4720047033e-3,-.4720047033e-3,-.1107876030e-1,-.9169410646e-1,-.7638008748e-2,-.5251649806e-1,-.1606240095e-5,-.1606240095e-5,-.7638008748e-2,-.5251649806e-1,-.2654790910e-3,-.1581160736e-1,.2530111540e-8,.2530111540e-8,-.2654790910e-3,-.1581160736e-1,-.1679563581e-2,-.1123601370,-.1825084025e-4,-.1825084025e-4,-.1679563581e-2,-.1123601370,-.5507385590e-4,-.6481175407e-1,-.2605756842e-6,-.2605756842e-6,-.5507385590e-4,-.6481175407e-1},
			{-.2905112760e-1,-.1318937122e-1,-.1318937122e-1,-.1541281795e-7,-.4933470527e-1,-.4933470527e-1,-.3617602516e-2,-.4685522442e-1,-.4685522442e-1,-.9280620982e-4,-.9813702180e-2,-.9813702180e-2,-.9849828315e-1,-.1626253042e-2,-.1626253042e-2,-.1752058944,-.1021198938e-3,-.1021198938e-3,-.1454455033,-.2740327927e-6,-.2740327927e-6,-.5417267383e-1,-.1430261805e-1,-.1183765824,-.6093554517e-3,-.6093554517e-3,-.1430261805e-1,-.1183765824,-.9860626894e-2,-.6779850746e-1,-.2073647046e-5,-.2073647046e-5,-.9860626894e-2,-.6779850746e-1,-.3427320328e-3,-.2041269733e-1,.3266359953e-8,.3266359953e-8,-.3427320328e-3,-.2041269733e-1,-.2168307259e-2,-.1450563132,-.2356173345e-4,-.2356173345e-4,-.2168307259e-2,-.1450563132,-.7110004225e-4,-.8367161471e-1,-.3364017618e-6,-.3364017618e-6,-.7110004225e-4,-.8367161471e-1},
			{-.5490147066e-2,-.3719247860e-2,-.3719247860e-2,-.4577893991e-10,-.1156480173e-1,-.1156480173e-1,-.3549462592e-3,-.1249745760e-1,-.1249745760e-1,-.3108711060e-5,-.1338404058e-2,-.1338404058e-2,-.2897665788e-1,-.1364242863e-3,-.1364242863e-3,-.6993680883e-1,-.4369173091e-5,-.4369173091e-5,-.7001415346e-1,-.2578450140e-8,-.2578450140e-8,-.2969356790e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1},
			{-.9509213660e-2,-.6441926259e-2,-.6441926259e-2,-.7929144985e-10,-.2003082417e-1,-.2003082417e-1,-.6147849549e-3,-.2164623154e-1,-.2164623154e-1,-.5384445502e-5,-.2318183830e-2,-.2318183830e-2,-.5018904368e-1,-.2362937952e-3,-.2362937952e-3,-.1211341062,-.7567629781e-5,-.7567629781e-5,-.1212680711,-.4466006648e-8,-.4466006648e-8,-.5143076827e-1,-.4165091360e-2,-.6973083633e-1,-.6197712293e-4,-.6197712293e-4,-.4165091360e-2,-.6973083633e-1,-.3270153304e-2,-.4275519802e-1,-.4089546907e-7,-.4089546907e-7,-.3270153304e-2,-.4275519802e-1,-.6890929850e-4,-.1602773282e-1,-.1392377369e-10,-.1392377369e-10,-.6890929850e-4,-.1602773282e-1,-.4031888160e-3,-.1094954884,-.9704139476e-6,-.9704139476e-6,-.4031888160e-3,-.1094954884,-.5954404436e-5,-.7398029375e-1,-.4729583171e-8,-.4729583171e-8,-.5954404436e-5,-.7398029375e-1},
			{-.2216299428e-2,-.2240317772e-2,-.2240317772e-2,-.2904510069e-12,-.5790916683e-2,-.5790916683e-2,-.7439227355e-4,-.7120472184e-2,-.7120472184e-2,-.2224370628e-6,-.3899106406e-3,-.3899106406e-3,-.1820922229e-1,-.2444661488e-4,-.2444661488e-4,-.5963296216e-1,-.3993113157e-6,-.3993113157e-6,-.7199376027e-1,-.5182489262e-10,-.5182489262e-10,-.3476706016e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1},
			{-.5412739307e-1,-.1162547668e-3,-.1103717978e-1,-.1103717978e-1,-.2500537776e-1,-.5974166429e-1,-.5974166429e-1,-.5503848602e-2,-.4382549273e-1,-.4382549273e-1,-.7573305578e-1,-.3510914219e-1,-.3510914219e-1,-.7316958648e-1,-.1537719318e-1,-.1537719318e-1,-.4176623208e-1,-.3712176777e-2,-.3712176777e-2,-.1199809046e-1,-.2059615826e-3,-.2059615826e-3,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3},
			{-.9375139486e-1,-.2013591627e-3,-.1911695615e-1,-.1911695615e-1,-.4331058473e-1,-.1034755978,-.1034755978,-.9532945415e-2,-.7590798006e-1,-.7590798006e-1,-.1311735004,-.6081081808e-1,-.6081081808e-1,-.1267334413,-.2663407986e-1,-.2663407986e-1,-.7234123601e-1,-.6429678784e-2,-.6429678784e-2,-.2078130226e-1,-.3567359253e-3,-.3567359253e-3,-.2036700860e-1,-.5831436180e-1,-.1179575144,-.5831436180e-1,-.1179575144,-.2036700860e-1,-.1843453093e-2,-.3099954510e-1,-.5894683929e-1,-.3099954510e-1,-.5894683929e-1,-.1843453093e-2,.6215215367e-4,-.2931476389e-2,-.1144814092e-1,-.2931476389e-2,-.1144814092e-1,.6215215367e-4,-.4802701816e-2,-.2168320548e-1,-.8802284926e-1,-.2168320548e-1,-.8802284926e-1,-.4802701816e-2,-.5884465473e-3,-.3505184516e-2,-.3700669622e-1,-.3505184516e-2,-.3700669622e-1,-.5884465473e-3},
			{-.1210325303,-.2599535613e-3,-.2467988427e-1,-.2467988427e-1,-.5591372447e-1,-.1335864224,-.1335864224,-.1230697961e-1,-.9799678087e-1,-.9799678087e-1,-.1693442608,-.7850642855e-1,-.7850642855e-1,-.1636121692,-.3438444924e-1,-.3438444924e-1,-.9339213407e-1,-.8300679616e-2,-.8300679616e-2,-.2682854586e-1,-.4605440992e-3,-.4605440992e-3,-.2629369504e-1,-.7528351732e-1,-.1522824962,-.7528351732e-1,-.1522824962,-.2629369504e-1,-.2379887709e-2,-.4002024064e-1,-.7610004229e-1,-.4002024064e-1,-.7610004229e-1,-.2379887709e-2,.8023808537e-4,-.3784519743e-2,-.1477948638e-1,-.3784519743e-2,-.1477948638e-1,.8023808537e-4,-.6200261386e-2,-.2799289790e-1,-.1136370098,-.2799289790e-1,-.1136370098,-.6200261386e-2,-.7596812259e-3,-.4525173753e-2,-.4777543936e-1,-.4525173753e-2,-.4777543936e-1,-.7596812259e-3},
			{-.1432076211,-.3075812017e-3,-.2920163288e-1,-.2920163288e-1,-.6615801099e-1,-.1580615866,-.1580615866,-.1456181465e-1,-.1159513548,-.1159513548,-.2003708316,-.9289005896e-1,-.9289005896e-1,-.1935885293,-.4068422900e-1,-.4068422900e-1,-.1105030633,-.9821496573e-2,-.9821496573e-2,-.3174396355e-1,-.5449231269e-3,-.5449231269e-3,-.3111111953e-1,-.8907665897e-1,-.1801830794,-.8907665897e-1,-.1801830794,-.3111111953e-1,-.2815921112e-2,-.4735258731e-1,-.9004278434e-1,-.4735258731e-1,-.9004278434e-1,-.2815921112e-2,.9493898295e-4,-.4477904149e-2,-.1748732412e-1,-.4477904149e-2,-.1748732412e-1,.9493898295e-4,-.7336248207e-2,-.3312164346e-1,-.1344571232,-.3312164346e-1,-.1344571232,-.7336248207e-2,-.8988669485e-3,-.5354257791e-2,-.5652866219e-1,-.5354257791e-2,-.5652866219e-1,-.8988669485e-3},
			{-.1623821792,-.3487643005e-3,-.3311153935e-1,-.3311153935e-1,-.7501613328e-1,-.1792249928,-.1792249928,-.1651154581e-1,-.1314764782,-.1314764782,-.2271991674,-.1053274266,-.1053274266,-.2195087594,-.4613157954e-1,-.4613157954e-1,-.1252986963,-.1113653033e-1,-.1113653033e-1,-.3599427137e-1,-.6178847477e-3,-.6178847477e-3,-.3527669370e-1,-.1010034375,-.2043084081,-.1010034375,-.2043084081,-.3527669370e-1,-.3192954419e-2,-.5369278714e-1,-.1020989206,-.5369278714e-1,-.1020989206,-.3192954419e-2,.1076506880e-3,-.5077466048e-2,-.1982876173e-1,-.5077466048e-2,-.1982876173e-1,.1076506880e-3,-.8318523561e-2,-.3755641357e-1,-.1524600472,-.3755641357e-1,-.1524600472,-.8318523561e-2,-.1019219318e-2,-.6071157673e-2,-.6409747809e-1,-.6071157673e-2,-.6409747809e-1,-.1019219318e-2},
			{-.6987816394e-1,-.2239469694e-3,-.2126143325e-1,-.2239469694e-3,-.4004261797e-1,-.9566792642e-1,-.4004261797e-1,-.1002844321e-1,-.7985348011e-1,-.1002844321e-1,-.7055750928e-1,-.3270980684e-1,-.7055750928e-1,-.4193122589e-1,-.8812193584e-2,-.4193122589e-1,-.1220724282e-1,-.1084978008e-2,-.1220724282e-1,-.7712083501e-3,-.1323871434e-4,-.7712083501e-3,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,.2845518738e-4,-.5241314676e-2,.1111245502e-3,.2845518738e-4,-.5241314676e-2,.1111245502e-3,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2},
			{-.1210325303,-.3878875292e-3,-.3682588264e-1,-.3878875292e-3,-.6935584881e-1,-.1657017093,-.6935584881e-1,-.1736977316e-1,-.1383102847,-.1736977316e-1,-.1222091909,-.5665504737e-1,-.1222091909,-.7262701368e-1,-.1526316702e-1,-.7262701368e-1,-.2114356478e-1,-.1879237035e-2,-.2114356478e-1,-.1335772046e-2,-.2293012587e-4,-.1335772046e-2,-.2339262725e-1,-.1354806795,-.4731829487e-1,-.2339262725e-1,-.1354806795,-.4731829487e-1,-.2411228903e-2,-.7710221821e-1,-.4585045432e-2,-.2411228903e-2,-.7710221821e-1,-.4585045432e-2,.4928583028e-4,-.9078223319e-2,.1924733670e-3,.4928583028e-4,-.9078223319e-2,.1924733670e-3,-.3522216439e-2,-.6455439842e-1,-.1429841759e-1,-.3522216439e-2,-.6455439842e-1,-.1429841759e-1,-.1943650851e-3,-.1222338663e-1,-.2052048795e-2,-.1943650851e-3,-.1222338663e-1,-.2052048795e-2},
			{-.1562523248,-.5007606471e-3,-.4754201006e-1,-.5007606471e-3,-.8953801576e-1,-.2139199867,-.8953801576e-1,-.2242428073e-1,-.1785578098,-.2242428073e-1,-.1577713871,-.7314135164e-1,-.1577713871,-.9376107150e-1,-.1970466389e-1,-.9376107150e-1,-.2729622476e-1,-.2426084580e-2,-.2729622476e-1,-.1724474296e-2,-.2960266522e-4,-.1724474296e-2,-.3019975193e-1,-.1749048052,-.6108765601e-1,-.3019975193e-1,-.1749048052,-.6108765601e-1,-.3112883129e-2,-.9953853571e-1,-.5919268200e-2,-.3112883129e-2,-.9953853571e-1,-.5919268200e-2,.6362773332e-4,-.1171993591e-1,.2484820485e-3,.6362773332e-4,-.1171993591e-1,.2484820485e-3,-.4547161871e-2,-.8333937002e-1,-.1845917774e-1,-.4547161871e-2,-.8333937002e-1,-.1845917774e-1,-.2509242459e-3,-.1578032429e-1,-.2649183603e-2,-.2509242459e-3,-.1578032429e-1,-.2649183603e-2},
			{-.1848802439,-.5925079880e-3,-.5625246492e-1,-.5925079880e-3,-.1059428090,-.2531135418,-.1059428090,-.2653276678e-1,-.2112724497,-.2653276678e-1,-.1866776227,-.8654201436e-1,-.1866776227,-.1109395959,-.2331487273e-1,-.1109395959,-.3229732869e-1,-.2870581987e-2,-.3229732869e-1,-.2040425504e-2,-.3502634584e-4,-.2040425504e-2,-.3573282836e-1,-.2069501564,-.7227988933e-1,-.3573282836e-1,-.2069501564,-.7227988933e-1,-.3683212988e-2,-.1177755837,-.7003772586e-2,-.3683212988e-2,-.1177755837,-.7003772586e-2,.7528534932e-4,-.1386721518e-1,.2940079245e-3,.7528534932e-4,-.1386721518e-1,.2940079245e-3,-.5380274482e-2,-.9860847240e-1,-.2184119364e-1,-.5380274482e-2,-.9860847240e-1,-.2184119364e-1,-.2968975716e-3,-.1867153150e-1,-.3134556311e-2,-.2968975716e-3,-.1867153150e-1,-.3134556311e-2},
			{-.5031802903e-1,-.2406231299e-3,-.2284466108e-1,-.2534486742e-5,-.3576590913e-1,-.8545021612e-1,-.1497012310e-1,-.1019197433e-1,-.8115562931e-1,-.1279964701e-2,-.3666559727e-1,-.1699783079e-1,-.7909044629e-1,-.1340300809e-1,-.2816752894e-2,-.6377578460e-1,-.1990066688e-2,-.1768768448e-3,-.2239052508e-1,-.2764961419e-4,-.4746387201e-6,-.1610701220e-2,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2},
			{-.8715338278e-1,-.4167714865e-3,-.3956811365e-1,-.4389859808e-5,-.6194837174e-1,-.1480041159,-.2592901380e-1,-.1765301736e-1,-.1405656732,-.2216963893e-2,-.6350667735e-1,-.2944110655e-1,-.1369886713,-.2321469098e-1,-.4878759123e-2,-.1104628991,-.3446896614e-2,-.3063596816e-3,-.3878152703e-1,-.4789053657e-4,-.8220983782e-6,-.2789816347e-2,-.1498609619e-1,-.1755646063,-.1058742437e-1,-.5234078544e-2,-.8679343590e-1,-.6131809964e-1,-.1759148700e-2,-.1069636429,-.1989227818e-3,-.1046114740e-3,-.5625109532e-1,-.6360817804e-2,.2179945987e-4,-.1568097578e-1,-.1804945451e-5,-.4621846472e-6,-.4015360274e-2,.3324626528e-3,-.1440801715e-2,-.1071978198,-.1295502609e-2,-.3191290614e-3,-.2640669295e-1,-.2374368331e-1,-.3580859223e-4,-.2377551727e-1,-.6346770766e-4,-.6011507348e-5,-.2251959335e-2,-.3991407866e-2},
			{-.1125145334,-.5380496754e-3,-.5108221507e-1,-.5667284643e-5,-.7997500406e-1,-.1910724920,-.3347421288e-1,-.2278994742e-1,-.1814695038,-.2862088080e-2,-.8198676790e-1,-.3800830512e-1,-.1768516141,-.2997003719e-1,-.6298450945e-2,-.1426069896,-.4449924395e-2,-.3955086483e-3,-.5006673611e-1,-.6182641688e-4,-.1061324443e-5,-.3601637417e-2,-.1934696698e-1,-.2266529321,-.1366830609e-1,-.6757166343e-2,-.1120498440,-.7916132626e-1,-.2271051205e-2,-.1380894693,-.2568082070e-3,-.1350528323e-3,-.7261985182e-1,-.8211780472e-2,.2814298168e-4,-.2024405269e-1,-.2330174558e-5,-.5966778137e-6,-.5183807822e-2,.4292074393e-3,-.1860067016e-2,-.1383917903,-.1672486677e-2,-.4119938467e-3,-.3409089400e-1,-.3065296333e-1,-.4622869379e-4,-.3069406081e-1,-.8193645830e-4,-.7760822615e-5,-.2907267000e-2,-.5152885397e-2},
			{-.2905112760e-1,-.2072940277e-3,-.1968040980e-1,-.2299810724e-7,-.2561375849e-1,-.6119517866e-1,-.4487304244e-2,-.8305013821e-2,-.6613032970e-1,-.1309844384e-3,-.1527676939e-1,-.7082168586e-2,-.7108239413e-1,-.3434980188e-2,-.7218894686e-3,-.7777343793e-1,-.2601206999e-3,-.2311949087e-4,-.3292821660e-1,-.7948106270e-6,-.1364387568e-7,-.2697214518e-2,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,.4463401949e-5,-.1253839410e-1,.7835272927e-8,.2006344762e-8,-.8221381159e-3,.2658347174e-3,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2},
			{-.5031802903e-1,-.3590437882e-3,-.3408746970e-1,-.3983389024e-7,-.4436433109e-1,-.1059931586,-.7772238941e-2,-.1438470590e-1,-.1145410910,-.2268717024e-3,-.2646014076e-1,-.1226667582e-1,-.1231183182,-.5949560210e-2,-.1250349237e-2,-.1347075460,-.4505422684e-3,-.4004413284e-4,-.5703334418e-1,-.1376652389e-5,-.2363188589e-7,-.4671712585e-2,-.7697598727e-2,-.1824121677,-.1899366584e-2,-.9389851147e-3,-.4458139289e-1,-.6370969475e-1,-.1029020197e-2,-.1189769237,-.6919626319e-5,-.3638961323e-5,-.3290427534e-1,-.7075212788e-2,.7730838953e-5,-.2171713563e-1,.1357109081e-7,.3475091067e-8,-.1423984988e-2,.4604392371e-3,-.4725520274e-3,-.1427260688,-.9411218238e-4,-.2318322805e-4,-.8660828315e-2,-.3161298039e-1,-.5289486092e-5,-.3707879714e-1,-.1573892910e-5,-.1490753193e-6,-.3326494240e-3,-.6224747955e-2},
			{-.1486738854e-1,-.1582954870e-3,-.1502850849e-1,-.1849807308e-9,-.1625959323e-1,-.3884665007e-1,-.1192279250e-2,-.5998666568e-2,-.4776557949e-1,-.1188156743e-4,-.5642045787e-2,-.2615600094e-2,-.5662815244e-1,-.7803303852e-3,-.1639928781e-3,-.8406962479e-1,-.3013800652e-4,-.2678661737e-5,-.4292442302e-1,-.2025212407e-7,-.3476519485e-9,-.4003578188e-2,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2},
			{-.2852764098e-1,-.9629909754e-6,-.8679943781e-2,-.8679943781e-2,-.6842313070e-2,-.3905626741e-1,-.3905626741e-1,-.5141591464e-3,-.3260004675e-1,-.3260004675e-1,-.6213451013e-1,-.1335372273e-1,-.1335372273e-1,-.8145457175e-1,-.3597562967e-2,-.3597562967e-2,-.5607103930e-1,-.4429404170e-3,-.4429404170e-3,-.1834095772e-1,-.5404682501e-5,-.5404682501e-5,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5},
			{-.4941132360e-1,-.1667949296e-5,-.1503410363e-1,-.1503410363e-1,-.1185123388e-1,-.6764743950e-1,-.6764743950e-1,-.8905497646e-3,-.5646493729e-1,-.5646493729e-1,-.1076201284,-.2312932623e-1,-.2312932623e-1,-.1410834568,-.6231161842e-2,-.6231161842e-2,-.9711788894e-1,-.7671953069e-3,-.7671953069e-3,-.3176747063e-1,-.9361184690e-5,-.9361184690e-5,-.3335455075e-2,-.2734334585e-1,-.1118798374,-.2734334585e-1,-.1118798374,-.3335455075e-2,-.5853822990e-4,-.1655335563e-1,-.5985445076e-1,-.1655335563e-1,-.5985445076e-1,-.5853822990e-4,-.4265954647e-6,-.9490228526e-3,-.1447350792e-1,-.9490228526e-3,-.1447350792e-1,-.4265954647e-6,-.3184949515e-3,-.6491996486e-2,-.1069848175,-.6491996486e-2,-.1069848175,-.3184949515e-3,-.1332105931e-4,-.4726574345e-3,-.5268480225e-1,-.4726574345e-3,-.5268480225e-1,-.1332105931e-4},
			{-.6378974448e-1,-.2153313281e-5,-.1940894434e-1,-.1940894434e-1,-.1529987715e-1,-.8733246888e-1,-.8733246888e-1,-.1149694802e-2,-.7289592064e-1,-.7289592064e-1,-.1389369884,-.2985983179e-1,-.2985983179e-1,-.1821379596,-.8044395345e-2,-.8044395345e-2,-.1253786555,-.9904448821e-3,-.9904448821e-3,-.4101162825e-1,-.1208523748e-4,-.1208523748e-4,-.4306053985e-2,-.3530010772e-1,-.1444362491,-.3530010772e-1,-.1444362491,-.4306053985e-2,-.7557252986e-4,-.2137029023e-1,-.7727176368e-1,-.2137029023e-1,-.7727176368e-1,-.7557252986e-4,-.5507323768e-6,-.1225183234e-2,-.1868521838e-1,-.1225183234e-2,-.1868521838e-1,-.5507323768e-6,-.4111752145e-3,-.8381131425e-2,-.1381168056,-.8381131425e-2,-.1381168056,-.4111752145e-3,-.1719741363e-4,-.6101981241e-3,-.6801578726e-1,-.6101981241e-3,-.6801578726e-1,-.1719741363e-4},
			{-.7547704355e-1,-.2547834636e-5,-.2296497265e-1,-.2296497265e-1,-.1810305878e-1,-.1033331707,-.1033331707,-.1360337236e-2,-.8625161644e-1,-.8625161644e-1,-.1643924617,-.3533062943e-1,-.3533062943e-1,-.2155085401,-.9518256939e-2,-.9518256939e-2,-.1483500258,-.1171910189e-2,-.1171910189e-2,-.4852561294e-1,-.1429944582e-4,-.1429944582e-4,-.5094991785e-2,-.4176765072e-1,-.1708992746,-.4176765072e-1,-.1708992746,-.5094991785e-2,-.8941862322e-4,-.2528566840e-1,-.9142918380e-1,-.2528566840e-1,-.9142918380e-1,-.8941862322e-4,-.6516353360e-6,-.1449656353e-2,-.2210864854e-1,-.1449656353e-2,-.2210864854e-1,-.6516353360e-6,-.4865090747e-3,-.9916688441e-2,-.1634220082,-.9916688441e-2,-.1634220082,-.4865090747e-3,-.2034825423e-4,-.7219961572e-3,-.8047736478e-1,-.7219961572e-3,-.8047736478e-1,-.2034825423e-4},
			{-.5031802903e-1,-.2534486742e-5,-.2284466108e-1,-.2406231299e-3,-.1497012310e-1,-.8545021612e-1,-.3576590913e-1,-.1279964701e-2,-.8115562931e-1,-.1019197433e-1,-.7909044629e-1,-.1699783079e-1,-.3666559727e-1,-.6377578460e-1,-.2816752894e-2,-.1340300809e-1,-.2239052508e-1,-.1768768448e-3,-.1990066688e-2,-.1610701220e-2,-.4746387201e-6,-.2764961419e-4,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4},
			{-.8715338278e-1,-.4389859808e-5,-.3956811365e-1,-.4167714865e-3,-.2592901380e-1,-.1480041159,-.6194837174e-1,-.2216963893e-2,-.1405656732,-.1765301736e-1,-.1369886713,-.2944110655e-1,-.6350667735e-1,-.1104628991,-.4878759123e-2,-.2321469098e-1,-.3878152703e-1,-.3063596816e-3,-.3446896614e-2,-.2789816347e-2,-.8220983782e-6,-.4789053657e-4,-.5234078544e-2,-.8679343590e-1,-.6131809964e-1,-.1498609619e-1,-.1755646063,-.1058742437e-1,-.1046114740e-3,-.5625109532e-1,-.6360817804e-2,-.1759148700e-2,-.1069636429,-.1989227818e-3,-.4621846472e-6,-.4015360274e-2,.3324626528e-3,.2179945987e-4,-.1568097578e-1,-.1804945451e-5,-.3191290614e-3,-.2640669295e-1,-.2374368331e-1,-.1440801715e-2,-.1071978198,-.1295502609e-2,-.6011507348e-5,-.2251959335e-2,-.3991407866e-2,-.3580859223e-4,-.2377551727e-1,-.6346770766e-4},
			{-.1125145334,-.5667284643e-5,-.5108221507e-1,-.5380496754e-3,-.3347421288e-1,-.1910724920,-.7997500406e-1,-.2862088080e-2,-.1814695038,-.2278994742e-1,-.1768516141,-.3800830512e-1,-.8198676790e-1,-.1426069896,-.6298450945e-2,-.2997003719e-1,-.5006673611e-1,-.3955086483e-3,-.4449924395e-2,-.3601637417e-2,-.1061324443e-5,-.6182641688e-4,-.6757166343e-2,-.1120498440,-.7916132626e-1,-.1934696698e-1,-.2266529321,-.1366830609e-1,-.1350528323e-3,-.7261985182e-1,-.8211780472e-2,-.2271051205e-2,-.1380894693,-.2568082070e-3,-.5966778137e-6,-.5183807822e-2,.4292074393e-3,.2814298168e-4,-.2024405269e-1,-.2330174558e-5,-.4119938467e-3,-.3409089400e-1,-.3065296333e-1,-.1860067016e-2,-.1383917903,-.1672486677e-2,-.7760822615e-5,-.2907267000e-2,-.5152885397e-2,-.4622869379e-4,-.3069406081e-1,-.8193645830e-4},
			{-.4593386590e-1,-.3452307302e-5,-.3111746013e-1,-.3452307302e-5,-.1695115978e-1,-.9675807323e-1,-.1695115978e-1,-.1649111558e-2,-.1045612321,-.1649111558e-2,-.5210348698e-1,-.1119789175e-1,-.5210348698e-1,-.2584328822e-1,-.1141407470e-2,-.2584328822e-1,-.4627448202e-2,-.3655512474e-4,-.4627448202e-2,-.7320817509e-4,-.2157286163e-7,-.7320817509e-4,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4},
			{-.7955978954e-1,-.5979571650e-5,-.5389702195e-1,-.5979571650e-5,-.2936026999e-1,-.1675898989,-.2936026999e-1,-.2856345005e-2,-.1811053665,-.2856345005e-2,-.9024588671e-1,-.1939531744e-1,-.9024588671e-1,-.4476188824e-1,-.1976975730e-2,-.4476188824e-1,-.8014975395e-2,-.6331533334e-4,-.8014975395e-2,-.1268002788e-3,-.3736529241e-7,-.1268002788e-3,-.4250861857e-2,-.1425849616,-.1739310676e-1,-.4250861857e-2,-.1425849616,-.1739310676e-1,-.9675438873e-4,-.9892989258e-1,-.3498493552e-3,-.9675438873e-4,-.9892989258e-1,-.3498493552e-3,-.2591590698e-6,-.8792734945e-2,-.3952424159e-5,-.2591590698e-6,-.8792734945e-2,-.3952424159e-5,-.1654938171e-3,-.5559060120e-1,-.2727254377e-2,-.1654938171e-3,-.5559060120e-1,-.2727254377e-2,-.1404040445e-5,-.5552981292e-2,-.1565014910e-3,-.1404040445e-5,-.5552981292e-2,-.1565014910e-3},
			{-.3211722999e-1,-.3601841323e-5,-.3246528886e-1,-.3793824429e-7,-.1470175463e-1,-.8391835536e-1,-.6153543472e-2,-.1627413015e-2,-.1031854450,-.2043795582e-3,-.2629088823e-1,-.5650341992e-2,-.5671141990e-1,-.8021127855e-2,-.3542651063e-3,-.3816708301e-1,-.7325116912e-3,-.5786570712e-5,-.8241593859e-2,-.2548589453e-5,-.7510140431e-9,-.1484655847e-3,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3},
			{-.1299205923e-1,-.6892821735e-8,-.5898466126e-2,-.5898466126e-2,-.1617841028e-2,-.2206315093e-1,-.2206315093e-1,-.4150419878e-4,-.2095429338e-1,-.2095429338e-1,-.4404977136e-1,-.4388821038e-2,-.4388821038e-2,-.7835445799e-1,-.7272824702e-3,-.7272824702e-3,-.6504520647e-1,-.4566940490e-4,-.4566940490e-4,-.2422675624e-1,-.1225511905e-6,-.1225511905e-6,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6},
			{-.2250290667e-1,-.1193871745e-7,-.1021644301e-1,-.1021644301e-1,-.2802182859e-2,-.3821449838e-1,-.3821449838e-1,-.7188738101e-4,-.3629390077e-1,-.3629390077e-1,-.7629644205e-1,-.7601661022e-2,-.7601661022e-2,-.1357139022,-.1259690190e-2,-.1259690190e-2,-.1126616024,-.7910172963e-4,-.7910172963e-4,-.4196197271e-1,-.2122648885e-6,-.2122648885e-6,-.4720047033e-3,-.1107876030e-1,-.9169410646e-1,-.1107876030e-1,-.9169410646e-1,-.4720047033e-3,-.1606240095e-5,-.7638008748e-2,-.5251649806e-1,-.7638008748e-2,-.5251649806e-1,-.1606240095e-5,.2530111540e-8,-.2654790910e-3,-.1581160736e-1,-.2654790910e-3,-.1581160736e-1,.2530111540e-8,-.1825084025e-4,-.1679563581e-2,-.1123601370,-.1679563581e-2,-.1123601370,-.1825084025e-4,-.2605756842e-6,-.5507385590e-4,-.6481175407e-1,-.5507385590e-4,-.6481175407e-1,-.2605756842e-6},
			{-.2905112760e-1,-.1541281795e-7,-.1318937122e-1,-.1318937122e-1,-.3617602516e-2,-.4933470527e-1,-.4933470527e-1,-.9280620982e-4,-.4685522442e-1,-.4685522442e-1,-.9849828315e-1,-.9813702180e-2,-.9813702180e-2,-.1752058944,-.1626253042e-2,-.1626253042e-2,-.1454455033,-.1021198938e-3,-.1021198938e-3,-.5417267383e-1,-.2740327927e-6,-.2740327927e-6,-.6093554517e-3,-.1430261805e-1,-.1183765824,-.1430261805e-1,-.1183765824,-.6093554517e-3,-.2073647046e-5,-.9860626894e-2,-.6779850746e-1,-.9860626894e-2,-.6779850746e-1,-.2073647046e-5,.3266359953e-8,-.3427320328e-3,-.2041269733e-1,-.3427320328e-3,-.2041269733e-1,.3266359953e-8,-.2356173345e-4,-.2168307259e-2,-.1450563132,-.2168307259e-2,-.1450563132,-.2356173345e-4,-.3364017618e-6,-.7110004225e-4,-.8367161471e-1,-.7110004225e-4,-.8367161471e-1,-.3364017618e-6},
			{-.2905112760e-1,-.2299810724e-7,-.1968040980e-1,-.2072940277e-3,-.4487304244e-2,-.6119517866e-1,-.2561375849e-1,-.1309844384e-3,-.6613032970e-1,-.8305013821e-2,-.7108239413e-1,-.7082168586e-2,-.1527676939e-1,-.7777343793e-1,-.7218894686e-3,-.3434980188e-2,-.3292821660e-1,-.2311949087e-4,-.2601206999e-3,-.2697214518e-2,-.1364387568e-7,-.7948106270e-6,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,.2006344762e-8,-.8221381159e-3,.2658347174e-3,.4463401949e-5,-.1253839410e-1,.7835272927e-8,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6},
			{-.5031802903e-1,-.3983389024e-7,-.3408746970e-1,-.3590437882e-3,-.7772238941e-2,-.1059931586,-.4436433109e-1,-.2268717024e-3,-.1145410910,-.1438470590e-1,-.1231183182,-.1226667582e-1,-.2646014076e-1,-.1347075460,-.1250349237e-2,-.5949560210e-2,-.5703334418e-1,-.4004413284e-4,-.4505422684e-3,-.4671712585e-2,-.2363188589e-7,-.1376652389e-5,-.9389851147e-3,-.4458139289e-1,-.6370969475e-1,-.7697598727e-2,-.1824121677,-.1899366584e-2,-.3638961323e-5,-.3290427534e-1,-.7075212788e-2,-.1029020197e-2,-.1189769237,-.6919626319e-5,.3475091067e-8,-.1423984988e-2,.4604392371e-3,.7730838953e-5,-.2171713563e-1,.1357109081e-7,-.2318322805e-4,-.8660828315e-2,-.3161298039e-1,-.4725520274e-3,-.1427260688,-.9411218238e-4,-.1490753193e-6,-.3326494240e-3,-.6224747955e-2,-.5289486092e-5,-.3707879714e-1,-.1573892910e-5},
			{-.3211722999e-1,-.3793824429e-7,-.3246528886e-1,-.3601841323e-5,-.6153543472e-2,-.8391835536e-1,-.1470175463e-1,-.2043795582e-3,-.1031854450,-.1627413015e-2,-.5671141990e-1,-.5650341992e-2,-.2629088823e-1,-.3816708301e-1,-.3542651063e-3,-.8021127855e-2,-.8241593859e-2,-.5786570712e-5,-.7325116912e-3,-.1484655847e-3,-.7510140431e-9,-.2548589453e-5,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5},
			{-.5490147066e-2,-.4577893991e-10,-.3719247860e-2,-.3719247860e-2,-.3549462592e-3,-.1156480173e-1,-.1156480173e-1,-.3108711060e-5,-.1249745760e-1,-.1249745760e-1,-.2897665788e-1,-.1338404058e-2,-.1338404058e-2,-.6993680883e-1,-.1364242863e-3,-.1364242863e-3,-.7001415346e-1,-.4369173091e-5,-.4369173091e-5,-.2969356790e-1,-.2578450140e-8,-.2578450140e-8,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8},
			{-.9509213660e-2,-.7929144985e-10,-.6441926259e-2,-.6441926259e-2,-.6147849549e-3,-.2003082417e-1,-.2003082417e-1,-.5384445502e-5,-.2164623154e-1,-.2164623154e-1,-.5018904368e-1,-.2318183830e-2,-.2318183830e-2,-.1211341062,-.2362937952e-3,-.2362937952e-3,-.1212680711,-.7567629781e-5,-.7567629781e-5,-.5143076827e-1,-.4466006648e-8,-.4466006648e-8,-.6197712293e-4,-.4165091360e-2,-.6973083633e-1,-.4165091360e-2,-.6973083633e-1,-.6197712293e-4,-.4089546907e-7,-.3270153304e-2,-.4275519802e-1,-.3270153304e-2,-.4275519802e-1,-.4089546907e-7,-.1392377369e-10,-.6890929850e-4,-.1602773282e-1,-.6890929850e-4,-.1602773282e-1,-.1392377369e-10,-.9704139476e-6,-.4031888160e-3,-.1094954884,-.4031888160e-3,-.1094954884,-.9704139476e-6,-.4729583171e-8,-.5954404436e-5,-.7398029375e-1,-.5954404436e-5,-.7398029375e-1,-.4729583171e-8},
			{-.1486738854e-1,-.1849807308e-9,-.1502850849e-1,-.1582954870e-3,-.1192279250e-2,-.3884665007e-1,-.1625959323e-1,-.1188156743e-4,-.4776557949e-1,-.5998666568e-2,-.5662815244e-1,-.2615600094e-2,-.5642045787e-2,-.8406962479e-1,-.1639928781e-3,-.7803303852e-3,-.4292442302e-1,-.2678661737e-5,-.3013800652e-4,-.4003578188e-2,-.3476519485e-9,-.2025212407e-7,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7},
			{-.2216299428e-2,-.2904510069e-12,-.2240317772e-2,-.2240317772e-2,-.7439227355e-4,-.5790916683e-2,-.5790916683e-2,-.2224370628e-6,-.7120472184e-2,-.7120472184e-2,-.1820922229e-1,-.3899106406e-3,-.3899106406e-3,-.5963296216e-1,-.2444661488e-4,-.2444661488e-4,-.7199376027e-1,-.3993113157e-6,-.3993113157e-6,-.3476706016e-1,-.5182489262e-10,-.5182489262e-10,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 52; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_bottom_fluxes_to_dofs(
		const T i_input[52],
		T o_output[56]
	) {
		static const T matrix[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-.1148215401,-.1569113387e-1,-.1569113387e-1,-.1569113387e-1,-.1021689440,-.1021689440,-.1021689440,-.6587041484e-1,-.6587041484e-1,-.6587041484e-1,-.1032031823,-.1032031823,-.1032031823,-.7348534591e-1,-.7348534591e-1,-.7348534591e-1,-.3478299668e-1,-.3478299668e-1,-.3478299668e-1,-.8775204783e-2,-.8775204783e-2,-.8775204783e-2,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{-.1753927996,-.2396860291e-1,-.2396860291e-1,-.2396860291e-1,-.1560656399,-.1560656399,-.1560656399,-.1006187207,-.1006187207,-.1006187207,-.1576454650,-.1576454650,-.1576454650,-.1122507200,-.1122507200,-.1122507200,-.5313190506e-1,-.5313190506e-1,-.5313190506e-1,-.1340434672e-1,-.1340434672e-1,-.1340434672e-1,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{-.2198665100,-.3004623384e-1,-.3004623384e-1,-.3004623384e-1,-.1956386332,-.1956386332,-.1956386332,-.1261322415,-.1261322415,-.1261322415,-.1976190488,-.1976190488,-.1976190488,-.1407137244,-.1407137244,-.1407137244,-.6660436781e-1,-.6660436781e-1,-.6660436781e-1,-.1680323787e-1,-.1680323787e-1,-.1680323787e-1,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1},
			{-.5412739307e-1,-.1103717978e-1,-.1103717978e-1,-.1162547668e-3,-.5974166429e-1,-.5974166429e-1,-.2500537776e-1,-.4382549273e-1,-.4382549273e-1,-.5503848602e-2,-.3510914219e-1,-.3510914219e-1,-.7573305578e-1,-.1537719318e-1,-.1537719318e-1,-.7316958648e-1,-.3712176777e-2,-.3712176777e-2,-.4176623208e-1,-.2059615826e-3,-.2059615826e-3,-.1199809046e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1},
			{-.9375139486e-1,-.1911695615e-1,-.1911695615e-1,-.2013591627e-3,-.1034755978,-.1034755978,-.4331058473e-1,-.7590798006e-1,-.7590798006e-1,-.9532945415e-2,-.6081081808e-1,-.6081081808e-1,-.1311735004,-.2663407986e-1,-.2663407986e-1,-.1267334413,-.6429678784e-2,-.6429678784e-2,-.7234123601e-1,-.3567359253e-3,-.3567359253e-3,-.2078130226e-1,-.5831436180e-1,-.1179575144,-.2036700860e-1,-.2036700860e-1,-.5831436180e-1,-.1179575144,-.3099954510e-1,-.5894683929e-1,-.1843453093e-2,-.1843453093e-2,-.3099954510e-1,-.5894683929e-1,-.2931476389e-2,-.1144814092e-1,.6215215367e-4,.6215215367e-4,-.2931476389e-2,-.1144814092e-1,-.2168320548e-1,-.8802284926e-1,-.4802701816e-2,-.4802701816e-2,-.2168320548e-1,-.8802284926e-1,-.3505184516e-2,-.3700669622e-1,-.5884465473e-3,-.5884465473e-3,-.3505184516e-2,-.3700669622e-1},
			{-.1210325303,-.2467988427e-1,-.2467988427e-1,-.2599535613e-3,-.1335864224,-.1335864224,-.5591372447e-1,-.9799678087e-1,-.9799678087e-1,-.1230697961e-1,-.7850642855e-1,-.7850642855e-1,-.1693442608,-.3438444924e-1,-.3438444924e-1,-.1636121692,-.8300679616e-2,-.8300679616e-2,-.9339213407e-1,-.4605440992e-3,-.4605440992e-3,-.2682854586e-1,-.7528351732e-1,-.1522824962,-.2629369504e-1,-.2629369504e-1,-.7528351732e-1,-.1522824962,-.4002024064e-1,-.7610004229e-1,-.2379887709e-2,-.2379887709e-2,-.4002024064e-1,-.7610004229e-1,-.3784519743e-2,-.1477948638e-1,.8023808537e-4,.8023808537e-4,-.3784519743e-2,-.1477948638e-1,-.2799289790e-1,-.1136370098,-.6200261386e-2,-.6200261386e-2,-.2799289790e-1,-.1136370098,-.4525173753e-2,-.4777543936e-1,-.7596812259e-3,-.7596812259e-3,-.4525173753e-2,-.4777543936e-1},
			{-.1432076211,-.2920163288e-1,-.2920163288e-1,-.3075812017e-3,-.1580615866,-.1580615866,-.6615801099e-1,-.1159513548,-.1159513548,-.1456181465e-1,-.9289005896e-1,-.9289005896e-1,-.2003708316,-.4068422900e-1,-.4068422900e-1,-.1935885293,-.9821496573e-2,-.9821496573e-2,-.1105030633,-.5449231269e-3,-.5449231269e-3,-.3174396355e-1,-.8907665897e-1,-.1801830794,-.3111111953e-1,-.3111111953e-1,-.8907665897e-1,-.1801830794,-.4735258731e-1,-.9004278434e-1,-.2815921112e-2,-.2815921112e-2,-.4735258731e-1,-.9004278434e-1,-.4477904149e-2,-.1748732412e-1,.9493898295e-4,.9493898295e-4,-.4477904149e-2,-.1748732412e-1,-.3312164346e-1,-.1344571232,-.7336248207e-2,-.7336248207e-2,-.3312164346e-1,-.1344571232,-.5354257791e-2,-.5652866219e-1,-.8988669485e-3,-.8988669485e-3,-.5354257791e-2,-.5652866219e-1},
			{-.1623821792,-.3311153935e-1,-.3311153935e-1,-.3487643005e-3,-.1792249928,-.1792249928,-.7501613328e-1,-.1314764782,-.1314764782,-.1651154581e-1,-.1053274266,-.1053274266,-.2271991674,-.4613157954e-1,-.4613157954e-1,-.2195087594,-.1113653033e-1,-.1113653033e-1,-.1252986963,-.6178847477e-3,-.6178847477e-3,-.3599427137e-1,-.1010034375,-.2043084081,-.3527669370e-1,-.3527669370e-1,-.1010034375,-.2043084081,-.5369278714e-1,-.1020989206,-.3192954419e-2,-.3192954419e-2,-.5369278714e-1,-.1020989206,-.5077466048e-2,-.1982876173e-1,.1076506880e-3,.1076506880e-3,-.5077466048e-2,-.1982876173e-1,-.3755641357e-1,-.1524600472,-.8318523561e-2,-.8318523561e-2,-.3755641357e-1,-.1524600472,-.6071157673e-2,-.6409747809e-1,-.1019219318e-2,-.1019219318e-2,-.6071157673e-2,-.6409747809e-1},
			{-.2852764098e-1,-.8679943781e-2,-.8679943781e-2,-.9629909754e-6,-.3905626741e-1,-.3905626741e-1,-.6842313070e-2,-.3260004675e-1,-.3260004675e-1,-.5141591464e-3,-.1335372273e-1,-.1335372273e-1,-.6213451013e-1,-.3597562967e-2,-.3597562967e-2,-.8145457175e-1,-.4429404170e-3,-.4429404170e-3,-.5607103930e-1,-.5404682501e-5,-.5404682501e-5,-.1834095772e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1},
			{-.4941132360e-1,-.1503410363e-1,-.1503410363e-1,-.1667949296e-5,-.6764743950e-1,-.6764743950e-1,-.1185123388e-1,-.5646493729e-1,-.5646493729e-1,-.8905497646e-3,-.2312932623e-1,-.2312932623e-1,-.1076201284,-.6231161842e-2,-.6231161842e-2,-.1410834568,-.7671953069e-3,-.7671953069e-3,-.9711788894e-1,-.9361184690e-5,-.9361184690e-5,-.3176747063e-1,-.2734334585e-1,-.1118798374,-.3335455075e-2,-.3335455075e-2,-.2734334585e-1,-.1118798374,-.1655335563e-1,-.5985445076e-1,-.5853822990e-4,-.5853822990e-4,-.1655335563e-1,-.5985445076e-1,-.9490228526e-3,-.1447350792e-1,-.4265954647e-6,-.4265954647e-6,-.9490228526e-3,-.1447350792e-1,-.6491996486e-2,-.1069848175,-.3184949515e-3,-.3184949515e-3,-.6491996486e-2,-.1069848175,-.4726574345e-3,-.5268480225e-1,-.1332105931e-4,-.1332105931e-4,-.4726574345e-3,-.5268480225e-1},
			{-.6378974448e-1,-.1940894434e-1,-.1940894434e-1,-.2153313281e-5,-.8733246888e-1,-.8733246888e-1,-.1529987715e-1,-.7289592064e-1,-.7289592064e-1,-.1149694802e-2,-.2985983179e-1,-.2985983179e-1,-.1389369884,-.8044395345e-2,-.8044395345e-2,-.1821379596,-.9904448821e-3,-.9904448821e-3,-.1253786555,-.1208523748e-4,-.1208523748e-4,-.4101162825e-1,-.3530010772e-1,-.1444362491,-.4306053985e-2,-.4306053985e-2,-.3530010772e-1,-.1444362491,-.2137029023e-1,-.7727176368e-1,-.7557252986e-4,-.7557252986e-4,-.2137029023e-1,-.7727176368e-1,-.1225183234e-2,-.1868521838e-1,-.5507323768e-6,-.5507323768e-6,-.1225183234e-2,-.1868521838e-1,-.8381131425e-2,-.1381168056,-.4111752145e-3,-.4111752145e-3,-.8381131425e-2,-.1381168056,-.6101981241e-3,-.6801578726e-1,-.1719741363e-4,-.1719741363e-4,-.6101981241e-3,-.6801578726e-1},
			{-.7547704355e-1,-.2296497265e-1,-.2296497265e-1,-.2547834636e-5,-.1033331707,-.1033331707,-.1810305878e-1,-.8625161644e-1,-.8625161644e-1,-.1360337236e-2,-.3533062943e-1,-.3533062943e-1,-.1643924617,-.9518256939e-2,-.9518256939e-2,-.2155085401,-.1171910189e-2,-.1171910189e-2,-.1483500258,-.1429944582e-4,-.1429944582e-4,-.4852561294e-1,-.4176765072e-1,-.1708992746,-.5094991785e-2,-.5094991785e-2,-.4176765072e-1,-.1708992746,-.2528566840e-1,-.9142918380e-1,-.8941862322e-4,-.8941862322e-4,-.2528566840e-1,-.9142918380e-1,-.1449656353e-2,-.2210864854e-1,-.6516353360e-6,-.6516353360e-6,-.1449656353e-2,-.2210864854e-1,-.9916688441e-2,-.1634220082,-.4865090747e-3,-.4865090747e-3,-.9916688441e-2,-.1634220082,-.7219961572e-3,-.8047736478e-1,-.2034825423e-4,-.2034825423e-4,-.7219961572e-3,-.8047736478e-1},
			{-.1299205923e-1,-.5898466126e-2,-.5898466126e-2,-.6892821735e-8,-.2206315093e-1,-.2206315093e-1,-.1617841028e-2,-.2095429338e-1,-.2095429338e-1,-.4150419878e-4,-.4388821038e-2,-.4388821038e-2,-.4404977136e-1,-.7272824702e-3,-.7272824702e-3,-.7835445799e-1,-.4566940490e-4,-.4566940490e-4,-.6504520647e-1,-.1225511905e-6,-.1225511905e-6,-.2422675624e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1},
			{-.2250290667e-1,-.1021644301e-1,-.1021644301e-1,-.1193871745e-7,-.3821449838e-1,-.3821449838e-1,-.2802182859e-2,-.3629390077e-1,-.3629390077e-1,-.7188738101e-4,-.7601661022e-2,-.7601661022e-2,-.7629644205e-1,-.1259690190e-2,-.1259690190e-2,-.1357139022,-.7910172963e-4,-.7910172963e-4,-.1126616024,-.2122648885e-6,-.2122648885e-6,-.4196197271e-1,-.1107876030e-1,-.9169410646e-1,-.4720047033e-3,-.4720047033e-3,-.1107876030e-1,-.9169410646e-1,-.7638008748e-2,-.5251649806e-1,-.1606240095e-5,-.1606240095e-5,-.7638008748e-2,-.5251649806e-1,-.2654790910e-3,-.1581160736e-1,.2530111540e-8,.2530111540e-8,-.2654790910e-3,-.1581160736e-1,-.1679563581e-2,-.1123601370,-.1825084025e-4,-.1825084025e-4,-.1679563581e-2,-.1123601370,-.5507385590e-4,-.6481175407e-1,-.2605756842e-6,-.2605756842e-6,-.5507385590e-4,-.6481175407e-1},
			{-.2905112760e-1,-.1318937122e-1,-.1318937122e-1,-.1541281795e-7,-.4933470527e-1,-.4933470527e-1,-.3617602516e-2,-.4685522442e-1,-.4685522442e-1,-.9280620982e-4,-.9813702180e-2,-.9813702180e-2,-.9849828315e-1,-.1626253042e-2,-.1626253042e-2,-.1752058944,-.1021198938e-3,-.1021198938e-3,-.1454455033,-.2740327927e-6,-.2740327927e-6,-.5417267383e-1,-.1430261805e-1,-.1183765824,-.6093554517e-3,-.6093554517e-3,-.1430261805e-1,-.1183765824,-.9860626894e-2,-.6779850746e-1,-.2073647046e-5,-.2073647046e-5,-.9860626894e-2,-.6779850746e-1,-.3427320328e-3,-.2041269733e-1,.3266359953e-8,.3266359953e-8,-.3427320328e-3,-.2041269733e-1,-.2168307259e-2,-.1450563132,-.2356173345e-4,-.2356173345e-4,-.2168307259e-2,-.1450563132,-.7110004225e-4,-.8367161471e-1,-.3364017618e-6,-.3364017618e-6,-.7110004225e-4,-.8367161471e-1},
			{-.5490147066e-2,-.3719247860e-2,-.3719247860e-2,-.4577893991e-10,-.1156480173e-1,-.1156480173e-1,-.3549462592e-3,-.1249745760e-1,-.1249745760e-1,-.3108711060e-5,-.1338404058e-2,-.1338404058e-2,-.2897665788e-1,-.1364242863e-3,-.1364242863e-3,-.6993680883e-1,-.4369173091e-5,-.4369173091e-5,-.7001415346e-1,-.2578450140e-8,-.2578450140e-8,-.2969356790e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1},
			{-.9509213660e-2,-.6441926259e-2,-.6441926259e-2,-.7929144985e-10,-.2003082417e-1,-.2003082417e-1,-.6147849549e-3,-.2164623154e-1,-.2164623154e-1,-.5384445502e-5,-.2318183830e-2,-.2318183830e-2,-.5018904368e-1,-.2362937952e-3,-.2362937952e-3,-.1211341062,-.7567629781e-5,-.7567629781e-5,-.1212680711,-.4466006648e-8,-.4466006648e-8,-.5143076827e-1,-.4165091360e-2,-.6973083633e-1,-.6197712293e-4,-.6197712293e-4,-.4165091360e-2,-.6973083633e-1,-.3270153304e-2,-.4275519802e-1,-.4089546907e-7,-.4089546907e-7,-.3270153304e-2,-.4275519802e-1,-.6890929850e-4,-.1602773282e-1,-.1392377369e-10,-.1392377369e-10,-.6890929850e-4,-.1602773282e-1,-.4031888160e-3,-.1094954884,-.9704139476e-6,-.9704139476e-6,-.4031888160e-3,-.1094954884,-.5954404436e-5,-.7398029375e-1,-.4729583171e-8,-.4729583171e-8,-.5954404436e-5,-.7398029375e-1},
			{-.2216299428e-2,-.2240317772e-2,-.2240317772e-2,-.2904510069e-12,-.5790916683e-2,-.5790916683e-2,-.7439227355e-4,-.7120472184e-2,-.7120472184e-2,-.2224370628e-6,-.3899106406e-3,-.3899106406e-3,-.1820922229e-1,-.2444661488e-4,-.2444661488e-4,-.5963296216e-1,-.3993113157e-6,-.3993113157e-6,-.7199376027e-1,-.5182489262e-10,-.5182489262e-10,-.3476706016e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1},
			{-.5412739307e-1,-.1162547668e-3,-.1103717978e-1,-.1103717978e-1,-.2500537776e-1,-.5974166429e-1,-.5974166429e-1,-.5503848602e-2,-.4382549273e-1,-.4382549273e-1,-.7573305578e-1,-.3510914219e-1,-.3510914219e-1,-.7316958648e-1,-.1537719318e-1,-.1537719318e-1,-.4176623208e-1,-.3712176777e-2,-.3712176777e-2,-.1199809046e-1,-.2059615826e-3,-.2059615826e-3,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3},
			{-.9375139486e-1,-.2013591627e-3,-.1911695615e-1,-.1911695615e-1,-.4331058473e-1,-.1034755978,-.1034755978,-.9532945415e-2,-.7590798006e-1,-.7590798006e-1,-.1311735004,-.6081081808e-1,-.6081081808e-1,-.1267334413,-.2663407986e-1,-.2663407986e-1,-.7234123601e-1,-.6429678784e-2,-.6429678784e-2,-.2078130226e-1,-.3567359253e-3,-.3567359253e-3,-.2036700860e-1,-.5831436180e-1,-.1179575144,-.5831436180e-1,-.1179575144,-.2036700860e-1,-.1843453093e-2,-.3099954510e-1,-.5894683929e-1,-.3099954510e-1,-.5894683929e-1,-.1843453093e-2,.6215215367e-4,-.2931476389e-2,-.1144814092e-1,-.2931476389e-2,-.1144814092e-1,.6215215367e-4,-.4802701816e-2,-.2168320548e-1,-.8802284926e-1,-.2168320548e-1,-.8802284926e-1,-.4802701816e-2,-.5884465473e-3,-.3505184516e-2,-.3700669622e-1,-.3505184516e-2,-.3700669622e-1,-.5884465473e-3},
			{-.1210325303,-.2599535613e-3,-.2467988427e-1,-.2467988427e-1,-.5591372447e-1,-.1335864224,-.1335864224,-.1230697961e-1,-.9799678087e-1,-.9799678087e-1,-.1693442608,-.7850642855e-1,-.7850642855e-1,-.1636121692,-.3438444924e-1,-.3438444924e-1,-.9339213407e-1,-.8300679616e-2,-.8300679616e-2,-.2682854586e-1,-.4605440992e-3,-.4605440992e-3,-.2629369504e-1,-.7528351732e-1,-.1522824962,-.7528351732e-1,-.1522824962,-.2629369504e-1,-.2379887709e-2,-.4002024064e-1,-.7610004229e-1,-.4002024064e-1,-.7610004229e-1,-.2379887709e-2,.8023808537e-4,-.3784519743e-2,-.1477948638e-1,-.3784519743e-2,-.1477948638e-1,.8023808537e-4,-.6200261386e-2,-.2799289790e-1,-.1136370098,-.2799289790e-1,-.1136370098,-.6200261386e-2,-.7596812259e-3,-.4525173753e-2,-.4777543936e-1,-.4525173753e-2,-.4777543936e-1,-.7596812259e-3},
			{-.1432076211,-.3075812017e-3,-.2920163288e-1,-.2920163288e-1,-.6615801099e-1,-.1580615866,-.1580615866,-.1456181465e-1,-.1159513548,-.1159513548,-.2003708316,-.9289005896e-1,-.9289005896e-1,-.1935885293,-.4068422900e-1,-.4068422900e-1,-.1105030633,-.9821496573e-2,-.9821496573e-2,-.3174396355e-1,-.5449231269e-3,-.5449231269e-3,-.3111111953e-1,-.8907665897e-1,-.1801830794,-.8907665897e-1,-.1801830794,-.3111111953e-1,-.2815921112e-2,-.4735258731e-1,-.9004278434e-1,-.4735258731e-1,-.9004278434e-1,-.2815921112e-2,.9493898295e-4,-.4477904149e-2,-.1748732412e-1,-.4477904149e-2,-.1748732412e-1,.9493898295e-4,-.7336248207e-2,-.3312164346e-1,-.1344571232,-.3312164346e-1,-.1344571232,-.7336248207e-2,-.8988669485e-3,-.5354257791e-2,-.5652866219e-1,-.5354257791e-2,-.5652866219e-1,-.8988669485e-3},
			{-.1623821792,-.3487643005e-3,-.3311153935e-1,-.3311153935e-1,-.7501613328e-1,-.1792249928,-.1792249928,-.1651154581e-1,-.1314764782,-.1314764782,-.2271991674,-.1053274266,-.1053274266,-.2195087594,-.4613157954e-1,-.4613157954e-1,-.1252986963,-.1113653033e-1,-.1113653033e-1,-.3599427137e-1,-.6178847477e-3,-.6178847477e-3,-.3527669370e-1,-.1010034375,-.2043084081,-.1010034375,-.2043084081,-.3527669370e-1,-.3192954419e-2,-.5369278714e-1,-.1020989206,-.5369278714e-1,-.1020989206,-.3192954419e-2,.1076506880e-3,-.5077466048e-2,-.1982876173e-1,-.5077466048e-2,-.1982876173e-1,.1076506880e-3,-.8318523561e-2,-.3755641357e-1,-.1524600472,-.3755641357e-1,-.1524600472,-.8318523561e-2,-.1019219318e-2,-.6071157673e-2,-.6409747809e-1,-.6071157673e-2,-.6409747809e-1,-.1019219318e-2},
			{-.6987816394e-1,-.2239469694e-3,-.2126143325e-1,-.2239469694e-3,-.4004261797e-1,-.9566792642e-1,-.4004261797e-1,-.1002844321e-1,-.7985348011e-1,-.1002844321e-1,-.7055750928e-1,-.3270980684e-1,-.7055750928e-1,-.4193122589e-1,-.8812193584e-2,-.4193122589e-1,-.1220724282e-1,-.1084978008e-2,-.1220724282e-1,-.7712083501e-3,-.1323871434e-4,-.7712083501e-3,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,.2845518738e-4,-.5241314676e-2,.1111245502e-3,.2845518738e-4,-.5241314676e-2,.1111245502e-3,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2},
			{-.1210325303,-.3878875292e-3,-.3682588264e-1,-.3878875292e-3,-.6935584881e-1,-.1657017093,-.6935584881e-1,-.1736977316e-1,-.1383102847,-.1736977316e-1,-.1222091909,-.5665504737e-1,-.1222091909,-.7262701368e-1,-.1526316702e-1,-.7262701368e-1,-.2114356478e-1,-.1879237035e-2,-.2114356478e-1,-.1335772046e-2,-.2293012587e-4,-.1335772046e-2,-.2339262725e-1,-.1354806795,-.4731829487e-1,-.2339262725e-1,-.1354806795,-.4731829487e-1,-.2411228903e-2,-.7710221821e-1,-.4585045432e-2,-.2411228903e-2,-.7710221821e-1,-.4585045432e-2,.4928583028e-4,-.9078223319e-2,.1924733670e-3,.4928583028e-4,-.9078223319e-2,.1924733670e-3,-.3522216439e-2,-.6455439842e-1,-.1429841759e-1,-.3522216439e-2,-.6455439842e-1,-.1429841759e-1,-.1943650851e-3,-.1222338663e-1,-.2052048795e-2,-.1943650851e-3,-.1222338663e-1,-.2052048795e-2},
			{-.1562523248,-.5007606471e-3,-.4754201006e-1,-.5007606471e-3,-.8953801576e-1,-.2139199867,-.8953801576e-1,-.2242428073e-1,-.1785578098,-.2242428073e-1,-.1577713871,-.7314135164e-1,-.1577713871,-.9376107150e-1,-.1970466389e-1,-.9376107150e-1,-.2729622476e-1,-.2426084580e-2,-.2729622476e-1,-.1724474296e-2,-.2960266522e-4,-.1724474296e-2,-.3019975193e-1,-.1749048052,-.6108765601e-1,-.3019975193e-1,-.1749048052,-.6108765601e-1,-.3112883129e-2,-.9953853571e-1,-.5919268200e-2,-.3112883129e-2,-.9953853571e-1,-.5919268200e-2,.6362773332e-4,-.1171993591e-1,.2484820485e-3,.6362773332e-4,-.1171993591e-1,.2484820485e-3,-.4547161871e-2,-.8333937002e-1,-.1845917774e-1,-.4547161871e-2,-.8333937002e-1,-.1845917774e-1,-.2509242459e-3,-.1578032429e-1,-.2649183603e-2,-.2509242459e-3,-.1578032429e-1,-.2649183603e-2},
			{-.1848802439,-.5925079880e-3,-.5625246492e-1,-.5925079880e-3,-.1059428090,-.2531135418,-.1059428090,-.2653276678e-1,-.2112724497,-.2653276678e-1,-.1866776227,-.8654201436e-1,-.1866776227,-.1109395959,-.2331487273e-1,-.1109395959,-.3229732869e-1,-.2870581987e-2,-.3229732869e-1,-.2040425504e-2,-.3502634584e-4,-.2040425504e-2,-.3573282836e-1,-.2069501564,-.7227988933e-1,-.3573282836e-1,-.2069501564,-.7227988933e-1,-.3683212988e-2,-.1177755837,-.7003772586e-2,-.3683212988e-2,-.1177755837,-.7003772586e-2,.7528534932e-4,-.1386721518e-1,.2940079245e-3,.7528534932e-4,-.1386721518e-1,.2940079245e-3,-.5380274482e-2,-.9860847240e-1,-.2184119364e-1,-.5380274482e-2,-.9860847240e-1,-.2184119364e-1,-.2968975716e-3,-.1867153150e-1,-.3134556311e-2,-.2968975716e-3,-.1867153150e-1,-.3134556311e-2},
			{-.5031802903e-1,-.2406231299e-3,-.2284466108e-1,-.2534486742e-5,-.3576590913e-1,-.8545021612e-1,-.1497012310e-1,-.1019197433e-1,-.8115562931e-1,-.1279964701e-2,-.3666559727e-1,-.1699783079e-1,-.7909044629e-1,-.1340300809e-1,-.2816752894e-2,-.6377578460e-1,-.1990066688e-2,-.1768768448e-3,-.2239052508e-1,-.2764961419e-4,-.4746387201e-6,-.1610701220e-2,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2},
			{-.8715338278e-1,-.4167714865e-3,-.3956811365e-1,-.4389859808e-5,-.6194837174e-1,-.1480041159,-.2592901380e-1,-.1765301736e-1,-.1405656732,-.2216963893e-2,-.6350667735e-1,-.2944110655e-1,-.1369886713,-.2321469098e-1,-.4878759123e-2,-.1104628991,-.3446896614e-2,-.3063596816e-3,-.3878152703e-1,-.4789053657e-4,-.8220983782e-6,-.2789816347e-2,-.1498609619e-1,-.1755646063,-.1058742437e-1,-.5234078544e-2,-.8679343590e-1,-.6131809964e-1,-.1759148700e-2,-.1069636429,-.1989227818e-3,-.1046114740e-3,-.5625109532e-1,-.6360817804e-2,.2179945987e-4,-.1568097578e-1,-.1804945451e-5,-.4621846472e-6,-.4015360274e-2,.3324626528e-3,-.1440801715e-2,-.1071978198,-.1295502609e-2,-.3191290614e-3,-.2640669295e-1,-.2374368331e-1,-.3580859223e-4,-.2377551727e-1,-.6346770766e-4,-.6011507348e-5,-.2251959335e-2,-.3991407866e-2},
			{-.1125145334,-.5380496754e-3,-.5108221507e-1,-.5667284643e-5,-.7997500406e-1,-.1910724920,-.3347421288e-1,-.2278994742e-1,-.1814695038,-.2862088080e-2,-.8198676790e-1,-.3800830512e-1,-.1768516141,-.2997003719e-1,-.6298450945e-2,-.1426069896,-.4449924395e-2,-.3955086483e-3,-.5006673611e-1,-.6182641688e-4,-.1061324443e-5,-.3601637417e-2,-.1934696698e-1,-.2266529321,-.1366830609e-1,-.6757166343e-2,-.1120498440,-.7916132626e-1,-.2271051205e-2,-.1380894693,-.2568082070e-3,-.1350528323e-3,-.7261985182e-1,-.8211780472e-2,.2814298168e-4,-.2024405269e-1,-.2330174558e-5,-.5966778137e-6,-.5183807822e-2,.4292074393e-3,-.1860067016e-2,-.1383917903,-.1672486677e-2,-.4119938467e-3,-.3409089400e-1,-.3065296333e-1,-.4622869379e-4,-.3069406081e-1,-.8193645830e-4,-.7760822615e-5,-.2907267000e-2,-.5152885397e-2},
			{-.2905112760e-1,-.2072940277e-3,-.1968040980e-1,-.2299810724e-7,-.2561375849e-1,-.6119517866e-1,-.4487304244e-2,-.8305013821e-2,-.6613032970e-1,-.1309844384e-3,-.1527676939e-1,-.7082168586e-2,-.7108239413e-1,-.3434980188e-2,-.7218894686e-3,-.7777343793e-1,-.2601206999e-3,-.2311949087e-4,-.3292821660e-1,-.7948106270e-6,-.1364387568e-7,-.2697214518e-2,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,.4463401949e-5,-.1253839410e-1,.7835272927e-8,.2006344762e-8,-.8221381159e-3,.2658347174e-3,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2},
			{-.5031802903e-1,-.3590437882e-3,-.3408746970e-1,-.3983389024e-7,-.4436433109e-1,-.1059931586,-.7772238941e-2,-.1438470590e-1,-.1145410910,-.2268717024e-3,-.2646014076e-1,-.1226667582e-1,-.1231183182,-.5949560210e-2,-.1250349237e-2,-.1347075460,-.4505422684e-3,-.4004413284e-4,-.5703334418e-1,-.1376652389e-5,-.2363188589e-7,-.4671712585e-2,-.7697598727e-2,-.1824121677,-.1899366584e-2,-.9389851147e-3,-.4458139289e-1,-.6370969475e-1,-.1029020197e-2,-.1189769237,-.6919626319e-5,-.3638961323e-5,-.3290427534e-1,-.7075212788e-2,.7730838953e-5,-.2171713563e-1,.1357109081e-7,.3475091067e-8,-.1423984988e-2,.4604392371e-3,-.4725520274e-3,-.1427260688,-.9411218238e-4,-.2318322805e-4,-.8660828315e-2,-.3161298039e-1,-.5289486092e-5,-.3707879714e-1,-.1573892910e-5,-.1490753193e-6,-.3326494240e-3,-.6224747955e-2},
			{-.1486738854e-1,-.1582954870e-3,-.1502850849e-1,-.1849807308e-9,-.1625959323e-1,-.3884665007e-1,-.1192279250e-2,-.5998666568e-2,-.4776557949e-1,-.1188156743e-4,-.5642045787e-2,-.2615600094e-2,-.5662815244e-1,-.7803303852e-3,-.1639928781e-3,-.8406962479e-1,-.3013800652e-4,-.2678661737e-5,-.4292442302e-1,-.2025212407e-7,-.3476519485e-9,-.4003578188e-2,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2},
			{-.2852764098e-1,-.9629909754e-6,-.8679943781e-2,-.8679943781e-2,-.6842313070e-2,-.3905626741e-1,-.3905626741e-1,-.5141591464e-3,-.3260004675e-1,-.3260004675e-1,-.6213451013e-1,-.1335372273e-1,-.1335372273e-1,-.8145457175e-1,-.3597562967e-2,-.3597562967e-2,-.5607103930e-1,-.4429404170e-3,-.4429404170e-3,-.1834095772e-1,-.5404682501e-5,-.5404682501e-5,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5},
			{-.4941132360e-1,-.1667949296e-5,-.1503410363e-1,-.1503410363e-1,-.1185123388e-1,-.6764743950e-1,-.6764743950e-1,-.8905497646e-3,-.5646493729e-1,-.5646493729e-1,-.1076201284,-.2312932623e-1,-.2312932623e-1,-.1410834568,-.6231161842e-2,-.6231161842e-2,-.9711788894e-1,-.7671953069e-3,-.7671953069e-3,-.3176747063e-1,-.9361184690e-5,-.9361184690e-5,-.3335455075e-2,-.2734334585e-1,-.1118798374,-.2734334585e-1,-.1118798374,-.3335455075e-2,-.5853822990e-4,-.1655335563e-1,-.5985445076e-1,-.1655335563e-1,-.5985445076e-1,-.5853822990e-4,-.4265954647e-6,-.9490228526e-3,-.1447350792e-1,-.9490228526e-3,-.1447350792e-1,-.4265954647e-6,-.3184949515e-3,-.6491996486e-2,-.1069848175,-.6491996486e-2,-.1069848175,-.3184949515e-3,-.1332105931e-4,-.4726574345e-3,-.5268480225e-1,-.4726574345e-3,-.5268480225e-1,-.1332105931e-4},
			{-.6378974448e-1,-.2153313281e-5,-.1940894434e-1,-.1940894434e-1,-.1529987715e-1,-.8733246888e-1,-.8733246888e-1,-.1149694802e-2,-.7289592064e-1,-.7289592064e-1,-.1389369884,-.2985983179e-1,-.2985983179e-1,-.1821379596,-.8044395345e-2,-.8044395345e-2,-.1253786555,-.9904448821e-3,-.9904448821e-3,-.4101162825e-1,-.1208523748e-4,-.1208523748e-4,-.4306053985e-2,-.3530010772e-1,-.1444362491,-.3530010772e-1,-.1444362491,-.4306053985e-2,-.7557252986e-4,-.2137029023e-1,-.7727176368e-1,-.2137029023e-1,-.7727176368e-1,-.7557252986e-4,-.5507323768e-6,-.1225183234e-2,-.1868521838e-1,-.1225183234e-2,-.1868521838e-1,-.5507323768e-6,-.4111752145e-3,-.8381131425e-2,-.1381168056,-.8381131425e-2,-.1381168056,-.4111752145e-3,-.1719741363e-4,-.6101981241e-3,-.6801578726e-1,-.6101981241e-3,-.6801578726e-1,-.1719741363e-4},
			{-.7547704355e-1,-.2547834636e-5,-.2296497265e-1,-.2296497265e-1,-.1810305878e-1,-.1033331707,-.1033331707,-.1360337236e-2,-.8625161644e-1,-.8625161644e-1,-.1643924617,-.3533062943e-1,-.3533062943e-1,-.2155085401,-.9518256939e-2,-.9518256939e-2,-.1483500258,-.1171910189e-2,-.1171910189e-2,-.4852561294e-1,-.1429944582e-4,-.1429944582e-4,-.5094991785e-2,-.4176765072e-1,-.1708992746,-.4176765072e-1,-.1708992746,-.5094991785e-2,-.8941862322e-4,-.2528566840e-1,-.9142918380e-1,-.2528566840e-1,-.9142918380e-1,-.8941862322e-4,-.6516353360e-6,-.1449656353e-2,-.2210864854e-1,-.1449656353e-2,-.2210864854e-1,-.6516353360e-6,-.4865090747e-3,-.9916688441e-2,-.1634220082,-.9916688441e-2,-.1634220082,-.4865090747e-3,-.2034825423e-4,-.7219961572e-3,-.8047736478e-1,-.7219961572e-3,-.8047736478e-1,-.2034825423e-4},
			{-.5031802903e-1,-.2534486742e-5,-.2284466108e-1,-.2406231299e-3,-.1497012310e-1,-.8545021612e-1,-.3576590913e-1,-.1279964701e-2,-.8115562931e-1,-.1019197433e-1,-.7909044629e-1,-.1699783079e-1,-.3666559727e-1,-.6377578460e-1,-.2816752894e-2,-.1340300809e-1,-.2239052508e-1,-.1768768448e-3,-.1990066688e-2,-.1610701220e-2,-.4746387201e-6,-.2764961419e-4,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4},
			{-.8715338278e-1,-.4389859808e-5,-.3956811365e-1,-.4167714865e-3,-.2592901380e-1,-.1480041159,-.6194837174e-1,-.2216963893e-2,-.1405656732,-.1765301736e-1,-.1369886713,-.2944110655e-1,-.6350667735e-1,-.1104628991,-.4878759123e-2,-.2321469098e-1,-.3878152703e-1,-.3063596816e-3,-.3446896614e-2,-.2789816347e-2,-.8220983782e-6,-.4789053657e-4,-.5234078544e-2,-.8679343590e-1,-.6131809964e-1,-.1498609619e-1,-.1755646063,-.1058742437e-1,-.1046114740e-3,-.5625109532e-1,-.6360817804e-2,-.1759148700e-2,-.1069636429,-.1989227818e-3,-.4621846472e-6,-.4015360274e-2,.3324626528e-3,.2179945987e-4,-.1568097578e-1,-.1804945451e-5,-.3191290614e-3,-.2640669295e-1,-.2374368331e-1,-.1440801715e-2,-.1071978198,-.1295502609e-2,-.6011507348e-5,-.2251959335e-2,-.3991407866e-2,-.3580859223e-4,-.2377551727e-1,-.6346770766e-4},
			{-.1125145334,-.5667284643e-5,-.5108221507e-1,-.5380496754e-3,-.3347421288e-1,-.1910724920,-.7997500406e-1,-.2862088080e-2,-.1814695038,-.2278994742e-1,-.1768516141,-.3800830512e-1,-.8198676790e-1,-.1426069896,-.6298450945e-2,-.2997003719e-1,-.5006673611e-1,-.3955086483e-3,-.4449924395e-2,-.3601637417e-2,-.1061324443e-5,-.6182641688e-4,-.6757166343e-2,-.1120498440,-.7916132626e-1,-.1934696698e-1,-.2266529321,-.1366830609e-1,-.1350528323e-3,-.7261985182e-1,-.8211780472e-2,-.2271051205e-2,-.1380894693,-.2568082070e-3,-.5966778137e-6,-.5183807822e-2,.4292074393e-3,.2814298168e-4,-.2024405269e-1,-.2330174558e-5,-.4119938467e-3,-.3409089400e-1,-.3065296333e-1,-.1860067016e-2,-.1383917903,-.1672486677e-2,-.7760822615e-5,-.2907267000e-2,-.5152885397e-2,-.4622869379e-4,-.3069406081e-1,-.8193645830e-4},
			{-.4593386590e-1,-.3452307302e-5,-.3111746013e-1,-.3452307302e-5,-.1695115978e-1,-.9675807323e-1,-.1695115978e-1,-.1649111558e-2,-.1045612321,-.1649111558e-2,-.5210348698e-1,-.1119789175e-1,-.5210348698e-1,-.2584328822e-1,-.1141407470e-2,-.2584328822e-1,-.4627448202e-2,-.3655512474e-4,-.4627448202e-2,-.7320817509e-4,-.2157286163e-7,-.7320817509e-4,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4},
			{-.7955978954e-1,-.5979571650e-5,-.5389702195e-1,-.5979571650e-5,-.2936026999e-1,-.1675898989,-.2936026999e-1,-.2856345005e-2,-.1811053665,-.2856345005e-2,-.9024588671e-1,-.1939531744e-1,-.9024588671e-1,-.4476188824e-1,-.1976975730e-2,-.4476188824e-1,-.8014975395e-2,-.6331533334e-4,-.8014975395e-2,-.1268002788e-3,-.3736529241e-7,-.1268002788e-3,-.4250861857e-2,-.1425849616,-.1739310676e-1,-.4250861857e-2,-.1425849616,-.1739310676e-1,-.9675438873e-4,-.9892989258e-1,-.3498493552e-3,-.9675438873e-4,-.9892989258e-1,-.3498493552e-3,-.2591590698e-6,-.8792734945e-2,-.3952424159e-5,-.2591590698e-6,-.8792734945e-2,-.3952424159e-5,-.1654938171e-3,-.5559060120e-1,-.2727254377e-2,-.1654938171e-3,-.5559060120e-1,-.2727254377e-2,-.1404040445e-5,-.5552981292e-2,-.1565014910e-3,-.1404040445e-5,-.5552981292e-2,-.1565014910e-3},
			{-.3211722999e-1,-.3601841323e-5,-.3246528886e-1,-.3793824429e-7,-.1470175463e-1,-.8391835536e-1,-.6153543472e-2,-.1627413015e-2,-.1031854450,-.2043795582e-3,-.2629088823e-1,-.5650341992e-2,-.5671141990e-1,-.8021127855e-2,-.3542651063e-3,-.3816708301e-1,-.7325116912e-3,-.5786570712e-5,-.8241593859e-2,-.2548589453e-5,-.7510140431e-9,-.1484655847e-3,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3},
			{-.1299205923e-1,-.6892821735e-8,-.5898466126e-2,-.5898466126e-2,-.1617841028e-2,-.2206315093e-1,-.2206315093e-1,-.4150419878e-4,-.2095429338e-1,-.2095429338e-1,-.4404977136e-1,-.4388821038e-2,-.4388821038e-2,-.7835445799e-1,-.7272824702e-3,-.7272824702e-3,-.6504520647e-1,-.4566940490e-4,-.4566940490e-4,-.2422675624e-1,-.1225511905e-6,-.1225511905e-6,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6},
			{-.2250290667e-1,-.1193871745e-7,-.1021644301e-1,-.1021644301e-1,-.2802182859e-2,-.3821449838e-1,-.3821449838e-1,-.7188738101e-4,-.3629390077e-1,-.3629390077e-1,-.7629644205e-1,-.7601661022e-2,-.7601661022e-2,-.1357139022,-.1259690190e-2,-.1259690190e-2,-.1126616024,-.7910172963e-4,-.7910172963e-4,-.4196197271e-1,-.2122648885e-6,-.2122648885e-6,-.4720047033e-3,-.1107876030e-1,-.9169410646e-1,-.1107876030e-1,-.9169410646e-1,-.4720047033e-3,-.1606240095e-5,-.7638008748e-2,-.5251649806e-1,-.7638008748e-2,-.5251649806e-1,-.1606240095e-5,.2530111540e-8,-.2654790910e-3,-.1581160736e-1,-.2654790910e-3,-.1581160736e-1,.2530111540e-8,-.1825084025e-4,-.1679563581e-2,-.1123601370,-.1679563581e-2,-.1123601370,-.1825084025e-4,-.2605756842e-6,-.5507385590e-4,-.6481175407e-1,-.5507385590e-4,-.6481175407e-1,-.2605756842e-6},
			{-.2905112760e-1,-.1541281795e-7,-.1318937122e-1,-.1318937122e-1,-.3617602516e-2,-.4933470527e-1,-.4933470527e-1,-.9280620982e-4,-.4685522442e-1,-.4685522442e-1,-.9849828315e-1,-.9813702180e-2,-.9813702180e-2,-.1752058944,-.1626253042e-2,-.1626253042e-2,-.1454455033,-.1021198938e-3,-.1021198938e-3,-.5417267383e-1,-.2740327927e-6,-.2740327927e-6,-.6093554517e-3,-.1430261805e-1,-.1183765824,-.1430261805e-1,-.1183765824,-.6093554517e-3,-.2073647046e-5,-.9860626894e-2,-.6779850746e-1,-.9860626894e-2,-.6779850746e-1,-.2073647046e-5,.3266359953e-8,-.3427320328e-3,-.2041269733e-1,-.3427320328e-3,-.2041269733e-1,.3266359953e-8,-.2356173345e-4,-.2168307259e-2,-.1450563132,-.2168307259e-2,-.1450563132,-.2356173345e-4,-.3364017618e-6,-.7110004225e-4,-.8367161471e-1,-.7110004225e-4,-.8367161471e-1,-.3364017618e-6},
			{-.2905112760e-1,-.2299810724e-7,-.1968040980e-1,-.2072940277e-3,-.4487304244e-2,-.6119517866e-1,-.2561375849e-1,-.1309844384e-3,-.6613032970e-1,-.8305013821e-2,-.7108239413e-1,-.7082168586e-2,-.1527676939e-1,-.7777343793e-1,-.7218894686e-3,-.3434980188e-2,-.3292821660e-1,-.2311949087e-4,-.2601206999e-3,-.2697214518e-2,-.1364387568e-7,-.7948106270e-6,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,.2006344762e-8,-.8221381159e-3,.2658347174e-3,.4463401949e-5,-.1253839410e-1,.7835272927e-8,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6},
			{-.5031802903e-1,-.3983389024e-7,-.3408746970e-1,-.3590437882e-3,-.7772238941e-2,-.1059931586,-.4436433109e-1,-.2268717024e-3,-.1145410910,-.1438470590e-1,-.1231183182,-.1226667582e-1,-.2646014076e-1,-.1347075460,-.1250349237e-2,-.5949560210e-2,-.5703334418e-1,-.4004413284e-4,-.4505422684e-3,-.4671712585e-2,-.2363188589e-7,-.1376652389e-5,-.9389851147e-3,-.4458139289e-1,-.6370969475e-1,-.7697598727e-2,-.1824121677,-.1899366584e-2,-.3638961323e-5,-.3290427534e-1,-.7075212788e-2,-.1029020197e-2,-.1189769237,-.6919626319e-5,.3475091067e-8,-.1423984988e-2,.4604392371e-3,.7730838953e-5,-.2171713563e-1,.1357109081e-7,-.2318322805e-4,-.8660828315e-2,-.3161298039e-1,-.4725520274e-3,-.1427260688,-.9411218238e-4,-.1490753193e-6,-.3326494240e-3,-.6224747955e-2,-.5289486092e-5,-.3707879714e-1,-.1573892910e-5},
			{-.3211722999e-1,-.3793824429e-7,-.3246528886e-1,-.3601841323e-5,-.6153543472e-2,-.8391835536e-1,-.1470175463e-1,-.2043795582e-3,-.1031854450,-.1627413015e-2,-.5671141990e-1,-.5650341992e-2,-.2629088823e-1,-.3816708301e-1,-.3542651063e-3,-.8021127855e-2,-.8241593859e-2,-.5786570712e-5,-.7325116912e-3,-.1484655847e-3,-.7510140431e-9,-.2548589453e-5,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5},
			{-.5490147066e-2,-.4577893991e-10,-.3719247860e-2,-.3719247860e-2,-.3549462592e-3,-.1156480173e-1,-.1156480173e-1,-.3108711060e-5,-.1249745760e-1,-.1249745760e-1,-.2897665788e-1,-.1338404058e-2,-.1338404058e-2,-.6993680883e-1,-.1364242863e-3,-.1364242863e-3,-.7001415346e-1,-.4369173091e-5,-.4369173091e-5,-.2969356790e-1,-.2578450140e-8,-.2578450140e-8,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8},
			{-.9509213660e-2,-.7929144985e-10,-.6441926259e-2,-.6441926259e-2,-.6147849549e-3,-.2003082417e-1,-.2003082417e-1,-.5384445502e-5,-.2164623154e-1,-.2164623154e-1,-.5018904368e-1,-.2318183830e-2,-.2318183830e-2,-.1211341062,-.2362937952e-3,-.2362937952e-3,-.1212680711,-.7567629781e-5,-.7567629781e-5,-.5143076827e-1,-.4466006648e-8,-.4466006648e-8,-.6197712293e-4,-.4165091360e-2,-.6973083633e-1,-.4165091360e-2,-.6973083633e-1,-.6197712293e-4,-.4089546907e-7,-.3270153304e-2,-.4275519802e-1,-.3270153304e-2,-.4275519802e-1,-.4089546907e-7,-.1392377369e-10,-.6890929850e-4,-.1602773282e-1,-.6890929850e-4,-.1602773282e-1,-.1392377369e-10,-.9704139476e-6,-.4031888160e-3,-.1094954884,-.4031888160e-3,-.1094954884,-.9704139476e-6,-.4729583171e-8,-.5954404436e-5,-.7398029375e-1,-.5954404436e-5,-.7398029375e-1,-.4729583171e-8},
			{-.1486738854e-1,-.1849807308e-9,-.1502850849e-1,-.1582954870e-3,-.1192279250e-2,-.3884665007e-1,-.1625959323e-1,-.1188156743e-4,-.4776557949e-1,-.5998666568e-2,-.5662815244e-1,-.2615600094e-2,-.5642045787e-2,-.8406962479e-1,-.1639928781e-3,-.7803303852e-3,-.4292442302e-1,-.2678661737e-5,-.3013800652e-4,-.4003578188e-2,-.3476519485e-9,-.2025212407e-7,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7},
			{-.2216299428e-2,-.2904510069e-12,-.2240317772e-2,-.2240317772e-2,-.7439227355e-4,-.5790916683e-2,-.5790916683e-2,-.2224370628e-6,-.7120472184e-2,-.7120472184e-2,-.1820922229e-1,-.3899106406e-3,-.3899106406e-3,-.5963296216e-1,-.2444661488e-4,-.2444661488e-4,-.7199376027e-1,-.3993113157e-6,-.3993113157e-6,-.3476706016e-1,-.5182489262e-10,-.5182489262e-10,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 52; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_top_fluxes_to_dofs(
		const T i_input[52],
		T o_output[56]
	) {
		static const T matrix[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5412739307e-1,-.1103717978e-1,-.1103717978e-1,-.1162547668e-3,-.5974166429e-1,-.5974166429e-1,-.2500537776e-1,-.4382549273e-1,-.4382549273e-1,-.5503848602e-2,-.3510914219e-1,-.3510914219e-1,-.7573305578e-1,-.1537719318e-1,-.1537719318e-1,-.7316958648e-1,-.3712176777e-2,-.3712176777e-2,-.4176623208e-1,-.2059615826e-3,-.2059615826e-3,-.1199809046e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2852764098e-1,-.8679943781e-2,-.8679943781e-2,-.9629909754e-6,-.3905626741e-1,-.3905626741e-1,-.6842313070e-2,-.3260004675e-1,-.3260004675e-1,-.5141591464e-3,-.1335372273e-1,-.1335372273e-1,-.6213451013e-1,-.3597562967e-2,-.3597562967e-2,-.8145457175e-1,-.4429404170e-3,-.4429404170e-3,-.5607103930e-1,-.5404682501e-5,-.5404682501e-5,-.1834095772e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1299205923e-1,-.5898466126e-2,-.5898466126e-2,-.6892821735e-8,-.2206315093e-1,-.2206315093e-1,-.1617841028e-2,-.2095429338e-1,-.2095429338e-1,-.4150419878e-4,-.4388821038e-2,-.4388821038e-2,-.4404977136e-1,-.7272824702e-3,-.7272824702e-3,-.7835445799e-1,-.4566940490e-4,-.4566940490e-4,-.6504520647e-1,-.1225511905e-6,-.1225511905e-6,-.2422675624e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5490147066e-2,-.3719247860e-2,-.3719247860e-2,-.4577893991e-10,-.1156480173e-1,-.1156480173e-1,-.3549462592e-3,-.1249745760e-1,-.1249745760e-1,-.3108711060e-5,-.1338404058e-2,-.1338404058e-2,-.2897665788e-1,-.1364242863e-3,-.1364242863e-3,-.6993680883e-1,-.4369173091e-5,-.4369173091e-5,-.7001415346e-1,-.2578450140e-8,-.2578450140e-8,-.2969356790e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2216299428e-2,-.2240317772e-2,-.2240317772e-2,-.2904510069e-12,-.5790916683e-2,-.5790916683e-2,-.7439227355e-4,-.7120472184e-2,-.7120472184e-2,-.2224370628e-6,-.3899106406e-3,-.3899106406e-3,-.1820922229e-1,-.2444661488e-4,-.2444661488e-4,-.5963296216e-1,-.3993113157e-6,-.3993113157e-6,-.7199376027e-1,-.5182489262e-10,-.5182489262e-10,-.3476706016e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1},
			{-.5412739307e-1,-.1162547668e-3,-.1103717978e-1,-.1103717978e-1,-.2500537776e-1,-.5974166429e-1,-.5974166429e-1,-.5503848602e-2,-.4382549273e-1,-.4382549273e-1,-.7573305578e-1,-.3510914219e-1,-.3510914219e-1,-.7316958648e-1,-.1537719318e-1,-.1537719318e-1,-.4176623208e-1,-.3712176777e-2,-.3712176777e-2,-.1199809046e-1,-.2059615826e-3,-.2059615826e-3,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.6987816394e-1,-.2239469694e-3,-.2126143325e-1,-.2239469694e-3,-.4004261797e-1,-.9566792642e-1,-.4004261797e-1,-.1002844321e-1,-.7985348011e-1,-.1002844321e-1,-.7055750928e-1,-.3270980684e-1,-.7055750928e-1,-.4193122589e-1,-.8812193584e-2,-.4193122589e-1,-.1220724282e-1,-.1084978008e-2,-.1220724282e-1,-.7712083501e-3,-.1323871434e-4,-.7712083501e-3,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,.2845518738e-4,-.5241314676e-2,.1111245502e-3,.2845518738e-4,-.5241314676e-2,.1111245502e-3,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5031802903e-1,-.2406231299e-3,-.2284466108e-1,-.2534486742e-5,-.3576590913e-1,-.8545021612e-1,-.1497012310e-1,-.1019197433e-1,-.8115562931e-1,-.1279964701e-2,-.3666559727e-1,-.1699783079e-1,-.7909044629e-1,-.1340300809e-1,-.2816752894e-2,-.6377578460e-1,-.1990066688e-2,-.1768768448e-3,-.2239052508e-1,-.2764961419e-4,-.4746387201e-6,-.1610701220e-2,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2905112760e-1,-.2072940277e-3,-.1968040980e-1,-.2299810724e-7,-.2561375849e-1,-.6119517866e-1,-.4487304244e-2,-.8305013821e-2,-.6613032970e-1,-.1309844384e-3,-.1527676939e-1,-.7082168586e-2,-.7108239413e-1,-.3434980188e-2,-.7218894686e-3,-.7777343793e-1,-.2601206999e-3,-.2311949087e-4,-.3292821660e-1,-.7948106270e-6,-.1364387568e-7,-.2697214518e-2,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,.4463401949e-5,-.1253839410e-1,.7835272927e-8,.2006344762e-8,-.8221381159e-3,.2658347174e-3,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1486738854e-1,-.1582954870e-3,-.1502850849e-1,-.1849807308e-9,-.1625959323e-1,-.3884665007e-1,-.1192279250e-2,-.5998666568e-2,-.4776557949e-1,-.1188156743e-4,-.5642045787e-2,-.2615600094e-2,-.5662815244e-1,-.7803303852e-3,-.1639928781e-3,-.8406962479e-1,-.3013800652e-4,-.2678661737e-5,-.4292442302e-1,-.2025212407e-7,-.3476519485e-9,-.4003578188e-2,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2},
			{-.2852764098e-1,-.9629909754e-6,-.8679943781e-2,-.8679943781e-2,-.6842313070e-2,-.3905626741e-1,-.3905626741e-1,-.5141591464e-3,-.3260004675e-1,-.3260004675e-1,-.6213451013e-1,-.1335372273e-1,-.1335372273e-1,-.8145457175e-1,-.3597562967e-2,-.3597562967e-2,-.5607103930e-1,-.4429404170e-3,-.4429404170e-3,-.1834095772e-1,-.5404682501e-5,-.5404682501e-5,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5031802903e-1,-.2534486742e-5,-.2284466108e-1,-.2406231299e-3,-.1497012310e-1,-.8545021612e-1,-.3576590913e-1,-.1279964701e-2,-.8115562931e-1,-.1019197433e-1,-.7909044629e-1,-.1699783079e-1,-.3666559727e-1,-.6377578460e-1,-.2816752894e-2,-.1340300809e-1,-.2239052508e-1,-.1768768448e-3,-.1990066688e-2,-.1610701220e-2,-.4746387201e-6,-.2764961419e-4,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.4593386590e-1,-.3452307302e-5,-.3111746013e-1,-.3452307302e-5,-.1695115978e-1,-.9675807323e-1,-.1695115978e-1,-.1649111558e-2,-.1045612321,-.1649111558e-2,-.5210348698e-1,-.1119789175e-1,-.5210348698e-1,-.2584328822e-1,-.1141407470e-2,-.2584328822e-1,-.4627448202e-2,-.3655512474e-4,-.4627448202e-2,-.7320817509e-4,-.2157286163e-7,-.7320817509e-4,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.3211722999e-1,-.3601841323e-5,-.3246528886e-1,-.3793824429e-7,-.1470175463e-1,-.8391835536e-1,-.6153543472e-2,-.1627413015e-2,-.1031854450,-.2043795582e-3,-.2629088823e-1,-.5650341992e-2,-.5671141990e-1,-.8021127855e-2,-.3542651063e-3,-.3816708301e-1,-.7325116912e-3,-.5786570712e-5,-.8241593859e-2,-.2548589453e-5,-.7510140431e-9,-.1484655847e-3,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3},
			{-.1299205923e-1,-.6892821735e-8,-.5898466126e-2,-.5898466126e-2,-.1617841028e-2,-.2206315093e-1,-.2206315093e-1,-.4150419878e-4,-.2095429338e-1,-.2095429338e-1,-.4404977136e-1,-.4388821038e-2,-.4388821038e-2,-.7835445799e-1,-.7272824702e-3,-.7272824702e-3,-.6504520647e-1,-.4566940490e-4,-.4566940490e-4,-.2422675624e-1,-.1225511905e-6,-.1225511905e-6,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2905112760e-1,-.2299810724e-7,-.1968040980e-1,-.2072940277e-3,-.4487304244e-2,-.6119517866e-1,-.2561375849e-1,-.1309844384e-3,-.6613032970e-1,-.8305013821e-2,-.7108239413e-1,-.7082168586e-2,-.1527676939e-1,-.7777343793e-1,-.7218894686e-3,-.3434980188e-2,-.3292821660e-1,-.2311949087e-4,-.2601206999e-3,-.2697214518e-2,-.1364387568e-7,-.7948106270e-6,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,.2006344762e-8,-.8221381159e-3,.2658347174e-3,.4463401949e-5,-.1253839410e-1,.7835272927e-8,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.3211722999e-1,-.3793824429e-7,-.3246528886e-1,-.3601841323e-5,-.6153543472e-2,-.8391835536e-1,-.1470175463e-1,-.2043795582e-3,-.1031854450,-.1627413015e-2,-.5671141990e-1,-.5650341992e-2,-.2629088823e-1,-.3816708301e-1,-.3542651063e-3,-.8021127855e-2,-.8241593859e-2,-.5786570712e-5,-.7325116912e-3,-.1484655847e-3,-.7510140431e-9,-.2548589453e-5,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5},
			{-.5490147066e-2,-.4577893991e-10,-.3719247860e-2,-.3719247860e-2,-.3549462592e-3,-.1156480173e-1,-.1156480173e-1,-.3108711060e-5,-.1249745760e-1,-.1249745760e-1,-.2897665788e-1,-.1338404058e-2,-.1338404058e-2,-.6993680883e-1,-.1364242863e-3,-.1364242863e-3,-.7001415346e-1,-.4369173091e-5,-.4369173091e-5,-.2969356790e-1,-.2578450140e-8,-.2578450140e-8,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1486738854e-1,-.1849807308e-9,-.1502850849e-1,-.1582954870e-3,-.1192279250e-2,-.3884665007e-1,-.1625959323e-1,-.1188156743e-4,-.4776557949e-1,-.5998666568e-2,-.5662815244e-1,-.2615600094e-2,-.5642045787e-2,-.8406962479e-1,-.1639928781e-3,-.7803303852e-3,-.4292442302e-1,-.2678661737e-5,-.3013800652e-4,-.4003578188e-2,-.3476519485e-9,-.2025212407e-7,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7},
			{-.2216299428e-2,-.2904510069e-12,-.2240317772e-2,-.2240317772e-2,-.7439227355e-4,-.5790916683e-2,-.5790916683e-2,-.2224370628e-6,-.7120472184e-2,-.7120472184e-2,-.1820922229e-1,-.3899106406e-3,-.3899106406e-3,-.5963296216e-1,-.2444661488e-4,-.2444661488e-4,-.7199376027e-1,-.3993113157e-6,-.3993113157e-6,-.3476706016e-1,-.5182489262e-10,-.5182489262e-10,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 52; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_top_fluxes_to_dofs(
		const T i_input[52],
		T o_output[56]
	) {
		static const T matrix[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5412739307e-1,-.1103717978e-1,-.1103717978e-1,-.1162547668e-3,-.5974166429e-1,-.5974166429e-1,-.2500537776e-1,-.4382549273e-1,-.4382549273e-1,-.5503848602e-2,-.3510914219e-1,-.3510914219e-1,-.7573305578e-1,-.1537719318e-1,-.1537719318e-1,-.7316958648e-1,-.3712176777e-2,-.3712176777e-2,-.4176623208e-1,-.2059615826e-3,-.2059615826e-3,-.1199809046e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2852764098e-1,-.8679943781e-2,-.8679943781e-2,-.9629909754e-6,-.3905626741e-1,-.3905626741e-1,-.6842313070e-2,-.3260004675e-1,-.3260004675e-1,-.5141591464e-3,-.1335372273e-1,-.1335372273e-1,-.6213451013e-1,-.3597562967e-2,-.3597562967e-2,-.8145457175e-1,-.4429404170e-3,-.4429404170e-3,-.5607103930e-1,-.5404682501e-5,-.5404682501e-5,-.1834095772e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1299205923e-1,-.5898466126e-2,-.5898466126e-2,-.6892821735e-8,-.2206315093e-1,-.2206315093e-1,-.1617841028e-2,-.2095429338e-1,-.2095429338e-1,-.4150419878e-4,-.4388821038e-2,-.4388821038e-2,-.4404977136e-1,-.7272824702e-3,-.7272824702e-3,-.7835445799e-1,-.4566940490e-4,-.4566940490e-4,-.6504520647e-1,-.1225511905e-6,-.1225511905e-6,-.2422675624e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5490147066e-2,-.3719247860e-2,-.3719247860e-2,-.4577893991e-10,-.1156480173e-1,-.1156480173e-1,-.3549462592e-3,-.1249745760e-1,-.1249745760e-1,-.3108711060e-5,-.1338404058e-2,-.1338404058e-2,-.2897665788e-1,-.1364242863e-3,-.1364242863e-3,-.6993680883e-1,-.4369173091e-5,-.4369173091e-5,-.7001415346e-1,-.2578450140e-8,-.2578450140e-8,-.2969356790e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2216299428e-2,-.2240317772e-2,-.2240317772e-2,-.2904510069e-12,-.5790916683e-2,-.5790916683e-2,-.7439227355e-4,-.7120472184e-2,-.7120472184e-2,-.2224370628e-6,-.3899106406e-3,-.3899106406e-3,-.1820922229e-1,-.2444661488e-4,-.2444661488e-4,-.5963296216e-1,-.3993113157e-6,-.3993113157e-6,-.7199376027e-1,-.5182489262e-10,-.5182489262e-10,-.3476706016e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1},
			{-.5412739307e-1,-.1162547668e-3,-.1103717978e-1,-.1103717978e-1,-.2500537776e-1,-.5974166429e-1,-.5974166429e-1,-.5503848602e-2,-.4382549273e-1,-.4382549273e-1,-.7573305578e-1,-.3510914219e-1,-.3510914219e-1,-.7316958648e-1,-.1537719318e-1,-.1537719318e-1,-.4176623208e-1,-.3712176777e-2,-.3712176777e-2,-.1199809046e-1,-.2059615826e-3,-.2059615826e-3,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.6987816394e-1,-.2239469694e-3,-.2126143325e-1,-.2239469694e-3,-.4004261797e-1,-.9566792642e-1,-.4004261797e-1,-.1002844321e-1,-.7985348011e-1,-.1002844321e-1,-.7055750928e-1,-.3270980684e-1,-.7055750928e-1,-.4193122589e-1,-.8812193584e-2,-.4193122589e-1,-.1220724282e-1,-.1084978008e-2,-.1220724282e-1,-.7712083501e-3,-.1323871434e-4,-.7712083501e-3,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,.2845518738e-4,-.5241314676e-2,.1111245502e-3,.2845518738e-4,-.5241314676e-2,.1111245502e-3,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5031802903e-1,-.2406231299e-3,-.2284466108e-1,-.2534486742e-5,-.3576590913e-1,-.8545021612e-1,-.1497012310e-1,-.1019197433e-1,-.8115562931e-1,-.1279964701e-2,-.3666559727e-1,-.1699783079e-1,-.7909044629e-1,-.1340300809e-1,-.2816752894e-2,-.6377578460e-1,-.1990066688e-2,-.1768768448e-3,-.2239052508e-1,-.2764961419e-4,-.4746387201e-6,-.1610701220e-2,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2905112760e-1,-.2072940277e-3,-.1968040980e-1,-.2299810724e-7,-.2561375849e-1,-.6119517866e-1,-.4487304244e-2,-.8305013821e-2,-.6613032970e-1,-.1309844384e-3,-.1527676939e-1,-.7082168586e-2,-.7108239413e-1,-.3434980188e-2,-.7218894686e-3,-.7777343793e-1,-.2601206999e-3,-.2311949087e-4,-.3292821660e-1,-.7948106270e-6,-.1364387568e-7,-.2697214518e-2,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,.4463401949e-5,-.1253839410e-1,.7835272927e-8,.2006344762e-8,-.8221381159e-3,.2658347174e-3,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1486738854e-1,-.1582954870e-3,-.1502850849e-1,-.1849807308e-9,-.1625959323e-1,-.3884665007e-1,-.1192279250e-2,-.5998666568e-2,-.4776557949e-1,-.1188156743e-4,-.5642045787e-2,-.2615600094e-2,-.5662815244e-1,-.7803303852e-3,-.1639928781e-3,-.8406962479e-1,-.3013800652e-4,-.2678661737e-5,-.4292442302e-1,-.2025212407e-7,-.3476519485e-9,-.4003578188e-2,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2},
			{-.2852764098e-1,-.9629909754e-6,-.8679943781e-2,-.8679943781e-2,-.6842313070e-2,-.3905626741e-1,-.3905626741e-1,-.5141591464e-3,-.3260004675e-1,-.3260004675e-1,-.6213451013e-1,-.1335372273e-1,-.1335372273e-1,-.8145457175e-1,-.3597562967e-2,-.3597562967e-2,-.5607103930e-1,-.4429404170e-3,-.4429404170e-3,-.1834095772e-1,-.5404682501e-5,-.5404682501e-5,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5031802903e-1,-.2534486742e-5,-.2284466108e-1,-.2406231299e-3,-.1497012310e-1,-.8545021612e-1,-.3576590913e-1,-.1279964701e-2,-.8115562931e-1,-.1019197433e-1,-.7909044629e-1,-.1699783079e-1,-.3666559727e-1,-.6377578460e-1,-.2816752894e-2,-.1340300809e-1,-.2239052508e-1,-.1768768448e-3,-.1990066688e-2,-.1610701220e-2,-.4746387201e-6,-.2764961419e-4,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.4593386590e-1,-.3452307302e-5,-.3111746013e-1,-.3452307302e-5,-.1695115978e-1,-.9675807323e-1,-.1695115978e-1,-.1649111558e-2,-.1045612321,-.1649111558e-2,-.5210348698e-1,-.1119789175e-1,-.5210348698e-1,-.2584328822e-1,-.1141407470e-2,-.2584328822e-1,-.4627448202e-2,-.3655512474e-4,-.4627448202e-2,-.7320817509e-4,-.2157286163e-7,-.7320817509e-4,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.3211722999e-1,-.3601841323e-5,-.3246528886e-1,-.3793824429e-7,-.1470175463e-1,-.8391835536e-1,-.6153543472e-2,-.1627413015e-2,-.1031854450,-.2043795582e-3,-.2629088823e-1,-.5650341992e-2,-.5671141990e-1,-.8021127855e-2,-.3542651063e-3,-.3816708301e-1,-.7325116912e-3,-.5786570712e-5,-.8241593859e-2,-.2548589453e-5,-.7510140431e-9,-.1484655847e-3,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3},
			{-.1299205923e-1,-.6892821735e-8,-.5898466126e-2,-.5898466126e-2,-.1617841028e-2,-.2206315093e-1,-.2206315093e-1,-.4150419878e-4,-.2095429338e-1,-.2095429338e-1,-.4404977136e-1,-.4388821038e-2,-.4388821038e-2,-.7835445799e-1,-.7272824702e-3,-.7272824702e-3,-.6504520647e-1,-.4566940490e-4,-.4566940490e-4,-.2422675624e-1,-.1225511905e-6,-.1225511905e-6,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2905112760e-1,-.2299810724e-7,-.1968040980e-1,-.2072940277e-3,-.4487304244e-2,-.6119517866e-1,-.2561375849e-1,-.1309844384e-3,-.6613032970e-1,-.8305013821e-2,-.7108239413e-1,-.7082168586e-2,-.1527676939e-1,-.7777343793e-1,-.7218894686e-3,-.3434980188e-2,-.3292821660e-1,-.2311949087e-4,-.2601206999e-3,-.2697214518e-2,-.1364387568e-7,-.7948106270e-6,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,.2006344762e-8,-.8221381159e-3,.2658347174e-3,.4463401949e-5,-.1253839410e-1,.7835272927e-8,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.3211722999e-1,-.3793824429e-7,-.3246528886e-1,-.3601841323e-5,-.6153543472e-2,-.8391835536e-1,-.1470175463e-1,-.2043795582e-3,-.1031854450,-.1627413015e-2,-.5671141990e-1,-.5650341992e-2,-.2629088823e-1,-.3816708301e-1,-.3542651063e-3,-.8021127855e-2,-.8241593859e-2,-.5786570712e-5,-.7325116912e-3,-.1484655847e-3,-.7510140431e-9,-.2548589453e-5,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5},
			{-.5490147066e-2,-.4577893991e-10,-.3719247860e-2,-.3719247860e-2,-.3549462592e-3,-.1156480173e-1,-.1156480173e-1,-.3108711060e-5,-.1249745760e-1,-.1249745760e-1,-.2897665788e-1,-.1338404058e-2,-.1338404058e-2,-.6993680883e-1,-.1364242863e-3,-.1364242863e-3,-.7001415346e-1,-.4369173091e-5,-.4369173091e-5,-.2969356790e-1,-.2578450140e-8,-.2578450140e-8,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1486738854e-1,-.1849807308e-9,-.1502850849e-1,-.1582954870e-3,-.1192279250e-2,-.3884665007e-1,-.1625959323e-1,-.1188156743e-4,-.4776557949e-1,-.5998666568e-2,-.5662815244e-1,-.2615600094e-2,-.5642045787e-2,-.8406962479e-1,-.1639928781e-3,-.7803303852e-3,-.4292442302e-1,-.2678661737e-5,-.3013800652e-4,-.4003578188e-2,-.3476519485e-9,-.2025212407e-7,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7},
			{-.2216299428e-2,-.2904510069e-12,-.2240317772e-2,-.2240317772e-2,-.7439227355e-4,-.5790916683e-2,-.5790916683e-2,-.2224370628e-6,-.7120472184e-2,-.7120472184e-2,-.1820922229e-1,-.3899106406e-3,-.3899106406e-3,-.5963296216e-1,-.2444661488e-4,-.2444661488e-4,-.7199376027e-1,-.3993113157e-6,-.3993113157e-6,-.3476706016e-1,-.5182489262e-10,-.5182489262e-10,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 52; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 362880;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[56][56] = 
		{
			{1959.99950,-5383.2139,9015.826058,-8519.319340,4200.00000,-835.7894471,-6584.2284,10097.4849,-9661.82590,4925.362810,-1028.78570,13230.11101,-11457.6123,5819.69066,-1229.634099,-14816.2076,6782.212097,-1419.859143,8640.1667,-1626.65301,-2047.25769,-6584.2284,10097.4849,-9661.82590,4925.362810,-1028.78570,10802.34025,-9355.10139,4751.757564,-1003.99205,-11476.58483,5253.47885,-1099.81816,6531.35214,-1229.63410,-1525.93584,13230.11101,-11457.6123,5819.69066,-1229.634099,-11476.58483,5253.47885,-1099.81816,6196.18489,-1166.53333,-1412.74207,-14816.2076,6782.212097,-1419.859143,6531.35214,-1229.63410,-1412.74207,8640.1667,-1626.65301,-1525.93584,-2047.25769},
			{-5383.2139,33096.0000,-83609.96448,103016.3016,-61834.21385,14476.29787,10097.4848,-40465.5705,63751.51762,-44899.88862,11879.39392,-11457.61230,30581.07650,-27719.99999,8519.154887,6782.21205,-12064.62597,4918.536370,-1626.653003,1878.297097,0.,10097.4848,-40465.5705,63751.51762,-44899.88862,11879.39392,-9355.10127,24969.34437,-22633.28523,6955.860840,5253.47884,-9345.21910,3809.881889,-1229.634091,1419.85915,0.,-11457.61230,30581.07650,-27719.99999,8519.154887,5253.47884,-9345.21910,3809.881889,-1166.53333,1346.99667,0.,6782.21205,-12064.62597,4918.536370,-1229.634091,1419.85915,0.,-1626.653003,1878.297097,0.,0.},
			{9015.826058,-83609.96448,277200.,-416728.6598,293014.3477,-78493.04477,-9661.825918,63751.51762,-141972.4255,128683.4255,-41407.82533,5819.690715,-27720.,39039.67213,-16497.27251,-1419.859148,4918.536370,-3809.881888,0.,0.,0.,-9661.825918,63751.51762,-141972.4255,128683.4255,-41407.82533,4751.757568,-22633.28523,31875.75880,-13469.96659,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,5819.690715,-27720.,39039.67213,-16497.27251,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,-1419.859148,4918.536370,-3809.881888,0.,0.,0.,0.,0.,0.,0.},
			{-8519.319340,103016.3016,-416728.6598,734000.,-588944.2418,176903.2820,4925.362927,-44899.88862,128683.4255,-143050.2010,54438.22186,-1229.634092,8519.154886,-16497.27251,9295.160030,0.,0.,0.,0.,0.,0.,4925.362927,-44899.88862,128683.4255,-143050.2010,54438.22186,-1003.992031,6955.860840,-13469.96659,7589.466382,0.,0.,0.,0.,0.,0.,-1229.634092,8519.154886,-16497.27251,9295.160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4200.00000,-61834.21385,293014.3477,-588944.2418,529200.,-175515.7839,-1028.785692,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1028.785692,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-835.7894471,14476.29787,-78493.04477,176903.2820,-175515.7839,63504.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-6584.2284,10097.4848,-9661.825918,4925.362927,-1028.785692,0.,39536.0001,-37827.9896,22915.22463,-7196.44356,840.,-106783.7919,62111.73800,-21382.90905,3011.976097,144131.6731,-46695.88421,6955.860840,-95851.44763,13281.56616,25073.68342,19768.0002,-18913.9949,11457.61232,-3598.221779,420.,-58125.73395,33809.34781,-11639.38143,1639.512123,83732.93531,-27127.85727,4040.989977,-57965.50700,8031.93625,15574.01680,-35594.59728,20703.91267,-7127.636351,1003.992032,55821.95698,-18085.23818,2693.993319,-41243.18141,5714.82284,11534.99025,36032.9184,-11673.97107,1738.965210,-28982.75339,4015.96813,8651.24267,-19170.28946,2656.313229,6229.60671,4178.947267},
			{10097.4849,-40465.5705,63751.51762,-44899.88862,11879.39392,0.,-37827.9896,112000.0001,-123625.6285,58656.96888,-9699.484525,62111.73799,-135471.9749,92590.71229,-20867.58252,-46695.88421,74658.13284,-24095.80875,13281.56616,-15336.23160,0.,-18913.9949,56000.0000,-61812.81421,29328.48444,-4849.742262,33809.34782,-73741.60290,50399.99999,-11358.87319,-27127.85727,43372.45579,-13998.39991,8031.93625,-9274.48112,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,-18085.23818,28914.97050,-9332.266609,5714.82284,-6598.90901,0.,-11673.97107,18664.53323,-6023.952191,4015.96813,-4637.24056,0.,2656.313229,-3067.24632,0.,0.},
			{-9661.82590,63751.51762,-141972.4255,128683.4255,-41407.82533,0.,22915.22463,-123625.6285,216720.,-147665.3514,33809.34781,-21382.90906,92590.71229,-114752.7317,40409.89977,6955.860840,-24095.80875,18664.53321,0.,0.,0.,11457.61232,-61812.81421,108360.,-73832.67568,16904.67391,-11639.38143,50400.,-62463.47540,21996.36334,4040.989977,-13998.39991,10843.11394,0.,0.,0.,-7127.636352,30863.57076,-38250.91056,13469.96659,2693.993318,-9332.266605,7228.742626,0.,0.,0.,1738.965210,-6023.952189,4666.133303,0.,0.,0.,0.,0.,0.,0.},
			{4925.362810,-44899.88862,128683.4255,-143050.2010,54438.22186,0.,-7196.44356,58656.96888,-147665.3514,140800.,-44448.62202,3011.976094,-20867.58252,40409.89977,-22768.39914,0.,0.,0.,0.,0.,0.,-3598.221779,29328.48444,-73832.67568,70400.,-22224.31101,1639.512123,-11358.87318,21996.36334,-12393.54671,0.,0.,0.,0.,0.,0.,1003.992031,-6955.860840,13469.96659,-7589.466382,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1028.78570,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,840.,-9699.484525,33809.34781,-44448.62202,19600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,420.,-4849.742262,16904.67391,-22224.31101,9800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{13230.11101,-11457.61230,5819.690715,-1229.634092,0.,0.,-106783.7919,62111.73799,-21382.90906,3011.976094,0.,347760.0000,-130070.0875,27047.47825,-1904.940944,-535300.7081,116428.4501,-10998.18167,392829.1232,-37800.,-111005.9285,-35594.59728,20703.91267,-7127.636352,1003.992031,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-248785.2872,54111.05382,-5111.492932,197967.2700,-19049.40944,-59099.23855,57960.00000,-21678.34791,4507.913042,-317.4901573,-124392.6436,27055.52691,-2555.746466,112684.9455,-10843.11394,-36476.84197,-53530.07079,11642.84501,-1099.818167,59390.18104,-5714.822832,-21886.10519,26188.60820,-2520.,-11819.84771,-5285.996592},
			{-11457.6123,30581.07650,-27720.,8519.154886,0.,0.,62111.73800,-135471.9749,92590.71229,-20867.58252,0.,-130070.0875,230832.,-105407.1147,13197.81800,116428.4501,-167230.2366,38098.81888,-37800.,43647.68036,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,54111.05382,-77721.59051,17706.73093,-19049.40944,21996.36334,0.,-21678.34791,38472.,-17567.85246,2199.636334,27055.52691,-38860.79526,8853.365465,-10843.11394,12520.54951,0.,11642.84501,-16723.02366,3809.881888,-5714.822832,6598.909002,0.,-2520.,2909.845357,0.,0.},
			{5819.69066,-27719.99999,39039.67213,-16497.27251,0.,0.,-21382.90905,92590.71229,-114752.7317,40409.89977,0.,27047.47825,-105407.1147,108864.,-25557.46466,-10998.18167,38098.81888,-29511.21822,0.,0.,0.,-7127.636351,30863.57076,-38250.91056,13469.96659,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-5111.492932,17706.73093,-13715.57480,0.,0.,0.,4507.913042,-17567.85246,18144.,-4259.577443,-2555.746466,8853.365465,-6857.787398,0.,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,0.,0.,0.,0.},
			{-1229.634099,8519.154887,-16497.27251,9295.160030,0.,0.,3011.976097,-20867.58252,40409.89977,-22768.39914,0.,-1904.940944,13197.81800,-25557.46466,14400.,0.,0.,0.,0.,0.,0.,1003.992032,-6955.860840,13469.96659,-7589.466382,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,-317.4901573,2199.636334,-4259.577443,2400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-14816.2076,6782.21205,-1419.859148,0.,0.,0.,144131.6731,-46695.88421,6955.860840,0.,0.,-535300.7081,116428.4501,-10998.18167,0.,909200.,-119165.0956,5366.563145,-720578.5174,43037.19323,216661.3874,36032.9184,-11673.97107,1738.965210,0.,0.,-174828.4790,38025.37259,-3591.991091,0.,352131.6458,-46152.44305,2078.460970,-311260.6624,18590.32006,100931.1647,-53530.07079,11642.84501,-1099.818167,0.,140852.6583,-18460.97722,831.3843878,-147643.8959,8818.163075,53396.62909,45460.,-5958.254780,268.3281572,-62252.13248,3718.064012,26698.31455,-20587.95764,1229.634092,11534.99025,3868.953347},
			{6782.212097,-12064.62597,4918.536370,0.,0.,0.,-46695.88421,74658.13284,-24095.80875,0.,0.,116428.4501,-167230.2366,38098.81888,0.,-119165.0956,153600.,-18590.32006,43037.19323,-49695.07017,0.,-11673.97107,18664.53323,-6023.952189,0.,0.,38025.37259,-54617.16651,12443.02214,0.,-46152.44305,59489.02419,-7200.,18590.32006,-21466.25258,0.,11642.84501,-16723.02366,3809.881888,0.,-18460.97722,23795.60968,-2880.,8818.163075,-10182.33765,0.,-5958.254780,7680.,-929.5160030,3718.064012,-4293.250516,0.,1229.634092,-1419.859148,0.,0.},
			{-1419.859143,4918.536370,-3809.881888,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,5366.563145,-18590.32006,14400.,0.,0.,0.,1738.965210,-6023.952191,4666.133303,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,268.3281572,-929.5160030,720.,0.,0.,0.,0.,0.,0.,0.},
			{8640.1667,-1626.653003,0.,0.,0.,0.,-95851.44763,13281.56616,0.,0.,0.,392829.1232,-37800.,0.,0.,-720578.5174,43037.19323,0.,607600.,-16974.09792,-192267.9080,-19170.28946,2656.313229,0.,0.,0.,106914.5452,-10287.85692,0.,0.,-239210.4513,14287.05708,0.,229651.2138,-6415.605974,-79615.57635,26188.60820,-2520.,0.,0.,-79736.81711,4762.352360,0.,93371.25892,-2608.447815,-36854.85042,-20587.95764,1229.634092,0.,32807.31626,-916.5151392,-15794.93589,8680.,-242.4871131,-5686.826882,-1525.935778},
			{-1626.65301,1878.297097,0.,0.,0.,0.,13281.56616,-15336.23160,0.,0.,0.,-37800.,43647.68036,0.,0.,43037.19323,-49695.07017,0.,-16974.09792,19600.,0.,2656.313229,-3067.24632,0.,0.,0.,-10287.85692,11879.39392,0.,0.,14287.05708,-16497.27251,0.,-6415.605974,7408.103671,0.,-2520.,2909.845357,0.,0.,4762.352360,-5499.090835,0.,-2608.447815,3011.976094,0.,1229.634092,-1419.859148,0.,-916.5151392,1058.300524,0.,-242.4871131,280.,0.,0.},
			{-2047.25769,0.,0.,0.,0.,0.,25073.68342,0.,0.,0.,0.,-111005.9285,0.,0.,0.,216661.3874,0.,0.,-192267.9080,0.,63504.,4178.947267,0.,0.,0.,0.,-25895.98888,0.,0.,0.,62934.44585,0.,0.,-64595.94539,0.,23666.54347,-5285.996592,0.,0.,0.,17981.27024,0.,0.,-22980.41043,0.,9738.205176,3868.953347,0.,0.,-6920.994149,0.,3651.826941,-1525.935778,0.,1126.978260,252.},
			{-6584.2284,10097.4848,-9661.825918,4925.362927,-1028.785692,0.,19768.0002,-18913.9949,11457.61232,-3598.221779,420.,-35594.59728,20703.91267,-7127.636351,1003.992032,36032.9184,-11673.97107,1738.965210,-19170.28946,2656.313232,4178.947267,39536.0001,-37827.9896,22915.22463,-7196.44356,840.,-58125.73395,33809.34781,-11639.38143,1639.512123,55821.95698,-18085.23818,2693.993319,-28982.75339,4015.968128,6229.60671,-106783.7919,62111.73800,-21382.90905,3011.976097,83732.93531,-27127.85727,4040.989977,-41243.18141,5714.82284,8651.24267,144131.6731,-46695.88421,6955.860840,-57965.50700,8031.936250,11534.99025,-95851.44763,13281.56616,15574.01680,25073.68342},
			{10097.4849,-40465.5705,63751.51762,-44899.88862,11879.39392,0.,-18913.9949,56000.0000,-61812.81421,29328.48444,-4849.742262,20703.91267,-45157.32496,30863.57076,-6955.860840,-11673.97107,18664.53323,-6023.952191,2656.313229,-3067.24632,0.,-37827.9896,112000.0001,-123625.6285,58656.96888,-9699.484525,33809.34782,-73741.60290,50399.99999,-11358.87319,-18085.23818,28914.97050,-9332.266609,4015.96813,-4637.24056,0.,62111.73799,-135471.9749,92590.71229,-20867.58252,-27127.85727,43372.45579,-13998.39991,5714.82284,-6598.90901,0.,-46695.88421,74658.13284,-24095.80875,8031.93625,-9274.48112,0.,13281.56616,-15336.23160,0.,0.},
			{-9661.82590,63751.51762,-141972.4255,128683.4255,-41407.82533,0.,11457.61232,-61812.81421,108360.,-73832.67568,16904.67391,-7127.636352,30863.57076,-38250.91056,13469.96659,1738.965210,-6023.952189,4666.133303,0.,0.,0.,22915.22463,-123625.6285,216720.,-147665.3514,33809.34781,-11639.38143,50400.,-62463.47540,21996.36334,2693.993318,-9332.266605,7228.742626,0.,0.,0.,-21382.90906,92590.71229,-114752.7317,40409.89977,4040.989977,-13998.39991,10843.11394,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,0.,0.,0.,0.,0.},
			{4925.362810,-44899.88862,128683.4255,-143050.2010,54438.22186,0.,-3598.221779,29328.48444,-73832.67568,70400.,-22224.31101,1003.992031,-6955.860840,13469.96659,-7589.466382,0.,0.,0.,0.,0.,0.,-7196.44356,58656.96888,-147665.3514,140800.,-44448.62202,1639.512123,-11358.87318,21996.36334,-12393.54671,0.,0.,0.,0.,0.,0.,3011.976094,-20867.58252,40409.89977,-22768.39914,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1028.78570,11879.39392,-41407.82533,54438.22186,-24004.99948,0.,420.,-4849.742262,16904.67391,-22224.31101,9800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,840.,-9699.484525,33809.34781,-44448.62202,19600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{10802.34025,-9355.10127,4751.757568,-1003.992031,0.,0.,-58125.73395,33809.34782,-11639.38143,1639.512123,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-174828.4790,38025.37259,-3591.991091,106914.5452,-10287.85692,-25895.98888,-58125.73395,33809.34782,-11639.38143,1639.512123,0.,154560.0000,-57808.92777,12021.10144,-846.6404195,-203132.3364,44181.49041,-4173.516504,129311.6793,-12443.02214,-32169.55082,141972.4255,-53100.89083,11042.08676,-777.6888838,-203132.3364,44181.49041,-4173.516504,138010.3092,-13280.04820,-35739.86008,-174828.4790,38025.37259,-3591.991091,129311.6793,-12443.02214,-35739.86008,106914.5452,-10287.85692,-32169.55082,-25895.98888},
			{-9355.10139,24969.34437,-22633.28523,6955.860840,0.,0.,33809.34781,-73741.60290,50400.,-11358.87318,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,38025.37259,-54617.16651,12443.02214,-10287.85692,11879.39392,0.,33809.34781,-73741.60290,50400.,-11358.87318,0.,-57808.92777,102592.,-46847.60655,5865.696891,44181.49041,-63459.41291,14457.48525,-12443.02214,14367.96436,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,44181.49041,-63459.41291,14457.48525,-13280.04820,15334.47879,0.,38025.37259,-54617.16651,12443.02214,-12443.02214,14367.96436,0.,-10287.85692,11879.39392,0.,0.},
			{4751.757564,-22633.28523,31875.75880,-13469.96659,0.,0.,-11639.38143,50399.99999,-62463.47540,21996.36334,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-3591.991091,12443.02214,-9638.323502,0.,0.,0.,-11639.38143,50399.99999,-62463.47540,21996.36334,0.,12021.10144,-46847.60655,48384.,-11358.87318,-4173.516504,14457.48525,-11198.71993,0.,0.,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-4173.516504,14457.48525,-11198.71993,0.,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,0.,0.,0.,0.,0.,0.},
			{-1003.99205,6955.860840,-13469.96659,7589.466382,0.,0.,1639.512123,-11358.87319,21996.36334,-12393.54671,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,1639.512123,-11358.87319,21996.36334,-12393.54671,0.,-846.6404195,5865.696891,-11358.87318,6400.,0.,0.,0.,0.,0.,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-11476.58483,5253.47884,-1099.818167,0.,0.,0.,83732.93531,-27127.85727,4040.989977,0.,0.,-248785.2872,54111.05382,-5111.492932,0.,352131.6458,-46152.44305,2078.460970,-239210.4513,14287.05708,62934.44585,55821.95698,-18085.23818,2693.993318,0.,0.,-203132.3364,44181.49041,-4173.516504,0.,327312.,-42899.43441,1931.962732,-241101.4725,14400.,67012.23766,-124392.6436,27055.52691,-2555.746466,0.,245484.,-32174.57581,1448.972049,-205856.0459,12294.93554,62041.27659,140852.6583,-18460.97722,831.3843878,-144660.8835,8640.,49633.02127,-79736.81711,4762.352360,33506.11883,17981.27024},
			{5253.47885,-9345.21910,3809.881888,0.,0.,0.,-27127.85727,43372.45579,-13998.39991,0.,0.,54111.05382,-77721.59051,17706.73093,0.,-46152.44305,59489.02419,-7200.,14287.05708,-16497.27251,0.,-18085.23818,28914.97050,-9332.266605,0.,0.,44181.49041,-63459.41291,14457.48525,0.,-42899.43441,55296.,-6692.515222,14400.,-16627.68776,0.,27055.52691,-38860.79526,8853.365465,0.,-32174.57581,41472.,-5019.386416,12294.93554,-14196.96869,0.,-18460.97722,23795.60968,-2880.,8640.,-9976.612654,0.,4762.352360,-5499.090835,0.,0.},
			{-1099.81816,3809.881889,-2951.121822,0.,0.,0.,4040.989977,-13998.39991,10843.11394,0.,0.,-5111.492932,17706.73093,-13715.57480,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,2693.993319,-9332.266609,7228.742626,0.,0.,-4173.516504,14457.48525,-11198.71993,0.,1931.962732,-6692.515222,5184.,0.,0.,0.,-2555.746466,8853.365465,-6857.787398,0.,1448.972049,-5019.386416,3888.,0.,0.,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,0.,0.,0.,0.},
			{6531.35214,-1229.634091,0.,0.,0.,0.,-57965.50700,8031.93625,0.,0.,0.,197967.2700,-19049.40944,0.,0.,-311260.6624,18590.32006,0.,229651.2138,-6415.605974,-64595.94539,-28982.75339,4015.96813,0.,0.,0.,129311.6793,-12443.02214,0.,0.,-241101.4725,14400.,0.,198400.,-5542.562586,-60183.71873,59390.18104,-5714.822832,0.,0.,-144660.8835,8640.,0.,141164.0747,-3943.602413,-47759.39698,-62252.13248,3718.064012,0.,79360.,-2217.025034,-31839.59798,32807.31626,-916.5151392,-17195.34821,-6920.994149},
			{-1229.63410,1419.85915,0.,0.,0.,0.,8031.93625,-9274.48112,0.,0.,0.,-19049.40944,21996.36334,0.,0.,18590.32006,-21466.25258,0.,-6415.605974,7408.103671,0.,4015.96813,-4637.24056,0.,0.,0.,-12443.02214,14367.96436,0.,0.,14400.,-16627.68776,0.,-5542.562586,6400.,0.,-5714.822832,6598.909002,0.,0.,8640.,-9976.612654,0.,-3943.602413,4553.679829,0.,3718.064012,-4293.250516,0.,-2217.025034,2560.,0.,-916.5151392,1058.300524,0.,0.},
			{-1525.93584,0.,0.,0.,0.,0.,15574.01680,0.,0.,0.,0.,-59099.23855,0.,0.,0.,100931.1647,0.,0.,-79615.57635,0.,23666.54347,6229.60671,0.,0.,0.,0.,-32169.55082,0.,0.,0.,67012.23766,0.,0.,-60183.71873,0.,19600.,-11819.84771,0.,0.,0.,33506.11883,0.,0.,-36704.11422,0.,13609.55547,11534.99025,0.,0.,-17195.34821,0.,7776.888838,-5686.826882,0.,3500.,1126.978260},
			{13230.11101,-11457.61230,5819.690715,-1229.634092,0.,0.,-35594.59728,20703.91267,-7127.636352,1003.992031,0.,57960.00000,-21678.34791,4507.913042,-317.4901573,-53530.07079,11642.84501,-1099.818167,26188.60820,-2520.,-5285.996592,-106783.7919,62111.73799,-21382.90906,3011.976094,0.,141972.4255,-53100.89083,11042.08676,-777.6888838,-124392.6436,27055.52691,-2555.746466,59390.18104,-5714.822832,-11819.84771,347760.0000,-130070.0875,27047.47825,-1904.940944,-248785.2872,54111.05382,-5111.492932,112684.9455,-10843.11394,-21886.10519,-535300.7081,116428.4501,-10998.18167,197967.2700,-19049.40944,-36476.84197,392829.1232,-37800.,-59099.23855,-111005.9285},
			{-11457.6123,30581.07650,-27720.,8519.154886,0.,0.,20703.91267,-45157.32496,30863.57076,-6955.860840,0.,-21678.34791,38472.,-17567.85246,2199.636334,11642.84501,-16723.02366,3809.881888,-2520.,2909.845357,0.,62111.73800,-135471.9749,92590.71229,-20867.58252,0.,-53100.89083,94236.76939,-43032.27438,5387.986636,27055.52691,-38860.79526,8853.365465,-5714.822832,6598.909002,0.,-130070.0875,230832.,-105407.1147,13197.81800,54111.05382,-77721.59051,17706.73093,-10843.11394,12520.54951,0.,116428.4501,-167230.2366,38098.81888,-19049.40944,21996.36334,0.,-37800.,43647.68036,0.,0.},
			{5819.69066,-27719.99999,39039.67213,-16497.27251,0.,0.,-7127.636351,30863.57076,-38250.91056,13469.96659,0.,4507.913042,-17567.85246,18144.,-4259.577443,-1099.818167,3809.881888,-2951.121822,0.,0.,0.,-21382.90905,92590.71229,-114752.7317,40409.89977,0.,11042.08676,-43032.27438,44443.54190,-10433.79126,-2555.746466,8853.365465,-6857.787398,0.,0.,0.,27047.47825,-105407.1147,108864.,-25557.46466,-5111.492932,17706.73093,-13715.57480,0.,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,0.,0.,0.,0.,0.,0.},
			{-1229.634099,8519.154887,-16497.27251,9295.160030,0.,0.,1003.992032,-6955.860840,13469.96659,-7589.466382,0.,-317.4901573,2199.636334,-4259.577443,2400.,0.,0.,0.,0.,0.,0.,3011.976097,-20867.58252,40409.89977,-22768.39914,0.,-777.6888838,5387.986636,-10433.79126,5878.775383,0.,0.,0.,0.,0.,0.,-1904.940944,13197.81800,-25557.46466,14400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-11476.58483,5253.47884,-1099.818167,0.,0.,0.,55821.95698,-18085.23818,2693.993318,0.,0.,-124392.6436,27055.52691,-2555.746466,0.,140852.6583,-18460.97722,831.3843878,-79736.81711,4762.352360,17981.27024,83732.93531,-27127.85727,4040.989977,0.,0.,-203132.3364,44181.49041,-4173.516504,0.,245484.,-32174.57581,1448.972049,-144660.8835,8640.,33506.11883,-248785.2872,54111.05382,-5111.492932,0.,327312.,-42899.43441,1931.962732,-205856.0459,12294.93554,49633.02127,352131.6458,-46152.44305,2078.460970,-241101.4725,14400.,62041.27659,-239210.4513,14287.05708,67012.23766,62934.44585},
			{5253.47885,-9345.21910,3809.881888,0.,0.,0.,-18085.23818,28914.97050,-9332.266605,0.,0.,27055.52691,-38860.79526,8853.365465,0.,-18460.97722,23795.60968,-2880.,4762.352360,-5499.090835,0.,-27127.85727,43372.45579,-13998.39991,0.,0.,44181.49041,-63459.41291,14457.48525,0.,-32174.57581,41472.,-5019.386416,8640.,-9976.612654,0.,54111.05382,-77721.59051,17706.73093,0.,-42899.43441,55296.,-6692.515222,12294.93554,-14196.96869,0.,-46152.44305,59489.02419,-7200.,14400.,-16627.68776,0.,14287.05708,-16497.27251,0.,0.},
			{-1099.81816,3809.881889,-2951.121822,0.,0.,0.,2693.993319,-9332.266609,7228.742626,0.,0.,-2555.746466,8853.365465,-6857.787398,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,4040.989977,-13998.39991,10843.11394,0.,0.,-4173.516504,14457.48525,-11198.71993,0.,1448.972049,-5019.386416,3888.,0.,0.,0.,-5111.492932,17706.73093,-13715.57480,0.,1931.962732,-6692.515222,5184.,0.,0.,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,0.,0.,0.,0.},
			{6196.18489,-1166.53333,0.,0.,0.,0.,-41243.18141,5714.82284,0.,0.,0.,112684.9455,-10843.11394,0.,0.,-147643.8959,8818.163075,0.,93371.25892,-2608.447815,-22980.41043,-41243.18141,5714.82284,0.,0.,0.,138010.3092,-13280.04820,0.,0.,-205856.0459,12294.93554,0.,141164.0747,-3943.602413,-36704.11422,112684.9455,-10843.11394,0.,0.,-205856.0459,12294.93554,0.,160704.,-4489.475694,-45308.54223,-147643.8959,8818.163075,0.,141164.0747,-3943.602413,-45308.54223,93371.25892,-2608.447815,-36704.11422,-22980.41043},
			{-1166.53333,1346.99667,0.,0.,0.,0.,5714.82284,-6598.90901,0.,0.,0.,-10843.11394,12520.54951,0.,0.,8818.163075,-10182.33765,0.,-2608.447815,3011.976094,0.,5714.82284,-6598.90901,0.,0.,0.,-13280.04820,15334.47879,0.,0.,12294.93554,-14196.96869,0.,-3943.602413,4553.679829,0.,-10843.11394,12520.54951,0.,0.,12294.93554,-14196.96869,0.,-4489.475694,5184.,0.,8818.163075,-10182.33765,0.,-3943.602413,4553.679829,0.,-2608.447815,3011.976094,0.,0.},
			{-1412.74207,0.,0.,0.,0.,0.,11534.99025,0.,0.,0.,0.,-36476.84197,0.,0.,0.,53396.62909,0.,0.,-36854.85042,0.,9738.205176,8651.24267,0.,0.,0.,0.,-35739.86008,0.,0.,0.,62041.27659,0.,0.,-47759.39698,0.,13609.55547,-21886.10519,0.,0.,0.,49633.02127,0.,0.,-45308.54223,0.,14400.,26698.31455,0.,0.,-31839.59798,0.,12000.,-15794.93589,0.,7776.888838,3651.826941},
			{-14816.2076,6782.21205,-1419.859148,0.,0.,0.,36032.9184,-11673.97107,1738.965210,0.,0.,-53530.07079,11642.84501,-1099.818167,0.,45460.,-5958.254780,268.3281572,-20587.95764,1229.634092,3868.953347,144131.6731,-46695.88421,6955.860840,0.,0.,-174828.4790,38025.37259,-3591.991091,0.,140852.6583,-18460.97722,831.3843878,-62252.13248,3718.064012,11534.99025,-535300.7081,116428.4501,-10998.18167,0.,352131.6458,-46152.44305,2078.460970,-147643.8959,8818.163075,26698.31455,909200.,-119165.0956,5366.563145,-311260.6624,18590.32006,53396.62909,-720578.5174,43037.19323,100931.1647,216661.3874},
			{6782.212097,-12064.62597,4918.536370,0.,0.,0.,-11673.97107,18664.53323,-6023.952189,0.,0.,11642.84501,-16723.02366,3809.881888,0.,-5958.254780,7680.,-929.5160030,1229.634092,-1419.859148,0.,-46695.88421,74658.13284,-24095.80875,0.,0.,38025.37259,-54617.16651,12443.02214,0.,-18460.97722,23795.60968,-2880.,3718.064012,-4293.250516,0.,116428.4501,-167230.2366,38098.81888,0.,-46152.44305,59489.02419,-7200.,8818.163075,-10182.33765,0.,-119165.0956,153600.,-18590.32006,18590.32006,-21466.25258,0.,43037.19323,-49695.07017,0.,0.},
			{-1419.859143,4918.536370,-3809.881888,0.,0.,0.,1738.965210,-6023.952191,4666.133303,0.,0.,-1099.818167,3809.881888,-2951.121822,0.,268.3281572,-929.5160030,720.,0.,0.,0.,6955.860840,-24095.80875,18664.53321,0.,0.,-3591.991091,12443.02214,-9638.323502,0.,831.3843878,-2880.,2230.838407,0.,0.,0.,-10998.18167,38098.81888,-29511.21822,0.,2078.460970,-7200.,5577.096018,0.,0.,0.,5366.563145,-18590.32006,14400.,0.,0.,0.,0.,0.,0.,0.},
			{6531.35214,-1229.634091,0.,0.,0.,0.,-28982.75339,4015.96813,0.,0.,0.,59390.18104,-5714.822832,0.,0.,-62252.13248,3718.064012,0.,32807.31626,-916.5151392,-6920.994149,-57965.50700,8031.93625,0.,0.,0.,129311.6793,-12443.02214,0.,0.,-144660.8835,8640.,0.,79360.,-2217.025034,-17195.34821,197967.2700,-19049.40944,0.,0.,-241101.4725,14400.,0.,141164.0747,-3943.602413,-31839.59798,-311260.6624,18590.32006,0.,198400.,-5542.562586,-47759.39698,229651.2138,-6415.605974,-60183.71873,-64595.94539},
			{-1229.63410,1419.85915,0.,0.,0.,0.,4015.96813,-4637.24056,0.,0.,0.,-5714.822832,6598.909002,0.,0.,3718.064012,-4293.250516,0.,-916.5151392,1058.300524,0.,8031.93625,-9274.48112,0.,0.,0.,-12443.02214,14367.96436,0.,0.,8640.,-9976.612654,0.,-2217.025034,2560.,0.,-19049.40944,21996.36334,0.,0.,14400.,-16627.68776,0.,-3943.602413,4553.679829,0.,18590.32006,-21466.25258,0.,-5542.562586,6400.,0.,-6415.605974,7408.103671,0.,0.},
			{-1412.74207,0.,0.,0.,0.,0.,8651.24267,0.,0.,0.,0.,-21886.10519,0.,0.,0.,26698.31455,0.,0.,-15794.93589,0.,3651.826941,11534.99025,0.,0.,0.,0.,-35739.86008,0.,0.,0.,49633.02127,0.,0.,-31839.59798,0.,7776.888838,-36476.84197,0.,0.,0.,62041.27659,0.,0.,-45308.54223,0.,12000.,53396.62909,0.,0.,-47759.39698,0.,14400.,-36854.85042,0.,13609.55547,9738.205176},
			{8640.1667,-1626.653004,0.,0.,0.,0.,-19170.28946,2656.313232,0.,0.,0.,26188.60820,-2520.,0.,0.,-20587.95764,1229.634092,0.,8680.,-242.4871131,-1525.935778,-95851.44763,13281.56616,0.,0.,0.,106914.5452,-10287.85692,0.,0.,-79736.81711,4762.352360,0.,32807.31626,-916.5151392,-5686.826882,392829.1232,-37800.,0.,0.,-239210.4513,14287.05708,0.,93371.25892,-2608.447815,-15794.93589,-720578.5174,43037.19323,0.,229651.2138,-6415.605974,-36854.85042,607600.,-16974.09792,-79615.57635,-192267.9080},
			{-1626.65301,1878.297097,0.,0.,0.,0.,2656.313229,-3067.24632,0.,0.,0.,-2520.,2909.845357,0.,0.,1229.634092,-1419.859148,0.,-242.4871131,280.,0.,13281.56616,-15336.23160,0.,0.,0.,-10287.85692,11879.39392,0.,0.,4762.352360,-5499.090835,0.,-916.5151392,1058.300524,0.,-37800.,43647.68036,0.,0.,14287.05708,-16497.27251,0.,-2608.447815,3011.976094,0.,43037.19323,-49695.07017,0.,-6415.605974,7408.103671,0.,-16974.09792,19600.,0.,0.},
			{-1525.93584,0.,0.,0.,0.,0.,6229.60671,0.,0.,0.,0.,-11819.84771,0.,0.,0.,11534.99025,0.,0.,-5686.826882,0.,1126.978260,15574.01680,0.,0.,0.,0.,-32169.55082,0.,0.,0.,33506.11883,0.,0.,-17195.34821,0.,3500.,-59099.23855,0.,0.,0.,67012.23766,0.,0.,-36704.11422,0.,7776.888838,100931.1647,0.,0.,-60183.71873,0.,13609.55547,-79615.57635,0.,19600.,23666.54347},
			{-2047.25769,0.,0.,0.,0.,0.,4178.947267,0.,0.,0.,0.,-5285.996592,0.,0.,0.,3868.953347,0.,0.,-1525.935778,0.,252.,25073.68342,0.,0.,0.,0.,-25895.98888,0.,0.,0.,17981.27024,0.,0.,-6920.994149,0.,1126.978260,-111005.9285,0.,0.,0.,62934.44585,0.,0.,-22980.41043,0.,3651.826941,216661.3874,0.,0.,-64595.94539,0.,9738.205176,-192267.9080,0.,23666.54347,63504.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,1.354006401,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.581138830,1.369306394,1.178511302,1.045825033,1.296148140,1.122497216,.9660917834,1.095445115,.9486832980,.9476070828,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.936491673,1.677050983,1.443375673,1.280868846,1.673320053,1.449137675,1.247219129,1.449137675,1.254990040,1.271348312,1.581138830,1.369306394,1.178511302,1.045825033,1.673320053,1.449137675,1.247219129,1.527525232,1.322875656,1.373213125,1.296148140,1.122497216,.9660917834,1.449137675,1.254990040,1.373213125,1.095445115,.9486832980,1.271348312,.9476070828},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,2.010178183,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.122497216,1.296148140,1.254990040,.9486832980,1.095445115,.8206518070,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.677050983,1.936491673,1.875000000,1.774823935,1.449137675,1.673320053,1.620185174,1.254990040,1.449137675,1.101019935,1.369306394,1.581138830,1.530931089,1.449137675,1.449137675,1.673320053,1.620185174,1.322875656,1.527525232,1.189237451,1.122497216,1.296148140,1.254990040,1.254990040,1.449137675,1.189237451,.9486832980,1.095445115,1.101019935,.8206518070},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,2.270737765,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.178511302,1.530931089,1.581138830,1.559023911,.9660917834,1.254990040,1.296148140,.8164965809,1.060660172,.7063046176,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.443375673,1.875000000,1.936491673,1.909406540,1.247219129,1.620185175,1.673320053,1.080123450,1.403121520,.9476070828,1.178511302,1.530931089,1.581138830,1.559023911,1.247219129,1.620185175,1.673320053,1.138550085,1.479019946,1.023532632,.9660917834,1.254990040,1.296148140,1.080123450,1.403121520,1.023532632,.8164965809,1.060660172,.9476070828,.7063046176},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,2.388242806,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.045825033,1.449137675,1.559023911,1.581138830,.8573214100,1.187939392,1.278019301,.7245688375,1.003992032,.6267831704,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.280868845,1.774823935,1.909406540,1.936491673,1.106797181,1.533623161,1.649915823,.9585144758,1.328156617,.8409178659,1.045825033,1.449137675,1.559023911,1.581138830,1.106797181,1.533623161,1.649915823,1.010362971,1.400000000,.9082951065,.8573214100,1.187939392,1.278019301,.9585144758,1.328156617,.9082951065,.7245688375,1.003992032,.8409178659,.6267831704},
			{1.469693846,2.121320343,2.347382390,2.430277762,2.449489743,2.437211522,1.200000000,1.732050808,1.916629694,1.984313483,2.,.9486832980,1.369306394,1.515228816,1.568737550,.7776888840,1.122497216,1.242118007,.6572670691,.9486832980,.5685642496,1.200000000,1.732050808,1.916629694,1.984313483,2.,1.161895004,1.677050983,1.855768722,1.921303269,1.003992032,1.449137675,1.603567452,.8694826050,1.254990040,.7628089876,.9486832980,1.369306394,1.515228816,1.568737550,1.003992032,1.449137675,1.603567452,.9165151392,1.322875656,.8239278748,.7776888840,1.122497216,1.242118007,.8694826050,1.254990040,.8239278748,.6572670691,.9486832980,.7628089876,.5685642496},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.748014747,3.872983346,3.354101966,2.886751347,2.561737690,2.323790008,3.674234614,3.181980514,2.738612787,2.430277762,3.346640106,2.898275348,2.494438258,3.030457633,2.624453295,2.752549839,1.936491673,1.677050983,1.443375673,1.280868845,1.161895004,3.,2.598076212,2.236067977,1.984313483,3.240370349,2.806243040,2.415229457,3.207134903,2.777460299,3.077444275,1.224744872,1.060660172,.9128709292,.8100925872,2.160246899,1.870828694,1.610152971,2.535462764,2.195775163,2.659215782,.8366600264,.7245688370,.6236095646,1.603567452,1.388730150,1.994411836,.6060915266,.5248906592,1.230977710,.4587583064},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,2.595128875,3.354101966,3.872983346,3.750000000,3.549647869,3.354101966,3.181980514,3.674234614,3.557562366,3.367491648,2.898275350,3.346640106,3.240370350,2.624453295,3.030457633,2.383778084,1.677050983,1.936491673,1.875000000,1.774823935,1.677050983,2.598076212,3.,2.904737510,2.749545418,2.806243040,3.240370350,3.137475099,2.777460300,3.207134903,2.665144921,1.060660172,1.224744872,1.185854122,1.122497216,1.870828694,2.160246900,2.091650066,2.195775165,2.535462764,2.302948421,.7245688375,.8366600264,.8100925875,1.388730150,1.603567452,1.727211316,.5248906592,.6060915266,1.066057968,.3972963474},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.931509850,2.886751347,3.750000000,3.872983346,3.818813079,3.711537446,2.738612788,3.557562368,3.674234614,3.622844188,2.494438258,3.240370349,3.346640106,2.258769757,2.934227987,2.051629517,1.443375673,1.875000000,1.936491673,1.909406540,1.855768722,2.236067977,2.904737510,3.,2.958039892,2.415229459,3.137475099,3.240370349,2.390457218,3.105295018,2.293791533,.9128709293,1.185854122,1.224744872,1.207614729,1.610152972,2.091650066,2.160246899,1.889822365,2.454951265,1.982062418,.6236095646,.8100925872,.8366600265,1.195228610,1.552647508,1.486546813,.4517539516,.5868455971,.9175166129,.3419382528},
			{2.091650066,2.898275350,3.118047822,3.162277659,3.137475099,3.083208205,2.561737692,3.549647870,3.818813080,3.872983346,3.842606539,2.430277762,3.367491647,3.622844188,3.674234614,2.213594361,3.067246321,3.299831644,2.004459314,2.777460300,1.820640586,1.280868846,1.774823935,1.909406540,1.936491673,1.921303269,1.984313483,2.749545418,2.958039892,3.,2.143303525,2.969848480,3.195048251,2.121320343,2.939387692,2.035538055,.8100925874,1.122497216,1.207614729,1.224744872,1.428869017,1.979898987,2.130032168,1.677050983,2.323790008,1.758905910,.5533985903,.7668115804,.8249579111,1.060660172,1.469693846,1.319179432,.4008918626,.5554920599,.8142152223,.3034400977},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,5.019960159,4.347413025,3.741657387,3.320391543,3.011976095,5.291502622,4.582575695,3.944053189,3.500000000,5.163977793,4.472135954,3.849001795,4.909902528,4.252100321,4.624812309,1.673320053,1.449137675,1.247219129,1.106797181,1.003992032,3.240370349,2.806243040,2.415229459,2.143303525,4.,3.464101616,2.981423969,4.330127020,3.750000000,4.432026303,.8819171036,.7637626160,.6573421980,.5833333333,2.,1.732050808,1.490711985,2.738612788,2.371708245,3.191423693,.5163977793,.4472135954,.3849001795,1.299038106,1.125000000,1.914854215,.3273268354,.2834733546,.8864052607,.2202291575},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,4.347413025,5.019960159,4.860555525,4.600869484,4.347413025,4.582575696,5.291502622,5.123475385,4.849742262,4.472135954,5.163977793,5.,4.252100321,4.909902529,4.005204947,1.449137675,1.673320053,1.620185175,1.533623161,1.449137675,2.806243040,3.240370350,3.137475099,2.969848480,3.464101616,4.,3.872983346,3.750000000,4.330127020,3.838247368,.7637626162,.8819171036,.8539125641,.8082903771,1.732050808,2.,1.936491673,2.371708245,2.738612788,2.763853992,.4472135954,.5163977793,.5000000000,1.125000000,1.299038106,1.658312395,.2834733546,.3273268354,.7676494735,.1907240451},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,3.741657387,4.860555524,5.019960158,4.949747467,4.810702356,3.944053188,5.123475380,5.291502622,5.217491946,3.849001795,5.,5.163977793,3.659625271,4.753992682,3.447131570,1.247219129,1.620185174,1.673320053,1.649915823,1.603567452,2.415229457,3.137475099,3.240370349,3.195048251,2.981423969,3.872983346,4.,3.227486122,4.192627457,3.303437362,.6573421979,.8539125633,.8819171036,.8695819911,1.490711985,1.936491673,2.,2.041241452,2.651650429,2.378746774,.3849001795,.5000000000,.5163977793,.9682458365,1.257788237,1.427248064,.2439750181,.3169328455,.6606874726,.1641491224},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,5.796550700,5.019960159,4.320493800,3.834057903,3.477930420,6.546536711,5.669467095,4.879500367,4.330127020,6.708203931,5.809475019,5.,6.614378278,5.728219620,6.408327913,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,3.207134903,2.777460300,2.390457218,2.121320343,4.330127020,3.750000000,3.227486122,5.,4.330127020,5.373546315,.6546536711,.5669467095,.4879500367,.4330127020,1.732050808,1.500000000,1.290994449,2.635231383,2.282177324,3.316624790,.3354101966,.2904737510,.2500000000,1.,.8660254040,1.658312395,.1889822365,.1636634177,.6141195788,.1144344270},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,5.019960159,5.796550700,5.612486080,5.312626469,5.019960159,5.669467095,6.546536709,6.338656908,6.,5.809475019,6.708203931,6.495190530,5.728219619,6.614378278,5.549774770,1.254990040,1.449137675,1.403121520,1.328156617,1.254990040,2.777460299,3.207134903,3.105295018,2.939387692,3.750000000,4.330127020,4.192627457,4.330127020,5.,4.653627619,.5669467095,.6546536709,.6338656908,.6000000000,1.500000000,1.732050808,1.677050983,2.282177323,2.635231383,2.872281324,.2904737510,.3354101966,.3247595265,.8660254040,1.,1.436140662,.1636634177,.1889822365,.5318431565,.9910312088e-1},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,6.356741560,5.505099678,4.738035416,4.204589330,3.814044936,7.538168182,6.528245145,5.618618828,4.986029592,8.010409897,6.937218462,5.970607017,8.124038403,7.035623637,8.049844717,1.271348312,1.101019935,.9476070828,.8409178659,.7628089876,3.077444275,2.665144921,2.293791533,2.035538055,4.432026303,3.838247368,3.303437362,5.373546315,4.653627619,6.,.5025445453,.4352163430,.3745745885,.3324019728,1.477342101,1.279415789,1.101145788,2.427520781,2.102294664,3.240370350,.2288688541,.1982062418,.1705887719,.7676494735,.6648039455,1.388730150,.1160576915,.1005089091,.4285714286,.6388765648e-1},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.427248064,1.581138830,1.369306394,1.178511302,1.045825033,.9486832980,1.,.8660254040,.7453559923,.6614378278,.6831300514,.5916079782,.5091750773,.4948716594,.4285714286,.3745745885,3.162277660,2.738612788,2.357022604,2.091650066,1.897366596,2.449489743,2.121320343,1.825741858,1.620185175,1.763834207,1.527525232,1.314684396,1.309307342,1.133893419,1.005089091,3.,2.598076212,2.236067977,1.984313483,2.645751311,2.291287848,1.972026594,2.070196679,1.792842914,1.628430445,2.732520206,2.366431913,2.036700309,2.618614683,2.267786838,2.171240595,2.474358298,2.142857143,2.512722727,2.247447532},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,1.369306394,1.581138830,1.530931089,1.449137675,1.369306394,.8660254040,1.,.9682458365,.9165151390,.5916079782,.6831300507,.6614378278,.4285714286,.4948716594,.3243911095,2.738612788,3.162277660,3.061862179,2.898275350,2.738612788,2.121320343,2.449489743,2.371708245,2.244994432,1.527525232,1.763834207,1.707825127,1.133893419,1.309307341,.8704326857,2.598076212,3.,2.904737510,2.749545417,2.291287848,2.645751311,2.561737690,1.792842914,2.070196679,1.410262133,2.366431913,2.732520203,2.645751311,2.267786838,2.618614683,1.880349512,2.142857143,2.474358298,2.176081715,1.946346656},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.393567770,1.178511302,1.530931089,1.581138830,1.559023911,1.515228816,.7453559923,.9682458365,1.,.9860132974,.5091750773,.6614378278,.6831300514,.3688555568,.4791574237,.2791914141,2.357022604,3.061862179,3.162277659,3.118047822,3.030457633,1.825741858,2.371708244,2.449489743,2.415229459,1.314684396,1.707825129,1.763834207,.9759000734,1.267731382,.7491491772,2.236067977,2.904737510,3.,2.958039892,1.972026594,2.561737692,2.645751311,1.543033500,2.004459314,1.213760391,2.036700309,2.645751311,2.732520206,1.951800147,2.535462764,1.618347188,1.844277784,2.395787117,1.872872943,1.675148485},
			{1.707825129,2.366431913,2.545875386,2.581988897,2.561737692,2.517428959,1.045825033,1.449137675,1.559023911,1.581138830,1.568737550,.6614378278,.9165151390,.9860132974,1.,.4518480572,.6260990336,.6735753142,.3273268354,.4535573675,.2477578022,2.091650066,2.898275350,3.118047822,3.162277660,3.137475099,1.620185174,2.244994432,2.415229459,2.449489743,1.166666667,1.616580754,1.739163982,.8660254040,1.200000000,.6648039455,1.984313483,2.749545417,2.958039892,3.,1.750000000,2.424871131,2.608745974,1.369306394,1.897366596,1.077105496,1.807392228,2.504396134,2.694301258,1.732050808,2.400000000,1.436140662,1.636634177,2.267786838,1.662009863,1.486546813},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,2.645751311,2.291287848,1.972026594,1.750000000,2.065591118,1.788854382,1.539600718,1.636634177,1.417366774,1.321374945,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,4.320493799,3.741657387,3.220305944,2.857738034,4.,3.464101616,2.981423969,3.464101616,3.,2.954684202,2.645751311,2.291287848,1.972026594,1.750000000,4.,3.464101616,2.981423969,4.107919181,3.557562368,3.829708432,2.065591118,1.788854382,1.539600718,3.464101616,3.,3.829708432,1.636634177,1.417366774,2.954684202,1.321374945},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,2.291287848,2.645751311,2.561737692,2.424871131,1.788854382,2.065591118,2.,1.417366774,1.636634177,1.144344270,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,3.741657387,4.320493800,4.183300132,3.959797974,3.464101616,4.,3.872983346,3.,3.464101616,2.558831578,2.291287848,2.645751311,2.561737692,2.424871131,3.464101616,4.,3.872983346,3.557562368,4.107919182,3.316624790,1.788854382,2.065591118,2.,3.,3.464101616,3.316624790,1.417366774,1.636634177,2.558831578,1.144344270},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,1.972026594,2.561737690,2.645751311,2.608745974,1.539600718,2.,2.065591118,1.219875090,1.584664228,.9848947341,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,3.220305942,4.183300132,4.320493799,4.260064335,2.981423969,3.872983346,4.,2.581988897,3.354101966,2.202291575,1.972026594,2.561737690,2.645751311,2.608745974,2.981423969,3.872983346,4.,3.061862179,3.977475643,2.854496129,1.539600718,2.,2.065591118,2.581988897,3.354101966,2.854496129,1.219875090,1.584664228,2.202291575,.9848947341},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,2.068278942,4.582575696,3.968626966,3.415650256,3.031088914,2.749545418,4.140393357,3.585685828,3.086067000,2.738612788,3.535533905,3.061862179,2.635231383,2.988071523,2.587745849,2.533114026,3.055050464,2.645751311,2.277100171,2.020725943,1.833030278,5.070925527,4.391550330,3.779644731,3.354101966,5.477225575,4.743416490,4.082482906,5.270462768,4.564354646,4.855041564,2.070196679,1.792842914,1.543033500,1.369306394,4.107919181,3.557562368,3.061862179,5.,4.330127020,5.244044240,1.414213562,1.224744872,1.054092553,3.162277660,2.738612788,4.195235392,.9960238407,.8625819493,2.427520781,.7237468643},
			{3.240370348,3.741657386,3.622844188,3.429285640,3.240370348,3.070597895,3.968626966,4.582575695,4.437059836,4.200000000,3.968626966,3.585685828,4.140393357,4.008918626,3.794733192,3.061862179,3.535533905,3.423265985,2.587745849,2.988071523,2.193741096,2.645751311,3.055050463,2.958039891,2.800000000,2.645751311,4.391550326,5.070925527,4.909902528,4.647580015,4.743416490,5.477225576,5.303300858,4.564354646,5.270462768,4.204589330,1.792842914,2.070196679,2.004459314,1.897366596,3.557562368,4.107919182,3.977475643,4.330127020,5.,4.541475532,1.224744872,1.414213562,1.369306394,2.738612788,3.162277660,3.633180426,.8625819493,.9960238407,2.102294664,.6267831704},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,5.492852500,4.756949802,4.094130528,3.633180426,3.295711500,5.428101485,4.700873779,4.045867967,3.590351654,4.944132325,4.281744193,3.685138655,4.387482194,3.799671040,3.864367134,2.746426250,2.378474901,2.047065264,1.816590213,1.647855750,5.318431565,4.605896844,3.964124837,3.517811820,6.382847385,5.527707984,4.757493549,6.633249580,5.744562647,6.480740700,1.628430445,1.410262133,1.213760391,1.077105496,3.829708432,3.316624790,2.854496129,5.244044240,4.541475532,6.,.9888264647,.8563488387,.7370277310,2.653299832,2.297825059,4.,.6267831705,.5428101485,1.851640200,.4140393357},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.462494065,1.296148140,1.122497216,.9660917834,.8573214100,.7776888840,.6831300514,.5916079782,.5091750773,.4518480572,.4000000000,.3464101616,.2981423969,.2535462764,.2195775165,.1705887719,3.888444420,3.367491648,2.898275350,2.571964230,2.333066652,2.509980079,2.173706512,1.870828694,1.660195771,1.549193338,1.341640786,1.154700539,1.006230590,.8714212528,.6866065626,4.098780308,3.549647869,3.055050464,2.711088342,3.098386677,2.683281572,2.309401077,2.121320343,1.837117307,1.483239697,4.,3.464101616,2.981423969,3.354101966,2.904737510,2.472066162,3.803194146,3.293662747,3.433032812,3.582364211},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,2.171240594,1.122497216,1.296148140,1.254990040,1.187939392,1.122497216,.5916079782,.6831300507,.6614378278,.6260990336,.3464101616,.4000000000,.3872983346,.2195775163,.2535462764,.1477342101,3.367491648,3.888444419,3.764970119,3.563818176,3.367491648,2.173706511,2.509980079,2.430277762,2.300434741,1.341640786,1.549193338,1.500000000,.8714212528,1.006230590,.5946187254,3.549647869,4.098780304,3.968626966,3.756594201,2.683281572,3.098386677,3.,1.837117307,2.121320343,1.284523258,3.464101616,4.,3.872983346,2.904737510,3.354101966,2.140872097,3.293662744,3.803194146,2.973093626,3.102418412},
			{1.972026594,2.561737692,2.645751311,2.608745974,2.535462764,2.452677109,.9660917834,1.254990040,1.296148140,1.278019301,1.242118007,.5091750773,.6614378278,.6831300514,.6735753142,.2981423969,.3872983346,.4000000000,.1889822365,.2454951266,.1271493634,2.898275350,3.764970119,3.888444420,3.834057903,3.726354021,1.870828694,2.430277762,2.509980080,2.474873734,1.154700539,1.500000000,1.549193338,.7500000000,.9742785795,.5117663157,3.055050464,3.968626966,4.098780308,4.041451885,2.309401077,3.,3.098386677,1.581138830,2.053959591,1.105541597,2.981423969,3.872983346,4.,2.500000000,3.247595265,1.842569328,2.834733546,3.682426899,2.558831578,2.670136631},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,2.898275350,2.509980080,2.160246900,1.917028952,1.738965210,1.963961013,1.700840129,1.463850110,1.299038106,1.341640786,1.161895004,1.,.9449111824,.8183170887,.6866065625,4.347413025,3.764970119,3.240370350,2.875543427,2.608447815,4.810702356,4.166190450,3.585685828,3.181980514,3.897114318,3.375000000,2.904737510,3.,2.598076212,2.302948420,3.927922025,3.401680258,2.927700220,2.598076212,5.196152424,4.500000000,3.872983346,4.743416490,4.107919182,3.979949748,3.354101966,2.904737510,2.500000000,5.,4.330127020,4.974937185,2.834733546,2.454951266,4.605896842,2.403122968},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,2.509980080,2.898275350,2.806243040,2.656313234,2.509980080,1.700840129,1.963961012,1.901597073,1.800000000,1.161895004,1.341640786,1.299038106,.8183170885,.9449111824,.5946187254,3.764970119,4.347413025,4.209364560,3.984469852,3.764970119,4.166190449,4.810702356,4.657942526,4.409081537,3.375000000,3.897114318,3.773364711,2.598076212,3.,1.994411836,3.401680258,3.927922024,3.803194146,3.600000000,4.500000000,5.196152424,5.031152948,4.107919181,4.743416490,3.446737588,2.904737510,3.354101966,3.247595265,4.330127020,5.,4.308421985,2.454951265,2.834733546,3.988823673,2.081165539},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,3.256860892,2.820524267,2.427520781,2.154210993,2.472066162,2.140872097,1.842569328,1.880349512,1.628430445,1.449137675,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,5.983235509,5.181633950,4.459640440,3.957538298,5.744562647,4.974937185,4.281744193,4.974937185,4.308421985,4.166190450,3.256860892,2.820524267,2.427520781,2.154210993,5.744562647,4.974937185,4.281744193,6.292853088,5.449770639,6.,2.472066162,2.140872097,1.842569328,4.974937185,4.308421985,6.,1.880349512,1.628430445,4.166190450,1.449137675},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.483239697,1.095445115,.9486832980,.8164965809,.7245688375,.6572670691,.4948716594,.4285714286,.3688555568,.3273268354,.2535462764,.2195775163,.1889822365,.1428571429,.1237179149,.8650429586e-1,4.381780461,3.794733192,3.265986323,2.898275350,2.629068276,2.424366106,2.099562637,1.807015806,1.603567451,1.309307341,1.133893419,.9759000724,.7559289460,.6546536709,.4642307659,4.948716594,4.285714286,3.688555568,3.273268354,3.273268354,2.834733546,2.439750181,1.992047682,1.725163898,1.253566341,5.070925527,4.391550326,3.779644731,3.779644731,3.273268354,2.507132682,5.,4.330127020,4.062019202,4.844240569},
			{2.323790008,2.683281572,2.598076212,2.459268185,2.323790008,2.202039870,.9486832980,1.095445115,1.060660172,1.003992032,.9486832980,.4285714286,.4948716594,.4791574237,.4535573675,.2195775165,.2535462764,.2454951266,.1237179149,.1428571429,.7491491772e-1,3.794733192,4.381780460,4.242640686,4.015968127,3.794733192,2.099562637,2.424366106,2.347382389,2.221968240,1.133893419,1.309307342,1.267731382,.6546536711,.7559289460,.4020356364,4.285714286,4.948716594,4.791574237,4.535573675,2.834733546,3.273268354,3.169328455,1.725163898,1.992047682,1.085620297,4.391550330,5.070925527,4.909902529,3.273268354,3.779644731,2.171240595,4.330127020,5.,3.517811818,4.195235392},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,2.542696624,2.202039870,1.895214166,1.681835732,1.525617974,1.507633636,1.305649029,1.123723765,.9972059183,.9154754167,.7928249671,.6823550877,.5802884574,.5025445453,.3833259389,5.085393248,4.404079742,3.790428332,3.363671463,3.051235949,4.923910840,4.264231874,3.670066451,3.256860890,3.545621043,3.070597895,2.642749891,2.456478316,2.127372625,1.714285714,5.025445453,4.352163430,3.745745885,3.324019728,5.909368405,5.117663157,4.404583150,4.855041564,4.204589330,3.703280400,4.577377083,3.964124837,3.411775437,6.141195788,5.318431565,5.554920600,4.062019202,3.517811818,6.,3.577708763},
			{2.708012801,2.345207879,2.018433570,1.791182105,1.624807681,1.496910398,.9476070828,.8206518070,.7063046176,.6267831704,.5685642496,.3745745885,.3243911095,.2791914141,.2477578022,.1705887719,.1477342101,.1271493634,.8650429586e-1,.7491491772e-1,.4761904762e-1,4.738035416,4.103259034,3.531523088,3.133915852,2.842821248,2.293791533,1.986481737,1.709691264,1.517200488,1.101145788,.9536202253,.8207456118,.5721721350,.4955156045,.3194382825,5.618618828,4.865866643,4.187871213,3.716367033,3.303437362,2.860860676,2.462236835,1.809367160,1.566957926,1.035098339,5.970607017,5.170697355,4.450227718,4.005204947,3.468609231,2.415229459,6.055300709,5.244044240,4.472135954,6.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,1.354006401,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.581138830,1.369306394,1.178511302,1.045825033,1.296148140,1.122497216,.9660917834,1.095445115,.9486832980,.9476070828,2.,1.732050808,1.490711985,1.322875656,1.200000000,1.936491673,1.677050983,1.443375673,1.280868846,1.673320053,1.449137675,1.247219129,1.449137675,1.254990040,1.271348312,1.581138830,1.369306394,1.178511302,1.045825033,1.673320053,1.449137675,1.247219129,1.527525232,1.322875656,1.373213125,1.296148140,1.122497216,.9660917834,1.449137675,1.254990040,1.373213125,1.095445115,.9486832980,1.271348312,.9476070828},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,2.010178183,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.122497216,1.296148140,1.254990040,.9486832980,1.095445115,.8206518070,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.677050983,1.936491673,1.875000000,1.774823935,1.449137675,1.673320053,1.620185174,1.254990040,1.449137675,1.101019935,1.369306394,1.581138830,1.530931089,1.449137675,1.449137675,1.673320053,1.620185174,1.322875656,1.527525232,1.189237451,1.122497216,1.296148140,1.254990040,1.254990040,1.449137675,1.189237451,.9486832980,1.095445115,1.101019935,.8206518070},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,2.270737765,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.178511302,1.530931089,1.581138830,1.559023911,.9660917834,1.254990040,1.296148140,.8164965809,1.060660172,.7063046176,1.490711985,1.936491673,2.,1.972026594,1.916629694,1.443375673,1.875000000,1.936491673,1.909406540,1.247219129,1.620185175,1.673320053,1.080123450,1.403121520,.9476070828,1.178511302,1.530931089,1.581138830,1.559023911,1.247219129,1.620185175,1.673320053,1.138550085,1.479019946,1.023532632,.9660917834,1.254990040,1.296148140,1.080123450,1.403121520,1.023532632,.8164965809,1.060660172,.9476070828,.7063046176},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,2.388242806,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.045825033,1.449137675,1.559023911,1.581138830,.8573214100,1.187939392,1.278019301,.7245688375,1.003992032,.6267831704,1.322875656,1.833030278,1.972026594,2.,1.984313483,1.280868845,1.774823935,1.909406540,1.936491673,1.106797181,1.533623161,1.649915823,.9585144758,1.328156617,.8409178659,1.045825033,1.449137675,1.559023911,1.581138830,1.106797181,1.533623161,1.649915823,1.010362971,1.400000000,.9082951065,.8573214100,1.187939392,1.278019301,.9585144758,1.328156617,.9082951065,.7245688375,1.003992032,.8409178659,.6267831704},
			{1.469693846,2.121320343,2.347382390,2.430277762,2.449489743,2.437211522,1.200000000,1.732050808,1.916629694,1.984313483,2.,.9486832980,1.369306394,1.515228816,1.568737550,.7776888840,1.122497216,1.242118007,.6572670691,.9486832980,.5685642496,1.200000000,1.732050808,1.916629694,1.984313483,2.,1.161895004,1.677050983,1.855768722,1.921303269,1.003992032,1.449137675,1.603567452,.8694826050,1.254990040,.7628089876,.9486832980,1.369306394,1.515228816,1.568737550,1.003992032,1.449137675,1.603567452,.9165151392,1.322875656,.8239278748,.7776888840,1.122497216,1.242118007,.8694826050,1.254990040,.8239278748,.6572670691,.9486832980,.7628089876,.5685642496},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.427248064,3.162277660,2.738612788,2.357022604,2.091650066,1.897366596,3.,2.598076212,2.236067977,1.984313483,2.732520206,2.366431913,2.036700309,2.474358298,2.142857143,2.247447532,1.581138830,1.369306394,1.178511302,1.045825033,.9486832980,2.449489743,2.121320343,1.825741858,1.620185175,2.645751311,2.291287848,1.972026594,2.618614683,2.267786838,2.512722727,1.,.8660254040,.7453559923,.6614378278,1.763834207,1.527525232,1.314684396,2.070196679,1.792842914,2.171240595,.6831300514,.5916079782,.5091750773,1.309307342,1.133893419,1.628430445,.4948716594,.4285714286,1.005089091,.3745745885},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,2.738612788,3.162277660,3.061862179,2.898275350,2.738612788,2.598076212,3.,2.904737510,2.749545417,2.366431913,2.732520203,2.645751311,2.142857143,2.474358298,1.946346656,1.369306394,1.581138830,1.530931089,1.449137675,1.369306394,2.121320343,2.449489743,2.371708245,2.244994432,2.291287848,2.645751311,2.561737690,2.267786838,2.618614683,2.176081715,.8660254040,1.,.9682458365,.9165151390,1.527525232,1.763834207,1.707825127,1.792842914,2.070196679,1.880349512,.5916079782,.6831300507,.6614378278,1.133893419,1.309307341,1.410262133,.4285714286,.4948716594,.8704326857,.3243911095},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.393567770,2.357022604,3.061862179,3.162277659,3.118047822,3.030457633,2.236067977,2.904737510,3.,2.958039892,2.036700309,2.645751311,2.732520206,1.844277784,2.395787117,1.675148485,1.178511302,1.530931089,1.581138830,1.559023911,1.515228816,1.825741858,2.371708244,2.449489743,2.415229459,1.972026594,2.561737692,2.645751311,1.951800147,2.535462764,1.872872943,.7453559923,.9682458365,1.,.9860132974,1.314684396,1.707825129,1.763834207,1.543033500,2.004459314,1.618347188,.5091750773,.6614378278,.6831300514,.9759000734,1.267731382,1.213760391,.3688555568,.4791574237,.7491491772,.2791914141},
			{1.707825129,2.366431913,2.545875386,2.581988897,2.561737692,2.517428959,2.091650066,2.898275350,3.118047822,3.162277660,3.137475099,1.984313483,2.749545417,2.958039892,3.,1.807392228,2.504396134,2.694301258,1.636634177,2.267786838,1.486546813,1.045825033,1.449137675,1.559023911,1.581138830,1.568737550,1.620185174,2.244994432,2.415229459,2.449489743,1.750000000,2.424871131,2.608745974,1.732050808,2.400000000,1.662009863,.6614378278,.9165151390,.9860132974,1.,1.166666667,1.616580754,1.739163982,1.369306394,1.897366596,1.436140662,.4518480572,.6260990336,.6735753142,.8660254040,1.200000000,1.077105496,.3273268354,.4535573675,.6648039455,.2477578022},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.462494065,3.888444420,3.367491648,2.898275350,2.571964230,2.333066652,4.098780308,3.549647869,3.055050464,2.711088342,4.,3.464101616,2.981423969,3.803194146,3.293662747,3.582364211,1.296148140,1.122497216,.9660917834,.8573214100,.7776888840,2.509980079,2.173706512,1.870828694,1.660195771,3.098386677,2.683281572,2.309401077,3.354101966,2.904737510,3.433032812,.6831300514,.5916079782,.5091750773,.4518480572,1.549193338,1.341640786,1.154700539,2.121320343,1.837117307,2.472066162,.4000000000,.3464101616,.2981423969,1.006230590,.8714212528,1.483239697,.2535462764,.2195775165,.6866065626,.1705887719},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,2.171240594,3.367491648,3.888444419,3.764970119,3.563818176,3.367491648,3.549647869,4.098780304,3.968626966,3.756594201,3.464101616,4.,3.872983346,3.293662744,3.803194146,3.102418412,1.122497216,1.296148140,1.254990040,1.187939392,1.122497216,2.173706511,2.509980079,2.430277762,2.300434741,2.683281572,3.098386677,3.,2.904737510,3.354101966,2.973093626,.5916079782,.6831300507,.6614378278,.6260990336,1.341640786,1.549193338,1.500000000,1.837117307,2.121320343,2.140872097,.3464101616,.4000000000,.3872983346,.8714212528,1.006230590,1.284523258,.2195775163,.2535462764,.5946187254,.1477342101},
			{1.972026594,2.561737692,2.645751311,2.608745974,2.535462764,2.452677109,2.898275350,3.764970119,3.888444420,3.834057903,3.726354021,3.055050464,3.968626966,4.098780308,4.041451885,2.981423969,3.872983346,4.,2.834733546,3.682426899,2.670136631,.9660917834,1.254990040,1.296148140,1.278019301,1.242118007,1.870828694,2.430277762,2.509980080,2.474873734,2.309401077,3.,3.098386677,2.500000000,3.247595265,2.558831578,.5091750773,.6614378278,.6831300514,.6735753142,1.154700539,1.500000000,1.549193338,1.581138830,2.053959591,1.842569328,.2981423969,.3872983346,.4000000000,.7500000000,.9742785795,1.105541597,.1889822365,.2454951266,.5117663157,.1271493634},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.483239697,4.381780461,3.794733192,3.265986323,2.898275350,2.629068276,4.948716594,4.285714286,3.688555568,3.273268354,5.070925527,4.391550326,3.779644731,5.,4.330127020,4.844240569,1.095445115,.9486832980,.8164965809,.7245688375,.6572670691,2.424366106,2.099562637,1.807015806,1.603567451,3.273268354,2.834733546,2.439750181,3.779644731,3.273268354,4.062019202,.4948716594,.4285714286,.3688555568,.3273268354,1.309307341,1.133893419,.9759000724,1.992047682,1.725163898,2.507132682,.2535462764,.2195775163,.1889822365,.7559289460,.6546536709,1.253566341,.1428571429,.1237179149,.4642307659,.8650429586e-1},
			{2.323790008,2.683281572,2.598076212,2.459268185,2.323790008,2.202039870,3.794733192,4.381780460,4.242640686,4.015968127,3.794733192,4.285714286,4.948716594,4.791574237,4.535573675,4.391550330,5.070925527,4.909902529,4.330127020,5.,4.195235392,.9486832980,1.095445115,1.060660172,1.003992032,.9486832980,2.099562637,2.424366106,2.347382389,2.221968240,2.834733546,3.273268354,3.169328455,3.273268354,3.779644731,3.517811818,.4285714286,.4948716594,.4791574237,.4535573675,1.133893419,1.309307342,1.267731382,1.725163898,1.992047682,2.171240595,.2195775165,.2535462764,.2454951266,.6546536711,.7559289460,1.085620297,.1237179149,.1428571429,.4020356364,.7491491772e-1},
			{2.708012801,2.345207879,2.018433570,1.791182105,1.624807681,1.496910398,4.738035416,4.103259034,3.531523088,3.133915852,2.842821248,5.618618828,4.865866643,4.187871213,3.716367033,5.970607017,5.170697355,4.450227718,6.055300709,5.244044240,6.,.9476070828,.8206518070,.7063046176,.6267831704,.5685642496,2.293791533,1.986481737,1.709691264,1.517200488,3.303437362,2.860860676,2.462236835,4.005204947,3.468609231,4.472135954,.3745745885,.3243911095,.2791914141,.2477578022,1.101145788,.9536202253,.8207456118,1.809367160,1.566957926,2.415229459,.1705887719,.1477342101,.1271493634,.5721721350,.4955156045,1.035098339,.8650429586e-1,.7491491772e-1,.3194382825,.4761904762e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.748014747,1.936491673,1.677050983,1.443375673,1.280868845,1.161895004,1.224744872,1.060660172,.9128709292,.8100925872,.8366600264,.7245688370,.6236095646,.6060915266,.5248906592,.4587583064,3.872983346,3.354101966,2.886751347,2.561737690,2.323790008,3.,2.598076212,2.236067977,1.984313483,2.160246899,1.870828694,1.610152971,1.603567452,1.388730150,1.230977710,3.674234614,3.181980514,2.738612787,2.430277762,3.240370349,2.806243040,2.415229457,2.535462764,2.195775163,1.994411836,3.346640106,2.898275348,2.494438258,3.207134903,2.777460299,2.659215782,3.030457633,2.624453295,3.077444275,2.752549839},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,2.595128875,1.677050983,1.936491673,1.875000000,1.774823935,1.677050983,1.060660172,1.224744872,1.185854122,1.122497216,.7245688375,.8366600264,.8100925875,.5248906592,.6060915266,.3972963474,3.354101966,3.872983346,3.750000000,3.549647869,3.354101966,2.598076212,3.,2.904737510,2.749545418,1.870828694,2.160246900,2.091650066,1.388730150,1.603567452,1.066057968,3.181980514,3.674234614,3.557562366,3.367491648,2.806243040,3.240370350,3.137475099,2.195775165,2.535462764,1.727211316,2.898275350,3.346640106,3.240370350,2.777460300,3.207134903,2.302948421,2.624453295,3.030457633,2.665144921,2.383778084},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.931509850,1.443375673,1.875000000,1.936491673,1.909406540,1.855768722,.9128709293,1.185854122,1.224744872,1.207614729,.6236095646,.8100925872,.8366600265,.4517539516,.5868455971,.3419382528,2.886751347,3.750000000,3.872983346,3.818813079,3.711537446,2.236067977,2.904737510,3.,2.958039892,1.610152972,2.091650066,2.160246899,1.195228610,1.552647508,.9175166129,2.738612788,3.557562368,3.674234614,3.622844188,2.415229459,3.137475099,3.240370349,1.889822365,2.454951265,1.486546813,2.494438258,3.240370349,3.346640106,2.390457218,3.105295018,1.982062418,2.258769757,2.934227987,2.293791533,2.051629517},
			{2.091650066,2.898275350,3.118047822,3.162277659,3.137475099,3.083208205,1.280868846,1.774823935,1.909406540,1.936491673,1.921303269,.8100925874,1.122497216,1.207614729,1.224744872,.5533985903,.7668115804,.8249579111,.4008918626,.5554920599,.3034400977,2.561737692,3.549647870,3.818813080,3.872983346,3.842606539,1.984313483,2.749545418,2.958039892,3.,1.428869017,1.979898987,2.130032168,1.060660172,1.469693846,.8142152223,2.430277762,3.367491647,3.622844188,3.674234614,2.143303525,2.969848480,3.195048251,1.677050983,2.323790008,1.319179432,2.213594361,3.067246321,3.299831644,2.121320343,2.939387692,1.758905910,2.004459314,2.777460300,2.035538055,1.820640586},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,2.645751311,2.291287848,1.972026594,1.750000000,2.065591118,1.788854382,1.539600718,1.636634177,1.417366774,1.321374945,3.346640106,2.898275350,2.494438258,2.213594362,2.007984064,4.320493799,3.741657387,3.220305944,2.857738034,4.,3.464101616,2.981423969,3.464101616,3.,2.954684202,2.645751311,2.291287848,1.972026594,1.750000000,4.,3.464101616,2.981423969,4.107919181,3.557562368,3.829708432,2.065591118,1.788854382,1.539600718,3.464101616,3.,3.829708432,1.636634177,1.417366774,2.954684202,1.321374945},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,2.291287848,2.645751311,2.561737692,2.424871131,1.788854382,2.065591118,2.,1.417366774,1.636634177,1.144344270,2.898275350,3.346640106,3.240370350,3.067246323,2.898275350,3.741657387,4.320493800,4.183300132,3.959797974,3.464101616,4.,3.872983346,3.,3.464101616,2.558831578,2.291287848,2.645751311,2.561737692,2.424871131,3.464101616,4.,3.872983346,3.557562368,4.107919182,3.316624790,1.788854382,2.065591118,2.,3.,3.464101616,3.316624790,1.417366774,1.636634177,2.558831578,1.144344270},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,1.972026594,2.561737690,2.645751311,2.608745974,1.539600718,2.,2.065591118,1.219875090,1.584664228,.9848947341,2.494438258,3.240370349,3.346640106,3.299831644,3.207134903,3.220305942,4.183300132,4.320493799,4.260064335,2.981423969,3.872983346,4.,2.581988897,3.354101966,2.202291575,1.972026594,2.561737690,2.645751311,2.608745974,2.981423969,3.872983346,4.,3.061862179,3.977475643,2.854496129,1.539600718,2.,2.065591118,2.581988897,3.354101966,2.854496129,1.219875090,1.584664228,2.202291575,.9848947341},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,4.347413025,3.764970119,3.240370350,2.875543427,2.608447815,3.927922025,3.401680258,2.927700220,2.598076212,3.354101966,2.904737510,2.500000000,2.834733546,2.454951266,2.403122968,2.898275350,2.509980080,2.160246900,1.917028952,1.738965210,4.810702356,4.166190450,3.585685828,3.181980514,5.196152424,4.500000000,3.872983346,5.,4.330127020,4.605896842,1.963961013,1.700840129,1.463850110,1.299038106,3.897114318,3.375000000,2.904737510,4.743416490,4.107919182,4.974937185,1.341640786,1.161895004,1.,3.,2.598076212,3.979949748,.9449111824,.8183170887,2.302948420,.6866065625},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,3.764970119,4.347413025,4.209364560,3.984469852,3.764970119,3.401680258,3.927922024,3.803194146,3.600000000,2.904737510,3.354101966,3.247595265,2.454951265,2.834733546,2.081165539,2.509980080,2.898275350,2.806243040,2.656313234,2.509980080,4.166190449,4.810702356,4.657942526,4.409081537,4.500000000,5.196152424,5.031152948,4.330127020,5.,3.988823673,1.700840129,1.963961012,1.901597073,1.800000000,3.375000000,3.897114318,3.773364711,4.107919181,4.743416490,4.308421985,1.161895004,1.341640786,1.299038106,2.598076212,3.,3.446737588,.8183170885,.9449111824,1.994411836,.5946187254},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,5.085393248,4.404079742,3.790428332,3.363671463,3.051235949,5.025445453,4.352163430,3.745745885,3.324019728,4.577377083,3.964124837,3.411775437,4.062019202,3.517811818,3.577708763,2.542696624,2.202039870,1.895214166,1.681835732,1.525617974,4.923910840,4.264231874,3.670066451,3.256860890,5.909368405,5.117663157,4.404583150,6.141195788,5.318431565,6.,1.507633636,1.305649029,1.123723765,.9972059183,3.545621043,3.070597895,2.642749891,4.855041564,4.204589330,5.554920600,.9154754167,.7928249671,.6823550877,2.456478316,2.127372625,3.703280400,.5802884574,.5025445453,1.714285714,.3833259389},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.888071720,1.673320053,1.449137675,1.247219129,1.106797181,1.003992032,.8819171036,.7637626160,.6573421980,.5833333333,.5163977793,.4472135954,.3849001795,.3273268354,.2834733546,.2202291575,5.019960159,4.347413025,3.741657387,3.320391543,3.011976095,3.240370349,2.806243040,2.415229459,2.143303525,2.,1.732050808,1.490711985,1.299038106,1.125000000,.8864052607,5.291502622,4.582575695,3.944053189,3.500000000,4.,3.464101616,2.981423969,2.738612788,2.371708245,1.914854215,5.163977793,4.472135954,3.849001795,4.330127020,3.750000000,3.191423693,4.909902528,4.252100321,4.432026303,4.624812309},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.803059554,1.449137675,1.673320053,1.620185175,1.533623161,1.449137675,.7637626162,.8819171036,.8539125641,.8082903771,.4472135954,.5163977793,.5000000000,.2834733546,.3273268354,.1907240451,4.347413025,5.019960159,4.860555525,4.600869484,4.347413025,2.806243040,3.240370350,3.137475099,2.969848480,1.732050808,2.,1.936491673,1.125000000,1.299038106,.7676494735,4.582575696,5.291502622,5.123475385,4.849742262,3.464101616,4.,3.872983346,2.371708245,2.738612788,1.658312395,4.472135954,5.163977793,5.,3.750000000,4.330127020,2.763853992,4.252100321,4.909902529,3.838247368,4.005204947},
			{2.545875386,3.307189139,3.415650253,3.367876570,3.273268354,3.166392531,1.247219129,1.620185174,1.673320053,1.649915823,1.603567452,.6573421979,.8539125633,.8819171036,.8695819911,.3849001795,.5000000000,.5163977793,.2439750181,.3169328455,.1641491224,3.741657387,4.860555524,5.019960158,4.949747467,4.810702356,2.415229457,3.137475099,3.240370349,3.195048251,1.490711985,1.936491673,2.,.9682458365,1.257788237,.6606874726,3.944053188,5.123475380,5.291502622,5.217491946,2.981423969,3.872983346,4.,2.041241452,2.651650429,1.427248064,3.849001795,5.,5.163977793,3.227486122,4.192627457,2.378746774,3.659625271,4.753992682,3.303437362,3.447131570},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,2.068278942,3.055050464,2.645751311,2.277100171,2.020725943,1.833030278,2.070196679,1.792842914,1.543033500,1.369306394,1.414213562,1.224744872,1.054092553,.9960238407,.8625819493,.7237468643,4.582575696,3.968626966,3.415650256,3.031088914,2.749545418,5.070925527,4.391550330,3.779644731,3.354101966,4.107919181,3.557562368,3.061862179,3.162277660,2.738612788,2.427520781,4.140393357,3.585685828,3.086067000,2.738612788,5.477225575,4.743416490,4.082482906,5.,4.330127020,4.195235392,3.535533905,3.061862179,2.635231383,5.270462768,4.564354646,5.244044240,2.988071523,2.587745849,4.855041564,2.533114026},
			{3.240370348,3.741657386,3.622844188,3.429285640,3.240370348,3.070597895,2.645751311,3.055050463,2.958039891,2.800000000,2.645751311,1.792842914,2.070196679,2.004459314,1.897366596,1.224744872,1.414213562,1.369306394,.8625819493,.9960238407,.6267831704,3.968626966,4.582575695,4.437059836,4.200000000,3.968626966,4.391550326,5.070925527,4.909902528,4.647580015,3.557562368,4.107919182,3.977475643,2.738612788,3.162277660,2.102294664,3.585685828,4.140393357,4.008918626,3.794733192,4.743416490,5.477225576,5.303300858,4.330127020,5.,3.633180426,3.061862179,3.535533905,3.423265985,4.564354646,5.270462768,4.541475532,2.587745849,2.988071523,4.204589330,2.193741096},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,3.256860892,2.820524267,2.427520781,2.154210993,2.472066162,2.140872097,1.842569328,1.880349512,1.628430445,1.449137675,4.119639374,3.567712353,3.070597897,2.724885320,2.471783625,5.983235509,5.181633950,4.459640440,3.957538298,5.744562647,4.974937185,4.281744193,4.974937185,4.308421985,4.166190450,3.256860892,2.820524267,2.427520781,2.154210993,5.744562647,4.974937185,4.281744193,6.292853088,5.449770639,6.,2.472066162,2.140872097,1.842569328,4.974937185,4.308421985,6.,1.880349512,1.628430445,4.166190450,1.449137675},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.962141687,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,.6546536711,.5669467095,.4879500367,.4330127020,.3354101966,.2904737510,.2500000000,.1889822365,.1636634177,.1144344270,5.796550700,5.019960159,4.320493800,3.834057903,3.477930420,3.207134903,2.777460300,2.390457218,2.121320343,1.732050808,1.500000000,1.290994449,1.,.8660254040,.6141195788,6.546536711,5.669467095,4.879500367,4.330127020,4.330127020,3.750000000,3.227486122,2.635231383,2.282177324,1.658312395,6.708203931,5.809475019,5.,5.,4.330127020,3.316624790,6.614378278,5.728219620,5.373546315,6.408327913},
			{3.074085231,3.549647870,3.436931771,3.253306011,3.074085231,2.913024938,1.254990040,1.449137675,1.403121520,1.328156617,1.254990040,.5669467095,.6546536709,.6338656908,.6000000000,.2904737510,.3354101966,.3247595265,.1636634177,.1889822365,.9910312088e-1,5.019960159,5.796550700,5.612486080,5.312626469,5.019960159,2.777460299,3.207134903,3.105295018,2.939387692,1.500000000,1.732050808,1.677050983,.8660254040,1.,.5318431565,5.669467095,6.546536709,6.338656908,6.,3.750000000,4.330127020,4.192627457,2.282177323,2.635231383,1.436140662,5.809475019,6.708203931,6.495190530,4.330127020,5.,2.872281324,5.728219619,6.614378278,4.653627619,5.549774770},
			{3.924283374,3.398529094,2.924988129,2.595669470,2.354570024,2.169229254,2.746426250,2.378474901,2.047065264,1.816590213,1.647855750,1.628430445,1.410262133,1.213760391,1.077105496,.9888264647,.8563488387,.7370277310,.6267831705,.5428101485,.4140393357,5.492852500,4.756949802,4.094130528,3.633180426,3.295711500,5.318431565,4.605896844,3.964124837,3.517811820,3.829708432,3.316624790,2.854496129,2.653299832,2.297825059,1.851640200,5.428101485,4.700873779,4.045867967,3.590351654,6.382847385,5.527707984,4.757493549,5.244044240,4.541475532,4.,4.944132325,4.281744193,3.685138655,6.633249580,5.744562647,6.,4.387482194,3.799671040,6.480740700,3.864367134},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.633180424,3.146426544,2.708012801,2.403122968,2.179908254,2.008316044,1.271348312,1.101019935,.9476070828,.8409178659,.7628089876,.5025445453,.4352163430,.3745745885,.3324019728,.2288688541,.1982062418,.1705887719,.1160576915,.1005089091,.6388765648e-1,6.356741560,5.505099678,4.738035416,4.204589330,3.814044936,3.077444275,2.665144921,2.293791533,2.035538055,1.477342101,1.279415789,1.101145788,.7676494735,.6648039455,.4285714286,7.538168182,6.528245145,5.618618828,4.986029592,4.432026303,3.838247368,3.303437362,2.427520781,2.102294664,1.388730150,8.010409897,6.937218462,5.970607017,5.373546315,4.653627619,3.240370350,8.124038403,7.035623637,6.,8.049844717},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,6.8546484667622e-008,1.09461601509022e-009,7.65361974719019e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,.122554596919714,.00195706935488603,2.32941342728397e-005,2.54112819850642e-007,.00244089433681864,3.89785419960172e-005,4.63944411923424e-007,3.89785419960172e-005,6.22446745908012e-007,5.51740069821357e-007,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,.262307293803493,.00418877446399711,4.98571369236328e-005,.00662303395559898,.000105762958835416,.000128085303901421,7.27823584467216,.116225851033514,.00138338509726576,.450140785461579,.0071882798268767,.0137645079246526,9.14177937313726,.145984701575574,.68472894504081,10.9691712594614},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,5.01144631725624e-005,8.00275816365055e-007,2.90963994598978e-006,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,.612365059546825,.00977883263593246,.000116393136443702,1.26971828042803e-006,.0634197097850823,.00101274675595133,1.20542784326106e-005,.00526617557576491,8.40953427875793e-005,.000387613365377294,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,1.2594400120878,.0201119461265218,.000239383633673578,.165355473459674,.00264055480463631,.0166285878943821,6.45780747384476,.103124463877826,.00122744505825102,2.07683410541886,.0331648790465492,.330223979417038,7.79428043908953,.124466546112493,3.03569721712443,8.98680735137037},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.0113956405123916,.000181976518058871,2.16598631798674e-006,.00166206328995312,2.65414208197256e-005,.000231576690958128,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,1.36595759702526,.0218129211013044,.000259629589389913,2.83226696917689e-006,.339486246269225,.00542124200448682,6.45266550483906e-005,.067649468233801,.00108029159656201,.0119492023316399,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,2.61131941894657,.0417000531735786,.000496337360406356,.822758173894117,.0131385916878024,.198554676115027,5.18623489409549,.0828187732700635,.000985755369367884,4.0025799484364,.0639170357987087,1.5272774292324,5.81833621896758,.0929127735338078,5.43816471058122,6.23568135536041},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.0657431866024243,.00104985026256264,1.24959051250218e-005,.0171974308138302,.000274625070541602,.00429748986298374,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,2.19713005267825,.0350858801127846,.000417611772616793,4.55567499953034e-006,.979365349166059,.0156394452705675,.000186149426512757,.350018452610409,.00558943038088415,.110884150015163,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.76697803717186,.0601547184591987,.000715995110407768,2.12867470143779,.0339927193989142,.921342735777177,3.7410698100085,.0597409910483066,.000711070695350591,5.17830525859385,.0826921452798168,3.543802987655,3.76406389114815,.0601081826012263,6.30978459469567,3.6179005979755},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,.236321146004057,.00377380273155657,4.4917911225712e-005,.0946962283839188,.00151220020487266,.0362494356624709,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.90029193047496,.0463146434325445,.000551262795170185,6.01365741772922e-006,1.98037740107218,.0316245658532948,.000376413273965959,1.08420509999131,.0173136269811071,.526147189418459,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,4.28496925234002,.0684264990231684,.00081445047001826,3.70920808974538,.0592321925476606,2.4592936780291,2.39387697960251,.0382277237453785,.000455007753107355,5.07587527588423,.0810564450674401,5.32120685033247,2.07554270972356,.0331442567927967,5.32974583920556,1.71909544841957},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,.626153181688297,.00999901459257298,.000119013865260662,.347192252628355,.00554429890730429,.183906859961414,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,3.30149822057238,.0527214902997539,.000627520670660238,6.84554511742325e-006,3.11944010216456,.0498142116158778,.000592916613348927,2.36319452013471,.0377377568190159,1.58691847335237,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,4.0126111958998,.0640772243410266,.000762682969712426,4.80639993236566,.076753204286926,4.40969381865368,1.33269841309056,.0212818065446243,.000253307966815722,3.91021223386622,.0624420195354773,5.67229573503973,.95054294779362,.0151791815252261,3.37758202463881,.647664476709677},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,1.33269841309056,.0212818065446243,.000253307966815722,.95054294779362,.0151791815252261,.647664476709677,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,3.30149822057238,.0527214902997539,.000627520670660238,6.84554511742325e-006,4.0126111958998,.0640772243410266,.000762682969712426,3.91021223386622,.0624420195354773,3.37758202463881,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,3.11944010216456,.0498142116158778,.000592916613348927,4.80639993236566,.076753204286926,5.67229573503973,.626153181688297,.00999901459257298,.000119013865260662,2.36319452013471,.0377377568190159,4.40969381865368,.347192252628355,.00554429890730429,1.58691847335237,.183906859961414},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,2.39387697960251,.0382277237453785,.000455007753107355,2.07554270972356,.0331442567927967,1.71909544841957,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,2.90029193047496,.0463146434325445,.000551262795170185,6.01365741772922e-006,4.28496925234002,.0684264990231684,.00081445047001826,5.07587527588423,.0810564450674401,5.32974583920556,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,1.98037740107218,.0316245658532948,.000376413273965959,3.70920808974538,.0592321925476606,5.32120685033247,.236321146004057,.00377380273155657,4.4917911225712e-005,1.08420509999131,.0173136269811071,2.4592936780291,.0946962283839188,.00151220020487266,.526147189418459,.0362494356624709},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.7410698100085,.0597409910483066,.000711070695350591,3.76406389114815,.0601081826012263,3.6179005979755,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,2.19713005267825,.0350858801127846,.000417611772616793,4.55567499953034e-006,3.76697803717186,.0601547184591987,.000715995110407768,5.17830525859385,.0826921452798168,6.30978459469567,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.979365349166059,.0156394452705675,.000186149426512757,2.12867470143779,.0339927193989142,3.543802987655,.0657431866024243,.00104985026256264,1.24959051250218e-005,.350018452610409,.00558943038088415,.921342735777177,.0171974308138302,.000274625070541602,.110884150015163,.00429748986298374},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,5.18623489409549,.0828187732700635,.000985755369367884,5.81833621896758,.0929127735338078,6.23568135536041,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,1.36595759702526,.0218129211013044,.000259629589389913,2.83226696917689e-006,2.61131941894657,.0417000531735786,.000496337360406356,4.0025799484364,.0639170357987087,5.43816471058122,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.339486246269225,.00542124200448682,6.45266550483906e-005,.822758173894117,.0131385916878024,1.5272774292324,.0113956405123916,.000181976518058871,2.16598631798674e-006,.067649468233801,.00108029159656201,.198554676115027,.00166206328995312,2.65414208197256e-005,.0119492023316399,.000231576690958128},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,6.45780747384476,.103124463877826,.00122744505825102,7.79428043908953,.124466546112493,8.98680735137037,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.612365059546825,.00977883263593246,.000116393136443702,1.26971828042803e-006,1.2594400120878,.0201119461265218,.000239383633673578,2.07683410541886,.0331648790465492,3.03569721712443,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.0634197097850823,.00101274675595133,1.20542784326106e-005,.165355473459674,.00264055480463631,.330223979417038,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,.00526617557576491,8.40953427875793e-005,.0166285878943821,5.01144631725624e-005,8.00275816365055e-007,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,7.27823584467216,.116225851033514,.00138338509726576,9.14177937313726,.145984701575574,10.9691712594614,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.122554596919714,.00195706935488603,2.32941342728397e-005,2.54112819850642e-007,.262307293803493,.00418877446399711,4.98571369236328e-005,.450140785461579,.0071882798268767,.68472894504081,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,.00244089433681864,3.89785419960172e-005,4.63944411923424e-007,.00662303395559898,.000105762958835416,.0137645079246526,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,3.89785419960172e-005,6.22446745908012e-007,.000128085303901421,6.8546484667622e-008,1.09461601509022e-009,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,6.8546484667622e-008,5.69188702865245e-009,7.65361974719019e-010,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,.122554596919714,.0101765528017885,.000629847662741689,3.57281062205031e-005,.00244089433681864,.0002026842788013,1.25445444792834e-005,3.89785419960172e-005,3.23665697396975e-006,5.51740069821357e-007,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,.262307293803493,.0217811823588299,.0013480819160108,.00662303395559898,.000549956153763738,.000128085303901421,7.27823584467216,.60436208201944,.037405205103521,.450140785461579,.037378291693961,.0137645079246526,9.14177937313726,.759104944404841,.68472894504081,10.9691712594614},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,5.01144631725624e-005,4.16134925463982e-006,2.90963994598978e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,.612365059546825,.0508488911805658,.0031471418550943,.000178521609495778,.0634197097850823,.00526617557576491,.000325934375061529,.00526617557576491,.000437286851447401,.000387613365377294,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,1.2594400120878,.104579983939912,.00647266905916808,.165355473459674,.0137306045533387,.0166285878943821,6.45780747384476,.53623626019628,.0331887586818224,2.07683410541886,.172453848883548,.330223979417038,7.79428043908953,.647212821256298,3.03569721712443,8.98680735137037},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.0113956405123916,.000946258567734596,5.85658777413804e-005,.00166206328995312,.000138012569524986,.000231576690958128,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,1.36595759702526,.11342487316276,.00702009734040598,.000398214994344052,.339486246269225,.0281898826795267,.0017447294838579,.067649468233801,.00561740157357991,.0119492023316399,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,2.61131941894657,.216835847896165,.0134204140287723,.822758173894117,.0683192814389796,.198554676115027,5.18623489409549,.430648825529863,.0266537364367886,4.0025799484364,.332361798070895,1.5272774292324,5.81833621896758,.483136554903789,5.43816471058122,6.23568135536041},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.0657431866024243,.0054591098696972,.000337875473054763,.0171974308138302,.00142802120123345,.00429748986298374,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,2.19713005267825,.182442850414858,.0112917610860855,.000640525103711722,.979365349166059,.0813234544750617,.00503327489386337,.350018452610409,.029064444374942,.110884150015163,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.76697803717186,.31279814746903,.0193597169803656,2.12867470143779,.176758530797994,.921342735777177,3.7410698100085,.310646808870845,.0192265662361566,5.17830525859385,.429990373428657,3.543802987655,3.76406389114815,.31255616594975,6.30978459469567,3.6179005979755},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,.236321146004057,.0196233734207421,.0012145307084342,.0946962283839188,.00786328046747136,.0362494356624709,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.90029193047496,.240831227166574,.0149055371205287,.000845516535217038,1.98037740107218,.164444383862132,.0101777991856447,1.08420509999131,.0900290215693292,.526147189418459,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,4.28496925234002,.355810527926517,.0220218411618753,3.70920808974538,.308001110702629,2.4592936780291,2.39387697960251,.198780103635972,.0123029070952925,5.07587527588423,.421484906038046,5.32120685033247,2.07554270972356,.172346615354148,5.32974583920556,1.71909544841957},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,.626153181688297,.0519938139714517,.00321800346775215,.347192252628355,.028829765505355,.183906859961414,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,3.30149822057238,.274146150459585,.0169674658481853,.000962479434277651,3.11944010216456,.259028610060664,.016031810358111,2.36319452013471,.196232327654089,1.58691847335237,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,4.0126111958998,.333194761478692,.0206221051299754,4.80639993236566,.399108510989818,4.40969381865368,1.33269841309056,.110663133873164,.00684916764813009,3.91021223386622,.32469187015097,5.67229573503973,.95054294779362,.0789301318799674,3.37758202463881,.647664476709677},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,1.33269841309056,.110663133873164,.00684916764813009,.95054294779362,.0789301318799674,.647664476709677,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,3.30149822057238,.274146150459585,.0169674658481853,.000962479434277651,4.0126111958998,.333194761478692,.0206221051299754,3.91021223386622,.32469187015097,3.37758202463881,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,3.11944010216456,.259028610060664,.016031810358111,4.80639993236566,.399108510989818,5.67229573503973,.626153181688297,.0519938139714517,.00321800346775215,2.36319452013471,.196232327654089,4.40969381865368,.347192252628355,.028829765505355,1.58691847335237,.183906859961414},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,2.39387697960251,.198780103635972,.0123029070952925,2.07554270972356,.172346615354148,1.71909544841957,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,2.90029193047496,.240831227166574,.0149055371205287,.000845516535217038,4.28496925234002,.355810527926517,.0220218411618753,5.07587527588423,.421484906038046,5.32974583920556,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,1.98037740107218,.164444383862132,.0101777991856447,3.70920808974538,.308001110702629,5.32120685033247,.236321146004057,.0196233734207421,.0012145307084342,1.08420509999131,.0900290215693292,2.4592936780291,.0946962283839188,.00786328046747136,.526147189418459,.0362494356624709},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.7410698100085,.310646808870845,.0192265662361566,3.76406389114815,.31255616594975,3.6179005979755,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,2.19713005267825,.182442850414858,.0112917610860855,.000640525103711722,3.76697803717186,.31279814746903,.0193597169803656,5.17830525859385,.429990373428657,6.30978459469567,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.979365349166059,.0813234544750617,.00503327489386337,2.12867470143779,.176758530797994,3.543802987655,.0657431866024243,.0054591098696972,.000337875473054763,.350018452610409,.029064444374942,.921342735777177,.0171974308138302,.00142802120123345,.110884150015163,.00429748986298374},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,5.18623489409549,.430648825529863,.0266537364367886,5.81833621896758,.483136554903789,6.23568135536041,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,1.36595759702526,.11342487316276,.00702009734040598,.000398214994344052,2.61131941894657,.216835847896165,.0134204140287723,4.0025799484364,.332361798070895,5.43816471058122,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.339486246269225,.0281898826795267,.0017447294838579,.822758173894117,.0683192814389796,1.5272774292324,.0113956405123916,.000946258567734596,5.85658777413804e-005,.067649468233801,.00561740157357991,.198554676115027,.00166206328995312,.000138012569524986,.0119492023316399,.000231576690958128},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,6.45780747384476,.53623626019628,.0331887586818224,7.79428043908953,.647212821256298,8.98680735137037,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.612365059546825,.0508488911805658,.0031471418550943,.000178521609495778,1.2594400120878,.104579983939912,.00647266905916808,2.07683410541886,.172453848883548,3.03569721712443,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.0634197097850823,.00526617557576491,.000325934375061529,.165355473459674,.0137306045533387,.330223979417038,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,.00526617557576491,.000437286851447401,.0166285878943821,5.01144631725624e-005,4.16134925463982e-006,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,7.27823584467216,.60436208201944,.037405205103521,9.14177937313726,.759104944404841,10.9691712594614,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.122554596919714,.0101765528017885,.000629847662741689,3.57281062205031e-005,.262307293803493,.0217811823588299,.0013480819160108,.450140785461579,.037378291693961,.68472894504081,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,.00244089433681864,.0002026842788013,1.25445444792834e-005,.00662303395559898,.000549956153763738,.0137645079246526,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,3.89785419960172e-005,3.23665697396975e-006,.000128085303901421,6.8546484667622e-008,5.69188702865245e-009,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,6.8546484667622e-008,1.36592668764407e-008,7.65361974719019e-010,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,.122554596919714,.0244214704020099,.00362725225603475,.000493768468568094,.00244089433681864,.000486397330423004,7.22432263467335e-005,3.89785419960172e-005,7.76725911325977e-006,5.51740069821357e-007,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,.262307293803493,.0522700083866766,.00776351721285218,.00662303395559898,.00131977283399776,.000128085303901421,7.27823584467216,1.45033499905202,.215414174926983,.450140785461579,.0896996126351677,.0137645079246526,9.14177937313726,1.82168355953934,.68472894504081,10.9691712594614},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.000824565926330002,.000164311358893045,2.44047036237916e-005,5.01144631725624e-005,9.98631556618579e-006,2.90963994598978e-006,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,.612365059546825,.122026064732152,.0181241879095991,.00246719882612911,.0634197097850823,.0126376537772916,.00187703514229927,.00526617557576491,.00104939148922157,.000387613365377294,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,1.2594400120878,.250968774218159,.0372756855923469,.165355473459674,.0329504066094101,.0166285878943821,6.45780747384476,1.28684813137965,.191131930663375,2.07683410541886,.413851001242982,.330223979417038,7.79428043908953,1.55316727245144,3.03569721712443,8.98680735137037},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.0113956405123916,.00227081076025246,.000337277130218073,.00166206328995312,.000331199567025036,.000231576690958128,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,1.36595759702526,.272194547284149,.0404282899212952,.0055033985485992,.339486246269225,.067649468233801,.0100477850929001,.067649468233801,.0134805182990133,.0119492023316399,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,2.61131941894657,.520357958596769,.0772873028549269,.822758173894117,.16395112785691,.198554676115027,5.18623489409549,1.03346169878758,.153497157076452,4.0025799484364,.797594624863601,1.5272774292324,5.81833621896758,1.1594206118243,5.43816471058122,6.23568135536041},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.0657431866024243,.0131006533057736,.00194580315907169,.0171974308138302,.00342693426532766,.00429748986298374,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,2.19713005267825,.437822390179288,.0650285272089779,.00885216522776848,.979365349166059,.195158259887658,.0289863069976914,.350018452610409,.0697482223969557,.110884150015163,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.76697803717186,.750646201027513,.111491367273716,2.12867470143779,.424181283413309,.921342735777177,3.7410698100085,.745483465381799,.110724560685561,5.17830525859385,1.0318815597352,3.543802987655,3.76406389114815,.750065499032834,6.30978459469567,3.6179005979755},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,.236321146004057,.0470917453597858,.00699440438187064,.0946962283839188,.0188701296931614,.0362494356624709,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.90029193047496,.577941548644503,.0858400314014041,.0116851814693591,1.98037740107218,.394630061032062,.0586132921362272,1.08420509999131,.216049690735243,.526147189418459,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,4.28496925234002,.853866377517717,.126822369537332,3.70920808974538,.739134376376297,2.4592936780291,2.39387697960251,.477028175651108,.0708516521644813,5.07587527588423,1.01147032381073,5.32120685033247,2.07554270972356,.413593664547998,5.32974583920556,1.71909544841957},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,.626153181688297,.124773625580585,.0185322753878643,.347192252628355,.0691850451424154,.183906859961414,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,3.30149822057238,.657889978038246,.0977145465764208,.0133016285094573,3.11944010216456,.621611232907719,.0923261666744144,2.36319452013471,.470914078181719,1.58691847335237,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,4.0126111958998,.799593552359549,.118761379587061,4.80639993236566,.957771936938711,4.40969381865368,1.33269841309056,.26556698531571,.0394439167964793,3.91021223386622,.77918849841486,5.67229573503973,.95054294779362,.189414816301507,3.37758202463881,.647664476709677},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,1.33269841309056,.26556698531571,.0394439167964793,.95054294779362,.189414816301507,.647664476709677,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,3.30149822057238,.657889978038246,.0977145465764208,.0133016285094573,4.0126111958998,.799593552359549,.118761379587061,3.91021223386622,.77918849841486,3.37758202463881,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,3.11944010216456,.621611232907719,.0923261666744144,4.80639993236566,.957771936938711,5.67229573503973,.626153181688297,.124773625580585,.0185322753878643,2.36319452013471,.470914078181719,4.40969381865368,.347192252628355,.0691850451424154,1.58691847335237,.183906859961414},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,2.39387697960251,.477028175651108,.0708516521644813,2.07554270972356,.413593664547998,1.71909544841957,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,2.90029193047496,.577941548644503,.0858400314014041,.0116851814693591,4.28496925234002,.853866377517717,.126822369537332,5.07587527588423,1.01147032381073,5.32974583920556,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,1.98037740107218,.394630061032062,.0586132921362272,3.70920808974538,.739134376376297,5.32120685033247,.236321146004057,.0470917453597858,.00699440438187064,1.08420509999131,.216049690735243,2.4592936780291,.0946962283839188,.0188701296931614,.526147189418459,.0362494356624709},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.7410698100085,.745483465381799,.110724560685561,3.76406389114815,.750065499032834,3.6179005979755,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,2.19713005267825,.437822390179288,.0650285272089779,.00885216522776848,3.76697803717186,.750646201027513,.111491367273716,5.17830525859385,1.0318815597352,6.30978459469567,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.979365349166059,.195158259887658,.0289863069976914,2.12867470143779,.424181283413309,3.543802987655,.0657431866024243,.0131006533057736,.00194580315907169,.350018452610409,.0697482223969557,.921342735777177,.0171974308138302,.00342693426532766,.110884150015163,.00429748986298374},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,5.18623489409549,1.03346169878758,.153497157076452,5.81833621896758,1.1594206118243,6.23568135536041,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,1.36595759702526,.272194547284149,.0404282899212952,.0055033985485992,2.61131941894657,.520357958596769,.0772873028549269,4.0025799484364,.797594624863601,5.43816471058122,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.339486246269225,.067649468233801,.0100477850929001,.822758173894117,.16395112785691,1.5272774292324,.0113956405123916,.00227081076025246,.000337277130218073,.067649468233801,.0134805182990133,.198554676115027,.00166206328995312,.000331199567025036,.0119492023316399,.000231576690958128},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,6.45780747384476,1.28684813137965,.191131930663375,7.79428043908953,1.55316727245144,8.98680735137037,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.612365059546825,.122026064732152,.0181241879095991,.00246719882612911,1.2594400120878,.250968774218159,.0372756855923469,2.07683410541886,.413851001242982,3.03569721712443,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.0634197097850823,.0126376537772916,.00187703514229927,.165355473459674,.0329504066094101,.330223979417038,.000824565926330002,.000164311358893045,2.44047036237916e-005,.00526617557576491,.00104939148922157,.0166285878943821,5.01144631725624e-005,9.98631556618579e-006,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,7.27823584467216,1.45033499905202,.215414174926983,9.14177937313726,1.82168355953934,10.9691712594614,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.122554596919714,.0244214704020099,.00362725225603475,.000493768468568094,.262307293803493,.0522700083866766,.00776351721285218,.450140785461579,.0896996126351677,.68472894504081,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,.00244089433681864,.000486397330423004,7.22432263467335e-005,.00662303395559898,.00131977283399776,.0137645079246526,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,3.89785419960172e-005,7.76725911325977e-006,.000128085303901421,6.8546484667622e-008,1.36592668764407e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,6.8546484667622e-008,2.44980430684131e-008,7.65361974719019e-010,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,.122554596919714,.0438001716427634,.0116677137630873,.00284862553642039,.00244089433681864,.000872358879642074,.00023238342051707,3.89785419960172e-005,1.39306633365776e-005,5.51740069821357e-007,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,.262307293803493,.0937468261090737,.0249727590585012,.00662303395559898,.0023670268704951,.000128085303901421,7.27823584467216,2.60119152746664,.692918704338266,.450140785461579,.160877226679454,.0137645079246526,9.14177937313726,3.26720919228747,.68472894504081,10.9691712594614},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.000824565926330002,.000294694201614443,7.85021488046878e-005,5.01144631725624e-005,1.79105651165762e-005,2.90963994598978e-006,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,.612365059546825,.218855068600592,.0582997326325419,.0142336459837523,.0634197097850823,.0226657689113441,.00603782345911206,.00526617557576491,.00188209500049501,.000387613365377294,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,1.2594400120878,.450115214472287,.119903993192299,.165355473459674,.0590969110964026,.0166285878943821,6.45780747384476,2.30797605978546,.614810468242533,2.07683410541886,.74224625260403,.330223979417038,7.79428043908953,2.78562232344933,3.03569721712443,8.98680735137037},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.0113956405123916,.00407272369067121,.00108491296894855,.00166206328995312,.000594010010245393,.000231576690958128,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,1.36595759702526,.488183868334511,.130044915940983,.0317499448437878,.339486246269225,.121330053899551,.0323205203668776,.067649468233801,.0241774555590068,.0119492023316399,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,2.61131941894657,.933267634054527,.248608606068695,.822758173894117,.294048123349947,.198554676115027,5.18623489409549,1.85352474883709,.493751403486214,4.0025799484364,1.43049459738903,1.5272774292324,5.81833621896758,2.07943342346341,5.43816471058122,6.23568135536041},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.0657431866024243,.0234961635798143,.00625902823868614,.0171974308138302,.0061462437174735,.00429748986298374,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,2.19713005267825,.785239198263813,.209176034185977,.0510694901064878,.979365349166059,.350018452610409,.0932397057677362,.350018452610409,.125094192220009,.110884150015163,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.76697803717186,1.3462920908025,.358632173497365,2.12867470143779,.760773725550181,.921342735777177,3.7410698100085,1.33703266851113,.356165601240813,5.17830525859385,1.85069075233423,3.543802987655,3.76406389114815,1.34525059548623,6.30978459469567,3.6179005979755},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,.236321146004057,.0844595552914031,.0224987683542483,.0946962283839188,.0338437819621889,.0362494356624709,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.90029193047496,1.03654442641711,.276120006304964,.0674135925038969,1.98037740107218,.707773288179075,.188540268799867,1.08420509999131,.387487460083106,.526147189418459,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,4.28496925234002,1.53141859517937,.407947118664331,3.70920808974538,1.32564550862114,2.4592936780291,2.39387697960251,.855555199093713,.227907170151078,5.07587527588423,1.81408297965651,5.32120685033247,2.07554270972356,.741784716454702,5.32974583920556,1.71909544841957},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,.626153181688297,.223782848737467,.0596124199094679,.347192252628355,.12408412771496,.183906859961414,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,3.30149822057238,1.17993280035085,.314316534794822,.0767391217950806,3.11944010216456,1.11486647807198,.296983835024965,2.36319452013471,.844589499068632,1.58691847335237,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,4.0126111958998,1.43407969550076,.382017484674746,4.80639993236566,1.71777434126509,4.40969381865368,1.33269841309056,.476297264169728,.126878501518793,3.91021223386622,1.39748301027071,5.67229573503973,.95054294779362,.339717524338981,3.37758202463881,.647664476709677},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,1.33269841309056,.476297264169728,.126878501518793,.95054294779362,.339717524338981,.647664476709677,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,3.30149822057238,1.17993280035085,.314316534794822,.0767391217950806,4.0126111958998,1.43407969550076,.382017484674746,3.91021223386622,1.39748301027071,3.37758202463881,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,3.11944010216456,1.11486647807198,.296983835024965,4.80639993236566,1.71777434126509,5.67229573503973,.626153181688297,.223782848737467,.0596124199094679,2.36319452013471,.844589499068632,4.40969381865368,.347192252628355,.12408412771496,1.58691847335237,.183906859961414},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,2.39387697960251,.855555199093713,.227907170151078,2.07554270972356,.741784716454702,1.71909544841957,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,2.90029193047496,1.03654442641711,.276120006304964,.0674135925038969,4.28496925234002,1.53141859517937,.407947118664331,5.07587527588423,1.81408297965651,5.32974583920556,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,1.98037740107218,.707773288179075,.188540268799867,3.70920808974538,1.32564550862114,5.32120685033247,.236321146004057,.0844595552914031,.0224987683542483,1.08420509999131,.387487460083106,2.4592936780291,.0946962283839188,.0338437819621889,.526147189418459,.0362494356624709},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.7410698100085,1.33703266851113,.356165601240813,3.76406389114815,1.34525059548623,3.6179005979755,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,2.19713005267825,.785239198263813,.209176034185977,.0510694901064878,3.76697803717186,1.3462920908025,.358632173497365,5.17830525859385,1.85069075233423,6.30978459469567,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.979365349166059,.350018452610409,.0932397057677362,2.12867470143779,.760773725550181,3.543802987655,.0657431866024243,.0234961635798143,.00625902823868614,.350018452610409,.125094192220009,.921342735777177,.0171974308138302,.0061462437174735,.110884150015163,.00429748986298374},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,5.18623489409549,1.85352474883709,.493751403486214,5.81833621896758,2.07943342346341,6.23568135536041,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,1.36595759702526,.488183868334511,.130044915940983,.0317499448437878,2.61131941894657,.933267634054527,.248608606068695,4.0025799484364,1.43049459738903,5.43816471058122,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.339486246269225,.121330053899551,.0323205203668776,.822758173894117,.294048123349947,1.5272774292324,.0113956405123916,.00407272369067121,.00108491296894855,.067649468233801,.0241774555590068,.198554676115027,.00166206328995312,.000594010010245393,.0119492023316399,.000231576690958128},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,6.45780747384476,2.30797605978546,.614810468242533,7.79428043908953,2.78562232344933,8.98680735137037,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.612365059546825,.218855068600592,.0582997326325419,.0142336459837523,1.2594400120878,.450115214472287,.119903993192299,2.07683410541886,.74224625260403,3.03569721712443,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.0634197097850823,.0226657689113441,.00603782345911206,.165355473459674,.0590969110964026,.330223979417038,.000824565926330002,.000294694201614443,7.85021488046878e-005,.00526617557576491,.00188209500049501,.0166285878943821,5.01144631725624e-005,1.79105651165762e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,7.27823584467216,2.60119152746664,.692918704338266,9.14177937313726,3.26720919228747,10.9691712594614,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.122554596919714,.0438001716427634,.0116677137630873,.00284862553642039,.262307293803493,.0937468261090737,.0249727590585012,.450140785461579,.160877226679454,.68472894504081,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,.00244089433681864,.000872358879642074,.00023238342051707,.00662303395559898,.0023670268704951,.0137645079246526,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,3.89785419960172e-005,1.39306633365776e-005,.000128085303901421,6.8546484667622e-008,2.44980430684131e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,6.8546484667622e-008,3.7527416869482e-008,7.65361974719019e-010,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,.122554596919714,.0670954531185465,.0273791727414708,.0102396991398416,.00244089433681864,.00133632614020231,.000545305270748019,3.89785419960172e-005,2.13397375798647e-005,5.51740069821357e-007,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,.262307293803493,.143606418429484,.058600467750326,.00662303395559898,.00362593876834461,.000128085303901421,7.27823584467216,3.98464475452196,1.62598614322311,.450140785461579,.246440368054962,.0137645079246526,9.14177937313726,5.00488635016176,.68472894504081,10.9691712594614},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.000824565926330002,.000451428390509428,.000184211229066994,5.01144631725624e-005,2.74363646769969e-005,2.90963994598978e-006,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,.612365059546825,.335253937240527,.13680473166712,.051164412687175,.0634197097850823,.0347206410009163,.0141682093704184,.00526617557576491,.00288309410977452,.000387613365377294,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,1.2594400120878,.689510637467093,.281363788027607,.165355473459674,.0905278193943952,.0166285878943821,6.45780747384476,3.53548156799627,1.44269926011819,2.07683410541886,1.13701263662145,.330223979417038,7.79428043908953,4.26716574385505,3.03569721712443,8.98680735137037},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.0113956405123916,.00623881668046779,.00254583033055523,.00166206328995312,.000909936406629402,.000231576690958128,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,1.36595759702526,.747826244112006,.305160230186869,.114128683728465,.339486246269225,.18585988679871,.0758425453509597,.067649468233801,.0370363237207337,.0119492023316399,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,2.61131941894657,1.42962943840673,.58337889571007,.822758173894117,.450438693157099,.198554676115027,5.18623489409549,2.83932866528416,1.15862500931079,4.0025799484364,2.19130838066721,1.5272774292324,5.81833621896758,3.18538769471191,5.43816471058122,6.23568135536041},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.0657431866024243,.0359926841107618,.0146872831147845,.0171974308138302,.00941514591687767,.00429748986298374,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,2.19713005267825,1.20287153766575,.490847383612732,.183574923144401,.979365349166059,.536177046731461,.218794021040783,.350018452610409,.191625995851247,.110884150015163,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.76697803717186,2.06232246300417,.841557517454501,2.12867470143779,1.16539401381336,.921342735777177,3.7410698100085,2.04813838310324,.835769519114577,5.17830525859385,2.83498739737667,3.543802987655,3.76406389114815,2.06072704459494,6.30978459469567,3.6179005979755},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,.236321146004057,.129379678661691,.0527950614618356,.0946962283839188,.0518437211734884,.0362494356624709,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.90029193047496,1.58783437049499,.647936476974322,.242325604524073,1.98037740107218,1.08420509999131,.442423999509308,1.08420509999131,.593574082861527,.526147189418459,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,4.28496925234002,2.34590917578526,.957278765839435,3.70920808974538,2.03069492058023,2.4592936780291,2.39387697960251,1.31058536069927,.534801410737989,5.07587527588423,2.77890964691703,5.32120685033247,2.07554270972356,1.13630563078316,5.32974583920556,1.71909544841957},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,.626153181688297,.342802575265255,.139885051637275,.347192252628355,.190078724845143,.183906859961414,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,3.30149822057238,1.80748437551057,.737567519771828,.275847249626467,3.11944010216456,1.70781229176476,.69689502893851,2.36319452013471,1.29378751301169,1.58691847335237,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,4.0126111958998,2.19680022632121,.896432918697558,4.80639993236566,2.63137890697367,4.40969381865368,1.33269841309056,.729617706050062,.297729999278181,3.91021223386622,2.14073946020884,5.67229573503973,.95054294779362,.520397532085943,3.37758202463881,.647664476709677},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,1.33269841309056,.729617706050062,.297729999278181,.95054294779362,.520397532085943,.647664476709677,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,3.30149822057238,1.80748437551057,.737567519771828,.275847249626467,4.0126111958998,2.19680022632121,.896432918697558,3.91021223386622,2.14073946020884,3.37758202463881,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,3.11944010216456,1.70781229176476,.69689502893851,4.80639993236566,2.63137890697367,5.67229573503973,.626153181688297,.342802575265255,.139885051637275,2.36319452013471,1.29378751301169,4.40969381865368,.347192252628355,.190078724845143,1.58691847335237,.183906859961414},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,2.39387697960251,1.31058536069927,.534801410737989,2.07554270972356,1.13630563078316,1.71909544841957,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,2.90029193047496,1.58783437049499,.647936476974322,.242325604524073,4.28496925234002,2.34590917578526,.957278765839435,5.07587527588423,2.77890964691703,5.32974583920556,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,1.98037740107218,1.08420509999131,.442423999509308,3.70920808974538,2.03069492058023,5.32120685033247,.236321146004057,.129379678661691,.0527950614618356,1.08420509999131,.593574082861527,2.4592936780291,.0946962283839188,.0518437211734884,.526147189418459,.0362494356624709},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.7410698100085,2.04813838310324,.835769519114577,3.76406389114815,2.06072704459494,3.6179005979755,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,2.19713005267825,1.20287153766575,.490847383612732,.183574923144401,3.76697803717186,2.06232246300417,.841557517454501,5.17830525859385,2.83498739737667,6.30978459469567,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.979365349166059,.536177046731461,.218794021040783,2.12867470143779,1.16539401381336,3.543802987655,.0657431866024243,.0359926841107618,.0146872831147845,.350018452610409,.191625995851247,.921342735777177,.0171974308138302,.00941514591687767,.110884150015163,.00429748986298374},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,5.18623489409549,2.83932866528416,1.15862500931079,5.81833621896758,3.18538769471191,6.23568135536041,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,1.36595759702526,.747826244112006,.305160230186869,.114128683728465,2.61131941894657,1.42962943840673,.58337889571007,4.0025799484364,2.19130838066721,5.43816471058122,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.339486246269225,.18585988679871,.0758425453509597,.822758173894117,.450438693157099,1.5272774292324,.0113956405123916,.00623881668046779,.00254583033055523,.067649468233801,.0370363237207337,.198554676115027,.00166206328995312,.000909936406629402,.0119492023316399,.000231576690958128},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,6.45780747384476,3.53548156799627,1.44269926011819,7.79428043908953,4.26716574385505,8.98680735137037,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.612365059546825,.335253937240527,.13680473166712,.051164412687175,1.2594400120878,.689510637467093,.281363788027607,2.07683410541886,1.13701263662145,3.03569721712443,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.0634197097850823,.0347206410009163,.0141682093704184,.165355473459674,.0905278193943952,.330223979417038,.000824565926330002,.000451428390509428,.000184211229066994,.00526617557576491,.00288309410977452,.0166285878943821,5.01144631725624e-005,2.74363646769969e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,7.27823584467216,3.98464475452196,1.62598614322311,9.14177937313726,5.00488635016176,10.9691712594614,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.122554596919714,.0670954531185465,.0273791727414708,.0102396991398416,.262307293803493,.143606418429484,.058600467750326,.450140785461579,.246440368054962,.68472894504081,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,.00244089433681864,.00133632614020231,.000545305270748019,.00662303395559898,.00362593876834461,.0137645079246526,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,3.89785419960172e-005,2.13397375798647e-005,.000128085303901421,6.8546484667622e-008,3.7527416869482e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,6.8546484667622e-008,5.19287666029906e-008,7.65361974719019e-010,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,.122554596919714,.0928436971090411,.0524250669352539,.0271309626935909,.00244089433681864,.001849148543354,.00104413911950708,3.89785419960172e-005,2.95289776007745e-005,5.51740069821357e-007,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,.262307293803493,.19871616027429,.112206949174718,.00662303395559898,.00501741243403364,.000128085303901421,7.27823584467216,5.51377378765633,3.11340423610211,.450140785461579,.341013195733391,.0137645079246526,9.14177937313726,6.92553862835673,.68472894504081,10.9691712594614},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.000824565926330002,.000624666483447503,.000352723805983911,5.01144631725624e-005,3.79652184082228e-005,2.90963994598978e-006,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,.612365059546825,.463909453726739,.26195083695297,.135564507599069,.0634197097850823,.0480448752738173,.0271289907701198,.00526617557576491,.00398949710931715,.000387613365377294,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,1.2594400120878,.954114083786716,.538749492535305,.165355473459674,.125268361016134,.0166285878943821,6.45780747384476,4.89224179250545,2.76245034842036,2.07683410541886,1.57334740190396,.330223979417038,7.79428043908953,5.90471373874725,3.03569721712443,8.98680735137037},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.0113956405123916,.00863299641447793,.00487470263420339,.00166206328995312,.00125912943722616,.000231576690958128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,1.36595759702526,1.03480862072507,.584314421936207,.302393753783036,.339486246269225,.25718462626926,.145221718497785,.067649468233801,.0512492137952295,.0119492023316399,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,2.61131941894657,1.97825748822484,1.11704169972566,.822758173894117,.623296984369123,.198554676115027,5.18623489409549,3.92893643816517,2.21851091848272,4.0025799484364,3.03223485513554,1.5272774292324,5.81833621896758,4.40779749680161,5.43816471058122,6.23568135536041},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.0657431866024243,.0498050718253105,.0281229023119214,.0171974308138302,.0130282592204804,.00429748986298374,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,2.19713005267825,1.66448001337449,.939864296992246,.48639753212384,.979365349166059,.741937896083684,.41894221225766,.350018452610409,.265163510940112,.110884150015163,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.76697803717186,2.85374989208327,1.61139672111394,2.12867470143779,1.61262028650446,.921342735777177,3.7410698100085,2.8341225946006,1.60031398303209,5.17830525859385,3.92292918515564,3.543802987655,3.76406389114815,2.85154222320761,6.30978459469567,3.6179005979755},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,.547225517982271,.414561685224313,.234086155283846,.121144008364367,.236321146004057,.179029831969503,.101090878717283,.0946962283839188,.0717390303204678,.0362494356624709,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.90029193047496,2.19717442094167,1.24065520517793,.642062510456358,1.98037740107218,1.50027468671861,.847144214584193,1.08420509999131,.821361356090746,.526147189418459,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,4.28496925234002,3.24616454377475,1.83297734554319,3.70920808974538,2.80998510873911,2.4592936780291,2.39387697960251,1.81352960012534,1.02402654819954,5.07587527588423,3.84533129384947,5.32120685033247,2.07554270972356,1.57236907874115,5.32974583920556,1.71909544841957},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,.626153181688297,.474354922529452,.267848969162523,.347192252628355,.263022466294621,.183906859961414,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,3.30149822057238,2.50111630653613,1.41227885000115,.730880989425352,3.11944010216456,2.36319452013471,1.33440001580502,2.36319452013471,1.79028548729353,1.58691847335237,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,4.0126111958998,3.03983422634135,1.71647099090401,4.80639993236566,3.6411848323989,4.40969381865368,1.33269841309056,1.0096124574376,.570087171302651,3.91021223386622,2.96225983721782,5.67229573503973,.95054294779362,.720102907121204,3.37758202463881,.647664476709677},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,1.33269841309056,1.0096124574376,.570087171302651,.95054294779362,.720102907121204,.647664476709677,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,3.30149822057238,2.50111630653613,1.41227885000115,.730880989425352,4.0126111958998,3.03983422634135,1.71647099090401,3.91021223386622,2.96225983721782,3.37758202463881,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,3.11944010216456,2.36319452013471,1.33440001580502,4.80639993236566,3.6411848323989,5.67229573503973,.626153181688297,.474354922529452,.267848969162523,2.36319452013471,1.79028548729353,4.40969381865368,.347192252628355,.263022466294621,1.58691847335237,.183906859961414},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,2.39387697960251,1.81352960012534,1.02402654819954,2.07554270972356,1.57236907874115,1.71909544841957,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,2.90029193047496,2.19717442094167,1.24065520517793,.642062510456358,4.28496925234002,3.24616454377475,1.83297734554319,5.07587527588423,3.84533129384947,5.32974583920556,.547225517982271,.414561685224313,.234086155283846,.121144008364367,1.98037740107218,1.50027468671861,.847144214584193,3.70920808974538,2.80998510873911,5.32120685033247,.236321146004057,.179029831969503,.101090878717283,1.08420509999131,.821361356090746,2.4592936780291,.0946962283839188,.0717390303204678,.526147189418459,.0362494356624709},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.7410698100085,2.8341225946006,1.60031398303209,3.76406389114815,2.85154222320761,3.6179005979755,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,2.19713005267825,1.66448001337449,.939864296992246,.48639753212384,3.76697803717186,2.85374989208327,1.61139672111394,5.17830525859385,3.92292918515564,6.30978459469567,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.979365349166059,.741937896083684,.41894221225766,2.12867470143779,1.61262028650446,3.543802987655,.0657431866024243,.0498050718253105,.0281229023119214,.350018452610409,.265163510940112,.921342735777177,.0171974308138302,.0130282592204804,.110884150015163,.00429748986298374},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,5.18623489409549,3.92893643816517,2.21851091848272,5.81833621896758,4.40779749680161,6.23568135536041,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,1.36595759702526,1.03480862072507,.584314421936207,.302393753783036,2.61131941894657,1.97825748822484,1.11704169972566,4.0025799484364,3.03223485513554,5.43816471058122,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.339486246269225,.25718462626926,.145221718497785,.822758173894117,.623296984369123,1.5272774292324,.0113956405123916,.00863299641447793,.00487470263420339,.067649468233801,.0512492137952295,.198554676115027,.00166206328995312,.00125912943722616,.0119492023316399,.000231576690958128},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,6.45780747384476,4.89224179250545,2.76245034842036,7.79428043908953,5.90471373874725,8.98680735137037,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.612365059546825,.463909453726739,.26195083695297,.135564507599069,1.2594400120878,.954114083786716,.538749492535305,2.07683410541886,1.57334740190396,3.03569721712443,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.0634197097850823,.0480448752738173,.0271289907701198,.165355473459674,.125268361016134,.330223979417038,.000824565926330002,.000624666483447503,.000352723805983911,.00526617557576491,.00398949710931715,.0166285878943821,5.01144631725624e-005,3.79652184082228e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,7.27823584467216,5.51377378765633,3.11340423610211,9.14177937313726,6.92553862835673,10.9691712594614,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.122554596919714,.0928436971090411,.0524250669352539,.0271309626935909,.262307293803493,.19871616027429,.112206949174718,.450140785461579,.341013195733391,.68472894504081,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,.00244089433681864,.001849148543354,.00104413911950708,.00662303395559898,.00501741243403364,.0137645079246526,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,3.89785419960172e-005,2.95289776007745e-005,.000128085303901421,6.8546484667622e-008,5.19287666029906e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,6.8546484667622e-008,6.67972275267736e-008,7.65361974719019e-010,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,.122554596919714,.119427091493105,.0867440815154734,.057745280204241,.00244089433681864,.00237860446263913,.00172766377213932,3.89785419960172e-005,3.79838375617868e-005,5.51740069821357e-007,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,.262307293803493,.255613399651286,.18566096935718,.00662303395559898,.00645401887770827,.000128085303901421,7.27823584467216,7.09250048322984,5.15153163620379,.450140785461579,.438653515481796,.0137645079246526,9.14177937313726,8.90848771819592,.68472894504081,10.9691712594614},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.000824565926330002,.000803523594969987,.000583627344081487,5.01144631725624e-005,4.88355780044979e-005,2.90963994598978e-006,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,.612365059546825,.596737942368568,.433431678432723,.288534194877923,.0634197097850823,.0618012842418601,.0448884383820082,.00526617557576491,.00513178655225592,.000387613365377294,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,1.2594400120878,1.22730000557206,.891431001025123,.165355473459674,.161135720335381,.0166285878943821,6.45780747384476,6.29300885631198,4.57083285180669,2.07683410541886,2.02383478866576,.330223979417038,7.79428043908953,7.59537599078757,3.03569721712443,8.98680735137037},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.0113956405123916,.0111048319353399,.00806582856989507,.00166206328995312,.00161964862649386,.000231576690958128,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,1.36595759702526,1.33109933870947,.966824094086502,.643611958831769,.339486246269225,.33082280056131,.24028819272175,.067649468233801,.0659231023263137,.0119492023316399,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,2.61131941894657,2.54468041880833,1.84829055872937,.822758173894117,.801761975206599,.198554676115027,5.18623489409549,5.05388589728677,3.6708144253493,4.0025799484364,3.90043697901805,1.5272774292324,5.81833621896758,5.66985645040998,5.43816471058122,6.23568135536041},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.0657431866024243,.0640654676075239,.0465329941039433,.0171974308138302,.0167585647100291,.00429748986298374,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,2.19713005267825,2.14106086934751,1.55512753646012,1.0352438319403,.979365349166059,.954372647329225,.693194296828135,.350018452610409,.341086232751863,.110884150015163,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.76697803717186,3.67084745731307,2.66626514187773,2.12867470143779,2.07435244929784,.921342735777177,3.7410698100085,3.6456003901969,2.64792731244254,5.17830525859385,5.04615862255375,3.543802987655,3.76406389114815,3.66800767948205,6.30978459469567,3.6179005979755},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,.547225517982271,.533260715077649,.387325943916553,.257841742921929,.236321146004057,.230290399761936,.167267987177829,.0946962283839188,.092279648532812,.0362494356624709,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.90029193047496,2.82627856027675,2.05282515677973,1.36655967551418,1.98037740107218,1.92983958905795,1.40171011886371,1.08420509999131,1.05653696348977,.526147189418459,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,4.28496925234002,4.17561990789474,3.03289906094317,3.70920808974538,3.61455175876414,2.4592936780291,2.39387697960251,2.33278695510428,1.69438491095747,5.07587527588423,4.94634257953458,5.32120685033247,2.07554270972356,2.02257634793674,5.32974583920556,1.71909544841957},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,.626153181688297,.610174201341807,.443190904144467,.347192252628355,.338332155278326,.183906859961414,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,3.30149822057238,3.21724635356535,2.3367987653383,1.55560007239588,3.11944010216456,3.03983422634135,2.20793812029263,2.36319452013471,2.30288749099116,1.58691847335237,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,4.0126111958998,3.91021223386622,2.84012416689539,4.80639993236566,4.6837440522375,4.40969381865368,1.33269841309056,1.29868890491693,.943283260267015,3.91021223386622,3.8104264222726,5.67229573503973,.95054294779362,.926285773305773,3.37758202463881,.647664476709677},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,1.33269841309056,1.29868890491693,.943283260267015,.95054294779362,.926285773305773,.647664476709677,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,3.30149822057238,3.21724635356535,2.3367987653383,1.55560007239588,4.0126111958998,3.91021223386622,2.84012416689539,3.91021223386622,3.8104264222726,3.37758202463881,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,3.11944010216456,3.03983422634135,2.20793812029263,4.80639993236566,4.6837440522375,5.67229573503973,.626153181688297,.610174201341807,.443190904144467,2.36319452013471,2.30288749099116,4.40969381865368,.347192252628355,.338332155278326,1.58691847335237,.183906859961414},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,2.39387697960251,2.33278695510428,1.69438491095747,2.07554270972356,2.02257634793674,1.71909544841957,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,2.90029193047496,2.82627856027675,2.05282515677973,1.36655967551418,4.28496925234002,4.17561990789474,3.03289906094317,5.07587527588423,4.94634257953458,5.32974583920556,.547225517982271,.533260715077649,.387325943916553,.257841742921929,1.98037740107218,1.92983958905795,1.40171011886371,3.70920808974538,3.61455175876414,5.32120685033247,.236321146004057,.230290399761936,.167267987177829,1.08420509999131,1.05653696348977,2.4592936780291,.0946962283839188,.092279648532812,.526147189418459,.0362494356624709},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.7410698100085,3.6456003901969,2.64792731244254,3.76406389114815,3.66800767948205,3.6179005979755,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,2.19713005267825,2.14106086934751,1.55512753646012,1.0352438319403,3.76697803717186,3.67084745731307,2.66626514187773,5.17830525859385,5.04615862255375,6.30978459469567,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.979365349166059,.954372647329225,.693194296828135,2.12867470143779,2.07435244929784,3.543802987655,.0657431866024243,.0640654676075239,.0465329941039433,.350018452610409,.341086232751863,.921342735777177,.0171974308138302,.0167585647100291,.110884150015163,.00429748986298374},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,5.18623489409549,5.05388589728677,3.6708144253493,5.81833621896758,5.66985645040998,6.23568135536041,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,1.36595759702526,1.33109933870947,.966824094086502,.643611958831769,2.61131941894657,2.54468041880833,1.84829055872937,4.0025799484364,3.90043697901805,5.43816471058122,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.339486246269225,.33082280056131,.24028819272175,.822758173894117,.801761975206599,1.5272774292324,.0113956405123916,.0111048319353399,.00806582856989507,.067649468233801,.0659231023263137,.198554676115027,.00166206328995312,.00161964862649386,.0119492023316399,.000231576690958128},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,6.45780747384476,6.29300885631198,4.57083285180669,7.79428043908953,7.59537599078757,8.98680735137037,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.612365059546825,.596737942368568,.433431678432723,.288534194877923,1.2594400120878,1.22730000557206,.891431001025123,2.07683410541886,2.02383478866576,3.03569721712443,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.0634197097850823,.0618012842418601,.0448884383820082,.165355473459674,.161135720335381,.330223979417038,.000824565926330002,.000803523594969987,.000583627344081487,.00526617557576491,.00513178655225592,.0166285878943821,5.01144631725624e-005,4.88355780044979e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,7.27823584467216,7.09250048322984,5.15153163620379,9.14177937313726,8.90848771819592,10.9691712594614,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.122554596919714,.119427091493105,.0867440815154734,.057745280204241,.262307293803493,.255613399651286,.18566096935718,.450140785461579,.438653515481796,.68472894504081,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,.00244089433681864,.00237860446263913,.00172766377213932,.00662303395559898,.00645401887770827,.0137645079246526,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,3.89785419960172e-005,3.79838375617868e-005,.000128085303901421,6.8546484667622e-008,6.67972275267736e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,6.8546484667622e-008,8.11985772602823e-008,7.65361974719019e-010,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,.122554596919714,.1451753354836,.128179867438098,.103725715888757,.00244089433681864,.00289142686579082,.00255293167466323,3.89785419960172e-005,4.61730775826967e-005,5.51740069821357e-007,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,.262307293803493,.310723141496092,.274347229515445,.00662303395559898,.00784549254339729,.000128085303901421,7.27823584467216,8.62162951636421,7.61230772976691,.450140785461579,.533226343160225,.0137645079246526,9.14177937313726,10.8291399963909,.68472894504081,10.9691712594614},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.000824565926330002,.000976761687908062,.000862413599759767,5.01144631725624e-005,5.93644317357238e-005,2.90963994598978e-006,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,.612365059546825,.72539345885478,.640472688330538,.518283326641486,.0634197097850823,.0751255185147611,.066330681941484,.00526617557576491,.00623818955179855,.000387613365377294,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,1.2594400120878,1.49190345189168,1.31724845713225,.165355473459674,.19587626195712,.0166285878943821,6.45780747384476,7.64976908082116,6.7542215997961,2.07683410541886,2.46016955394827,.330223979417038,7.79428043908953,9.23292398567978,3.03569721712443,8.98680735137037},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.0113956405123916,.0134990116693501,.0119187017581499,.00166206328995312,.00196884165709061,.000231576690958128,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,1.36595759702526,1.61808171532254,1.4286552125618,1.15609641079355,.339486246269225,.402147540031861,.355068704998853,.067649468233801,.0801359924008096,.0119492023316399,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,2.61131941894657,3.09330846862644,2.73117928815418,.822758173894117,.974620266418623,.198554676115027,5.18623489409549,6.14349367016778,5.42428368841739,4.0025799484364,4.74136345348638,1.5272774292324,5.81833621896758,6.89226625249968,5.43816471058122,6.23568135536041},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,.23320173943763,.276245376362,.243905726875982,.197373399115395,.0657431866024243,.0778778553220726,.0687608066341364,.0171974308138302,.0203716780136318,.00429748986298374,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,2.19713005267825,2.60266934505625,2.29797858240319,1.85957029228411,.979365349166059,1.16013349668145,1.02431833416123,.350018452610409,.414623747840729,.110884150015163,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.76697803717186,4.46227488639218,3.93988277320967,2.12867470143779,2.52157872198894,.921342735777177,3.7410698100085,4.43158460169426,3.91278535624427,5.17830525859385,6.13410041033272,3.543802987655,3.76406389114815,4.45882285809472,6.30978459469567,3.6179005979755},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,.547225517982271,.648230667294543,.572343234018839,.463151607819505,.236321146004057,.279940553069749,.247168314523763,.0946962283839188,.112174957679791,.0362494356624709,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.90029193047496,3.43561861072344,3.03341567369851,2.45470071573063,1.98037740107218,2.34590917578526,2.07127695727047,1.08420509999131,1.28432423671899,.526147189418459,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,4.28496925234002,5.07587527588423,4.48164984622588,3.70920808974538,4.39384194692302,2.4592936780291,2.39387697960251,2.83573119453035,2.50375621577018,5.07587527588423,6.01276422646702,5.32120685033247,2.07554270972356,2.45863979589474,5.32974583920556,1.71909544841957},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,.626153181688297,.741726548606004,.654893686699248,.347192252628355,.411275896727804,.183906859961414,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,3.30149822057238,3.91087828459092,3.45303738004469,2.79426700459607,3.11944010216456,3.6952164547113,3.26262277063996,2.36319452013471,2.79938546527299,1.58691847335237,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,4.0126111958998,4.75324623388636,4.19679052288369,4.80639993236566,5.69354997766273,4.40969381865368,1.33269841309056,1.57868365630447,1.39386942769156,3.91021223386622,4.63194679928158,5.67229573503973,.95054294779362,1.12599114834103,3.37758202463881,.647664476709677},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,1.33269841309056,1.57868365630447,1.39386942769156,.95054294779362,1.12599114834103,.647664476709677,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,3.30149822057238,3.91087828459092,3.45303738004469,2.79426700459607,4.0126111958998,4.75324623388636,4.19679052288369,3.91021223386622,4.63194679928158,3.37758202463881,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,3.11944010216456,3.6952164547113,3.26262277063996,4.80639993236566,5.69354997766273,5.67229573503973,.626153181688297,.741726548606004,.654893686699248,2.36319452013471,2.79938546527299,4.40969381865368,.347192252628355,.411275896727804,1.58691847335237,.183906859961414},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,2.39387697960251,2.83573119453035,2.50375621577018,2.07554270972356,2.45863979589474,1.71909544841957,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,2.90029193047496,3.43561861072344,3.03341567369851,2.45470071573063,4.28496925234002,5.07587527588423,4.48164984622588,5.07587527588423,6.01276422646702,5.32974583920556,.547225517982271,.648230667294543,.572343234018839,.463151607819505,1.98037740107218,2.34590917578526,2.07127695727047,3.70920808974538,4.39384194692302,5.32120685033247,.236321146004057,.279940553069749,.247168314523763,1.08420509999131,1.28432423671899,2.4592936780291,.0946962283839188,.112174957679791,.526147189418459,.0362494356624709},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.7410698100085,4.43158460169426,3.91278535624427,3.76406389114815,4.45882285809472,3.6179005979755,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,2.19713005267825,2.60266934505625,2.29797858240319,1.85957029228411,3.76697803717186,4.46227488639218,3.93988277320967,5.17830525859385,6.13410041033272,6.30978459469567,.23320173943763,.276245376362,.243905726875982,.197373399115395,.979365349166059,1.16013349668145,1.02431833416123,2.12867470143779,2.52157872198894,3.543802987655,.0657431866024243,.0778778553220726,.0687608066341364,.350018452610409,.414623747840729,.921342735777177,.0171974308138302,.0203716780136318,.110884150015163,.00429748986298374},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,5.18623489409549,6.14349367016778,5.42428368841739,5.81833621896758,6.89226625249968,6.23568135536041,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,1.36595759702526,1.61808171532254,1.4286552125618,1.15609641079355,2.61131941894657,3.09330846862644,2.73117928815418,4.0025799484364,4.74136345348638,5.43816471058122,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.339486246269225,.402147540031861,.355068704998853,.822758173894117,.974620266418623,1.5272774292324,.0113956405123916,.0134990116693501,.0119187017581499,.067649468233801,.0801359924008096,.198554676115027,.00166206328995312,.00196884165709061,.0119492023316399,.000231576690958128},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,6.45780747384476,7.64976908082116,6.7542215997961,7.79428043908953,9.23292398567978,8.98680735137037,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.612365059546825,.72539345885478,.640472688330538,.518283326641486,1.2594400120878,1.49190345189168,1.31724845713225,2.07683410541886,2.46016955394827,3.03569721712443,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.0634197097850823,.0751255185147611,.066330681941484,.165355473459674,.19587626195712,.330223979417038,.000824565926330002,.000976761687908062,.000862413599759767,.00526617557576491,.00623818955179855,.0166285878943821,5.01144631725624e-005,5.93644317357238e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,7.27823584467216,8.62162951636421,7.61230772976691,9.14177937313726,10.8291399963909,10.9691712594614,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.122554596919714,.1451753354836,.128179867438098,.103725715888757,.262307293803493,.310723141496092,.274347229515445,.450140785461579,.533226343160225,.68472894504081,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,.00244089433681864,.00289142686579082,.00255293167466323,.00662303395559898,.00784549254339729,.0137645079246526,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,3.89785419960172e-005,4.61730775826967e-005,.000128085303901421,6.8546484667622e-008,8.11985772602823e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,6.8546484667622e-008,9.42279510613512e-008,7.65361974719019e-010,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,.122554596919714,.168470616959383,.172616566596884,.162099033300108,.00244089433681864,.00335539412635106,.00343796813996252,3.89785419960172e-005,5.35821518259837e-005,5.51740069821357e-007,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,.262307293803493,.360582733816502,.369456434624521,.00662303395559898,.0091044044412468,.000128085303901421,7.27823584467216,10.0050827434195,10.2513011633895,.450140785461579,.618789484535732,.0137645079246526,9.14177937313726,12.5668171542652,.68472894504081,10.9691712594614},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.000824565926330002,.00113349587680305,.00116139045508752,5.01144631725624e-005,6.88902312961445e-005,2.90963994598978e-006,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,.612365059546825,.841792327494715,.862508275818624,.809955617122475,.0634197097850823,.0871803906043333,.0893258419252021,.00526617557576491,.00723918866107805,.000387613365377294,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,1.2594400120878,1.73129887488649,1.77390498656132,.165355473459674,.227307170255112,.0166285878943821,6.45780747384476,8.87727458903197,9.0957384017764,2.07683410541886,2.85493593796569,.330223979417038,7.79428043908953,10.7144674060855,3.03569721712443,8.98680735137037},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.0113956405123916,.0156651046591466,.0160506124472133,.00166206328995312,.00228476805347462,.000231576690958128,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,1.36595759702526,1.87772409110003,1.92393362992246,1.80670828815818,.339486246269225,.46667737293102,.478161992112396,.067649468233801,.0929948605625365,.0119492023316399,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,2.61131941894657,3.58967027297864,3.67800966645072,.822758173894117,1.13101083622578,.198554676115027,5.18623489409549,7.12929758661485,7.30474484999971,4.0025799484364,5.50217723676456,1.5272774292324,5.81833621896758,7.99822052374818,5.43816471058122,6.23568135536041},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.0657431866024243,.0903743758530202,.0925984290266871,.0171974308138302,.023640580213036,.00429748986298374,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,2.19713005267825,3.02030168445819,3.09462929659509,2.90607342788681,.979365349166059,1.3462920908025,1.37942344206699,.350018452610409,.481155551471966,.110884150015163,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.76697803717186,5.17830525859385,5.30573990048865,2.12867470143779,2.92619901025212,.921342735777177,3.7410698100085,5.14269031628637,5.26924849841671,5.17830525859385,7.11839705537515,3.543802987655,3.76406389114815,5.17429930720343,6.30978459469567,3.6179005979755},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,.547225517982271,.752247756737171,.770760072903028,.723797635481732,.236321146004057,.324860676440036,.332855281233889,.0946962283839188,.130174896891091,.0362494356624709,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.90029193047496,3.98690855480132,4.08502371800223,3.83612308338039,1.98037740107218,2.72234098759749,2.78933598529375,1.08420509999131,1.49041085949741,.526147189418459,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,4.28496925234002,5.89036585649012,6.03532383522368,3.70920808974538,5.09889135888211,2.4592936780291,2.39387697960251,3.29076135613591,3.37174480048962,5.07587527588423,6.97759089372753,5.32120685033247,2.07554270972356,2.8531607102232,5.32974583920556,1.71909544841957},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,.626153181688297,.860746275133792,.881928667453122,.347192252628355,.477270493857987,.183906859961414,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,3.30149822057238,4.53842985975064,4.65011759480767,4.36678577097684,3.11944010216456,4.28816226840408,4.39369108444944,2.36319452013471,3.24858347921606,1.58691847335237,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,4.0126111958998,5.51596676470681,5.65171103126921,4.80639993236566,6.60715454337131,4.40969381865368,1.33269841309056,1.83200409818481,1.87708849838432,3.91021223386622,5.37520324921971,5.67229573503973,.95054294779362,1.306671156088,3.37758202463881,.647664476709677},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,1.33269841309056,1.83200409818481,1.87708849838432,.95054294779362,1.306671156088,.647664476709677,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,3.30149822057238,4.53842985975064,4.65011759480767,4.36678577097684,4.0126111958998,5.51596676470681,5.65171103126921,3.91021223386622,5.37520324921971,3.37758202463881,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,3.11944010216456,4.28816226840408,4.39369108444944,4.80639993236566,6.60715454337131,5.67229573503973,.626153181688297,.860746275133792,.881928667453122,2.36319452013471,3.24858347921606,4.40969381865368,.347192252628355,.477270493857987,1.58691847335237,.183906859961414},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,2.39387697960251,3.29076135613591,3.37174480048962,2.07554270972356,2.8531607102232,1.71909544841957,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,2.90029193047496,3.98690855480132,4.08502371800223,3.83612308338039,4.28496925234002,5.89036585649012,6.03532383522368,5.07587527588423,6.97759089372753,5.32974583920556,.547225517982271,.752247756737171,.770760072903028,.723797635481732,1.98037740107218,2.72234098759749,2.78933598529375,3.70920808974538,5.09889135888211,5.32120685033247,.236321146004057,.324860676440036,.332855281233889,1.08420509999131,1.49041085949741,2.4592936780291,.0946962283839188,.130174896891091,.526147189418459,.0362494356624709},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.7410698100085,5.14269031628637,5.26924849841671,3.76406389114815,5.17429930720343,3.6179005979755,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,2.19713005267825,3.02030168445819,3.09462929659509,2.90607342788681,3.76697803717186,5.17830525859385,5.30573990048865,5.17830525859385,7.11839705537515,6.30978459469567,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.979365349166059,1.3462920908025,1.37942344206699,2.12867470143779,2.92619901025212,3.543802987655,.0657431866024243,.0903743758530202,.0925984290266871,.350018452610409,.481155551471966,.921342735777177,.0171974308138302,.023640580213036,.110884150015163,.00429748986298374},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,5.18623489409549,7.12929758661485,7.30474484999971,5.81833621896758,7.99822052374818,6.23568135536041,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,1.36595759702526,1.87772409110003,1.92393362992246,1.80670828815818,2.61131941894657,3.58967027297864,3.67800966645072,4.0025799484364,5.50217723676456,5.43816471058122,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.339486246269225,.46667737293102,.478161992112396,.822758173894117,1.13101083622578,1.5272774292324,.0113956405123916,.0156651046591466,.0160506124472133,.067649468233801,.0929948605625365,.198554676115027,.00166206328995312,.00228476805347462,.0119492023316399,.000231576690958128},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,6.45780747384476,8.87727458903197,9.0957384017764,7.79428043908953,10.7144674060855,8.98680735137037,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.612365059546825,.841792327494715,.862508275818624,.809955617122475,1.2594400120878,1.73129887488649,1.77390498656132,2.07683410541886,2.85493593796569,3.03569721712443,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.0634197097850823,.0871803906043333,.0893258419252021,.165355473459674,.227307170255112,.330223979417038,.000824565926330002,.00113349587680305,.00116139045508752,.00526617557576491,.00723918866107805,.0166285878943821,5.01144631725624e-005,6.88902312961445e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,7.27823584467216,10.0050827434195,10.2513011633895,9.14177937313726,12.5668171542652,10.9691712594614,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.122554596919714,.168470616959383,.172616566596884,.162099033300108,.262307293803493,.360582733816502,.369456434624521,.450140785461579,.618789484535732,.68472894504081,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,.00244089433681864,.00335539412635106,.00343796813996252,.00662303395559898,.0091044044412468,.0137645079246526,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,3.89785419960172e-005,5.35821518259837e-005,.000128085303901421,6.8546484667622e-008,9.42279510613512e-008,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,6.8546484667622e-008,1.05066727253324e-007,7.65361974719019e-010,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,.122554596919714,.187849318200136,.214611696543555,.22471744861619,.00244089433681864,.00374135567557013,.00427437638070464,3.89785419960172e-005,5.97455560493016e-005,5.51740069821357e-007,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,.262307293803493,.402059551538899,.459339875638177,.00662303395559898,.0101516584777441,.000128085303901421,7.27823584467216,11.1559392718342,12.7452954129935,.450140785461579,.689967098580019,.0137645079246526,9.14177937313726,14.0123427870133,.68472894504081,10.9691712594614},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.000824565926330002,.00126387871952445,.00144394006224153,5.01144631725624e-005,7.68144808465349e-005,2.90963994598978e-006,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,.612365059546825,.938621331363155,1.07234414405062,1.12283926724684,.0634197097850823,.0972085057383859,.111057535546505,.00526617557576491,.0080718921723515,.000387613365377294,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,1.2594400120878,1.93044531514061,2.2054705766571,.165355473459674,.253453674742105,.0166285878943821,6.45780747384476,9.89840251743778,11.3086008382978,2.07683410541886,3.18333118932674,.330223979417038,7.79428043908953,11.9469224570834,3.03569721712443,8.98680735137037},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.0113956405123916,.0174670175895654,.0199554957891379,.00166206328995312,.00254757849669498,.000231576690958128,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,1.36595759702526,2.0937134121504,2.39199903285711,2.50463478185567,.339486246269225,.520357958596769,.594491933034081,.067649468233801,.10369179782253,.0119492023316399,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,2.61131941894657,4.0025799484364,4.57281656090381,.822758173894117,1.26110783171881,.198554676115027,5.18623489409549,7.94936063666436,9.0818842940915,4.0025799484364,6.13507720928999,1.5272774292324,5.81833621896758,8.91823333538729,5.43816471058122,6.23568135536041},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.0657431866024243,.100769886127061,.115126296058794,.0171974308138302,.0263598896651818,.00429748986298374,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,2.19713005267825,3.36771849254271,3.8475081309354,4.02868168249338,.979365349166059,1.50115228352525,1.71501734155852,.350018452610409,.53650152129502,.110884150015163,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.76697803717186,5.77395114836884,6.59655016845447,2.12867470143779,3.262791452389,.921342735777177,3.7410698100085,5.7342395194157,6.55118093268348,5.17830525859385,7.93720624797419,3.543802987655,3.76406389114815,5.76948440365682,6.30978459469567,3.6179005979755},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,.236321146004057,.362228486371654,.41383418762992,.0946962283839188,.145148549160118,.0362494356624709,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.90029193047496,4.44551143257392,5.07885128191947,5.3180069882276,1.98037740107218,3.0354842147445,3.46794134442417,1.08420509999131,1.66184862884528,.526147189418459,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,4.28496925234002,6.56791807415177,7.50363138951749,3.70920808974538,5.68540249112695,2.4592936780291,2.39387697960251,3.66928837957851,4.19204185444651,5.07587527588423,7.78020354957332,5.32120685033247,2.07554270972356,3.1813517621299,5.32974583920556,1.71909544841957},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,.626153181688297,.959755498290674,1.09648923787555,.347192252628355,.532169576430532,.183906859961414,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,3.30149822057238,5.06047268206324,5.78142437787735,6.0536632275323,3.11944010216456,4.78141751356834,5.46261298270436,2.36319452013471,3.62225890010297,1.58691847335237,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,4.0126111958998,6.15045290784802,7.02669110333533,4.80639993236566,7.36715694769769,4.40969381865368,1.33269841309056,2.04273437703882,2.33375715403626,3.91021223386622,5.99349776107556,5.67229573503973,.95054294779362,1.45697386412547,3.37758202463881,.647664476709677},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,1.33269841309056,2.04273437703882,2.33375715403626,.95054294779362,1.45697386412547,.647664476709677,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,3.30149822057238,5.06047268206324,5.78142437787735,6.0536632275323,4.0126111958998,6.15045290784802,7.02669110333533,3.91021223386622,5.99349776107556,3.37758202463881,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,3.11944010216456,4.78141751356834,5.46261298270436,4.80639993236566,7.36715694769769,5.67229573503973,.626153181688297,.959755498290674,1.09648923787555,2.36319452013471,3.62225890010297,4.40969381865368,.347192252628355,.532169576430532,1.58691847335237,.183906859961414},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,2.39387697960251,3.66928837957851,4.19204185444651,2.07554270972356,3.1813517621299,1.71909544841957,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,2.90029193047496,4.44551143257392,5.07885128191947,5.3180069882276,4.28496925234002,6.56791807415177,7.50363138951749,5.07587527588423,7.78020354957332,5.32974583920556,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,1.98037740107218,3.0354842147445,3.46794134442417,3.70920808974538,5.68540249112695,5.32120685033247,.236321146004057,.362228486371654,.41383418762992,1.08420509999131,1.66184862884528,2.4592936780291,.0946962283839188,.145148549160118,.526147189418459,.0362494356624709},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.7410698100085,5.7342395194157,6.55118093268348,3.76406389114815,5.76948440365682,3.6179005979755,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,2.19713005267825,3.36771849254271,3.8475081309354,4.02868168249338,3.76697803717186,5.77395114836884,6.59655016845447,5.17830525859385,7.93720624797419,6.30978459469567,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.979365349166059,1.50115228352525,1.71501734155852,2.12867470143779,3.262791452389,3.543802987655,.0657431866024243,.100769886127061,.115126296058794,.350018452610409,.53650152129502,.921342735777177,.0171974308138302,.0263598896651818,.110884150015163,.00429748986298374},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,5.18623489409549,7.94936063666436,9.0818842940915,5.81833621896758,8.91823333538729,6.23568135536041,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,1.36595759702526,2.0937134121504,2.39199903285711,2.50463478185567,2.61131941894657,4.0025799484364,4.57281656090381,4.0025799484364,6.13507720928999,5.43816471058122,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.339486246269225,.520357958596769,.594491933034081,.822758173894117,1.26110783171881,1.5272774292324,.0113956405123916,.0174670175895654,.0199554957891379,.067649468233801,.10369179782253,.198554676115027,.00166206328995312,.00254757849669498,.0119492023316399,.000231576690958128},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,6.45780747384476,9.89840251743778,11.3086008382978,7.79428043908953,11.9469224570834,8.98680735137037,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.612365059546825,.938621331363155,1.07234414405062,1.12283926724684,1.2594400120878,1.93044531514061,2.2054705766571,2.07683410541886,3.18333118932674,3.03569721712443,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.0634197097850823,.0972085057383859,.111057535546505,.165355473459674,.253453674742105,.330223979417038,.000824565926330002,.00126387871952445,.00144394006224153,.00526617557576491,.0080718921723515,.0166285878943821,5.01144631725624e-005,7.68144808465349e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,7.27823584467216,11.1559392718342,12.7452954129935,9.14177937313726,14.0123427870133,10.9691712594614,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.122554596919714,.187849318200136,.214611696543555,.22471744861619,.262307293803493,.402059551538899,.459339875638177,.450140785461579,.689967098580019,.68472894504081,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,.00244089433681864,.00374135567557013,.00427437638070464,.00662303395559898,.0101516584777441,.0137645079246526,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,3.89785419960172e-005,5.97455560493016e-005,.000128085303901421,6.8546484667622e-008,1.05066727253324e-007,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,6.8546484667622e-008,1.13034107101112e-007,7.65361974719019e-010,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,.122554596919714,.202094235800358,.248394511041402,.279814171322846,.00244089433681864,.00402506872719183,.00494722164817595,3.89785419960172e-005,6.42761581885916e-005,5.51740069821357e-007,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,.262307293803493,.432548377566746,.531646250640432,.00662303395559898,.0109214751579782,.000128085303901421,7.27823584467216,12.0019121888667,14.7515791225583,.450140785461579,.742288419521226,.0137645079246526,9.14177937313726,15.0749214021478,.68472894504081,10.9691712594614},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.000824565926330002,.00135972068684592,.00167123596481511,5.01144631725624e-005,8.26394471580808e-005,2.90963994598978e-006,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,.612365059546825,1.00979850491474,1.24114576986956,1.39813949040533,.0634197097850823,.104579983939912,.128539509653141,.00526617557576491,.00868399681012567,.000387613365377294,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,1.2594400120878,2.07683410541886,2.55264179132826,.165355473459674,.272673476798176,.0166285878943821,6.45780747384476,10.6490143886212,13.0887291840651,2.07683410541886,3.42472834168617,.330223979417038,7.79428043908953,12.8528769082785,3.03569721712443,8.98680735137037},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.0113956405123916,.0187915697820833,.0230967635919399,.00166206328995312,.00274076549419503,.000231576690958128,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,1.36595759702526,2.25248308627179,2.76853237613478,3.11872669553275,.339486246269225,.559817544151044,.688073089222744,.067649468233801,.111554914547963,.0119492023316399,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,2.61131941894657,4.306102059137,5.29264039202637,.822758173894117,1.35673967813674,.198554676115027,5.18623489409549,8.55217350992208,10.5114970194908,4.0025799484364,6.60031003608269,1.5272774292324,5.81833621896758,9.5945173923078,5.43816471058122,6.23568135536041},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.0657431866024243,.108411429563137,.133248748684711,.0171974308138302,.0283588027292761,.00429748986298374,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,2.19713005267825,3.62309803230714,4.45315850116099,5.01644279717582,.979365349166059,1.61498708893785,1.98498451316933,.350018452610409,.577185299317033,.110884150015163,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.76697803717186,6.21179920192732,7.63493733120357,2.12867470143779,3.51021420500431,.921342735777177,3.7410698100085,6.16907617592665,7.58242635758397,5.17830525859385,8.53909743428073,3.543802987655,3.76406389114815,6.20699373673991,6.30978459469567,3.6179005979755},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,.236321146004057,.389696858310697,.47897734533629,.0946962283839188,.156155398385808,.0362494356624709,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.90029193047496,4.78262175405185,5.87833189496431,6.62188774242252,1.98037740107218,3.26566989191443,4.01384271427013,1.08420509999131,1.78786929801119,.526147189418459,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,4.28496925234002,7.06597392374297,8.68480553507875,3.70920808974538,6.11653575680062,2.4592936780291,2.39387697960251,3.94753645159365,4.85192654207924,5.07587527588423,8.370188967346,5.32120685033247,2.07554270972356,3.42259881132375,5.32974583920556,1.71909544841957},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,.626153181688297,1.03253530989981,1.26909163149447,.347192252628355,.572524856067592,.183906859961414,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,3.30149822057238,5.4442165096419,6.69149959948354,7.53791381092379,3.11944010216456,5.1440001364154,6.3225029329745,2.36319452013471,3.8969406506306,1.58691847335237,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,4.0126111958998,6.61685169872888,8.13278832870006,4.80639993236566,7.92582037364658,4.40969381865368,1.33269841309056,2.19763822848137,2.70112242949704,3.91021223386622,6.44799438933945,5.67229573503973,.95054294779362,1.56745854854701,3.37758202463881,.647664476709677},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,1.33269841309056,2.19763822848137,2.70112242949704,.95054294779362,1.56745854854701,.647664476709677,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,3.30149822057238,5.4442165096419,6.69149959948354,7.53791381092379,4.0126111958998,6.61685169872888,8.13278832870006,3.91021223386622,6.44799438933945,3.37758202463881,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,3.11944010216456,5.1440001364154,6.3225029329745,4.80639993236566,7.92582037364658,5.67229573503973,.626153181688297,1.03253530989981,1.26909163149447,2.36319452013471,3.8969406506306,4.40969381865368,.347192252628355,.572524856067592,1.58691847335237,.183906859961414},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,2.39387697960251,3.94753645159365,4.85192654207924,2.07554270972356,3.42259881132375,1.71909544841957,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,2.90029193047496,4.78262175405185,5.87833189496431,6.62188774242252,4.28496925234002,7.06597392374297,8.68480553507875,5.07587527588423,8.370188967346,5.32974583920556,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,1.98037740107218,3.26566989191443,4.01384271427013,3.70920808974538,6.11653575680062,5.32120685033247,.236321146004057,.389696858310697,.47897734533629,1.08420509999131,1.78786929801119,2.4592936780291,.0946962283839188,.156155398385808,.526147189418459,.0362494356624709},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.7410698100085,6.16907617592665,7.58242635758397,3.76406389114815,6.20699373673991,3.6179005979755,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,2.19713005267825,3.62309803230714,4.45315850116099,5.01644279717582,3.76697803717186,6.21179920192732,7.63493733120357,5.17830525859385,8.53909743428073,6.30978459469567,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.979365349166059,1.61498708893785,1.98498451316933,2.12867470143779,3.51021420500431,3.543802987655,.0657431866024243,.108411429563137,.133248748684711,.350018452610409,.577185299317033,.921342735777177,.0171974308138302,.0283588027292761,.110884150015163,.00429748986298374},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,5.18623489409549,8.55217350992208,10.5114970194908,5.81833621896758,9.5945173923078,6.23568135536041,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,1.36595759702526,2.25248308627179,2.76853237613478,3.11872669553275,2.61131941894657,4.306102059137,5.29264039202637,4.0025799484364,6.60031003608269,5.43816471058122,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.339486246269225,.559817544151044,.688073089222744,.822758173894117,1.35673967813674,1.5272774292324,.0113956405123916,.0187915697820833,.0230967635919399,.067649468233801,.111554914547963,.198554676115027,.00166206328995312,.00274076549419503,.0119492023316399,.000231576690958128},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,6.45780747384476,10.6490143886212,13.0887291840651,7.79428043908953,12.8528769082785,8.98680735137037,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.612365059546825,1.00979850491474,1.24114576986956,1.39813949040533,1.2594400120878,2.07683410541886,2.55264179132826,2.07683410541886,3.42472834168617,3.03569721712443,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.0634197097850823,.104579983939912,.128539509653141,.165355473459674,.272673476798176,.330223979417038,.000824565926330002,.00135972068684592,.00167123596481511,.00526617557576491,.00868399681012567,.0166285878943821,5.01144631725624e-005,8.26394471580808e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,7.27823584467216,12.0019121888667,14.7515791225583,9.14177937313726,15.0749214021478,10.9691712594614,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.122554596919714,.202094235800358,.248394511041402,.279814171322846,.262307293803493,.432548377566746,.531646250640432,.450140785461579,.742288419521226,.68472894504081,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,.00244089433681864,.00402506872719183,.00494722164817595,.00662303395559898,.0109214751579782,.0137645079246526,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,3.89785419960172e-005,6.42761581885916e-005,.000128085303901421,6.8546484667622e-008,1.13034107101112e-007,5.51740069821357e-007,7.65361974719019e-010},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,6.8546484667622e-008,1.17631378114674e-007,7.65361974719019e-010,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,.122554596919714,.21031371924726,.269010572516923,.315362998946248,.00244089433681864,.00418877446399711,.00535782744298286,3.89785419960172e-005,6.68903684166533e-005,5.51740069821357e-007,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,.262307293803493,.450140785461579,.575771427724586,.00662303395559898,.0113656683529065,.000128085303901421,7.27823584467216,12.4900484198527,15.9759196314393,.450140785461579,.77247843138831,.0137645079246526,9.14177937313726,15.6880416449771,.68472894504081,10.9691712594614},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.000824565926330002,.00141502261880692,.00180994395496465,5.01144631725624e-005,8.60005205963556e-005,2.90963994598978e-006,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,.612365059546825,1.05086856345938,1.34415745633734,1.57576530364745,.0634197097850823,.108833412759726,.139207935545212,.00526617557576491,.00903718831878549,.000387613365377294,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,1.2594400120878,2.16130214323225,2.76450404172331,.165355473459674,.283763526546879,.0166285878943821,6.45780747384476,11.0821261849396,14.1750577199246,2.07683410541886,3.56401731152317,.330223979417038,7.79428043908953,13.3756231834223,3.03569721712443,8.98680735137037},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.0113956405123916,.019555851831759,.0250137314673599,.00166206328995312,.00285223664290029,.000231576690958128,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,1.36595759702526,2.34409503833324,2.99831295149481,3.51494350320858,.339486246269225,.582586184826084,.745181263826102,.067649468233801,.116092024524981,.0119492023316399,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,2.61131941894657,4.48123785385959,5.73191499287154,.822758173894117,1.41192036788792,.198554676115027,5.18623489409549,8.90000356218188,11.3839223715851,4.0025799484364,6.86875479835488,1.5272774292324,5.81833621896758,9.98474117367778,5.43816471058122,6.23568135536041},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.0657431866024243,.112820689170272,.144308028468726,.0171974308138302,.0295121988599679,.00429748986298374,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,2.19713005267825,3.77045500260921,4.82275841315291,5.6537538362715,.979365349166059,1.68067109814234,2.14973277020564,.350018452610409,.600660313311091,.110884150015163,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.76697803717186,6.46444263093715,8.26861613804166,2.12867470143779,3.65298001640339,.921342735777177,3.7410698100085,6.41998199374919,8.21174689798638,5.17830525859385,8.88639566242957,3.543802987655,3.76406389114815,6.45944172008843,6.30978459469567,3.6179005979755},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,.236321146004057,.405546428999883,.518731148089159,.0946962283839188,.162506478648407,.0362494356624709,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.90029193047496,4.97713833778588,6.36621728024122,7.46316159493692,1.98037740107218,3.39848970992327,4.34698062381386,1.08420509999131,1.86058469259941,.526147189418459,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,4.28496925234002,7.35335795264632,9.40562051630952,3.70920808974538,6.36530467495559,2.4592936780291,2.39387697960251,4.10808883148424,5.25462310508641,5.07587527588423,8.71061742831661,5.32120685033247,2.07554270972356,3.5618011698851,5.32974583920556,1.71909544841957},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,.626153181688297,1.07453010927869,1.3744227476422,.347192252628355,.595810322665643,.183906859961414,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,3.30149822057238,5.66564116980173,7.24687566849574,8.49556365917046,3.11944010216456,5.35321453486018,6.84725329319339,2.36319452013471,4.05543522146567,1.58691847335237,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,4.0126111958998,6.88596923586654,8.80778739952872,4.80639993236566,8.24817568034947,4.40969381865368,1.33269841309056,2.28701955580991,2.92530816462441,3.91021223386622,6.71024423995495,5.67229573503973,.95054294779362,1.63120949890175,3.37758202463881,.647664476709677},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,1.33269841309056,2.28701955580991,2.92530816462441,.95054294779362,1.63120949890175,.647664476709677,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,3.30149822057238,5.66564116980173,7.24687566849574,8.49556365917046,4.0126111958998,6.88596923586654,8.80778739952872,3.91021223386622,6.71024423995495,3.37758202463881,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,3.11944010216456,5.35321453486018,6.84725329319339,4.80639993236566,8.24817568034947,5.67229573503973,.626153181688297,1.07453010927869,1.3744227476422,2.36319452013471,4.05543522146567,4.40969381865368,.347192252628355,.595810322665643,1.58691847335237,.183906859961414},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,2.39387697960251,4.10808883148424,5.25462310508641,2.07554270972356,3.5618011698851,1.71909544841957,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,2.90029193047496,4.97713833778588,6.36621728024122,7.46316159493692,4.28496925234002,7.35335795264632,9.40562051630952,5.07587527588423,8.71061742831661,5.32974583920556,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,1.98037740107218,3.39848970992327,4.34698062381386,3.70920808974538,6.36530467495559,5.32120685033247,.236321146004057,.405546428999883,.518731148089159,1.08420509999131,1.86058469259941,2.4592936780291,.0946962283839188,.162506478648407,.526147189418459,.0362494356624709},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.7410698100085,6.41998199374919,8.21174689798638,3.76406389114815,6.45944172008843,3.6179005979755,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,2.19713005267825,3.77045500260921,4.82275841315291,5.6537538362715,3.76697803717186,6.46444263093715,8.26861613804166,5.17830525859385,8.88639566242957,6.30978459469567,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.979365349166059,1.68067109814234,2.14973277020564,2.12867470143779,3.65298001640339,3.543802987655,.0657431866024243,.112820689170272,.144308028468726,.350018452610409,.600660313311091,.921342735777177,.0171974308138302,.0295121988599679,.110884150015163,.00429748986298374},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,5.18623489409549,8.90000356218188,11.3839223715851,5.81833621896758,9.98474117367778,6.23568135536041,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,1.36595759702526,2.34409503833324,2.99831295149481,3.51494350320858,2.61131941894657,4.48123785385959,5.73191499287154,4.0025799484364,6.86875479835488,5.43816471058122,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.339486246269225,.582586184826084,.745181263826102,.822758173894117,1.41192036788792,1.5272774292324,.0113956405123916,.019555851831759,.0250137314673599,.067649468233801,.116092024524981,.198554676115027,.00166206328995312,.00285223664290029,.0119492023316399,.000231576690958128},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,6.45780747384476,11.0821261849396,14.1750577199246,7.79428043908953,13.3756231834223,8.98680735137037,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.612365059546825,1.05086856345938,1.34415745633734,1.57576530364745,1.2594400120878,2.16130214323225,2.76450404172331,2.07683410541886,3.56401731152317,3.03569721712443,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.0634197097850823,.108833412759726,.139207935545212,.165355473459674,.283763526546879,.330223979417038,.000824565926330002,.00141502261880692,.00180994395496465,.00526617557576491,.00903718831878549,.0166285878943821,5.01144631725624e-005,8.60005205963556e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,7.27823584467216,12.4900484198527,15.9759196314393,9.14177937313726,15.6880416449771,10.9691712594614,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.122554596919714,.21031371924726,.269010572516923,.315362998946248,.262307293803493,.450140785461579,.575771427724586,.450140785461579,.77247843138831,.68472894504081,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,.00244089433681864,.00418877446399711,.00535782744298286,.00662303395559898,.0113656683529065,.0137645079246526,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,3.89785419960172e-005,6.68903684166533e-005,.000128085303901421,6.8546484667622e-008,1.17631378114674e-007,5.51740069821357e-007,7.65361974719019e-010}
		};
		for (int j = 0; j < 144; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,0.,0.,0.,6.8546484667622e-008,1.09461601509022e-009,0.,7.65361974719019e-010},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,0.,0.,0.,0.,0.,0.,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,0.,0.,0.,5.01144631725624e-005,8.00275816365055e-007,0.,2.90963994598978e-006},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,0.,0.,0.,0.,0.,0.,.0113956405123916,.000181976518058871,2.16598631798674e-006,0.,0.,0.,.00166206328995312,2.65414208197256e-005,0.,.000231576690958128},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,0.,0.,0.,0.,0.,0.,.0657431866024243,.00104985026256264,1.24959051250218e-005,0.,0.,0.,.0171974308138302,.000274625070541602,0.,.00429748986298374},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,0.,0.,0.,0.,0.,0.,.236321146004057,.00377380273155657,4.4917911225712e-005,0.,0.,0.,.0946962283839188,.00151220020487266,0.,.0362494356624709},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,0.,0.,0.,0.,0.,0.,.626153181688297,.00999901459257298,.000119013865260662,0.,0.,0.,.347192252628355,.00554429890730429,0.,.183906859961414},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,0.,0.,0.,0.,0.,0.,1.33269841309056,.0212818065446243,.000253307966815722,0.,0.,0.,.95054294779362,.0151791815252261,0.,.647664476709677},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,0.,0.,0.,0.,0.,0.,2.39387697960251,.0382277237453785,.000455007753107355,0.,0.,0.,2.07554270972356,.0331442567927967,0.,1.71909544841957},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,0.,0.,0.,0.,0.,0.,3.7410698100085,.0597409910483066,.000711070695350591,0.,0.,0.,3.76406389114815,.0601081826012263,0.,3.6179005979755},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,0.,0.,0.,0.,0.,0.,5.18623489409549,.0828187732700635,.000985755369367884,0.,0.,0.,5.81833621896758,.0929127735338078,0.,6.23568135536041},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,0.,0.,0.,0.,0.,0.,6.45780747384476,.103124463877826,.00122744505825102,0.,0.,0.,7.79428043908953,.124466546112493,0.,8.98680735137037},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,0.,0.,0.,0.,0.,0.,7.27823584467216,.116225851033514,.00138338509726576,0.,0.,0.,9.14177937313726,.145984701575574,0.,10.9691712594614},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,0.,0.,0.,6.8546484667622e-008,5.69188702865245e-009,0.,7.65361974719019e-010},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,0.,0.,0.,0.,0.,0.,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,0.,0.,0.,5.01144631725624e-005,4.16134925463982e-006,0.,2.90963994598978e-006},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,0.,0.,0.,0.,0.,0.,.0113956405123916,.000946258567734596,5.85658777413804e-005,0.,0.,0.,.00166206328995312,.000138012569524986,0.,.000231576690958128},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,0.,0.,0.,0.,0.,0.,.0657431866024243,.0054591098696972,.000337875473054763,0.,0.,0.,.0171974308138302,.00142802120123345,0.,.00429748986298374},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,0.,0.,0.,0.,0.,0.,.236321146004057,.0196233734207421,.0012145307084342,0.,0.,0.,.0946962283839188,.00786328046747136,0.,.0362494356624709},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,0.,0.,0.,0.,0.,0.,.626153181688297,.0519938139714517,.00321800346775215,0.,0.,0.,.347192252628355,.028829765505355,0.,.183906859961414},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,0.,0.,0.,0.,0.,0.,1.33269841309056,.110663133873164,.00684916764813009,0.,0.,0.,.95054294779362,.0789301318799674,0.,.647664476709677},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,0.,0.,0.,0.,0.,0.,2.39387697960251,.198780103635972,.0123029070952925,0.,0.,0.,2.07554270972356,.172346615354148,0.,1.71909544841957},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,0.,0.,0.,0.,0.,0.,3.7410698100085,.310646808870845,.0192265662361566,0.,0.,0.,3.76406389114815,.31255616594975,0.,3.6179005979755},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,0.,0.,0.,0.,0.,0.,5.18623489409549,.430648825529863,.0266537364367886,0.,0.,0.,5.81833621896758,.483136554903789,0.,6.23568135536041},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,0.,0.,0.,0.,0.,0.,6.45780747384476,.53623626019628,.0331887586818224,0.,0.,0.,7.79428043908953,.647212821256298,0.,8.98680735137037},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,0.,0.,0.,0.,0.,0.,7.27823584467216,.60436208201944,.037405205103521,0.,0.,0.,9.14177937313726,.759104944404841,0.,10.9691712594614},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,0.,0.,0.,6.8546484667622e-008,1.36592668764407e-008,0.,7.65361974719019e-010},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,0.,0.,0.,0.,0.,0.,.000824565926330002,.000164311358893045,2.44047036237916e-005,0.,0.,0.,5.01144631725624e-005,9.98631556618579e-006,0.,2.90963994598978e-006},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,0.,0.,0.,0.,0.,0.,.0113956405123916,.00227081076025246,.000337277130218073,0.,0.,0.,.00166206328995312,.000331199567025036,0.,.000231576690958128},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,0.,0.,0.,0.,0.,0.,.0657431866024243,.0131006533057736,.00194580315907169,0.,0.,0.,.0171974308138302,.00342693426532766,0.,.00429748986298374},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,0.,0.,0.,0.,0.,0.,.236321146004057,.0470917453597858,.00699440438187064,0.,0.,0.,.0946962283839188,.0188701296931614,0.,.0362494356624709},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,0.,0.,0.,0.,0.,0.,.626153181688297,.124773625580585,.0185322753878643,0.,0.,0.,.347192252628355,.0691850451424154,0.,.183906859961414},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,0.,0.,0.,0.,0.,0.,1.33269841309056,.26556698531571,.0394439167964793,0.,0.,0.,.95054294779362,.189414816301507,0.,.647664476709677},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,0.,0.,0.,0.,0.,0.,2.39387697960251,.477028175651108,.0708516521644813,0.,0.,0.,2.07554270972356,.413593664547998,0.,1.71909544841957},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,0.,0.,0.,0.,0.,0.,3.7410698100085,.745483465381799,.110724560685561,0.,0.,0.,3.76406389114815,.750065499032834,0.,3.6179005979755},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,0.,0.,0.,0.,0.,0.,5.18623489409549,1.03346169878758,.153497157076452,0.,0.,0.,5.81833621896758,1.1594206118243,0.,6.23568135536041},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,0.,0.,0.,0.,0.,0.,6.45780747384476,1.28684813137965,.191131930663375,0.,0.,0.,7.79428043908953,1.55316727245144,0.,8.98680735137037},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,0.,0.,0.,0.,0.,0.,7.27823584467216,1.45033499905202,.215414174926983,0.,0.,0.,9.14177937313726,1.82168355953934,0.,10.9691712594614},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,0.,0.,0.,6.8546484667622e-008,2.44980430684131e-008,0.,7.65361974719019e-010},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,0.,0.,0.,0.,0.,0.,.000824565926330002,.000294694201614443,7.85021488046878e-005,0.,0.,0.,5.01144631725624e-005,1.79105651165762e-005,0.,2.90963994598978e-006},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,0.,0.,0.,0.,0.,0.,.0113956405123916,.00407272369067121,.00108491296894855,0.,0.,0.,.00166206328995312,.000594010010245393,0.,.000231576690958128},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,0.,0.,0.,0.,0.,0.,.0657431866024243,.0234961635798143,.00625902823868614,0.,0.,0.,.0171974308138302,.0061462437174735,0.,.00429748986298374},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,0.,0.,0.,0.,0.,0.,.236321146004057,.0844595552914031,.0224987683542483,0.,0.,0.,.0946962283839188,.0338437819621889,0.,.0362494356624709},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,0.,0.,0.,0.,0.,0.,.626153181688297,.223782848737467,.0596124199094679,0.,0.,0.,.347192252628355,.12408412771496,0.,.183906859961414},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,0.,0.,0.,0.,0.,0.,1.33269841309056,.476297264169728,.126878501518793,0.,0.,0.,.95054294779362,.339717524338981,0.,.647664476709677},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,0.,0.,0.,0.,0.,0.,2.39387697960251,.855555199093713,.227907170151078,0.,0.,0.,2.07554270972356,.741784716454702,0.,1.71909544841957},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,0.,0.,0.,0.,0.,0.,3.7410698100085,1.33703266851113,.356165601240813,0.,0.,0.,3.76406389114815,1.34525059548623,0.,3.6179005979755},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,0.,0.,0.,0.,0.,0.,5.18623489409549,1.85352474883709,.493751403486214,0.,0.,0.,5.81833621896758,2.07943342346341,0.,6.23568135536041},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,0.,0.,0.,0.,0.,0.,6.45780747384476,2.30797605978546,.614810468242533,0.,0.,0.,7.79428043908953,2.78562232344933,0.,8.98680735137037},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,0.,0.,0.,0.,0.,0.,7.27823584467216,2.60119152746664,.692918704338266,0.,0.,0.,9.14177937313726,3.26720919228747,0.,10.9691712594614},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,0.,0.,0.,6.8546484667622e-008,3.7527416869482e-008,0.,7.65361974719019e-010},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,0.,0.,0.,0.,0.,0.,.000824565926330002,.000451428390509428,.000184211229066994,0.,0.,0.,5.01144631725624e-005,2.74363646769969e-005,0.,2.90963994598978e-006},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,0.,0.,0.,0.,0.,0.,.0113956405123916,.00623881668046779,.00254583033055523,0.,0.,0.,.00166206328995312,.000909936406629402,0.,.000231576690958128},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,0.,0.,0.,0.,0.,0.,.0657431866024243,.0359926841107618,.0146872831147845,0.,0.,0.,.0171974308138302,.00941514591687767,0.,.00429748986298374},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,0.,0.,0.,0.,0.,0.,.236321146004057,.129379678661691,.0527950614618356,0.,0.,0.,.0946962283839188,.0518437211734884,0.,.0362494356624709},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,0.,0.,0.,0.,0.,0.,.626153181688297,.342802575265255,.139885051637275,0.,0.,0.,.347192252628355,.190078724845143,0.,.183906859961414},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,0.,0.,0.,0.,0.,0.,1.33269841309056,.729617706050062,.297729999278181,0.,0.,0.,.95054294779362,.520397532085943,0.,.647664476709677},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,0.,0.,0.,0.,0.,0.,2.39387697960251,1.31058536069927,.534801410737989,0.,0.,0.,2.07554270972356,1.13630563078316,0.,1.71909544841957},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,0.,0.,0.,0.,0.,0.,3.7410698100085,2.04813838310324,.835769519114577,0.,0.,0.,3.76406389114815,2.06072704459494,0.,3.6179005979755},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,0.,0.,0.,0.,0.,0.,5.18623489409549,2.83932866528416,1.15862500931079,0.,0.,0.,5.81833621896758,3.18538769471191,0.,6.23568135536041},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,0.,0.,0.,0.,0.,0.,6.45780747384476,3.53548156799627,1.44269926011819,0.,0.,0.,7.79428043908953,4.26716574385505,0.,8.98680735137037},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,0.,0.,0.,0.,0.,0.,7.27823584467216,3.98464475452196,1.62598614322311,0.,0.,0.,9.14177937313726,5.00488635016176,0.,10.9691712594614},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,0.,0.,0.,6.8546484667622e-008,5.19287666029906e-008,0.,7.65361974719019e-010},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,0.,0.,0.,0.,0.,0.,.000824565926330002,.000624666483447503,.000352723805983911,0.,0.,0.,5.01144631725624e-005,3.79652184082228e-005,0.,2.90963994598978e-006},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,0.,0.,0.,0.,0.,0.,.0113956405123916,.00863299641447793,.00487470263420339,0.,0.,0.,.00166206328995312,.00125912943722616,0.,.000231576690958128},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,0.,0.,0.,0.,0.,0.,.0657431866024243,.0498050718253105,.0281229023119214,0.,0.,0.,.0171974308138302,.0130282592204804,0.,.00429748986298374},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.414561685224313,.234086155283846,.121144008364367,0.,0.,0.,0.,0.,0.,.236321146004057,.179029831969503,.101090878717283,0.,0.,0.,.0946962283839188,.0717390303204678,0.,.0362494356624709},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,0.,0.,0.,0.,0.,0.,.626153181688297,.474354922529452,.267848969162523,0.,0.,0.,.347192252628355,.263022466294621,0.,.183906859961414},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,0.,0.,0.,0.,0.,0.,1.33269841309056,1.0096124574376,.570087171302651,0.,0.,0.,.95054294779362,.720102907121204,0.,.647664476709677},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,0.,0.,0.,0.,0.,0.,2.39387697960251,1.81352960012534,1.02402654819954,0.,0.,0.,2.07554270972356,1.57236907874115,0.,1.71909544841957},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,0.,0.,0.,0.,0.,0.,3.7410698100085,2.8341225946006,1.60031398303209,0.,0.,0.,3.76406389114815,2.85154222320761,0.,3.6179005979755},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,0.,0.,0.,0.,0.,0.,5.18623489409549,3.92893643816517,2.21851091848272,0.,0.,0.,5.81833621896758,4.40779749680161,0.,6.23568135536041},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,0.,0.,0.,0.,0.,0.,6.45780747384476,4.89224179250545,2.76245034842036,0.,0.,0.,7.79428043908953,5.90471373874725,0.,8.98680735137037},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,0.,0.,0.,0.,0.,0.,7.27823584467216,5.51377378765633,3.11340423610211,0.,0.,0.,9.14177937313726,6.92553862835673,0.,10.9691712594614},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,0.,0.,0.,6.8546484667622e-008,6.67972275267736e-008,0.,7.65361974719019e-010},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,0.,0.,0.,0.,0.,0.,.000824565926330002,.000803523594969987,.000583627344081487,0.,0.,0.,5.01144631725624e-005,4.88355780044979e-005,0.,2.90963994598978e-006},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,0.,0.,0.,0.,0.,0.,.0113956405123916,.0111048319353399,.00806582856989507,0.,0.,0.,.00166206328995312,.00161964862649386,0.,.000231576690958128},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.227250598233051,.16506007282648,.109880005542746,0.,0.,0.,0.,0.,0.,.0657431866024243,.0640654676075239,.0465329941039433,0.,0.,0.,.0171974308138302,.0167585647100291,0.,.00429748986298374},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.533260715077649,.387325943916553,.257841742921929,0.,0.,0.,0.,0.,0.,.236321146004057,.230290399761936,.167267987177829,0.,0.,0.,.0946962283839188,.092279648532812,0.,.0362494356624709},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,0.,0.,0.,0.,0.,0.,.626153181688297,.610174201341807,.443190904144467,0.,0.,0.,.347192252628355,.338332155278326,0.,.183906859961414},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,0.,0.,0.,0.,0.,0.,1.33269841309056,1.29868890491693,.943283260267015,0.,0.,0.,.95054294779362,.926285773305773,0.,.647664476709677},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,0.,0.,0.,0.,0.,0.,2.39387697960251,2.33278695510428,1.69438491095747,0.,0.,0.,2.07554270972356,2.02257634793674,0.,1.71909544841957},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,0.,0.,0.,0.,0.,0.,3.7410698100085,3.6456003901969,2.64792731244254,0.,0.,0.,3.76406389114815,3.66800767948205,0.,3.6179005979755},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,0.,0.,0.,0.,0.,0.,5.18623489409549,5.05388589728677,3.6708144253493,0.,0.,0.,5.81833621896758,5.66985645040998,0.,6.23568135536041},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,0.,0.,0.,0.,0.,0.,6.45780747384476,6.29300885631198,4.57083285180669,0.,0.,0.,7.79428043908953,7.59537599078757,0.,8.98680735137037},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,0.,0.,0.,0.,0.,0.,7.27823584467216,7.09250048322984,5.15153163620379,0.,0.,0.,9.14177937313726,8.90848771819592,0.,10.9691712594614},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,0.,0.,0.,6.8546484667622e-008,8.11985772602823e-008,0.,7.65361974719019e-010},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,0.,0.,0.,0.,0.,0.,.000824565926330002,.000976761687908062,.000862413599759767,0.,0.,0.,5.01144631725624e-005,5.93644317357238e-005,0.,2.90963994598978e-006},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,0.,0.,0.,0.,0.,0.,.0113956405123916,.0134990116693501,.0119187017581499,0.,0.,0.,.00166206328995312,.00196884165709061,0.,.000231576690958128},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.276245376362,.243905726875982,.197373399115395,0.,0.,0.,0.,0.,0.,.0657431866024243,.0778778553220726,.0687608066341364,0.,0.,0.,.0171974308138302,.0203716780136318,0.,.00429748986298374},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.648230667294543,.572343234018839,.463151607819505,0.,0.,0.,0.,0.,0.,.236321146004057,.279940553069749,.247168314523763,0.,0.,0.,.0946962283839188,.112174957679791,0.,.0362494356624709},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,0.,0.,0.,0.,0.,0.,.626153181688297,.741726548606004,.654893686699248,0.,0.,0.,.347192252628355,.411275896727804,0.,.183906859961414},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,0.,0.,0.,0.,0.,0.,1.33269841309056,1.57868365630447,1.39386942769156,0.,0.,0.,.95054294779362,1.12599114834103,0.,.647664476709677},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,0.,0.,0.,0.,0.,0.,2.39387697960251,2.83573119453035,2.50375621577018,0.,0.,0.,2.07554270972356,2.45863979589474,0.,1.71909544841957},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,0.,0.,0.,0.,0.,0.,3.7410698100085,4.43158460169426,3.91278535624427,0.,0.,0.,3.76406389114815,4.45882285809472,0.,3.6179005979755},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,0.,0.,0.,0.,0.,0.,5.18623489409549,6.14349367016778,5.42428368841739,0.,0.,0.,5.81833621896758,6.89226625249968,0.,6.23568135536041},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,0.,0.,0.,0.,0.,0.,6.45780747384476,7.64976908082116,6.7542215997961,0.,0.,0.,7.79428043908953,9.23292398567978,0.,8.98680735137037},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,0.,0.,0.,0.,0.,0.,7.27823584467216,8.62162951636421,7.61230772976691,0.,0.,0.,9.14177937313726,10.8291399963909,0.,10.9691712594614},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,0.,0.,0.,6.8546484667622e-008,9.42279510613512e-008,0.,7.65361974719019e-010},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,0.,0.,0.,0.,0.,0.,.000824565926330002,.00113349587680305,.00116139045508752,0.,0.,0.,5.01144631725624e-005,6.88902312961445e-005,0.,2.90963994598978e-006},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,0.,0.,0.,0.,0.,0.,.0113956405123916,.0156651046591466,.0160506124472133,0.,0.,0.,.00166206328995312,.00228476805347462,0.,.000231576690958128},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.320572560296515,.3284616374485,.308448458722372,0.,0.,0.,0.,0.,0.,.0657431866024243,.0903743758530202,.0925984290266871,0.,0.,0.,.0171974308138302,.023640580213036,0.,.00429748986298374},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.752247756737171,.770760072903028,.723797635481732,0.,0.,0.,0.,0.,0.,.236321146004057,.324860676440036,.332855281233889,0.,0.,0.,.0946962283839188,.130174896891091,0.,.0362494356624709},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,0.,0.,0.,0.,0.,0.,.626153181688297,.860746275133792,.881928667453122,0.,0.,0.,.347192252628355,.477270493857987,0.,.183906859961414},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,0.,0.,0.,0.,0.,0.,1.33269841309056,1.83200409818481,1.87708849838432,0.,0.,0.,.95054294779362,1.306671156088,0.,.647664476709677},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,0.,0.,0.,0.,0.,0.,2.39387697960251,3.29076135613591,3.37174480048962,0.,0.,0.,2.07554270972356,2.8531607102232,0.,1.71909544841957},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,0.,0.,0.,0.,0.,0.,3.7410698100085,5.14269031628637,5.26924849841671,0.,0.,0.,3.76406389114815,5.17429930720343,0.,3.6179005979755},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,0.,0.,0.,0.,0.,0.,5.18623489409549,7.12929758661485,7.30474484999971,0.,0.,0.,5.81833621896758,7.99822052374818,0.,6.23568135536041},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,0.,0.,0.,0.,0.,0.,6.45780747384476,8.87727458903197,9.0957384017764,0.,0.,0.,7.79428043908953,10.7144674060855,0.,8.98680735137037},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,0.,0.,0.,0.,0.,0.,7.27823584467216,10.0050827434195,10.2513011633895,0.,0.,0.,9.14177937313726,12.5668171542652,0.,10.9691712594614},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,0.,0.,0.,6.8546484667622e-008,1.05066727253324e-007,0.,7.65361974719019e-010},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,0.,0.,0.,0.,0.,0.,.000824565926330002,.00126387871952445,.00144394006224153,0.,0.,0.,5.01144631725624e-005,7.68144808465349e-005,0.,2.90963994598978e-006},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,0.,0.,0.,0.,0.,0.,.0113956405123916,.0174670175895654,.0199554957891379,0.,0.,0.,.00166206328995312,.00254757849669498,0.,.000231576690958128},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.357447120288584,.408371633453457,.427601258703136,0.,0.,0.,0.,0.,0.,.0657431866024243,.100769886127061,.115126296058794,0.,0.,0.,.0171974308138302,.0263598896651818,0.,.00429748986298374},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,0.,0.,0.,0.,0.,0.,.236321146004057,.362228486371654,.41383418762992,0.,0.,0.,.0946962283839188,.145148549160118,0.,.0362494356624709},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,0.,0.,0.,0.,0.,0.,.626153181688297,.959755498290674,1.09648923787555,0.,0.,0.,.347192252628355,.532169576430532,0.,.183906859961414},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,0.,0.,0.,0.,0.,0.,1.33269841309056,2.04273437703882,2.33375715403626,0.,0.,0.,.95054294779362,1.45697386412547,0.,.647664476709677},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,0.,0.,0.,0.,0.,0.,2.39387697960251,3.66928837957851,4.19204185444651,0.,0.,0.,2.07554270972356,3.1813517621299,0.,1.71909544841957},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,0.,0.,0.,0.,0.,0.,3.7410698100085,5.7342395194157,6.55118093268348,0.,0.,0.,3.76406389114815,5.76948440365682,0.,3.6179005979755},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,0.,0.,0.,0.,0.,0.,5.18623489409549,7.94936063666436,9.0818842940915,0.,0.,0.,5.81833621896758,8.91823333538729,0.,6.23568135536041},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,0.,0.,0.,0.,0.,0.,6.45780747384476,9.89840251743778,11.3086008382978,0.,0.,0.,7.79428043908953,11.9469224570834,0.,8.98680735137037},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,0.,0.,0.,0.,0.,0.,7.27823584467216,11.1559392718342,12.7452954129935,0.,0.,0.,9.14177937313726,14.0123427870133,0.,10.9691712594614},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,0.,0.,0.,6.8546484667622e-008,1.13034107101112e-007,0.,7.65361974719019e-010},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,0.,0.,0.,0.,0.,0.,.000824565926330002,.00135972068684592,.00167123596481511,0.,0.,0.,5.01144631725624e-005,8.26394471580808e-005,0.,2.90963994598978e-006},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,0.,0.,0.,0.,0.,0.,.0113956405123916,.0187915697820833,.0230967635919399,0.,0.,0.,.00166206328995312,.00274076549419503,0.,.000231576690958128},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.384552913504838,.472654910466463,.532441484172332,0.,0.,0.,0.,0.,0.,.0657431866024243,.108411429563137,.133248748684711,0.,0.,0.,.0171974308138302,.0283588027292761,0.,.00429748986298374},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,0.,0.,0.,0.,0.,0.,.236321146004057,.389696858310697,.47897734533629,0.,0.,0.,.0946962283839188,.156155398385808,0.,.0362494356624709},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,0.,0.,0.,0.,0.,0.,.626153181688297,1.03253530989981,1.26909163149447,0.,0.,0.,.347192252628355,.572524856067592,0.,.183906859961414},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,0.,0.,0.,0.,0.,0.,1.33269841309056,2.19763822848137,2.70112242949704,0.,0.,0.,.95054294779362,1.56745854854701,0.,.647664476709677},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,0.,0.,0.,0.,0.,0.,2.39387697960251,3.94753645159365,4.85192654207924,0.,0.,0.,2.07554270972356,3.42259881132375,0.,1.71909544841957},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,0.,0.,0.,0.,0.,0.,3.7410698100085,6.16907617592665,7.58242635758397,0.,0.,0.,3.76406389114815,6.20699373673991,0.,3.6179005979755},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,0.,0.,0.,0.,0.,0.,5.18623489409549,8.55217350992208,10.5114970194908,0.,0.,0.,5.81833621896758,9.5945173923078,0.,6.23568135536041},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,0.,0.,0.,0.,0.,0.,6.45780747384476,10.6490143886212,13.0887291840651,0.,0.,0.,7.79428043908953,12.8528769082785,0.,8.98680735137037},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,0.,0.,0.,0.,0.,0.,7.27823584467216,12.0019121888667,14.7515791225583,0.,0.,0.,9.14177937313726,15.0749214021478,0.,10.9691712594614},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,0.,0.,0.,0.,0.,0.,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,0.,0.,0.,6.8546484667622e-008,1.17631378114674e-007,0.,7.65361974719019e-010},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,0.,0.,0.,0.,0.,0.,.000824565926330002,.00141502261880692,.00180994395496465,0.,0.,0.,5.01144631725624e-005,8.60005205963556e-005,0.,2.90963994598978e-006},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,0.,0.,0.,0.,0.,0.,.0113956405123916,.019555851831759,.0250137314673599,0.,0.,0.,.00166206328995312,.00285223664290029,0.,.000231576690958128},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.23320173943763,.400193272045958,.511883968508167,.600085200896572,0.,0.,0.,0.,0.,0.,.0657431866024243,.112820689170272,.144308028468726,0.,0.,0.,.0171974308138302,.0295121988599679,0.,.00429748986298374},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,0.,0.,0.,0.,0.,0.,.236321146004057,.405546428999883,.518731148089159,0.,0.,0.,.0946962283839188,.162506478648407,0.,.0362494356624709},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,0.,0.,0.,0.,0.,0.,.626153181688297,1.07453010927869,1.3744227476422,0.,0.,0.,.347192252628355,.595810322665643,0.,.183906859961414},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,0.,0.,0.,0.,0.,0.,1.33269841309056,2.28701955580991,2.92530816462441,0.,0.,0.,.95054294779362,1.63120949890175,0.,.647664476709677},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,0.,0.,0.,0.,0.,0.,2.39387697960251,4.10808883148424,5.25462310508641,0.,0.,0.,2.07554270972356,3.5618011698851,0.,1.71909544841957},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,0.,0.,0.,0.,0.,0.,3.7410698100085,6.41998199374919,8.21174689798638,0.,0.,0.,3.76406389114815,6.45944172008843,0.,3.6179005979755},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,0.,0.,0.,0.,0.,0.,5.18623489409549,8.90000356218188,11.3839223715851,0.,0.,0.,5.81833621896758,9.98474117367778,0.,6.23568135536041},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,0.,0.,0.,0.,0.,0.,6.45780747384476,11.0821261849396,14.1750577199246,0.,0.,0.,7.79428043908953,13.3756231834223,0.,8.98680735137037},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,0.,0.,0.,0.,0.,0.,7.27823584467216,12.4900484198527,15.9759196314393,0.,0.,0.,9.14177937313726,15.6880416449771,0.,10.9691712594614}
		};
		for (int j = 0; j < 144; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[144][56] = 
		{
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.43216369764802,.0548080819457638,.000652356451628676,7.11647557031877e-006,7.43966158485903e-008,5.37669458746301,.0858602163438174,.00102195632637702,1.11483947361089e-005,7.27823584467216,.116225851033514,.00138338509726576,9.14177937313726,.145984701575574,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.29802783242322,.0526660717910194,.000626861048500277,6.83834938169028e-006,7.14890463630537e-008,4.96464287844228,.0792801794259956,.000943637046012817,1.02940194263112e-005,6.45780747384476,.103124463877826,.00122744505825102,7.79428043908953,.124466546112493,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,3.06556135691961,.0489538241357645,.000582675800230771,6.35633799191501e-006,6.64500329012046e-008,4.2894282422279,.068497704468415,.000815297997194491,8.89398467007483e-006,5.18623489409549,.0828187732700635,.000985755369367884,5.81833621896758,.0929127735338078,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.74931534527504,.0439037044887576,.000522566451093507,5.70061256202903e-006,5.95949889363459e-008,3.45007465931297,.0550941013725511,.00065576081497714,7.15361334838549e-006,3.7410698100085,.0597409910483066,.000711070695350591,3.76406389114815,.0601081826012263,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,2.36915365304765,.0378329179483261,.000450308553617182,4.91236012600413e-006,5.13544893947963e-008,2.56192161158569,.0409112216155168,.00048694824599985,5.31205798363975e-006,2.39387697960251,.0382277237453785,.000455007753107355,2.07554270972356,.0331442567927967,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.94896143440088,.0311228855660286,.000370441994529039,4.04110574472654e-006,4.22462760847341e-008,1.73374759666501,.027686144584506,.000329535980876457,3.59486711881878e-006,1.33269841309056,.0212818065446243,.000253307966815722,.95054294779362,.0151791815252261,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.51514018159912,.02419521169381,.000287984944676549,3.14159202119275e-006,3.28426357182316e-008,1.04781596872157,.0167325304235648,.000199159937521975,2.17261104198562e-006,.626153181688297,.00999901459257298,.000119013865260662,.347192252628355,.00554429890730429,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,1.09494796295235,.0174851793115125,.000208118385588407,2.27033763991516e-006,2.37344224081693e-008,.547225517982271,.0087386219541596,.000104011967010535,1.1346536398707e-006,.236321146004057,.00377380273155657,4.4917911225712e-005,.0946962283839188,.00151220020487266,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.714786270724957,.011414392771081,.000135860488112082,1.48208520389026e-006,1.54939228666198e-008,.23320173943763,.00372398905575873,4.43250009952421e-005,4.83535934970157e-007,.0657431866024243,.00104985026256264,1.24959051250218e-005,.0171974308138302,.000274625070541602,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.398540259080393,.0063642731240741,7.57511389748173e-005,8.26359774004284e-007,8.63887890176104e-009,.0724976248699598,.00115771161156557,1.37797312415549e-005,1.50321378001502e-007,.0113956405123916,.000181976518058871,2.16598631798674e-006,.00166206328995312,2.65414208197256e-005,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.166073783576779,.00265202546881917,3.15658907053118e-005,3.44348384229007e-007,3.59986543991196e-009,.0125887190510939,.000201028740546051,2.39275652837828e-006,2.61022840186632e-008,.000824565926330002,1.31674596105725e-005,1.56726470334577e-007,5.01144631725624e-005,8.00275816365055e-007,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0225835186400418,.000268801672326118,2.93232408400406e-006,3.06549198606349e-008,3.12457717751721e-010,.0319379183519753,.000510015314074827,6.0704875769121e-006,6.62221956005248e-008,6.92295954375316e-010,.000465578153648717,7.43479852667174e-006,8.84931311987991e-008,9.65360585942034e-010,5.86464816800813e-006,9.36523271415897e-008,1.11470238797886e-009,6.8546484667622e-008,1.09461601509022e-009,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.43216369764802,.284996205419761,.017638997934852,.00100057208936333,5.43915050229927e-005,5.37669458746301,.446464006507695,.0276325703309419,.00156745744490148,7.27823584467216,.60436208201944,.037405205103521,9.14177937313726,.759104944404841,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.29802783242322,.273857980099692,.0169496303935219,.000961467718258155,5.22657755328352e-005,4.96464287844228,.412248513344535,.0255149034922769,.00144733280168761,6.45780747384476,.53623626019628,.0331887586818224,7.79428043908953,.647212821256298,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,3.06556135691961,.25455468653848,.0157549100822091,.000893697213237962,4.85817433642365e-005,4.2894282422279,.356180788679677,.0220447573606405,.001250488775823,5.18623489409549,.430648825529863,.0266537364367886,5.81833621896758,.483136554903789,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.74931534527504,.228294600704113,.0141296196713449,.000801502620992592,4.35700078975819e-005,3.45007465931297,.286483476063375,.01773104815973,.001005793638124,3.7410698100085,.310646808870845,.0192265662361566,3.76406389114815,.31255616594975,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,2.36915365304765,.196727155420251,.0121758459312683,.000690674813173132,3.75453632669935e-005,2.56192161158569,.212734007569305,.0131665427453459,.000746872086187954,2.39387697960251,.198780103635972,.0123029070952925,2.07554270972356,.172346615354148,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.94896143440088,.161835699647526,.0100163423848523,.000568176982887904,3.0886331476986e-005,1.73374759666501,.143965011530511,.00891028895571934,.000505436106471301,1.33269841309056,.110663133873164,.00684916764813009,.95054294779362,.0789301318799674,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.51514018159912,.125812531241055,.00778679482932368,.000441705906457738,2.40113124133503e-005,1.04781596872157,.0870073811834108,.00538506474165289,.000305467776594366,.626153181688297,.0519938139714517,.00321800346775215,.347192252628355,.028829765505355,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,1.09494796295235,.0909210754683297,.00562729128290768,.000319208076172511,1.73522806233428e-005,.547225517982271,.0454399061072379,.00281236870842355,.000159531604082818,.236321146004057,.0196233734207421,.0012145307084342,.0946962283839188,.00786328046747136,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.714786270724957,.0593536301844674,.0036735175428311,.000208380268353051,1.13276359927544e-005,.23320173943763,.0193643475968783,.00119849907066209,6.79848551371716e-005,.0657431866024243,.0054591098696972,.000337875473054763,.0171974308138302,.00142802120123345,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.398540259080393,.0330935443501007,.00204822713196687,.00011618567610768,6.31590052609987e-006,.0724976248699598,.00601997742947992,.000372588713280563,2.1135093316452e-005,.0113956405123916,.000946258567734596,5.85658777413804e-005,.00166206328995312,.000138012569524986,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.166073783576779,.0137902507888884,.000853506820654063,4.8415171087487e-005,2.63186835750109e-006,.0125887190510939,.00104532810129413,6.46974937663248e-005,3.66996508308664e-006,.000824565926330002,6.8469391571569e-005,4.23771065598628e-006,5.01144631725624e-005,4.16134925463982e-006,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.117431898525612,.00726810033258317,.000412282963165001,2.24118692572326e-005,1.18785553376488e-006,.0319379183519753,.00265202546881917,.000164139279324007,9.31079998231151e-006,5.06138867343602e-007,.000465578153648717,3.86601627522499e-005,2.39275652837828e-006,1.3572910479643e-007,5.86464816800813e-006,4.86982155122583e-007,3.01403214001255e-008,6.8546484667622e-008,5.69188702865245e-009,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.43216369764802,.683927704293011,.101581856753867,.0138280754431197,.00180391284393747,5.37669458746301,1.0714146266282,.15913419863567,.0216625269007544,7.27823584467216,1.45033499905202,.215414174926983,9.14177937313726,1.82168355953934,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.29802783242322,.657198433066984,.0976118333379795,.0132876464230156,.00173341229925847,4.96464287844228,.989305029218415,.146938691257518,.0200023840218978,6.45780747384476,1.28684813137965,.191131930663375,7.79428043908953,1.55316727245144,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,3.06556135691961,.610874808402686,.0907315157613829,.0123510465855827,.00161123011394102,4.2894282422279,.854754921231925,.126954342454893,.017281966303755,5.18623489409549,1.03346169878758,.153497157076452,5.81833621896758,1.1594206118243,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.74931534527504,.547856424727075,.0813715726223437,.0110769017332834,.00144501419520715,3.45007465931297,.687496824083399,.102111968089684,.0139002381298067,3.7410698100085,.745483465381799,.110724560685561,3.76406389114815,.750065499032834,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,2.36915365304765,.472101555108414,.0701199150776834,.00954524269140647,.001245204798047,2.56192161158569,.51051445126307,.0758252744311929,.0103218985058226,2.39387697960251,.477028175651108,.0708516521644813,2.07554270972356,.413593664547998,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.94896143440088,.388369797308574,.057683472785343,.00785230196598564,.00102435573404142,1.73374759666501,.345484108076311,.0513137820911617,.00698521244622072,1.33269841309056,.26556698531571,.0394439167964793,.95054294779362,.189414816301507,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.51514018159912,.301922180108511,.0448436515410674,.0061044503070795,.00079634337832586,1.04781596872157,.208798423760314,.0310122421476135,.00422161631839709,.626153181688297,.124773625580585,.0185322753878643,.347192252628355,.0691850451424154,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,1.09494796295235,.21819042230867,.032407209248727,.00441150958165867,.000575494314320284,.547225517982271,.109045699824108,.0161962508490161,.00220475374065531,.236321146004057,.0470917453597858,.00699440438187064,.0946962283839188,.0188701296931614,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.714786270724957,.142435552690009,.0211555517040667,.00287985053978171,.000375684917160134,.23320173943763,.0464701408131328,.00690207920910791,.000939562192289973,.0657431866024243,.0131006533057736,.00194580315907169,.0171974308138302,.00342693426532766,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.398540259080393,.0794171690143981,.0117956085650275,.00160570568748242,.000209468998426257,.0724976248699598,.0144466110949646,.00214571448108123,.000292090562973066,.0113956405123916,.00227081076025246,.000337277130218073,.00166206328995312,.000331199567025036,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.166073783576779,.0330935443501007,.00491529098843092,.000669105850049561,8.72868131088095e-005,.0125887190510939,.0025085556753223,.000372588713280563,5.07195379343716e-005,.000824565926330002,.000164311358893045,2.44047036237916e-005,5.01144631725624e-005,9.98631556618579e-006,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.281810519726389,.0418565232324459,.00569782025619581,.000743297299691906,9.45407881908902e-005,.0319379183519753,.0063642731240741,.000945267572543323,.000128676829945482,1.67862684298058e-005,.000465578153648717,9.27758189615077e-005,1.37797312415549e-005,1.87579917619568e-006,5.86464816800813e-006,1.16864919971161e-006,1.73576176055534e-007,6.8546484667622e-008,1.36592668764407e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.43216369764802,1.22663174436907,.326756438335703,.0797762743762283,.0186651534362866,5.37669458746301,1.92159373593738,.511883968508167,.124974418661917,7.27823584467216,2.60119152746664,.692918704338266,9.14177937313726,3.26720919228747,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.29802783242322,1.17869250695568,.313986139062399,.0766584570077852,.0179356816726159,4.96464287844228,1.7743292837621,.472654910466463,.115396801418492,6.45780747384476,2.30797605978546,.614810468242533,7.79428043908953,2.78562232344933,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,3.06556135691961,1.09561058445014,.291854351578,.071255069825012,.0166714580468485,4.2894282422279,1.53301220795345,.408371633453457,.0997022970608015,5.18623489409549,1.85352474883709,.493751403486214,5.81833621896758,2.07943342346341,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.74931534527504,.982586430858874,.261746399421271,.0639043340158099,.0149516157400831,3.45007465931297,1.23303299936556,.3284616374485,.0801925919119944,3.7410698100085,1.33703266851113,.356165601240813,3.76406389114815,1.34525059548623,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,2.36915365304765,.846719251796657,.225553405296596,.0550679596064988,.0128841804598588,2.56192161158569,.915613197049483,.243905726875982,.0595486053479555,2.39387697960251,.855555199093713,.227907170151078,2.07554270972356,.741784716454702,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.94896143440088,.696545437394333,.185549336471009,.0453011265884548,.0105990469625404,1.73374759666501,.619629489317892,.16506007282648,.0402987550204044,1.33269841309056,.476297264169728,.126878501518793,.95054294779362,.339717524338981,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.51514018159912,.541500699745815,.144247726196132,.0352175040277159,.00823979461889025,1.04781596872157,.374482234234657,.0997564930665907,.0243551478376682,.626153181688297,.223782848737467,.0596124199094679,.347192252628355,.12408412771496,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,1.09494796295235,.39132688534349,.104243657370545,.0254506710096719,.00595466112157189,.547225517982271,.195574643564791,.0520981739351268,.0127195602938403,.236321146004057,.0844595552914031,.0224987683542483,.0946962283839188,.0338437819621889,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.714786270724957,.255459706281274,.0680506632458702,.0166142966003608,.00388722584134757,.23320173943763,.0833447008052014,.0222017877163202,.00542047745935247,.0657431866024243,.0234961635798143,.00625902823868614,.0171974308138302,.0061462437174735,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.398540259080393,.142435552690009,.0379427110891412,.00926356079115871,.00216738353458216,.0724976248699598,.0259101534510232,.00690207920910791,.00168511496703183,.0113956405123916,.00407272369067121,.00108491296894855,.00166206328995312,.000594010010245393,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.166073783576779,.0593536301844674,.0158109236047423,.00386017360838543,.000903159908814793,.0125887190510939,.00449912177055081,.00119849907066209,.000292608743069969,.000824565926330002,.000294694201614443,7.85021488046878e-005,5.01144631725624e-005,1.79105651165762e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.505430219047594,.134639087033905,.0328715933012122,.00769092486599526,.00175444288977385,.0319379183519753,.011414392771081,.00304062433143834,.000742356239942391,.000173688145144045,.000465578153648717,.000166394436040081,4.43250009952421e-005,1.08217712848354e-005,5.86464816800813e-006,2.09598499578199e-006,5.58339204266198e-007,6.8546484667622e-008,2.44980430684131e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.43216369764802,1.87902032369395,.766758694225435,.286764630052617,.102778121439049,5.37669458746301,2.94360039211812,1.20117444444972,.449234352115903,7.27823584467216,3.98464475452196,1.62598614322311,9.14177937313726,5.00488635016176,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.29802783242322,1.80558442753713,.736792221198799,.275557291138585,.0987613455915411,4.96464287844228,2.71801280247251,1.1091204929715,.414806549024428,6.45780747384476,3.53548156799627,1.44269926011819,7.79428043908953,4.26716574385505,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,3.06556135691961,1.67831508069679,.684858308101796,.256134219070923,.0918000029066945,4.2894282422279,2.34835035734144,.958274921896979,.358390919550843,5.18623489409549,2.83932866528416,1.15862500931079,5.81833621896758,3.18538769471191,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.74931534527504,1.50517861766197,.614207721386262,.229711187268269,.0823298336919529,3.45007465931297,1.88882610957123,.770760072903028,.288261129419902,3.7410698100085,2.04813838310324,.835769519114577,3.76406389114815,2.06072704459494,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,2.36915365304765,1.29704998251713,.529278123498329,.197947863419097,.0709456725585557,2.56192161158569,1.40258542451411,.572343234018839,.214054039453166,2.39387697960251,1.31058536069927,.534801410737989,2.07554270972356,1.13630563078316,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.94896143440088,1.0670056757038,.435405550603841,.162839903325643,.0583627742237777,1.73374759666501,.949181699343098,.387325943916553,.14485832617988,1.33269841309056,.729617706050062,.297729999278181,.95054294779362,.520397532085943,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.51514018159912,.829499827301664,.338488198569183,.126593208229511,.0453717466006328,1.04781596872157,.573651980082395,.234086155283846,.0875472690866454,.626153181688297,.342802575265255,.139885051637275,.347192252628355,.190078724845143,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,1.09494796295235,.599455520488332,.244615625674694,.0914852481360572,.0327888482658548,.547225517982271,.299591733007419,.122252305177188,.0457218644341953,.236321146004057,.129379678661691,.0527950614618356,.0946962283839188,.0518437211734884,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.714786270724957,.39132688534349,.159686027786761,.0597219242868854,.0214046871324576,.23320173943763,.127671884739716,.0520981739351268,.0194845049545575,.0657431866024243,.0359926841107618,.0146872831147845,.0171974308138302,.00941514591687767,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.398540259080393,.21819042230867,.089035441071228,.0332988924842312,.011934517917716,.0724976248699598,.0396905633235048,.0161962508490161,.00605733188088067,.0113956405123916,.00623881668046779,.00254583033055523,.00166206328995312,.000909936406629402,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.166073783576779,.0909210754683297,.0371015279742252,.0138758204165696,.00497317523286939,.0125887190510939,.0068919961385699,.00281236870842355,.00105181444749975,.000824565926330002,.000451428390509428,.000184211229066994,5.01144631725624e-005,2.74363646769969e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.774245129525821,.315940800100654,.118160573002028,.0423494407669399,.0147987701388505,.0319379183519753,.0174851793115125,.00713505494758908,.00266848150253762,.000956399385361249,.000465578153648717,.000254891925391777,.000104011967010535,3.8900052218214e-005,5.86464816800813e-006,3.21074228130114e-006,1.3101851684348e-006,6.8546484667622e-008,3.7527416869482e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.43216369764802,2.60010456277178,1.46817349989066,.759807526915221,.376823534710004,5.37669458746301,4.07322300563352,2.29998368635187,1.19028501535882,7.27823584467216,5.51377378765633,3.11340423610211,9.14177937313726,6.92553862835673,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.29802783242322,2.49848724322454,1.41079432452006,.730112719497692,.362096512540277,4.96464287844228,3.76106495510795,2.12372070669689,1.09906559293822,6.45780747384476,4.89224179250545,2.76245034842036,7.79428043908953,5.90471373874725,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,3.06556135691961,2.32237759435708,1.31135235466843,.67864962117166,.336573592680459,4.2894282422279,3.24954254199156,1.83488476431321,.94958753526484,5.18623489409549,3.92893643816517,2.21851091848272,5.81833621896758,4.40779749680161,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.74931534527504,2.08279907471978,1.17607208990109,.608639528072336,.301852364194909,3.45007465931297,2.6136733721559,1.475835256036,.763772630572106,3.7410698100085,2.8341225946006,1.60031398303209,3.76406389114815,2.85154222320761,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,2.36915365304765,1.79479994716393,1.01345067339222,.52447987234346,.260113716144808,2.56192161158569,1.94083518734219,1.09591084568925,.56715456963183,2.39387697960251,1.81352960012534,1.02402654819954,2.07554270972356,1.57236907874115,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.94896143440088,1.4764748900889,.833705435512044,.431458315505189,.213980043325934,1.73374759666501,1.31343532384456,.741643610905347,.38381458183186,1.33269841309056,1.0096124574376,.570087171302651,.95054294779362,.720102907121204,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.51514018159912,1.14782488437669,.648130118259726,.335419582434137,.166350013900149,1.04781596872157,.793794038333268,.448223270850827,.231964011742662,.626153181688297,.474354922529452,.267848969162523,.347192252628355,.263022466294621,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,1.09494796295235,.829499827301664,.468384880379551,.242398025595865,.120216341081275,.547225517982271,.414561685224313,.234086155283846,.121144008364367,.236321146004057,.179029831969503,.101090878717283,.0946962283839188,.0717390303204678,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.714786270724957,.541500699745815,.305763463870679,.15823836986699,.0784776930311739,.23320173943763,.176666662868666,.0997564930665907,.0516258700383423,.0657431866024243,.0498050718253105,.0281229023119214,.0171974308138302,.0130282592204804,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.398540259080393,.301922180108511,.170483199103343,.0882282767676657,.0437564645456243,.0724976248699598,.0549220322394108,.0310122421476135,.0160494212806935,.0113956405123916,.00863299641447793,.00487470263420339,.00166206328995312,.00125912943722616,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.166073783576779,.125812531241055,.0710412292517083,.0367651784416334,.0182335446858059,.0125887190510939,.00953683703179566,.00538506474165289,.0027868727533861,.000824565926330002,.000624666483447503,.000352723805983911,5.01144631725624e-005,3.79652184082228e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.07136589668511,.604956831576084,.313076590844149,.155269095594961,.0750796611806128,.0319379183519753,.02419521169381,.0136620538811056,.0070703710241047,.00350652251607874,.000465578153648717,.000352708084030703,.000199159937521975,.000103069030749612,5.86464816800813e-006,4.44288204349147e-006,2.50871513910765e-006,6.8546484667622e-008,5.19287666029906e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.43216369764802,3.34457733996838,2.42928372243492,1.61716703673679,1.03166747175281,5.37669458746301,5.23948519622497,3.80562170038934,2.53339118523212,7.27823584467216,7.09250048322984,5.15153163620379,9.14177937313726,8.90848771819592,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.29802783242322,3.21386452588688,2.33434242513939,1.55396489406674,.991347830518138,4.96464287844228,4.83794875141232,3.51397170985646,2.33924064338676,6.45780747384476,6.29300885631198,4.57083285180669,7.79428043908953,7.59537599078757,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,3.06556135691961,2.98733042822583,2.16980277181815,1.44443133027185,.921471180632688,4.2894282422279,4.1799651086425,3.03605513300837,2.02109298227284,5.18623489409549,5.05388589728677,3.6708144253493,5.81833621896758,5.66985645040998,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.74931534527504,2.67915478814007,1.94596400535734,1.29542252108203,.826411401430188,3.45007465931297,3.36203122741823,2.44196109298446,1.62560632524607,3.7410698100085,3.6456003901969,2.64792731244254,3.76406389114815,3.66800767948205,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,2.36915365304765,2.30869454983061,1.67688575263467,1.11629791880228,.712139330973213,2.56192161158569,2.49654320873854,1.81332681653159,1.20712633430595,2.39387697960251,2.33278695510428,1.69438491095747,2.07554270972356,2.02257634793674,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.94896143440088,1.89922533544551,1.37947391363879,.918311731385135,.585834561722673,1.73374759666501,1.68950360094811,1.2271456690605,.816907266603923,1.33269841309056,1.29868890491693,.943283260267015,.95054294779362,.926285773305773,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.51514018159912,1.4764748900889,1.07241545118845,.713903815076353,.455433067359994,1.04781596872157,1.02107645639498,.741643610905347,.493709972869637,.626153181688297,.610174201341807,.443190904144467,.347192252628355,.338332155278326,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,1.09494796295235,1.0670056757038,.77500361219257,.515917627659212,.329128298109454,.547225517982271,.533260715077649,.387325943916553,.257841742921929,.236321146004057,.230290399761936,.167267987177829,.0946962283839188,.092279648532812,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.714786270724957,.696545437394333,.505925359469895,.336793025379457,.214856227652479,.23320173943763,.227250598233051,.16506007282648,.109880005542746,.0657431866024243,.0640654676075239,.0465329941039433,.0171974308138302,.0167585647100291,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.398540259080393,.388369797308574,.282086593009086,.187784216189635,.119796448449979,.0724976248699598,.0706475374579269,.0513137820911617,.0341594339808844,.0113956405123916,.0111048319353399,.00806582856989507,.00166206328995312,.00161964862649386,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.166073783576779,.161835699647526,.117546939687847,.0782506523947475,.0499197985645293,.0125887190510939,.0122674639659539,.00891028895571934,.00593155317986048,.000824565926330002,.000803523594969987,.000583627344081487,5.01144631725624e-005,4.88355780044979e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.37812384631489,1.00097964159756,.66634920654528,.425095729270401,.264407915399588,.0319379183519753,.0311228855660286,.0226056423923128,.0150485097246962,.00960015732985275,.000465578153648717,.000453696932947651,.000329535980876457,.00021937113431002,5.86464816800813e-006,5.71498655076136e-006,4.15099499624134e-006,6.8546484667622e-008,6.67972275267736e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.43216369764802,4.06566157904621,3.58970041611072,2.90485747067004,2.25268085447981,5.37669458746301,6.36910780974037,5.62348550533029,4.55063709769945,7.27823584467216,8.62162951636421,7.61230772976691,9.14177937313726,10.8291399963909,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.29802783242322,3.90676734157429,3.44940769885402,2.79132979410552,2.1646415527128,4.96464287844228,5.88100090404776,5.19252057410142,4.20189164401264,6.45780747384476,7.64976908082116,6.7542215997961,7.79428043908953,9.23292398567978,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,3.06556135691961,3.63139294188613,3.206270984711,2.59457869551726,2.01206352182394,4.2894282422279,5.08115729329262,4.4863135061768,3.6304147407808,5.18623489409549,6.14349367016778,5.42428368841739,5.81833621896758,6.89226625249968,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.74931534527504,3.25677524519788,2.87550924383838,2.32691967035921,1.80449727542797,3.45007465931297,4.0868784900029,3.60843349447312,2.92001665318886,3.7410698100085,4.43158460169426,3.91278535624427,3.76406389114815,4.45882285809472,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,2.36915365304765,2.80644451447741,2.47789808510688,2.00516475742016,1.55498034059349,2.56192161158569,3.03479297156663,2.67951411674704,2.16831648839865,2.39387697960251,2.83573119453035,2.50375621577018,2.07554270972356,2.45863979589474,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.94896143440088,2.30869454983061,2.03841899407272,1.64952947513746,1.27918960054353,1.73374759666501,2.05375722544957,1.81332681653159,1.46738037712382,1.33269841309056,1.57868365630447,1.39386942769156,.95054294779362,1.12599114834103,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.51514018159912,1.79479994716393,1.58468529460864,1.28235907822422,.994453522505445,1.04781596872157,1.24121851464586,1.09591084568925,.886833005159806,.626153181688297,.741726548606004,.654893686699248,.347192252628355,.411275896727804,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,1.09494796295235,1.29704998251713,1.14520620357448,.926723795941524,.718662782455483,.547225517982271,.648230667294543,.572343234018839,.463151607819505,.236321146004057,.279940553069749,.247168314523763,.0946962283839188,.112174957679791,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.714786270724957,.846719251796657,.747595044842976,.604968883002474,.469145847621007,.23320173943763,.276245376362,.243905726875982,.197373399115395,.0657431866024243,.0778778553220726,.0687608066341364,.0171974308138302,.0203716780136318,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.398540259080393,.472101555108414,.416833303970362,.337309857844425,.261579601225039,.0724976248699598,.0858790063738329,.0758252744311929,.0613593306931733,.0113956405123916,.0134990116693501,.0119187017581499,.00166206328995312,.00196884165709061,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.166073783576779,.196727155420251,.173696589827337,.140558759256161,.109001570336171,.0125887190510939,.0149123048591796,.0131665427453459,.0106546300881587,.000824565926330002,.000976761687908062,.000862413599759767,5.01144631725624e-005,5.93644317357238e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.67524461347418,1.47912613202691,1.19693848980126,.928210917633743,.701817777931477,.0319379183519753,.0378329179483261,.0334038725706356,.0270310826916448,.0209622685691644,.000465578153648717,.000551513091586578,.00048694824599985,.000394048273229513,5.86464816800813e-006,6.94712631295169e-006,6.13383621174789e-006,6.8546484667622e-008,8.11985772602823e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.43216369764802,4.71805015837109,4.8341582287849,4.53961280320503,4.08530965077445,5.37669458746301,7.3911144659211,7.57300486916358,7.11158142526392,7.27823584467216,10.0050827434195,10.2513011633895,9.14177937313726,12.5668171542652,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.29802783242322,4.53365926215574,4.64522959548689,4.36219559797068,3.92564752711369,4.96464287844228,6.82468442275817,6.99263535997903,6.56657384998658,6.45780747384476,8.87727458903197,9.0957384017764,7.79428043908953,10.7144674060855,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,3.06556135691961,4.21409743813278,4.31780356792227,4.05471964942098,3.64894232901744,4.2894282422279,5.89649544268062,6.04160426743658,5.67348911421506,5.18623489409549,7.12929758661485,7.30474484999971,5.81833621896758,7.99822052374818,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.74931534527504,3.77936743200098,3.87237514603213,3.63643119645238,3.27251422208425,3.45007465931297,4.74267160020857,4.85938558884804,4.56330305986767,3.7410698100085,5.14269031628637,5.26924849841671,3.76406389114815,5.17429930720343,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,2.36915365304765,3.25677524519788,3.33692231375341,3.1336035234873,2.82000718368873,2.56192161158569,3.52176519903125,3.60843349447312,3.38857151909233,2.39387697960251,3.29076135613591,3.37174480048962,2.07554270972356,2.8531607102232,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.94896143440088,2.67915478814007,2.74508700215838,2.57782875759162,2.31985174902974,1.73374759666501,2.38330943547478,2.44196109298446,2.29317232064629,1.33269841309056,1.83200409818481,1.87708849838432,.95054294779362,1.306671156088,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.51514018159912,2.08279907471978,2.13405537202647,2.00402730550153,1.80347365436117,1.04781596872157,1.44038826049359,1.475835256036,1.38591256375812,.626153181688297,.860746275133792,.881928667453122,.347192252628355,.477270493857987,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,1.09494796295235,1.50517861766197,1.54222006043145,1.44825253960585,1.30331821970218,.547225517982271,.752247756737171,.770760072903028,.723797635481732,.236321146004057,.324860676440036,.332855281233889,.0946962283839188,.130174896891091,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.714786270724957,.982586430858874,1.00676722815273,.945424866640774,.850811181306659,.23320173943763,.320572560296515,.3284616374485,.308448458722372,.0657431866024243,.0903743758530202,.0925984290266871,.0171974308138302,.023640580213036,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.398540259080393,.547856424727075,.561338806262592,.527136413672172,.474383074373468,.0724976248699598,.0996594162463145,.102111968089684,.0958902824056873,.0113956405123916,.0156651046591466,.0160506124472133,.00166206328995312,.00228476805347462,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.166073783576779,.228294600704113,.233912778697964,.219660465122477,.197677876277223,.0125887190510939,.0173051792271987,.01773104815973,.0166506947931123,.000824565926330002,.00113349587680305,.00116139045508752,5.01144631725624e-005,6.88902312961445e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94405952395241,1.99190153319138,1.87053490500425,1.68334054609746,1.47700173412851,.0319379183519753,.0439037044887576,.0449841453999596,.0422432598881281,.0380157526164569,.000465578153648717,.000640010580938273,.00065576081497714,.000615805286247286,5.86464816800813e-006,8.06188359847084e-006,8.26028118222923e-006,6.8546484667622e-008,9.42279510613512e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.43216369764802,5.26075419844715,6.01023945321738,6.29325287185994,6.31490479284833,5.37669458746301,8.24129357523028,9.41540812070088,9.8587659714966,7.27823584467216,11.1559392718342,12.7452954129935,9.14177937313726,14.0123427870133,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.29802783242322,5.05515333604444,5.77534719856822,6.04729988318872,6.06810560352602,4.96464287844228,7.60970867730185,8.69384305053515,9.10322345344791,6.45780747384476,9.89840251743778,11.3086008382978,7.79428043908953,11.9469224570834,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,3.06556135691961,4.69883321418023,5.36826312400043,5.6210468126907,5.64038601038998,4.2894282422279,6.57475272940214,7.51143976063047,7.86514251530264,5.18623489409549,7.94936063666436,9.0818842940915,5.81833621896758,8.91823333538729,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.74931534527504,4.21409743813278,4.81446836840316,5.04117466895809,5.05851881797664,3.45007465931297,5.28820777549073,6.04160426743658,6.32609461018441,3.7410698100085,5.7342395194157,6.55118093268348,3.76406389114815,5.76948440365682,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,2.36915365304765,3.63139294188613,4.14874755712812,4.34410603466752,4.35905192076855,2.56192161158569,3.92686394481766,4.4863135061768,4.69756747292958,2.39387697960251,3.66928837957851,4.19204185444651,2.07554270972356,3.1813517621299,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.94896143440088,2.98733042822583,3.41292721960274,3.57363698957388,3.585932078825,1.73374759666501,2.65745481671636,3.03605513300837,3.17901854589628,1.33269841309056,2.04273437703882,2.33375715403626,.95054294779362,1.45697386412547,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.51514018159912,2.32237759435708,2.65324037511448,2.77817759847913,2.78773591165655,1.04781596872157,1.60607207096794,1.83488476431321,1.92128681456264,.626153181688297,.959755498290674,1.09648923787555,.347192252628355,.532169576430532,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,1.09494796295235,1.67831508069679,1.9174200375891,2.00770855338548,2.014616069713,.547225517982271,.838776700477854,.958274921896979,1.00339869182784,.236321146004057,.362228486371654,.41383418762992,.0946962283839188,.145148549160118,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.714786270724957,1.09561058445014,1.25169922631407,1.31063991909491,1.31514917250492,.23320173943763,.357447120288584,.408371633453457,.427601258703136,.0657431866024243,.100769886127061,.115126296058794,.0171974308138302,.0263598896651818,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.398540259080393,.610874808402686,.697904470716799,.730767775362299,.733281980091582,.0724976248699598,.111122958602373,.126954342454893,.132932437477267,.0113956405123916,.0174670175895654,.0199554957891379,.00166206328995312,.00254757849669498,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.166073783576779,.25455468653848,.290820396149008,.304514704864287,.305562386955532,.0125887190510939,.0192957453224272,.0220447573606405,.0230828128670437,.000824565926330002,.00126387871952445,.00144394006224153,5.01144631725624e-005,7.68144808465349e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.16767922327361,2.47650255021136,2.59311744704774,2.60203905976421,2.54570625309438,.0319379183519753,.0489538241357645,.0559281414998414,.058561716193063,.0587631976332298,.000465578153648717,.000713629198016847,.000815297997194491,.00085368919204875,5.86464816800813e-006,8.98921939454121e-006,1.02698888837534e-005,6.8546484667622e-008,1.05066727253324e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.43216369764802,5.6596856973204,6.95633329528564,7.83624657590448,8.45948687893877,5.37669458746301,8.86624419535079,10.897522055244,12.2759600892934,7.27823584467216,12.0019121888667,14.7515791225583,9.14177937313726,15.0749214021478,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.29802783242322,5.43849378901173,6.68446578908404,7.52999028769359,8.12887310528866,4.96464287844228,8.18676519317573,10.0623727801811,11.3351719800974,6.45780747384476,10.6490143886212,13.0887291840651,7.79428043908953,12.8528769082785,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,3.06556135691961,5.05515333604444,6.21330111687109,6.99922754350219,7.55589720071174,4.2894282422279,7.07332686195439,8.69384305053515,9.79353561019799,5.18623489409549,8.55217350992208,10.5114970194908,5.81833621896758,9.5945173923078,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.74931534527504,4.53365926215574,5.57233149709104,6.27718106081507,6.77642418552391,3.45007465931297,5.68922112351076,6.99263535997903,7.87714052450818,3.7410698100085,6.16907617592665,7.58242635758397,3.76406389114815,6.20699373673991,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,2.36915365304765,3.90676734157429,4.80181713058638,5.40920359195237,5.83941385309815,2.56192161158569,4.22464438851143,5.19252057410142,5.84932748998933,2.39387697960251,3.94753645159365,4.85192654207924,2.07554270972356,3.42259881132375,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.94896143440088,3.21386452588688,3.9501686142305,4.44982923668813,4.80373756449036,1.73374759666501,2.85897391326216,3.51397170985646,3.95845736731918,1.33269841309056,2.19763822848137,2.70112242949704,.95054294779362,1.56745854854701,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.51514018159912,2.49848724322454,3.07089667649178,3.45933734693588,3.73446887011071,1.04781596872157,1.72786311354484,2.12372070669689,2.39235217915172,.626153181688297,1.03253530989981,1.26909163149447,.347192252628355,.572524856067592,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,1.09494796295235,1.80558442753713,2.21924816013591,2.49996299167164,2.69879258150293,.547225517982271,.902382494194724,1.1091204929715,1.24941420966279,.236321146004057,.389696858310697,.47897734533629,.0946962283839188,.156155398385808,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.714786270724957,1.17869250695568,1.44873379363125,1.63198552280894,1.76178224907717,.23320173943763,.384552913504838,.472654910466463,.532441484172332,.0657431866024243,.108411429563137,.133248748684711,.0171974308138302,.0283588027292761,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.398540259080393,.657198433066984,.807764173851191,.909939040121824,.982309233889342,.0724976248699598,.119549592267858,.146938691257518,.165525107479116,.0113956405123916,.0187915697820833,.0230967635919399,.00166206328995312,.00274076549419503,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.166073783576779,.273857980099692,.336599501638247,.379176295930421,.409333329312416,.0125887190510939,.0207589728964554,.0255149034922769,.028742308147258,.000824565926330002,.00135972068684592,.00167123596481511,5.01144631725624e-005,8.26394471580808e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.33205784447439,2.86633790217342,3.22890373692238,3.48570817876623,3.6688487377681,.0319379183519753,.0526660717910194,.0647319954366421,.0729200077195298,.078719555662307,.000465578153648717,.000767744854226105,.000943637046012817,.00106299860251787,5.86464816800813e-006,9.67088643913024e-006,1.18865097700383e-005,6.8546484667622e-008,1.13034107101112e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.43216369764802,5.8898738207944,7.53368983291028,8.83179793566707,9.92198872770369,5.37669458746301,9.22684798551467,11.8019864239608,13.8355522538466,7.27823584467216,12.4900484198527,15.9759196314393,9.14177937313726,15.6880416449771,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.29802783242322,5.6596856973204,7.23925807117198,8.48663349656903,9.53421743822432,4.96464287844228,8.51973352709427,10.897522055244,12.7752422699513,6.45780747384476,11.0821261849396,14.1750577199246,7.79428043908953,13.3756231834223,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,3.06556135691961,5.26075419844715,6.72898802659509,7.88844031019171,8.86218371468821,4.2894282422279,7.36100994616565,9.41540812070088,11.0377496097416,5.18623489409549,8.90000356218188,11.3839223715851,5.81833621896758,9.98474117367778,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.74931534527504,4.71805015837109,6.03481969066818,7.07466185471787,7.94795303132399,3.45007465931297,5.92061049820158,7.57300486916358,8.87788723203632,3.7410698100085,6.41998199374919,8.21174689798638,3.76406389114815,6.45944172008843,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,2.36915365304765,4.06566157904621,5.20035475021148,6.09641269634165,6.84894949965393,2.56192161158569,4.39646717446522,5.62348550533029,6.5924518774048,2.39387697960251,4.10808883148424,5.25462310508641,2.07554270972356,3.5618011698851,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.94896143440088,3.34457733996838,4.2780217485378,5.01515518762464,5.63422234088262,1.73374759666501,2.97525278020816,3.80562170038934,4.46135726670653,1.33269841309056,2.28701955580991,2.92530816462441,.95054294779362,1.63120949890175,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.51514018159912,2.60010456277178,3.3257726574549,3.89882683546337,4.38009521894869,1.04781596872157,1.79813796430469,2.29998368635187,2.69628716153333,.626153181688297,1.07453010927869,1.3744227476422,.347192252628355,.595810322665643,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,1.09494796295235,1.87902032369395,2.40343965578121,2.81756932674636,3.16536806017738,.547225517982271,.939083778347803,1.20117444444972,1.40814530666032,.236321146004057,.405546428999883,.518731148089159,.0946962283839188,.162506478648407,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.714786270724957,1.22663174436907,1.56897471532452,1.83932016837014,2.06636452850732,.23320173943763,.400193272045958,.511883968508167,.600085200896572,.0657431866024243,.112820689170272,.144308028468726,.0171974308138302,.0295121988599679,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.398540259080393,.683927704293011,.874806379397611,1.02554171289629,1.1521338451431,.0724976248699598,.124411858085772,.15913419863567,.186554147878685,.0113956405123916,.019555851831759,.0250137314673599,.00166206328995312,.00285223664290029,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.166073783576779,.284996205419761,.364536334820714,.427348526518979,.480100121606987,.0125887190510939,.0216032722572035,.0276325703309419,.032393857863254,.000824565926330002,.00141502261880692,.00180994395496465,5.01144631725624e-005,8.60005205963556e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.42690622435996,3.10423606728616,3.63911792233608,4.08832802186718,4.47814541428892,.0319379183519753,.0548080819457638,.0701045730824201,.0821840874209392,.0923288321276096,.000465578153648717,.000798970218451683,.00102195632637702,.00119804663781278,5.86464816800813e-006,1.00642162671112e-005,1.287305739994e-005,6.8546484667622e-008,1.17631378114674e-007,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 144; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[56][144] = 
		{
			{-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505},
			{-1.77695599756377e-005,-4.02808883404862e-005,-6.02967929985932e-005,-7.65271814689107e-005,-8.7949757770287e-005,-9.38463500146473e-005,-9.38463500146473e-005,-8.7949757770287e-005,-7.65271814689107e-005,-6.02967929985932e-005,-4.02808883404862e-005,-1.77695599756377e-005,-.000209456341481459,-.000474805651630584,-.000710740484393843,-.000902054045026543,-.00103669615466006,-.00110620145700885,-.00110620145700885,-.00103669615466006,-.000902054045026543,-.000710740484393843,-.000474805651630584,-.000209456341481459,-.000752419091267727,-.00170561957877169,-.00255315597328018,-.00324040169914516,-.00372406952729629,-.00397374979986132,-.00397374979986132,-.00372406952729629,-.00324040169914516,-.00255315597328018,-.00170561957877169,-.000752419091267727,-.00171271583978708,-.00388246617226489,-.00581169553993682,-.00737605861122963,-.00847702157201008,-.00904536355411474,-.00904536355411474,-.00847702157201008,-.00737605861122963,-.00581169553993682,-.00388246617226489,-.00171271583978708,-.00301523746950065,-.00683508449255492,-.0102314941838372,-.012985556497555,-.0149238025829595,-.0159243690517937,-.0159243690517937,-.0149238025829595,-.012985556497555,-.0102314941838372,-.00683508449255492,-.00301523746950065,-.00445208652099934,-.0100922026364427,-.0151071011176736,-.0191735548643253,-.0220354320327057,-.0235127977574022,-.0235127977574022,-.0220354320327057,-.0191735548643253,-.0151071011176736,-.0100922026364427,-.00445208652099934,-.00572682649263902,-.012981844165618,-.0194326292850773,-.0246634070198705,-.0283447087891094,-.0302450800266409,-.0302450800266409,-.0283447087891094,-.0246634070198705,-.0194326292850773,-.012981844165618,-.00572682649263902,-.00652410991880577,-.0147891643642708,-.022138021752572,-.0280970234695782,-.0322908326268278,-.0344557717700214,-.0344557717700214,-.0322908326268278,-.0280970234695782,-.022138021752572,-.0147891643642708,-.00652410991880577,-.0065876977962188,-.0149333084057324,-.0223537921535752,-.028370873865462,-.0326055583951231,-.034791598330081,-.034791598330081,-.0326055583951231,-.028370873865462,-.0223537921535752,-.0149333084057324,-.0065876977962188,-.00578758817420658,-.0131195816511808,-.0196388096598727,-.0249250859943669,-.0286454464091129,-.0305659806028896,-.0305659806028896,-.0286454464091129,-.0249250859943669,-.0196388096598727,-.0131195816511809,-.00578758817420658,-.00415955383809286,-.00942907556113789,-.0141144607455587,-.0179137205329708,-.0205875527021657,-.0219678453450519,-.0219678453450519,-.0205875527021657,-.0179137205329708,-.0141144607455587,-.0094290755611379,-.00415955383809286,-.00190958089376513,-.00432872929123383,-.00647971047247572,-.00822388645453697,-.00945139763053613,-.0100850666636237,-.0100850666636237,-.00945139763053613,-.00822388645453697,-.00647971047247572,-.00432872929123383,-.00190958089376513},
			{-2.11503243320183e-007,-4.79445666606905e-007,-7.17686160968121e-007,-9.10869324002267e-007,-1.0468272170579e-006,-1.11701175656978e-006,-1.11701175656978e-006,-1.0468272170579e-006,-9.10869324002267e-007,-7.17686160968121e-007,-4.79445666606905e-007,-2.11503243320183e-007,-1.29636812850387e-005,-2.93866926947107e-005,-4.39891819502224e-005,-5.58299694289237e-005,-6.41632449195998e-005,-6.84650702111951e-005,-6.84650702111951e-005,-6.41632449195998e-005,-5.58299694289237e-005,-4.39891819502224e-005,-2.93866926947107e-005,-1.29636812850387e-005,-.000111754689657294,-.00025333087492222,-.000379213069886165,-.000481287743035328,-.000553125564084201,-.000590209872148528,-.000590209872148528,-.000553125564084201,-.000481287743035328,-.000379213069886165,-.00025333087492222,-.000111754689657294,-.000456242005972593,-.0010342311978472,-.00154814918484807,-.00196487222152227,-.00225815236645889,-.00240955021073066,-.00240955021073066,-.00225815236645889,-.00196487222152227,-.00154814918484807,-.0010342311978472,-.000456242005972593,-.00123040688572486,-.0027891451698972,-.00417509434078899,-.00529892530561013,-.00608985184251931,-.00649814601017004,-.00649814601017004,-.00608985184251931,-.00529892530561013,-.00417509434078899,-.0027891451698972,-.00123040688572486,-.00251391253350483,-.00569865715294523,-.00853036675399556,-.0108265281139349,-.0124425139778682,-.0132767224314477,-.0132767224314477,-.0124425139778682,-.0108265281139349,-.00853036675399556,-.00569865715294523,-.00251391253350483,-.00415959476024,-.00942916832541804,-.0141145996051725,-.0179138967701193,-.0205877552447685,-.0219680614671252,-.0219680614671252,-.0205877552447685,-.0179138967701193,-.0141145996051725,-.00942916832541803,-.00415959476024,-.00576034173845761,-.0130578181276279,-.0195463553335665,-.0248077452759809,-.0285105912170028,-.030422084043298,-.030422084043298,-.0285105912170028,-.0248077452759809,-.0195463553335665,-.0130578181276279,-.00576034173845761,-.00674981664852096,-.0153008071730865,-.0229039040804791,-.0290690621631277,-.0334079594567994,-.0356477963776399,-.0356477963776399,-.0334079594567994,-.0290690621631277,-.0229039040804791,-.0153008071730865,-.00674981664852096,-.006612130023256,-.0149886925463945,-.0224366971291726,-.0284760947867864,-.0327264847688013,-.0349206322134925,-.0349206322134925,-.0327264847688013,-.0284760947867864,-.0224366971291726,-.0149886925463945,-.006612130023256,-.00511251762065263,-.0115892994367119,-.0173481176290686,-.0220177969659965,-.0253042105122013,-.02700073151736,-.02700073151736,-.0253042105122013,-.0220177969659965,-.0173481176290686,-.0115892994367119,-.00511251762065263,-.00244252943287481,-.005536842526707,-.00828814510932977,-.01051910646122,-.0120892060502712,-.0128997269708937,-.0128997269708937,-.0120892060502712,-.01051910646122,-.00828814510932977,-.005536842526707,-.00244252943287481},
			{-2.30726263295083e-009,-5.23021327582048e-009,-7.82914926217426e-009,-9.9365604128826e-009,-1.14197081952885e-008,-1.21853426266313e-008,-1.21853426266313e-008,-1.14197081952885e-008,-9.9365604128826e-009,-7.82914926217426e-009,-5.23021327582048e-009,-2.30726263295083e-009,-7.35364770593756e-007,-1.66696002908487e-006,-2.49528617544501e-006,-3.16695479923119e-006,-3.63965981910605e-006,-3.88368084208039e-006,-3.88368084208039e-006,-3.63965981910605e-006,-3.16695479923119e-006,-2.49528617544501e-006,-1.66696002908487e-006,-7.35364770593756e-007,-1.52128768774711e-005,-3.44852768261633e-005,-5.16212944637669e-005,-6.55164557289296e-005,-7.52955525177414e-005,-8.03437434652432e-005,-8.03437434652432e-005,-7.52955525177414e-005,-6.55164557289296e-005,-5.16212944637669e-005,-3.44852768261633e-005,-1.52128768774711e-005,-.000111389656582294,-.000252503400490955,-.000377974416605436,-.000479715675273885,-.000551318846834677,-.000588282023525919,-.000588282023525919,-.000551318846834677,-.000479715675273885,-.000377974416605436,-.000252503400490955,-.000111389656582294,-.000460167166116478,-.00104312893857298,-.00156146828610925,-.00198177649169201,-.00227757979653243,-.00243028015301575,-.00243028015301575,-.00227757979653243,-.00198177649169201,-.00156146828610925,-.00104312893857298,-.000460167166116478,-.00130099723581861,-.00294916275130836,-.00441462597427551,-.00560293286342162,-.0064392360816438,-.00687095471852507,-.00687095471852507,-.0064392360816438,-.00560293286342162,-.00441462597427551,-.00294916275130836,-.00130099723581861,-.00276903001105212,-.00627696965144704,-.00939604748863851,-.011925228449054,-.0137052081798638,-.0146240739767634,-.0146240739767634,-.0137052081798638,-.011925228449054,-.00939604748863851,-.00627696965144704,-.00276903001105212,-.00466138390221013,-.0105666479493299,-.0158173022080392,-.0200749243239431,-.0230713414195831,-.0246181596977759,-.0246181596977759,-.0230713414195831,-.0200749243239431,-.0158173022080392,-.0105666479493299,-.00466138390221013,-.00633855009047991,-.0143685284714563,-.0215083684256039,-.0272978832165217,-.0313724113504692,-.0334757749315177,-.0334757749315177,-.0313724113504692,-.0272978832165217,-.0215083684256039,-.0143685284714563,-.00633855009047991,-.00692348559220745,-.0156944882399467,-.0234932085068186,-.0298169926007543,-.0342675272542232,-.036564993905222,-.036564993905222,-.0342675272542232,-.0298169926007543,-.0234932085068186,-.0156944882399467,-.00692348559220746,-.00575920488574972,-.0130552410555424,-.0195424976931696,-.0248028492552804,-.0285049644079869,-.0304160799848247,-.0304160799848247,-.0285049644079869,-.0248028492552804,-.0195424976931696,-.0130552410555424,-.00575920488574972,-.00286339454936457,-.00649087969965739,-.0097162511988505,-.0123316229887931,-.014172261854691,-.0151224412691208,-.0151224412691208,-.014172261854691,-.0123316229887931,-.0097162511988505,-.00649087969965739,-.00286339454936457},
			{-2.41204413706157e-011,-5.46773700026945e-011,-8.18470047853013e-011,-1.03878171232717e-010,-1.19383202440944e-010,-1.27387250245815e-010,-1.27387250245815e-010,-1.19383202440944e-010,-1.03878171232717e-010,-8.18470047853013e-011,-5.46773700026945e-011,-2.41204413706157e-011,-3.99747274956451e-008,-9.06166239851226e-008,-1.35644769610767e-007,-1.72156948704618e-007,-1.97853385508442e-007,-2.11118467392541e-007,-2.11118467392541e-007,-1.97853385508442e-007,-1.72156948704618e-007,-1.35644769610767e-007,-9.06166239851226e-008,-3.99747274956451e-008,-1.98456423704136e-006,-4.49870511967568e-006,-6.73414868783207e-006,-8.54681307319903e-006,-9.82252482147343e-006,-1.04810760801775e-005,-1.04810760801775e-005,-9.82252482147343e-006,-8.54681307319903e-006,-6.73414868783207e-006,-4.49870511967567e-006,-1.98456423704136e-006,-2.60616962562586e-005,-5.90778994134517e-005,-8.84341934471251e-005,-.00011223846631682,-.000128991369283385,-.000137639596714465,-.000137639596714465,-.000128991369283385,-.00011223846631682,-8.84341934471251e-005,-5.90778994134517e-005,-2.60616962562586e-005,-.000164926605038453,-.000373863515530918,-.00055963937078906,-.000710280291134566,-.000816297926504351,-.000871026627805438,-.000871026627805438,-.000816297926504351,-.000710280291134566,-.00055963937078906,-.000373863515530918,-.000164926605038453,-.00064522442811524,-.00146262559000318,-.0021894162732808,-.00277875237013443,-.00319351364006811,-.00340762278874215,-.00340762278874215,-.00319351364006811,-.00277875237013443,-.0021894162732808,-.00146262559000318,-.00064522442811524,-.00176649543659496,-.00400437633419262,-.00599418386382994,-.00760766822733064,-.00874320163042015,-.00932938965676434,-.00932938965676434,-.00874320163042015,-.00760766822733064,-.00599418386382994,-.00400437633419262,-.00176649543659496,-.00361484526436693,-.00819430411680041,-.0122661212166373,-.0155678542354177,-.0178915384406828,-.0190910767848879,-.0190910767848879,-.0178915384406828,-.0155678542354177,-.0122661212166373,-.00819430411680041,-.00361484526436693,-.00570421773357935,-.0129305935493704,-.0193559118161673,-.0245660391272654,-.028232807600475,-.0301256764218797,-.0301256764218797,-.028232807600475,-.0245660391272654,-.0193559118161673,-.0129305935493704,-.00570421773357935,-.00694730582706703,-.0157484850874735,-.0235740368319136,-.0299195778892043,-.0343854245960146,-.0366907956752786,-.0366907956752786,-.0343854245960146,-.0299195778892043,-.0235740368319136,-.0157484850874735,-.00694730582706703,-.00621725180422221,-.01409356371536,-.0210967714210865,-.0267754945936724,-.0307720501198624,-.03283516247719,-.03283516247719,-.0307720501198624,-.0267754945936724,-.0210967714210865,-.01409356371536,-.00621725180422221,-.00321684991523739,-.00729210922619538,-.0109156182658738,-.0138538296704664,-.0159216756755027,-.0169891445541618,-.0169891445541618,-.0159216756755027,-.0138538296704664,-.0109156182658738,-.00729210922619538,-.00321684991523739},
			{-2.45853458176161e-013,-5.57312376361246e-013,-8.34245479120496e-013,-1.05880349510068e-012,-1.21684229228106e-012,-1.29842549393142e-012,-1.29842549393142e-012,-1.21684229228106e-012,-1.05880349510068e-012,-8.34245479120496e-013,-5.57312376361246e-013,-2.45853458176161e-013,-2.11870775687607e-009,-4.80278806553758e-009,-7.18933295005722e-009,-9.12452155328201e-009,-1.04864630445972e-008,-1.11895280470165e-008,-1.11895280470165e-008,-1.04864630445972e-008,-9.12452155328201e-009,-7.18933295005722e-009,-4.80278806553757e-009,-2.11870775687607e-009,-2.52418873530302e-007,-5.72195173861672e-007,-8.56523661084552e-007,-1.08707840640484e-006,-1.24933756458098e-006,-1.33309941203413e-006,-1.33309941203413e-006,-1.24933756458098e-006,-1.08707840640484e-006,-8.56523661084552e-007,-5.72195173861672e-007,-2.52418873530302e-007,-5.94515724243734e-006,-1.34767667504191e-005,-2.01734829721616e-005,-2.56036799885296e-005,-2.94253284884657e-005,-3.1398149882767e-005,-3.1398149882767e-005,-2.94253284884657e-005,-2.56036799885296e-005,-2.01734829721616e-005,-1.34767667504191e-005,-5.94515724243734e-006,-5.76326598948289e-005,-.000130644469596975,-.000195562780868023,-.000248203389861898,-.000285250646856585,-.000304375278857807,-.000304375278857807,-.000285250646856585,-.000248203389861898,-.000195562780868023,-.000130644469596975,-5.7632659894829e-005,-.000311995321652138,-.000707245915568759,-.00105868222690829,-.00134365300155203,-.00154420891695623,-.00164774041669247,-.00164774041669247,-.00154420891695623,-.00134365300155203,-.00105868222690829,-.000707245915568759,-.000311995321652138,-.00109875339493098,-.00249070674094487,-.00372835940234233,-.00473194049592369,-.00543823792293946,-.00580284462990898,-.00580284462990898,-.00543823792293946,-.00473194049592369,-.00372835940234233,-.00249070674094487,-.00109875339493098,-.00273317477899627,-.00619569129673009,-.00927438125108664,-.0117708127035932,-.0135277440795465,-.0144347117943581,-.0144347117943581,-.0135277440795465,-.0117708127035932,-.00927438125108664,-.00619569129673009,-.00273317477899627,-.00500501190886783,-.0113456003480285,-.0169833224592084,-.0215548080610388,-.0247721151016589,-.0264329617655731,-.0264329617655731,-.0247721151016589,-.0215548080610388,-.0169833224592084,-.0113456003480285,-.00500501190886783,-.00679690022731426,-.0154075384811573,-.0230636710532136,-.0292718344086679,-.0336409978300346,-.0358964588105553,-.0358964588105553,-.0336409978300346,-.0292718344086679,-.0230636710532136,-.0154075384811573,-.00679690022731426,-.00654390879914259,-.0148340454130535,-.0222052045635651,-.0281822902128456,-.0323888264281449,-.0345603355665664,-.0345603355665664,-.0323888264281449,-.0281822902128456,-.0222052045635651,-.0148340454130535,-.00654390879914259,-.00352357286688459,-.00798740348129818,-.0119564099539485,-.0151747764476273,-.0174397891986819,-.0186090400111835,-.0186090400111835,-.0174397891986819,-.0151747764476273,-.0119564099539485,-.00798740348129818,-.00352357286688459},
			{-2.51299527189796e-005,-.00029621599889564,-.00106408128362993,-.00242214596950661,-.00426418972382729,-.00629620113986558,-.00809895569654666,-.00922648473106781,-.00931641156972225,-.00818488569070588,-.00588249745219549,-.00270055519884434,-5.69657786046887e-005,-.000671476592135066,-.00241211034093489,-.00549063631715242,-.00966626919088759,-.014272529844964,-.0183591000865756,-.0209150368234713,-.0211188872818742,-.0185538903067371,-.0133347265413407,-.00612174767248644,-8.52725424398906e-005,-.00100513883251861,-.00361070780484583,-.00821897865428013,-.0144695178404447,-.0213646672921825,-.0274818878919328,-.0313078906116195,-.0316130360391258,-.0277734709743103,-.0199608618151533,-.00916369443189591,-.000108225777940879,-.00127569706467466,-.00458262003120292,-.0104313221265329,-.0183643501167498,-.0271155013323826,-.034879324707423,-.0397351916593658,-.0401224746033481,-.035249394662203,-.025333826534352,-.011630331761288,-.000124379740266118,-.00146610976221544,-.00526662963356749,-.0119883188775888,-.0211054440183851,-.0311628068384012,-.0400854715975038,-.045666133448504,-.0461112228985269,-.0405107788184967,-.0291151962521434,-.013366294714629,-.000132718780991214,-.00156440510346971,-.00561973086134261,-.0127920758168761,-.0225204586888932,-.0332521174831873,-.0427730023755956,-.0487278197470118,-.0492027502229294,-.0432268243227726,-.0310672248280699,-.014262438055421,-.000132718780991214,-.00156440510346971,-.00561973086134261,-.0127920758168761,-.0225204586888932,-.0332521174831873,-.0427730023755956,-.0487278197470118,-.0492027502229294,-.0432268243227726,-.0310672248280699,-.014262438055421,-.000124379740266118,-.00146610976221544,-.00526662963356749,-.0119883188775888,-.0211054440183851,-.0311628068384012,-.0400854715975038,-.045666133448504,-.0461112228985269,-.0405107788184967,-.0291151962521434,-.013366294714629,-.000108225777940879,-.00127569706467466,-.00458262003120292,-.0104313221265329,-.0183643501167498,-.0271155013323826,-.034879324707423,-.0397351916593658,-.0401224746033481,-.035249394662203,-.025333826534352,-.011630331761288,-8.52725424398906e-005,-.00100513883251861,-.00361070780484583,-.00821897865428013,-.0144695178404447,-.0213646672921825,-.0274818878919328,-.0313078906116195,-.0316130360391258,-.0277734709743103,-.0199608618151533,-.00916369443189591,-5.69657786046887e-005,-.000671476592135066,-.00241211034093489,-.00549063631715242,-.00966626919088758,-.014272529844964,-.0183591000865756,-.0209150368234713,-.0211188872818742,-.0185538903067371,-.0133347265413407,-.00612174767248644,-2.51299527189796e-005,-.00029621599889564,-.00106408128362993,-.00242214596950661,-.00426418972382729,-.00629620113986558,-.00809895569654666,-.00922648473106781,-.00931641156972225,-.00818488569070588,-.00588249745219549,-.00270055519884434},
			{-4.01299188864111e-007,-4.73026119128408e-006,-1.69922705697598e-005,-3.86791500860789e-005,-6.80946715845654e-005,-.000100543778916268,-.000129331892821903,-.000147337357934948,-.000148773395949532,-.000130704105389941,-9.39373616201976e-005,-4.31250557013781e-005,-4.73026119128408e-006,-5.5757329091799e-005,-.000200294145262156,-.000455925373489819,-.000802656948401598,-.00118514651569285,-.00152448260646228,-.00173671964859619,-.00175364675704319,-.00154065738088721,-.00110727424428985,-.000508330898783026,-1.69922705697598e-005,-.000200294145262156,-.000719506211645245,-.00163779694031119,-.00288334269301423,-.00425734001677702,-.00547631935750839,-.00623872741470196,-.00629953378353948,-.00553442315607657,-.00397760351765944,-.00182605057560164,-3.86791500860789e-005,-.000455925373489819,-.00163779694031119,-.00372808291892168,-.0065632926638522,-.00969089403328137,-.0124656312103062,-.0142010847243002,-.0143394969898535,-.0125978916716827,-.0090541356913339,-.0041566007314033,-6.80946715845654e-005,-.000802656948401598,-.00288334269301423,-.0065632926638522,-.0115546814618157,-.0170608259252993,-.0219457527239106,-.0250010198852289,-.0252446947784246,-.0221785973614222,-.015939812405686,-.00731769858135901,-.000100543778916268,-.00118514651569285,-.00425734001677702,-.00969089403328137,-.0170608259252993,-.0251908096484753,-.0324035472773186,-.0369147388118314,-.0372745319354167,-.0327473492108772,-.0235356003221748,-.0108047964872896,-.000129331892821903,-.00152448260646228,-.00547631935750839,-.0124656312103062,-.0219457527239106,-.0324035472773186,-.0416814660110368,-.0474843207110434,-.0479471313016055,-.0421237067473831,-.0302744115168138,-.0138984708583546,-.000147337357934948,-.00173671964859619,-.00623872741470196,-.0142010847243002,-.0250010198852289,-.0369147388118314,-.0474843207110434,-.0540950434131128,-.0546222860610777,-.0479881297889341,-.0344891867628131,-.0158334029675552,-.000148773395949532,-.00175364675704319,-.00629953378353948,-.0143394969898535,-.0252446947784246,-.0372745319354167,-.0479471313016055,-.0546222860610777,-.0551546675312396,-.0484558508041033,-.0348253390053783,-.0159877247830154,-.000130704105389941,-.00154065738088721,-.00553442315607657,-.0125978916716827,-.0221785973614222,-.0327473492108772,-.0421237067473831,-.0479881297889341,-.0484558508041033,-.0425706396619946,-.0305956233004421,-.0140459337615274,-9.39373616201976e-005,-.00110727424428985,-.00397760351765944,-.0090541356913339,-.015939812405686,-.0235356003221748,-.0302744115168138,-.0344891867628131,-.0348253390053783,-.0305956233004421,-.0219891496246005,-.0100948470984408,-4.31250557013781e-005,-.000508330898783026,-.00182605057560164,-.0041566007314033,-.00731769858135901,-.0108047964872896,-.0138984708583546,-.0158334029675552,-.0159877247830154,-.0140459337615274,-.0100948470984408,-.00463437375617701},
			{-4.77648743752639e-009,-5.63022148645256e-008,-2.02251510004898e-007,-4.60380881913306e-007,-8.10500874190154e-007,-1.19672830207889e-006,-1.53938053820615e-006,-1.75369165645863e-006,-1.77078418424541e-006,-1.55571337982341e-006,-1.11809502771057e-006,-5.13298537640478e-007,-2.92765537918228e-007,-3.45093511422209e-006,-1.23966142266297e-005,-2.82181536753749e-005,-4.96780798691598e-005,-7.33511413319109e-005,-9.43533458892899e-005,-.000107489130425036,-.00010853678378063,-9.53544357538297e-005,-6.85314672157709e-005,-3.14616387984846e-005,-2.52381411676342e-006,-2.97491255946283e-005,-.000106866232302167,-.00024325736936566,-.000428254773970375,-.000632330728850647,-.000813382299066098,-.000926620621724616,-.000935652020526277,-.000822012292713028,-.000590782254057447,-.000271218151906612,-1.03035498453369e-005,-.000121451733066827,-.000436284726357267,-.000993105797236321,-.0017483634713027,-.00258150992189795,-.00332065860400601,-.0037829575879782,-.0038198285552606,-.00335589082222505,-.00241188697772547,-.00110725656401101,-2.7786916836135e-005,-.000327534612535476,-.00117658550597994,-.00267823697769347,-.00471503812818561,-.00696189202635823,-.00895524997265581,-.0102019914951301,-.0103014262061688,-.00905026525693164,-.00650444787225885,-.00298608212919577,-5.67729905547983e-005,-.000669204128457476,-.00240394708818667,-.00547205447566498,-.00963355584556567,-.0142242276315438,-.0182969677820565,-.0208442545176493,-.021047415089364,-.0184910987778976,-.0132895981153155,-.00610102997451746,-9.39382857945563e-005,-.00110728513787305,-.00397764265011023,-.00905422476771312,-.0159399692247158,-.0235358318700697,-.0302747093624672,-.0344895260742596,-.0348256816239572,-.0305959243062406,-.0219893659582073,-.0100949464135448,-.000130088785011913,-.0015334043731933,-.00550836855481976,-.0125385841278022,-.0220741863877053,-.0325931833471828,-.0419253994709405,-.0477622143590056,-.0482277334672722,-.0423702283435318,-.0304515871935496,-.0139798092182615,-.000152434610085198,-.00179680283515025,-.00645456111210986,-.0146923824553127,-.0258659498945255,-.0381918333275861,-.0491270782522392,-.0559665041221956,-.0565119871455732,-.0496483170026962,-.0356823673916168,-.0163811720361289,-.000149325161617318,-.00176015062198346,-.00632289727835408,-.0143926788244271,-.0253383214364947,-.0374127744408963,-.0481249559775311,-.0548248673220092,-.0553592232703062,-.04863556220147,-.0349544980281293,-.016047019508292,-.000115458636973295,-.001360953435314,-.0048888820448881,-.0111284599424805,-.0195916617438277,-.0289276629306809,-.0372103519687145,-.0423907423550166,-.0428039078843937,-.0376051541440983,-.027026916660946,-.0124076008346097,-5.51609089712794e-005,-.000650201929690796,-.00233568648062049,-.00531667428240711,-.00936000890342516,-.0138203275519371,-.0177774213479609,-.0202523773150973,-.0204497690975875,-.0179660399513751,-.0129122370468795,-.00592778988329973},
			{-5.21061086714588e-011,-6.14193874587905e-010,-2.20633662228147e-009,-5.02223790536332e-009,-8.84165344957659e-009,-1.30549605277777e-008,-1.67929112469387e-008,-1.91308046389391e-008,-1.9317264903302e-008,-1.6971084189159e-008,-1.21971598964522e-008,-5.59951004436151e-009,-1.66071240016209e-008,-1.95754280954478e-007,-7.03197893870279e-007,-1.60067465766628e-006,-2.81798888768115e-006,-4.16084320723466e-006,-5.35219317920171e-006,-6.09732050598647e-006,-6.15674863851456e-006,-5.40897999788056e-006,-3.88744721171111e-006,-1.78466133868042e-006,-3.43560288483173e-007,-4.04967152831355e-006,-1.45474238197567e-005,-3.31139965656822e-005,-5.82972148036964e-005,-8.60775467486887e-005,-.000110723628756234,-.000126138468756338,-.00012736788971769,-.000111898407472009,-8.04216603301225e-005,-3.69202255792134e-005,-2.51557038538368e-006,-2.96519536996954e-005,-.000106517166772019,-.00024246279880047,-.000426855931918081,-.000630265297550785,-.000810725484866511,-.000923593928337934,-.000932595827175548,-.000819327289690436,-.000588852535789363,-.00027033225085123,-1.03921937720791e-005,-.000122496611646568,-.000440038189183426,-.00100164972615974,-.00176340506432601,-.00260371927496907,-.00334922703163026,-.00381550328536708,-.00385269146248218,-.00338476236112819,-.0024326370236585,-.0011167825595386,-2.93810953216598e-005,-.000346325780898917,-.0012440880400349,-.00283189158407368,-.00498554717338517,-.00736130656207001,-.00946902654323017,-.0107872955591597,-.0108924349936834,-.00956949300162804,-.0068776181278623,-.00315739828904347,-6.253444086072e-005,-.000737116463054399,-.00264790502577821,-.00602737082635835,-.0106111906809188,-.0156677341271221,-.0201537850748433,-.0229595761766608,-.0231833539384622,-.0203676169191767,-.0146382563131457,-.00672017616832581,-.00010527045022929,-.00124086152956873,-.00445748215529598,-.010146473397001,-.0178628737232513,-.026375056735671,-.0339268729271671,-.0386501404333965,-.039026847819149,-.0342868373598303,-.0246420342366014,-.0113127416048292,-.000143146764099252,-.00168732357718739,-.00606128448364179,-.0137971750917403,-.0242899366861979,-.0358648036217386,-.0461337637005683,-.0525564631192487,-.0530687098434273,-.0466232433562013,-.0335082395307456,-.0153830666658602,-.000156356665903898,-.00184303355014811,-.00662063329842476,-.0150704090994405,-.0265314660737799,-.039174487477097,-.0503910900341255,-.0574064904417204,-.0579660084330403,-.0509257399611718,-.0366004544098804,-.0168026502756542,-.000130063110871032,-.0015331017426483,-.00550728143089673,-.0125361095303546,-.0220698298571165,-.0325867508020486,-.0419171251326689,-.0477527880751015,-.0482182153091671,-.0423618662144552,-.0304455773108406,-.0139770501826433,-6.46655241703613e-005,-.000762236326125946,-.0027381417997588,-.00623277490756985,-.0109728047138306,-.0162016678481028,-.0208405969245687,-.0237420053295095,-.0239734091138143,-.0210617158489217,-.015137106919017,-.00694919005368325},
			{-5.44724437242996e-013,-6.4208673651397e-012,-2.30653469542085e-011,-5.25031668349854e-011,-9.24318630275365e-011,-1.36478355571743e-010,-1.75555407262082e-010,-1.99996066807805e-010,-2.01945348094836e-010,-1.77418051741924e-010,-1.27510789616824e-010,-5.85380492906022e-011,-9.02770003435125e-010,-1.06412821914539e-008,-3.82261230182231e-008,-8.70133242853409e-008,-1.53187020074262e-007,-2.26185126101403e-007,-2.9094739428102e-007,-3.31452818296366e-007,-3.34683355709183e-007,-2.94034348800585e-007,-2.11323208782435e-007,-9.70149149662404e-008,-4.48184434349486e-008,-5.28291483055767e-007,-1.89775394143646e-006,-4.31981760329904e-006,-7.60504200188573e-006,-1.12290674717028e-005,-1.44442208796392e-005,-1.64551317961879e-005,-1.6615513352675e-005,-1.45974741955391e-005,-1.04912405632317e-005,-4.81635129901971e-006,-5.88564803133452e-007,-6.93762988830925e-006,-2.49216860143408e-005,-5.67287126102093e-005,-9.98709394081982e-005,-.000147462369938119,-.00018968441040975,-.0002160920965543,-.000218198259374459,-.000191696963742466,-.000137773078747959,-6.3249292855151e-005,-3.7246230587395e-006,-4.39035024137147e-005,-.000157712261925103,-.000358997123096203,-.000632014693772654,-.000933188224038868,-.00120038256642976,-.00136749870422512,-.00138082716451303,-.00121311863645989,-.000871871344066475,-.000400261404288981,-1.45714378978438e-005,-.000171758899848453,-.000617000537806862,-.00140446541897081,-.00247256238164128,-.00365081084424778,-.00469612622392589,-.00534991652302709,-.00540206000930644,-.00474595216619707,-.00341092748034312,-.00156589918054001,-3.98936826157651e-005,-.000470241515286636,-.00168922406982674,-.00384514541818346,-.00676938127811489,-.00999518991411039,-.0128570543562206,-.0146470014343651,-.0147897598708826,-.0129934678207622,-.00933843724141566,-.00428711876993237,-8.16359254002543e-005,-.000962272689432589,-.00345672199472992,-.00786846397549439,-.01385243649599,-.0204535285962103,-.0263098681663614,-.0299727033964202,-.0302648352907365,-.0265890161087102,-.0191095911936167,-.00877289047128288,-.000128821307499823,-.00151846414940855,-.0054546995681281,-.0124164185359414,-.0218591137752465,-.0322756223296346,-.0415169130590646,-.0472968588510568,-.0477578423262084,-.0419574078870825,-.0301548921174003,-.0138436015212964,-.000156894610627413,-.00184937450253303,-.00664341157096352,-.015122258804788,-.0266227475173882,-.0393092672046696,-.0505644604551389,-.0576039973305382,-.0581654403421276,-.0511009498439408,-.036726378183032,-.0168604597524974,-.000140407421995964,-.00165503394391524,-.00594529211811034,-.0135331440961878,-.0238251099283638,-.0351784732865376,-.0452509204027591,-.0515507112035252,-.0520531552679814,-.045731033082912,-.0328670058156056,-.0150886871005195,-7.26477900154669e-005,-.000856326230599691,-.00307613605632261,-.00700214416426768,-.0123272798443727,-.0182015900873001,-.0234131452361595,-.0266727014101101,-.0269326695112991,-.0236615589216733,-.0170056205219542,-.00780699308131397},
			{-3.66334363412211e-007,-2.24537546398006e-005,-.000193564800479007,-.000790234334926298,-.00213112724015142,-.00435422423400534,-.00720462946394863,-.00997720456040535,-.0116910253775248,-.0114525451470328,-.00885514027394949,-.00423058507690692,-8.30424254068273e-007,-5.08992448158822e-005,-.000438781946510387,-.00179134098152266,-.00483094114415894,-.00987036372424941,-.0163317986134588,-.0226168044340364,-.0265017754217613,-.0259611770305218,-.0200732554493939,-.00959009257018474,-1.2430688947403e-006,-7.61914981245153e-005,-.000656816303965783,-.00268147304597068,-.00723147552495667,-.0147750286257641,-.0244472036457215,-.0338552805420154,-.0396707255608121,-.0388614993834644,-.0300478211505447,-.0143554884304917,-1.57767194829697e-006,-9.670034364015e-005,-.000833614824033866,-.00340325851820642,-.00917800785523132,-.0187520967637297,-.03102777936875,-.0429682752411084,-.0503490925971213,-.0493220429740223,-.0381359430175126,-.0182196268418573,-1.81315792676965e-006,-.000111134000184102,-.000958041580001006,-.00391123463011006,-.0105479328022725,-.0215510663844978,-.035659038099294,-.0493817925438395,-.057864283158913,-.0566839343751764,-.0438281782544715,-.0209391191031562,-1.93472111511538e-006,-.000118584980154756,-.00102227348573477,-.00417346338855665,-.0112551190451086,-.0229959578102643,-.038049798606524,-.0526925952374313,-.0617437945066473,-.0604843092288456,-.0467666388316428,-.0223429825183335,-1.93472111511538e-006,-.000118584980154756,-.00102227348573477,-.00417346338855665,-.0112551190451086,-.0229959578102643,-.038049798606524,-.0526925952374313,-.0617437945066473,-.0604843092288456,-.0467666388316428,-.0223429825183335,-1.81315792676965e-006,-.000111134000184102,-.000958041580001006,-.00391123463011006,-.0105479328022725,-.0215510663844978,-.035659038099294,-.0493817925438395,-.057864283158913,-.0566839343751764,-.0438281782544715,-.0209391191031562,-1.57767194829697e-006,-9.670034364015e-005,-.000833614824033866,-.00340325851820642,-.00917800785523132,-.0187520967637297,-.03102777936875,-.0429682752411084,-.0503490925971213,-.0493220429740223,-.0381359430175126,-.0182196268418573,-1.2430688947403e-006,-7.61914981245153e-005,-.000656816303965783,-.00268147304597068,-.00723147552495667,-.0147750286257641,-.0244472036457215,-.0338552805420154,-.0396707255608121,-.0388614993834644,-.0300478211505447,-.0143554884304917,-8.30424254068273e-007,-5.08992448158822e-005,-.000438781946510386,-.00179134098152266,-.00483094114415894,-.00987036372424941,-.0163317986134588,-.0226168044340364,-.0265017754217613,-.0259611770305218,-.020073255449394,-.00959009257018474,-3.66334363412211e-007,-2.24537546398006e-005,-.000193564800479007,-.000790234334926298,-.00213112724015142,-.00435422423400534,-.00720462946394863,-.00997720456040535,-.0116910253775248,-.0114525451470328,-.00885514027394949,-.00423058507690693},
			{-5.84997849237861e-009,-3.58563091085661e-007,-3.09102839585264e-006,-1.26192198302079e-005,-3.40318838868514e-005,-6.95324290155738e-005,-.000115050433754236,-.000159325572268953,-.00018669350692448,-.000182885225860587,-.00014140737349095,-6.75580403638936e-005,-6.89558489033362e-008,-4.22651508264995e-006,-3.64350890004165e-005,-.000148747387195263,-.000401146336906179,-.000819604324242625,-.00135614179377857,-.0018780291418385,-.00220062505725282,-.00215573544715162,-.00166682074010411,-.000796331478757963,-2.4770649960978e-007,-1.51826896966895e-005,-.000130883869937062,-.000534337481072711,-.00144101706420306,-.00294422186735428,-.00487159743595245,-.00674634613725228,-.00790519061914258,-.00774393601400238,-.00598763321131698,-.00286062003831949,-5.63849124010232e-007,-3.45600389940658e-005,-.000297928215558894,-.00121630123191481,-.0032801570028024,-.00670187065504814,-.0110891153486868,-.0153565666050425,-.0179944200606735,-.0176273595758088,-.013629524241044,-.00651156956023967,-9.92656788923096e-007,-6.0842973540354e-005,-.000524502838069343,-.00214129919480704,-.0057747187659129,-.0117986481151247,-.0195223955581165,-.0270352465684925,-.0316791894770017,-.031032979229028,-.0239947872427992,-.0114636229006531,-1.46568685042074e-006,-8.98364341570052e-005,-.000774443817183375,-.00316169103729199,-.00852654155446687,-.0174210498412452,-.0288253893768114,-.0399183341468002,-.0467752519970179,-.0458211036210322,-.0354290068157069,-.0169263752901921,-1.88534841928859e-006,-.00011555877647658,-.00099618579926221,-.00406695959489997,-.0109678964760433,-.0224091174530978,-.0370787950245641,-.0513479179830939,-.0601681371359018,-.0589407930194454,-.045573187736373,-.0217728056224891,-2.1478248622506e-006,-.000131646761218445,-.00113487385418326,-.00463315790456963,-.012494836761645,-.0255288938184798,-.042240870176189,-.0584965270825159,-.0685446888933965,-.0671464750773758,-.0519178442938048,-.0248039952501605,-2.16875884799657e-006,-.000132929869292705,-.00114593501354709,-.004678315432652,-.012616618914028,-.0257777136867565,-.042652574961671,-.0590666692229003,-.0692127664287096,-.0678009247845504,-.0524238666569454,-.0250457498234202,-1.90535198329458e-006,-.000116784856153978,-.00100675533972143,-.0041101100734715,-.0110842659581639,-.0226468783946324,-.0374722013795649,-.0518927198634471,-.0608065216222948,-.0595661553735154,-.0460567196758888,-.0220038152896698,-1.36938115091738e-006,-8.39335630014821e-005,-.000723557535766244,-.00295394620634634,-.00796628917278541,-.0162763671345931,-.0269313631824755,-.0372954252410452,-.0437017964620066,-.042810342192032,-.0331010775700317,-.015814196101363,-6.28659751468171e-007,-3.85324807640305e-005,-.00033217304057603,-.00135610679808757,-.00365718877328773,-.00747220517149892,-.0123637338469658,-.0171216996431039,-.0200627564386384,-.0196535048439074,-.0151961454884669,-.00726003025826054},
			{-6.96297165792815e-011,-4.26781849550599e-009,-3.67911491336473e-008,-1.50201013794113e-007,-4.05066519951766e-007,-8.27613867594724e-007,-1.36939462342768e-006,-1.89638209018582e-006,-2.22213055847631e-006,-2.17680226684608e-006,-1.68310966462922e-006,-8.04113589360719e-007,-4.26781849550599e-009,-2.61587661208472e-007,-2.25504216385422e-006,-9.20627996502583e-006,-2.48277670898173e-005,-5.07269876251226e-005,-8.39343887728836e-005,-.00011623506394466,-.000136201184821656,-.000133422875058339,-.000103162944064127,-4.92865836277514e-005,-3.67911491336473e-008,-2.25504216385422e-006,-1.94398127850062e-005,-7.93636419908757e-005,-.000214030208318087,-.000437296986453802,-.000723564654409836,-.00100201580190209,-.00117413571083906,-.001150185018243,-.000889326306666056,-.000424879842112779,-1.50201013794113e-007,-9.20627996502583e-006,-7.93636419908757e-005,-.000324004543650437,-.000873785000712069,-.00178527858577817,-.00295397526843083,-.00409076076250558,-.00479344565888028,-.00469566620937377,-.0036307023836027,-.00173458520673544,-4.05066519951766e-007,-2.48277670898173e-005,-.000214030208318087,-.000873785000712069,-.00235644913761801,-.00481459189667545,-.00796636754827121,-.0110320841661877,-.0129271055006463,-.0126634109999641,-.0097913851734879,-.00467788049829846,-8.27613867594724e-007,-5.07269876251226e-005,-.000437296986453802,-.00178527858577817,-.00481459189667545,-.0098369596701605,-.016276527267894,-.0225402628819985,-.0264120860481604,-.025873317192124,-.0200053219740424,-.00955763703157627,-1.36939462342768e-006,-8.39343887728836e-005,-.000723564654409836,-.00295397526843083,-.00796636754827121,-.016276527267894,-.0269316281438179,-.037295792168106,-.0437022264174669,-.0428107633770146,-.0331014032314506,-.0158143516876436,-1.89638209018582e-006,-.00011623506394466,-.00100201580190209,-.00409076076250558,-.0110320841661877,-.0225402628819985,-.037295792168106,-.0516484226656698,-.0605202606038321,-.0592857336712341,-.0458398968231796,-.0219002271480226,-2.22213055847631e-006,-.000136201184821656,-.00117413571083906,-.00479344565888028,-.0129271055006463,-.0264120860481604,-.0437022264174669,-.0605202606038321,-.0709160465028163,-.0694694603763257,-.0537139830919874,-.0256621090417623,-2.17680226684608e-006,-.000133422875058339,-.001150185018243,-.00469566620937377,-.0126634109999641,-.025873317192124,-.0428107633770146,-.0592857336712341,-.0694694603763257,-.0680523825420278,-.0526182945056766,-.0251386386461752,-1.68310966462922e-006,-.000103162944064127,-.000889326306666056,-.0036307023836027,-.0097913851734879,-.0200053219740424,-.0331014032314506,-.0458398968231796,-.0537139830919874,-.0526182945056766,-.0406846140173892,-.0194372664460252,-8.04113589360719e-007,-4.92865836277514e-005,-.000424879842112779,-.00173458520673544,-.00467788049829846,-.00955763703157627,-.0158143516876436,-.0219002271480226,-.0256621090417623,-.0251386386461752,-.0194372664460252,-.00928624582089677},
			{-7.59581936873143e-013,-4.65571023163414e-011,-4.01350079989265e-010,-1.63852421901125e-009,-4.41882039598848e-009,-9.02833697182859e-009,-1.49385560003323e-008,-2.06873968742175e-008,-2.42409464882796e-008,-2.37464657802842e-008,-1.83608344516711e-008,-8.77197535304017e-009,-2.42091987612831e-010,-1.48385590679724e-008,-1.2791725800267e-007,-5.22226195326764e-007,-1.40835499192186e-006,-2.8774881763854e-006,-4.76117787775939e-006,-6.59343355166069e-006,-7.72601168100707e-006,-7.56841207045341e-006,-5.85191760211322e-006,-2.79578126521843e-006,-5.00828398075496e-009,-3.06973057681128e-007,-2.64629143836613e-006,-1.08035673306481e-005,-2.91353787244659e-005,-5.95281078102025e-005,-9.84969850916432e-005,-.000136401819657762,-.000159832057717645,-.000156571711876826,-.000121061690113565,-5.7837794064797e-005,-3.66709753307118e-008,-2.24767634356886e-006,-1.93763150067161e-005,-7.91044103306138e-005,-.00021333110473795,-.000435868609164376,-.000721201218685817,-.000998742839454559,-.00117030053969446,-.00114642807911561,-.0008864214307152,-.000423492023911399,-1.51493229393287e-007,-9.28548381512419e-006,-8.00464265713343e-005,-.000326792032995098,-.000881302384115301,-.00180063776864345,-.00297938903112802,-.00412595456537195,-.00483468482962708,-.00473606415990016,-.00366193819311915,-.00174950826221703,-4.28305814057175e-007,-2.62521745709628e-005,-.000226309453117529,-.000923915400575431,-.00249164227715486,-.00509081249643689,-.00842341040243774,-.011665011934605,-.0136687535803171,-.0133899305174026,-.0103531321176025,-.00494625775323145,-9.11601977606186e-007,-5.58748760112742e-005,-.000481674864645602,-.00196645265756991,-.00530318747211388,-.0108352363826531,-.0179283057316278,-.0248276992732258,-.0290924437312574,-.0284989994042999,-.0220355068809891,-.010527567456728,-1.53459036799021e-006,-9.40597416919936e-005,-.000810851255203673,-.00331032553849853,-.00892738345711226,-.0182400321589689,-.0301805019800266,-.0417949380321651,-.0489742069762943,-.0479752030574144,-.0370945625876605,-.0177221024244434,-2.08673606807239e-006,-.00012790244200432,-.00110259558209743,-.00450138085207253,-.0121394565234153,-.0248027967481441,-.0410394482774791,-.0568327590696347,-.0665951293806949,-.0652366837960255,-.0504411882777659,-.0240985158662193,-2.27930471413928e-006,-.000139705563856799,-.00120434555501086,-.00491677828990747,-.0132597125742255,-.027091654003041,-.044826659852073,-.0620774124945382,-.0727406760531766,-.0712568700882474,-.0550960133326791,-.02632238051474,-1.89600782307658e-006,-.000116212123967743,-.00100181804557465,-.00408995341613109,-.011029906890695,-.0225358143697108,-.0372884315268124,-.0516382294108676,-.0605083164164748,-.0592740331282503,-.0458308499302774,-.0218959049522221,-9.42668054679449e-007,-5.77790109816517e-005,-.000498089647453146,-.00203346652033204,-.00548391243189871,-.0112044855690737,-.0185392764637387,-.025673791359598,-.0300838721412506,-.0294702040898467,-.0227864450885946,-.0108863317310921},
			{-4.61452526590166e-009,-1.47072954118751e-006,-3.04257537549422e-005,-.000222779313164588,-.000920334332232957,-.00260199447163722,-.00553806002210424,-.00932276780442027,-.0126771001809598,-.0138469711844149,-.0115184097714994,-.00572678909872914,-1.0460426551641e-008,-3.33392005816974e-006,-6.89705536523267e-005,-.000505006800981909,-.00208625787714596,-.00589832550261673,-.0125539393028941,-.0211332958986598,-.0287370569429126,-.0313889764798934,-.0261104821110847,-.0129817593993148,-1.56582985243485e-008,-4.99057235089002e-006,-.000103242588927534,-.000755948833210871,-.0031229365722185,-.00882925194855102,-.018792094977277,-.0316346044160784,-.0430167368512079,-.0469864170136373,-.0390849953863391,-.019432502397701,-1.98731208257652e-008,-6.33390959846238e-006,-.000131032911457859,-.000959431350547771,-.00396355298338403,-.0112058657268432,-.023850456898108,-.0401498486478861,-.0545957664330435,-.0596339852015086,-.0496056985105609,-.0246632459775863,-2.2839416390577e-008,-7.2793196382121e-006,-.000150591105035483,-.00110263769366935,-.00455515959306486,-.0128784721632876,-.0274104163597275,-.0461426828391663,-.0627448227009384,-.0685350545084463,-.0570099288159739,-.028344523709382,-2.43706852532627e-008,-7.76736168416077e-006,-.000160687486930486,-.00117656404705184,-.00486056030603151,-.0137419094370501,-.0292481479535269,-.0492363193955517,-.0669515498630354,-.073129987810444,-.0608321599696494,-.0302448825382416,-2.43706852532627e-008,-7.76736168416077e-006,-.000160687486930486,-.00117656404705184,-.00486056030603151,-.0137419094370501,-.0292481479535269,-.0492363193955517,-.0669515498630354,-.073129987810444,-.0608321599696494,-.0302448825382416,-2.2839416390577e-008,-7.2793196382121e-006,-.000150591105035483,-.00110263769366935,-.00455515959306486,-.0128784721632876,-.0274104163597275,-.0461426828391663,-.0627448227009384,-.0685350545084463,-.0570099288159739,-.028344523709382,-1.98731208257652e-008,-6.33390959846238e-006,-.000131032911457859,-.000959431350547771,-.00396355298338403,-.0112058657268432,-.023850456898108,-.0401498486478861,-.0545957664330435,-.0596339852015086,-.0496056985105609,-.0246632459775863,-1.56582985243485e-008,-4.99057235089002e-006,-.000103242588927534,-.000755948833210871,-.0031229365722185,-.00882925194855102,-.018792094977277,-.0316346044160784,-.0430167368512079,-.0469864170136373,-.0390849953863391,-.019432502397701,-1.0460426551641e-008,-3.33392005816974e-006,-6.89705536523267e-005,-.000505006800981909,-.00208625787714596,-.00589832550261673,-.0125539393028941,-.0211332958986598,-.0287370569429126,-.0313889764798934,-.0261104821110847,-.0129817593993148,-4.61452526590166e-009,-1.47072954118751e-006,-3.04257537549422e-005,-.000222779313164588,-.000920334332232957,-.00260199447163722,-.00553806002210424,-.00932276780442027,-.0126771001809598,-.0138469711844149,-.0115184097714994,-.00572678909872914},
			{-7.36891655603157e-011,-2.34860199933996e-008,-4.85867619430457e-007,-3.55755375585554e-006,-1.46967813742169e-005,-4.15511434782526e-005,-8.84370543742305e-005,-.000148874898420579,-.000202440095184293,-.000221121717472685,-.0001839370153449,-9.14508612930534e-005,-8.68601307305661e-010,-2.76838358998326e-007,-5.72710039808159e-006,-4.19341950699251e-005,-.000173236369522755,-.000489778616311158,-.00104244009902437,-.00175484320401577,-.00238623588679712,-.00260644304233894,-.00216813327679371,-.00107796495006253,-3.12023117416446e-009,-9.94471998471375e-007,-2.05731640619964e-005,-.000150638021863614,-.000622307975060138,-.00175940617887584,-.00374470319905967,-.00630383171759815,-.00857195072154575,-.00936298940145173,-.00778847209081444,-.00387231726863855,-7.10251695871362e-009,-2.26369580965429e-006,-4.68302630441653e-005,-.000342894178409679,-.0014165466273796,-.00400489948507659,-.00852398956746932,-.0143492802872553,-.0195121521357594,-.0213127769373929,-.0177287361159355,-.00881447480486433,-1.25039862212867e-008,-3.98523810328614e-006,-8.24447118179819e-005,-.000603665448054148,-.00249383135774626,-.00705062842794049,-.0150064897727934,-.0252619182805468,-.0343511578883119,-.0375211591484582,-.0312114527009358,-.0155178892424445,-1.84625022332888e-008,-5.88432089415219e-006,-.000121732034018966,-.00089132973162475,-.00368221511100628,-.01041045957611,-.0221575220926239,-.0372999629412389,-.0507204916900297,-.055401091485105,-.0460846249345851,-.0229126184021882,-2.37487628354249e-008,-7.56914418191853e-006,-.000156586857450866,-.00114653897597652,-.00473652229125437,-.0133912257589351,-.0285017561839274,-.0479798438190263,-.0652429943039245,-.0712637629438524,-.0592797668540085,-.0294730548165372,-2.70550433775344e-008,-8.62291335307587e-006,-.000178386733239242,-.00130615905948616,-.00539593649305364,-.0152555396799836,-.0324697440130155,-.0546595527845674,-.074326062928209,-.0811850373450308,-.0675326405320358,-.03357626593249,-2.7318738012221e-008,-8.7069574240775e-006,-.000180125397036526,-.0013188896669085,-.00544852851748593,-.0154042292942165,-.0327862135586896,-.055192297478506,-.0750504891927709,-.0819763152766897,-.0681908540387014,-.0339035203026146,-2.40007374267426e-008,-7.64945287105235e-006,-.000158248245443382,-.00115870376501852,-.00478677683617009,-.0135333067869547,-.0288041600782291,-.0484889104016337,-.0659352230713762,-.0720198721217481,-.0599087257235696,-.0297857641982319,-1.72493889467441e-008,-5.49768056941236e-006,-.000113733402739168,-.000832763242290672,-.0034402682709391,-.00972642083254007,-.0207016206060418,-.0348490990193348,-.0473878067921877,-.0517608590117147,-.0430565483441448,-.0214071018984119,-7.91890304680574e-009,-2.52389226922161e-006,-5.22130837362085e-005,-.000382307535472911,-.00157936904180661,-.00446523548185483,-.00950376427809313,-.0159986326040109,-.0217549415081703,-.0237625359019177,-.0197665339288291,-.00982763882071387},
			{-8.77089671184217e-013,-2.79543748348947e-010,-5.78306820717772e-009,-4.23439949444061e-008,-1.74929313488118e-007,-4.94564953932593e-007,-1.05262729400979e-006,-1.77199232357178e-006,-2.40955526052686e-006,-2.63191437974773e-006,-2.18932125341295e-006,-1.08849930992072e-006,-5.37595110937315e-011,-1.71340921393568e-008,-3.5446195486468e-007,-2.5953930832331e-006,-1.07219531571797e-005,-3.03134001015112e-005,-6.45187493924991e-005,-.000108610834338575,-.000147689018597559,-.000161318089751033,-.000134190202070998,-6.67174550672731e-005,-4.63439153301422e-010,-1.47706126638825e-007,-3.05567414766182e-006,-2.23738413632665e-005,-9.24296518291858e-005,-.000261319647275767,-.000556190225361124,-.000936290380644839,-.00127316771187585,-.00139065846034338,-.0011567998362318,-.000575144383900619,-1.89200479726517e-009,-6.03014868716438e-007,-1.24748850093275e-005,-9.13419051691941e-005,-.000377346936322897,-.00106684561013735,-.00227066394170372,-.00382243468035404,-.00519774689175974,-.00567740653672345,-.00472267141011844,-.00234804488509053,-5.1024142886996e-009,-1.62622826690742e-006,-3.36426374888076e-005,-.000246333541419145,-.00101764033710382,-.00287710067800549,-.00612359342731638,-.0103084544811151,-.0140174369789623,-.0153109972435626,-.0127362394210907,-.00633227663561512,-1.04250255588739e-008,-3.32263714544313e-006,-6.87371381162648e-005,-.000503297717511878,-.00207919740024719,-.00587836392863835,-.0125114532807165,-.0210617749476202,-.0286398027496958,-.0312827474533958,-.0260221169776175,-.012937825514943,-1.7249558650954e-008,-5.49773465709859e-006,-.000113734521679514,-.000832771435243388,-.00344030211724,-.00972651652373048,-.0207018242742517,-.0348494418743096,-.0473882730064016,-.0517613682492133,-.0430569719462525,-.0214073125073398,-2.388774830107e-008,-7.61344126955836e-006,-.000157503254546897,-.00115324889407038,-.00476424195651651,-.0134695955569983,-.0286685577088058,-.0482606374326126,-.0656248174753565,-.071680821606866,-.0596266912835726,-.0296455406961615,-2.79910339523387e-008,-8.92122984487403e-006,-.000184558163040781,-.00135134666284035,-.00558261317396905,-.0157833171133724,-.0335930604290494,-.0565505431450506,-.076897431725972,-.0839936977751664,-.0698689855213075,-.0347378633474854,-2.74200568127716e-008,-8.73924948977462e-006,-.000180793439944228,-.00132378111976588,-.00546873583357464,-.0154613599726102,-.0329078099454216,-.0553969927824272,-.0753288338784716,-.0822803462291326,-.068443757944315,-.0340292605183584,-2.12012654199985e-008,-6.75721240365912e-006,-.000139789998708784,-.00102355130296691,-.00422844200182213,-.0119547672264766,-.0254444116512108,-.0428331113779068,-.0582444672431377,-.0636193962384749,-.0529209070727989,-.0263115204035591,-1.0129004659729e-008,-3.22829013115797e-006,-6.67853319250079e-005,-.000489006467861309,-.00202015813166747,-.00571144648888444,-.0121561878064161,-.0204637212045026,-.0278265692364139,-.0303944669425635,-.0252832132289538,-.0125704531070497},
			{-5.3934946534894e-011,-8.93862080339512e-008,-4.43762053833248e-006,-5.82757244194688e-005,-.000368787100047311,-.00144276568155165,-.00395000387691709,-.00808303973670482,-.0127550186066991,-.0155346480848768,-.0139021976630662,-.00719309508160456,-1.22262316118168e-010,-2.02624931058027e-007,-1.00594104551317e-005,-.000132102199014488,-.000835984234769121,-.00327053024384089,-.00895405768790711,-.0183230210276625,-.028913686156645,-.0352146831870672,-.0315141665027775,-.0163056519239563,-1.83015466396657e-010,-3.03310925545804e-007,-1.50580142318092e-005,-.00019774486802044,-.00125139167557278,-.00489568361654787,-.0134034025849063,-.0274278808519569,-.0432811345737185,-.0527131488435292,-.0471738149883672,-.0244080644511933,-2.32278652181071e-010,-3.84954639980414e-007,-1.91112550165974e-005,-.000250972840295157,-.00158823501355166,-.00621347919028916,-.017011263301183,-.0348107803231648,-.0549313334090684,-.0669022099971482,-.0598718260236465,-.0309781048820443,-2.6694895594493e-010,-4.42413619435074e-007,-2.19638332045296e-005,-.000288433470136975,-.00182529765297712,-.00714091358400096,-.0195503931804077,-.0400066961627327,-.0631304769703183,-.0768881468075034,-.0688083958532261,-.0356019491148408,-2.84846550926105e-010,-4.72075244245615e-007,-2.3436398585193e-005,-.000307771494551617,-.00194767474936783,-.00761967619488893,-.0208611494544942,-.0426889454411423,-.0673630603261273,-.0820431132524654,-.0734216553279677,-.0379888820906312,-2.84846550926105e-010,-4.72075244245615e-007,-2.3436398585193e-005,-.000307771494551617,-.00194767474936783,-.00761967619488893,-.0208611494544942,-.0426889454411423,-.0673630603261273,-.0820431132524654,-.0734216553279677,-.0379888820906312,-2.6694895594493e-010,-4.42413619435074e-007,-2.19638332045296e-005,-.000288433470136975,-.00182529765297712,-.00714091358400096,-.0195503931804077,-.0400066961627327,-.0631304769703183,-.0768881468075034,-.0688083958532261,-.0356019491148408,-2.32278652181071e-010,-3.84954639980414e-007,-1.91112550165974e-005,-.000250972840295156,-.00158823501355166,-.00621347919028916,-.017011263301183,-.0348107803231648,-.0549313334090684,-.0669022099971482,-.0598718260236465,-.0309781048820443,-1.83015466396657e-010,-3.03310925545804e-007,-1.50580142318092e-005,-.00019774486802044,-.00125139167557278,-.00489568361654787,-.0134034025849063,-.0274278808519569,-.0432811345737185,-.0527131488435292,-.0471738149883672,-.0244080644511933,-1.22262316118168e-010,-2.02624931058027e-007,-1.00594104551317e-005,-.000132102199014488,-.00083598423476912,-.00327053024384089,-.00895405768790711,-.0183230210276625,-.028913686156645,-.0352146831870672,-.0315141665027775,-.0163056519239563,-5.3934946534894e-011,-8.93862080339512e-008,-4.43762053833248e-006,-5.82757244194688e-005,-.000368787100047311,-.00144276568155165,-.00395000387691709,-.00808303973670482,-.0127550186066991,-.0155346480848768,-.0139021976630662,-.00719309508160456},
			{-8.61284959198155e-013,-1.42740470669776e-009,-7.08641812006219e-008,-9.30602664014744e-007,-5.88914614407857e-006,-2.30394662644891e-005,-6.30774506425453e-005,-.000129077731546852,-.000203684371377566,-.000248072161029854,-.000222003626892483,-.000114866241683583,-1.015228271122e-011,-1.68253444704445e-008,-8.35302177246445e-007,-1.09693560023246e-005,-6.94175324250859e-005,-.000271574665892772,-.000743517119145248,-.00152148671399835,-.00240090262810037,-.00292411783656719,-.00261683843315572,-.00135397065407101,-3.64695156892415e-011,-6.04408074120732e-008,-3.00061244597533e-006,-3.94046454582605e-005,-.000249364981245768,-.00097556350825723,-.00267089776882591,-.00546555736926158,-.0086246372913827,-.0105041559963674,-.00940033222170924,-.00486379816401721,-8.30147957637377e-011,-1.37580145726718e-007,-6.83023134969281e-006,-8.96959702655164e-005,-.000567624291069279,-.00222065480887151,-.00607970872643953,-.0124411339215587,-.0196320814726822,-.0239103905886559,-.0213977796180793,-.0110713620291109,-1.46147607732105e-010,-2.42209945641729e-007,-1.20246272104963e-005,-.000157909820254493,-.00099930297324955,-.0039094643904085,-.0107033315917073,-.0219026252294268,-.0345622935723416,-.0420942598523953,-.0376708064290284,-.0194911408262166,-2.15791227404772e-010,-3.57630085574026e-007,-1.77547146005579e-005,-.000233158479025165,-.00147550013642398,-.00577243878564134,-.015803782883183,-.0323398682673508,-.0510322397173339,-.0621534087434013,-.0556220500820546,-.0287792408488707,-2.77577471181613e-010,-4.60028222490692e-007,-2.28383184972103e-005,-.000299917386683,-.00189797148624789,-.00742522752170763,-.0203287878778718,-.041599554161483,-.0656440033259579,-.0799494318272223,-.0715479873273676,-.0370194330577289,-3.16221547012239e-010,-5.24072921213834e-007,-2.60178478303842e-005,-.000341671604648038,-.00216220530078157,-.00845896075008184,-.0231589427061906,-.0473910051704331,-.0747829000510975,-.0910799169238503,-.0815088311812927,-.0421732438918713,-3.19303631345124e-010,-5.29180849357959e-007,-2.62714333369098e-005,-.00034500174046461,-.00218327944688257,-.00854140684095277,-.0233846636134322,-.0478529062519234,-.0755117789234986,-.0919676362701244,-.0823032650013445,-.0425842895511383,-2.80522570694571e-010,-4.64909126147018e-007,-2.30806332657561e-005,-.000303099512904151,-.00191810898111,-.00750400925375776,-.0205444765035491,-.0420409258123548,-.0663404868028083,-.0807976960315661,-.072307112128577,-.0374122095817176,-2.01612260665772e-010,-3.34131331017576e-007,-1.65881078259945e-005,-.000217838364492369,-.00137854963658506,-.00539314988437847,-.0147653657308921,-.0302149166554566,-.0476790708315039,-.058069502618547,-.0519672991102317,-.026888246929992,-9.25567827468422e-011,-1.53394049110811e-007,-7.61532005623914e-006,-.00010000591288281,-.000632868848341838,-.0024759039977092,-.00677852995457166,-.0138711577726375,-.0218886559078795,-.0266587276008582,-.0238573090634585,-.0123439399038752},
			{-6.02215524056038e-013,-5.18975291864507e-009,-6.18297441623807e-007,-1.45626016852063e-005,-.000141170609267733,-.00076422934021594,-.00269138517084675,-.00669488358667141,-.0122597253338038,-.0166489373902391,-.016029237481894,-.00863095559575211,-1.36513094948138e-012,-1.1764380103799e-008,-1.40158620930416e-006,-3.30112019224449e-005,-.000320012288242827,-.00173239161588508,-.00610096061448634,-.0151762822814405,-.0277909316794019,-.0377406074727463,-.0363358420848087,-.0195650628997474,-2.0434757441563e-012,-1.76101973193715e-008,-2.09804592236745e-006,-4.94147396186346e-005,-.000479029025826867,-.00259323125578963,-.0091325781138374,-.0227175017461691,-.0416004741639895,-.0564942256781886,-.0543914208171814,-.0292871035439603,-2.59352830098303e-012,-2.23504219535243e-008,-2.66278740620364e-006,-6.27159515120887e-005,-.00060797165761674,-.0032912642453023,-.0115908397087212,-.0288324849829067,-.0527982812554331,-.0717010581395466,-.0690322308074567,-.0371704592590807,-2.98064271365472e-012,-2.56864836669144e-008,-3.06023954984572e-006,-7.20770403136054e-005,-.00069871853362736,-.00378252390296041,-.0133209080116245,-.0331360703672624,-.0606790418511532,-.0824032791251857,-.0793360981199193,-.0427185847602993,-3.18047992928925e-012,-2.7408634178924e-008,-3.26541333602756e-006,-7.69094460834964e-005,-.000745564123550859,-.00403612324963012,-.0142140084005345,-.0353576784818238,-.0647472687189207,-.087928007662455,-.0846551874810701,-.0455826526323854,-3.18047992928925e-012,-2.7408634178924e-008,-3.26541333602756e-006,-7.69094460834964e-005,-.000745564123550859,-.00403612324963012,-.0142140084005345,-.0353576784818238,-.0647472687189207,-.087928007662455,-.0846551874810701,-.0455826526323854,-2.98064271365472e-012,-2.56864836669144e-008,-3.06023954984572e-006,-7.20770403136054e-005,-.00069871853362736,-.00378252390296041,-.0133209080116245,-.0331360703672624,-.0606790418511532,-.0824032791251857,-.0793360981199193,-.0427185847602993,-2.59352830098303e-012,-2.23504219535243e-008,-2.66278740620364e-006,-6.27159515120887e-005,-.00060797165761674,-.0032912642453023,-.0115908397087212,-.0288324849829067,-.0527982812554331,-.0717010581395466,-.0690322308074567,-.0371704592590807,-2.0434757441563e-012,-1.76101973193715e-008,-2.09804592236745e-006,-4.94147396186346e-005,-.000479029025826867,-.00259323125578962,-.0091325781138374,-.0227175017461691,-.0416004741639895,-.0564942256781886,-.0543914208171814,-.0292871035439603,-1.36513094948138e-012,-1.1764380103799e-008,-1.40158620930416e-006,-3.30112019224449e-005,-.000320012288242827,-.00173239161588507,-.00610096061448634,-.0151762822814405,-.0277909316794019,-.0377406074727463,-.0363358420848087,-.0195650628997474,-6.02215524056038e-013,-5.18975291864507e-009,-6.18297441623807e-007,-1.45626016852063e-005,-.000141170609267733,-.00076422934021594,-.00269138517084675,-.00669488358667141,-.0122597253338038,-.0166489373902391,-.016029237481894,-.00863095559575211},
			{-.00270055519884434,-.00588249745219549,-.00818488569070588,-.00931641156972225,-.00922648473106781,-.00809895569654666,-.00629620113986558,-.00426418972382729,-.00242214596950661,-.00106408128362993,-.00029621599889564,-2.51299527189796e-005,-.00612174767248644,-.0133347265413407,-.0185538903067371,-.0211188872818742,-.0209150368234713,-.0183591000865756,-.014272529844964,-.00966626919088759,-.00549063631715242,-.00241211034093489,-.000671476592135066,-5.69657786046887e-005,-.00916369443189591,-.0199608618151533,-.0277734709743103,-.0316130360391258,-.0313078906116195,-.0274818878919328,-.0213646672921825,-.0144695178404447,-.00821897865428013,-.00361070780484583,-.00100513883251861,-8.52725424398906e-005,-.011630331761288,-.025333826534352,-.035249394662203,-.0401224746033481,-.0397351916593658,-.034879324707423,-.0271155013323826,-.0183643501167498,-.0104313221265329,-.00458262003120292,-.00127569706467466,-.000108225777940879,-.013366294714629,-.0291151962521434,-.0405107788184967,-.0461112228985269,-.045666133448504,-.0400854715975038,-.0311628068384012,-.0211054440183851,-.0119883188775888,-.00526662963356749,-.00146610976221544,-.000124379740266118,-.014262438055421,-.03106722482807,-.0432268243227726,-.0492027502229294,-.0487278197470118,-.0427730023755956,-.0332521174831873,-.0225204586888932,-.0127920758168761,-.00561973086134261,-.00156440510346971,-.000132718780991214,-.014262438055421,-.03106722482807,-.0432268243227726,-.0492027502229294,-.0487278197470118,-.0427730023755956,-.0332521174831873,-.0225204586888932,-.0127920758168761,-.00561973086134261,-.00156440510346971,-.000132718780991214,-.013366294714629,-.0291151962521434,-.0405107788184967,-.0461112228985269,-.045666133448504,-.0400854715975038,-.0311628068384012,-.0211054440183851,-.0119883188775888,-.00526662963356749,-.00146610976221544,-.000124379740266118,-.011630331761288,-.025333826534352,-.035249394662203,-.0401224746033481,-.0397351916593658,-.034879324707423,-.0271155013323826,-.0183643501167498,-.0104313221265329,-.00458262003120292,-.00127569706467466,-.000108225777940879,-.00916369443189591,-.0199608618151533,-.0277734709743103,-.0316130360391258,-.0313078906116195,-.0274818878919328,-.0213646672921825,-.0144695178404447,-.00821897865428013,-.00361070780484583,-.00100513883251861,-8.52725424398906e-005,-.00612174767248644,-.0133347265413407,-.0185538903067371,-.0211188872818742,-.0209150368234713,-.0183591000865756,-.014272529844964,-.00966626919088758,-.00549063631715242,-.00241211034093489,-.000671476592135066,-5.69657786046887e-005,-.00270055519884434,-.00588249745219549,-.00818488569070588,-.00931641156972225,-.00922648473106781,-.00809895569654666,-.00629620113986558,-.00426418972382729,-.00242214596950661,-.00106408128362993,-.00029621599889564,-2.51299527189796e-005},
			{-4.31250557013781e-005,-9.39373616201976e-005,-.000130704105389941,-.000148773395949532,-.000147337357934948,-.000129331892821903,-.000100543778916268,-6.80946715845654e-005,-3.86791500860789e-005,-1.69922705697598e-005,-4.73026119128408e-006,-4.01299188864111e-007,-.000508330898783026,-.00110727424428985,-.00154065738088721,-.00175364675704319,-.00173671964859619,-.00152448260646228,-.00118514651569285,-.000802656948401598,-.000455925373489819,-.000200294145262156,-5.5757329091799e-005,-4.73026119128408e-006,-.00182605057560164,-.00397760351765945,-.00553442315607657,-.00629953378353948,-.00623872741470196,-.00547631935750839,-.00425734001677702,-.00288334269301423,-.00163779694031119,-.000719506211645245,-.000200294145262156,-1.69922705697598e-005,-.0041566007314033,-.0090541356913339,-.0125978916716827,-.0143394969898535,-.0142010847243002,-.0124656312103062,-.00969089403328137,-.0065632926638522,-.00372808291892168,-.00163779694031119,-.000455925373489819,-3.86791500860789e-005,-.00731769858135901,-.015939812405686,-.0221785973614222,-.0252446947784247,-.0250010198852289,-.0219457527239106,-.0170608259252993,-.0115546814618157,-.0065632926638522,-.00288334269301423,-.000802656948401598,-6.80946715845654e-005,-.0108047964872896,-.0235356003221748,-.0327473492108772,-.0372745319354167,-.0369147388118314,-.0324035472773186,-.0251908096484753,-.0170608259252993,-.00969089403328137,-.00425734001677702,-.00118514651569285,-.000100543778916268,-.0138984708583546,-.0302744115168138,-.0421237067473831,-.0479471313016055,-.0474843207110434,-.0416814660110368,-.0324035472773186,-.0219457527239106,-.0124656312103062,-.00547631935750839,-.00152448260646228,-.000129331892821903,-.0158334029675552,-.0344891867628131,-.0479881297889341,-.0546222860610778,-.0540950434131128,-.0474843207110434,-.0369147388118314,-.0250010198852289,-.0142010847243002,-.00623872741470196,-.00173671964859619,-.000147337357934948,-.0159877247830154,-.0348253390053783,-.0484558508041033,-.0551546675312396,-.0546222860610777,-.0479471313016055,-.0372745319354167,-.0252446947784246,-.0143394969898535,-.00629953378353948,-.00175364675704319,-.000148773395949532,-.0140459337615274,-.0305956233004421,-.0425706396619946,-.0484558508041033,-.0479881297889341,-.0421237067473831,-.0327473492108772,-.0221785973614222,-.0125978916716827,-.00553442315607657,-.00154065738088721,-.000130704105389941,-.0100948470984408,-.0219891496246005,-.0305956233004421,-.0348253390053783,-.0344891867628131,-.0302744115168138,-.0235356003221748,-.015939812405686,-.0090541356913339,-.00397760351765944,-.00110727424428985,-9.39373616201976e-005,-.00463437375617701,-.0100948470984408,-.0140459337615274,-.0159877247830154,-.0158334029675552,-.0138984708583546,-.0108047964872896,-.00731769858135901,-.0041566007314033,-.00182605057560164,-.000508330898783026,-4.31250557013781e-005},
			{-5.13298537640478e-007,-1.11809502771057e-006,-1.55571337982341e-006,-1.77078418424541e-006,-1.75369165645863e-006,-1.53938053820615e-006,-1.19672830207889e-006,-8.10500874190154e-007,-4.60380881913306e-007,-2.02251510004898e-007,-5.63022148645256e-008,-4.77648743752639e-009,-3.14616387984846e-005,-6.85314672157709e-005,-9.53544357538297e-005,-.00010853678378063,-.000107489130425036,-9.43533458892899e-005,-7.33511413319109e-005,-4.96780798691598e-005,-2.82181536753749e-005,-1.23966142266297e-005,-3.45093511422209e-006,-2.92765537918228e-007,-.000271218151906612,-.000590782254057447,-.000822012292713028,-.000935652020526277,-.000926620621724616,-.000813382299066098,-.000632330728850647,-.000428254773970375,-.00024325736936566,-.000106866232302167,-2.97491255946283e-005,-2.52381411676342e-006,-.00110725656401101,-.00241188697772547,-.00335589082222505,-.0038198285552606,-.0037829575879782,-.00332065860400601,-.00258150992189795,-.0017483634713027,-.000993105797236321,-.000436284726357267,-.000121451733066827,-1.03035498453369e-005,-.00298608212919577,-.00650444787225884,-.00905026525693164,-.0103014262061688,-.0102019914951301,-.00895524997265581,-.00696189202635823,-.00471503812818561,-.00267823697769347,-.00117658550597994,-.000327534612535476,-2.7786916836135e-005,-.00610102997451746,-.0132895981153155,-.0184910987778976,-.021047415089364,-.0208442545176493,-.0182969677820565,-.0142242276315438,-.00963355584556567,-.00547205447566498,-.00240394708818667,-.000669204128457476,-5.67729905547983e-005,-.0100949464135448,-.0219893659582073,-.0305959243062406,-.0348256816239572,-.0344895260742596,-.0302747093624672,-.0235358318700697,-.0159399692247158,-.00905422476771311,-.00397764265011023,-.00110728513787305,-9.39382857945563e-005,-.0139798092182615,-.0304515871935496,-.0423702283435318,-.0482277334672722,-.0477622143590056,-.0419253994709405,-.0325931833471828,-.0220741863877053,-.0125385841278022,-.00550836855481976,-.0015334043731933,-.000130088785011913,-.0163811720361289,-.0356823673916168,-.0496483170026962,-.0565119871455732,-.0559665041221956,-.0491270782522392,-.0381918333275861,-.0258659498945255,-.0146923824553127,-.00645456111210986,-.00179680283515025,-.000152434610085198,-.0160470195082919,-.0349544980281293,-.04863556220147,-.0553592232703062,-.0548248673220092,-.0481249559775311,-.0374127744408963,-.0253383214364947,-.0143926788244271,-.00632289727835408,-.00176015062198346,-.000149325161617318,-.0124076008346097,-.027026916660946,-.0376051541440983,-.0428039078843937,-.0423907423550166,-.0372103519687145,-.0289276629306809,-.0195916617438277,-.0111284599424805,-.0048888820448881,-.00136095343531401,-.000115458636973295,-.00592778988329973,-.0129122370468795,-.0179660399513751,-.0204497690975875,-.0202523773150973,-.0177774213479609,-.0138203275519371,-.00936000890342516,-.00531667428240711,-.00233568648062049,-.000650201929690796,-5.51609089712794e-005},
			{-5.59951004436151e-009,-1.21971598964522e-008,-1.6971084189159e-008,-1.9317264903302e-008,-1.91308046389391e-008,-1.67929112469387e-008,-1.30549605277777e-008,-8.84165344957659e-009,-5.02223790536332e-009,-2.20633662228147e-009,-6.14193874587905e-010,-5.21061086714588e-011,-1.78466133868042e-006,-3.88744721171111e-006,-5.40897999788056e-006,-6.15674863851456e-006,-6.09732050598647e-006,-5.35219317920171e-006,-4.16084320723466e-006,-2.81798888768115e-006,-1.60067465766628e-006,-7.03197893870279e-007,-1.95754280954478e-007,-1.66071240016209e-008,-3.69202255792134e-005,-8.04216603301225e-005,-.000111898407472009,-.00012736788971769,-.000126138468756338,-.000110723628756234,-8.60775467486887e-005,-5.82972148036964e-005,-3.31139965656822e-005,-1.45474238197567e-005,-4.04967152831355e-006,-3.43560288483173e-007,-.00027033225085123,-.000588852535789363,-.000819327289690437,-.000932595827175549,-.000923593928337934,-.000810725484866511,-.000630265297550785,-.000426855931918081,-.00024246279880047,-.000106517166772019,-2.96519536996954e-005,-2.51557038538368e-006,-.0011167825595386,-.0024326370236585,-.00338476236112819,-.00385269146248218,-.00381550328536708,-.00334922703163026,-.00260371927496907,-.00176340506432601,-.00100164972615974,-.000440038189183426,-.000122496611646568,-1.03921937720791e-005,-.00315739828904347,-.0068776181278623,-.00956949300162804,-.0108924349936834,-.0107872955591597,-.00946902654323017,-.00736130656207001,-.00498554717338517,-.00283189158407368,-.0012440880400349,-.000346325780898917,-2.93810953216598e-005,-.00672017616832581,-.0146382563131457,-.0203676169191767,-.0231833539384622,-.0229595761766608,-.0201537850748433,-.0156677341271221,-.0106111906809188,-.00602737082635835,-.00264790502577821,-.000737116463054399,-6.253444086072e-005,-.0113127416048293,-.0246420342366014,-.0342868373598303,-.039026847819149,-.0386501404333965,-.0339268729271671,-.026375056735671,-.0178628737232513,-.010146473397001,-.00445748215529598,-.00124086152956873,-.00010527045022929,-.0153830666658602,-.0335082395307456,-.0466232433562013,-.0530687098434273,-.0525564631192487,-.0461337637005683,-.0358648036217386,-.0242899366861979,-.0137971750917403,-.00606128448364179,-.00168732357718739,-.000143146764099252,-.0168026502756542,-.0366004544098804,-.0509257399611718,-.0579660084330403,-.0574064904417204,-.0503910900341255,-.039174487477097,-.0265314660737799,-.0150704090994405,-.00662063329842476,-.00184303355014811,-.000156356665903898,-.0139770501826433,-.0304455773108406,-.0423618662144552,-.0482182153091671,-.0477527880751015,-.0419171251326689,-.0325867508020486,-.0220698298571165,-.0125361095303546,-.00550728143089673,-.0015331017426483,-.000130063110871032,-.00694919005368325,-.015137106919017,-.0210617158489217,-.0239734091138143,-.0237420053295095,-.0208405969245687,-.0162016678481028,-.0109728047138306,-.00623277490756985,-.0027381417997588,-.000762236326125946,-6.46655241703613e-005},
			{-5.85380492906022e-011,-1.27510789616824e-010,-1.77418051741924e-010,-2.01945348094836e-010,-1.99996066807805e-010,-1.75555407262082e-010,-1.36478355571743e-010,-9.24318630275365e-011,-5.25031668349854e-011,-2.30653469542085e-011,-6.4208673651397e-012,-5.44724437242996e-013,-9.70149149662404e-008,-2.11323208782435e-007,-2.94034348800585e-007,-3.34683355709183e-007,-3.31452818296366e-007,-2.9094739428102e-007,-2.26185126101403e-007,-1.53187020074262e-007,-8.70133242853409e-008,-3.82261230182231e-008,-1.06412821914539e-008,-9.02770003435125e-010,-4.81635129901971e-006,-1.04912405632317e-005,-1.45974741955391e-005,-1.6615513352675e-005,-1.64551317961879e-005,-1.44442208796392e-005,-1.12290674717028e-005,-7.60504200188573e-006,-4.31981760329904e-006,-1.89775394143646e-006,-5.28291483055767e-007,-4.48184434349486e-008,-6.3249292855151e-005,-.000137773078747959,-.000191696963742466,-.000218198259374459,-.0002160920965543,-.00018968441040975,-.000147462369938119,-9.98709394081982e-005,-5.67287126102093e-005,-2.49216860143408e-005,-6.93762988830925e-006,-5.88564803133452e-007,-.000400261404288981,-.000871871344066475,-.00121311863645989,-.00138082716451303,-.00136749870422512,-.00120038256642976,-.000933188224038868,-.000632014693772654,-.000358997123096203,-.000157712261925103,-4.39035024137147e-005,-3.7246230587395e-006,-.00156589918054001,-.00341092748034313,-.00474595216619707,-.00540206000930644,-.0053499165230271,-.00469612622392589,-.00365081084424778,-.00247256238164128,-.00140446541897081,-.000617000537806862,-.000171758899848453,-1.45714378978438e-005,-.00428711876993237,-.00933843724141566,-.0129934678207622,-.0147897598708826,-.0146470014343651,-.0128570543562206,-.00999518991411039,-.00676938127811489,-.00384514541818346,-.00168922406982674,-.000470241515286636,-3.98936826157651e-005,-.00877289047128288,-.0191095911936167,-.0265890161087102,-.0302648352907365,-.0299727033964202,-.0263098681663614,-.0204535285962103,-.01385243649599,-.00786846397549439,-.00345672199472992,-.000962272689432589,-8.16359254002543e-005,-.0138436015212964,-.0301548921174003,-.0419574078870825,-.0477578423262084,-.0472968588510568,-.0415169130590646,-.0322756223296346,-.0218591137752465,-.0124164185359414,-.0054546995681281,-.00151846414940855,-.000128821307499823,-.0168604597524974,-.036726378183032,-.0511009498439408,-.0581654403421276,-.0576039973305382,-.0505644604551389,-.0393092672046696,-.0266227475173882,-.015122258804788,-.00664341157096352,-.00184937450253303,-.000156894610627413,-.0150886871005195,-.0328670058156056,-.045731033082912,-.0520531552679814,-.0515507112035252,-.0452509204027591,-.0351784732865376,-.0238251099283638,-.0135331440961878,-.00594529211811034,-.00165503394391524,-.000140407421995964,-.00780699308131397,-.0170056205219542,-.0236615589216733,-.0269326695112991,-.0266727014101101,-.0234131452361595,-.0182015900873001,-.0123272798443727,-.00700214416426768,-.00307613605632261,-.000856326230599691,-7.26477900154669e-005},
			{-9.64305560601928e-005,-.00109223938839564,-.0036470340953271,-.00744526010016809,-.0112949614634506,-.0137194545508608,-.0137194545508608,-.0112949614634506,-.00744526010016809,-.0036470340953271,-.00109223938839564,-9.64305560601928e-005,-.000218593396043406,-.00247594047941341,-.00826727129817659,-.0168772716472393,-.025603958800506,-.0310999156767677,-.0310999156767677,-.025603958800506,-.0168772716472393,-.00826727129817659,-.00247594047941341,-.000218593396043406,-.000327214252095852,-.00370625566402853,-.0123753463904701,-.0252637267155745,-.0383267764774421,-.0465537195204054,-.0465537195204054,-.0383267764774421,-.0252637267155745,-.0123753463904701,-.00370625566402854,-.000327214252095852,-.000415292144143345,-.00470388698413707,-.0157064800939949,-.0320640900252994,-.0486433860367337,-.0590848163661698,-.0590848163661698,-.0486433860367337,-.0320640900252994,-.0157064800939949,-.00470388698413707,-.000415292144143345,-.000477279350685986,-.00540599709662285,-.0180508558289433,-.0368500302339686,-.0559039799577005,-.0679039158056228,-.0679039158056228,-.0559039799577005,-.0368500302339686,-.0180508558289433,-.00540599709662285,-.000477279350685986,-.000509278548739481,-.00576844221712266,-.019261075608775,-.0393206408188163,-.059652062760935,-.0724565344080177,-.0724565344080177,-.059652062760935,-.0393206408188163,-.019261075608775,-.00576844221712266,-.000509278548739481,-.000509278548739481,-.00576844221712266,-.019261075608775,-.0393206408188163,-.059652062760935,-.0724565344080177,-.0724565344080177,-.059652062760935,-.0393206408188163,-.019261075608775,-.00576844221712266,-.000509278548739481,-.000477279350685986,-.00540599709662285,-.0180508558289433,-.0368500302339686,-.0559039799577005,-.0679039158056228,-.0679039158056228,-.0559039799577005,-.0368500302339686,-.0180508558289433,-.00540599709662285,-.000477279350685986,-.000415292144143345,-.00470388698413707,-.0157064800939949,-.0320640900252994,-.0486433860367337,-.0590848163661699,-.0590848163661699,-.0486433860367337,-.0320640900252994,-.0157064800939949,-.00470388698413707,-.000415292144143345,-.000327214252095852,-.00370625566402853,-.0123753463904701,-.0252637267155745,-.0383267764774421,-.0465537195204054,-.0465537195204054,-.0383267764774421,-.0252637267155745,-.0123753463904701,-.00370625566402854,-.000327214252095852,-.000218593396043406,-.00247594047941341,-.0082672712981766,-.0168772716472393,-.025603958800506,-.0310999156767677,-.0310999156767677,-.025603958800506,-.0168772716472393,-.0082672712981766,-.00247594047941341,-.000218593396043406,-9.64305560601928e-005,-.00109223938839564,-.0036470340953271,-.00744526010016809,-.0112949614634506,-.0137194545508608,-.0137194545508608,-.0112949614634506,-.00744526010016809,-.0036470340953271,-.00109223938839564,-9.64305560601928e-005},
			{-1.53989561292143e-006,-1.74419262023187e-005,-5.82393385771156e-005,-.000118893054584807,-.000180368778489956,-.000219085409622217,-.000219085409622217,-.000180368778489956,-.000118893054584807,-5.82393385771156e-005,-1.74419262023187e-005,-1.53989561292143e-006,-1.81513161714798e-005,-.000205594401647313,-.000686488512117749,-.00140143617934442,-.00212607315610414,-.00258244033247591,-.00258244033247591,-.00212607315610414,-.00140143617934442,-.000686488512117749,-.000205594401647313,-1.81513161714798e-005,-6.52040263973911e-005,-.000738546046221744,-.00246603687813908,-.00503430609488318,-.00763738171291507,-.00927676571867675,-.00927676571867675,-.00763738171291507,-.00503430609488318,-.00246603687813908,-.000738546046221744,-6.52040263973911e-005,-.000148422561475078,-.00168113691751877,-.00561338816585801,-.0114594856657818,-.0173848122489431,-.0211165077717546,-.0211165077717546,-.0173848122489431,-.0114594856657818,-.005613388165858,-.00168113691751877,-.000148422561475078,-.000261298026375791,-.00295964275410309,-.00988237390894377,-.0201744327681154,-.0306059744853902,-.0371756271409923,-.0371756271409923,-.0306059744853902,-.0201744327681154,-.00988237390894377,-.00295964275410309,-.000261298026375791,-.000385814196380375,-.0043699992938526,-.0145916147966848,-.0297881414330556,-.0451906185986146,-.0548909033461696,-.0548909033461696,-.0451906185986146,-.0297881414330556,-.0145916147966848,-.0043699992938526,-.000385814196380375,-.000496282125391267,-.0056212357084272,-.0187695467718091,-.0383172062628759,-.0581297848971723,-.0706074956102091,-.070607495610209,-.0581297848971723,-.0383172062628759,-.0187695467718091,-.0056212357084272,-.000496282125391267,-.00056537405855632,-.00640381888440892,-.0213826255122005,-.0436516918684151,-.066222559163116,-.080437405087215,-.080437405087215,-.066222559163116,-.0436516918684151,-.0213826255122005,-.00640381888440892,-.00056537405855632,-.000570884532287664,-.00646623433345328,-.0215910333696339,-.0440771473659381,-.0668680038331856,-.0812213961477219,-.0812213961477219,-.0668680038331856,-.0440771473659381,-.0215910333696339,-.00646623433345328,-.000570884532287664,-.0005015476832896,-.00568087707080664,-.0189686918350622,-.038723752172464,-.0587465424476329,-.0713566417856807,-.0713566417856807,-.0587465424476329,-.038723752172464,-.0189686918350622,-.00568087707080664,-.0005015476832896,-.000360463552039073,-.00408286030594201,-.0136328454187137,-.027830855811784,-.0422212843688097,-.0512841937399486,-.0512841937399486,-.0422212843688097,-.027830855811784,-.0136328454187137,-.00408286030594202,-.000360463552039073,-.000165482726913838,-.00187437218884842,-.00625860901253902,-.0127766757166634,-.0193830783490639,-.0235437069286394,-.0235437069286394,-.0193830783490639,-.0127766757166634,-.00625860901253902,-.00187437218884842,-.000165482726913838},
			{-1.83286990233606e-008,-2.076035629087e-007,-6.93197187616125e-007,-1.41513164604626e-006,-2.14685010231423e-006,-2.6076771046559e-006,-2.6076771046559e-006,-2.14685010231423e-006,-1.41513164604626e-006,-6.93197187616125e-007,-2.076035629087e-007,-1.83286990233606e-008,-1.1234220751336e-006,-1.27246579340282e-005,-4.24881778022507e-005,-8.67377509097766e-005,-.000131587015197788,-.000159832512959983,-.000159832512959983,-.000131587015197788,-8.67377509097766e-005,-4.24881778022507e-005,-1.27246579340282e-005,-1.1234220751336e-006,-9.68457050125128e-006,-.000109694165348983,-.00036627351598611,-.000747731313456525,-.00113435880773542,-.00137785189949114,-.00137785189949114,-.00113435880773542,-.000747731313456525,-.00036627351598611,-.000109694165348983,-9.68457050125128e-006,-3.95375611173333e-005,-.000447829851219516,-.00149532305248749,-.00305263640770766,-.00463105521138285,-.00562512335262842,-.00562512335262842,-.00463105521138285,-.00305263640770766,-.00149532305248749,-.000447829851219516,-3.95375611173333e-005,-.000106626059868886,-.001207720738907,-.00403263127041889,-.00823243981590666,-.0124891661566992,-.015169998412074,-.015169998412074,-.0124891661566992,-.00823243981590666,-.00403263127041889,-.001207720738907,-.000106626059868886,-.000217853615265421,-.00246756120900883,-.00823929255543563,-.0168201542714382,-.0255173078910835,-.0309946649224858,-.0309946649224858,-.0255173078910835,-.0168201542714382,-.00823929255543563,-.00246756120900883,-.000217853615265422,-.000360467098389462,-.00408290047436723,-.0136329795428058,-.0278311296202336,-.0422216997546855,-.0512846982894915,-.0512846982894915,-.0422216997546855,-.0278311296202336,-.0136329795428058,-.00408290047436723,-.000360467098389462,-.000499186529428583,-.00565413300383832,-.0188793922500835,-.038541450987549,-.0584699792610344,-.0710207135860798,-.0710207135860798,-.0584699792610344,-.038541450987549,-.0188793922500835,-.00565413300383832,-.000499186529428583,-.000584933620267579,-.00662536405218185,-.0221223743155372,-.0451618565956071,-.0685135808557914,-.0832202005920431,-.0832202005920431,-.0685135808557914,-.0451618565956071,-.0221223743155372,-.00662536405218185,-.000584933620267579,-.000573001809320343,-.00649021608224453,-.0216711094559173,-.0442406191829269,-.0671160015993357,-.0815226272844922,-.0815226272844922,-.0671160015993357,-.0442406191829269,-.0216711094559173,-.00649021608224453,-.000573001809320343,-.000443046618338206,-.00501825341691917,-.0167561630764652,-.0342069717815257,-.0518942821144625,-.0630335258090745,-.0630335258090745,-.0518942821144625,-.0342069717815257,-.0167561630764652,-.00501825341691917,-.000443046618338206,-.000211667613829872,-.00239749426445693,-.00800533602681095,-.016342542282569,-.0247927383077269,-.0301145645786321,-.0301145645786321,-.0247927383077269,-.016342542282569,-.00800533602681095,-.00239749426445693,-.000211667613829872},
			{-1.99945502952237e-010,-2.26472150301291e-009,-7.56200209007358e-009,-1.54374955010019e-008,-2.34197213300945e-008,-2.84468259074429e-008,-2.84468259074429e-008,-2.34197213300945e-008,-1.54374955010019e-008,-7.56200209007358e-009,-2.26472150301291e-009,-1.99945502952237e-010,-6.37261128446745e-008,-7.21806171840919e-007,-2.4101417206608e-006,-4.92019858316594e-006,-7.46427292555284e-006,-9.0664986763096e-006,-9.0664986763096e-006,-7.46427292555284e-006,-4.92019858316594e-006,-2.4101417206608e-006,-7.21806171840918e-007,-6.37261128446745e-008,-1.31833553544195e-006,-1.49323830304632e-005,-4.98598664497681e-005,-.000101786729867376,-.0001544173307413,-.000187563415583651,-.000187563415583651,-.0001544173307413,-.000101786729867376,-4.98598664497681e-005,-1.49323830304632e-005,-1.31833553544195e-006,-9.65293703064032e-006,-.000109335862711265,-.000365077128133699,-.000745288940144662,-.0011306535627788,-.00137335131399144,-.00137335131399144,-.0011306535627788,-.000745288940144662,-.000365077128133699,-.000109335862711265,-9.65293703064033e-006,-3.98777122973611e-005,-.000451682639505874,-.00150818767757949,-.0030788989754264,-.00467089729699484,-.00567351764636142,-.00567351764636142,-.00467089729699484,-.0030788989754264,-.00150818767757949,-.000451682639505874,-3.98777122973611e-005,-.000112743362173094,-.00127700954942894,-.00426398957619228,-.0087047476468162,-.0132056889748729,-.0160403247315041,-.0160403247315041,-.0132056889748729,-.0087047476468162,-.00426398957619228,-.00127700954942894,-.000112743362173094,-.000239961888318527,-.00271797484991882,-.00907543442693066,-.0185271012182458,-.0281068614761773,-.0341400729730291,-.0341400729730291,-.0281068614761773,-.0185271012182458,-.00907543442693066,-.00271797484991882,-.000239961888318527,-.000403951737210287,-.004575437666423,-.0152775823210325,-.0311885140387246,-.0473150782417566,-.0574713838208962,-.0574713838208962,-.0473150782417566,-.0311885140387246,-.0152775823210325,-.004575437666423,-.000403951737210287,-.000549293594811998,-.00622168039425801,-.0207744572931185,-.042410143131176,-.0643390460348379,-.0781495909283993,-.0781495909283993,-.0643390460348379,-.042410143131176,-.0207744572931185,-.00622168039425801,-.000549293594811998,-.000599983629581883,-.00679583090045496,-.0226915704225263,-.0463238454758227,-.0702763963177652,-.0853614090140783,-.0853614090140783,-.0702763963177652,-.0463238454758227,-.0226915704225263,-.00679583090045496,-.000599983629581883,-.000499088010632534,-.00565301711159485,-.0188756662380921,-.0385338444975389,-.058458439702929,-.0710066970315766,-.0710066970315766,-.058458439702929,-.0385338444975389,-.0188756662380921,-.00565301711159485,-.000499088010632534,-.000248139442448811,-.00281059950217384,-.00938471210765786,-.0191584780345857,-.0290647026682284,-.0353035172878956,-.0353035172878956,-.0290647026682284,-.0191584780345857,-.00938471210765786,-.00281059950217384,-.000248139442448811},
			{-1.92058726559064e-006,-.000113117990564565,-.000906410211953001,-.00331870648655728,-.00771242584002074,-.0129629077002378,-.0166745014700866,-.0166874796528932,-.0127648935683974,-.00697208388856876,-.00224638876284317,-.000206393222578878,-4.35367906124191e-006,-.000256421270615505,-.00205469401540682,-.00752300257311598,-.0174828957229149,-.0293849390956891,-.0377985573514871,-.0378279769169244,-.0289360711920697,-.0158046531822386,-.00509222147590975,-.000467862027215704,-6.51705798837019e-006,-.000383839108612685,-.00307568836802217,-.0112612444155839,-.0261702903746801,-.0439865569734266,-.0565809713282148,-.0566250097176437,-.0433146429173246,-.0236581153148544,-.0076225882021403,-.000700346515907722,-8.27128698753606e-006,-.000487158995675827,-.00390358674443833,-.0142924897345245,-.0332146779455411,-.0558266378740331,-.071811153548789,-.0718670459711192,-.0549738613606283,-.0300262881966959,-.00967439828220347,-.000888862280208502,-9.50587324711357e-006,-.000559873169805834,-.00448624269205744,-.0164258108812803,-.0381723568496402,-.0641594160912017,-.0825298075610779,-.0825940425807241,-.063179352715725,-.0345080626641537,-.0111184153023947,-.00102153536475457,-1.01431945983696e-005,-.000597409871152132,-.00478702287082038,-.0175270795090219,-.0407316228334812,-.0684609846789651,-.0880630191983885,-.0881315608554635,-.067415212946293,-.0368216560137206,-.0118638495492168,-.00109002421181751,-1.01431945983696e-005,-.000597409871152132,-.00478702287082038,-.0175270795090219,-.0407316228334812,-.0684609846789651,-.0880630191983885,-.0881315608554635,-.067415212946293,-.0368216560137206,-.0118638495492168,-.00109002421181751,-9.50587324711357e-006,-.000559873169805834,-.00448624269205744,-.0164258108812803,-.0381723568496403,-.0641594160912017,-.0825298075610779,-.0825940425807241,-.063179352715725,-.0345080626641537,-.0111184153023947,-.00102153536475457,-8.27128698753606e-006,-.000487158995675827,-.00390358674443833,-.0142924897345245,-.0332146779455411,-.0558266378740331,-.071811153548789,-.0718670459711192,-.0549738613606283,-.0300262881966959,-.00967439828220347,-.000888862280208502,-6.51705798837019e-006,-.000383839108612685,-.00307568836802217,-.0112612444155839,-.0261702903746801,-.0439865569734266,-.0565809713282148,-.0566250097176437,-.0433146429173246,-.0236581153148544,-.0076225882021403,-.000700346515907722,-4.35367906124191e-006,-.000256421270615505,-.00205469401540682,-.00752300257311598,-.0174828957229149,-.0293849390956891,-.0377985573514871,-.0378279769169244,-.0289360711920697,-.0158046531822386,-.00509222147590975,-.000467862027215704,-1.92058726559064e-006,-.000113117990564565,-.000906410211953001,-.00331870648655728,-.00771242584002074,-.0129629077002378,-.0166745014700866,-.0166874796528932,-.0127648935683974,-.00697208388856876,-.00224638876284317,-.000206393222578878},
			{-3.06697796211909e-008,-1.80637657239712e-006,-1.44744276633774e-005,-5.29962883716334e-005,-.000123159413319067,-.000207004143752805,-.000266274433108618,-.000266481681179921,-.000203842064166924,-.000111336922926592,-3.58724904274172e-005,-3.29588494374136e-006,-3.61515976889952e-007,-2.1292425288575e-005,-.000170615404521307,-.000624686750242411,-.00145172531948917,-.00244003400654881,-.00313867471481826,-.0031411176240886,-.00240276141102928,-.00131236927532137,-.000422842243423972,-3.88498085043389e-005,-1.29865498884588e-006,-7.6487668853577e-005,-.000612892819182374,-.00224402963223718,-.00521495715018533,-.00876520690127838,-.0112748974798531,-.0112836730154145,-.00863131450035304,-.00471435570086283,-.00151895413763271,-.000139558140870839,-2.95610118832487e-006,-.00017410728078841,-.0013951151049833,-.00510803771553511,-.0118706978844523,-.0199520571355211,-.025664813306616,-.0256847888746645,-.0196472806638003,-.0107311892759846,-.00345756353291146,-.000317673277053615,-5.20421827112933e-006,-.000306515993225909,-.00245610114713186,-.00899270407717994,-.0208983721752537,-.0351256109572267,-.0451829223109582,-.0452180893128947,-.034589051759252,-.0188922664493339,-.00608704308994404,-.000559264033051857,-7.68418084863826e-006,-.000452579850083181,-.00362650534890298,-.0132779912077635,-.0308570591913224,-.0518639943508112,-.0667139094133409,-.0667658345226861,-.051071748964773,-.0278949852742899,-.00898769757526597,-.000825769740663992,-9.88434754146089e-006,-.000582164399381254,-.0046648614778342,-.0170798010009427,-.0396921810095751,-.0667139094133409,-.0858157140598609,-.0858825065829686,-.0656948250258218,-.0358819937409303,-.0115610926110904,-.0010622075751108,-1.12604371581811e-005,-.000663212782377944,-.00531429912823647,-.0194576349160754,-.0452180893128947,-.0760017574629395,-.0977628974803007,-.0978389887959829,-.0748407971004666,-.0408774513376,-.0131706171076067,-.00121008711989414,-1.13701881137203e-005,-.000669676850830108,-.00536609546608301,-.0196472806638003,-.0456588118567917,-.0767425160486888,-.0987157531522446,-.0987925860994006,-.0755702402721397,-.0412758674276048,-.0132989858194334,-.0012218813527315,-9.98922055945599e-006,-.000588341168993646,-.00471435570086283,-.017261017846082,-.040113315413826,-.067421744603583,-.0867262195724516,-.0867937207648416,-.0663918477213734,-.0362627020233443,-.01168375590964,-.0010734776072168,-7.17927735473982e-006,-.000422842243423972,-.00338821902309026,-.0124055359279088,-.0288295383268322,-.0484561734690455,-.0623303470508818,-.06237886032367,-.0477159840498455,-.0260620930240078,-.00839714407356662,-.000771510993319342,-3.29588494374136e-006,-.000194119730276028,-.00155547411146173,-.00569517196000836,-.0132351539871473,-.0222454106000665,-.0286148092951926,-.0286370809191255,-.0219056021427923,-.0119646666030396,-.00385498697918716,-.000354187662792186},
			{-3.65048874112027e-010,-2.15005044744533e-008,-1.72282735226472e-007,-6.30791470989705e-007,-1.46591223424896e-006,-2.46387911966847e-006,-3.16934726012697e-006,-3.1718140426088e-006,-2.4262422795302e-006,-1.32519433994785e-006,-4.26974448589619e-007,-3.92294663599167e-008,-2.23749630651595e-008,-1.31783174148499e-006,-1.05597362732147e-005,-3.86631403796111e-005,-8.98502486218331e-005,-.00015101869423293,-.000194258996301528,-.000194410192951716,-.000148711817078381,-8.12252180824206e-005,-2.61705711055768e-005,-2.40449409138482e-006,-1.92885569959679e-007,-1.1360498152647e-005,-9.10312452249138e-005,-.000333299404644144,-.000774562906136249,-.00130187150820567,-.00167462878541062,-.00167593219100564,-.00128198484678476,-.000700209981991998,-.000225605982417451,-2.07281778267435e-005,-7.87461355147049e-007,-4.63795880236092e-005,-.000371637897746933,-.00136070521452511,-.00316217722165629,-.00531493103550084,-.00683672424538222,-.00684204543937499,-.00523374312001741,-.00285862908989082,-.000921043459502259,-8.46234324557743e-005,-2.12364898657175e-006,-.00012507783964276,-.00100224403364473,-.00366958991823735,-.00852785271078683,-.0143334626308517,-.0184374794017228,-.0184518297544813,-.0141145126680233,-.00770923516443232,-.00248389714184165,-.000228214966233319,-4.33894499945229e-006,-.000255553469660892,-.00204774035893469,-.00749754263838111,-.0174237287374253,-.0292854922824927,-.0376706365121829,-.0376999565136218,-.0288381434728016,-.0157511658366454,-.00507498798101565,-.000466278651886231,-7.17934798611665e-006,-.000422846403442921,-.00338825235716593,-.0124056579764159,-.0288298219584409,-.0484566501919911,-.0623309602711284,-.0623794740212014,-.0477164534906379,-.0260623494288551,-.00839722668659549,-.000771518583621402,-9.94219394991092e-006,-.000585571414309697,-.00469216176055673,-.0171797575373556,-.0399244725296093,-.0671043408542665,-.0863179354585031,-.0863851188731245,-.0660792924543047,-.0360919868089839,-.0116287518756503,-.00106842395843715,-1.16500008667718e-005,-.000686157151895624,-.00549815149985173,-.0201307871491398,-.0467824447922306,-.0786310982319312,-.101145082058928,-.101223805813716,-.0774299735296571,-.0422916390211788,-.0136262649987847,-.00125195103863209,-1.14123574777427e-005,-.000672160525384795,-.00538599705705276,-.0197201478250184,-.0458281496934892,-.0770271360622618,-.0990818667545695,-.0991589846571268,-.0758505125894862,-.0414289499502059,-.0133483086422873,-.00122641302441892,-8.82406705447926e-006,-.000519716417833643,-.00416446814598573,-.015247674029803,-.0354344548588437,-.0595575992912496,-.0766103793918513,-.0766700071720531,-.0586478306968206,-.0320329811846719,-.0103209499661349,-.000948266016470748,-4.21573969958762e-006,-.000248296972548114,-.00198959431997835,-.00728464826218446,-.0169289781185245,-.0284539242725722,-.0366009704831913,-.0366294579367253,-.028019278030959,-.0153039080100614,-.00493088257841441,-.00045303856676568},
			{-3.06697796211909e-008,-9.39296633663941e-006,-.000180620480252622,-.00118608291592308,-.00422235247914266,-.00982031115806883,-.0162489801425508,-.0197676016789817,-.0175473588745023,-.0106866754670937,-.00370432632910279,-.000354187662792186,-6.95237231559063e-008,-2.1292425288575e-005,-.000409439142389331,-.00268866947546018,-.00957143052386794,-.0222611509665594,-.0368339652566651,-.0448101448253052,-.0397771922581837,-.024225067014107,-.00839714407356662,-.000802889532215207,-1.04070632906319e-007,-3.18728065084108e-005,-.000612892819182374,-.00402469144754472,-.0143275530598965,-.0333228999419048,-.0551370654893481,-.0670766742761412,-.0595428061909519,-.0362627020233444,-.0125697540159339,-.00120185194317225,-1.32083844163853e-007,-4.04521687853206e-005,-.000777868235805777,-.00510803771553511,-.0181841719682459,-.0422925910998948,-.0699785843745455,-.085132037201067,-.0755702402721397,-.0460237143683289,-.0159532173895116,-.00152536042432723,-1.51798901731387e-007,-4.64901277906996e-005,-.000893974161901044,-.00587047204848792,-.0208983721752537,-.0486052546469995,-.0804237060180922,-.0978389887959829,-.0868499819149511,-.0528932916734667,-.0183344215497376,-.00175303829641842,-1.61976260365957e-007,-4.96070587967628e-005,-.000953910667053138,-.00626405789601532,-.0222995037123093,-.0518639943508112,-.0858157140598609,-.104398604617067,-.0926728396779039,-.0564395228555362,-.0195636529956887,-.00187057076364506,-1.61976260365957e-007,-4.96070587967628e-005,-.000953910667053138,-.00626405789601532,-.0222995037123093,-.0518639943508112,-.0858157140598609,-.104398604617067,-.0926728396779039,-.0564395228555362,-.0195636529956887,-.00187057076364506,-1.51798901731387e-007,-4.64901277906996e-005,-.000893974161901044,-.00587047204848792,-.0208983721752537,-.0486052546469995,-.0804237060180922,-.0978389887959829,-.0868499819149511,-.0528932916734667,-.0183344215497376,-.00175303829641842,-1.32083844163853e-007,-4.04521687853206e-005,-.000777868235805777,-.00510803771553511,-.0181841719682459,-.0422925910998948,-.0699785843745455,-.085132037201067,-.0755702402721397,-.0460237143683289,-.0159532173895116,-.00152536042432723,-1.04070632906319e-007,-3.18728065084108e-005,-.000612892819182374,-.00402469144754472,-.0143275530598965,-.0333228999419048,-.0551370654893481,-.0670766742761412,-.0595428061909519,-.0362627020233443,-.0125697540159339,-.00120185194317225,-6.95237231559063e-008,-2.1292425288575e-005,-.000409439142389331,-.00268866947546018,-.00957143052386794,-.0222611509665594,-.0368339652566651,-.0448101448253052,-.0397771922581837,-.024225067014107,-.00839714407356662,-.000802889532215207,-3.06697796211909e-008,-9.39296633663941e-006,-.000180620480252622,-.00118608291592308,-.00422235247914266,-.00982031115806883,-.0162489801425508,-.0197676016789817,-.0175473588745023,-.0106866754670937,-.00370432632910279,-.000354187662792186},
			{-4.89764458734161e-010,-1.49995895979437e-007,-2.8843210756594e-006,-1.89405096647495e-005,-6.74265743697526e-005,-.000156820147987287,-.00025947930056204,-.000315668024236135,-.00028021305752868,-.000170655084270396,-5.91542359272188e-005,-5.6560083280767e-006,-5.77303387673914e-009,-1.76805681469661e-006,-3.3998553762376e-005,-.000223258756300761,-.000794781840714042,-.00184849678399644,-.00305857798732252,-.00372089514708473,-.00330297440936967,-.00201157426832706,-.000697272743806459,-6.66694512082365e-005,-2.07381685017084e-008,-6.35129828208852e-006,-.000122131231479417,-.000802000786155484,-.00285505335439599,-.00664025858844208,-.0109871701831305,-.0133663775728643,-.0118651027035077,-.00722607333006517,-.00250477651118074,-.000239493192417188,-4.72058591990341e-008,-1.44573274351898e-005,-.000278004767708733,-.00182557761480488,-.00649889823405786,-.0151150817366226,-.0250098656792164,-.030425605696751,-.0270082851124352,-.0164485595800252,-.00570157038226907,-.00054515334463801,-8.31059491191454e-008,-2.54521353623046e-005,-.000489427593779599,-.00321393070570632,-.0114413150217837,-.0266101122837137,-.044029886541206,-.0535643007426736,-.0475481053927164,-.0289577009874395,-.0100376187644472,-.000959742855872972,-1.22708370279532e-007,-3.75808240390211e-005,-.000722654190693695,-.00474546290931794,-.0168934370530446,-.0392906109117239,-.0650114182959186,-.0790892603834303,-.0702061715732879,-.0427568944566942,-.0148208383771267,-.00141708846333921,-1.57842742379509e-007,-4.83411222371217e-005,-.000929567551026009,-.00610420363142845,-.0217304363719754,-.0505404624146264,-.0836257585887017,-.101734427107537,-.0903078952792718,-.0549992267137261,-.0190644026033435,-.00182283513943086,-1.79817461292618e-007,-5.50711280460104e-005,-.00105898107575704,-.00695402514977553,-.0247557273922317,-.0575766583052787,-.0952680584574148,-.115897798864594,-.102880475966099,-.062656167598386,-.0217185308967278,-.0020766085420598,-1.81570069821245e-007,-5.56078841986148e-005,-.00106930253870999,-.00702180323816678,-.0249970115180785,-.0581378348543761,-.0961965979360501,-.11702740813212,-.103883210618848,-.0632668520831664,-.0219302126889477,-.00209684841095391,-1.59517455322627e-007,-4.88540219870558e-005,-.000939430271260956,-.00616896928789685,-.0219609965010198,-.0510766971839986,-.0845130286536747,-.102813830311442,-.0912660628757668,-.0555827690131719,-.0192666760896712,-.00184217543696469,-1.14645586997212e-007,-3.51115056126793e-005,-.000675170843678735,-.00443365338137005,-.0157834221327753,-.0367089477367546,-.0607397213008226,-.0738925524084332,-.0655931435851624,-.0399474726296093,-.0138470074345037,-.00132397601187602,-5.2631852118241e-008,-1.61190990378048e-005,-.000309959527704379,-.00203541536332734,-.00724590244919067,-.0168524577290889,-.0278845798860605,-.0339228224379417,-.0301127040609837,-.0183392097943569,-.00635692717584237,-.000607815019228362},
			{-4.34129792523551e-010,-6.91363066088356e-007,-3.19037343548278e-005,-.000375745321421736,-.00204903932812416,-.00659447754217011,-.0140356226122099,-.020756280852719,-.0213815233314377,-.0145196103881163,-.00541459382780606,-.000538770430325433,-9.84106174935942e-010,-1.56721486103326e-006,-7.23209107571181e-005,-.000851757463741354,-.00464486033951284,-.0149486770580804,-.0318166205582331,-.0470513300576056,-.0484686596091265,-.032913747194243,-.0122740602291765,-.0012213105768994,-1.47311662585924e-009,-2.34597681315439e-006,-.000108257766028679,-.00127500275172926,-.00695292963828415,-.0223767976156558,-.0476265609511484,-.0704315228803041,-.0725531351410981,-.04926885883459,-.0183731416903004,-.00182818984576065,-1.86964277444457e-009,-2.97745509129004e-006,-.000137398048789906,-.00161820160082515,-.00882448438314045,-.0284000718224987,-.0604464398750644,-.0893899271345745,-.0920826243480438,-.0625308032698983,-.0233187318655643,-.00232029282368979,-2.14870881134135e-009,-3.42187506483889e-006,-.000157906313511514,-.00185973710365748,-.0101416418199092,-.0326391144885199,-.0694687775381858,-.102732418569255,-.105827032314693,-.0718642565322132,-.0267993251511533,-.00266662364773693,-2.29276901154607e-009,-3.65129470714422e-006,-.000168493143619897,-.00198442319330673,-.0108215882804389,-.0348274041920398,-.0741263123085118,-.109620114430362,-.112922206580213,-.0766823962117034,-.0285960861297693,-.00284540745247398,-2.29276901154607e-009,-3.65129470714422e-006,-.000168493143619897,-.00198442319330673,-.0108215882804389,-.0348274041920399,-.0741263123085118,-.109620114430362,-.112922206580213,-.0766823962117034,-.0285960861297693,-.00284540745247398,-2.14870881134135e-009,-3.42187506483889e-006,-.000157906313511514,-.00185973710365748,-.0101416418199092,-.0326391144885199,-.0694687775381858,-.102732418569255,-.105827032314693,-.0718642565322132,-.0267993251511533,-.00266662364773693,-1.86964277444457e-009,-2.97745509129004e-006,-.000137398048789906,-.00161820160082515,-.00882448438314046,-.0284000718224987,-.0604464398750644,-.0893899271345745,-.0920826243480438,-.0625308032698983,-.0233187318655643,-.00232029282368979,-1.47311662585924e-009,-2.34597681315439e-006,-.000108257766028679,-.00127500275172926,-.00695292963828415,-.0223767976156558,-.0476265609511484,-.0704315228803041,-.0725531351410981,-.04926885883459,-.0183731416903004,-.00182818984576065,-9.84106174935942e-010,-1.56721486103326e-006,-7.23209107571182e-005,-.000851757463741354,-.00464486033951284,-.0149486770580804,-.0318166205582331,-.0470513300576056,-.0484686596091265,-.032913747194243,-.0122740602291765,-.0012213105768994,-4.34129792523551e-010,-6.91363066088356e-007,-3.19037343548278e-005,-.000375745321421736,-.00204903932812416,-.00659447754217011,-.0140356226122099,-.020756280852719,-.0213815233314377,-.0145196103881163,-.00541459382780606,-.000538770430325433},
			{-.00423058507690693,-.00885514027394949,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394863,-.00435422423400534,-.00213112724015142,-.000790234334926298,-.000193564800479007,-2.24537546398006e-005,-3.66334363412211e-007,-.00959009257018474,-.020073255449394,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134588,-.00987036372424941,-.00483094114415894,-.00179134098152266,-.000438781946510387,-5.08992448158822e-005,-8.30424254068273e-007,-.0143554884304917,-.0300478211505447,-.0388614993834644,-.0396707255608121,-.0338552805420154,-.0244472036457215,-.0147750286257641,-.00723147552495667,-.00268147304597068,-.000656816303965783,-7.61914981245153e-005,-1.2430688947403e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971213,-.0429682752411084,-.03102777936875,-.0187520967637297,-.00917800785523132,-.00340325851820642,-.000833614824033866,-9.670034364015e-005,-1.57767194829697e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.057864283158913,-.0493817925438395,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011006,-.000958041580001006,-.000111134000184102,-1.81315792676965e-006,-.0223429825183335,-.0467666388316429,-.0604843092288455,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102643,-.0112551190451086,-.00417346338855665,-.00102227348573477,-.000118584980154756,-1.93472111511538e-006,-.0223429825183335,-.0467666388316429,-.0604843092288455,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102643,-.0112551190451086,-.00417346338855665,-.00102227348573477,-.000118584980154756,-1.93472111511538e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.057864283158913,-.0493817925438395,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011006,-.000958041580001006,-.000111134000184102,-1.81315792676965e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971213,-.0429682752411084,-.03102777936875,-.0187520967637297,-.00917800785523132,-.00340325851820642,-.000833614824033866,-9.670034364015e-005,-1.57767194829697e-006,-.0143554884304917,-.0300478211505447,-.0388614993834644,-.0396707255608121,-.0338552805420154,-.0244472036457215,-.0147750286257641,-.00723147552495667,-.00268147304597068,-.000656816303965783,-7.61914981245153e-005,-1.2430688947403e-006,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134588,-.00987036372424941,-.00483094114415894,-.00179134098152266,-.000438781946510386,-5.08992448158822e-005,-8.30424254068273e-007,-.00423058507690692,-.00885514027394949,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394863,-.00435422423400534,-.00213112724015142,-.000790234334926298,-.000193564800479007,-2.24537546398006e-005,-3.66334363412211e-007},
			{-6.75580403638936e-005,-.00014140737349095,-.000182885225860587,-.00018669350692448,-.000159325572268953,-.000115050433754236,-6.95324290155738e-005,-3.40318838868514e-005,-1.26192198302079e-005,-3.09102839585264e-006,-3.58563091085661e-007,-5.84997849237861e-009,-.000796331478757963,-.00166682074010411,-.00215573544715162,-.00220062505725282,-.0018780291418385,-.00135614179377857,-.000819604324242625,-.000401146336906179,-.000148747387195263,-3.64350890004165e-005,-4.22651508264995e-006,-6.89558489033362e-008,-.00286062003831949,-.00598763321131698,-.00774393601400238,-.00790519061914258,-.00674634613725228,-.00487159743595245,-.00294422186735428,-.00144101706420306,-.000534337481072711,-.000130883869937062,-1.51826896966895e-005,-2.4770649960978e-007,-.00651156956023967,-.013629524241044,-.0176273595758088,-.0179944200606735,-.0153565666050425,-.0110891153486868,-.00670187065504814,-.0032801570028024,-.00121630123191481,-.000297928215558894,-3.45600389940658e-005,-5.63849124010232e-007,-.0114636229006531,-.0239947872427992,-.031032979229028,-.0316791894770017,-.0270352465684925,-.0195223955581165,-.0117986481151247,-.0057747187659129,-.00214129919480704,-.000524502838069343,-6.0842973540354e-005,-9.92656788923096e-007,-.0169263752901921,-.0354290068157069,-.0458211036210322,-.0467752519970179,-.0399183341468002,-.0288253893768114,-.0174210498412452,-.00852654155446687,-.00316169103729199,-.000774443817183375,-8.98364341570052e-005,-1.46568685042074e-006,-.0217728056224891,-.045573187736373,-.0589407930194454,-.0601681371359018,-.0513479179830939,-.0370787950245641,-.0224091174530978,-.0109678964760433,-.00406695959489998,-.000996185799262211,-.00011555877647658,-1.88534841928859e-006,-.0248039952501605,-.0519178442938048,-.0671464750773758,-.0685446888933965,-.0584965270825159,-.042240870176189,-.0255288938184798,-.012494836761645,-.00463315790456963,-.00113487385418326,-.000131646761218445,-2.1478248622506e-006,-.0250457498234202,-.0524238666569454,-.0678009247845504,-.0692127664287096,-.0590666692229003,-.042652574961671,-.0257777136867565,-.012616618914028,-.004678315432652,-.00114593501354709,-.000132929869292705,-2.16875884799657e-006,-.0220038152896698,-.0460567196758888,-.0595661553735154,-.0608065216222948,-.0518927198634471,-.0374722013795649,-.0226468783946324,-.0110842659581639,-.0041101100734715,-.00100675533972143,-.000116784856153978,-1.90535198329458e-006,-.015814196101363,-.0331010775700317,-.042810342192032,-.0437017964620066,-.0372954252410452,-.0269313631824755,-.0162763671345931,-.00796628917278541,-.00295394620634634,-.000723557535766244,-8.39335630014821e-005,-1.36938115091738e-006,-.00726003025826053,-.0151961454884669,-.0196535048439074,-.0200627564386384,-.0171216996431039,-.0123637338469658,-.00747220517149892,-.00365718877328773,-.00135610679808757,-.00033217304057603,-3.85324807640305e-005,-6.28659751468172e-007},
			{-8.04113589360719e-007,-1.68310966462922e-006,-2.17680226684608e-006,-2.22213055847631e-006,-1.89638209018582e-006,-1.36939462342768e-006,-8.27613867594724e-007,-4.05066519951766e-007,-1.50201013794113e-007,-3.67911491336473e-008,-4.26781849550599e-009,-6.96297165792815e-011,-4.92865836277514e-005,-.000103162944064127,-.000133422875058339,-.000136201184821656,-.00011623506394466,-8.39343887728836e-005,-5.07269876251226e-005,-2.48277670898173e-005,-9.20627996502583e-006,-2.25504216385422e-006,-2.61587661208472e-007,-4.26781849550599e-009,-.000424879842112779,-.000889326306666056,-.001150185018243,-.00117413571083906,-.00100201580190209,-.000723564654409836,-.000437296986453802,-.000214030208318087,-7.93636419908757e-005,-1.94398127850062e-005,-2.25504216385422e-006,-3.67911491336473e-008,-.00173458520673544,-.0036307023836027,-.00469566620937377,-.00479344565888028,-.00409076076250558,-.00295397526843083,-.00178527858577817,-.000873785000712069,-.000324004543650437,-7.93636419908757e-005,-9.20627996502583e-006,-1.50201013794113e-007,-.00467788049829846,-.0097913851734879,-.0126634109999641,-.0129271055006463,-.0110320841661877,-.00796636754827121,-.00481459189667545,-.00235644913761801,-.000873785000712069,-.000214030208318087,-2.48277670898173e-005,-4.05066519951766e-007,-.00955763703157626,-.0200053219740424,-.025873317192124,-.0264120860481604,-.0225402628819985,-.016276527267894,-.0098369596701605,-.00481459189667545,-.00178527858577817,-.000437296986453802,-5.07269876251226e-005,-8.27613867594725e-007,-.0158143516876436,-.0331014032314506,-.0428107633770146,-.0437022264174669,-.037295792168106,-.0269316281438179,-.016276527267894,-.00796636754827121,-.00295397526843083,-.000723564654409835,-8.39343887728836e-005,-1.36939462342768e-006,-.0219002271480226,-.0458398968231796,-.0592857336712341,-.0605202606038321,-.0516484226656698,-.037295792168106,-.0225402628819985,-.0110320841661877,-.00409076076250558,-.00100201580190209,-.00011623506394466,-1.89638209018582e-006,-.0256621090417623,-.0537139830919874,-.0694694603763257,-.0709160465028163,-.0605202606038321,-.0437022264174669,-.0264120860481604,-.0129271055006463,-.00479344565888028,-.00117413571083906,-.000136201184821656,-2.22213055847631e-006,-.0251386386461752,-.0526182945056766,-.0680523825420277,-.0694694603763257,-.0592857336712341,-.0428107633770146,-.025873317192124,-.0126634109999641,-.00469566620937377,-.001150185018243,-.000133422875058339,-2.17680226684608e-006,-.0194372664460252,-.0406846140173892,-.0526182945056766,-.0537139830919874,-.0458398968231796,-.0331014032314506,-.0200053219740424,-.0097913851734879,-.0036307023836027,-.000889326306666056,-.000103162944064127,-1.68310966462922e-006,-.00928624582089677,-.0194372664460252,-.0251386386461752,-.0256621090417623,-.0219002271480226,-.0158143516876436,-.00955763703157627,-.00467788049829846,-.00173458520673544,-.000424879842112779,-4.92865836277514e-005,-8.04113589360719e-007},
			{-8.77197535304017e-009,-1.83608344516711e-008,-2.37464657802842e-008,-2.42409464882796e-008,-2.06873968742175e-008,-1.49385560003323e-008,-9.02833697182859e-009,-4.41882039598848e-009,-1.63852421901125e-009,-4.01350079989265e-010,-4.65571023163414e-011,-7.59581936873143e-013,-2.79578126521843e-006,-5.85191760211322e-006,-7.56841207045341e-006,-7.72601168100707e-006,-6.59343355166069e-006,-4.76117787775939e-006,-2.8774881763854e-006,-1.40835499192186e-006,-5.22226195326764e-007,-1.2791725800267e-007,-1.48385590679724e-008,-2.42091987612831e-010,-5.7837794064797e-005,-.000121061690113565,-.000156571711876826,-.000159832057717645,-.000136401819657762,-9.84969850916432e-005,-5.95281078102025e-005,-2.91353787244659e-005,-1.08035673306481e-005,-2.64629143836613e-006,-3.06973057681128e-007,-5.00828398075496e-009,-.000423492023911399,-.000886421430715201,-.00114642807911561,-.00117030053969446,-.000998742839454559,-.000721201218685817,-.000435868609164376,-.00021333110473795,-7.91044103306138e-005,-1.93763150067161e-005,-2.24767634356886e-006,-3.66709753307118e-008,-.00174950826221703,-.00366193819311915,-.00473606415990016,-.00483468482962708,-.00412595456537195,-.00297938903112802,-.00180063776864345,-.000881302384115301,-.000326792032995098,-8.00464265713343e-005,-9.28548381512419e-006,-1.51493229393287e-007,-.00494625775323145,-.0103531321176025,-.0133899305174026,-.0136687535803171,-.011665011934605,-.00842341040243774,-.00509081249643689,-.00249164227715485,-.000923915400575431,-.000226309453117529,-2.62521745709628e-005,-4.28305814057175e-007,-.010527567456728,-.0220355068809891,-.0284989994042999,-.0290924437312573,-.0248276992732258,-.0179283057316278,-.0108352363826531,-.00530318747211388,-.00196645265756991,-.000481674864645602,-5.58748760112742e-005,-9.11601977606186e-007,-.0177221024244434,-.0370945625876605,-.0479752030574144,-.0489742069762943,-.0417949380321651,-.0301805019800266,-.0182400321589689,-.00892738345711226,-.00331032553849853,-.000810851255203673,-9.40597416919936e-005,-1.53459036799021e-006,-.0240985158662193,-.0504411882777659,-.0652366837960255,-.0665951293806949,-.0568327590696347,-.0410394482774791,-.0248027967481441,-.0121394565234153,-.00450138085207253,-.00110259558209743,-.00012790244200432,-2.08673606807239e-006,-.02632238051474,-.0550960133326791,-.0712568700882474,-.0727406760531766,-.0620774124945382,-.044826659852073,-.027091654003041,-.0132597125742255,-.00491677828990747,-.00120434555501086,-.000139705563856799,-2.27930471413928e-006,-.0218959049522221,-.0458308499302774,-.0592740331282503,-.0605083164164748,-.0516382294108676,-.0372884315268124,-.0225358143697108,-.011029906890695,-.00408995341613109,-.00100181804557465,-.000116212123967743,-1.89600782307658e-006,-.0108863317310921,-.0227864450885946,-.0294702040898467,-.0300838721412506,-.025673791359598,-.0185392764637387,-.0112044855690737,-.00548391243189871,-.00203346652033204,-.000498089647453146,-5.77790109816517e-005,-9.42668054679449e-007},
			{-.000206393222578878,-.00224638876284317,-.00697208388856876,-.0127648935683974,-.0166874796528932,-.0166745014700866,-.0129629077002378,-.00771242584002074,-.00331870648655728,-.000906410211953001,-.000113117990564565,-1.92058726559064e-006,-.000467862027215704,-.00509222147590975,-.0158046531822386,-.0289360711920697,-.0378279769169244,-.0377985573514871,-.0293849390956891,-.0174828957229149,-.00752300257311598,-.00205469401540682,-.000256421270615505,-4.35367906124192e-006,-.000700346515907722,-.0076225882021403,-.0236581153148544,-.0433146429173246,-.0566250097176437,-.0565809713282148,-.0439865569734266,-.0261702903746801,-.0112612444155839,-.00307568836802217,-.000383839108612685,-6.51705798837019e-006,-.000888862280208502,-.00967439828220347,-.0300262881966959,-.0549738613606283,-.0718670459711192,-.071811153548789,-.0558266378740331,-.0332146779455411,-.0142924897345245,-.00390358674443833,-.000487158995675827,-8.27128698753606e-006,-.00102153536475457,-.0111184153023947,-.0345080626641537,-.063179352715725,-.0825940425807241,-.0825298075610779,-.0641594160912017,-.0381723568496403,-.0164258108812803,-.00448624269205744,-.000559873169805834,-9.50587324711357e-006,-.00109002421181751,-.0118638495492168,-.0368216560137206,-.067415212946293,-.0881315608554635,-.0880630191983885,-.0684609846789651,-.0407316228334812,-.0175270795090219,-.00478702287082038,-.000597409871152132,-1.01431945983696e-005,-.00109002421181751,-.0118638495492168,-.0368216560137206,-.067415212946293,-.0881315608554635,-.0880630191983885,-.0684609846789651,-.0407316228334812,-.0175270795090219,-.00478702287082038,-.000597409871152132,-1.01431945983696e-005,-.00102153536475457,-.0111184153023947,-.0345080626641537,-.063179352715725,-.0825940425807241,-.0825298075610779,-.0641594160912017,-.0381723568496402,-.0164258108812803,-.00448624269205744,-.000559873169805834,-9.50587324711357e-006,-.000888862280208502,-.00967439828220347,-.0300262881966959,-.0549738613606283,-.0718670459711192,-.071811153548789,-.0558266378740331,-.0332146779455411,-.0142924897345245,-.00390358674443833,-.000487158995675827,-8.27128698753606e-006,-.000700346515907722,-.0076225882021403,-.0236581153148544,-.0433146429173246,-.0566250097176437,-.0565809713282148,-.0439865569734266,-.0261702903746801,-.0112612444155839,-.00307568836802217,-.000383839108612685,-6.51705798837019e-006,-.000467862027215704,-.00509222147590975,-.0158046531822386,-.0289360711920697,-.0378279769169244,-.0377985573514871,-.0293849390956891,-.0174828957229149,-.00752300257311598,-.00205469401540682,-.000256421270615505,-4.35367906124192e-006,-.000206393222578878,-.00224638876284317,-.00697208388856876,-.0127648935683974,-.0166874796528932,-.0166745014700866,-.0129629077002378,-.00771242584002074,-.00331870648655728,-.000906410211953001,-.000113117990564565,-1.92058726559064e-006},
			{-3.29588494374136e-006,-3.58724904274172e-005,-.000111336922926592,-.000203842064166924,-.000266481681179921,-.000266274433108618,-.000207004143752805,-.000123159413319067,-5.29962883716334e-005,-1.44744276633774e-005,-1.80637657239712e-006,-3.06697796211909e-008,-3.88498085043389e-005,-.000422842243423972,-.00131236927532137,-.00240276141102928,-.0031411176240886,-.00313867471481826,-.00244003400654881,-.00145172531948917,-.000624686750242411,-.000170615404521307,-2.1292425288575e-005,-3.61515976889952e-007,-.000139558140870839,-.00151895413763271,-.00471435570086283,-.00863131450035304,-.0112836730154145,-.0112748974798531,-.00876520690127838,-.00521495715018533,-.00224402963223718,-.000612892819182374,-7.6487668853577e-005,-1.29865498884588e-006,-.000317673277053615,-.00345756353291146,-.0107311892759846,-.0196472806638003,-.0256847888746645,-.025664813306616,-.0199520571355211,-.0118706978844523,-.00510803771553511,-.0013951151049833,-.00017410728078841,-2.95610118832487e-006,-.000559264033051857,-.00608704308994404,-.0188922664493339,-.034589051759252,-.0452180893128947,-.0451829223109582,-.0351256109572267,-.0208983721752537,-.00899270407717994,-.00245610114713186,-.000306515993225909,-5.20421827112933e-006,-.000825769740663992,-.00898769757526597,-.0278949852742899,-.051071748964773,-.0667658345226861,-.0667139094133409,-.0518639943508112,-.0308570591913224,-.0132779912077635,-.00362650534890298,-.000452579850083181,-7.68418084863826e-006,-.0010622075751108,-.0115610926110904,-.0358819937409303,-.0656948250258218,-.0858825065829686,-.0858157140598609,-.0667139094133409,-.0396921810095751,-.0170798010009427,-.0046648614778342,-.000582164399381254,-9.88434754146089e-006,-.00121008711989414,-.0131706171076067,-.0408774513376,-.0748407971004666,-.0978389887959829,-.0977628974803007,-.0760017574629395,-.0452180893128947,-.0194576349160754,-.00531429912823647,-.000663212782377944,-1.12604371581811e-005,-.0012218813527315,-.0132989858194334,-.0412758674276049,-.0755702402721397,-.0987925860994006,-.0987157531522446,-.0767425160486888,-.0456588118567917,-.0196472806638003,-.00536609546608301,-.000669676850830108,-1.13701881137203e-005,-.0010734776072168,-.01168375590964,-.0362627020233444,-.0663918477213734,-.0867937207648416,-.0867262195724516,-.067421744603583,-.040113315413826,-.017261017846082,-.00471435570086283,-.000588341168993646,-9.98922055945599e-006,-.000771510993319342,-.00839714407356662,-.0260620930240078,-.0477159840498455,-.06237886032367,-.0623303470508818,-.0484561734690455,-.0288295383268322,-.0124055359279088,-.00338821902309026,-.000422842243423972,-7.17927735473982e-006,-.000354187662792186,-.00385498697918716,-.0119646666030396,-.0219056021427923,-.0286370809191255,-.0286148092951926,-.0222454106000665,-.0132351539871473,-.00569517196000836,-.00155547411146173,-.000194119730276028,-3.29588494374136e-006},
			{-3.92294663599167e-008,-4.26974448589619e-007,-1.32519433994785e-006,-2.4262422795302e-006,-3.1718140426088e-006,-3.16934726012697e-006,-2.46387911966847e-006,-1.46591223424896e-006,-6.30791470989705e-007,-1.72282735226472e-007,-2.15005044744533e-008,-3.65048874112027e-010,-2.40449409138482e-006,-2.61705711055768e-005,-8.12252180824206e-005,-.000148711817078381,-.000194410192951716,-.000194258996301528,-.00015101869423293,-8.98502486218331e-005,-3.86631403796111e-005,-1.05597362732147e-005,-1.31783174148499e-006,-2.23749630651595e-008,-2.07281778267435e-005,-.000225605982417451,-.000700209981991998,-.00128198484678476,-.00167593219100564,-.00167462878541062,-.00130187150820567,-.000774562906136249,-.000333299404644144,-9.10312452249138e-005,-1.1360498152647e-005,-1.92885569959679e-007,-8.46234324557744e-005,-.000921043459502259,-.00285862908989082,-.00523374312001741,-.00684204543937499,-.00683672424538222,-.00531493103550084,-.00316217722165629,-.00136070521452511,-.000371637897746933,-4.63795880236092e-005,-7.87461355147049e-007,-.000228214966233319,-.00248389714184165,-.00770923516443232,-.0141145126680233,-.0184518297544813,-.0184374794017228,-.0143334626308517,-.00852785271078683,-.00366958991823735,-.00100224403364473,-.00012507783964276,-2.12364898657175e-006,-.000466278651886231,-.00507498798101565,-.0157511658366454,-.0288381434728016,-.0376999565136218,-.0376706365121829,-.0292854922824927,-.0174237287374253,-.00749754263838111,-.00204774035893469,-.000255553469660892,-4.3389449994523e-006,-.000771518583621402,-.00839722668659549,-.0260623494288551,-.0477164534906379,-.0623794740212014,-.0623309602711284,-.0484566501919911,-.0288298219584409,-.0124056579764159,-.00338825235716593,-.000422846403442921,-7.17934798611665e-006,-.00106842395843715,-.0116287518756503,-.0360919868089839,-.0660792924543047,-.0863851188731245,-.0863179354585031,-.0671043408542665,-.0399244725296093,-.0171797575373556,-.00469216176055673,-.000585571414309697,-9.94219394991092e-006,-.00125195103863209,-.0136262649987847,-.0422916390211788,-.077429973529657,-.101223805813716,-.101145082058928,-.0786310982319312,-.0467824447922306,-.0201307871491398,-.00549815149985173,-.000686157151895624,-1.16500008667718e-005,-.00122641302441892,-.0133483086422873,-.0414289499502059,-.0758505125894862,-.0991589846571268,-.0990818667545695,-.0770271360622618,-.0458281496934892,-.0197201478250184,-.00538599705705277,-.000672160525384795,-1.14123574777427e-005,-.000948266016470748,-.0103209499661349,-.0320329811846719,-.0586478306968206,-.0766700071720531,-.0766103793918513,-.0595575992912496,-.0354344548588437,-.015247674029803,-.00416446814598573,-.000519716417833643,-8.82406705447926e-006,-.00045303856676568,-.00493088257841441,-.0153039080100614,-.028019278030959,-.0366294579367253,-.0366009704831913,-.0284539242725722,-.0169289781185245,-.00728464826218446,-.00198959431997835,-.000248296972548114,-4.21573969958762e-006},
			{-5.21125166412433e-006,-.000294934791565546,-.00219672054164462,-.00721329026542284,-.0144452225630722,-.0199731094854017,-.0199731094854017,-.0144452225630722,-.00721329026542284,-.00219672054164462,-.000294934791565546,-5.21125166412433e-006,-1.18131144881787e-005,-.000668572290088461,-.00497963117683019,-.0163514313324231,-.032745121315009,-.0452759997488288,-.0452759997488288,-.032745121315009,-.0163514313324231,-.00497963117683019,-.000668572290088461,-1.18131144881787e-005,-1.76831482201068e-005,-.00100079135890211,-.0074540508575846,-.0244765920241849,-.049016441369304,-.0677740163420273,-.0677740163420273,-.049016441369304,-.0244765920241849,-.0074540508575846,-.00100079135890211,-1.76831482201068e-005,-2.24430094120152e-005,-.00127017935990404,-.00946049490012139,-.0310650783636037,-.0622104413366804,-.0860170863084584,-.0860170863084584,-.0622104413366804,-.0310650783636037,-.00946049490012139,-.00127017935990404,-2.24430094120152e-005,-2.57928908857684e-005,-.00145976847551561,-.0108725843403854,-.0357019044050888,-.0714960768359992,-.0988561418273082,-.0988561418273082,-.0714960768359992,-.0357019044050888,-.0108725843403854,-.00145976847551561,-2.57928908857684e-005,-2.75221754706548e-005,-.00155763866515012,-.0116015368483061,-.0380955388841488,-.0762895319046813,-.105483952681873,-.105483952681873,-.0762895319046813,-.0380955388841488,-.0116015368483061,-.00155763866515012,-2.75221754706548e-005,-2.75221754706548e-005,-.00155763866515012,-.0116015368483061,-.0380955388841488,-.0762895319046813,-.105483952681873,-.105483952681873,-.0762895319046813,-.0380955388841488,-.0116015368483061,-.00155763866515012,-2.75221754706548e-005,-2.57928908857684e-005,-.00145976847551561,-.0108725843403854,-.0357019044050888,-.0714960768359992,-.0988561418273082,-.0988561418273082,-.0714960768359992,-.0357019044050888,-.0108725843403854,-.00145976847551561,-2.57928908857684e-005,-2.24430094120152e-005,-.00127017935990404,-.00946049490012139,-.0310650783636037,-.0622104413366804,-.0860170863084584,-.0860170863084584,-.0622104413366804,-.0310650783636037,-.00946049490012139,-.00127017935990404,-2.24430094120152e-005,-1.76831482201068e-005,-.00100079135890211,-.0074540508575846,-.0244765920241849,-.049016441369304,-.0677740163420273,-.0677740163420273,-.049016441369304,-.0244765920241849,-.0074540508575846,-.00100079135890211,-1.76831482201068e-005,-1.18131144881787e-005,-.000668572290088461,-.00497963117683019,-.0163514313324231,-.032745121315009,-.0452759997488288,-.0452759997488288,-.032745121315009,-.0163514313324231,-.00497963117683019,-.000668572290088461,-1.18131144881787e-005,-5.21125166412433e-006,-.000294934791565546,-.00219672054164462,-.00721329026542284,-.0144452225630722,-.0199731094854017,-.0199731094854017,-.0144452225630722,-.00721329026542284,-.00219672054164462,-.000294934791565546,-5.21125166412433e-006},
			{-8.32182650623191e-008,-4.7098016450767e-006,-3.50793406430393e-005,-.000115188737748337,-.000230675169348194,-.000318949769921382,-.000318949769921382,-.000230675169348194,-.000115188737748337,-3.50793406430393e-005,-4.7098016450767e-006,-8.32182650623191e-008,-9.80924309227986e-007,-5.55161648928631e-005,-.000413493095087696,-.00135777203384397,-.00271905309464788,-.00375957829094687,-.00375957829094687,-.00271905309464788,-.00135777203384397,-.000413493095087696,-5.55161648928631e-005,-9.80924309227986e-007,-3.52372323574208e-006,-.000199427823688284,-.00148536967967095,-.00487745366231389,-.00976751261916201,-.0135053370130334,-.0135053370130334,-.00976751261916201,-.00487745366231389,-.00148536967967095,-.000199427823688284,-3.52372323574208e-006,-8.02097749900636e-006,-.000453953383811275,-.00338111593370866,-.011102445754264,-.0222335846768227,-.0307419161639197,-.0307419161639197,-.0222335846768227,-.011102445754264,-.00338111593370866,-.000453953383811275,-8.02097749900635e-006,-1.41209366639782e-005,-.000799185259152228,-.00595245703648738,-.0195458637467361,-.039142241836921,-.0541211655479415,-.0541211655479415,-.039142241836921,-.0195458637467361,-.00595245703648738,-.000799185259152228,-1.41209366639782e-005,-2.08499769658255e-005,-.00118002046473714,-.00878897732169731,-.0288600408453219,-.0577946675996675,-.0799114875939334,-.0799114875939334,-.0577946675996675,-.0288600408453219,-.00878897732169731,-.00118002046473714,-2.08499769658255e-005,-2.68198292857976e-005,-.0015178888432285,-.0113054739461356,-.0371233680459864,-.0743426777493866,-.102792077840484,-.102792077840484,-.0743426777493866,-.0371233680459864,-.0113054739461356,-.0015178888432285,-2.68198292857976e-005,-3.05536608257741e-005,-.00172920790781427,-.0128794114512791,-.0422916486119502,-.0846925957890945,-.117102694743021,-.117102694743021,-.0846925957890945,-.0422916486119502,-.0128794114512791,-.00172920790781427,-3.0553660825774e-005,-3.08514550786741e-005,-.00174606180234275,-.0130049418986058,-.0427038483144384,-.0855180604832598,-.11824404764644,-.11824404764644,-.0855180604832598,-.0427038483144384,-.0130049418986058,-.00174606180234275,-3.08514550786741e-005,-2.7104387920297e-005,-.00153399365776509,-.0114254251283075,-.037517247321153,-.0751314542415842,-.103882702728358,-.103882702728358,-.0751314542415842,-.037517247321153,-.0114254251283075,-.00153399365776509,-2.7104387920297e-005,-1.94799901806221e-005,-.00110248501011246,-.00821148110642986,-.0269637378113508,-.0539971607248856,-.0746607536401705,-.0746607536401705,-.0539971607248856,-.0269637378113508,-.00821148110642986,-.00110248501011246,-1.94799901806221e-005,-8.94293438853622e-006,-.00050613224228358,-.00376975224764522,-.0123785964921314,-.0247891842373852,-.0342754906451081,-.0342754906451081,-.0247891842373852,-.0123785964921314,-.00376975224764522,-.00050613224228358,-8.94293438853622e-006},
			{-1.00782323868636e-007,-2.96594300874774e-005,-.000530130419242249,-.00312208933690556,-.00957752805114946,-.0183245877738932,-.0235713601331056,-.0207230264243475,-.0120086359724961,-.00407774946276834,-.000589001007959142,-1.08304313867452e-005,-2.28458191424719e-007,-6.72334145152776e-005,-.00120172498658768,-.00707730141552942,-.0217107986092594,-.0415390518964802,-.0534326863949788,-.0469759473290577,-.0272217503074272,-.00924363749104905,-.00133517565243957,-2.45509398075641e-005,-3.41980945423285e-007,-.000100642251068274,-.00179887201456491,-.0105940706876594,-.0324990729725514,-.0621801483718135,-.0799838276575082,-.0703186818540462,-.0407484618877329,-.0138368769735303,-.00199863541368309,-3.67505036876191e-005,-4.34033661954923e-007,-.00012773262768919,-.00228308336566367,-.0134457295270728,-.0412470105167922,-.0789174890001959,-.101513473425809,-.0892467121265429,-.0517169285858085,-.0175614181527209,-.00253661807513852,-4.66428200392335e-005,-4.98818259086074e-007,-.00014679821533987,-.00262386024318778,-.0154526618157401,-.0474036089408846,-.0906968466390758,-.116665545847229,-.102567827047345,-.0594362846564907,-.020182665073873,-.00291523797142323,-5.3604806102002e-005,-5.32261533433169e-007,-.000156640302913549,-.00279977697510962,-.0164886856562537,-.0505817842982924,-.0967776174795907,-.124487388343073,-.109444487868491,-.0634211908577612,-.0215358120223382,-.00311069012556852,-5.7198740774876e-005,-5.32261533433169e-007,-.000156640302913549,-.00279977697510962,-.0164886856562537,-.0505817842982924,-.0967776174795907,-.124487388343073,-.109444487868491,-.0634211908577612,-.0215358120223382,-.00311069012556852,-5.7198740774876e-005,-4.98818259086074e-007,-.00014679821533987,-.00262386024318778,-.0154526618157401,-.0474036089408846,-.0906968466390758,-.116665545847229,-.102567827047345,-.0594362846564907,-.020182665073873,-.00291523797142323,-5.3604806102002e-005,-4.34033661954923e-007,-.00012773262768919,-.00228308336566367,-.0134457295270728,-.0412470105167922,-.0789174890001959,-.101513473425809,-.0892467121265429,-.0517169285858086,-.0175614181527209,-.00253661807513852,-4.66428200392335e-005,-3.41980945423285e-007,-.000100642251068274,-.00179887201456491,-.0105940706876594,-.0324990729725514,-.0621801483718135,-.0799838276575082,-.0703186818540462,-.0407484618877329,-.0138368769735303,-.00199863541368309,-3.67505036876191e-005,-2.28458191424719e-007,-6.72334145152776e-005,-.00120172498658768,-.00707730141552942,-.0217107986092594,-.0415390518964802,-.0534326863949788,-.0469759473290577,-.0272217503074272,-.00924363749104905,-.00133517565243957,-2.45509398075641e-005,-1.00782323868636e-007,-2.96594300874774e-005,-.000530130419242249,-.00312208933690556,-.00957752805114946,-.0183245877738932,-.0235713601331056,-.0207230264243475,-.0120086359724961,-.00407774946276834,-.000589001007959142,-1.08304313867452e-005},
			{-.00572678909872914,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442027,-.00553806002210424,-.00260199447163722,-.000920334332232957,-.000222779313164588,-3.04257537549422e-005,-1.47072954118751e-006,-4.61452526590166e-009,-.0129817593993148,-.0261104821110847,-.0313889764798934,-.0287370569429126,-.0211332958986598,-.0125539393028941,-.00589832550261673,-.00208625787714596,-.000505006800981909,-6.89705536523267e-005,-3.33392005816974e-006,-1.0460426551641e-008,-.019432502397701,-.0390849953863391,-.0469864170136373,-.0430167368512079,-.0316346044160784,-.018792094977277,-.00882925194855102,-.0031229365722185,-.000755948833210871,-.000103242588927534,-4.99057235089002e-006,-1.56582985243485e-008,-.0246632459775863,-.0496056985105609,-.0596339852015086,-.0545957664330435,-.0401498486478861,-.023850456898108,-.0112058657268432,-.00396355298338403,-.000959431350547771,-.000131032911457859,-6.33390959846238e-006,-1.98731208257652e-008,-.028344523709382,-.0570099288159739,-.0685350545084464,-.0627448227009384,-.0461426828391663,-.0274104163597275,-.0128784721632876,-.00455515959306486,-.00110263769366935,-.000150591105035483,-7.2793196382121e-006,-2.2839416390577e-008,-.0302448825382416,-.0608321599696494,-.073129987810444,-.0669515498630354,-.0492363193955517,-.0292481479535269,-.0137419094370501,-.00486056030603151,-.00117656404705184,-.000160687486930486,-7.76736168416077e-006,-2.43706852532627e-008,-.0302448825382416,-.0608321599696494,-.073129987810444,-.0669515498630354,-.0492363193955517,-.0292481479535269,-.0137419094370501,-.00486056030603151,-.00117656404705184,-.000160687486930486,-7.76736168416077e-006,-2.43706852532627e-008,-.028344523709382,-.0570099288159739,-.0685350545084464,-.0627448227009384,-.0461426828391663,-.0274104163597275,-.0128784721632876,-.00455515959306486,-.00110263769366935,-.000150591105035483,-7.2793196382121e-006,-2.2839416390577e-008,-.0246632459775863,-.0496056985105609,-.0596339852015086,-.0545957664330435,-.0401498486478861,-.023850456898108,-.0112058657268432,-.00396355298338403,-.000959431350547771,-.000131032911457859,-6.33390959846238e-006,-1.98731208257652e-008,-.019432502397701,-.0390849953863391,-.0469864170136373,-.0430167368512079,-.0316346044160784,-.018792094977277,-.00882925194855102,-.0031229365722185,-.000755948833210871,-.000103242588927534,-4.99057235089002e-006,-1.56582985243485e-008,-.0129817593993148,-.0261104821110847,-.0313889764798934,-.0287370569429126,-.0211332958986598,-.0125539393028941,-.00589832550261673,-.00208625787714596,-.000505006800981909,-6.89705536523267e-005,-3.33392005816974e-006,-1.0460426551641e-008,-.00572678909872914,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442027,-.00553806002210424,-.00260199447163722,-.000920334332232957,-.000222779313164588,-3.04257537549422e-005,-1.47072954118751e-006,-4.61452526590166e-009},
			{-9.14508612930534e-005,-.0001839370153449,-.000221121717472685,-.000202440095184293,-.000148874898420579,-8.84370543742305e-005,-4.15511434782526e-005,-1.46967813742169e-005,-3.55755375585554e-006,-4.85867619430457e-007,-2.34860199933996e-008,-7.36891655603157e-011,-.00107796495006253,-.00216813327679371,-.00260644304233894,-.00238623588679712,-.00175484320401577,-.00104244009902437,-.000489778616311158,-.000173236369522755,-4.19341950699251e-005,-5.72710039808159e-006,-2.76838358998326e-007,-8.68601307305661e-010,-.00387231726863855,-.00778847209081444,-.00936298940145173,-.00857195072154575,-.00630383171759815,-.00374470319905967,-.00175940617887584,-.000622307975060138,-.000150638021863614,-2.05731640619964e-005,-9.94471998471375e-007,-3.12023117416446e-009,-.00881447480486433,-.0177287361159355,-.0213127769373929,-.0195121521357594,-.0143492802872553,-.00852398956746932,-.00400489948507659,-.0014165466273796,-.000342894178409679,-4.68302630441653e-005,-2.26369580965429e-006,-7.10251695871362e-009,-.0155178892424445,-.0312114527009358,-.0375211591484582,-.0343511578883119,-.0252619182805468,-.0150064897727934,-.00705062842794049,-.00249383135774626,-.000603665448054148,-8.24447118179819e-005,-3.98523810328615e-006,-1.25039862212867e-008,-.0229126184021882,-.0460846249345851,-.055401091485105,-.0507204916900297,-.0372999629412389,-.0221575220926239,-.01041045957611,-.00368221511100628,-.00089132973162475,-.000121732034018965,-5.88432089415219e-006,-1.84625022332888e-008,-.0294730548165372,-.0592797668540085,-.0712637629438524,-.0652429943039245,-.0479798438190263,-.0285017561839274,-.0133912257589351,-.00473652229125437,-.00114653897597652,-.000156586857450866,-7.56914418191853e-006,-2.37487628354249e-008,-.03357626593249,-.0675326405320358,-.0811850373450308,-.074326062928209,-.0546595527845674,-.0324697440130155,-.0152555396799836,-.00539593649305364,-.00130615905948616,-.000178386733239242,-8.62291335307587e-006,-2.70550433775344e-008,-.0339035203026146,-.0681908540387014,-.0819763152766897,-.0750504891927709,-.055192297478506,-.0327862135586896,-.0154042292942165,-.00544852851748593,-.0013188896669085,-.000180125397036526,-8.7069574240775e-006,-2.7318738012221e-008,-.0297857641982319,-.0599087257235696,-.0720198721217481,-.0659352230713762,-.0484889104016337,-.0288041600782291,-.0135333067869547,-.00478677683617009,-.00115870376501852,-.000158248245443382,-7.64945287105236e-006,-2.40007374267426e-008,-.0214071018984119,-.0430565483441448,-.0517608590117147,-.0473878067921877,-.0348490990193348,-.0207016206060418,-.00972642083254007,-.0034402682709391,-.000832763242290672,-.000113733402739168,-5.49768056941236e-006,-1.72493889467441e-008,-.00982763882071387,-.0197665339288291,-.0237625359019177,-.0217549415081703,-.0159986326040109,-.00950376427809313,-.00446523548185483,-.00157936904180661,-.000382307535472911,-5.22130837362085e-005,-2.52389226922161e-006,-7.91890304680574e-009},
			{-1.08849930992072e-006,-2.18932125341295e-006,-2.63191437974773e-006,-2.40955526052686e-006,-1.77199232357178e-006,-1.05262729400979e-006,-4.94564953932593e-007,-1.74929313488118e-007,-4.23439949444061e-008,-5.78306820717772e-009,-2.79543748348947e-010,-8.77089671184216e-013,-6.67174550672731e-005,-.000134190202070998,-.000161318089751033,-.000147689018597559,-.000108610834338575,-6.45187493924991e-005,-3.03134001015112e-005,-1.07219531571797e-005,-2.5953930832331e-006,-3.5446195486468e-007,-1.71340921393568e-008,-5.37595110937315e-011,-.000575144383900619,-.0011567998362318,-.00139065846034338,-.00127316771187585,-.000936290380644839,-.000556190225361124,-.000261319647275767,-9.24296518291858e-005,-2.23738413632665e-005,-3.05567414766182e-006,-1.47706126638825e-007,-4.63439153301422e-010,-.00234804488509053,-.00472267141011844,-.00567740653672345,-.00519774689175974,-.00382243468035404,-.00227066394170372,-.00106684561013735,-.000377346936322897,-9.13419051691941e-005,-1.24748850093275e-005,-6.03014868716438e-007,-1.89200479726517e-009,-.00633227663561512,-.0127362394210907,-.0153109972435626,-.0140174369789623,-.0103084544811151,-.00612359342731638,-.00287710067800549,-.00101764033710382,-.000246333541419145,-3.36426374888076e-005,-1.62622826690742e-006,-5.1024142886996e-009,-.012937825514943,-.0260221169776175,-.0312827474533958,-.0286398027496958,-.0210617749476202,-.0125114532807165,-.00587836392863835,-.00207919740024719,-.000503297717511878,-6.87371381162648e-005,-3.32263714544313e-006,-1.04250255588739e-008,-.0214073125073398,-.0430569719462525,-.0517613682492133,-.0473882730064016,-.0348494418743096,-.0207018242742517,-.00972651652373047,-.00344030211724,-.000832771435243388,-.000113734521679514,-5.49773465709859e-006,-1.7249558650954e-008,-.0296455406961615,-.0596266912835726,-.071680821606866,-.0656248174753565,-.0482606374326126,-.0286685577088058,-.0134695955569983,-.00476424195651651,-.00115324889407038,-.000157503254546897,-7.61344126955836e-006,-2.388774830107e-008,-.0347378633474854,-.0698689855213075,-.0839936977751664,-.076897431725972,-.0565505431450506,-.0335930604290494,-.0157833171133724,-.00558261317396905,-.00135134666284035,-.000184558163040781,-8.92122984487403e-006,-2.79910339523387e-008,-.0340292605183584,-.068443757944315,-.0822803462291326,-.0753288338784716,-.0553969927824272,-.0329078099454216,-.0154613599726102,-.00546873583357464,-.00132378111976588,-.000180793439944228,-8.73924948977462e-006,-2.74200568127716e-008,-.0263115204035591,-.0529209070727989,-.0636193962384749,-.0582444672431377,-.0428331113779068,-.0254444116512108,-.0119547672264766,-.00422844200182213,-.00102355130296691,-.000139789998708784,-6.75721240365912e-006,-2.12012654199985e-008,-.0125704531070497,-.0252832132289538,-.0303944669425635,-.0278265692364139,-.0204637212045026,-.0121561878064161,-.00571144648888444,-.00202015813166747,-.000489006467861309,-6.67853319250079e-005,-3.22829013115797e-006,-1.0129004659729e-008},
			{-.000354187662792186,-.00370432632910279,-.0106866754670937,-.0175473588745023,-.0197676016789817,-.0162489801425508,-.00982031115806883,-.00422235247914266,-.00118608291592308,-.000180620480252622,-9.39296633663941e-006,-3.06697796211909e-008,-.000802889532215207,-.00839714407356662,-.024225067014107,-.0397771922581837,-.0448101448253052,-.0368339652566651,-.0222611509665594,-.00957143052386794,-.00268866947546018,-.000409439142389331,-2.1292425288575e-005,-6.95237231559063e-008,-.00120185194317225,-.0125697540159339,-.0362627020233443,-.0595428061909519,-.0670766742761412,-.0551370654893481,-.0333228999419048,-.0143275530598965,-.00402469144754472,-.000612892819182374,-3.18728065084108e-005,-1.04070632906319e-007,-.00152536042432723,-.0159532173895116,-.0460237143683289,-.0755702402721397,-.085132037201067,-.0699785843745455,-.0422925910998948,-.0181841719682459,-.00510803771553511,-.000777868235805777,-4.04521687853206e-005,-1.32083844163853e-007,-.00175303829641842,-.0183344215497376,-.0528932916734667,-.0868499819149511,-.0978389887959829,-.0804237060180922,-.0486052546469995,-.0208983721752537,-.00587047204848792,-.000893974161901044,-4.64901277906996e-005,-1.51798901731387e-007,-.00187057076364506,-.0195636529956887,-.0564395228555362,-.0926728396779039,-.104398604617067,-.0858157140598609,-.0518639943508112,-.0222995037123093,-.00626405789601532,-.000953910667053138,-4.96070587967628e-005,-1.61976260365957e-007,-.00187057076364506,-.0195636529956887,-.0564395228555362,-.0926728396779039,-.104398604617067,-.0858157140598609,-.0518639943508112,-.0222995037123093,-.00626405789601532,-.000953910667053138,-4.96070587967628e-005,-1.61976260365957e-007,-.00175303829641842,-.0183344215497376,-.0528932916734667,-.0868499819149511,-.0978389887959829,-.0804237060180922,-.0486052546469995,-.0208983721752537,-.00587047204848792,-.000893974161901044,-4.64901277906996e-005,-1.51798901731387e-007,-.00152536042432723,-.0159532173895116,-.0460237143683289,-.0755702402721397,-.085132037201067,-.0699785843745455,-.0422925910998948,-.0181841719682459,-.00510803771553511,-.000777868235805777,-4.04521687853206e-005,-1.32083844163853e-007,-.00120185194317225,-.0125697540159339,-.0362627020233444,-.0595428061909519,-.0670766742761412,-.0551370654893481,-.0333228999419048,-.0143275530598965,-.00402469144754472,-.000612892819182374,-3.18728065084108e-005,-1.04070632906319e-007,-.000802889532215207,-.00839714407356662,-.024225067014107,-.0397771922581837,-.0448101448253052,-.0368339652566651,-.0222611509665594,-.00957143052386794,-.00268866947546018,-.000409439142389331,-2.1292425288575e-005,-6.95237231559063e-008,-.000354187662792186,-.00370432632910279,-.0106866754670937,-.0175473588745023,-.0197676016789817,-.0162489801425508,-.00982031115806883,-.00422235247914266,-.00118608291592308,-.000180620480252622,-9.39296633663941e-006,-3.06697796211909e-008},
			{-5.6560083280767e-006,-5.91542359272188e-005,-.000170655084270396,-.00028021305752868,-.000315668024236135,-.00025947930056204,-.000156820147987287,-6.74265743697526e-005,-1.89405096647495e-005,-2.8843210756594e-006,-1.49995895979437e-007,-4.89764458734161e-010,-6.66694512082365e-005,-.000697272743806459,-.00201157426832706,-.00330297440936967,-.00372089514708473,-.00305857798732252,-.00184849678399644,-.000794781840714042,-.000223258756300761,-3.3998553762376e-005,-1.76805681469661e-006,-5.77303387673915e-009,-.000239493192417188,-.00250477651118074,-.00722607333006517,-.0118651027035077,-.0133663775728643,-.0109871701831305,-.00664025858844208,-.00285505335439599,-.000802000786155484,-.000122131231479417,-6.35129828208852e-006,-2.07381685017084e-008,-.00054515334463801,-.00570157038226907,-.0164485595800252,-.0270082851124352,-.030425605696751,-.0250098656792164,-.0151150817366226,-.00649889823405786,-.00182557761480488,-.000278004767708733,-1.44573274351898e-005,-4.72058591990341e-008,-.000959742855872972,-.0100376187644472,-.0289577009874395,-.0475481053927164,-.0535643007426736,-.044029886541206,-.0266101122837137,-.0114413150217837,-.00321393070570632,-.000489427593779599,-2.54521353623046e-005,-8.31059491191454e-008,-.00141708846333921,-.0148208383771267,-.0427568944566942,-.0702061715732879,-.0790892603834303,-.0650114182959186,-.0392906109117239,-.0168934370530446,-.00474546290931794,-.000722654190693695,-3.75808240390211e-005,-1.22708370279532e-007,-.00182283513943086,-.0190644026033435,-.0549992267137261,-.0903078952792718,-.101734427107537,-.0836257585887017,-.0505404624146264,-.0217304363719754,-.00610420363142845,-.000929567551026009,-4.83411222371217e-005,-1.57842742379509e-007,-.0020766085420598,-.0217185308967278,-.062656167598386,-.102880475966099,-.115897798864594,-.0952680584574148,-.0575766583052787,-.0247557273922317,-.00695402514977553,-.00105898107575704,-5.50711280460104e-005,-1.79817461292618e-007,-.00209684841095391,-.0219302126889477,-.0632668520831664,-.103883210618848,-.11702740813212,-.0961965979360501,-.0581378348543761,-.0249970115180785,-.00702180323816678,-.00106930253870999,-5.56078841986148e-005,-1.81570069821245e-007,-.00184217543696469,-.0192666760896712,-.0555827690131719,-.0912660628757668,-.102813830311442,-.0845130286536747,-.0510766971839986,-.0219609965010198,-.00616896928789685,-.000939430271260956,-4.88540219870558e-005,-1.59517455322627e-007,-.00132397601187602,-.0138470074345037,-.0399474726296093,-.0655931435851624,-.0738925524084332,-.0607397213008226,-.0367089477367546,-.0157834221327753,-.00443365338137005,-.000675170843678735,-3.51115056126793e-005,-1.14645586997212e-007,-.000607815019228362,-.00635692717584237,-.0183392097943569,-.0301127040609837,-.0339228224379417,-.0278845798860605,-.0168524577290889,-.00724590244919067,-.00203541536332734,-.000309959527704379,-1.61190990378048e-005,-5.2631852118241e-008},
			{-1.08304313867452e-005,-.000589001007959142,-.00407774946276834,-.0120086359724961,-.0207230264243475,-.0235713601331056,-.0183245877738932,-.00957752805114946,-.00312208933690556,-.000530130419242249,-2.96594300874774e-005,-1.00782323868636e-007,-2.45509398075641e-005,-.00133517565243957,-.00924363749104905,-.0272217503074272,-.0469759473290577,-.0534326863949788,-.0415390518964802,-.0217107986092594,-.00707730141552942,-.00120172498658768,-6.72334145152776e-005,-2.28458191424719e-007,-3.67505036876191e-005,-.00199863541368309,-.0138368769735303,-.0407484618877329,-.0703186818540462,-.0799838276575082,-.0621801483718135,-.0324990729725514,-.0105940706876594,-.00179887201456491,-.000100642251068273,-3.41980945423285e-007,-4.66428200392335e-005,-.00253661807513852,-.0175614181527209,-.0517169285858086,-.0892467121265429,-.101513473425809,-.0789174890001959,-.0412470105167922,-.0134457295270728,-.00228308336566367,-.00012773262768919,-4.34033661954923e-007,-5.3604806102002e-005,-.00291523797142323,-.020182665073873,-.0594362846564907,-.102567827047345,-.116665545847229,-.0906968466390758,-.0474036089408846,-.0154526618157401,-.00262386024318778,-.00014679821533987,-4.98818259086074e-007,-5.7198740774876e-005,-.00311069012556852,-.0215358120223382,-.0634211908577612,-.109444487868491,-.124487388343073,-.0967776174795907,-.0505817842982924,-.0164886856562537,-.00279977697510962,-.000156640302913549,-5.32261533433169e-007,-5.7198740774876e-005,-.00311069012556852,-.0215358120223382,-.0634211908577612,-.109444487868491,-.124487388343073,-.0967776174795907,-.0505817842982924,-.0164886856562537,-.00279977697510962,-.000156640302913549,-5.32261533433169e-007,-5.3604806102002e-005,-.00291523797142323,-.020182665073873,-.0594362846564907,-.102567827047345,-.116665545847229,-.0906968466390758,-.0474036089408846,-.0154526618157401,-.00262386024318778,-.00014679821533987,-4.98818259086074e-007,-4.66428200392335e-005,-.00253661807513852,-.0175614181527209,-.0517169285858085,-.0892467121265429,-.101513473425809,-.0789174890001959,-.0412470105167922,-.0134457295270728,-.00228308336566367,-.00012773262768919,-4.34033661954923e-007,-3.67505036876191e-005,-.00199863541368309,-.0138368769735303,-.0407484618877329,-.0703186818540462,-.0799838276575082,-.0621801483718135,-.0324990729725514,-.0105940706876594,-.00179887201456491,-.000100642251068273,-3.41980945423285e-007,-2.45509398075641e-005,-.00133517565243957,-.00924363749104905,-.0272217503074272,-.0469759473290577,-.0534326863949788,-.0415390518964802,-.0217107986092594,-.00707730141552942,-.00120172498658768,-6.72334145152776e-005,-2.28458191424719e-007,-1.08304313867452e-005,-.000589001007959142,-.00407774946276834,-.0120086359724961,-.0207230264243475,-.0235713601331056,-.0183245877738932,-.00957752805114946,-.00312208933690556,-.000530130419242249,-2.96594300874774e-005,-1.00782323868636e-007},
			{-.00719309508160456,-.0139021976630662,-.0155346480848768,-.0127550186066991,-.00808303973670482,-.00395000387691709,-.00144276568155165,-.000368787100047311,-5.82757244194688e-005,-4.43762053833248e-006,-8.93862080339512e-008,-5.3934946534894e-011,-.0163056519239563,-.0315141665027775,-.0352146831870672,-.028913686156645,-.0183230210276625,-.00895405768790711,-.00327053024384089,-.000835984234769121,-.000132102199014488,-1.00594104551317e-005,-2.02624931058027e-007,-1.22262316118168e-010,-.0244080644511933,-.0471738149883672,-.0527131488435292,-.0432811345737185,-.0274278808519569,-.0134034025849063,-.00489568361654787,-.00125139167557278,-.00019774486802044,-1.50580142318092e-005,-3.03310925545804e-007,-1.83015466396657e-010,-.0309781048820443,-.0598718260236465,-.0669022099971482,-.0549313334090684,-.0348107803231648,-.017011263301183,-.00621347919028916,-.00158823501355166,-.000250972840295156,-1.91112550165974e-005,-3.84954639980414e-007,-2.32278652181071e-010,-.0356019491148408,-.0688083958532261,-.0768881468075034,-.0631304769703183,-.0400066961627327,-.0195503931804077,-.00714091358400096,-.00182529765297712,-.000288433470136975,-2.19638332045296e-005,-4.42413619435074e-007,-2.6694895594493e-010,-.0379888820906312,-.0734216553279677,-.0820431132524654,-.0673630603261273,-.0426889454411423,-.0208611494544942,-.00761967619488893,-.00194767474936783,-.000307771494551617,-2.3436398585193e-005,-4.72075244245615e-007,-2.84846550926105e-010,-.0379888820906312,-.0734216553279677,-.0820431132524654,-.0673630603261273,-.0426889454411423,-.0208611494544942,-.00761967619488893,-.00194767474936783,-.000307771494551617,-2.3436398585193e-005,-4.72075244245615e-007,-2.84846550926105e-010,-.0356019491148408,-.0688083958532261,-.0768881468075034,-.0631304769703183,-.0400066961627327,-.0195503931804077,-.00714091358400096,-.00182529765297712,-.000288433470136975,-2.19638332045296e-005,-4.42413619435074e-007,-2.6694895594493e-010,-.0309781048820443,-.0598718260236465,-.0669022099971482,-.0549313334090684,-.0348107803231648,-.017011263301183,-.00621347919028916,-.00158823501355166,-.000250972840295157,-1.91112550165974e-005,-3.84954639980414e-007,-2.32278652181071e-010,-.0244080644511933,-.0471738149883672,-.0527131488435292,-.0432811345737185,-.0274278808519569,-.0134034025849063,-.00489568361654787,-.00125139167557278,-.00019774486802044,-1.50580142318092e-005,-3.03310925545804e-007,-1.83015466396657e-010,-.0163056519239563,-.0315141665027775,-.0352146831870672,-.028913686156645,-.0183230210276625,-.00895405768790711,-.00327053024384089,-.00083598423476912,-.000132102199014488,-1.00594104551317e-005,-2.02624931058027e-007,-1.22262316118168e-010,-.00719309508160456,-.0139021976630662,-.0155346480848768,-.0127550186066991,-.00808303973670482,-.00395000387691709,-.00144276568155165,-.000368787100047311,-5.82757244194688e-005,-4.43762053833248e-006,-8.93862080339512e-008,-5.3934946534894e-011},
			{-.000114866241683583,-.000222003626892483,-.000248072161029854,-.000203684371377566,-.000129077731546852,-6.30774506425453e-005,-2.30394662644891e-005,-5.88914614407857e-006,-9.30602664014744e-007,-7.08641812006219e-008,-1.42740470669776e-009,-8.61284959198155e-013,-.00135397065407101,-.00261683843315572,-.00292411783656719,-.00240090262810037,-.00152148671399835,-.000743517119145248,-.000271574665892772,-6.94175324250859e-005,-1.09693560023246e-005,-8.35302177246445e-007,-1.68253444704445e-008,-1.015228271122e-011,-.00486379816401721,-.00940033222170924,-.0105041559963674,-.0086246372913827,-.00546555736926158,-.00267089776882591,-.00097556350825723,-.000249364981245768,-3.94046454582605e-005,-3.00061244597533e-006,-6.04408074120732e-008,-3.64695156892415e-011,-.0110713620291109,-.0213977796180793,-.0239103905886559,-.0196320814726822,-.0124411339215587,-.00607970872643953,-.00222065480887151,-.000567624291069279,-8.96959702655164e-005,-6.8302313496928e-006,-1.37580145726718e-007,-8.30147957637377e-011,-.0194911408262166,-.0376708064290284,-.0420942598523953,-.0345622935723416,-.0219026252294268,-.0107033315917073,-.0039094643904085,-.00099930297324955,-.000157909820254493,-1.20246272104963e-005,-2.42209945641729e-007,-1.46147607732105e-010,-.0287792408488707,-.0556220500820546,-.0621534087434013,-.0510322397173339,-.0323398682673508,-.015803782883183,-.00577243878564134,-.00147550013642398,-.000233158479025165,-1.77547146005579e-005,-3.57630085574026e-007,-2.15791227404772e-010,-.0370194330577289,-.0715479873273676,-.0799494318272223,-.0656440033259579,-.041599554161483,-.0203287878778718,-.00742522752170763,-.00189797148624789,-.000299917386683,-2.28383184972103e-005,-4.60028222490692e-007,-2.77577471181613e-010,-.0421732438918713,-.0815088311812927,-.0910799169238503,-.0747829000510975,-.0473910051704331,-.0231589427061906,-.00845896075008184,-.00216220530078157,-.000341671604648038,-2.60178478303842e-005,-5.24072921213834e-007,-3.16221547012239e-010,-.0425842895511383,-.0823032650013445,-.0919676362701244,-.0755117789234986,-.0478529062519234,-.0233846636134322,-.00854140684095277,-.00218327944688257,-.00034500174046461,-2.62714333369098e-005,-5.29180849357959e-007,-3.19303631345124e-010,-.0374122095817176,-.072307112128577,-.0807976960315661,-.0663404868028083,-.0420409258123548,-.0205444765035491,-.00750400925375776,-.00191810898111,-.000303099512904151,-2.30806332657561e-005,-4.64909126147018e-007,-2.80522570694571e-010,-.026888246929992,-.0519672991102317,-.058069502618547,-.0476790708315039,-.0302149166554566,-.0147653657308921,-.00539314988437847,-.00137854963658506,-.000217838364492369,-1.65881078259945e-005,-3.34131331017576e-007,-2.01612260665772e-010,-.0123439399038752,-.0238573090634585,-.0266587276008582,-.0218886559078795,-.0138711577726375,-.00677852995457166,-.0024759039977092,-.000632868848341838,-.00010000591288281,-7.61532005623914e-006,-1.53394049110811e-007,-9.25567827468422e-011},
			{-.000538770430325433,-.00541459382780606,-.0145196103881163,-.0213815233314377,-.020756280852719,-.0140356226122099,-.00659447754217011,-.00204903932812416,-.000375745321421736,-3.19037343548278e-005,-6.91363066088356e-007,-4.34129792523551e-010,-.0012213105768994,-.0122740602291765,-.032913747194243,-.0484686596091265,-.0470513300576056,-.0318166205582331,-.0149486770580804,-.00464486033951284,-.000851757463741354,-7.23209107571181e-005,-1.56721486103326e-006,-9.84106174935942e-010,-.00182818984576065,-.0183731416903004,-.04926885883459,-.0725531351410981,-.0704315228803041,-.0476265609511484,-.0223767976156558,-.00695292963828415,-.00127500275172926,-.000108257766028679,-2.34597681315439e-006,-1.47311662585924e-009,-.00232029282368979,-.0233187318655643,-.0625308032698983,-.0920826243480438,-.0893899271345745,-.0604464398750644,-.0284000718224987,-.00882448438314045,-.00161820160082515,-.000137398048789906,-2.97745509129004e-006,-1.86964277444457e-009,-.00266662364773693,-.0267993251511533,-.0718642565322132,-.105827032314693,-.102732418569255,-.0694687775381858,-.0326391144885199,-.0101416418199092,-.00185973710365748,-.000157906313511514,-3.4218750648389e-006,-2.14870881134135e-009,-.00284540745247398,-.0285960861297693,-.0766823962117034,-.112922206580213,-.109620114430362,-.0741263123085118,-.0348274041920399,-.0108215882804389,-.00198442319330673,-.000168493143619897,-3.65129470714422e-006,-2.29276901154607e-009,-.00284540745247398,-.0285960861297693,-.0766823962117034,-.112922206580213,-.109620114430362,-.0741263123085118,-.0348274041920398,-.0108215882804389,-.00198442319330673,-.000168493143619897,-3.65129470714422e-006,-2.29276901154607e-009,-.00266662364773693,-.0267993251511533,-.0718642565322132,-.105827032314693,-.102732418569255,-.0694687775381858,-.0326391144885199,-.0101416418199092,-.00185973710365748,-.000157906313511514,-3.4218750648389e-006,-2.14870881134135e-009,-.00232029282368979,-.0233187318655643,-.0625308032698983,-.0920826243480438,-.0893899271345745,-.0604464398750644,-.0284000718224987,-.00882448438314046,-.00161820160082515,-.000137398048789906,-2.97745509129004e-006,-1.86964277444457e-009,-.00182818984576065,-.0183731416903004,-.04926885883459,-.0725531351410981,-.0704315228803041,-.0476265609511484,-.0223767976156558,-.00695292963828415,-.00127500275172926,-.000108257766028679,-2.34597681315439e-006,-1.47311662585924e-009,-.0012213105768994,-.0122740602291765,-.032913747194243,-.0484686596091265,-.0470513300576056,-.0318166205582331,-.0149486770580804,-.00464486033951284,-.000851757463741354,-7.23209107571182e-005,-1.56721486103326e-006,-9.84106174935942e-010,-.000538770430325433,-.00541459382780606,-.0145196103881163,-.0213815233314377,-.020756280852719,-.0140356226122099,-.00659447754217011,-.00204903932812416,-.000375745321421736,-3.19037343548278e-005,-6.91363066088356e-007,-4.34129792523551e-010},
			{-.00863095559575211,-.016029237481894,-.0166489373902391,-.0122597253338038,-.00669488358667141,-.00269138517084675,-.00076422934021594,-.000141170609267733,-1.45626016852063e-005,-6.18297441623807e-007,-5.18975291864507e-009,-6.02215524056038e-013,-.0195650628997474,-.0363358420848087,-.0377406074727463,-.0277909316794019,-.0151762822814405,-.00610096061448634,-.00173239161588508,-.000320012288242827,-3.30112019224449e-005,-1.40158620930416e-006,-1.1764380103799e-008,-1.36513094948138e-012,-.0292871035439603,-.0543914208171814,-.0564942256781886,-.0416004741639895,-.0227175017461691,-.0091325781138374,-.00259323125578963,-.000479029025826867,-4.94147396186346e-005,-2.09804592236745e-006,-1.76101973193715e-008,-2.0434757441563e-012,-.0371704592590807,-.0690322308074568,-.0717010581395466,-.0527982812554331,-.0288324849829067,-.0115908397087212,-.0032912642453023,-.00060797165761674,-6.27159515120887e-005,-2.66278740620364e-006,-2.23504219535243e-008,-2.59352830098302e-012,-.0427185847602993,-.0793360981199193,-.0824032791251857,-.0606790418511532,-.0331360703672624,-.0133209080116245,-.00378252390296041,-.00069871853362736,-7.20770403136054e-005,-3.06023954984572e-006,-2.56864836669144e-008,-2.98064271365472e-012,-.0455826526323854,-.0846551874810701,-.087928007662455,-.0647472687189207,-.0353576784818238,-.0142140084005345,-.00403612324963012,-.000745564123550858,-7.69094460834964e-005,-3.26541333602756e-006,-2.7408634178924e-008,-3.18047992928925e-012,-.0455826526323854,-.0846551874810701,-.087928007662455,-.0647472687189207,-.0353576784818238,-.0142140084005345,-.00403612324963012,-.000745564123550858,-7.69094460834964e-005,-3.26541333602756e-006,-2.7408634178924e-008,-3.18047992928925e-012,-.0427185847602993,-.0793360981199193,-.0824032791251857,-.0606790418511532,-.0331360703672624,-.0133209080116245,-.00378252390296041,-.00069871853362736,-7.20770403136054e-005,-3.06023954984572e-006,-2.56864836669144e-008,-2.98064271365472e-012,-.0371704592590807,-.0690322308074568,-.0717010581395466,-.0527982812554331,-.0288324849829067,-.0115908397087212,-.0032912642453023,-.00060797165761674,-6.27159515120887e-005,-2.66278740620364e-006,-2.23504219535243e-008,-2.59352830098302e-012,-.0292871035439603,-.0543914208171814,-.0564942256781886,-.0416004741639895,-.0227175017461691,-.0091325781138374,-.00259323125578962,-.000479029025826867,-4.94147396186346e-005,-2.09804592236745e-006,-1.76101973193715e-008,-2.0434757441563e-012,-.0195650628997474,-.0363358420848087,-.0377406074727463,-.0277909316794019,-.0151762822814405,-.00610096061448634,-.00173239161588507,-.000320012288242827,-3.30112019224449e-005,-1.40158620930416e-006,-1.1764380103799e-008,-1.36513094948138e-012,-.00863095559575211,-.016029237481894,-.0166489373902391,-.0122597253338038,-.00669488358667141,-.00269138517084675,-.00076422934021594,-.000141170609267733,-1.45626016852063e-005,-6.18297441623807e-007,-5.18975291864507e-009,-6.02215524056038e-013}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 144; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{-1.25649763607893e-005,-2.84828893052902e-005,-4.2636271224355e-005,-5.41128889760362e-005,-6.2189870139491e-005,-6.63593905024702e-005,-6.63593905024702e-005,-6.2189870139491e-005,-5.41128889760362e-005,-4.2636271224355e-005,-2.84828893052902e-005,-1.25649763607893e-005,-.000148107999463138,-.000335738296102257,-.000502569416311285,-.000637848532403299,-.000733054881183536,-.000782202551815753,-.000782202551815753,-.000733054881183536,-.000637848532403299,-.000502569416311285,-.000335738296102257,-.000148107999463138,-.000532040641869991,-.00120605517059218,-.00180535390260964,-.00229131001583844,-.0026333148170561,-.00280986543096192,-.00280986543096192,-.0026333148170561,-.00229131001583844,-.00180535390260964,-.00120605517059218,-.000532040641869991,-.00121107298487856,-.00274531815886015,-.00410948932756509,-.00521566106380589,-.00599415943941434,-.00639603790909957,-.00639603790909957,-.00599415943941434,-.00521566106380588,-.00410948932756509,-.00274531815886015,-.00121107298487856,-.00213209486213416,-.00483313459594366,-.00723475892097063,-.00918217505932459,-.010552722010284,-.0112602293456112,-.0112602293456112,-.010552722010284,-.00918217505932459,-.00723475892097063,-.00483313459594367,-.00213209486213416,-.00314810057025839,-.0071362649232201,-.0106823336471961,-.0135577506675935,-.0155814034208121,-.0166260587433132,-.0166260587433132,-.0155814034208121,-.0135577506675935,-.0106823336471961,-.0071362649232201,-.00314810057025839,-.00404947784869215,-.00917955004423722,-.0137409439473876,-.0174396623555152,-.0200427358008248,-.0213865011900097,-.0213865011900097,-.0200427358008248,-.0174396623555152,-.0137409439473876,-.00917955004423722,-.00404947784869215,-.00461324236601103,-.0104575184128172,-.0156539453074288,-.0198675958317378,-.0228330667266135,-.0243639098760258,-.0243639098760258,-.0228330667266135,-.0198675958317378,-.0156539453074288,-.0104575184128172,-.00461324236601103,-.00465820578534291,-.0105594436420292,-.0158065180211977,-.0200612373037489,-.023055611451648,-.0246013751140091,-.0246013751140091,-.023055611451648,-.0200612373037489,-.0158065180211977,-.0105594436420292,-.00465820578534291,-.00409244284577621,-.00927694515432801,-.0138867354885914,-.0176246973329244,-.0202553894113433,-.0216134121636217,-.0216134121636217,-.0202553894113433,-.0176246973329244,-.0138867354885914,-.00927694515432801,-.00409244284577621,-.00294124872640195,-.00666736327135993,-.0099804309086089,-.0126669132684861,-.0145575981275773,-.0155336124156416,-.0155336124156416,-.0145575981275773,-.0126669132684861,-.0099804309086089,-.00666736327135993,-.00294124872640195,-.00135027759956182,-.0030608738365598,-.00458184721642184,-.00581516588124543,-.0066831473580057,-.00713121902844807,-.00713121902844807,-.0066831473580057,-.00581516588124543,-.00458184721642184,-.0030608738365598,-.00135027759956182},
			{-1.49555377634105e-007,-3.39019282157687e-007,-5.07480751318181e-007,-6.4408187594673e-007,-7.40218624107564e-007,-7.89846587943962e-007,-7.89846587943962e-007,-7.40218624107564e-007,-6.4408187594673e-007,-5.07480751318181e-007,-3.39019282157687e-007,-1.49555377634105e-007,-9.16670694821036e-006,-2.07795296865571e-005,-3.11050488640572e-005,-3.94777499870446e-005,-4.53702655975518e-005,-4.84121154335211e-005,-4.84121154335211e-005,-4.53702655975518e-005,-3.94777499870446e-005,-3.11050488640572e-005,-2.07795296865571e-005,-9.16670694821036e-006,-7.90224989069185e-005,-.000179131979588681,-.000268144133301817,-.000340321826892032,-.000391118837314757,-.000417341403029572,-.000417341403029572,-.000391118837314757,-.000340321826892032,-.000268144133301817,-.000179131979588681,-7.90224989069185e-005,-.000322611816370485,-.000731311893505376,-.0010947067871833,-.00138937447237002,-.00159675485169678,-.00170380929406662,-.00170380929406662,-.00159675485169678,-.00138937447237002,-.0010947067871833,-.000731311893505376,-.000322611816370485,-.000870029052744199,-.00197222346386832,-.00295223752124433,-.00374690601758642,-.00430617553540284,-.00459488311014376,-.00459488311014376,-.00430617553540284,-.00374690601758643,-.00295223752124433,-.00197222346386832,-.000870029052744198,-.00177760460022009,-.00402955911756787,-.00603188017934985,-.00765551144808981,-.0087981860110801,-.00938806046568492,-.00938806046568492,-.0087981860110801,-.00765551144808981,-.00603188017934986,-.00402955911756787,-.00177760460022009,-.0029412776627297,-.00666742886561148,-.00998052909718346,-.0126670378869689,-.0145577413468253,-.0155337652370252,-.0155337652370252,-.0145577413468253,-.0126670378869689,-.00998052909718346,-.00666742886561148,-.0029412776627297,-.00407317670628986,-.00923327174798222,-.013821360407493,-.0175417249152225,-.0201600323904989,-.0215116619305183,-.0215116619305183,-.0201600323904989,-.0175417249152225,-.013821360407493,-.00923327174798222,-.00407317670628986,-.00477284112519419,-.0108193045125715,-.0161955058952257,-.0205549309837037,-.0236229946837403,-.0252067985596364,-.0252067985596364,-.0236229946837403,-.0205549309837037,-.0161955058952257,-.0108193045125715,-.00477284112519419,-.00467548197876496,-.0105986061434719,-.0158651406916522,-.0201356397307597,-.0231411193105227,-.0246926158479963,-.0246926158479963,-.0231411193105227,-.0201356397307597,-.0158651406916522,-.0105986061434719,-.00467548197876496,-.00361509587945292,-.0081948722230624,-.0122669716195725,-.015568933545552,-.0178927788504699,-.0190924003579595,-.0190924003579595,-.0178927788504699,-.015568933545552,-.0122669716195725,-.00819487222306241,-.00361509587945292,-.00172712912568916,-.00391513889802946,-.00586060361181134,-.00743813151271418,-.00854835957956341,-.00912148441898038,-.00912148441898038,-.00854835957956341,-.00743813151271418,-.00586060361181134,-.00391513889802946,-.00172712912568916},
			{-1.63148105416827e-009,-3.69831927536025e-009,-5.53604453566559e-009,-7.02620925147273e-009,-8.07495310619039e-009,-8.61633840464567e-009,-8.61633840464567e-009,-8.07495310619039e-009,-7.02620925147273e-009,-5.53604453566559e-009,-3.69831927536025e-009,-1.63148105416827e-009,-5.19981416069715e-007,-1.1787187408438e-006,-1.7644337761237e-006,-2.23937521483844e-006,-2.57362813998106e-006,-2.74617706012381e-006,-2.74617706012381e-006,-2.57362813998106e-006,-2.23937521483844e-006,-1.7644337761237e-006,-1.1787187408438e-006,-5.19981416069715e-007,-1.07571284042538e-005,-2.43847731013085e-005,-3.6501767378587e-005,-4.63271301374563e-005,-5.3241995792529e-005,-5.68116058451738e-005,-5.68116058451738e-005,-5.3241995792529e-005,-4.63271301374563e-005,-3.6501767378587e-005,-2.43847731013085e-005,-1.07571284042538e-005,-7.87643815441602e-005,-.000178546866806921,-.000267268273167243,-.000339210207117138,-.000389841295295595,-.000415978208195064,-.000415978208195064,-.000389841295295595,-.000339210207117138,-.000267268273167243,-.000178546866806921,-7.87643815441602e-005,-.000325387323726201,-.000737603546311472,-.00110412481400688,-.0014013275964412,-.00161049211924643,-.00171846757683388,-.00171846757683388,-.00161049211924644,-.0014013275964412,-.00110412481400688,-.000737603546311472,-.000325387323726201,-.000919943967994991,-.00208537298082308,-.00312161196363602,-.0039618718233036,-.00455322750019265,-.00485849867597654,-.00485849867597654,-.00455322750019265,-.0039618718233036,-.00312161196363602,-.00208537298082308,-.000919943967994991,-.00195799989864057,-.00443848780701132,-.00664400889731993,-.00843240990574942,-.00969104564411167,-.0103407818802712,-.0103407818802712,-.00969104564411167,-.00843240990574942,-.00664400889731993,-.00443848780701132,-.00195799989864057,-.00329609616783617,-.00747174842135327,-.0111845216543321,-.0141951151250118,-.0163139019731612,-.0174076676672231,-.0174076676672231,-.0163139019731612,-.0141951151250118,-.0111845216543321,-.00747174842135327,-.00329609616783617,-.00448203175305139,-.0101600839205191,-.0152087131700155,-.0193025183395333,-.022183644813943,-.0236709474657956,-.0236709474657956,-.022183644813943,-.0193025183395333,-.0152087131700155,-.0101600839205191,-.00448203175305139,-.00489564361298881,-.0110976790646466,-.0166122070513835,-.0210837976681447,-.0242308009023486,-.0258553551512484,-.0258553551512484,-.0242308009023486,-.0210837976681447,-.0166122070513835,-.0110976790646466,-.00489564361298881,-.00407237283003069,-.00923144948283444,-.0138186326438083,-.0175382629057834,-.0201560536356863,-.021507416420056,-.021507416420056,-.0201560536356863,-.0175382629057834,-.0138186326438083,-.00923144948283444,-.00407237283003069,-.00202472570360245,-.0045897450527047,-.00687042711223165,-.00871977424071199,-.0100213024648473,-.0106931807723117,-.0106931807723117,-.0100213024648473,-.00871977424071199,-.00687042711223165,-.0045897450527047,-.00202472570360245},
			{-1.70557276538753e-011,-3.86627390961513e-011,-5.7874572088226e-011,-7.34529592765333e-011,-8.44166719834875e-011,-9.00763884617597e-011,-9.00763884617597e-011,-8.44166719834875e-011,-7.34529592765333e-011,-5.7874572088226e-011,-3.86627390961513e-011,-1.70557276538753e-011,-2.82664008807978e-008,-6.40756292912074e-008,-9.59153363989562e-008,-1.21733345825304e-007,-1.39903470536826e-007,-1.49283299887593e-007,-1.49283299887593e-007,-1.39903470536826e-007,-1.21733345825304e-007,-9.59153363989562e-008,-6.40756292912074e-008,-2.82664008807978e-008,-1.40329882934204e-006,-3.18106489584209e-006,-4.76176220142831e-006,-6.04350947999848e-006,-6.94557390780468e-006,-7.41123996847038e-006,-7.41123996847038e-006,-6.94557390780468e-006,-6.04350947999848e-006,-4.76176220142831e-006,-3.18106489584208e-006,-1.40329882934204e-006,-1.84284021471628e-005,-4.17743832824876e-005,-6.2532417858728e-005,-7.93645806216638e-005,-9.12106719107567e-005,-9.73258921709034e-005,-9.73258921709034e-005,-9.12106719107567e-005,-7.93645806216638e-005,-6.2532417858728e-005,-4.17743832824876e-005,-1.84284021471628e-005,-.000116620720789999,-.000264361427000411,-.000395724793999518,-.000502244010271906,-.000577209799147466,-.000615908834952788,-.000615908834952788,-.000577209799147466,-.000502244010271906,-.000395724793999518,-.000264361427000411,-.000116620720789999,-.000456242568387133,-.00103423247275538,-.0015481510932686,-.00196487464364187,-.00225815515010815,-.00240955318100971,-.00240955318100971,-.00225815515010815,-.00196487464364187,-.0015481510932686,-.00103423247275538,-.000456242568387133,-.00124910090182185,-.00283152165958353,-.00423852805667493,-.00537943379114375,-.00618237716052034,-.00659687468888933,-.00659687468888933,-.00618237716052034,-.00537943379114375,-.00423852805667493,-.00283152165958353,-.00124910090182185,-.0025560815986996,-.00579424800656579,-.00867345748885218,-.0110081352954835,-.012651228153929,-.0134994298511859,-.0134994298511859,-.012651228153929,-.0110081352954835,-.00867345748885218,-.00579424800656579,-.0025560815986996,-.00403349103971441,-.00914331038111467,-.0136866964976499,-.0173708128492007,-.0199636097009642,-.021302070080123,-.021302070080123,-.0199636097009642,-.0173708128492007,-.0136866964976499,-.00914331038111467,-.00403349103971441,-.00491248705999991,-.0111358605958299,-.0166693612993899,-.021156336410114,-.0243141668994061,-.025944310422275,-.025944310422275,-.0243141668994061,-.021156336410114,-.0166693612993899,-.0111358605958299,-.00491248705999991,-.00439626090995001,-.00996565447158663,-.0149176701290573,-.0189331337918146,-.0217591253050265,-.0232179660428578,-.0232179660428578,-.0217591253050265,-.0189331337918146,-.0149176701290573,-.00996565447158663,-.00439626090995001,-.00227465638852363,-.00515629988163542,-.00771850769460683,-.00979613690280579,-.0112583248350307,-.012013139317637,-.012013139317637,-.0112583248350307,-.00979613690280579,-.00771850769460683,-.00515629988163542,-.00227465638852363},
			{-1.7384464750039e-013,-3.94079360668192e-013,-5.8990063561595e-013,-7.48687131527224e-013,-8.60437436733519e-013,-9.18125471866621e-013,-9.18125471866621e-013,-8.60437436733519e-013,-7.48687131527224e-013,-5.8990063561595e-013,-3.94079360668192e-013,-1.7384464750039e-013,-1.49815262263485e-009,-3.39608401063939e-009,-5.0836260825345e-009,-6.45201106711068e-009,-7.41504913145304e-009,-7.91219116240981e-009,-7.91219116240981e-009,-7.41504913145304e-009,-6.45201106711068e-009,-5.0836260825345e-009,-3.39608401063939e-009,-1.49815262263485e-009,-1.78487097219834e-007,-4.04603087706546e-007,-6.05653689159397e-007,-7.68680513053121e-007,-8.83415064139356e-007,-9.42643634493817e-007,-9.42643634493817e-007,-8.83415064139356e-007,-7.68680513053121e-007,-6.05653689159397e-007,-4.04603087706546e-007,-1.78487097219834e-007,-4.20386100245681e-006,-9.52951316020476e-006,-1.42648066135301e-005,-1.81045357479959e-005,-2.0806849318325e-005,-2.22018447046734e-005,-2.22018447046734e-005,-2.0806849318325e-005,-1.81045357479959e-005,-1.42648066135301e-005,-9.52951316020476e-006,-4.20386100245681e-006,-4.07524446402028e-005,-9.23795904009122e-005,-.00013828376853596,-.000175506300131138,-.000201702666783354,-.000215225823762682,-.000215225823762682,-.000201702666783354,-.000175506300131138,-.00013828376853596,-9.23795904009122e-005,-4.07524446402028e-005,-.000220614007696907,-.000500098382997093,-.000748601381966019,-.000950106149209752,-.00109192059703656,-.00116512842258577,-.00116512842258577,-.00109192059703656,-.000950106149209752,-.00074860138196602,-.000500098382997093,-.000220614007696907,-.000776935976612406,-.0017611956269338,-.0026363482167924,-.00334598721372161,-.00384541491403083,-.00410323078906309,-.00410323078906309,-.00384541491403083,-.00334598721372161,-.0026363482167924,-.0017611956269338,-.000776935976612406,-.00193264642090617,-.00438101533121211,-.00655797787568285,-.00832322148498332,-.00956555957532708,-.0102068825969568,-.0102068825969568,-.00956555957532708,-.00832322148498332,-.00655797787568285,-.00438101533121211,-.00193264642090617,-.00353907786161354,-.00802255094483993,-.0120090224811522,-.015241550951156,-.0175165305773379,-.0186909265162125,-.0186909265162125,-.0175165305773379,-.015241550951156,-.0120090224811522,-.00802255094483993,-.00353907786161354,-.00480613424305024,-.0108947749442933,-.0163084782050857,-.0206983126135993,-.0237877776977751,-.0253826294522236,-.0253826294522236,-.0237877776977751,-.0206983126135993,-.0163084782050857,-.0108947749442933,-.00480613424305024,-.00462724228856079,-.0104892541067666,-.0157014507286737,-.0199278885241277,-.0229023588080574,-.0244378476456489,-.0244378476456489,-.0229023588080574,-.0199278885241277,-.0157014507286737,-.0104892541067666,-.00462724228856079,-.00249154226883633,-.00564794716718901,-.00845445855931378,-.010730187331938,-.0123317932081052,-.0131585783867511,-.0131585783867511,-.0123317932081052,-.010730187331938,-.00845445855931378,-.00564794716718901,-.00249154226883633},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.77695599831757e-005,-.000209456341570312,-.000752419091586911,-.00171271584051363,-.00301523747077974,-.00445208652288796,-.0057268264950684,-.00652410992157336,-.00658769779901337,-.00578758817666173,-.00415955383985738,-.0019095808945752,-4.02808883575737e-005,-.000474805651832001,-.00170561957949523,-.00388246617391187,-.00683508449545443,-.010092202640724,-.012981844171125,-.0147891643705445,-.0149333084120673,-.0131195816567463,-.0094290755651378,-.00432872929307012,-6.02967930241717e-005,-.000710740484695346,-.00255315597436325,-.0058116955424022,-.0102314941881775,-.0151071011240822,-.0194326292933208,-.0221380217619632,-.0223537921630579,-.0196388096682036,-.0141144607515462,-.00647971047522447,-7.65271815013742e-005,-.000902054045409203,-.00324040170051977,-.00737605861435863,-.0129855565030636,-.0191735548724589,-.0246634070303329,-.0280970234814972,-.0283708738774972,-.0249250860049403,-.0179137205405699,-.00822388645802562,-8.79497578075961e-005,-.00103669615509983,-.00372406952887607,-.00847702157560611,-.0149238025892903,-.0220354320420534,-.0283447088011335,-.0322908326405259,-.0326055584089547,-.0286454464212646,-.0205875527108991,-.00945139763454551,-9.38463500544579e-005,-.00110620145747811,-.00397374980154702,-.00904536355795187,-.015924369058549,-.0235127977673766,-.0302450800394712,-.0344557717846379,-.0347915983448399,-.030565980615856,-.0219678453543708,-.0100850666679019,-9.38463500544579e-005,-.00110620145747811,-.00397374980154702,-.00904536355795187,-.015924369058549,-.0235127977673766,-.0302450800394711,-.0344557717846379,-.0347915983448399,-.030565980615856,-.0219678453543708,-.0100850666679019,-8.79497578075961e-005,-.00103669615509983,-.00372406952887607,-.00847702157560611,-.0149238025892903,-.0220354320420534,-.0283447088011335,-.0322908326405259,-.0326055584089547,-.0286454464212646,-.0205875527108991,-.00945139763454551,-7.65271815013742e-005,-.000902054045409203,-.00324040170051977,-.00737605861435863,-.0129855565030636,-.0191735548724589,-.0246634070303329,-.0280970234814972,-.0283708738774972,-.0249250860049403,-.0179137205405699,-.00822388645802562,-6.02967930241717e-005,-.000710740484695346,-.00255315597436325,-.0058116955424022,-.0102314941881775,-.0151071011240822,-.0194326292933208,-.0221380217619632,-.0223537921630579,-.0196388096682036,-.0141144607515462,-.00647971047522447,-4.02808883575737e-005,-.000474805651832001,-.00170561957949523,-.00388246617391187,-.00683508449545443,-.010092202640724,-.012981844171125,-.0147891643705445,-.0149333084120673,-.0131195816567463,-.0094290755651378,-.00432872929307012,-1.77695599831757e-005,-.000209456341570312,-.000752419091586911,-.00171271584051363,-.00301523747077974,-.00445208652288796,-.0057268264950684,-.00652410992157336,-.00658769779901337,-.00578758817666173,-.00415955383985738,-.0019095808945752},
			{-2.83761377805336e-007,-3.34479976602295e-006,-1.20153497508037e-005,-2.73502893236142e-005,-4.815020405282e-005,-7.10951878965703e-005,-9.14514584621857e-005,-.000104183244945397,-.000105198677163819,-9.2421759274531e-005,-6.64237454259384e-005,-3.04940193335369e-005,-3.34479976602295e-006,-3.94263855120624e-005,-.000141629348384198,-.000322387923394712,-.000567564171331005,-.00083802513816711,-.00107797198911481,-.00122804624086627,-.00124001551403817,-.00108940928179785,-.000782961126977123,-.000359444225710958,-1.20153497508037e-005,-.000141629348384198,-.000508767721494418,-.00115809732300615,-.00203883117125292,-.00301039399647408,-.00387234255465894,-.004411446462074,-.00445444305782967,-.00391342814465003,-.00281259042095047,-.00129121274513816,-2.73502893236142e-005,-.000322387923394712,-.00115809732300615,-.00263615271349073,-.00464094875074617,-.00685249688850132,-.00881453236290365,-.0100416833114065,-.0101395555630045,-.00890805463205047,-.00640224074681438,-.00293916056463564,-4.815020405282e-005,-.000567564171331005,-.00203883117125292,-.00464094875074617,-.00817039361825585,-.0120638257076051,-.0155179905734142,-.0176783907020889,-.017850694871518,-.0156826365956051,-.0112711494458756,-.00517439429092323,-7.10951878965703e-005,-.00083802513816711,-.00301039399647408,-.00685249688850132,-.0120638257076051,-.0178125923307157,-.0229127680203357,-.0261026621464626,-.0263570743040411,-.0231558726990041,-.0166421825914966,-.00764014486751864,-9.14514584621857e-005,-.00107797198911481,-.00387234255465894,-.00881453236290365,-.0155179905734142,-.0229127680203357,-.0294732472739763,-.0335764851836737,-.0339037416907514,-.0297859586976462,-.0214072416856188,-.00982770299465891,-.000104183244945397,-.00122804624086627,-.004411446462074,-.0100416833114065,-.0176783907020889,-.0261026621464626,-.0335764851836737,-.038250972036084,-.0386237888878892,-.0339327319991675,-.0243875378440283,-.0111959066105712,-.000105198677163819,-.00124001551403817,-.00445444305782967,-.0101395555630045,-.017850694871518,-.0263570743040411,-.0339037416907514,-.0386237888878892,-.039000239435718,-.0342634607007844,-.02462523337432,-.0113050286127969,-9.2421759274531e-005,-.00108940928179785,-.00391342814465003,-.00890805463205047,-.0156826365956051,-.0231558726990041,-.0297859586976462,-.0339327319991675,-.0342634607007844,-.0301019879923868,-.0216343727160793,-.00993197501349335,-6.64237454259384e-005,-.000782961126977123,-.00281259042095047,-.00640224074681438,-.0112711494458756,-.0166421825914966,-.0214072416856188,-.0243875378440283,-.02462523337432,-.0216343727160793,-.0155486768161827,-.00713813484023199,-3.04940193335369e-005,-.000359444225710958,-.00129121274513816,-.00293916056463564,-.00517439429092323,-.00764014486751864,-.00982770299465891,-.0111959066105712,-.0113050286127969,-.00993197501349335,-.00713813484023199,-.00327699711041027},
			{-3.37748665821831e-009,-3.98116779370311e-008,-1.43013414267412e-007,-3.25538443615425e-007,-5.73110664448679e-007,-8.46214697861095e-007,-1.08850641767933e-006,-1.24004726271932e-006,-1.25213350502815e-006,-1.10005548074599e-006,-7.90612576313682e-007,-3.62956876834475e-007,-2.07016497214321e-007,-2.44017962134497e-006,-8.76572998571608e-006,-1.99532478216857e-005,-3.51277071610772e-005,-5.18670894572505e-005,-6.67178907235581e-005,-7.60062930474401e-005,-7.67470958397088e-005,-6.74257681555381e-005,-4.84590652057195e-005,-2.22467381475193e-005,-1.78460607688856e-006,-2.10358084478815e-005,-7.5565837560655e-005,-.000172008935497438,-.000302821854829854,-.000447125346440885,-.000575148139518477,-.000655219725381629,-.000661605888719569,-.000581250466549428,-.000417746138158903,-.000191780194444643,-7.28570996785344e-006,-8.58793440610684e-005,-.000308499888616729,-.000702231843846736,-.00123627966686315,-.00182540317195597,-.00234806021751757,-.00267495496410619,-.00270102667510736,-.00237297315794306,-.00170546163785514,-.000782948625132055,-1.96483173282813e-005,-.000231601945658244,-.000831971590143711,-.00189379952905124,-.00333403543487275,-.00492280106302518,-.00633231798455615,-.00721389736971712,-.00728420832819645,-.00639950393640167,-.00459933919956203,-.00211147892329134,-4.01445666201286e-005,-.000473198777355153,-.0016998472881189,-.0038693268277857,-.0068119526671359,-.0100580478180592,-.0129379099972572,-.0147391137220966,-.0148827699400637,-.0130751813408912,-.00939716494906272,-.00431407966834182,-6.64243989158946e-005,-.000782968829903678,-.00281261809177179,-.00640230373332618,-.011271260333675,-.0166423463205833,-.0214074522943001,-.0243877777734532,-.0246254756422405,-.0216345855593206,-.0155488297871431,-.00713820506661556,-9.19866620625105e-005,-.00108428063087214,-.00389500475941537,-.00886611786558599,-.0156088068880401,-.0230468609713296,-.029645734277678,-.0337729856666483,-.0341021573849623,-.0299602757900378,-.021532523808133,-.00988521790053479,-.000107787546507207,-.00127053146952515,-.00456406393315995,-.0103890832686787,-.0182899885770756,-.0270057043390071,-.034738090181205,-.0395742945945498,-.039960009339504,-.0351066616363676,-.0252312439580584,-.0115832378335861,-.000105588834409239,-.00124461444104258,-.00447096354344966,-.0101771607988773,-.0179168989163566,-.0264548265171398,-.0340294827249939,-.0387670354712729,-.0391448821859808,-.0343905358485524,-.0247165625951827,-.0113469563151396,-8.16415851719115e-005,-.000962339403243543,-.00345696164727354,-.00786900949156683,-.0138533968774284,-.0204549466275595,-.0263116922143577,-.0299747813866719,-.0302669335343241,-.0265908595098725,-.0191109160505604,-.00877349869072301,-3.90046528002956e-005,-.000459762193746235,-.0016515797496082,-.003759456439442,-.00661852576932447,-.00977244733277217,-.01257053519047,-.014320593338432,-.0144601704064181,-.0127039086840373,-.0091303303785454,-.00419158042503607},
			{-3.68445827925519e-011,-4.34300653798924e-010,-1.56011558760704e-009,-3.55125848055141e-009,-6.25199311274643e-009,-9.2312511197497e-009,-1.18743814217069e-008,-1.35275216933177e-008,-1.36593690107054e-008,-1.20003687174081e-008,-8.62469447627331e-009,-3.95945152473478e-009,-1.174301000065e-008,-1.38419179545726e-007,-4.97235999402952e-007,-1.13184790520788e-006,-1.99261905231337e-006,-2.9421604480658e-006,-3.78457209223235e-006,-4.31145667798826e-006,-4.35347871350321e-006,-3.82472643681277e-006,-2.74884028563086e-006,-1.26194613503531e-006,-2.42933809796949e-007,-2.86355020000406e-006,-1.02865820344585e-005,-2.34151315299593e-005,-4.12223559228577e-005,-6.08660170299574e-005,-7.82934287517701e-005,-8.91933666496246e-005,-9.0062698548559e-005,-7.91241227483076e-005,-5.68667013887132e-005,-2.61065418768862e-005,-1.77877687852613e-006,-2.09670975420156e-005,-7.53190109571436e-005,-.000171447089262513,-.000301832724128607,-.000445664865962291,-.000573269488181101,-.000653079529962769,-.000659444833676081,-.000579351882704146,-.000416381621285402,-.000191153767800758,-7.34839068958039e-006,-8.66181847905149e-005,-.000311153987634738,-.000708273313928056,-.00124691567929258,-.00184110755612247,-.00236826114642384,-.00269796824743438,-.00272426425965937,-.00239338841889015,-.00172013413604819,-.000789684521168945,-2.07755717461149e-005,-.000244889308237957,-.000879703089733841,-.00200244974321189,-.00352531421515612,-.00520522978980597,-.00669561288171989,-.0076277698425577,-.00770211464969917,-.00676665339575356,-.00486321041790597,-.00223261774167844,-4.42185272019898e-005,-.000521220049687516,-.00187235160015967,-.00426199478516835,-.00750324488892071,-.0110787610500387,-.0142508780967575,-.016234872011969,-.0163931067848593,-.0144020800439593,-.0103508103065028,-.00475188214064504,-7.44374492353296e-005,-.000877421602303038,-.00315191585985935,-.00717464014604113,-.0126309591445223,-.0186499814768931,-.0239899219175831,-.0273297764014771,-.027596148748536,-.0242444552089724,-.0174245495155288,-.00799931630469631,-.000101220047626196,-.00119311794379991,-.00428597536221463,-.00975607607116152,-.0171755789499336,-.0253602458535457,-.0326214971629358,-.0371630314766057,-.0375252446090086,-.032967611546778,-.0236939034040642,-.0108774707577442,-.000110560858773534,-.00130322152160786,-.00468149470230067,-.0106563884722811,-.01876057958054,-.0277005457518707,-.0356318814839124,-.0405925186861703,-.0409881576521328,-.0360099360729874,-.0258804295145632,-.0118812679549556,-9.19685077033871e-005,-.00108406663876152,-.0038942360467172,-.00886436806094961,-.0156057263557175,-.0230423124750437,-.0296398834369748,-.0337663202773765,-.03409542703082,-.0299543628718611,-.0215282741793143,-.00988326696773915,-4.57254306619082e-005,-.00053898247521257,-.00193615863497058,-.00440723740391471,-.00775894462383229,-.0114563092049479,-.0147365274132258,-.0167881329718923,-.0169517601570097,-.0148928821041257,-.0107035509528065,-.00491381941200989},
			{-3.85178343554166e-013,-4.54023885618748e-012,-1.63096632460441e-011,-3.71253453125811e-011,-6.53591971617201e-011,-9.65047707354281e-011,-1.24136418981734e-010,-1.41418575087745e-010,-1.42796925104609e-010,-1.25453507524717e-010,-9.01637440362945e-011,-4.13926516217373e-011,-6.38354791449189e-010,-7.5245228000818e-009,-2.70299508117877e-008,-6.15277116719808e-008,-1.08319580712847e-007,-1.5993703651203e-007,-2.05730875518941e-007,-2.34372535522585e-007,-2.36656870434667e-007,-2.07913681993516e-007,-1.49428073991582e-007,-6.85999042669627e-008,-3.16914252834386e-008,-3.73558490210382e-007,-1.34191468136724e-006,-3.05457232158762e-006,-5.37757677216062e-006,-7.94014975773712e-006,-1.02136065356437e-005,-1.16355352814725e-005,-1.17489421676717e-005,-1.03219729945844e-005,-7.41842734727761e-006,-3.40567466501195e-006,-4.16178163573184e-007,-4.90564514068014e-006,-1.76222931839914e-005,-4.01132573852444e-005,-7.06194185176384e-005,-.000104271641780592,-.000134127132921491,-.000152800186874675,-.000154289468887485,-.00013555022303093,-9.74202782733315e-005,-4.47240038949301e-005,-2.6337062228933e-006,-3.10444642827677e-005,-.00011151940991293,-.000253849300234756,-.000446901875894084,-.000659863721515399,-.000848798652964526,-.000966967607276502,-.000976392251931343,-.000857804414450873,-.00061650613987428,-.000283027553294656,-1.03035625519223e-005,-.000121451882844024,-.000436285264394079,-.000993107021958261,-.00174836562742656,-.00258151310547804,-.00332066269912202,-.00378296225321243,-.00381983326596499,-.00335589496079028,-.00241188995212247,-.00110725792950642,-2.82090935115534e-005,-.00033251096434234,-.00119446179503315,-.00271892840056321,-.00478667540745511,-.00706766656937942,-.0090913103237661,-.010356994041021,-.0104579394995808,-.00918776920961406,-.00660327230083207,-.00303145075477106,-5.77253164541879e-005,-.00068042954422791,-.00244427156379506,-.00556384423606199,-.00979515178485467,-.0144628287733888,-.0186038861974662,-.0211939018276931,-.0214004702712195,-.0188012736005075,-.013512521522274,-.00620337034448753,-9.10904201184736e-005,-.00107371629731872,-.00385705505497627,-.00877973374713072,-.0154567275852829,-.0228223114223215,-.0293568907657418,-.0334439296312284,-.033769894172609,-.029668367645793,-.0213227287077879,-.00978890451433557,-.000110941243135535,-.0013077052520396,-.00469760137328079,-.0106930517505446,-.0188251253083289,-.0277958494112289,-.0357544728843004,-.0407321771466212,-.0411291773074706,-.0361338281692569,-.0259694710684948,-.011922145428059,-9.92830402484601e-005,-.00117028572514512,-.00420395637395983,-.00956937796371369,-.0168468967973055,-.024874937019263,-.0319972326801651,-.0364518574766186,-.0368071390818563,-.0323367236121244,-.0232404826956437,-.0106693129707941,-5.13697449717053e-005,-.000605514084724702,-.00217515666585201,-.00495126362270572,-.00871670317383983,-.0128704677825032,-.0165555937697616,-.0188604480446285,-.01904427325192,-.0167312487713743,-.0120247895925313,-.00552037774992995},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.5903751261877e-007,-1.58772021730906e-005,-.000136870983053836,-.000558780057100243,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567,-5.8719862146837e-007,-3.59912011760796e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-8.78982445184828e-007,-5.38755250068202e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-1.11558253342289e-006,-6.83774687490587e-005,-.000589454695127486,-.00240646717698951,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-1.2820962657192e-006,-7.85836051712971e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.3680544205638e-006,-8.38522436364219e-005,-.00072285651418097,-.00295108426386074,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.3680544205638e-006,-8.38522436364219e-005,-.00072285651418097,-.00295108426386074,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.2820962657192e-006,-7.85836051712971e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.11558253342289e-006,-6.83774687490587e-005,-.000589454695127486,-.00240646717698951,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-8.78982445184828e-007,-5.38755250068202e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-5.8719862146837e-007,-3.59912011760796e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-2.5903751261877e-007,-1.58772021730906e-005,-.000136870983053836,-.000558780057100243,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567},
			{-4.13655946284767e-009,-2.53542393256769e-007,-2.1856871401242e-006,-8.92313591757784e-006,-2.40641758792944e-005,-4.91668520822556e-005,-8.13529419075361e-005,-.000112660192597526,-.000132012244784625,-.000129319383418971,-9.99901127316091e-005,-4.77707484775864e-005,-4.87591483748876e-008,-2.98859747651745e-006,-2.57634985121267e-005,-.0001051802861973,-.000283653295149336,-.000579547775714673,-.0009589370588843,-.0013279671418103,-.00155607690124302,-.00152433515352727,-.00117862024866094,-.00056309138885062,-1.75154945664267e-007,-1.07357828440125e-005,-9.25488720048507e-005,-.000377833656408332,-.00101895293817233,-.00208187924827317,-.00344473958308176,-.00477038710314128,-.0055898138948427,-.00547578967002042,-.00423389604809699,-.00202276382802748,-3.98701539258914e-007,-2.44376379372225e-005,-.000210667061584079,-.000860054849279415,-.00231942126065002,-.00473893818806989,-.00784118866248488,-.0108587323850331,-.0127239764517783,-.0124644254937564,-.00963752901773119,-.00460437499342809,-7.01914347023563e-007,-4.30224791892881e-005,-.000370879513648267,-.00151412718159681,-.00408334279989948,-.00834290409323953,-.0138044182877922,-.0191168061846747,-.0224005697075911,-.0219436300590561,-.0169668767769879,-.00810600549215572,-1.03639711130188e-006,-6.35239518067959e-005,-.00054761447492283,-.0022356531730757,-.00602917535482301,-.0123185424813832,-.0203826283040634,-.0282265247763196,-.0330750978875268,-.0324004131004311,-.0250520909767,-.0119687547517608,-1.33314265253001e-006,-8.17123944937675e-005,-.000704409734165886,-.00287577470912415,-.00775547397560826,-.0158456389156716,-.0262186674070122,-.0363084610152352,-.0425452977913825,-.0416774344435578,-.0322251100971785,-.0153956985051803,-1.51874152529913e-006,-9.30883176033672e-005,-.000802476998296003,-.00327613737349354,-.00883518380630904,-.0180516539400008,-.0298688057526838,-.0413632909868219,-.0484684143436297,-.047479727872512,-.0367114597744218,-.0175390732465345,-1.53354408858128e-006,-9.39956120239112e-005,-.000810298419092016,-.00330806856783063,-.00892129679210925,-.0182275961561996,-.0301599249984219,-.0417664423606343,-.0489408164993326,-.0479424936985227,-.037069271618925,-.0177100195447144,-1.34728730829027e-006,-8.25793637481591e-005,-.000711883527900599,-.00290628670514157,-.00783775962556066,-.0160137612897764,-.0264968477084686,-.0366936941193377,-.0429967037908343,-.0421196324049362,-.0325670188106206,-.0155590470074065,-9.68298698098172e-007,-5.93499915831539e-005,-.000511632440253917,-.00208875539431883,-.00563301709645563,-.0115091295769889,-.0190433495379502,-.0263718481021364,-.0309018366364716,-.0302714832768886,-.0234059964205263,-.0111823253052378,-4.44529573439469e-007,-2.72465784513743e-005,-.000234881809580631,-.000958912313193877,-.0025860229823533,-.005283646948578,-.00874248004628156,-.012106869926272,-.0141865111307979,-.0138971265528759,-.0107452975256271,-.00513361662858989},
			{-4.92356447782966e-011,-3.0178033998418e-009,-2.60152710469109e-008,-1.06208155422931e-007,-2.86425283165094e-007,-5.85211378134644e-007,-9.68308224601569e-007,-1.34094463604488e-006,-1.57128358699498e-006,-1.53923164459519e-006,-1.19013825765391e-006,-5.68594172031225e-007,-3.0178033998418e-009,-1.84970409164038e-007,-1.59455560634358e-006,-6.50982299448902e-006,-1.75558824755616e-005,-3.58693969483533e-005,-5.93505754917118e-005,-8.21906019486044e-005,-9.63087814184433e-005,-9.4344219744047e-005,-7.29472173341576e-005,-3.48508775138952e-005,-2.60152710469109e-008,-1.59455560634358e-006,-1.37460234489013e-005,-5.61185694462147e-005,-.000151342211720415,-.000309215664595502,-.000511637473895075,-.00070853216856797,-.000830239323386619,-.000813303626233361,-.000628848662297057,-.000300435417626676,-1.06208155422931e-007,-6.50982299448902e-006,-5.61185694462147e-005,-.000229105810010919,-.000617859299465599,-.00126238259464391,-.00208877594431585,-.00289260467614267,-.00338947793153767,-.00332033741971269,-.00256729427659293,-.00122653696255208,-2.86425283165094e-007,-1.75558824755616e-005,-.000151342211720415,-.000617859299465599,-.00166626116517048,-.00340443057968316,-.00563307251629313,-.00780086152659004,-.00914084396303246,-.00895438379338928,-.00692355485520927,-.0033077610227998,-5.85211378134644e-007,-3.58693969483533e-005,-.000309215664595502,-.00126238259464391,-.00340443057968316,-.00695578089086413,-.0115092428083319,-.0159383727377934,-.0186761651548639,-.018295198043168,-.0141458988313976,-.0067582699589302,-9.68308224601569e-007,-5.93505754917118e-005,-.000511637473895075,-.00208877594431585,-.00563307251629313,-.0115092428083319,-.0190435368939121,-.0263721075587493,-.0309021406608933,-.030271781099646,-.023406226697924,-.0111824353213518,-1.34094463604488e-006,-8.21906019486044e-005,-.00070853216856797,-.00289260467614267,-.00780086152659004,-.0159383727377934,-.0263721075587493,-.036520949914119,-.0427942866834367,-.0419213443176089,-.0324137019011133,-.0154857991299779,-1.57128358699498e-006,-9.63087814184433e-005,-.000830239323386619,-.00338947793153767,-.00914084396303246,-.0186761651548639,-.0309021406608933,-.0427942866834367,-.0501452173903112,-.0491223265304294,-.037981521698904,-.0181458513277659,-1.53923164459519e-006,-9.4344219744047e-005,-.000813303626233361,-.00332033741971269,-.00895438379338928,-.018295198043168,-.030271781099646,-.0419213443176089,-.0491223265304294,-.0481203011840638,-.0372067528692506,-.0177757018611982,-1.19013825765391e-006,-7.29472173341576e-005,-.000628848662297057,-.00256729427659293,-.00692355485520927,-.0141458988313976,-.023406226697924,-.0324137019011133,-.037981521698904,-.0372067528692506,-.0287683664692428,-.0137442229153402,-5.68594172031225e-007,-3.48508775138952e-005,-.000300435417626676,-.00122653696255208,-.0033077610227998,-.0067582699589302,-.0111824353213518,-.0154857991299779,-.0181458513277659,-.0177757018611982,-.0137442229153402,-.00656636739345367},
			{-5.37105538571509e-013,-3.2920842768966e-011,-2.83797363265043e-010,-1.15861158670691e-009,-3.1245778676732e-009,-6.38399829730143e-009,-1.05631542517567e-008,-1.46281986187158e-008,-1.71409376487648e-008,-1.67912869868831e-008,-1.29830705524454e-008,-6.20272325817235e-009,-1.71184886157124e-010,-1.04924457427685e-008,-9.045116058834e-008,-3.69269684126226e-007,-9.95857365368599e-007,-2.03469140284302e-006,-3.36666116468723e-006,-4.66226157691217e-006,-5.46311525260784e-006,-5.35167549924359e-006,-4.13793062049084e-006,-1.9769158918718e-006,-3.54139156583407e-009,-2.1706273078516e-007,-1.87121062155825e-006,-7.6392757225221e-006,-2.06018238739433e-005,-4.20927287149029e-005,-6.96478861031055e-005,-9.64506516716335e-005,-.000113018331892963,-.000110712919239298,-8.56035420437903e-005,-4.08974964028785e-005,-2.59302953359118e-008,-1.58934718486943e-006,-1.37011237392702e-005,-5.59352649812969e-005,-.000150847870838019,-.000308205649327789,-.000509966272467282,-.000706217834626139,-.000827527447862548,-.00081064706909918,-.000626794604813159,-.000299454081965167,-1.07121889836103e-007,-6.56582857400443e-006,-5.6601371053274e-005,-.000231076862629534,-.000623174892248206,-.0012732431770043,-.00210674618825924,-.00291749045281179,-.00341863842883094,-.00334890308448347,-.00258938132932369,-.00123708915628192,-3.02857945621352e-007,-1.85630906649181e-005,-.000160024948988241,-.000653306845161927,-.0017618571509321,-.00359974803892942,-.0059562506178524,-.00824840904375729,-.00966526834956002,-.00946811067096992,-.00732076992880829,-.00349753239972921,-6.44599940278459e-007,-3.95095037359528e-005,-.000340595563207873,-.00139049200941683,-.00374991982442457,-.00766166912395449,-.0126772265613641,-.0175558345219898,-.0205714642490873,-.0201518357411283,-.0155814563465409,-.0074441143400151,-1.08511925583571e-006,-6.65102812046103e-005,-.000573358421239403,-.00234075363682485,-.00631261338244207,-.0128976504320703,-.0213408376153209,-.0295534841096122,-.0346299938653073,-.0339235914196489,-.0262298167578035,-.0125314188045125,-1.47554522466982e-006,-9.04406840954337e-005,-.00077965281321311,-.00318295692604349,-.0085838920298908,-.0175382257776313,-.0290192721808159,-.040186829342282,-.047089867591508,-.0461293015064619,-.0356673062917253,-.0170402239900308,-1.61171182018355e-006,-9.87867515986945e-005,-.000851600909064724,-.00347668727130158,-.00937603268029293,-.0191566922641638,-.0316972351677059,-.0438953593449828,-.0514354253188647,-.0503862160588214,-.0389587646541599,-.0186127337638557,-1.34067998923389e-006,-8.2174380935362e-005,-.000708392333727775,-.00289203379604635,-.00779932196032425,-.0159352271645878,-.0263669027993754,-.0365137421945241,-.0427858408675583,-.0419130707843193,-.0324073047817918,-.015482742876017,-6.66566974047622e-007,-4.08559304861565e-005,-.000352202567445854,-.00143787796622194,-.0038777116690518,-.00792276772768902,-.0131092481092602,-.0181541119739297,-.0212725100010394,-.0208385811603797,-.0161124498455309,-.0076977989913323},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.26296210833654e-009,-1.03996283213943e-006,-2.15142568085075e-005,-.00015752876308832,-.000650774647452403,-.00183988793598998,-.00391599979728114,-.00659219233567233,-.00896406350610278,-.00979128722597762,-.00814474566006138,-.00404945140720489,-7.3966385507205e-009,-2.35743748168761e-006,-4.87695462026171e-005,-.000357093733613841,-.00147520709262294,-.00417074596164616,-.00887697561402263,-.0149434968427065,-.0203201678410382,-.0221953581292932,-.0184628989656689,-.0091794901054094,-1.10720890713312e-008,-3.5288675522474e-006,-7.3003534757174e-005,-.000534536546334486,-.00220824962801375,-.00624322392727204,-.0132880177946399,-.0223690433086643,-.030417426340031,-.0332244141027671,-.0276372652876165,-.0137408542244633,-1.40524185029455e-008,-4.47875042967689e-006,-9.26542602749126e-005,-.000678420414234276,-.00280265519288241,-.00792374364660721,-.0168648198114988,-.0283902302500237,-.0386050366790666,-.0421675953362895,-.0350765258115668,-.017439548481424,-1.61499062123808e-008,-5.14725627996212e-006,-.000106483991585058,-.000779682590591189,-.00322098423849287,-.00910645500038529,-.0193820912882233,-.0326278039463224,-.0443672896278861,-.0484616018046971,-.0403121072713726,-.0200426049296945,-1.72326768092913e-008,-5.49235412024762e-006,-.000113623211690348,-.000831956416390127,-.00343693515366777,-.00971699735195308,-.0206815637605425,-.0348153353344463,-.0473418949315912,-.0517107103024967,-.043014832840112,-.0213863615446234,-1.72326768092913e-008,-5.49235412024762e-006,-.000113623211690348,-.000831956416390127,-.00343693515366777,-.00971699735195308,-.0206815637605425,-.0348153353344463,-.0473418949315912,-.0517107103024967,-.043014832840112,-.0213863615446234,-1.61499062123808e-008,-5.14725627996212e-006,-.000106483991585058,-.000779682590591189,-.00322098423849287,-.00910645500038529,-.0193820912882233,-.0326278039463224,-.0443672896278861,-.0484616018046971,-.0403121072713726,-.0200426049296945,-1.40524185029455e-008,-4.47875042967689e-006,-9.26542602749126e-005,-.000678420414234276,-.00280265519288241,-.00792374364660721,-.0168648198114988,-.0283902302500237,-.0386050366790666,-.0421675953362895,-.0350765258115668,-.017439548481424,-1.10720890713312e-008,-3.5288675522474e-006,-7.3003534757174e-005,-.000534536546334486,-.00220824962801375,-.00624322392727204,-.0132880177946399,-.0223690433086643,-.030417426340031,-.0332244141027671,-.0276372652876165,-.0137408542244633,-7.3966385507205e-009,-2.35743748168761e-006,-4.87695462026171e-005,-.000357093733613841,-.00147520709262294,-.00417074596164616,-.00887697561402263,-.0149434968427065,-.0203201678410382,-.0221953581292932,-.0184628989656689,-.0091794901054094,-3.26296210833654e-009,-1.03996283213943e-006,-2.15142568085075e-005,-.00015752876308832,-.000650774647452403,-.00183988793598998,-.00391599979728114,-.00659219233567233,-.00896406350610278,-.00979128722597762,-.00814474566006138,-.00404945140720489},
			{-5.21061086814239e-011,-1.66071240047969e-008,-3.43560288548878e-007,-2.51557038586478e-006,-1.03921937740666e-005,-2.93810953272788e-005,-6.25344408726795e-005,-.000105270450249422,-.000143146764126628,-.000156356665933801,-.000130063110895906,-6.46655241827284e-005,-6.14193874705368e-010,-1.95754280991916e-007,-4.04967152908804e-006,-2.96519537053663e-005,-.000122496611669995,-.000346325780965151,-.00073711646319537,-.00124086152980604,-.00168732357751008,-.00184303355050058,-.0015331017429415,-.000762236326271721,-2.20633662270343e-009,-7.03197894004764e-007,-1.45474238225389e-005,-.00010651716679239,-.000440038189267583,-.00124408804027283,-.00264790502628461,-.00445748215614846,-.00606128448480099,-.00662063329969093,-.00550728143194998,-.00273814180028246,-5.02223790632381e-009,-1.6006746579724e-006,-3.31139965720151e-005,-.00024246279884684,-.0010016497263513,-.00283189158461527,-.00602737082751107,-.0101464733989415,-.013797175094379,-.0150704091023226,-.0125361095327521,-.00623277490876185,-8.84165345126754e-009,-2.81798888822008e-006,-5.82972148148455e-005,-.000426855931999716,-.00176340506466325,-.00498554717433864,-.0106111906829482,-.0178628737266676,-.0242899366908432,-.0265314660788539,-.0220698298613373,-.0109728047159291,-1.30549605302744e-008,-4.16084320803041e-006,-8.60775467651508e-005,-.000630265297671321,-.00260371927546702,-.00736130656347784,-.0156677341301185,-.0263750567407152,-.0358648036285977,-.039174487484589,-.0325867508082808,-.0162016678512013,-1.67929112501503e-008,-5.3521931802253e-006,-.00011072362877741,-.00081072548502156,-.00334922703227079,-.0094690265450411,-.0201537850786976,-.0339268729336555,-.0461337637093912,-.0503910900437627,-.0419171251406854,-.0208405969285544,-1.91308046425978e-008,-6.09732050715256e-006,-.000126138468780461,-.000923593928514568,-.00381550328609678,-.0107872955612227,-.0229595761810517,-.0386501404407882,-.0525564631293,-.0574064904526992,-.0477527880842341,-.0237420053340501,-1.93172649069964e-008,-6.15674863969202e-006,-.000127367889742049,-.000932595827353905,-.00385269146321899,-.0108924349957666,-.0231833539428959,-.0390268478266128,-.0530687098535765,-.0579660084441261,-.0482182153183887,-.0239734091183992,-1.69710841924047e-008,-5.40897999891501e-006,-.00011189840749341,-.000819327289847131,-.00338476236177552,-.00956949300345818,-.020367616923072,-.0342868373663876,-.0466232433651179,-.0509257399709112,-.0423618662225568,-.0210617158529497,-1.21971598987849e-008,-3.88744721245458e-006,-8.04216603455029e-005,-.00058885253590198,-.00243263702412373,-.00687761812917763,-.0146382563159452,-.0246420342413141,-.0335082395371539,-.0366004544168802,-.0304455773166633,-.0151371069219119,-5.5995100454324e-009,-1.78466133902173e-006,-3.69202255862743e-005,-.000270332250902931,-.00111678255975218,-.00315739828964731,-.00672017616961103,-.0113127416069928,-.0153830666688022,-.0168026502788677,-.0139770501853164,-.00694919005501226},
			{-6.20196054366658e-013,-1.97667280147994e-010,-4.08924674643852e-009,-2.99417259756176e-008,-1.23693703828388e-007,-3.49710232755209e-007,-7.44319897852733e-007,-1.25298778853868e-006,-1.70381286481175e-006,-1.86104450591298e-006,-1.54808390489254e-006,-7.69685243564878e-007,-3.80137148576797e-011,-1.21156327444106e-008,-2.50642452023579e-007,-1.83522004948295e-006,-7.58156578700641e-006,-2.14348107782544e-005,-4.56216452211473e-005,-7.67994574913963e-005,-.000104431906584671,-.000114069115221109,-9.48868018782288e-005,-4.71763649140235e-005,-3.27700968053241e-010,-1.04444003796667e-007,-2.16068791147813e-006,-1.58206949533316e-005,-6.53576336083714e-005,-.00018478089469472,-.000393285880086281,-.000662057277488362,-.000900265522892682,-.000983344027882673,-.00081798100889079,-.000406688494124778,-1.33784942253663e-009,-4.263959029382e-007,-8.82107578694501e-006,-6.45884805686753e-005,-.000266824577604282,-.000754373765606237,-.00160560187139806,-.00270286948383404,-.00367536207502424,-.00401453266272909,-.00333943298029159,-.00166031846121592,-3.60795174491446e-009,-1.14991703559086e-006,-2.37889371116125e-005,-.000174184117617127,-.000719580383364912,-.00203441740011081,-.00433003443882712,-.00728917806907233,-.00991182474529423,-.0108265099805079,-.00900588126384461,-.00447759575057386,-7.37160626866756e-009,-2.34945925758498e-006,-4.86044964941899e-005,-.000355885229102249,-.0014702145815281,-.00415663099731916,-.00884693345962729,-.0148929238932162,-.0202513987414977,-.0221202428642781,-.0184004153805573,-.00914842415783806,-1.21972798977825e-008,-3.8874854582243e-006,-8.04224515558099e-005,-.000588858329194405,-.00243266095707262,-.00687768579306728,-.0146384003311175,-.0246422766763918,-.0335085692003862,-.0366008145021687,-.0304458768485863,-.0151372558449131,-1.68911888154202e-008,-5.3835159512905e-006,-.000111371619378443,-.000815470113608189,-.00336882779555505,-.00952444236070643,-.0202717315680825,-.0341254240019886,-.0464037534631961,-.05068599505261,-.0421624377574542,-.0209625628637279,-1.97926499253432e-008,-6.30826212149848e-006,-.000130502328643898,-.00095554638928031,-.00394750363309629,-.0111604905634277,-.023753880836457,-.0399872725481971,-.0543746954436094,-.0593925132894224,-.0494048334697752,-.0245633787434188,-1.93889081179464e-008,-6.179582578331e-006,-.000127840267412333,-.000936054606840123,-.00386698019345867,-.0109328324858832,-.0232693355725446,-.0391715892641307,-.0532655292683946,-.0581809907923459,-.0483970453850838,-.0240623208776429,-1.49915585521718e-008,-4.77807071380575e-006,-9.88464560551176e-005,-.000723760067411173,-.00298996001413125,-.00845329697557845,-.0179919160266198,-.030287583522627,-.0411850577650857,-.0449857065070876,-.0374207322675915,-.0186050545055932,-7.16228788345407e-009,-2.28274584398164e-006,-4.72243611004261e-005,-.000345779789560036,-.00142846751434807,-.00403860254373974,-.00859572283356178,-.0144700360358322,-.0196763558094162,-.021492133691307,-.0178779315290952,-.00888865263692734},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.81377664478188e-011,-6.32055938620332e-008,-3.13787157581542e-006,-4.12071599264361e-005,-.000260771859326372,-.00102018939735755,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404,-8.64525128335375e-011,-1.43277462826387e-007,-7.11307734943901e-006,-9.34103607574427e-005,-.000591130121526243,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-1.29411477385236e-010,-2.14473212317988e-007,-1.06476239773248e-005,-.000139826737158981,-.000884867539951351,-.00346177108471816,-.00947763686126095,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-1.64245810125438e-010,-2.72204036451189e-007,-1.35136980227866e-005,-.000177464597313172,-.00112305174849657,-.00439359327137407,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.88761417029128e-010,-3.12833670474357e-007,-1.55307754038705e-005,-.000203953262708829,-.00129068034844452,-.00504938842044633,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-2.01416927810585e-010,-3.33807606524435e-007,-1.65720363705528e-005,-.000217627310910781,-.00137721402318714,-.00538792470927311,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-2.01416927810585e-010,-3.33807606524435e-007,-1.65720363705528e-005,-.000217627310910781,-.00137721402318714,-.00538792470927311,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-1.88761417029128e-010,-3.12833670474357e-007,-1.55307754038705e-005,-.000203953262708829,-.00129068034844452,-.00504938842044633,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-1.64245810125438e-010,-2.72204036451189e-007,-1.35136980227866e-005,-.000177464597313172,-.00112305174849657,-.00439359327137407,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.29411477385236e-010,-2.14473212317988e-007,-1.06476239773248e-005,-.000139826737158981,-.000884867539951351,-.00346177108471816,-.00947763686126095,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-8.64525128335375e-011,-1.43277462826387e-007,-7.11307734943901e-006,-9.34103607574427e-005,-.000591130121526243,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-3.81377664478188e-011,-6.32055938620331e-008,-3.13787157581542e-006,-4.12071599264361e-005,-.000260771859326372,-.00102018939735755,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404},
			{-6.09020435343665e-013,-1.00932754786986e-009,-5.01085430834115e-008,-6.58035454488694e-007,-4.16425517497517e-006,-1.62913628348369e-005,-4.46024931010705e-005,-9.12717393010351e-005,-.000144026600260793,-.000175413507334088,-.000156980270065097,-8.12226984453021e-005,-7.17874795152048e-012,-1.18973151739896e-008,-5.90647834026672e-007,-7.75650601653937e-006,-4.90856079239649e-005,-.000192032287901912,-.000525745997014592,-.0010758535732373,-.00169769452974626,-.00206766355177068,-.00185038420184216,-.000957401831273779,-2.57878418572552e-011,-4.27381047927422e-008,-2.12175340882166e-006,-2.78632920211386e-005,-.000176327669275857,-.000689827572348814,-.00188860992469107,-.00386473267978856,-.00609853951561999,-.00742755993763223,-.00664703866113062,-.00343922466500658,-5.87003250388414e-011,-9.72838540256606e-008,-4.82970290571489e-006,-6.34246288365862e-005,-.000401370985487183,-.00157024007444182,-.00429900326923858,-.00879721016390503,-.0138819779418027,-.016907199330518,-.0151305150742709,-.00782863516982087,-1.03341964508827e-010,-1.71268295079276e-007,-8.50269544402536e-006,-.000111659104747358,-.000706613909031054,-.00276440878199448,-.0075683983517811,-.0154874948296011,-.0244392321648105,-.0297651365985099,-.0266372826857591,-.0137823178549157,-1.52587440258738e-010,-2.52882658732434e-007,-1.25544790953984e-005,-.00016486794165333,-.00104333615238232,-.00408173061038807,-.0111749620480468,-.0228677401605563,-.0360852427727842,-.0439490968079129,-.0393307288068947,-.0203499963670061,-1.96276912228913e-010,-3.25289075746179e-007,-1.6149129884536e-005,-.000212073617975246,-.00134206850877865,-.0052504287338376,-.0143746237655383,-.029415326849682,-.0464173199082629,-.0565327854119549,-.0505920670327779,-.0261766921577074,-2.23602400308645e-010,-3.7057551652431e-007,-1.83973966375979e-005,-.000241598308649254,-.0015289100309035,-.00598138850975171,-.0163758454369784,-.0335105011321007,-.0528794957568774,-.0644032269037527,-.0576354472700868,-.0298209867484436,-2.25781763041192e-010,-3.74187367153787e-007,-1.85767086689201e-005,-.000243953070268046,-.00154381170252316,-.00603968669970425,-.0165354542211866,-.033837114519146,-.0533948909503518,-.0650309392734592,-.0581971968115976,-.0301116399215653,-1.98359412066345e-010,-3.28740395820796e-007,-1.63204723006016e-005,-.000214323721005407,-.00135630786795552,-.00530613583081857,-.0145271386554198,-.0297274237371193,-.046909808097858,-.0571325987832424,-.0511288493276215,-.026454427101384,-1.42561396724728e-010,-2.36266530031746e-007,-1.17295635339088e-005,-.000154034984775778,-.000974781796488713,-.00381353285620554,-.0104406902377677,-.0213651724656962,-.0337141943145245,-.0410613390925373,-.0367464296104888,-.0190128617434316,-6.54475287423684e-011,-1.08465972348532e-007,-5.38484445409324e-006,-7.0714859176842e-005,-.000447505854382295,-.00175072850680893,-.00479314449861828,-.00980838972652806,-.0154776170276039,-.0188505670693449,-.0168696650240853,-.0087284836148921},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-4.25830680908177e-013,-3.66970948242474e-009,-4.37202313877829e-007,-1.02973144060446e-005,-9.98226951437858e-005,-.000540391748990977,-.00190309670559272,-.00473399758463878,-.00866893492130421,-.0117725765312941,-.011334382523687,-.00610300723148638,-9.65293351840578e-013,-8.31867295004696e-009,-9.91071113277978e-007,-2.33424447406373e-005,-.000226282859139225,-.00122498585958623,-.00431403062339345,-.0107312521172389,-.0196511562511815,-.026686639477119,-.0256933203450701,-.0138345886544025,-1.44495555626435e-012,-1.24522899458459e-008,-1.48354249933819e-006,-3.49414974841223e-005,-.000338724672636725,-.00183369140663751,-.00645770791571394,-.0160636995405713,-.0294159773896932,-.0399474500854692,-.0384605425083467,-.0207091095227104,-1.83390144930814e-012,-1.58041349298872e-008,-1.88287503228147e-006,-4.43468746144642e-005,-.000429900881983438,-.00232727526714406,-.00819596135984531,-.0203876456552513,-.0373340227205607,-.0507003044420999,-.0488131585372653,-.0262834838088483,-2.10763267567556e-012,-1.81630867905044e-008,-2.16391613832206e-006,-5.09661639870523e-005,-.000494068613398971,-.00267464830248913,-.00941930438906687,-.0234307400647473,-.0429065619801723,-.0582679174767988,-.0560990929882762,-.030206600974669,-2.24893892602145e-012,-1.93808310960915e-008,-2.30899591389123e-006,-5.43831908772887e-005,-.0005271934477113,-.00285397012027107,-.0100508217305121,-.0250016542281072,-.0457832327865349,-.0621744904907474,-.0598602571462754,-.0322318027893339,-2.24893892602145e-012,-1.93808310960915e-008,-2.30899591389123e-006,-5.43831908772887e-005,-.0005271934477113,-.00285397012027107,-.0100508217305121,-.0250016542281072,-.0457832327865349,-.0621744904907474,-.0598602571462754,-.0322318027893339,-2.10763267567556e-012,-1.81630867905044e-008,-2.16391613832206e-006,-5.09661639870523e-005,-.000494068613398971,-.00267464830248913,-.00941930438906688,-.0234307400647473,-.0429065619801723,-.0582679174767988,-.0560990929882762,-.030206600974669,-1.83390144930814e-012,-1.58041349298872e-008,-1.88287503228147e-006,-4.43468746144641e-005,-.000429900881983438,-.00232727526714406,-.00819596135984532,-.0203876456552513,-.0373340227205607,-.0507003044420999,-.0488131585372653,-.0262834838088483,-1.44495555626435e-012,-1.24522899458459e-008,-1.48354249933819e-006,-3.49414974841223e-005,-.000338724672636725,-.00183369140663751,-.00645770791571394,-.0160636995405713,-.0294159773896932,-.0399474500854692,-.0384605425083467,-.0207091095227104,-9.65293351840578e-013,-8.31867295004696e-009,-9.91071113277978e-007,-2.33424447406373e-005,-.000226282859139225,-.00122498585958623,-.00431403062339345,-.0107312521172389,-.0196511562511815,-.026686639477119,-.0256933203450701,-.0138345886544025,-4.25830680908177e-013,-3.66970948242474e-009,-4.37202313877829e-007,-1.02973144060446e-005,-9.98226951437858e-005,-.000540391748990977,-.00190309670559272,-.00473399758463878,-.0086689349213042,-.0117725765312941,-.011334382523687,-.00610300723148638}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 144; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{-1.25649763607893e-005,-2.84828893052902e-005,-4.2636271224355e-005,-5.41128889760362e-005,-6.2189870139491e-005,-6.63593905024702e-005,-6.63593905024702e-005,-6.2189870139491e-005,-5.41128889760362e-005,-4.2636271224355e-005,-2.84828893052902e-005,-1.25649763607893e-005,-.000148107999463138,-.000335738296102257,-.000502569416311285,-.000637848532403299,-.000733054881183536,-.000782202551815753,-.000782202551815753,-.000733054881183536,-.000637848532403299,-.000502569416311285,-.000335738296102257,-.000148107999463138,-.000532040641869991,-.00120605517059218,-.00180535390260964,-.00229131001583844,-.0026333148170561,-.00280986543096192,-.00280986543096192,-.0026333148170561,-.00229131001583844,-.00180535390260964,-.00120605517059218,-.000532040641869991,-.00121107298487856,-.00274531815886015,-.00410948932756509,-.00521566106380589,-.00599415943941434,-.00639603790909957,-.00639603790909957,-.00599415943941434,-.00521566106380588,-.00410948932756509,-.00274531815886015,-.00121107298487856,-.00213209486213416,-.00483313459594366,-.00723475892097063,-.00918217505932459,-.010552722010284,-.0112602293456112,-.0112602293456112,-.010552722010284,-.00918217505932459,-.00723475892097063,-.00483313459594367,-.00213209486213416,-.00314810057025839,-.0071362649232201,-.0106823336471961,-.0135577506675935,-.0155814034208121,-.0166260587433132,-.0166260587433132,-.0155814034208121,-.0135577506675935,-.0106823336471961,-.0071362649232201,-.00314810057025839,-.00404947784869215,-.00917955004423722,-.0137409439473876,-.0174396623555152,-.0200427358008248,-.0213865011900097,-.0213865011900097,-.0200427358008248,-.0174396623555152,-.0137409439473876,-.00917955004423722,-.00404947784869215,-.00461324236601103,-.0104575184128172,-.0156539453074288,-.0198675958317378,-.0228330667266135,-.0243639098760258,-.0243639098760258,-.0228330667266135,-.0198675958317378,-.0156539453074288,-.0104575184128172,-.00461324236601103,-.00465820578534291,-.0105594436420292,-.0158065180211977,-.0200612373037489,-.023055611451648,-.0246013751140091,-.0246013751140091,-.023055611451648,-.0200612373037489,-.0158065180211977,-.0105594436420292,-.00465820578534291,-.00409244284577621,-.00927694515432801,-.0138867354885914,-.0176246973329244,-.0202553894113433,-.0216134121636217,-.0216134121636217,-.0202553894113433,-.0176246973329244,-.0138867354885914,-.00927694515432801,-.00409244284577621,-.00294124872640195,-.00666736327135993,-.0099804309086089,-.0126669132684861,-.0145575981275773,-.0155336124156416,-.0155336124156416,-.0145575981275773,-.0126669132684861,-.0099804309086089,-.00666736327135993,-.00294124872640195,-.00135027759956182,-.0030608738365598,-.00458184721642184,-.00581516588124543,-.0066831473580057,-.00713121902844807,-.00713121902844807,-.0066831473580057,-.00581516588124543,-.00458184721642184,-.0030608738365598,-.00135027759956182},
			{-1.49555377634105e-007,-3.39019282157687e-007,-5.07480751318181e-007,-6.4408187594673e-007,-7.40218624107564e-007,-7.89846587943962e-007,-7.89846587943962e-007,-7.40218624107564e-007,-6.4408187594673e-007,-5.07480751318181e-007,-3.39019282157687e-007,-1.49555377634105e-007,-9.16670694821036e-006,-2.07795296865571e-005,-3.11050488640572e-005,-3.94777499870446e-005,-4.53702655975518e-005,-4.84121154335211e-005,-4.84121154335211e-005,-4.53702655975518e-005,-3.94777499870446e-005,-3.11050488640572e-005,-2.07795296865571e-005,-9.16670694821036e-006,-7.90224989069185e-005,-.000179131979588681,-.000268144133301817,-.000340321826892032,-.000391118837314757,-.000417341403029572,-.000417341403029572,-.000391118837314757,-.000340321826892032,-.000268144133301817,-.000179131979588681,-7.90224989069185e-005,-.000322611816370485,-.000731311893505376,-.0010947067871833,-.00138937447237002,-.00159675485169678,-.00170380929406662,-.00170380929406662,-.00159675485169678,-.00138937447237002,-.0010947067871833,-.000731311893505376,-.000322611816370485,-.000870029052744199,-.00197222346386832,-.00295223752124433,-.00374690601758642,-.00430617553540284,-.00459488311014376,-.00459488311014376,-.00430617553540284,-.00374690601758643,-.00295223752124433,-.00197222346386832,-.000870029052744198,-.00177760460022009,-.00402955911756787,-.00603188017934985,-.00765551144808981,-.0087981860110801,-.00938806046568492,-.00938806046568492,-.0087981860110801,-.00765551144808981,-.00603188017934986,-.00402955911756787,-.00177760460022009,-.0029412776627297,-.00666742886561148,-.00998052909718346,-.0126670378869689,-.0145577413468253,-.0155337652370252,-.0155337652370252,-.0145577413468253,-.0126670378869689,-.00998052909718346,-.00666742886561148,-.0029412776627297,-.00407317670628986,-.00923327174798222,-.013821360407493,-.0175417249152225,-.0201600323904989,-.0215116619305183,-.0215116619305183,-.0201600323904989,-.0175417249152225,-.013821360407493,-.00923327174798222,-.00407317670628986,-.00477284112519419,-.0108193045125715,-.0161955058952257,-.0205549309837037,-.0236229946837403,-.0252067985596364,-.0252067985596364,-.0236229946837403,-.0205549309837037,-.0161955058952257,-.0108193045125715,-.00477284112519419,-.00467548197876496,-.0105986061434719,-.0158651406916522,-.0201356397307597,-.0231411193105227,-.0246926158479963,-.0246926158479963,-.0231411193105227,-.0201356397307597,-.0158651406916522,-.0105986061434719,-.00467548197876496,-.00361509587945292,-.0081948722230624,-.0122669716195725,-.015568933545552,-.0178927788504699,-.0190924003579595,-.0190924003579595,-.0178927788504699,-.015568933545552,-.0122669716195725,-.00819487222306241,-.00361509587945292,-.00172712912568916,-.00391513889802946,-.00586060361181134,-.00743813151271418,-.00854835957956341,-.00912148441898038,-.00912148441898038,-.00854835957956341,-.00743813151271418,-.00586060361181134,-.00391513889802946,-.00172712912568916},
			{-1.63148105416827e-009,-3.69831927536025e-009,-5.53604453566559e-009,-7.02620925147273e-009,-8.07495310619039e-009,-8.61633840464567e-009,-8.61633840464567e-009,-8.07495310619039e-009,-7.02620925147273e-009,-5.53604453566559e-009,-3.69831927536025e-009,-1.63148105416827e-009,-5.19981416069715e-007,-1.1787187408438e-006,-1.7644337761237e-006,-2.23937521483844e-006,-2.57362813998106e-006,-2.74617706012381e-006,-2.74617706012381e-006,-2.57362813998106e-006,-2.23937521483844e-006,-1.7644337761237e-006,-1.1787187408438e-006,-5.19981416069715e-007,-1.07571284042538e-005,-2.43847731013085e-005,-3.6501767378587e-005,-4.63271301374563e-005,-5.3241995792529e-005,-5.68116058451738e-005,-5.68116058451738e-005,-5.3241995792529e-005,-4.63271301374563e-005,-3.6501767378587e-005,-2.43847731013085e-005,-1.07571284042538e-005,-7.87643815441602e-005,-.000178546866806921,-.000267268273167243,-.000339210207117138,-.000389841295295595,-.000415978208195064,-.000415978208195064,-.000389841295295595,-.000339210207117138,-.000267268273167243,-.000178546866806921,-7.87643815441602e-005,-.000325387323726201,-.000737603546311472,-.00110412481400688,-.0014013275964412,-.00161049211924643,-.00171846757683388,-.00171846757683388,-.00161049211924644,-.0014013275964412,-.00110412481400688,-.000737603546311472,-.000325387323726201,-.000919943967994991,-.00208537298082308,-.00312161196363602,-.0039618718233036,-.00455322750019265,-.00485849867597654,-.00485849867597654,-.00455322750019265,-.0039618718233036,-.00312161196363602,-.00208537298082308,-.000919943967994991,-.00195799989864057,-.00443848780701132,-.00664400889731993,-.00843240990574942,-.00969104564411167,-.0103407818802712,-.0103407818802712,-.00969104564411167,-.00843240990574942,-.00664400889731993,-.00443848780701132,-.00195799989864057,-.00329609616783617,-.00747174842135327,-.0111845216543321,-.0141951151250118,-.0163139019731612,-.0174076676672231,-.0174076676672231,-.0163139019731612,-.0141951151250118,-.0111845216543321,-.00747174842135327,-.00329609616783617,-.00448203175305139,-.0101600839205191,-.0152087131700155,-.0193025183395333,-.022183644813943,-.0236709474657956,-.0236709474657956,-.022183644813943,-.0193025183395333,-.0152087131700155,-.0101600839205191,-.00448203175305139,-.00489564361298881,-.0110976790646466,-.0166122070513835,-.0210837976681447,-.0242308009023486,-.0258553551512484,-.0258553551512484,-.0242308009023486,-.0210837976681447,-.0166122070513835,-.0110976790646466,-.00489564361298881,-.00407237283003069,-.00923144948283444,-.0138186326438083,-.0175382629057834,-.0201560536356863,-.021507416420056,-.021507416420056,-.0201560536356863,-.0175382629057834,-.0138186326438083,-.00923144948283444,-.00407237283003069,-.00202472570360245,-.0045897450527047,-.00687042711223165,-.00871977424071199,-.0100213024648473,-.0106931807723117,-.0106931807723117,-.0100213024648473,-.00871977424071199,-.00687042711223165,-.0045897450527047,-.00202472570360245},
			{-1.70557276538753e-011,-3.86627390961513e-011,-5.7874572088226e-011,-7.34529592765333e-011,-8.44166719834875e-011,-9.00763884617597e-011,-9.00763884617597e-011,-8.44166719834875e-011,-7.34529592765333e-011,-5.7874572088226e-011,-3.86627390961513e-011,-1.70557276538753e-011,-2.82664008807978e-008,-6.40756292912074e-008,-9.59153363989562e-008,-1.21733345825304e-007,-1.39903470536826e-007,-1.49283299887593e-007,-1.49283299887593e-007,-1.39903470536826e-007,-1.21733345825304e-007,-9.59153363989562e-008,-6.40756292912074e-008,-2.82664008807978e-008,-1.40329882934204e-006,-3.18106489584209e-006,-4.76176220142831e-006,-6.04350947999848e-006,-6.94557390780468e-006,-7.41123996847038e-006,-7.41123996847038e-006,-6.94557390780468e-006,-6.04350947999848e-006,-4.76176220142831e-006,-3.18106489584208e-006,-1.40329882934204e-006,-1.84284021471628e-005,-4.17743832824876e-005,-6.2532417858728e-005,-7.93645806216638e-005,-9.12106719107567e-005,-9.73258921709034e-005,-9.73258921709034e-005,-9.12106719107567e-005,-7.93645806216638e-005,-6.2532417858728e-005,-4.17743832824876e-005,-1.84284021471628e-005,-.000116620720789999,-.000264361427000411,-.000395724793999518,-.000502244010271906,-.000577209799147466,-.000615908834952788,-.000615908834952788,-.000577209799147466,-.000502244010271906,-.000395724793999518,-.000264361427000411,-.000116620720789999,-.000456242568387133,-.00103423247275538,-.0015481510932686,-.00196487464364187,-.00225815515010815,-.00240955318100971,-.00240955318100971,-.00225815515010815,-.00196487464364187,-.0015481510932686,-.00103423247275538,-.000456242568387133,-.00124910090182185,-.00283152165958353,-.00423852805667493,-.00537943379114375,-.00618237716052034,-.00659687468888933,-.00659687468888933,-.00618237716052034,-.00537943379114375,-.00423852805667493,-.00283152165958353,-.00124910090182185,-.0025560815986996,-.00579424800656579,-.00867345748885218,-.0110081352954835,-.012651228153929,-.0134994298511859,-.0134994298511859,-.012651228153929,-.0110081352954835,-.00867345748885218,-.00579424800656579,-.0025560815986996,-.00403349103971441,-.00914331038111467,-.0136866964976499,-.0173708128492007,-.0199636097009642,-.021302070080123,-.021302070080123,-.0199636097009642,-.0173708128492007,-.0136866964976499,-.00914331038111467,-.00403349103971441,-.00491248705999991,-.0111358605958299,-.0166693612993899,-.021156336410114,-.0243141668994061,-.025944310422275,-.025944310422275,-.0243141668994061,-.021156336410114,-.0166693612993899,-.0111358605958299,-.00491248705999991,-.00439626090995001,-.00996565447158663,-.0149176701290573,-.0189331337918146,-.0217591253050265,-.0232179660428578,-.0232179660428578,-.0217591253050265,-.0189331337918146,-.0149176701290573,-.00996565447158663,-.00439626090995001,-.00227465638852363,-.00515629988163542,-.00771850769460683,-.00979613690280579,-.0112583248350307,-.012013139317637,-.012013139317637,-.0112583248350307,-.00979613690280579,-.00771850769460683,-.00515629988163542,-.00227465638852363},
			{-1.7384464750039e-013,-3.94079360668192e-013,-5.8990063561595e-013,-7.48687131527224e-013,-8.60437436733519e-013,-9.18125471866621e-013,-9.18125471866621e-013,-8.60437436733519e-013,-7.48687131527224e-013,-5.8990063561595e-013,-3.94079360668192e-013,-1.7384464750039e-013,-1.49815262263485e-009,-3.39608401063939e-009,-5.0836260825345e-009,-6.45201106711068e-009,-7.41504913145304e-009,-7.91219116240981e-009,-7.91219116240981e-009,-7.41504913145304e-009,-6.45201106711068e-009,-5.0836260825345e-009,-3.39608401063939e-009,-1.49815262263485e-009,-1.78487097219834e-007,-4.04603087706546e-007,-6.05653689159397e-007,-7.68680513053121e-007,-8.83415064139356e-007,-9.42643634493817e-007,-9.42643634493817e-007,-8.83415064139356e-007,-7.68680513053121e-007,-6.05653689159397e-007,-4.04603087706546e-007,-1.78487097219834e-007,-4.20386100245681e-006,-9.52951316020476e-006,-1.42648066135301e-005,-1.81045357479959e-005,-2.0806849318325e-005,-2.22018447046734e-005,-2.22018447046734e-005,-2.0806849318325e-005,-1.81045357479959e-005,-1.42648066135301e-005,-9.52951316020476e-006,-4.20386100245681e-006,-4.07524446402028e-005,-9.23795904009122e-005,-.00013828376853596,-.000175506300131138,-.000201702666783354,-.000215225823762682,-.000215225823762682,-.000201702666783354,-.000175506300131138,-.00013828376853596,-9.23795904009122e-005,-4.07524446402028e-005,-.000220614007696907,-.000500098382997093,-.000748601381966019,-.000950106149209752,-.00109192059703656,-.00116512842258577,-.00116512842258577,-.00109192059703656,-.000950106149209752,-.00074860138196602,-.000500098382997093,-.000220614007696907,-.000776935976612406,-.0017611956269338,-.0026363482167924,-.00334598721372161,-.00384541491403083,-.00410323078906309,-.00410323078906309,-.00384541491403083,-.00334598721372161,-.0026363482167924,-.0017611956269338,-.000776935976612406,-.00193264642090617,-.00438101533121211,-.00655797787568285,-.00832322148498332,-.00956555957532708,-.0102068825969568,-.0102068825969568,-.00956555957532708,-.00832322148498332,-.00655797787568285,-.00438101533121211,-.00193264642090617,-.00353907786161354,-.00802255094483993,-.0120090224811522,-.015241550951156,-.0175165305773379,-.0186909265162125,-.0186909265162125,-.0175165305773379,-.015241550951156,-.0120090224811522,-.00802255094483993,-.00353907786161354,-.00480613424305024,-.0108947749442933,-.0163084782050857,-.0206983126135993,-.0237877776977751,-.0253826294522236,-.0253826294522236,-.0237877776977751,-.0206983126135993,-.0163084782050857,-.0108947749442933,-.00480613424305024,-.00462724228856079,-.0104892541067666,-.0157014507286737,-.0199278885241277,-.0229023588080574,-.0244378476456489,-.0244378476456489,-.0229023588080574,-.0199278885241277,-.0157014507286737,-.0104892541067666,-.00462724228856079,-.00249154226883633,-.00564794716718901,-.00845445855931378,-.010730187331938,-.0123317932081052,-.0131585783867511,-.0131585783867511,-.0123317932081052,-.010730187331938,-.00845445855931378,-.00564794716718901,-.00249154226883633},
			{-.0019095808945752,-.00415955383985738,-.00578758817666173,-.00658769779901337,-.00652410992157336,-.0057268264950684,-.00445208652288796,-.00301523747077974,-.00171271584051363,-.000752419091586911,-.000209456341570312,-1.77695599831757e-005,-.00432872929307012,-.0094290755651378,-.0131195816567463,-.0149333084120673,-.0147891643705445,-.012981844171125,-.010092202640724,-.00683508449545443,-.00388246617391187,-.00170561957949523,-.000474805651832001,-4.02808883575737e-005,-.00647971047522447,-.0141144607515462,-.0196388096682036,-.0223537921630579,-.0221380217619632,-.0194326292933208,-.0151071011240822,-.0102314941881775,-.0058116955424022,-.00255315597436325,-.000710740484695346,-6.02967930241717e-005,-.00822388645802562,-.0179137205405699,-.0249250860049403,-.0283708738774972,-.0280970234814972,-.0246634070303329,-.0191735548724589,-.0129855565030636,-.00737605861435863,-.00324040170051977,-.000902054045409203,-7.65271815013742e-005,-.00945139763454551,-.0205875527108991,-.0286454464212646,-.0326055584089547,-.0322908326405259,-.0283447088011335,-.0220354320420534,-.0149238025892903,-.00847702157560611,-.00372406952887607,-.00103669615509983,-8.79497578075961e-005,-.0100850666679019,-.0219678453543708,-.030565980615856,-.0347915983448399,-.0344557717846379,-.0302450800394711,-.0235127977673766,-.015924369058549,-.00904536355795187,-.00397374980154702,-.00110620145747811,-9.38463500544579e-005,-.0100850666679019,-.0219678453543708,-.030565980615856,-.0347915983448399,-.0344557717846379,-.0302450800394712,-.0235127977673766,-.015924369058549,-.00904536355795187,-.00397374980154702,-.00110620145747811,-9.38463500544579e-005,-.00945139763454551,-.0205875527108991,-.0286454464212646,-.0326055584089547,-.0322908326405259,-.0283447088011335,-.0220354320420534,-.0149238025892903,-.00847702157560611,-.00372406952887607,-.00103669615509983,-8.79497578075961e-005,-.00822388645802562,-.0179137205405699,-.0249250860049403,-.0283708738774972,-.0280970234814972,-.0246634070303329,-.0191735548724589,-.0129855565030636,-.00737605861435863,-.00324040170051977,-.000902054045409203,-7.65271815013742e-005,-.00647971047522447,-.0141144607515462,-.0196388096682036,-.0223537921630579,-.0221380217619632,-.0194326292933208,-.0151071011240822,-.0102314941881775,-.0058116955424022,-.00255315597436325,-.000710740484695346,-6.02967930241717e-005,-.00432872929307012,-.0094290755651378,-.0131195816567463,-.0149333084120673,-.0147891643705445,-.012981844171125,-.010092202640724,-.00683508449545443,-.00388246617391187,-.00170561957949523,-.000474805651832001,-4.02808883575737e-005,-.0019095808945752,-.00415955383985738,-.00578758817666173,-.00658769779901337,-.00652410992157336,-.0057268264950684,-.00445208652288796,-.00301523747077974,-.00171271584051363,-.000752419091586911,-.000209456341570312,-1.77695599831757e-005},
			{-3.04940193335369e-005,-6.64237454259384e-005,-9.2421759274531e-005,-.000105198677163819,-.000104183244945397,-9.14514584621857e-005,-7.10951878965703e-005,-4.815020405282e-005,-2.73502893236142e-005,-1.20153497508037e-005,-3.34479976602295e-006,-2.83761377805336e-007,-.000359444225710958,-.000782961126977123,-.00108940928179785,-.00124001551403817,-.00122804624086627,-.00107797198911481,-.00083802513816711,-.000567564171331005,-.000322387923394712,-.000141629348384198,-3.94263855120624e-005,-3.34479976602295e-006,-.00129121274513816,-.00281259042095047,-.00391342814465003,-.00445444305782967,-.004411446462074,-.00387234255465894,-.00301039399647408,-.00203883117125292,-.00115809732300615,-.000508767721494418,-.000141629348384198,-1.20153497508037e-005,-.00293916056463564,-.00640224074681438,-.00890805463205047,-.0101395555630045,-.0100416833114065,-.00881453236290365,-.00685249688850132,-.00464094875074617,-.00263615271349073,-.00115809732300615,-.000322387923394712,-2.73502893236141e-005,-.00517439429092323,-.0112711494458756,-.0156826365956051,-.017850694871518,-.0176783907020889,-.0155179905734142,-.0120638257076051,-.00817039361825585,-.00464094875074617,-.00203883117125292,-.000567564171331005,-4.815020405282e-005,-.00764014486751864,-.0166421825914966,-.0231558726990041,-.0263570743040411,-.0261026621464626,-.0229127680203357,-.0178125923307157,-.0120638257076051,-.00685249688850132,-.00301039399647408,-.00083802513816711,-7.10951878965703e-005,-.00982770299465891,-.0214072416856188,-.0297859586976462,-.0339037416907514,-.0335764851836737,-.0294732472739763,-.0229127680203357,-.0155179905734142,-.00881453236290365,-.00387234255465894,-.00107797198911481,-9.14514584621857e-005,-.0111959066105712,-.0243875378440283,-.0339327319991675,-.0386237888878892,-.0382509720360841,-.0335764851836737,-.0261026621464626,-.0176783907020889,-.0100416833114065,-.004411446462074,-.00122804624086627,-.000104183244945397,-.0113050286127969,-.02462523337432,-.0342634607007844,-.039000239435718,-.0386237888878892,-.0339037416907514,-.0263570743040411,-.017850694871518,-.0101395555630045,-.00445444305782967,-.00124001551403817,-.000105198677163819,-.00993197501349335,-.0216343727160793,-.0301019879923868,-.0342634607007844,-.0339327319991675,-.0297859586976462,-.0231558726990041,-.0156826365956051,-.00890805463205047,-.00391342814465003,-.00108940928179785,-9.2421759274531e-005,-.00713813484023199,-.0155486768161827,-.0216343727160793,-.02462523337432,-.0243875378440283,-.0214072416856188,-.0166421825914966,-.0112711494458756,-.00640224074681438,-.00281259042095047,-.000782961126977123,-6.64237454259384e-005,-.00327699711041027,-.00713813484023199,-.00993197501349335,-.0113050286127969,-.0111959066105712,-.00982770299465891,-.00764014486751864,-.00517439429092323,-.00293916056463564,-.00129121274513816,-.000359444225710958,-3.04940193335369e-005},
			{-3.62956876834475e-007,-7.90612576313682e-007,-1.10005548074599e-006,-1.25213350502815e-006,-1.24004726271932e-006,-1.08850641767933e-006,-8.46214697861095e-007,-5.73110664448679e-007,-3.25538443615425e-007,-1.43013414267412e-007,-3.98116779370311e-008,-3.37748665821831e-009,-2.22467381475193e-005,-4.84590652057195e-005,-6.74257681555381e-005,-7.67470958397088e-005,-7.60062930474401e-005,-6.67178907235581e-005,-5.18670894572505e-005,-3.51277071610772e-005,-1.99532478216857e-005,-8.76572998571608e-006,-2.44017962134497e-006,-2.07016497214321e-007,-.000191780194444643,-.000417746138158903,-.000581250466549428,-.000661605888719569,-.000655219725381629,-.000575148139518477,-.000447125346440885,-.000302821854829854,-.000172008935497438,-7.5565837560655e-005,-2.10358084478815e-005,-1.78460607688856e-006,-.000782948625132055,-.00170546163785514,-.00237297315794306,-.00270102667510736,-.00267495496410619,-.00234806021751757,-.00182540317195597,-.00123627966686315,-.000702231843846736,-.000308499888616729,-8.58793440610684e-005,-7.28570996785344e-006,-.00211147892329134,-.00459933919956203,-.00639950393640168,-.00728420832819645,-.00721389736971712,-.00633231798455615,-.00492280106302518,-.00333403543487275,-.00189379952905124,-.000831971590143711,-.000231601945658244,-1.96483173282813e-005,-.00431407966834182,-.00939716494906272,-.0130751813408912,-.0148827699400637,-.0147391137220966,-.0129379099972572,-.0100580478180592,-.0068119526671359,-.0038693268277857,-.0016998472881189,-.000473198777355153,-4.01445666201286e-005,-.00713820506661555,-.0155488297871431,-.0216345855593206,-.0246254756422405,-.0243877777734532,-.0214074522943001,-.0166423463205833,-.011271260333675,-.00640230373332618,-.00281261809177179,-.000782968829903678,-6.64243989158946e-005,-.00988521790053478,-.021532523808133,-.0299602757900378,-.0341021573849623,-.0337729856666483,-.029645734277678,-.0230468609713296,-.0156088068880401,-.00886611786558599,-.00389500475941537,-.00108428063087214,-9.19866620625105e-005,-.0115832378335861,-.0252312439580584,-.0351066616363676,-.039960009339504,-.0395742945945498,-.034738090181205,-.0270057043390071,-.0182899885770756,-.0103890832686787,-.00456406393315995,-.00127053146952515,-.000107787546507207,-.0113469563151396,-.0247165625951827,-.0343905358485524,-.0391448821859808,-.0387670354712729,-.0340294827249939,-.0264548265171398,-.0179168989163566,-.0101771607988773,-.00447096354344966,-.00124461444104258,-.000105588834409239,-.00877349869072301,-.0191109160505604,-.0265908595098725,-.0302669335343241,-.0299747813866719,-.0263116922143577,-.0204549466275595,-.0138533968774284,-.00786900949156683,-.00345696164727354,-.000962339403243543,-8.16415851719116e-005,-.00419158042503607,-.0091303303785454,-.0127039086840373,-.0144601704064181,-.014320593338432,-.01257053519047,-.00977244733277217,-.00661852576932447,-.003759456439442,-.0016515797496082,-.000459762193746235,-3.90046528002957e-005},
			{-3.95945152473478e-009,-8.62469447627331e-009,-1.20003687174081e-008,-1.36593690107054e-008,-1.35275216933177e-008,-1.18743814217069e-008,-9.2312511197497e-009,-6.25199311274643e-009,-3.55125848055141e-009,-1.56011558760704e-009,-4.34300653798924e-010,-3.68445827925519e-011,-1.26194613503531e-006,-2.74884028563086e-006,-3.82472643681277e-006,-4.35347871350321e-006,-4.31145667798826e-006,-3.78457209223235e-006,-2.9421604480658e-006,-1.99261905231337e-006,-1.13184790520788e-006,-4.97235999402952e-007,-1.38419179545726e-007,-1.174301000065e-008,-2.61065418768862e-005,-5.68667013887132e-005,-7.91241227483076e-005,-9.0062698548559e-005,-8.91933666496246e-005,-7.82934287517701e-005,-6.08660170299574e-005,-4.12223559228577e-005,-2.34151315299593e-005,-1.02865820344585e-005,-2.86355020000406e-006,-2.42933809796949e-007,-.000191153767800758,-.000416381621285402,-.000579351882704146,-.000659444833676081,-.000653079529962769,-.000573269488181101,-.000445664865962291,-.000301832724128607,-.000171447089262513,-7.53190109571436e-005,-2.09670975420157e-005,-1.77877687852613e-006,-.000789684521168946,-.00172013413604819,-.00239338841889015,-.00272426425965937,-.00269796824743438,-.00236826114642384,-.00184110755612247,-.00124691567929258,-.000708273313928056,-.000311153987634738,-8.66181847905149e-005,-7.34839068958039e-006,-.00223261774167844,-.00486321041790597,-.00676665339575356,-.00770211464969917,-.0076277698425577,-.00669561288171989,-.00520522978980597,-.00352531421515612,-.00200244974321189,-.000879703089733842,-.000244889308237957,-2.0775571746115e-005,-.00475188214064504,-.0103508103065028,-.0144020800439593,-.0163931067848593,-.016234872011969,-.0142508780967576,-.0110787610500387,-.00750324488892071,-.00426199478516835,-.00187235160015967,-.000521220049687516,-4.42185272019898e-005,-.00799931630469631,-.0174245495155288,-.0242444552089724,-.027596148748536,-.0273297764014771,-.0239899219175831,-.0186499814768931,-.0126309591445223,-.00717464014604113,-.00315191585985935,-.000877421602303038,-7.44374492353296e-005,-.0108774707577442,-.0236939034040642,-.032967611546778,-.0375252446090086,-.0371630314766057,-.0326214971629358,-.0253602458535457,-.0171755789499336,-.00975607607116152,-.00428597536221463,-.00119311794379991,-.000101220047626196,-.0118812679549556,-.0258804295145632,-.0360099360729874,-.0409881576521328,-.0405925186861703,-.0356318814839124,-.0277005457518707,-.01876057958054,-.0106563884722811,-.00468149470230067,-.00130322152160786,-.000110560858773534,-.00988326696773915,-.0215282741793143,-.0299543628718611,-.03409542703082,-.0337663202773765,-.0296398834369748,-.0230423124750437,-.0156057263557175,-.00886436806094961,-.0038942360467172,-.00108406663876152,-9.19685077033871e-005,-.00491381941200989,-.0107035509528065,-.0148928821041257,-.0169517601570097,-.0167881329718923,-.0147365274132258,-.0114563092049479,-.00775894462383229,-.00440723740391471,-.00193615863497058,-.00053898247521257,-4.57254306619082e-005},
			{-4.13926516217373e-011,-9.01637440362945e-011,-1.25453507524717e-010,-1.42796925104609e-010,-1.41418575087745e-010,-1.24136418981734e-010,-9.65047707354281e-011,-6.53591971617201e-011,-3.71253453125811e-011,-1.63096632460441e-011,-4.54023885618748e-012,-3.85178343554166e-013,-6.85999042669627e-008,-1.49428073991582e-007,-2.07913681993516e-007,-2.36656870434667e-007,-2.34372535522585e-007,-2.05730875518941e-007,-1.5993703651203e-007,-1.08319580712847e-007,-6.15277116719808e-008,-2.70299508117877e-008,-7.5245228000818e-009,-6.38354791449189e-010,-3.40567466501195e-006,-7.41842734727761e-006,-1.03219729945844e-005,-1.17489421676717e-005,-1.16355352814725e-005,-1.02136065356437e-005,-7.94014975773712e-006,-5.37757677216062e-006,-3.05457232158762e-006,-1.34191468136724e-006,-3.73558490210382e-007,-3.16914252834386e-008,-4.47240038949301e-005,-9.74202782733315e-005,-.00013555022303093,-.000154289468887485,-.000152800186874675,-.000134127132921491,-.000104271641780592,-7.06194185176384e-005,-4.01132573852444e-005,-1.76222931839914e-005,-4.90564514068014e-006,-4.16178163573184e-007,-.000283027553294656,-.00061650613987428,-.000857804414450873,-.000976392251931343,-.000966967607276502,-.000848798652964526,-.000659863721515399,-.000446901875894084,-.000253849300234756,-.00011151940991293,-3.10444642827677e-005,-2.6337062228933e-006,-.00110725792950642,-.00241188995212247,-.00335589496079028,-.00381983326596499,-.00378296225321243,-.00332066269912202,-.00258151310547804,-.00174836562742656,-.000993107021958261,-.000436285264394079,-.000121451882844024,-1.03035625519223e-005,-.00303145075477106,-.00660327230083207,-.00918776920961406,-.0104579394995808,-.010356994041021,-.0090913103237661,-.00706766656937942,-.00478667540745511,-.00271892840056321,-.00119446179503315,-.00033251096434234,-2.82090935115534e-005,-.00620337034448753,-.013512521522274,-.0188012736005075,-.0214004702712195,-.0211939018276931,-.0186038861974662,-.0144628287733888,-.00979515178485467,-.00556384423606199,-.00244427156379506,-.00068042954422791,-5.77253164541879e-005,-.00978890451433557,-.0213227287077879,-.029668367645793,-.033769894172609,-.0334439296312284,-.0293568907657418,-.0228223114223215,-.0154567275852829,-.00877973374713072,-.00385705505497627,-.00107371629731872,-9.10904201184736e-005,-.011922145428059,-.0259694710684948,-.0361338281692569,-.0411291773074706,-.0407321771466212,-.0357544728843004,-.0277958494112289,-.0188251253083289,-.0106930517505446,-.00469760137328079,-.0013077052520396,-.000110941243135535,-.0106693129707941,-.0232404826956437,-.0323367236121244,-.0368071390818563,-.0364518574766186,-.0319972326801651,-.024874937019263,-.0168468967973055,-.00956937796371369,-.00420395637395983,-.00117028572514512,-9.92830402484601e-005,-.00552037774992994,-.0120247895925313,-.0167312487713743,-.01904427325192,-.0188604480446285,-.0165555937697616,-.0128704677825032,-.00871670317383983,-.00495126362270572,-.00217515666585201,-.000605514084724702,-5.13697449717053e-005},
			{-.0029914753970567,-.00626152973771969,-.00809817233744839,-.00826680332565275,-.00705494900380919,-.00509444235123848,-.00307890148348425,-.00150693452348,-.000558780057100243,-.000136870983053836,-1.58772021730906e-005,-2.5903751261877e-007,-.00678121949037336,-.0141939350525009,-.0183573243307094,-.0187395851191542,-.0159924957882962,-.0115483255515964,-.00697940112403575,-.00341599124344908,-.00126666935578621,-.000310265689921581,-3.59912011760796e-005,-5.8719862146837e-007,-.0101508632191237,-.0212470181010361,-.027479229748374,-.0280514390660412,-.0239392984565476,-.0172867834834987,-.0104475229362594,-.00511342538303042,-.00189608777487504,-.000464439262650617,-5.38755250068202e-005,-8.78982445184828e-007,-.0128832216939646,-.0269661839217411,-.0348759510581064,-.0356021848114063,-.0303831588068934,-.0219399532025914,-.0132597347865977,-.0064898315939296,-.00240646717698951,-.000589454695127486,-6.83774687490587e-005,-1.11558253342289e-006,-.0148061931138207,-.0309912020589656,-.0400815943915947,-.0409162270209603,-.0349182003841082,-.0252147476572524,-.0152389051863002,-.00745851481395462,-.00276566053049211,-.000677437698056106,-7.85836051712971e-005,-1.2820962657192e-006,-.0157988744548141,-.033069007459881,-.042768865222384,-.0436594558033572,-.0372592914205353,-.0269052706245537,-.0162605977118075,-.00795857100195779,-.00295108426386074,-.00072285651418097,-8.38522436364219e-005,-1.3680544205638e-006,-.0157988744548141,-.033069007459881,-.042768865222384,-.0436594558033572,-.0372592914205353,-.0269052706245537,-.0162605977118075,-.00795857100195779,-.00295108426386074,-.00072285651418097,-8.38522436364219e-005,-1.3680544205638e-006,-.0148061931138207,-.0309912020589656,-.0400815943915947,-.0409162270209603,-.0349182003841082,-.0252147476572524,-.0152389051863002,-.00745851481395462,-.00276566053049211,-.000677437698056106,-7.85836051712971e-005,-1.2820962657192e-006,-.0128832216939646,-.0269661839217411,-.0348759510581064,-.0356021848114063,-.0303831588068934,-.0219399532025914,-.0132597347865977,-.0064898315939296,-.00240646717698951,-.000589454695127486,-6.83774687490587e-005,-1.11558253342289e-006,-.0101508632191237,-.0212470181010361,-.027479229748374,-.0280514390660412,-.0239392984565476,-.0172867834834987,-.0104475229362594,-.00511342538303042,-.00189608777487504,-.000464439262650617,-5.38755250068202e-005,-8.78982445184828e-007,-.00678121949037336,-.0141939350525009,-.0183573243307094,-.0187395851191542,-.0159924957882962,-.0115483255515964,-.00697940112403575,-.00341599124344908,-.00126666935578621,-.000310265689921581,-3.59912011760796e-005,-5.8719862146837e-007,-.0029914753970567,-.00626152973771969,-.00809817233744839,-.00826680332565275,-.00705494900380919,-.00509444235123848,-.00307890148348425,-.00150693452348,-.000558780057100243,-.000136870983053836,-1.58772021730906e-005,-2.5903751261877e-007},
			{-4.77707484775864e-005,-9.99901127316091e-005,-.000129319383418971,-.000132012244784625,-.000112660192597526,-8.13529419075361e-005,-4.91668520822556e-005,-2.40641758792944e-005,-8.92313591757784e-006,-2.1856871401242e-006,-2.53542393256769e-007,-4.13655946284767e-009,-.000563091388850621,-.00117862024866094,-.00152433515352727,-.00155607690124302,-.0013279671418103,-.0009589370588843,-.000579547775714673,-.000283653295149336,-.0001051802861973,-2.57634985121267e-005,-2.98859747651745e-006,-4.87591483748876e-008,-.00202276382802748,-.004233896048097,-.00547578967002042,-.0055898138948427,-.00477038710314128,-.00344473958308176,-.00208187924827317,-.00101895293817233,-.000377833656408332,-9.25488720048507e-005,-1.07357828440126e-005,-1.75154945664267e-007,-.00460437499342809,-.00963752901773119,-.0124644254937564,-.0127239764517783,-.0108587323850331,-.00784118866248488,-.00473893818806989,-.00231942126065002,-.000860054849279415,-.000210667061584079,-2.44376379372225e-005,-3.98701539258914e-007,-.00810600549215572,-.0169668767769879,-.0219436300590561,-.0224005697075911,-.0191168061846747,-.0138044182877922,-.00834290409323953,-.00408334279989948,-.00151412718159681,-.000370879513648267,-4.30224791892881e-005,-7.01914347023563e-007,-.0119687547517608,-.0250520909767,-.0324004131004311,-.0330750978875267,-.0282265247763196,-.0203826283040634,-.0123185424813832,-.00602917535482301,-.0022356531730757,-.00054761447492283,-6.35239518067959e-005,-1.03639711130188e-006,-.0153956985051803,-.0322251100971785,-.0416774344435578,-.0425452977913825,-.0363084610152352,-.0262186674070122,-.0158456389156716,-.00775547397560826,-.00287577470912415,-.000704409734165886,-8.17123944937675e-005,-1.33314265253001e-006,-.0175390732465345,-.0367114597744218,-.047479727872512,-.0484684143436297,-.0413632909868219,-.0298688057526838,-.0180516539400008,-.00883518380630904,-.00327613737349354,-.000802476998296004,-9.30883176033672e-005,-1.51874152529913e-006,-.0177100195447144,-.037069271618925,-.0479424936985227,-.0489408164993326,-.0417664423606343,-.0301599249984219,-.0182275961561996,-.00892129679210925,-.00330806856783064,-.000810298419092016,-9.39956120239112e-005,-1.53354408858128e-006,-.0155590470074065,-.0325670188106206,-.0421196324049362,-.0429967037908343,-.0366936941193377,-.0264968477084686,-.0160137612897764,-.00783775962556066,-.00290628670514157,-.000711883527900599,-8.25793637481591e-005,-1.34728730829027e-006,-.0111823253052378,-.0234059964205263,-.0302714832768886,-.0309018366364716,-.0263718481021364,-.0190433495379502,-.0115091295769889,-.00563301709645563,-.00208875539431883,-.000511632440253917,-5.93499915831539e-005,-9.68298698098172e-007,-.00513361662858989,-.0107452975256271,-.0138971265528759,-.0141865111307979,-.012106869926272,-.00874248004628156,-.005283646948578,-.0025860229823533,-.000958912313193877,-.000234881809580631,-2.72465784513743e-005,-4.44529573439469e-007},
			{-5.68594172031225e-007,-1.19013825765391e-006,-1.53923164459519e-006,-1.57128358699498e-006,-1.34094463604488e-006,-9.68308224601569e-007,-5.85211378134644e-007,-2.86425283165094e-007,-1.06208155422931e-007,-2.60152710469109e-008,-3.0178033998418e-009,-4.92356447782966e-011,-3.48508775138951e-005,-7.29472173341577e-005,-9.4344219744047e-005,-9.63087814184433e-005,-8.21906019486044e-005,-5.93505754917118e-005,-3.58693969483533e-005,-1.75558824755616e-005,-6.50982299448902e-006,-1.59455560634358e-006,-1.84970409164038e-007,-3.0178033998418e-009,-.000300435417626676,-.000628848662297057,-.000813303626233361,-.000830239323386619,-.00070853216856797,-.000511637473895075,-.000309215664595502,-.000151342211720415,-5.61185694462147e-005,-1.37460234489013e-005,-1.59455560634358e-006,-2.60152710469109e-008,-.00122653696255208,-.00256729427659293,-.00332033741971269,-.00338947793153767,-.00289260467614267,-.00208877594431585,-.00126238259464391,-.000617859299465599,-.000229105810010919,-5.61185694462147e-005,-6.50982299448902e-006,-1.06208155422931e-007,-.0033077610227998,-.00692355485520927,-.00895438379338928,-.00914084396303245,-.00780086152659004,-.00563307251629313,-.00340443057968316,-.00166626116517048,-.000617859299465599,-.000151342211720415,-1.75558824755616e-005,-2.86425283165094e-007,-.0067582699589302,-.0141458988313976,-.018295198043168,-.0186761651548639,-.0159383727377934,-.0115092428083319,-.00695578089086413,-.00340443057968316,-.00126238259464391,-.000309215664595502,-3.58693969483533e-005,-5.85211378134644e-007,-.0111824353213518,-.023406226697924,-.030271781099646,-.0309021406608933,-.0263721075587493,-.0190435368939121,-.0115092428083319,-.00563307251629313,-.00208877594431585,-.000511637473895075,-5.93505754917118e-005,-9.68308224601569e-007,-.0154857991299779,-.0324137019011133,-.0419213443176089,-.0427942866834367,-.036520949914119,-.0263721075587493,-.0159383727377934,-.00780086152659004,-.00289260467614267,-.00070853216856797,-8.21906019486044e-005,-1.34094463604488e-006,-.0181458513277659,-.037981521698904,-.0491223265304294,-.0501452173903112,-.0427942866834367,-.0309021406608933,-.0186761651548639,-.00914084396303246,-.00338947793153767,-.000830239323386619,-9.63087814184433e-005,-1.57128358699498e-006,-.0177757018611982,-.0372067528692506,-.0481203011840638,-.0491223265304294,-.0419213443176089,-.030271781099646,-.018295198043168,-.00895438379338928,-.00332033741971269,-.000813303626233361,-9.4344219744047e-005,-1.53923164459519e-006,-.0137442229153402,-.0287683664692428,-.0372067528692506,-.037981521698904,-.0324137019011133,-.023406226697924,-.0141458988313976,-.00692355485520927,-.00256729427659293,-.000628848662297057,-7.29472173341577e-005,-1.19013825765392e-006,-.00656636739345367,-.0137442229153402,-.0177757018611982,-.0181458513277659,-.0154857991299779,-.0111824353213518,-.0067582699589302,-.0033077610227998,-.00122653696255208,-.000300435417626676,-3.48508775138952e-005,-5.68594172031225e-007},
			{-6.20272325817235e-009,-1.29830705524454e-008,-1.67912869868831e-008,-1.71409376487648e-008,-1.46281986187158e-008,-1.05631542517567e-008,-6.38399829730143e-009,-3.1245778676732e-009,-1.15861158670691e-009,-2.83797363265043e-010,-3.2920842768966e-011,-5.37105538571509e-013,-1.9769158918718e-006,-4.13793062049084e-006,-5.35167549924359e-006,-5.46311525260784e-006,-4.66226157691217e-006,-3.36666116468723e-006,-2.03469140284302e-006,-9.95857365368599e-007,-3.69269684126226e-007,-9.045116058834e-008,-1.04924457427685e-008,-1.71184886157124e-010,-4.08974964028785e-005,-8.56035420437903e-005,-.000110712919239298,-.000113018331892963,-9.64506516716335e-005,-6.96478861031055e-005,-4.20927287149029e-005,-2.06018238739433e-005,-7.6392757225221e-006,-1.87121062155825e-006,-2.1706273078516e-007,-3.54139156583407e-009,-.000299454081965167,-.000626794604813159,-.00081064706909918,-.000827527447862548,-.000706217834626139,-.000509966272467282,-.000308205649327789,-.000150847870838019,-5.59352649812969e-005,-1.37011237392702e-005,-1.58934718486943e-006,-2.59302953359118e-008,-.00123708915628192,-.00258938132932369,-.00334890308448347,-.00341863842883094,-.00291749045281179,-.00210674618825924,-.0012732431770043,-.000623174892248206,-.000231076862629534,-5.6601371053274e-005,-6.56582857400444e-006,-1.07121889836103e-007,-.00349753239972921,-.00732076992880828,-.00946811067096992,-.00966526834956002,-.00824840904375729,-.00595625061785239,-.00359974803892942,-.0017618571509321,-.000653306845161927,-.000160024948988241,-1.85630906649181e-005,-3.02857945621352e-007,-.0074441143400151,-.0155814563465409,-.0201518357411283,-.0205714642490873,-.0175558345219898,-.0126772265613641,-.00766166912395449,-.00374991982442457,-.00139049200941683,-.000340595563207872,-3.95095037359528e-005,-6.44599940278459e-007,-.0125314188045125,-.0262298167578035,-.033923591419649,-.0346299938653073,-.0295534841096122,-.0213408376153209,-.0128976504320703,-.00631261338244207,-.00234075363682485,-.000573358421239403,-6.65102812046103e-005,-1.08511925583571e-006,-.0170402239900308,-.0356673062917253,-.0461293015064619,-.047089867591508,-.040186829342282,-.0290192721808159,-.0175382257776313,-.0085838920298908,-.00318295692604349,-.00077965281321311,-9.04406840954337e-005,-1.47554522466982e-006,-.0186127337638557,-.0389587646541599,-.0503862160588214,-.0514354253188647,-.0438953593449828,-.0316972351677059,-.0191566922641638,-.00937603268029293,-.00347668727130158,-.000851600909064724,-9.87867515986945e-005,-1.61171182018355e-006,-.015482742876017,-.0324073047817918,-.0419130707843193,-.0427858408675583,-.0365137421945241,-.0263669027993754,-.0159352271645878,-.00779932196032425,-.00289203379604635,-.000708392333727775,-8.2174380935362e-005,-1.34067998923389e-006,-.00769779899133229,-.0161124498455309,-.0208385811603797,-.0212725100010394,-.0181541119739297,-.0131092481092602,-.00792276772768902,-.0038777116690518,-.00143787796622194,-.000352202567445854,-4.08559304861565e-005,-6.66566974047622e-007},
			{-.00404945140720489,-.00814474566006138,-.00979128722597762,-.00896406350610278,-.00659219233567233,-.00391599979728114,-.00183988793598998,-.000650774647452403,-.00015752876308832,-2.15142568085075e-005,-1.03996283213943e-006,-3.26296210833654e-009,-.00917949010540939,-.0184628989656689,-.0221953581292932,-.0203201678410382,-.0149434968427065,-.00887697561402263,-.00417074596164616,-.00147520709262294,-.000357093733613841,-4.87695462026171e-005,-2.35743748168761e-006,-7.3966385507205e-009,-.0137408542244633,-.0276372652876165,-.0332244141027671,-.030417426340031,-.0223690433086643,-.0132880177946399,-.00624322392727204,-.00220824962801375,-.000534536546334486,-7.3003534757174e-005,-3.5288675522474e-006,-1.10720890713312e-008,-.017439548481424,-.0350765258115668,-.0421675953362895,-.0386050366790666,-.0283902302500237,-.0168648198114988,-.00792374364660721,-.00280265519288241,-.000678420414234276,-9.26542602749126e-005,-4.47875042967689e-006,-1.40524185029455e-008,-.0200426049296945,-.0403121072713726,-.0484616018046971,-.044367289627886,-.0326278039463224,-.0193820912882233,-.00910645500038529,-.00322098423849287,-.000779682590591189,-.000106483991585058,-5.14725627996212e-006,-1.61499062123808e-008,-.0213863615446234,-.043014832840112,-.0517107103024967,-.0473418949315912,-.0348153353344463,-.0206815637605425,-.00971699735195308,-.00343693515366777,-.000831956416390127,-.000113623211690348,-5.49235412024762e-006,-1.72326768092913e-008,-.0213863615446234,-.043014832840112,-.0517107103024967,-.0473418949315912,-.0348153353344463,-.0206815637605425,-.00971699735195308,-.00343693515366777,-.000831956416390127,-.000113623211690348,-5.49235412024762e-006,-1.72326768092913e-008,-.0200426049296945,-.0403121072713726,-.0484616018046971,-.044367289627886,-.0326278039463224,-.0193820912882233,-.00910645500038529,-.00322098423849287,-.000779682590591189,-.000106483991585058,-5.14725627996212e-006,-1.61499062123808e-008,-.017439548481424,-.0350765258115668,-.0421675953362895,-.0386050366790666,-.0283902302500237,-.0168648198114988,-.00792374364660721,-.00280265519288241,-.000678420414234276,-9.26542602749126e-005,-4.47875042967689e-006,-1.40524185029455e-008,-.0137408542244633,-.0276372652876165,-.0332244141027671,-.030417426340031,-.0223690433086643,-.0132880177946399,-.00624322392727204,-.00220824962801375,-.000534536546334486,-7.3003534757174e-005,-3.5288675522474e-006,-1.10720890713312e-008,-.00917949010540939,-.0184628989656689,-.0221953581292932,-.0203201678410382,-.0149434968427065,-.00887697561402263,-.00417074596164616,-.00147520709262294,-.000357093733613841,-4.87695462026171e-005,-2.35743748168761e-006,-7.3966385507205e-009,-.00404945140720489,-.00814474566006138,-.00979128722597762,-.00896406350610278,-.00659219233567233,-.00391599979728114,-.00183988793598998,-.000650774647452403,-.00015752876308832,-2.15142568085075e-005,-1.03996283213943e-006,-3.26296210833654e-009},
			{-6.46655241827284e-005,-.000130063110895906,-.000156356665933801,-.000143146764126628,-.000105270450249422,-6.25344408726795e-005,-2.93810953272788e-005,-1.03921937740666e-005,-2.51557038586478e-006,-3.43560288548878e-007,-1.66071240047969e-008,-5.21061086814239e-011,-.000762236326271721,-.0015331017429415,-.00184303355050058,-.00168732357751008,-.00124086152980604,-.00073711646319537,-.000346325780965151,-.000122496611669995,-2.96519537053663e-005,-4.04967152908804e-006,-1.95754280991916e-007,-6.14193874705368e-010,-.00273814180028246,-.00550728143194998,-.00662063329969093,-.00606128448480099,-.00445748215614846,-.00264790502628461,-.00124408804027283,-.000440038189267583,-.00010651716679239,-1.45474238225389e-005,-7.03197894004764e-007,-2.20633662270343e-009,-.00623277490876185,-.0125361095327521,-.0150704091023226,-.013797175094379,-.0101464733989415,-.00602737082751107,-.00283189158461527,-.0010016497263513,-.00024246279884684,-3.31139965720151e-005,-1.6006746579724e-006,-5.02223790632381e-009,-.0109728047159291,-.0220698298613373,-.0265314660788539,-.0242899366908432,-.0178628737266676,-.0106111906829482,-.00498554717433864,-.00176340506466325,-.000426855931999716,-5.82972148148455e-005,-2.81798888822008e-006,-8.84165345126754e-009,-.0162016678512013,-.0325867508082808,-.039174487484589,-.0358648036285977,-.0263750567407152,-.0156677341301185,-.00736130656347784,-.00260371927546702,-.000630265297671321,-8.60775467651508e-005,-4.16084320803041e-006,-1.30549605302744e-008,-.0208405969285544,-.0419171251406854,-.0503910900437627,-.0461337637093912,-.0339268729336555,-.0201537850786976,-.00946902654504109,-.00334922703227079,-.00081072548502156,-.00011072362877741,-5.3521931802253e-006,-1.67929112501503e-008,-.0237420053340501,-.0477527880842341,-.0574064904526992,-.0525564631293,-.0386501404407882,-.0229595761810517,-.0107872955612227,-.00381550328609678,-.000923593928514568,-.000126138468780461,-6.09732050715256e-006,-1.91308046425978e-008,-.0239734091183992,-.0482182153183887,-.0579660084441261,-.0530687098535765,-.0390268478266128,-.0231833539428959,-.0108924349957666,-.00385269146321899,-.000932595827353905,-.000127367889742049,-6.15674863969202e-006,-1.93172649069964e-008,-.0210617158529497,-.0423618662225568,-.0509257399709112,-.0466232433651179,-.0342868373663876,-.020367616923072,-.00956949300345818,-.00338476236177552,-.000819327289847131,-.00011189840749341,-5.40897999891501e-006,-1.69710841924047e-008,-.0151371069219119,-.0304455773166633,-.0366004544168802,-.0335082395371539,-.0246420342413141,-.0146382563159452,-.00687761812917763,-.00243263702412373,-.00058885253590198,-8.04216603455029e-005,-3.88744721245458e-006,-1.21971598987849e-008,-.00694919005501226,-.0139770501853164,-.0168026502788677,-.0153830666688022,-.0113127416069928,-.00672017616961103,-.00315739828964731,-.00111678255975218,-.000270332250902931,-3.69202255862743e-005,-1.78466133902173e-006,-5.5995100454324e-009},
			{-7.69685243564879e-007,-1.54808390489254e-006,-1.86104450591298e-006,-1.70381286481175e-006,-1.25298778853868e-006,-7.44319897852733e-007,-3.49710232755209e-007,-1.23693703828388e-007,-2.99417259756176e-008,-4.08924674643852e-009,-1.97667280147994e-010,-6.20196054366658e-013,-4.71763649140235e-005,-9.48868018782288e-005,-.000114069115221109,-.000104431906584671,-7.67994574913963e-005,-4.56216452211473e-005,-2.14348107782544e-005,-7.58156578700641e-006,-1.83522004948295e-006,-2.50642452023579e-007,-1.21156327444106e-008,-3.80137148576797e-011,-.000406688494124778,-.00081798100889079,-.000983344027882673,-.000900265522892682,-.000662057277488362,-.000393285880086281,-.00018478089469472,-6.53576336083714e-005,-1.58206949533316e-005,-2.16068791147813e-006,-1.04444003796667e-007,-3.27700968053241e-010,-.00166031846121593,-.00333943298029159,-.00401453266272909,-.00367536207502424,-.00270286948383404,-.00160560187139806,-.000754373765606237,-.000266824577604282,-6.45884805686753e-005,-8.82107578694501e-006,-4.263959029382e-007,-1.33784942253663e-009,-.00447759575057386,-.00900588126384461,-.0108265099805079,-.00991182474529423,-.00728917806907233,-.00433003443882712,-.00203441740011081,-.000719580383364912,-.000174184117617127,-2.37889371116125e-005,-1.14991703559086e-006,-3.60795174491446e-009,-.00914842415783806,-.0184004153805573,-.0221202428642781,-.0202513987414977,-.0148929238932162,-.00884693345962729,-.00415663099731916,-.0014702145815281,-.000355885229102249,-4.86044964941899e-005,-2.34945925758498e-006,-7.37160626866756e-009,-.0151372558449131,-.0304458768485863,-.0366008145021687,-.0335085692003862,-.0246422766763918,-.0146384003311175,-.00687768579306728,-.00243266095707262,-.000588858329194405,-8.04224515558099e-005,-3.8874854582243e-006,-1.21972798977825e-008,-.0209625628637279,-.0421624377574542,-.05068599505261,-.0464037534631962,-.0341254240019886,-.0202717315680825,-.00952444236070643,-.00336882779555505,-.000815470113608189,-.000111371619378443,-5.3835159512905e-006,-1.68911888154202e-008,-.0245633787434188,-.0494048334697752,-.0593925132894224,-.0543746954436094,-.0399872725481971,-.023753880836457,-.0111604905634277,-.00394750363309629,-.00095554638928031,-.000130502328643898,-6.30826212149847e-006,-1.97926499253432e-008,-.0240623208776429,-.0483970453850838,-.0581809907923459,-.0532655292683946,-.0391715892641307,-.0232693355725446,-.0109328324858832,-.00386698019345867,-.000936054606840123,-.000127840267412333,-6.179582578331e-006,-1.93889081179464e-008,-.0186050545055932,-.0374207322675915,-.0449857065070876,-.0411850577650857,-.030287583522627,-.0179919160266198,-.00845329697557845,-.00298996001413125,-.000723760067411173,-9.88464560551176e-005,-4.77807071380575e-006,-1.49915585521718e-008,-.00888865263692734,-.0178779315290952,-.021492133691307,-.0196763558094162,-.0144700360358322,-.00859572283356178,-.00403860254373974,-.00142846751434807,-.000345779789560036,-4.72243611004261e-005,-2.28274584398164e-006,-7.16228788345407e-009},
			{-.00508628631126404,-.00983033824354329,-.0109846550070609,-.00901916015333691,-.00571557221193217,-.00279307452781809,-.00102018939735755,-.000260771859326372,-4.12071599264361e-005,-3.13787157581542e-006,-6.32055938620332e-008,-3.81377664478188e-011,-.0115298370501388,-.0222838808434348,-.0249005412854803,-.0204450635558571,-.012956332423902,-.00633147491192501,-.00231261411410569,-.000591130121526243,-9.34103607574427e-005,-7.11307734943901e-006,-1.43277462826387e-007,-8.64525128335375e-011,-.0172591078936303,-.0333569244815142,-.0372738250147888,-.0306043837625979,-.019394440549112,-.00947763686126095,-.00346177108471816,-.000884867539951351,-.000139826737158981,-1.06476239773248e-005,-2.14473212317988e-007,-1.29411477385236e-010,-.0219048280361805,-.0423357741945106,-.0473070063778303,-.0388423183634187,-.0246149388313989,-.0120287796399897,-.00439359327137407,-.00112305174849657,-.000177464597313172,-1.35136980227866e-005,-2.72204036451189e-007,-1.64245810125438e-010,-.0251743796492038,-.0486548833232205,-.0543681300147957,-.0446399883770301,-.0282890061570013,-.0138242155963766,-.00504938842044633,-.00129068034844452,-.000203953262708829,-1.55307754038705e-005,-3.12833670474356e-007,-1.88761417029128e-010,-.0268621961430683,-.051916950382044,-.0580132417457791,-.0476328767706495,-.0301856428110978,-.0147510602465105,-.00538792470927311,-.00137721402318714,-.000217627310910781,-1.65720363705528e-005,-3.33807606524435e-007,-2.01416927810585e-010,-.0268621961430683,-.051916950382044,-.0580132417457791,-.0476328767706495,-.0301856428110978,-.0147510602465105,-.00538792470927311,-.00137721402318714,-.000217627310910781,-1.65720363705528e-005,-3.33807606524435e-007,-2.01416927810585e-010,-.0251743796492038,-.0486548833232205,-.0543681300147957,-.0446399883770301,-.0282890061570013,-.0138242155963766,-.00504938842044633,-.00129068034844452,-.000203953262708829,-1.55307754038705e-005,-3.12833670474356e-007,-1.88761417029128e-010,-.0219048280361805,-.0423357741945106,-.0473070063778303,-.0388423183634187,-.0246149388313989,-.0120287796399897,-.00439359327137407,-.00112305174849657,-.000177464597313172,-1.35136980227866e-005,-2.72204036451189e-007,-1.64245810125438e-010,-.0172591078936303,-.0333569244815142,-.0372738250147888,-.0306043837625979,-.019394440549112,-.00947763686126095,-.00346177108471816,-.000884867539951351,-.000139826737158981,-1.06476239773248e-005,-2.14473212317988e-007,-1.29411477385236e-010,-.0115298370501388,-.0222838808434348,-.0249005412854803,-.0204450635558571,-.012956332423902,-.00633147491192501,-.00231261411410569,-.000591130121526243,-9.34103607574427e-005,-7.11307734943901e-006,-1.43277462826387e-007,-8.64525128335375e-011,-.00508628631126404,-.00983033824354329,-.0109846550070609,-.00901916015333691,-.00571557221193217,-.00279307452781809,-.00102018939735755,-.000260771859326372,-4.12071599264361e-005,-3.13787157581542e-006,-6.32055938620331e-008,-3.81377664478188e-011},
			{-8.12226984453021e-005,-.000156980270065097,-.000175413507334088,-.000144026600260793,-9.12717393010351e-005,-4.46024931010705e-005,-1.62913628348369e-005,-4.16425517497517e-006,-6.58035454488694e-007,-5.01085430834115e-008,-1.00932754786986e-009,-6.09020435343665e-013,-.000957401831273778,-.00185038420184216,-.00206766355177068,-.00169769452974626,-.0010758535732373,-.000525745997014592,-.000192032287901912,-4.90856079239649e-005,-7.75650601653937e-006,-5.90647834026672e-007,-1.18973151739896e-008,-7.17874795152048e-012,-.00343922466500658,-.00664703866113062,-.00742755993763223,-.00609853951561999,-.00386473267978856,-.00188860992469107,-.000689827572348814,-.000176327669275857,-2.78632920211386e-005,-2.12175340882166e-006,-4.27381047927422e-008,-2.57878418572552e-011,-.00782863516982087,-.0151305150742709,-.016907199330518,-.0138819779418027,-.00879721016390503,-.00429900326923858,-.00157024007444182,-.000401370985487183,-6.34246288365862e-005,-4.82970290571489e-006,-9.72838540256606e-008,-5.87003250388414e-011,-.0137823178549157,-.0266372826857591,-.0297651365985099,-.0244392321648105,-.0154874948296011,-.0075683983517811,-.00276440878199448,-.000706613909031054,-.000111659104747359,-8.50269544402536e-006,-1.71268295079276e-007,-1.03341964508827e-010,-.0203499963670061,-.0393307288068947,-.0439490968079129,-.0360852427727842,-.0228677401605563,-.0111749620480468,-.00408173061038807,-.00104333615238232,-.00016486794165333,-1.25544790953984e-005,-2.52882658732434e-007,-1.52587440258738e-010,-.0261766921577074,-.0505920670327779,-.0565327854119549,-.0464173199082629,-.029415326849682,-.0143746237655383,-.0052504287338376,-.00134206850877865,-.000212073617975246,-1.6149129884536e-005,-3.25289075746179e-007,-1.96276912228913e-010,-.0298209867484436,-.0576354472700868,-.0644032269037527,-.0528794957568774,-.0335105011321007,-.0163758454369784,-.00598138850975171,-.0015289100309035,-.000241598308649254,-1.83973966375979e-005,-3.7057551652431e-007,-2.23602400308645e-010,-.0301116399215653,-.0581971968115976,-.0650309392734592,-.0533948909503518,-.033837114519146,-.0165354542211866,-.00603968669970425,-.00154381170252316,-.000243953070268046,-1.85767086689201e-005,-3.74187367153787e-007,-2.25781763041192e-010,-.026454427101384,-.0511288493276215,-.0571325987832424,-.046909808097858,-.0297274237371193,-.0145271386554198,-.00530613583081857,-.00135630786795552,-.000214323721005407,-1.63204723006016e-005,-3.28740395820796e-007,-1.98359412066345e-010,-.0190128617434316,-.0367464296104888,-.0410613390925373,-.0337141943145245,-.0213651724656962,-.0104406902377677,-.00381353285620554,-.000974781796488713,-.000154034984775778,-1.17295635339088e-005,-2.36266530031746e-007,-1.42561396724728e-010,-.0087284836148921,-.0168696650240853,-.0188505670693449,-.0154776170276039,-.00980838972652806,-.00479314449861828,-.00175072850680893,-.000447505854382295,-7.0714859176842e-005,-5.38484445409324e-006,-1.08465972348532e-007,-6.54475287423684e-011},
			{-.00610300723148638,-.011334382523687,-.0117725765312941,-.00866893492130421,-.00473399758463878,-.00190309670559272,-.000540391748990977,-9.98226951437858e-005,-1.02973144060446e-005,-4.37202313877829e-007,-3.66970948242474e-009,-4.25830680908177e-013,-.0138345886544025,-.0256933203450701,-.026686639477119,-.0196511562511815,-.0107312521172389,-.00431403062339345,-.00122498585958623,-.000226282859139225,-2.33424447406373e-005,-9.91071113277978e-007,-8.31867295004696e-009,-9.65293351840578e-013,-.0207091095227104,-.0384605425083467,-.0399474500854692,-.0294159773896932,-.0160636995405713,-.00645770791571394,-.00183369140663751,-.000338724672636725,-3.49414974841223e-005,-1.48354249933819e-006,-1.24522899458459e-008,-1.44495555626435e-012,-.0262834838088483,-.0488131585372653,-.0507003044420999,-.0373340227205607,-.0203876456552513,-.00819596135984531,-.00232727526714406,-.000429900881983438,-4.43468746144641e-005,-1.88287503228147e-006,-1.58041349298872e-008,-1.83390144930814e-012,-.030206600974669,-.0560990929882762,-.0582679174767988,-.0429065619801723,-.0234307400647473,-.00941930438906687,-.00267464830248913,-.000494068613398971,-5.09661639870523e-005,-2.16391613832206e-006,-1.81630867905044e-008,-2.10763267567556e-012,-.0322318027893339,-.0598602571462754,-.0621744904907474,-.0457832327865349,-.0250016542281072,-.0100508217305121,-.00285397012027107,-.0005271934477113,-5.43831908772887e-005,-2.30899591389123e-006,-1.93808310960915e-008,-2.24893892602145e-012,-.0322318027893339,-.0598602571462754,-.0621744904907474,-.0457832327865349,-.0250016542281072,-.0100508217305121,-.00285397012027107,-.0005271934477113,-5.43831908772887e-005,-2.30899591389123e-006,-1.93808310960915e-008,-2.24893892602145e-012,-.030206600974669,-.0560990929882762,-.0582679174767988,-.0429065619801723,-.0234307400647473,-.00941930438906688,-.00267464830248913,-.000494068613398971,-5.09661639870523e-005,-2.16391613832206e-006,-1.81630867905044e-008,-2.10763267567556e-012,-.0262834838088483,-.0488131585372653,-.0507003044420999,-.0373340227205607,-.0203876456552513,-.00819596135984532,-.00232727526714406,-.000429900881983438,-4.43468746144642e-005,-1.88287503228147e-006,-1.58041349298872e-008,-1.83390144930814e-012,-.0207091095227104,-.0384605425083467,-.0399474500854692,-.0294159773896932,-.0160636995405713,-.00645770791571394,-.00183369140663751,-.000338724672636725,-3.49414974841223e-005,-1.48354249933819e-006,-1.24522899458459e-008,-1.44495555626435e-012,-.0138345886544025,-.0256933203450701,-.026686639477119,-.0196511562511815,-.0107312521172389,-.00431403062339345,-.00122498585958623,-.000226282859139225,-2.33424447406373e-005,-9.91071113277978e-007,-8.31867295004696e-009,-9.65293351840578e-013,-.00610300723148638,-.011334382523687,-.0117725765312941,-.0086689349213042,-.00473399758463878,-.00190309670559272,-.000540391748990977,-9.98226951437858e-005,-1.02973144060446e-005,-4.37202313877829e-007,-3.66970948242474e-009,-4.25830680908177e-013},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 144; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,3.354101966,0.,0.,4.062019202},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,3.354101966,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,-7.245688375,0.,0.,-5.477225576,0.,-11.12429773,-8.291561975},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370349,0.,0.,-7.245688375,0.,0.,-5.477225575,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,7.245688375,0.,0.,2.049390154,0.,0.,13.74772709,0.,15.19868416,5.196152424,0.,28.14249455,10.48808848},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,7.245688375,0.,0.,2.049390152,0.,0.,13.74772708,0.,0.,5.196152424,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,5.291502622,0.,0.,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-11.12429773,-.5000000000,0.,0.,-10.06230590,0.,-33.37289319,-2.535462764,0.,-30.89729531,-7.676494735},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,-10.06230590,0.,0.,-2.535462764,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,17.54992877,0.,0.,0.,2.645751311,0.,26.32489316,.5000000000,0.,16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,-3.354101966,-.5000000000},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,7.245688375,0.,0.,-5.477225576,0.,11.12429773,-8.291561975},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370349,0.,0.,7.245688375,0.,0.,-5.477225575,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,-5.916079782,0.,0.,1.673320053,0.,0.,0.,0.,-12.40967365,4.242640686,0.,0.,8.563488387},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246900,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,-5.916079782,0.,0.,1.673320053,0.,0.,0.,0.,0.,4.242640686,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,2.738612788,0.,8.616843970,-.3872983346,0.,0.,-2.598076212,0.,8.616843970,-1.963961012,0.,-7.977647347,-5.946187254},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,-2.598076212,0.,0.,-1.963961013,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-3.070597895,0.,0.,0.,0.,0.,0.,0.,0.,0.,-6.633249580,0.,0.,0.,1.,0.,0.,.3779644731,0.,6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,-1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,7.245688375,0.,0.,2.049390154,0.,0.,-13.74772709,0.,15.19868416,5.196152424,0.,-28.14249455,10.48808848},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,7.245688375,0.,0.,2.049390152,0.,0.,-13.74772708,0.,0.,5.196152424,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,-5.291502622,0.,0.,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,-8.616843970,-.3872983346,0.,0.,2.598076212,0.,8.616843970,-1.963961012,0.,7.977647347,-5.946187254},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,2.598076212,0.,0.,-1.963961013,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.913024938,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,-6.292853088,.3585685828,0.,0.,2.171240594},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,-10.60660172,0.,11.12429773,-.5000000000,0.,0.,10.06230590,0.,-33.37289319,-2.535462764,0.,30.89729531,-7.676494735},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,10.06230590,0.,0.,-2.535462764,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,-3.070597895,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.633249580,0.,0.,0.,-1.,0.,0.,.3779644731,0.,-6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,1.,0.,0.,.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,-17.54992877,0.,0.,0.,-2.645751311,0.,26.32489316,.5000000000,0.,-16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,3.354101966,-.5000000000}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,3.354101966,0.,4.062019202,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,-5.477225576,0.,-8.291561975,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,7.245688375,0.,11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370349,0.,0.,-5.477225575,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,5.196152424,0.,10.48808848,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,-13.74772709,0.,-28.14249455,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,15.19868416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390152,0.,0.,5.196152424,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,-13.74772708,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,-5.291502622,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,-7.676494735,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,10.06230590,0.,30.89729531,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-33.37289319,.5000000000,0.,0.,3.354101966,0.,11.12429773,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,.5000000000,0.,0.,3.354101966,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,3.027650356,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,-16.24807681,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,26.32489316,0.,0.,0.,-2.645751311,0.,-17.54992877,.5000000000,0.,4.062019202,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,-2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,7.245688375,0.,0.,-3.354101966,.5000000000},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,-5.477225576,0.,-8.291561975,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,-7.245688375,0.,-11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370349,0.,0.,-5.477225575,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,-7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,-3.240370350,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,4.242640686,0.,8.563488387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246899,0.,0.,-5.916079782,0.,-12.40967365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246900,0.,0.,-5.916079782,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-2.160246899,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961012,0.,-5.946187254,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,7.977647347,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,8.616843970,-.3872983346,0.,0.,-2.598076212,0.,-8.616843970,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961013,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,-.3872983346,0.,0.,-2.598076212,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,2.288688541,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.141195788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,6.633249580,-.3779644731,0.,-3.070597895,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,-.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,-1.080123450,0.,0.,1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,5.196152424,0.,10.48808848,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,13.74772709,0.,28.14249455,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,15.19868416,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390152,0.,0.,5.196152424,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,13.74772708,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.049390154,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,5.291502622,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961012,0.,-5.946187254,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.598076212,0.,-7.977647347,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,8.616843970,.3872983346,0.,0.,2.598076212,0.,8.616843970,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,-1.963961013,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.598076212,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.738612788,0.,0.,.3872983346,0.,0.,2.598076212,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,2.171240594,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.292853088,0.,0.,0.,0.,0.,0.,.3585685828,0.,2.913024938,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,-1.,0.,0.,-.4629100500,.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,-7.676494735,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.06230590,0.,-30.89729531,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,-33.37289319,-.5000000000,0.,0.,-3.354101966,0.,-11.12429773,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.06230590,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-10.60660172,0.,0.,-.5000000000,0.,0.,-3.354101966,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,2.288688541,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,6.141195788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,-6.633249580,-.3779644731,0.,-3.070597895,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,-.3779644731,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4629100500,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,3.027650356,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,16.24807681,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,26.32489316,0.,0.,0.,2.645751311,0.,17.54992877,.5000000000,0.,4.062019202,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.183300132,0.,0.,0.,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.080123450,0.,0.,0.,0.,0.,1.080123450,0.,0.,1.500000000,.3726779962},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,-7.245688375,0.,0.,0.,0.,0.,-7.245688375,0.,0.,-3.354101966,-.5000000000}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,.2538762001,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,1.703053324,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,3.679015663,.6614378278,0.,0.,2.218529919,0.,3.679015663,.4192627457,0.,1.703053324,.2538762001},
			{0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,0.,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,0.,.6614378278,0.,0.,2.218529919,0.,0.,.4192627457,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708245,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,-1.811422094,0.,-2.085805824,.7905694150,0.,0.,0.,1.045825033,0.,0.,0.,0.,-1.501951855,.8100925875,0.,0.,1.811422094,0.,1.501951855,.6846531970,0.,2.085805824,.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,-1.811422094,0.,0.,.7905694150,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,.8100925872,0.,0.,1.811422094,0.,0.,.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,.7905694148,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,.8794529546,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,-1.899835520,.5123475385,0.,0.,0.,0.,-1.899835520,.6495190530,0.,.8794529546,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,0.,.5123475380,0.,0.,0.,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,.8385254914,0.,.6436936521,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,1.390537216,.1250000000,0.,0.,-.8385254914,0.,-1.390537216,.3169328455,0.,-.6436936521,.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,.8385254914,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.8385254914,0.,0.,.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,-.7616286003,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,.5484352742,0.,0.,0.,-.3307189139,0.,.5484352742,.6250000000e-1,0.,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,-.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,.4528555234,0.,0.,-.2096313728,.3125000000e-1},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,-3.476343041,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219620,0.,-7.509759277,-.8100925875,0.,0.,-3.622844188,0.,-7.509759277,-.6846531970,0.,-3.476343041,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,0.,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219619,0.,0.,-.8100925872,0.,0.,-3.622844188,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,.5303300858,0.,.5352180242,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,0.,1.403121520,0.,2.154210993,-.2041241452,0.,0.,0.,-.5400617248,0.,0.,0.,0.,1.551209206,-.4183300132,0.,0.,-1.403121520,0.,-1.551209206,-.5303300858,0.,-2.154210993,-.5352180242},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,.5303300858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617250,0.,0.,1.403121520,0.,0.,-.2041241452,0.,0.,0.,-.5400617250,0.,0.,0.,0.,0.,-.4183300132,0.,0.,-1.403121520,0.,0.,-.5303300858,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.5400617248,0.,0.,0.,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.4986029592,0.,0.,0.,0.,.1250000000,0.,0.,.6846531969,0.,1.077105496,-.9682458365e-1,0.,0.,0.,0.,1.077105496,-.2454951265,0.,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6846531970,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.1919123684,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4145780988,0.,0.,0.,.1250000000,0.,.4145780988,-.4724555913e-1,0.,.1919123684,-.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931772,0.,4.397264773,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,9.499177600,.5123475385,0.,0.,3.436931772,0.,9.499177600,.6495190530,0.,4.397264773,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931771,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,0.,.5123475380,0.,0.,3.436931771,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.6495190530,0.,-1.495808878,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-1.077105496,.9682458365e-1,0.,0.,.6495190530,0.,1.077105496,.2454951265,0.,1.495808878,.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.6495190530,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.6495190530,0.,0.,.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,.1357025371,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1820640586,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3933033180,0.,0.,0.,0.,0.,-.3933033180,.4482107285e-1,0.,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,-.6250000000e-1,0.,0.,-.2893187812e-1,.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,-3.218468261,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,-6.952686082,-.1250000000,0.,0.,-1.677050983,0.,-6.952686082,-.3169328455,0.,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,0.,-.1250000000,0.,0.,-1.677050983,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.5757371053,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4145780988,0.,0.,0.,-.1250000000,0.,-.4145780988,-.4724555913e-1,0.,-.5757371053,-.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,1.269381000,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,2.742176371,0.,0.,0.,.3307189139,0.,2.742176371,.6250000000e-1,0.,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,.6750771565e-1,0.,0.,.9375000000e-1,.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,-.2096313728,-.3125000000e-1}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,.2538762001,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,1.703053324,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,3.679015663,.6614378278,0.,0.,2.218529919,0.,3.679015663,.4192627457,0.,1.703053324,.2538762001},
			{0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,.4192627457,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,2.561737692,0.,0.,2.218529919,0.,0.,.9682458365,0.,0.,0.,2.561737692,0.,0.,3.507803799,0.,0.,.6614378278,0.,0.,2.218529919,0.,0.,.4192627457,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708245,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,2.561737690,0.,0.,0.,0.,0.,.6614378278,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,2.371708244,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925875,0.,0.,-.6846531970,0.,-.5182226234,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,-3.476343041,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219620,0.,-7.509759277,-.8100925875,0.,0.,-3.622844188,0.,-7.509759277,-.6846531970,0.,-3.476343041,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,-.8100925872,0.,0.,-.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,-3.622844188,0.,0.,-.7905694150,0.,0.,0.,-3.137475099,0.,0.,-5.728219619,0.,0.,-.8100925872,0.,0.,-3.622844188,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-3.137475099,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931772,0.,4.397264773,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,9.499177600,.5123475385,0.,0.,3.436931772,0.,9.499177600,.6495190530,0.,4.397264773,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,3.436931771,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,5.434266281,0.,0.,.5123475380,0.,0.,3.436931771,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,1.984313483,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,-.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,-3.218468261,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,-6.952686082,-.1250000000,0.,0.,-1.677050983,0.,-6.952686082,-.3169328455,0.,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-1.677050983,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-2.651650429,0.,0.,-.1250000000,0.,0.,-1.677050983,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,1.269381000,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,2.742176371,0.,0.,0.,.3307189139,0.,2.742176371,.6250000000e-1,0.,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,-.2096313728,-.3125000000e-1},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,.8100925875,0.,0.,.6846531970,0.,.5182226234,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,1.811422094,0.,2.085805824,-.7905694150,0.,0.,0.,-1.045825033,0.,0.,0.,0.,1.501951855,-.8100925875,0.,0.,-1.811422094,0.,-1.501951855,-.6846531970,0.,-2.085805824,-.5182226234},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,.8100925872,0.,0.,.6846531969,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,1.811422094,0.,0.,-.7905694150,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,-.8100925872,0.,0.,-1.811422094,0.,0.,-.6846531969,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694148,0.,0.,0.,.8100925875,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,-1.045825033,0.,0.,0.,0.,0.,-.8100925875,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694150,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,-.5303300858,0.,-.5352180242,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,0.,-1.403121520,0.,-2.154210993,.2041241452,0.,0.,0.,.5400617248,0.,0.,0.,0.,-1.551209206,.4183300132,0.,0.,1.403121520,0.,1.551209206,.5303300858,0.,2.154210993,.5352180242},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,-.5303300858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617250,0.,0.,-1.403121520,0.,0.,.2041241452,0.,0.,0.,.5400617250,0.,0.,0.,0.,0.,.4183300132,0.,0.,1.403121520,0.,0.,.5303300858,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,.5400617248,0.,0.,0.,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2454951265,0.,.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6495190530,0.,1.495808878,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,1.077105496,-.9682458365e-1,0.,0.,-.6495190530,0.,-1.077105496,-.2454951265,0.,-1.495808878,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.6495190530,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,-.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.5757371053,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4145780988,0.,0.,0.,.1250000000,0.,.4145780988,.4724555913e-1,0.,.5757371053,.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,-.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,.6495190530,0.,.6555055300,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,.8794529546,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,-1.899835520,.5123475385,0.,0.,0.,0.,-1.899835520,.6495190530,0.,.8794529546,.6555055300},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475380,0.,0.,.6495190530,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,-1.811422094,0.,0.,.5123475380,0.,0.,0.,0.,0.,.6495190530,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,-.6614378278,0.,0.,0.,0.,0.,.5123475385,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951265,0.,-.3716367033,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.4986029592,0.,0.,0.,0.,.1250000000,0.,0.,.6846531969,0.,1.077105496,-.9682458365e-1,0.,0.,0.,0.,1.077105496,-.2454951265,0.,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,-.2454951266,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.6846531970,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.2454951266,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,.1357025371,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1820640586,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3933033180,0.,0.,0.,0.,0.,-.3933033180,.4482107285e-1,0.,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.3169328455,0.,.4797809210,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.8385254914,0.,-.6436936521,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,-1.390537216,-.1250000000,0.,0.,.8385254914,0.,1.390537216,-.3169328455,0.,.6436936521,-.4797809210},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,.3169328455,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.8385254914,0.,0.,-.3169328455,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,-.1430430338,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.1919123684,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4145780988,0.,0.,0.,-.1250000000,0.,-.4145780988,.4724555913e-1,0.,-.1919123684,.1430430338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4724555913e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2156454873e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6250000000e-1,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,.1892281472,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,-.7616286003,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,.5484352742,0.,0.,0.,-.3307189139,0.,.5484352742,.6250000000e-1,0.,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5229125166,0.,0.,0.,0.,0.,-.3307189139,0.,0.,.6250000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,0.,0.,0.,-.6750771565e-1,0.,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4528555234,0.,0.,0.,0.,0.,-.4528555234,0.,0.,.2096313728,-.3125000000e-1}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[56][56] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2357022604,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2132007163},
			{0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9622504490e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6900655595e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5976143047e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6085806194e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8703882798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9622504490e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2817180849e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1543033500e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1150109266e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6900655595e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1543033500e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7273929673e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5976143047e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1150109266e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6085806194e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.8703882798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[56][56] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,4.690415760,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,14.49137675,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.73320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.43167672,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,11.48912529,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,30.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,35.49647870,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,64.80740698,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,86.94826050,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,14.49137675,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,64.80740698,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,137.4772708,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.73320053,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,86.94826050,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,16.43167672,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,11.48912529}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][56] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_w:" << std::endl;
		static const T timestep_stiffness_w[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,.9574271080,1.414213562,1.224744872,1.054092553,.9354143465,.8485281372,1.118033988,.9682458365,.8333333333,.7395099729,.9165151392,.7937253933,.6831300514,.7745966692,.6708203931,.6700593942,1.414213562,1.224744872,1.054092553,.9354143465,.8485281372,1.369306394,1.185854122,1.020620726,.9057110469,1.183215957,1.024695077,.8819171036,1.024695077,.8874119674,.8989790125,1.118033988,.9682458365,.8333333333,.7395099729,1.183215957,1.024695077,.8819171036,1.080123450,.9354143465,.9710083125,.9165151392,.7937253933,.6831300514,1.024695077,.8874119674,.9710083125,.7745966692,.6708203931,.8989790125,.6700593942},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.118913853,1.825741858,2.108185107,2.041241452,1.932183566,1.825741858,1.443375673,1.666666667,1.613743061,1.527525232,1.183215956,1.366260101,1.322875656,1.,1.154700539,.8650429586,1.825741858,2.108185107,2.041241452,1.932183566,1.825741858,1.767766952,2.041241452,1.976423538,1.870828693,1.527525232,1.763834207,1.707825127,1.322875656,1.527525232,1.160576915,1.443375673,1.666666667,1.613743061,1.527525232,1.527525232,1.763834207,1.707825127,1.394433378,1.610152972,1.253566341,1.183215956,1.366260101,1.322875656,1.322875656,1.527525232,1.253566341,1.,1.154700539,1.160576915,.8650429586},
			{2.645751311,3.436931772,3.549647869,3.500000000,3.401680258,3.290611645,2.160246900,2.806243040,2.898275350,2.857738034,2.777460300,1.707825129,2.218529918,2.291287848,2.259240285,1.400000000,1.818653348,1.878297101,1.183215956,1.537042616,1.023532631,2.160246900,2.806243040,2.898275350,2.857738034,2.777460300,2.091650066,2.717133141,2.806243040,2.766992952,1.807392228,2.347871376,2.424871131,1.565247584,2.033316257,1.373213125,1.707825129,2.218529918,2.291287848,2.259240285,1.807392228,2.347871376,2.424871131,1.649915823,2.143303525,1.483239697,1.400000000,1.818653348,1.878297101,1.565247584,2.033316257,1.483239697,1.183215956,1.537042616,1.373213125,1.023532631},
			{3.,4.156921939,4.472135954,4.535573675,4.500000000,4.422166386,2.449489743,3.394112549,3.651483716,3.703280398,3.674234614,1.936491673,2.683281572,2.886751347,2.927700220,1.587450787,2.199636334,2.366431913,1.341640786,1.859032006,1.160576915,2.449489743,3.394112549,3.651483716,3.703280398,3.674234614,2.371708244,3.286335344,3.535533905,3.585685827,2.049390154,2.839718296,3.055050463,1.774823935,2.459268185,1.557077324,1.936491673,2.683281572,2.886751347,2.927700220,2.049390154,2.839718296,3.055050463,1.870828693,2.592296279,1.681835732,1.587450787,2.199636334,2.366431913,1.774823935,2.459268185,1.681835732,1.341640786,1.859032006,1.557077324,1.160576915},
			{3.316624790,4.787135539,5.297284632,5.484352741,5.527707984,5.500000000,2.708012802,3.908679800,4.325214791,4.477955262,4.513354671,2.140872097,3.090082702,3.419382528,3.540134475,1.754992877,2.533114026,2.803059554,1.483239697,2.140872097,1.283066056,2.708012802,3.908679800,4.325214791,4.477955262,4.513354671,2.622022120,3.784562941,4.187871214,4.335761539,2.265686063,3.270236146,3.618734322,1.962141687,2.832107579,1.721413752,2.140872097,3.090082702,3.419382528,3.540134475,2.265686063,3.270236146,3.618734322,2.068278942,2.985303509,1.859339360,1.754992877,2.533114026,2.803059554,1.962141687,2.832107579,1.859339360,1.483239697,2.140872097,1.721413752,1.283066056},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.7817359602,1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.643167672,1.423024947,1.224744872,1.086853256,1.496662955,1.296148140,1.115546702,1.355261854,1.173691195,1.230977710,.8660254040,.7500000000,.6454972245,.5728219619,.5196152424,1.341640786,1.161895004,1.,.8874119674,1.449137675,1.254990040,1.080123450,1.434274331,1.242118007,1.376274919,.5477225575,.4743416490,.4082482906,.3622844188,.9660917834,.8366600265,.7200822997,1.133893419,.9819805061,1.189237451,.3741657387,.3240370349,.2788866755,.7171371655,.6210590036,.8919280880,.2710523708,.2347382389,.5505099678,.2051629517},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.730085916,2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.121320343,2.449489743,2.371708244,2.244994432,1.932183566,2.231093404,2.160246900,1.749635531,2.020305089,1.589185390,1.118033988,1.290994449,1.250000000,1.183215956,1.118033988,1.732050808,2.,1.936491673,1.833030278,1.870828694,2.160246900,2.091650066,1.851640200,2.138089935,1.776763280,.7071067810,.8164965809,.7905694148,.7483314774,1.247219129,1.440164600,1.394433378,1.463850110,1.690308509,1.535298948,.4830458916,.5577733510,.5400617250,.9258201003,1.069044968,1.151474211,.3499271062,.4040610177,.7107053122,.2648642316},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.686773158,2.645751311,3.436931771,3.549647869,3.500000000,3.401680258,2.509980080,3.260559769,3.367491647,3.320391543,2.286190427,2.969848480,3.067246323,2.070196679,2.689264371,1.880349511,1.322875656,1.718465886,1.774823935,1.750000000,1.700840129,2.049390152,2.662235902,2.749545417,2.711088342,2.213594362,2.875543427,2.969848480,2.190890230,2.846049894,2.102294664,.8366600265,1.086853256,1.122497216,1.106797181,1.475729575,1.917028952,1.979898987,1.732050808,2.250000000,1.816590213,.5715476066,.7424621200,.7668115806,1.095445115,1.423024947,1.362442660,.4140393357,.5378528742,.8409178659,.3133915852},
			{2.449489743,3.394112549,3.651483718,3.703280400,3.674234614,3.610683736,3.,4.156921939,4.472135954,4.535573675,4.500000000,2.846049894,3.943602415,4.242640686,4.302822994,2.592296280,3.591991092,3.864367134,2.347382390,3.252628452,2.132115937,1.500000000,2.078460970,2.236067977,2.267786838,2.250000000,2.323790008,3.219937887,3.464101616,3.513240264,2.509980080,3.477930420,3.741657387,2.484236014,3.442258395,2.383778085,.9486832980,1.314534138,1.414213562,1.434274331,1.673320053,2.318620280,2.494438258,1.963961013,2.721344207,2.059819687,.6480740700,.8979977729,.9660917834,1.242118007,1.721129197,1.544864765,.4694764779,.6505256900,.9535112343,.3553526559},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.6180165404,1.643167673,1.423024947,1.224744872,1.086853256,.9859006037,1.732050808,1.500000000,1.290994449,1.145643924,1.690308509,1.463850109,1.259881577,1.607142857,1.391826542,1.513825177,.5477225576,.4743416490,.4082482906,.3622844188,.3286335346,1.060660172,.9185586536,.7905694150,.7015607599,1.309307341,1.133893419,.9759000724,1.417366774,1.227475633,1.450721144,.2886751347,.2500000000,.2151657415,.1909406540,.6546536709,.5669467095,.4879500362,.8964214566,.7763237544,1.044638617,.1690308509,.1463850109,.1259881577,.4252100321,.3682426898,.6267831705,.1071428571,.9278843614e-1,.2901442286,.7208691321e-1},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,1.367753011,2.121320343,2.449489743,2.371708244,2.244994432,2.121320343,2.236067977,2.581988897,2.500000000,2.366431913,2.182178903,2.519763153,2.439750183,2.074812507,2.395787117,1.954339899,.7071067810,.8164965809,.7905694148,.7483314772,.7071067810,1.369306394,1.581138830,1.530931089,1.449137675,1.690308509,1.951800147,1.889822365,1.829812638,2.112885637,1.872872943,.3726779962,.4303314828,.4166666667,.3944053189,.8451542550,.9759000734,.9449111824,1.157275125,1.336306209,1.348622656,.2182178903,.2519763153,.2439750183,.5489437912,.6338656908,.8091735938,.1383208338,.1597191412,.3745745885,.9306380472e-1},
			{1.707825129,2.218529919,2.291287848,2.259240285,2.195775165,2.124080684,2.509980080,3.260559769,3.367491648,3.320391543,3.227117245,2.645751311,3.436931771,3.549647870,3.500000000,2.581988897,3.354101966,3.464101616,2.454951265,3.189075241,2.312406154,.8366600265,1.086853256,1.122497216,1.106797181,1.075705749,1.620185174,2.104682280,2.173706512,2.143303525,2.,2.598076212,2.683281572,2.165063510,2.812500000,2.216013152,.4409585519,.5728219619,.5916079783,.5833333333,1.,1.299038106,1.341640786,1.369306394,1.778781184,1.595711847,.2581988897,.3354101966,.3464101616,.6495190530,.8437500000,.9574271080,.1636634177,.2126050161,.4432026303,.1101145788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.5066228049,1.496662955,1.296148140,1.115546702,.9899494934,.8979977729,1.690308509,1.463850109,1.259881577,1.118033988,1.732050808,1.500000000,1.290994449,1.707825127,1.479019946,1.654623153,.3741657387,.3240370349,.2788866754,.2474873734,.2244994432,.8280786708,.7171371654,.6172133998,.5477225574,1.118033988,.9682458365,.8333333333,1.290994449,1.118033988,1.387443692,.1690308509,.1463850109,.1259881577,.1118033988,.4472135954,.3872983346,.3333333333,.6804138176,.5892556509,.8563488387,.8660254040e-1,.7500000000e-1,.6454972245e-1,.2581988897,.2236067977,.4281744193,.4879500362e-1,.4225771273e-1,.1585649934,.2954684202e-1},
			{1.183215957,1.366260102,1.322875656,1.252198067,1.183215957,1.121223821,1.932183566,2.231093404,2.160246900,2.044830882,1.932183566,2.182178903,2.519763153,2.439750183,2.309401077,2.236067977,2.581988897,2.500000000,2.204792759,2.545875387,2.136109305,.4830458916,.5577733510,.5400617250,.5112077204,.4830458916,1.069044968,1.234426800,1.195228610,1.131370850,1.443375673,1.666666667,1.613743061,1.666666667,1.924500898,1.791182105,.2182178903,.2519763153,.2439750183,.2309401077,.5773502693,.6666666667,.6454972245,.8784104612,1.014301033,1.105541597,.1118033988,.1290994449,.1250000000,.3333333333,.3849001795,.5527707984,.6299407883e-1,.7273929675e-1,.2047065263,.3814480901e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.7745966692,.6708203931,.5773502693,.5123475385,.4647580015,.4281744193,1.355261854,1.173691195,1.010152544,.8964214566,.8131571124,1.607142857,1.391826542,1.197893559,1.063025080,1.707825129,1.479019946,1.272937693,1.732050808,1.500000000,1.716232660,.2710523708,.2347382389,.2020305089,.1792842914,.1626314225,.6561133241,.5682108062,.4890379978,.4339781718,.9449111824,.8183170887,.7042952119,1.145643924,.9921567416,1.279204298,.1071428571,.9278843614e-1,.7985957060e-1,.7086833870e-1,.3149703940,.2727723628,.2347650707,.5175491696,.4482107285,.6908492799,.4879500367e-1,.4225771273e-1,.3636964838e-1,.1636634177,.1417366774,.2960782628,.2474358298e-1,.2142857143e-1,.9137173554e-1,.1362089413e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.7817359602,.8660254040,.7500000000,.6454972245,.5728219619,.5196152424,.5477225575,.4743416490,.4082482906,.3622844188,.3741657387,.3240370349,.2788866755,.2710523708,.2347382389,.2051629517,1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.341640786,1.161895004,1.,.8874119674,.9660917834,.8366600265,.7200822997,.7171371655,.6210590036,.5505099678,1.643167672,1.423024947,1.224744872,1.086853256,1.449137675,1.254990040,1.080123450,1.133893419,.9819805061,.8919280880,1.496662955,1.296148140,1.115546702,1.434274331,1.242118007,1.189237451,1.355261854,1.173691195,1.376274919,1.230977710},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.730085916,1.118033988,1.290994449,1.250000000,1.183215956,1.118033988,.7071067810,.8164965809,.7905694148,.7483314774,.4830458916,.5577733510,.5400617250,.3499271062,.4040610177,.2648642316,2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.732050808,2.,1.936491673,1.833030278,1.247219129,1.440164600,1.394433378,.9258201003,1.069044968,.7107053122,2.121320343,2.449489743,2.371708244,2.244994432,1.870828694,2.160246900,2.091650066,1.463850110,1.690308509,1.151474211,1.932183566,2.231093404,2.160246900,1.851640200,2.138089935,1.535298948,1.749635531,2.020305089,1.776763280,1.589185390},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.686773158,1.322875656,1.718465886,1.774823935,1.750000000,1.700840129,.8366600265,1.086853256,1.122497216,1.106797181,.5715476066,.7424621200,.7668115806,.4140393357,.5378528742,.3133915852,2.645751311,3.436931771,3.549647869,3.500000000,3.401680258,2.049390152,2.662235902,2.749545417,2.711088342,1.475729575,1.917028952,1.979898987,1.095445115,1.423024947,.8409178659,2.509980080,3.260559769,3.367491647,3.320391543,2.213594362,2.875543427,2.969848480,1.732050808,2.250000000,1.362442660,2.286190427,2.969848480,3.067246323,2.190890230,2.846049894,1.816590213,2.070196679,2.689264371,2.102294664,1.880349511},
			{2.449489743,3.394112549,3.651483718,3.703280400,3.674234614,3.610683736,1.500000000,2.078460970,2.236067977,2.267786838,2.250000000,.9486832980,1.314534138,1.414213562,1.434274331,.6480740700,.8979977729,.9660917834,.4694764779,.6505256900,.3553526559,3.,4.156921939,4.472135954,4.535573675,4.500000000,2.323790008,3.219937887,3.464101616,3.513240264,1.673320053,2.318620280,2.494438258,1.242118007,1.721129197,.9535112343,2.846049894,3.943602415,4.242640686,4.302822994,2.509980080,3.477930420,3.741657387,1.963961013,2.721344207,1.544864765,2.592296280,3.591991092,3.864367134,2.484236014,3.442258395,2.059819687,2.347382390,3.252628452,2.383778085,2.132115937},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,.8215838360,.7569125884,1.341640786,1.161895004,1.,.8874119673,.8049844717,1.060660172,.9185586536,.7905694148,.7015607601,.8280786714,.7171371654,.6172134000,.6561133241,.5682108062,.5297284632,1.341640786,1.161895004,1.,.8874119673,.8049844717,1.732050808,1.500000000,1.290994449,1.145643924,1.603567452,1.388730150,1.195228610,1.388730150,1.202675589,1.184508853,1.060660172,.9185586536,.7905694148,.7015607601,1.603567452,1.388730150,1.195228610,1.646831374,1.426197805,1.535298948,.8280786714,.7171371654,.6172134000,1.388730150,1.202675589,1.535298948,.6561133241,.5682108062,1.184508853,.5297284632},
			{1.767766952,2.041241452,1.976423538,1.870828694,1.767766952,1.675148486,1.732050808,2.,1.936491673,1.833030278,1.732050808,1.369306394,1.581138830,1.530931089,1.449137675,1.069044968,1.234426800,1.195228610,.8470386591,.9780759957,.6838765053,1.732050808,2.,1.936491673,1.833030278,1.732050808,2.236067977,2.581988897,2.500000000,2.366431913,2.070196679,2.390457218,2.314550249,1.792842914,2.070196679,1.529194354,1.369306394,1.581138830,1.530931089,1.449137675,2.070196679,2.390457218,2.314550249,2.126050161,2.454951265,1.982062418,1.069044968,1.234426800,1.195228610,1.792842914,2.070196679,1.982062418,.8470386591,.9780759957,1.529194354,.6838765053},
			{2.091650066,2.717133141,2.806243040,2.766992952,2.689264370,2.601456923,2.049390154,2.662235902,2.749545418,2.711088342,2.634930198,1.620185175,2.104682280,2.173706512,2.143303525,1.264911064,1.643167672,1.697056274,1.002229657,1.301934515,.8091735938,2.049390154,2.662235902,2.749545418,2.711088342,2.634930198,2.645751311,3.436931772,3.549647870,3.500000000,2.449489743,3.181980514,3.286335344,2.121320343,2.755675961,1.809367160,1.620185175,2.104682280,2.173706512,2.143303525,2.449489743,3.181980514,3.286335344,2.515576474,3.267829698,2.345207880,1.264911064,1.643167672,1.697056274,2.121320343,2.755675961,2.345207880,1.002229657,1.301934515,1.809367160,.8091735938},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.6540472289,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,1.309307342,1.133893419,.9759000734,.8660254040,1.118033988,.9682458365,.8333333333,.9449111824,.8183170887,.8010409895,.9660917834,.8366600265,.7200822999,.6390096507,.5796550700,1.603567452,1.388730150,1.195228610,1.060660172,1.732050808,1.500000000,1.290994449,1.666666667,1.443375673,1.535298947,.6546536711,.5669467095,.4879500367,.4330127020,1.299038106,1.125000000,.9682458365,1.581138830,1.369306394,1.658312395,.4472135954,.3872983346,.3333333333,1.,.8660254040,1.326649916,.3149703940,.2727723628,.7676494735,.2288688541},
			{1.527525232,1.763834207,1.707825127,1.616580754,1.527525232,1.447493729,1.870828694,2.160246899,2.091650066,1.979898987,1.870828694,1.690308509,1.951800145,1.889822365,1.788854382,1.443375673,1.666666667,1.613743061,1.219875090,1.408590424,1.034139471,1.247219129,1.440164599,1.394433378,1.319932658,1.247219129,2.070196678,2.390457218,2.314550249,2.190890230,2.236067977,2.581988897,2.500000000,2.151657415,2.484519974,1.982062418,.8451542548,.9759000724,.9449111824,.8944271908,1.677050983,1.936491673,1.875000000,2.041241452,2.357022604,2.140872097,.5773502693,.6666666667,.6454972245,1.290994449,1.490711985,1.712697677,.4066250301,.4695301414,.9910312088,.2954684202},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.024695077,.8874119674,.7637626160,.6777720856,.6148170462,.5664215159,1.434274331,1.242118007,1.069044968,.9486832980,.8605645990,1.417366774,1.227475633,1.056442819,.9375000000,1.290994449,1.118033988,.9622504490,1.145643924,.9921567416,1.009049958,.7171371655,.6210590036,.5345224840,.4743416490,.4302822994,1.388730150,1.202675589,1.035098339,.9185586536,1.666666667,1.443375673,1.242259987,1.732050808,1.500000000,1.692228225,.4252100321,.3682426898,.3169328455,.2812500000,1.,.8660254040,.7453559923,1.369306394,1.185854122,1.566698904,.2581988897,.2236067977,.1924500898,.6928203232,.6000000000,1.044465936,.1636634177,.1417366774,.4834937786,.1081124955},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.6180165404,.5477225576,.4743416490,.4082482906,.3622844188,.3286335346,.2886751347,.2500000000,.2151657415,.1909406540,.1690308509,.1463850109,.1259881577,.1071428571,.9278843614e-1,.7208691321e-1,1.643167673,1.423024947,1.224744872,1.086853256,.9859006037,1.060660172,.9185586536,.7905694150,.7015607599,.6546536709,.5669467095,.4879500362,.4252100321,.3682426898,.2901442286,1.732050808,1.500000000,1.290994449,1.145643924,1.309307341,1.133893419,.9759000724,.8964214566,.7763237544,.6267831705,1.690308509,1.463850109,1.259881577,1.417366774,1.227475633,1.044638617,1.607142857,1.391826542,1.450721144,1.513825177},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,1.367753011,.7071067810,.8164965809,.7905694148,.7483314772,.7071067810,.3726779962,.4303314828,.4166666667,.3944053189,.2182178903,.2519763153,.2439750183,.1383208338,.1597191412,.9306380472e-1,2.121320343,2.449489743,2.371708244,2.244994432,2.121320343,1.369306394,1.581138830,1.530931089,1.449137675,.8451542550,.9759000734,.9449111824,.5489437912,.6338656908,.3745745885,2.236067977,2.581988897,2.500000000,2.366431913,1.690308509,1.951800147,1.889822365,1.157275125,1.336306209,.8091735938,2.182178903,2.519763153,2.439750183,1.829812638,2.112885637,1.348622656,2.074812507,2.395787117,1.872872943,1.954339899},
			{1.707825129,2.218529919,2.291287848,2.259240285,2.195775165,2.124080684,.8366600265,1.086853256,1.122497216,1.106797181,1.075705749,.4409585519,.5728219619,.5916079783,.5833333333,.2581988897,.3354101966,.3464101616,.1636634177,.2126050161,.1101145788,2.509980080,3.260559769,3.367491648,3.320391543,3.227117245,1.620185174,2.104682280,2.173706512,2.143303525,1.,1.299038106,1.341640786,.6495190530,.8437500000,.4432026303,2.645751311,3.436931771,3.549647870,3.500000000,2.,2.598076212,2.683281572,1.369306394,1.778781184,.9574271080,2.581988897,3.354101966,3.464101616,2.165063510,2.812500000,1.595711847,2.454951265,3.189075241,2.216013152,2.312406154},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.6540472289,.9660917834,.8366600265,.7200822999,.6390096507,.5796550700,.6546536711,.5669467095,.4879500367,.4330127020,.4472135954,.3872983346,.3333333333,.3149703940,.2727723628,.2288688541,1.449137675,1.254990040,1.080123450,.9585144758,.8694826050,1.603567452,1.388730150,1.195228610,1.060660172,1.299038106,1.125000000,.9682458365,1.,.8660254040,.7676494735,1.309307342,1.133893419,.9759000734,.8660254040,1.732050808,1.500000000,1.290994449,1.581138830,1.369306394,1.326649916,1.118033988,.9682458365,.8333333333,1.666666667,1.443375673,1.658312395,.9449111824,.8183170887,1.535298947,.8010409895},
			{1.527525232,1.763834207,1.707825127,1.616580754,1.527525232,1.447493729,1.247219129,1.440164599,1.394433378,1.319932658,1.247219129,.8451542548,.9759000724,.9449111824,.8944271908,.5773502693,.6666666667,.6454972245,.4066250301,.4695301414,.2954684202,1.870828694,2.160246899,2.091650066,1.979898987,1.870828694,2.070196678,2.390457218,2.314550249,2.190890230,1.677050983,1.936491673,1.875000000,1.290994449,1.490711985,.9910312088,1.690308509,1.951800145,1.889822365,1.788854382,2.236067977,2.581988897,2.500000000,2.041241452,2.357022604,1.712697677,1.443375673,1.666666667,1.613743061,2.151657415,2.484519974,2.140872097,1.219875090,1.408590424,1.982062418,1.034139471},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.080123450,.9354143465,.8050764862,.7144345085,.6480740697,.5970607017,1.133893419,.9819805061,.8451542548,.7500000000,.6803360513,.8964214566,.7763237544,.6681531047,.5929270612,.6804138176,.5892556509,.5071505163,.5175491696,.4482107285,.3988620175,1.133893419,.9819805061,.8451542548,.7500000000,.6803360513,1.646831372,1.426197805,1.227475633,1.089276566,1.581138830,1.369306394,1.178511302,1.369306394,1.185854122,1.146706181,.8964214566,.7763237544,.6681531047,.5929270612,1.581138830,1.369306394,1.178511302,1.732050808,1.500000000,1.651445648,.6804138176,.5892556509,.5071505163,1.369306394,1.185854122,1.651445648,.5175491696,.4482107285,1.146706181,.3988620175},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.5066228049,.3741657387,.3240370349,.2788866754,.2474873734,.2244994432,.1690308509,.1463850109,.1259881577,.1118033988,.8660254040e-1,.7500000000e-1,.6454972245e-1,.4879500362e-1,.4225771273e-1,.2954684202e-1,1.496662955,1.296148140,1.115546702,.9899494934,.8979977729,.8280786708,.7171371654,.6172133998,.5477225574,.4472135954,.3872983346,.3333333333,.2581988897,.2236067977,.1585649934,1.690308509,1.463850109,1.259881577,1.118033988,1.118033988,.9682458365,.8333333333,.6804138176,.5892556509,.4281744193,1.732050808,1.500000000,1.290994449,1.290994449,1.118033988,.8563488387,1.707825127,1.479019946,1.387443692,1.654623153},
			{1.183215957,1.366260102,1.322875656,1.252198067,1.183215957,1.121223821,.4830458916,.5577733510,.5400617250,.5112077204,.4830458916,.2182178903,.2519763153,.2439750183,.2309401077,.1118033988,.1290994449,.1250000000,.6299407883e-1,.7273929675e-1,.3814480901e-1,1.932183566,2.231093404,2.160246900,2.044830882,1.932183566,1.069044968,1.234426800,1.195228610,1.131370850,.5773502693,.6666666667,.6454972245,.3333333333,.3849001795,.2047065263,2.182178903,2.519763153,2.439750183,2.309401077,1.443375673,1.666666667,1.613743061,.8784104612,1.014301033,.5527707984,2.236067977,2.581988897,2.500000000,1.666666667,1.924500898,1.105541597,2.204792759,2.545875387,1.791182105,2.136109305},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.024695077,.8874119674,.7637626160,.6777720856,.6148170462,.5664215159,.7171371655,.6210590036,.5345224840,.4743416490,.4302822994,.4252100321,.3682426898,.3169328455,.2812500000,.2581988897,.2236067977,.1924500898,.1636634177,.1417366774,.1081124955,1.434274331,1.242118007,1.069044968,.9486832980,.8605645990,1.388730150,1.202675589,1.035098339,.9185586536,1.,.8660254040,.7453559923,.6928203232,.6000000000,.4834937786,1.417366774,1.227475633,1.056442819,.9375000000,1.666666667,1.443375673,1.242259987,1.369306394,1.185854122,1.044465936,1.290994449,1.118033988,.9622504490,1.732050808,1.500000000,1.566698904,1.145643924,.9921567416,1.692228225,1.009049958},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.7745966692,.6708203931,.5773502693,.5123475385,.4647580015,.4281744193,.2710523708,.2347382389,.2020305089,.1792842914,.1626314225,.1071428571,.9278843614e-1,.7985957060e-1,.7086833870e-1,.4879500367e-1,.4225771273e-1,.3636964838e-1,.2474358298e-1,.2142857143e-1,.1362089413e-1,1.355261854,1.173691195,1.010152544,.8964214566,.8131571124,.6561133241,.5682108062,.4890379978,.4339781718,.3149703940,.2727723628,.2347650707,.1636634177,.1417366774,.9137173554e-1,1.607142857,1.391826542,1.197893559,1.063025080,.9449111824,.8183170887,.7042952119,.5175491696,.4482107285,.2960782628,1.707825129,1.479019946,1.272937693,1.145643924,.9921567416,.6908492799,1.732050808,1.500000000,1.279204298,1.716232660},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_stiffness_w[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_stiffness_w[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_bottom_face:" << std::endl;
		static const T timestep_dofs_to_bottom_face[52][56] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.154700539,2.,2.581988897,3.055050463,3.464101616,.6085806194,1.054092553,1.360827635,1.610152972,.2771598064,.4800548665,.6197481677,.1171213948,.2028602065,.4728035100e-1,1.154700539,2.,2.581988897,3.055050463,3.464101616,1.490711985,2.581988897,3.333333333,3.944053189,1.073435315,1.859244503,2.400274333,.6197481677,1.073435315,.3171662365,.6085806194,1.054092553,1.360827635,1.610152972,1.073435315,1.859244503,2.400274333,.9799078930,1.697250258,.6851573789,.2771598064,.4800548665,.6197481677,.6197481677,1.073435315,.6851573789,.1171213948,.2028602065,.3171662365,.4728035100e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,1.354996613,2.346922978,3.029864538,3.584984067,.9207895610,1.594854303,2.058948051,.5805991817,1.005627282,.3497284165,.1814813775e-1,.3143349662e-1,.4058046965e-1,.4801545922e-1,.5444441322e-1,.3495960260e-1,.6055180793e-1,.7817204790e-1,.9249441444e-1,.3756286152e-1,.6506078461e-1,.8399311177e-1,.3235996830e-1,.5604910924e-1,.2471097213e-1,.1503292582e-3,.2603779131e-3,.3361464403e-3,.3977338322e-3,.3956501378e-3,.6852861406e-3,.8847006032e-3,.5389279955e-3,.9334506698e-3,.5622712446e-3,.1076015046e-5,.1863712730e-5,.2406042788e-5,.3590156597e-5,.6218333635e-5,.5922410770e-5,.7146395199e-8,.1237791958e-7,.2887671511e-7,.4534132255e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,1.354996613,2.346922978,3.029864538,3.584984067,.9207895610,1.594854303,2.058948051,.5805991817,1.005627282,.3497284165,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,3.319050305,5.748763763,7.421622105,8.781381698,3.566202636,6.176844154,7.974271513,3.072242092,5.321279398,2.346049539,1.354996613,2.346922978,3.029864538,3.584984067,3.566202636,6.176844154,7.974271513,4.857641268,8.413681482,5.068046242,.9207895610,1.594854303,2.058948051,3.072242092,5.321279398,5.068046242,.5805991817,1.005627282,2.346049539,.3497284165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1814813775e-1,.3143349662e-1,.4058046965e-1,.4801545922e-1,.5444441322e-1,.1503292582e-3,.2603779131e-3,.3361464403e-3,.3977338322e-3,.1076015046e-5,.1863712730e-5,.2406042788e-5,.7146395199e-8,.1237791958e-7,.4534132255e-10,1.722976739,2.984283251,3.852693111,4.558567965,5.168930216,.3495960260e-1,.6055180793e-1,.7817204790e-1,.9249441444e-1,.3956501378e-3,.6852861406e-3,.8847006032e-3,.3590156597e-5,.6218333635e-5,.2887671511e-7,1.354996613,2.346922978,3.029864538,3.584984067,.3756286152e-1,.6506078461e-1,.8399311177e-1,.5389279955e-3,.9334506698e-3,.5922410770e-5,.9207895610,1.594854303,2.058948051,.3235996830e-1,.5604910924e-1,.5622712446e-3,.5805991817,1.005627282,.2471097213e-1,.3497284165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9363699245,1.621840284,2.093786803,2.477401956,.5289617350,.9161886001,1.182794397,.2772649114,.4802369137,.1388366216,.5995013155,1.038366737,1.340525694,1.586131391,1.798503946,.9600175758,1.662799218,2.146664559,2.539967760,.8574839288,1.485205731,1.917392354,.6140871797,1.063630196,.3898220464,.1640437399,.2841320922,.3668129538,.4340189401,.3589071350,.6216453928,.8025407512,.4064022821,.7039094010,.3524730288,.3878756937e-1,.6718204084e-1,.8673164178e-1,.1075826497,.1863386154,.1475305609,.8509799423e-2,.1473940496e-1,.2858477025e-1,.1783546975e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9363699245,1.621840284,2.093786803,2.477401956,.5289617350,.9161886001,1.182794397,.2772649114,.4802369137,.1388366216,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,2.293628526,3.972681141,5.128709298,6.068370681,2.048659991,3.548383191,4.580943002,1.467148005,2.541174888,.9313443707,.9363699245,1.621840284,2.093786803,2.477401956,2.048659991,3.548383191,4.580943002,2.319764682,4.017950291,2.011933789,.5289617350,.9161886001,1.182794397,1.467148005,2.541174888,2.011933789,.2772649114,.4802369137,.9313443707,.1388366216},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5995013155,1.038366737,1.340525694,1.586131391,1.798503946,.1640437399,.2841320922,.3668129538,.4340189401,.3878756937e-1,.6718204084e-1,.8673164178e-1,.8509799423e-2,.1473940496e-1,.1783546975e-2,1.432300150,2.480816631,3.202720499,3.789509999,4.296900450,.9600175758,1.662799218,2.146664559,2.539967760,.3589071350,.6216453928,.8025407512,.1075826497,.1863386154,.2858477025e-1,.9363699245,1.621840284,2.093786803,2.477401956,.8574839288,1.485205731,1.917392354,.4064022821,.7039094010,.1475305609,.5289617350,.9161886001,1.182794397,.6140871797,1.063630196,.3524730288,.2772649114,.4802369137,.3898220464,.1388366216},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,1.212281421,2.099733015,2.710743667,3.207395161,.7792166912,1.349642899,1.742381491,.4647366240,.8049474449,.2647853914,.2046688294,.3544968112,.4576534152,.5415028235,.6140064883,.3729226366,.6459209540,.8338803660,.9866605549,.3790037856,.6564538126,.8474782279,.3088343411,.5349167702,.2230693744,.1911977567e-1,.3311642289e-1,.4275311811e-1,.5058617157e-1,.4759739883e-1,.8244111306e-1,.1064310193,.6132467595e-1,.1062174545,.6051778323e-1,.1543395612e-2,.2673239616e-2,.3451137504e-2,.4870851947e-2,.8436563051e-2,.7600159080e-2,.1156020631e-3,.2002286468e-3,.4418338282e-3,.8271654356e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,1.212281421,2.099733015,2.710743667,3.207395161,.7792166912,1.349642899,1.742381491,.4647366240,.8049474449,.2647853914,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,2.969470906,5.143274482,6.639938805,7.856481545,3.017893269,5.227144472,6.748214496,2.459155065,4.259381518,1.776234404,1.212281421,2.099733015,2.710743667,3.207395161,3.017893269,5.227144472,6.748214496,3.888265562,6.734673506,3.837104864,.7792166912,1.349642899,1.742381491,2.459155065,4.259381518,3.837104864,.4647366240,.8049474449,1.776234404,.2647853914},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2046688294,.3544968112,.4576534152,.5415028235,.6140064883,.1911977567e-1,.3311642289e-1,.4275311811e-1,.5058617157e-1,.1543395612e-2,.2673239616e-2,.3451137504e-2,.1156020631e-3,.2002286468e-3,.8271654356e-5,1.629716393,2.822751595,3.644156639,4.311824283,4.889149181,.3729226366,.6459209540,.8338803660,.9866605549,.4759739883e-1,.8244111306e-1,.1064310193,.4870851947e-2,.8436563051e-2,.4418338282e-3,1.212281421,2.099733015,2.710743667,3.207395161,.3790037856,.6564538126,.8474782279,.6132467595e-1,.1062174545,.7600159080e-2,.7792166912,1.349642899,1.742381491,.3088343411,.5349167702,.6051778323e-1,.4647366240,.8049474449,.2230693744,.2647853914},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.3169457192,.5489660888,.7087121734,.8385595524,.1041670603,.1804226409,.2329246278,.3176653023e-1,.5502124434e-1,.9254386281e-2,1.797496348,3.113355001,4.019324022,4.755728318,5.392489045,1.674656646,2.900590397,3.744646101,4.430725019,.8702445254,1.507307732,1.945925915,.3625885274,.6280217519,.1339118908,1.474739846,2.554324341,3.297618546,3.901794882,1.877182782,3.251375951,4.197508304,1.236657185,2.141953076,.6240046052,1.045505193,1.810868114,2.337820683,1.687114596,2.922168200,1.346024785,.6877503643,1.191218574,1.344048463,.4321892234},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.3169457192,.5489660888,.7087121734,.8385595524,.1041670603,.1804226409,.2329246278,.3176653023e-1,.5502124434e-1,.9254386281e-2,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,.7763552880,1.344686804,1.735983199,2.054043021,.4034372898,.6987738835,.9021132044,.1680926781,.2911450589,.6208031047e-1,.3169457192,.5489660888,.7087121734,.8385595524,.4034372898,.6987738835,.9021132044,.2657778603,.4603407575,.1341087982,.1041670603,.1804226409,.2329246278,.1680926781,.2911450589,.1341087982,.3176653023e-1,.5502124434e-1,.6208031047e-1,.9254386281e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.797496348,3.113355001,4.019324022,4.755728318,5.392489045,1.474739846,2.554324341,3.297618546,3.901794882,1.045505193,1.810868114,2.337820683,.6877503643,1.191218574,.4321892234,.8333026341,1.443322500,1.863321335,2.204711536,2.499907902,1.674656646,2.900590397,3.744646101,4.430725019,1.877182782,3.251375951,4.197508304,1.687114596,2.922168200,1.344048463,.3169457192,.5489660888,.7087121734,.8385595524,.8702445254,1.507307732,1.945925915,1.236657185,2.141953076,1.346024785,.1041670603,.1804226409,.2329246278,.3625885274,.6280217519,.6240046052,.3176653023e-1,.5502124434e-1,.1339118908,.9254386281e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.1199176989,.2077035473,.2681441264,.3172724093,.2424253337e-1,.4198929950e-1,.5420795256e-1,.4547435761e-2,.7876389783e-2,.8148799146e-3,2.438964522,4.224410470,5.453690464,6.452893580,7.316893568,1.397695098,2.420878923,3.125341251,3.697953638,.4467629626,.7738161499,.9989923539,.1144983212,.1983169097,.2601078148e-1,2.715128242,4.702740065,6.071211317,7.183554108,2.125840578,3.682063888,4.753524039,.8614352786,1.492049670,.2673685503,2.611792036,4.523756504,5.840144534,2.592424872,4.490211595,1.272224784,2.331206226,4.037767626,2.802295901,1.987747725},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.1199176989,.2077035473,.2681441264,.3172724093,.2424253337e-1,.4198929950e-1,.5420795256e-1,.4547435761e-2,.7876389783e-2,.8148799146e-3,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,.2937371735,.5087677086,.6568162875,.7771555119,.9389092801e-1,.1626238576,.2099464974,.2406276826e-1,.4167793720e-1,.5466380647e-2,.1199176989,.2077035473,.2681441264,.3172724093,.9389092801e-1,.1626238576,.2099464974,.3804657725e-1,.6589860485e-1,.1180873184e-1,.2424253337e-1,.4198929950e-1,.5420795256e-1,.2406276826e-1,.4167793720e-1,.1180873184e-1,.4547435761e-2,.7876389783e-2,.5466380647e-2,.8148799146e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.438964522,4.224410470,5.453690464,6.452893580,7.316893568,2.715128242,4.702740065,6.071211317,7.183554108,2.611792036,4.523756504,5.840144534,2.331206226,4.037767626,1.987747725,.5125685468,.8877947652,1.146138113,1.356128904,1.537705640,1.397695098,2.420878923,3.125341251,3.697953638,2.125840578,3.682063888,4.753524039,2.592424872,4.490211595,2.802295901,.1199176989,.2077035473,.2681441264,.3172724093,.4467629626,.7738161499,.9989923539,.8614352786,1.492049670,1.272224784,.2424253337e-1,.4198929950e-1,.5420795256e-1,.1144983212,.1983169097,.2673685503,.4547435761e-2,.7876389783e-2,.2601078148e-1,.8148799146e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.3216132869e-2,.5570505532e-2,.7191491718e-2,.3076861022e-3,.5329279618e-3,.2812031928e-4,2.941263458,5.094417748,6.576865031,7.781851649,8.823790377,.8596589978,1.488973062,1.922255957,2.274443921,.1401445651,.2427375071,.3133727741,.1831823152e-1,.3172810771e-1,.2122380039e-2,3.948637231,6.839240305,8.829421268,10.44711213,1.576786555,2.731074426,3.525801923,.3258743636,.5644309546,.5158497099e-1,4.580616431,7.933860388,10.24256972,2.318872338,4.016404707,.5803898901,4.930539837,8.539945507,3.022825632,5.069947794},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.3216132869e-2,.5570505532e-2,.7191491718e-2,.3076861022e-3,.5329279618e-3,.2812031928e-4,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.7640636963e-1,.1323397142,.1708498365,.2021522527,.1245602905e-1,.2157447516e-1,.2785252767e-1,.1628121817e-2,.2819989708e-2,.1886368363e-3,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.1245602905e-1,.2157447516e-1,.2785252767e-1,.2574286624e-2,.4458795226e-2,.4075021408e-3,.3216132869e-2,.5570505532e-2,.7191491718e-2,.1628121817e-2,.2819989708e-2,.4075021408e-3,.3076861022e-3,.5329279618e-3,.1886368363e-3,.2812031928e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.941263458,5.094417748,6.576865031,7.781851649,8.823790377,3.948637231,6.839240305,8.829421268,10.44711213,4.580616431,7.933860388,10.24256972,4.930539837,8.539945507,5.069947794,.2614190785,.4527911260,.5845508299,.6916498695,.7842572356,.8596589978,1.488973062,1.922255957,2.274443921,1.576786555,2.731074426,3.525801923,2.318872338,4.016404707,3.022825632,.3119276979e-1,.5402746212e-1,.6974915366e-1,.8252831160e-1,.1401445651,.2427375071,.3133727741,.3258743636,.5644309546,.5803898901,.3216132869e-2,.5570505532e-2,.7191491718e-2,.1831823152e-1,.3172810771e-1,.5158497099e-1,.3076861022e-3,.5329279618e-3,.2122380039e-2,.2812031928e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.3420864716e-4,.5925111493e-4,.7649286046e-4,.7197424251e-6,.1246630449e-5,.1446627697e-7,3.349118367,5.800843169,7.488856330,8.860934307,10.04735510,.2152732603,.3728642244,.4813656438,.5695595107,.7718047390e-2,.1336805021e-1,.1725807861e-1,.2218615436e-3,.3842754658e-3,.5653129632e-5,5.119651213,8.867496018,11.44788813,13.54532391,.4496072984,.7787426841,1.005352482,.2043515547e-1,.3539472753e-1,.7114071839e-3,6.762598981,11.71316502,15.12163102,.7528940303,1.304050714,.4144232937e-1,8.288591754,14.35626204,1.117549271,9.704794288},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.3420864716e-4,.5925111493e-4,.7649286046e-4,.7197424251e-6,.1246630449e-5,.1446627697e-7,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.3695423161e-2,.6400660671e-2,.8263217397e-2,.9777170675e-2,.1324895208e-3,.2294785814e-3,.2962555747e-3,.3808518930e-5,.6596548290e-5,.9704273609e-7,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.1324895208e-3,.2294785814e-3,.2962555747e-3,.6021797165e-5,.1043005864e-4,.2096362696e-6,.3420864716e-4,.5925111493e-4,.7649286046e-4,.3808518930e-5,.6596548290e-5,.2096362696e-6,.7197424251e-6,.1246630449e-5,.9704273609e-7,.1446627697e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.349118367,5.800843169,7.488856330,8.860934307,10.04735510,5.119651213,8.867496018,11.44788813,13.54532391,6.762598981,11.71316502,15.12163102,8.288591754,14.35626204,9.704794288,.5749162514e-1,.9957841572e-1,.1285551819,.1521085425,.1724748754,.2152732603,.3728642244,.4813656438,.5695595107,.4496072984,.7787426841,1.005352482,.7528940303,1.304050714,1.117549271,.1508650189e-2,.2613058777e-2,.3373444378e-2,.3991513216e-2,.7718047390e-2,.1336805021e-1,.1725807861e-1,.2043515547e-1,.3539472753e-1,.4144232937e-1,.3420864716e-4,.5925111493e-4,.7649286046e-4,.2218615436e-3,.3842754658e-3,.7114071839e-3,.7197424251e-6,.1246630449e-5,.5653129632e-5,.1446627697e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4816958184,.8343216312,1.077104595,1.274447343,.1951696965,.3380438305,.4364127086,.7337460098e-1,.1270885369,.2635224308e-1,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.4120977289,.7137742043,.9214785357,1.090308107,.2640035315,.4572675299,.5903298427,.1356052452,.2348751745,.6174127088e-1,.5875925975e-1,.1017740233,.1313896991,.1554623886,.9220648278e-1,.1597063129,.2061799634,.7488558248e-1,.1297056336,.4658334547e-1,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1654169192e-1,.2865105086e-1,.1626980677e-1,.1091823989e-2,.1891094623e-2,.2630454079e-2,.1369544625e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,1.970938383,3.413765418,4.407152207,5.214612814,1.615335150,2.797842550,3.611999201,1.228417782,2.127682012,.8924173718,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,2.386704142,4.133892837,5.336832705,6.314625614,3.092845228,5.356965074,6.915812173,3.213475734,5.565903242,2.959539815,.4816958184,.8343216312,1.077104595,1.274447343,1.529002171,2.648309444,3.418952791,2.511857186,4.350664268,3.160657901,.1951696965,.3380438305,.4364127086,.7853709873,1.360302453,1.562526553,.7337460098e-1,.1270885369,.3575805848,.2635224308e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.5875925975e-1,.1017740233,.1313896991,.1554623886,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1091823989e-2,.1891094623e-2,.1369544625e-3,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,.8335857979,1.443812955,1.863954510,2.205460718,.1865140454,.3230518028,.4170580840,.3346031412e-1,.5795496410e-1,.5320847480e-2,1.970938383,3.413765418,4.407152207,5.214612814,1.080214264,1.870985987,2.415432523,.3064067886,.5307121257,.6657061456e-1,1.615335150,2.797842550,3.611999201,1.122345952,1.943960214,.3855501990,1.228417782,2.127682012,1.033655707,.8924173718},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.5875925975e-1,.1017740233,.1313896991,.1554623886,.8315101345e-2,.1440217800e-1,.1859313185e-1,.1091823989e-2,.1891094623e-2,.1369544625e-3,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4120977289,.7137742043,.9214785357,1.090308107,.9220648278e-1,.1597063129,.2061799634,.1654169192e-1,.2865105086e-1,.2630454079e-2,.4816958184,.8343216312,1.077104595,1.274447343,.2640035315,.4572675299,.5903298427,.7488558248e-1,.1297056336,.1626980677e-1,.1951696965,.3380438305,.4364127086,.1356052452,.2348751745,.4658334547e-1,.7337460098e-1,.1270885369,.6174127088e-1,.2635224308e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.027298721,1.779333580,2.297109774,2.717976938,3.081896165,.4816958184,.8343216312,1.077104595,1.274447343,.1951696965,.3380438305,.4364127086,.7337460098e-1,.1270885369,.2635224308e-1,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,2.386704142,4.133892837,5.336832705,6.314625614,1.529002171,2.648309444,3.418952791,.7853709873,1.360302453,.3575805848,1.970938383,3.413765418,4.407152207,5.214612814,3.092845228,5.356965074,6.915812173,2.511857186,4.350664268,1.562526553,1.615335150,2.797842550,3.611999201,3.213475734,5.565903242,3.160657901,1.228417782,2.127682012,2.959539815,.8924173718},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.078006171,3.599212267,4.646563055,5.497887550,6.234018515,1.970938383,3.413765418,4.407152207,5.214612814,1.615335150,2.797842550,3.611999201,1.228417782,2.127682012,.8924173718,.3587967227,.6214541532,.8022938619,.9492868993,1.076390168,.8335857979,1.443812955,1.863954510,2.205460718,1.080214264,1.870985987,2.415432523,1.122345952,1.943960214,1.033655707,.5875925975e-1,.1017740233,.1313896991,.1554623886,.1865140454,.3230518028,.4170580840,.3064067886,.5307121257,.3855501990,.8315101345e-2,.1440217800e-1,.1859313185e-1,.3346031412e-1,.5795496410e-1,.6657061456e-1,.1091823989e-2,.1891094623e-2,.5320847480e-2,.1369544625e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.6247151922,1.082038453,1.396905637,1.652841039,.2882545515,.4992715287,.6445567719,.1234139165,.2137591738,.5047665480e-1,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.9099854801e-1,.1576141086,.2034789393,.2407595278,.6638937401e-1,.1149897688,.1484511532,.3883469697e-1,.6726366826e-1,.2013601810e-1,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.3947983634e-2,.6838108240e-2,.8827959777e-2,.3651461244e-2,.6324516397e-2,.2586745805e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1373325428e-3,.2378669418e-3,.1538262752e-3,.1543374128e-5,.2673202406e-5,.4234517225e-5,.3753828413e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.258876421,3.912488728,5.051001230,5.976425253,1.981945831,3.432830877,4.431765606,1.613559350,2.794766775,1.254921551,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,2.909798360,5.039918600,6.506506935,7.698602827,4.036753290,6.991861793,9.026454761,4.490128369,7.777130471,4.427082670,.6247151922,1.082038453,1.396905637,1.652841039,2.122887626,3.676949226,4.746921039,3.733563649,6.466721933,5.029395863,.2882545515,.4992715287,.6445567719,1.241790556,2.150844335,2.644908288,.1234139165,.2137591738,.6438756851,.5047665480e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1543374128e-5,.2673202406e-5,.3753828413e-7,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,.1730372742,.2997093506,.3869231079,.4578135952,.7507244251e-2,.1300292847e-1,.1678670846e-1,.2611431652e-3,.4523132303e-3,.8052099006e-5,2.258876421,3.912488728,5.051001230,5.976425253,.2400540172,.4157857543,.5367771006,.1320313609e-1,.2286850254e-1,.5562127353e-3,1.981945831,3.432830877,4.431765606,.2670149191,.4624834065,.1778563440e-1,1.613559350,2.794766775,.2632657742,1.254921551},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.6061868131e-4,.1049946359e-3,.1355474921e-3,.1543374128e-5,.2673202406e-5,.3753828413e-7,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.9099854801e-1,.1576141086,.2034789393,.2407595278,.3947983634e-2,.6838108240e-2,.8827959777e-2,.1373325428e-3,.2378669418e-3,.4234517225e-5,.6247151922,1.082038453,1.396905637,1.652841039,.6638937401e-1,.1149897688,.1484511532,.3651461244e-2,.6324516397e-2,.1538262752e-3,.2882545515,.4992715287,.6445567719,.3883469697e-1,.6726366826e-1,.2586745805e-2,.1234139165,.2137591738,.2013601810e-1,.5047665480e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.169907010,2.026338380,2.615991600,3.095283004,3.509721028,.6247151922,1.082038453,1.396905637,1.652841039,.2882545515,.4992715287,.6445567719,.1234139165,.2137591738,.5047665480e-1,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.909798360,5.039918600,6.506506935,7.698602827,2.122887626,3.676949226,4.746921039,1.241790556,2.150844335,.6438756851,2.258876421,3.912488728,5.051001230,5.976425253,4.036753290,6.991861793,9.026454761,3.733563649,6.466721933,2.644908288,1.981945831,3.432830877,4.431765606,4.490128369,7.777130471,5.029395863,1.613559350,2.794766775,4.427082670,1.254921551},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224623628,3.853161149,4.974409654,5.885800878,6.673870881,2.258876421,3.912488728,5.051001230,5.976425253,1.981945831,3.432830877,4.431765606,1.613559350,2.794766775,1.254921551,.6957097879e-1,.1205004700,.1555654378,.1840675083,.2087129364,.1730372742,.2997093506,.3869231079,.4578135952,.2400540172,.4157857543,.5367771006,.2670149191,.4624834065,.2632657742,.2209202917e-2,.3826451697e-2,.4939927900e-2,.5845001515e-2,.7507244251e-2,.1300292847e-1,.1678670846e-1,.1320313609e-1,.2286850254e-1,.1778563440e-1,.6061868131e-4,.1049946359e-3,.1355474921e-3,.2611431652e-3,.4523132303e-3,.5562127353e-3,.1543374128e-5,.2673202406e-5,.8052099006e-5,.3753828413e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,.2296154773,.3977056730,.5134358159,.6075054502,.6423249772e-1,.1112539495,.1436282312,.1667256107e-1,.2887772287e-1,.4134169639e-2,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,-.1192467538e-1,-.2065414363e-1,-.2666438477e-1,-.3154972553e-1,-.5274365494e-2,-.9135469009e-2,-.1179383978e-1,-.1870471581e-2,-.3239751814e-2,-.5879824074e-3,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,.1118252823e-3,.1936870705e-3,.2500489327e-3,.6270337257e-4,.1086054271e-3,.2693007136e-4,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,-.8407960796e-6,-.1456301529e-5,-.5709621067e-6,.3368848215e-8,.5835016271e-8,.5603684524e-8,-.1771076019e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,3.501856062,6.065392619,7.830388205,9.265040271,3.825608374,6.626148073,8.554320378,3.877899792,6.716719466,3.755183165,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,2.196470374,3.804398286,4.911457068,5.811314373,3.794002148,6.571404483,8.483646708,5.254447192,9.100969504,6.450432215,.2296154773,.3977056730,.5134358159,.6075054502,.9715138718,1.682711386,2.172371058,2.127396643,3.684759074,3.568154409,.6423249772e-1,.1112539495,.1436282312,.3445322647,.5967473875,.9136820088,.1667256107e-1,.2887772287e-1,.1083036558,.4134169639e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,.3368848215e-8,.5835016271e-8,-.1771076019e-10,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,-.4656880908e-1,-.8065954338e-1,-.1041310228,-.1232094877,.4367054075e-3,.7563959535e-3,.9765029769e-3,-.3283516813e-5,-.5687217949e-5,.2188377514e-7,3.501856062,6.065392619,7.830388205,9.265040271,-.8043912806e-1,-.1393246566,-.1798673583,.9562865187e-3,.1656336837e-2,-.8707719261e-5,3.825608374,6.626148073,8.554320378,-.1114029814,-.1929556239,.1603921848e-2,3.877899792,6.716719466,-.1367598443,3.755183165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,.3368848215e-8,.5835016271e-8,-.1771076019e-10,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,-.1192467538e-1,-.2065414363e-1,-.2666438477e-1,-.3154972553e-1,.1118252823e-3,.1936870705e-3,.2500489327e-3,-.8407960796e-6,-.1456301529e-5,.5603684524e-8,.2296154773,.3977056730,.5134358159,.6075054502,-.5274365494e-2,-.9135469009e-2,-.1179383978e-1,.6270337257e-4,.1086054271e-3,-.5709621067e-6,.6423249772e-1,.1112539495,.1436282312,-.1870471581e-2,-.3239751814e-2,.2693007136e-4,.1667256107e-1,.2887772287e-1,-.5879824074e-3,.4134169639e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.7092688532,1.228489690,1.585973370,1.876548998,2.127806560,.2296154773,.3977056730,.5134358159,.6075054502,.6423249772e-1,.1112539495,.1436282312,.1667256107e-1,.2887772287e-1,.4134169639e-2,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,2.196470374,3.804398286,4.911457068,5.811314373,.9715138718,1.682711386,2.172371058,.3445322647,.5967473875,.1083036558,3.501856062,6.065392619,7.830388205,9.265040271,3.794002148,6.571404483,8.483646708,2.127396643,3.684759074,.9136820088,3.825608374,6.626148073,8.554320378,5.254447192,9.100969504,3.568154409,3.877899792,6.716719466,6.450432215,3.755183165},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.769870437,4.797556326,6.193618584,7.328388338,8.309611311,3.501856062,6.065392619,7.830388205,9.265040271,3.825608374,6.626148073,8.554320378,3.877899792,6.716719466,3.755183165,-.1503767417e-1,-.2604601568e-1,-.3362526166e-1,-.3978594614e-1,-.4511302250e-1,-.4656880908e-1,-.8065954338e-1,-.1041310228,-.1232094877,-.8043912806e-1,-.1393246566,-.1798673583,-.1114029814,-.1929556239,-.1367598443,.1032145021e-3,.1787727617e-3,.2307946430e-3,.2730799044e-3,.4367054075e-3,.7563959535e-3,.9765029769e-3,.9562865187e-3,.1656336837e-2,.1603921848e-2,-.6121588826e-6,-.1060290287e-5,-.1368828874e-5,-.3283516813e-5,-.5687217949e-5,-.8707719261e-5,.3368848215e-8,.5835016271e-8,.2188377514e-7,-.1771076019e-10},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1963948976,.3401659410,.4391523414,.5196120579,.5080990387e-1,.8800533503e-1,.1136143990,.1219720719e-1,.2112618256e-1,.2797118711e-2,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.1065535599,.1845561795,.2382610033,.2819142209,.4358691597e-1,.7549475298e-1,.9746330700e-1,.1429557258e-1,.2476065805e-1,.4156035932e-2,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.9654244188e-2,.1672164144e-1,.2158754627e-1,.5006493970e-2,.8671501924e-2,.1988588036e-2,.5521216640e-3,.9563027739e-3,.1234581572e-2,.7013355149e-3,.1214748745e-2,.4404604930e-3,.2935681623e-4,.5084749726e-4,.4516122448e-4,.1491150883e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,3.236488548,5.605762603,7.237008402,8.562943823,3.399101902,5.887417193,7.600622914,3.312440982,5.737316079,3.083689087,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,1.952889915,3.382504555,4.366794603,5.166861053,3.242932261,5.616923440,7.251416980,4.317727482,7.478523374,5.095712344,.1963948976,.3401659410,.4391523414,.5196120579,.7988512884,1.383651019,1.786285784,1.681718474,2.912821842,2.711667007,.5080990387e-1,.8800533503e-1,.1136143990,.2620060705,.4538078261,.6679814775,.1219720719e-1,.2112618256e-1,.7617090096e-1,.2797118711e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2935681623e-4,.5084749726e-4,.1491150883e-5,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,.4325535702,.7492047606,.9672191871,1.144429176,.3919134934e-1,.6788140826e-1,.8763452124e-1,.2847067532e-2,.4931265619e-2,.1833317338e-3,3.236488548,5.605762603,7.237008402,8.562943823,.7182903227,1.244115334,1.606145989,.8250448766e-1,.1429019645,.7258566080e-2,3.399101902,5.887417193,7.600622914,.9563511098,1.656448712,.1330333825,3.312440982,5.737316079,1.128670157,3.083689087},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2935681623e-4,.5084749726e-4,.1491150883e-5,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1065535599,.1845561795,.2382610033,.2819142209,.9654244188e-2,.1672164144e-1,.2158754627e-1,.7013355149e-3,.1214748745e-2,.4516122448e-4,.1963948976,.3401659410,.4391523414,.5196120579,.4358691597e-1,.7549475298e-1,.9746330700e-1,.5006493970e-2,.8671501924e-2,.4404604930e-3,.5080990387e-1,.8800533503e-1,.1136143990,.1429557258e-1,.2476065805e-1,.1988588036e-2,.1219720719e-1,.2112618256e-1,.4156035932e-2,.2797118711e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.6559570585,1.136150953,1.466764573,1.735499247,1.967871176,.1963948976,.3401659410,.4391523414,.5196120579,.5080990387e-1,.8800533503e-1,.1136143990,.1219720719e-1,.2112618256e-1,.2797118711e-2,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,1.952889915,3.382504555,4.366794603,5.166861053,.7988512884,1.383651019,1.786285784,.2620060705,.4538078261,.7617090096e-1,3.236488548,5.605762603,7.237008402,8.562943823,3.242932261,5.616923440,7.251416980,1.681718474,2.912821842,.6679814775,3.399101902,5.887417193,7.600622914,4.317727482,7.478523374,2.711667007,3.312440982,5.737316079,5.095712344,3.083689087},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.662853947,4.612198328,5.954322439,7.045249321,7.988561841,3.236488548,5.605762603,7.237008402,8.562943823,3.399101902,5.887417193,7.600622914,3.312440982,5.737316079,3.083689087,.1452906103,.2516507188,.3248796810,.3844028226,.4358718308,.4325535702,.7492047606,.9672191871,1.144429176,.7182903227,1.244115334,1.606145989,.9563511098,1.656448712,1.128670157,.9635061189e-2,.1668841551e-1,.2154465179e-1,.2549197578e-1,.3919134934e-1,.6788140826e-1,.8763452124e-1,.8250448766e-1,.1429019645,.1330333825,.5521216640e-3,.9563027739e-3,.1234581572e-2,.2847067532e-2,.4931265619e-2,.7258566080e-2,.2935681623e-4,.5084749726e-4,.1833317338e-3,.1491150883e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.4641847056e-2,.8039914942e-2,.1037948556e-1,.5018612597e-3,.8692492001e-3,.5183403142e-4,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1638187454e-1,.2837423903e-1,.3663098508e-1,.4334236605e-1,.3018092809e-2,.5227490086e-2,.6748660682e-2,.4458192555e-3,.7721816018e-3,.5837372111e-4,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.5066741239e-3,.8775853253e-3,.1132957783e-2,.1183382006e-3,.2049677758e-3,.2116980724e-4,.2196237131e-4,.3803994296e-4,.4910935520e-4,.1256467012e-4,.2176264704e-4,.3553964112e-5,.3986283756e-6,.6904445999e-6,.2761886212e-6,.6911871564e-8},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,4.440487963,7.691150763,9.929232940,11.74842685,5.462596453,9.461494597,12.21473700,6.235358013,10.79995688,6.799267692,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,1.030236404,1.784421795,2.303678632,2.725749316,2.003896641,3.470850795,4.480849108,3.125150808,5.412919982,4.320151396,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.1898042300,.3287505697,.4244151605,.4680276955,.8106477480,.8839606376,.4641847056e-2,.8039914942e-2,.1037948556e-1,.2803703724e-1,.4856157301e-1,.8372660778e-1,.5018612597e-3,.8692492001e-3,.3671053173e-2,.5183403142e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.2196237131e-4,.3803994296e-4,.4910935520e-4,.3986283756e-6,.6904445999e-6,.6911871564e-8,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,.1729549620,.2995667817,.3867390522,.4575958176,.5349314793e-2,.9265285003e-2,.1196143151e-1,.1326540524e-3,.2297635587e-3,.2915917367e-5,4.440487963,7.691150763,9.929232940,11.74842685,.3364119790,.5826826397,.7522400533,.1319057787e-1,.2284675106e-1,.3961429206e-3,5.462596453,9.461494597,12.21473700,.5246469034,.9087150930,.2491295222e-1,6.235358013,10.79995688,.7252622967,6.799267692},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.2196237131e-4,.3803994296e-4,.4910935520e-4,.3986283756e-6,.6904445999e-6,.6911871564e-8,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.1638187454e-1,.2837423903e-1,.3663098508e-1,.4334236605e-1,.5066741239e-3,.8775853253e-3,.1132957783e-2,.1256467012e-4,.2176264704e-4,.2761886212e-6,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.3018092809e-2,.5227490086e-2,.6748660682e-2,.1183382006e-3,.2049677758e-3,.3553964112e-5,.4641847056e-2,.8039914942e-2,.1037948556e-1,.4458192555e-3,.7721816018e-3,.2116980724e-4,.5018612597e-3,.8692492001e-3,.5837372111e-4,.5183403142e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.2954311110,.5117016941,.6606040466,.7816372490,.8862933327,.3983747795e-1,.6900053586e-1,.8907930878e-1,.1054000596,.4641847056e-2,.8039914942e-2,.1037948556e-1,.5018612597e-3,.8692492001e-3,.5183403142e-4,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,1.030236404,1.784421795,2.303678632,2.725749316,.1898042300,.3287505697,.4244151605,.2803703724e-1,.4856157301e-1,.3671053173e-2,4.440487963,7.691150763,9.929232940,11.74842685,2.003896641,3.470850795,4.480849108,.4680276955,.8106477480,.8372660778e-1,5.462596453,9.461494597,12.21473700,3.125150808,5.412919982,.8839606376,6.235358013,10.79995688,4.320151396,6.799267692},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.119073853,5.402394385,6.974461159,8.252293733,9.357221560,4.440487963,7.691150763,9.929232940,11.74842685,5.462596453,9.461494597,12.21473700,6.235358013,10.79995688,6.799267692,.4959665215e-1,.8590392138e-1,.1109014856,.1312204074,.1487899564,.1729549620,.2995667817,.3867390522,.4575958176,.3364119790,.5826826397,.7522400533,.5246469034,.9087150930,.7252622967,.1122752692e-2,.1944664706e-2,.2510551341e-2,.2970524407e-2,.5349314793e-2,.9265285003e-2,.1196143151e-1,.1319057787e-1,.2284675106e-1,.2491295222e-1,.2196237131e-4,.3803994296e-4,.4910935520e-4,.1326540524e-3,.2297635587e-3,.3961429206e-3,.3986283756e-6,.6904445999e-6,.2915917367e-5,.6911871564e-8}
		};
		for (int j = 0; j < 52; j++)
		{
			std::cout << timestep_dofs_to_bottom_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_bottom_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_top_face:" << std::endl;
		static const T timestep_dofs_to_top_face[52][56] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,1.154700539,0.,0.,0.,0.,.6085806194,0.,0.,0.,.2771598064,0.,0.,.1171213948,0.,.4728035100e-1,1.154700539,0.,0.,0.,0.,1.490711985,0.,0.,0.,1.073435315,0.,0.,.6197481677,0.,.3171662365,.6085806194,0.,0.,0.,1.073435315,0.,0.,.9799078930,0.,.6851573789,.2771598064,0.,0.,.6197481677,0.,.6851573789,.1171213948,0.,.3171662365,.4728035100e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.722976739,0.,0.,0.,0.,1.354996613,0.,0.,0.,.9207895610,0.,0.,.5805991817,0.,.3497284165,.1814813775e-1,0.,0.,0.,0.,.3495960260e-1,0.,0.,0.,.3756286152e-1,0.,0.,.3235996830e-1,0.,.2471097213e-1,.1503292582e-3,0.,0.,0.,.3956501378e-3,0.,0.,.5389279955e-3,0.,.5622712446e-3,.1076015046e-5,0.,0.,.3590156597e-5,0.,.5922410770e-5,.7146395199e-8,0.,.2887671511e-7,.4534132255e-10},
			{1.414213562,0.,0.,0.,0.,0.,1.722976739,0.,0.,0.,0.,1.354996613,0.,0.,0.,.9207895610,0.,0.,.5805991817,0.,.3497284165,1.722976739,0.,0.,0.,0.,3.319050305,0.,0.,0.,3.566202636,0.,0.,3.072242092,0.,2.346049539,1.354996613,0.,0.,0.,3.566202636,0.,0.,4.857641268,0.,5.068046242,.9207895610,0.,0.,3.072242092,0.,5.068046242,.5805991817,0.,2.346049539,.3497284165},
			{1.414213562,0.,0.,0.,0.,0.,.1814813775e-1,0.,0.,0.,0.,.1503292582e-3,0.,0.,0.,.1076015046e-5,0.,0.,.7146395199e-8,0.,.4534132255e-10,1.722976739,0.,0.,0.,0.,.3495960260e-1,0.,0.,0.,.3956501378e-3,0.,0.,.3590156597e-5,0.,.2887671511e-7,1.354996613,0.,0.,0.,.3756286152e-1,0.,0.,.5389279955e-3,0.,.5922410770e-5,.9207895610,0.,0.,.3235996830e-1,0.,.5622712446e-3,.5805991817,0.,.2471097213e-1,.3497284165},
			{1.414213562,0.,0.,0.,0.,0.,1.432300150,0.,0.,0.,0.,.9363699245,0.,0.,0.,.5289617350,0.,0.,.2772649114,0.,.1388366216,.5995013155,0.,0.,0.,0.,.9600175758,0.,0.,0.,.8574839288,0.,0.,.6140871797,0.,.3898220464,.1640437399,0.,0.,0.,.3589071350,0.,0.,.4064022821,0.,.3524730288,.3878756937e-1,0.,0.,.1075826497,0.,.1475305609,.8509799423e-2,0.,.2858477025e-1,.1783546975e-2},
			{1.414213562,0.,0.,0.,0.,0.,1.432300150,0.,0.,0.,0.,.9363699245,0.,0.,0.,.5289617350,0.,0.,.2772649114,0.,.1388366216,1.432300150,0.,0.,0.,0.,2.293628526,0.,0.,0.,2.048659991,0.,0.,1.467148005,0.,.9313443707,.9363699245,0.,0.,0.,2.048659991,0.,0.,2.319764682,0.,2.011933789,.5289617350,0.,0.,1.467148005,0.,2.011933789,.2772649114,0.,.9313443707,.1388366216},
			{1.414213562,0.,0.,0.,0.,0.,.5995013155,0.,0.,0.,0.,.1640437399,0.,0.,0.,.3878756937e-1,0.,0.,.8509799423e-2,0.,.1783546975e-2,1.432300150,0.,0.,0.,0.,.9600175758,0.,0.,0.,.3589071350,0.,0.,.1075826497,0.,.2858477025e-1,.9363699245,0.,0.,0.,.8574839288,0.,0.,.4064022821,0.,.1475305609,.5289617350,0.,0.,.6140871797,0.,.3524730288,.2772649114,0.,.3898220464,.1388366216},
			{1.414213562,0.,0.,0.,0.,0.,1.629716393,0.,0.,0.,0.,1.212281421,0.,0.,0.,.7792166912,0.,0.,.4647366240,0.,.2647853914,.2046688294,0.,0.,0.,0.,.3729226366,0.,0.,0.,.3790037856,0.,0.,.3088343411,0.,.2230693744,.1911977567e-1,0.,0.,0.,.4759739883e-1,0.,0.,.6132467595e-1,0.,.6051778323e-1,.1543395612e-2,0.,0.,.4870851947e-2,0.,.7600159080e-2,.1156020631e-3,0.,.4418338282e-3,.8271654356e-5},
			{1.414213562,0.,0.,0.,0.,0.,1.629716393,0.,0.,0.,0.,1.212281421,0.,0.,0.,.7792166912,0.,0.,.4647366240,0.,.2647853914,1.629716393,0.,0.,0.,0.,2.969470906,0.,0.,0.,3.017893269,0.,0.,2.459155065,0.,1.776234404,1.212281421,0.,0.,0.,3.017893269,0.,0.,3.888265562,0.,3.837104864,.7792166912,0.,0.,2.459155065,0.,3.837104864,.4647366240,0.,1.776234404,.2647853914},
			{1.414213562,0.,0.,0.,0.,0.,.2046688294,0.,0.,0.,0.,.1911977567e-1,0.,0.,0.,.1543395612e-2,0.,0.,.1156020631e-3,0.,.8271654356e-5,1.629716393,0.,0.,0.,0.,.3729226366,0.,0.,0.,.4759739883e-1,0.,0.,.4870851947e-2,0.,.4418338282e-3,1.212281421,0.,0.,0.,.3790037856,0.,0.,.6132467595e-1,0.,.7600159080e-2,.7792166912,0.,0.,.3088343411,0.,.6051778323e-1,.4647366240,0.,.2230693744,.2647853914},
			{1.414213562,0.,0.,0.,0.,0.,.8333026341,0.,0.,0.,0.,.3169457192,0.,0.,0.,.1041670603,0.,0.,.3176653023e-1,0.,.9254386281e-2,1.797496348,0.,0.,0.,0.,1.674656646,0.,0.,0.,.8702445254,0.,0.,.3625885274,0.,.1339118908,1.474739846,0.,0.,0.,1.877182782,0.,0.,1.236657185,0.,.6240046052,1.045505193,0.,0.,1.687114596,0.,1.346024785,.6877503643,0.,1.344048463,.4321892234},
			{1.414213562,0.,0.,0.,0.,0.,.8333026341,0.,0.,0.,0.,.3169457192,0.,0.,0.,.1041670603,0.,0.,.3176653023e-1,0.,.9254386281e-2,.8333026341,0.,0.,0.,0.,.7763552880,0.,0.,0.,.4034372898,0.,0.,.1680926781,0.,.6208031047e-1,.3169457192,0.,0.,0.,.4034372898,0.,0.,.2657778603,0.,.1341087982,.1041670603,0.,0.,.1680926781,0.,.1341087982,.3176653023e-1,0.,.6208031047e-1,.9254386281e-2},
			{1.414213562,0.,0.,0.,0.,0.,1.797496348,0.,0.,0.,0.,1.474739846,0.,0.,0.,1.045505193,0.,0.,.6877503643,0.,.4321892234,.8333026341,0.,0.,0.,0.,1.674656646,0.,0.,0.,1.877182782,0.,0.,1.687114596,0.,1.344048463,.3169457192,0.,0.,0.,.8702445254,0.,0.,1.236657185,0.,1.346024785,.1041670603,0.,0.,.3625885274,0.,.6240046052,.3176653023e-1,0.,.1339118908,.9254386281e-2},
			{1.414213562,0.,0.,0.,0.,0.,.5125685468,0.,0.,0.,0.,.1199176989,0.,0.,0.,.2424253337e-1,0.,0.,.4547435761e-2,0.,.8148799146e-3,2.438964522,0.,0.,0.,0.,1.397695098,0.,0.,0.,.4467629626,0.,0.,.1144983212,0.,.2601078148e-1,2.715128242,0.,0.,0.,2.125840578,0.,0.,.8614352786,0.,.2673685503,2.611792036,0.,0.,2.592424872,0.,1.272224784,2.331206226,0.,2.802295901,1.987747725},
			{1.414213562,0.,0.,0.,0.,0.,.5125685468,0.,0.,0.,0.,.1199176989,0.,0.,0.,.2424253337e-1,0.,0.,.4547435761e-2,0.,.8148799146e-3,.5125685468,0.,0.,0.,0.,.2937371735,0.,0.,0.,.9389092801e-1,0.,0.,.2406276826e-1,0.,.5466380647e-2,.1199176989,0.,0.,0.,.9389092801e-1,0.,0.,.3804657725e-1,0.,.1180873184e-1,.2424253337e-1,0.,0.,.2406276826e-1,0.,.1180873184e-1,.4547435761e-2,0.,.5466380647e-2,.8148799146e-3},
			{1.414213562,0.,0.,0.,0.,0.,2.438964522,0.,0.,0.,0.,2.715128242,0.,0.,0.,2.611792036,0.,0.,2.331206226,0.,1.987747725,.5125685468,0.,0.,0.,0.,1.397695098,0.,0.,0.,2.125840578,0.,0.,2.592424872,0.,2.802295901,.1199176989,0.,0.,0.,.4467629626,0.,0.,.8614352786,0.,1.272224784,.2424253337e-1,0.,0.,.1144983212,0.,.2673685503,.4547435761e-2,0.,.2601078148e-1,.8148799146e-3},
			{1.414213562,0.,0.,0.,0.,0.,.2614190785,0.,0.,0.,0.,.3119276979e-1,0.,0.,0.,.3216132869e-2,0.,0.,.3076861022e-3,0.,.2812031928e-4,2.941263458,0.,0.,0.,0.,.8596589978,0.,0.,0.,.1401445651,0.,0.,.1831823152e-1,0.,.2122380039e-2,3.948637231,0.,0.,0.,1.576786555,0.,0.,.3258743636,0.,.5158497099e-1,4.580616431,0.,0.,2.318872338,0.,.5803898901,4.930539837,0.,3.022825632,5.069947794},
			{1.414213562,0.,0.,0.,0.,0.,.2614190785,0.,0.,0.,0.,.3119276979e-1,0.,0.,0.,.3216132869e-2,0.,0.,.3076861022e-3,0.,.2812031928e-4,.2614190785,0.,0.,0.,0.,.7640636963e-1,0.,0.,0.,.1245602905e-1,0.,0.,.1628121817e-2,0.,.1886368363e-3,.3119276979e-1,0.,0.,0.,.1245602905e-1,0.,0.,.2574286624e-2,0.,.4075021408e-3,.3216132869e-2,0.,0.,.1628121817e-2,0.,.4075021408e-3,.3076861022e-3,0.,.1886368363e-3,.2812031928e-4},
			{1.414213562,0.,0.,0.,0.,0.,2.941263458,0.,0.,0.,0.,3.948637231,0.,0.,0.,4.580616431,0.,0.,4.930539837,0.,5.069947794,.2614190785,0.,0.,0.,0.,.8596589978,0.,0.,0.,1.576786555,0.,0.,2.318872338,0.,3.022825632,.3119276979e-1,0.,0.,0.,.1401445651,0.,0.,.3258743636,0.,.5803898901,.3216132869e-2,0.,0.,.1831823152e-1,0.,.5158497099e-1,.3076861022e-3,0.,.2122380039e-2,.2812031928e-4},
			{1.414213562,0.,0.,0.,0.,0.,.5749162514e-1,0.,0.,0.,0.,.1508650189e-2,0.,0.,0.,.3420864716e-4,0.,0.,.7197424251e-6,0.,.1446627697e-7,3.349118367,0.,0.,0.,0.,.2152732603,0.,0.,0.,.7718047390e-2,0.,0.,.2218615436e-3,0.,.5653129632e-5,5.119651213,0.,0.,0.,.4496072984,0.,0.,.2043515547e-1,0.,.7114071839e-3,6.762598981,0.,0.,.7528940303,0.,.4144232937e-1,8.288591754,0.,1.117549271,9.704794288},
			{1.414213562,0.,0.,0.,0.,0.,.5749162514e-1,0.,0.,0.,0.,.1508650189e-2,0.,0.,0.,.3420864716e-4,0.,0.,.7197424251e-6,0.,.1446627697e-7,.5749162514e-1,0.,0.,0.,0.,.3695423161e-2,0.,0.,0.,.1324895208e-3,0.,0.,.3808518930e-5,0.,.9704273609e-7,.1508650189e-2,0.,0.,0.,.1324895208e-3,0.,0.,.6021797165e-5,0.,.2096362696e-6,.3420864716e-4,0.,0.,.3808518930e-5,0.,.2096362696e-6,.7197424251e-6,0.,.9704273609e-7,.1446627697e-7},
			{1.414213562,0.,0.,0.,0.,0.,3.349118367,0.,0.,0.,0.,5.119651213,0.,0.,0.,6.762598981,0.,0.,8.288591754,0.,9.704794288,.5749162514e-1,0.,0.,0.,0.,.2152732603,0.,0.,0.,.4496072984,0.,0.,.7528940303,0.,1.117549271,.1508650189e-2,0.,0.,0.,.7718047390e-2,0.,0.,.2043515547e-1,0.,.4144232937e-1,.3420864716e-4,0.,0.,.2218615436e-3,0.,.7114071839e-3,.7197424251e-6,0.,.5653129632e-5,.1446627697e-7},
			{1.414213562,0.,0.,0.,0.,0.,1.027298721,0.,0.,0.,0.,.4816958184,0.,0.,0.,.1951696965,0.,0.,.7337460098e-1,0.,.2635224308e-1,.3587967227,0.,0.,0.,0.,.4120977289,0.,0.,0.,.2640035315,0.,0.,.1356052452,0.,.6174127088e-1,.5875925975e-1,0.,0.,0.,.9220648278e-1,0.,0.,.7488558248e-1,0.,.4658334547e-1,.8315101345e-2,0.,0.,.1654169192e-1,0.,.1626980677e-1,.1091823989e-2,0.,.2630454079e-2,.1369544625e-3},
			{1.414213562,0.,0.,0.,0.,0.,2.078006171,0.,0.,0.,0.,1.970938383,0.,0.,0.,1.615335150,0.,0.,1.228417782,0.,.8924173718,1.027298721,0.,0.,0.,0.,2.386704142,0.,0.,0.,3.092845228,0.,0.,3.213475734,0.,2.959539815,.4816958184,0.,0.,0.,1.529002171,0.,0.,2.511857186,0.,3.160657901,.1951696965,0.,0.,.7853709873,0.,1.562526553,.7337460098e-1,0.,.3575805848,.2635224308e-1},
			{1.414213562,0.,0.,0.,0.,0.,.3587967227,0.,0.,0.,0.,.5875925975e-1,0.,0.,0.,.8315101345e-2,0.,0.,.1091823989e-2,0.,.1369544625e-3,2.078006171,0.,0.,0.,0.,.8335857979,0.,0.,0.,.1865140454,0.,0.,.3346031412e-1,0.,.5320847480e-2,1.970938383,0.,0.,0.,1.080214264,0.,0.,.3064067886,0.,.6657061456e-1,1.615335150,0.,0.,1.122345952,0.,.3855501990,1.228417782,0.,1.033655707,.8924173718},
			{1.414213562,0.,0.,0.,0.,0.,.3587967227,0.,0.,0.,0.,.5875925975e-1,0.,0.,0.,.8315101345e-2,0.,0.,.1091823989e-2,0.,.1369544625e-3,1.027298721,0.,0.,0.,0.,.4120977289,0.,0.,0.,.9220648278e-1,0.,0.,.1654169192e-1,0.,.2630454079e-2,.4816958184,0.,0.,0.,.2640035315,0.,0.,.7488558248e-1,0.,.1626980677e-1,.1951696965,0.,0.,.1356052452,0.,.4658334547e-1,.7337460098e-1,0.,.6174127088e-1,.2635224308e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.027298721,0.,0.,0.,0.,.4816958184,0.,0.,0.,.1951696965,0.,0.,.7337460098e-1,0.,.2635224308e-1,2.078006171,0.,0.,0.,0.,2.386704142,0.,0.,0.,1.529002171,0.,0.,.7853709873,0.,.3575805848,1.970938383,0.,0.,0.,3.092845228,0.,0.,2.511857186,0.,1.562526553,1.615335150,0.,0.,3.213475734,0.,3.160657901,1.228417782,0.,2.959539815,.8924173718},
			{1.414213562,0.,0.,0.,0.,0.,2.078006171,0.,0.,0.,0.,1.970938383,0.,0.,0.,1.615335150,0.,0.,1.228417782,0.,.8924173718,.3587967227,0.,0.,0.,0.,.8335857979,0.,0.,0.,1.080214264,0.,0.,1.122345952,0.,1.033655707,.5875925975e-1,0.,0.,0.,.1865140454,0.,0.,.3064067886,0.,.3855501990,.8315101345e-2,0.,0.,.3346031412e-1,0.,.6657061456e-1,.1091823989e-2,0.,.5320847480e-2,.1369544625e-3},
			{1.414213562,0.,0.,0.,0.,0.,1.169907010,0.,0.,0.,0.,.6247151922,0.,0.,0.,.2882545515,0.,0.,.1234139165,0.,.5047665480e-1,.6957097879e-1,0.,0.,0.,0.,.9099854801e-1,0.,0.,0.,.6638937401e-1,0.,0.,.3883469697e-1,0.,.2013601810e-1,.2209202917e-2,0.,0.,0.,.3947983634e-2,0.,0.,.3651461244e-2,0.,.2586745805e-2,.6061868131e-4,0.,0.,.1373325428e-3,0.,.1538262752e-3,.1543374128e-5,0.,.4234517225e-5,.3753828413e-7},
			{1.414213562,0.,0.,0.,0.,0.,2.224623628,0.,0.,0.,0.,2.258876421,0.,0.,0.,1.981945831,0.,0.,1.613559350,0.,1.254921551,1.169907010,0.,0.,0.,0.,2.909798360,0.,0.,0.,4.036753290,0.,0.,4.490128369,0.,4.427082670,.6247151922,0.,0.,0.,2.122887626,0.,0.,3.733563649,0.,5.029395863,.2882545515,0.,0.,1.241790556,0.,2.644908288,.1234139165,0.,.6438756851,.5047665480e-1},
			{1.414213562,0.,0.,0.,0.,0.,.6957097879e-1,0.,0.,0.,0.,.2209202917e-2,0.,0.,0.,.6061868131e-4,0.,0.,.1543374128e-5,0.,.3753828413e-7,2.224623628,0.,0.,0.,0.,.1730372742,0.,0.,0.,.7507244251e-2,0.,0.,.2611431652e-3,0.,.8052099006e-5,2.258876421,0.,0.,0.,.2400540172,0.,0.,.1320313609e-1,0.,.5562127353e-3,1.981945831,0.,0.,.2670149191,0.,.1778563440e-1,1.613559350,0.,.2632657742,1.254921551},
			{1.414213562,0.,0.,0.,0.,0.,.6957097879e-1,0.,0.,0.,0.,.2209202917e-2,0.,0.,0.,.6061868131e-4,0.,0.,.1543374128e-5,0.,.3753828413e-7,1.169907010,0.,0.,0.,0.,.9099854801e-1,0.,0.,0.,.3947983634e-2,0.,0.,.1373325428e-3,0.,.4234517225e-5,.6247151922,0.,0.,0.,.6638937401e-1,0.,0.,.3651461244e-2,0.,.1538262752e-3,.2882545515,0.,0.,.3883469697e-1,0.,.2586745805e-2,.1234139165,0.,.2013601810e-1,.5047665480e-1},
			{1.414213562,0.,0.,0.,0.,0.,1.169907010,0.,0.,0.,0.,.6247151922,0.,0.,0.,.2882545515,0.,0.,.1234139165,0.,.5047665480e-1,2.224623628,0.,0.,0.,0.,2.909798360,0.,0.,0.,2.122887626,0.,0.,1.241790556,0.,.6438756851,2.258876421,0.,0.,0.,4.036753290,0.,0.,3.733563649,0.,2.644908288,1.981945831,0.,0.,4.490128369,0.,5.029395863,1.613559350,0.,4.427082670,1.254921551},
			{1.414213562,0.,0.,0.,0.,0.,2.224623628,0.,0.,0.,0.,2.258876421,0.,0.,0.,1.981945831,0.,0.,1.613559350,0.,1.254921551,.6957097879e-1,0.,0.,0.,0.,.1730372742,0.,0.,0.,.2400540172,0.,0.,.2670149191,0.,.2632657742,.2209202917e-2,0.,0.,0.,.7507244251e-2,0.,0.,.1320313609e-1,0.,.1778563440e-1,.6061868131e-4,0.,0.,.2611431652e-3,0.,.5562127353e-3,.1543374128e-5,0.,.8052099006e-5,.3753828413e-7},
			{1.414213562,0.,0.,0.,0.,0.,.7092688532,0.,0.,0.,0.,.2296154773,0.,0.,0.,.6423249772e-1,0.,0.,.1667256107e-1,0.,.4134169639e-2,-.1503767417e-1,-0.,-0.,-0.,-0.,-.1192467538e-1,-0.,-0.,-0.,-.5274365494e-2,-0.,-0.,-.1870471581e-2,-0.,-.5879824074e-3,.1032145021e-3,0.,0.,0.,.1118252823e-3,0.,0.,.6270337257e-4,0.,.2693007136e-4,-.6121588826e-6,-0.,-0.,-.8407960796e-6,-0.,-.5709621067e-6,.3368848215e-8,0.,.5603684524e-8,-.1771076019e-10},
			{1.414213562,0.,0.,0.,0.,0.,2.769870437,0.,0.,0.,0.,3.501856062,0.,0.,0.,3.825608374,0.,0.,3.877899792,0.,3.755183165,.7092688532,0.,0.,0.,0.,2.196470374,0.,0.,0.,3.794002148,0.,0.,5.254447192,0.,6.450432215,.2296154773,0.,0.,0.,.9715138718,0.,0.,2.127396643,0.,3.568154409,.6423249772e-1,0.,0.,.3445322647,0.,.9136820088,.1667256107e-1,0.,.1083036558,.4134169639e-2},
			{1.414213562,0.,0.,0.,0.,0.,-.1503767417e-1,-0.,-0.,-0.,-0.,.1032145021e-3,0.,0.,0.,-.6121588826e-6,-0.,-0.,.3368848215e-8,0.,-.1771076019e-10,2.769870437,0.,0.,0.,0.,-.4656880908e-1,-0.,-0.,-0.,.4367054075e-3,0.,0.,-.3283516813e-5,-0.,.2188377514e-7,3.501856062,0.,0.,0.,-.8043912806e-1,-0.,-0.,.9562865187e-3,0.,-.8707719261e-5,3.825608374,0.,0.,-.1114029814,-0.,.1603921848e-2,3.877899792,0.,-.1367598443,3.755183165},
			{1.414213562,0.,0.,0.,0.,0.,-.1503767417e-1,-0.,-0.,-0.,-0.,.1032145021e-3,0.,0.,0.,-.6121588826e-6,-0.,-0.,.3368848215e-8,0.,-.1771076019e-10,.7092688532,0.,0.,0.,0.,-.1192467538e-1,-0.,-0.,-0.,.1118252823e-3,0.,0.,-.8407960796e-6,-0.,.5603684524e-8,.2296154773,0.,0.,0.,-.5274365494e-2,-0.,-0.,.6270337257e-4,0.,-.5709621067e-6,.6423249772e-1,0.,0.,-.1870471581e-2,-0.,.2693007136e-4,.1667256107e-1,0.,-.5879824074e-3,.4134169639e-2},
			{1.414213562,0.,0.,0.,0.,0.,.7092688532,0.,0.,0.,0.,.2296154773,0.,0.,0.,.6423249772e-1,0.,0.,.1667256107e-1,0.,.4134169639e-2,2.769870437,0.,0.,0.,0.,2.196470374,0.,0.,0.,.9715138718,0.,0.,.3445322647,0.,.1083036558,3.501856062,0.,0.,0.,3.794002148,0.,0.,2.127396643,0.,.9136820088,3.825608374,0.,0.,5.254447192,0.,3.568154409,3.877899792,0.,6.450432215,3.755183165},
			{1.414213562,0.,0.,0.,0.,0.,2.769870437,0.,0.,0.,0.,3.501856062,0.,0.,0.,3.825608374,0.,0.,3.877899792,0.,3.755183165,-.1503767417e-1,-0.,-0.,-0.,-0.,-.4656880908e-1,-0.,-0.,-0.,-.8043912806e-1,-0.,-0.,-.1114029814,-0.,-.1367598443,.1032145021e-3,0.,0.,0.,.4367054075e-3,0.,0.,.9562865187e-3,0.,.1603921848e-2,-.6121588826e-6,-0.,-0.,-.3283516813e-5,-0.,-.8707719261e-5,.3368848215e-8,0.,.2188377514e-7,-.1771076019e-10},
			{1.414213562,0.,0.,0.,0.,0.,.6559570585,0.,0.,0.,0.,.1963948976,0.,0.,0.,.5080990387e-1,0.,0.,.1219720719e-1,0.,.2797118711e-2,.1452906103,0.,0.,0.,0.,.1065535599,0.,0.,0.,.4358691597e-1,0.,0.,.1429557258e-1,0.,.4156035932e-2,.9635061189e-2,0.,0.,0.,.9654244188e-2,0.,0.,.5006493970e-2,0.,.1988588036e-2,.5521216640e-3,0.,0.,.7013355149e-3,0.,.4404604930e-3,.2935681623e-4,0.,.4516122448e-4,.1491150883e-5},
			{1.414213562,0.,0.,0.,0.,0.,2.662853947,0.,0.,0.,0.,3.236488548,0.,0.,0.,3.399101902,0.,0.,3.312440982,0.,3.083689087,.6559570585,0.,0.,0.,0.,1.952889915,0.,0.,0.,3.242932261,0.,0.,4.317727482,0.,5.095712344,.1963948976,0.,0.,0.,.7988512884,0.,0.,1.681718474,0.,2.711667007,.5080990387e-1,0.,0.,.2620060705,0.,.6679814775,.1219720719e-1,0.,.7617090096e-1,.2797118711e-2},
			{1.414213562,0.,0.,0.,0.,0.,.1452906103,0.,0.,0.,0.,.9635061189e-2,0.,0.,0.,.5521216640e-3,0.,0.,.2935681623e-4,0.,.1491150883e-5,2.662853947,0.,0.,0.,0.,.4325535702,0.,0.,0.,.3919134934e-1,0.,0.,.2847067532e-2,0.,.1833317338e-3,3.236488548,0.,0.,0.,.7182903227,0.,0.,.8250448766e-1,0.,.7258566080e-2,3.399101902,0.,0.,.9563511098,0.,.1330333825,3.312440982,0.,1.128670157,3.083689087},
			{1.414213562,0.,0.,0.,0.,0.,.1452906103,0.,0.,0.,0.,.9635061189e-2,0.,0.,0.,.5521216640e-3,0.,0.,.2935681623e-4,0.,.1491150883e-5,.6559570585,0.,0.,0.,0.,.1065535599,0.,0.,0.,.9654244188e-2,0.,0.,.7013355149e-3,0.,.4516122448e-4,.1963948976,0.,0.,0.,.4358691597e-1,0.,0.,.5006493970e-2,0.,.4404604930e-3,.5080990387e-1,0.,0.,.1429557258e-1,0.,.1988588036e-2,.1219720719e-1,0.,.4156035932e-2,.2797118711e-2},
			{1.414213562,0.,0.,0.,0.,0.,.6559570585,0.,0.,0.,0.,.1963948976,0.,0.,0.,.5080990387e-1,0.,0.,.1219720719e-1,0.,.2797118711e-2,2.662853947,0.,0.,0.,0.,1.952889915,0.,0.,0.,.7988512884,0.,0.,.2620060705,0.,.7617090096e-1,3.236488548,0.,0.,0.,3.242932261,0.,0.,1.681718474,0.,.6679814775,3.399101902,0.,0.,4.317727482,0.,2.711667007,3.312440982,0.,5.095712344,3.083689087},
			{1.414213562,0.,0.,0.,0.,0.,2.662853947,0.,0.,0.,0.,3.236488548,0.,0.,0.,3.399101902,0.,0.,3.312440982,0.,3.083689087,.1452906103,0.,0.,0.,0.,.4325535702,0.,0.,0.,.7182903227,0.,0.,.9563511098,0.,1.128670157,.9635061189e-2,0.,0.,0.,.3919134934e-1,0.,0.,.8250448766e-1,0.,.1330333825,.5521216640e-3,0.,0.,.2847067532e-2,0.,.7258566080e-2,.2935681623e-4,0.,.1833317338e-3,.1491150883e-5},
			{1.414213562,0.,0.,0.,0.,0.,.2954311110,0.,0.,0.,0.,.3983747795e-1,0.,0.,0.,.4641847056e-2,0.,0.,.5018612597e-3,0.,.5183403142e-4,.4959665215e-1,0.,0.,0.,0.,.1638187454e-1,0.,0.,0.,.3018092809e-2,0.,0.,.4458192555e-3,0.,.5837372111e-4,.1122752692e-2,0.,0.,0.,.5066741239e-3,0.,0.,.1183382006e-3,0.,.2116980724e-4,.2196237131e-4,0.,0.,.1256467012e-4,0.,.3553964112e-5,.3986283756e-6,0.,.2761886212e-6,.6911871564e-8},
			{1.414213562,0.,0.,0.,0.,0.,3.119073853,0.,0.,0.,0.,4.440487963,0.,0.,0.,5.462596453,0.,0.,6.235358013,0.,6.799267692,.2954311110,0.,0.,0.,0.,1.030236404,0.,0.,0.,2.003896641,0.,0.,3.125150808,0.,4.320151396,.3983747795e-1,0.,0.,0.,.1898042300,0.,0.,.4680276955,0.,.8839606376,.4641847056e-2,0.,0.,.2803703724e-1,0.,.8372660778e-1,.5018612597e-3,0.,.3671053173e-2,.5183403142e-4},
			{1.414213562,0.,0.,0.,0.,0.,.4959665215e-1,0.,0.,0.,0.,.1122752692e-2,0.,0.,0.,.2196237131e-4,0.,0.,.3986283756e-6,0.,.6911871564e-8,3.119073853,0.,0.,0.,0.,.1729549620,0.,0.,0.,.5349314793e-2,0.,0.,.1326540524e-3,0.,.2915917367e-5,4.440487963,0.,0.,0.,.3364119790,0.,0.,.1319057787e-1,0.,.3961429206e-3,5.462596453,0.,0.,.5246469034,0.,.2491295222e-1,6.235358013,0.,.7252622967,6.799267692},
			{1.414213562,0.,0.,0.,0.,0.,.4959665215e-1,0.,0.,0.,0.,.1122752692e-2,0.,0.,0.,.2196237131e-4,0.,0.,.3986283756e-6,0.,.6911871564e-8,.2954311110,0.,0.,0.,0.,.1638187454e-1,0.,0.,0.,.5066741239e-3,0.,0.,.1256467012e-4,0.,.2761886212e-6,.3983747795e-1,0.,0.,0.,.3018092809e-2,0.,0.,.1183382006e-3,0.,.3553964112e-5,.4641847056e-2,0.,0.,.4458192555e-3,0.,.2116980724e-4,.5018612597e-3,0.,.5837372111e-4,.5183403142e-4},
			{1.414213562,0.,0.,0.,0.,0.,.2954311110,0.,0.,0.,0.,.3983747795e-1,0.,0.,0.,.4641847056e-2,0.,0.,.5018612597e-3,0.,.5183403142e-4,3.119073853,0.,0.,0.,0.,1.030236404,0.,0.,0.,.1898042300,0.,0.,.2803703724e-1,0.,.3671053173e-2,4.440487963,0.,0.,0.,2.003896641,0.,0.,.4680276955,0.,.8372660778e-1,5.462596453,0.,0.,3.125150808,0.,.8839606376,6.235358013,0.,4.320151396,6.799267692},
			{1.414213562,0.,0.,0.,0.,0.,3.119073853,0.,0.,0.,0.,4.440487963,0.,0.,0.,5.462596453,0.,0.,6.235358013,0.,6.799267692,.4959665215e-1,0.,0.,0.,0.,.1729549620,0.,0.,0.,.3364119790,0.,0.,.5246469034,0.,.7252622967,.1122752692e-2,0.,0.,0.,.5349314793e-2,0.,0.,.1319057787e-1,0.,.2491295222e-1,.2196237131e-4,0.,0.,.1326540524e-3,0.,.3961429206e-3,.3986283756e-6,0.,.2915917367e-5,.6911871564e-8}
		};
		for (int j = 0; j < 52; j++)
		{
			std::cout << timestep_dofs_to_top_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_top_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_bottom_fluxes_to_dofs:" << std::endl;
		static const T timestep_bottom_fluxes_to_dofs[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-.1148215401,-.1569113387e-1,-.1569113387e-1,-.1569113387e-1,-.1021689440,-.1021689440,-.1021689440,-.6587041484e-1,-.6587041484e-1,-.6587041484e-1,-.1032031823,-.1032031823,-.1032031823,-.7348534591e-1,-.7348534591e-1,-.7348534591e-1,-.3478299668e-1,-.3478299668e-1,-.3478299668e-1,-.8775204783e-2,-.8775204783e-2,-.8775204783e-2,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{-.1753927996,-.2396860291e-1,-.2396860291e-1,-.2396860291e-1,-.1560656399,-.1560656399,-.1560656399,-.1006187207,-.1006187207,-.1006187207,-.1576454650,-.1576454650,-.1576454650,-.1122507200,-.1122507200,-.1122507200,-.5313190506e-1,-.5313190506e-1,-.5313190506e-1,-.1340434672e-1,-.1340434672e-1,-.1340434672e-1,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{-.2198665100,-.3004623384e-1,-.3004623384e-1,-.3004623384e-1,-.1956386332,-.1956386332,-.1956386332,-.1261322415,-.1261322415,-.1261322415,-.1976190488,-.1976190488,-.1976190488,-.1407137244,-.1407137244,-.1407137244,-.6660436781e-1,-.6660436781e-1,-.6660436781e-1,-.1680323787e-1,-.1680323787e-1,-.1680323787e-1,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1},
			{-.5412739307e-1,-.1103717978e-1,-.1103717978e-1,-.1162547668e-3,-.5974166429e-1,-.5974166429e-1,-.2500537776e-1,-.4382549273e-1,-.4382549273e-1,-.5503848602e-2,-.3510914219e-1,-.3510914219e-1,-.7573305578e-1,-.1537719318e-1,-.1537719318e-1,-.7316958648e-1,-.3712176777e-2,-.3712176777e-2,-.4176623208e-1,-.2059615826e-3,-.2059615826e-3,-.1199809046e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1},
			{-.9375139486e-1,-.1911695615e-1,-.1911695615e-1,-.2013591627e-3,-.1034755978,-.1034755978,-.4331058473e-1,-.7590798006e-1,-.7590798006e-1,-.9532945415e-2,-.6081081808e-1,-.6081081808e-1,-.1311735004,-.2663407986e-1,-.2663407986e-1,-.1267334413,-.6429678784e-2,-.6429678784e-2,-.7234123601e-1,-.3567359253e-3,-.3567359253e-3,-.2078130226e-1,-.5831436180e-1,-.1179575144,-.2036700860e-1,-.2036700860e-1,-.5831436180e-1,-.1179575144,-.3099954510e-1,-.5894683929e-1,-.1843453093e-2,-.1843453093e-2,-.3099954510e-1,-.5894683929e-1,-.2931476389e-2,-.1144814092e-1,.6215215367e-4,.6215215367e-4,-.2931476389e-2,-.1144814092e-1,-.2168320548e-1,-.8802284926e-1,-.4802701816e-2,-.4802701816e-2,-.2168320548e-1,-.8802284926e-1,-.3505184516e-2,-.3700669622e-1,-.5884465473e-3,-.5884465473e-3,-.3505184516e-2,-.3700669622e-1},
			{-.1210325303,-.2467988427e-1,-.2467988427e-1,-.2599535613e-3,-.1335864224,-.1335864224,-.5591372447e-1,-.9799678087e-1,-.9799678087e-1,-.1230697961e-1,-.7850642855e-1,-.7850642855e-1,-.1693442608,-.3438444924e-1,-.3438444924e-1,-.1636121692,-.8300679616e-2,-.8300679616e-2,-.9339213407e-1,-.4605440992e-3,-.4605440992e-3,-.2682854586e-1,-.7528351732e-1,-.1522824962,-.2629369504e-1,-.2629369504e-1,-.7528351732e-1,-.1522824962,-.4002024064e-1,-.7610004229e-1,-.2379887709e-2,-.2379887709e-2,-.4002024064e-1,-.7610004229e-1,-.3784519743e-2,-.1477948638e-1,.8023808537e-4,.8023808537e-4,-.3784519743e-2,-.1477948638e-1,-.2799289790e-1,-.1136370098,-.6200261386e-2,-.6200261386e-2,-.2799289790e-1,-.1136370098,-.4525173753e-2,-.4777543936e-1,-.7596812259e-3,-.7596812259e-3,-.4525173753e-2,-.4777543936e-1},
			{-.1432076211,-.2920163288e-1,-.2920163288e-1,-.3075812017e-3,-.1580615866,-.1580615866,-.6615801099e-1,-.1159513548,-.1159513548,-.1456181465e-1,-.9289005896e-1,-.9289005896e-1,-.2003708316,-.4068422900e-1,-.4068422900e-1,-.1935885293,-.9821496573e-2,-.9821496573e-2,-.1105030633,-.5449231269e-3,-.5449231269e-3,-.3174396355e-1,-.8907665897e-1,-.1801830794,-.3111111953e-1,-.3111111953e-1,-.8907665897e-1,-.1801830794,-.4735258731e-1,-.9004278434e-1,-.2815921112e-2,-.2815921112e-2,-.4735258731e-1,-.9004278434e-1,-.4477904149e-2,-.1748732412e-1,.9493898295e-4,.9493898295e-4,-.4477904149e-2,-.1748732412e-1,-.3312164346e-1,-.1344571232,-.7336248207e-2,-.7336248207e-2,-.3312164346e-1,-.1344571232,-.5354257791e-2,-.5652866219e-1,-.8988669485e-3,-.8988669485e-3,-.5354257791e-2,-.5652866219e-1},
			{-.1623821792,-.3311153935e-1,-.3311153935e-1,-.3487643005e-3,-.1792249928,-.1792249928,-.7501613328e-1,-.1314764782,-.1314764782,-.1651154581e-1,-.1053274266,-.1053274266,-.2271991674,-.4613157954e-1,-.4613157954e-1,-.2195087594,-.1113653033e-1,-.1113653033e-1,-.1252986963,-.6178847477e-3,-.6178847477e-3,-.3599427137e-1,-.1010034375,-.2043084081,-.3527669370e-1,-.3527669370e-1,-.1010034375,-.2043084081,-.5369278714e-1,-.1020989206,-.3192954419e-2,-.3192954419e-2,-.5369278714e-1,-.1020989206,-.5077466048e-2,-.1982876173e-1,.1076506880e-3,.1076506880e-3,-.5077466048e-2,-.1982876173e-1,-.3755641357e-1,-.1524600472,-.8318523561e-2,-.8318523561e-2,-.3755641357e-1,-.1524600472,-.6071157673e-2,-.6409747809e-1,-.1019219318e-2,-.1019219318e-2,-.6071157673e-2,-.6409747809e-1},
			{-.2852764098e-1,-.8679943781e-2,-.8679943781e-2,-.9629909754e-6,-.3905626741e-1,-.3905626741e-1,-.6842313070e-2,-.3260004675e-1,-.3260004675e-1,-.5141591464e-3,-.1335372273e-1,-.1335372273e-1,-.6213451013e-1,-.3597562967e-2,-.3597562967e-2,-.8145457175e-1,-.4429404170e-3,-.4429404170e-3,-.5607103930e-1,-.5404682501e-5,-.5404682501e-5,-.1834095772e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1},
			{-.4941132360e-1,-.1503410363e-1,-.1503410363e-1,-.1667949296e-5,-.6764743950e-1,-.6764743950e-1,-.1185123388e-1,-.5646493729e-1,-.5646493729e-1,-.8905497646e-3,-.2312932623e-1,-.2312932623e-1,-.1076201284,-.6231161842e-2,-.6231161842e-2,-.1410834568,-.7671953069e-3,-.7671953069e-3,-.9711788894e-1,-.9361184690e-5,-.9361184690e-5,-.3176747063e-1,-.2734334585e-1,-.1118798374,-.3335455075e-2,-.3335455075e-2,-.2734334585e-1,-.1118798374,-.1655335563e-1,-.5985445076e-1,-.5853822990e-4,-.5853822990e-4,-.1655335563e-1,-.5985445076e-1,-.9490228526e-3,-.1447350792e-1,-.4265954647e-6,-.4265954647e-6,-.9490228526e-3,-.1447350792e-1,-.6491996486e-2,-.1069848175,-.3184949515e-3,-.3184949515e-3,-.6491996486e-2,-.1069848175,-.4726574345e-3,-.5268480225e-1,-.1332105931e-4,-.1332105931e-4,-.4726574345e-3,-.5268480225e-1},
			{-.6378974448e-1,-.1940894434e-1,-.1940894434e-1,-.2153313281e-5,-.8733246888e-1,-.8733246888e-1,-.1529987715e-1,-.7289592064e-1,-.7289592064e-1,-.1149694802e-2,-.2985983179e-1,-.2985983179e-1,-.1389369884,-.8044395345e-2,-.8044395345e-2,-.1821379596,-.9904448821e-3,-.9904448821e-3,-.1253786555,-.1208523748e-4,-.1208523748e-4,-.4101162825e-1,-.3530010772e-1,-.1444362491,-.4306053985e-2,-.4306053985e-2,-.3530010772e-1,-.1444362491,-.2137029023e-1,-.7727176368e-1,-.7557252986e-4,-.7557252986e-4,-.2137029023e-1,-.7727176368e-1,-.1225183234e-2,-.1868521838e-1,-.5507323768e-6,-.5507323768e-6,-.1225183234e-2,-.1868521838e-1,-.8381131425e-2,-.1381168056,-.4111752145e-3,-.4111752145e-3,-.8381131425e-2,-.1381168056,-.6101981241e-3,-.6801578726e-1,-.1719741363e-4,-.1719741363e-4,-.6101981241e-3,-.6801578726e-1},
			{-.7547704355e-1,-.2296497265e-1,-.2296497265e-1,-.2547834636e-5,-.1033331707,-.1033331707,-.1810305878e-1,-.8625161644e-1,-.8625161644e-1,-.1360337236e-2,-.3533062943e-1,-.3533062943e-1,-.1643924617,-.9518256939e-2,-.9518256939e-2,-.2155085401,-.1171910189e-2,-.1171910189e-2,-.1483500258,-.1429944582e-4,-.1429944582e-4,-.4852561294e-1,-.4176765072e-1,-.1708992746,-.5094991785e-2,-.5094991785e-2,-.4176765072e-1,-.1708992746,-.2528566840e-1,-.9142918380e-1,-.8941862322e-4,-.8941862322e-4,-.2528566840e-1,-.9142918380e-1,-.1449656353e-2,-.2210864854e-1,-.6516353360e-6,-.6516353360e-6,-.1449656353e-2,-.2210864854e-1,-.9916688441e-2,-.1634220082,-.4865090747e-3,-.4865090747e-3,-.9916688441e-2,-.1634220082,-.7219961572e-3,-.8047736478e-1,-.2034825423e-4,-.2034825423e-4,-.7219961572e-3,-.8047736478e-1},
			{-.1299205923e-1,-.5898466126e-2,-.5898466126e-2,-.6892821735e-8,-.2206315093e-1,-.2206315093e-1,-.1617841028e-2,-.2095429338e-1,-.2095429338e-1,-.4150419878e-4,-.4388821038e-2,-.4388821038e-2,-.4404977136e-1,-.7272824702e-3,-.7272824702e-3,-.7835445799e-1,-.4566940490e-4,-.4566940490e-4,-.6504520647e-1,-.1225511905e-6,-.1225511905e-6,-.2422675624e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1},
			{-.2250290667e-1,-.1021644301e-1,-.1021644301e-1,-.1193871745e-7,-.3821449838e-1,-.3821449838e-1,-.2802182859e-2,-.3629390077e-1,-.3629390077e-1,-.7188738101e-4,-.7601661022e-2,-.7601661022e-2,-.7629644205e-1,-.1259690190e-2,-.1259690190e-2,-.1357139022,-.7910172963e-4,-.7910172963e-4,-.1126616024,-.2122648885e-6,-.2122648885e-6,-.4196197271e-1,-.1107876030e-1,-.9169410646e-1,-.4720047033e-3,-.4720047033e-3,-.1107876030e-1,-.9169410646e-1,-.7638008748e-2,-.5251649806e-1,-.1606240095e-5,-.1606240095e-5,-.7638008748e-2,-.5251649806e-1,-.2654790910e-3,-.1581160736e-1,.2530111540e-8,.2530111540e-8,-.2654790910e-3,-.1581160736e-1,-.1679563581e-2,-.1123601370,-.1825084025e-4,-.1825084025e-4,-.1679563581e-2,-.1123601370,-.5507385590e-4,-.6481175407e-1,-.2605756842e-6,-.2605756842e-6,-.5507385590e-4,-.6481175407e-1},
			{-.2905112760e-1,-.1318937122e-1,-.1318937122e-1,-.1541281795e-7,-.4933470527e-1,-.4933470527e-1,-.3617602516e-2,-.4685522442e-1,-.4685522442e-1,-.9280620982e-4,-.9813702180e-2,-.9813702180e-2,-.9849828315e-1,-.1626253042e-2,-.1626253042e-2,-.1752058944,-.1021198938e-3,-.1021198938e-3,-.1454455033,-.2740327927e-6,-.2740327927e-6,-.5417267383e-1,-.1430261805e-1,-.1183765824,-.6093554517e-3,-.6093554517e-3,-.1430261805e-1,-.1183765824,-.9860626894e-2,-.6779850746e-1,-.2073647046e-5,-.2073647046e-5,-.9860626894e-2,-.6779850746e-1,-.3427320328e-3,-.2041269733e-1,.3266359953e-8,.3266359953e-8,-.3427320328e-3,-.2041269733e-1,-.2168307259e-2,-.1450563132,-.2356173345e-4,-.2356173345e-4,-.2168307259e-2,-.1450563132,-.7110004225e-4,-.8367161471e-1,-.3364017618e-6,-.3364017618e-6,-.7110004225e-4,-.8367161471e-1},
			{-.5490147066e-2,-.3719247860e-2,-.3719247860e-2,-.4577893991e-10,-.1156480173e-1,-.1156480173e-1,-.3549462592e-3,-.1249745760e-1,-.1249745760e-1,-.3108711060e-5,-.1338404058e-2,-.1338404058e-2,-.2897665788e-1,-.1364242863e-3,-.1364242863e-3,-.6993680883e-1,-.4369173091e-5,-.4369173091e-5,-.7001415346e-1,-.2578450140e-8,-.2578450140e-8,-.2969356790e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1},
			{-.9509213660e-2,-.6441926259e-2,-.6441926259e-2,-.7929144985e-10,-.2003082417e-1,-.2003082417e-1,-.6147849549e-3,-.2164623154e-1,-.2164623154e-1,-.5384445502e-5,-.2318183830e-2,-.2318183830e-2,-.5018904368e-1,-.2362937952e-3,-.2362937952e-3,-.1211341062,-.7567629781e-5,-.7567629781e-5,-.1212680711,-.4466006648e-8,-.4466006648e-8,-.5143076827e-1,-.4165091360e-2,-.6973083633e-1,-.6197712293e-4,-.6197712293e-4,-.4165091360e-2,-.6973083633e-1,-.3270153304e-2,-.4275519802e-1,-.4089546907e-7,-.4089546907e-7,-.3270153304e-2,-.4275519802e-1,-.6890929850e-4,-.1602773282e-1,-.1392377369e-10,-.1392377369e-10,-.6890929850e-4,-.1602773282e-1,-.4031888160e-3,-.1094954884,-.9704139476e-6,-.9704139476e-6,-.4031888160e-3,-.1094954884,-.5954404436e-5,-.7398029375e-1,-.4729583171e-8,-.4729583171e-8,-.5954404436e-5,-.7398029375e-1},
			{-.2216299428e-2,-.2240317772e-2,-.2240317772e-2,-.2904510069e-12,-.5790916683e-2,-.5790916683e-2,-.7439227355e-4,-.7120472184e-2,-.7120472184e-2,-.2224370628e-6,-.3899106406e-3,-.3899106406e-3,-.1820922229e-1,-.2444661488e-4,-.2444661488e-4,-.5963296216e-1,-.3993113157e-6,-.3993113157e-6,-.7199376027e-1,-.5182489262e-10,-.5182489262e-10,-.3476706016e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1},
			{-.5412739307e-1,-.1162547668e-3,-.1103717978e-1,-.1103717978e-1,-.2500537776e-1,-.5974166429e-1,-.5974166429e-1,-.5503848602e-2,-.4382549273e-1,-.4382549273e-1,-.7573305578e-1,-.3510914219e-1,-.3510914219e-1,-.7316958648e-1,-.1537719318e-1,-.1537719318e-1,-.4176623208e-1,-.3712176777e-2,-.3712176777e-2,-.1199809046e-1,-.2059615826e-3,-.2059615826e-3,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3},
			{-.9375139486e-1,-.2013591627e-3,-.1911695615e-1,-.1911695615e-1,-.4331058473e-1,-.1034755978,-.1034755978,-.9532945415e-2,-.7590798006e-1,-.7590798006e-1,-.1311735004,-.6081081808e-1,-.6081081808e-1,-.1267334413,-.2663407986e-1,-.2663407986e-1,-.7234123601e-1,-.6429678784e-2,-.6429678784e-2,-.2078130226e-1,-.3567359253e-3,-.3567359253e-3,-.2036700860e-1,-.5831436180e-1,-.1179575144,-.5831436180e-1,-.1179575144,-.2036700860e-1,-.1843453093e-2,-.3099954510e-1,-.5894683929e-1,-.3099954510e-1,-.5894683929e-1,-.1843453093e-2,.6215215367e-4,-.2931476389e-2,-.1144814092e-1,-.2931476389e-2,-.1144814092e-1,.6215215367e-4,-.4802701816e-2,-.2168320548e-1,-.8802284926e-1,-.2168320548e-1,-.8802284926e-1,-.4802701816e-2,-.5884465473e-3,-.3505184516e-2,-.3700669622e-1,-.3505184516e-2,-.3700669622e-1,-.5884465473e-3},
			{-.1210325303,-.2599535613e-3,-.2467988427e-1,-.2467988427e-1,-.5591372447e-1,-.1335864224,-.1335864224,-.1230697961e-1,-.9799678087e-1,-.9799678087e-1,-.1693442608,-.7850642855e-1,-.7850642855e-1,-.1636121692,-.3438444924e-1,-.3438444924e-1,-.9339213407e-1,-.8300679616e-2,-.8300679616e-2,-.2682854586e-1,-.4605440992e-3,-.4605440992e-3,-.2629369504e-1,-.7528351732e-1,-.1522824962,-.7528351732e-1,-.1522824962,-.2629369504e-1,-.2379887709e-2,-.4002024064e-1,-.7610004229e-1,-.4002024064e-1,-.7610004229e-1,-.2379887709e-2,.8023808537e-4,-.3784519743e-2,-.1477948638e-1,-.3784519743e-2,-.1477948638e-1,.8023808537e-4,-.6200261386e-2,-.2799289790e-1,-.1136370098,-.2799289790e-1,-.1136370098,-.6200261386e-2,-.7596812259e-3,-.4525173753e-2,-.4777543936e-1,-.4525173753e-2,-.4777543936e-1,-.7596812259e-3},
			{-.1432076211,-.3075812017e-3,-.2920163288e-1,-.2920163288e-1,-.6615801099e-1,-.1580615866,-.1580615866,-.1456181465e-1,-.1159513548,-.1159513548,-.2003708316,-.9289005896e-1,-.9289005896e-1,-.1935885293,-.4068422900e-1,-.4068422900e-1,-.1105030633,-.9821496573e-2,-.9821496573e-2,-.3174396355e-1,-.5449231269e-3,-.5449231269e-3,-.3111111953e-1,-.8907665897e-1,-.1801830794,-.8907665897e-1,-.1801830794,-.3111111953e-1,-.2815921112e-2,-.4735258731e-1,-.9004278434e-1,-.4735258731e-1,-.9004278434e-1,-.2815921112e-2,.9493898295e-4,-.4477904149e-2,-.1748732412e-1,-.4477904149e-2,-.1748732412e-1,.9493898295e-4,-.7336248207e-2,-.3312164346e-1,-.1344571232,-.3312164346e-1,-.1344571232,-.7336248207e-2,-.8988669485e-3,-.5354257791e-2,-.5652866219e-1,-.5354257791e-2,-.5652866219e-1,-.8988669485e-3},
			{-.1623821792,-.3487643005e-3,-.3311153935e-1,-.3311153935e-1,-.7501613328e-1,-.1792249928,-.1792249928,-.1651154581e-1,-.1314764782,-.1314764782,-.2271991674,-.1053274266,-.1053274266,-.2195087594,-.4613157954e-1,-.4613157954e-1,-.1252986963,-.1113653033e-1,-.1113653033e-1,-.3599427137e-1,-.6178847477e-3,-.6178847477e-3,-.3527669370e-1,-.1010034375,-.2043084081,-.1010034375,-.2043084081,-.3527669370e-1,-.3192954419e-2,-.5369278714e-1,-.1020989206,-.5369278714e-1,-.1020989206,-.3192954419e-2,.1076506880e-3,-.5077466048e-2,-.1982876173e-1,-.5077466048e-2,-.1982876173e-1,.1076506880e-3,-.8318523561e-2,-.3755641357e-1,-.1524600472,-.3755641357e-1,-.1524600472,-.8318523561e-2,-.1019219318e-2,-.6071157673e-2,-.6409747809e-1,-.6071157673e-2,-.6409747809e-1,-.1019219318e-2},
			{-.6987816394e-1,-.2239469694e-3,-.2126143325e-1,-.2239469694e-3,-.4004261797e-1,-.9566792642e-1,-.4004261797e-1,-.1002844321e-1,-.7985348011e-1,-.1002844321e-1,-.7055750928e-1,-.3270980684e-1,-.7055750928e-1,-.4193122589e-1,-.8812193584e-2,-.4193122589e-1,-.1220724282e-1,-.1084978008e-2,-.1220724282e-1,-.7712083501e-3,-.1323871434e-4,-.7712083501e-3,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,.2845518738e-4,-.5241314676e-2,.1111245502e-3,.2845518738e-4,-.5241314676e-2,.1111245502e-3,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2},
			{-.1210325303,-.3878875292e-3,-.3682588264e-1,-.3878875292e-3,-.6935584881e-1,-.1657017093,-.6935584881e-1,-.1736977316e-1,-.1383102847,-.1736977316e-1,-.1222091909,-.5665504737e-1,-.1222091909,-.7262701368e-1,-.1526316702e-1,-.7262701368e-1,-.2114356478e-1,-.1879237035e-2,-.2114356478e-1,-.1335772046e-2,-.2293012587e-4,-.1335772046e-2,-.2339262725e-1,-.1354806795,-.4731829487e-1,-.2339262725e-1,-.1354806795,-.4731829487e-1,-.2411228903e-2,-.7710221821e-1,-.4585045432e-2,-.2411228903e-2,-.7710221821e-1,-.4585045432e-2,.4928583028e-4,-.9078223319e-2,.1924733670e-3,.4928583028e-4,-.9078223319e-2,.1924733670e-3,-.3522216439e-2,-.6455439842e-1,-.1429841759e-1,-.3522216439e-2,-.6455439842e-1,-.1429841759e-1,-.1943650851e-3,-.1222338663e-1,-.2052048795e-2,-.1943650851e-3,-.1222338663e-1,-.2052048795e-2},
			{-.1562523248,-.5007606471e-3,-.4754201006e-1,-.5007606471e-3,-.8953801576e-1,-.2139199867,-.8953801576e-1,-.2242428073e-1,-.1785578098,-.2242428073e-1,-.1577713871,-.7314135164e-1,-.1577713871,-.9376107150e-1,-.1970466389e-1,-.9376107150e-1,-.2729622476e-1,-.2426084580e-2,-.2729622476e-1,-.1724474296e-2,-.2960266522e-4,-.1724474296e-2,-.3019975193e-1,-.1749048052,-.6108765601e-1,-.3019975193e-1,-.1749048052,-.6108765601e-1,-.3112883129e-2,-.9953853571e-1,-.5919268200e-2,-.3112883129e-2,-.9953853571e-1,-.5919268200e-2,.6362773332e-4,-.1171993591e-1,.2484820485e-3,.6362773332e-4,-.1171993591e-1,.2484820485e-3,-.4547161871e-2,-.8333937002e-1,-.1845917774e-1,-.4547161871e-2,-.8333937002e-1,-.1845917774e-1,-.2509242459e-3,-.1578032429e-1,-.2649183603e-2,-.2509242459e-3,-.1578032429e-1,-.2649183603e-2},
			{-.1848802439,-.5925079880e-3,-.5625246492e-1,-.5925079880e-3,-.1059428090,-.2531135418,-.1059428090,-.2653276678e-1,-.2112724497,-.2653276678e-1,-.1866776227,-.8654201436e-1,-.1866776227,-.1109395959,-.2331487273e-1,-.1109395959,-.3229732869e-1,-.2870581987e-2,-.3229732869e-1,-.2040425504e-2,-.3502634584e-4,-.2040425504e-2,-.3573282836e-1,-.2069501564,-.7227988933e-1,-.3573282836e-1,-.2069501564,-.7227988933e-1,-.3683212988e-2,-.1177755837,-.7003772586e-2,-.3683212988e-2,-.1177755837,-.7003772586e-2,.7528534932e-4,-.1386721518e-1,.2940079245e-3,.7528534932e-4,-.1386721518e-1,.2940079245e-3,-.5380274482e-2,-.9860847240e-1,-.2184119364e-1,-.5380274482e-2,-.9860847240e-1,-.2184119364e-1,-.2968975716e-3,-.1867153150e-1,-.3134556311e-2,-.2968975716e-3,-.1867153150e-1,-.3134556311e-2},
			{-.5031802903e-1,-.2406231299e-3,-.2284466108e-1,-.2534486742e-5,-.3576590913e-1,-.8545021612e-1,-.1497012310e-1,-.1019197433e-1,-.8115562931e-1,-.1279964701e-2,-.3666559727e-1,-.1699783079e-1,-.7909044629e-1,-.1340300809e-1,-.2816752894e-2,-.6377578460e-1,-.1990066688e-2,-.1768768448e-3,-.2239052508e-1,-.2764961419e-4,-.4746387201e-6,-.1610701220e-2,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2},
			{-.8715338278e-1,-.4167714865e-3,-.3956811365e-1,-.4389859808e-5,-.6194837174e-1,-.1480041159,-.2592901380e-1,-.1765301736e-1,-.1405656732,-.2216963893e-2,-.6350667735e-1,-.2944110655e-1,-.1369886713,-.2321469098e-1,-.4878759123e-2,-.1104628991,-.3446896614e-2,-.3063596816e-3,-.3878152703e-1,-.4789053657e-4,-.8220983782e-6,-.2789816347e-2,-.1498609619e-1,-.1755646063,-.1058742437e-1,-.5234078544e-2,-.8679343590e-1,-.6131809964e-1,-.1759148700e-2,-.1069636429,-.1989227818e-3,-.1046114740e-3,-.5625109532e-1,-.6360817804e-2,.2179945987e-4,-.1568097578e-1,-.1804945451e-5,-.4621846472e-6,-.4015360274e-2,.3324626528e-3,-.1440801715e-2,-.1071978198,-.1295502609e-2,-.3191290614e-3,-.2640669295e-1,-.2374368331e-1,-.3580859223e-4,-.2377551727e-1,-.6346770766e-4,-.6011507348e-5,-.2251959335e-2,-.3991407866e-2},
			{-.1125145334,-.5380496754e-3,-.5108221507e-1,-.5667284643e-5,-.7997500406e-1,-.1910724920,-.3347421288e-1,-.2278994742e-1,-.1814695038,-.2862088080e-2,-.8198676790e-1,-.3800830512e-1,-.1768516141,-.2997003719e-1,-.6298450945e-2,-.1426069896,-.4449924395e-2,-.3955086483e-3,-.5006673611e-1,-.6182641688e-4,-.1061324443e-5,-.3601637417e-2,-.1934696698e-1,-.2266529321,-.1366830609e-1,-.6757166343e-2,-.1120498440,-.7916132626e-1,-.2271051205e-2,-.1380894693,-.2568082070e-3,-.1350528323e-3,-.7261985182e-1,-.8211780472e-2,.2814298168e-4,-.2024405269e-1,-.2330174558e-5,-.5966778137e-6,-.5183807822e-2,.4292074393e-3,-.1860067016e-2,-.1383917903,-.1672486677e-2,-.4119938467e-3,-.3409089400e-1,-.3065296333e-1,-.4622869379e-4,-.3069406081e-1,-.8193645830e-4,-.7760822615e-5,-.2907267000e-2,-.5152885397e-2},
			{-.2905112760e-1,-.2072940277e-3,-.1968040980e-1,-.2299810724e-7,-.2561375849e-1,-.6119517866e-1,-.4487304244e-2,-.8305013821e-2,-.6613032970e-1,-.1309844384e-3,-.1527676939e-1,-.7082168586e-2,-.7108239413e-1,-.3434980188e-2,-.7218894686e-3,-.7777343793e-1,-.2601206999e-3,-.2311949087e-4,-.3292821660e-1,-.7948106270e-6,-.1364387568e-7,-.2697214518e-2,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,.4463401949e-5,-.1253839410e-1,.7835272927e-8,.2006344762e-8,-.8221381159e-3,.2658347174e-3,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2},
			{-.5031802903e-1,-.3590437882e-3,-.3408746970e-1,-.3983389024e-7,-.4436433109e-1,-.1059931586,-.7772238941e-2,-.1438470590e-1,-.1145410910,-.2268717024e-3,-.2646014076e-1,-.1226667582e-1,-.1231183182,-.5949560210e-2,-.1250349237e-2,-.1347075460,-.4505422684e-3,-.4004413284e-4,-.5703334418e-1,-.1376652389e-5,-.2363188589e-7,-.4671712585e-2,-.7697598727e-2,-.1824121677,-.1899366584e-2,-.9389851147e-3,-.4458139289e-1,-.6370969475e-1,-.1029020197e-2,-.1189769237,-.6919626319e-5,-.3638961323e-5,-.3290427534e-1,-.7075212788e-2,.7730838953e-5,-.2171713563e-1,.1357109081e-7,.3475091067e-8,-.1423984988e-2,.4604392371e-3,-.4725520274e-3,-.1427260688,-.9411218238e-4,-.2318322805e-4,-.8660828315e-2,-.3161298039e-1,-.5289486092e-5,-.3707879714e-1,-.1573892910e-5,-.1490753193e-6,-.3326494240e-3,-.6224747955e-2},
			{-.1486738854e-1,-.1582954870e-3,-.1502850849e-1,-.1849807308e-9,-.1625959323e-1,-.3884665007e-1,-.1192279250e-2,-.5998666568e-2,-.4776557949e-1,-.1188156743e-4,-.5642045787e-2,-.2615600094e-2,-.5662815244e-1,-.7803303852e-3,-.1639928781e-3,-.8406962479e-1,-.3013800652e-4,-.2678661737e-5,-.4292442302e-1,-.2025212407e-7,-.3476519485e-9,-.4003578188e-2,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2},
			{-.2852764098e-1,-.9629909754e-6,-.8679943781e-2,-.8679943781e-2,-.6842313070e-2,-.3905626741e-1,-.3905626741e-1,-.5141591464e-3,-.3260004675e-1,-.3260004675e-1,-.6213451013e-1,-.1335372273e-1,-.1335372273e-1,-.8145457175e-1,-.3597562967e-2,-.3597562967e-2,-.5607103930e-1,-.4429404170e-3,-.4429404170e-3,-.1834095772e-1,-.5404682501e-5,-.5404682501e-5,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5},
			{-.4941132360e-1,-.1667949296e-5,-.1503410363e-1,-.1503410363e-1,-.1185123388e-1,-.6764743950e-1,-.6764743950e-1,-.8905497646e-3,-.5646493729e-1,-.5646493729e-1,-.1076201284,-.2312932623e-1,-.2312932623e-1,-.1410834568,-.6231161842e-2,-.6231161842e-2,-.9711788894e-1,-.7671953069e-3,-.7671953069e-3,-.3176747063e-1,-.9361184690e-5,-.9361184690e-5,-.3335455075e-2,-.2734334585e-1,-.1118798374,-.2734334585e-1,-.1118798374,-.3335455075e-2,-.5853822990e-4,-.1655335563e-1,-.5985445076e-1,-.1655335563e-1,-.5985445076e-1,-.5853822990e-4,-.4265954647e-6,-.9490228526e-3,-.1447350792e-1,-.9490228526e-3,-.1447350792e-1,-.4265954647e-6,-.3184949515e-3,-.6491996486e-2,-.1069848175,-.6491996486e-2,-.1069848175,-.3184949515e-3,-.1332105931e-4,-.4726574345e-3,-.5268480225e-1,-.4726574345e-3,-.5268480225e-1,-.1332105931e-4},
			{-.6378974448e-1,-.2153313281e-5,-.1940894434e-1,-.1940894434e-1,-.1529987715e-1,-.8733246888e-1,-.8733246888e-1,-.1149694802e-2,-.7289592064e-1,-.7289592064e-1,-.1389369884,-.2985983179e-1,-.2985983179e-1,-.1821379596,-.8044395345e-2,-.8044395345e-2,-.1253786555,-.9904448821e-3,-.9904448821e-3,-.4101162825e-1,-.1208523748e-4,-.1208523748e-4,-.4306053985e-2,-.3530010772e-1,-.1444362491,-.3530010772e-1,-.1444362491,-.4306053985e-2,-.7557252986e-4,-.2137029023e-1,-.7727176368e-1,-.2137029023e-1,-.7727176368e-1,-.7557252986e-4,-.5507323768e-6,-.1225183234e-2,-.1868521838e-1,-.1225183234e-2,-.1868521838e-1,-.5507323768e-6,-.4111752145e-3,-.8381131425e-2,-.1381168056,-.8381131425e-2,-.1381168056,-.4111752145e-3,-.1719741363e-4,-.6101981241e-3,-.6801578726e-1,-.6101981241e-3,-.6801578726e-1,-.1719741363e-4},
			{-.7547704355e-1,-.2547834636e-5,-.2296497265e-1,-.2296497265e-1,-.1810305878e-1,-.1033331707,-.1033331707,-.1360337236e-2,-.8625161644e-1,-.8625161644e-1,-.1643924617,-.3533062943e-1,-.3533062943e-1,-.2155085401,-.9518256939e-2,-.9518256939e-2,-.1483500258,-.1171910189e-2,-.1171910189e-2,-.4852561294e-1,-.1429944582e-4,-.1429944582e-4,-.5094991785e-2,-.4176765072e-1,-.1708992746,-.4176765072e-1,-.1708992746,-.5094991785e-2,-.8941862322e-4,-.2528566840e-1,-.9142918380e-1,-.2528566840e-1,-.9142918380e-1,-.8941862322e-4,-.6516353360e-6,-.1449656353e-2,-.2210864854e-1,-.1449656353e-2,-.2210864854e-1,-.6516353360e-6,-.4865090747e-3,-.9916688441e-2,-.1634220082,-.9916688441e-2,-.1634220082,-.4865090747e-3,-.2034825423e-4,-.7219961572e-3,-.8047736478e-1,-.7219961572e-3,-.8047736478e-1,-.2034825423e-4},
			{-.5031802903e-1,-.2534486742e-5,-.2284466108e-1,-.2406231299e-3,-.1497012310e-1,-.8545021612e-1,-.3576590913e-1,-.1279964701e-2,-.8115562931e-1,-.1019197433e-1,-.7909044629e-1,-.1699783079e-1,-.3666559727e-1,-.6377578460e-1,-.2816752894e-2,-.1340300809e-1,-.2239052508e-1,-.1768768448e-3,-.1990066688e-2,-.1610701220e-2,-.4746387201e-6,-.2764961419e-4,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4},
			{-.8715338278e-1,-.4389859808e-5,-.3956811365e-1,-.4167714865e-3,-.2592901380e-1,-.1480041159,-.6194837174e-1,-.2216963893e-2,-.1405656732,-.1765301736e-1,-.1369886713,-.2944110655e-1,-.6350667735e-1,-.1104628991,-.4878759123e-2,-.2321469098e-1,-.3878152703e-1,-.3063596816e-3,-.3446896614e-2,-.2789816347e-2,-.8220983782e-6,-.4789053657e-4,-.5234078544e-2,-.8679343590e-1,-.6131809964e-1,-.1498609619e-1,-.1755646063,-.1058742437e-1,-.1046114740e-3,-.5625109532e-1,-.6360817804e-2,-.1759148700e-2,-.1069636429,-.1989227818e-3,-.4621846472e-6,-.4015360274e-2,.3324626528e-3,.2179945987e-4,-.1568097578e-1,-.1804945451e-5,-.3191290614e-3,-.2640669295e-1,-.2374368331e-1,-.1440801715e-2,-.1071978198,-.1295502609e-2,-.6011507348e-5,-.2251959335e-2,-.3991407866e-2,-.3580859223e-4,-.2377551727e-1,-.6346770766e-4},
			{-.1125145334,-.5667284643e-5,-.5108221507e-1,-.5380496754e-3,-.3347421288e-1,-.1910724920,-.7997500406e-1,-.2862088080e-2,-.1814695038,-.2278994742e-1,-.1768516141,-.3800830512e-1,-.8198676790e-1,-.1426069896,-.6298450945e-2,-.2997003719e-1,-.5006673611e-1,-.3955086483e-3,-.4449924395e-2,-.3601637417e-2,-.1061324443e-5,-.6182641688e-4,-.6757166343e-2,-.1120498440,-.7916132626e-1,-.1934696698e-1,-.2266529321,-.1366830609e-1,-.1350528323e-3,-.7261985182e-1,-.8211780472e-2,-.2271051205e-2,-.1380894693,-.2568082070e-3,-.5966778137e-6,-.5183807822e-2,.4292074393e-3,.2814298168e-4,-.2024405269e-1,-.2330174558e-5,-.4119938467e-3,-.3409089400e-1,-.3065296333e-1,-.1860067016e-2,-.1383917903,-.1672486677e-2,-.7760822615e-5,-.2907267000e-2,-.5152885397e-2,-.4622869379e-4,-.3069406081e-1,-.8193645830e-4},
			{-.4593386590e-1,-.3452307302e-5,-.3111746013e-1,-.3452307302e-5,-.1695115978e-1,-.9675807323e-1,-.1695115978e-1,-.1649111558e-2,-.1045612321,-.1649111558e-2,-.5210348698e-1,-.1119789175e-1,-.5210348698e-1,-.2584328822e-1,-.1141407470e-2,-.2584328822e-1,-.4627448202e-2,-.3655512474e-4,-.4627448202e-2,-.7320817509e-4,-.2157286163e-7,-.7320817509e-4,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4},
			{-.7955978954e-1,-.5979571650e-5,-.5389702195e-1,-.5979571650e-5,-.2936026999e-1,-.1675898989,-.2936026999e-1,-.2856345005e-2,-.1811053665,-.2856345005e-2,-.9024588671e-1,-.1939531744e-1,-.9024588671e-1,-.4476188824e-1,-.1976975730e-2,-.4476188824e-1,-.8014975395e-2,-.6331533334e-4,-.8014975395e-2,-.1268002788e-3,-.3736529241e-7,-.1268002788e-3,-.4250861857e-2,-.1425849616,-.1739310676e-1,-.4250861857e-2,-.1425849616,-.1739310676e-1,-.9675438873e-4,-.9892989258e-1,-.3498493552e-3,-.9675438873e-4,-.9892989258e-1,-.3498493552e-3,-.2591590698e-6,-.8792734945e-2,-.3952424159e-5,-.2591590698e-6,-.8792734945e-2,-.3952424159e-5,-.1654938171e-3,-.5559060120e-1,-.2727254377e-2,-.1654938171e-3,-.5559060120e-1,-.2727254377e-2,-.1404040445e-5,-.5552981292e-2,-.1565014910e-3,-.1404040445e-5,-.5552981292e-2,-.1565014910e-3},
			{-.3211722999e-1,-.3601841323e-5,-.3246528886e-1,-.3793824429e-7,-.1470175463e-1,-.8391835536e-1,-.6153543472e-2,-.1627413015e-2,-.1031854450,-.2043795582e-3,-.2629088823e-1,-.5650341992e-2,-.5671141990e-1,-.8021127855e-2,-.3542651063e-3,-.3816708301e-1,-.7325116912e-3,-.5786570712e-5,-.8241593859e-2,-.2548589453e-5,-.7510140431e-9,-.1484655847e-3,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3},
			{-.1299205923e-1,-.6892821735e-8,-.5898466126e-2,-.5898466126e-2,-.1617841028e-2,-.2206315093e-1,-.2206315093e-1,-.4150419878e-4,-.2095429338e-1,-.2095429338e-1,-.4404977136e-1,-.4388821038e-2,-.4388821038e-2,-.7835445799e-1,-.7272824702e-3,-.7272824702e-3,-.6504520647e-1,-.4566940490e-4,-.4566940490e-4,-.2422675624e-1,-.1225511905e-6,-.1225511905e-6,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6},
			{-.2250290667e-1,-.1193871745e-7,-.1021644301e-1,-.1021644301e-1,-.2802182859e-2,-.3821449838e-1,-.3821449838e-1,-.7188738101e-4,-.3629390077e-1,-.3629390077e-1,-.7629644205e-1,-.7601661022e-2,-.7601661022e-2,-.1357139022,-.1259690190e-2,-.1259690190e-2,-.1126616024,-.7910172963e-4,-.7910172963e-4,-.4196197271e-1,-.2122648885e-6,-.2122648885e-6,-.4720047033e-3,-.1107876030e-1,-.9169410646e-1,-.1107876030e-1,-.9169410646e-1,-.4720047033e-3,-.1606240095e-5,-.7638008748e-2,-.5251649806e-1,-.7638008748e-2,-.5251649806e-1,-.1606240095e-5,.2530111540e-8,-.2654790910e-3,-.1581160736e-1,-.2654790910e-3,-.1581160736e-1,.2530111540e-8,-.1825084025e-4,-.1679563581e-2,-.1123601370,-.1679563581e-2,-.1123601370,-.1825084025e-4,-.2605756842e-6,-.5507385590e-4,-.6481175407e-1,-.5507385590e-4,-.6481175407e-1,-.2605756842e-6},
			{-.2905112760e-1,-.1541281795e-7,-.1318937122e-1,-.1318937122e-1,-.3617602516e-2,-.4933470527e-1,-.4933470527e-1,-.9280620982e-4,-.4685522442e-1,-.4685522442e-1,-.9849828315e-1,-.9813702180e-2,-.9813702180e-2,-.1752058944,-.1626253042e-2,-.1626253042e-2,-.1454455033,-.1021198938e-3,-.1021198938e-3,-.5417267383e-1,-.2740327927e-6,-.2740327927e-6,-.6093554517e-3,-.1430261805e-1,-.1183765824,-.1430261805e-1,-.1183765824,-.6093554517e-3,-.2073647046e-5,-.9860626894e-2,-.6779850746e-1,-.9860626894e-2,-.6779850746e-1,-.2073647046e-5,.3266359953e-8,-.3427320328e-3,-.2041269733e-1,-.3427320328e-3,-.2041269733e-1,.3266359953e-8,-.2356173345e-4,-.2168307259e-2,-.1450563132,-.2168307259e-2,-.1450563132,-.2356173345e-4,-.3364017618e-6,-.7110004225e-4,-.8367161471e-1,-.7110004225e-4,-.8367161471e-1,-.3364017618e-6},
			{-.2905112760e-1,-.2299810724e-7,-.1968040980e-1,-.2072940277e-3,-.4487304244e-2,-.6119517866e-1,-.2561375849e-1,-.1309844384e-3,-.6613032970e-1,-.8305013821e-2,-.7108239413e-1,-.7082168586e-2,-.1527676939e-1,-.7777343793e-1,-.7218894686e-3,-.3434980188e-2,-.3292821660e-1,-.2311949087e-4,-.2601206999e-3,-.2697214518e-2,-.1364387568e-7,-.7948106270e-6,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,.2006344762e-8,-.8221381159e-3,.2658347174e-3,.4463401949e-5,-.1253839410e-1,.7835272927e-8,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6},
			{-.5031802903e-1,-.3983389024e-7,-.3408746970e-1,-.3590437882e-3,-.7772238941e-2,-.1059931586,-.4436433109e-1,-.2268717024e-3,-.1145410910,-.1438470590e-1,-.1231183182,-.1226667582e-1,-.2646014076e-1,-.1347075460,-.1250349237e-2,-.5949560210e-2,-.5703334418e-1,-.4004413284e-4,-.4505422684e-3,-.4671712585e-2,-.2363188589e-7,-.1376652389e-5,-.9389851147e-3,-.4458139289e-1,-.6370969475e-1,-.7697598727e-2,-.1824121677,-.1899366584e-2,-.3638961323e-5,-.3290427534e-1,-.7075212788e-2,-.1029020197e-2,-.1189769237,-.6919626319e-5,.3475091067e-8,-.1423984988e-2,.4604392371e-3,.7730838953e-5,-.2171713563e-1,.1357109081e-7,-.2318322805e-4,-.8660828315e-2,-.3161298039e-1,-.4725520274e-3,-.1427260688,-.9411218238e-4,-.1490753193e-6,-.3326494240e-3,-.6224747955e-2,-.5289486092e-5,-.3707879714e-1,-.1573892910e-5},
			{-.3211722999e-1,-.3793824429e-7,-.3246528886e-1,-.3601841323e-5,-.6153543472e-2,-.8391835536e-1,-.1470175463e-1,-.2043795582e-3,-.1031854450,-.1627413015e-2,-.5671141990e-1,-.5650341992e-2,-.2629088823e-1,-.3816708301e-1,-.3542651063e-3,-.8021127855e-2,-.8241593859e-2,-.5786570712e-5,-.7325116912e-3,-.1484655847e-3,-.7510140431e-9,-.2548589453e-5,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5},
			{-.5490147066e-2,-.4577893991e-10,-.3719247860e-2,-.3719247860e-2,-.3549462592e-3,-.1156480173e-1,-.1156480173e-1,-.3108711060e-5,-.1249745760e-1,-.1249745760e-1,-.2897665788e-1,-.1338404058e-2,-.1338404058e-2,-.6993680883e-1,-.1364242863e-3,-.1364242863e-3,-.7001415346e-1,-.4369173091e-5,-.4369173091e-5,-.2969356790e-1,-.2578450140e-8,-.2578450140e-8,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8},
			{-.9509213660e-2,-.7929144985e-10,-.6441926259e-2,-.6441926259e-2,-.6147849549e-3,-.2003082417e-1,-.2003082417e-1,-.5384445502e-5,-.2164623154e-1,-.2164623154e-1,-.5018904368e-1,-.2318183830e-2,-.2318183830e-2,-.1211341062,-.2362937952e-3,-.2362937952e-3,-.1212680711,-.7567629781e-5,-.7567629781e-5,-.5143076827e-1,-.4466006648e-8,-.4466006648e-8,-.6197712293e-4,-.4165091360e-2,-.6973083633e-1,-.4165091360e-2,-.6973083633e-1,-.6197712293e-4,-.4089546907e-7,-.3270153304e-2,-.4275519802e-1,-.3270153304e-2,-.4275519802e-1,-.4089546907e-7,-.1392377369e-10,-.6890929850e-4,-.1602773282e-1,-.6890929850e-4,-.1602773282e-1,-.1392377369e-10,-.9704139476e-6,-.4031888160e-3,-.1094954884,-.4031888160e-3,-.1094954884,-.9704139476e-6,-.4729583171e-8,-.5954404436e-5,-.7398029375e-1,-.5954404436e-5,-.7398029375e-1,-.4729583171e-8},
			{-.1486738854e-1,-.1849807308e-9,-.1502850849e-1,-.1582954870e-3,-.1192279250e-2,-.3884665007e-1,-.1625959323e-1,-.1188156743e-4,-.4776557949e-1,-.5998666568e-2,-.5662815244e-1,-.2615600094e-2,-.5642045787e-2,-.8406962479e-1,-.1639928781e-3,-.7803303852e-3,-.4292442302e-1,-.2678661737e-5,-.3013800652e-4,-.4003578188e-2,-.3476519485e-9,-.2025212407e-7,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7},
			{-.2216299428e-2,-.2904510069e-12,-.2240317772e-2,-.2240317772e-2,-.7439227355e-4,-.5790916683e-2,-.5790916683e-2,-.2224370628e-6,-.7120472184e-2,-.7120472184e-2,-.1820922229e-1,-.3899106406e-3,-.3899106406e-3,-.5963296216e-1,-.2444661488e-4,-.2444661488e-4,-.7199376027e-1,-.3993113157e-6,-.3993113157e-6,-.3476706016e-1,-.5182489262e-10,-.5182489262e-10,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_bottom_fluxes_to_dofs[j][0];
			for (int i = 1; i < 52; i++)
			{
				std::cout << ", " << timestep_bottom_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_top_fluxes_to_dofs:" << std::endl;
		static const T timestep_top_fluxes_to_dofs[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5412739307e-1,-.1103717978e-1,-.1103717978e-1,-.1162547668e-3,-.5974166429e-1,-.5974166429e-1,-.2500537776e-1,-.4382549273e-1,-.4382549273e-1,-.5503848602e-2,-.3510914219e-1,-.3510914219e-1,-.7573305578e-1,-.1537719318e-1,-.1537719318e-1,-.7316958648e-1,-.3712176777e-2,-.3712176777e-2,-.4176623208e-1,-.2059615826e-3,-.2059615826e-3,-.1199809046e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2852764098e-1,-.8679943781e-2,-.8679943781e-2,-.9629909754e-6,-.3905626741e-1,-.3905626741e-1,-.6842313070e-2,-.3260004675e-1,-.3260004675e-1,-.5141591464e-3,-.1335372273e-1,-.1335372273e-1,-.6213451013e-1,-.3597562967e-2,-.3597562967e-2,-.8145457175e-1,-.4429404170e-3,-.4429404170e-3,-.5607103930e-1,-.5404682501e-5,-.5404682501e-5,-.1834095772e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1299205923e-1,-.5898466126e-2,-.5898466126e-2,-.6892821735e-8,-.2206315093e-1,-.2206315093e-1,-.1617841028e-2,-.2095429338e-1,-.2095429338e-1,-.4150419878e-4,-.4388821038e-2,-.4388821038e-2,-.4404977136e-1,-.7272824702e-3,-.7272824702e-3,-.7835445799e-1,-.4566940490e-4,-.4566940490e-4,-.6504520647e-1,-.1225511905e-6,-.1225511905e-6,-.2422675624e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5490147066e-2,-.3719247860e-2,-.3719247860e-2,-.4577893991e-10,-.1156480173e-1,-.1156480173e-1,-.3549462592e-3,-.1249745760e-1,-.1249745760e-1,-.3108711060e-5,-.1338404058e-2,-.1338404058e-2,-.2897665788e-1,-.1364242863e-3,-.1364242863e-3,-.6993680883e-1,-.4369173091e-5,-.4369173091e-5,-.7001415346e-1,-.2578450140e-8,-.2578450140e-8,-.2969356790e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2216299428e-2,-.2240317772e-2,-.2240317772e-2,-.2904510069e-12,-.5790916683e-2,-.5790916683e-2,-.7439227355e-4,-.7120472184e-2,-.7120472184e-2,-.2224370628e-6,-.3899106406e-3,-.3899106406e-3,-.1820922229e-1,-.2444661488e-4,-.2444661488e-4,-.5963296216e-1,-.3993113157e-6,-.3993113157e-6,-.7199376027e-1,-.5182489262e-10,-.5182489262e-10,-.3476706016e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1},
			{-.5412739307e-1,-.1162547668e-3,-.1103717978e-1,-.1103717978e-1,-.2500537776e-1,-.5974166429e-1,-.5974166429e-1,-.5503848602e-2,-.4382549273e-1,-.4382549273e-1,-.7573305578e-1,-.3510914219e-1,-.3510914219e-1,-.7316958648e-1,-.1537719318e-1,-.1537719318e-1,-.4176623208e-1,-.3712176777e-2,-.3712176777e-2,-.1199809046e-1,-.2059615826e-3,-.2059615826e-3,-.1175889790e-1,-.3366781247e-1,-.6810280266e-1,-.3366781247e-1,-.6810280266e-1,-.1175889790e-1,-.1064318140e-2,-.1789759571e-1,-.3403297354e-1,-.1789759571e-1,-.3403297354e-1,-.1064318140e-2,.3588356267e-4,-.1692488682e-2,-.6609587246e-2,-.1692488682e-2,-.6609587246e-2,.3588356267e-4,-.2772841188e-2,-.1251880452e-1,-.5082001574e-1,-.1251880452e-1,-.5082001574e-1,-.2772841188e-2,-.3397397726e-3,-.2023719225e-2,-.2136582602e-1,-.2023719225e-2,-.2136582602e-1,-.3397397726e-3},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.6987816394e-1,-.2239469694e-3,-.2126143325e-1,-.2239469694e-3,-.4004261797e-1,-.9566792642e-1,-.4004261797e-1,-.1002844321e-1,-.7985348011e-1,-.1002844321e-1,-.7055750928e-1,-.3270980684e-1,-.7055750928e-1,-.4193122589e-1,-.8812193584e-2,-.4193122589e-1,-.1220724282e-1,-.1084978008e-2,-.1220724282e-1,-.7712083501e-3,-.1323871434e-4,-.7712083501e-3,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1350573964e-1,-.7821980679e-1,-.2731923027e-1,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,-.1392123656e-2,-.4451498643e-1,-.2647177214e-2,.2845518738e-4,-.5241314676e-2,.1111245502e-3,.2845518738e-4,-.5241314676e-2,.1111245502e-3,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.2033552609e-2,-.3727049930e-1,-.8255195243e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2,-.1122167342e-3,-.7057175562e-2,-.1184750924e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5031802903e-1,-.2406231299e-3,-.2284466108e-1,-.2534486742e-5,-.3576590913e-1,-.8545021612e-1,-.1497012310e-1,-.1019197433e-1,-.8115562931e-1,-.1279964701e-2,-.3666559727e-1,-.1699783079e-1,-.7909044629e-1,-.1340300809e-1,-.2816752894e-2,-.6377578460e-1,-.1990066688e-2,-.1768768448e-3,-.2239052508e-1,-.2764961419e-4,-.4746387201e-6,-.1610701220e-2,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2905112760e-1,-.2072940277e-3,-.1968040980e-1,-.2299810724e-7,-.2561375849e-1,-.6119517866e-1,-.4487304244e-2,-.8305013821e-2,-.6613032970e-1,-.1309844384e-3,-.1527676939e-1,-.7082168586e-2,-.7108239413e-1,-.3434980188e-2,-.7218894686e-3,-.7777343793e-1,-.2601206999e-3,-.2311949087e-4,-.3292821660e-1,-.7948106270e-6,-.1364387568e-7,-.2697214518e-2,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,.4463401949e-5,-.1253839410e-1,.7835272927e-8,.2006344762e-8,-.8221381159e-3,.2658347174e-3,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1486738854e-1,-.1582954870e-3,-.1502850849e-1,-.1849807308e-9,-.1625959323e-1,-.3884665007e-1,-.1192279250e-2,-.5998666568e-2,-.4776557949e-1,-.1188156743e-4,-.5642045787e-2,-.2615600094e-2,-.5662815244e-1,-.7803303852e-3,-.1639928781e-3,-.8406962479e-1,-.3013800652e-4,-.2678661737e-5,-.4292442302e-1,-.2025212407e-7,-.3476519485e-9,-.4003578188e-2,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2},
			{-.2852764098e-1,-.9629909754e-6,-.8679943781e-2,-.8679943781e-2,-.6842313070e-2,-.3905626741e-1,-.3905626741e-1,-.5141591464e-3,-.3260004675e-1,-.3260004675e-1,-.6213451013e-1,-.1335372273e-1,-.1335372273e-1,-.8145457175e-1,-.3597562967e-2,-.3597562967e-2,-.5607103930e-1,-.4429404170e-3,-.4429404170e-3,-.1834095772e-1,-.5404682501e-5,-.5404682501e-5,-.1925725884e-2,-.1578668809e-1,-.6459385425e-1,-.1578668809e-1,-.6459385425e-1,-.1925725884e-2,-.3379706279e-4,-.9557084327e-2,-.3455698326e-1,-.9557084327e-2,-.3455698326e-1,-.3379706279e-4,-.2462950063e-6,-.5479185995e-3,-.8356283694e-2,-.5479185995e-3,-.8356283694e-2,-.2462950063e-6,-.1838831460e-3,-.3748155920e-2,-.6176771321e-1,-.3748155920e-2,-.6176771321e-1,-.1838831460e-3,-.7690917183e-5,-.2728888970e-3,-.3041758476e-1,-.2728888970e-3,-.3041758476e-1,-.7690917183e-5},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.5031802903e-1,-.2534486742e-5,-.2284466108e-1,-.2406231299e-3,-.1497012310e-1,-.8545021612e-1,-.3576590913e-1,-.1279964701e-2,-.8115562931e-1,-.1019197433e-1,-.7909044629e-1,-.1699783079e-1,-.3666559727e-1,-.6377578460e-1,-.2816752894e-2,-.1340300809e-1,-.2239052508e-1,-.1768768448e-3,-.1990066688e-2,-.1610701220e-2,-.4746387201e-6,-.2764961419e-4,-.3021896657e-2,-.5011021362e-1,-.3540202135e-1,-.8652226669e-2,-.1013622727,-.6112652313e-2,-.6039746270e-4,-.3247658505e-1,-.3672419871e-2,-.1015644975e-2,-.6175548808e-1,-.1148481216e-3,-.2668424305e-6,-.2318269335e-2,.1919474022e-3,.1258592403e-4,-.9053415595e-2,-.1042085743e-5,-.1842492495e-3,-.1524591129e-1,-.1370842195e-1,-.8318472582e-3,-.6189069014e-1,-.7479587802e-3,-.3470745387e-5,-.1300169329e-2,-.2304440407e-2,-.2067410037e-4,-.1372680130e-1,-.3664309813e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.4593386590e-1,-.3452307302e-5,-.3111746013e-1,-.3452307302e-5,-.1695115978e-1,-.9675807323e-1,-.1695115978e-1,-.1649111558e-2,-.1045612321,-.1649111558e-2,-.5210348698e-1,-.1119789175e-1,-.5210348698e-1,-.2584328822e-1,-.1141407470e-2,-.2584328822e-1,-.4627448202e-2,-.3655512474e-4,-.4627448202e-2,-.7320817509e-4,-.2157286163e-7,-.7320817509e-4,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.2454236237e-2,-.8232146596e-1,-.1004191487e-1,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.5586117238e-4,-.5711720011e-1,-.2019856194e-3,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.1496255587e-6,-.5076487887e-2,-.2281933152e-5,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.9554789986e-4,-.3209524856e-1,-.1574581049e-2,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4,-.8106231288e-6,-.3206015244e-2,-.9035617794e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.3211722999e-1,-.3601841323e-5,-.3246528886e-1,-.3793824429e-7,-.1470175463e-1,-.8391835536e-1,-.6153543472e-2,-.1627413015e-2,-.1031854450,-.2043795582e-3,-.2629088823e-1,-.5650341992e-2,-.5671141990e-1,-.8021127855e-2,-.3542651063e-3,-.3816708301e-1,-.7325116912e-3,-.5786570712e-5,-.8241593859e-2,-.2548589453e-5,-.7510140431e-9,-.1484655847e-3,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3},
			{-.1299205923e-1,-.6892821735e-8,-.5898466126e-2,-.5898466126e-2,-.1617841028e-2,-.2206315093e-1,-.2206315093e-1,-.4150419878e-4,-.2095429338e-1,-.2095429338e-1,-.4404977136e-1,-.4388821038e-2,-.4388821038e-2,-.7835445799e-1,-.7272824702e-3,-.7272824702e-3,-.6504520647e-1,-.4566940490e-4,-.4566940490e-4,-.2422675624e-1,-.1225511905e-6,-.1225511905e-6,-.2725120425e-3,-.6396325243e-2,-.5293961706e-1,-.6396325243e-2,-.5293961706e-1,-.2725120425e-3,-.9273631513e-6,-.4409806408e-2,-.3032041429e-1,-.4409806408e-2,-.3032041429e-1,-.9273631513e-6,.1460760579e-8,-.1532744247e-3,-.9128835766e-2,-.1532744247e-3,-.9128835766e-2,.1460760579e-8,-.1053712753e-4,-.9696964856e-3,-.6487115538e-1,-.9696964856e-3,-.6487115538e-1,-.1053712753e-4,-.1504434415e-6,-.3179690554e-4,-.3741908366e-1,-.3179690554e-4,-.3741908366e-1,-.1504434415e-6},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.2905112760e-1,-.2299810724e-7,-.1968040980e-1,-.2072940277e-3,-.4487304244e-2,-.6119517866e-1,-.2561375849e-1,-.1309844384e-3,-.6613032970e-1,-.8305013821e-2,-.7108239413e-1,-.7082168586e-2,-.1527676939e-1,-.7777343793e-1,-.7218894686e-3,-.3434980188e-2,-.3292821660e-1,-.2311949087e-4,-.2601206999e-3,-.2697214518e-2,-.1364387568e-7,-.7948106270e-6,-.5421233086e-3,-.2573907918e-1,-.3678280940e-1,-.4444210696e-2,-.1053157141,-.1096599808e-2,-.2100955299e-5,-.1899729222e-1,-.4084876007e-2,-.5941050873e-3,-.6869135890e-1,-.3995048117e-5,.2006344762e-8,-.8221381159e-3,.2658347174e-3,.4463401949e-5,-.1253839410e-1,.7835272927e-8,-.1338484295e-4,-.5000331558e-2,-.1825176273e-1,-.2728280401e-3,-.8240293419e-1,-.5433569382e-4,-.8606867573e-7,-.1920552345e-3,-.3593859906e-2,-.3053886218e-5,-.2140745350e-1,-.9086874950e-6},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.3211722999e-1,-.3793824429e-7,-.3246528886e-1,-.3601841323e-5,-.6153543472e-2,-.8391835536e-1,-.1470175463e-1,-.2043795582e-3,-.1031854450,-.1627413015e-2,-.5671141990e-1,-.5650341992e-2,-.2629088823e-1,-.3816708301e-1,-.3542651063e-3,-.8021127855e-2,-.8241593859e-2,-.5786570712e-5,-.7325116912e-3,-.1484655847e-3,-.7510140431e-9,-.2548589453e-5,-.5332127766e-3,-.5120891311e-1,-.1263569352e-1,-.1526682850e-2,-.1035847075,-.2181728568e-2,-.2353281468e-5,-.4046261701e-1,-.2720900819e-3,-.3957282952e-4,-.7694123815e-1,-.8509112765e-5,.1362455012e-8,-.2180268388e-2,-.3827349198e-5,-.6426172639e-7,-.8514487739e-2,.2077874452e-7,-.8406097227e-5,-.1274828807e-1,-.2538914536e-2,-.3795179057e-4,-.5175160286e-1,-.1385282294e-3,-.2434484803e-7,-.5735318303e-3,-.1706550817e-3,-.1450143344e-6,-.6055178584e-2,-.2713600615e-5},
			{-.5490147066e-2,-.4577893991e-10,-.3719247860e-2,-.3719247860e-2,-.3549462592e-3,-.1156480173e-1,-.1156480173e-1,-.3108711060e-5,-.1249745760e-1,-.1249745760e-1,-.2897665788e-1,-.1338404058e-2,-.1338404058e-2,-.6993680883e-1,-.1364242863e-3,-.1364242863e-3,-.7001415346e-1,-.4369173091e-5,-.4369173091e-5,-.2969356790e-1,-.2578450140e-8,-.2578450140e-8,-.3578250861e-4,-.2404716618e-2,-.4025911712e-1,-.2404716618e-2,-.4025911712e-1,-.3578250861e-4,-.2361101008e-7,-.1888023890e-2,-.2468472508e-1,-.1888023890e-2,-.2468472508e-1,-.2361101008e-7,-.8038894490e-11,-.3978480204e-4,-.9253615860e-2,-.3978480204e-4,-.9253615860e-2,-.8038894490e-11,-.5602687538e-6,-.2327811714e-3,-.6321724968e-1,-.2327811714e-3,-.6321724968e-1,-.5602687538e-6,-.2730626117e-8,-.3437777004e-5,-.4271254251e-1,-.3437777004e-5,-.4271254251e-1,-.2730626117e-8},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1486738854e-1,-.1849807308e-9,-.1502850849e-1,-.1582954870e-3,-.1192279250e-2,-.3884665007e-1,-.1625959323e-1,-.1188156743e-4,-.4776557949e-1,-.5998666568e-2,-.5662815244e-1,-.2615600094e-2,-.5642045787e-2,-.8406962479e-1,-.1639928781e-3,-.7803303852e-3,-.4292442302e-1,-.2678661737e-5,-.3013800652e-4,-.4003578188e-2,-.3476519485e-9,-.2025212407e-7,-.8620825942e-4,-.1171904123e-1,-.3387615092e-1,-.2023455775e-2,-.9699343477e-1,-.1743809191e-3,-.6478094133e-7,-.9850207418e-2,-.4027520438e-2,-.3080469716e-3,-.6772686648e-1,-.1231834765e-6,-.1337175965e-10,-.2584389697e-3,.3263423843e-3,.1403069605e-5,-.1539230641e-1,-.5222003136e-10,-.8618926100e-6,-.1453705859e-2,-.2154043605e-1,-.7931708447e-4,-.9725061405e-1,-.3498848146e-5,-.1891907121e-8,-.2514691448e-4,-.4968086294e-2,-.3998631737e-6,-.2959327271e-1,-.1997419301e-7},
			{-.2216299428e-2,-.2904510069e-12,-.2240317772e-2,-.2240317772e-2,-.7439227355e-4,-.5790916683e-2,-.5790916683e-2,-.2224370628e-6,-.7120472184e-2,-.7120472184e-2,-.1820922229e-1,-.3899106406e-3,-.3899106406e-3,-.5963296216e-1,-.2444661488e-4,-.2444661488e-4,-.7199376027e-1,-.3993113157e-6,-.3993113157e-6,-.3476706016e-1,-.5182489262e-10,-.5182489262e-10,-.4488428794e-5,-.8636459480e-3,-.2924732612e-1,-.8636459480e-3,-.2924732612e-1,-.4488428794e-5,-.5742721667e-9,-.7722073236e-3,-.1919817421e-1,-.7722073236e-3,-.1919817421e-1,-.5742721667e-9,.4226219864e-13,-.9865138295e-5,-.8960784022e-2,-.9865138295e-5,-.8960784022e-2,.4226219864e-13,-.2845830558e-7,-.5338243093e-4,-.5885156716e-1,-.5338243093e-4,-.5885156716e-1,-.2845830558e-7,-.4734669723e-10,-.3550659426e-6,-.4657535457e-1,-.3550659426e-6,-.4657535457e-1,-.4734669723e-10}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_top_fluxes_to_dofs[j][0];
			for (int i = 1; i < 52; i++)
			{
				std::cout << ", " << timestep_top_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
